#GrainMaker.py

import sys
from Numeric import *
sys.path.append("..")
from MDWebServiceDecorator import *


class BulkEnergy(MDWebServiceDecorator):

    def __init__(self, DIMENSION, Parameters,ioControl,MD):
        self.Parameters = Parameters
        self.DIMENSION = DIMENSION
        self.ioControl = ioControl
        MDWebServiceDecorator.__init__(self, DIMENSION, Parameters,ioControl,MD)

    def ScaleLength(self):
        # necessary because previously calculated geometries are in units of 1 lattice constant...
        self.SetPotentialAndLattice()
        self.Parameters["length"]*=self.latticeConst


    def MakeLeaves(self,angle,length,lattice):

        # get rotation and size

        if self.DIMENSION==2:
            rotation = self.GetRotationMatrix(angle)
            self.ioControl.Print(6,"Rotation: " + repr(rotation))
            rectSize1 = array([self.Parameters["width"],length])
            rectSize2 = array([self.constrLayerWidth,length])
        elif self.DIMENSION==3:
            rotationString = "rotation"+repr(angle)
            rotation = array(self.Parameters[rotationString])
            rectSize1 = array([self.Parameters["length"],self.Parameters["width"],self.Parameters["height"]])
            rectSize2 = array([self.Parameters["length"]+(self.constrLayerWidth*2.1),self.Parameters["width"]+(self.constrLayerWidth*2.1),self.Parameters["height"]+(self.constrLayerWidth*2.1)])
        
        # declare the rectangular cluster initializer
        rectangle = self.MD.RectangularClusterInitializer(rectSize1, lattice)

        # build the first crystal
        rectCenter = self.origin[:]
        self.CreateCrystal(rectangle, rectangle, lattice, rotation, rectSize1, rectCenter, self.bulk)
        
        # build the constrained zone/s
        if self.Parameters["BC"]=="periodic" and self.DIMENSION == 2:
            rectCenter[0] = self.origin[0] - rectSize1[0]/2 - self.constrLayerWidth/2.
            self.ioControl.Print(6,"center of constrZone 1 = "+repr(rectCenter[0])+" " + repr(rectCenter[1]))
            self.CreateCrystal(rectangle, rectangle, lattice, rotation, rectSize2, rectCenter, self.constrZone1)
        
            rectCenter[0] = self.origin[0] + rectSize1[0]/2 + self.constrLayerWidth/2.
            self.ioControl.Print(6,"center of constrZone2 = "+repr(rectCenter[0])+" " + repr(rectCenter[1]))
            self.CreateCrystal(rectangle, rectangle, lattice, rotation, rectSize2, rectCenter, self.constrZone2)
        else:
            rectangle2 = self.MD.RectangularClusterInitializer(rectSize2, lattice)
            shell = self.MD.CompositeClusterInitializer(rectangle2,rectangle,lattice,"difference")
            shell.Create(self.constrZone1)
            


    def MakeSharedObjects(self):

        #set PotentialAndLattice
        self.ioControl.Print(3,"Setting Potential And Lattice")
        self.SetPotentialAndLattice()
        self.lattice2 = self.lattice1 = self.lattice
        self.constrLayerWidth = 2*self.potential.GetCutoffDistance()

        # Shift by tiny amount to avoid problems
        if self.DIMENSION==2:
            self.origin=[1.0e-6,1.3e-6]
        if self.DIMENSION==3:
            self.origin=[1.0e-6,1.3e-6,1.4e-6]
        self.lattice1.SetOrigin(self.Parameters["latticeOrigin1"])
        self.lattice2.SetOrigin(self.Parameters["latticeOrigin2"])

        #boundary conditions
        if self.Parameters["BC"]=="periodic":
            self.boundaryConditions = self.MD.SmartPeriodicBoundaryConditions([0,1])
            boxSize = [0,self.Parameters["length"]]
            self.boundaryConditions.SetLength(boxSize)
        else:
            self.boundaryConditions = self.MD.FreeBoundaryConditions()
        
        self.neighborLocator = self.MD.CellNeighborList(self.potential.GetCutoffDistance(), 0.05 * self.potential.GetCutoffDistance())

        #movers
        tol = 1.0e-10
        qmStepsBetweenUpdate = 100
        maxQMIterations = 600
        self.qm = self.MD.QuickMinAtomsMover(self.potential,self.Parameters["timestep"], tol)
        self.qm.SetMaxIterations(maxQMIterations)
        if self.ioControl.GetVerbosity() > 5:
            self.qm.Verbose()
            self.qm.SetStepsBetweenUpdate(qmStepsBetweenUpdate)

        self.verlet = self.MD.VerletAtomsMover(self.potential, self.Parameters["timestep"], 100)

    def MakeListsOfAtoms(self):

        # ListOfAtoms
        self.ioControl.Print(3,"Making lists of atoms")
        self.atoms = self.MD.DynamicListOfAtoms(0)
        self.bulk = self.MD.DynamicListOfAtoms(0)
        self.constrZone1 = self.MD.DynamicListOfAtoms(0)
        self.constrZone2 = self.MD.DynamicListOfAtoms(0)

        self.atoms.Attach(self.bulk)
        self.atoms.Attach(self.constrZone1)
        if self.Parameters["BC"]=="periodic":
            self.atoms.Attach(self.constrZone2)

        # set neighborLocator and boundaryConditions
        self.atoms.SetBoundaryConditions(self.boundaryConditions)
        self.atoms.SetNeighborLocator(self.neighborLocator)
        self.neighborLocator.thisown = 0

##         if self.Parameters["plotObserver"]==1:
##             self.AddPlotAtomsObserver()


    def DeleteListsOfAtoms(self):
        del self.atoms
        del self.bulk
        del self.constrZone1
        del self.constrZone2


    def MeasureBulkEnergies(self):

        self.MakeSharedObjects()
        self.MakeListsOfAtoms()

        if self.Parameters["yystrain"] == 0 or self.DIMENSION == 3:
            length1 = self.Parameters["length"]
            length2 = self.Parameters["length"]
        else:
            length1 = self.Parameters["repeat1"]*self.Parameters["numRepeats1"]*self.latticeConst
            length2 = self.Parameters["repeat2"]*self.Parameters["numRepeats2"]*self.latticeConst

        if self.DIMENSION==2:
            angle1 = self.Parameters["angle1"]
            angle2 = self.Parameters["angle2"]
        else:
            angle1=1
            angle2=2

        # first crystal
        self.ioControl.Print(3,"First Crystal")
        self.MakeLeaves(angle1,length1,self.lattice1)
        energy1 = self.MeasureEnergy()

        self.DeleteListsOfAtoms()


        if self.Parameters["angle2"] != self.Parameters["angle1"] or self.DIMENSION==3:
            # second crystal
            self.MakeListsOfAtoms()
            self.ioControl.Print(3,"Second Crystal")
            self.MakeLeaves(angle2,length2,self.lattice1)
            energy2 = self.MeasureEnergy()
        else:
            energy2 = energy1
            
        self.ioControl.Print(4,"Bulk Energies: " + repr(energy1) + " " + repr(energy2))

        return energy1,energy2



    def MeasureEnergy(self):
        
        # prune
        self.MD.OverlapPruner(0.7*self.lattice1.GetMinInteratomicSpacing()).Transform(self.atoms)

        # shift to help visualize periodic boundary
        if self.Parameters["BC"]=="periodic":
            self.MD.Translator(array([0,self.Parameters["length"]/2.0])).Transform(self.atoms)
            self.atoms.Notify()

        self.ioControl.Print(4,self.atoms.GetShape())

        # put on constraint
        self.fixedConstraint1 = self.MD.FixedBody(self.constrZone1.GetCartesianPositions())
        self.constrZone1.SetConstraint(self.fixedConstraint1)
        if self.Parameters["BC"]=="periodic":
            self.fixedConstraint2 = self.MD.FixedBody(self.constrZone2.GetCartesianPositions())
            self.constrZone2.SetConstraint(self.fixedConstraint2)

        # get list of relevant atoms:
        #lower = [-(self.Parameters["width"]/2.0-self.constrLayerWidth),-self.Parameters["length"]/2.0]
        #upper = [(self.Parameters["width"]/2.0-self.constrLayerWidth),self.Parameters["length"]/2.0]
        #atomsLocator = self.MD.RectangleAtomsLocator(lower,upper,[1,1])
        #print "RELEVANT ATOMS ",atomsLocator.GetAtomNumbers(self.atoms)

        # relax and measure energy
        #nMajorSteps = 100
        #for i in xrange(nMajorSteps):
        self.qm.Move(self.atoms)
        #self.Notify()
        numberOfAtoms = self.atoms.GetLeaf(0).GetNumber()
        PE = 0
        for j in xrange(numberOfAtoms):
            PE = PE + self.potential.CalculateAtomicEnergy(self.atoms,j)
        AvgEnergyPerAtom = PE/numberOfAtoms
        Deviation = 0
        for j in xrange(numberOfAtoms):
            temp = self.potential.CalculateAtomicEnergy(self.atoms,j)-AvgEnergyPerAtom
            Deviation += temp*temp
        Deviation /= numberOfAtoms
        self.ioControl.Print(4,repr(PE)+' / '+repr(numberOfAtoms)+' = '+repr(AvgEnergyPerAtom) + " Deviation: " + repr(Deviation))

##         for i in xrange(1000):
##             for j in xrange(10000):
##                 x=1+2

        return AvgEnergyPerAtom        



# end of BulkEnergy class


if __name__ == "__main__":

    from IOControl import *

    Parameters = {"width":6.0,
                  "length":19.0656714558,
                  "angle1":3.00449159888,
                  "angle2":0.0,
                  "yystrain": 0.000688231571464,
                  "numRepeats1": 2,
                  "numRepeats2": 11,
                  "repeat2": 1.73205080757,
                  "repeat1": 9.53939201417,
                  "verbose":4,
                  "material":"CLJ",
                  "timestep":0.001,
                  "latticeOrigin1":[1.0e-6,1.3e-6],
                  "latticeOrigin2":[1.0e-6,1.3e-6],
                  "plotObserver":1,
                  "BC":"periodic"}

    
    
    UserParameters = ProcessCommandLine(Parameters)
    if UserParameters == None:
        UserParameters = {}
    Parameters.update(UserParameters)
    ioControl = IOControl(Parameters["verbose"],'.')

    print "Initializing Object"
    bulkEnergyMeasurer = BulkEnergy(2,Parameters,ioControl)
    #forceObserver = ForceObserver.ForceObserver("forcetest")
    #bulkEnergyMeasurer.AddObserver(forceObserver)
    #energyObserver = EnergyObserver.EnergyObserver("energytest")
    #bulkEnergyMeasurer.AddObserver(energyObserver)
    #asciiObserver = ASCIIObserver.ASCIIObserver("asciitest",bulkEnergyMeasurer.Parameters)
    #bulkEnergyMeasurer.AddObserver(asciiObserver)
    print "Measuring Energies"
    bulkEnergyMeasurer.ScaleLength()
    bulkEnergyMeasurer.MeasureBulkEnergies()

