
#!/afs/msc.cornell.edu/i386_linux22/usr/local/pub/dm/bin/python

import sys
import os
import string
import math
sys.path.append("/home/val/work/MDWebServices")
sys.path.append("/afs/msc.cornell.edu/home/jsethna/val/ASP/MDWebServices")
from MDWebService import *
from GrainMaker import *
from MinimumFinder import *


# Right now this has lots of if statements and is somewhat messy since I am trying to
# accomodate 2D, 3D, and three types of boundary conditions, free, periodic, and "rollered".
# Periodic boundary conditions and grain boundary energy measurement are not working yet for 3D.



class GrainBreaker(MDWebService):

    def __init__(self,DIMENSION):

        self.DIMENSION=DIMENSION

        # Initializes Parameter dictionary, ioControl
        
        # Defaults overwritten by UserParameters
        self.Parameters = {"DMDIR":"/afs/msc.cornell.edu/home/jsethna/val/ASP/DigitalMaterial",
                           "material":"MEAM_Si",
                           "brittleFactor":2.0,
                           "verbose":3,
                           "directory":".",
                           "timestep":0.01,
                           "QMtolerance":0.01,
                           "maxQMIterations":100000,
                           "MDIterations":100,
                           "moverType":"qm",
                           "angle1":2.384222916664901,
                           "angle2":5.98175107659759,
                           "length":10,
                           "width":10,
                           "height":10,
                           "minimizerWidth":10,
                           "lengthScaled":0,
                           "numRepeats":1,
                           "measureBulkEnergies":1,
                           "BC":"rollered",
                           "constrLayerWidth":2.0,
                           "maxStrainSteps":30,
                           "strainIncrement":.01,
                           "strainSubdivisions":10,
                           "minSearchSteps":10,
                           "annealSteps":0,
                           "minStress":0.01,
                           "randomNumberSeed":1,
                           "andersenFrequency":0.01,
                           "andersenSigma":0.01,
                           "temperature":0.0,
                           "friction":0.0,
                           "grainOverlap":-0.4,
                           "plotObserver":0,
                           "parallel":0,
                           "yystrain":0}

        if self.DIMENSION==2:
            self.Parameters["latticeOrigin1"]=[1.0e-6,1.3e-6]
            self.Parameters["latticeOrigin2"]=[1.0e-6,1.3e-6]
        if self.DIMENSION==3:
            self.Parameters["latticeOrigin1"]=[1.0e-6,1.3e-6,1.4e-6]
            self.Parameters["latticeOrigin2"]=[1.0e-6,1.3e-6,1.4e-6]
            self.Parameters["rotation1"]=[[0.029245643430524545, 0.85710715101868318, 0.51430732447921734], [0.59795214268700791, 0.39731034590822351, -0.69613053667419589], [-0.80099808201789779, 0.32788995212864347, -0.50088946075628815]]
            self.Parameters["rotation2"]=[[0.024714749269945629, 0.94270921702219279, 0.33269883274506368], [0.44111632315723898, 0.28836595362395823, -0.84985967443736465], [-0.89710956444169176, 0.16776295458756479, -0.40871753137746442]]
            
        
        self.ioControl = IOControl(self.Parameters["verbose"],self.Parameters["directory"])
        MDWebService.__init__(self,self.DIMENSION,self.Parameters,self.ioControl)



    def MakeGrains(self):
        self.ImportDigitalMaterial()
        self.grainMaker = GrainMaker(self.DIMENSION,self.Parameters,self.ioControl,self.MD)
        # must multiply length by latticeConst if bc is periodic
        # only do bulk energy and grain boundary energy measurement if bc is periodic
        if self.Parameters["lengthScaled"]==0 and self.Parameters["BC"] == "periodic":
            self.grainMaker.ScaleLength()
        if self.Parameters["measureBulkEnergies"]:
            self.ioControl.Print(3,"Calculating Bulk Energies for each grain")
            self.grainMaker.CalculateBulkEnergies()
            self.ioControl.Print(3,"BulkEnergies = "+repr(self.grainMaker.bulkEnergy1)+" "+repr(self.grainMaker.bulkEnergy2))
        self.grainMaker.MakeGrains()
        
        self.atoms = self.grainMaker.atoms
        self.constrZone1 = self.grainMaker.constrZone1
        self.constrZone2 = self.grainMaker.constrZone2
        self.bulk = self.grainMaker.bulk
        self.potential = self.grainMaker.potential
        self.moverpotential = self.grainMaker.moverpotential
        self.qm = self.grainMaker.qm
        self.verlet = self.grainMaker.verlet
        self.langevin = self.grainMaker.langevin
        if self.Parameters["moverType"] == "qm":
            self.mover = self.qm
        elif self.Parameters["moverType"] == "verlet":
            self.mover = self.verlet
        elif self.Parameters["moverType"] == "verletThermalize":
            self.mover = self.grainMaker.verletThermalize
        elif self.Parameters["moverType"] == "hoover":
            self.mover = self.grainMaker.hoover
        elif self.Parameters["moverType"] == "langevin":
            self.mover = self.grainMaker.langevin
        self.origin = self.grainMaker.origin

        if self.Parameters["plotObserver"]==1:
            self.AddPlotAtomsObserver()

        self.Notify()


    def ShiftToMinimize(self,minOffset=0,maxOffset=None):
        self.ioControl.Print(1,"Finding Minimizing Offset...")
        self.MinimizedAtomicPositions = self.atoms.GetCartesianPositions().ConstructFromArray(self.atoms.GetCartesianPositions())
        width = self.Parameters["width"]
        minFinder = MinimumFinder(self.Parameters,self.ioControl.directory)
        self.Parameters["width"]=width
        minimizingOffset,self.minFinderEnergy = minFinder.FindMinimum(self.grainMaker.bulkEnergy1,self.grainMaker.bulkEnergy2,self.MinimizedAtomicPositions,minOffset,maxOffset)
        print "MIN FINDER ENERGY = ", self.minFinderEnergy
        if self.Parameters["width"] == self.Parameters["minimizerWidth"]:
            self.atoms.SetCartesianPositions(self.MinimizedAtomicPositions)
            self.atoms.Notify()
            self.Notify()
            self.Save()
        else:
            self.MD.Translator(array([0,minimizingOffset])).Transform(self.atoms.GetLeaf(0))     
            self.MD.Translator(array([0,minimizingOffset])).Transform(self.atoms.GetLeaf(2))

        if self.Parameters["strainSubdivisions"] > 1:
            self.AtomicPositionsBack1 = self.atoms.GetCartesianPositions().ConstructFromArray(self.atoms.GetCartesianPositions())
            self.AtomicPositionsBack2 = self.atoms.GetCartesianPositions().ConstructFromArray(self.atoms.GetCartesianPositions())
            self.AtomicPositionsBack1.Copy(self.atoms.GetCartesianPositions())



    def RelaxNewGrains(self): 

        if self.Parameters["BC"] == "periodic":
            # put rigid body constraint on constrained zones to get better energy measurement
            self.rigidConstraint1 = self.MD.NonRotatingRigidBody()
            self.rigidConstraint2 = self.MD.NonRotatingRigidBody()
            self.constrZone1.SetConstraint(self.rigidConstraint1)
            self.constrZone2.SetConstraint(self.rigidConstraint2)

        #if self.Parameters["BC"] == "rollered":
            #put rigid body constraint on endcaps and rollered constraint on other sides
            #self.grainMaker.ImplementRolleredAndRigidConstraints()
        
        if self.Parameters["annealSteps"] > 0:
            self.ioControl.Print(1,"Annealing...")
            self.Anneal(self.Parameters["annealSteps"])

        thisMover = self.qm
        if self.Parameters["temperature"] > 0:
            thisMover = self.mover
        if self.Parameters["moverType"] == "hoover":
            thisMover = self.langevin


##         thisMover.SetDt(0.0002)
##         thisMover.SetMaxIterations(1)
##         self.Relax(thisMover)

##         thisMover.SetDt(self.Parameters["timestep"])
##         thisMover.SetMaxIterations(self.Parameters["maxQMIterations"])

        self.Relax(thisMover)

        if self.Parameters["strainSubdivisions"] > 1:
            self.AtomicPositionsBack1 = self.atoms.GetCartesianPositions().ConstructFromArray(self.atoms.GetCartesianPositions())
            self.AtomicPositionsBack2 = self.atoms.GetCartesianPositions().ConstructFromArray(self.atoms.GetCartesianPositions())
            self.AtomicPositionsBack1.Copy(self.atoms.GetCartesianPositions())


    def CheckMinimized(self,GBEnergy):
        if self.minFinderEnergy < GBEnergy:
            self.ioControl.Print(1,"WARNING: grains may not be optimized! minFinderEnergy = " + repr(self.minFinderEnergy) + " current potential energy = " + repr(GBEnergy))
      


    def Anneal(self, nSteps):
        # We relax using Verlet, with a bit of randomization using AndersonDynamics
        # (which is implemented as a constraint)
        andersen = self.MD.AndersenDynamics(self.Parameters["andersenFrequency"], self.Parameters["andersenSigma"],self.Parameters["randomNumberSeed"])
        self.bulk.SetConstraint(andersen)
        for i in xrange(nSteps):
            self.verlet.Move(self.atoms) 

        #we remove the Andersen constraint
        constraint = self.MD.Constraint()
        self.bulk.SetConstraint(constraint)


    

    def MoveConstrainedZones(self,increment):
        #moves entire constrained zone by the same amount
        #positions1 = self.constrZone1.GetCartesianPositions()
        #avg1 = positions1.Average()
        #disp1 = increment * (avg1[0] - self.origin[0])
        #positions2 = self.constrZone2.GetCartesianPositions()
        #avg2 = positions2.Average()
        #disp2 = increment * (avg2[0] - self.origin[0])
        #disp=(disp1-disp2)/2.0 #otherwise it will just move off to the side with rollered bc!
        disp = -increment * (self.Parameters["width"]+self.grainMaker.constrLayerWidth)

        for i in xrange(self.constrZone1.NumberOfLeaves()):
            numAtoms1 = self.constrZone1.GetLeaf(i).GetNumber()
            positions1 = self.constrZone1.GetLeaf(i).GetCartesianPositions()
            self.ioControl.Print(4,"Moving " + repr(numAtoms1) + " by " + repr(disp))
            for j in xrange(numAtoms1):
                positions1[j][0] += disp
            self.constrZone1.GetLeaf(i).GetConstraint().SetFixedPositions(positions1)
        
        for i in xrange(self.constrZone2.NumberOfLeaves()):
            numAtoms2 = self.constrZone2.GetLeaf(i).GetNumber()
            positions2 = self.constrZone2.GetLeaf(i).GetCartesianPositions()
            self.ioControl.Print(4,"Moving " + repr(numAtoms2) + " by " + repr(-disp))
            for j in xrange(numAtoms2):
                positions2[j][0] -= disp
            self.constrZone2.GetLeaf(i).GetConstraint().SetFixedPositions(positions2)

        
    def IncrementStrain(self,increment):
        self.ioControl.Print(2,"Straining by " + repr(increment))

        if self.DIMENSION == 2:
            strainTensor = array([[increment, 0.],[0.,0.]])
        if self.DIMENSION == 3:
            strainTensor = array([[increment, 0.,0.],[0.,0.,0.],[0.,0.,0.]])

        #Must stretch bulk and contrained zones separately.  Adding and removing constraints causes Seg Fault 
        self.MD.LinearPositionTransformer(strainTensor).Transform(self.bulk)
        self.MoveConstrainedZones(increment)
        
        self.Relax(self.mover)
        self.atoms.Notify()
        
            


    def CalculateStress(self):
        self.ioControl.Print(3,"Calculating Stress")

        stress1=0
        stress2=0

        for i in xrange(self.constrZone1.NumberOfLeaves()):
            stress1+=self.constrZone1.GetLeaf(i).GetConstraint().GetTotalForce(0)
            stress2+=self.constrZone2.GetLeaf(i).GetConstraint().GetTotalForce(0)

        if self.Parameters["BC"]=="rollered":
            length = self.Parameters["length"] + 2*self.grainMaker.constrLayerWidth
            height = self.Parameters["height"] + 2*self.grainMaker.constrLayerWidth
        else:
            length = self.Parameters["length"]
            height = self.Parameters["height"]

        if self.DIMENSION == 2:
            stress1 /= length
            stress2 /= length
        if self.DIMENSION == 3:
            stress1 /= length*height
            stress2 /= length*height
        return stress1



    def CalculateStressStrainCurve(self,stressStrainFileName,initialStrain=0):

        if self.Parameters["BC"]!="rollered":
            self.grainMaker.FixSides()
        else:
            self.grainMaker.ImplementRolleredAndRigidConstraints()
            
        self.ioControl.OpenFile(stressStrainFileName)

        stress = self.CalculateStress()
        strain = initialStrain

        self.ioControl.Print(1,"stress = " + repr(stress) + " strain = " + repr(strain))
        if initialStrain == 0:
            self.ioControl.WriteToFile(stressStrainFileName,repr(strain) + " " + repr(stress) + "\n")
            self.ioControl.FlushFile(stressStrainFileName)

        stressStrain = [(strain,stress)]
        self.jumpBackFrames = []
        peakStress = 0

        for i in xrange(initialStrain/self.Parameters["strainIncrement"],self.Parameters["maxStrainSteps"]):
            if self.Parameters["strainSubdivisions"] > 1:
                self.SaveAtomicPositions()
            self.IncrementStrain(self.Parameters["strainIncrement"])
            stress = self.CalculateStress()
            strain = self.Parameters["strainIncrement"]*(i+1)
            stressStrain.append((strain,stress))
            if stress > peakStress: peakStress = stress
            self.ioControl.Print(1,"stress = " + repr(stress) + " strain = " + repr(strain))
            self.ioControl.WriteToFile(stressStrainFileName,repr(strain) + " " + repr(stress) + "\n")
            self.ioControl.FlushFile(stressStrainFileName)
            if (stress-stressStrain[-2][1] <= 0 and stressStrain[-2][1] > .5*peakStress and self.Parameters["strainSubdivisions"] > 1):
                if abs(strain-stressStrain[-3][0]-2*self.Parameters["strainIncrement"]) < 1e-6:
                    # go back two big steps
                    self.ioControl.Print(1,"Going Back two steps...")
                    self.RestoreAtomicPositions(2)
                    stressStrain = stressStrain[0:-2]
                    self.jumpBackFrames.append((self.Parameters["frameNumber"],2))
                else:
                    # go back one big step
                    self.ioControl.Print(1,"Going Back one step...")
                    self.RestoreAtomicPositions(1)
                    stressStrain = stressStrain[0:-1]
                    self.jumpBackFrames.append((self.Parameters["frameNumber"],1))
                #self.Notify()
                maxStrain = strain
                strain = stressStrain[-1][0]
                self.ioControl.Print(1,"stress = " + repr(stressStrain[-1][1]) + " strain = " + repr(strain))
                while strain < maxStrain:
                    if (maxStrain - strain) <= 2*self.Parameters["strainIncrement"]/self.Parameters["strainSubdivisions"]:
                        self.SaveAtomicPositions()
                    self.IncrementStrain(self.Parameters["strainIncrement"]/self.Parameters["strainSubdivisions"])
                    stress = self.CalculateStress()
                    strain += self.Parameters["strainIncrement"]/self.Parameters["strainSubdivisions"]
                    stressStrain.append((strain,stress))
                    self.ioControl.Print(1,"stress = " + repr(stress) + " strain = " + repr(strain))
                    self.ioControl.WriteToFile(stressStrainFileName,repr(strain) + " " + repr(stress) + "\n")
                    self.ioControl.FlushFile(stressStrainFileName)
            elif stress < self.Parameters["minStress"]:
                break

                    

    def SaveAtomicPositions(self):
        self.AtomicPositionsBack2.Copy(self.AtomicPositionsBack1)
        self.AtomicPositionsBack1.Copy(self.atoms.GetCartesianPositions())


    def RestoreAtomicPositions(self,num):
        if num==2:
            fixedPositions1 = self.AtomicPositionsBack2.GetLeaf(2)
            fixedPositions2 = self.AtomicPositionsBack2.GetLeaf(3)
            self.constrZone1.GetConstraint().SetFixedPositions(fixedPositions1)
            self.constrZone2.GetConstraint().SetFixedPositions(fixedPositions2)
            self.atoms.SetCartesianPositions(self.AtomicPositionsBack2)
        if num==1:
            fixedPositions1 = self.AtomicPositionsBack1.GetLeaf(2)
            fixedPositions2 = self.AtomicPositionsBack1.GetLeaf(3)
            self.constrZone1.GetConstraint().SetFixedPositions(fixedPositions1)
            self.constrZone2.GetConstraint().SetFixedPositions(fixedPositions2)
            self.atoms.SetCartesianPositions(self.AtomicPositionsBack1)


    def CalculateGrainBoundaryEnergy(self):
        numberOfAtoms1 = self.atoms.GetLeaf(0).GetNumber()
        totalEnergy = 0
        grainEnergy1 = 0
        grainEnergy2 = 0

        for i in xrange(numberOfAtoms1):
            grainEnergy1 += self.potential.CalculateAtomicEnergy(self.atoms,i) - self.grainMaker.bulkEnergy1
            self.ioControl.Print(6,repr(self.potential.CalculateAtomicEnergy(self.atoms,i)) + "- " +  repr(self.grainMaker.bulkEnergy1) + " = " + repr(self.potential.CalculateAtomicEnergy(self.atoms,i) - self.grainMaker.bulkEnergy1))

        numberOfAtoms2 = self.atoms.GetLeaf(1).GetNumber()
        for i in xrange(numberOfAtoms2):
            grainEnergy2 += self.potential.CalculateAtomicEnergy(self.atoms,i+numberOfAtoms1) - self.grainMaker.bulkEnergy2
            self.ioControl.Print(6,repr(self.potential.CalculateAtomicEnergy(self.atoms,i+numberOfAtoms1)) + "- " +  repr(self.grainMaker.bulkEnergy2) + " = " + repr(self.potential.CalculateAtomicEnergy(self.atoms,i+numberOfAtoms1) - self.grainMaker.bulkEnergy2))

        if self.DIMENSION == 2:
            GBE1 = grainEnergy1/self.Parameters["length"]
            GBE2 = grainEnergy2/self.Parameters["length"]
            GrainBoundaryEnergy = GBE1+GBE2
            self.ioControl.Print(2,"Grain Boundary Energy per Length =" + repr(GBE1)+"+"+repr(GBE2)+"="+repr(GrainBoundaryEnergy))
        if self.DIMENSION == 3:
            GBE1 = grainEnergy1/(self.Parameters["length"]*self.Parameters["width"])
            GBE2 = grainEnergy2/(self.Parameters["length"]*self.Parameters["width"])
            GrainBoundaryEnergy = GBE1+GBE2
            self.ioControl.Print(2,"Grain Boundary Energy per Area =" + repr(GBE1)+"+"+repr(GBE2)+"="+repr(GrainBoundaryEnergy))
 
                       
        return GrainBoundaryEnergy

    def MigrateBack(self):
        self.grainMaker.migrator.MigrateBack()

    def Migrate(self):
        self.grainMaker.migrator.Migrate()

### END OF GrainBreaker CLASS ###


# if calling directly from command line or web service
if __name__ == "__main__":

    from IOControl import *

    grainBreaker = GrainBreaker(2)

    UserParameters = ProcessCommandLine(grainBreaker.Parameters)
    if UserParameters == None:
        UserParameters = {}
    grainBreaker.SetParameters(UserParameters)
    grainBreaker.MakeGrains()

    #grainBreaker.AddPlotAtomsObserver()
    #asciiObserver = ASCIIObserver.ASCIIObserver("test/",grainBreaker.Parameters)
    #grainBreaker.AddObserver(asciiObserver)
    #dxObserver = DXObserver.DXObserver("test/",grainBreaker.Parameters)
    #grainBreaker.AddObserver(dxObserver)

    grainBreaker.RelaxNewGrains()
    #grainBreaker.CalculateGrainBoundaryEnergy()
    grainBreaker.CalculateStressStrainCurve("test/stressStrain.dat")

    #dxObserver.WriteTail(grainBreaker.Parameters["numFrames"]+1,grainBreaker)

    grainBreaker.ioControl.Print(3,"It Worked!!!!")


