#GrainMaker.py

import sys
import BulkEnergy
sys.path.append("..")
from MDWebServiceDecorator import *

class GrainMaker(MDWebServiceDecorator):

    def __init__(self, DIMENSION, Parameters,ioControl,MD):
        self.Parameters = Parameters
        self.DIMENSION = DIMENSION
        self.ioControl = ioControl
        MDWebServiceDecorator.__init__(self, DIMENSION, Parameters,ioControl,MD)


    def MakeLeaves(self):


        # get rotations and sizes
        if self.DIMENSION == 2:
            rotation1 = self.GetRotationMatrix(self.Parameters["angle1"])
            rotation2 = self.GetRotationMatrix(self.Parameters["angle2"])
            if self.Parameters["yystrain"] == 0:
                length1 = self.Parameters["length"]
                length2 = self.Parameters["length"]
            else:
                length1 = self.Parameters["repeat1"]*self.Parameters["numRepeats1"]*self.latticeConst*self.Parameters["numRepeats"]
                length2 = self.Parameters["repeat2"]*self.Parameters["numRepeats2"]*self.latticeConst*self.Parameters["numRepeats"]
            rectSize1 = [self.Parameters["width"],length1]
            rectSize2 = [self.Parameters["width"],length2]
        if self.DIMENSION == 3:
            rotation1 = array(self.Parameters["rotation1"])
            rotation2 = array(self.Parameters["rotation2"])
            rectSize1 = [self.Parameters["width"],self.Parameters["length"],self.Parameters["height"]]
            rectSize2 = [self.Parameters["width"],self.Parameters["length"],self.Parameters["height"]]
        
        # declare the rectangular cluster initializer
        rectangle = self.MD.RectangularClusterInitializer(rectSize1, self.lattice1)
        if self.Parameters["parallel"]:
            pararect = self.MD.ParallelAtomsInitializer(rectangle)
            initializer = pararect
        else:
            initializer = rectangle

        # build the first crystal
        rectCenter1 = self.origin[:]
        rectCenter1[0] = self.origin[0] - (rectSize1[0] - self.Parameters["grainOverlap"])/2.
        self.ioControl.Print(4,"center of grain 1 = "+repr(rectCenter1[0])+" " + repr(rectCenter1[1]))
        rectSize1[0] = rectSize1[0]
        grain1 = self.MD.DynamicListOfAtoms();
        self.CreateCrystal(rectangle, initializer, self.lattice1, rotation1, rectSize1, rectCenter1, grain1)

        # build the constrained zone(s)
        if self.Parameters["BC"] != "rollered":
            rectCenter1[0] = self.origin[0] - rectSize1[0] + self.Parameters["grainOverlap"]/2. - self.constrLayerWidth/2.
            self.ioControl.Print(4,"center of constrZone 1 = "+repr(rectCenter1[0])+" " + repr(rectCenter1[1]))
            rectSize1[0] = self.constrLayerWidth
            self.CreateCrystal(rectangle, initializer, self.lattice1, rotation1, rectSize1, rectCenter1, self.constrZone1)
        else:
            self.MakeRolleredLeaves(rectangle,initializer,self.lattice1,rotation1,rectSize1,rectCenter1,self.rolleredLOAs1)
            rectCenter1[0] = self.origin[0] - rectSize1[0] + self.Parameters["grainOverlap"]/2. - self.constrLayerWidth/2.
            self.ioControl.Print(4,"center of constrZone 1 = "+repr(rectCenter1[0])+" " + repr(rectCenter1[1]))
            rectSize1[0] = self.constrLayerWidth
            for i in xrange(len(rectSize1)-1):
                rectSize1[i+1] += 2*self.constrLayerWidth
            self.CreateCrystal(rectangle, initializer, self.lattice1, rotation1, rectSize1, rectCenter1, self.constrZone1)
        
        # now build the second crystal
        rectangle.SetBravaisLattice(self.lattice2)
        rectCenter2 = self.origin[:]
        rectCenter2[0] =  self.origin[0] + (rectSize2[0] - self.Parameters["grainOverlap"])/2.
        self.ioControl.Print(4,"center of grain 2 = "+repr(rectCenter2[0])+" " + repr(rectCenter2[1]))
        rectSize2[0] = rectSize2[0]
        grain2 = self.MD.DynamicListOfAtoms()
        self.CreateCrystal(rectangle, initializer, self.lattice2, rotation2, rectSize2, rectCenter2, grain2)
        
        # build the constrained zone
        if self.Parameters["BC"] != "rollered":
            rectCenter2[0] = self.origin[0] + rectSize2[0] - self.Parameters["grainOverlap"]/2. + self.constrLayerWidth/2.
            self.ioControl.Print(4,"center of constrZone2 = "+repr(rectCenter2[0])+" " + repr(rectCenter2[1]))
            rectSize2[0] = self.constrLayerWidth
            self.CreateCrystal(rectangle, initializer, self.lattice2, rotation2, rectSize2, rectCenter2, self.constrZone2)
        else:
            self.MakeRolleredLeaves(rectangle,initializer,self.lattice2,rotation2,rectSize2,rectCenter2,self.rolleredLOAs2)
            rectCenter2[0] = self.origin[0] + rectSize2[0] - self.Parameters["grainOverlap"]/2. + self.constrLayerWidth/2.
            self.ioControl.Print(4,"center of constrZone2 = "+repr(rectCenter2[0])+" " + repr(rectCenter2[1]))
            rectSize2[0] = self.constrLayerWidth
            for i in xrange(len(rectSize2)-1):
                rectSize2[i+1] += 2*self.constrLayerWidth
            self.CreateCrystal(rectangle, initializer, self.lattice2, rotation2, rectSize2, rectCenter2, self.constrZone2)

        # now put together the bulk
        self.bulk.Attach(grain1)
        self.bulk.Attach(grain2)
        del grain1, grain2


    def MakeRolleredLeaves(self,rectangle, initializer, lattice, rotation, grainSize, grainCenter, rolleredLOAs):

        if grainCenter[0] > 0: sign=1
        else: sign=-1

        centers=[]
        sizes=[]
        self.constrDirs=[]

        xCenter=grainCenter[0]
        yCenters=[grainCenter[1],grainCenter[1]+self.Parameters["length"]/2.0+self.constrLayerWidth/2.0,grainCenter[1]-self.Parameters["length"]/2.0-self.constrLayerWidth/2.0]
        if self.DIMENSION == 3:
            zCenters=[grainCenter[2],grainCenter[2]+self.Parameters["height"]/2.0+self.constrLayerWidth/2.0,grainCenter[2]-self.Parameters["height"]/2.0-self.constrLayerWidth/2.0]
        else:
            zCenters=[]

        # first just find centers
        tempCenter=grainCenter[:]
        tempCenter[0]=xCenter
        for j in xrange(3):
            tempCenter[1]=yCenters[j]
            if self.DIMENSION==2:
                if tempCenter != grainCenter: centers.append(tempCenter[:])
            for k in xrange(len(zCenters)):
                tempCenter[2]=zCenters[k]
                if tempCenter != grainCenter: centers.append(tempCenter[:])

        # now find corresponding sizes
        for i in xrange(len(centers)):
            tempSize = grainSize[:]
            tempDir = zeros(self.DIMENSION, Int)
            temp=array(centers[i])-array(grainCenter)
            for coord in xrange(self.DIMENSION):
                if temp[coord] != 0:
                    tempSize[coord] = self.constrLayerWidth
                    tempDir[coord]=1
            sizes.append(tempSize[:])
            self.constrDirs.append(tempDir[:])


        for i in xrange(len(sizes)):
            self.CreateCrystal(rectangle, initializer, lattice, rotation, sizes[i], centers[i], rolleredLOAs[i])
            numOfAtoms=rolleredLOAs[i].GetNumber()
            self.ioControl.Print(4,"rolleredLOAs"+repr(i)+" has center="+repr(centers[i])+", size="+repr(sizes[i])+", numOfAtoms="+repr(numOfAtoms)+", and constrDirs " + repr(self.constrDirs[i]))


        

    def ScaleLength(self):
        # necessary because previously calculated geometries are in units of 1 lattice constant...
        self.SetPotentialAndLattice()
        self.Parameters["length"]*=self.latticeConst*self.Parameters["numRepeats"]
        self.Parameters["width"]*=self.latticeConst
        self.Parameters["lengthScaled"]=1


             
    def MakeGrains(self):

        # set PotentialAndLattice
        self.ioControl.Print(3,"Setting Potential And Lattice")
        self.SetPotentialAndLattice()
        self.lattice2 = self.lattice1 = self.lattice
        self.constrLayerWidth = self.Parameters["constrLayerWidth"] * self.potential.GetCutoffDistance()
        self.ioControl.Print(4,"constrLayerWidth = " + repr(self.constrLayerWidth))

        # Shift by tiny amount to avoid problems
        if self.DIMENSION==2:
            self.origin=[1.0e-6,1.3e-6]
        if self.DIMENSION==3:
            self.origin=[1.0e-6,1.3e-6,1.4e-6]
        self.lattice1.SetOrigin(self.Parameters["latticeOrigin1"])
        self.lattice2.SetOrigin(self.Parameters["latticeOrigin2"])

        #boundary conditions
        self.ioControl.Print(3,"Making Boundary Conditions")
        if self.Parameters["BC"]!="periodic":
            self.boundaryConditions = self.MD.FreeBoundaryConditions()
        if self.Parameters["BC"]=="periodic":
            self.boundaryConditions = self.MD.SmartPeriodicBoundaryConditions([0,1])
            self.boundaryConditions.SetLength(array([0,self.Parameters["length"]]))

        # ListOfAtoms
        self.ioControl.Print(3,"Making lists of atoms")
        self.atoms = self.MD.DynamicListOfAtoms(0)
        self.bulk = self.MD.DynamicListOfAtoms(0)
        self.constrZone1 = self.MD.DynamicListOfAtoms(0)
        self.constrZone2 = self.MD.DynamicListOfAtoms(0)
        if self.Parameters["BC"]=="rollered":
            self.rolleredZone1 = self.MD.DynamicListOfAtoms(0)
            self.rolleredZone2 = self.MD.DynamicListOfAtoms(0)
            self.rolleredLOAs1 = []
            self.rolleredLOAs2 = []
            if self.DIMENSION == 2: numLOAs = 2
            else: numLOAs = 8
            for i in xrange(numLOAs):
                self.rolleredLOAs1.append(self.MD.DynamicListOfAtoms(0))
                self.rolleredLOAs2.append(self.MD.DynamicListOfAtoms(0))

        self.MakeLeaves()


        #  attach Lists of Atoms
        self.atoms.Attach(self.bulk)
        if self.Parameters["BC"]=="rollered":
            for i in xrange(len(self.rolleredLOAs1)):
                self.rolleredZone1.Attach(self.rolleredLOAs1[i])
                self.rolleredZone2.Attach(self.rolleredLOAs2[i])
            self.atoms.Attach(self.rolleredZone1)
            self.atoms.Attach(self.rolleredZone2)                
        self.atoms.Attach(self.constrZone1)
        self.atoms.Attach(self.constrZone2)

        numBefore = self.atoms.GetNumber()
        self.MD.OverlapPruner(0.8*self.lattice1.GetMinInteratomicSpacing()).Transform(self.atoms)
        numAfter = self.atoms.GetNumber()
        self.ioControl.Print(2,"Pruning Removed " + repr(numBefore-numAfter) + " atoms")
        self.ioControl.Print(3,"Shape of Atoms = " + repr(self.atoms.GetShape()))


        # boundary conditions
        self.atoms.SetBoundaryConditions(self.boundaryConditions)

        # neighbor locator
        self.neighborLocator = self.MD.CellNeighborLocator(self.potential.GetCutoffDistance(), 0.05 * self.potential.GetCutoffDistance())
        if self.Parameters["parallel"]:
            self.domain = self.MD.HomogeneousDomainSubdivision(self.Parameters["cpulayout"],self.Parameters["myProc"], self.potential.GetCutoffDistance())
            self.synchronizer = self.MD.Synchronizer(self.atoms.GetCartesianPositions())
            self.migrator = self.MD.Migrator(self.domain, self.atoms.GetCartesianPositions())
            self.paraNeighborLocator = self.MD.ParallelNeighborLocator(self.migrator, self.neighborLocator)


        self.atoms.SetBoundaryConditions(self.boundaryConditions)
        if self.Parameters["parallel"]:
            Scientific.MPI.world.barrier() #wait till all processors reach this fuction call
            self.atoms.SetNeighborLocator(self.paraNeighborLocator)
        else:
           self.atoms.SetNeighborLocator(self.neighborLocator)
        self.neighborLocator.thisown = 0
        print 'on processor ', self.Parameters["myProc"], ', we have:', \
              len(self.atoms.GetCartesianPositions()), ' and ', \
              len(self.atoms.GetCartesianPositions().GetExtended().GetBranch(1)), \
              ' ghost atoms after migration'

        #movers
        qmStepsBetweenUpdate = 100
        self.qm = self.MD.QuickMinAtomsMover(self.moverpotential,self.Parameters["timestep"], self.Parameters["QMtolerance"])
        self.qm.SetMaxIterations(self.Parameters["maxQMIterations"])
        if self.ioControl.GetVerbosity() > 5:
            self.qm.Verbose()
            self.qm.SetStepsBetweenUpdate(qmStepsBetweenUpdate)

        self.verlet = self.MD.VerletAtomsMover(self.moverpotential, self.Parameters["timestep"], self.Parameters["MDIterations"])

        if self.Parameters["moverType"] == "verletThermalize":
            self.verletThermalize = self.MD.VerletThermalizeAtomsMover(self.moverpotential, self.Parameters["temperature"], self.Parameters["timestep"], self.Parameters["MDIterations"])
        elif self.Parameters["moverType"] == "hoover":
            self.hoover = self.MD.HooverAtomsMover(self.moverpotential, self.Parameters["temperature"],self.Parameters["timestep"], 1.0, self.Parameters["MDIterations"])

        self.langevin = self.MD.LangevinAtomsMover(self.moverpotential, self.Parameters["timestep"], self.Parameters["MDIterations"], self.Parameters["friction"], self.Parameters["temperature"])


    def FixSides(self):
        self.ioControl.Print(3,"Fixing Sides")
        self.fixedConstraint1 = self.MD.FixedBody(self.constrZone1.GetCartesianPositions())
        self.fixedConstraint2 = self.MD.FixedBody(self.constrZone2.GetCartesianPositions())
        self.constrZone1.SetConstraint(self.fixedConstraint1)
        self.constrZone2.SetConstraint(self.fixedConstraint2)

        

    def ImplementRolleredAndRigidConstraints(self):
        self.ioControl.Print(3,"Implementing Rollered And Rigid BC's")
        self.constraints1=[]
        self.constraints2=[]
        
        #the endcaps
        self.constraints1.append(self.MD.NonRotatingRigidBody())
        self.constrZone1.SetConstraint(self.constraints1[0])
        self.constraints2.append(self.MD.NonRotatingRigidBody())
        self.constrZone2.SetConstraint(self.constraints2[0])

        #the rest
        for i in xrange(len(self.constrDirs)):
            self.constraints1.append(self.MD.PartialFixedBody(self.rolleredLOAs1[i].GetCartesianPositions(),self.constrDirs[i]))
            self.rolleredLOAs1[i].SetConstraint(self.constraints1[i+1])
            self.constraints2.append(self.MD.PartialFixedBody(self.rolleredLOAs2[i].GetCartesianPositions(),self.constrDirs[i]))
            self.rolleredLOAs2[i].SetConstraint(self.constraints2[i+1])


    def ImplementRolleredAndFixedConstraints(self):
        self.ioControl.Print(3,"Implementing Rollered And Fixed BC's")
        self.constraints1=[]
        self.constraints2=[]
        
        #the endcaps
        self.constraints1.append(self.MD.FixedBody(self.constrZone1.GetCartesianPositions()))
        self.constrZone1.SetConstraint(self.constraints1[0])
        self.constraints2.append(self.MD.FixedBody(self.constrZone2.GetCartesianPositions()))
        self.constrZone2.SetConstraint(self.constraints2[0])

        #the rest
        for i in xrange(len(self.constrDirs)):
            self.constraints1.append(self.MD.PartialFixedBody(self.rolleredLOAs1[i].GetCartesianPositions(),self.constrDirs[i]))
            self.rolleredLOAs1[i].SetConstraint(self.constraints1[i+1])
            self.constraints2.append(self.MD.PartialFixedBody(self.rolleredLOAs2[i].GetCartesianPositions(),self.constrDirs[i]))
            self.rolleredLOAs2[i].SetConstraint(self.constraints2[i+1])

            
    def ImplementRolleredConstraints(self):
        self.ioControl.Print(3,"Implementing Rollered BC's")
        self.constraints1=[]
        self.constraints2=[]
        
        #the endcaps
        if self.DIMENSION==2:
            dir = (1,0)
        elif self.DIMENSION==3:
            dir = (1,0,0)
        self.constraints1.append(self.MD.PartialFixedBody(self.constrZone1.GetCartesianPositions(),dir))
        self.constrZone1.SetConstraint(self.constraints1[0])
        self.constraints2.append(self.MD.PartialFixedBody(self.constrZone2.GetCartesianPositions(),dir))
        self.constrZone2.SetConstraint(self.constraints2[0])

        #the rest
        for i in xrange(len(self.constrDirs)):
            self.constraints1.append(self.MD.PartialFixedBody(self.rolleredLOAs1[i].GetCartesianPositions(),self.constrDirs[i]))
            self.rolleredLOAs1[i].SetConstraint(self.constraints1[i+1])
            self.constraints2.append(self.MD.PartialFixedBody(self.rolleredLOAs2[i].GetCartesianPositions(),self.constrDirs[i]))
            self.rolleredLOAs2[i].SetConstraint(self.constraints2[i+1])            



    def FixAllSides(self):
        self.constraints1=[]
        self.constraints2=[]
        
        self.constraints1.append(self.MD.FixedBody(self.constrZone1.GetCartesianPositions()))
        self.constrLOAs1[i].SetConstraint(self.constraints1[i])
        self.constraints2.append(self.MD.FixedBody(self.constrZone2.GetCartesianPositions()))
        self.constrLOAs2[i].SetConstraint(self.constraints2[i])

        for i in xrange(len(self.constrDirs)):
            self.constraints1.append(self.MD.FixedBody(self.rolleredLOAs1[i].GetCartesianPositions()))
            self.rolleredLOAs1[i].SetConstraint(self.constraints1[i+1])
            self.constraints2.append(self.MD.FixedBody(self.rolleredLOAs2[i].GetCartesianPositions()))
            self.rolleredLOAs2[i].SetConstraint(self.constraints2[i+1])


    def CalculateBulkEnergies(self):

        bulkEnergyMeasurer = BulkEnergy.BulkEnergy(self.DIMENSION,self.Parameters,self.ioControl,self.MD)
        self.bulkEnergy1,self.bulkEnergy2 = bulkEnergyMeasurer.MeasureBulkEnergies()
        self.Parameters["measureBulkEnergies"]=0


# end of GrainMaker class
