import  sys
sys.path.append("/afs/msc.cornell.edu/home/jsethna/val/ASP/MDWebServices/GrainBreaker/")
sys.path.append("/afs/msc.cornell.edu/home/jsethna/val/ASP/MDWebServices/Utilities/")
import GrainBreaker
import ASCIIObserver
import DXObserver
from Numeric import *
from IOControl import *


class MinimumFinder:

    def __init__(self,Parameters,dir):
        self.Parameters = Parameters
        self.ioControl = IOControl(self.Parameters["verbose"],dir)

    def FindMaximumOffset(self):

        # find Maximum Offset to search through

        repeat1 = self.Parameters["repeat1"]
        repeat2 = self.Parameters["repeat2"]
        numRepeats1 = self.Parameters["numRepeats1"]
        numRepeats2 = self.Parameters["numRepeats2"]

        if repeat1*numRepeats1 > repeat2*numRepeats2:
            repeat1 *= (1-self.Parameters["yystrain"])
            repeat2 *= (1+self.Parameters["yystrain"])
        else:
            repeat1 *= (1+self.Parameters["yystrain"])
            repeat2 *= (1-self.Parameters["yystrain"])

        if self.Parameters["repeat2"] > self.Parameters["repeat1"]:
            l1=repeat1
            l2=repeat2
        else:
            l1=repeat2
            l2=repeat1

        if l2 % l1 <= 1e-3 or abs(l2 % l1 - l1) <= 1e-3:
            maxOffset = l1

        else:
            offsets = (abs(l2 - l1*floor(l2/l1)),abs(l2 - l1*ceil(l2/l1)))
            maxOffset = min(offsets)
##             offsets = [l1,l2]
##             numRepeats1 = self.Parameters["numRepeats1"]
##             numRepeats2 = self.Parameters["numRepeats2"]
##             for i in xrange(1,numRepeats1+1):
##                 for j in xrange(i+1,numRepeats2):
##                     offsets.append(abs(i*l1-j*l2))


        return maxOffset

    def FindMinimum(self,bulkEnergy1,bulkEnergy2,positions,minOffset=0,maxOffset=None):

        if maxOffset==None:
            maxOffset = self.FindMaximumOffset()

        if minOffset != maxOffset:
            numOffsets = self.Parameters["minSearchSteps"]*ceil(maxOffset-minOffset+1)
            offset = (maxOffset-minOffset)/(numOffsets-1)
        else:
            numOffsets = 1
            offset = 0
        minEnergy = 1e6
        minimizingOffset = 0
        oldWidth = self.Parameters["width"]
        self.Parameters["width"]=self.Parameters["minimizerWidth"]
        oldPlotObserver = self.Parameters["plotObserver"]
        self.Parameters["plotObserver"]=0

        self.ioControl.OpenFile("MinFinderEnergy.dat")

            
        for i in xrange(numOffsets):

            grain = GrainBreaker.GrainBreaker(2)
            grain.SetParameters(self.Parameters)


            if i == 0:
                grain.ioControl.Print(2,"angles = " + repr(grain.Parameters["angle1"]) + " " + repr(grain.Parameters["angle2"]) + " length = " + repr(grain.Parameters["length"]) + " maxOffset = " + repr(maxOffset) + " offset = " + repr(offset))

            if self.Parameters["verbose"] >= 5:
                grain.ioControl.SetDirectory("origin-"+repr(i))
                asciiObserver = ASCIIObserver.ASCIIObserver(self.Parameters,grain.ioControl)
                grain.AddObserver(asciiObserver)    

            grain.MakeGrains()
            grain.grainMaker.bulkEnergy1 = bulkEnergy1
            grain.grainMaker.bulkEnergy2 = bulkEnergy2
            delta0 = self.FindCurrentOffset(grain)
            delta1 = minOffset + i*offset*grain.grainMaker.latticeConst
            grain.MD.Translator(array([0,delta1])).Transform(grain.atoms.GetLeaf(0))     
            grain.MD.Translator(array([0,delta1])).Transform(grain.atoms.GetLeaf(2))
            grain.Notify()
            grain.atoms.Notify()
            energy1 = grain.CalculateGrainBoundaryEnergy()     
            grain.RelaxNewGrains()
            delta2 = self.FindCurrentOffset(grain) - delta0
            if delta2 < 0: delta2 += grain.Parameters["length"]
            energy2 = grain.CalculateGrainBoundaryEnergy()
            grain.ioControl.Print(2,"Initial Offset = " + repr(delta1) + " Final Offset = " + repr(delta2) + " Energy = " + repr(energy2))
            self.ioControl.WriteToFile("MinFinderEnergy.dat",repr(delta1)+" "+repr(delta2)+" "+repr(energy1)+" "+repr(energy2)+" "+repr(grain.atoms.GetNumber())+"\n")
            self.ioControl.FlushFile("MinFinderEnergy.dat")

            if energy2 < minEnergy:
                minEnergy = energy2
                minimizingOffset = delta1
                # save minimized positions if minimizer width = width
                if oldWidth == self.Parameters["minimizerWidth"]:
                    positions.Copy(grain.atoms.GetCartesianPositions())
            #energiesAndOffsets.append((delta1,energy))

        self.Parameters["width"]= oldWidth
        self.Parameters["plotObserver"]= oldPlotObserver

        grain.ioControl.Print(2,"angles = " + repr(grain.Parameters["angle1"]) + " " + repr(grain.Parameters["angle2"]) + " minimizing offset = " + repr(minimizingOffset))
        return minimizingOffset,minEnergy


    def FindCurrentOffset(self,grain):

        constrZonePos1 = grain.constrZone1.GetCartesianPosition(0,1)
        constrZonePos2 = grain.constrZone2.GetCartesianPosition(0,1)
        return constrZonePos1-constrZonePos2
        
