# base class for WebServices

import sys
# some of these are for linux some are for windows, extras don't hurt
HOMEDIR = "/afs/msc.cornell.edu/home/jsethna/val/ASP/"
#HOMEDIR = "/home/val/work/"
#HOMEDIR = "H:/Users/vrc3/"
#sys.path.append("H:/Users/vrc3/DigitalMaterial/src/MD/SWIGBuild/")
#HOMEDIR = "T:/vrc3/"

#sys.path.append(HOMEDIR + "DigitalMaterial/src/MD/SWIGBuild/i686/obj2D")
#sys.path.append(HOMEDIR + "DigitalMaterial/src/MD/SWIGBuild/i686/obj3D")
sys.path.append(HOMEDIR + "DigitalMaterial/src/MD/SWIGBuild/")
sys.path.append(HOMEDIR + "MDWebServices/Utilities/")
sys.path.append(HOMEDIR + "MDWebServices/Observers/")
sys.path.append(HOMEDIR + "Python22Copy/Lib/site-packages/Numeric/")
sys.path.append(HOMEDIR + "FemLib-0.8/")

from Numeric import *
from IOControl import *
from Utilities import *
import ASCIIObserver
import DXObserver
import DXStressObserver
import ForceObserver
import EnergyObserver
import OFEMDASCIIObserver
import FemLib
#from XMLFemModel import *
import Scientific.MPI

class MDWebService:

    def __init__(self,DIMENSION,Parameters,ioControl):
        self.DIMENION = DIMENSION
        self.Parameters = Parameters
        self.ioControl = ioControl
        #self.ImportDigitalMaterial()
        self.Observers = []
        self.Parameters["frameNumber"] = 0
        #self.SetParameters({})
        self.Parameters["myProc"] = 0
        self.Parameters["nProcs"] = 1
        self.mpiworld = Scientific.MPI.world

    def SetParameters(self,UserParameters):

        #Overwrite the Default Parameters
        for key in UserParameters.keys():
            self.Parameters[key] = UserParameters[key]

        self.ioControl.SetVerbosity(self.Parameters["verbose"])
        self.ioControl.Print(4,"Parameters:")
        self.ioControl.Print(4,self.Parameters)
	if self.Parameters["parallel"]:
	    self.GetCPULayout()
            self.ioControl.SetMyProc(self.Parameters["myProc"])
            self.ioControl.Print(2,"nProcs: " + repr(self.Parameters["nProcs"]))
            #self.ioControl.SetDirectory(self.ioControl.directory + "/proc" + repr(self.Parameters["myProc"]))
        self.ioControl.SetDirectory(self.Parameters["directory"])



    def ImportDigitalMaterial(self):
        # Import proper version of DigitalMaterial
        if self.DIMENSION==2:
            sys.path.append(self.Parameters["DMDIR"]+"/src/MD/SWIGBuild/i686/obj2D")
            import MD2D
            self.MD = MD2D
        if self.DIMENSION==3:
            sys.path.append(self.Parameters["DMDIR"]+"/src/MD/SWIGBuild/i686/obj3D")
            import MD3D
            self.MD = MD3D


    def GetCPULayout(self):
	self.ioControl.Print(1,"Getting CPU layout")
	# dmProtocol is a global variable. It is set to a MPIProtocol if we
	# use mpipython, or a LocalProtocol otherwise.
	self.Parameters["nProcs"] = self.MD.cvar.dmProtocol.GetNumberOfProcessors()
	self.Parameters["myProc"] = self.MD.cvar.dmProtocol.GetProcessorNumber()

	# Choose a cpu layout (ie how many processors in each direction)
	tmp0 = int(self.Parameters["nProcs"]**(1/3.))
	tmp1 = int(sqrt(self.Parameters["nProcs"]/tmp0))
	self.Parameters["cpulayout"] = array([tmp0, tmp1, self.Parameters["nProcs"]/tmp0/tmp1])


    def AddObserver(self,observer):
        self.Observers.append(observer)


    def Notify(self):
        for observer in self.Observers:
            observer.Update(self)
        if len(self.Observers) > 0:
            self.Parameters["frameNumber"] += 1


    def AddPlotAtomsObserver(self):           
        self.plotter = self.MD.PlotAtomsObserver(-30,30,-30,30,1)
        self.atoms.AddObserver(self.plotter)
        self.atoms.Notify()




    def Relax(self,mover):
        if self.Parameters["parallel"]: self.mpiworld.barrier()
        mover.Move(self.atoms)
        if self.Parameters["parallel"]: self.mpiworld.barrier()
        E = self.moverpotential.CalculateEnergy(self.atoms)
        K = self.atoms.GetKineticEnergy()
        string = "Energy = " + repr(E) + "+" + repr(K) + "=" + repr(E+K)
        if self.Parameters["moverType"] == "qm":
            string +=", Convergence = " + repr(mover.GetConvergence())
        self.ioControl.Print(2, string)
        self.atoms.Notify()
        self.Notify()
        self.Save()



    def Save(self):
        if self.Parameters["parallel"]: self.MigrateBack()
        
        if self.ioControl.myProc == 0:
            parameterFile = open(self.ioControl.directory+"/parameters.pickle","w")
            pickle.dump(self.Parameters,parameterFile)
            filesFile = open(self.ioControl.directory+"/files.pickle","w")
            pickle.dump(self.ioControl.files.keys(),filesFile)

            atomsFile = open(self.ioControl.directory+"/lastpositions.dat","w")#+repr(self.Parameters["myProc"])+".dat","w")
            for atomIdx in xrange(self.atoms.GetNumber()):
                line = ""
                for coord in xrange(self.DIMENSION):
                    line = line + repr(self.atoms.GetCartesianPosition(atomIdx,coord)) + " "
                line = line + "\n"
                atomsFile.write(line)

        if self.Parameters["parallel"]: self.Migrate()


    def LoadParameters(self):
        parameterFile = open(self.ioControl.directory+"/parameters.pickle","r")
        Parameters = pickle.load(parameterFile)
        self.SetParameters(Parameters)
        filesFile = open(self.ioControl.directory+"/files.pickle","r")
        fileKeys = pickle.load(filesFile)
        for key in fileKeys:
            self.ioControl.files[key] = 0


    def LoadAtomicPositions(self):
        if self.Parameters["parallel"]:
            self.MigrateBack()
            self.mpiworld.barrier()
            print "after first barrier ", repr(self.Parameters["myProc"])
            sys.stdout.flush()

        if self.ioControl.myProc == 0:
            positions = self.ioControl.GetDMArrayFromFile(self.ioControl.directory+"/lastpositions.dat", self.MD,self.atoms.GetCartesianPositions())
        else:
            positions = self.atoms.GetCartesianPositions().ConstructFromArray(self.atoms.GetCartesianPositions())

        self.atoms.SetCartesianPositions(positions)
        if self.Parameters["parallel"]:
            self.Migrate()
            self.mpiworld.barrier()
            print "after second barrier ", repr(self.Parameters["myProc"])
            sys.stdout.flush()

##         for proc in xrange(self.Parameters["nProcs"]):
##             if self.Parameters["parallel"]: self.mpiworld.barrier()
##             if self.Parameters["myProc"] == proc:
##                 atomsFile = open(self.ioControl.directory+"/lastpositions"+repr(self.Parameters["myProc"])+".dat","r")
##                 print "opened positions file " + repr(self.Parameters["myProc"])
##                 sys.stdout.flush()        
##                 for atomIdx in xrange(self.atoms.GetNumber()):
##                     print "getting atom " + repr(atomIdx) + " out of " + repr(self.atoms.GetNumber()) + " for proc " + repr(self.Parameters["myProc"])
##                     line = atomsFile.readline()
##                     print line
##                     position = string.splitfields(line)
##                     print position
##                     for coord in xrange(3):
##                         print "setting coord"
##                         self.atoms.SetCartesianPosition(atomIdx,coord,float(position[coord]))
##                         print "set coord"
##                 print "loaded positions for " + repr(self.Parameters["myProc"])
##                 sys.stdout.flush()


        
