# base class for WebServiceDecorators

import sys
# some of these are for linux some are for windows, extras don't hurt
HOMEDIR = "/afs/msc.cornell.edu/home/jsethna/val/ASP/"
#HOMEDIR = "H:/Users/vrc3/"
#sys.path.append("H:/Users/vrc3/DigitalMaterial/src/MD/SWIGBuild/")
#HOMEDIR = "T:/vrc3/"

#sys.path.append(HOMEDIR + "DigitalMaterial/src/MD/SWIGBuild/i686/obj2D")
#sys.path.append(HOMEDIR + "DigitalMaterial/src/MD/SWIGBuild/i686/obj3D")
sys.path.append(HOMEDIR + "DigitalMaterial/src/MD/SWIGBuild/")
sys.path.append(HOMEDIR + "MDWebServices/Utilities/")
sys.path.append(HOMEDIR + "MDWebServices/Observers/")
sys.path.append(HOMEDIR + "Python22Copy/Lib/site-packages/Numeric/")
sys.path.append(HOMEDIR + "FemLib-0.8/")

from Numeric import *
from IOControl import *
from Utilities import *
import ASCIIObserver
import DXObserver
import ForceObserver
import EnergyObserver
import OFEMDASCIIObserver
#import FemLib
#from XMLFemModel import *
import Scientific.MPI

class MDWebServiceDecorator:

    def __init__(self,DIMENSION,Parameters,ioControl,MD):
        self.DIMENION = DIMENSION
        self.Parameters = Parameters
        self.ioControl = ioControl
        self.MD = MD


    def SetPotentialAndLattice(self):
        if self.DIMENSION == 2:
            self.potential = self.MD.CutLennardJonesPotential()
            self.latticeConst = self.potential.GetLengthScale()
            self.lattice = self.MD.TriangularLattice(self.latticeConst)
        if self.DIMENSION == 3:
            if self.Parameters["material"][0:4] == "EDIP":
                self.potential = self.MD.SiliconEDIPPotential()
                self.latticeType = "Diamond"
                nearestNeighborFactor =  4./sqrt(3.)
                nAtomsPerCubicCell = 8
            elif self.Parameters["material"] == "SW_nat":
                self.potential = self.MD.StillingerWeberPotential(self.Parameters["brittleFactor"], "natural")
                self.latticeType = "Diamond"
                nearestNeighborFactor =  4./sqrt(3.)
                nAtomsPerCubicCell = 8
                self.latticeConst = 2.5922152
            elif self.Parameters["material"] == "SW_evA":
                self.potential = self.MD.StillingerWeberPotential(self.Parameters["brittleFactor"], "ev_angstroms1")
                self.latticeType = "Diamond"
                nearestNeighborFactor =  4./sqrt(3.)
                nAtomsPerCubicCell = 8
                self.latticeConst = 5.43095006
            elif self.Parameters["material"] == "CLJ":
                self.potential = self.MD.CutLennardJonesPotential()
                self.latticeType = "FCC"
                nearestNeighborFactor =  sqrt(2.)
                nAtomsPerCubicCell = 4
            elif self.Parameters["material"] == "HolianCLJ":
                self.potential = self.MD.HolianCutLennardJonesPotential()
                self.latticeType = "FCC"
                nearestNeighborFactor =  sqrt(2.)
                nAtomsPerCubicCell = 4
            elif self.Parameters["material"] == "MEAM_Al":
                AlParameters = self.MD.MEAM_ParameterList(3.580, 2.86, 4.61, 1.07, 2.21, 2.2, 6.0, 2.2, 1, -1.78, -2.21, 8.01, 12, "FCC", 4.0, 0.1, 2.0, 2.8)
                self.potential = self.MD.MEAMPotential(AlParameters)
                self.latticeType = "FCC"
                nearestNeighborFactor =  sqrt(2.)
                nAtomsPerCubicCell = 4
            elif self.Parameters["material"] == "MEAM_Si":
                SiParameters = self.MD.MEAM_ParameterList(4.63, 2.35, 4.87, 1.0, 4.4, 5.5, 5.5, 5.5, 1., 3.13, 4.47, -1.80, 4, "Diamond", 4.0, 0.1, 2.0, 2.8)
                self.potential = self.MD.MEAMPotential(SiParameters)
                self.latticeType = "Diamond"
                nearestNeighborFactor =  4./sqrt(3.)
                nAtomsPerCubicCell = 8
            elif self.Parameters["material"] == "MEAM_W":
                WParameters = self.MD.MEAM_ParameterList(8.660, 2.74, 5.63, 0.98, 3.98, 1.0, 1.0, 1.0, 1, 3.16, 8.25, -2.70, 8, "BCC", 4.0, 0.1, 2.0, 2.8)
                self.potential = self.MD.MEAMPotential(WParameters)
                self.latticeType = "BCC"
                nearestNeighborFactor =  2./sqrt(3.)
                nAtomsPerCubicCell = 1
            elif self.Parameters["material"] == "MEAM_Fe":
                FeParameters = self.MD.MEAM_ParameterList(4.290, 2.48, 5.07, 0.89, 2.94, 1.0, 1.0, 1.0, 1, 3.94, 4.12, -1.50, 8, "BCC", 4.0, 0.1, 2.0, 2.8)
                self.potential = self.MD.MEAMPotential(FeParameters)
                self.latticeType = "BCC"
                nearestNeighborFactor =  2./sqrt(3.)
                nAtomsPerCubicCell = 1
            else:
                self.ioControl.Print(0, "trying to create potential: unknown material")
                raise StandardError

            # now can make the lattice
            if self.Parameters["material"][0:2] != "SW":
                self.latticeConst = self.potential.GetLengthScale() * nearestNeighborFactor
            makeLatticeCommand = "self.lattice = self.MD." + self.latticeType + "Lattice(self.latticeConst)"
            exec makeLatticeCommand

        if self.Parameters["parallel"]:
            self.parapotential = self.MD.ParallelPotential(self.potential)
            self.moverpotential = self.parapotential
        else:
            self.moverpotential = self.potential


    def GetRotationMatrix(self,angle): #currently only for 2D!!
        pi = 4*arctan(1)
        theta = (pi/180.0)*angle
        sintheta = sin(theta)
        costheta = cos(theta)
        rotation = array([[costheta,sintheta],[-sintheta,costheta]])
        return rotation



    def CreateCrystal(self,rect, init, lattice, rotation, size, center, crystal):
        rect.SetCenter(center)
        rect.SetRectangleSize(size)
        lattice.ResetOrientation()
        self.ioControl.Print(6,"LatticeVectors before rotation: " + repr(lattice.GetLatticeVectors()))
        lattice.RotateLatticeVectors(rotation)
        self.ioControl.Print(6,"LatticeVectors after rotation: " + repr(lattice.GetLatticeVectors()))
        init.Create(crystal) 


