#AtomsFiller.py

import sys
sys.path.append("..")
from MDWebServiceDecorator import *
import __builtin__
#import Scientific.MPI

class AtomsDecorator(MDWebServiceDecorator):

    def __init__(self,Parameters,ioControl,MD,model):
        self.Parameters = Parameters
        self.DIMENSION = 3
        self.ioControl = ioControl
        self.model = model
        self.MD = MD
        MDWebServiceDecorator.__init__(self,self.DIMENSION,self.Parameters,self.ioControl,self.MD)
        
                    
    def Decorate(self):
        self.InitializeLOAs()
        self.InitializeConstrainedLeaves()
        self.InitializeNeighborLocator()
        self.InitializeMover()
        

    def InitializeLOAs(self):

        #set PotentialAndLattice
        self.SetPotentialAndLattice()

        # ListsOfAtoms (leaves will be created in a separate function)
        self.BulkLOAs = []
        for i in xrange(self.model.NumMats()):
            self.BulkLOAs.append(self.MD.DynamicListOfAtoms(0))

        self.MakeLeaves()

        
        # attach Lists of Atoms
        self.bulk = self.MD.DynamicListOfAtoms(0)
        for i in xrange(self.model.NumMats()):
            self.bulk.Attach(self.BulkLOAs[i])
        self.ioControl.Print(2,"shape of atoms" + repr(self.bulk.GetShape()))
            

        # Constrained Lists of Atoms (leaves will be created separately)
        # numConstrainedLeaves set by child class
        self.ConstrLOAs = []
        for i in xrange(self.numConstrainedLeaves):
            self.ConstrLOAs.append(self.MD.DynamicListOfAtoms(0))


        self.atoms = self.MD.DynamicListOfAtoms(0)
        self.atoms.Attach(self.bulk)
        #self.atoms.Attach(self.constrZones)
        self.ioControl.Print(3,"shape of atoms" + repr(self.atoms.GetShape()))



    def InitializeNeighborLocator(self):

        # boundary conditions
        self.boundaryConditions = self.MD.FreeBoundaryConditions()

        # neighbor locator
        self.neighborLocator = self.MD.CellNeighborList(self.potential.GetCutoffDistance(), 0.05 * self.potential.GetCutoffDistance())

        if self.Parameters["parallel"]:
            self.domain = self.MD.HomogeneousDomainSubdivision(self.Parameters["cpulayout"],self.Parameters["myProc"], self.potential.GetCutoffDistance())
            self.synchronizer = self.MD.Synchronizer(self.atoms.GetCartesianPositions())
            self.migrator = self.MD.Migrator(self.domain, self.atoms.GetCartesianPositions())
            self.paraNeighborLocator = self.MD.ParallelNeighborLocator(self.migrator, self.neighborLocator)
 
	# prune just in case
        numBefore = self.atoms.GetNumber()
        self.MD.OverlapPruner(0.7*self.lattice.GetMinInteratomicSpacing()).Transform(self.atoms)
        numRemoved = numBefore-self.atoms.GetNumber()
        self.ioControl.Print(3,repr(numRemoved) + " atoms removed by pruner")
        self.ioControl.Print(3,self.atoms.GetShape())


        # set NL and BC etc
        self.atoms.SetBoundaryConditions(self.boundaryConditions)
        if self.Parameters["parallel"]:
            Scientific.MPI.world.barrier() #wait till all processors reach this fuction call
            self.atoms.SetNeighborLocator(self.paraNeighborLocator)
        else:
            self.atoms.SetNeighborLocator(self.neighborLocator)
        self.neighborLocator.thisown = 0

        # print number of real and ghost atoms on each processor
        print 'on processor ', self.Parameters["myProc"], ', we have:', \
              len(self.atoms.GetCartesianPositions()), ' and ', \
              len(self.atoms.GetCartesianPositions().GetExtended().GetBranch(1)), \
              ' ghost atoms after migration'
    


    def InitializeConstrainedLeaves(self):

        self.MakeConstrainedLeaves()

        # attach Lists of Atoms
        self.constrZones = self.MD.DynamicListOfAtoms(0)
        for i in xrange(len(self.ConstrLOAs)):
            self.constrZones.Attach(self.ConstrLOAs[i])

        #self.atoms = self.MD.DynamicListOfAtoms(0)
        #self.atoms.Attach(self.bulk)
        self.atoms.Attach(self.constrZones)
        self.ioControl.Print(3,"shape of atoms" + repr(self.atoms.GetShape()))



    def InitializeMover(self):

        #mover parameters
        self.tol = 1.0e-10
        self.stepsPerStressIntensity = 5
        self.maxQMIterations = 600
        self.maxCGIterations = 100
        self.maxMDIterations = 200
        self.qmStepsBetweenUpdate = 100
        self.temperature = 0.5
        self.friction = 0.1
        
        # movers -- quick min, conjugate gradient, verlet, langevin
        # if self.Parameters["MDMoverType"] == "qm" or self.Parameters["MDMoverType"] == "combo":
        # for now only quick min is available
        self.qm = self.MD.QuickMinAtomsMover(self.moverpotential,self.Parameters["timestep"], self.Parameters["QMtolerance"])
        self.qm.SetMaxIterations(self.Parameters["maxQMIterations"])
        if self.ioControl.GetVerbosity() > 5:
            self.qm.Verbose()
        self.qm.SetStepsBetweenUpdate(self.qmStepsBetweenUpdate)

##         if self.Parameters["MDMoverType"] == "cg" or self.Parameters["MDMoverType"] == "combo":
##             cg_min = self.MD.ConjugateGradientMinimizer(1.e-10,100)
##             cg_min.SetLineMinimizer("NEWTON")
##             cg_min.SetMaxIterations(self.maxCGIterations)
##             cg_min.SetFiniteDifferenceParameter(1.e-5)
##             self.mover = self.MD.MinimizerAtomsMover(cg_min, self.potential)
##             if self.ioControl.GetVerbosity() > 5:
##                 cg_min.Verbose()
##                 cg_min.SetStepsBetweenUpdate(self.qmStepsBetweenUpdate)

##         self.verlet = self.MD.VerletAtomsMover(self.potential, self.Parameters["timestep"], self.maxMDIterations)
##         if self.Parameters["MDMoverType"] == "verlet":
##             self.mover = self.verlet

##         if self.Parameters["MDMoverType"] == "langevin":
##             self.mover = self.MD.LangevinAtomsMover(self.potential, self.Parameters["timestep"], self.maxMDIterations)
##             self.mover.SetFriction(self.friction)
##             self.mover.SetTemperature(self.temperature)

##         if self.Parameters["MDMoverType"] == "combo":
##             self.mover = CombinedQMCG(qm,cg)


        
    def RemoveExtraAtoms(self,elements,atoms):
        # searches elements in list, removes atoms that are not in any element

        # find normals for each tet
        self.ioControl.Print(3,'Finding normals for each element')
        elemNormals = {}
        for ID in elements:
           node_coords = self.model.GetElemCoords(ID,vec3d=0)[0:4]
           normals = []
           axes = []
           # find all normals and axes being careful about sign, using NumPy arrays!
           for i in xrange(4):
               a = array(node_coords[i])
               b = array(node_coords[(i+1)%4])
               c = array(node_coords[(i+2)%4])
               d = array(node_coords[(i+3)%4])
               axis = GetNormalizedVector(c-b)
               temp1 = a-b
               temp2 = d-b
               # tangent vectors
               t1 = GetNormalizedVector(temp1-dot(axis,temp1)*axis)
               t2 = GetNormalizedVector(temp2-dot(axis,temp2)*axis)
               n1 = CrossProduct(t1,axis)
               sign = dot(n1,t2)
               if sign < 0:
                   n1 = -n1
               normals.append((n1,a))
           elemNormals[ID] = normals

        removeList = []

        # loop over atoms to find if they are in an element
        self.ioControl.Print(1,'Looping over atom positions to identify elements')     
        for atomIdx in xrange(atoms.GetNumber()):
            position = atoms.GetCartesianPositions()[atomIdx]
            cartesian = array([position[0],position[1],position[2]])
            inRegion = 0
            for ID in elements:
                inElement = 1
                for i in xrange(4):
                    normal = elemNormals[ID][i][0]
                    point = elemNormals[ID][i][1]
                    if dot((cartesian - point),normal) < 0:
                        inElement = 0
                        break #break because we know atom is not in this element
                if inElement:
                    inRegion = 1
                    break  #break because we found an element that contains the atom

            if not inRegion:                
                removeList.append(atomIdx)
        

        self.ioControl.Print(3,"Removing " + repr(len(removeList)) + " out of " + repr(atoms.GetNumber()) + " atoms")
        atoms.RemoveAtoms(removeList)


            
class CombinedQMCG:
    def __init__(self,qm,cg):
        self.qm = qm
        self.cg = cg
    def Move(self,atoms):
        self.qm.Move(atoms)
        self.cg.Move(atoms)

def CrossProd(op1,op2):
    return Vec3D((op1.y()*op2.z()) - (op1.z()*op2.y()), 
                 (op1.z()*op2.x()) - (op1.x()*op2.z()), 
                 (op1.x()*op2.y()) - (op1.y()*op2.x()))
