#AtomsDeformer.py

import sys
sys.path.append("..")
from MDWebService import *
sys.path.append("/afs/msc.cornell.edu/home/jsethna/val/ASP/FemLib-0.8/")
# we want to use Vec3D here because we are passing stuff to FemLib methods
from Vec3D import Vec3D
import __builtin__


class AtomsDeformer(MDWebService):
    def __init__(self,atoms,model,elem_hash,Parameters,ioControl):
        self.DIMENSION = 3
        self.atoms = atoms
	self.model = model
	self.elem_hash = elem_hash
        self.ioControl = ioControl
        self.undeformedPositions = self.atoms.GetCartesianPositions()
        self.nAtoms = len(self.undeformedPositions)
        self.Parameters = Parameters
        MDWebService.__init__(self,self.DIMENSION,self.Parameters,self.ioControl)

    def FindNaturalCoordinatesWithinElements(self):
        TINY =1.e-10
        self.ioControl.Print(3,'Calculating bounding boxes of elements')
        boundingBoxes = {}
        for elementID in self.model.GetElemIds():
            node_coords = array(self.model.GetElemCoords(elementID))
            numNodes = len(node_coords)
            box = [[1.e99,-1.e99],[1.e99,-1.e99],[1.e99,-1.e99]]
            for nd in xrange(numNodes):
                for coord in xrange(3):
                    # check if smaller than current minimum
                    if node_coords[nd][coord] < box[coord][0]:
                        box[coord][0] = node_coords[nd][coord]
                    # check if greater than current maximum
                    if node_coords[nd][coord] > box[coord][1]:
                        box[coord][1] = node_coords[nd][coord]
                        
            boundingBoxes[elementID]=box

        self.naturalCoordList = []
        self.elementList = []
        
        # loop over atoms to find which elements they're in
        self.ioControl.Print(1,'Looping over atom positions to identify elements')     
        for atomIdx in xrange(self.nAtoms):
            position = self.undeformedPositions[atomIdx]
            cartesian = Vec3D(float(position[0]),float(position[1]),float(position[2]))
            for ID in self.model.GetElemIds():
                el_type = self.model.GetElemData(ID)[0]
                node_coords = array(self.model.GetElemCoords(ID))
                element = self.elem_hash[el_type]
                # see if in the bounding box
                box = boundingBoxes[ID]
                if (position[0] >= box[0][0]-TINY and position[0] <= box[0][1]+TINY) and (position[1] >= box[1][0]-TINY and position[1] <= box[1][1]+TINY) and (position[2] >= box[2][0]-TINY and position[2] <= box[2][1]+TINY):
                    # find out the coordinates within the element
                    naturalCoords = element.FindNaturalCoords(cartesian,node_coords)
                    #Often have trouble here when making changes elsewhere
                    if atomIdx == 0:
                        self.ioControl.Print(4,"cartesian:")
                        self.ioControl.Print(4,cartesian)
                        self.ioControl.Print(4,"node_ID's:")
                        self.ioControl.Print(4,self.model.GetElemData(ID)[2])
                    if __builtin__.type(naturalCoords) != __builtin__.type(0):
                        #exit loop because coordinate found
                        break
                else:
                    naturalCoords = 0

            if __builtin__.type(naturalCoords) == __builtin__.type(0):
                self.ioControl.Print(0,["Natural coords not found for atom:",atomIdx])
                self.ioControl.Print(0,["position",'%16.16f\t%16.16f\t%16.16f\n ' % (position[0],position[1],position[2])])
            assert(__builtin__.type(naturalCoords) != __builtin__.type(0))
            self.elementList.append(ID)

            self.naturalCoordList.append(naturalCoords)
            
    def GetNodeDisplacementsFromRepository(self):
        self.ioControl.Print(3,"Getting Node Displacements From XML Repository")

        self.node_disps = {}
        scalar = self.Parameters["lengthScale"]*self.Parameters["displacementScale"]
        for nodeID in self.model.CornerNodeIds:

            xqueryText = """
            default namespace="http://www.tc.cornell.edu/~heber/2004/04/Displacement"
            <Displacements xmlns="http://www.tc.cornell.edu/~heber/Displacements.xsd">
            { for $d in /Displacement/d
            where $d/attribute::NodeID = """ + repr(nodeID) + """
            return <d NodeID="{data($d/attribute::NodeID)}"
            dx="{data($d/attribute::dx)}"
            dy="{data($d/attribute::dy)}"
            dz="{data($d/attribute::dz)}" />}
            </Displacements>
            """

            self.model.ConstructSoapEnvelope(self.Parameters["dispsUUID"],xqueryText)
            walker = self.model.CallWebService()

            while  1:
                node = walker.currentNode
                # print node.tagName
                if node.tagName == "soap:Header":
                    next = walker.nextSibling()
                    node = walker.currentNode
                elif node.tagName == "detail":
                    break
                elif node.tagName == "faultstring":
                    print "ERROR"
                elif node.tagName == "d":
                    atts = node.attributes
                    for k in node.attributes.keys():
                        if atts[k].localName == "NodeID":
                            id = int(atts[k].value)
                        elif atts[k].localName == "dx":
                            dx = float(atts[k].value)
                        elif atts[k].localName == "dy":
                            dy = float(atts[k].value)
                        elif atts[k].localName == "dz":
                            dz = float(atts[k].value)
                    self.node_disps[id] = scalar*Vec3D(dx,dy,dz)
                next = walker.nextNode()
                if next is None:
                    next = walker.previousNode()
                    next = walker.nextSibling()
                    next = walker.nextSibling()
                    if next is None: break


        # Get displacements for edge nodes
        # Currently edge node displacements are stored in the
        # repository according to the edge id's.  However, there is no
        # data for mapping these id's to the appropriate tets and
        # coordinates!  So we are making up our own id's (in
        # XMLFemModel.py) and averaging the displacements here.

        for pair in self.model.edgeNodeMap.keys():
            averagedDisp = 0.5*(self.node_disps[pair[0]]+self.node_disps[pair[1]])
            edgeNodeId = self.model.edgeNodeMap[pair]
            self.node_disps[edgeNodeId] = averagedDisp
            

    
    def DeformAtoms(self,leavesList):

        for leaf in leavesList:
            atomsLeaf = self.atoms.GetLeaf(leaf)
            nAtomsInLeaf = atomsLeaf.GetNumber()
            displacements = self.MD.ArrayDoubleVector(atomsLeaf.GetNumber())
            displacements.Zero()
            baseIdx = 0
            for l in xrange(leaf):
                baseIdx += self.atoms.GetLeaf(l).GetNumber()

            endIdx = baseIdx + atomsLeaf.GetNumber()
            pyListDisps = self.GetDisplacements(baseIdx,endIdx)

            displacements[:] = pyListDisps

            atomsLeaf.SetCartesianPositions(self.undeformedPositions.GetLeaf(leaf))
            atomsLeaf.IncrementCartesianPositions(displacements)


    def GetDisplacements(self, startIdx, endIdx):
        pyListDisps = zeros((endIdx-startIdx,3),Float64)
        for atomIdx in xrange( startIdx, endIdx):
            # get the element
            elementIdx = self.elementList[atomIdx]
            el_type = self.model.GetElemData(elementIdx)[0]
            element = self.elem_hash[el_type]
            numShapeFuncs = element.NumVolumeShapeFunc()
            # get the natural coords of this atom in that element
            natural = self.naturalCoordList[atomIdx]
            naturalCoord = Vec3D(natural[0], natural[1], natural[2])
            # get the node numbers for that element
            node_nums = self.model.GetElemData(elementIdx)[2]
            # find atomic displacement
            node_values = []
            for t in xrange(numShapeFuncs):
                node_values.append(self.node_disps[node_nums[t]])                    
            displacement = element.InterpolateVolume(naturalCoord, node_values)
            pyListDisps[atomIdx-startIdx] = array([displacement[0],displacement[1],displacement[2]])
        return pyListDisps
