#AtomsFiller.py

import sys
sys.path.append("..")
from MDWebService import *
from AtomsDecorator import *


class CylinderDecorator(AtomsDecorator):

    def __init__(self,Parameters,ioControl,MD,model):
        AtomsDecorator.__init__(self,Parameters,ioControl,MD,model)
        self.numConstrainedLeaves = 3*len(self.model.GetMaterialIds()) # needed by base class!

        
    def MakeLeaves(self):

        # declare spherical cluster initializer
        cylinder = self.MD.CylindricalClusterInitializer(self.Parameters["radius"], self.Parameters["length"], self.Parameters["cylinderAxis"], self.lattice)
        if self.Parameters["parallel"]:
            paracylinder = self.MD.ParallelAtomsInitializer(cylinder)
            initializer = paracylinder
        else:
            initializer = cylinder
            
        # for each material (or grain)
        LOAId = 0
        for matID in self.model.GetMaterialIds():

            # get rotation
            rotation = array(self.model.GetMaterialData(matID))


            # get list of elems with current material id
            elems = []
            for elemID in self.model.GetElemIds():
                if self.model.GetElemData(elemID)[1] == matID:
                    elems.append(elemID)

            self.ioControl.Print(4,"Rotation = " + repr(rotation))

            # generate spherical cluster of atoms
            cylinder.SetCenter(self.Parameters["center"])
            self.lattice.ResetOrientation()
            self.lattice.RotateLatticeVectors(rotation)
            paracylinder.Create(self.BulkLOAs[LOAId])

            # remove atoms outside material region
            self.ioControl.Print(3,"Removing extra atoms...")
            self.RemoveExtraAtoms(elems,self.BulkLOAs[LOAId])

            # increment ListOfAtoms
            LOAId += 1
            

    def MakeConstrainedLeaves(self):
        
        self.constrZoneMap = []

        # declare spherical cluster initializers
        constrZoneWidth = self.potential.GetCutoffDistance()
        cylinder1 = self.MD.CylindricalClusterInitializer(self.Parameters["radius"], self.Parameters["length"] + 2*constrZoneWidth, self.Parameters["cylinderAxis"], self.lattice)
        cylinder2 = self.MD.CylindricalClusterInitializer(self.Parameters["radius"] + constrZoneWidth, self.Parameters["length"] + 2*constrZoneWidth, self.Parameters["cylinderAxis"], self.lattice)
        cylinder1.SetCenter(self.Parameters["center"])
        cylinder2.SetCenter(self.Parameters["center"])
        shell = self.MD.CompositeClusterInitializer(cylinder2,cylinder1,self.lattice,"difference")

        endcap1 = self.MD.CylindricalClusterInitializer(self.Parameters["radius"], constrZoneWidth,self.Parameters["cylinderAxis"], self.lattice)
        endcap2 = self.MD.CylindricalClusterInitializer(self.Parameters["radius"], constrZoneWidth,self.Parameters["cylinderAxis"], self.lattice)
        if self.Parameters["parallel"]:
            parashell = self.MD.ParallelAtomsInitializer(shell)
            paraendcap1 = self.MD.ParallelAtomsInitializer(endcap1)
            paraendcap2 = self.MD.ParallelAtomsInitializer(endcap2)
            initializer1 = parashell
            initializer2 = paraendcap1
            initializer3 = paraendcap2
        else:
            initializer1 = shell
            initializer2 = endcap1
            initializer3 = endcap2
                
        center1 = array(self.Parameters["center"]) + 0.5*(self.Parameters["length"]+constrZoneWidth)*array(self.Parameters["cylinderAxis"])
        center2 = array(self.Parameters["center"]) - 0.5*(self.Parameters["length"]+constrZoneWidth)*array(self.Parameters["cylinderAxis"])
        endcap1.SetCenter(center1)
        endcap2.SetCenter(center2)
        

        # for each material (or grain)
        LOAId = 0
        for matID in self.model.GetMaterialIds():

            # get rotation
            rotation = array(self.model.GetMaterialData(matID))

            # get list of elems with current material id
            elems = []
            for elemID in self.model.GetElemIds():
                if self.model.GetElemData(elemID)[1] == matID:
                    elems.append(elemID)

            self.ioControl.Print(4,"Rotation = " + repr(rotation))

            # generate spherical clusters of atoms
            self.lattice.ResetOrientation()
            self.lattice.RotateLatticeVectors(rotation)
            initializer1.Create(self.ConstrLOAs[LOAId])
            initializer2.Create(self.ConstrLOAs[LOAId+1])
            initializer3.Create(self.ConstrLOAs[LOAId+2])

            # remove atoms outside material region
            self.ioControl.Print(3,"Removing extra atoms...")
            self.RemoveExtraAtoms(elems,self.ConstrLOAs[LOAId])
            self.RemoveExtraAtoms(elems,self.ConstrLOAs[LOAId+1])
            self.RemoveExtraAtoms(elems,self.ConstrLOAs[LOAId+2])

            # increment ListOfAtoms
            LOAId += 3





    def ImplementConstraints(self):

        # implement constraints
        self.ioControl.Print(3,"Implementing Constraints...")
        self.constraints = []
        for LOAId in xrange(len(self.ConstrLOAs)):
            if LOAId % 3 == 0:
                self.constraints.append(self.MD.FixedBody(self.ConstrLOAs[LOAId].GetCartesianPositions()))
            else:
                self.constraints.append(self.MD.FixedInPlane(self.ConstrLOAs[LOAId].GetCartesianPositions(),self.Parameters["cylinderAxis"]))
            self.ConstrLOAs[LOAId].SetConstraint(self.constraints[LOAId])
 
                    
