#AtomsFiller.py

import sys
sys.path.append("..")
from MDWebService import *
from AtomsDecorator import *


class FullDecorator(AtomsDecorator):

    def __init__(self,Parameters,ioControl,MD,model):
        AtomsDecorator.__init__(self,Parameters,ioControl,MD,model)
        self.numConstrainedLeaves = 0 # needed by base class!

        
    def MakeLeaves(self):

        # declare rectangular cluster initializer
        size = array([1,1,1])
        rectangle = self.MD.RectangularClusterInitializer(size, self.lattice)
        if self.Parameters["parallel"]:
            pararect = self.MD.ParallelAtomsInitializer(rectangle)
            initializer = pararect
        else:
            initializer = rectangle
            
        # for each material (or grain)
        LOAId = 0
        for matID in self.model.GetMaterialIds():

            # get rotation
            rotation = array(self.model.GetMaterialData(matID))


            # get list of elems with current material id
            elems = []
            for elemID in self.model.GetElemIds():
                if self.model.GetElemData(elemID)[1] == matID:
                    elems.append(elemID)

            # find box containing material region
            min = array([1.0e20,1.0e20,1.0e20])
            max = array([-1.0e20,-1.0e20,-1.0e20])
            for elemID in elems:
                for vertId in self.model.GetElemData(elemID)[2]:
                    for coord in xrange(self.DIMENSION):
                        temp = self.model.GetNodeCoord(vertId,0)[coord]
                        if temp > max[coord]:
                            max[coord] = temp
                        if temp < min[coord]:
                            min[coord] = temp
            size = max-min
            center = (max+min)/2
            self.ioControl.Print(4,"Material ID = " + repr(matID))
            self.ioControl.Print(4,"Bounding box = " + repr(max) + repr(min))
            self.ioControl.Print(4,"Rotation = " + repr(rotation))

            # generate rectangular cluster of atoms
            self.CreateCrystal(rectangle, initializer, self.lattice, rotation, size, center, self.BulkLOAs[LOAId])

            # remove atoms outside material region
            self.ioControl.Print(3,"Removing extra atoms...")
            self.RemoveExtraAtoms(elems,self.BulkLOAs[LOAId])

            # increment ListOfAtoms
            LOAId += 1
            



    def MakeConstrainedLeaves(self):
        # does not initialize new clusters, searches for atoms near fixed faces

        # find all triangle norms
        self.ioControl.Print(3,'Finding normals for each surface triangle')
        triangleNormals = {}
        for faceID in self.model.GetFaceIds():
            triangleNormals[faceID] = []
            for j in xrange(len(self.model.GetFaceTriangles(faceID))):
                normals = []
                points = []
                points.append(array(self.model.GetNodeCoord(self.model.GetFaceTriangles(faceID)[j][0],0)))
                points.append(array(self.model.GetNodeCoord(self.model.GetFaceTriangles(faceID)[j][1],0)))
                points.append(array(self.model.GetNodeCoord(self.model.GetFaceTriangles(faceID)[j][2],0)))
                surfaceNormal = GetNormalizedVector(CrossProduct(points[0]-points[1],points[2]-points[1]))
                normals.append(surfaceNormal)

                # find all normals being careful about sign
                for i in xrange(3):
                    t1 = GetNormalizedVector(points[(i+1)%3]-points[i])
                    t2 = GetNormalizedVector(points[(i+2)%3]-points[i])
                    n1 = CrossProduct(t1,surfaceNormal)
                    sign = dot(n1,t2)
                    if sign < 0:
                        n1 = -n1
                    normals.append((n1,points[i]))
                triangleNormals[faceID].append(normals)        


        # make constrZoneMap
        self.constrZoneMap = []
        
        # a ParallelAtomsInitializer is neccessary to make the
        # lists of atoms on different processor compatible with one another
        if self.Parameters["parallel"]:
            empty = self.MD.EmptyInitializer()
            paraEmpty = self.MD.ParallelAtomsInitializer(empty)


        # search for list of faces each atom is within cutoff distance of
        # and make array of atoms to put in appropriate constrained zone
        self.constrZoneWidth = self.potential.GetCutoffDistance()
        self.ioControl.Print(3,"Constrained zone width = " + repr(self.constrZoneWidth))
        removeList = []
        self.ioControl.Print(3,"Finding atoms in constrained zones...")
        # loop over atoms
        for matID in xrange(len(self.model.GetMaterialIds())):
            tempConstrZoneMap = []
            removeList = []
            tempConstrLOAs = []
            constrAtomMap = {}

            for atomIdx in xrange(self.bulk.GetBranch(matID).GetNumber()):
                adjacentFaces = []
                position = self.bulk.GetBranch(matID).GetCartesianPositions()[atomIdx]
                cartesian = array([position[0],position[1],position[2]])
                adjacentFaces = []
                # loop over faces
                for faceID in self.model.GetFaceIds():
                    # loop over triangles on that face
                    for j in xrange(len(self.model.GetFaceTriangles(faceID))):
                        surfaceNormal = triangleNormals[faceID][j][0]
                        surfacePoint = array(self.model.GetNodeCoord(self.model.GetFaceTriangles(faceID)[j][0],0))
                        perpDist = abs(dot(surfaceNormal,(surfacePoint-cartesian)))

                        if perpDist < self.constrZoneWidth:
                            # test whether this is the closest triangle
                            inTriangle = 1
                            for k in xrange(3):
                                normal = triangleNormals[faceID][j][k+1][0]
                                point = triangleNormals[faceID][j][k+1][1]
                                if dot((cartesian - point),normal) < 0:
                                    inTriangle = 0
                            if inTriangle:
                                adjacentFaces.append(faceID)
                                break # break out of triangle loop since we found triangle

                # find which list of atoms this atom belongs in and save position
                # in constrAtomMap
                if len(adjacentFaces) > 0:
                    if adjacentFaces not in tempConstrZoneMap:
                        tempConstrZoneMap.append(adjacentFaces)
                        LOAId = len(tempConstrZoneMap) - 1
                        constrAtomMap[LOAId] = [cartesian]
                    else:
                        for k in xrange(len(tempConstrZoneMap)):
                            if adjacentFaces == tempConstrZoneMap[k]: LOAId = k
                        constrAtomMap[LOAId].append(cartesian)
                    removeList.append(atomIdx)
                    self.ioControl.Print(4,"Moving Atom: " + repr(atomIdx) + " to constrained zone for face(s): " + repr(adjacentFaces))


            # create lists of atoms for each new constrained zone and set positions
            # being careful for this to work in parallel
            numZones = array(len(constrAtomMap))
            if self.Parameters["parallel"]:
                Scientific.MPI.world.barrier()
                Scientific.MPI.world.broadcast(numZones,0)
            for LOAId in xrange(numZones[0]):
                numConstrAtoms = 0
                if self.Parameters["myProc"] == 0:
                    numConstrAtoms = len(constrAtomMap[LOAId])
                tempConstrLOAs.append(self.MD.DynamicListOfAtoms(numConstrAtoms))
                # make lists of atoms on different processors compatible
                if self.Parameters["parallel"]: paraEmpty.Create(tempConstrLOAs[LOAId])
                if self.Parameters["myProc"] == 0:
                    for atomIdx in xrange(len(constrAtomMap[LOAId])):
                        for coord in xrange(3):
                            tempConstrLOAs[LOAId].SetCartesianPosition(atomIdx,coord,constrAtomMap[LOAId][atomIdx][coord])


            # append tempConstrZoneMap and LOAs to self.constrZoneMap and LOAs
            # be careful so that constrZoneMap is the same on all processors
            numFaces = array(0)
            for faceID in xrange(numZones[0]):
                if self.Parameters["myProc"] == 0: numFaces = array(len(tempConstrZoneMap[faceID]))
                if self.Parameters["parallel"]:
                    Scientific.MPI.world.barrier()
                    Scientific.MPI.world.broadcast(numFaces,0)
                if self.Parameters["myProc"] == 0:
                    faces = array(tempConstrZoneMap[faceID])
                else:
                    faces = zeros(numFaces)
                if self.Parameters["parallel"]: Scientific.MPI.world.broadcast(faces,0)
                self.constrZoneMap.append(faces)


            self.ConstrLOAs.extend(tempConstrLOAs)

            # remove atoms from bulkLOA
            self.ioControl.Print(3,"Moving " + repr(len(removeList)) + " out of " + repr(self.bulk.GetBranch(matID).GetNumber()) + " atoms to constrained Zones")
            self.bulk.GetBranch(matID).RemoveAtoms(removeList)

                        
        self.ioControl.Print(4,"ConstrZoneMap: " + repr(self.constrZoneMap))
        self.ioControl.Print(3,"length of constrZoneMap: " + repr(len(self.constrZoneMap)))
        self.ioControl.Print(3,"length of ConstrLOAs: " + repr(len(self.ConstrLOAs)))



    def ImplementConstraints(self):
        
        # implement constraints
        self.ioControl.Print(3,"Implementing Constraints...")
        self.constraints = []
        for LOAId in xrange(len(self.constrZoneMap)):
            directions = [0,0,0]
            # loop over faces adjacent to that constrained zone
            for faceID in self.constrZoneMap[LOAId]:
                # get constr directions
                dim = self.model.GetDispBCs(faceID)[0]
                if dim == 'x': directions[0] = 1
                if dim == 'y': directions[1] = 1
                if dim == 'z': directions[2] = 1
            self.constraints.append(self.MD.PartialFixedBody(self.ConstrLOAs[LOAId].GetCartesianPositions(),directions))
            self.ConstrLOAs[LOAId].SetConstraint(self.constraints[LOAId])
 
                    
