# GrainBoundaryGeometryFinder.py.
import string
import sys
sys.path.append("..")
from MDWebService import *


class GrainBoundaryGeometryFinder:
    def __init__(self, Parameters, ioControl):
        self.ioControl = ioControl
        self.Parameters = Parameters
            
    def MakeModel(self):

        self.model = XMLFemModel(self.Parameters["soapAction"],self.Parameters["webService"],self.Parameters["meshUUID"],self.Parameters["attributeUUID"],self.Parameters["lengthScale"],self.Parameters["randomNumberSeed"],self.Parameters["myProc"],self.Parameters["parallel"],self.Parameters["topologyUUID"])

        self.model.GetInternalFaces()

        return self.model
    
    def FindGrainBoundaryGeometries(self):
        negX = array((-1,0,0))

        vectorsAndRegions={}

        for faceID in self.model.internalFaces.keys():
            # find vector which defines the plane of the face
            vertices = []
            for j in xrange(3):
                vertexID = self.model.internalFaces[faceID][3][j]
                vertexCoord = self.model.topologicalVertices[vertexID]
                vertices.append(array(vertexCoord))
            # print vertices

            vector1 = vertices[1]-vertices[0]
            vector2 = vertices[2]-vertices[0]
            planeVector = GetNormalizedVector(CrossProduct(vector1,vector2))
            # if the loop number is negative, 
            if self.model.internalFaces[faceID][0] < 0:
                planeVector = -1.0 * planeVector
            thisTuple = (planeVector,self.model.internalFaces[faceID][1],self.model.internalFaces[faceID][2],vector1)

            duplicateBoundary = 0
            for id in vectorsAndRegions.keys():
                # the annoying thing about numpy arrays is that we cannot compare them directly!
                if (vectorsAndRegions[id][0][0],vectorsAndRegions[id][0][1],vectorsAndRegions[id][0][2],vectorsAndRegions[id][1],vectorsAndRegions[id][2]) == (planeVector[0],planeVector[1],planeVector[2],self.model.internalFaces[faceID][1],self.model.internalFaces[faceID][2]):
                    duplicateBoundary = 1
                    break
            if not duplicateBoundary:
                vectorsAndRegions[faceID] = thisTuple

        faceIDstring = ""
        for faceID in vectorsAndRegions.keys():
            faceIDstring += repr(faceID) + " "

            # find matrix which rotates planeVector to negative x axis
            row1 = vectorsAndRegions[faceID][0]
            row2 = GetNormalizedVector(vectorsAndRegions[faceID][3])
            row3 = CrossProduct(row1,row2)
            rotationMatrix = [row1,row2,row3]

            # region in (grain on the negative x / left side)
            region_in = vectorsAndRegions[faceID][1]
            latticeOrientation1 = matrixmultiply(rotationMatrix,self.model.materials[region_in])

            # region out (grain on positive x / right side)
            region_out = vectorsAndRegions[faceID][2]
            latticeOrientation2 = matrixmultiply(rotationMatrix,self.model.materials[region_out])

            # write it all to files
            filename = "face"
            if faceID != 0:
                temp = int(log10(faceID))
            else:
                temp = 0
            for i in xrange(int(log10(max(vectorsAndRegions.keys())))-temp):
                filename += "0"
            filename += repr(faceID) + ".geo"

            geoString1 = ""
            geoString2 = ""
            self.ioControl.OpenFile(filename)
            for i in xrange(3):
                for j in xrange(3):
                    geoString1 += repr(latticeOrientation1[i][j]) + " "
                    geoString2 += repr(latticeOrientation2[i][j]) + " "


            self.ioControl.Print(3, repr(faceID)+" "+repr(vectorsAndRegions[faceID][1])+" "+repr(vectorsAndRegions[faceID][2])+" "+repr(vectorsAndRegions[faceID][0])) 
            self.ioControl.WriteToFile(filename,geoString1)
            self.ioControl.WriteToFile(filename,"\n")
            self.ioControl.WriteToFile(filename,geoString2)
            self.ioControl.CloseFile(filename)
        self.ioControl.OpenFile("faceIDs.list")
        self.ioControl.WriteToFile("faceIDs.list",faceIDstring)
        self.ioControl.CloseFile("faceIDs.list")

                


