# DatabaseMesher.py.
import string
import sys
sys.path.append("..")
from MDWebService import *


class Mesher:
    def __init__(self, Parameters, ioControl):
        self.ioControl = ioControl
        self.Parameters = Parameters
            
    def MakeModel(self):

        self.model = XMLFemModel(self.Parameters["soapAction"],self.Parameters["webService"],self.Parameters["meshUUID"],self.Parameters["attributeUUID"],self.Parameters["lengthScale"],self.Parameters["randomNumberSeed"],self.Parameters["myProc"],self.Parameters["parallel"])
        self.ioControl.Print(5,self.model)
        self.ResetGeometryParameters()

        if self.Parameters["atomsGeometry"] == "Full":
            self.model.GetNodesAndElements()
            self.model.GetKinematicBCs(self.Parameters["caseID"])
            self.model.GetFaceTrianglesFromRepository()

        elif self.Parameters["subMesh"]=="cube":
            scale = 1.0
            self.SetMeshLimits(scale)
            self.model.GetNodesAndElements(meshPortion="cube",meshMax=self.meshMax,meshMin=self.meshMin)
            while self.model.NumElems() == 0:
                print "SCALING LIMITS"
                scale = scale*2.0
                self.SetMeshLimits(scale)
                self.model.GetNodesAndElements(meshPortion="cube",meshMax=self.meshMax,meshMin=self.meshMin)

        elif self.Parameters["subMesh"]=="connectivity":
            self.model.GetNodesAndElements(meshPortion="connectivity",subMeshNodes=self.subMeshNodes)

        if self.Parameters["deformationMethod"] == "meshDisp":
            self.model.Make10Noded()

            

        return self.model
    

    def MakeElementHash(self):
        # for each type of element that appears in the mesh, we create
        # one Element object which handles element operations (interpolations,
        # integrations etc)
        self.ioControl.Print(3,'creating elem_hash')

        elem_hash = {}
        for id in self.model.GetElemIds():
            el_type,mat,conn = self.model.GetElemData(id)

            # get the formulator object for this element type
            ef = elem_hash.get(el_type)
            if ef == None:
                ef = FemLib.Element(el_type+"ShapeFunc")
                elem_hash[el_type] = ef
            # end loop for elem_hash
        return elem_hash


    def ResetGeometryParameters(self):
        # This function overrides the user set geometry parameters based on
        # which feature of the mesh is to be decorated

        resetParams = []

        if self.Parameters["decoratedFeature"] == "vertex":
            self.Parameters["center"] = array(self.model.GetNodeCoord(self.Parameters["vertexToDecorate"],0))
            self.subMeshNodes=[self.Parameters["vertexToDecorate"]]
            resetParams.append("center")

        elif self.Parameters["decoratedFeature"] == "edge":
            leftPos = array(self.model.GetNodeCoord(self.Parameters["edgeToDecorate"][0],0))
            rightPos = array(self.model.GetNodeCoord(self.Parameters["edgeToDecorate"][1],0))
            self.Parameters["center"] = 0.5 * (rightPos + leftPos)
            resetParams.append("center")
            if self.Parameters["atomsGeometry"] == "Cylinder":
                self.Parameters["cylinderAxis"] = GetNormalizedVector(rightPos - leftPos)
                resetParams.append("cylinderAxis")
            elif self.Parameters["atomsGeometry"] == "Rectangle": #XXX NOT TESTED
                axis1 = GetNormalizedVector(rightPos - leftPos)
                temp = self.Parameters["rectangleAxes"][1]
                axis2 = GetNormalizedVector(temp - dot(axis1,temp)*axis1)
                if axis2 == array([0,0,0]):  
                    temp = self.Parameters["rectangleAxes"][2]
                    axis2 = GetNormalizedVector(temp - dot(axis1,temp)*axis1)
                axis3 = CrossProduct(axis1,axis2)
                self.Parameters["rectangleAxes"] = [axis1,axis2,axis3]
                resetParams.append("rectangleAxes")
            self.subMeshNodes = self.Parameters["edgeToDecorate"]

        elif self.Parameters["decoratedFeature"] == "face":
            point1 = array(self.model.GetNodeCoord(self.Parameters["faceToDecorate"][0],0))
            point2 = array(self.model.GetNodeCoord(self.Parameters["faceToDecorate"][1],0))
            point3 = array(self.model.GetNodeCoord(self.Parameters["faceToDecorate"][2],0))
            self.Parameters["center"] = (1.0/3.0) * (point1 + point2 + point3)
            resetParams.append("center")
            side1 = point2 - point1
            side2 = point3 - point1
            normal = GetNormalizedVector(CrossProduct(side1,side2))
            if self.Parameters["atomsGeometry"] == "Cylinder":
                self.Parameters["cylinderAxis"] = normal
                resetParams.append("cylinderAxis")
            elif self.Parameters["atomsGeometry"] == "Rectangle":
                axis3 = GetNormalizedVector(CrossProduct(normal,side1))
                self.Parameters["rectangleAxes"] = [normal,GetNormalizedVector(side1),axis3]
                resetParams.append("rectangleAxes")
            self.subMeshNodes = self.Parameters["faceToDecorate"]

        for param in resetParams:
            self.ioControl.Print(3,param + " reset to " + repr(self.Parameters[param]))


    def SetMeshLimits(self,scale):

        cutoffEstimate = 6.0
        scale = scale/self.Parameters["lengthScale"]

        if self.Parameters["atomsGeometry"] == "Sphere":
            temp = scale*(self.Parameters["radius"]+cutoffEstimate)
        elif self.Parameters["atomsGeometry"] == "Cylinder":
            temp = scale*(sqrt(self.Parameters["radius"]*self.Parameters["radius"]+0.25*self.Parameters["length"]*self.Parameters["length"])+cutoffEstimate)
        elif self.Parameters["atomsGeometry"] == "Rectangle":
            tempWidth = self.Parameters["width"]+cutoffEstimate
            tempLength = self.Parameters["length"]+cutoffEstimate
            tempHeight = self.Parameters["height"]+cutoffEstimate
            temp = scale*sqrt(tempWidth*tempWidth+tempLength*tempLength+tempHeight*tempHeight)

        tempArray = array([temp,temp,temp])
        self.meshMax = array(self.Parameters["center"])/self.Parameters["lengthScale"] + tempArray
        self.meshMin = array(self.Parameters["center"])/self.Parameters["lengthScale"] - tempArray



