
import sys
sys.path.append("/afs/msc.cornell.edu/home/jsethna/val/ASP/MDWebServices")
from MDWebService import *
import Mesher
import FullDecorator
import SphereDecorator
import CylinderDecorator
import RectangleDecorator
import AtomsDeformer
import GrainBoundaryGeometryFinder


# This code needs to be cleaned up and retested for all test cases!!!
# things to do: resurrect and test database mesh retrieval, database
# retrieval in parallel, more efficient retrieval in parallel for both
# cases, separate user parameters from internal parameters

class OFEMD(MDWebService):

    def __init__(self):

        self.DIMENSION = 3

        #Default Parameter settings
        self.Parameters = {
            #atoms geometry parameters
            "atomsGeometry":"Sphere", #choices: "Full","Rectangle","Cylinder","Sphere"
            "decoratedFeature":"vertex", #choices: "entire","point","edge","face","vertex"
            "edgeToDecorate":[0,1],
            "faceToDecorate":[0,1,2],
            "vertexToDecorate":0,
            "center":array([0.0,0.0,0.0]),
            "radius":5.0,
            "cylinderAxis":array([0.0,0.0,1.0]),
            "length":10.0,
            "width":10.0,
            "height":10.0,
            "rectangleAxes":[array([1.0,0.0,0.0]),array([0.0,1.0,0.0]),array([0.0,0.0,1.0])],

            #how to deform / what to do
            "deformationMethod":"meshDisp", #choices: "kinematicBC", "meshDisp", "cohesiveLaw"
            "findGrainBoundaryGeometries" : 0,
            "strainIncrement":0.01,            

            #mesh/geometry data access parameters
            "meshSource":"xml", # "database"
            "soapAction": "http://www.tc.cornell.edu/~heber",
            "webService":"http://cmiws.tc.cornell.edu/XMLRepository/default.asmx",
	    # new web service is http://cmiws.tc.cornell.edu/XMLRepository/default.asmx
            "meshUUID": "fd4aafc5-b7c0-4f79-a7c5-70a648e9b649", #"d5185a2e-9e54-4d3b-841f-73721cf43ddd", #"1c681426-6ec3-4d6d-af4c-b488485d7199",
            "dispsUUID": "797ce9da-01af-4966-ac0a-f5cdecc2f420",
            "attributeUUID": "6bb65170-bb99-4402-a0f3-0a44280f9368", #"44750e44-4dee-422a-adbb-d89ef8aacf54",
            "topologyUUID" : "6bc35972-fe4f-40ed-a154-e07e26c0784d",
            "caseID":4711,
            "databaseServer": "na",
            "database": "na",
            "databaseUserName": "na",
            "databasePassword": "na",

            #scale parameters
            "lengthScale":1000000.0,
            "displacementScale":1.0,
            "subMesh":"connectivity",#"cube", "connectivity"

            #mover parameters
            "nSteps":2,
            "QMtolerance":0.01,
            "maxQMIterations":100,
            "timestep":0.001,

            #md parameters
            "material":"SW_evA",
            "brittleFactor":2.0,
            "bc":"constrained", #choices: "rollered", "constrained"
            "randomNumberSeed": 5,
            "moverType":"qm",

            #system parameters
            "checkpointing":0,
            "verbose":3,
            "verboseMover": 0,
            "directory":".",
            "parallel":0
            }

        self.ioControl = IOControl(self.Parameters["verbose"],".")
        MDWebService.__init__(self,self.DIMENSION,self.Parameters,self.ioControl)


    def ParameterCheck(self):

        # Check that all parameters are consistent.  Some parameter choices won't work.

        goodParameters=1
        params1 = []
        params2 = []

        if self.Parameters["atomsGeometry"] == "Full" and self.Parameters["decoratedFeature"] != "entire":
            goodParameters=0
            params1.append("atomsGeometry")
            params2.append("decoratedFeature")

        if self.Parameters["decoratedFeature"] == "entire" and self.Parameters["atomsGeometry"] != "Full":
            goodParameters=0
            params1.append("atomsGeometry")
            params2.append("decoratedFeature")

        if self.Parameters["bc"] == "rollered" and not (self.Parameters["atomsGeometry"] == "Full" or self.Parameters["atomsGeometry"] == "Rectangle"):
            goodParameters=0
            params1.append("bc")
            params2.append("atomsGeometry")

        if self.Parameters["deformationMethod"] == "kinematicBC" and self.Parameters["decoratedFeature"] != "entire":
            goodParamters=0
            params1.append("deformationMethod")
            params2.append("decoratedFeature")

        if not goodParameters:
            for i in xrange(len(params1)):
                print "Parameters:", params1[i], "=", self.Parameters[params1[i]], "and", params2[i], "=", self.Parameters[params2[i]], "not consistent"
        assert(goodParameters)



    def GetMesh(self):
        self.ioControl.Print(3,'creating FE model')
        self.mesher = Mesher.Mesher(self.Parameters, self.ioControl)
        self.model = self.mesher.MakeModel()
        self.ioControl.Print(3,"Number of vertices: " + repr(self.model.NumNodes()))
        self.ioControl.Print(3,"Number of elements: " + repr(self.model.NumElems()))
        self.ioControl.Print(3,"Number of materials: " + repr(self.model.NumMats()))
        #self.ioControl.Print(3,"Number of faces: " + repr(len(self.model.GetFaceIds())))
        #self.ioControl.Print(3,"Number of face constraints: " + repr(self.model.NumBCs()))
        self.elem_hash = self.mesher.MakeElementHash()


    def DecorateWithAtoms(self):
        self.ioControl.Print(3,"Decorating With Atoms")
        # Make and execute atomsDecorator command based on which geometry we are using.
        # There is a separate class for each type of geometry
        exec("self.atomsDecorator = " + self.Parameters["atomsGeometry"] + "Decorator." + self.Parameters["atomsGeometry"] + "Decorator(self.Parameters,self.ioControl,self.MD,self.model)")
        self.atomsDecorator.Decorate()
        # We need to make pointers to all the MD Objects, otherwise Python garbage collection
        # will delete them
        # actuall we don't since atomsDecorator is a member of the class
        # but certain things like observers expect to see self.atoms and self.qm
        self.potential = self.atomsDecorator.potential
        self.moverpotential = self.atomsDecorator.moverpotential
        self.atoms = self.atomsDecorator.atoms
        #self.bulk = self.atomsDecorator.bulk
        self.constrZones = self.atomsDecorator.constrZones
        #self.neighborLocator = self.atomsDecorator.neighborLocator
        #self.boundaryConditions = self.atomsDecorator.boundaryConditions
        self.qm = self.atomsDecorator.qm
        if self.Parameters["verboseMover"]:
            self.qm.SetStepsBetweenUpdate(1)
            self.qm.Verbose()
        self.constrZoneMap = self.atomsDecorator.constrZoneMap
        if self.Parameters["deformationMethod"] == "cohesiveLaw":
            self.endcap1 = self.atomsDecorator.endcap1
            self.endcap2 = self.atomsDecorator.endcap2

        self.ioControl.Print(3,"Done Decorating With Atoms")
        for proc in xrange(self.Parameters["nProcs"]):
            nAtoms = self.atoms.GetNumber()
            nLeaves = self.atoms.NumberOfLeaves()
            self.ioControl.Print(3,"Processor "+repr(proc)+" has "+repr(nAtoms)+" atoms and "+repr(nLeaves)+" leaves", proc)
            if self.Parameters["parallel"]: self.mpiworld.barrier()

        
        self.Notify()


    def IncrementBoundaryDisplacements(self):
        self.ioControl.Print(3,"Incrementing Boundary Displacements...")
        dispFactor = self.Parameters["lengthScale"]*self.Parameters["displacementScale"]/self.Parameters["nSteps"]
        # moves all constrained zones with non-zero displacement bc
        for LOAId in xrange(len(self.constrZoneMap)):
            numAtoms = self.constrZones.GetLeaf(LOAId).GetNumber()
            positions = self.constrZones.GetLeaf(LOAId).GetCartesianPositions()
            increment = [0.0,0.0,0.0]
            # loop over faces adjacent to that constrained zone
            for faceID in self.constrZoneMap[LOAId]:
                dim = self.model.GetDispBCs(faceID)[0]
                disp = self.model.GetDispBCs(faceID)[1]*dispFactor
                if dim == 'x': increment[0] += disp
                if dim == 'y': increment[1] += disp
                if dim == 'z': increment[2] += disp
            if (increment != [0,0,0]):
                self.ioControl.Print(4,"Displacing faces: " + repr(self.constrZoneMap[LOAId]) + " by " + repr(increment))
            for atomID in xrange(numAtoms):
                for coord in xrange(3):
                    positions[atomID][coord] += increment[coord]
            self.constraints[LOAId].SetFixedPositions(positions)

        #stretch bulk - must find strain tensor from displacement
        strainTensor = array([[0.0,0.0,0.0],[0.0,0.0,0.0],[0.0,0.0,0.0]])
        #loop over faces to get geometry
        faceGeometry = {}
        center = (0,0,0)
        for faceID in self.model.GetFaceIds():
            # find normal and surface point
            points = []
            points.append(array(self.model.GetNodeCoord(self.model.GetFaceTriangles(faceID)[1][0],0)))
            points.append(array(self.model.GetNodeCoord(self.model.GetFaceTriangles(faceID)[1][1],0)))
            points.append(array(self.model.GetNodeCoord(self.model.GetFaceTriangles(faceID)[1][2],0)))
            surfaceNormal = GetNormalizedVector(CrossProduct(points[0]-points[1],points[2]-points[1]))            
            faceGeometry[faceID] = (surfaceNormal,points[0])


        self.ioControl.Print(5,"faceGeometry used to find strain matrix: " + repr(faceGeometry))
        for faceID in faceGeometry.keys():
            increment = [0.0,0.0,0.0]
            dim = self.model.GetDispBCs(faceID)[0]
            disp = self.model.GetDispBCs(faceID)[1]*dispFactor
            if dim == 'x': increment[0] += disp
            if dim == 'y': increment[1] += disp
            if dim == 'z': increment[2] += disp
            if (increment != [0,0,0]):
                normal1 = faceGeometry[faceID][0]
                surface1 = faceGeometry[faceID][1]
                # find length
                for faceID2 in faceGeometry.keys():
                    normal2 = faceGeometry[faceID2][0]
                    if faceID != faceID2 and abs(dot(normal1,normal2)) == 1:
                        surface2 = faceGeometry[faceID2][1]
                        length = abs(dot(surface1 - surface2,normal1)) - 2*self.atomsDecorator.constrZoneWidth
                        center -= (length / 2) * normal1
                
                NonSymmStrain = OuterProduct(increment,normal1)
                strainTensor += (0.5/length) * (NonSymmStrain + Transpose(NonSymmStrain))

        # transform all atoms - rollered constraints will hold automatically
        self.ioControl.Print(5,"Straining all atoms by " + repr(strainTensor))
        transformer = self.MD.LinearPositionTransformer(strainTensor)
        transformer.SetCenter(center)
        transformer.Transform(self.atoms)
        self.atoms.Notify()
        #self.Notify()


    def MoveConstrainedZones(self):
        #moves entire endcap by the same amount
        positions1 = self.endcap1.GetCartesianPositions()
        avg1 = positions1.Average()
        disp1 = self.Parameters["strainIncrement"] * (avg1[0] - self.Parameters["center"][0])
        positions2 = self.endcap2.GetCartesianPositions()
        avg2 = positions2.Average()
        disp2 = self.Parameters["strainIncrement"] * (avg2[0] - self.Parameters["center"][0])
        disp=(disp1-disp2)/2.0 #otherwise it will just move off to the side with rolled bc!

        for i in xrange(self.endcap1.NumberOfLeaves()):
            numAtoms1 = self.endcap1.GetLeaf(i).GetNumber()
            positions1 = self.endcap1.GetLeaf(i).GetCartesianPositions()
            self.ioControl.Print(4,"Moving " + repr(numAtoms1) + " by " + repr(disp))
            for j in xrange(numAtoms1):
                positions1[j][0] += disp
            self.endcap1.GetLeaf(i).GetConstraint().SetFixedPositions(positions1)
        
        for i in xrange(self.endcap2.NumberOfLeaves()):
            numAtoms2 = self.endcap2.GetLeaf(i).GetNumber()
            positions2 = self.endcap2.GetLeaf(i).GetCartesianPositions()
            self.ioControl.Print(4,"Moving " + repr(numAtoms2) + " by " + repr(-disp))
            for j in xrange(numAtoms2):
                positions2[j][0] -= disp
            self.endcap2.GetLeaf(i).GetConstraint().SetFixedPositions(positions2)


    def CalculateStress(self):
        self.ioControl.Print(2,"Calculating Stress")

        stress1=0
        stress2=0


        for i in xrange(self.endcap1.NumberOfLeaves()):
            stress1+=self.endcap1.GetLeaf(i).GetConstraint().GetTotalForce(0)
            stress2+=self.endcap2.GetLeaf(i).GetConstraint().GetTotalForce(0)

        stress1 /= self.Parameters["width"]*self.Parameters["height"]
        stress2 /= self.Parameters["width"]*self.Parameters["height"]
        return stress1
    

    def BreakApart(self,start=0):

        if start==0:
            self.Relax(self.qm)

##         # tracking a bad atoms - temporary debugging 10/17
##         positions = self.atoms.GetCartesianPositions()

##         # set id numbers and processors in an array to track the atoms as they are moved across processors
##         atomIDs = positions.ConstructFromArray(positions)
##         #forces = positions.ConstructFromArray(positions)
##         number = self.atoms.GetNumber()

##         for i in xrange(number):
##             atomIDs[i][0] = i
##             atomIDs[i][1] = self.Parameters["myProc"]
##         if self.Parameters["parallel"]:  self.mpiworld.barrier()
        
##         tempBadAtomID = array([0,0]) #6906
##         badAtomID = array([0,0]) #6906
##         foundAtom = 0
##         # this is not playing nicely with parallel...
##         print "finding forces"
##         sys.stdout.flush()
##         forces = self.potential.Forces(self.atoms)
##         for i in xrange(number):
##             if (abs(positions[i][0]+0.10661)<1e-3 and abs(positions[i][1]-3.66541)<1e-3 and abs(positions[i][2]-1.140593)) :
##                 line1 = "BAD ATOM: " + repr(i) + " position: "
##                 line2 = "force on atom: "
##                 tempBadAtomID[0] = atomIDs[i][0]
##                 tempBadAtomID[1] = atomIDs[i][1]
##                 foundAtom = 1
##                 print str(atomIDs[i])
##                 for j in xrange(3):
##                     line1 += " " + repr(positions[i][j])
##                     line2 += " " + repr(forces[i][j])
##                 line1 += " myProc: " + repr(self.Parameters["myProc"])
##                 print line1
##                 print line2
##                 sys.stdout.flush()
##         print "PROCESSOR : " + repr(self.Parameters["myProc"]) + " has "+ repr(number) +" atoms"
##         sys.stdout.flush()
        
##         if self.Parameters["parallel"]:
##             print "trying to send stuff!" + str(self.Parameters["myProc"])
##             sys.stdout.flush()
##             Scientific.MPI.world.allreduce(tempBadAtomID,badAtomID,Scientific.MPI.sum)

##         print str(badAtomID) + " " + str(self.Parameters["myProc"])
##         sys.stdout.flush()

        if self.Parameters["deformationMethod"] == "kinematicBC":

            self.atomsDecorator.ImplementConstraints()
            self.constraints = self.atomsDecorator.constraints
           
            for i in xrange(start,self.Parameters["nSteps"]):
                self.ioControl.Print(3,"MD step " + repr(i+1))
                self.IncrementBoundaryDisplacements()
                self.Relax(self.qm)

##                 # tracking a bad atoms - temporary debugging 10/17
##                 positions = self.atoms.GetCartesianPositions()
##                 #self.potential.CalculateForces(self.atoms,forces)
##                 number = self.atoms.GetNumber()
##                 for k in xrange(number):
##                     #if (abs(positions[k][0]-oldPositionOfBadAtom[0]) < .5 and abs(positions[k][1] - oldPositionOfBadAtom[1]) < .5 and abs(positions[k][2]-oldPositionOfBadAtom[2]) < .5):
##                     if (atomIDs[k][0] == badAtomID[0] and atomIDs[k][1]== badAtomID[1]):
##                         line1 = "BAD ATOM: " + repr(k) + " position: "
##                         line2 = "force on atom: "
##                         for j in xrange(3):
##                             line1 += " " + repr(positions[k][j])
##                             line2 += " " + repr(forces[k][j])
##                         line1 += " myProc: " + repr(self.Parameters["myProc"])
##                         print line1
##                         print line2
##                         sys.stdout.flush()
##                     #if k == badAtomID[0]:
##                     #    for j in xrange(3):
##                     #        print positions[k][j]-oldPositionOfBadAtom[0]
##                 print "PROCESSOR : " + repr(self.Parameters["myProc"]) + " has "+ repr(number) +" atoms"
##                 sys.stdout.flush()
        
 
        elif self.Parameters["deformationMethod"] == "meshDisp":

            self.atomsDeformer = AtomsDeformer.AtomsDeformer(self.atoms,self.model,self.elem_hash,self.Parameters,self.ioControl)
            self.atomsDeformer.FindNaturalCoordinatesWithinElements()
            self.atomsDeformer.GetNodeDisplacementsFromRepository()
            self.ioControl.Print(3,"LEAVES LIST " + repr(xrange(self.atoms.NumberOfLeaves())))
            self.atomsDeformer.DeformAtoms(xrange(self.atoms.NumberOfLeaves()))
            self.Notify()        
            self.atomsDecorator.ImplementConstraints()
            self.constraints = self.atomsDecorator.constraints
            self.Relax(self.qm)        

            

        elif self.Parameters["deformationMethod"] == "cohesiveLaw":

            self.atomsDecorator.ImplementConstraints()
            self.constraints = self.atomsDecorator.constraints

            stressStrainFile = open("stressStrain.txt","w")
            stress = self.CalculateStress()
            strain = 0
            self.ioControl.Print(1,"stress = " + repr(stress) + " strain = " + repr(strain))
            stressStrainFile.write(repr(strain) + " " + repr(stress) + "\n")
            stressStrainFile.flush()
            for i in xrange(self.Parameters["nSteps"]):
                self.ioControl.Print(2,"Stretching by " + repr(self.Parameters["strainIncrement"]))
                self.MoveConstrainedZones()
                self.Relax(self.qm)
                stress = self.CalculateStress()
                strain = self.Parameters["strainIncrement"]*(i+1)
                self.ioControl.Print(1,"stress = " + repr(stress) + " strain = " + repr(strain))
                stressStrainFile.write(repr(strain) + " " + repr(stress) + "\n")
                stressStrainFile.flush()


    def FindGrainBoundaryGeometries(self):
        self.geometryFinder = GrainBoundaryGeometryFinder.GrainBoundaryGeometryFinder(self.Parameters,self.ioControl)
        self.geometryFinder.MakeModel()
        self.geometryFinder.FindGrainBoundaryGeometries()


    def MigrateBack(self):
        self.atomsDecorator.migrator.MigrateBack()

    def Migrate(self):
        self.atomsDecorator.migrator.Migrate()

# end of OFEMD class

# if calling directly from command line or web service
if __name__ == "__main__":


    ofemd = OFEMD()
    UserParameters = ProcessCommandLine(ofemd.Parameters)
    if UserParameters == None:
        UserParameters = {}
    ofemd.SetParameters(UserParameters)
    ofemd.ParameterCheck()

    if ofemd.Parameters["findGrainBoundaryGeometries"]:
        ofemd.ioControl.Print(3,"Finding Grain Boundary Geometries")
        ofemd.FindGrainBoundaryGeometries()

    else:
    
        #asciiObserver = ASCIIObserver.ASCIIObserver(ofemd.Parameters,ofemd.ioControl)
        #ofemd.AddObserver(asciiObserver)
        dxObserver = DXObserver.DXObserver(ofemd.Parameters,ofemd.ioControl)
        ofemd.AddObserver(dxObserver)
        dxStressObserver = DXStressObserver.DXStressObserver(ofemd.Parameters,ofemd.ioControl)
        ofemd.AddObserver(dxStressObserver)
        #meshObserver = OFEMDASCIIObserver.OFEMDASCIIObserver(ofemd.Parameters,ofemd.ioControl)
        #ofemd.AddObserver(meshObserver)

        ofemd.GetMesh()
        ofemd.DecorateWithAtoms()
        # ofemd.AddPlotAtomsObserver()
    ##     for i in xrange(10000):
    ##         for j in xrange(10000):
    ##             ofemd.atoms.Notify()
    ##             x=1+2
    #    ofemd.atoms.Notify()
        ofemd.BreakApart()
        dxObserver.WriteTail(ofemd)
        dxStressObserver.WriteTail(ofemd)

    ofemd.ioControl.Print(1,"It Worked!!!!")

