#AtomsFiller.py

import sys
sys.path.append("..")
from MDWebService import *
from AtomsDecorator import *


class SphereDecorator(AtomsDecorator):

    def __init__(self,Parameters,ioControl,MD,model):
        AtomsDecorator.__init__(self,Parameters,ioControl,MD,model)
        self.numConstrainedLeaves = len(self.model.GetMaterialIds()) # needed by base class!

        
    def MakeLeaves(self):

        # declare spherical cluster initializer
        sphere = self.MD.SphericalClusterInitializer(self.Parameters["radius"], self.lattice)
        if self.Parameters["parallel"]:
            parasphere = self.MD.ParallelAtomsInitializer(sphere)
            initializer = parasphere
        else:
            initializer = sphere
            

        # for each material (or grain)
        LOAId = 0
        for matID in self.model.GetMaterialIds():

            # get rotation
            rotation = array(self.model.GetMaterialData(matID))


            # get list of elems with current material id
            elems = []
            for elemID in self.model.GetElemIds():
                if self.model.GetElemData(elemID)[1] == matID:
                    elems.append(elemID)

            self.ioControl.Print(4,"Rotation = " + repr(rotation))

            # generate spherical cluster of atoms
            sphere.SetCenter(self.Parameters["center"])
            self.lattice.ResetOrientation()
            self.lattice.RotateLatticeVectors(rotation)
            initializer.Create(self.BulkLOAs[LOAId])

            # remove atoms outside material region
            self.ioControl.Print(3,"Removing extra atoms...")
            self.RemoveExtraAtoms(elems,self.BulkLOAs[LOAId])

            # increment ListOfAtoms
            LOAId += 1
            

    def MakeConstrainedLeaves(self):
        
        self.constrZoneMap = []

        # declare spherical cluster initializers
        constrZoneWidth = self.potential.GetCutoffDistance()
        sphere1 = self.MD.SphericalClusterInitializer(self.Parameters["radius"], self.lattice)
        sphere2 = self.MD.SphericalClusterInitializer(self.Parameters["radius"] + constrZoneWidth, self.lattice)

        # for each material (or grain)
        LOAId = 0
        for matID in self.model.GetMaterialIds():

            # get rotation
            rotation = array(self.model.GetMaterialData(matID))

            # get list of elems with current material id
            elems = []
            for elemID in self.model.GetElemIds():
                if self.model.GetElemData(elemID)[1] == matID:
                    elems.append(elemID)

            self.ioControl.Print(4,"Rotation = " + repr(rotation))

            # generate spherical cluster of atoms
            sphere1.SetCenter(self.Parameters["center"])
            sphere2.SetCenter(self.Parameters["center"])
            shell = self.MD.CompositeClusterInitializer(sphere2,sphere1,self.lattice,"difference")
            if self.Parameters["parallel"]:
                parashell = self.MD.ParallelAtomsInitializer(shell)
                initializer1 = parashell
            else:
                initializer1 = shell

            self.lattice.ResetOrientation()
            self.lattice.RotateLatticeVectors(rotation)
            initializer1.Create(self.ConstrLOAs[LOAId])

            # remove atoms outside material region
            self.ioControl.Print(3,"Removing extra atoms...")
            self.RemoveExtraAtoms(elems,self.ConstrLOAs[LOAId])

            # increment ListOfAtoms
            LOAId += 1



    def ImplementConstraints(self):

        # implement constraints
        self.ioControl.Print(3,"Implementing Constraints...")
        self.constraints = []
        for LOAId in xrange(len(self.ConstrLOAs)):
            self.constraints.append(self.MD.FixedBody(self.ConstrLOAs[LOAId].GetCartesianPositions()))
            self.ConstrLOAs[LOAId].SetConstraint(self.constraints[LOAId])
 
                    
