#!/afs/msc.cornell.edu/i386_linux22/usr/local/pub/dm/bin/python

import sys
import os
from Numeric import *
import Scientific.MPI

class DXObserver:

    def __init__(self, Parameters, ioControl):
        self.ioControl = ioControl


    def Update(self, mdobj):
        mdobj.ioControl.Print(3,"Writing frame " + repr(mdobj.Parameters["frameNumber"]) + " to DX file")
        nAtoms = mdobj.atoms.GetNumber()
        nTotalAtoms = array(nAtoms)
        if mdobj.Parameters["parallel"]:
            mdobj.mpiworld.reduce(array(nAtoms), nTotalAtoms, Scientific.MPI.sum, 0)
        if nTotalAtoms[0] > 0:
            if mdobj.Parameters["parallel"]: mdobj.mpiworld.barrier()
            if mdobj.Parameters["myProc"] == 0:
                self.ioControl.OpenFile("atomdata.dx")
                self.ioControl.WriteToFile("atomdata.dx","object " + repr(mdobj.Parameters["frameNumber"]) + " class array type float rank 1 shape " + repr(mdobj.DIMENSION) + " items " + repr(nTotalAtoms[0]) + " data follows\n")
                self.ioControl.CloseFile("atomdata.dx")
            

            for leafIdx in xrange(mdobj.atoms.NumberOfLeaves()):
                positions = mdobj.atoms.GetLeaf(leafIdx).GetCartesianPositions()[:]
                for proc in xrange(mdobj.Parameters["nProcs"]):
                    if mdobj.Parameters["parallel"]: mdobj.mpiworld.barrier()
                    if mdobj.Parameters["myProc"] == proc:
                        self.ioControl.OpenFile("atomdata.dx")
                        for atIdx in xrange(mdobj.atoms.GetLeaf(leafIdx).GetNumber()):
                            line = ""
                            for coord in xrange(len(positions[atIdx])):
                                line = line + repr(positions[atIdx][coord]) + "   "
                            line = line + "\n"
                            self.ioControl.WriteToFile("atomdata.dx",line,proc)
                            self.ioControl.FlushFile("atomdata.dx")
                        self.ioControl.CloseFile("atomdata.dx")


        if mdobj.Parameters["parallel"]: mdobj.mpiworld.barrier()
        if mdobj.Parameters["myProc"] == 0:
            self.ioControl.OpenFile("atomdata.dx")
            self.ioControl.WriteToFile("atomdata.dx","\n")
            self.ioControl.FlushFile("atomdata.dx")
            self.ioControl.CloseFile("atomdata.dx")
            if mdobj.Parameters["frameNumber"] > 0:
                self.ioControl.RemoveFile("tempTail.txt")
        # self.ioControl.OpenFile("tempTail.txt")
        self.WriteTail(mdobj,"tempTail.txt")
        # self.ioControl.CloseFile("tempTail.txt")

    def WriteTail(self, mdobj, tailFile=None):
        if tailFile == None:
            tailFile = "atomdata.dx"
            tailNumber = mdobj.Parameters["frameNumber"]
        else:
            tailNumber = mdobj.Parameters["frameNumber"]+1
        
        # color info
        nAtoms = mdobj.atoms.GetNumber()
        nTotalAtoms = array(nAtoms)
        if mdobj.Parameters["parallel"]:
            mdobj.mpiworld.reduce(array(nAtoms), nTotalAtoms, Scientific.MPI.sum, 0)
        

        if mdobj.Parameters["myProc"] == 0:
            self.ioControl.OpenFile(tailFile)
            self.ioControl.WriteToFile(tailFile,"object " + repr(tailNumber) + " class array type float rank 0 items " + repr(nTotalAtoms[0]) + " data follows\n")
            self.ioControl.CloseFile(tailFile)


        for leafIdx in xrange(mdobj.atoms.NumberOfLeaves()):
            for proc in xrange(mdobj.Parameters["nProcs"]):
                if mdobj.Parameters["parallel"]: mdobj.mpiworld.barrier()
                if mdobj.Parameters["myProc"] == proc:                    
                    self.ioControl.OpenFile(tailFile)
                    for i in xrange(mdobj.atoms.GetLeaf(leafIdx).GetNumber()):
                        self.ioControl.WriteToFile(tailFile,repr(leafIdx) + "\n",proc)
                    self.ioControl.FlushFile(tailFile)
                    self.ioControl.CloseFile(tailFile)


        if mdobj.Parameters["parallel"]: mdobj.mpiworld.barrier()
        if mdobj.Parameters["myProc"] == 0:
            self.ioControl.OpenFile(tailFile)
            self.ioControl.WriteToFile(tailFile,"\nattribute \"dep\" string \"positions\"\n\n")
            for frameIdx in xrange(tailNumber):
                objName = "time_"
                if frameIdx < 10:
                    objName = objName + "0"
                objName = objName + repr(frameIdx)
                self.ioControl.WriteToFile(tailFile,"object \"" + objName + "\" class field\n")
                self.ioControl.WriteToFile(tailFile,"component \"positions\" value " + repr(frameIdx) + "\n")
                self.ioControl.WriteToFile(tailFile,"component \"data\" value " + repr(tailNumber) + "\n\n")

            self.ioControl.WriteToFile(tailFile,"\nobject \"series\" class series\n")
            for frameIdx in xrange(tailNumber):
                objName = "time_"
                if frameIdx < 10:
                    objName = objName + "0"
                objName = objName + repr(frameIdx)
                self.ioControl.WriteToFile(tailFile,"member " + repr(frameIdx) + " value \"" + objName + "\"\n")

            self.ioControl.WriteToFile(tailFile,"\nend\n\n")
            self.ioControl.CloseFile(tailFile)


    def WriteGifScript(self, mdobj):

        self.ioControl.OpenFile("makeGif")

        self.ioControl.Print(3,"jumpBackFrames = " + str(mdobj.jumpBackFrames))

        convertString = """#!/bin/tcsh

foreach file (*.tiff)
	echo $file " to " $file:s/tiff/gif/
	convert $file $file:s/tiff/gif/
	end

gifsicle --colors 16 -U frame.*.gif > temp.gif
"""
        self.ioControl.WriteToFile("makeGif",convertString)

        firstFrame = 0
        partNum = 1
        for jumpFrame,jumpSize in mdobj.jumpBackFrames:
            numFrames = jumpSize*mdobj.Parameters["strainSubdivisions"]
            if (firstFrame==0 or jumpSize==2):
                gifString = "gifsicle -d 50 temp.gif '#" + repr(firstFrame) + "-" + repr(jumpFrame-3) + "' > part" + repr(1000+partNum) + ".gif\n"
                self.ioControl.WriteToFile("makeGif",gifString)
                partNum += 1
            gifString = "gifsicle -d " + str(50/mdobj.Parameters["strainSubdivisions"]) + " temp.gif '#" + repr(jumpFrame) + "-" + repr(jumpFrame+numFrames-1) + "' > part" + repr(1000+partNum) + ".gif\n"
            self.ioControl.WriteToFile("makeGif",gifString)
            firstFrame = jumpFrame + numFrames
            partNum += 1
        gifString = "gifsicle -d 50 temp.gif '#" + repr(firstFrame) + "-' > part" + repr(1000+partNum) + ".gif\n"
        self.ioControl.WriteToFile("makeGif",gifString)
        gifString = """gifsicle -l part*.gif > temp2.gif
gifsicle -t '#000000' temp2.gif > temp3.gif
gifsicle --crop-transparency temp3.gif > temp4.gif
gifsicle --resize-width 300 temp4.gif > atoms.gif"""
        self.ioControl.WriteToFile("makeGif",gifString)
         
            
            
        
