import os
import sys
import time
import tempfile
import string
from types import *
from Numeric import *



class IOControl:
    def __init__(self,verbose, directory, html = 0, recordPrintStatements = 0, **keywordArgs):
        self.verbose = verbose
        self.html = html
        self.files = {}
        self.SetDirectory(directory)
        self.recordPrintStatements = recordPrintStatements
        self.printRecord = ""
        self.myProc = 0

        for item in keywordArgs.items():
            setattr(self,item[0],item[1])

                    
        if self.html:
            self.PrintHTMLHeader()


    def __del__(self):
        if self.html:
            self.PrintHTMLFooter()


    def SetVerbosity(self,verbose):
        self.verbose=verbose

    def SetDirectory(self,directory):
        if not os.access(directory,os.F_OK) and self.myProc == 0:
            os.mkdir(directory)
        self.directory=directory

    def SetMyProc(self,myProc):
        self.myProc = myProc

    def GetVerbosity(self): 
        return self.verbose

    def Print(self,verbosity,printable,proc=0):
        # Takes a verbosity argument and optionally adds an HTML line break tag
        # printable can be a string or a number or a list of strings and
        # numbers
        if self.verbose < verbosity or self.myProc != proc:
            return
        

        if type(printable) == type(""): # string
            printString = printable
        elif type(printable) == type([]): # list
            printString = ""
            for item in printable:
                printString += repr(item) + " "
        else:
            printString = repr(printable) 
                
        if self.html:
            printString += "<BR>"

        print printString
        sys.stdout.flush()

        if self.recordPrintStatements:
            self.printRecord += printString


    def OpenFile(self,name):
        tries = 0
        fileopen = 0

        while(not fileopen and tries < 10):
            try:
                if name not in self.files.keys() and self.myProc == 0:
                    self.files[name]=open(self.directory+"/"+name,"w")
                else:
                    self.files[name]=open(self.directory+"/"+name,"a")
                if self.files[name]:
                    fileopen = 1
                tries = tries+1

            except:
                print "could not open", name, ". trying again."
                time.sleep(3)

        if(not self.files[name]):
            print "Error, could not open ", name, " after 10 tries"



    def RemoveFile(self,name):
        if name in self.files.keys() and self.myProc == 0:
            os.remove(self.directory + "/" + name)
            del self.files[name]

    def WriteToFile(self,fileName,string,proc=0):
        if self.myProc == proc:
            self.files[fileName].write(string)
            self.files[fileName].flush()
        
    def FlushFile(self,filename):
        self.files[filename].flush()

    def CloseFile(self,filename):
        self.files[filename].close()

                
    def PrintHTMLHeader(self):
        print "Content-type: text/html\n\n\n"
    
        print "<HTML>"
        print "<HEAD> <TITLE>",
        if hasattr(self, "htmlHeaderString"):
            print self.htmlHeaderString,
        print  "</TITLE> </HEAD>"
        print "<BODY BGCOLOR=#ffffff>"


    def PrintHTMLFooter(self):
        if hasattr(self,"htmlFooterString"):
            print self.htmlFooterString + "<BR>"
        print "</BODY>"
        print "</HTML>"

    def GetDMArrayFromFile(self,filename,md,array=None):
        dmarrayfile = open(filename,"r")
        allLines = dmarrayfile.readlines()
        num =  len(allLines)
        print "lines in ",filename, " = ", num
        if array != None:
            print "size of array = ", array.GetShape()
            dmarray = array.ConstructFromArray(array)
        else:
            dmarray = md.ArrayDoubleVector(num)
        for i in xrange(num):
            stringVector = string.splitfields(allLines[i])
            for coord in xrange(len(stringVector)):
                dmarray[i][coord] = float(stringVector[coord])

        return dmarray
        

def ProcessCommandLine(defaults, start=1):
    UserParameters = {}

    if len(sys.argv) > 1:

        if sys.argv[1] == "--help":
            print "usage:"
            print "python ScriptName.py [option] [value] ..."
            print "options (order not significant):"
            print defaults.keys()
            sys.exit()

        ParametersLeft = 1
        i = start
        while ParametersLeft:
            if i == len(sys.argv):
                ParametersLeft=0
            else:
                key = sys.argv[i]
                defaultValue = defaults[key]
                if isinstance(defaultValue,IntType):
                    UserParameters[key] = int(sys.argv[i+1])
                    i = i+2
                elif isinstance(defaultValue,FloatType):
                    UserParameters[key] = float(sys.argv[i+1])
                    i = i+2
                elif isinstance(defaultValue,StringType):
                    UserParameters[key] = sys.argv[i+1]
                    i = i+2
                elif isinstance(defaultValue,ListType):
                    length = len(defaultValue)
                    length2 = 0
                    totalLength = 0
                    newList = list(xrange(length))
                    for j in xrange(length):
                        if isinstance(defaultValue[j],ListType):
                            length2 = len(defaultValue[j])
                            newList[j] = []
                            for k in xrange(length2):
                                newList[j].append(float(sys.argv[i+1+length2*j+k]))
                            totalLength += length2
                        else:
                            newList[j] = float(sys.argv[i+1+j])
                            totalLength = length
                    UserParameters[key]=newList
                    i = i + totalLength + 1
                elif isinstance(defaultValue,ArrayType):
                    length = len(defaultValue)
                    newList = list(xrange(length))
                    for j in xrange(length):
                        newList[j] = float(sys.argv[i+1+j])
                    UserParameters[key]=array(newList)
                    i = i + length + 1
                elif key not in defaults.keys():
                    print "Invalid option: ", key
                    sys.exit()


        print UserParameters
        return UserParameters

