import sys
sys.path.append("/afs/msc.cornell.edu/home/jsethna/val/ASP/MDWebServices/")
from MDWebService import *
from random import *


def Transpose(array):

    transposeOfArray = []
    for i in xrange(len(array[0])):
        transposeOfArray.append([])
        for j in xrange(len(array)):
            transposeOfArray[i].append(array[j][i])

    return transposeOfArray

def GetArrayNorm(array1D,coordSep=0):
    # if coordSep is = 0 then we sum over coords; otherwise we return a
    # separate norm for each coordinates 
    sum = [0.,0.,0.]
    for i in xrange(0,len(array1D),3):
        for c in xrange(3):
            sum[c] += array1D[i+c] * array1D[i+c]

    if coordSep == 0:
        return sqrt(sum[0] + sum[1] + sum[2])
    else:
        return sqrt(sum)

def CrossProduct(vect1,vect2):
    return array([vect1[1]*vect2[2]-vect1[2]*vect2[1],vect1[2]*vect2[0]-vect1[0]*vect2[2],vect1[0]*vect2[1]-vect1[1]*vect2[0]])

def GetNormalizedVector(vect):
    normVec = vect[:]
    norm = GetArrayNorm(vect)
    for cd in xrange(len(normVec)):
        normVec[cd] /= norm
    return array(normVec)

def OuterProduct(vect1,vect2):
    answer = array([[0.0,0.0,0.0],[0.0,0.0,0.0],[0.0,0.0,0.0]])
    for i in xrange(3):
        for j in xrange(3):
            answer[i][j]=vect1[i]*vect2[j]
    return answer

def GetRandomRotationMatrix(randomNumberSeed=None):
    if randomNumberSeed != None:
        seed(randomNumberSeed)
    vect1 = GetNormalizedVector([random(),random(),random()])
    tempvect = GetNormalizedVector([random(),random(),random()])
    vect2 = GetNormalizedVector(CrossProduct(vect1,tempvect))
    vect3 = GetNormalizedVector(CrossProduct(vect1,vect2))
    rotMatrix = [vect1,vect2,vect3]
    return rotMatrix

def MaxInMatrix(mat):
    max = -1.e99
    for i in xrange(len(mat)):
        for j in xrange(len(mat[i])):
            if abs(mat[i][j]) > max:
                max = abs(mat[i][j])
                imax = i
                jmax = j
    return (imax,jmax,max)


def DiagonalPart(mat):
    prod = 1.
    length = len(mat)
    diag = zeros((length,length),Float64)
    for i in xrange(length):
        diag[i][i] = mat[i][i]
        prod *= mat[i][i]
    print prod
    return diag

def SubMatrix(mat, size):
    if size > len(mat):
        return
    sub = zeros((size,size),Float64)
    for i in xrange(size):
        for j in xrange(size):
            sub[i][j] = mat[i][j]
    return sub

def SaveNodeCoords(directory, model):
    fileName = 'nodes'
    file = open(directory + "/" + fileName,'w')
    for i in xrange(model.NumNodes()):
        node = model.nodes[i]
        s = '%15.8f\t%15.8f\t%15.8f\n' % node
        file.write(s)



def WriteNodesByElement(directory, model,node_disps):
    for elementIdx in xrange(model.NumElems()):
        nodeFile = open(directory + "/" + 'nodesElem%d'%elementIdx,'w')
        nodeNums = model.GetElemData(elementIdx)[3]
        nodeCoords = model.GetElemCoords(elementIdx)
        for i in xrange(len(nodeCoords)):
            eq = nodeNums[i] * 3
            nodeFile.write('%10.8f\t%10.8f\t%10.8f\n'%(nodeCoords[i][0]+node_disps[eq],nodeCoords[i][1]+node_disps[eq+1],nodeCoords[i][2]+node_disps[eq+2]))
        nodeFile.close()
        
def DisplayNodeVectors(nodeVectors):
    assert(len(nodeVectors) %3 == 0)
    nNodes = len(nodeVectors) / 3

    for i in xrange(nNodes):
        print nodeVectors[3*i],nodeVectors[3*i+1],nodeVectors[3*i+2]




class FaceForceDispObserver:
    def __init__(self, elIdx, faceIdx, coord, ioControl):
        self.elIdx = elIdx
        self.faceIdx = faceIdx
        self.coord = coord
        self.faceForce = 0.
        self.ioControl = ioControl
        import ShapeFunc
        self.ShapeFunc = ShapeFunc
        
    def Update(self, mainObjects, nodeForces = None):
        self.faceForce = 0.
        self.faceDisp = 0.

        if not nodeForces:
            nodeForces = mainObjects.fullElasticForces + (mainObjects.atomisticForces - mainObjects.partialElasticForces)

            #nodeForces = mover.forceCorrection
        nodeDisps = mainObjects.node_disps

        # DisplayNodeVectors(nodeForces)
        # get list of nodes corresponding to the face!
        elemData =  mainObjects.FEModel.GetElemData(self.elIdx) 
        elemType = elemData[1]
        connectivity = elemData[3]
        faceMapCommand = "faceMap = self.ShapeFunc." + elemType + "Face_map(" + `self.faceIdx` + ")"
        exec(faceMapCommand)

        faceNodes = [] 
        for idx in xrange(len(faceMap)):
            faceNodes.append(connectivity[faceMap[idx]])

        for node in faceNodes:
            self.faceForce += nodeForces[3*node + self.coord]

            self.faceDisp += nodeDisps[3*node + self.coord]

        self.faceDisp /= len(faceNodes)

        self.Display()

    def Display(self):
        self.ioControl.Print(3,['force[%d] on element %d, face %d:' %(self.coord, self.elIdx, self.faceIdx),self.faceForce])
        self.ioControl.Print(3,['ave. displacement[%d] on element %d, face %d:' %(self.coord, self.elIdx, self.faceIdx),self.faceDisp])
        
    def GetFaceForce(self):
        return self.faceForce

    def GetFaceDisp(self):
        return self.faceDisp

# given a disp for once face, linearly interpolate other node disps
def ComputeNodeDispsX(face_disp):
    node_disps = zeros(96,Float64)
    nodePos = fem_md.FEModel.nodes
    for key in nodePos.keys():
        xPos = nodePos[key][0]
        node_disps[key*3] = face_disp * xPos/width[0]

    return node_disps



def GetNbrDisps(mainObj, node_disps, atomIdx):
    mainObj.DisplaceAtoms(node_disps, [0,1,2])
    nl = mainObj.atoms.GetNeighborLocator()
    nbrs = nl.Neighbors(atomIdx)
    disps = nbrs[2]
    return disps

def AtomisticEnergy(mainObj,node_disps):
    mainObj.DisplaceAtoms(node_disps, [0,1,2])
    return mainObj.CalculateWeightedEnergy()

def CompareEnergies(faceDisp):
    node_dispsZero = zeros(96,Float64)

    node_disps = ComputeNodeDispsX(faceDisp)

    elEnergy = PartialElasticEnergy(node_disps)
    
    atEnergy = AtomisticEnergy(node_disps)
    atEnergy0 = AtomisticEnergy(node_dispsZero)

    print elEnergy,atEnergy - atEnergy0


def GetNodesAffectingAtoms(model,coupling):
    nodesAffectingAtoms = []
    for el in coupling.atomContainingElements:
        nodes = model.GetElemData(el)[3]
        for nd in nodes:
            if nd not in nodesAffectingAtoms:
                nodesAffectingAtoms.append(nd)
    return nodesAffectingAtoms

def CrackTestDisps(model,nodesToSet,val):
    node_disps = zeros(model.NumNodes()*3,Float64)
    strain = array([[0.,0.,0.],[0.,val,0.],[0.,0.,0.]])
    
    count = 0
    for node in nodesToSet:
        nodePos = model.nodes[node]
        disp = matrixmultiply(strain,nodePos)
        

        node_disps[3*node:3*node+3] = disp
        #if count == 0:
        #    print 'node',node,'nodePosY',nodePosY,'disp',disp
        count += 1
        
    return node_disps


def ForceDispCurve(maxDisp, nPoints):
    global gPlot
    fullPoints = []
    elasticPoints = []
    partialElasticPoints = []
    atomicPoints = []
    diffPoints = []
    for i in xrange(nPoints):
        disp = i * maxDisp / nPoints
        print disp,
        #node_disps = zeros(24,Float64)
        #node_disps[4*3] = disp
        #node_disps[5*3] = disp
        #node_disps[6*3] = disp
        #node_disps[7*3] = disp

        node_disps = ComputeNodeDispsX(disp)


        fem_md.coupling.DisplaceAtoms(node_disps,fem_md.atoms, [1,2])

        atomicForces = fem_md.coupling.CalculateNodalForces(fem_md.atoms,fem_md.eqn_nums)
        partialElasticForces = matrixmultiply(fem_md.partialStiffnessMatrix,node_disps)

        fullElasticForces = matrixmultiply(fem_md.globalMatrix,node_disps)
        # we're really adding the atomicForces here, not subtracting
        # it's just that the quantity we want is really the negative
        # of the forces on the nodes due to the material
        # rather, it's the external force needed to balance at the spacified
        # displacement. Otherwise we would have to add a minus sign to the
        # elastic terms
        fullForces = fullElasticForces - partialElasticForces - atomicForces

        nd = 7
        # print fullElasticForces[nd*3],partialElasticForces[nd*3],atomicForces[nd*3],fullForces[nd*3]

        # add up contributions from nodes on the right most face
        fullForceFace = FaceXForce(fullForces)
        fullElasticForceFace = FaceXForce(fullElasticForces)
        atomicForceFace = FaceXForce(atomicForces)
        partialElasticForceFace = FaceXForce(partialElasticForces)
        
        print fullElasticForceFace, partialElasticForceFace, atomicForceFace, fullForceFace

        fullPoints.append([disp,fullForces[nd*3]])
        elasticPoints.append([disp,fullElasticForces[nd*3]])
        diffPoints.append([disp,fullForces[nd*3]-fullElasticForces[nd*3]])
        partialElasticPoints.append([disp,partialElasticForces[nd*3]])
        atomicPoints.append([disp,-atomicForces[nd*3]])
    from Gnuplot import Gnuplot
    gPlot = Gnuplot()
    gPlot.plot(fullPoints,elasticPoints)
    #gPlot.plot(partialElasticPoints,atomicPoints)
    #gPlot.plot(diffPoints)

    fullPtsFile = open("fullPts.dat","w")
    elasticPtsFile = open("elasticPts.dat","w")
    for i in xrange(nPoints):
        fullPtsFile.write("%15.12f %15.12f\n" %(fullPoints[i][0],fullPoints[i][1]))
        elasticPtsFile.write("%15.12f %15.12f\n" %(elasticPoints[i][0],elasticPoints[i][1]))


def TestCSi(nPoints,range,fileName):
    nodesToSet = GetNodesAffectingAtoms(model,coupling)
    print nodesToSet
    testFile = open(fileName,"w")
    atomisticGSE =  AtomisticEnergy(mainObjects,zeros(model.NumNodes()*3,Float64))
    nbrDisps0 = GetNbrDisps(mainObjects,zeros(model.NumNodes()*3,Float64),0)
    print 'GSE: ',atomisticGSE

    for i in xrange(nPoints):
        val = -range + i * (2. * range) / nPoints
        print val
        testDisps = CrackTestDisps(model, nodesToSet, val)
        partialEl = PartialElasticEnergy(mainObjects,testDisps)
        fullEl = FullElasticEnergy(mainObjects,testDisps)

        nbrDisps = GetNbrDisps(mainObjects,testDisps,0)

        dispDiffs = nbrDisps - nbrDisps0
        # print 'strains',dispDiffs[0][1]/nbrDisps0[0][1],dispDiffs[1][1]/nbrDisps0[1][1],dispDiffs[2][1]/nbrDisps0[2][1],dispDiffs[3][1]/nbrDisps0[3][1]
        atomicEl = AtomisticEnergy(mainObjects,testDisps) - atomisticGSE
        
        print "partial elastic: ",partialEl
        #print 'partial "elastic"',mainObjects.partialStiffnessMatrix[0][0] 
        print "full elastic: ",FullElasticEnergy(mainObjects,testDisps)
        print "Atomistic: ",atomicEl
        
        testFile.write("%10.8f %10.8f %10.8f %10.8f\n" %(val,partialEl,fullEl,atomicEl))
    
    testFile.close()
