(***********************************************************************

                    Mathematica-Compatible Notebook

This notebook can be used on any computer system with Mathematica 3.0,
MathReader 3.0, or any compatible application. The data for the notebook 
starts with the line of stars above.

To get the notebook into a Mathematica-compatible application, do one of 
the following:

* Save the data starting with the line of stars above into a file
  with a name ending in .nb, then open the file inside the application;

* Copy the data starting with the line of stars above to the
  clipboard, then use the Paste menu command inside the application.

Data for notebooks contains only printable 7-bit ASCII and can be
sent directly in email or through ftp in text mode.  Newlines can be
CR, LF or CRLF (Unix, Macintosh or MS-DOS style).

NOTE: If you modify the data for this notebook not in a Mathematica-
compatible application, you must delete the line below containing the 
word CacheID, otherwise Mathematica-compatible applications may try to 
use invalid cache data.

For more information on notebooks and Mathematica-compatible 
applications, contact Wolfram Research:
  web: http://www.wolfram.com
  email: info@wolfram.com
  phone: +1-217-398-0700 (U.S.)

Notebook reader applications are available free of charge from 
Wolfram Research.
***********************************************************************)

(*CacheID: 232*)


(*NotebookFileLineBreakTest
NotebookFileLineBreakTest*)
(*NotebookOptionsPosition[   1186325,      42775]*)
(*NotebookOutlinePosition[   1187099,      42802]*)
(*  CellTagsIndexPosition[   1187055,      42798]*)
(*WindowFrame->Normal*)



Notebook[{
Cell[BoxData[
    \(FindAngles[SizeLimit_] := 
      Module[\n\t\t{RationalAnglesList, AnglesAndRepeatsList, RotMatrix, i, 
          j, \[Theta], MillerInd, SurfaceVect, ReducedMillerInd, a, b, 
          RepeatDistance, maxi, maxj, temp}, \n\t\t\n\t\t
        RationalAnglesList = {}; \nMIAndRepeatsList = {}; \n
        RotMatrix\  = \ {{1, 2}, {\(-2\), \(-1\)}}; \n\t\t\n\t\t
        maxi\  = \ Floor[\((2/Sqrt[3])\) SizeLimit]; \n
        For[i = 1, i <= maxi, \(i++\), \n\t\t\t\n\t\t\t
            temp = N[Sqrt[4*SizeLimit*SizeLimit\  - \ 3*i*i]/2]; \n\t\t\t
            maxj = Floor[i/2 + temp]; \n\t\t
            For[j = \(-maxj\), j <= maxj, \(j++\), \n\t\t\t\t\n\t\t
              SurfaceVect\  = \ {i, j}; \n\t\t
              MillerInd = RotMatrix . SurfaceVect; \n
              ReducedMillerInd = 
                MillerInd/GCD[MillerInd[\([1]\)], MillerInd[\([2]\)]]; \n
              a = ReducedMillerInd[\([1]\)]; \nb = ReducedMillerInd[\([2]\)]; 
              \n\t\tIf[\((a + b/2)\)\  != 0, 
                \[Theta] = \ 
                  N[\((180/Pi)\)*ArcTan[\((Sqrt[3]*b/2)\)/\((a + b/2)\)]], 
                \[Theta] = 0]; \n\t\t\t
              RepeatDistance\  = \ N[Sqrt[i*i + j*j + i*j]]; \n\t\t\n\t\t\t\t
              If[Count[RationalAnglesList, \[Theta]] == 0\ \  && \ 
                  RepeatDistance\  < \ SizeLimit\  && \ \[Theta]\  >= 0\  && 
                  \ \[Theta] <= 60, \n\t\t\t
                RationalAnglesList = Append[RationalAnglesList, \[Theta]]; \n
                \t\t\tMIAndRepeatsList = 
                  Append[MIAndRepeatsList, {\[Theta], {a, b}, 
                      RepeatDistance}]; \n\t\t\t\t\n\t\t\t\t\t\t\t\t\t\t\t] (*
                end\ of\ If\ Count\  == \ 0*) \n\t\t\t] (*
              end\ of\ j\ For\ loop*) \n\t\t] (*end\ of\ i\ For\ loop*) \n
          RationalAnglesList; 
        \  (*without\ this\ line, \ get\ Null\ \(?? \(?? ?\)\)*) \n\t\t
        MIAndRepeatsList\n\t\t]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(X = FindAngles[10]\), 
    \(Length[X]\), 
    \(Angles = Sort[\(Transpose[X]\)[\([1]\)]]; \n
    AnglesAndRepeats = 
      Transpose[{\(Transpose[X]\)[\([1]\)], \t\t\(Transpose[X]\)[\([3]\)]}]; 
    \nListPlot[Angles, PlotLabel -> "\<Rational Angles\>", 
      AxesLabel -> {"\<\>", "\<Angle\>"}]; \n
    ListPlot[AnglesAndRepeats, \ 
      PlotLabel -> "\<Angle vs. Repeat Distance\>", \ 
      AxesLabel -> {"\<Angle\>", "\<Distance\>"}]; \)}], "Input"],

Cell[BoxData[
    \({{0, {\(-1\), 0}, 1.73205080756887745`}, {30.`, {\(-1\), \(-1\)}, 
        1.`}, {10.8933946491309052`, {\(-4\), \(-1\)}, 
        2.64575131106459027`}, {60.`, {0, \(-1\)}, 1.73205080756887745`}, {
        6.58677555362946165`, {\(-7\), \(-1\)}, 4.35889894354067308`}, {
        16.102113751986014`, {\(-5\), \(-2\)}, 3.60555127546398912`}, {
        49.1066053508690991`, {\(-1\), \(-4\)}, 2.64575131106459027`}, {
        4.71500395394821403`, {\(-10\), \(-1\)}, 6.08276253029821933`}, {
        19.1066053508690956`, {\(-2\), \(-1\)}, 4.58257569495583982`}, {
        43.8978862480139842`, {\(-2\), \(-5\)}, 3.60555127546398912`}, {
        3.67049650831510598`, {\(-13\), \(-1\)}, 7.81024967590665397`}, {
        8.21321070173818768`, {\(-11\), \(-2\)}, 7.`}, {
        13.8978862480139842`, {\(-3\), \(-1\)}, 6.2449979983983983`}, {
        21.0517244353729182`, {\(-7\), \(-4\)}, 5.56776436283002151`}, {
        40.8933946491309097`, {\(-1\), \(-2\)}, 4.58257569495583982`}, {
        53.413224446370533`, {\(-1\), \(-7\)}, 4.35889894354067308`}, {
        3.00449159888307892`, {\(-16\), \(-1\)}, 9.53939201416945614`}, {
        22.4109105310250544`, {\(-8\), \(-5\)}, 6.55743852430200035`}, {
        38.9482755646270817`, {\(-4\), \(-7\)}, 5.56776436283002151`}, {
        8.94827556462708351`, {\(-5\), \(-1\)}, 9.64365076099295492`}, {
        13.0039119428228167`, {\(-13\), \(-4\)}, 8.88819441731558868`}, {
        17.7836511602727327`, {\(-11\), \(-5\)}, 8.18535277187245036`}, {
        23.4132244463705374`, {\(-3\), \(-2\)}, 7.54983443527074982`}, {
        37.5890894689749455`, {\(-5\), \(-8\)}, 6.55743852430200035`}, {
        46.1021137519860246`, {\(-1\), \(-3\)}, 6.2449979983983983`}, {
        55.2849960460517841`, {\(-1\), \(-10\)}, 6.08276253029821933`}, {
        14.7046556985952729`, {\(-14\), \(-5\)}, 9.8488578017961057`}, {
        24.1824743555564181`, {\(-10\), \(-7\)}, 8.54400374531753037`}, {
        36.5867755536294669`, {\(-2\), \(-3\)}, 7.54983443527074982`}, {
        51.7867892982618105`, {\(-2\), \(-11\)}, 7.`}, {
        24.7912808971448894`, {\(-11\), \(-8\)}, 9.53939201416945614`}, {
        35.8175256444435774`, {\(-7\), \(-10\)}, 8.54400374531753037`}, {
        42.216348839727269`, {\(-5\), \(-11\)}, 8.18535277187245036`}, {
        56.329503491684898`, {\(-1\), \(-13\)}, 7.81024967590665397`}, {
        35.208719102855106`, {\(-8\), \(-11\)}, 9.53939201416945614`}, {
        46.9960880571771788`, {\(-4\), \(-13\)}, 8.88819441731558868`}, {
        45.2953443014047252`, {\(-5\), \(-14\)}, 9.8488578017961057`}, {
        51.0517244353729182`, {\(-1\), \(-5\)}, 9.64365076099295492`}, {
        56.9955084011169255`, {\(-1\), \(-16\)}, 9.53939201416945614`}}\)], 
  "Output"],

Cell[BoxData[
    \(39\)], "Output"],

Cell[CellGroupData[{

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.02442 0.0147151 0.00981006 [
[.26801 .00222 -6 -9 ]
[.26801 .00222 6 0 ]
[.51221 .00222 -6 -9 ]
[.51221 .00222 6 0 ]
[.75641 .00222 -6 -9 ]
[.75641 .00222 6 0 ]
[.01131 .11282 -12 -4.5 ]
[.01131 .11282 0 4.5 ]
[.01131 .21092 -12 -4.5 ]
[.01131 .21092 0 4.5 ]
[.01131 .30902 -12 -4.5 ]
[.01131 .30902 0 4.5 ]
[.01131 .40712 -12 -4.5 ]
[.01131 .40712 0 4.5 ]
[.01131 .50522 -12 -4.5 ]
[.01131 .50522 0 4.5 ]
[.01131 .60332 -12 -4.5 ]
[.01131 .60332 0 4.5 ]
[.02381 .64303 -17 0 ]
[.02381 .64303 17 9.8125 ]
[.5 .63053 -47 0 ]
[.5 .63053 47 10.125 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.26801 .01472 m
.26801 .02097 L
s
[(10)] .26801 .00222 0 1 Mshowa
.51221 .01472 m
.51221 .02097 L
s
[(20)] .51221 .00222 0 1 Mshowa
.75641 .01472 m
.75641 .02097 L
s
[(30)] .75641 .00222 0 1 Mshowa
.125 Mabswid
.07265 .01472 m
.07265 .01847 L
s
.12149 .01472 m
.12149 .01847 L
s
.17033 .01472 m
.17033 .01847 L
s
.21917 .01472 m
.21917 .01847 L
s
.31685 .01472 m
.31685 .01847 L
s
.36569 .01472 m
.36569 .01847 L
s
.41453 .01472 m
.41453 .01847 L
s
.46337 .01472 m
.46337 .01847 L
s
.56105 .01472 m
.56105 .01847 L
s
.60989 .01472 m
.60989 .01847 L
s
.65873 .01472 m
.65873 .01847 L
s
.70757 .01472 m
.70757 .01847 L
s
.80525 .01472 m
.80525 .01847 L
s
.85409 .01472 m
.85409 .01847 L
s
.90293 .01472 m
.90293 .01847 L
s
.95177 .01472 m
.95177 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .11282 m
.03006 .11282 L
s
[(10)] .01131 .11282 1 0 Mshowa
.02381 .21092 m
.03006 .21092 L
s
[(20)] .01131 .21092 1 0 Mshowa
.02381 .30902 m
.03006 .30902 L
s
[(30)] .01131 .30902 1 0 Mshowa
.02381 .40712 m
.03006 .40712 L
s
[(40)] .01131 .40712 1 0 Mshowa
.02381 .50522 m
.03006 .50522 L
s
[(50)] .01131 .50522 1 0 Mshowa
.02381 .60332 m
.03006 .60332 L
s
[(60)] .01131 .60332 1 0 Mshowa
.125 Mabswid
.02381 .03434 m
.02756 .03434 L
s
.02381 .05396 m
.02756 .05396 L
s
.02381 .07358 m
.02756 .07358 L
s
.02381 .0932 m
.02756 .0932 L
s
.02381 .13244 m
.02756 .13244 L
s
.02381 .15206 m
.02756 .15206 L
s
.02381 .17168 m
.02756 .17168 L
s
.02381 .1913 m
.02756 .1913 L
s
.02381 .23054 m
.02756 .23054 L
s
.02381 .25016 m
.02756 .25016 L
s
.02381 .26978 m
.02756 .26978 L
s
.02381 .2894 m
.02756 .2894 L
s
.02381 .32864 m
.02756 .32864 L
s
.02381 .34826 m
.02756 .34826 L
s
.02381 .36788 m
.02756 .36788 L
s
.02381 .3875 m
.02756 .3875 L
s
.02381 .42674 m
.02756 .42674 L
s
.02381 .44636 m
.02756 .44636 L
s
.02381 .46598 m
.02756 .46598 L
s
.02381 .4856 m
.02756 .4856 L
s
.02381 .52484 m
.02756 .52484 L
s
.02381 .54446 m
.02756 .54446 L
s
.02381 .56408 m
.02756 .56408 L
s
.02381 .5837 m
.02756 .5837 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
gsave
.02381 .64303 -78 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Angle) show
93.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
gsave
.5 .63053 -108 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.125000 moveto
 558.000000 14.125000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.125000 lineto
closepath
clip newpath
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Rational) show
117.000000 11.562500 moveto
(Angles) show
153.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.008 w
.04823 .01472 Mdot
.07265 .04419 Mdot
.09707 .05072 Mdot
.12149 .06097 Mdot
.14591 .07933 Mdot
.17033 .09529 Mdot
.19475 .1025 Mdot
.21917 .12158 Mdot
.24359 .14228 Mdot
.26801 .15105 Mdot
.29243 .15897 Mdot
.31685 .17268 Mdot
.34127 .18917 Mdot
.36569 .20215 Mdot
.39011 .22123 Mdot
.41453 .23457 Mdot
.43895 .2444 Mdot
.46337 .25195 Mdot
.48779 .25792 Mdot
.51221 .30902 Mdot
.53663 .36011 Mdot
.56105 .36609 Mdot
.58547 .37363 Mdot
.60989 .38347 Mdot
.63431 .3968 Mdot
.65873 .41588 Mdot
.68315 .42886 Mdot
.70757 .44536 Mdot
.73199 .45907 Mdot
.75641 .46698 Mdot
.78083 .47575 Mdot
.80525 .49645 Mdot
.82967 .51554 Mdot
.85409 .52275 Mdot
.87851 .5387 Mdot
.90293 .55706 Mdot
.92735 .56731 Mdot
.95177 .57384 Mdot
.97619 .60332 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/B000`400?l00000o`00003oo`3ooolQ
0?ooo`00o`3ooolQ0?ooo`00o`3ooolQ0?ooo`00EP3oool4000000<0oooo0P00000d0?ooo`@00000
0`3oool2000003@0oooo0P0000040?ooo`800000B03oool005P0oooo00D000000?ooo`3oool0oooo
000000020?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool00`3oool010000000oooo
0?ooo`00000b0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool00000A`3oool0
05P0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003oool0oooo0380oooo00<00000
0?ooo`3oool00P3oool010000000oooo0?ooo`00000e0?ooo`040000003oool0oooo00000080oooo
00<000000?ooo`3oool0A@3oool005P0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000
003oool0oooo03<0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003oool0oooo0340
oooo0P0000030?ooo`040000003oool0oooo000004L0oooo001H0?ooo`050000003oool0oooo0?oo
o`0000000P3oool00`000000oooo0?ooo`0d0?ooo`040000003oool0oooo00000080oooo00<00000
0?ooo`3oool0<`3oool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo04D0oooo000O
0?ooo`030000003oool0oooo03@0oooo0`0000030?ooo`040000003oool0oooo000003<0oooo00@0
00000?ooo`3oool000000P3oool010000000oooo0?ooo`00000b0?ooo`040000003oool0oooo0000
0080oooo00@000000?ooo`3oool00000A`3oool001l0oooo00<000000?ooo`3oool0=P3oool00`00
0000oooo0?ooo`020?ooo`800000=@3oool2000000@0oooo0P00000d0?ooo`800000103oool20000
04P0oooo000O0?ooo`030000003oool0oooo0?h0oooo000O0?ooo`030000003oool0oooo0?h0oooo
000I0?ooooP000003`3oool001l0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0I
0?ooo`007`3oool00`000000oooo0?ooo`3n0?ooo`007`3oool00`000000oooo0?ooo`3n0?ooo`00
7`3oool200000?l0oooo000O0?ooo`030000003oool0oooo0?h0oooo000O0?ooo`030000003oool0
oooo0?h0oooo000O0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0lP3oool001l0
oooo00<000000?ooo`3oool0oP3oool001l0oooo0P00000@0?ooo`030000003oool0oooo0>`0oooo
000O0?ooo`030000003oool0oooo0?h0oooo000O0?ooo`030000003oool0oooo01D0oooo00<00000
0?ooo`3oool0iP3oool001l0oooo00<000000?ooo`3oool0oP3oool001l0oooo00<000000?ooo`3o
ool0oP3oool001l0oooo0P00003o0?ooo`007`3oool00`000000oooo0?ooo`3n0?ooo`007`3oool0
0`000000oooo0?ooo`0K0?ooo`030000003oool0oooo0>00oooo000O0?ooo`030000003oool0oooo
0?h0oooo000O0?ooo`030000003oool0oooo0?h0oooo000O0?ooo`800000o`3oool001l0oooo00<0
00000?ooo`3oool08@3oool00`000000oooo0?ooo`3J0?ooo`007`3oool00`000000oooo0?ooo`3n
0?ooo`007`3oool00`000000oooo0?ooo`0W0?ooo`030000003oool0oooo0=@0oooo000A0?ooo`@0
00000`3oool2000000D0oooo00<000000?ooo`3oool0oP3oool001<0oooo00D000000?ooo`3oool0
oooo000000020?ooo`030000003oool0oooo0080oooo0`00003n0?ooo`004`3oool01@000000oooo
0?ooo`3oool000000080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`3n0?ooo`00
4`3oool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`0]0?ooo`030000003oool0oooo0<h0oooo000C0?ooo`050000003oool0oooo0?ooo`00
00000P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0?h0oooo000A0?ooo`<00000
0`3oool010000000oooo0?ooo`0000040?ooo`030000003oool0oooo0?h0oooo000C0?ooo`030000
003oool0oooo0080oooo0P0000050?ooo`800000o`3oool001l0oooo00<000000?ooo`3oool0oP3o
ool001l0oooo00<000000?ooo`3oool0oP3oool001l0oooo00<000000?ooo`3oool0<`3oool00`00
0000oooo0?ooo`380?ooo`007`3oool00`000000oooo0?ooo`3n0?ooo`007`3oool2000003X0oooo
00<000000?ooo`3oool0`P3oool001l0oooo00<000000?ooo`3oool0oP3oool001l0oooo00<00000
0?ooo`3oool0?`3oool00`000000oooo0?ooo`2l0?ooo`007`3oool00`000000oooo0?ooo`3n0?oo
o`007`3oool200000?l0oooo000O0?ooo`030000003oool0oooo04D0oooo00<000000?ooo`3oool0
]P3oool001l0oooo00<000000?ooo`3oool0oP3oool001l0oooo00<000000?ooo`3oool0oP3oool0
01l0oooo00<000000?ooo`3oool0oP3oool001l0oooo0P00001<0?ooo`030000003oool0oooo0;00
oooo000O0?ooo`030000003oool0oooo0?h0oooo000O0?ooo`030000003oool0oooo0?h0oooo000O
0?ooo`030000003oool0oooo0540oooo00<000000?ooo`3oool0ZP3oool00140oooo100000030?oo
o`8000001@3oool00`000000oooo0?ooo`3n0?ooo`004@3oool00`000000oooo0?ooo`030?ooo`04
0000003oool0oooo000000@0oooo0`00003n0?ooo`004P3oool00`000000oooo0?ooo`020?ooo`04
0000003oool0oooo000000@0oooo00<000000?ooo`3oool0oP3oool001<0oooo00D000000?ooo`3o
ool0oooo000000020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0oP3oool001@0
oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
05L0oooo00<000000?ooo`3oool0Y03oool00140oooo00@000000?ooo`3oool000000P3oool01000
0000oooo0?ooo`0000040?ooo`030000003oool0oooo0?h0oooo000B0?ooo`800000103oool20000
00D0oooo0P00003o0?ooo`007`3oool00`000000oooo0?ooo`1M0?ooo`030000003oool0oooo09h0
oooo000O0?ooo`030000003oool0oooo0?h0oooo000O0?ooo`030000003oool0oooo0?h0oooo000O
0?ooo`030000003oool0oooo06<0oooo00<000000?ooo`3oool0V03oool001l0oooo0P00003o0?oo
o`007`3oool00`000000oooo0?ooo`1Y0?ooo`030000003oool0oooo0980oooo000O0?ooo`030000
003oool0oooo06l0oooo00<000000?ooo`3oool0S03oool001l0oooo00<000000?ooo`3oool0oP3o
ool001l0oooo00<000000?ooo`3oool0oP3oool001l0oooo0P00003o0?ooo`007`3oool00`000000
oooo0?ooo`3n0?ooo`007`3oool00`000000oooo0?ooo`3n0?ooo`007`3oool00`000000oooo0?oo
o`3n0?ooo`007`3oool200000?l0oooo000O0?ooo`030000003oool0oooo0?h0oooo000O0?ooo`03
0000003oool0oooo0?h0oooo000O0?ooo`030000003oool0oooo0?h0oooo000B0?ooo`800000103o
ool2000000D0oooo00<000000?ooo`3oool0oP3oool00140oooo00@000000?ooo`3oool000000P3o
ool010000000oooo0?ooo`0000040?ooo`<00000oP3oool001@0oooo00@000000?ooo`3oool00000
0P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo07D0oooo00<000000?ooo`3oool0
QP3oool00180oooo0P0000030?ooo`040000003oool0oooo000000@0oooo00<000000?ooo`3oool0
oP3oool001@0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo0?h0oooo000A0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool0
0000103oool00`000000oooo0?ooo`3n0?ooo`004P3oool2000000@0oooo0P0000050?ooo`800000
o`3oool001l0oooo00<000000?ooo`3oool0oP3oool001l0oooo00<000000?ooo`3oool0oP3oool0
01l0oooo00<000000?ooo`3oool0oP3oool001l0oooo00<000000?ooo`3oool0oP3oool001l0oooo
0P00003o0?ooo`007`3oool00`000000oooo0?ooo`3n0?ooo`007`3oool00`000000oooo0?ooo`3n
0?ooo`007`3oool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo0800oooo000O0?ooo`03
0000003oool0oooo0?h0oooo000O0?ooo`800000PP3oool00`000000oooo0?ooo`1j0?ooo`007`3o
ool00`000000oooo0?ooo`3n0?ooo`007`3oool00`000000oooo0?ooo`270?ooo`030000003oool0
oooo07@0oooo000O0?ooo`030000003oool0oooo0?h0oooo000O0?ooo`030000003oool0oooo08d0
oooo00<000000?ooo`3oool0KP3oool001l0oooo0P00003o0?ooo`007`3oool00`000000oooo0?oo
o`3n0?ooo`007`3oool00`000000oooo0?ooo`2C0?ooo`030000003oool0oooo06P0oooo000C0?oo
o`030000003oool0oooo0080oooo0P0000050?ooo`030000003oool0oooo0?h0oooo000C0?ooo`05
0000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`020?ooo`<00000oP3oool00100
oooo1@0000020?ooo`040000003oool0oooo000000@0oooo00<000000?ooo`3oool0oP3oool00100
oooo00@000000?ooo`3oool000000`3oool010000000oooo0?ooo`0000040?ooo`030000003oool0
oooo0?h0oooo000A0?ooo`030000003oool0000000<0oooo00@000000?ooo`3oool00000103oool0
0`000000oooo0?ooo`2I0?ooo`030000003oool0oooo0680oooo000B0?ooo`8000000`3oool01000
0000oooo0?ooo`0000040?ooo`030000003oool0oooo0?h0oooo000C0?ooo`030000003oool0oooo
0080oooo0P0000050?ooo`800000o`3oool001l0oooo00<000000?ooo`3oool0W`3oool00`000000
oooo0?ooo`1L0?ooo`007`3oool00`000000oooo0?ooo`3n0?ooo`007`3oool00`000000oooo0?oo
o`3n0?ooo`007`3oool00`000000oooo0?ooo`3n0?ooo`007`3oool200000:H0oooo00<000000?oo
o`3oool0EP3oool001l0oooo00<000000?ooo`3oool0oP3oool001l0oooo00<000000?ooo`3oool0
oP3oool001l0oooo00<000000?ooo`3oool0oP3oool001l0oooo00<000000?ooo`3oool0Z`3oool0
0`000000oooo0?ooo`1@0?ooo`007`3oool200000?l0oooo000O0?ooo`030000003oool0oooo0;40
oooo00<000000?ooo`3oool0BP3oool001l0oooo00<000000?ooo`3oool0oP3oool001l0oooo00<0
00000?ooo`3oool0]`3oool00`000000oooo0?ooo`140?ooo`007`3oool00`000000oooo0?ooo`3n
0?ooo`007`3oool200000?l0oooo000O0?ooo`030000003oool0oooo0?h0oooo000O0?ooo`030000
003oool0oooo0?h0oooo000A0?ooo`<00000103oool2000000D0oooo00<000000?ooo`3oool0_@3o
ool00`000000oooo0?ooo`0n0?ooo`00503oool010000000oooo0?ooo`0000020?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool0oP3oool001@0oooo00@000000?ooo`3oool000000P3o
ool00`000000oooo0?ooo`020?ooo`<00000oP3oool00140oooo0`0000030?ooo`040000003oool0
oooo000000@0oooo00<000000?ooo`3oool0oP3oool00140oooo00<000000?ooo`3oool00`3oool0
10000000oooo0?ooo`0000040?ooo`030000003oool0oooo0?h0oooo000A0?ooo`030000003oool0
oooo00<0oooo00@000000?ooo`3oool00000103oool00`000000oooo0?ooo`330?ooo`030000003o
ool0oooo03P0oooo000A0?ooo`@000000`3oool2000000D0oooo00<000000?ooo`3oool0bP3oool0
0`000000oooo0?ooo`0a0?ooo`007`3oool200000?l0oooo000O0?ooo`030000003oool0oooo0?h0
oooo000O0?ooo`030000003oool0oooo0?h0oooo000O0?ooo`030000003oool0oooo0=00oooo00<0
00000?ooo`3oool0:`3oool001l0oooo0P00003o0?ooo`007`3oool00`000000oooo0?ooo`3n0?oo
o`007`3oool00`000000oooo0?ooo`3n0?ooo`007`3oool00`000000oooo0?ooo`3n0?ooo`007`3o
ool00`000000oooo0?ooo`3F0?ooo`030000003oool0oooo02D0oooo000O0?ooo`800000o`3oool0
01l0oooo00<000000?ooo`3oool0g03oool00`000000oooo0?ooo`0O0?ooo`007`3oool00`000000
oooo0?ooo`3n0?ooo`007`3oool00`000000oooo0?ooo`3R0?ooo`030000003oool0oooo01T0oooo
000O0?ooo`030000003oool0oooo0?h0oooo000O0?ooo`800000o`3oool001l0oooo00<000000?oo
o`3oool0oP3oool001l0oooo00<000000?ooo`3oool0oP3oool001l0oooo00<000000?ooo`3oool0
oP3oool00180oooo0P0000040?ooo`8000001@3oool00`000000oooo0?ooo`3n0?ooo`004@3oool0
10000000oooo0?ooo`0000020?ooo`040000003oool0oooo000000@0oooo0`00003X0?ooo`030000
003oool0oooo01<0oooo000A0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool0
0000103oool00`000000oooo0?ooo`3n0?ooo`004@3oool3000000<0oooo00@000000?ooo`3oool0
0000103oool00`000000oooo0?ooo`3n0?ooo`004@3oool00`000000oooo0?ooo`030?ooo`040000
003oool0oooo000000@0oooo00<000000?ooo`3oool0oP3oool00140oooo00<000000?ooo`3oool0
0`3oool010000000oooo0?ooo`0000040?ooo`030000003oool0oooo0?h0oooo000B0?ooo`<00000
0`3oool200000?l0oooo1`3oool00?l0oooo8@3oool00?l0oooo8@3oool00?l0oooo8@3oool00?l0
oooo8@3oool00:/0oooo0`00001b0?ooo`00[P3oool00`000000oooo0?ooo`1_0?ooo`00J03oool3
00000080oooo00<000000?ooo`3oool00P0000000`3oool000000?ooo`020?ooo`8000000P3oool5
00000080oooo0P0000020?ooo`<0000000<0oooo0000000000000P3oool2000000030?ooo`000000
oooo00D000001P3oool200000080oooo1@0000000`3oool00000000000020?ooo`<000000P3oool5
00000080oooo0`0000020?ooo`<00000H03oool001d0oooo0`0000190?ooo`040000003oool0oooo
00000080oooo00@000000?ooo`3oool000000`3oool010000000oooo0?ooo`0000030?ooo`050000
003oool0oooo0?ooo`0000000P3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo
00000080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`070?ooo`040000003oool0
oooo00000080oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`0000040?ooo`05
0000003oool0oooo0?ooo`000000203oool00`000000oooo0?ooo`1M0?ooo`00803oool00`000000
oooo0?ooo`160?ooo`<00000103oool3000000<0oooo00<000000?ooo`3oool0103oool01@000000
oooo0?ooo`3oool000000080oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`3o
ool3000000@0oooo00<000000?ooo`3oool01`3oool400000080oooo00@000000?ooo`3oool00000
0P3oool010000000oooo0?ooo`0000040?ooo`040000003oool0oooo0?ooo`<00000103oool20000
0600oooo000@0?ooo`8000000P3oool5000000030?ooo`00000000000080oooo0`0000020?ooo`D0
00000P3oool3000003`0oooo00@000000?ooo`3oool000001@3oool01@000000oooo0?ooo`3oool0
000000H0oooo00D000000?ooo`3oool0oooo000000020?ooo`040000003oool0oooo00000080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00L0oooo
00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo
000000@0oooo00D000000?ooo`3oool0oooo000000020?ooo`040000003oool0oooo00000680oooo
000A0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool000000P3oool010000000
oooo0?ooo`0000040?ooo`050000003oool0oooo0?ooo`000000?`3oool010000000oooo0?ooo`00
00030?ooo`8000000`3oool4000000<0oooo0P0000040?ooo`8000000P3oool010000000oooo0000
000000040?ooo`8000001@3oool00`000000oooo0?ooo`080?ooo`060000003oool000000?ooo`00
0000oooo0P0000040?ooo`80000000<0oooo0000003oool00P3oool00`000000oooo0?ooo`020?oo
o`800000103oool3000005l0oooo000A0?ooo`@000000P3oool010000000oooo0?ooo`0000020?oo
o`040000003oool0oooo000000@0oooo00@000000?ooo`3oool0oooo0`00000l0?ooo`@000002P3o
ool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo00L0oooo0`00000A0?ooo`030000003o
ool0oooo06X0oooo000A0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool00000
0P3oool010000000oooo0?ooo`0000040?ooo`050000003oool0oooo0?ooo`0000000P3oool00`00
0000oooo0?ooo`170?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0503oool20000
01`0oooo0P00001/0?ooo`004P3oool01P000000oooo0000003oool000000?ooo`800000103oool2
000000030?ooo`000000oooo0080oooo00<000000?ooo`3oool00P3oool200000?@0oooo0000\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-5.23746, -4.99198, 
  0.165954, 0.413106}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.015873 0.0147151 0.0597637 [
[.18254 .00222 -6 -9 ]
[.18254 .00222 6 0 ]
[.34127 .00222 -6 -9 ]
[.34127 .00222 6 0 ]
[.5 .00222 -6 -9 ]
[.5 .00222 6 0 ]
[.65873 .00222 -6 -9 ]
[.65873 .00222 6 0 ]
[.81746 .00222 -6 -9 ]
[.81746 .00222 6 0 ]
[.97619 .00222 -6 -9 ]
[.97619 .00222 6 0 ]
[1.025 .01472 0 -4.90625 ]
[1.025 .01472 34 4.90625 ]
[.01131 .13424 -6 -4.5 ]
[.01131 .13424 0 4.5 ]
[.01131 .25377 -6 -4.5 ]
[.01131 .25377 0 4.5 ]
[.01131 .3733 -6 -4.5 ]
[.01131 .3733 0 4.5 ]
[.01131 .49282 -6 -4.5 ]
[.01131 .49282 0 4.5 ]
[.01131 .61235 -12 -4.5 ]
[.01131 .61235 0 4.5 ]
[.02381 .64303 -26 0 ]
[.02381 .64303 26 10.125 ]
[.5 .63053 -77 0 ]
[.5 .63053 77 10.125 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.18254 .01472 m
.18254 .02097 L
s
[(10)] .18254 .00222 0 1 Mshowa
.34127 .01472 m
.34127 .02097 L
s
[(20)] .34127 .00222 0 1 Mshowa
.5 .01472 m
.5 .02097 L
s
[(30)] .5 .00222 0 1 Mshowa
.65873 .01472 m
.65873 .02097 L
s
[(40)] .65873 .00222 0 1 Mshowa
.81746 .01472 m
.81746 .02097 L
s
[(50)] .81746 .00222 0 1 Mshowa
.97619 .01472 m
.97619 .02097 L
s
[(60)] .97619 .00222 0 1 Mshowa
.125 Mabswid
.05556 .01472 m
.05556 .01847 L
s
.0873 .01472 m
.0873 .01847 L
s
.11905 .01472 m
.11905 .01847 L
s
.15079 .01472 m
.15079 .01847 L
s
.21429 .01472 m
.21429 .01847 L
s
.24603 .01472 m
.24603 .01847 L
s
.27778 .01472 m
.27778 .01847 L
s
.30952 .01472 m
.30952 .01847 L
s
.37302 .01472 m
.37302 .01847 L
s
.40476 .01472 m
.40476 .01847 L
s
.43651 .01472 m
.43651 .01847 L
s
.46825 .01472 m
.46825 .01847 L
s
.53175 .01472 m
.53175 .01847 L
s
.56349 .01472 m
.56349 .01847 L
s
.59524 .01472 m
.59524 .01847 L
s
.62698 .01472 m
.62698 .01847 L
s
.69048 .01472 m
.69048 .01847 L
s
.72222 .01472 m
.72222 .01847 L
s
.75397 .01472 m
.75397 .01847 L
s
.78571 .01472 m
.78571 .01847 L
s
.84921 .01472 m
.84921 .01847 L
s
.88095 .01472 m
.88095 .01847 L
s
.9127 .01472 m
.9127 .01847 L
s
.94444 .01472 m
.94444 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
gsave
1.025 .01472 -61 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Angle) show
93.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.02381 .13424 m
.03006 .13424 L
s
[(2)] .01131 .13424 1 0 Mshowa
.02381 .25377 m
.03006 .25377 L
s
[(4)] .01131 .25377 1 0 Mshowa
.02381 .3733 m
.03006 .3733 L
s
[(6)] .01131 .3733 1 0 Mshowa
.02381 .49282 m
.03006 .49282 L
s
[(8)] .01131 .49282 1 0 Mshowa
.02381 .61235 m
.03006 .61235 L
s
[(10)] .01131 .61235 1 0 Mshowa
.125 Mabswid
.02381 .0446 m
.02756 .0446 L
s
.02381 .07448 m
.02756 .07448 L
s
.02381 .10436 m
.02756 .10436 L
s
.02381 .16412 m
.02756 .16412 L
s
.02381 .19401 m
.02756 .19401 L
s
.02381 .22389 m
.02756 .22389 L
s
.02381 .28365 m
.02756 .28365 L
s
.02381 .31353 m
.02756 .31353 L
s
.02381 .34342 m
.02756 .34342 L
s
.02381 .40318 m
.02756 .40318 L
s
.02381 .43306 m
.02756 .43306 L
s
.02381 .46294 m
.02756 .46294 L
s
.02381 .52271 m
.02756 .52271 L
s
.02381 .55259 m
.02756 .55259 L
s
.02381 .58247 m
.02756 .58247 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
gsave
.02381 .64303 -87 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.125000 moveto
 558.000000 14.125000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.125000 lineto
closepath
clip newpath
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Distance) show
111.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
gsave
.5 .63053 -138 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.125000 moveto
 558.000000 14.125000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.125000 lineto
closepath
clip newpath
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Angle) show
99.000000 11.562500 moveto
(vs) show
111.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(.) show
123.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Repeat) show
165.000000 11.562500 moveto
(Distance) show
213.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.008 w
.02381 .11823 Mdot
.5 .07448 Mdot
.19672 .17283 Mdot
.97619 .11823 Mdot
.12836 .27522 Mdot
.2794 .2302 Mdot
.80328 .17283 Mdot
.09865 .37824 Mdot
.32709 .28859 Mdot
.7206 .2302 Mdot
.08207 .48148 Mdot
.15418 .43306 Mdot
.24441 .38794 Mdot
.35796 .34747 Mdot
.67291 .28859 Mdot
.87164 .27522 Mdot
.0715 .58482 Mdot
.37954 .40661 Mdot
.64204 .34747 Mdot
.16585 .59105 Mdot
.23022 .54591 Mdot
.30609 .5039 Mdot
.39545 .46592 Mdot
.62046 .40661 Mdot
.75559 .38794 Mdot
.90135 .37824 Mdot
.25722 .60332 Mdot
.40766 .52534 Mdot
.60455 .46592 Mdot
.84582 .43306 Mdot
.41732 .58482 Mdot
.59234 .52534 Mdot
.69391 .5039 Mdot
.91793 .48148 Mdot
.58268 .58482 Mdot
.76978 .54591 Mdot
.74278 .60332 Mdot
.83415 .59105 Mdot
.9285 .58482 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 177.938},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000/B000`400?l00000o`00003oo`3ooolQ
0?ooo`00o`3ooolQ0?ooo`00o`3ooolQ0?ooo`00o`3ooolQ0?ooo`00o`3ooolQ0?ooo`00o`3ooolQ
0?ooo`00o`3ooolQ0?ooo`00o`3ooolQ0?ooo`00o`3ooolQ0?ooo`00o`3ooolQ0?ooo`00>03oool4
000000<0oooo0P00000K0?ooo`@000000`3oool2000001`0oooo0P0000040?ooo`8000007@3oool0
0`000000oooo0?ooo`020?ooo`8000006`3oool3000000@0oooo0P00000L0?ooo`800000103oool2
000002/0oooo000j0?ooo`050000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`0H
0?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool000006P3oool010000000oooo0?oo
o`0000020?ooo`040000003oool0oooo000001`0oooo00D000000?ooo`3oool0oooo000000020?oo
o`030000003oool0oooo01/0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`0H
0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool00000:P3oool003X0oooo00D0
00000?ooo`3oool0oooo000000020?ooo`030000003oool0oooo01T0oooo00<000000?ooo`3oool0
0P3oool010000000oooo0?ooo`00000M0?ooo`040000003oool0oooo00000080oooo00<000000?oo
o`3oool05`3oool500000080oooo00@000000?ooo`3oool000007@3oool010000000oooo0?ooo`00
00020?ooo`030000003oool0oooo01P0oooo00@000000?ooo`3oool000000P3oool010000000oooo
0?ooo`00000Z0?ooo`00>P3oool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`3o
ool06P3oool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`3oool06@3oool20000
00<0oooo00@000000?ooo`3oool000006@3oool010000000oooo0?ooo`0000030?ooo`040000003o
ool0oooo000001X0oooo0`0000030?ooo`040000003oool0oooo000001X0oooo0`0000030?ooo`04
0000003oool0oooo000002X0oooo000J0?ooo`030000003oool0oooo01d0oooo00D000000?ooo`3o
ool0oooo000000020?ooo`030000003oool0oooo01/0oooo00@000000?ooo`3oool000000P3oool0
0`000000oooo0?ooo`0K0?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3oool0603o
ool00`000000oooo000000030?ooo`040000003oool0oooo000001X0oooo00<000000?ooo`3oool0
0`3oool010000000oooo0?ooo`00000J0?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3o
ool00000:P3oool001X0oooo00<000000?ooo`3oool06`3oool3000000<0oooo00@000000?ooo`3o
ool000006P3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000001X0oooo00@0
00000?ooo`3oool000000P3oool010000000oooo0?ooo`00000K0?ooo`8000000`3oool010000000
oooo0?ooo`00000J0?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool000006P3oool0
0`000000oooo0?ooo`030?ooo`040000003oool0oooo000002X0oooo000J0?ooo`030000003oool0
oooo01d0oooo00<000000?ooo`3oool00P3oool2000001`0oooo0P0000040?ooo`800000703oool2
000000@0oooo0P00000M0?ooo`030000003oool0oooo0080oooo0P00000K0?ooo`@000000`3oool2
000001`0oooo0`0000030?ooo`8000005`3oool300000140oooo000J0?ooo`030000003oool0oooo
0?80oooo00<000000?ooo`3oool03P3oool001@0ooooi00000070?ooo`8000000P3oool500000003
0?ooo`00000000000080oooo0`0000020?ooo`D000000P3oool3000000@0oooo000J0?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo00D0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`050?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo00`0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`0000020?oo
o`040000003oool0oooo000000@0oooo00D000000?ooo`3oool0oooo000000070?ooo`006P3oool0
0`000000oooo0?ooo`3S0?ooo`@000000P3oool010000000oooo0?ooo`0000020?ooo`040000003o
ool0oooo000000@0oooo00@000000?ooo`3oool0oooo0`0000050?ooo`006P3oool00`000000oooo
0?ooo`3S0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool000000P3oool01000
0000oooo0?ooo`0000040?ooo`050000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?oo
o`020?ooo`006P3oool00`000000oooo0?ooo`3T0?ooo`060000003oool000000?ooo`000000oooo
0P0000040?ooo`80000000<0oooo0000003oool00P3oool00`000000oooo0?ooo`020?ooo`800000
1@3oool001X0oooo00<000000?ooo`3oool0h`3oool300000140oooo00<000000?ooo`3oool02@3o
ool001X0oooo00<000000?ooo`3oool0mP3oool2000000/0oooo000J0?ooo`030000003oool0oooo
0?l0oooo103oool001X0oooo00<000000?ooo`3oool0o`3oool40?ooo`006P3oool00`000000oooo
0?ooo`3o0?ooo`@0oooo000J0?ooo`030000003oool0oooo0?l0oooo103oool001X0oooo00<00000
0?ooo`3oool0o`3oool40?ooo`006P3oool00`000000oooo0?ooo`3o0?ooo`@0oooo000J0?ooo`03
0000003oool0oooo0?l0oooo103oool001X0oooo00<000000?ooo`3oool0J03oool00`000000oooo
0?ooo`2H0?ooo`006P3oool00`000000oooo0?ooo`3o0?ooo`@0oooo000J0?ooo`030000003oool0
oooo0?l0oooo103oool001X0oooo00<000000?ooo`3oool0o`3oool40?ooo`006P3oool00`000000
oooo0?ooo`3o0?ooo`@0oooo000J0?ooo`030000003oool0oooo0?l0oooo103oool001X0oooo00<0
00000?ooo`3oool0o`3oool40?ooo`006P3oool00`000000oooo0?ooo`3o0?ooo`@0oooo000J0?oo
o`030000003oool0oooo0?l0oooo103oool001T0oooo0P00003F0?ooo`030000003oool0oooo02`0
oooo000J0?ooo`030000003oool0oooo0?l0oooo103oool00180oooo100000040?ooo`030000003o
ool0oooo0?l0oooo103oool00180oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`3o
0?ooo`@0oooo000C0?ooo`030000003oool0oooo00@0oooo0P00003o0?ooo`D0oooo000D0?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool0o`3oool40?ooo`005@3oool00`000000oooo
0?ooo`020?ooo`030000003oool0oooo0?l0oooo103oool00180oooo00@000000?ooo`3oool00000
103oool00`000000oooo0?ooo`3o0?ooo`@0oooo000C0?ooo`8000001@3oool00`000000oooo0?oo
o`3o0?ooo`@0oooo000J0?ooo`030000003oool0oooo0?l0oooo103oool001X0oooo00<000000?oo
o`3oool0o`3oool40?ooo`006P3oool00`000000oooo0?ooo`3o0?ooo`@0oooo000J0?ooo`030000
003oool0oooo0?l0oooo103oool001X0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?oo
o`270?ooo`030000003oool0oooo05<0oooo000J0?ooo`030000003oool0oooo0?l0oooo103oool0
01X0oooo00<000000?ooo`3oool0o`3oool40?ooo`006P3oool00`000000oooo0?ooo`3o0?ooo`@0
oooo000J0?ooo`030000003oool0oooo0?l0oooo103oool001X0oooo00<000000?ooo`3oool0o`3o
ool40?ooo`006P3oool00`000000oooo0?ooo`3o0?ooo`@0oooo000J0?ooo`030000003oool0oooo
0?l0oooo103oool001X0oooo00<000000?ooo`3oool0o`3oool40?ooo`006P3oool00`000000oooo
0?ooo`3o0?ooo`@0oooo000J0?ooo`030000003oool0oooo0?l0oooo103oool001X0oooo00<00000
0?ooo`3oool0o`3oool40?ooo`006P3oool00`000000oooo0?ooo`3o0?ooo`@0oooo000J0?ooo`03
0000003oool0oooo03H0oooo00<000000?ooo`3oool0H@3oool00`000000oooo0?ooo`1V0?ooo`00
6P3oool00`000000oooo0?ooo`3o0?ooo`@0oooo000J0?ooo`030000003oool0oooo0?l0oooo103o
ool001X0oooo00<000000?ooo`3oool0o`3oool40?ooo`00503oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo0?l0oooo103oool001@0oooo00<000000?ooo`3oool00`3oool200000?l0
oooo1@3oool00140oooo1@0000040?ooo`030000003oool0oooo0?l0oooo103oool00140oooo00@0
00000?ooo`3oool000001@3oool00`000000oooo0?ooo`3o0?ooo`@0oooo000B0?ooo`030000003o
ool0000000D0oooo00<000000?ooo`3oool0o`3oool40?ooo`004`3oool2000000D0oooo00<00000
0?ooo`3oool0o`3oool40?ooo`00503oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
01@0oooo00<000000?ooo`3oool0Y@3oool00`000000oooo0?ooo`140?ooo`006P3oool00`000000
oooo0?ooo`3o0?ooo`@0oooo000J0?ooo`030000003oool0oooo0?l0oooo103oool001X0oooo00<0
00000?ooo`3oool0@@3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo0740oooo000J
0?ooo`030000003oool0oooo0?l0oooo103oool001X0oooo00<000000?ooo`3oool0o`3oool40?oo
o`006P3oool00`000000oooo0?ooo`3o0?ooo`@0oooo000J0?ooo`030000003oool0oooo0?l0oooo
103oool001X0oooo00<000000?ooo`3oool0o`3oool40?ooo`006P3oool00`000000oooo0?ooo`3o
0?ooo`@0oooo000J0?ooo`030000003oool0oooo0?l0oooo103oool001X0oooo00<000000?ooo`3o
ool0o`3oool40?ooo`006P3oool00`000000oooo0?ooo`3o0?ooo`@0oooo000J0?ooo`030000003o
ool0oooo0?l0oooo103oool001X0oooo00<000000?ooo`3oool0o`3oool40?ooo`006P3oool00`00
0000oooo0?ooo`3o0?ooo`@0oooo000J0?ooo`030000003oool0oooo04P0oooo00<000000?ooo`3o
ool0?@3oool00`000000oooo0?ooo`1h0?ooo`006P3oool00`000000oooo0?ooo`3o0?ooo`@0oooo
000J0?ooo`030000003oool0oooo0?l0oooo103oool001X0oooo00<000000?ooo`3oool0o`3oool4
0?ooo`006P3oool00`000000oooo0?ooo`3o0?ooo`@0oooo000C0?ooo`8000001@3oool00`000000
oooo0?ooo`3o0?ooo`@0oooo000B0?ooo`040000003oool0oooo000000@0oooo0P00003o0?ooo`D0
oooo000B0?ooo`040000003oool0oooo000000@0oooo00<000000?ooo`3oool03@3oool00`000000
oooo0?ooo`2c0?ooo`030000003oool0oooo03d0oooo000B0?ooo`<000001@3oool00`000000oooo
0?ooo`3o0?ooo`@0oooo000B0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0o`3o
ool40?ooo`004P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo02h0oooo00<00000
0?ooo`3oool0L@3oool00`000000oooo0?ooo`1N0?ooo`004`3oool3000000@0oooo00<000000?oo
o`3oool0o`3oool40?ooo`006P3oool00`000000oooo0?ooo`3o0?ooo`@0oooo000J0?ooo`030000
003oool0oooo0?l0oooo103oool001X0oooo00<000000?ooo`3oool0C@3oool00`000000oooo0?oo
o`0c0?ooo`030000003oool0oooo07d0oooo000J0?ooo`030000003oool0oooo0?l0oooo103oool0
01X0oooo00<000000?ooo`3oool0o`3oool40?ooo`006P3oool00`000000oooo0?ooo`3o0?ooo`@0
oooo000J0?ooo`030000003oool0oooo0?l0oooo103oool001X0oooo00<000000?ooo`3oool0o`3o
ool40?ooo`006P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo09X0oooo00<00000
0?ooo`3oool0B@3oool001X0oooo00<000000?ooo`3oool0o`3oool40?ooo`006P3oool00`000000
oooo0?ooo`3o0?ooo`@0oooo000J0?ooo`030000003oool0oooo0?l0oooo103oool001X0oooo00<0
00000?ooo`3oool0o`3oool40?ooo`006P3oool00`000000oooo0?ooo`3o0?ooo`@0oooo000J0?oo
o`030000003oool0oooo0?l0oooo103oool001X0oooo00<000000?ooo`3oool0D03oool00`000000
oooo0?ooo`0]0?ooo`030000003oool0oooo0800oooo000J0?ooo`030000003oool0oooo0?l0oooo
103oool001X0oooo00<000000?ooo`3oool0o`3oool40?ooo`006P3oool00`000000oooo0?ooo`3o
0?ooo`@0oooo000J0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0^`3oool00`00
0000oooo0?ooo`0i0?ooo`004`3oool2000000D0oooo00<000000?ooo`3oool0o`3oool40?ooo`00
4P3oool010000000oooo0?ooo`0000040?ooo`800000o`3oool50?ooo`004P3oool010000000oooo
0?ooo`0000040?ooo`030000003oool0oooo0?l0oooo103oool001<0oooo0P0000050?ooo`030000
003oool0oooo0?l0oooo103oool00180oooo00@000000?ooo`3oool00000103oool00`000000oooo
0?ooo`0l0?ooo`030000003oool0oooo05D0oooo00<000000?ooo`3oool0K03oool00180oooo00@0
00000?ooo`3oool00000103oool00`000000oooo0?ooo`3o0?ooo`@0oooo000C0?ooo`8000001@3o
ool00`000000oooo0?ooo`3o0?ooo`@0oooo000J0?ooo`030000003oool0oooo0?l0oooo103oool0
01X0oooo00<000000?ooo`3oool0o`3oool40?ooo`006P3oool00`000000oooo0?ooo`1C0?ooo`03
0000003oool0oooo02L0oooo00<000000?ooo`3oool0P`3oool001X0oooo00<000000?ooo`3oool0
o`3oool40?ooo`006P3oool00`000000oooo0?ooo`3o0?ooo`@0oooo000J0?ooo`030000003oool0
oooo0?l0oooo103oool001X0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`1g0?oo
o`030000003oool0oooo05/0oooo000J0?ooo`030000003oool0oooo0?l0oooo103oool001X0oooo
00<000000?ooo`3oool0o`3oool40?ooo`006P3oool00`000000oooo0?ooo`3o0?ooo`@0oooo000J
0?ooo`030000003oool0oooo0?l0oooo103oool001X0oooo00<000000?ooo`3oool0o`3oool40?oo
o`006P3oool00`000000oooo0?ooo`3o0?ooo`@0oooo000J0?ooo`030000003oool0oooo0?l0oooo
103oool001X0oooo00<000000?ooo`3oool0o`3oool40?ooo`006P3oool00`000000oooo0?ooo`07
0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`1;
0?ooo`030000003oool0oooo03L0oooo000J0?ooo`030000003oool0oooo0?l0oooo103oool001X0
oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`2E0?ooo`030000003oool0oooo04`0
oooo000J0?ooo`030000003oool0oooo0?l0oooo103oool001X0oooo00<000000?ooo`3oool0<@3o
ool00`000000oooo0?ooo`1[0?ooo`030000003oool0oooo0640oooo000<0?ooo`@000000`3oool2
000000D0oooo00<000000?ooo`3oool0o`3oool40?ooo`003P3oool01@000000oooo0?ooo`3oool0
00000080oooo00<000000?ooo`3oool00P3oool200000?l0oooo1@3oool000h0oooo00D000000?oo
o`3oool0oooo000000020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0o`3oool4
0?ooo`003P3oool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`3oool0o`3oool9
0?ooo`003P3oool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`3oool0o`3oool9
0?ooo`00303oool3000000<0oooo00@000000?ooo`3oool00000o`3oool;0?ooo`003P3oool00`00
0000oooo0?ooo`020?ooo`800000o`3oool<0?ooo`00o`3ooolQ0?ooo`00o`3ooolQ0?ooo`00B03o
ool3000003P0oooo0`00002J0?ooo`00B`3oool00`000000oooo0?ooo`0f0?ooo`030000003oool0
oooo09T0oooo000k0?ooo`8000000P3oool5000000030?ooo`00000000000080oooo0`0000020?oo
o`D000000P3oool3000000X0oooo00@000000?ooo`3oool0oooo0`0000040?ooo`030000003oool0
oooo00L0oooo0`0000020?ooo`030000003oool0oooo00<000000P3oool3000000@0oooo0`000003
0?ooo`80000000<0oooo0000003oool00P3oool2000000L0oooo100000030?ooo`D0000000@0oooo
00000000000000001@3oool2000000<0oooo0P0000000`3oool0000000000002000000030?ooo`00
000000000080oooo0P0000040?ooo`<00000D03oool003`0oooo00@000000?ooo`3oool000000P3o
ool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000000@0oooo00D000000?ooo`3o
ool0oooo0000000<0?ooo`030000003oool0000000D0oooo00<000000?ooo`3oool0303oool01000
0000oooo0?ooo`0000020?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool000000P3o
ool00`000000oooo0?ooo`030?ooo`040000003oool0oooo000000<0oooo00@000000?ooo`3oool0
00001`3oool010000000oooo0?ooo`0000040?ooo`030000003oool0oooo00@0oooo00D000000?oo
o`3oool0oooo000000020?ooo`030000003oool000000080oooo00@000000?ooo`3oool000000P3o
ool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000005<0oooo000l0?ooo`@00000
0P3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000000@0oooo00@000000?oo
o`3oool0oooo0`00000:0?ooo`030000003oool0000000<0oooo0P00000?0?ooo`<000000`3oool3
000000<0oooo00@000000?ooo`3oool000000P3oool3000000@0oooo0`0000030?ooo`030000003o
ool0oooo00P0oooo00@000000?ooo`3oool00000103oool00`000000oooo0?ooo`020?ooo`800000
103oool00`000000oooo0?ooo`030?ooo`<000000P3oool010000000oooo0?ooo`0000020?ooo`03
0000003oool0oooo00<0oooo0`00001A0?ooo`000P3oool4000000<0oooo1@000000103oool00000
0000000000050?ooo`8000000`3oool2000000030?ooo`00000000000080000000<0oooo00000000
00000P3oool2000000@0oooo0`00000;0?ooo`040000003oool0oooo00000080oooo00@000000?oo
o`3oool000000P3oool010000000oooo0?ooo`0000040?ooo`050000003oool0oooo0?ooo`000000
0P3oool00`000000oooo0?ooo`060?ooo`070000003oool0oooo0?ooo`000000oooo0000000A0?oo
o`040000003oool0oooo00000080oooo00@000000?ooo`3oool000000P3oool010000000oooo0?oo
o`0000020?ooo`040000003oool0oooo000000D0oooo00D000000?ooo`3oool0oooo0000000:0?oo
o`040000003oool0oooo000000@0oooo00D000000?ooo`3oool0oooo000000060?ooo`030000003o
ool0oooo00D0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`0000020?ooo`04
0000003oool0oooo00000080oooo00<000000?ooo`3oool0CP3oool000<0oooo00@000000?ooo`3o
ool00000103oool00`000000oooo0?ooo`040?ooo`050000003oool0oooo0?ooo`0000000P3oool0
0`000000oooo000000020?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool00000
0P3oool010000000oooo0?ooo`00000?0?ooo`060000003oool000000?ooo`000000oooo0P000004
0?ooo`80000000<0oooo0000003oool00P3oool00`000000oooo0?ooo`020?ooo`800000203oool2
00000080oooo0P0000020?ooo`<000003P3oool010000000oooo0?ooo`0000030?ooo`8000000P3o
ool010000000oooo0000000000040?ooo`800000103oool2000000<0oooo100000080?ooo`040000
003oool0oooo000000<0oooo0P0000040?ooo`<000000P3oool4000000<0oooo0P0000020?ooo`04
0000003oool00000000000@0oooo0P0000040?ooo`800000D@3oool000<0oooo00@000000?ooo`3o
ool00000103oool00`000000oooo0?ooo`020?ooo`800000103oool00`000000oooo0?ooo`030?oo
o`<000000P3oool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo00<0oooo0`00000<
0?ooo`<000004@3oool00`000000oooo0?ooo`0T0?ooo`@00000703oool00`000000oooo0?ooo`07
0?ooo`@00000403oool00`000000oooo0?ooo`1X0?ooo`000`3oool010000000oooo0?ooo`000004
0?ooo`050000003oool0oooo0?ooo`0000001P3oool00`000000oooo0?ooo`050?ooo`040000003o
ool0oooo00000080oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`0000020?oo
o`030000003oool0oooo01`0oooo0P0000160?ooo`030000003oool0oooo0100oooo00<000000?oo
o`3oool0203oool00`000000oooo0?ooo`1X0?ooo`000`3oool010000000oooo0?ooo`0000030?oo
o`800000103oool300000080oooo100000030?ooo`8000000P3oool010000000oooo000000000004
0?ooo`800000103oool200000?00oooo00020?ooo`@00000403oool00`000000oooo0?ooo`3o0?oo
o`P0oooo000;0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0o`3oool80?ooo`00
o`3ooolQ0?ooo`00o`3ooolQ0?ooo`00o`3ooolQ0?ooo`00o`3ooolQ0?ooo`00\
\>"],
  ImageRangeCache->{{{0, 287}, {176.938, 0}} -> {-7.22481, -1.33538, 
  0.277866, 0.0738002}}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(X = FindAngles[50]; \nLength[X]\), 
    \(Angles = Sort[\(Transpose[X]\)[\([1]\)]]; \n
    AnglesAndRepeats = 
      Transpose[{\(Transpose[X]\)[\([1]\)], \t\t\(Transpose[X]\)[\([3]\)]}]; 
    \nListPlot[Angles, PlotLabel -> "\<Rational Angles\>", 
      AxesLabel -> {"\<\>", "\<Angle\>"}]; \n
    ListPlot[AnglesAndRepeats, \ 
      PlotLabel -> "\<Angle vs. Repeat Distance\>", \ 
      AxesLabel -> {"\<Angle\>", "\<Distance\>"}]; \)}], "Input"],

Cell[BoxData[
    \(917\)], "Output"],

Cell[CellGroupData[{

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.00103858 0.0147151 0.00981006 [
[.23153 .00222 -9 -9 ]
[.23153 .00222 9 0 ]
[.43924 .00222 -9 -9 ]
[.43924 .00222 9 0 ]
[.64696 .00222 -9 -9 ]
[.64696 .00222 9 0 ]
[.85468 .00222 -9 -9 ]
[.85468 .00222 9 0 ]
[.01131 .11282 -12 -4.5 ]
[.01131 .11282 0 4.5 ]
[.01131 .21092 -12 -4.5 ]
[.01131 .21092 0 4.5 ]
[.01131 .30902 -12 -4.5 ]
[.01131 .30902 0 4.5 ]
[.01131 .40712 -12 -4.5 ]
[.01131 .40712 0 4.5 ]
[.01131 .50522 -12 -4.5 ]
[.01131 .50522 0 4.5 ]
[.01131 .60332 -12 -4.5 ]
[.01131 .60332 0 4.5 ]
[.02381 .64303 -17 0 ]
[.02381 .64303 17 9.8125 ]
[.5 .63053 -47 0 ]
[.5 .63053 47 10.125 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.23153 .01472 m
.23153 .02097 L
s
[(200)] .23153 .00222 0 1 Mshowa
.43924 .01472 m
.43924 .02097 L
s
[(400)] .43924 .00222 0 1 Mshowa
.64696 .01472 m
.64696 .02097 L
s
[(600)] .64696 .00222 0 1 Mshowa
.85468 .01472 m
.85468 .02097 L
s
[(800)] .85468 .00222 0 1 Mshowa
.125 Mabswid
.07574 .01472 m
.07574 .01847 L
s
.12767 .01472 m
.12767 .01847 L
s
.1796 .01472 m
.1796 .01847 L
s
.28346 .01472 m
.28346 .01847 L
s
.33538 .01472 m
.33538 .01847 L
s
.38731 .01472 m
.38731 .01847 L
s
.49117 .01472 m
.49117 .01847 L
s
.5431 .01472 m
.5431 .01847 L
s
.59503 .01472 m
.59503 .01847 L
s
.69889 .01472 m
.69889 .01847 L
s
.75082 .01472 m
.75082 .01847 L
s
.80275 .01472 m
.80275 .01847 L
s
.90661 .01472 m
.90661 .01847 L
s
.95853 .01472 m
.95853 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .11282 m
.03006 .11282 L
s
[(10)] .01131 .11282 1 0 Mshowa
.02381 .21092 m
.03006 .21092 L
s
[(20)] .01131 .21092 1 0 Mshowa
.02381 .30902 m
.03006 .30902 L
s
[(30)] .01131 .30902 1 0 Mshowa
.02381 .40712 m
.03006 .40712 L
s
[(40)] .01131 .40712 1 0 Mshowa
.02381 .50522 m
.03006 .50522 L
s
[(50)] .01131 .50522 1 0 Mshowa
.02381 .60332 m
.03006 .60332 L
s
[(60)] .01131 .60332 1 0 Mshowa
.125 Mabswid
.02381 .03434 m
.02756 .03434 L
s
.02381 .05396 m
.02756 .05396 L
s
.02381 .07358 m
.02756 .07358 L
s
.02381 .0932 m
.02756 .0932 L
s
.02381 .13244 m
.02756 .13244 L
s
.02381 .15206 m
.02756 .15206 L
s
.02381 .17168 m
.02756 .17168 L
s
.02381 .1913 m
.02756 .1913 L
s
.02381 .23054 m
.02756 .23054 L
s
.02381 .25016 m
.02756 .25016 L
s
.02381 .26978 m
.02756 .26978 L
s
.02381 .2894 m
.02756 .2894 L
s
.02381 .32864 m
.02756 .32864 L
s
.02381 .34826 m
.02756 .34826 L
s
.02381 .36788 m
.02756 .36788 L
s
.02381 .3875 m
.02756 .3875 L
s
.02381 .42674 m
.02756 .42674 L
s
.02381 .44636 m
.02756 .44636 L
s
.02381 .46598 m
.02756 .46598 L
s
.02381 .4856 m
.02756 .4856 L
s
.02381 .52484 m
.02756 .52484 L
s
.02381 .54446 m
.02756 .54446 L
s
.02381 .56408 m
.02756 .56408 L
s
.02381 .5837 m
.02756 .5837 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
gsave
.02381 .64303 -78 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Angle) show
93.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
gsave
.5 .63053 -108 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.125000 moveto
 558.000000 14.125000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.125000 lineto
closepath
clip newpath
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Rational) show
117.000000 11.562500 moveto
(Angles) show
153.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.008 w
.02485 .01472 Mdot
.02589 .02041 Mdot
.02693 .02062 Mdot
.02796 .02084 Mdot
.029 .02108 Mdot
.03004 .02134 Mdot
.03108 .02162 Mdot
.03212 .02193 Mdot
.03316 .02226 Mdot
.0342 .02263 Mdot
.03523 .02304 Mdot
.03627 .02349 Mdot
.03731 .02399 Mdot
.03835 .02455 Mdot
.03939 .02518 Mdot
.04043 .0259 Mdot
.04147 .0263 Mdot
.0425 .02673 Mdot
.04354 .02719 Mdot
.04458 .02769 Mdot
.04562 .02823 Mdot
.04666 .02882 Mdot
.0477 .02946 Mdot
.04874 .03016 Mdot
.04977 .03094 Mdot
.05081 .03179 Mdot
.05185 .03241 Mdot
.05289 .03274 Mdot
.05393 .03308 Mdot
.05497 .0338 Mdot
.05601 .03458 Mdot
.05704 .03499 Mdot
.05808 .03542 Mdot
.05912 .03634 Mdot
.06016 .03734 Mdot
.0612 .03788 Mdot
.06224 .03845 Mdot
.06328 .03874 Mdot
.06431 .03966 Mdot
.06535 .04066 Mdot
.06639 .04101 Mdot
.06743 .04174 Mdot
.06847 .04251 Mdot
.06951 .04291 Mdot
.07055 .04316 Mdot
.07158 .04419 Mdot
.07262 .0453 Mdot
.07366 .04559 Mdot
.0747 .04609 Mdot
.07574 .0465 Mdot
.07678 .04713 Mdot
.07782 .04779 Mdot
.07885 .04825 Mdot
.07989 .04883 Mdot
.08093 .04919 Mdot
.08197 .04944 Mdot
.08301 .05072 Mdot
.08405 .0521 Mdot
.08509 .05239 Mdot
.08612 .05283 Mdot
.08716 .05359 Mdot
.0882 .05422 Mdot
.08924 .0545 Mdot
.09028 .05521 Mdot
.09132 .05594 Mdot
.09236 .05624 Mdot
.09339 .05696 Mdot
.09443 .05749 Mdot
.09547 .0579 Mdot
.09651 .05848 Mdot
.09755 .05888 Mdot
.09859 .05916 Mdot
.09963 .06097 Mdot
.10066 .06267 Mdot
.1017 .06293 Mdot
.10274 .06327 Mdot
.10378 .06376 Mdot
.10482 .06451 Mdot
.10586 .06506 Mdot
.1069 .06581 Mdot
.10793 .06649 Mdot
.10897 .06691 Mdot
.11001 .06739 Mdot
.11105 .06766 Mdot
.11209 .06863 Mdot
.11313 .06965 Mdot
.11417 .06994 Mdot
.1152 .07048 Mdot
.11624 .07096 Mdot
.11728 .07179 Mdot
.11832 .07246 Mdot
.11936 .07275 Mdot
.1204 .0735 Mdot
.12144 .0741 Mdot
.12247 .07458 Mdot
.12351 .07533 Mdot
.12455 .07587 Mdot
.12559 .07628 Mdot
.12663 .0766 Mdot
.12767 .07687 Mdot
.12871 .07708 Mdot
.12975 .07933 Mdot
.13078 .08175 Mdot
.13182 .082 Mdot
.13286 .08231 Mdot
.1339 .0827 Mdot
.13494 .08321 Mdot
.13598 .0839 Mdot
.13702 .08435 Mdot
.13805 .08489 Mdot
.13909 .08557 Mdot
.14013 .08644 Mdot
.14117 .08716 Mdot
.14221 .08758 Mdot
.14325 .08805 Mdot
.14429 .0883 Mdot
.14532 .08916 Mdot
.14636 .09021 Mdot
.1474 .09061 Mdot
.14844 .09095 Mdot
.14948 .0915 Mdot
.15052 .09211 Mdot
.15156 .09254 Mdot
.15259 .09314 Mdot
.15363 .09352 Mdot
.15467 .09379 Mdot
.15571 .09529 Mdot
.15675 .09684 Mdot
.15779 .09713 Mdot
.15883 .09756 Mdot
.15986 .09823 Mdot
.1609 .09874 Mdot
.16194 .09947 Mdot
.16298 .10015 Mdot
.16402 .10058 Mdot
.16506 .10109 Mdot
.1661 .10139 Mdot
.16713 .1025 Mdot
.16817 .10372 Mdot
.16921 .1041 Mdot
.17025 .1048 Mdot
.17129 .10545 Mdot
.17233 .10575 Mdot
.17337 .10661 Mdot
.1744 .10761 Mdot
.17544 .10806 Mdot
.17648 .10849 Mdot
.17752 .10926 Mdot
.17856 .10995 Mdot
.1796 .11026 Mdot
.18064 .11111 Mdot
.18167 .11185 Mdot
.18271 .11248 Mdot
.18375 .11304 Mdot
.18479 .11353 Mdot
.18583 .11397 Mdot
.18687 .11437 Mdot
.18791 .11505 Mdot
.18894 .11561 Mdot
.18998 .11608 Mdot
.19102 .11648 Mdot
.19206 .11683 Mdot
.1931 .11713 Mdot
.19414 .1174 Mdot
.19518 .11764 Mdot
.19621 .11785 Mdot
.19725 .12158 Mdot
.19829 .12537 Mdot
.19933 .12559 Mdot
.20037 .12583 Mdot
.20141 .12611 Mdot
.20245 .12642 Mdot
.20348 .12679 Mdot
.20452 .12721 Mdot
.20556 .1277 Mdot
.2066 .12829 Mdot
.20764 .12901 Mdot
.20868 .12943 Mdot
.20972 .1299 Mdot
.21075 .13043 Mdot
.21179 .13103 Mdot
.21283 .13172 Mdot
.21387 .13252 Mdot
.21491 .13312 Mdot
.21595 .13345 Mdot
.21699 .1338 Mdot
.21802 .13456 Mdot
.21906 .13542 Mdot
.2201 .13589 Mdot
.22114 .1364 Mdot
.22218 .13667 Mdot
.22322 .13754 Mdot
.22426 .13851 Mdot
.22529 .13886 Mdot
.22633 .1396 Mdot
.22737 .14008 Mdot
.22841 .14042 Mdot
.22945 .14085 Mdot
.23049 .14112 Mdot
.23153 .14228 Mdot
.23256 .14359 Mdot
.2336 .14395 Mdot
.23464 .14456 Mdot
.23568 .14508 Mdot
.23672 .1459 Mdot
.23776 .14678 Mdot
.2388 .14741 Mdot
.23983 .14787 Mdot
.24087 .14823 Mdot
.24191 .14875 Mdot
.24295 .14911 Mdot
.24399 .14937 Mdot
.24503 .15105 Mdot
.24607 .1527 Mdot
.2471 .15295 Mdot
.24814 .15329 Mdot
.24918 .15378 Mdot
.25022 .15454 Mdot
.25126 .15511 Mdot
.2523 .1559 Mdot
.25334 .15662 Mdot
.25438 .15706 Mdot
.25541 .15759 Mdot
.25645 .15789 Mdot
.25749 .15897 Mdot
.25853 .16012 Mdot
.25957 .16046 Mdot
.26061 .16109 Mdot
.26165 .16166 Mdot
.26268 .16264 Mdot
.26372 .16347 Mdot
.26476 .16383 Mdot
.2658 .16416 Mdot
.26684 .16476 Mdot
.26788 .16552 Mdot
.26892 .16614 Mdot
.26995 .16667 Mdot
.27099 .16711 Mdot
.27203 .16783 Mdot
.27307 .16839 Mdot
.27411 .16883 Mdot
.27515 .16919 Mdot
.27619 .16949 Mdot
.27722 .16974 Mdot
.27826 .16995 Mdot
.2793 .17268 Mdot
.28034 .1755 Mdot
.28138 .17573 Mdot
.28242 .176 Mdot
.28346 .17632 Mdot
.28449 .17672 Mdot
.28553 .17721 Mdot
.28657 .17783 Mdot
.28761 .17821 Mdot
.28865 .17865 Mdot
.28969 .17917 Mdot
.29073 .17978 Mdot
.29176 .18053 Mdot
.2928 .18081 Mdot
.29384 .18145 Mdot
.29488 .18219 Mdot
.29592 .18261 Mdot
.29696 .18307 Mdot
.298 .18413 Mdot
.29903 .18508 Mdot
.30007 .18543 Mdot
.30111 .1862 Mdot
.30215 .1867 Mdot
.30319 .18706 Mdot
.30423 .18754 Mdot
.30527 .18784 Mdot
.3063 .18917 Mdot
.30734 .19048 Mdot
.30838 .19077 Mdot
.30942 .19121 Mdot
.31046 .19201 Mdot
.3115 .1927 Mdot
.31254 .19301 Mdot
.31357 .19383 Mdot
.31461 .19472 Mdot
.31565 .1951 Mdot
.31669 .19544 Mdot
.31773 .19603 Mdot
.31877 .19674 Mdot
.31981 .19731 Mdot
.32084 .19777 Mdot
.32188 .19815 Mdot
.32292 .19874 Mdot
.32396 .19917 Mdot
.325 .19951 Mdot
.32604 .19978 Mdot
.32708 .2 Mdot
.32811 .20215 Mdot
.32915 .20442 Mdot
.33019 .20467 Mdot
.33123 .20497 Mdot
.33227 .20537 Mdot
.33331 .20588 Mdot
.33435 .2066 Mdot
.33538 .20707 Mdot
.33642 .20765 Mdot
.33746 .20839 Mdot
.3385 .20869 Mdot
.33954 .20936 Mdot
.34058 .2102 Mdot
.34162 .21069 Mdot
.34265 .21125 Mdot
.34369 .21156 Mdot
.34473 .21262 Mdot
.34577 .21366 Mdot
.34681 .21395 Mdot
.34785 .21447 Mdot
.34889 .21493 Mdot
.34992 .21567 Mdot
.35096 .21651 Mdot
.352 .21713 Mdot
.35304 .2176 Mdot
.35408 .21798 Mdot
.35512 .21854 Mdot
.35616 .21893 Mdot
.35719 .21923 Mdot
.35823 .21945 Mdot
.35927 .22123 Mdot
.36031 .22315 Mdot
.36135 .22342 Mdot
.36239 .22378 Mdot
.36343 .22427 Mdot
.36446 .22499 Mdot
.3655 .2255 Mdot
.36654 .22618 Mdot
.36758 .22675 Mdot
.36862 .2271 Mdot
.36966 .22749 Mdot
.3707 .22844 Mdot
.37173 .22938 Mdot
.37277 .22975 Mdot
.37381 .23007 Mdot
.37485 .23059 Mdot
.37589 .23119 Mdot
.37693 .23163 Mdot
.37797 .23223 Mdot
.37901 .23263 Mdot
.38004 .23291 Mdot
.38108 .23457 Mdot
.38212 .23615 Mdot
.38316 .23641 Mdot
.3842 .23677 Mdot
.38524 .23729 Mdot
.38628 .23767 Mdot
.38731 .23816 Mdot
.38835 .23884 Mdot
.38939 .23912 Mdot
.39043 .23983 Mdot
.39147 .24052 Mdot
.39251 .24079 Mdot
.39355 .24142 Mdot
.39458 .24219 Mdot
.39562 .24265 Mdot
.39666 .24295 Mdot
.3977 .2444 Mdot
.39874 .24579 Mdot
.39978 .24606 Mdot
.40082 .24647 Mdot
.40185 .24713 Mdot
.40289 .24765 Mdot
.40393 .24841 Mdot
.40497 .24915 Mdot
.40601 .24963 Mdot
.40705 .25023 Mdot
.40809 .25058 Mdot
.40912 .25195 Mdot
.41016 .25325 Mdot
.4112 .25357 Mdot
.41224 .25409 Mdot
.41328 .2545 Mdot
.41432 .2551 Mdot
.41536 .25569 Mdot
.41639 .25607 Mdot
.41743 .25655 Mdot
.41847 .25683 Mdot
.41951 .25792 Mdot
.42055 .25897 Mdot
.42159 .25922 Mdot
.42263 .25964 Mdot
.42366 .25997 Mdot
.4247 .26046 Mdot
.42574 .26125 Mdot
.42678 .26164 Mdot
.42782 .26276 Mdot
.42886 .26383 Mdot
.4299 .26418 Mdot
.43093 .26486 Mdot
.43197 .26551 Mdot
.43301 .26583 Mdot
.43405 .26677 Mdot
.43509 .26766 Mdot
.43613 .26795 Mdot
.43717 .26852 Mdot
.4382 .26908 Mdot
.43924 .26935 Mdot
.44028 .27014 Mdot
.44132 .27114 Mdot
.44236 .27163 Mdot
.4434 .2721 Mdot
.44444 .27301 Mdot
.44547 .27387 Mdot
.44651 .27429 Mdot
.44755 .2747 Mdot
.44859 .27549 Mdot
.44963 .27624 Mdot
.45067 .2766 Mdot
.45171 .27696 Mdot
.45274 .27765 Mdot
.45378 .2783 Mdot
.45482 .27862 Mdot
.45586 .27894 Mdot
.4569 .27954 Mdot
.45794 .28041 Mdot
.45898 .28122 Mdot
.46001 .28199 Mdot
.46105 .28272 Mdot
.46209 .28342 Mdot
.46313 .28407 Mdot
.46417 .28469 Mdot
.46521 .28529 Mdot
.46625 .28585 Mdot
.46728 .28639 Mdot
.46832 .2869 Mdot
.46936 .28739 Mdot
.4704 .28786 Mdot
.47144 .28831 Mdot
.47248 .28874 Mdot
.47352 .28916 Mdot
.47455 .28994 Mdot
.47559 .29065 Mdot
.47663 .29132 Mdot
.47767 .29194 Mdot
.47871 .29252 Mdot
.47975 .29306 Mdot
.48079 .29357 Mdot
.48182 .29404 Mdot
.48286 .29449 Mdot
.4839 .29491 Mdot
.48494 .29531 Mdot
.48598 .29568 Mdot
.48702 .29604 Mdot
.48806 .29638 Mdot
.48909 .2967 Mdot
.49013 .297 Mdot
.49117 .29729 Mdot
.49221 .29757 Mdot
.49325 .29783 Mdot
.49429 .29808 Mdot
.49533 .29832 Mdot
.49636 .29855 Mdot
.4974 .29877 Mdot
.49844 .29898 Mdot
.49948 .29918 Mdot
.50052 .30902 Mdot
.50156 .31885 Mdot
.5026 .31905 Mdot
.50364 .31926 Mdot
.50467 .31948 Mdot
.50571 .31971 Mdot
.50675 .31995 Mdot
.50779 .32021 Mdot
.50883 .32047 Mdot
.50987 .32074 Mdot
.51091 .32103 Mdot
.51194 .32134 Mdot
.51298 .32166 Mdot
.51402 .322 Mdot
.51506 .32235 Mdot
.5161 .32273 Mdot
.51714 .32312 Mdot
.51818 .32354 Mdot
.51921 .32399 Mdot
.52025 .32447 Mdot
.52129 .32497 Mdot
.52233 .32551 Mdot
.52337 .32609 Mdot
.52441 .32671 Mdot
.52545 .32738 Mdot
.52648 .3281 Mdot
.52752 .32888 Mdot
.52856 .32929 Mdot
.5296 .32972 Mdot
.53064 .33017 Mdot
.53168 .33064 Mdot
.53272 .33113 Mdot
.53375 .33165 Mdot
.53479 .33218 Mdot
.53583 .33275 Mdot
.53687 .33334 Mdot
.53791 .33396 Mdot
.53895 .33462 Mdot
.53999 .33531 Mdot
.54102 .33604 Mdot
.54206 .33681 Mdot
.5431 .33763 Mdot
.54414 .33849 Mdot
.54518 .3391 Mdot
.54622 .33941 Mdot
.54726 .33973 Mdot
.54829 .34039 Mdot
.54933 .34108 Mdot
.55037 .34143 Mdot
.55141 .3418 Mdot
.55245 .34255 Mdot
.55349 .34333 Mdot
.55453 .34374 Mdot
.55556 .34416 Mdot
.5566 .34502 Mdot
.55764 .34593 Mdot
.55868 .34641 Mdot
.55972 .34689 Mdot
.56076 .3479 Mdot
.5618 .34869 Mdot
.56283 .34896 Mdot
.56387 .34951 Mdot
.56491 .35008 Mdot
.56595 .35037 Mdot
.56699 .35127 Mdot
.56803 .3522 Mdot
.56907 .35252 Mdot
.5701 .35318 Mdot
.57114 .35385 Mdot
.57218 .3542 Mdot
.57322 .35527 Mdot
.57426 .35639 Mdot
.5753 .35678 Mdot
.57634 .35757 Mdot
.57737 .35806 Mdot
.57841 .35839 Mdot
.57945 .35881 Mdot
.58049 .35907 Mdot
.58153 .36011 Mdot
.58257 .36121 Mdot
.58361 .36149 Mdot
.58464 .36196 Mdot
.58568 .36235 Mdot
.58672 .36294 Mdot
.58776 .36354 Mdot
.5888 .36395 Mdot
.58984 .36447 Mdot
.59088 .36478 Mdot
.59191 .36609 Mdot
.59295 .36745 Mdot
.59399 .3678 Mdot
.59503 .3684 Mdot
.59607 .36888 Mdot
.59711 .36963 Mdot
.59815 .37039 Mdot
.59918 .37091 Mdot
.60022 .37157 Mdot
.60126 .37197 Mdot
.6023 .37224 Mdot
.60334 .37363 Mdot
.60438 .37509 Mdot
.60542 .37539 Mdot
.60645 .37584 Mdot
.60749 .37661 Mdot
.60853 .37724 Mdot
.60957 .37751 Mdot
.61061 .3782 Mdot
.61165 .37891 Mdot
.61269 .3792 Mdot
.61372 .37987 Mdot
.61476 .38037 Mdot
.6158 .38074 Mdot
.61684 .38127 Mdot
.61788 .38163 Mdot
.61892 .38188 Mdot
.61996 .38347 Mdot
.62099 .38512 Mdot
.62203 .3854 Mdot
.62307 .3858 Mdot
.62411 .38641 Mdot
.62515 .38685 Mdot
.62619 .38744 Mdot
.62723 .38797 Mdot
.62827 .38829 Mdot
.6293 .38865 Mdot
.63034 .38959 Mdot
.63138 .39055 Mdot
.63242 .39094 Mdot
.63346 .39128 Mdot
.6345 .39186 Mdot
.63554 .39253 Mdot
.63657 .39304 Mdot
.63761 .39377 Mdot
.63865 .39426 Mdot
.63969 .39461 Mdot
.64073 .39488 Mdot
.64177 .3968 Mdot
.64281 .39858 Mdot
.64384 .39881 Mdot
.64488 .3991 Mdot
.64592 .3995 Mdot
.64696 .40006 Mdot
.648 .40043 Mdot
.64904 .40091 Mdot
.65008 .40153 Mdot
.65111 .40236 Mdot
.65215 .40311 Mdot
.65319 .40356 Mdot
.65423 .40408 Mdot
.65527 .40438 Mdot
.65631 .40541 Mdot
.65735 .40648 Mdot
.65838 .40678 Mdot
.65942 .40734 Mdot
.66046 .40784 Mdot
.6615 .40867 Mdot
.66254 .40935 Mdot
.66358 .40964 Mdot
.66462 .41038 Mdot
.66565 .41097 Mdot
.66669 .41144 Mdot
.66773 .41215 Mdot
.66877 .41267 Mdot
.66981 .41306 Mdot
.67085 .41337 Mdot
.67189 .41361 Mdot
.67292 .41588 Mdot
.67396 .41803 Mdot
.675 .41825 Mdot
.67604 .41852 Mdot
.67708 .41886 Mdot
.67812 .4193 Mdot
.67916 .41989 Mdot
.68019 .42027 Mdot
.68123 .42073 Mdot
.68227 .42129 Mdot
.68331 .422 Mdot
.68435 .4226 Mdot
.68539 .42294 Mdot
.68643 .42331 Mdot
.68746 .4242 Mdot
.6885 .42502 Mdot
.68954 .42533 Mdot
.69058 .42602 Mdot
.69162 .42682 Mdot
.69266 .42727 Mdot
.6937 .42755 Mdot
.69473 .42886 Mdot
.69577 .4302 Mdot
.69681 .4305 Mdot
.69785 .43097 Mdot
.69889 .43133 Mdot
.69993 .43184 Mdot
.70097 .4326 Mdot
.702 .43296 Mdot
.70304 .4339 Mdot
.70408 .43496 Mdot
.70512 .43542 Mdot
.70616 .43584 Mdot
.7072 .43659 Mdot
.70824 .43722 Mdot
.70927 .43751 Mdot
.71031 .43825 Mdot
.71135 .43887 Mdot
.71239 .43938 Mdot
.71343 .43982 Mdot
.71447 .44021 Mdot
.71551 .44083 Mdot
.71654 .44132 Mdot
.71758 .44171 Mdot
.71862 .44203 Mdot
.71966 .4423 Mdot
.7207 .44253 Mdot
.72174 .44536 Mdot
.72278 .44808 Mdot
.72381 .4483 Mdot
.72485 .44855 Mdot
.72589 .44885 Mdot
.72693 .4492 Mdot
.72797 .44964 Mdot
.72901 .4502 Mdot
.73005 .45092 Mdot
.73108 .45137 Mdot
.73212 .45189 Mdot
.73316 .45251 Mdot
.7342 .45327 Mdot
.73524 .45387 Mdot
.73628 .45421 Mdot
.73732 .45457 Mdot
.73835 .45539 Mdot
.73939 .45638 Mdot
.74043 .45694 Mdot
.74147 .45757 Mdot
.74251 .45792 Mdot
.74355 .45907 Mdot
.74459 .46015 Mdot
.74562 .46045 Mdot
.74666 .46097 Mdot
.7477 .46142 Mdot
.74874 .46214 Mdot
.74978 .46292 Mdot
.75082 .46349 Mdot
.75186 .46425 Mdot
.7529 .46474 Mdot
.75393 .46508 Mdot
.75497 .46533 Mdot
.75601 .46698 Mdot
.75705 .46866 Mdot
.75809 .46892 Mdot
.75913 .46928 Mdot
.76017 .4698 Mdot
.7612 .47016 Mdot
.76224 .47063 Mdot
.76328 .47125 Mdot
.76432 .47213 Mdot
.76536 .47295 Mdot
.7664 .47347 Mdot
.76744 .47409 Mdot
.76847 .47444 Mdot
.76951 .47575 Mdot
.77055 .47691 Mdot
.77159 .47718 Mdot
.77263 .47762 Mdot
.77367 .47795 Mdot
.77471 .47843 Mdot
.77574 .47918 Mdot
.77678 .47953 Mdot
.77782 .4805 Mdot
.77886 .48136 Mdot
.7799 .48163 Mdot
.78094 .48214 Mdot
.78198 .48261 Mdot
.78301 .48348 Mdot
.78405 .48424 Mdot
.78509 .48458 Mdot
.78613 .48491 Mdot
.78717 .48552 Mdot
.78821 .48631 Mdot
.78925 .487 Mdot
.79028 .4876 Mdot
.79132 .48813 Mdot
.79236 .4886 Mdot
.7934 .48902 Mdot
.79444 .48974 Mdot
.79548 .49033 Mdot
.79652 .49083 Mdot
.79755 .49125 Mdot
.79859 .49161 Mdot
.79963 .49193 Mdot
.80067 .4922 Mdot
.80171 .49245 Mdot
.80275 .49267 Mdot
.80379 .49645 Mdot
.80482 .50018 Mdot
.80586 .5004 Mdot
.8069 .50063 Mdot
.80794 .5009 Mdot
.80898 .5012 Mdot
.81002 .50155 Mdot
.81106 .50195 Mdot
.81209 .50243 Mdot
.81313 .50299 Mdot
.81417 .50367 Mdot
.81521 .50406 Mdot
.81625 .5045 Mdot
.81729 .50499 Mdot
.81833 .50555 Mdot
.81936 .50619 Mdot
.8204 .50692 Mdot
.82144 .50777 Mdot
.82248 .50809 Mdot
.82352 .50878 Mdot
.82456 .50955 Mdot
.8256 .50997 Mdot
.82663 .51043 Mdot
.82767 .51143 Mdot
.82871 .51228 Mdot
.82975 .51259 Mdot
.83079 .51323 Mdot
.83183 .51394 Mdot
.83287 .51431 Mdot
.8339 .51554 Mdot
.83494 .51664 Mdot
.83598 .51694 Mdot
.83702 .51746 Mdot
.83806 .51789 Mdot
.8391 .51857 Mdot
.84014 .51929 Mdot
.84117 .51981 Mdot
.84221 .52048 Mdot
.84325 .5209 Mdot
.84429 .52119 Mdot
.84533 .52275 Mdot
.84637 .52425 Mdot
.84741 .52451 Mdot
.84844 .5249 Mdot
.84948 .52549 Mdot
.85052 .52593 Mdot
.85156 .52653 Mdot
.8526 .52709 Mdot
.85364 .52743 Mdot
.85468 .52783 Mdot
.85571 .52887 Mdot
.85675 .52973 Mdot
.85779 .52998 Mdot
.85883 .53045 Mdot
.85987 .53087 Mdot
.86091 .5316 Mdot
.86195 .53246 Mdot
.86298 .53314 Mdot
.86402 .53368 Mdot
.86506 .53413 Mdot
.8661 .53482 Mdot
.86714 .53533 Mdot
.86818 .53573 Mdot
.86922 .53603 Mdot
.87025 .53629 Mdot
.87129 .5387 Mdot
.87233 .54095 Mdot
.87337 .54117 Mdot
.87441 .54143 Mdot
.87545 .54175 Mdot
.87649 .54217 Mdot
.87753 .54271 Mdot
.87856 .54345 Mdot
.8796 .54394 Mdot
.88064 .54453 Mdot
.88168 .54528 Mdot
.88272 .54557 Mdot
.88376 .54625 Mdot
.8848 .54707 Mdot
.88583 .54755 Mdot
.88687 .54809 Mdot
.88791 .54839 Mdot
.88895 .5494 Mdot
.88999 .55037 Mdot
.89103 .55065 Mdot
.89207 .55113 Mdot
.8931 .55154 Mdot
.89414 .55222 Mdot
.89518 .55297 Mdot
.89622 .55352 Mdot
.89726 .55427 Mdot
.8983 .55476 Mdot
.89934 .55511 Mdot
.90037 .55536 Mdot
.90141 .55706 Mdot
.90245 .55887 Mdot
.90349 .55916 Mdot
.90453 .55955 Mdot
.90557 .56014 Mdot
.90661 .56054 Mdot
.90764 .56107 Mdot
.90868 .56179 Mdot
.90972 .56209 Mdot
.91076 .56282 Mdot
.9118 .56353 Mdot
.91284 .56381 Mdot
.91388 .56444 Mdot
.91491 .5652 Mdot
.91595 .56564 Mdot
.91699 .56593 Mdot
.91803 .56731 Mdot
.91907 .56859 Mdot
.92011 .56884 Mdot
.92115 .5692 Mdot
.92218 .56979 Mdot
.92322 .57024 Mdot
.92426 .5709 Mdot
.9253 .57154 Mdot
.92634 .57195 Mdot
.92738 .57244 Mdot
.92842 .57273 Mdot
.92945 .57384 Mdot
.93049 .57488 Mdot
.93153 .57512 Mdot
.93257 .57553 Mdot
.93361 .5763 Mdot
.93465 .57703 Mdot
.93569 .57738 Mdot
.93672 .57837 Mdot
.93776 .5793 Mdot
.9388 .57959 Mdot
.93984 .58015 Mdot
.94088 .58069 Mdot
.94192 .5817 Mdot
.94296 .58261 Mdot
.94399 .58305 Mdot
.94503 .58346 Mdot
.94607 .58424 Mdot
.94711 .58496 Mdot
.94815 .5853 Mdot
.94919 .58562 Mdot
.95023 .58624 Mdot
.95126 .5871 Mdot
.9523 .58787 Mdot
.95334 .58857 Mdot
.95438 .58921 Mdot
.95542 .5898 Mdot
.95646 .59034 Mdot
.9575 .59084 Mdot
.95853 .5913 Mdot
.95957 .59173 Mdot
.96061 .59213 Mdot
.96165 .59285 Mdot
.96269 .59349 Mdot
.96373 .59405 Mdot
.96477 .59455 Mdot
.9658 .595 Mdot
.96684 .5954 Mdot
.96788 .59577 Mdot
.96892 .59611 Mdot
.96996 .59641 Mdot
.971 .5967 Mdot
.97204 .59696 Mdot
.97307 .5972 Mdot
.97411 .59742 Mdot
.97515 .59763 Mdot
.97619 .60332 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{360, 222.375},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40005X0000gR000`400?l00000o`00003oo`3ooomY
0?ooo`00o`3ooomY0?ooo`00o`3ooomY0?ooo`00o`3ooomY0?ooo`00F`3oool4000000<0oooo0P00
00040?ooo`800000=@3oool00`000000oooo0?ooo`020?ooo`800000103oool2000003<0oooo0P00
00040?ooo`800000103oool2000003@0oooo0P0000040?ooo`800000103oool2000003T0oooo001K
0?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?oo
o`00000d0?ooo`050000003oool0oooo0?ooo`0000000P3oool010000000oooo0?ooo`0000020?oo
o`030000003oool0oooo02l0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`00
00020?ooo`040000003oool0oooo00000380oooo00@000000?ooo`3oool000000P3oool010000000
oooo0?ooo`0000020?ooo`040000003oool0oooo000003P0oooo001L0?ooo`030000003oool0oooo
0080oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`00000a0?ooo`D000000P3o
ool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo00000340oooo00@000000?ooo`3o
ool000000P3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo00000380oooo00@0
00000?ooo`3oool000000P3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo0000
03P0oooo001M0?ooo`050000003oool0oooo0?ooo`0000000P3oool010000000oooo0?ooo`000002
0?ooo`030000003oool0oooo02l0oooo00@000000?ooo`3oool000000`3oool010000000oooo0?oo
o`0000020?ooo`040000003oool0oooo00000340oooo0`0000030?ooo`040000003oool0oooo0000
0080oooo00@000000?ooo`3oool00000<`3oool2000000<0oooo00@000000?ooo`3oool000000P3o
ool010000000oooo0?ooo`00000h0?ooo`008@3oool00`000000oooo0?ooo`0j0?ooo`040000003o
ool0oooo00000080oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`0`0?ooo`03
0000003oool0000000<0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`00000a
0?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?oo
o`00000b0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool000000P3oool01000
0000oooo0?ooo`00000h0?ooo`008@3oool00`000000oooo0?ooo`0g0?ooo`040000003oool0oooo
00000080oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`00000c0?ooo`800000
0`3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo00000340oooo00<000000?oo
o`3oool00`3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo00000380oooo00@0
00000?ooo`3oool000000P3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo0000
03P0oooo000Q0?ooo`030000003oool0oooo03P0oooo0P0000040?ooo`800000103oool2000003D0
oooo00<000000?ooo`3oool00P3oool2000000@0oooo0P00000c0?ooo`<000000`3oool2000000@0
oooo0P00000d0?ooo`800000103oool2000000@0oooo0P00000i0?ooo`008@3oool00`000000oooo
0?ooo`3o0?ooodD0oooo000Q0?ooo`030000003oool0oooo0?l0ooooA@3oool001X0ooooo`00000n
00000140oooo000Q0?ooo`8000003`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
00d0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
00d0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo
00d0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
00d0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
00h0oooo00<000000?ooo`3oool0703oool00240oooo1@00000m0?ooo`030000003oool0oooo03l0
oooo00<000000?ooo`3oool0?P3oool00`000000oooo0?ooo`0o0?ooo`030000003oool0oooo03d0
oooo000Q0?ooo`L00000o`3ooom10?ooo`008@3oool010000000oooo0?ooo`3oool500000?l0oooo
?`3oool00240oooo00<000000?ooo`3oool00`3oool400000?l0oooo?P3oool00240oooo0P000006
0?ooo`@00000o`3oooll0?ooo`008@3oool00`000000oooo0?ooo`070?ooo`@00000o`3ooolj0?oo
o`008@3oool00`000000oooo0?ooo`090?ooo`<00000o`3oooli0?ooo`008@3oool00`000000oooo
0?ooo`0:0?ooo`@00000o`3ooolg0?ooo`008@3oool00`000000oooo0?ooo`0<0?ooo`@00000o`3o
oole0?ooo`008@3oool00`000000oooo0?ooo`0=0?ooo`@00000o`3ooold0?ooo`008@3oool20000
0100oooo1000003o0?oooc80oooo000Q0?ooo`030000003oool0oooo0140oooo1000003o0?oooc00
oooo000Q0?ooo`030000003oool0oooo0180oooo1000003o0?ooobl0oooo000Q0?ooo`030000003o
ool0oooo01@0oooo1000003o0?ooobd0oooo000Q0?ooo`030000003oool0oooo01H0oooo1000003o
0?ooob/0oooo000Q0?ooo`030000003oool0oooo01L0oooo1000003o0?ooobX0oooo000Q0?ooo`03
0000003oool0oooo01T0oooo1000003o0?ooobP0oooo000Q0?ooo`800000703oool400000?l0oooo
9P3oool00240oooo00<000000?ooo`3oool0703oool400000?l0oooo9@3oool00240oooo00<00000
0?ooo`3oool07P3oool400000?l0oooo8`3oool00240oooo00<000000?ooo`3oool07`3oool40000
0?l0oooo8P3oool00240oooo00<000000?ooo`3oool08@3oool400000?l0oooo803oool00240oooo
00<000000?ooo`3oool08`3oool400000?l0oooo7P3oool00240oooo0P00000U0?ooo`@00000o`3o
oolM0?ooo`008@3oool00`000000oooo0?ooo`0V0?ooo`@00000o`3ooolK0?ooo`008@3oool00`00
0000oooo0?ooo`0W0?ooo`@00000o`3ooolJ0?ooo`008@3oool00`000000oooo0?ooo`0Y0?ooo`@0
0000o`3ooolH0?ooo`008@3oool00`000000oooo0?ooo`0[0?ooo`@00000o`3ooolF0?ooo`004P3o
ool4000000<0oooo0P0000060?ooo`030000003oool0oooo02`0oooo1000003o0?oooaD0oooo000D
0?ooo`050000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`030?ooo`<00000;P3o
ool400000?l0oooo4`3oool001@0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool0;`3oool600000?l0oooo403oool001@0oooo00D00000
0?ooo`3oool0oooo000000020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0<@3o
ool400000?l0oooo403oool001@0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool0<`3oool200000?l0oooo403oool00180oooo0`000003
0?ooo`040000003oool0oooo000000D0oooo00<000000?ooo`3oool0=03oool400000?l0oooo3@3o
ool001@0oooo00<000000?ooo`3oool00P3oool2000000H0oooo00<000000?ooo`3oool0=@3oool5
00000?l0oooo2`3oool00240oooo0P00000h0?ooo`D00000o`3oool90?ooo`008@3oool00`000000
oooo0?ooo`0i0?ooo`@00000o`3oool80?ooo`008@3oool00`000000oooo0?ooo`0k0?ooo`@00000
o`3oool60?ooo`008@3oool00`000000oooo0?ooo`0l0?ooo`@00000o`3oool50?ooo`008@3oool0
0`000000oooo0?ooo`0n0?ooo`@00000o`3oool30?ooo`008@3oool00`000000oooo0?ooo`0o0?oo
o`D00000o`3oool10?ooo`008@3oool200000480oooo1000003o0?ooo`008@3oool00`000000oooo
0?ooo`130?ooo`@00000o@3oool00240oooo00<000000?ooo`3oool0A03oool400000?`0oooo000Q
0?ooo`030000003oool0oooo04H0oooo1000003j0?ooo`008@3oool00`000000oooo0?ooo`180?oo
o`@00000n03oool00240oooo00<000000?ooo`3oool0B@3oool400000?L0oooo000Q0?ooo`030000
003oool0oooo04/0oooo1000003e0?ooo`008@3oool2000004h0oooo0P00003e0?ooo`008@3oool0
0`000000oooo0?ooo`1>0?ooo`@00000lP3oool00240oooo00<000000?ooo`3oool0C`3oool40000
0?40oooo000Q0?ooo`030000003oool0oooo0540oooo1000003_0?ooo`008@3oool00`000000oooo
0?ooo`1C0?ooo`@00000k@3oool00240oooo00<000000?ooo`3oool0E03oool400000>`0oooo000Q
0?ooo`800000E`3oool400000>X0oooo000Q0?ooo`030000003oool0oooo05P0oooo1000003X0?oo
o`008@3oool00`000000oooo0?ooo`1I0?ooo`D00000iP3oool00240oooo00<000000?ooo`3oool0
F`3oool400000>D0oooo000Q0?ooo`030000003oool0oooo05d0oooo1000003S0?ooo`004P3oool4
000000<0oooo0P0000060?ooo`030000003oool0oooo05h0oooo1000003R0?ooo`004P3oool00`00
0000oooo0?ooo`030?ooo`040000003oool0oooo000000D0oooo0`00001P0?ooo`@00000h03oool0
01<0oooo00<000000?ooo`3oool00P3oool010000000oooo0?ooo`0000050?ooo`030000003oool0
oooo0680oooo1000003N0?ooo`00503oool01@000000oooo0?ooo`3oool000000080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`1S0?ooo`D00000g03oool001D0oooo00@000000?oo
o`3oool000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo06D0oooo1000003K
0?ooo`004P3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000000D0oooo00<0
00000?ooo`3oool0I`3oool300000=X0oooo000C0?ooo`800000103oool2000000H0oooo00<00000
0?ooo`3oool0J03oool400000=P0oooo000Q0?ooo`800000J`3oool400000=H0oooo000Q0?ooo`03
0000003oool0oooo06/0oooo1000003E0?ooo`008@3oool00`000000oooo0?ooo`1]0?ooo`@00000
d`3oool00240oooo00<000000?ooo`3oool0K`3oool400000=40oooo000Q0?ooo`030000003oool0
oooo0740oooo1000003?0?ooo`008@3oool00`000000oooo0?ooo`1c0?ooo`<00000cP3oool00240
oooo0P00001e0?ooo`@00000c03oool00240oooo00<000000?ooo`3oool0MP3oool300000</0oooo
000Q0?ooo`030000003oool0oooo07L0oooo100000390?ooo`008@3oool00`000000oooo0?ooo`1i
0?ooo`@00000a`3oool00240oooo00<000000?ooo`3oool0N`3oool300000<H0oooo000Q0?ooo`03
0000003oool0oooo07`0oooo100000340?ooo`008@3oool00`000000oooo0?ooo`1n0?ooo`@00000
`P3oool00240oooo0P0000200?ooo`@00000`@3oool00240oooo00<000000?ooo`3oool0P@3oool4
00000;l0oooo000Q0?ooo`030000003oool0oooo08<0oooo1000002m0?ooo`008@3oool00`000000
oooo0?ooo`240?ooo`@00000_03oool00240oooo00<000000?ooo`3oool0QP3oool400000;X0oooo
000Q0?ooo`030000003oool0oooo08P0oooo1000002h0?ooo`008@3oool2000008X0oooo1@00002f
0?ooo`008@3oool00`000000oooo0?ooo`2<0?ooo`@00000]03oool00240oooo00<000000?ooo`3o
ool0S@3oool700000;00oooo000Q0?ooo`030000003oool0oooo08l0oooo1P00002_0?ooo`008@3o
ool00`000000oooo0?ooo`2C0?ooo`800000[`3oool001<0oooo0P0000040?ooo`8000001P3oool0
0`000000oooo0?ooo`3o0?ooodD0oooo000B0?ooo`040000003oool0oooo00000080oooo00@00000
0?ooo`3oool000001@3oool3000009<0oooo0P00002_0?ooo`005@3oool010000000oooo0?ooo`00
00020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0U03oool00`000000oooo0?oo
o`2]0?ooo`004`3oool2000000<0oooo00@000000?ooo`3oool000001@3oool00`000000oooo0?oo
o`3o0?ooodD0oooo000E0?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`2C0?ooo`D00000[03oool00180oooo00@000000?ooo`3oool000000P3o
ool010000000oooo0?ooo`0000050?ooo`030000003oool0oooo09@0oooo1`00002Y0?ooo`004`3o
ool2000000@0oooo0P0000060?ooo`030000003oool0oooo09L0oooo1P00002W0?ooo`008@3oool2
000009/0oooo1000002V0?ooo`008@3oool00`000000oooo0?ooo`2L0?ooo`@00000Y03oool00240
oooo00<000000?ooo`3oool0WP3oool400000:80oooo000Q0?ooo`030000003oool0oooo0:00oooo
1000002P0?ooo`008@3oool00`000000oooo0?ooo`2Q0?ooo`D00000WP3oool00240oooo00<00000
0?ooo`3oool0X`3oool4000009d0oooo000Q0?ooo`030000003oool0oooo0:D0oooo1000002K0?oo
o`008@3oool200000:L0oooo1000002J0?ooo`008@3oool00`000000oooo0?ooo`2X0?ooo`@00000
V03oool00240oooo00<000000?ooo`3oool0Z@3oool5000009H0oooo000Q0?ooo`030000003oool0
oooo0:/0oooo1000002E0?ooo`008@3oool00`000000oooo0?ooo`2]0?ooo`D00000TP3oool00240
oooo00<000000?ooo`3oool0[`3oool400000940oooo000Q0?ooo`800000/P3oool300000900oooo
000Q0?ooo`030000003oool0oooo0;80oooo1000002>0?ooo`008@3oool00`000000oooo0?ooo`2d
0?ooo`<00000S@3oool00240oooo00<000000?ooo`3oool0]@3oool4000008/0oooo000Q0?ooo`03
0000003oool0oooo0;L0oooo100000290?ooo`008@3oool00`000000oooo0?ooo`2i0?ooo`@00000
Q`3oool00240oooo0P00002k0?ooo`@00000QP3oool00240oooo00<000000?ooo`3oool0_03oool4
000008@0oooo000Q0?ooo`030000003oool0oooo0;h0oooo0`0000230?ooo`008@3oool00`000000
oooo0?ooo`2o0?ooo`@00000P@3oool00240oooo00<000000?ooo`3oool0`@3oool4000007l0oooo
000D0?ooo`030000003oool0oooo0080oooo0P0000060?ooo`030000003oool0oooo0<80oooo1000
001n0?ooo`00503oool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`3oool00`3o
ool300000<@0oooo1000001l0?ooo`004@3oool500000080oooo00@000000?ooo`3oool000001@3o
ool00`000000oooo0?ooo`360?ooo`@00000NP3oool00140oooo00@000000?ooo`3oool000000`3o
ool010000000oooo0?ooo`0000050?ooo`030000003oool0oooo0<L0oooo1000001i0?ooo`004P3o
ool00`000000oooo000000030?ooo`040000003oool0oooo000000D0oooo00<000000?ooo`3oool0
b@3oool4000007L0oooo000C0?ooo`8000000`3oool010000000oooo0?ooo`0000050?ooo`030000
003oool0oooo0</0oooo1000001e0?ooo`00503oool00`000000oooo0?ooo`020?ooo`8000001P3o
ool00`000000oooo0?ooo`3<0?ooo`@00000M03oool00240oooo0P00003?0?ooo`@00000LP3oool0
0240oooo00<000000?ooo`3oool0d03oool400000700oooo000Q0?ooo`030000003oool0oooo0=80
oooo0`00001_0?ooo`008@3oool00`000000oooo0?ooo`3C0?ooo`@00000K@3oool00240oooo00<0
00000?ooo`3oool0e03oool5000006/0oooo000Q0?ooo`030000003oool0oooo0=H0oooo1@00001Y
0?ooo`008@3oool00`000000oooo0?ooo`3H0?ooo`<00000J@3oool00240oooo0P00003J0?ooo`@0
0000I`3oool00240oooo00<000000?ooo`3oool0fP3oool5000006D0oooo000Q0?ooo`030000003o
ool0oooo0=d0oooo1000001S0?ooo`008@3oool00`000000oooo0?ooo`3N0?ooo`@00000HP3oool0
0240oooo00<000000?ooo`3oool0h03oool400000600oooo000Q0?ooo`030000003oool0oooo0>40
oooo1@00001N0?ooo`008@3oool200000>@0oooo1000001M0?ooo`008@3oool00`000000oooo0?oo
o`3U0?ooo`@00000F`3oool00240oooo00<000000?ooo`3oool0iP3oool4000005X0oooo000Q0?oo
o`030000003oool0oooo0>P0oooo1000001H0?ooo`008@3oool00`000000oooo0?ooo`3Y0?ooo`@0
0000E`3oool00240oooo00<000000?ooo`3oool0j`3oool4000005D0oooo000Q0?ooo`800000kP3o
ool4000005<0oooo000Q0?ooo`030000003oool0oooo0>l0oooo1000001A0?ooo`008@3oool00`00
0000oooo0?ooo`3`0?ooo`D00000C`3oool00240oooo00<000000?ooo`3oool0lP3oool3000004l0
oooo000Q0?ooo`030000003oool0oooo0?<0oooo0`00001>0?ooo`004P3oool3000000@0oooo0P00
00060?ooo`030000003oool0oooo0?@0oooo1000001<0?ooo`005@3oool010000000oooo0?ooo`00
00020?ooo`030000003oool0oooo00<0oooo0`00003e0?ooo`D00000BP3oool001D0oooo00@00000
0?ooo`3oool000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0?P0oooo1000
00180?ooo`004P3oool3000000<0oooo00@000000?ooo`3oool000001@3oool00`000000oooo0?oo
o`3j0?ooo`@00000AP3oool00180oooo00<000000?ooo`3oool00`3oool010000000oooo0?ooo`00
00050?ooo`030000003oool0oooo0?/0oooo100000150?ooo`004P3oool00`000000oooo0?ooo`03
0?ooo`040000003oool0oooo000000D0oooo00<000000?ooo`3oool0o@3oool4000004<0oooo000B
0?ooo`@000000`3oool2000000H0oooo00<000000?ooo`3oool0oP3oool400000480oooo000Q0?oo
o`800000o`3oool20?ooo`@00000@03oool00240oooo00<000000?ooo`3oool0o`3oool20?ooo`@0
0000?`3oool00240oooo00<000000?ooo`3oool0o`3oool40?ooo`@00000?@3oool00240oooo00<0
00000?ooo`3oool0o`3oool60?ooo`@00000>`3oool00240oooo00<000000?ooo`3oool0o`3oool8
0?ooo`<00000>P3oool00240oooo00<000000?ooo`3oool0o`3oool90?ooo`@00000>03oool00240
oooo00<000000?ooo`3oool0o`3oool;0?ooo`<00000=`3oool00240oooo0P00003o0?ooo`d0oooo
1000000e0?ooo`008@3oool00`000000oooo0?ooo`3o0?ooo`h0oooo1000000c0?ooo`008@3oool0
0`000000oooo0?ooo`3o0?oooa00oooo1000000a0?ooo`008@3oool00`000000oooo0?ooo`3o0?oo
oa40oooo1000000`0?ooo`008@3oool00`000000oooo0?ooo`3o0?oooa<0oooo1000000^0?ooo`00
8@3oool00`000000oooo0?ooo`3o0?oooa@0oooo1000000]0?ooo`008@3oool200000?l0oooo5`3o
ool4000002/0oooo000Q0?ooo`030000003oool0oooo0?l0oooo603oool4000002T0oooo000Q0?oo
o`030000003oool0oooo0?l0oooo6@3oool4000002P0oooo000Q0?ooo`030000003oool0oooo0?l0
oooo6`3oool4000002H0oooo000Q0?ooo`030000003oool0oooo0?l0oooo7@3oool4000002@0oooo
000Q0?ooo`030000003oool0oooo0?l0oooo7P3oool4000002<0oooo000Q0?ooo`800000o`3ooolQ
0?ooo`@000008@3oool00240oooo00<000000?ooo`3oool0o`3ooolR0?ooo`@000007`3oool00240
oooo00<000000?ooo`3oool0o`3ooolT0?ooo`@000007@3oool00240oooo00<000000?ooo`3oool0
o`3ooolU0?ooo`D000006`3oool00240oooo00<000000?ooo`3oool0o`3ooolW0?ooo`D000006@3o
ool001<0oooo0P0000040?ooo`8000001P3oool00`000000oooo0?ooo`3o0?ooobT0oooo0`00000I
0?ooo`004P3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000000D0oooo0`00
003o0?ooob/0oooo0P00000H0?ooo`004P3oool010000000oooo0?ooo`0000020?ooo`040000003o
ool0oooo000000D0oooo00<000000?ooo`3oool0o`3oool/0?ooo`030000003oool0oooo01H0oooo
000B0?ooo`<000000`3oool010000000oooo0?ooo`0000050?ooo`030000003oool0oooo0?l0oooo
A@3oool00180oooo00<000000?ooo`3oool00`3oool010000000oooo0?ooo`0000050?ooo`030000
003oool0oooo0?l0ooooA@3oool00180oooo00<000000?ooo`3oool00`3oool010000000oooo0?oo
o`0000050?ooo`030000003oool0oooo0?l0ooooA@3oool001<0oooo0`0000030?ooo`8000001P3o
ool00`000000oooo0?ooo`3o0?ooodD0oooo003o0?ooofT0oooo003o0?ooofT0oooo003o0?ooofT0
oooo003o0?ooofT0oooo003o0?ooofT0oooo003o0?ooofT0oooo003>0?ooo`<00000U`3oool00=40
oooo00<000000?ooo`3oool0U03oool008/0oooo0`0000020?ooo`030000003oool0oooo00800000
00<0oooo0000003oool00P3oool200000080oooo1@0000020?ooo`8000000P3oool3000000030?oo
o`00000000000080oooo0P0000000`3oool000000?ooo`05000000H0oooo0P0000020?ooo`D00000
00<0oooo0000000000000P3oool300000080oooo1@0000020?ooo`<000000P3oool3000008D0oooo
002<0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool000000`3oool010000000
oooo0?ooo`0000030?ooo`050000003oool0oooo0?ooo`0000000P3oool010000000oooo0?ooo`00
00020?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`070?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool000000P3oool01000
0000oooo0?ooo`0000040?ooo`050000003oool0oooo0?ooo`000000203oool00`000000oooo0?oo
o`220?ooo`007`3oool3000006X0oooo0`0000040?ooo`<000000`3oool00`000000oooo0?ooo`04
0?ooo`050000003oool0oooo0?ooo`0000000P3oool010000000oooo0?ooo`0000020?ooo`040000
003oool0oooo0?ooo`<00000103oool00`000000oooo0?ooo`070?ooo`@000000P3oool010000000
oooo0?ooo`0000020?ooo`040000003oool0oooo000000@0oooo00@000000?ooo`3oool0oooo0`00
00040?ooo`800000Q@3oool00280oooo00<000000?ooo`3oool0I`3oool010000000oooo0?ooo`00
00050?ooo`050000003oool0oooo0?ooo`0000001P3oool01@000000oooo0?ooo`3oool000000080
oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool01`3oool010000000oooo0?ooo`0000020?ooo`040000003oool0
oooo00000080oooo00@000000?ooo`3oool00000103oool01@000000oooo0?ooo`3oool000000080
oooo00@000000?ooo`3oool00000Q`3oool00180oooo0P0000020?ooo`D0000000<0oooo00000000
00000P3oool300000080oooo1@0000020?ooo`<00000G@3oool010000000oooo0?ooo`0000030?oo
o`8000000`3oool4000000<0oooo0P0000040?ooo`8000000P3oool010000000oooo000000000004
0?ooo`8000001@3oool00`000000oooo0?ooo`080?ooo`060000003oool000000?ooo`000000oooo
0P0000040?ooo`80000000<0oooo0000003oool00P3oool00`000000oooo0?ooo`020?ooo`800000
103oool3000008@0oooo000C0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool0
00000P3oool010000000oooo0?ooo`0000040?ooo`050000003oool0oooo0?ooo`000000G`3oool4
000000X0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`070?ooo`<000004@3oool0
0`000000oooo0?ooo`2?0?ooo`004`3oool400000080oooo00@000000?ooo`3oool000000P3oool0
10000000oooo0?ooo`0000040?ooo`040000003oool0oooo0?ooo`<00000J`3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo01@0oooo0P00000L0?ooo`800000T@3oool001<0oooo00@0
00000?ooo`3oool000000P3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo0000
00@0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003oool0oooo0?l0oooo>03oool0
01@0oooo00H000000?ooo`000000oooo0000003oool2000000@0oooo0P0000000`3oool000000?oo
o`020?ooo`030000003oool0oooo0080oooo0P00003o0?oooc/0oooo0000\
\>"],
  ImageRangeCache->{{{0, 359}, {221.375, 0}} -> {-102.69, -4.17635, 3.04603, 
  0.32248}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.015873 0.0147151 0.0118028 [
[.18254 .00222 -6 -9 ]
[.18254 .00222 6 0 ]
[.34127 .00222 -6 -9 ]
[.34127 .00222 6 0 ]
[.5 .00222 -6 -9 ]
[.5 .00222 6 0 ]
[.65873 .00222 -6 -9 ]
[.65873 .00222 6 0 ]
[.81746 .00222 -6 -9 ]
[.81746 .00222 6 0 ]
[.97619 .00222 -6 -9 ]
[.97619 .00222 6 0 ]
[1.025 .01472 0 -4.90625 ]
[1.025 .01472 34 4.90625 ]
[.01131 .13274 -12 -4.5 ]
[.01131 .13274 0 4.5 ]
[.01131 .25077 -12 -4.5 ]
[.01131 .25077 0 4.5 ]
[.01131 .3688 -12 -4.5 ]
[.01131 .3688 0 4.5 ]
[.01131 .48683 -12 -4.5 ]
[.01131 .48683 0 4.5 ]
[.01131 .60486 -12 -4.5 ]
[.01131 .60486 0 4.5 ]
[.02381 .64303 -26 0 ]
[.02381 .64303 26 10.125 ]
[.5 .63053 -77 0 ]
[.5 .63053 77 10.125 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.18254 .01472 m
.18254 .02097 L
s
[(10)] .18254 .00222 0 1 Mshowa
.34127 .01472 m
.34127 .02097 L
s
[(20)] .34127 .00222 0 1 Mshowa
.5 .01472 m
.5 .02097 L
s
[(30)] .5 .00222 0 1 Mshowa
.65873 .01472 m
.65873 .02097 L
s
[(40)] .65873 .00222 0 1 Mshowa
.81746 .01472 m
.81746 .02097 L
s
[(50)] .81746 .00222 0 1 Mshowa
.97619 .01472 m
.97619 .02097 L
s
[(60)] .97619 .00222 0 1 Mshowa
.125 Mabswid
.05556 .01472 m
.05556 .01847 L
s
.0873 .01472 m
.0873 .01847 L
s
.11905 .01472 m
.11905 .01847 L
s
.15079 .01472 m
.15079 .01847 L
s
.21429 .01472 m
.21429 .01847 L
s
.24603 .01472 m
.24603 .01847 L
s
.27778 .01472 m
.27778 .01847 L
s
.30952 .01472 m
.30952 .01847 L
s
.37302 .01472 m
.37302 .01847 L
s
.40476 .01472 m
.40476 .01847 L
s
.43651 .01472 m
.43651 .01847 L
s
.46825 .01472 m
.46825 .01847 L
s
.53175 .01472 m
.53175 .01847 L
s
.56349 .01472 m
.56349 .01847 L
s
.59524 .01472 m
.59524 .01847 L
s
.62698 .01472 m
.62698 .01847 L
s
.69048 .01472 m
.69048 .01847 L
s
.72222 .01472 m
.72222 .01847 L
s
.75397 .01472 m
.75397 .01847 L
s
.78571 .01472 m
.78571 .01847 L
s
.84921 .01472 m
.84921 .01847 L
s
.88095 .01472 m
.88095 .01847 L
s
.9127 .01472 m
.9127 .01847 L
s
.94444 .01472 m
.94444 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
gsave
1.025 .01472 -61 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Angle) show
93.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.02381 .13274 m
.03006 .13274 L
s
[(10)] .01131 .13274 1 0 Mshowa
.02381 .25077 m
.03006 .25077 L
s
[(20)] .01131 .25077 1 0 Mshowa
.02381 .3688 m
.03006 .3688 L
s
[(30)] .01131 .3688 1 0 Mshowa
.02381 .48683 m
.03006 .48683 L
s
[(40)] .01131 .48683 1 0 Mshowa
.02381 .60486 m
.03006 .60486 L
s
[(50)] .01131 .60486 1 0 Mshowa
.125 Mabswid
.02381 .03832 m
.02756 .03832 L
s
.02381 .06193 m
.02756 .06193 L
s
.02381 .08553 m
.02756 .08553 L
s
.02381 .10914 m
.02756 .10914 L
s
.02381 .15635 m
.02756 .15635 L
s
.02381 .17995 m
.02756 .17995 L
s
.02381 .20356 m
.02756 .20356 L
s
.02381 .22717 m
.02756 .22717 L
s
.02381 .27438 m
.02756 .27438 L
s
.02381 .29798 m
.02756 .29798 L
s
.02381 .32159 m
.02756 .32159 L
s
.02381 .34519 m
.02756 .34519 L
s
.02381 .3924 m
.02756 .3924 L
s
.02381 .41601 m
.02756 .41601 L
s
.02381 .43962 m
.02756 .43962 L
s
.02381 .46322 m
.02756 .46322 L
s
.02381 .51043 m
.02756 .51043 L
s
.02381 .53404 m
.02756 .53404 L
s
.02381 .55764 m
.02756 .55764 L
s
.02381 .58125 m
.02756 .58125 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
gsave
.02381 .64303 -87 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.125000 moveto
 558.000000 14.125000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.125000 lineto
closepath
clip newpath
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Distance) show
111.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
gsave
.5 .63053 -138 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.125000 moveto
 558.000000 14.125000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.125000 lineto
closepath
clip newpath
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Angle) show
99.000000 11.562500 moveto
(vs) show
111.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(.) show
123.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Repeat) show
165.000000 11.562500 moveto
(Distance) show
213.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.008 w
.02381 .03516 Mdot
.5 .02652 Mdot
.19672 .04594 Mdot
.97619 .03516 Mdot
.12836 .06616 Mdot
.2794 .05727 Mdot
.80328 .04594 Mdot
.09865 .08651 Mdot
.32709 .0688 Mdot
.7206 .05727 Mdot
.08207 .1069 Mdot
.15418 .09733 Mdot
.24441 .08842 Mdot
.35796 .08043 Mdot
.67291 .0688 Mdot
.87164 .06616 Mdot
.0715 .12731 Mdot
.37954 .09211 Mdot
.64204 .08043 Mdot
.06417 .14773 Mdot
.11105 .13794 Mdot
.16585 .12854 Mdot
.23022 .11962 Mdot
.30609 .11133 Mdot
.39545 .10382 Mdot
.62046 .09211 Mdot
.75559 .08842 Mdot
.90135 .08651 Mdot
.0588 .16815 Mdot
.14427 .14877 Mdot
.25722 .13096 Mdot
.40766 .11556 Mdot
.60455 .10382 Mdot
.84582 .09733 Mdot
.05468 .18858 Mdot
.08933 .17868 Mdot
.17249 .15975 Mdot
.22254 .15083 Mdot
.34402 .1345 Mdot
.41732 .12731 Mdot
.59234 .11556 Mdot
.69391 .11133 Mdot
.91793 .1069 Mdot
.05144 .20901 Mdot
.11615 .18938 Mdot
.29793 .15387 Mdot
.42516 .13907 Mdot
.58268 .12731 Mdot
.76978 .11962 Mdot
.04881 .22945 Mdot
.07626 .21949 Mdot
.10649 .20973 Mdot
.13986 .20021 Mdot
.17678 .19097 Mdot
.21772 .18205 Mdot
.26316 .17351 Mdot
.31363 .1654 Mdot
.36963 .15782 Mdot
.43164 .15083 Mdot
.57484 .13907 Mdot
.65598 .1345 Mdot
.74278 .13096 Mdot
.83415 .12854 Mdot
.9285 .12731 Mdot
.04663 .24988 Mdot
.16094 .21115 Mdot
.23608 .19332 Mdot
.43709 .1626 Mdot
.56836 .15083 Mdot
.88895 .13794 Mdot
.04481 .27032 Mdot
.06753 .26032 Mdot
.09217 .25048 Mdot
.11893 .24082 Mdot
.14805 .23139 Mdot
.17978 .22219 Mdot
.21442 .21327 Mdot
.25225 .20466 Mdot
.29359 .19642 Mdot
.33876 .18858 Mdot
.38805 .18121 Mdot
.44174 .17438 Mdot
.56291 .1626 Mdot
.63037 .15782 Mdot
.70207 .15387 Mdot
.77746 .15083 Mdot
.85573 .14877 Mdot
.93583 .14773 Mdot
.04325 .29076 Mdot
.08672 .27087 Mdot
.13736 .25165 Mdot
.26659 .21606 Mdot
.34896 .20021 Mdot
.44575 .18616 Mdot
.55826 .17438 Mdot
.68637 .1654 Mdot
.82751 .15975 Mdot
.04191 .3112 Mdot
.06129 .30116 Mdot
.10438 .28152 Mdot
.182 .25341 Mdot
.21201 .24449 Mdot
.31718 .21949 Mdot
.40193 .20466 Mdot
.44924 .19794 Mdot
.55425 .18616 Mdot
.61195 .18121 Mdot
.73684 .17351 Mdot
.9412 .16815 Mdot
.04075 .33164 Mdot
.07806 .31167 Mdot
.12068 .29227 Mdot
.16957 .27357 Mdot
.22588 .25574 Mdot
.2909 .23897 Mdot
.36596 .22353 Mdot
.45231 .20973 Mdot
.55076 .19794 Mdot
.66124 .18858 Mdot
.78228 .18205 Mdot
.91067 .17868 Mdot
.03972 .35208 Mdot
.05661 .34202 Mdot
.07457 .33208 Mdot
.09368 .32227 Mdot
.11404 .3126 Mdot
.13576 .3031 Mdot
.15894 .29377 Mdot
.1837 .28464 Mdot
.21018 .27571 Mdot
.23851 .26703 Mdot
.26883 .25861 Mdot
.30127 .25048 Mdot
.33599 .24267 Mdot
.37311 .23521 Mdot
.41276 .22815 Mdot
.45503 .22152 Mdot
.54769 .20973 Mdot
.59807 .20466 Mdot
.65104 .20021 Mdot
.70641 .19642 Mdot
.76392 .19332 Mdot
.82322 .19097 Mdot
.88385 .18938 Mdot
.94532 .18858 Mdot
.03881 .37252 Mdot
.10826 .33295 Mdot
.14974 .314 Mdot
.25006 .27837 Mdot
.31068 .26201 Mdot
.45746 .23331 Mdot
.54497 .22152 Mdot
.74775 .20466 Mdot
.86014 .20021 Mdot
.038 .39296 Mdot
.05297 .38288 Mdot
.06878 .37291 Mdot
.08549 .36305 Mdot
.10316 .35331 Mdot
.12188 .34371 Mdot
.14171 .33426 Mdot
.16274 .32497 Mdot
.18505 .31586 Mdot
.20874 .30694 Mdot
.23391 .29823 Mdot
.26065 .28975 Mdot
.28906 .28152 Mdot
.31925 .27357 Mdot
.35132 .26592 Mdot
.38535 .25861 Mdot
.42143 .25165 Mdot
.45964 .2451 Mdot
.54254 .23331 Mdot
.58724 .22815 Mdot
.63404 .22353 Mdot
.68282 .21949 Mdot
.73341 .21606 Mdot
.78558 .21327 Mdot
.83906 .21115 Mdot
.89351 .20973 Mdot
.94856 .20901 Mdot
.03727 .4134 Mdot
.06635 .39333 Mdot
.13464 .35455 Mdot
.17485 .336 Mdot
.21988 .31816 Mdot
.2704 .30116 Mdot
.39063 .27032 Mdot
.4616 .25689 Mdot
.54036 .2451 Mdot
.62689 .23521 Mdot
.82022 .22219 Mdot
.92374 .21949 Mdot
.03662 .43384 Mdot
.05006 .42375 Mdot
.07901 .40385 Mdot
.09462 .39406 Mdot
.1466 .36544 Mdot
.18615 .34709 Mdot
.20758 .33816 Mdot
.25413 .32091 Mdot
.33428 .29675 Mdot
.36405 .28924 Mdot
.42855 .27518 Mdot
.46339 .26868 Mdot
.5384 .25689 Mdot
.57857 .25165 Mdot
.66401 .24267 Mdot
.7091 .23897 Mdot
.85195 .23139 Mdot
.95119 .22945 Mdot
.03602 .45428 Mdot
.06221 .43417 Mdot
.091 .41445 Mdot
.12276 .39516 Mdot
.15785 .37639 Mdot
.23985 .34074 Mdot
.28773 .32407 Mdot
.34091 .30837 Mdot
.39986 .29377 Mdot
.46501 .28047 Mdot
.53661 .26868 Mdot
.61465 .25861 Mdot
.69873 .25048 Mdot
.78799 .24449 Mdot
.88107 .24082 Mdot
.03548 .47472 Mdot
.04767 .46462 Mdot
.06042 .4546 Mdot
.07377 .44467 Mdot
.08774 .43484 Mdot
.10237 .42511 Mdot
.11772 .41549 Mdot
.13381 .40599 Mdot
.1507 .39662 Mdot
.16843 .38739 Mdot
.18706 .37831 Mdot
.20663 .36939 Mdot
.2272 .36064 Mdot
.24882 .35208 Mdot
.27156 .34371 Mdot
.29547 .33557 Mdot
.32061 .32766 Mdot
.34703 .32 Mdot
.37478 .3126 Mdot
.40392 .30551 Mdot
.43448 .29872 Mdot
.4665 .29227 Mdot
.53499 .28047 Mdot
.57145 .27518 Mdot
.60937 .27032 Mdot
.64868 .26592 Mdot
.68932 .26201 Mdot
.73117 .25861 Mdot
.77412 .25574 Mdot
.818 .25341 Mdot
.86264 .25165 Mdot
.90783 .25048 Mdot
.95337 .24988 Mdot
.03498 .49516 Mdot
.08477 .45523 Mdot
.11317 .43583 Mdot
.17841 .39844 Mdot
.21593 .3806 Mdot
.30267 .34709 Mdot
.3527 .33164 Mdot
.46787 .30406 Mdot
.5335 .29227 Mdot
.68075 .27357 Mdot
.76149 .26703 Mdot
.93247 .26032 Mdot
.03452 .5156 Mdot
.04568 .50549 Mdot
.05731 .49545 Mdot
.06943 .4855 Mdot
.09526 .46585 Mdot
.10904 .45618 Mdot
.12342 .44661 Mdot
.13846 .43715 Mdot
.17062 .41861 Mdot
.18782 .40954 Mdot
.20583 .40061 Mdot
.22467 .39185 Mdot
.26508 .37485 Mdot
.28673 .36663 Mdot
.30939 .35862 Mdot
.33312 .35084 Mdot
.38395 .336 Mdot
.41112 .32899 Mdot
.4395 .32227 Mdot
.46913 .31586 Mdot
.53213 .30406 Mdot
.56552 .29872 Mdot
.60014 .29377 Mdot
.63595 .28924 Mdot
.71094 .28152 Mdot
.74994 .27837 Mdot
.78982 .27571 Mdot
.83043 .27357 Mdot
.91328 .27087 Mdot
.95519 .27032 Mdot
.0341 .53605 Mdot
.05594 .51588 Mdot
.0796 .49603 Mdot
.10527 .47654 Mdot
.13318 .45744 Mdot
.16357 .4388 Mdot
.23291 .40313 Mdot
.27246 .38627 Mdot
.31567 .37017 Mdot
.36287 .35496 Mdot
.41433 .34074 Mdot
.47029 .32766 Mdot
.53087 .31586 Mdot
.59608 .30551 Mdot
.66572 .29675 Mdot
.73935 .28975 Mdot
.8163 .28464 Mdot
.89562 .28152 Mdot
.03372 .55649 Mdot
.044 .54637 Mdot
.06579 .52634 Mdot
.07733 .51644 Mdot
.10182 .4969 Mdot
.11482 .48727 Mdot
.14247 .46832 Mdot
.15717 .45901 Mdot
.18847 .44076 Mdot
.20514 .43184 Mdot
.24069 .41445 Mdot
.25963 .40599 Mdot
.30003 .38963 Mdot
.32156 .38174 Mdot
.36745 .36663 Mdot
.39188 .35943 Mdot
.44381 .34583 Mdot
.47137 .33945 Mdot
.52971 .32766 Mdot
.5605 .32227 Mdot
.62522 .3126 Mdot
.65909 .30837 Mdot
.7296 .30116 Mdot
.76609 .29823 Mdot
.84106 .29377 Mdot
.87932 .29227 Mdot
.95675 .29076 Mdot
.03336 .57693 Mdot
.05352 .55675 Mdot
.07523 .53685 Mdot
.12395 .49805 Mdot
.15132 .47924 Mdot
.18097 .46089 Mdot
.21313 .44304 Mdot
.24803 .42578 Mdot
.28593 .40918 Mdot
.37174 .37831 Mdot
.42011 .36425 Mdot
.47237 .35125 Mdot
.52863 .33945 Mdot
.58888 .32899 Mdot
.65297 .32 Mdot
.79126 .30694 Mdot
.86424 .3031 Mdot
.93871 .30116 Mdot
.03302 .59737 Mdot
.04256 .58724 Mdot
.05244 .57718 Mdot
.06268 .56718 Mdot
.0733 .55726 Mdot
.08431 .54741 Mdot
.09573 .53765 Mdot
.10758 .52797 Mdot
.11989 .51838 Mdot
.13268 .50888 Mdot
.14596 .49949 Mdot
.15976 .49021 Mdot
.17411 .48104 Mdot
.18904 .47199 Mdot
.20456 .46307 Mdot
.2207 .45428 Mdot
.2375 .44564 Mdot
.25498 .43715 Mdot
.27317 .42882 Mdot
.2921 .42067 Mdot
.3118 .4127 Mdot
.33229 .40492 Mdot
.3536 .39735 Mdot
.37576 .39 Mdot
.39879 .38288 Mdot
.42272 .376 Mdot
.44755 .36939 Mdot
.47331 .36305 Mdot
.52763 .35125 Mdot
.55619 .34583 Mdot
.58567 .34074 Mdot
.61605 .336 Mdot
.6473 .33164 Mdot
.67939 .32766 Mdot
.71227 .32407 Mdot
.74587 .32091 Mdot
.78012 .31816 Mdot
.81495 .31586 Mdot
.85026 .314 Mdot
.88596 .3126 Mdot
.92194 .31167 Mdot
.95809 .3112 Mdot
.09303 .55803 Mdot
.14103 .51976 Mdot
.16783 .50121 Mdot
.2279 .46555 Mdot
.26157 .44854 Mdot
.33719 .41653 Mdot
.47418 .37485 Mdot
.52669 .36305 Mdot
.70453 .33557 Mdot
.83726 .32497 Mdot
.90632 .32227 Mdot
.06983 .59809 Mdot
.08 .58821 Mdot
.09052 .57841 Mdot
.10141 .56869 Mdot
.11269 .55905 Mdot
.12437 .5495 Mdot
.13648 .54004 Mdot
.14903 .53067 Mdot
.16204 .52141 Mdot
.17553 .51226 Mdot
.18953 .50321 Mdot
.20405 .49429 Mdot
.21911 .4855 Mdot
.23475 .47684 Mdot
.25097 .46832 Mdot
.26782 .45995 Mdot
.2853 .45174 Mdot
.30344 .44369 Mdot
.32227 .43583 Mdot
.34181 .42815 Mdot
.36208 .42067 Mdot
.38309 .4134 Mdot
.40488 .40635 Mdot
.42745 .39953 Mdot
.45082 .39296 Mdot
.475 .38664 Mdot
.52582 .37485 Mdot
.55245 .36939 Mdot
.57989 .36425 Mdot
.60812 .35943 Mdot
.63713 .35496 Mdot
.66688 .35084 Mdot
.69733 .34709 Mdot
.72844 .34371 Mdot
.76015 .34074 Mdot
.79242 .33816 Mdot
.82515 .336 Mdot
.85829 .33426 Mdot
.89174 .33295 Mdot
.92543 .33208 Mdot
.95925 .33164 Mdot
.08818 .5988 Mdot
.10948 .5794 Mdot
.13227 .56033 Mdot
.1567 .54163 Mdot
.1829 .52333 Mdot
.21104 .50549 Mdot
.24127 .48815 Mdot
.27375 .47138 Mdot
.30867 .45523 Mdot
.34618 .43978 Mdot
.38645 .42511 Mdot
.42961 .4113 Mdot
.47577 .39844 Mdot
.525 .38664 Mdot
.57728 .376 Mdot
.63255 .36663 Mdot
.69061 .35862 Mdot
.75118 .35208 Mdot
.81385 .34709 Mdot
.87812 .34371 Mdot
.94339 .34202 Mdot
.11724 .59015 Mdot
.15175 .56186 Mdot
.16405 .55262 Mdot
.18996 .53444 Mdot
.2036 .52552 Mdot
.23234 .50804 Mdot
.24748 .49949 Mdot
.29621 .47472 Mdot
.33165 .45901 Mdot
.35032 .45142 Mdot
.38961 .43682 Mdot
.41028 .42983 Mdot
.45371 .41653 Mdot
.47649 .41024 Mdot
.52423 .39844 Mdot
.54918 .39296 Mdot
.60121 .38288 Mdot
.62826 .37831 Mdot
.68433 .37017 Mdot
.71327 .36663 Mdot
.7728 .36064 Mdot
.86536 .35455 Mdot
.89684 .35331 Mdot
.96028 .35208 Mdot
.12472 .60095 Mdot
.14716 .58211 Mdot
.17111 .56364 Mdot
.22411 .52797 Mdot
.25341 .51085 Mdot
.28477 .49429 Mdot
.31834 .47834 Mdot
.35424 .46307 Mdot
.39261 .44854 Mdot
.43356 .43484 Mdot
.47718 .42204 Mdot
.52351 .41024 Mdot
.57255 .39953 Mdot
.62424 .39 Mdot
.67844 .38174 Mdot
.73492 .37485 Mdot
.79337 .36939 Mdot
.8534 .36544 Mdot
.91451 .36305 Mdot
.14288 .60237 Mdot
.1779 .5747 Mdot
.19034 .56567 Mdot
.2032 .55675 Mdot
.21649 .54794 Mdot
.25908 .52224 Mdot
.27423 .51393 Mdot
.2899 .50577 Mdot
.32282 .48992 Mdot
.34011 .48223 Mdot
.37641 .4674 Mdot
.4151 .45333 Mdot
.43537 .44661 Mdot
.45628 .44011 Mdot
.47782 .43384 Mdot
.52282 .42204 Mdot
.54629 .41653 Mdot
.57039 .4113 Mdot
.59512 .40635 Mdot
.6464 .39735 Mdot
.69997 .38963 Mdot
.72754 .38627 Mdot
.78407 .3806 Mdot
.81294 .37831 Mdot
.84215 .37639 Mdot
.93122 .37291 Mdot
.96119 .37252 Mdot
.18441 .58578 Mdot
.20942 .56794 Mdot
.26449 .53364 Mdot
.29479 .51727 Mdot
.3615 .48638 Mdot
.39814 .47199 Mdot
.47843 .44564 Mdot
.52218 .43384 Mdot
.61691 .4134 Mdot
.66771 .40492 Mdot
.77533 .39185 Mdot
.83157 .38739 Mdot
.94703 .38288 Mdot
.19069 .59689 Mdot
.20285 .58797 Mdot
.2154 .57916 Mdot
.22834 .57045 Mdot
.24169 .56186 Mdot
.25546 .55339 Mdot
.26968 .54505 Mdot
.28434 .53685 Mdot
.29946 .52878 Mdot
.31507 .52086 Mdot
.33116 .5131 Mdot
.34776 .50549 Mdot
.36487 .49805 Mdot
.38252 .49079 Mdot
.4007 .48372 Mdot
.41943 .47684 Mdot
.43872 .47016 Mdot
.45858 .46369 Mdot
.479 .45744 Mdot
.52157 .44564 Mdot
.54372 .44011 Mdot
.56644 .43484 Mdot
.58972 .42983 Mdot
.61355 .42511 Mdot
.63792 .42067 Mdot
.66281 .41653 Mdot
.6882 .4127 Mdot
.71407 .40918 Mdot
.74037 .40599 Mdot
.76709 .40313 Mdot
.79417 .40061 Mdot
.82159 .39844 Mdot
.8493 .39662 Mdot
.87724 .39516 Mdot
.90538 .39406 Mdot
.93365 .39333 Mdot
.962 .39296 Mdot
.22114 .5904 Mdot
.24708 .5732 Mdot
.27464 .55649 Mdot
.30393 .5403 Mdot
.33504 .5247 Mdot
.36808 .50973 Mdot
.40313 .49545 Mdot
.44027 .48193 Mdot
.47955 .46924 Mdot
.521 .45744 Mdot
.56463 .44661 Mdot
.61039 .43682 Mdot
.65819 .42815 Mdot
.7079 .42067 Mdot
.75931 .41445 Mdot
.81218 .40954 Mdot
.86619 .40599 Mdot
.92099 .40385 Mdot
.22666 .60166 Mdot
.23926 .59305 Mdot
.26562 .57619 Mdot
.3082 .55184 Mdot
.32326 .544 Mdot
.35472 .52878 Mdot
.37114 .52141 Mdot
.40545 .50719 Mdot
.42334 .50035 Mdot
.46064 .48727 Mdot
.48006 .48104 Mdot
.52045 .46924 Mdot
.54142 .46369 Mdot
.5849 .45333 Mdot
.60739 .44854 Mdot
.65382 .43978 Mdot
.67773 .43583 Mdot
.72683 .42882 Mdot
.75197 .42578 Mdot
.82938 .41861 Mdot
.88228 .41549 Mdot
.909 .41445 Mdot
.96273 .4134 Mdot
.28396 .5794 Mdot
.3123 .56339 Mdot
.34231 .54794 Mdot
.37407 .5331 Mdot
.44314 .50549 Mdot
.48056 .49284 Mdot
.51994 .48104 Mdot
.56128 .47016 Mdot
.64968 .45142 Mdot
.69656 .44369 Mdot
.74502 .43715 Mdot
.79486 .43184 Mdot
.89763 .42511 Mdot
.94994 .42375 Mdot
.27499 .59904 Mdot
.28835 .59088 Mdot
.30209 .58285 Mdot
.31623 .57494 Mdot
.33076 .56718 Mdot
.34571 .55957 Mdot
.36107 .5521 Mdot
.37686 .54479 Mdot
.39309 .53765 Mdot
.40977 .53067 Mdot
.42689 .52388 Mdot
.44447 .51727 Mdot
.46252 .51085 Mdot
.48102 .50464 Mdot
.51944 .49284 Mdot
.53936 .48727 Mdot
.55973 .48193 Mdot
.58057 .47684 Mdot
.60186 .47199 Mdot
.62359 .4674 Mdot
.64576 .46307 Mdot
.66835 .45901 Mdot
.69133 .45523 Mdot
.7147 .45174 Mdot
.73843 .44854 Mdot
.7625 .44564 Mdot
.78687 .44304 Mdot
.81153 .44076 Mdot
.83643 .4388 Mdot
.86154 .43715 Mdot
.88683 .43583 Mdot
.91226 .43484 Mdot
.93779 .43417 Mdot
.96338 .43384 Mdot
.29256 .60237 Mdot
.31999 .58651 Mdot
.41178 .54242 Mdot
.48147 .51644 Mdot
.51898 .50464 Mdot
.5993 .48372 Mdot
.73218 .45995 Mdot
.7793 .45428 Mdot
.87658 .44661 Mdot
.92623 .44467 Mdot
.32361 .59809 Mdot
.33766 .5904 Mdot
.35209 .58285 Mdot
.3669 .57544 Mdot
.3821 .56819 Mdot
.3977 .5611 Mdot
.41371 .55417 Mdot
.43013 .54741 Mdot
.44696 .54083 Mdot
.46422 .53444 Mdot
.4819 .52824 Mdot
.51853 .51644 Mdot
.53748 .51085 Mdot
.55686 .50549 Mdot
.57666 .50035 Mdot
.59687 .49545 Mdot
.61748 .49079 Mdot
.6385 .48638 Mdot
.65989 .48223 Mdot
.68166 .47834 Mdot
.70379 .47472 Mdot
.72625 .47138 Mdot
.74903 .46832 Mdot
.7721 .46555 Mdot
.79544 .46307 Mdot
.81903 .46089 Mdot
.84283 .45901 Mdot
.86682 .45744 Mdot
.89096 .45618 Mdot
.91523 .45523 Mdot
.93958 .4546 Mdot
.96398 .45428 Mdot
.35508 .5945 Mdot
.38455 .5799 Mdot
.41555 .56592 Mdot
.44813 .55262 Mdot
.4823 .54004 Mdot
.5181 .52824 Mdot
.55553 .51727 Mdot
.59455 .50719 Mdot
.63513 .49805 Mdot
.67718 .48992 Mdot
.76525 .47684 Mdot
.81096 .47199 Mdot
.85753 .46832 Mdot
.90474 .46585 Mdot
.95233 .46462 Mdot
.37226 .5988 Mdot
.38691 .59161 Mdot
.43309 .57095 Mdot
.46577 .55803 Mdot
.48269 .55184 Mdot
.5177 .54004 Mdot
.53578 .53444 Mdot
.57311 .52388 Mdot
.63192 .50973 Mdot
.65224 .50549 Mdot
.71523 .49429 Mdot
.75873 .48815 Mdot
.78089 .4855 Mdot
.82589 .48104 Mdot
.84868 .47924 Mdot
.89473 .47654 Mdot
.96452 .47472 Mdot
.38917 .60332 Mdot
.41902 .58943 Mdot
.45031 .57619 Mdot
.48306 .56364 Mdot
.51731 .55184 Mdot
.55304 .54083 Mdot
.59023 .53067 Mdot
.62886 .52141 Mdot
.66884 .5131 Mdot
.7101 .50577 Mdot
.75252 .49949 Mdot
.79595 .49429 Mdot
.84024 .49021 Mdot
.88518 .48727 Mdot
.93057 .4855 Mdot
.42065 .60118 Mdot
.43581 .5945 Mdot
.45133 .58797 Mdot
.4672 .58162 Mdot
.48342 .57544 Mdot
.51694 .56364 Mdot
.53423 .55803 Mdot
.55187 .55262 Mdot
.56987 .54741 Mdot
.58822 .54242 Mdot
.60691 .53765 Mdot
.62593 .5331 Mdot
.64528 .52878 Mdot
.66496 .5247 Mdot
.68493 .52086 Mdot
.70521 .51727 Mdot
.72577 .51393 Mdot
.74659 .51085 Mdot
.76766 .50804 Mdot
.78896 .50549 Mdot
.81047 .50321 Mdot
.83217 .50121 Mdot
.85404 .49949 Mdot
.87605 .49805 Mdot
.89818 .4969 Mdot
.9204 .49603 Mdot
.94269 .49545 Mdot
.96502 .49516 Mdot
.48376 .58724 Mdot
.51658 .57544 Mdot
.58629 .55417 Mdot
.62314 .54479 Mdot
.70054 .52878 Mdot
.74092 .52224 Mdot
.82447 .51226 Mdot
.86732 .50888 Mdot
.95432 .50549 Mdot
.48409 .59904 Mdot
.51624 .58724 Mdot
.5328 .58162 Mdot
.54969 .57619 Mdot
.56691 .57095 Mdot
.58445 .56592 Mdot
.6023 .5611 Mdot
.63893 .5521 Mdot
.65769 .54794 Mdot
.67674 .544 Mdot
.69607 .5403 Mdot
.71566 .53685 Mdot
.73551 .53364 Mdot
.77589 .52797 Mdot
.7964 .52552 Mdot
.8171 .52333 Mdot
.83796 .52141 Mdot
.85897 .51976 Mdot
.88011 .51838 Mdot
.92267 .51644 Mdot
.94406 .51588 Mdot
.96548 .5156 Mdot
.51591 .59904 Mdot
.54867 .58797 Mdot
.6179 .56819 Mdot
.65429 .55957 Mdot
.6918 .55184 Mdot
.73032 .54505 Mdot
.81004 .53444 Mdot
.85097 .53067 Mdot
.89242 .52797 Mdot
.93421 .52634 Mdot
.56419 .5945 Mdot
.58098 .58943 Mdot
.61545 .5799 Mdot
.6331 .57544 Mdot
.66924 .56718 Mdot
.6877 .56339 Mdot
.72536 .55649 Mdot
.74454 .55339 Mdot
.78351 .54794 Mdot
.8433 .54163 Mdot
.86352 .54004 Mdot
.90427 .53765 Mdot
.92477 .53685 Mdot
.9659 .53605 Mdot
.57935 .60118 Mdot
.61309 .59161 Mdot
.64791 .58285 Mdot
.68377 .57494 Mdot
.75831 .56186 Mdot
.7968 .55675 Mdot
.83595 .55262 Mdot
.87563 .5495 Mdot
.91569 .54741 Mdot
.956 .54637 Mdot
.61083 .60332 Mdot
.62774 .5988 Mdot
.64492 .5945 Mdot
.66234 .5904 Mdot
.68001 .58651 Mdot
.69791 .58285 Mdot
.71604 .5794 Mdot
.73438 .57619 Mdot
.75292 .5732 Mdot
.77166 .57045 Mdot
.79058 .56794 Mdot
.80966 .56567 Mdot
.82889 .56364 Mdot
.84825 .56186 Mdot
.86773 .56033 Mdot
.88731 .55905 Mdot
.90697 .55803 Mdot
.9267 .55726 Mdot
.94648 .55675 Mdot
.96628 .55649 Mdot
.67639 .59809 Mdot
.71165 .59088 Mdot
.7846 .57916 Mdot
.8221 .5747 Mdot
.89859 .56869 Mdot
.93732 .56718 Mdot
.70744 .60237 Mdot
.72501 .59904 Mdot
.76074 .59305 Mdot
.77886 .5904 Mdot
.79715 .58797 Mdot
.81559 .58578 Mdot
.85284 .58211 Mdot
.89052 .5794 Mdot
.90948 .57841 Mdot
.94756 .57718 Mdot
.96664 .57693 Mdot
.77334 .60166 Mdot
.80931 .59689 Mdot
.88276 .59015 Mdot
.92 .58821 Mdot
.95744 .58724 Mdot
.85712 .60237 Mdot
.87528 .60095 Mdot
.91182 .5988 Mdot
.93017 .59809 Mdot
.96698 .59737 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{360, 222.375},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40005X0000gR000`400?l00000o`00003oo`3ooomY
0?ooo`00o`3ooomY0?ooo`00o`3ooomY0?ooo`00o`3ooomY0?ooo`00o`3ooomY0?ooo`00o`3ooomY
0?ooo`00o`3ooomY0?ooo`00o`3ooomY0?ooo`00o`3ooomY0?ooo`00A03oool4000000<0oooo0P00
000V0?ooo`@000000`3oool2000002P0oooo0P0000040?ooo`800000:03oool00`000000oooo0?oo
o`020?ooo`8000009`3oool3000000@0oooo0P00000W0?ooo`800000103oool2000002h0oooo0016
0?ooo`050000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`0S0?ooo`030000003o
ool0oooo00<0oooo00@000000?ooo`3oool000009P3oool010000000oooo0?ooo`0000020?ooo`04
0000003oool0oooo000002L0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003oool0
oooo02L0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`0S0?ooo`040000003o
ool0oooo00000080oooo00@000000?ooo`3oool00000;@3oool004H0oooo00D000000?ooo`3oool0
oooo000000020?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool00P3oool010000000
oooo0?ooo`00000Y0?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3oool08P3oool5
00000080oooo00@000000?ooo`3oool00000:@3oool010000000oooo0?ooo`0000020?ooo`030000
003oool0oooo02<0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`00000]0?oo
o`00AP3oool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`3oool09@3oool01@00
0000oooo0?ooo`3oool000000080oooo00<000000?ooo`3oool09@3oool2000000<0oooo00@00000
0?ooo`3oool00000903oool010000000oooo0?ooo`0000030?ooo`040000003oool0oooo000002H0
oooo0`0000030?ooo`040000003oool0oooo000002D0oooo0`0000030?ooo`040000003oool0oooo
000002d0oooo00160?ooo`050000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`0V
0?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3oool09`3oool010000000oooo0?oo
o`0000020?ooo`030000003oool0oooo02<0oooo00<000000?ooo`0000000`3oool010000000oooo
0?ooo`00000V0?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool000009@3oool00`00
0000oooo0?ooo`030?ooo`040000003oool0oooo000002d0oooo000J0?ooo`030000003oool0oooo
02L0oooo0`0000030?ooo`040000003oool0oooo000002D0oooo00@000000?ooo`3oool000000P3o
ool010000000oooo0?ooo`00000V0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3o
ool000009P3oool2000000<0oooo00@000000?ooo`3oool000009P3oool00`000000oooo0?ooo`03
0?ooo`040000003oool0oooo000002D0oooo00<000000?ooo`3oool00`3oool010000000oooo0?oo
o`00000]0?ooo`006P3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo0080oooo0P00
000W0?ooo`800000103oool2000002P0oooo0P0000040?ooo`800000:03oool00`000000oooo0?oo
o`020?ooo`8000009`3oool4000000<0oooo0P00000W0?ooo`<000000`3oool2000002h0oooo000J
0?ooo`030000003oool0oooo0?l0ooooC03oool001X0oooo00<000000?ooo`3oool0o`3ooolh0?oo
o`<000004@3oool00180ooooo`00000]000001T0oooo00<000000?ooo`3oool03P3oool001X0oooo
00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo
00<000000?ooo`3oool03P3oool200000080oooo1@0000000`3oool00000000000020?ooo`<00000
0P3oool500000080oooo0`0000040?ooo`006P3oool00`000000oooo0?ooo`0/0?ooo`030000003o
ool0oooo02`0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`0/0?ooo`030000003o
ool0oooo02d0oooo00<000000?ooo`3oool0;03oool00`000000oooo0?ooo`0?0?ooo`040000003o
ool0oooo00000080oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`0000040?oo
o`050000003oool0oooo0?ooo`0000001`3oool001X0oooo00<000000?ooo`3oool0o`3oool/0?oo
o`@000000P3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000000@0oooo00@0
00000?ooo`3oool0oooo0`0000050?ooo`006P3oool00`000000oooo0?ooo`2:0?ooo`800000W`3o
ool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3o
ool00000103oool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`3oool00P3oool0
01X0oooo00<000000?ooo`3oool0R`3oool00`000000oooo0?ooo`2N0?ooo`060000003oool00000
0?ooo`000000oooo0P0000040?ooo`80000000<0oooo0000003oool00P3oool00`000000oooo0?oo
o`020?ooo`8000001@3oool001T0oooo0P00003o0?oooa/0oooo0P00000A0?ooo`<000004@3oool0
0`000000oooo0?ooo`090?ooo`006P3oool200000?l0oooo6`3oool00`000000oooo0?ooo`0R0?oo
o`8000002`3oool001X0oooo00<000000?ooo`3oool0o`3ooom<0?ooo`006P3oool00`000000oooo
0?ooo`0_0?ooo`800000]03oool2000006@0oooo000J0?ooo`030000003oool0oooo0300oooo00<0
00000?ooo`3oool0/`3oool00`000000oooo0?ooo`1R0?ooo`006P3oool00`000000oooo0?ooo`3o
0?oood`0oooo000J0?ooo`030000003oool0oooo0?l0ooooC03oool001X0oooo00<000000?ooo`3o
ool0B03oool200000880oooo0P00001m0?ooo`006P3oool2000004X0oooo00<000000?ooo`3oool0
P@3oool00`000000oooo0?ooo`1k0?ooo`006P3oool00`000000oooo0?ooo`0K0?ooo`800000g03o
ool200000500oooo000J0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool0=`3oool2
000006H0oooo0P00000j0?ooo`030000003oool0oooo04h0oooo000J0?ooo`030000003oool0oooo
05L0oooo00<000000?ooo`3oool0I@3oool00`000000oooo0?ooo`290?ooo`006P3oool00`000000
oooo0?ooo`3o0?oood`0oooo000J0?ooo`030000003oool0oooo0?l0ooooC03oool001X0oooo00<0
00000?ooo`3oool0H03oool200000580oooo0P00002E0?ooo`006P3oool2000001<0oooo0P00001=
0?ooo`030000003oool0oooo0540oooo00<000000?ooo`3oool0BP3oool2000004L0oooo000J0?oo
o`030000003oool0oooo01<0oooo00<000000?ooo`3oool0:03oool2000009H0oooo0P00000[0?oo
o`030000003oool0oooo04D0oooo000J0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3o
ool0903oool2000004H0oooo0P00000W0?ooo`030000003oool0oooo0740oooo000J0?ooo`030000
003oool0oooo06L0oooo00<000000?ooo`3oool0A@3oool00`000000oooo0?ooo`2I0?ooo`006P3o
ool00`000000oooo0?ooo`0S0?ooo`800000c03oool2000005P0oooo000J0?ooo`030000003oool0
oooo02@0oooo00<000000?ooo`3oool0b`3oool00`000000oooo0?ooo`1F0?ooo`006P3oool00`00
0000oooo0?ooo`1[0?ooo`800000?03oool200000:00oooo000J0?ooo`8000003P3oool2000005d0
oooo00<000000?ooo`3oool0>`3oool00`000000oooo0?ooo`1J0?ooo`800000@P3oool001X0oooo
00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0o0?ooo`800000LP3oool200000480oooo
00<000000?ooo`3oool0@03oool001X0oooo00<000000?ooo`3oool0D@3oool00`000000oooo0?oo
o`0J0?ooo`800000=P3oool2000001d0oooo00<000000?ooo`3oool0P`3oool001X0oooo00<00000
0?ooo`3oool0>@3oool2000003@0oooo00<000000?ooo`3oool0=@3oool00`000000oooo0?ooo`0a
0?ooo`800000KP3oool001X0oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`2O0?oo
o`030000003oool0oooo06`0oooo000J0?ooo`030000003oool0oooo0?l0ooooC03oool000/0oooo
100000030?ooo`8000001P3oool00`000000oooo0?ooo`0:0?ooo`8000006P3oool2000004T0oooo
0P00000`0?ooo`800000B@3oool2000001X0oooo0P00000o0?ooo`003@3oool01@000000oooo0?oo
o`3oool000000080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0;0?ooo`030000
003oool0oooo01T0oooo00<000000?ooo`3oool0603oool2000002h0oooo00<000000?ooo`3oool0
;`3oool00`000000oooo0?ooo`0/0?ooo`8000006P3oool00`000000oooo0?ooo`0I0?ooo`030000
003oool0oooo03d0oooo000=0?ooo`050000003oool0oooo0?ooo`0000000P3oool00`000000oooo
0?ooo`030?ooo`800000A03oool00`000000oooo0?ooo`0E0?ooo`800000G03oool2000001T0oooo
00<000000?ooo`3oool0M03oool000d0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool05P3oool2000004@0oooo00<000000?ooo`3oool0
5@3oool2000002X0oooo0P00000H0?ooo`030000003oool0oooo0440oooo0P00001;0?ooo`003@3o
ool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`0G0?ooo`030000003oool0oooo05/0oooo00<000000?ooo`3oool0:@3oool00`000000oooo
0?ooo`1K0?ooo`030000003oool0oooo04T0oooo000;0?ooo`<000000`3oool010000000oooo0?oo
o`0000050?ooo`030000003oool0oooo0?l0ooooC03oool000d0oooo00<000000?ooo`3oool00P3o
ool2000000H0oooo00<000000?ooo`3oool0203oool2000001H0oooo0P00003B0?ooo`8000005P3o
ool2000003d0oooo000J0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool05@3oool0
0`000000oooo0?ooo`0C0?ooo`800000?@3oool2000002H0oooo0P00000m0?ooo`8000005P3oool0
0`000000oooo0?ooo`0E0?ooo`030000003oool0oooo03/0oooo000J0?ooo`030000003oool0oooo
03P0oooo00<000000?ooo`3oool04`3oool2000002L0oooo00<000000?ooo`3oool09@3oool00`00
0000oooo0?ooo`0T0?ooo`8000005P3oool00`000000oooo0?ooo`1Z0?ooo`006P3oool2000002T0
oooo0P00000U0?ooo`030000003oool0oooo0140oooo0P00001<0?ooo`800000503oool00`000000
oooo0?ooo`0R0?ooo`800000G@3oool001X0oooo00<000000?ooo`3oool0:@3oool00`000000oooo
0?ooo`0h0?ooo`030000003oool0oooo0100oooo0P00000T0?ooo`8000004`3oool00`000000oooo
0?ooo`0h0?ooo`030000003oool0oooo05/0oooo000J0?ooo`030000003oool0oooo0580oooo0P00
000T0?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool08@3oool2000008L0oooo000J
0?ooo`030000003oool0oooo00H0oooo0P00001;0?ooo`030000003oool0oooo06d0oooo00<00000
0?ooo`3oool0B03oool2000003/0oooo000J0?ooo`030000003oool0oooo00L0oooo00<000000?oo
o`3oool0>@3oool2000008`0oooo0P00000l0?ooo`030000003oool0oooo03T0oooo000J0?ooo`03
0000003oool0oooo04@0oooo00<000000?ooo`3oool0<P3oool200000200oooo0P00000e0?ooo`03
0000003oool0oooo07H0oooo000J0?ooo`030000003oool0oooo00l0oooo0P00001Y0?ooo`030000
003oool0oooo01l0oooo00<000000?ooo`3oool0IP3oool2000004@0oooo000J0?ooo`8000004@3o
ool00`000000oooo0?ooo`0S0?ooo`800000<@3oool200000400oooo0P00000a0?ooo`8000009P3o
ool00`000000oooo0?ooo`120?ooo`006P3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0
oooo0300oooo00<000000?ooo`3oool03@3oool2000001h0oooo0P00000@0?ooo`030000003oool0
oooo0300oooo00<000000?ooo`3oool0J@3oool001X0oooo00<000000?ooo`3oool01@3oool20000
0100oooo0P0000110?ooo`8000007`3oool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo
01`0oooo0P0000110?ooo`800000403oool2000003X0oooo000J0?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0>0?ooo`800000403oool2000001h0
oooo00<000000?ooo`3oool0G@3oool00`000000oooo0?ooo`0K0?ooo`800000403oool200000140
oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0h0?ooo`006P3oool00`000000oooo
0?ooo`0Z0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03@3oool2000007X0oooo
0P00000@0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool0G03oool001X0oooo00<0
00000?ooo`3oool07P3oool2000002d0oooo00<000000?ooo`3oool03@3oool2000001`0oooo0P00
000L0?ooo`800000703oool200000100oooo00<000000?ooo`3oool0:P3oool2000005<0oooo000J
0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool0?03oool00`000000oooo0?ooo`0K
0?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool06`3oool00`000000oooo0?ooo`0l
0?ooo`030000003oool0oooo0540oooo000J0?ooo`800000@@3oool2000002/0oooo0P00000h0?oo
o`800000:`3oool2000007D0oooo000J0?ooo`030000003oool0oooo00@0oooo0P00000?0?ooo`80
0000:P3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo00/0oooo0P00000J0?ooo`80
00003P3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo02P0oooo0P00000>0?ooo`80
0000>@3oool001X0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0>0?ooo`030000
003oool0oooo00`0oooo0P00000>0?ooo`800000AP3oool00`000000oooo0?ooo`0I0?ooo`030000
003oool0oooo04<0oooo0P00000>0?ooo`800000403oool00`000000oooo0?ooo`0=0?ooo`030000
003oool0oooo03L0oooo000J0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool03@3o
ool00`000000oooo0?ooo`0;0?ooo`800000RP3oool2000000h0oooo00<000000?ooo`3oool03@3o
ool00`000000oooo0?ooo`1H0?ooo`006P3oool00`000000oooo0?ooo`0;0?ooo`800000>03oool0
0`000000oooo0?ooo`0;0?ooo`800000K03oool2000000h0oooo00<000000?ooo`3oool0=@3oool2
00000400oooo000J0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool06`3oool20000
02P0oooo00<000000?ooo`3oool02`3oool2000001T0oooo0P00000H0?ooo`8000006@3oool20000
00h0oooo00<000000?ooo`3oool09@3oool2000001h0oooo00<000000?ooo`3oool0?P3oool001X0
oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?ooo`0e0?ooo`030000003oool0oooo01P0
oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo03D0
oooo00<000000?ooo`3oool0G@3oool001X0oooo0P0000040?ooo`800000J`3oool200000380oooo
0P00001[0?ooo`800000>03oool001X0oooo00<000000?ooo`3oool0103oool00`000000oooo0?oo
o`0J0?ooo`800000CP3oool00`000000oooo0?ooo`090?ooo`800000603oool2000000`0oooo00<0
00000?ooo`3oool0B`3oool2000001d0oooo00<000000?ooo`3oool0=P3oool001X0oooo00<00000
0?ooo`3oool08P3oool00`000000oooo0?ooo`0o0?ooo`800000603oool00`000000oooo0?ooo`0G
0?ooo`030000003oool0oooo01D0oooo0P0000120?ooo`030000003oool0oooo05@0oooo000J0?oo
o`030000003oool0oooo04`0oooo0P00000G0?ooo`030000003oool0oooo04T0oooo00<000000?oo
o`3oool0503oool200000840oooo000J0?ooo`030000003oool0oooo01P0oooo0P00000c0?ooo`03
0000003oool0oooo00T0oooo0P00001P0?ooo`800000303oool00`000000oooo0?ooo`0`0?ooo`80
0000C@3oool000/0oooo100000030?ooo`8000001P3oool00`000000oooo0?ooo`0I0?ooo`030000
003oool0oooo01P0oooo0P00000T0?ooo`030000003oool0oooo0240oooo0P00000F0?ooo`800000
903oool00`000000oooo0?ooo`0Q0?ooo`8000006`3oool00`000000oooo0?ooo`1;0?ooo`002`3o
ool00`000000oooo0?ooo`030?ooo`040000003oool0oooo000000D0oooo00<000000?ooo`3oool0
0`3oool2000000/0oooo0P00000S0?ooo`030000003oool0oooo01L0oooo0P00000^0?ooo`030000
003oool0oooo01D0oooo00<000000?ooo`3oool0:`3oool2000001X0oooo00<000000?ooo`3oool0
803oool2000000/0oooo0P00000h0?ooo`00303oool00`000000oooo0?ooo`020?ooo`040000003o
ool0oooo000000D0oooo0P0000050?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0
2P3oool2000000/0oooo0P00000S0?ooo`030000003oool0oooo0200oooo0P00000/0?ooo`800000
8`3oool00`000000oooo0?ooo`0P0?ooo`8000002`3oool2000000d0oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`0f0?ooo`003@3oool01@000000oooo0?ooo`3oool000000080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool02@3oool2000003X0oooo00<000000?ooo`3oool0203oool2000001@0oooo
0P00000;0?ooo`030000003oool0oooo03L0oooo0P00000<0?ooo`030000003oool0oooo00X0oooo
00<000000?ooo`3oool0D@3oool000h0oooo00@000000?ooo`3oool000000P3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00T0oooo0P00000^0?ooo`030000003oool0oooo00T0oooo
0P00000Q0?ooo`8000005P3oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo01<0oooo
0P00000Q0?ooo`800000303oool00`000000oooo0?ooo`0[0?ooo`800000?P3oool000/0oooo00@0
00000?ooo`3oool000000P3oool010000000oooo0?ooo`0000050?ooo`030000003oool0oooo00X0
oooo00<000000?ooo`3oool0>@3oool00`000000oooo0?ooo`080?ooo`8000005P3oool00`000000
oooo0?ooo`110?ooo`030000003oool0oooo01<0oooo0P00000;0?ooo`030000003oool0oooo03T0
oooo00<000000?ooo`3oool0?03oool000`0oooo0P0000040?ooo`8000001P3oool00`000000oooo
0?ooo`0l0?ooo`800000503oool00`000000oooo0?ooo`090?ooo`800000EP3oool2000000`0oooo
00<000000?ooo`3oool04@3oool200000740oooo000J0?ooo`030000003oool0oooo0080oooo0P00
000;0?ooo`800000;03oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo00L0oooo0P00
000D0?ooo`800000503oool2000001@0oooo0P00000:0?ooo`030000003oool0oooo01l0oooo00<0
00000?ooo`3oool0:@3oool2000000/0oooo0P00000g0?ooo`006P3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0503oool2000002/0oooo0P00000D
0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0C
0?ooo`030000003oool0oooo0140oooo0P00000[0?ooo`8000005`3oool00`000000oooo0?ooo`0:
0?ooo`030000003oool0oooo03D0oooo000J0?ooo`800000:@3oool00`000000oooo0?ooo`080?oo
o`800000803oool00`000000oooo0?ooo`0M0?ooo`800000:03oool200000200oooo00<000000?oo
o`3oool07@3oool2000000/0oooo00<000000?ooo`3oool0FP3oool001X0oooo00<000000?ooo`3o
ool0=03oool00`000000oooo0?ooo`080?ooo`800000=@3oool00`000000oooo0?ooo`070?ooo`80
00004P3oool2000000X0oooo00<000000?ooo`3oool0<P3oool2000000/0oooo00<000000?ooo`3o
ool0IP3oool001X0oooo00<000000?ooo`3oool0503oool2000002X0oooo00<000000?ooo`3oool0
203oool2000003@0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0a0?ooo`800000
2`3oool00`000000oooo0?ooo`0W0?ooo`800000B@3oool001X0oooo00<000000?ooo`3oool05@3o
ool00`000000oooo0?ooo`0D0?ooo`8000007P3oool00`000000oooo0?ooo`1k0?ooo`030000003o
ool0oooo01/0oooo0P00000G0?ooo`030000003oool0oooo04L0oooo000J0?ooo`030000003oool0
oooo0080oooo0P00000Y0?ooo`030000003oool0oooo01<0oooo0P00000L0?ooo`800000D03oool2
000001d0oooo0P00000E0?ooo`030000003oool0oooo02H0oooo0P00000g0?ooo`006P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo01<0oooo0P0000090?ooo`8000007P3oool00`00
0000oooo0?ooo`0K0?ooo`030000003oool0oooo00L0oooo0P00000B0?ooo`8000004P3oool20000
0180oooo0P00000:0?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3oool06P3oool20000
00T0oooo0P00000F0?ooo`030000003oool0oooo03D0oooo000J0?ooo`030000003oool0oooo01X0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0a0?ooo`8000004P3oool00`000000
oooo0?ooo`0A0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3oool00`000000
oooo0?ooo`0@0?ooo`800000<`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo04`0
oooo000J0?ooo`800000?03oool2000001d0oooo00<000000?ooo`3oool06@3oool2000002H0oooo
0P00000M0?ooo`030000003oool0oooo01T0oooo0P00001`0?ooo`006P3oool00`000000oooo0?oo
o`070?ooo`800000503oool2000001d0oooo00<000000?ooo`3oool01P3oool200000300oooo00<0
00000?ooo`3oool01P3oool200000180oooo0P0000090?ooo`030000003oool0oooo02d0oooo0P00
00090?ooo`030000003oool0oooo01X0oooo0P00000D0?ooo`800000?03oool001X0oooo00<00000
0?ooo`3oool0203oool00`000000oooo0?ooo`0C0?ooo`030000003oool0oooo0180oooo0P00000A
0?ooo`030000003oool0oooo02@0oooo0P00000B0?ooo`030000003oool0oooo0140oooo00<00000
0?ooo`3oool03`3oool2000002L0oooo00<000000?ooo`3oool03P3oool2000001D0oooo00<00000
0?ooo`3oool04`3oool00`000000oooo0?ooo`0j0?ooo`006P3oool00`000000oooo0?ooo`0d0?oo
o`030000003oool0oooo02@0oooo0P00000A0?ooo`030000003oool0oooo03L0oooo00<000000?oo
o`3oool03`3oool2000002H0oooo00<000000?ooo`3oool0IP3oool001X0oooo00@000000?ooo`3o
ool0oooo0P0000090?ooo`8000002@3oool2000000P0oooo0P00000i0?ooo`030000003oool0oooo
00H0oooo0P0000180?ooo`8000002P3oool00`000000oooo0?ooo`0e0?ooo`800000203oool20000
00T0oooo0P0000090?ooo`800000=P3oool001X0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool01`3oool00`000000
oooo0?ooo`070?ooo`800000>03oool00`000000oooo0?ooo`0H0?ooo`800000403oool2000001/0
oooo00<000000?ooo`3oool0=@3oool2000000X0oooo00<000000?ooo`3oool01`3oool00`000000
oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0=03oool001X0oooo
00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`060?ooo`8000002@3oool2000001T0oooo
0P00000T0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool08@3oool2000001T0oooo
0P0000090?ooo`8000002@3oool00`000000oooo0?ooo`1O0?ooo`006P3oool200000180oooo0P00
000T0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool01P3oool200000100oooo00<0
00000?ooo`3oool0603oool200000280oooo0P00000K0?ooo`030000003oool0oooo00d0oooo0P00
00090?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool08@3oool2000004H0oooo000J
0?ooo`030000003oool0oooo0180oooo00<000000?ooo`3oool0403oool2000002D0oooo00<00000
0?ooo`3oool01@3oool2000002<0oooo00<000000?ooo`3oool01@3oool200000100oooo0P000008
0?ooo`030000003oool0oooo0200oooo0P0000080?ooo`030000003oool0oooo0280oooo0P00000C
0?ooo`030000003oool0oooo04@0oooo000J0?ooo`030000003oool0oooo02H0oooo00<000000?oo
o`3oool0;03oool00`000000oooo0?ooo`0G0?ooo`8000004@3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo00h0oooo0P00000J0?ooo`030000003oool0oooo02`0oooo00<000000?oo
o`3oool0F03oool001X0oooo00@000000?ooo`3oool0oooo0P0000080?ooo`800000203oool20000
00P0oooo0P00000m0?ooo`800000403oool00`000000oooo0?ooo`0c0?ooo`030000003oool0oooo
00d0oooo0P00000m0?ooo`800000203oool2000000P0oooo0P0000080?ooo`800000=P3oool001X0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0
oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`8000008P3oool200000100
oooo00<000000?ooo`3oool01@3oool2000004@0oooo0P0000080?ooo`030000003oool0oooo00d0
oooo0P00000R0?ooo`8000002@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0
oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0d0?ooo`006P3oool00`000000oooo
0?ooo`0Z0?ooo`030000003oool0oooo00H0oooo0P0000070?ooo`800000403oool00`000000oooo
0?ooo`0G0?ooo`030000003oool0oooo00D0oooo0P00000?0?ooo`800000403oool2000000l0oooo
0P0000080?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool03@3oool2000000L0oooo
0P0000090?ooo`030000003oool0oooo05`0oooo000J0?ooo`030000003oool0oooo00H0oooo0P00
000A0?ooo`8000006@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo0P00
000Y0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?oo
o`0>0?ooo`030000003oool0oooo02H0oooo0P0000090?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool05P3oool200000140oooo0P00000k0?ooo`006P3oool2000000P0oooo00<00000
0?ooo`3oool0403oool00`000000oooo0?ooo`0?0?ooo`8000006@3oool00`000000oooo0?ooo`05
0?ooo`800000:03oool200000200oooo0P00000X0?ooo`800000203oool00`000000oooo0?ooo`0F
0?ooo`8000004P3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo03T0oooo000J0?oo
o`030000003oool0oooo02d0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`050?oo
o`800000803oool00`000000oooo0?ooo`050?ooo`8000003P3oool2000000P0oooo00<000000?oo
o`3oool07@3oool2000000P0oooo00<000000?ooo`3oool0803oool00`000000oooo0?ooo`1O0?oo
o`006P3oool010000000oooo0?ooo`3oool200000140oooo0P0000070?ooo`800000803oool20000
01P0oooo00<000000?ooo`3oool09`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
02L0oooo00<000000?ooo`3oool05@3oool200000200oooo0P0000080?ooo`800000403oool20000
03H0oooo000J0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0403oool00`000000
oooo0?ooo`060?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool07P3oool200000500
oooo0P00000Q0?ooo`030000003oool0oooo01l0oooo00<000000?ooo`3oool01`3oool00`000000
oooo0?ooo`0?0?ooo`030000003oool0oooo03@0oooo000J0?ooo`030000003oool0oooo03T0oooo
0P00000F0?ooo`8000003`3oool00`000000oooo0?ooo`050?ooo`800000?P3oool2000000P0oooo
00<000000?ooo`3oool0303oool2000001H0oooo0P00001^0?ooo`006P3oool00`000000oooo0?oo
o`0>0?ooo`800000:P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo01H0oooo00<0
00000?ooo`3oool0103oool2000000h0oooo0P00000>0?ooo`8000003P3oool2000000L0oooo00<0
00000?ooo`3oool05P3oool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo02L0oooo0P00
00130?ooo`00303oool2000000@0oooo0P0000060?ooo`030000003oool0oooo00l0oooo00<00000
0?ooo`3oool03P3oool2000002P0oooo0P00000F0?ooo`8000003`3oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0<
0?ooo`8000005P3oool2000002P0oooo0P00000A0?ooo`030000003oool0oooo0440oooo000;0?oo
o`040000003oool0oooo00000080oooo00@000000?ooo`3oool000001@3oool200000280oooo00<0
00000?ooo`3oool03P3oool2000001L0oooo00<000000?ooo`3oool01@3oool2000000h0oooo00<0
00000?ooo`3oool0503oool2000001h0oooo0P00000G0?ooo`030000003oool0oooo00/0oooo0P00
00080?ooo`030000003oool0oooo01@0oooo0P00000A0?ooo`030000003oool0oooo05<0oooo000>
0?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`02000000L0oooo0P00000X0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool0903o
ool00`000000oooo0?ooo`0M0?ooo`030000003oool0oooo02@0oooo00<000000?ooo`3oool07P3o
ool00`000000oooo0?ooo`0U0?ooo`8000001`3oool2000003D0oooo000<0?ooo`8000000`3oool0
10000000oooo0?ooo`0000050?ooo`050000003oool0oooo0?ooo`000000203oool00`000000oooo
0?ooo`0G0?ooo`8000007P3oool2000002d0oooo0P00000=0?ooo`8000003P3oool2000000d0oooo
0P00000]0?ooo`8000007P3oool2000001X0oooo00<000000?ooo`3oool01P3oool00`000000oooo
0?ooo`0c0?ooo`003P3oool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool09@3oool00`000000oooo0?ooo`050?ooo`8000001P3oool2000000h0oooo
00<000000?ooo`3oool0703oool2000000h0oooo00<000000?ooo`3oool0303oool00`000000oooo
0?ooo`0=0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool02`3oool2000001l0oooo
00<000000?ooo`3oool02`3oool2000000H0oooo0P0000080?ooo`030000003oool0oooo05L0oooo
000;0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool000001@3oool00`000000
oooo0?ooo`050?ooo`8000009`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo01`0
oooo0P00000>0?ooo`030000003oool0oooo00@0oooo0P00000j0?ooo`8000001`3oool00`000000
oooo0?ooo`0;0?ooo`8000007`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo02D0
oooo0P00000i0?ooo`00303oool2000000@0oooo0P0000060?ooo`030000003oool0oooo00H0oooo
00<000000?ooo`3oool07P3oool2000001d0oooo0P00000>0?ooo`030000003oool0oooo01@0oooo
00<000000?ooo`3oool04`3oool2000000`0oooo0P00000F0?ooo`030000003oool0oooo01@0oooo
00<000000?ooo`3oool02`3oool2000001d0oooo0P00000R0?ooo`030000003oool0oooo03L0oooo
000J0?ooo`030000003oool0oooo02P0oooo00<000000?ooo`3oool0703oool00`000000oooo0?oo
o`040?ooo`8000005@3oool2000001d0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?oo
o`0J0?ooo`8000005@3oool2000000L0oooo00<000000?ooo`3oool0703oool00`000000oooo0?oo
o`1J0?ooo`006P3oool2000000030?ooo`000000000000L0oooo0P0000060?ooo`8000009@3oool2
000001D0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0B0?ooo`800000703oool2
000001D0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`0B0?ooo`8000009@3oool2
000000L0oooo0P0000060?ooo`800000=@3oool001X0oooo00D000000?ooo`3oool0oooo00000008
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0103oool2000000L0oooo0P00000E
0?ooo`030000003oool0oooo0280oooo0P00000L0?ooo`030000003oool0oooo01/0oooo00<00000
0?ooo`3oool06@3oool2000002D0oooo00<000000?ooo`3oool04P3oool2000000L0oooo0P000007
0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0c
0?ooo`006P3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool0103oool2000000H0oooo0P00000[0?ooo`030000003oool0oooo0180oooo0P00000=0?oo
o`800000303oool2000000d0oooo0P00000E0?ooo`030000003oool0oooo02P0oooo0P0000060?oo
o`8000001`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo04`0oooo000J0?ooo`03
0000003oool0oooo00`0oooo0P00000M0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool0103oool2000003T0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0;0?ooo`03
0000003oool0oooo00`0oooo00<000000?ooo`3oool0=P3oool2000000L0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`0J0?ooo`800000@@3oool001X0oooo00<000000?ooo`3oool0
3@3oool00`000000oooo0?ooo`0=0?ooo`800000703oool00`000000oooo0?ooo`040?ooo`800000
503oool2000001@0oooo0P00000f0?ooo`800000503oool2000001@0oooo0P0000070?ooo`030000
003oool0oooo01T0oooo0P00000@0?ooo`030000003oool0oooo03l0oooo000J0?ooo`030000003o
ool0oooo01h0oooo00<000000?ooo`3oool0303oool2000001@0oooo00<000000?ooo`3oool0103o
ool2000000d0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`030?ooo`800000303o
ool2000000`0oooo0P00000<0?ooo`8000001P3oool00`000000oooo0?ooo`0C0?ooo`030000003o
ool0oooo00X0oooo0P0000070?ooo`030000003oool0oooo0140oooo0P00000?0?ooo`030000003o
ool0oooo0500oooo000J0?ooo`030000003oool0oooo008000003P3oool2000001`0oooo00<00000
0?ooo`3oool0303oool2000000`0oooo00<000000?ooo`3oool0103oool2000001<0oooo0P00000=
0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0;
0?ooo`030000003oool0oooo00X0oooo0P00000C0?ooo`8000001`3oool00`000000oooo0?ooo`09
0?ooo`8000003`3oool00`000000oooo0?ooo`0I0?ooo`8000003P3oool2000003D0oooo000J0?oo
o`8000000P3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00@0oooo0P00000T0?oo
o`030000003oool0oooo0180oooo00<000000?ooo`3oool00`3oool2000000d0oooo00<000000?oo
o`3oool04@3oool2000001X0oooo0P00000D0?ooo`030000003oool0oooo00X0oooo0P0000060?oo
o`030000003oool0oooo0180oooo00<000000?ooo`3oool08@3oool2000000L0oooo00<000000?oo
o`3oool03@3oool00`000000oooo0?ooo`0c0?ooo`006P3oool00`000000oooo0?ooo`0I0?ooo`03
0000003oool0oooo00`0oooo0P00000R0?ooo`800000303oool00`000000oooo0?ooo`040?ooo`80
00006P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo01L0oooo0P0000070?ooo`03
0000003oool0oooo00T0oooo0P00000R0?ooo`8000003`3oool00`000000oooo0?ooo`1;0?ooo`00
6P3oool00`000000oooo0?ooo`040?ooo`8000008`3oool00`000000oooo0?ooo`0Q0?ooo`030000
003oool0oooo0180oooo00<000000?ooo`3oool07P3oool2000000`0oooo0P00000Q0?ooo`030000
003oool0oooo0180oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0P0?ooo`800000
>@3oool001X0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0;0?ooo`800000:P3o
ool2000001X0oooo0P00000;0?ooo`8000006P3oool00`000000oooo0?ooo`0;0?ooo`030000003o
ool0oooo01L0oooo0P00000;0?ooo`8000006P3oool2000002X0oooo0P00000>0?ooo`030000003o
ool0oooo03L0oooo000J0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool0:@3oool0
0`000000oooo0?ooo`030?ooo`800000503oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo00<0oooo0P00000d0?ooo`8000001P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo0140oooo0P0000060?ooo`030000003oool0oooo02T0oooo00<000000?ooo`3oool0AP3oool0
01X0oooo00<000000?ooo`3oool00P0000050?ooo`8000001P3oool200000240oooo0P00000C0?oo
o`030000003oool0oooo02H0oooo00<000000?ooo`3oool00`3oool2000000/0oooo0P00000<0?oo
o`8000002`3oool2000000H0oooo00<000000?ooo`3oool09P3oool00`000000oooo0?ooo`0@0?oo
o`8000008@3oool2000000H0oooo0P0000050?ooo`800000=@3oool001X0oooo00D000000?ooo`3o
ool0oooo000000060?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0103oool20000
00D0oooo0P00000C0?ooo`030000003oool0oooo00/0oooo0P00000B0?ooo`8000004P3oool20000
00`0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo
00X0oooo00<000000?ooo`3oool02@3oool200000180oooo0P00000B0?ooo`8000003P3oool00`00
0000oooo0?ooo`0@0?ooo`8000001P3oool2000000H0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo03<0oooo000J0?ooo`8000006@3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo0P0000050?ooo`8000004`3oool00`000000
oooo0?ooo`0A0?ooo`030000003oool0oooo00<0oooo0P00000<0?ooo`030000003oool0oooo0100
oooo0P00000H0?ooo`8000004`3oool00`000000oooo0?ooo`090?ooo`8000001P3oool00`000000
oooo0?ooo`0A0?ooo`030000003oool0oooo0100oooo0P0000050?ooo`8000001`3oool00`000000
oooo0?ooo`050?ooo`030000003oool0oooo04T0oooo000J0?ooo`030000003oool0oooo00/0oooo
0P00000J0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00`3oool2000001T0oooo
0P00000<0?ooo`030000003oool0oooo01h0oooo00<000000?ooo`3oool05`3oool00`000000oooo
0?ooo`0N0?ooo`030000003oool0oooo00T0oooo0P00000I0?ooo`8000001P3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo01L0oooo0P0000100?ooo`006P3oool00`000000oooo0?oo
o`0<0?ooo`030000003oool0oooo00/0oooo0P00000I0?ooo`030000003oool0oooo00@0oooo0P00
000B0?ooo`030000003oool0oooo02@0oooo0P00000;0?ooo`8000002P3oool2000000/0oooo0P00
000W0?ooo`030000003oool0oooo00l0oooo0P0000070?ooo`030000003oool0oooo01H0oooo0P00
000>0?ooo`030000003oool0oooo03h0oooo000J0?ooo`030000003oool0oooo01/0oooo00<00000
0?ooo`3oool07`3oool00`000000oooo0?ooo`030?ooo`8000001@3oool200000140oooo0P00000;
0?ooo`800000303oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00T0oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`090?ooo`8000002`3oool200000140oooo0P000005
0?ooo`8000001P3oool00`000000oooo0?ooo`0O0?ooo`030000003oool0oooo04d0oooo000J0?oo
o`040000003oool00000000000H0oooo0P00000^0?ooo`8000002`3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`03
0?ooo`800000<03oool2000000H0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0@
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0203oool2000002h0oooo0P000006
0?ooo`800000=03oool001X0oooo00@000000?ooo`3oool000001`3oool00`000000oooo0?ooo`03
0?ooo`8000001@3oool2000000D0oooo0P00000J0?ooo`030000003oool0oooo01H0oooo0P00000O
0?ooo`030000003oool0oooo00l0oooo0P00000:0?ooo`8000004P3oool00`000000oooo0?ooo`0L
0?ooo`8000006@3oool00`000000oooo0?ooo`0G0?ooo`8000001@3oool2000000D0oooo0P000006
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0<P3oool001X0oooo00<000000?oo
o`3oool03`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool0103oool2000000D0oooo0P00000H0?ooo`8000002`3oool00`000000oooo0?ooo`030?oo
o`800000:`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo02P0oooo0P0000060?oo
o`030000003oool0oooo00P0oooo0P00000H0?ooo`8000001@3oool2000000L0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0440oooo000J0?ooo`80
0000103oool200000200oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`80
00004P3oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo0080oooo0P00000N0?ooo`80
00005P3oool2000001h0oooo0P0000050?ooo`030000003oool0oooo0100oooo00<000000?ooo`3o
ool03`3oool2000000H0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0M0?ooo`80
0000>03oool001X0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0:0?ooo`800000
303oool200000180oooo00<000000?ooo`3oool00`3oool2000000D0oooo0P00000M0?ooo`030000
003oool0oooo00<0oooo0P00000H0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool0
5@3oool2000000H0oooo00<000000?ooo`3oool06P3oool2000000D0oooo0P0000060?ooo`030000
003oool0oooo00l0oooo0P00000<0?ooo`8000003@3oool00`000000oooo0?ooo`0f0?ooo`006P3o
ool00`000000oooo0?ooo`0B0?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool05`3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo0P00000N0?ooo`030000003o
ool0oooo00l0oooo0P00000:0?ooo`8000002P3oool2000000X0oooo0P00000B0?ooo`030000003o
ool0oooo01/0oooo0P0000050?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool05`3o
ool00`000000oooo0?ooo`0;0?ooo`030000003oool0oooo04@0oooo000J0?ooo`040000003oool0
0000000002`0oooo0P00000H0?ooo`030000003oool0oooo00<0oooo0P00000M0?ooo`8000002`3o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3o
ool00`000000oooo0?ooo`080?ooo`8000007@3oool2000000H0oooo00<000000?ooo`3oool05@3o
ool2000002`0oooo0P00000d0?ooo`006P3oool010000000oooo0?ooo`00000C0?ooo`800000603o
ool00`000000oooo0?ooo`0:0?ooo`8000004@3oool00`000000oooo0?ooo`030?ooo`8000005`3o
ool00`000000oooo0?ooo`020?ooo`800000;P3oool2000000D0oooo00<000000?ooo`3oool0503o
ool2000000H0oooo00<000000?ooo`3oool03P3oool2000000d0oooo00<000000?ooo`3oool05@3o
ool2000001@0oooo00<000000?ooo`3oool0<P3oool001X0oooo00<000000?ooo`3oool05@3oool0
0`000000oooo0?ooo`0:0?ooo`8000001@3oool200000140oooo00<000000?ooo`3oool05P3oool0
0`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00h0oooo0P00000:0?ooo`8000004@3oool0
0`000000oooo0?ooo`0K0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool03P3oool2
000000D0oooo0P00000=0?ooo`030000003oool0oooo04L0oooo000=0?ooo`030000003oool0oooo
0080oooo0P0000060?ooo`030000003oool0oooo02<0oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`0]0?ooo`800000403oool2000000X0oooo0P00000;0?ooo`030000003oool0oooo
00T0oooo00<000000?ooo`3oool0203oool2000000X0oooo0P00000@0?ooo`800000<03oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo05D0oooo000=0?ooo`050000003oool0oooo0?oo
o`0000000P3oool00`000000oooo0?ooo`030?ooo`8000002P3oool2000000`0oooo0P00000M0?oo
o`8000001@3oool2000001`0oooo00<000000?ooo`3oool00`3oool2000000X0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`020?ooo`8000005P3oool2000000D0oooo00<000000?oo
o`3oool02@3oool00`000000oooo0?ooo`070?ooo`8000001P3oool00`000000oooo0?ooo`0I0?oo
o`8000001@3oool2000001d0oooo0P00000<0?ooo`800000?P3oool000X0oooo1@0000020?ooo`04
0000003oool0oooo000000D0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0;0?oo
o`030000003oool0oooo00T0oooo0P00000A0?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool0503oool2000000/0oooo00<000000?ooo`3oool00P3oool2000001H0oooo00<000000?oo
o`3oool05@3oool00`000000oooo0?ooo`0C0?ooo`8000001@3oool00`000000oooo0?ooo`080?oo
o`8000005`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00h0oooo0P00000<0?oo
o`030000003oool0oooo00/0oooo00<000000?ooo`3oool0?03oool000X0oooo00@000000?ooo`3o
ool000000`3oool010000000oooo0?ooo`0000050?ooo`040000003oool00000000002@0oooo00<0
00000?ooo`3oool02P3oool2000001H0oooo0P00000:0?ooo`030000003oool0oooo00l0oooo00<0
00000?ooo`3oool06@3oool2000000X0oooo0P00000L0?ooo`030000003oool0oooo00l0oooo00<0
00000?ooo`3oool01`3oool2000001H0oooo0P00000=0?ooo`030000003oool0oooo0240oooo0P00
000d0?ooo`002`3oool00`000000oooo000000030?ooo`040000003oool0oooo000000D0oooo00@0
00000?ooo`3oool000001@3oool2000000@0oooo0P0000050?ooo`8000001@3oool2000001L0oooo
00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0D0?ooo`8000002@3oool2000001H0oooo
00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0C0?ooo`8000002@3oool2000001L0oooo
00<000000?ooo`3oool05@3oool00`000000oooo0?ooo`0D0?ooo`8000001@3oool2000000D0oooo
0P0000040?ooo`8000001P3oool00`000000oooo0?ooo`0b0?ooo`00303oool2000000<0oooo00@0
00000?ooo`3oool000001@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo
0P0000050?ooo`8000005P3oool200000280oooo00<000000?ooo`3oool0203oool00`000000oooo
0?ooo`020?ooo`800000;03oool2000000D0oooo00<000000?ooo`3oool0203oool00`000000oooo
0?ooo`0O0?ooo`8000005P3oool2000000D0oooo0P0000050?ooo`030000003oool0oooo00@0oooo
00<000000?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo03T0oooo
000=0?ooo`030000003oool0oooo0080oooo0P0000060?ooo`030000003oool0oooo0240oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`8000004@3oool00`000000oooo0?oo
o`0a0?ooo`030000003oool0oooo00d0oooo0P00000:0?ooo`800000403oool00`000000oooo0?oo
o`0a0?ooo`030000003oool0oooo00h0oooo0P0000050?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool0D`3oool001X0oooo00<000000?ooo`3oool00P3oool2000002X0oooo00<00000
0?ooo`3oool00`3oool2000000@0oooo0P00000?0?ooo`800000403oool2000000l0oooo0P000009
0?ooo`8000002P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00L0oooo0P000009
0?ooo`8000003`3oool200000100oooo0P00000?0?ooo`800000103oool2000000H0oooo00<00000
0?ooo`3oool09`3oool2000003L0oooo000J0?ooo`800000103oool00`000000oooo0?ooo`0F0?oo
o`8000005`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo0P00000:0?oo
o`030000003oool0oooo00l0oooo00<000000?ooo`3oool00P3oool2000000X0oooo00<000000?oo
o`3oool0203oool00`000000oooo0?ooo`020?ooo`800000503oool2000000D0oooo00<000000?oo
o`3oool0203oool00`000000oooo0?ooo`070?ooo`8000001@3oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo00L0oooo0P0000050?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool0503oool2000001T0oooo00<000000?ooo`3oool0=@3oool001X0oooo00<000000?ooo`3o
ool07@3oool00`000000oooo0?ooo`090?ooo`8000005P3oool00`000000oooo0?ooo`030?ooo`80
00003`3oool2000000X0oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0C0?ooo`03
0000003oool0oooo01T0oooo00<000000?ooo`3oool01`3oool2000000l0oooo0P0000060?ooo`03
0000003oool0oooo01<0oooo0P00000<0?ooo`030000003oool0oooo04l0oooo000J0?ooo`040000
003oool00000000000@0oooo0P0000050?ooo`8000002`3oool2000000l0oooo00<000000?ooo`3o
ool06`3oool00`000000oooo0?ooo`020?ooo`8000002P3oool00`000000oooo0?ooo`0M0?ooo`80
00002@3oool2000000X0oooo0P0000090?ooo`800000803oool00`000000oooo0?ooo`070?ooo`80
00001@3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00d0oooo0P00000:0?ooo`80
00001@3oool2000000@0oooo0P00000d0?ooo`006P3oool010000000oooo0?ooo`0000050?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`020?ooo`80
0000103oool2000002L0oooo00<000000?ooo`3oool00P3oool2000000l0oooo0P00000D0?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`03
0000003oool0oooo0180oooo0P00000>0?ooo`8000001@3oool00`000000oooo0?ooo`0T0?ooo`80
0000103oool2000000H0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool0<P3oool001X0oooo00<000000?ooo`3oool0
803oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo0P00000E0?ooo`800000
403oool00`000000oooo0?ooo`020?ooo`8000002P3oool00`000000oooo0?ooo`0<0?ooo`800000
:P3oool200000100oooo00<000000?ooo`3oool01P3oool2000000D0oooo00<000000?ooo`3oool0
3@3oool2000001D0oooo0P0000050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
DP3oool001X0oooo00<000000?ooo`3oool0203oool2000000/0oooo0P00000E0?ooo`030000003o
ool0oooo00<0oooo0P0000040?ooo`8000002@3oool00`000000oooo0?ooo`0D0?ooo`030000003o
ool0oooo01P0oooo00<000000?ooo`3oool03@3oool2000000P0oooo0P00000@0?ooo`030000003o
ool0oooo01P0oooo00<000000?ooo`3oool0503oool00`000000oooo0?ooo`060?ooo`8000001@3o
ool2000000D0oooo00<000000?ooo`3oool04P3oool2000000/0oooo0P00000m0?ooo`006P3oool0
0`000000oooo0?ooo`090?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0203oool2
00000100oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0B0?ooo`8000003`3oool2
000000h0oooo0P00000E0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool04P3oool2
000000h0oooo0P00000?0?ooo`8000005@3oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo00`0oooo0P00000;0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool0>`3oool0
01X0oooo0P00000S0?ooo`030000003oool0oooo00P0oooo0P00000E0?ooo`8000002@3oool00`00
0000oooo0?ooo`0>0?ooo`040000003oool0oooo0?ooo`8000002P3oool00`000000oooo0?ooo`0<
0?ooo`800000503oool2000000l0oooo00<000000?ooo`3oool01`3oool2000000@0oooo00<00000
0?ooo`3oool03P3oool00`000000oooo0?ooo`060?ooo`8000005@3oool2000000/0oooo00<00000
0?ooo`3oool0E03oool001X0oooo00@000000?ooo`00000000002P3oool2000000@0oooo0P00000K
0?ooo`030000003oool0oooo01@0oooo00<000000?ooo`3oool00P3oool2000001T0oooo00<00000
0?ooo`3oool00P3oool2000001@0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0A
0?ooo`8000001@3oool00`000000oooo0?ooo`0F0?ooo`8000001@3oool00`000000oooo0?ooo`0D
0?ooo`030000003oool0oooo01P0oooo0P0000040?ooo`8000002P3oool2000003@0oooo000J0?oo
o`040000003oool0oooo000000/0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`08
0?ooo`8000001@3oool2000002D0oooo00@000000?ooo`3oool0oooo0P00000J0?ooo`030000003o
ool0oooo00`0oooo0P0000090?ooo`800000203oool2000000T0oooo0P00000?0?ooo`030000003o
ool0oooo01L0oooo0P0000040?ooo`030000003oool0oooo02<0oooo0P0000040?ooo`8000002`3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0<P3o
ool001X0oooo00<000000?ooo`3oool07P3oool00`000000oooo0?ooo`040?ooo`030000003oool0
oooo01d0oooo0P0000090?ooo`030000003oool0oooo0080oooo0P00000>0?ooo`8000002@3oool2
000000T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool01P3oool2000000T0oooo0P00000>0?ooo`8000001@3oool0
0`000000oooo0?ooo`060?ooo`8000008@3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo0500oooo000J0?ooo`030000003oool0oooo0080oooo0P00000:0?ooo`8000009@3oool20000
00l0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`020?ooo`8000002@3oool00`00
0000oooo0?ooo`080?ooo`040000003oool0oooo0?ooo`800000:03oool2000000@0oooo00<00000
0?ooo`3oool0203oool00`000000oooo0?ooo`060?ooo`8000001@3oool00`000000oooo0?ooo`0=
0?ooo`030000003oool0oooo00`0oooo0P00000U0?ooo`8000002P3oool2000003L0oooo000J0?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?oo
o`8000006P3oool00`000000oooo0?ooo`0H0?ooo`8000002@3oool00`000000oooo0?ooo`020?oo
o`8000004`3oool00`000000oooo0?ooo`0<0?ooo`800000203oool2000000l0oooo00<000000?oo
o`3oool04@3oool2000000@0oooo00<000000?ooo`3oool01P3oool2000001/0oooo00<000000?oo
o`3oool05`3oool2000000/0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0e0?oo
o`006P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo00P0oooo0P00000I0?ooo`80
00003`3oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo00/0oooo0P0000090?ooo`80
00002@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo0P0000080?ooo`80
00003`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00`0oooo0P00000I0?ooo`80
00002`3oool00`000000oooo0?ooo`1=0?ooo`006P3oool4000000@0oooo0P0000040?ooo`800000
0`3oool2000001D0oooo00<000000?ooo`3oool01`3oool2000000l0oooo00<000000?ooo`3oool0
0P3oool2000002P0oooo00<000000?ooo`3oool0203oool010000000oooo0?ooo`3oool200000180
oooo0P0000050?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool09@3oool2000000D0
oooo00<000000?ooo`3oool0303oool2000000X0oooo00<000000?ooo`3oool04P3oool2000000<0
oooo0P0000040?ooo`800000103oool2000003@0oooo000J0?ooo`040000003oool0oooo000000D0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080
oooo0P0000040?ooo`800000103oool2000000h0oooo00<000000?ooo`3oool0203oool2000000T0
oooo00<000000?ooo`3oool04@3oool2000000d0oooo0P00000D0?ooo`030000003oool0oooo0140
oooo00<000000?ooo`3oool04@3oool2000000d0oooo0P00000D0?ooo`030000003oool0oooo00H0
oooo0P00000;0?ooo`030000003oool0oooo00/0oooo0P0000040?ooo`800000103oool2000000D0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool0<P3oool001X0oooo00<000000?ooo`3oool05`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool2000001@0oooo
00<000000?ooo`3oool04@3oool2000000T0oooo00<000000?ooo`3oool0303oool00`000000oooo
0?ooo`0F0?ooo`800000203oool2000001T0oooo00<000000?ooo`3oool0303oool00`000000oooo
0?ooo`060?ooo`800000503oool00`000000oooo0?ooo`0A0?ooo`8000001@3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0B@3oool001X0oooo00<0
00000?ooo`3oool01`3oool2000000X0oooo0P00000D0?ooo`030000003oool0oooo0080oooo0P00
00030?ooo`8000007P3oool010000000oooo0?ooo`3oool2000000h0oooo0P0000080?ooo`800000
4`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0100oooo0P0000080?ooo`800000
3P3oool2000000@0oooo00<000000?ooo`3oool06`3oool2000000@0oooo0P0000040?ooo`030000
003oool0oooo0140oooo0P00000:0?ooo`800000?03oool001X0oooo00<000000?ooo`3oool0203o
ool00`000000oooo0?ooo`090?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`020?ooo`8000007@3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo00L0oooo00@000000?ooo`3oool0oooo0P00000V0?ooo`800000103oool00`000000oooo
0?ooo`070?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool06P3oool2000000D0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo00T0oooo
00<000000?ooo`3oool0>P3oool001X0oooo00<000000?ooo`3oool0:P3oool2000000h0oooo00@0
00000?ooo`3oool0oooo0P00000C0?ooo`8000003@3oool2000001<0oooo00<000000?ooo`3oool0
2`3oool2000000P0oooo0P00000>0?ooo`030000003oool0oooo0100oooo0P00000=0?ooo`800000
4`3oool2000000@0oooo00<000000?ooo`3oool02`3oool2000005l0oooo000J0?ooo`040000003o
ool00000000000<0oooo0P00000:0?ooo`8000006@3oool00`000000oooo0?ooo`0A0?ooo`030000
003oool0oooo0080oooo0P00000>0?ooo`030000003oool0oooo00`0oooo00<000000?ooo`3oool0
5@3oool2000000T0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`800000
603oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00/0oooo0P0000050?ooo`030000
003oool0oooo0140oooo00<000000?ooo`3oool05P3oool2000000X0oooo0P0000030?ooo`800000
=03oool001X0oooo0P0000000`3oool000000?ooo`030?ooo`030000003oool0oooo00T0oooo00@0
00000?ooo`3oool0oooo0P00000:0?ooo`8000004P3oool2000000h0oooo00<000000?ooo`3oool0
0P3oool2000001`0oooo0P00000B0?ooo`040000003oool0oooo0?ooo`8000004P3oool2000000@0
oooo00<000000?ooo`3oool0403oool2000001`0oooo0P0000040?ooo`030000003oool0oooo00/0
oooo0P00000B0?ooo`8000002P3oool2000000@0oooo00<000000?ooo`3oool02@3oool00`000000
oooo0?ooo`020?ooo`030000003oool0oooo0380oooo000J0?ooo`030000003oool0oooo01H0oooo
00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0180oooo
00@000000?ooo`3oool0oooo0P00000=0?ooo`8000002@3oool00`000000oooo0?ooo`0E0?ooo`03
0000003oool0oooo0140oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`050?ooo`80
00003@3oool2000000D0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0A0?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool0B03oool001X0oooo00@000000?ooo`3oool0
oooo0P00000:0?ooo`8000007@3oool2000000<0oooo0P00000M0?ooo`040000003oool0oooo0?oo
o`8000002@3oool00`000000oooo0?ooo`0I0?ooo`800000203oool2000000P0oooo0P0000080?oo
o`800000703oool00`000000oooo0?ooo`060?ooo`800000103oool00`000000oooo0?ooo`0J0?oo
o`8000000`3oool2000001d0oooo0P00000:0?ooo`800000=P3oool001X0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00L0oooo0P00000C0?ooo`03
0000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool200000180oooo0P0000080?ooo`03
0000003oool0oooo0080oooo0P00000<0?ooo`800000203oool2000000P0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool01@3oool2000000P0oooo0P00000=0?ooo`800000103oool00`000000oooo0?ooo`050?ooo`80
00004P3oool2000000D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0@0?ooo`80
00002P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo03@0oooo000J0?ooo`030000
003oool0oooo01T0oooo00<000000?ooo`3oool0703oool010000000oooo0?ooo`3oool2000000h0
oooo00<000000?ooo`3oool0303oool010000000oooo0?ooo`3oool2000000P0oooo00<000000?oo
o`3oool01`3oool010000000oooo0?ooo`3oool2000002@0oooo0P0000040?ooo`030000003oool0
oooo00L0oooo00<000000?ooo`3oool01@3oool2000000D0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`0;0?ooo`800000103oool00`000000oooo0?ooo`0L0?ooo`030000003oool0
oooo04/0oooo000J0?ooo`040000003oool00000000000T0oooo0P0000030?ooo`8000007@3oool2
000000d0oooo00<000000?ooo`3oool01`3oool2000000`0oooo0P0000090?ooo`040000003oool0
oooo0?ooo`8000004P3oool00`000000oooo0?ooo`0:0?ooo`800000203oool2000000d0oooo00<0
00000?ooo`3oool03`3oool2000000@0oooo00<000000?ooo`3oool01P3oool2000000`0oooo0P00
000:0?ooo`030000003oool0oooo00X0oooo0P00000M0?ooo`8000000`3oool2000000T0oooo0P00
000d0?ooo`002`3oool3000000@0oooo0P0000060?ooo`040000003oool0oooo000000X0oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`070?ooo`8000000`3oool2000000h0oooo00<0
00000?ooo`3oool01P3oool2000000h0oooo00@000000?ooo`3oool0oooo0P0000080?ooo`030000
003oool0oooo00`0oooo00<000000?ooo`3oool02P3oool200000180oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`0?0?ooo`8000003@3oool00`000000oooo0?ooo`0<0?ooo`030000
003oool0oooo00D0oooo0P0000040?ooo`030000003oool0oooo00/0oooo0P0000090?ooo`030000
003oool0oooo00/0oooo0P0000030?ooo`8000002P3oool00`000000oooo0?ooo`020?ooo`030000
003oool0oooo00P0oooo00<000000?ooo`3oool0<P3oool000h0oooo00@000000?ooo`3oool00000
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo01/0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`0F0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool0
6@3oool2000000P0oooo00<000000?ooo`3oool0;@3oool00`000000oooo0?ooo`050?ooo`800000
703oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo01H0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`1=0?ooo`003P3oool010000000oooo0?ooo`0000020?ooo`030000
003oool0oooo00<0oooo0P00001[0?ooo`030000003oool0oooo03l0oooo00<000000?ooo`3oool0
W03oool000/0oooo0`0000030?ooo`040000003oool0oooo000000D0oooo00<000000?ooo`3oool0
o`3ooom<0?ooo`002`3oool00`000000oooo0?ooo`030?ooo`040000003oool0oooo000000D0oooo
00<000000?ooo`3oool0o`3ooom<0?ooo`002`3oool00`000000oooo0?ooo`030?ooo`040000003o
ool0oooo000000D0oooo00<000000?ooo`3oool0o`3ooom<0?ooo`002`3oool4000000<0oooo0P00
003o0?oooeD0oooo003o0?ooofT0oooo003o0?ooofT0oooo003o0?ooofT0oooo003o0?ooofT0oooo
003o0?ooofT0oooo003o0?ooofT0oooo001Z0?ooo`<00000>03oool300000<00oooo001]0?ooo`03
0000003oool0oooo03H0oooo00<000000?ooo`3oool0_`3oool005d0oooo0P0000020?ooo`D00000
00<0oooo0000000000000P3oool300000080oooo1@0000020?ooo`<000002P3oool010000000oooo
0?ooo`3oool3000000@0oooo00<000000?ooo`3oool01`3oool300000080oooo00<000000?ooo`3o
ool00`0000020?ooo`<00000103oool3000000<0oooo0P0000000`3oool000000?ooo`020?ooo`80
00001`3oool4000000<0oooo1@000000103oool000000000000000050?ooo`8000000`3oool20000
00030?ooo`00000000000080000000<0oooo0000000000000P3oool2000000@0oooo0`00001f0?oo
o`00GP3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo00000080oooo00@00000
0?ooo`3oool00000103oool01@000000oooo0?ooo`3oool0000000`0oooo00<000000?ooo`000000
1@3oool00`000000oooo0?ooo`0<0?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3o
ool00`3oool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo00<0oooo00@000000?oo
o`3oool000000`3oool010000000oooo0?ooo`0000070?ooo`040000003oool0oooo000000@0oooo
00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`00
00000P3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo00000080oooo00@00000
0?ooo`3oool00000N@3oool005h0oooo100000020?ooo`040000003oool0oooo00000080oooo00@0
00000?ooo`3oool00000103oool010000000oooo0?ooo`3oool3000000X0oooo00<000000?ooo`00
00000`3oool2000000l0oooo0`0000030?ooo`<000000`3oool010000000oooo0?ooo`0000020?oo
o`<00000103oool3000000<0oooo00<000000?ooo`3oool0203oool010000000oooo0?ooo`000004
0?ooo`030000003oool0oooo0080oooo0P0000040?ooo`030000003oool0oooo00<0oooo0`000002
0?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3oool00`3oool3000007L0oooo0002
0?ooo`@000000`3oool5000000040?ooo`0000000000000000D0oooo0P0000030?ooo`80000000<0
oooo0000000000000P0000000`3oool00000000000020?ooo`800000103oool3000002d0oooo00@0
00000?ooo`3oool000000P3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo0000
00@0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003oool0oooo00H0oooo00L00000
0?ooo`3oool0oooo0000003oool000000140oooo00@000000?ooo`3oool000000P3oool010000000
oooo0?ooo`0000020?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool000001@3o
ool01@000000oooo0?ooo`3oool0000000X0oooo00@000000?ooo`3oool00000103oool01@000000
oooo0?ooo`3oool0000000H0oooo00<000000?ooo`3oool01@3oool010000000oooo0?ooo`000002
0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool000000P3oool00`000000oooo
0?ooo`1d0?ooo`000`3oool010000000oooo0?ooo`0000040?ooo`030000003oool0oooo00@0oooo
00D000000?ooo`3oool0oooo000000020?ooo`030000003oool000000080oooo00@000000?ooo`3o
ool000000P3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo00000340oooo00H0
00000?ooo`000000oooo0000003oool2000000@0oooo0P0000000`3oool000000?ooo`020?ooo`03
0000003oool0oooo0080oooo0P0000080?ooo`8000000P3oool200000080oooo0`00000>0?ooo`04
0000003oool0oooo000000<0oooo0P0000020?ooo`040000003oool00000000000@0oooo0P000004
0?ooo`8000000`3oool4000000P0oooo00@000000?ooo`3oool000000`3oool2000000@0oooo0`00
00020?ooo`@000000`3oool200000080oooo00@000000?ooo`0000000000103oool2000000@0oooo
0P00001g0?ooo`000`3oool010000000oooo0?ooo`0000040?ooo`030000003oool0oooo0080oooo
0P0000040?ooo`030000003oool0oooo00<0oooo0`0000020?ooo`040000003oool0oooo00000080
oooo00<000000?ooo`3oool00`3oool3000002h0oooo0`00000A0?ooo`030000003oool0oooo02@0
oooo1000000L0?ooo`030000003oool0oooo00L0oooo1000000@0?ooo`030000003oool0oooo08h0
oooo00030?ooo`040000003oool0oooo000000@0oooo00D000000?ooo`3oool0oooo000000060?oo
o`030000003oool0oooo00D0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`00
00020?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3oool0?P3oool2000004H0oooo
00<000000?ooo`3oool0403oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo08h0oooo
00030?ooo`040000003oool0oooo000000<0oooo0P0000040?ooo`<000000P3oool4000000<0oooo
0P0000020?ooo`040000003oool00000000000@0oooo0P0000040?ooo`800000o`3oooli0?ooo`00
0P3oool400000100oooo00<000000?ooo`3oool0o`3ooom@0?ooo`002`3oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo0?l0ooooD03oool00?l0ooooJ@3oool00?l0ooooJ@3oool00?l0
ooooJ@3oool00?l0ooooJ@3oool00001\
\>"],
  ImageRangeCache->{{{0, 359}, {221.375, 0}} -> {-5.48512, -5.15497, 
  0.210955, 0.283703}}]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(X = FindAngles[100]; \nLength[X]\), 
    \(Angles = Sort[\(Transpose[X]\)[\([1]\)]]; \n
    AnglesAndRepeats = 
      Transpose[{\(Transpose[X]\)[\([1]\)], \t\t\(Transpose[X]\)[\([3]\)]}]; 
    \nListPlot[Angles, PlotLabel -> "\<Rational Angles\>", 
      AxesLabel -> {"\<\>", "\<Angle\>"}]; \n
    ListPlot[AnglesAndRepeats, \ 
      PlotLabel -> "\<Angle vs. Repeat Distance\>", \ 
      AxesLabel -> {"\<Angle\>", "\<Distance\>"}]; \)}], "Input"],

Cell[BoxData[
    \(3689\)], "Output"],

Cell[CellGroupData[{

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.000258168 0.0147151 0.00981006 [
[.15289 .00222 -9 -9 ]
[.15289 .00222 9 0 ]
[.28198 .00222 -12 -9 ]
[.28198 .00222 12 0 ]
[.41106 .00222 -12 -9 ]
[.41106 .00222 12 0 ]
[.54015 .00222 -12 -9 ]
[.54015 .00222 12 0 ]
[.66923 .00222 -12 -9 ]
[.66923 .00222 12 0 ]
[.79831 .00222 -12 -9 ]
[.79831 .00222 12 0 ]
[.9274 .00222 -12 -9 ]
[.9274 .00222 12 0 ]
[.01131 .11282 -12 -4.5 ]
[.01131 .11282 0 4.5 ]
[.01131 .21092 -12 -4.5 ]
[.01131 .21092 0 4.5 ]
[.01131 .30902 -12 -4.5 ]
[.01131 .30902 0 4.5 ]
[.01131 .40712 -12 -4.5 ]
[.01131 .40712 0 4.5 ]
[.01131 .50522 -12 -4.5 ]
[.01131 .50522 0 4.5 ]
[.01131 .60332 -12 -4.5 ]
[.01131 .60332 0 4.5 ]
[.02381 .64303 -17 0 ]
[.02381 .64303 17 9.8125 ]
[.5 .63053 -47 0 ]
[.5 .63053 47 10.125 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.15289 .01472 m
.15289 .02097 L
s
[(500)] .15289 .00222 0 1 Mshowa
.28198 .01472 m
.28198 .02097 L
s
[(1000)] .28198 .00222 0 1 Mshowa
.41106 .01472 m
.41106 .02097 L
s
[(1500)] .41106 .00222 0 1 Mshowa
.54015 .01472 m
.54015 .02097 L
s
[(2000)] .54015 .00222 0 1 Mshowa
.66923 .01472 m
.66923 .02097 L
s
[(2500)] .66923 .00222 0 1 Mshowa
.79831 .01472 m
.79831 .02097 L
s
[(3000)] .79831 .00222 0 1 Mshowa
.9274 .01472 m
.9274 .02097 L
s
[(3500)] .9274 .00222 0 1 Mshowa
.125 Mabswid
.04963 .01472 m
.04963 .01847 L
s
.07544 .01472 m
.07544 .01847 L
s
.10126 .01472 m
.10126 .01847 L
s
.12708 .01472 m
.12708 .01847 L
s
.17871 .01472 m
.17871 .01847 L
s
.20453 .01472 m
.20453 .01847 L
s
.23034 .01472 m
.23034 .01847 L
s
.25616 .01472 m
.25616 .01847 L
s
.30779 .01472 m
.30779 .01847 L
s
.33361 .01472 m
.33361 .01847 L
s
.35943 .01472 m
.35943 .01847 L
s
.38524 .01472 m
.38524 .01847 L
s
.43688 .01472 m
.43688 .01847 L
s
.46269 .01472 m
.46269 .01847 L
s
.48851 .01472 m
.48851 .01847 L
s
.51433 .01472 m
.51433 .01847 L
s
.56596 .01472 m
.56596 .01847 L
s
.59178 .01472 m
.59178 .01847 L
s
.6176 .01472 m
.6176 .01847 L
s
.64341 .01472 m
.64341 .01847 L
s
.69505 .01472 m
.69505 .01847 L
s
.72086 .01472 m
.72086 .01847 L
s
.74668 .01472 m
.74668 .01847 L
s
.7725 .01472 m
.7725 .01847 L
s
.82413 .01472 m
.82413 .01847 L
s
.84995 .01472 m
.84995 .01847 L
s
.87576 .01472 m
.87576 .01847 L
s
.90158 .01472 m
.90158 .01847 L
s
.95321 .01472 m
.95321 .01847 L
s
.97903 .01472 m
.97903 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
.02381 .11282 m
.03006 .11282 L
s
[(10)] .01131 .11282 1 0 Mshowa
.02381 .21092 m
.03006 .21092 L
s
[(20)] .01131 .21092 1 0 Mshowa
.02381 .30902 m
.03006 .30902 L
s
[(30)] .01131 .30902 1 0 Mshowa
.02381 .40712 m
.03006 .40712 L
s
[(40)] .01131 .40712 1 0 Mshowa
.02381 .50522 m
.03006 .50522 L
s
[(50)] .01131 .50522 1 0 Mshowa
.02381 .60332 m
.03006 .60332 L
s
[(60)] .01131 .60332 1 0 Mshowa
.125 Mabswid
.02381 .03434 m
.02756 .03434 L
s
.02381 .05396 m
.02756 .05396 L
s
.02381 .07358 m
.02756 .07358 L
s
.02381 .0932 m
.02756 .0932 L
s
.02381 .13244 m
.02756 .13244 L
s
.02381 .15206 m
.02756 .15206 L
s
.02381 .17168 m
.02756 .17168 L
s
.02381 .1913 m
.02756 .1913 L
s
.02381 .23054 m
.02756 .23054 L
s
.02381 .25016 m
.02756 .25016 L
s
.02381 .26978 m
.02756 .26978 L
s
.02381 .2894 m
.02756 .2894 L
s
.02381 .32864 m
.02756 .32864 L
s
.02381 .34826 m
.02756 .34826 L
s
.02381 .36788 m
.02756 .36788 L
s
.02381 .3875 m
.02756 .3875 L
s
.02381 .42674 m
.02756 .42674 L
s
.02381 .44636 m
.02756 .44636 L
s
.02381 .46598 m
.02756 .46598 L
s
.02381 .4856 m
.02756 .4856 L
s
.02381 .52484 m
.02756 .52484 L
s
.02381 .54446 m
.02756 .54446 L
s
.02381 .56408 m
.02756 .56408 L
s
.02381 .5837 m
.02756 .5837 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
gsave
.02381 .64303 -78 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Angle) show
93.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
gsave
.5 .63053 -108 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.125000 moveto
 558.000000 14.125000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.125000 lineto
closepath
clip newpath
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Rational) show
117.000000 11.562500 moveto
(Angles) show
153.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.008 w
.02407 .01472 Mdot
.02433 .01754 Mdot
.02458 .01759 Mdot
.02484 .01764 Mdot
.0251 .01769 Mdot
.02536 .01775 Mdot
.02562 .01781 Mdot
.02587 .01787 Mdot
.02613 .01793 Mdot
.02639 .01799 Mdot
.02665 .01806 Mdot
.02691 .01813 Mdot
.02717 .0182 Mdot
.02742 .01828 Mdot
.02768 .01836 Mdot
.02794 .01845 Mdot
.0282 .01853 Mdot
.02846 .01862 Mdot
.02871 .01872 Mdot
.02897 .01882 Mdot
.02923 .01893 Mdot
.02949 .01904 Mdot
.02975 .01916 Mdot
.03001 .01929 Mdot
.03026 .01942 Mdot
.03052 .01956 Mdot
.03078 .01971 Mdot
.03104 .01987 Mdot
.0313 .02003 Mdot
.03155 .02022 Mdot
.03181 .02041 Mdot
.03207 .02051 Mdot
.03233 .02062 Mdot
.03259 .02072 Mdot
.03285 .02084 Mdot
.0331 .02096 Mdot
.03336 .02108 Mdot
.03362 .02121 Mdot
.03388 .02134 Mdot
.03414 .02148 Mdot
.03439 .02162 Mdot
.03465 .02177 Mdot
.03491 .02193 Mdot
.03517 .02209 Mdot
.03543 .02226 Mdot
.03569 .02244 Mdot
.03594 .02263 Mdot
.0362 .02283 Mdot
.03646 .02304 Mdot
.03672 .02318 Mdot
.03698 .02325 Mdot
.03723 .02333 Mdot
.03749 .02349 Mdot
.03775 .02365 Mdot
.03801 .02373 Mdot
.03827 .02381 Mdot
.03853 .02399 Mdot
.03878 .02417 Mdot
.03904 .02426 Mdot
.0393 .02435 Mdot
.03956 .02455 Mdot
.03982 .02475 Mdot
.04007 .02486 Mdot
.04033 .02496 Mdot
.04059 .02518 Mdot
.04085 .02541 Mdot
.04111 .02553 Mdot
.04136 .02565 Mdot
.04162 .0259 Mdot
.04188 .0261 Mdot
.04214 .02617 Mdot
.0424 .0263 Mdot
.04266 .02644 Mdot
.04291 .02651 Mdot
.04317 .02673 Mdot
.04343 .02696 Mdot
.04369 .02704 Mdot
.04395 .02719 Mdot
.0442 .02736 Mdot
.04446 .02744 Mdot
.04472 .02769 Mdot
.04498 .02796 Mdot
.04524 .02805 Mdot
.0455 .02823 Mdot
.04575 .02842 Mdot
.04601 .02852 Mdot
.04627 .02882 Mdot
.04653 .02907 Mdot
.04679 .02913 Mdot
.04704 .02924 Mdot
.0473 .02933 Mdot
.04756 .02946 Mdot
.04782 .0296 Mdot
.04808 .02969 Mdot
.04834 .02981 Mdot
.04859 .02988 Mdot
.04885 .03016 Mdot
.04911 .03046 Mdot
.04937 .03054 Mdot
.04963 .03067 Mdot
.04988 .03078 Mdot
.05014 .03094 Mdot
.0504 .0311 Mdot
.05066 .03121 Mdot
.05092 .03135 Mdot
.05118 .03144 Mdot
.05143 .03179 Mdot
.05169 .03209 Mdot
.05195 .03216 Mdot
.05221 .03225 Mdot
.05247 .03241 Mdot
.05272 .03254 Mdot
.05298 .03274 Mdot
.05324 .03294 Mdot
.0535 .03308 Mdot
.05376 .03325 Mdot
.05402 .03336 Mdot
.05427 .03343 Mdot
.05453 .0338 Mdot
.05479 .03418 Mdot
.05505 .03426 Mdot
.05531 .03437 Mdot
.05556 .03446 Mdot
.05582 .03458 Mdot
.05608 .03474 Mdot
.05634 .03481 Mdot
.0566 .03499 Mdot
.05686 .03517 Mdot
.05711 .03524 Mdot
.05737 .03542 Mdot
.05763 .03555 Mdot
.05789 .03564 Mdot
.05815 .03578 Mdot
.0584 .03587 Mdot
.05866 .03593 Mdot
.05892 .03634 Mdot
.05918 .03676 Mdot
.05944 .03683 Mdot
.05969 .03693 Mdot
.05995 .03708 Mdot
.06021 .03719 Mdot
.06047 .03734 Mdot
.06073 .03748 Mdot
.06099 .03756 Mdot
.06124 .03765 Mdot
.0615 .03788 Mdot
.06176 .03812 Mdot
.06202 .03822 Mdot
.06228 .0383 Mdot
.06253 .03845 Mdot
.06279 .03861 Mdot
.06305 .03874 Mdot
.06331 .03892 Mdot
.06357 .03904 Mdot
.06383 .03913 Mdot
.06408 .03919 Mdot
.06434 .03966 Mdot
.0646 .04009 Mdot
.06486 .04015 Mdot
.06512 .04022 Mdot
.06537 .04032 Mdot
.06563 .04045 Mdot
.06589 .04054 Mdot
.06615 .04066 Mdot
.06641 .04081 Mdot
.06667 .04101 Mdot
.06692 .04119 Mdot
.06718 .04129 Mdot
.06744 .04142 Mdot
.0677 .04149 Mdot
.06796 .04174 Mdot
.06821 .04199 Mdot
.06847 .04206 Mdot
.06873 .04219 Mdot
.06899 .04231 Mdot
.06925 .04251 Mdot
.06951 .04274 Mdot
.06976 .04291 Mdot
.07002 .04305 Mdot
.07028 .04316 Mdot
.07054 .04332 Mdot
.0708 .04344 Mdot
.07105 .04354 Mdot
.07131 .04361 Mdot
.07157 .04366 Mdot
.07183 .04419 Mdot
.07209 .04473 Mdot
.07235 .0448 Mdot
.0726 .04487 Mdot
.07286 .04497 Mdot
.07312 .04511 Mdot
.07338 .0453 Mdot
.07364 .04543 Mdot
.07389 .04559 Mdot
.07415 .04572 Mdot
.07441 .0458 Mdot
.07467 .04589 Mdot
.07493 .04609 Mdot
.07518 .04627 Mdot
.07544 .04634 Mdot
.0757 .0465 Mdot
.07596 .04667 Mdot
.07622 .04678 Mdot
.07648 .04684 Mdot
.07673 .04713 Mdot
.07699 .04743 Mdot
.07725 .04749 Mdot
.07751 .0476 Mdot
.07777 .04779 Mdot
.07802 .04796 Mdot
.07828 .04804 Mdot
.07854 .04825 Mdot
.0788 .04848 Mdot
.07906 .04858 Mdot
.07932 .04867 Mdot
.07957 .04883 Mdot
.07983 .04903 Mdot
.08009 .04919 Mdot
.08035 .04933 Mdot
.08061 .04944 Mdot
.08086 .04962 Mdot
.08112 .04975 Mdot
.08138 .04986 Mdot
.08164 .04994 Mdot
.0819 .05001 Mdot
.08216 .05007 Mdot
.08241 .05072 Mdot
.08267 .05135 Mdot
.08293 .0514 Mdot
.08319 .05146 Mdot
.08345 .05154 Mdot
.0837 .05163 Mdot
.08396 .05175 Mdot
.08422 .0519 Mdot
.08448 .05199 Mdot
.08474 .0521 Mdot
.085 .05223 Mdot
.08525 .05239 Mdot
.08551 .05259 Mdot
.08577 .05266 Mdot
.08603 .05283 Mdot
.08629 .05304 Mdot
.08654 .05316 Mdot
.0868 .05329 Mdot
.08706 .05336 Mdot
.08732 .05359 Mdot
.08758 .05388 Mdot
.08784 .05399 Mdot
.08809 .05408 Mdot
.08835 .05422 Mdot
.08861 .05439 Mdot
.08887 .0545 Mdot
.08913 .05466 Mdot
.08938 .05476 Mdot
.08964 .05482 Mdot
.0899 .05521 Mdot
.09016 .0556 Mdot
.09042 .05567 Mdot
.09067 .05578 Mdot
.09093 .05594 Mdot
.09119 .05607 Mdot
.09145 .05624 Mdot
.09171 .05641 Mdot
.09197 .05651 Mdot
.09222 .05663 Mdot
.09248 .0567 Mdot
.09274 .05696 Mdot
.093 .05725 Mdot
.09326 .05733 Mdot
.09351 .05749 Mdot
.09377 .05764 Mdot
.09403 .05771 Mdot
.09429 .0579 Mdot
.09455 .05812 Mdot
.09481 .05822 Mdot
.09506 .05831 Mdot
.09532 .05848 Mdot
.09558 .05869 Mdot
.09584 .05888 Mdot
.0961 .05903 Mdot
.09635 .05916 Mdot
.09661 .05928 Mdot
.09687 .05938 Mdot
.09713 .05955 Mdot
.09739 .05969 Mdot
.09765 .0598 Mdot
.0979 .0599 Mdot
.09816 .05998 Mdot
.09842 .06005 Mdot
.09868 .06011 Mdot
.09894 .06016 Mdot
.09919 .06097 Mdot
.09945 .06181 Mdot
.09971 .06186 Mdot
.09997 .06193 Mdot
.10023 .062 Mdot
.10049 .06209 Mdot
.10074 .0622 Mdot
.101 .06232 Mdot
.10126 .06248 Mdot
.10152 .06257 Mdot
.10178 .06267 Mdot
.10203 .06279 Mdot
.10229 .06293 Mdot
.10255 .06309 Mdot
.10281 .06327 Mdot
.10307 .06341 Mdot
.10333 .06349 Mdot
.10358 .06357 Mdot
.10384 .06376 Mdot
.1041 .06397 Mdot
.10436 .06409 Mdot
.10462 .06422 Mdot
.10487 .06429 Mdot
.10513 .06451 Mdot
.10539 .06477 Mdot
.10565 .06486 Mdot
.10591 .06506 Mdot
.10617 .06519 Mdot
.10642 .06528 Mdot
.10668 .0654 Mdot
.10694 .06548 Mdot
.1072 .06581 Mdot
.10746 .06613 Mdot
.10771 .0662 Mdot
.10797 .0663 Mdot
.10823 .06649 Mdot
.10849 .06665 Mdot
.10875 .06672 Mdot
.109 .06691 Mdot
.10926 .0671 Mdot
.10952 .06719 Mdot
.10978 .06739 Mdot
.11004 .06754 Mdot
.1103 .06766 Mdot
.11055 .06783 Mdot
.11081 .06796 Mdot
.11107 .06805 Mdot
.11133 .06811 Mdot
.11159 .06863 Mdot
.11184 .06911 Mdot
.1121 .06917 Mdot
.11236 .06924 Mdot
.11262 .06933 Mdot
.11288 .06946 Mdot
.11314 .06965 Mdot
.11339 .06977 Mdot
.11365 .06994 Mdot
.11391 .07008 Mdot
.11417 .07017 Mdot
.11443 .07026 Mdot
.11468 .07048 Mdot
.11494 .0707 Mdot
.1152 .07078 Mdot
.11546 .07096 Mdot
.11572 .07109 Mdot
.11598 .07119 Mdot
.11623 .07131 Mdot
.11649 .0714 Mdot
.11675 .07145 Mdot
.11701 .07179 Mdot
.11727 .07214 Mdot
.11752 .07221 Mdot
.11778 .0723 Mdot
.11804 .07246 Mdot
.1183 .07258 Mdot
.11856 .07275 Mdot
.11882 .07292 Mdot
.11907 .07302 Mdot
.11933 .07315 Mdot
.11959 .07322 Mdot
.11985 .0735 Mdot
.12011 .07382 Mdot
.12036 .07391 Mdot
.12062 .0741 Mdot
.12088 .07427 Mdot
.12114 .07435 Mdot
.1214 .07458 Mdot
.12166 .07479 Mdot
.12191 .07486 Mdot
.12217 .07499 Mdot
.12243 .07511 Mdot
.12269 .07533 Mdot
.12295 .07552 Mdot
.1232 .07562 Mdot
.12346 .0757 Mdot
.12372 .07587 Mdot
.12398 .07609 Mdot
.12424 .07628 Mdot
.12449 .07645 Mdot
.12475 .0766 Mdot
.12501 .07674 Mdot
.12527 .07687 Mdot
.12553 .07698 Mdot
.12579 .07708 Mdot
.12604 .07726 Mdot
.1263 .07742 Mdot
.12656 .07755 Mdot
.12682 .07767 Mdot
.12708 .07777 Mdot
.12733 .07786 Mdot
.12759 .07794 Mdot
.12785 .07801 Mdot
.12811 .07808 Mdot
.12837 .07813 Mdot
.12863 .07819 Mdot
.12888 .07933 Mdot
.12914 .08047 Mdot
.1294 .08052 Mdot
.12966 .08058 Mdot
.12992 .08064 Mdot
.13017 .08071 Mdot
.13043 .08079 Mdot
.13069 .08087 Mdot
.13095 .08097 Mdot
.13121 .08109 Mdot
.13147 .08121 Mdot
.13172 .08136 Mdot
.13198 .08154 Mdot
.13224 .08164 Mdot
.1325 .08175 Mdot
.13276 .08187 Mdot
.13301 .082 Mdot
.13327 .08215 Mdot
.13353 .08231 Mdot
.13379 .08249 Mdot
.13405 .0827 Mdot
.13431 .08285 Mdot
.13456 .08294 Mdot
.13482 .08302 Mdot
.13508 .08321 Mdot
.13534 .08342 Mdot
.1356 .08353 Mdot
.13585 .08365 Mdot
.13611 .0839 Mdot
.13637 .08412 Mdot
.13663 .08419 Mdot
.13689 .08435 Mdot
.13715 .08452 Mdot
.1374 .08461 Mdot
.13766 .08489 Mdot
.13792 .08515 Mdot
.13818 .08521 Mdot
.13844 .08533 Mdot
.13869 .08542 Mdot
.13895 .08557 Mdot
.13921 .08573 Mdot
.13947 .08584 Mdot
.13973 .08598 Mdot
.13999 .08606 Mdot
.14024 .08612 Mdot
.1405 .08644 Mdot
.14076 .08678 Mdot
.14102 .08685 Mdot
.14128 .08697 Mdot
.14153 .08716 Mdot
.14179 .08732 Mdot
.14205 .0874 Mdot
.14231 .08758 Mdot
.14257 .08778 Mdot
.14282 .08786 Mdot
.14308 .08805 Mdot
.14334 .08819 Mdot
.1436 .0883 Mdot
.14386 .08846 Mdot
.14412 .08857 Mdot
.14437 .08865 Mdot
.14463 .08871 Mdot
.14489 .08916 Mdot
.14515 .0896 Mdot
.14541 .08966 Mdot
.14566 .08973 Mdot
.14592 .08984 Mdot
.14618 .08998 Mdot
.14644 .09008 Mdot
.1467 .09021 Mdot
.14696 .09038 Mdot
.14721 .09061 Mdot
.14747 .09082 Mdot
.14773 .09095 Mdot
.14799 .0911 Mdot
.14825 .09119 Mdot
.1485 .0915 Mdot
.14876 .09177 Mdot
.14902 .09183 Mdot
.14928 .09193 Mdot
.14954 .09211 Mdot
.1498 .09227 Mdot
.15005 .09234 Mdot
.15031 .09254 Mdot
.15057 .09278 Mdot
.15083 .09288 Mdot
.15109 .09297 Mdot
.15134 .09314 Mdot
.1516 .09335 Mdot
.15186 .09352 Mdot
.15212 .09366 Mdot
.15238 .09379 Mdot
.15264 .09389 Mdot
.15289 .09398 Mdot
.15315 .09414 Mdot
.15341 .09425 Mdot
.15367 .09435 Mdot
.15393 .09443 Mdot
.15418 .0945 Mdot
.15444 .09456 Mdot
.1547 .09529 Mdot
.15496 .09603 Mdot
.15522 .09609 Mdot
.15548 .09616 Mdot
.15573 .09624 Mdot
.15599 .09635 Mdot
.15625 .09647 Mdot
.15651 .09663 Mdot
.15677 .09673 Mdot
.15702 .09684 Mdot
.15728 .09698 Mdot
.15754 .09713 Mdot
.1578 .09732 Mdot
.15806 .0974 Mdot
.15831 .09756 Mdot
.15857 .09774 Mdot
.15883 .09785 Mdot
.15909 .09796 Mdot
.15935 .09823 Mdot
.15961 .09846 Mdot
.15986 .09855 Mdot
.16012 .09874 Mdot
.16038 .09886 Mdot
.16064 .09895 Mdot
.1609 .09907 Mdot
.16115 .09914 Mdot
.16141 .09947 Mdot
.16167 .09978 Mdot
.16193 .09985 Mdot
.16219 .09996 Mdot
.16245 .10015 Mdot
.1627 .10031 Mdot
.16296 .10038 Mdot
.16322 .10058 Mdot
.16348 .10079 Mdot
.16374 .10088 Mdot
.16399 .10096 Mdot
.16425 .10109 Mdot
.16451 .10126 Mdot
.16477 .10139 Mdot
.16503 .10158 Mdot
.16529 .10172 Mdot
.16554 .10182 Mdot
.1658 .1019 Mdot
.16606 .10196 Mdot
.16632 .1025 Mdot
.16658 .10301 Mdot
.16683 .10306 Mdot
.16709 .10313 Mdot
.16735 .10322 Mdot
.16761 .10333 Mdot
.16787 .10349 Mdot
.16813 .10359 Mdot
.16838 .10372 Mdot
.16864 .10389 Mdot
.1689 .1041 Mdot
.16916 .10428 Mdot
.16942 .10438 Mdot
.16967 .10451 Mdot
.16993 .10457 Mdot
.17019 .1048 Mdot
.17045 .10508 Mdot
.17071 .1052 Mdot
.17097 .10529 Mdot
.17122 .10545 Mdot
.17148 .10562 Mdot
.17174 .10575 Mdot
.172 .10593 Mdot
.17226 .10605 Mdot
.17251 .10613 Mdot
.17277 .10619 Mdot
.17303 .10661 Mdot
.17329 .10705 Mdot
.17355 .10712 Mdot
.17381 .10722 Mdot
.17406 .10737 Mdot
.17432 .10747 Mdot
.17458 .10761 Mdot
.17484 .10779 Mdot
.1751 .10787 Mdot
.17535 .10806 Mdot
.17561 .10825 Mdot
.17587 .10832 Mdot
.17613 .10849 Mdot
.17639 .1086 Mdot
.17664 .10869 Mdot
.1769 .10881 Mdot
.17716 .10888 Mdot
.17742 .10926 Mdot
.17768 .10961 Mdot
.17794 .10968 Mdot
.17819 .10978 Mdot
.17845 .10995 Mdot
.17871 .11007 Mdot
.17897 .11026 Mdot
.17923 .11044 Mdot
.17948 .11056 Mdot
.17974 .1107 Mdot
.18 .11079 Mdot
.18026 .11111 Mdot
.18052 .11142 Mdot
.18078 .11149 Mdot
.18103 .11161 Mdot
.18129 .11171 Mdot
.18155 .11185 Mdot
.18181 .11198 Mdot
.18207 .11207 Mdot
.18232 .11217 Mdot
.18258 .11224 Mdot
.18284 .11248 Mdot
.1831 .11277 Mdot
.18336 .11286 Mdot
.18362 .11304 Mdot
.18387 .11321 Mdot
.18413 .1133 Mdot
.18439 .11353 Mdot
.18465 .11376 Mdot
.18491 .11383 Mdot
.18516 .11397 Mdot
.18542 .11411 Mdot
.18568 .11418 Mdot
.18594 .11437 Mdot
.1862 .11461 Mdot
.18646 .11472 Mdot
.18671 .11484 Mdot
.18697 .11505 Mdot
.18723 .11524 Mdot
.18749 .11534 Mdot
.18775 .11543 Mdot
.188 .11561 Mdot
.18826 .11577 Mdot
.18852 .11585 Mdot
.18878 .11593 Mdot
.18904 .11608 Mdot
.1893 .11629 Mdot
.18955 .11648 Mdot
.18981 .11666 Mdot
.19007 .11683 Mdot
.19033 .11699 Mdot
.19059 .11713 Mdot
.19084 .11727 Mdot
.1911 .1174 Mdot
.19136 .11752 Mdot
.19162 .11764 Mdot
.19188 .11775 Mdot
.19213 .11785 Mdot
.19239 .11804 Mdot
.19265 .11821 Mdot
.19291 .11837 Mdot
.19317 .11851 Mdot
.19343 .11864 Mdot
.19368 .11876 Mdot
.19394 .11887 Mdot
.1942 .11897 Mdot
.19446 .11906 Mdot
.19472 .11915 Mdot
.19497 .11923 Mdot
.19523 .11931 Mdot
.19549 .11938 Mdot
.19575 .11945 Mdot
.19601 .11951 Mdot
.19627 .11957 Mdot
.19652 .11963 Mdot
.19678 .11968 Mdot
.19704 .11973 Mdot
.1973 .12158 Mdot
.19756 .12344 Mdot
.19781 .12349 Mdot
.19807 .12355 Mdot
.19833 .1236 Mdot
.19859 .12366 Mdot
.19885 .12373 Mdot
.19911 .1238 Mdot
.19936 .12387 Mdot
.19962 .12395 Mdot
.19988 .12403 Mdot
.20014 .12412 Mdot
.2004 .12422 Mdot
.20065 .12432 Mdot
.20091 .12443 Mdot
.20117 .12456 Mdot
.20143 .12469 Mdot
.20169 .12484 Mdot
.20195 .125 Mdot
.2022 .12517 Mdot
.20246 .12527 Mdot
.20272 .12537 Mdot
.20298 .12547 Mdot
.20324 .12559 Mdot
.20349 .12571 Mdot
.20375 .12583 Mdot
.20401 .12597 Mdot
.20427 .12611 Mdot
.20453 .12626 Mdot
.20479 .12642 Mdot
.20504 .1266 Mdot
.2053 .12679 Mdot
.20556 .12699 Mdot
.20582 .12721 Mdot
.20608 .12736 Mdot
.20633 .12744 Mdot
.20659 .12753 Mdot
.20685 .1277 Mdot
.20711 .12789 Mdot
.20737 .12798 Mdot
.20762 .12808 Mdot
.20788 .12829 Mdot
.20814 .12852 Mdot
.2084 .12863 Mdot
.20866 .12876 Mdot
.20892 .12901 Mdot
.20917 .12922 Mdot
.20943 .12929 Mdot
.20969 .12943 Mdot
.20995 .12958 Mdot
.21021 .12966 Mdot
.21046 .1299 Mdot
.21072 .13016 Mdot
.21098 .13025 Mdot
.21124 .13043 Mdot
.2115 .13062 Mdot
.21176 .13072 Mdot
.21201 .13078 Mdot
.21227 .13103 Mdot
.21253 .1313 Mdot
.21279 .13136 Mdot
.21305 .13148 Mdot
.2133 .13157 Mdot
.21356 .13172 Mdot
.21382 .13187 Mdot
.21408 .13197 Mdot
.21434 .1321 Mdot
.2146 .13218 Mdot
.21485 .13252 Mdot
.21511 .13281 Mdot
.21537 .13287 Mdot
.21563 .13296 Mdot
.21589 .13312 Mdot
.21614 .13325 Mdot
.2164 .13345 Mdot
.21666 .13366 Mdot
.21692 .1338 Mdot
.21718 .13398 Mdot
.21744 .13409 Mdot
.21769 .13417 Mdot
.21795 .13456 Mdot
.21821 .13491 Mdot
.21847 .13497 Mdot
.21873 .13506 Mdot
.21898 .13519 Mdot
.21924 .13529 Mdot
.2195 .13542 Mdot
.21976 .13561 Mdot
.22002 .13569 Mdot
.22028 .13589 Mdot
.22053 .13611 Mdot
.22079 .13619 Mdot
.22105 .1364 Mdot
.22131 .13655 Mdot
.22157 .13667 Mdot
.22182 .13684 Mdot
.22208 .13695 Mdot
.22234 .13703 Mdot
.2226 .13709 Mdot
.22286 .13754 Mdot
.22312 .13801 Mdot
.22337 .13808 Mdot
.22363 .13817 Mdot
.22389 .1383 Mdot
.22415 .13839 Mdot
.22441 .13851 Mdot
.22466 .13865 Mdot
.22492 .13886 Mdot
.22518 .13904 Mdot
.22544 .13915 Mdot
.2257 .13927 Mdot
.22595 .13935 Mdot
.22621 .1396 Mdot
.22647 .13987 Mdot
.22673 .13994 Mdot
.22699 .14008 Mdot
.22725 .14021 Mdot
.2275 .14042 Mdot
.22776 .14059 Mdot
.22802 .14066 Mdot
.22828 .14085 Mdot
.22854 .141 Mdot
.22879 .14112 Mdot
.22905 .14131 Mdot
.22931 .14144 Mdot
.22957 .14154 Mdot
.22983 .14162 Mdot
.23009 .14169 Mdot
.23034 .14228 Mdot
.2306 .14286 Mdot
.23086 .14292 Mdot
.23112 .14299 Mdot
.23138 .14308 Mdot
.23163 .1432 Mdot
.23189 .14336 Mdot
.23215 .14347 Mdot
.23241 .14359 Mdot
.23267 .14375 Mdot
.23293 .14395 Mdot
.23318 .14411 Mdot
.23344 .14421 Mdot
.2337 .14431 Mdot
.23396 .14456 Mdot
.23422 .1448 Mdot
.23447 .14489 Mdot
.23473 .14508 Mdot
.23499 .14521 Mdot
.23525 .14531 Mdot
.23551 .14544 Mdot
.23577 .14552 Mdot
.23602 .14558 Mdot
.23628 .1459 Mdot
.23654 .14623 Mdot
.2368 .1463 Mdot
.23706 .14639 Mdot
.23731 .14653 Mdot
.23757 .14663 Mdot
.23783 .14678 Mdot
.23809 .14693 Mdot
.23835 .14701 Mdot
.23861 .14712 Mdot
.23886 .14741 Mdot
.23912 .14766 Mdot
.23938 .14773 Mdot
.23964 .14787 Mdot
.2399 .148 Mdot
.24015 .14823 Mdot
.24041 .14843 Mdot
.24067 .14852 Mdot
.24093 .1486 Mdot
.24119 .14875 Mdot
.24144 .14895 Mdot
.2417 .14911 Mdot
.24196 .14925 Mdot
.24222 .14937 Mdot
.24248 .14948 Mdot
.24274 .14957 Mdot
.24299 .14973 Mdot
.24325 .14986 Mdot
.24351 .14996 Mdot
.24377 .15005 Mdot
.24403 .15012 Mdot
.24428 .15019 Mdot
.24454 .15024 Mdot
.2448 .15105 Mdot
.24506 .15186 Mdot
.24532 .15191 Mdot
.24558 .15197 Mdot
.24583 .15205 Mdot
.24609 .15213 Mdot
.24635 .15223 Mdot
.24661 .15236 Mdot
.24687 .15251 Mdot
.24712 .1527 Mdot
.24738 .15282 Mdot
.24764 .15295 Mdot
.2479 .15311 Mdot
.24816 .15329 Mdot
.24842 .15343 Mdot
.24867 .15351 Mdot
.24893 .1536 Mdot
.24919 .15378 Mdot
.24945 .154 Mdot
.24971 .15412 Mdot
.24996 .15425 Mdot
.25022 .15432 Mdot
.25048 .15454 Mdot
.25074 .15481 Mdot
.251 .1549 Mdot
.25126 .15498 Mdot
.25151 .15511 Mdot
.25177 .15525 Mdot
.25203 .15534 Mdot
.25229 .15547 Mdot
.25255 .15555 Mdot
.2528 .1559 Mdot
.25306 .15623 Mdot
.25332 .15631 Mdot
.25358 .15642 Mdot
.25384 .15662 Mdot
.2541 .15679 Mdot
.25435 .15686 Mdot
.25461 .15706 Mdot
.25487 .15728 Mdot
.25513 .15737 Mdot
.25539 .15745 Mdot
.25564 .15759 Mdot
.2559 .15775 Mdot
.25616 .15789 Mdot
.25642 .15808 Mdot
.25668 .15821 Mdot
.25694 .15831 Mdot
.25719 .15839 Mdot
.25745 .15845 Mdot
.25771 .15897 Mdot
.25797 .1595 Mdot
.25823 .15957 Mdot
.25848 .15965 Mdot
.25874 .15976 Mdot
.259 .1599 Mdot
.25926 .16 Mdot
.25952 .16012 Mdot
.25977 .16027 Mdot
.26003 .16046 Mdot
.26029 .16062 Mdot
.26055 .16072 Mdot
.26081 .16083 Mdot
.26107 .16109 Mdot
.26132 .16134 Mdot
.26158 .16144 Mdot
.26184 .16152 Mdot
.2621 .16166 Mdot
.26236 .16181 Mdot
.26261 .16192 Mdot
.26287 .16207 Mdot
.26313 .16217 Mdot
.26339 .16224 Mdot
.26365 .16264 Mdot
.26391 .16301 Mdot
.26416 .16307 Mdot
.26442 .16315 Mdot
.26468 .16327 Mdot
.26494 .16347 Mdot
.2652 .16361 Mdot
.26545 .16383 Mdot
.26571 .16403 Mdot
.26597 .16416 Mdot
.26623 .16432 Mdot
.26649 .16441 Mdot
.26675 .16447 Mdot
.267 .16476 Mdot
.26726 .16508 Mdot
.26752 .16516 Mdot
.26778 .16528 Mdot
.26804 .16538 Mdot
.26829 .16552 Mdot
.26855 .16565 Mdot
.26881 .16574 Mdot
.26907 .16585 Mdot
.26933 .16591 Mdot
.26959 .16614 Mdot
.26984 .16642 Mdot
.2701 .1665 Mdot
.27036 .16667 Mdot
.27062 .16683 Mdot
.27088 .1669 Mdot
.27113 .16711 Mdot
.27139 .16738 Mdot
.27165 .1675 Mdot
.27191 .16762 Mdot
.27217 .16783 Mdot
.27243 .16803 Mdot
.27268 .16813 Mdot
.27294 .16822 Mdot
.2732 .16839 Mdot
.27346 .16855 Mdot
.27372 .16862 Mdot
.27397 .16883 Mdot
.27423 .16902 Mdot
.27449 .16919 Mdot
.27475 .16934 Mdot
.27501 .16949 Mdot
.27526 .16962 Mdot
.27552 .16974 Mdot
.27578 .16985 Mdot
.27604 .16995 Mdot
.2763 .17014 Mdot
.27656 .1703 Mdot
.27681 .17044 Mdot
.27707 .17057 Mdot
.27733 .17068 Mdot
.27759 .17078 Mdot
.27785 .17087 Mdot
.2781 .17095 Mdot
.27836 .17103 Mdot
.27862 .1711 Mdot
.27888 .17116 Mdot
.27914 .17122 Mdot
.2794 .17128 Mdot
.27965 .17133 Mdot
.27991 .17268 Mdot
.28017 .17405 Mdot
.28043 .1741 Mdot
.28069 .17416 Mdot
.28094 .17422 Mdot
.2812 .17429 Mdot
.28146 .17436 Mdot
.28172 .17444 Mdot
.28198 .17453 Mdot
.28224 .17462 Mdot
.28249 .17473 Mdot
.28275 .17485 Mdot
.28301 .17498 Mdot
.28327 .17513 Mdot
.28353 .1753 Mdot
.28378 .1754 Mdot
.28404 .1755 Mdot
.2843 .17561 Mdot
.28456 .17573 Mdot
.28482 .17586 Mdot
.28508 .176 Mdot
.28533 .17615 Mdot
.28559 .17632 Mdot
.28585 .17651 Mdot
.28611 .17672 Mdot
.28637 .17687 Mdot
.28662 .17695 Mdot
.28688 .17703 Mdot
.28714 .17721 Mdot
.2874 .1774 Mdot
.28766 .1775 Mdot
.28792 .1776 Mdot
.28817 .17783 Mdot
.28843 .17808 Mdot
.28869 .17821 Mdot
.28895 .17835 Mdot
.28921 .17842 Mdot
.28946 .17865 Mdot
.28972 .1789 Mdot
.28998 .17899 Mdot
.29024 .17917 Mdot
.2905 .17936 Mdot
.29076 .17946 Mdot
.29101 .17952 Mdot
.29127 .17978 Mdot
.29153 .18006 Mdot
.29179 .18014 Mdot
.29205 .18026 Mdot
.2923 .18037 Mdot
.29256 .18053 Mdot
.29282 .1807 Mdot
.29308 .18081 Mdot
.29334 .18096 Mdot
.29359 .18106 Mdot
.29385 .18112 Mdot
.29411 .18145 Mdot
.29437 .1818 Mdot
.29463 .18188 Mdot
.29489 .18199 Mdot
.29514 .18219 Mdot
.2954 .18235 Mdot
.29566 .18243 Mdot
.29592 .18261 Mdot
.29618 .1828 Mdot
.29643 .18288 Mdot
.29669 .18307 Mdot
.29695 .18321 Mdot
.29721 .18332 Mdot
.29747 .18347 Mdot
.29773 .18357 Mdot
.29798 .18365 Mdot
.29824 .18371 Mdot
.2985 .18413 Mdot
.29876 .18458 Mdot
.29902 .18465 Mdot
.29927 .18474 Mdot
.29953 .18488 Mdot
.29979 .18496 Mdot
.30005 .18508 Mdot
.30031 .18523 Mdot
.30057 .18543 Mdot
.30082 .18561 Mdot
.30108 .18573 Mdot
.30134 .18586 Mdot
.3016 .18593 Mdot
.30186 .1862 Mdot
.30211 .18647 Mdot
.30237 .18655 Mdot
.30263 .1867 Mdot
.30289 .18684 Mdot
.30315 .18706 Mdot
.30341 .18725 Mdot
.30366 .18733 Mdot
.30392 .18754 Mdot
.30418 .1877 Mdot
.30444 .18784 Mdot
.3047 .18795 Mdot
.30495 .18805 Mdot
.30521 .1882 Mdot
.30547 .18831 Mdot
.30573 .1884 Mdot
.30599 .18848 Mdot
.30625 .18854 Mdot
.3065 .18917 Mdot
.30676 .1898 Mdot
.30702 .18986 Mdot
.30728 .18993 Mdot
.30754 .19002 Mdot
.30779 .19013 Mdot
.30805 .19028 Mdot
.30831 .19037 Mdot
.30857 .19048 Mdot
.30883 .19061 Mdot
.30908 .19077 Mdot
.30934 .19096 Mdot
.3096 .19104 Mdot
.30986 .19121 Mdot
.31012 .19143 Mdot
.31038 .19155 Mdot
.31063 .19169 Mdot
.31089 .19176 Mdot
.31115 .19201 Mdot
.31141 .19225 Mdot
.31167 .19232 Mdot
.31192 .19244 Mdot
.31218 .19254 Mdot
.31244 .1927 Mdot
.3127 .19288 Mdot
.31296 .19301 Mdot
.31322 .19319 Mdot
.31347 .1933 Mdot
.31373 .19338 Mdot
.31399 .19344 Mdot
.31425 .19383 Mdot
.31451 .19424 Mdot
.31476 .1943 Mdot
.31502 .19439 Mdot
.31528 .19452 Mdot
.31554 .19461 Mdot
.3158 .19472 Mdot
.31606 .19488 Mdot
.31631 .1951 Mdot
.31657 .19531 Mdot
.31683 .19544 Mdot
.31709 .1956 Mdot
.31735 .19569 Mdot
.3176 .19575 Mdot
.31786 .19603 Mdot
.31812 .19634 Mdot
.31838 .19641 Mdot
.31864 .19652 Mdot
.3189 .19661 Mdot
.31915 .19674 Mdot
.31941 .19695 Mdot
.31967 .19704 Mdot
.31993 .19731 Mdot
.32019 .19755 Mdot
.32044 .19762 Mdot
.3207 .19777 Mdot
.32096 .1979 Mdot
.32122 .19815 Mdot
.32148 .19836 Mdot
.32174 .19846 Mdot
.32199 .19856 Mdot
.32225 .19874 Mdot
.32251 .1989 Mdot
.32277 .19897 Mdot
.32303 .19917 Mdot
.32328 .19935 Mdot
.32354 .19951 Mdot
.3238 .19966 Mdot
.32406 .19978 Mdot
.32432 .1999 Mdot
.32457 .2 Mdot
.32483 .20019 Mdot
.32509 .20034 Mdot
.32535 .20047 Mdot
.32561 .20058 Mdot
.32587 .20068 Mdot
.32612 .20077 Mdot
.32638 .20085 Mdot
.32664 .20092 Mdot
.3269 .20098 Mdot
.32716 .20104 Mdot
.32741 .20109 Mdot
.32767 .20215 Mdot
.32793 .20324 Mdot
.32819 .2033 Mdot
.32845 .20336 Mdot
.32871 .20342 Mdot
.32896 .2035 Mdot
.32922 .20358 Mdot
.32948 .20368 Mdot
.32974 .20378 Mdot
.33 .20391 Mdot
.33025 .20405 Mdot
.33051 .20422 Mdot
.33077 .20432 Mdot
.33103 .20442 Mdot
.33129 .20454 Mdot
.33155 .20467 Mdot
.3318 .20481 Mdot
.33206 .20497 Mdot
.33232 .20516 Mdot
.33258 .20537 Mdot
.33284 .20552 Mdot
.33309 .2056 Mdot
.33335 .20569 Mdot
.33361 .20588 Mdot
.33387 .20609 Mdot
.33413 .20621 Mdot
.33439 .20633 Mdot
.33464 .2066 Mdot
.3349 .20682 Mdot
.33516 .2069 Mdot
.33542 .20707 Mdot
.33568 .20725 Mdot
.33593 .20734 Mdot
.33619 .20765 Mdot
.33645 .20793 Mdot
.33671 .208 Mdot
.33697 .20812 Mdot
.33723 .20823 Mdot
.33748 .20839 Mdot
.33774 .20857 Mdot
.338 .20869 Mdot
.33826 .20884 Mdot
.33852 .20894 Mdot
.33877 .20901 Mdot
.33903 .20936 Mdot
.33929 .2097 Mdot
.33955 .20976 Mdot
.33981 .20984 Mdot
.34007 .20997 Mdot
.34032 .21007 Mdot
.34058 .2102 Mdot
.34084 .21039 Mdot
.3411 .21047 Mdot
.34136 .21069 Mdot
.34161 .21092 Mdot
.34187 .21102 Mdot
.34213 .2111 Mdot
.34239 .21125 Mdot
.34265 .21142 Mdot
.3429 .21156 Mdot
.34316 .21175 Mdot
.34342 .21189 Mdot
.34368 .21198 Mdot
.34394 .21206 Mdot
.3442 .21212 Mdot
.34445 .21262 Mdot
.34471 .21311 Mdot
.34497 .21317 Mdot
.34523 .21325 Mdot
.34549 .21334 Mdot
.34574 .21347 Mdot
.346 .21366 Mdot
.34626 .21379 Mdot
.34652 .21395 Mdot
.34678 .21409 Mdot
.34704 .21417 Mdot
.34729 .21426 Mdot
.34755 .21447 Mdot
.34781 .21468 Mdot
.34807 .21475 Mdot
.34833 .21493 Mdot
.34858 .21504 Mdot
.34884 .21513 Mdot
.3491 .21525 Mdot
.34936 .21532 Mdot
.34962 .21567 Mdot
.34988 .21598 Mdot
.35013 .21604 Mdot
.35039 .21612 Mdot
.35065 .21626 Mdot
.35091 .21636 Mdot
.35117 .21651 Mdot
.35142 .21665 Mdot
.35168 .21673 Mdot
.35194 .21684 Mdot
.3522 .21713 Mdot
.35246 .21738 Mdot
.35272 .21746 Mdot
.35297 .2176 Mdot
.35323 .21774 Mdot
.35349 .21798 Mdot
.35375 .21819 Mdot
.35401 .21828 Mdot
.35426 .21837 Mdot
.35452 .21854 Mdot
.35478 .21875 Mdot
.35504 .21893 Mdot
.3553 .21909 Mdot
.35556 .21923 Mdot
.35581 .21935 Mdot
.35607 .21945 Mdot
.35633 .21964 Mdot
.35659 .21978 Mdot
.35685 .21991 Mdot
.3571 .22001 Mdot
.35736 .2201 Mdot
.35762 .22017 Mdot
.35788 .22024 Mdot
.35814 .2203 Mdot
.35839 .22035 Mdot
.35865 .22123 Mdot
.35891 .22215 Mdot
.35917 .2222 Mdot
.35943 .22227 Mdot
.35969 .22234 Mdot
.35994 .22243 Mdot
.3602 .22252 Mdot
.36046 .22264 Mdot
.36072 .22278 Mdot
.36098 .22294 Mdot
.36123 .22304 Mdot
.36149 .22315 Mdot
.36175 .22328 Mdot
.36201 .22342 Mdot
.36227 .22359 Mdot
.36253 .22378 Mdot
.36278 .22392 Mdot
.36304 .224 Mdot
.3633 .22408 Mdot
.36356 .22427 Mdot
.36382 .22448 Mdot
.36407 .22459 Mdot
.36433 .22471 Mdot
.36459 .22478 Mdot
.36485 .22499 Mdot
.36511 .22523 Mdot
.36537 .22532 Mdot
.36562 .2255 Mdot
.36588 .22562 Mdot
.36614 .22571 Mdot
.3664 .22582 Mdot
.36666 .22588 Mdot
.36691 .22618 Mdot
.36717 .22651 Mdot
.36743 .2266 Mdot
.36769 .22675 Mdot
.36795 .22689 Mdot
.36821 .2271 Mdot
.36846 .22733 Mdot
.36872 .22749 Mdot
.36898 .22761 Mdot
.36924 .2277 Mdot
.3695 .22784 Mdot
.36975 .22793 Mdot
.37001 .228 Mdot
.37027 .22844 Mdot
.37053 .22888 Mdot
.37079 .22895 Mdot
.37105 .22904 Mdot
.3713 .22917 Mdot
.37156 .22926 Mdot
.37182 .22938 Mdot
.37208 .22953 Mdot
.37234 .22975 Mdot
.37259 .22994 Mdot
.37285 .23007 Mdot
.37311 .23021 Mdot
.37337 .23029 Mdot
.37363 .23059 Mdot
.37389 .23085 Mdot
.37414 .23091 Mdot
.3744 .23101 Mdot
.37466 .23119 Mdot
.37492 .23135 Mdot
.37518 .23142 Mdot
.37543 .23163 Mdot
.37569 .23186 Mdot
.37595 .23196 Mdot
.37621 .23206 Mdot
.37647 .23223 Mdot
.37672 .23245 Mdot
.37698 .23263 Mdot
.37724 .23278 Mdot
.3775 .23291 Mdot
.37776 .23303 Mdot
.37802 .23313 Mdot
.37827 .23329 Mdot
.37853 .23342 Mdot
.37879 .23352 Mdot
.37905 .23361 Mdot
.37931 .23369 Mdot
.37956 .23375 Mdot
.37982 .23381 Mdot
.38008 .23457 Mdot
.38034 .23531 Mdot
.3806 .23537 Mdot
.38086 .23543 Mdot
.38111 .2355 Mdot
.38137 .23558 Mdot
.38163 .23568 Mdot
.38189 .2358 Mdot
.38215 .23596 Mdot
.3824 .23615 Mdot
.38266 .23627 Mdot
.38292 .23641 Mdot
.38318 .23657 Mdot
.38344 .23677 Mdot
.3837 .23692 Mdot
.38395 .237 Mdot
.38421 .23709 Mdot
.38447 .23729 Mdot
.38473 .23753 Mdot
.38499 .23767 Mdot
.38524 .23782 Mdot
.3855 .2379 Mdot
.38576 .23816 Mdot
.38602 .2384 Mdot
.38628 .23847 Mdot
.38654 .23858 Mdot
.38679 .23868 Mdot
.38705 .23884 Mdot
.38731 .239 Mdot
.38757 .23912 Mdot
.38783 .23928 Mdot
.38808 .23938 Mdot
.38834 .23945 Mdot
.3886 .23983 Mdot
.38886 .2402 Mdot
.38912 .24027 Mdot
.38938 .24037 Mdot
.38963 .24052 Mdot
.38989 .24064 Mdot
.39015 .24079 Mdot
.39041 .24094 Mdot
.39067 .24103 Mdot
.39092 .24114 Mdot
.39118 .24142 Mdot
.39144 .24166 Mdot
.3917 .24173 Mdot
.39196 .24187 Mdot
.39221 .24198 Mdot
.39247 .24219 Mdot
.39273 .24237 Mdot
.39299 .24245 Mdot
.39325 .24265 Mdot
.39351 .24281 Mdot
.39376 .24295 Mdot
.39402 .24306 Mdot
.39428 .24316 Mdot
.39454 .24331 Mdot
.3948 .24344 Mdot
.39505 .24353 Mdot
.39531 .24361 Mdot
.39557 .24368 Mdot
.39583 .24374 Mdot
.39609 .2444 Mdot
.39635 .24505 Mdot
.3966 .2451 Mdot
.39686 .24517 Mdot
.39712 .24524 Mdot
.39738 .24534 Mdot
.39764 .24545 Mdot
.39789 .2456 Mdot
.39815 .24579 Mdot
.39841 .24592 Mdot
.39867 .24606 Mdot
.39893 .24624 Mdot
.39919 .24647 Mdot
.39944 .24665 Mdot
.3997 .24675 Mdot
.39996 .24686 Mdot
.40022 .24713 Mdot
.40048 .24736 Mdot
.40073 .24745 Mdot
.40099 .24765 Mdot
.40125 .24777 Mdot
.40151 .24786 Mdot
.40177 .24799 Mdot
.40203 .24806 Mdot
.40228 .24841 Mdot
.40254 .24875 Mdot
.4028 .24882 Mdot
.40306 .24894 Mdot
.40332 .24903 Mdot
.40357 .24915 Mdot
.40383 .24933 Mdot
.40409 .24941 Mdot
.40435 .24963 Mdot
.40461 .24987 Mdot
.40487 .24998 Mdot
.40512 .25007 Mdot
.40538 .25023 Mdot
.40564 .25043 Mdot
.4059 .25058 Mdot
.40616 .25071 Mdot
.40641 .25081 Mdot
.40667 .25098 Mdot
.40693 .2511 Mdot
.40719 .2512 Mdot
.40745 .25127 Mdot
.40771 .25133 Mdot
.40796 .25195 Mdot
.40822 .25255 Mdot
.40848 .25261 Mdot
.40874 .25268 Mdot
.409 .25277 Mdot
.40925 .25288 Mdot
.40951 .25304 Mdot
.40977 .25313 Mdot
.41003 .25325 Mdot
.41029 .25339 Mdot
.41054 .25357 Mdot
.4108 .25371 Mdot
.41106 .25379 Mdot
.41132 .25388 Mdot
.41158 .25409 Mdot
.41184 .25427 Mdot
.41209 .25434 Mdot
.41235 .2545 Mdot
.41261 .25467 Mdot
.41287 .25476 Mdot
.41313 .25483 Mdot
.41338 .2551 Mdot
.41364 .25537 Mdot
.4139 .25543 Mdot
.41416 .25552 Mdot
.41442 .25569 Mdot
.41468 .25583 Mdot
.41493 .25607 Mdot
.41519 .25626 Mdot
.41545 .25634 Mdot
.41571 .25655 Mdot
.41597 .2567 Mdot
.41622 .25683 Mdot
.41648 .25701 Mdot
.41674 .25714 Mdot
.417 .25724 Mdot
.41726 .25732 Mdot
.41752 .25738 Mdot
.41777 .25792 Mdot
.41803 .25845 Mdot
.41829 .25851 Mdot
.41855 .25858 Mdot
.41881 .25867 Mdot
.41906 .2588 Mdot
.41932 .25897 Mdot
.41958 .25908 Mdot
.41984 .25922 Mdot
.4201 .2594 Mdot
.42036 .25948 Mdot
.42061 .25964 Mdot
.42087 .25985 Mdot
.42113 .25997 Mdot
.42139 .26011 Mdot
.42165 .26019 Mdot
.4219 .26046 Mdot
.42216 .26073 Mdot
.42242 .2608 Mdot
.42268 .26094 Mdot
.42294 .26106 Mdot
.4232 .26125 Mdot
.42345 .26147 Mdot
.42371 .26164 Mdot
.42397 .26177 Mdot
.42423 .26187 Mdot
.42449 .26202 Mdot
.42474 .26213 Mdot
.425 .26221 Mdot
.42526 .26227 Mdot
.42552 .26276 Mdot
.42578 .26324 Mdot
.42603 .2633 Mdot
.42629 .26338 Mdot
.42655 .26348 Mdot
.42681 .26362 Mdot
.42707 .26372 Mdot
.42733 .26383 Mdot
.42758 .26398 Mdot
.42784 .26418 Mdot
.4281 .26435 Mdot
.42836 .26445 Mdot
.42862 .26457 Mdot
.42887 .26463 Mdot
.42913 .26486 Mdot
.42939 .26508 Mdot
.42965 .26514 Mdot
.42991 .26525 Mdot
.43017 .26535 Mdot
.43042 .26551 Mdot
.43068 .2657 Mdot
.43094 .26583 Mdot
.4312 .26602 Mdot
.43146 .26615 Mdot
.43171 .26624 Mdot
.43197 .26631 Mdot
.43223 .26677 Mdot
.43249 .26722 Mdot
.43275 .26729 Mdot
.43301 .26737 Mdot
.43326 .26749 Mdot
.43352 .26766 Mdot
.43378 .26779 Mdot
.43404 .26795 Mdot
.4343 .2681 Mdot
.43455 .26818 Mdot
.43481 .26828 Mdot
.43507 .26852 Mdot
.43533 .26876 Mdot
.43559 .26886 Mdot
.43585 .26894 Mdot
.4361 .26908 Mdot
.43636 .26923 Mdot
.43662 .26935 Mdot
.43688 .26951 Mdot
.43714 .26961 Mdot
.43739 .26969 Mdot
.43765 .26975 Mdot
.43791 .27014 Mdot
.43817 .27058 Mdot
.43843 .27065 Mdot
.43869 .27075 Mdot
.43894 .2709 Mdot
.4392 .271 Mdot
.43946 .27114 Mdot
.43972 .27134 Mdot
.43998 .27142 Mdot
.44023 .27163 Mdot
.44049 .27183 Mdot
.44075 .27191 Mdot
.44101 .2721 Mdot
.44127 .27223 Mdot
.44152 .27233 Mdot
.44178 .27247 Mdot
.44204 .27256 Mdot
.4423 .27263 Mdot
.44256 .27301 Mdot
.44282 .27339 Mdot
.44307 .27345 Mdot
.44333 .27353 Mdot
.44359 .27366 Mdot
.44385 .27375 Mdot
.44411 .27387 Mdot
.44436 .27404 Mdot
.44462 .27411 Mdot
.44488 .27429 Mdot
.44514 .27447 Mdot
.4454 .27454 Mdot
.44566 .2747 Mdot
.44591 .27481 Mdot
.44617 .2749 Mdot
.44643 .27502 Mdot
.44669 .2751 Mdot
.44695 .27549 Mdot
.4472 .27587 Mdot
.44746 .27594 Mdot
.44772 .27605 Mdot
.44798 .27624 Mdot
.44824 .27638 Mdot
.4485 .2766 Mdot
.44875 .27682 Mdot
.44901 .27696 Mdot
.44927 .27713 Mdot
.44953 .27724 Mdot
.44979 .2773 Mdot
.45004 .27765 Mdot
.4503 .27798 Mdot
.45056 .27804 Mdot
.45082 .27814 Mdot
.45108 .2783 Mdot
.45134 .27843 Mdot
.45159 .27862 Mdot
.45185 .27881 Mdot
.45211 .27894 Mdot
.45237 .27909 Mdot
.45263 .27918 Mdot
.45288 .27924 Mdot
.45314 .27954 Mdot
.4534 .2799 Mdot
.45366 .27998 Mdot
.45392 .28013 Mdot
.45418 .28024 Mdot
.45443 .28041 Mdot
.45469 .28058 Mdot
.45495 .28069 Mdot
.45521 .28082 Mdot
.45547 .2809 Mdot
.45572 .28122 Mdot
.45598 .28154 Mdot
.45624 .28161 Mdot
.4565 .28174 Mdot
.45676 .28184 Mdot
.45702 .28199 Mdot
.45727 .28214 Mdot
.45753 .28224 Mdot
.45779 .28236 Mdot
.45805 .28244 Mdot
.45831 .28272 Mdot
.45856 .28301 Mdot
.45882 .28307 Mdot
.45908 .28319 Mdot
.45934 .28328 Mdot
.4596 .28342 Mdot
.45985 .28355 Mdot
.46011 .28364 Mdot
.46037 .28375 Mdot
.46063 .28381 Mdot
.46089 .28407 Mdot
.46115 .28432 Mdot
.4614 .28439 Mdot
.46166 .28449 Mdot
.46192 .28457 Mdot
.46218 .28469 Mdot
.46244 .28489 Mdot
.46269 .28499 Mdot
.46295 .28529 Mdot
.46321 .28557 Mdot
.46347 .28567 Mdot
.46373 .28585 Mdot
.46399 .28603 Mdot
.46424 .28612 Mdot
.4645 .28639 Mdot
.46476 .28665 Mdot
.46502 .28673 Mdot
.46528 .2869 Mdot
.46553 .28707 Mdot
.46579 .28715 Mdot
.46605 .28739 Mdot
.46631 .28763 Mdot
.46657 .28771 Mdot
.46683 .28786 Mdot
.46708 .28802 Mdot
.46734 .28809 Mdot
.4676 .28831 Mdot
.46786 .28853 Mdot
.46812 .2886 Mdot
.46837 .28874 Mdot
.46863 .28888 Mdot
.46889 .28895 Mdot
.46915 .28916 Mdot
.46941 .28936 Mdot
.46967 .28942 Mdot
.46992 .28955 Mdot
.47018 .28968 Mdot
.47044 .28994 Mdot
.4707 .29018 Mdot
.47096 .2903 Mdot
.47121 .29042 Mdot
.47147 .29065 Mdot
.47173 .29088 Mdot
.47199 .29099 Mdot
.47225 .29111 Mdot
.47251 .29132 Mdot
.47276 .29153 Mdot
.47302 .29164 Mdot
.47328 .29174 Mdot
.47354 .29194 Mdot
.4738 .29214 Mdot
.47405 .29224 Mdot
.47431 .29233 Mdot
.47457 .29252 Mdot
.47483 .29271 Mdot
.47509 .2928 Mdot
.47534 .29289 Mdot
.4756 .29306 Mdot
.47586 .29323 Mdot
.47612 .29332 Mdot
.47638 .2934 Mdot
.47664 .29357 Mdot
.47689 .29373 Mdot
.47715 .29381 Mdot
.47741 .29389 Mdot
.47767 .29404 Mdot
.47793 .29419 Mdot
.47818 .29427 Mdot
.47844 .29434 Mdot
.4787 .29449 Mdot
.47896 .2947 Mdot
.47922 .29491 Mdot
.47948 .29511 Mdot
.47973 .29531 Mdot
.47999 .2955 Mdot
.48025 .29568 Mdot
.48051 .29586 Mdot
.48077 .29604 Mdot
.48102 .29621 Mdot
.48128 .29638 Mdot
.48154 .29654 Mdot
.4818 .2967 Mdot
.48206 .29685 Mdot
.48232 .297 Mdot
.48257 .29715 Mdot
.48283 .29729 Mdot
.48309 .29743 Mdot
.48335 .29757 Mdot
.48361 .2977 Mdot
.48386 .29783 Mdot
.48412 .29796 Mdot
.48438 .29808 Mdot
.48464 .2982 Mdot
.4849 .29832 Mdot
.48516 .29844 Mdot
.48541 .29855 Mdot
.48567 .29866 Mdot
.48593 .29877 Mdot
.48619 .29888 Mdot
.48645 .29898 Mdot
.4867 .29908 Mdot
.48696 .29918 Mdot
.48722 .29938 Mdot
.48748 .29957 Mdot
.48774 .29975 Mdot
.488 .29992 Mdot
.48825 .30009 Mdot
.48851 .30025 Mdot
.48877 .3004 Mdot
.48903 .30055 Mdot
.48929 .3007 Mdot
.48954 .30084 Mdot
.4898 .30097 Mdot
.49006 .3011 Mdot
.49032 .30123 Mdot
.49058 .30135 Mdot
.49084 .30147 Mdot
.49109 .30159 Mdot
.49135 .3017 Mdot
.49161 .30181 Mdot
.49187 .30191 Mdot
.49213 .30201 Mdot
.49238 .30211 Mdot
.49264 .30221 Mdot
.4929 .3023 Mdot
.49316 .30239 Mdot
.49342 .30248 Mdot
.49367 .30257 Mdot
.49393 .30265 Mdot
.49419 .30274 Mdot
.49445 .30282 Mdot
.49471 .30289 Mdot
.49497 .30297 Mdot
.49522 .30304 Mdot
.49548 .30312 Mdot
.49574 .30319 Mdot
.496 .30326 Mdot
.49626 .30332 Mdot
.49651 .30339 Mdot
.49677 .30345 Mdot
.49703 .30352 Mdot
.49729 .30358 Mdot
.49755 .30364 Mdot
.49781 .3037 Mdot
.49806 .30375 Mdot
.49832 .30381 Mdot
.49858 .30387 Mdot
.49884 .30392 Mdot
.4991 .30397 Mdot
.49935 .30402 Mdot
.49961 .30408 Mdot
.49987 .30412 Mdot
.50013 .30902 Mdot
.50039 .31391 Mdot
.50065 .31396 Mdot
.5009 .31401 Mdot
.50116 .31406 Mdot
.50142 .31411 Mdot
.50168 .31417 Mdot
.50194 .31422 Mdot
.50219 .31428 Mdot
.50245 .31434 Mdot
.50271 .3144 Mdot
.50297 .31446 Mdot
.50323 .31452 Mdot
.50349 .31458 Mdot
.50374 .31464 Mdot
.504 .31471 Mdot
.50426 .31478 Mdot
.50452 .31485 Mdot
.50478 .31492 Mdot
.50503 .31499 Mdot
.50529 .31506 Mdot
.50555 .31514 Mdot
.50581 .31522 Mdot
.50607 .3153 Mdot
.50633 .31538 Mdot
.50658 .31546 Mdot
.50684 .31555 Mdot
.5071 .31564 Mdot
.50736 .31573 Mdot
.50762 .31582 Mdot
.50787 .31592 Mdot
.50813 .31602 Mdot
.50839 .31612 Mdot
.50865 .31623 Mdot
.50891 .31634 Mdot
.50916 .31645 Mdot
.50942 .31656 Mdot
.50968 .31668 Mdot
.50994 .3168 Mdot
.5102 .31693 Mdot
.51046 .31706 Mdot
.51071 .3172 Mdot
.51097 .31734 Mdot
.51123 .31748 Mdot
.51149 .31763 Mdot
.51175 .31779 Mdot
.512 .31795 Mdot
.51226 .31811 Mdot
.51252 .31829 Mdot
.51278 .31847 Mdot
.51304 .31866 Mdot
.5133 .31885 Mdot
.51355 .31895 Mdot
.51381 .31905 Mdot
.51407 .31916 Mdot
.51433 .31926 Mdot
.51459 .31937 Mdot
.51484 .31948 Mdot
.5151 .3196 Mdot
.51536 .31971 Mdot
.51562 .31983 Mdot
.51588 .31995 Mdot
.51614 .32008 Mdot
.51639 .32021 Mdot
.51665 .32034 Mdot
.51691 .32047 Mdot
.51717 .32061 Mdot
.51743 .32074 Mdot
.51768 .32089 Mdot
.51794 .32103 Mdot
.5182 .32118 Mdot
.51846 .32134 Mdot
.51872 .3215 Mdot
.51898 .32166 Mdot
.51923 .32182 Mdot
.51949 .322 Mdot
.51975 .32217 Mdot
.52001 .32235 Mdot
.52027 .32254 Mdot
.52052 .32273 Mdot
.52078 .32292 Mdot
.52104 .32312 Mdot
.5213 .32333 Mdot
.52156 .32354 Mdot
.52182 .32369 Mdot
.52207 .32376 Mdot
.52233 .32384 Mdot
.52259 .32399 Mdot
.52285 .32415 Mdot
.52311 .32422 Mdot
.52336 .3243 Mdot
.52362 .32447 Mdot
.52388 .32463 Mdot
.52414 .32472 Mdot
.5244 .3248 Mdot
.52466 .32497 Mdot
.52491 .32515 Mdot
.52517 .32524 Mdot
.52543 .32533 Mdot
.52569 .32551 Mdot
.52595 .3257 Mdot
.5262 .3258 Mdot
.52646 .32589 Mdot
.52672 .32609 Mdot
.52698 .32629 Mdot
.52724 .3264 Mdot
.52749 .3265 Mdot
.52775 .32671 Mdot
.52801 .32693 Mdot
.52827 .32704 Mdot
.52853 .32715 Mdot
.52879 .32738 Mdot
.52904 .32761 Mdot
.5293 .32773 Mdot
.52956 .32785 Mdot
.52982 .3281 Mdot
.53008 .32835 Mdot
.53033 .32848 Mdot
.53059 .32861 Mdot
.53085 .32868 Mdot
.53111 .32888 Mdot
.53137 .32908 Mdot
.53163 .32915 Mdot
.53188 .32929 Mdot
.53214 .32943 Mdot
.5324 .3295 Mdot
.53266 .32972 Mdot
.53292 .32994 Mdot
.53317 .33002 Mdot
.53343 .33017 Mdot
.53369 .33033 Mdot
.53395 .3304 Mdot
.53421 .33064 Mdot
.53447 .33088 Mdot
.53472 .33097 Mdot
.53498 .33113 Mdot
.53524 .3313 Mdot
.5355 .33139 Mdot
.53576 .33165 Mdot
.53601 .33191 Mdot
.53627 .332 Mdot
.53653 .33218 Mdot
.53679 .33237 Mdot
.53705 .33246 Mdot
.53731 .33275 Mdot
.53756 .33304 Mdot
.53782 .33314 Mdot
.53808 .33334 Mdot
.53834 .33346 Mdot
.5386 .33354 Mdot
.53885 .33365 Mdot
.53911 .33371 Mdot
.53937 .33396 Mdot
.53963 .33422 Mdot
.53989 .33429 Mdot
.54015 .3344 Mdot
.5404 .33448 Mdot
.54066 .33462 Mdot
.54092 .33475 Mdot
.54118 .33485 Mdot
.54144 .33496 Mdot
.54169 .33503 Mdot
.54195 .33531 Mdot
.54221 .3356 Mdot
.54247 .33567 Mdot
.54273 .33579 Mdot
.54298 .33589 Mdot
.54324 .33604 Mdot
.5435 .33619 Mdot
.54376 .33629 Mdot
.54402 .33642 Mdot
.54428 .3365 Mdot
.54453 .33681 Mdot
.54479 .33713 Mdot
.54505 .33721 Mdot
.54531 .33735 Mdot
.54557 .33746 Mdot
.54582 .33763 Mdot
.54608 .33779 Mdot
.54634 .33791 Mdot
.5466 .33805 Mdot
.54686 .33814 Mdot
.54712 .33849 Mdot
.54737 .33879 Mdot
.54763 .33885 Mdot
.54789 .33894 Mdot
.54815 .3391 Mdot
.54841 .33922 Mdot
.54866 .33941 Mdot
.54892 .3396 Mdot
.54918 .33973 Mdot
.54944 .33989 Mdot
.5497 .33999 Mdot
.54996 .34006 Mdot
.55021 .34039 Mdot
.55047 .34073 Mdot
.55073 .3408 Mdot
.55099 .3409 Mdot
.55125 .34108 Mdot
.5515 .34122 Mdot
.55176 .34143 Mdot
.55202 .34165 Mdot
.55228 .3418 Mdot
.55254 .34198 Mdot
.5528 .34209 Mdot
.55305 .34217 Mdot
.55331 .34255 Mdot
.55357 .34294 Mdot
.55383 .34302 Mdot
.55409 .34313 Mdot
.55434 .34322 Mdot
.5546 .34333 Mdot
.55486 .3435 Mdot
.55512 .34357 Mdot
.55538 .34374 Mdot
.55564 .34392 Mdot
.55589 .34399 Mdot
.55615 .34416 Mdot
.55641 .34428 Mdot
.55667 .34437 Mdot
.55693 .3445 Mdot
.55718 .34459 Mdot
.55744 .34465 Mdot
.5577 .34502 Mdot
.55796 .34541 Mdot
.55822 .34547 Mdot
.55848 .34556 Mdot
.55873 .3457 Mdot
.55899 .3458 Mdot
.55925 .34593 Mdot
.55951 .34612 Mdot
.55977 .3462 Mdot
.56002 .34641 Mdot
.56028 .34661 Mdot
.56054 .34669 Mdot
.5608 .34689 Mdot
.56106 .34703 Mdot
.56131 .34714 Mdot
.56157 .34729 Mdot
.56183 .34739 Mdot
.56209 .34746 Mdot
.56235 .3479 Mdot
.56261 .34829 Mdot
.56286 .34834 Mdot
.56312 .34842 Mdot
.56338 .34853 Mdot
.56364 .34869 Mdot
.5639 .3488 Mdot
.56415 .34896 Mdot
.56441 .3491 Mdot
.56467 .34918 Mdot
.56493 .34927 Mdot
.56519 .34951 Mdot
.56545 .34975 Mdot
.5657 .34985 Mdot
.56596 .34994 Mdot
.56622 .35008 Mdot
.56648 .35024 Mdot
.56674 .35037 Mdot
.56699 .35055 Mdot
.56725 .35066 Mdot
.56751 .35075 Mdot
.56777 .35081 Mdot
.56803 .35127 Mdot
.56829 .35173 Mdot
.56854 .35179 Mdot
.5688 .35188 Mdot
.56906 .35201 Mdot
.56932 .3522 Mdot
.56958 .35234 Mdot
.56983 .35252 Mdot
.57009 .35268 Mdot
.57035 .35278 Mdot
.57061 .35289 Mdot
.57087 .35296 Mdot
.57113 .35318 Mdot
.57138 .3534 Mdot
.57164 .35347 Mdot
.5719 .35358 Mdot
.57216 .35368 Mdot
.57242 .35385 Mdot
.57267 .35405 Mdot
.57293 .3542 Mdot
.57319 .35432 Mdot
.57345 .35441 Mdot
.57371 .35455 Mdot
.57397 .35465 Mdot
.57422 .35473 Mdot
.57448 .35479 Mdot
.57474 .35527 Mdot
.575 .35576 Mdot
.57526 .35583 Mdot
.57551 .35591 Mdot
.57577 .35601 Mdot
.57603 .35617 Mdot
.57629 .35627 Mdot
.57655 .35639 Mdot
.5768 .35656 Mdot
.57706 .35678 Mdot
.57732 .35698 Mdot
.57758 .35709 Mdot
.57784 .35723 Mdot
.5781 .35731 Mdot
.57835 .35757 Mdot
.57861 .35784 Mdot
.57887 .35792 Mdot
.57913 .35806 Mdot
.57939 .35818 Mdot
.57964 .35839 Mdot
.5799 .35856 Mdot
.58016 .35863 Mdot
.58042 .35881 Mdot
.58068 .35895 Mdot
.58094 .35907 Mdot
.58119 .35924 Mdot
.58145 .35936 Mdot
.58171 .35945 Mdot
.58197 .35953 Mdot
.58223 .35959 Mdot
.58248 .36011 Mdot
.58274 .36066 Mdot
.583 .36072 Mdot
.58326 .36079 Mdot
.58352 .36089 Mdot
.58378 .36102 Mdot
.58403 .36121 Mdot
.58429 .36133 Mdot
.58455 .36149 Mdot
.58481 .36169 Mdot
.58507 .36177 Mdot
.58532 .36196 Mdot
.58558 .3622 Mdot
.58584 .36235 Mdot
.5861 .36251 Mdot
.58636 .36261 Mdot
.58662 .36267 Mdot
.58687 .36294 Mdot
.58713 .36321 Mdot
.58739 .36327 Mdot
.58765 .36337 Mdot
.58791 .36354 Mdot
.58816 .36369 Mdot
.58842 .36376 Mdot
.58868 .36395 Mdot
.58894 .36415 Mdot
.5892 .36424 Mdot
.58946 .36432 Mdot
.58971 .36447 Mdot
.58997 .36464 Mdot
.59023 .36478 Mdot
.59049 .3649 Mdot
.59075 .365 Mdot
.591 .36515 Mdot
.59126 .36527 Mdot
.59152 .36536 Mdot
.59178 .36543 Mdot
.59204 .36549 Mdot
.59229 .36609 Mdot
.59255 .3667 Mdot
.59281 .36676 Mdot
.59307 .36684 Mdot
.59333 .36693 Mdot
.59359 .36706 Mdot
.59384 .36722 Mdot
.5941 .36733 Mdot
.59436 .36745 Mdot
.59462 .36761 Mdot
.59488 .3678 Mdot
.59513 .36797 Mdot
.59539 .36806 Mdot
.59565 .36816 Mdot
.59591 .3684 Mdot
.59617 .36862 Mdot
.59643 .3687 Mdot
.59668 .36888 Mdot
.59694 .36901 Mdot
.5972 .3691 Mdot
.59746 .36921 Mdot
.59772 .36929 Mdot
.59797 .36963 Mdot
.59823 .36997 Mdot
.59849 .37005 Mdot
.59875 .37017 Mdot
.59901 .37026 Mdot
.59927 .37039 Mdot
.59952 .37058 Mdot
.59978 .37067 Mdot
.60004 .37091 Mdot
.6003 .37117 Mdot
.60056 .37128 Mdot
.60081 .37139 Mdot
.60107 .37157 Mdot
.60133 .37179 Mdot
.60159 .37197 Mdot
.60185 .37212 Mdot
.60211 .37224 Mdot
.60236 .37244 Mdot
.60262 .37258 Mdot
.60288 .3727 Mdot
.60314 .37279 Mdot
.6034 .37287 Mdot
.60365 .37293 Mdot
.60391 .37298 Mdot
.60417 .37363 Mdot
.60443 .3743 Mdot
.60469 .37435 Mdot
.60495 .37442 Mdot
.6052 .3745 Mdot
.60546 .3746 Mdot
.60572 .37472 Mdot
.60598 .37488 Mdot
.60624 .37497 Mdot
.60649 .37509 Mdot
.60675 .37522 Mdot
.60701 .37539 Mdot
.60727 .37559 Mdot
.60753 .37567 Mdot
.60779 .37584 Mdot
.60804 .37605 Mdot
.6083 .37617 Mdot
.60856 .3763 Mdot
.60882 .37637 Mdot
.60908 .37661 Mdot
.60933 .3769 Mdot
.60959 .377 Mdot
.60985 .37709 Mdot
.61011 .37724 Mdot
.61037 .3774 Mdot
.61062 .37751 Mdot
.61088 .37766 Mdot
.61114 .37776 Mdot
.6114 .37783 Mdot
.61166 .3782 Mdot
.61192 .37858 Mdot
.61217 .37865 Mdot
.61243 .37875 Mdot
.61269 .37891 Mdot
.61295 .37903 Mdot
.61321 .3792 Mdot
.61346 .37935 Mdot
.61372 .37945 Mdot
.61398 .37956 Mdot
.61424 .37963 Mdot
.6145 .37987 Mdot
.61476 .38014 Mdot
.61501 .38022 Mdot
.61527 .38037 Mdot
.61553 .3805 Mdot
.61579 .38074 Mdot
.61605 .38094 Mdot
.6163 .38103 Mdot
.61656 .38112 Mdot
.61682 .38127 Mdot
.61708 .38146 Mdot
.61734 .38163 Mdot
.6176 .38176 Mdot
.61785 .38188 Mdot
.61811 .38208 Mdot
.61837 .38223 Mdot
.61863 .38235 Mdot
.61889 .38245 Mdot
.61914 .38253 Mdot
.6194 .38261 Mdot
.61966 .38267 Mdot
.61992 .38272 Mdot
.62018 .38347 Mdot
.62044 .38423 Mdot
.62069 .38428 Mdot
.62095 .38435 Mdot
.62121 .38442 Mdot
.62147 .38451 Mdot
.62173 .38462 Mdot
.62198 .38475 Mdot
.62224 .38491 Mdot
.6225 .38501 Mdot
.62276 .38512 Mdot
.62302 .38525 Mdot
.62328 .3854 Mdot
.62353 .38558 Mdot
.62379 .3858 Mdot
.62405 .38597 Mdot
.62431 .38607 Mdot
.62457 .38617 Mdot
.62482 .38641 Mdot
.62508 .38661 Mdot
.62534 .38669 Mdot
.6256 .38685 Mdot
.62586 .38702 Mdot
.62611 .38712 Mdot
.62637 .38718 Mdot
.62663 .38744 Mdot
.62689 .38774 Mdot
.62715 .38782 Mdot
.62741 .38797 Mdot
.62766 .38809 Mdot
.62792 .38829 Mdot
.62818 .3885 Mdot
.62844 .38865 Mdot
.6287 .38877 Mdot
.62895 .38886 Mdot
.62921 .38899 Mdot
.62947 .38908 Mdot
.62973 .38915 Mdot
.62999 .38959 Mdot
.63025 .39003 Mdot
.6305 .3901 Mdot
.63076 .3902 Mdot
.63102 .39033 Mdot
.63128 .39043 Mdot
.63154 .39055 Mdot
.63179 .39071 Mdot
.63205 .39094 Mdot
.63231 .39115 Mdot
.63257 .39128 Mdot
.63283 .39144 Mdot
.63309 .39153 Mdot
.63334 .39186 Mdot
.6336 .39215 Mdot
.63386 .39222 Mdot
.63412 .39233 Mdot
.63438 .39241 Mdot
.63463 .39253 Mdot
.63489 .39272 Mdot
.63515 .3928 Mdot
.63541 .39304 Mdot
.63567 .39325 Mdot
.63593 .39332 Mdot
.63618 .39344 Mdot
.63644 .39356 Mdot
.6367 .39377 Mdot
.63696 .39395 Mdot
.63722 .39403 Mdot
.63747 .39411 Mdot
.63773 .39426 Mdot
.63799 .39445 Mdot
.63825 .39461 Mdot
.63851 .39475 Mdot
.63877 .39488 Mdot
.63902 .39499 Mdot
.63928 .39509 Mdot
.63954 .39526 Mdot
.6398 .3954 Mdot
.64006 .39551 Mdot
.64031 .39561 Mdot
.64057 .39569 Mdot
.64083 .39577 Mdot
.64109 .39583 Mdot
.64135 .39589 Mdot
.64161 .3968 Mdot
.64186 .39768 Mdot
.64212 .39773 Mdot
.64238 .39779 Mdot
.64264 .39786 Mdot
.6429 .39794 Mdot
.64315 .39802 Mdot
.64341 .39813 Mdot
.64367 .39825 Mdot
.64393 .3984 Mdot
.64419 .39858 Mdot
.64444 .39869 Mdot
.6447 .39881 Mdot
.64496 .39894 Mdot
.64522 .3991 Mdot
.64548 .39928 Mdot
.64574 .3995 Mdot
.64599 .39966 Mdot
.64625 .39975 Mdot
.64651 .39985 Mdot
.64677 .40006 Mdot
.64703 .4003 Mdot
.64728 .40043 Mdot
.64754 .40058 Mdot
.6478 .40066 Mdot
.64806 .40091 Mdot
.64832 .4012 Mdot
.64858 .4013 Mdot
.64883 .40139 Mdot
.64909 .40153 Mdot
.64935 .40167 Mdot
.64961 .40178 Mdot
.64987 .40191 Mdot
.65012 .40199 Mdot
.65038 .40205 Mdot
.65064 .40236 Mdot
.6509 .40271 Mdot
.65116 .40279 Mdot
.65142 .4029 Mdot
.65167 .40299 Mdot
.65193 .40311 Mdot
.65219 .40328 Mdot
.65245 .40336 Mdot
.65271 .40356 Mdot
.65296 .40378 Mdot
.65322 .40386 Mdot
.65348 .40395 Mdot
.65374 .40408 Mdot
.654 .40425 Mdot
.65426 .40438 Mdot
.65451 .40456 Mdot
.65477 .40469 Mdot
.65503 .40479 Mdot
.65529 .40486 Mdot
.65555 .40492 Mdot
.6558 .40541 Mdot
.65606 .40591 Mdot
.65632 .40597 Mdot
.65658 .40605 Mdot
.65684 .40615 Mdot
.6571 .40628 Mdot
.65735 .40648 Mdot
.65761 .40661 Mdot
.65787 .40678 Mdot
.65813 .40693 Mdot
.65839 .40702 Mdot
.65864 .40711 Mdot
.6589 .40734 Mdot
.65916 .40756 Mdot
.65942 .40765 Mdot
.65968 .40784 Mdot
.65993 .40797 Mdot
.66019 .40806 Mdot
.66045 .40819 Mdot
.66071 .40828 Mdot
.66097 .40834 Mdot
.66123 .40867 Mdot
.66148 .40903 Mdot
.66174 .40909 Mdot
.662 .40919 Mdot
.66226 .40935 Mdot
.66252 .40947 Mdot
.66277 .40964 Mdot
.66303 .40981 Mdot
.66329 .40991 Mdot
.66355 .41003 Mdot
.66381 .41011 Mdot
.66407 .41038 Mdot
.66432 .41069 Mdot
.66458 .41078 Mdot
.66484 .41097 Mdot
.6651 .41113 Mdot
.66536 .41121 Mdot
.66561 .41144 Mdot
.66587 .4117 Mdot
.66613 .41183 Mdot
.66639 .41194 Mdot
.66665 .41215 Mdot
.66691 .41234 Mdot
.66716 .41243 Mdot
.66742 .41251 Mdot
.66768 .41267 Mdot
.66794 .41288 Mdot
.6682 .41306 Mdot
.66845 .41322 Mdot
.66871 .41337 Mdot
.66897 .4135 Mdot
.66923 .41361 Mdot
.66949 .41372 Mdot
.66975 .41381 Mdot
.67 .41398 Mdot
.67026 .41413 Mdot
.67052 .41425 Mdot
.67078 .41436 Mdot
.67104 .41445 Mdot
.67129 .41454 Mdot
.67155 .41461 Mdot
.67181 .41468 Mdot
.67207 .41474 Mdot
.67233 .41479 Mdot
.67259 .41588 Mdot
.67284 .41695 Mdot
.6731 .417 Mdot
.67336 .41705 Mdot
.67362 .41712 Mdot
.67388 .41719 Mdot
.67413 .41726 Mdot
.67439 .41735 Mdot
.67465 .41745 Mdot
.67491 .41756 Mdot
.67517 .41769 Mdot
.67543 .41785 Mdot
.67568 .41803 Mdot
.67594 .41814 Mdot
.6762 .41825 Mdot
.67646 .41838 Mdot
.67672 .41852 Mdot
.67697 .41868 Mdot
.67723 .41886 Mdot
.67749 .41906 Mdot
.67775 .41914 Mdot
.67801 .4193 Mdot
.67826 .41947 Mdot
.67852 .41957 Mdot
.67878 .41967 Mdot
.67904 .41989 Mdot
.6793 .42013 Mdot
.67956 .42027 Mdot
.67981 .42041 Mdot
.68007 .42048 Mdot
.68033 .42073 Mdot
.68059 .42099 Mdot
.68085 .42109 Mdot
.6811 .42129 Mdot
.68136 .42142 Mdot
.68162 .42151 Mdot
.68188 .42163 Mdot
.68214 .4217 Mdot
.6824 .422 Mdot
.68265 .42229 Mdot
.68291 .42235 Mdot
.68317 .42244 Mdot
.68343 .4226 Mdot
.68369 .42273 Mdot
.68394 .42294 Mdot
.6842 .42316 Mdot
.68446 .42331 Mdot
.68472 .42343 Mdot
.68498 .42352 Mdot
.68524 .42364 Mdot
.68549 .42373 Mdot
.68575 .4238 Mdot
.68601 .4242 Mdot
.68627 .42459 Mdot
.68653 .42465 Mdot
.68678 .42473 Mdot
.68704 .42485 Mdot
.6873 .42502 Mdot
.68756 .42515 Mdot
.68782 .42533 Mdot
.68808 .4255 Mdot
.68833 .4256 Mdot
.68859 .42571 Mdot
.68885 .42578 Mdot
.68911 .42602 Mdot
.68937 .42627 Mdot
.68962 .42635 Mdot
.68988 .42649 Mdot
.69014 .42661 Mdot
.6904 .42682 Mdot
.69066 .42699 Mdot
.69092 .42707 Mdot
.69117 .42727 Mdot
.69143 .42742 Mdot
.69169 .42755 Mdot
.69195 .42766 Mdot
.69221 .42775 Mdot
.69246 .4279 Mdot
.69272 .42801 Mdot
.69298 .4281 Mdot
.69324 .42817 Mdot
.6935 .42823 Mdot
.69375 .42886 Mdot
.69401 .4295 Mdot
.69427 .42956 Mdot
.69453 .42963 Mdot
.69479 .42972 Mdot
.69505 .42984 Mdot
.6953 .42999 Mdot
.69556 .43008 Mdot
.69582 .4302 Mdot
.69608 .43033 Mdot
.69634 .4305 Mdot
.69659 .4307 Mdot
.69685 .43079 Mdot
.69711 .43097 Mdot
.69737 .4312 Mdot
.69763 .43133 Mdot
.69789 .43148 Mdot
.69814 .43156 Mdot
.6984 .43184 Mdot
.69866 .4321 Mdot
.69892 .43218 Mdot
.69918 .43231 Mdot
.69943 .43242 Mdot
.69969 .4326 Mdot
.69995 .43281 Mdot
.70021 .43296 Mdot
.70047 .43307 Mdot
.70073 .43316 Mdot
.70098 .43329 Mdot
.70124 .43338 Mdot
.7015 .43345 Mdot
.70176 .4339 Mdot
.70202 .43433 Mdot
.70227 .43438 Mdot
.70253 .43446 Mdot
.70279 .43456 Mdot
.70305 .43472 Mdot
.70331 .43482 Mdot
.70357 .43496 Mdot
.70382 .43515 Mdot
.70408 .43523 Mdot
.70434 .43542 Mdot
.7046 .43561 Mdot
.70486 .43568 Mdot
.70511 .43584 Mdot
.70537 .43604 Mdot
.70563 .43616 Mdot
.70589 .43623 Mdot
.70615 .43659 Mdot
.70641 .43692 Mdot
.70666 .43698 Mdot
.70692 .43707 Mdot
.70718 .43722 Mdot
.70744 .43734 Mdot
.7077 .43751 Mdot
.70795 .43767 Mdot
.70821 .43777 Mdot
.70847 .4379 Mdot
.70873 .43797 Mdot
.70899 .43825 Mdot
.70924 .43851 Mdot
.7095 .43857 Mdot
.70976 .43867 Mdot
.71002 .43887 Mdot
.71028 .43905 Mdot
.71054 .43914 Mdot
.71079 .43938 Mdot
.71105 .43961 Mdot
.71131 .43969 Mdot
.71157 .43982 Mdot
.71183 .43996 Mdot
.71208 .44021 Mdot
.71234 .44043 Mdot
.7126 .44054 Mdot
.71286 .44064 Mdot
.71312 .44083 Mdot
.71338 .441 Mdot
.71363 .44109 Mdot
.71389 .44117 Mdot
.71415 .44132 Mdot
.71441 .44152 Mdot
.71467 .44171 Mdot
.71492 .44188 Mdot
.71518 .44203 Mdot
.71544 .44217 Mdot
.7157 .4423 Mdot
.71596 .44242 Mdot
.71622 .44253 Mdot
.71647 .44264 Mdot
.71673 .44273 Mdot
.71699 .4429 Mdot
.71725 .44305 Mdot
.71751 .44319 Mdot
.71776 .44331 Mdot
.71802 .44341 Mdot
.71828 .44351 Mdot
.71854 .4436 Mdot
.7188 .44367 Mdot
.71906 .44375 Mdot
.71931 .44381 Mdot
.71957 .44387 Mdot
.71983 .44393 Mdot
.72009 .44398 Mdot
.72035 .44536 Mdot
.7206 .44671 Mdot
.72086 .44676 Mdot
.72112 .44681 Mdot
.72138 .44687 Mdot
.72164 .44693 Mdot
.7219 .447 Mdot
.72215 .44708 Mdot
.72241 .44716 Mdot
.72267 .44725 Mdot
.72293 .44736 Mdot
.72319 .44747 Mdot
.72344 .44759 Mdot
.7237 .44774 Mdot
.72396 .4479 Mdot
.72422 .44808 Mdot
.72448 .44819 Mdot
.72474 .4483 Mdot
.72499 .44842 Mdot
.72525 .44855 Mdot
.72551 .44869 Mdot
.72577 .44885 Mdot
.72603 .44902 Mdot
.72628 .4492 Mdot
.72654 .44941 Mdot
.7268 .44949 Mdot
.72706 .44964 Mdot
.72732 .44982 Mdot
.72757 .44991 Mdot
.72783 .45 Mdot
.72809 .4502 Mdot
.72835 .45042 Mdot
.72861 .45053 Mdot
.72887 .45066 Mdot
.72912 .45092 Mdot
.72938 .45113 Mdot
.72964 .45121 Mdot
.7299 .45137 Mdot
.73016 .45153 Mdot
.73041 .45162 Mdot
.73067 .45189 Mdot
.73093 .45213 Mdot
.73119 .45219 Mdot
.73145 .45229 Mdot
.73171 .45238 Mdot
.73196 .45251 Mdot
.73222 .45265 Mdot
.73248 .45275 Mdot
.73274 .45287 Mdot
.733 .45295 Mdot
.73325 .45327 Mdot
.73351 .45356 Mdot
.73377 .45362 Mdot
.73403 .45371 Mdot
.73429 .45387 Mdot
.73455 .454 Mdot
.7348 .45421 Mdot
.73506 .45442 Mdot
.73532 .45457 Mdot
.73558 .45476 Mdot
.73584 .45488 Mdot
.73609 .45496 Mdot
.73635 .45502 Mdot
.73661 .45539 Mdot
.73687 .45579 Mdot
.73713 .45586 Mdot
.73739 .45596 Mdot
.73764 .45611 Mdot
.7379 .45622 Mdot
.73816 .45638 Mdot
.73842 .45651 Mdot
.73868 .4566 Mdot
.73893 .45669 Mdot
.73919 .45694 Mdot
.73945 .45721 Mdot
.73971 .45731 Mdot
.73997 .45741 Mdot
.74023 .45757 Mdot
.74048 .45777 Mdot
.74074 .45792 Mdot
.741 .45803 Mdot
.74126 .45813 Mdot
.74152 .45828 Mdot
.74177 .45838 Mdot
.74203 .45847 Mdot
.74229 .45853 Mdot
.74255 .45907 Mdot
.74281 .45959 Mdot
.74306 .45965 Mdot
.74332 .45972 Mdot
.74358 .45982 Mdot
.74384 .45996 Mdot
.7441 .46015 Mdot
.74436 .46028 Mdot
.74461 .46045 Mdot
.74487 .46059 Mdot
.74513 .46067 Mdot
.74539 .46076 Mdot
.74565 .46097 Mdot
.7459 .46117 Mdot
.74616 .46125 Mdot
.74642 .46142 Mdot
.74668 .46162 Mdot
.74694 .46173 Mdot
.7472 .4618 Mdot
.74745 .46214 Mdot
.74771 .46249 Mdot
.74797 .46256 Mdot
.74823 .46269 Mdot
.74849 .46279 Mdot
.74874 .46292 Mdot
.749 .46305 Mdot
.74926 .46313 Mdot
.74952 .46323 Mdot
.74978 .46349 Mdot
.75004 .46372 Mdot
.75029 .46379 Mdot
.75055 .46392 Mdot
.75081 .46404 Mdot
.75107 .46425 Mdot
.75133 .46444 Mdot
.75158 .46452 Mdot
.75184 .4646 Mdot
.7521 .46474 Mdot
.75236 .46493 Mdot
.75262 .46508 Mdot
.75288 .46522 Mdot
.75313 .46533 Mdot
.75339 .46552 Mdot
.75365 .46568 Mdot
.75391 .4658 Mdot
.75417 .4659 Mdot
.75442 .46599 Mdot
.75468 .46606 Mdot
.75494 .46612 Mdot
.7552 .46618 Mdot
.75546 .46698 Mdot
.75572 .46779 Mdot
.75597 .46785 Mdot
.75623 .46791 Mdot
.75649 .46798 Mdot
.75675 .46807 Mdot
.75701 .46818 Mdot
.75726 .4683 Mdot
.75752 .46846 Mdot
.75778 .46856 Mdot
.75804 .46866 Mdot
.7583 .46878 Mdot
.75856 .46892 Mdot
.75881 .46909 Mdot
.75907 .46928 Mdot
.75933 .46943 Mdot
.75959 .46952 Mdot
.75985 .4696 Mdot
.7601 .4698 Mdot
.76036 .47003 Mdot
.76062 .47016 Mdot
.76088 .4703 Mdot
.76114 .47038 Mdot
.76139 .47063 Mdot
.76165 .47091 Mdot
.76191 .47102 Mdot
.76217 .47111 Mdot
.76243 .47125 Mdot
.76269 .4714 Mdot
.76294 .47151 Mdot
.7632 .47165 Mdot
.76346 .47174 Mdot
.76372 .4718 Mdot
.76398 .47213 Mdot
.76423 .47245 Mdot
.76449 .47251 Mdot
.76475 .47259 Mdot
.76501 .47272 Mdot
.76527 .47282 Mdot
.76553 .47295 Mdot
.76578 .47315 Mdot
.76604 .47324 Mdot
.7663 .47347 Mdot
.76656 .47372 Mdot
.76682 .47383 Mdot
.76707 .47392 Mdot
.76733 .47409 Mdot
.76759 .47428 Mdot
.76785 .47444 Mdot
.76811 .47457 Mdot
.76837 .47467 Mdot
.76862 .47483 Mdot
.76888 .47495 Mdot
.76914 .47504 Mdot
.7694 .47511 Mdot
.76966 .47517 Mdot
.76991 .47575 Mdot
.77017 .47635 Mdot
.77043 .47641 Mdot
.77069 .47649 Mdot
.77095 .47659 Mdot
.77121 .47673 Mdot
.77146 .47691 Mdot
.77172 .47703 Mdot
.77198 .47718 Mdot
.77224 .47737 Mdot
.7725 .47745 Mdot
.77275 .47762 Mdot
.77301 .47783 Mdot
.77327 .47795 Mdot
.77353 .47809 Mdot
.77379 .47817 Mdot
.77405 .47843 Mdot
.7743 .47869 Mdot
.77456 .47876 Mdot
.77482 .47889 Mdot
.77508 .479 Mdot
.77534 .47918 Mdot
.77559 .47938 Mdot
.77585 .47953 Mdot
.77611 .47964 Mdot
.77637 .47973 Mdot
.77663 .47986 Mdot
.77688 .47996 Mdot
.77714 .48003 Mdot
.7774 .4805 Mdot
.77766 .48094 Mdot
.77792 .481 Mdot
.77818 .48109 Mdot
.77843 .4812 Mdot
.77869 .48136 Mdot
.77895 .48148 Mdot
.77921 .48163 Mdot
.77947 .48184 Mdot
.77972 .48193 Mdot
.77998 .48214 Mdot
.78024 .48235 Mdot
.7805 .48243 Mdot
.78076 .48261 Mdot
.78102 .48274 Mdot
.78127 .48284 Mdot
.78153 .48297 Mdot
.78179 .48306 Mdot
.78205 .48312 Mdot
.78231 .48348 Mdot
.78256 .48387 Mdot
.78282 .48394 Mdot
.78308 .48405 Mdot
.78334 .48424 Mdot
.7836 .48438 Mdot
.78386 .48458 Mdot
.78411 .48478 Mdot
.78437 .48491 Mdot
.78463 .48507 Mdot
.78489 .48516 Mdot
.78515 .48522 Mdot
.7854 .48552 Mdot
.78566 .48585 Mdot
.78592 .48593 Mdot
.78618 .48606 Mdot
.78644 .48616 Mdot
.7867 .48631 Mdot
.78695 .48646 Mdot
.78721 .48655 Mdot
.78747 .48667 Mdot
.78773 .48674 Mdot
.78799 .487 Mdot
.78824 .48725 Mdot
.7885 .48731 Mdot
.78876 .48741 Mdot
.78902 .4876 Mdot
.78928 .48779 Mdot
.78954 .48788 Mdot
.78979 .48813 Mdot
.79005 .48838 Mdot
.79031 .48845 Mdot
.79057 .4886 Mdot
.79083 .48875 Mdot
.79108 .48882 Mdot
.79134 .48902 Mdot
.7916 .48928 Mdot
.79186 .4894 Mdot
.79212 .48952 Mdot
.79238 .48974 Mdot
.79263 .48995 Mdot
.79289 .49005 Mdot
.79315 .49015 Mdot
.79341 .49033 Mdot
.79367 .49051 Mdot
.79392 .49059 Mdot
.79418 .49067 Mdot
.79444 .49083 Mdot
.7947 .49105 Mdot
.79496 .49125 Mdot
.79521 .49144 Mdot
.79547 .49161 Mdot
.79573 .49177 Mdot
.79599 .49193 Mdot
.79625 .49207 Mdot
.79651 .4922 Mdot
.79676 .49233 Mdot
.79702 .49245 Mdot
.79728 .49256 Mdot
.79754 .49267 Mdot
.7978 .49277 Mdot
.79805 .49286 Mdot
.79831 .49304 Mdot
.79857 .4932 Mdot
.79883 .49334 Mdot
.79909 .49348 Mdot
.79935 .4936 Mdot
.7996 .49371 Mdot
.79986 .49382 Mdot
.80012 .49391 Mdot
.80038 .494 Mdot
.80064 .49409 Mdot
.80089 .49416 Mdot
.80115 .49424 Mdot
.80141 .4943 Mdot
.80167 .49437 Mdot
.80193 .49443 Mdot
.80219 .49449 Mdot
.80244 .49454 Mdot
.8027 .49459 Mdot
.80296 .49645 Mdot
.80322 .4983 Mdot
.80348 .49835 Mdot
.80373 .4984 Mdot
.80399 .49846 Mdot
.80425 .49852 Mdot
.80451 .49858 Mdot
.80477 .49865 Mdot
.80503 .49872 Mdot
.80528 .4988 Mdot
.80554 .49888 Mdot
.8058 .49897 Mdot
.80606 .49906 Mdot
.80632 .49917 Mdot
.80657 .49928 Mdot
.80683 .4994 Mdot
.80709 .49953 Mdot
.80735 .49967 Mdot
.80761 .49982 Mdot
.80787 .49999 Mdot
.80812 .50018 Mdot
.80838 .50029 Mdot
.80864 .5004 Mdot
.8089 .50051 Mdot
.80916 .50063 Mdot
.80941 .50076 Mdot
.80967 .5009 Mdot
.80993 .50105 Mdot
.81019 .5012 Mdot
.81045 .50137 Mdot
.8107 .50155 Mdot
.81096 .50174 Mdot
.81122 .50195 Mdot
.81148 .5021 Mdot
.81174 .50218 Mdot
.812 .50226 Mdot
.81225 .50243 Mdot
.81251 .5026 Mdot
.81277 .50269 Mdot
.81303 .50279 Mdot
.81329 .50299 Mdot
.81354 .5032 Mdot
.8138 .50331 Mdot
.81406 .50342 Mdot
.81432 .50367 Mdot
.81458 .50386 Mdot
.81484 .50392 Mdot
.81509 .50406 Mdot
.81535 .5042 Mdot
.81561 .50427 Mdot
.81587 .5045 Mdot
.81613 .50474 Mdot
.81638 .50482 Mdot
.81664 .50499 Mdot
.8169 .50517 Mdot
.81716 .50526 Mdot
.81742 .50555 Mdot
.81768 .5058 Mdot
.81793 .50586 Mdot
.81819 .50597 Mdot
.81845 .50605 Mdot
.81871 .50619 Mdot
.81897 .50633 Mdot
.81922 .50642 Mdot
.81948 .50654 Mdot
.81974 .50662 Mdot
.82 .50692 Mdot
.82026 .50725 Mdot
.82052 .50733 Mdot
.82077 .50747 Mdot
.82103 .50759 Mdot
.82129 .50777 Mdot
.82155 .50796 Mdot
.82181 .50809 Mdot
.82206 .50825 Mdot
.82232 .50835 Mdot
.82258 .50842 Mdot
.82284 .50878 Mdot
.8231 .50915 Mdot
.82336 .50923 Mdot
.82361 .50935 Mdot
.82387 .50943 Mdot
.82413 .50955 Mdot
.82439 .50972 Mdot
.82465 .50979 Mdot
.8249 .50997 Mdot
.82516 .51016 Mdot
.82542 .51024 Mdot
.82568 .51043 Mdot
.82594 .51056 Mdot
.82619 .51066 Mdot
.82645 .51081 Mdot
.82671 .51091 Mdot
.82697 .51098 Mdot
.82723 .51143 Mdot
.82749 .51184 Mdot
.82774 .5119 Mdot
.828 .51199 Mdot
.82826 .5121 Mdot
.82852 .51228 Mdot
.82878 .51241 Mdot
.82903 .51259 Mdot
.82929 .51274 Mdot
.82955 .51284 Mdot
.82981 .51295 Mdot
.83007 .51323 Mdot
.83033 .51346 Mdot
.83058 .51353 Mdot
.83084 .51365 Mdot
.8311 .51376 Mdot
.83136 .51394 Mdot
.83162 .51415 Mdot
.83187 .51431 Mdot
.83213 .51444 Mdot
.83239 .51454 Mdot
.83265 .5147 Mdot
.83291 .51482 Mdot
.83317 .5149 Mdot
.83342 .51497 Mdot
.83368 .51503 Mdot
.83394 .51554 Mdot
.8342 .51607 Mdot
.83446 .51614 Mdot
.83471 .51621 Mdot
.83497 .51631 Mdot
.83523 .51645 Mdot
.83549 .51664 Mdot
.83575 .51677 Mdot
.83601 .51694 Mdot
.83626 .51708 Mdot
.83652 .51716 Mdot
.83678 .51725 Mdot
.83704 .51746 Mdot
.8373 .51765 Mdot
.83755 .51772 Mdot
.83781 .51789 Mdot
.83807 .51808 Mdot
.83833 .51818 Mdot
.83859 .51825 Mdot
.83885 .51857 Mdot
.8391 .51889 Mdot
.83936 .51897 Mdot
.83962 .51908 Mdot
.83988 .51917 Mdot
.84014 .51929 Mdot
.84039 .51948 Mdot
.84065 .51957 Mdot
.84091 .51981 Mdot
.84117 .52007 Mdot
.84143 .52018 Mdot
.84169 .52029 Mdot
.84194 .52048 Mdot
.8422 .52064 Mdot
.84246 .52071 Mdot
.84272 .5209 Mdot
.84298 .52106 Mdot
.84323 .52119 Mdot
.84349 .5213 Mdot
.84375 .5214 Mdot
.84401 .52156 Mdot
.84427 .52169 Mdot
.84452 .52179 Mdot
.84478 .52187 Mdot
.84504 .52194 Mdot
.8453 .522 Mdot
.84556 .52275 Mdot
.84582 .52348 Mdot
.84607 .52353 Mdot
.84633 .5236 Mdot
.84659 .52368 Mdot
.84685 .52378 Mdot
.84711 .5239 Mdot
.84736 .52405 Mdot
.84762 .52414 Mdot
.84788 .52425 Mdot
.84814 .52437 Mdot
.8484 .52451 Mdot
.84866 .52469 Mdot
.84891 .5249 Mdot
.84917 .52506 Mdot
.84943 .52516 Mdot
.84969 .52526 Mdot
.84995 .52549 Mdot
.8502 .52569 Mdot
.85046 .52577 Mdot
.85072 .52593 Mdot
.85098 .52611 Mdot
.85124 .52621 Mdot
.8515 .52627 Mdot
.85175 .52653 Mdot
.85201 .52685 Mdot
.85227 .52693 Mdot
.85253 .52709 Mdot
.85279 .52722 Mdot
.85304 .52743 Mdot
.8533 .52766 Mdot
.85356 .52783 Mdot
.85382 .52795 Mdot
.85408 .52805 Mdot
.85434 .5282 Mdot
.85459 .5283 Mdot
.85485 .52837 Mdot
.85511 .52843 Mdot
.85537 .52887 Mdot
.85563 .52932 Mdot
.85588 .52938 Mdot
.85614 .52946 Mdot
.8564 .52957 Mdot
.85666 .52973 Mdot
.85692 .52984 Mdot
.85718 .52998 Mdot
.85743 .53018 Mdot
.85769 .53026 Mdot
.85795 .53045 Mdot
.85821 .53064 Mdot
.85847 .53071 Mdot
.85872 .53087 Mdot
.85898 .53107 Mdot
.85924 .53118 Mdot
.8595 .53125 Mdot
.85976 .5316 Mdot
.86001 .53191 Mdot
.86027 .53197 Mdot
.86053 .53206 Mdot
.86079 .5322 Mdot
.86105 .53231 Mdot
.86131 .53246 Mdot
.86156 .53261 Mdot
.86182 .53271 Mdot
.86208 .53282 Mdot
.86234 .53289 Mdot
.8626 .53314 Mdot
.86285 .53343 Mdot
.86311 .53351 Mdot
.86337 .53368 Mdot
.86363 .53384 Mdot
.86389 .53392 Mdot
.86415 .53413 Mdot
.8644 .53439 Mdot
.86466 .53451 Mdot
.86492 .53462 Mdot
.86518 .53482 Mdot
.86544 .53501 Mdot
.86569 .5351 Mdot
.86595 .53518 Mdot
.86621 .53533 Mdot
.86647 .53554 Mdot
.86673 .53573 Mdot
.86699 .53589 Mdot
.86724 .53603 Mdot
.8675 .53617 Mdot
.86776 .53629 Mdot
.86802 .5364 Mdot
.86828 .53649 Mdot
.86853 .53667 Mdot
.86879 .53682 Mdot
.86905 .53695 Mdot
.86931 .53706 Mdot
.86957 .53716 Mdot
.86983 .53725 Mdot
.87008 .53733 Mdot
.87034 .5374 Mdot
.8706 .53746 Mdot
.87086 .53752 Mdot
.87112 .53757 Mdot
.87137 .5387 Mdot
.87163 .53985 Mdot
.87189 .5399 Mdot
.87215 .53996 Mdot
.87241 .54002 Mdot
.87267 .54009 Mdot
.87292 .54018 Mdot
.87318 .54026 Mdot
.87344 .54037 Mdot
.8737 .54048 Mdot
.87396 .54061 Mdot
.87421 .54077 Mdot
.87447 .54095 Mdot
.87473 .54105 Mdot
.87499 .54117 Mdot
.87525 .54129 Mdot
.87551 .54143 Mdot
.87576 .54158 Mdot
.87602 .54175 Mdot
.87628 .54195 Mdot
.87654 .54217 Mdot
.8768 .54233 Mdot
.87705 .54242 Mdot
.87731 .54251 Mdot
.87757 .54271 Mdot
.87783 .54293 Mdot
.87809 .54305 Mdot
.87834 .54317 Mdot
.8786 .54324 Mdot
.87886 .54345 Mdot
.87912 .54368 Mdot
.87938 .54376 Mdot
.87964 .54394 Mdot
.87989 .54412 Mdot
.88015 .54422 Mdot
.88041 .54453 Mdot
.88067 .54481 Mdot
.88093 .54488 Mdot
.88118 .54501 Mdot
.88144 .54512 Mdot
.8817 .54528 Mdot
.88196 .54545 Mdot
.88222 .54557 Mdot
.88248 .54573 Mdot
.88273 .54583 Mdot
.88299 .5459 Mdot
.88325 .54625 Mdot
.88351 .54658 Mdot
.88377 .54664 Mdot
.88402 .54672 Mdot
.88428 .54685 Mdot
.88454 .54694 Mdot
.8848 .54707 Mdot
.88506 .54726 Mdot
.88532 .54734 Mdot
.88557 .54755 Mdot
.88583 .54778 Mdot
.88609 .54787 Mdot
.88635 .54795 Mdot
.88661 .54809 Mdot
.88686 .54826 Mdot
.88712 .54839 Mdot
.88738 .54857 Mdot
.88764 .5487 Mdot
.8879 .5488 Mdot
.88816 .54887 Mdot
.88841 .54892 Mdot
.88867 .5494 Mdot
.88893 .54992 Mdot
.88919 .54999 Mdot
.88945 .55008 Mdot
.8897 .5502 Mdot
.88996 .55037 Mdot
.89022 .55049 Mdot
.89048 .55065 Mdot
.89074 .55085 Mdot
.891 .55093 Mdot
.89125 .55113 Mdot
.89151 .55131 Mdot
.89177 .55138 Mdot
.89203 .55154 Mdot
.89229 .55173 Mdot
.89254 .55184 Mdot
.8928 .5519 Mdot
.89306 .55222 Mdot
.89332 .55255 Mdot
.89358 .55263 Mdot
.89383 .55275 Mdot
.89409 .55284 Mdot
.89435 .55297 Mdot
.89461 .55318 Mdot
.89487 .55327 Mdot
.89513 .55352 Mdot
.89538 .55375 Mdot
.89564 .55382 Mdot
.8959 .55394 Mdot
.89616 .55406 Mdot
.89642 .55427 Mdot
.89667 .55446 Mdot
.89693 .55454 Mdot
.89719 .55462 Mdot
.89745 .55476 Mdot
.89771 .55495 Mdot
.89797 .55511 Mdot
.89822 .55524 Mdot
.89848 .55536 Mdot
.89874 .55546 Mdot
.899 .55556 Mdot
.89926 .55571 Mdot
.89951 .55584 Mdot
.89977 .55594 Mdot
.90003 .55603 Mdot
.90029 .55611 Mdot
.90055 .55617 Mdot
.90081 .55623 Mdot
.90106 .55706 Mdot
.90132 .55787 Mdot
.90158 .55793 Mdot
.90184 .55799 Mdot
.9021 .55805 Mdot
.90235 .55814 Mdot
.90261 .55823 Mdot
.90287 .55834 Mdot
.90313 .55848 Mdot
.90339 .55865 Mdot
.90365 .55875 Mdot
.9039 .55887 Mdot
.90416 .559 Mdot
.90442 .55916 Mdot
.90468 .55934 Mdot
.90494 .55955 Mdot
.90519 .55972 Mdot
.90545 .55981 Mdot
.90571 .55991 Mdot
.90597 .56014 Mdot
.90623 .56033 Mdot
.90649 .56039 Mdot
.90674 .56054 Mdot
.907 .5607 Mdot
.90726 .56079 Mdot
.90752 .56107 Mdot
.90778 .56133 Mdot
.90803 .5614 Mdot
.90829 .56152 Mdot
.90855 .56163 Mdot
.90881 .56179 Mdot
.90907 .56197 Mdot
.90933 .56209 Mdot
.90958 .56226 Mdot
.90984 .56236 Mdot
.9101 .56243 Mdot
.91036 .56282 Mdot
.91062 .56321 Mdot
.91087 .56328 Mdot
.91113 .56338 Mdot
.91139 .56353 Mdot
.91165 .56365 Mdot
.91191 .56381 Mdot
.91216 .56396 Mdot
.91242 .56405 Mdot
.91268 .56416 Mdot
.91294 .56444 Mdot
.9132 .56468 Mdot
.91346 .56475 Mdot
.91371 .56488 Mdot
.91397 .56499 Mdot
.91423 .5652 Mdot
.91449 .56537 Mdot
.91475 .56545 Mdot
.915 .56564 Mdot
.91526 .5658 Mdot
.91552 .56593 Mdot
.91578 .56604 Mdot
.91604 .56613 Mdot
.9163 .56629 Mdot
.91655 .5664 Mdot
.91681 .5665 Mdot
.91707 .56657 Mdot
.91733 .56663 Mdot
.91759 .56669 Mdot
.91784 .56731 Mdot
.9181 .56796 Mdot
.91836 .56802 Mdot
.91862 .56809 Mdot
.91888 .56818 Mdot
.91914 .56828 Mdot
.91939 .56842 Mdot
.91965 .56859 Mdot
.91991 .56871 Mdot
.92017 .56884 Mdot
.92043 .569 Mdot
.92068 .5692 Mdot
.92094 .56936 Mdot
.9212 .56946 Mdot
.92146 .56956 Mdot
.92172 .56979 Mdot
.92198 .57 Mdot
.92223 .57007 Mdot
.92249 .57024 Mdot
.92275 .57043 Mdot
.92301 .57054 Mdot
.92327 .57061 Mdot
.92352 .5709 Mdot
.92378 .57119 Mdot
.92404 .57126 Mdot
.9243 .57136 Mdot
.92456 .57154 Mdot
.92482 .57169 Mdot
.92507 .57176 Mdot
.92533 .57195 Mdot
.92559 .57215 Mdot
.92585 .57223 Mdot
.92611 .57231 Mdot
.92636 .57244 Mdot
.92662 .57261 Mdot
.92688 .57273 Mdot
.92714 .57293 Mdot
.9274 .57306 Mdot
.92765 .57316 Mdot
.92791 .57324 Mdot
.92817 .5733 Mdot
.92843 .57384 Mdot
.92869 .57437 Mdot
.92895 .57443 Mdot
.9292 .5745 Mdot
.92946 .57459 Mdot
.92972 .57471 Mdot
.92998 .57488 Mdot
.93024 .57499 Mdot
.93049 .57512 Mdot
.93075 .5753 Mdot
.93101 .57553 Mdot
.93127 .57572 Mdot
.93153 .57584 Mdot
.93179 .57597 Mdot
.93204 .57604 Mdot
.9323 .5763 Mdot
.93256 .57654 Mdot
.93282 .57661 Mdot
.93308 .57674 Mdot
.93333 .57685 Mdot
.93359 .57703 Mdot
.93385 .57723 Mdot
.93411 .57738 Mdot
.93437 .57749 Mdot
.93463 .57758 Mdot
.93488 .57772 Mdot
.93514 .57781 Mdot
.9354 .57788 Mdot
.93566 .57794 Mdot
.93592 .57837 Mdot
.93617 .57884 Mdot
.93643 .57891 Mdot
.93669 .579 Mdot
.93695 .57912 Mdot
.93721 .5793 Mdot
.93747 .57942 Mdot
.93772 .57959 Mdot
.93798 .57973 Mdot
.93824 .57982 Mdot
.9385 .57991 Mdot
.93876 .58015 Mdot
.93901 .58039 Mdot
.93927 .58048 Mdot
.93953 .58056 Mdot
.93979 .58069 Mdot
.94005 .58084 Mdot
.94031 .58095 Mdot
.94056 .5811 Mdot
.94082 .5812 Mdot
.94108 .58128 Mdot
.94134 .5817 Mdot
.9416 .5821 Mdot
.94185 .58216 Mdot
.94211 .58226 Mdot
.94237 .58239 Mdot
.94263 .58249 Mdot
.94289 .58261 Mdot
.94314 .58279 Mdot
.9434 .58286 Mdot
.94366 .58305 Mdot
.94392 .58322 Mdot
.94418 .5833 Mdot
.94444 .58346 Mdot
.94469 .58357 Mdot
.94495 .58366 Mdot
.94521 .58378 Mdot
.94547 .58386 Mdot
.94573 .58424 Mdot
.94598 .5846 Mdot
.94624 .58468 Mdot
.9465 .58478 Mdot
.94676 .58496 Mdot
.94702 .58509 Mdot
.94728 .5853 Mdot
.94753 .58549 Mdot
.94779 .58562 Mdot
.94805 .58578 Mdot
.94831 .58588 Mdot
.94857 .58594 Mdot
.94882 .58624 Mdot
.94908 .5866 Mdot
.94934 .58668 Mdot
.9496 .58682 Mdot
.94986 .58693 Mdot
.95012 .5871 Mdot
.95037 .58726 Mdot
.95063 .58736 Mdot
.95089 .58749 Mdot
.95115 .58757 Mdot
.95141 .58787 Mdot
.95166 .58816 Mdot
.95192 .58823 Mdot
.95218 .58834 Mdot
.95244 .58844 Mdot
.9527 .58857 Mdot
.95296 .5887 Mdot
.95321 .58879 Mdot
.95347 .5889 Mdot
.95373 .58896 Mdot
.95399 .58921 Mdot
.95425 .58951 Mdot
.9545 .58961 Mdot
.95476 .5898 Mdot
.95502 .58999 Mdot
.95528 .59008 Mdot
.95554 .59034 Mdot
.9558 .5906 Mdot
.95605 .59068 Mdot
.95631 .59084 Mdot
.95657 .591 Mdot
.95683 .59108 Mdot
.95709 .5913 Mdot
.95734 .59152 Mdot
.9576 .59159 Mdot
.95786 .59173 Mdot
.95812 .59187 Mdot
.95838 .59193 Mdot
.95864 .59213 Mdot
.95889 .59238 Mdot
.95915 .5925 Mdot
.95941 .59262 Mdot
.95967 .59285 Mdot
.95993 .59307 Mdot
.96018 .59318 Mdot
.96044 .59328 Mdot
.9607 .59349 Mdot
.96096 .59368 Mdot
.96122 .59378 Mdot
.96147 .59387 Mdot
.96173 .59405 Mdot
.96199 .59422 Mdot
.96225 .59431 Mdot
.96251 .59439 Mdot
.96277 .59455 Mdot
.96302 .5947 Mdot
.96328 .59478 Mdot
.96354 .59485 Mdot
.9638 .595 Mdot
.96406 .59521 Mdot
.96431 .5954 Mdot
.96457 .59559 Mdot
.96483 .59577 Mdot
.96509 .59594 Mdot
.96535 .59611 Mdot
.96561 .59626 Mdot
.96586 .59641 Mdot
.96612 .59656 Mdot
.96638 .5967 Mdot
.96664 .59683 Mdot
.9669 .59696 Mdot
.96715 .59708 Mdot
.96741 .5972 Mdot
.96767 .59731 Mdot
.96793 .59742 Mdot
.96819 .59752 Mdot
.96845 .59763 Mdot
.9687 .59782 Mdot
.96896 .598 Mdot
.96922 .59817 Mdot
.96948 .59833 Mdot
.96974 .59848 Mdot
.96999 .59862 Mdot
.97025 .59875 Mdot
.97051 .59887 Mdot
.97077 .59899 Mdot
.97103 .5991 Mdot
.97129 .59921 Mdot
.97154 .59931 Mdot
.9718 .59941 Mdot
.97206 .5995 Mdot
.97232 .59959 Mdot
.97258 .59967 Mdot
.97283 .59975 Mdot
.97309 .59983 Mdot
.97335 .5999 Mdot
.97361 .59997 Mdot
.97387 .60004 Mdot
.97413 .60011 Mdot
.97438 .60017 Mdot
.97464 .60023 Mdot
.9749 .60029 Mdot
.97516 .60034 Mdot
.97542 .6004 Mdot
.97567 .60045 Mdot
.97593 .6005 Mdot
.97619 .60332 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{345, 213.125},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40005I0000eB000`400?l00000o`00003oo`3ooomJ
0?ooo`00o`3ooomJ0?ooo`00o`3ooomJ0?ooo`00o`3ooomJ0?ooo`00@03oool3000000@0oooo0P00
00040?ooo`8000005@3oool4000000<0oooo0P0000040?ooo`800000103oool200000180oooo1000
00020?ooo`<00000103oool2000000@0oooo0P00000B0?ooo`@000000`3oool2000000@0oooo0P00
00040?ooo`8000004@3oool400000080oooo0`0000040?ooo`800000103oool2000001<0oooo0P00
00040?ooo`800000103oool2000000@0oooo0P00000C0?ooo`8000000`3oool3000000@0oooo0P00
00040?ooo`8000007P3oool004<0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?oo
o`0000020?ooo`030000003oool0oooo01@0oooo00D000000?ooo`3oool0oooo000000020?ooo`04
0000003oool0oooo00000080oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`0A
0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?oo
o`0000020?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool00`3oool010000000oooo
0?ooo`0000020?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool00000403oool0
0`000000oooo0?ooo`060?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool00000
0P3oool00`000000oooo0?ooo`0?0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3o
ool000000P3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo00000140oooo00@0
00000?ooo`3oool000001@3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo0000
0080oooo00<000000?ooo`3oool06`3oool004<0oooo00@000000?ooo`3oool000000P3oool01000
0000oooo0?ooo`0000020?ooo`030000003oool0oooo01@0oooo00D000000?ooo`3oool0oooo0000
00020?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool000000P3oool00`000000
oooo0?ooo`0A0?ooo`030000003oool0oooo00@0oooo00@000000?ooo`3oool000000P3oool01000
0000oooo0?ooo`0000020?ooo`030000003oool0oooo0100oooo00<000000?ooo`3oool00P3oool0
10000000oooo0?ooo`0000020?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool0
00004@3oool00`000000oooo0?ooo`050?ooo`040000003oool0oooo00000080oooo00@000000?oo
o`3oool000000P3oool00`000000oooo0?ooo`0B0?ooo`040000003oool0oooo00000080oooo00@0
00000?ooo`3oool000000P3oool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo0180
oooo00<000000?ooo`3oool00`3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo
00000080oooo00<000000?ooo`3oool06`3oool00400oooo0`0000030?ooo`040000003oool0oooo
00000080oooo00@000000?ooo`3oool000005P3oool01@000000oooo0?ooo`3oool000000080oooo
00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo
0140oooo00@000000?ooo`3oool0oooo0`0000030?ooo`040000003oool0oooo00000080oooo00@0
00000?ooo`3oool000004`3oool01@000000oooo0?ooo`3oool000000080oooo00@000000?ooo`3o
ool000000P3oool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo0100oooo00@00000
0?ooo`3oool0oooo0`0000030?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool0
00004P3oool2000000<0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`000002
0?ooo`040000003oool0oooo00000180oooo0P0000030?ooo`<000000`3oool010000000oooo0?oo
o`0000020?ooo`040000003oool0oooo000001d0oooo00100?ooo`030000003oool0oooo00<0oooo
00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`00000F0?ooo`050000003oool0oooo
0?ooo`0000000P3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo00000080oooo
00<000000?ooo`3oool04@3oool01@000000oooo0?ooo`3oool0000000D0oooo00@000000?ooo`3o
ool000000P3oool010000000oooo0?ooo`00000D0?ooo`040000003oool0oooo00000080oooo00@0
00000?ooo`3oool000000P3oool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo0140
oooo00@000000?ooo`3oool000001@3oool010000000oooo0?ooo`0000020?ooo`040000003oool0
oooo000001@0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`0000020?ooo`04
0000003oool0oooo00000080oooo00<000000?ooo`3oool04P3oool010000000oooo0?ooo`000005
0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool000007@3oool00240oooo00<0
00000?ooo`3oool0703oool00`000000oooo0?ooo`030?ooo`040000003oool0oooo00000080oooo
00@000000?ooo`3oool00000503oool3000000<0oooo00@000000?ooo`3oool000000P3oool01000
0000oooo0?ooo`0000020?ooo`040000003oool0oooo00000140oooo0`0000030?ooo`030000003o
ool0oooo00<0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`00000A0?ooo`04
0000003oool0oooo00000080oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`00
00020?ooo`040000003oool0oooo00000100oooo00@000000?ooo`3oool000000P3oool00`000000
oooo0?ooo`030?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool000004@3oool0
10000000oooo0?ooo`0000020?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool0
00000P3oool010000000oooo0?ooo`00000A0?ooo`040000003oool0oooo00000080oooo00<00000
0?ooo`3oool00`3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000001d0oooo
000Q0?ooo`030000003oool0oooo01`0oooo100000030?ooo`800000103oool2000001L0oooo00<0
00000?ooo`3oool00P3oool2000000@0oooo0P0000040?ooo`800000503oool010000000oooo0?oo
o`3oool4000000<0oooo0P0000040?ooo`8000004`3oool2000000@0oooo0P0000040?ooo`800000
103oool200000180oooo0P0000030?ooo`@000000`3oool2000000@0oooo0P00000C0?ooo`800000
103oool2000000@0oooo0P0000040?ooo`8000004`3oool2000000<0oooo100000030?ooo`800000
103oool2000001h0oooo000Q0?ooo`030000003oool0oooo0?l0oooo=P3oool00240oooo00<00000
0?ooo`3oool0o`3ooolf0?ooo`006P3ooooo000002l000004@3oool00200oooo100000050?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo00@0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`03
0000003oool0oooo01D0oooo000Q0?ooo`D000008P3oool00`000000oooo0?ooo`0T0?ooo`030000
003oool0oooo02@0oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`0S0?ooo`030000
003oool0oooo02@0oooo00<000000?ooo`3oool0903oool00`000000oooo0?ooo`0U0?ooo`008@3o
ool00`000000oooo0000000400000?l0oooo<P3oool00240oooo00@000000?ooo`3oool0oooo1@00
003o0?oooc00oooo000Q0?ooo`030000003oool0oooo00<0oooo1000003o0?ooobl0oooo000Q0?oo
o`8000001P3oool400000?l0oooo;@3oool00240oooo00<000000?ooo`3oool01P3oool400000?l0
oooo;03oool00240oooo00<000000?ooo`3oool0203oool400000?l0oooo:P3oool00240oooo00<0
00000?ooo`3oool02@3oool500000?l0oooo:03oool00240oooo00<000000?ooo`3oool02`3oool5
00000?l0oooo9P3oool00240oooo00<000000?ooo`3oool03@3oool400000?l0oooo9@3oool00240
oooo0P00000@0?ooo`@00000o`3ooolS0?ooo`008@3oool00`000000oooo0?ooo`0@0?ooo`D00000
o`3ooolQ0?ooo`008@3oool00`000000oooo0?ooo`0B0?ooo`@00000o`3ooolP0?ooo`008@3oool0
0`000000oooo0?ooo`0D0?ooo`@00000o`3ooolN0?ooo`008@3oool00`000000oooo0?ooo`0E0?oo
o`@00000o`3ooolM0?ooo`008@3oool00`000000oooo0?ooo`0G0?ooo`@00000o`3ooolK0?ooo`00
8@3oool2000001T0oooo1@00003o0?oooaT0oooo000Q0?ooo`030000003oool0oooo01X0oooo1000
003o0?oooaP0oooo000Q0?ooo`030000003oool0oooo01`0oooo1000003o0?oooaH0oooo000Q0?oo
o`030000003oool0oooo01d0oooo1000003o0?oooaD0oooo000Q0?ooo`030000003oool0oooo01l0
oooo1000003o0?oooa<0oooo000Q0?ooo`030000003oool0oooo0200oooo1@00003o0?oooa40oooo
000Q0?ooo`8000008`3oool400000?l0oooo403oool00240oooo00<000000?ooo`3oool0903oool4
00000?l0oooo3P3oool00240oooo00<000000?ooo`3oool09@3oool400000?l0oooo3@3oool00240
oooo00<000000?ooo`3oool09`3oool400000?l0oooo2`3oool00180oooo100000030?ooo`800000
1P3oool00`000000oooo0?ooo`0X0?ooo`D00000o`3oool90?ooo`00503oool01@000000oooo0?oo
o`3oool000000080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`0Z0?ooo`@00000
o`3oool80?ooo`00503oool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`3oool0
0`3oool3000002`0oooo1000003o0?ooo`H0oooo000D0?ooo`050000003oool0oooo0?ooo`000000
0P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo02d0oooo1@00003o0?ooo`@0oooo
000D0?ooo`050000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo02l0oooo0`00003o0?ooo`@0oooo000B0?ooo`<000000`3oool010000000oooo0?oo
o`0000050?ooo`030000003oool0oooo0300oooo1@00003o0?ooo`40oooo000D0?ooo`030000003o
ool0oooo0080oooo0P0000060?ooo`030000003oool0oooo0340oooo1@00003o0?ooo`008@3oool0
0`000000oooo0?ooo`0d0?ooo`@00000o@3oool00240oooo0P00000f0?ooo`@00000o03oool00240
oooo00<000000?ooo`3oool0=`3oool400000?X0oooo000Q0?ooo`030000003oool0oooo03T0oooo
0`00003i0?ooo`008@3oool00`000000oooo0?ooo`0j0?ooo`@00000m`3oool00240oooo00<00000
0?ooo`3oool0?03oool400000?D0oooo000Q0?ooo`030000003oool0oooo03d0oooo1@00003c0?oo
o`008@3oool200000400oooo1000003b0?ooo`008@3oool00`000000oooo0?ooo`110?ooo`@00000
l03oool00240oooo00<000000?ooo`3oool0@P3oool400000>l0oooo000Q0?ooo`030000003oool0
oooo04@0oooo1000003]0?ooo`008@3oool00`000000oooo0?ooo`160?ooo`@00000j`3oool00240
oooo0P0000180?ooo`@00000jP3oool00240oooo00<000000?ooo`3oool0B@3oool400000>P0oooo
000Q0?ooo`030000003oool0oooo04X0oooo1@00003V0?ooo`008@3oool00`000000oooo0?ooo`1<
0?ooo`@00000i@3oool00240oooo00<000000?ooo`3oool0CP3oool400000><0oooo000Q0?ooo`03
0000003oool0oooo04l0oooo1000003R0?ooo`008@3oool200000580oooo1000003P0?ooo`008@3o
ool00`000000oooo0?ooo`1B0?ooo`D00000gP3oool00240oooo00<000000?ooo`3oool0E03oool4
00000=d0oooo000Q0?ooo`030000003oool0oooo05H0oooo1000003K0?ooo`008@3oool00`000000
oooo0?ooo`1G0?ooo`D00000f@3oool00180oooo100000030?ooo`8000001P3oool00`000000oooo
0?ooo`1I0?ooo`@00000f03oool00180oooo00<000000?ooo`3oool00`3oool010000000oooo0?oo
o`0000050?ooo`<00000F`3oool400000=H0oooo000C0?ooo`030000003oool0oooo0080oooo00@0
00000?ooo`3oool000001@3oool00`000000oooo0?ooo`1L0?ooo`@00000e@3oool001@0oooo00D0
00000?ooo`3oool0oooo000000020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
GP3oool400000=<0oooo000E0?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`1O0?ooo`@00000dP3oool00180oooo00@000000?ooo`3oool00000
0P3oool010000000oooo0?ooo`0000050?ooo`030000003oool0oooo0640oooo1000003@0?ooo`00
4`3oool2000000@0oooo0P0000060?ooo`030000003oool0oooo06<0oooo1000003>0?ooo`008@3o
ool2000006D0oooo1000003=0?ooo`008@3oool00`000000oooo0?ooo`1V0?ooo`@00000b`3oool0
0240oooo00<000000?ooo`3oool0J03oool400000<T0oooo000Q0?ooo`030000003oool0oooo06T0
oooo100000380?ooo`008@3oool00`000000oooo0?ooo`1[0?ooo`@00000aP3oool00240oooo00<0
00000?ooo`3oool0K03oool400000<D0oooo000Q0?ooo`800000K`3oool400000<<0oooo000Q0?oo
o`030000003oool0oooo0700oooo100000310?ooo`008@3oool00`000000oooo0?ooo`1a0?ooo`@0
0000`03oool00240oooo00<000000?ooo`3oool0L`3oool400000;h0oooo000Q0?ooo`030000003o
ool0oooo07@0oooo1@00002l0?ooo`008@3oool00`000000oooo0?ooo`1f0?ooo`@00000^`3oool0
0240oooo0P00001i0?ooo`@00000^@3oool00240oooo00<000000?ooo`3oool0N@3oool500000;L0
oooo000Q0?ooo`030000003oool0oooo07/0oooo1000002f0?ooo`008@3oool00`000000oooo0?oo
o`1m0?ooo`@00000]03oool00240oooo00<000000?ooo`3oool0OP3oool500000;80oooo000Q0?oo
o`030000003oool0oooo0800oooo1000002a0?ooo`008@3oool2000008<0oooo1000002_0?ooo`00
8@3oool00`000000oooo0?ooo`230?ooo`D00000[@3oool00240oooo00<000000?ooo`3oool0Q@3o
ool400000:`0oooo000Q0?ooo`030000003oool0oooo08L0oooo1000002Z0?ooo`008@3oool00`00
0000oooo0?ooo`280?ooo`H00000Y`3oool001<0oooo0P0000040?ooo`8000001P3oool00`000000
oooo0?ooo`2:0?ooo`@00000Y`3oool00180oooo00@000000?ooo`3oool000000P3oool010000000
oooo0?ooo`0000050?ooo`<00000S03oool200000:L0oooo000E0?ooo`040000003oool0oooo0000
0080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`2<0?ooo`<00000YP3oool001<0
oooo0P0000030?ooo`040000003oool0oooo000000D0oooo00<000000?ooo`3oool0S@3oool50000
0:<0oooo000E0?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`2?0?ooo`D00000X@3oool00180oooo00@000000?ooo`3oool000000P3oool01000
0000oooo0?ooo`0000050?ooo`030000003oool0oooo0940oooo1000002P0?ooo`004`3oool20000
00@0oooo0P0000060?ooo`030000003oool0oooo09<0oooo1000002N0?ooo`008@3oool2000009D0
oooo1@00002L0?ooo`008@3oool00`000000oooo0?ooo`2F0?ooo`@00000V`3oool00240oooo00<0
00000?ooo`3oool0V03oool4000009T0oooo000Q0?ooo`030000003oool0oooo09T0oooo1000002H
0?ooo`008@3oool00`000000oooo0?ooo`2K0?ooo`@00000UP3oool00240oooo00<000000?ooo`3o
ool0W03oool5000009@0oooo000Q0?ooo`800000W`3oool4000009<0oooo000Q0?ooo`030000003o
ool0oooo0:00oooo1000002A0?ooo`008@3oool00`000000oooo0?ooo`2Q0?ooo`D00000S`3oool0
0240oooo00<000000?ooo`3oool0X`3oool4000008h0oooo000Q0?ooo`030000003oool0oooo0:D0
oooo1000002<0?ooo`008@3oool00`000000oooo0?ooo`2V0?ooo`D00000RP3oool00240oooo0P00
002Y0?ooo`@00000R@3oool00240oooo00<000000?ooo`3oool0ZP3oool4000008L0oooo000Q0?oo
o`030000003oool0oooo0:/0oooo100000260?ooo`008@3oool00`000000oooo0?ooo`2]0?ooo`@0
0000Q03oool00240oooo00<000000?ooo`3oool0[P3oool500000880oooo000Q0?ooo`030000003o
ool0oooo0;00oooo100000210?ooo`008@3oool200000;<0oooo1000001o0?ooo`008@3oool00`00
0000oooo0?ooo`2c0?ooo`D00000O@3oool00240oooo00<000000?ooo`3oool0]@3oool4000007`0
oooo000Q0?ooo`030000003oool0oooo0;L0oooo1000001j0?ooo`00503oool00`000000oooo0?oo
o`020?ooo`8000001P3oool00`000000oooo0?ooo`2h0?ooo`@00000N@3oool001@0oooo00D00000
0?ooo`3oool0oooo000000020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0^P3o
ool4000007L0oooo000A0?ooo`D000000P3oool010000000oooo0?ooo`0000050?ooo`<00000^`3o
ool5000007D0oooo000A0?ooo`040000003oool0oooo000000<0oooo00@000000?ooo`3oool00000
1@3oool00`000000oooo0?ooo`2m0?ooo`D00000L`3oool00180oooo00<000000?ooo`0000000`3o
ool010000000oooo0?ooo`0000050?ooo`030000003oool0oooo0;l0oooo1000001b0?ooo`004`3o
ool2000000<0oooo00@000000?ooo`3oool000001@3oool00`000000oooo0?ooo`310?ooo`<00000
L@3oool001@0oooo00<000000?ooo`3oool00P3oool2000000H0oooo00<000000?ooo`3oool0`P3o
ool4000006l0oooo000Q0?ooo`800000a03oool5000006d0oooo000Q0?ooo`030000003oool0oooo
0<D0oooo1000001/0?ooo`008@3oool00`000000oooo0?ooo`370?ooo`@00000JP3oool00240oooo
00<000000?ooo`3oool0b03oool5000006P0oooo000Q0?ooo`030000003oool0oooo0<X0oooo1000
001W0?ooo`008@3oool00`000000oooo0?ooo`3<0?ooo`@00000I@3oool00240oooo0P00003>0?oo
o`@00000I03oool00240oooo00<000000?ooo`3oool0c`3oool400000680oooo000Q0?ooo`030000
003oool0oooo0=00oooo1@00001P0?ooo`008@3oool00`000000oooo0?ooo`3B0?ooo`@00000G`3o
ool00240oooo00<000000?ooo`3oool0e03oool4000005d0oooo000Q0?ooo`030000003oool0oooo
0=D0oooo1@00001K0?ooo`008@3oool200000=P0oooo1000001J0?ooo`008@3oool00`000000oooo
0?ooo`3I0?ooo`@00000F03oool00240oooo00<000000?ooo`3oool0fP3oool4000005L0oooo000Q
0?ooo`030000003oool0oooo0=`0oooo1000001E0?ooo`008@3oool00`000000oooo0?ooo`3N0?oo
o`@00000D`3oool00240oooo00<000000?ooo`3oool0g`3oool400000580oooo000Q0?ooo`800000
hP3oool400000500oooo000Q0?ooo`030000003oool0oooo0>80oooo1@00001>0?ooo`008@3oool0
0`000000oooo0?ooo`3T0?ooo`D00000C03oool00240oooo00<000000?ooo`3oool0iP3oool30000
04`0oooo000Q0?ooo`030000003oool0oooo0>L0oooo1000001:0?ooo`004P3oool3000000@0oooo
0P0000060?ooo`030000003oool0oooo0>P0oooo1@0000180?ooo`005@3oool010000000oooo0?oo
o`0000020?ooo`030000003oool0oooo00<0oooo0`00003Z0?ooo`@00000A`3oool001D0oooo00@0
00000?ooo`3oool000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0>`0oooo
100000150?ooo`004P3oool3000000<0oooo00@000000?ooo`3oool000001@3oool00`000000oooo
0?ooo`3^0?ooo`@00000@`3oool00180oooo00<000000?ooo`3oool00`3oool010000000oooo0?oo
o`0000050?ooo`030000003oool0oooo0>l0oooo100000120?ooo`004P3oool00`000000oooo0?oo
o`030?ooo`040000003oool0oooo000000D0oooo00<000000?ooo`3oool0l@3oool400000400oooo
000B0?ooo`@000000`3oool2000000H0oooo00<000000?ooo`3oool0lP3oool5000003h0oooo000Q
0?ooo`800000m@3oool4000003d0oooo000Q0?ooo`030000003oool0oooo0?H0oooo1000000k0?oo
o`008@3oool00`000000oooo0?ooo`3g0?ooo`@00000>P3oool00240oooo00<000000?ooo`3oool0
n@3oool4000003P0oooo000Q0?ooo`030000003oool0oooo0?/0oooo1000000f0?ooo`008@3oool0
0`000000oooo0?ooo`3l0?ooo`@00000=@3oool00240oooo0P00003o0?ooo`@00000<`3oool00240
oooo00<000000?ooo`3oool0o`3oool400000380oooo000Q0?ooo`030000003oool0oooo0?l0oooo
0P3oool400000300oooo000Q0?ooo`030000003oool0oooo0?l0oooo103oool4000002h0oooo000Q
0?ooo`030000003oool0oooo0?l0oooo1@3oool4000002d0oooo000Q0?ooo`030000003oool0oooo
0?l0oooo1`3oool4000002/0oooo000Q0?ooo`800000o`3oool:0?ooo`@00000:@3oool00240oooo
00<000000?ooo`3oool0o`3oool:0?ooo`@00000:03oool00240oooo00<000000?ooo`3oool0o`3o
ool<0?ooo`@000009P3oool00240oooo00<000000?ooo`3oool0o`3oool=0?ooo`D00000903oool0
0240oooo00<000000?ooo`3oool0o`3oool?0?ooo`@000008`3oool00240oooo00<000000?ooo`3o
ool0o`3ooolA0?ooo`@000008@3oool00240oooo0P00003o0?oooa<0oooo1@00000O0?ooo`008@3o
ool00`000000oooo0?ooo`3o0?oooa@0oooo1000000N0?ooo`008@3oool00`000000oooo0?ooo`3o
0?oooaH0oooo1000000L0?ooo`008@3oool00`000000oooo0?ooo`3o0?oooaL0oooo1@00000J0?oo
o`004`3oool2000000@0oooo0P0000060?ooo`030000003oool0oooo0?l0oooo6@3oool5000001P0
oooo000B0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool000001@3oool00`00
0000oooo0?ooo`3o0?oooa/0oooo0`00000H0?ooo`004P3oool010000000oooo0?ooo`0000020?oo
o`040000003oool0oooo000000D0oooo0`00003o0?oooad0oooo00<000000?ooo`3oool05P3oool0
0180oooo0`0000030?ooo`040000003oool0oooo000000D0oooo00<000000?ooo`3oool0o`3ooolf
0?ooo`004P3oool00`000000oooo0?ooo`030?ooo`040000003oool0oooo000000D0oooo00<00000
0?ooo`3oool0o`3ooolf0?ooo`004P3oool00`000000oooo0?ooo`030?ooo`040000003oool0oooo
000000D0oooo00<000000?ooo`3oool0o`3ooolf0?ooo`004`3oool3000000<0oooo0P0000060?oo
o`030000003oool0oooo0?l0oooo=P3oool00?l0ooooFP3oool00?l0ooooFP3oool00?l0ooooFP3o
ool00?l0ooooFP3oool00?l0ooooFP3oool00?l0ooooFP3oool00<H0oooo0`00002@0?ooo`00b@3o
ool00`000000oooo0?ooo`2=0?ooo`00P`3oool300000080oooo00<000000?ooo`3oool00P000000
0`3oool000000?ooo`020?ooo`8000000P3oool500000080oooo0P0000020?ooo`<0000000<0oooo
0000000000000P3oool2000000030?ooo`000000oooo00D000001P3oool200000080oooo1@000000
0`3oool00000000000020?ooo`<000000P3oool500000080oooo0`0000020?ooo`<00000OP3oool0
08@0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`0000030?ooo`040000003o
ool0oooo000000<0oooo00D000000?ooo`3oool0oooo000000020?ooo`040000003oool0oooo0000
0080oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00L0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`0000020?ooo`040000
003oool0oooo000000@0oooo00D000000?ooo`3oool0oooo000000080?ooo`030000003oool0oooo
07/0oooo000O0?ooo`<00000HP3oool3000000@0oooo0`0000030?ooo`030000003oool0oooo00@0
oooo00D000000?ooo`3oool0oooo000000020?ooo`040000003oool0oooo00000080oooo00@00000
0?ooo`3oool0oooo0`0000040?ooo`030000003oool0oooo00L0oooo100000020?ooo`040000003o
ool0oooo00000080oooo00@000000?ooo`3oool00000103oool010000000oooo0?ooo`3oool30000
00@0oooo0P00001n0?ooo`008P3oool00`000000oooo0?ooo`1O0?ooo`040000003oool0oooo0000
00D0oooo00D000000?ooo`3oool0oooo000000060?ooo`050000003oool0oooo0?ooo`0000000P3o
ool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`070?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3o
ool000000P3oool010000000oooo0?ooo`0000040?ooo`050000003oool0oooo0?ooo`0000000P3o
ool010000000oooo0?ooo`0000200?ooo`004P3oool200000080oooo1@0000000`3oool000000000
00020?ooo`<000000P3oool500000080oooo0`00001E0?ooo`040000003oool0oooo000000<0oooo
0P0000030?ooo`@000000`3oool2000000@0oooo0P0000020?ooo`040000003oool00000000000@0
oooo0P0000050?ooo`030000003oool0oooo00P0oooo00H000000?ooo`000000oooo0000003oool2
000000@0oooo0P0000000`3oool000000?ooo`020?ooo`030000003oool0oooo0080oooo0P000004
0?ooo`<00000O@3oool001<0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`00
00020?ooo`040000003oool0oooo000000@0oooo00D000000?ooo`3oool0oooo0000001G0?ooo`@0
00002P3oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo00L0oooo0`00000A0?ooo`03
0000003oool0oooo08P0oooo000C0?ooo`@000000P3oool010000000oooo0?ooo`0000020?ooo`04
0000003oool0oooo000000@0oooo00@000000?ooo`3oool0oooo0`00001S0?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool0503oool2000001`0oooo0P00002:0?ooo`004`3oool01000
0000oooo0?ooo`0000020?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool00000
103oool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`3oool0o`3ooolY0?ooo`00
503oool01P000000oooo0000003oool000000?ooo`800000103oool2000000030?ooo`000000oooo
0080oooo00<000000?ooo`3oool00P3oool200000?l0oooo;03oool00001\
\>"],
  ImageRangeCache->{{{0, 344}, {212.125, 0}} -> {-427.041, -4.31523, 12.8402, 
  0.337911}}],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .61803 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0238095 0.015873 0.0147151 0.00588692 [
[.18254 .00222 -6 -9 ]
[.18254 .00222 6 0 ]
[.34127 .00222 -6 -9 ]
[.34127 .00222 6 0 ]
[.5 .00222 -6 -9 ]
[.5 .00222 6 0 ]
[.65873 .00222 -6 -9 ]
[.65873 .00222 6 0 ]
[.81746 .00222 -6 -9 ]
[.81746 .00222 6 0 ]
[.97619 .00222 -6 -9 ]
[.97619 .00222 6 0 ]
[1.025 .01472 0 -4.90625 ]
[1.025 .01472 34 4.90625 ]
[.01131 .13245 -12 -4.5 ]
[.01131 .13245 0 4.5 ]
[.01131 .25019 -12 -4.5 ]
[.01131 .25019 0 4.5 ]
[.01131 .36793 -12 -4.5 ]
[.01131 .36793 0 4.5 ]
[.01131 .48567 -12 -4.5 ]
[.01131 .48567 0 4.5 ]
[.01131 .60341 -18 -4.5 ]
[.01131 .60341 0 4.5 ]
[.02381 .64303 -26 0 ]
[.02381 .64303 26 10.125 ]
[.5 .63053 -77 0 ]
[.5 .63053 77 10.125 ]
[ 0 0 0 0 ]
[ 1 .61803 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.18254 .01472 m
.18254 .02097 L
s
[(10)] .18254 .00222 0 1 Mshowa
.34127 .01472 m
.34127 .02097 L
s
[(20)] .34127 .00222 0 1 Mshowa
.5 .01472 m
.5 .02097 L
s
[(30)] .5 .00222 0 1 Mshowa
.65873 .01472 m
.65873 .02097 L
s
[(40)] .65873 .00222 0 1 Mshowa
.81746 .01472 m
.81746 .02097 L
s
[(50)] .81746 .00222 0 1 Mshowa
.97619 .01472 m
.97619 .02097 L
s
[(60)] .97619 .00222 0 1 Mshowa
.125 Mabswid
.05556 .01472 m
.05556 .01847 L
s
.0873 .01472 m
.0873 .01847 L
s
.11905 .01472 m
.11905 .01847 L
s
.15079 .01472 m
.15079 .01847 L
s
.21429 .01472 m
.21429 .01847 L
s
.24603 .01472 m
.24603 .01847 L
s
.27778 .01472 m
.27778 .01847 L
s
.30952 .01472 m
.30952 .01847 L
s
.37302 .01472 m
.37302 .01847 L
s
.40476 .01472 m
.40476 .01847 L
s
.43651 .01472 m
.43651 .01847 L
s
.46825 .01472 m
.46825 .01847 L
s
.53175 .01472 m
.53175 .01847 L
s
.56349 .01472 m
.56349 .01847 L
s
.59524 .01472 m
.59524 .01847 L
s
.62698 .01472 m
.62698 .01847 L
s
.69048 .01472 m
.69048 .01847 L
s
.72222 .01472 m
.72222 .01847 L
s
.75397 .01472 m
.75397 .01847 L
s
.78571 .01472 m
.78571 .01847 L
s
.84921 .01472 m
.84921 .01847 L
s
.88095 .01472 m
.88095 .01847 L
s
.9127 .01472 m
.9127 .01847 L
s
.94444 .01472 m
.94444 .01847 L
s
.25 Mabswid
0 .01472 m
1 .01472 L
s
gsave
1.025 .01472 -61 -8.90625 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 13.812500 moveto
 558.000000 13.812500 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 13.812500 lineto
closepath
clip newpath
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Angle) show
93.000000 11.250000 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
.02381 .13245 m
.03006 .13245 L
s
[(20)] .01131 .13245 1 0 Mshowa
.02381 .25019 m
.03006 .25019 L
s
[(40)] .01131 .25019 1 0 Mshowa
.02381 .36793 m
.03006 .36793 L
s
[(60)] .01131 .36793 1 0 Mshowa
.02381 .48567 m
.03006 .48567 L
s
[(80)] .01131 .48567 1 0 Mshowa
.02381 .60341 m
.03006 .60341 L
s
[(100)] .01131 .60341 1 0 Mshowa
.125 Mabswid
.02381 .04415 m
.02756 .04415 L
s
.02381 .07358 m
.02756 .07358 L
s
.02381 .10302 m
.02756 .10302 L
s
.02381 .16189 m
.02756 .16189 L
s
.02381 .19132 m
.02756 .19132 L
s
.02381 .22076 m
.02756 .22076 L
s
.02381 .27963 m
.02756 .27963 L
s
.02381 .30906 m
.02756 .30906 L
s
.02381 .3385 m
.02756 .3385 L
s
.02381 .39736 m
.02756 .39736 L
s
.02381 .4268 m
.02756 .4268 L
s
.02381 .45623 m
.02756 .45623 L
s
.02381 .5151 m
.02756 .5151 L
s
.02381 .54454 m
.02756 .54454 L
s
.02381 .57397 m
.02756 .57397 L
s
.25 Mabswid
.02381 0 m
.02381 .61803 L
s
gsave
.02381 .64303 -87 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.125000 moveto
 558.000000 14.125000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.125000 lineto
closepath
clip newpath
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Distance) show
111.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
gsave
.5 .63053 -138 -4 Mabsadd m
1 1 Mabs scale
currentpoint translate
0 20 translate 1 -1 scale
gsave
0.000000 0.000000 0.000000 setrgbcolor
1.000000 setlinewidth
gsave
newpath
 61.000000 14.125000 moveto
 558.000000 14.125000 lineto
 558.000000 4.000000 lineto
 61.000000 4.000000 lineto
 61.000000 14.125000 lineto
closepath
clip newpath
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
63.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Angle) show
99.000000 11.562500 moveto
(vs) show
111.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(.) show
123.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
(Repeat) show
165.000000 11.562500 moveto
(Distance) show
213.000000 11.562500 moveto
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10.000000 scalefont
[1 0 0 -1 0 0 ] makefont setfont
0.000000 0.000000 0.000000 setrgbcolor
0.000000 0.000000 rmoveto
1.000000 setlinewidth
grestore
grestore
%%DocumentNeededResources: font Courier
%%DocumentSuppliedResources:
%%DocumentNeededFonts: Courier
%%DocumentSuppliedFonts:
%%DocumentFonts: font Courier
grestore
0 0 m
1 0 L
1 .61803 L
0 .61803 L
closepath
clip
newpath
.008 w
.02381 .02491 Mdot
.5 .0206 Mdot
.19672 .03029 Mdot
.97619 .02491 Mdot
.12836 .04038 Mdot
.2794 .03594 Mdot
.80328 .03029 Mdot
.09865 .05052 Mdot
.32709 .04169 Mdot
.7206 .03594 Mdot
.08207 .06069 Mdot
.15418 .05592 Mdot
.24441 .05148 Mdot
.35796 .04749 Mdot
.67291 .04169 Mdot
.87164 .04038 Mdot
.0715 .07087 Mdot
.37954 .05332 Mdot
.64204 .04749 Mdot
.06417 .08106 Mdot
.11105 .07618 Mdot
.16585 .07149 Mdot
.23022 .06704 Mdot
.30609 .0629 Mdot
.39545 .05916 Mdot
.62046 .05332 Mdot
.75559 .05148 Mdot
.90135 .05052 Mdot
.0588 .09125 Mdot
.14427 .08158 Mdot
.25722 .07269 Mdot
.40766 .06501 Mdot
.60455 .05916 Mdot
.84582 .05592 Mdot
.05468 .10143 Mdot
.08933 .0965 Mdot
.17249 .08705 Mdot
.22254 .08261 Mdot
.34402 .07446 Mdot
.41732 .07087 Mdot
.59234 .06501 Mdot
.69391 .0629 Mdot
.91793 .06069 Mdot
.05144 .11163 Mdot
.11615 .10183 Mdot
.29793 .08412 Mdot
.42516 .07674 Mdot
.58268 .07087 Mdot
.76978 .06704 Mdot
.04881 .12182 Mdot
.07626 .11685 Mdot
.10649 .11198 Mdot
.13986 .10724 Mdot
.17678 .10263 Mdot
.21772 .09818 Mdot
.26316 .09392 Mdot
.31363 .08987 Mdot
.36963 .08609 Mdot
.43164 .08261 Mdot
.57484 .07674 Mdot
.65598 .07446 Mdot
.74278 .07269 Mdot
.83415 .07149 Mdot
.9285 .07087 Mdot
.04663 .13201 Mdot
.16094 .11269 Mdot
.23608 .1038 Mdot
.43709 .08848 Mdot
.56836 .08261 Mdot
.88895 .07618 Mdot
.04481 .1422 Mdot
.06753 .13721 Mdot
.09217 .13231 Mdot
.11893 .12749 Mdot
.14805 .12278 Mdot
.17978 .1182 Mdot
.21442 .11375 Mdot
.25225 .10946 Mdot
.29359 .10534 Mdot
.33876 .10143 Mdot
.38805 .09776 Mdot
.44174 .09435 Mdot
.56291 .08848 Mdot
.63037 .08609 Mdot
.70207 .08412 Mdot
.77746 .08261 Mdot
.85573 .08158 Mdot
.93583 .08106 Mdot
.04325 .1524 Mdot
.08672 .14248 Mdot
.13736 .13289 Mdot
.26659 .11514 Mdot
.34896 .10724 Mdot
.44575 .10023 Mdot
.55826 .09435 Mdot
.68637 .08987 Mdot
.82751 .08705 Mdot
.04191 .16259 Mdot
.06129 .15759 Mdot
.10438 .14779 Mdot
.182 .13377 Mdot
.21201 .12932 Mdot
.31718 .11685 Mdot
.40193 .10946 Mdot
.44924 .1061 Mdot
.55425 .10023 Mdot
.61195 .09776 Mdot
.73684 .09392 Mdot
.9412 .09125 Mdot
.04075 .17279 Mdot
.07806 .16283 Mdot
.12068 .15315 Mdot
.16957 .14383 Mdot
.22588 .13493 Mdot
.2909 .12657 Mdot
.36596 .11887 Mdot
.45231 .11198 Mdot
.55076 .1061 Mdot
.66124 .10143 Mdot
.78228 .09818 Mdot
.91067 .0965 Mdot
.03972 .18298 Mdot
.05661 .17796 Mdot
.07457 .17301 Mdot
.09368 .16811 Mdot
.11404 .16329 Mdot
.13576 .15855 Mdot
.15894 .1539 Mdot
.1837 .14934 Mdot
.21018 .14489 Mdot
.23851 .14056 Mdot
.26883 .13636 Mdot
.30127 .13231 Mdot
.33599 .12841 Mdot
.37311 .12469 Mdot
.41276 .12117 Mdot
.45503 .11786 Mdot
.54769 .11198 Mdot
.59807 .10946 Mdot
.65104 .10724 Mdot
.70641 .10534 Mdot
.76392 .1038 Mdot
.82322 .10263 Mdot
.88385 .10183 Mdot
.94532 .10143 Mdot
.03881 .19318 Mdot
.10826 .17344 Mdot
.14974 .16399 Mdot
.25006 .14622 Mdot
.31068 .13806 Mdot
.45746 .12374 Mdot
.54497 .11786 Mdot
.74775 .10946 Mdot
.86014 .10724 Mdot
.038 .20337 Mdot
.05297 .19835 Mdot
.06878 .19337 Mdot
.08549 .18845 Mdot
.10316 .1836 Mdot
.12188 .17881 Mdot
.14171 .1741 Mdot
.16274 .16946 Mdot
.18505 .16492 Mdot
.20874 .16047 Mdot
.23391 .15612 Mdot
.26065 .15189 Mdot
.28906 .14779 Mdot
.31925 .14383 Mdot
.35132 .14001 Mdot
.38535 .13636 Mdot
.42143 .13289 Mdot
.45964 .12962 Mdot
.54254 .12374 Mdot
.58724 .12117 Mdot
.63404 .11887 Mdot
.68282 .11685 Mdot
.73341 .11514 Mdot
.78558 .11375 Mdot
.83906 .11269 Mdot
.89351 .11198 Mdot
.94856 .11163 Mdot
.03727 .21357 Mdot
.06635 .20356 Mdot
.13464 .18421 Mdot
.17485 .17496 Mdot
.21988 .16607 Mdot
.2704 .15759 Mdot
.39063 .1422 Mdot
.4616 .1355 Mdot
.54036 .12962 Mdot
.62689 .12469 Mdot
.82022 .1182 Mdot
.92374 .11685 Mdot
.03662 .22376 Mdot
.05006 .21873 Mdot
.07901 .20881 Mdot
.09462 .20392 Mdot
.1466 .18965 Mdot
.18615 .18049 Mdot
.20758 .17604 Mdot
.25413 .16744 Mdot
.33428 .15539 Mdot
.36405 .15164 Mdot
.42855 .14463 Mdot
.46339 .14139 Mdot
.5384 .1355 Mdot
.57857 .13289 Mdot
.66401 .12841 Mdot
.7091 .12657 Mdot
.85195 .12278 Mdot
.95119 .12182 Mdot
.03602 .23396 Mdot
.06221 .22393 Mdot
.091 .21409 Mdot
.12276 .20447 Mdot
.15785 .19511 Mdot
.23985 .17733 Mdot
.28773 .16902 Mdot
.34091 .16118 Mdot
.39986 .1539 Mdot
.46501 .14727 Mdot
.53661 .14139 Mdot
.61465 .13636 Mdot
.69873 .13231 Mdot
.78799 .12932 Mdot
.88107 .12749 Mdot
.03548 .24415 Mdot
.04767 .23911 Mdot
.06042 .23412 Mdot
.07377 .22916 Mdot
.08774 .22426 Mdot
.10237 .21941 Mdot
.11772 .21461 Mdot
.13381 .20987 Mdot
.1507 .2052 Mdot
.16843 .2006 Mdot
.18706 .19607 Mdot
.20663 .19162 Mdot
.2272 .18725 Mdot
.24882 .18298 Mdot
.27156 .17881 Mdot
.29547 .17475 Mdot
.32061 .1708 Mdot
.34703 .16698 Mdot
.37478 .16329 Mdot
.40392 .15975 Mdot
.43448 .15637 Mdot
.4665 .15315 Mdot
.53499 .14727 Mdot
.57145 .14463 Mdot
.60937 .1422 Mdot
.64868 .14001 Mdot
.68932 .13806 Mdot
.73117 .13636 Mdot
.77412 .13493 Mdot
.818 .13377 Mdot
.86264 .13289 Mdot
.90783 .13231 Mdot
.95337 .13201 Mdot
.03498 .25435 Mdot
.08477 .23443 Mdot
.11317 .22476 Mdot
.17841 .20611 Mdot
.21593 .19721 Mdot
.30267 .18049 Mdot
.3527 .17279 Mdot
.46787 .15903 Mdot
.5335 .15315 Mdot
.68075 .14383 Mdot
.76149 .14056 Mdot
.93247 .13721 Mdot
.03452 .26455 Mdot
.04568 .2595 Mdot
.05731 .25449 Mdot
.06943 .24953 Mdot
.09526 .23973 Mdot
.10904 .2349 Mdot
.12342 .23013 Mdot
.13846 .22541 Mdot
.17062 .21616 Mdot
.18782 .21164 Mdot
.20583 .20719 Mdot
.22467 .20282 Mdot
.26508 .19434 Mdot
.28673 .19024 Mdot
.30939 .18625 Mdot
.33312 .18236 Mdot
.38395 .17496 Mdot
.41112 .17147 Mdot
.4395 .16811 Mdot
.46913 .16492 Mdot
.53213 .15903 Mdot
.56552 .15637 Mdot
.60014 .1539 Mdot
.63595 .15164 Mdot
.71094 .14779 Mdot
.74994 .14622 Mdot
.78982 .14489 Mdot
.83043 .14383 Mdot
.91328 .14248 Mdot
.95519 .1422 Mdot
.0341 .27474 Mdot
.05594 .26468 Mdot
.0796 .25478 Mdot
.10527 .24506 Mdot
.13318 .23553 Mdot
.16357 .22624 Mdot
.23291 .20845 Mdot
.27246 .20004 Mdot
.31567 .19201 Mdot
.36287 .18442 Mdot
.41433 .17733 Mdot
.47029 .1708 Mdot
.53087 .16492 Mdot
.59608 .15975 Mdot
.66572 .15539 Mdot
.73935 .15189 Mdot
.8163 .14934 Mdot
.89562 .14779 Mdot
.03372 .28494 Mdot
.044 .27989 Mdot
.06579 .2699 Mdot
.07733 .26496 Mdot
.10182 .25522 Mdot
.11482 .25041 Mdot
.14247 .24096 Mdot
.15717 .23632 Mdot
.18847 .22722 Mdot
.20514 .22277 Mdot
.24069 .21409 Mdot
.25963 .20987 Mdot
.30003 .20171 Mdot
.32156 .19778 Mdot
.36745 .19024 Mdot
.39188 .18665 Mdot
.44381 .17986 Mdot
.47137 .17669 Mdot
.52971 .1708 Mdot
.5605 .16811 Mdot
.62522 .16329 Mdot
.65909 .16118 Mdot
.7296 .15759 Mdot
.76609 .15612 Mdot
.84106 .1539 Mdot
.87932 .15315 Mdot
.95675 .1524 Mdot
.03336 .29513 Mdot
.05352 .28507 Mdot
.07523 .27514 Mdot
.12395 .25579 Mdot
.15132 .24641 Mdot
.18097 .23725 Mdot
.21313 .22835 Mdot
.24803 .21975 Mdot
.28593 .21147 Mdot
.37174 .19607 Mdot
.42011 .18905 Mdot
.47237 .18257 Mdot
.52863 .17669 Mdot
.58888 .17147 Mdot
.65297 .16698 Mdot
.79126 .16047 Mdot
.86424 .15855 Mdot
.93871 .15759 Mdot
.03302 .30533 Mdot
.04256 .30028 Mdot
.05244 .29526 Mdot
.06268 .29027 Mdot
.0733 .28532 Mdot
.08431 .28041 Mdot
.09573 .27554 Mdot
.10758 .27071 Mdot
.11989 .26593 Mdot
.13268 .26119 Mdot
.14596 .25651 Mdot
.15976 .25188 Mdot
.17411 .2473 Mdot
.18904 .24279 Mdot
.20456 .23834 Mdot
.2207 .23396 Mdot
.2375 .22965 Mdot
.25498 .22541 Mdot
.27317 .22126 Mdot
.2921 .21719 Mdot
.3118 .21322 Mdot
.33229 .20934 Mdot
.3536 .20556 Mdot
.37576 .2019 Mdot
.39879 .19835 Mdot
.42272 .19492 Mdot
.44755 .19162 Mdot
.47331 .18845 Mdot
.52763 .18257 Mdot
.55619 .17986 Mdot
.58567 .17733 Mdot
.61605 .17496 Mdot
.6473 .17279 Mdot
.67939 .1708 Mdot
.71227 .16902 Mdot
.74587 .16744 Mdot
.78012 .16607 Mdot
.81495 .16492 Mdot
.85026 .16399 Mdot
.88596 .16329 Mdot
.92194 .16283 Mdot
.95809 .16259 Mdot
.03271 .31552 Mdot
.09303 .28571 Mdot
.14103 .26662 Mdot
.16783 .25737 Mdot
.2279 .23958 Mdot
.26157 .23109 Mdot
.33719 .21513 Mdot
.47418 .19434 Mdot
.52669 .18845 Mdot
.70453 .17475 Mdot
.83726 .16946 Mdot
.90632 .16811 Mdot
.03242 .32572 Mdot
.04131 .32067 Mdot
.0505 .31564 Mdot
.06 .31065 Mdot
.06983 .30569 Mdot
.08 .30076 Mdot
.09052 .29587 Mdot
.10141 .29102 Mdot
.11269 .28622 Mdot
.12437 .28145 Mdot
.13648 .27673 Mdot
.14903 .27206 Mdot
.16204 .26744 Mdot
.17553 .26288 Mdot
.18953 .25837 Mdot
.20405 .25392 Mdot
.21911 .24953 Mdot
.23475 .24521 Mdot
.25097 .24096 Mdot
.26782 .23679 Mdot
.2853 .23269 Mdot
.30344 .22868 Mdot
.32227 .22476 Mdot
.34181 .22093 Mdot
.36208 .21719 Mdot
.38309 .21357 Mdot
.40488 .21005 Mdot
.42745 .20665 Mdot
.45082 .20337 Mdot
.475 .20022 Mdot
.52582 .19434 Mdot
.55245 .19162 Mdot
.57989 .18905 Mdot
.60812 .18665 Mdot
.63713 .18442 Mdot
.66688 .18236 Mdot
.69733 .18049 Mdot
.72844 .17881 Mdot
.76015 .17733 Mdot
.79242 .17604 Mdot
.82515 .17496 Mdot
.85829 .1741 Mdot
.89174 .17344 Mdot
.92543 .17301 Mdot
.95925 .17279 Mdot
.03214 .33592 Mdot
.04963 .32583 Mdot
.06827 .31587 Mdot
.08818 .30604 Mdot
.10948 .29637 Mdot
.13227 .28685 Mdot
.1567 .27753 Mdot
.1829 .2684 Mdot
.21104 .2595 Mdot
.24127 .25085 Mdot
.27375 .24249 Mdot
.30867 .23443 Mdot
.34618 .22673 Mdot
.38645 .21941 Mdot
.42961 .21252 Mdot
.47577 .20611 Mdot
.525 .20022 Mdot
.57728 .19492 Mdot
.63255 .19024 Mdot
.69061 .18625 Mdot
.75118 .18298 Mdot
.81385 .18049 Mdot
.87812 .17881 Mdot
.94339 .17796 Mdot
.03189 .34611 Mdot
.04022 .34105 Mdot
.05767 .33102 Mdot
.06682 .32605 Mdot
.08601 .31622 Mdot
.09608 .31135 Mdot
.11724 .30173 Mdot
.15175 .28762 Mdot
.16405 .28301 Mdot
.18996 .27394 Mdot
.2036 .26949 Mdot
.23234 .26077 Mdot
.24748 .25651 Mdot
.29621 .24415 Mdot
.33165 .23632 Mdot
.35032 .23253 Mdot
.38961 .22525 Mdot
.41028 .22176 Mdot
.45371 .21513 Mdot
.47649 .21199 Mdot
.52423 .20611 Mdot
.54918 .20337 Mdot
.60121 .19835 Mdot
.62826 .19607 Mdot
.68433 .19201 Mdot
.71327 .19024 Mdot
.7728 .18725 Mdot
.86536 .18421 Mdot
.89684 .1836 Mdot
.96028 .18298 Mdot
.03165 .35631 Mdot
.04804 .34622 Mdot
.06545 .33624 Mdot
.08398 .32639 Mdot
.1037 .31667 Mdot
.12472 .30711 Mdot
.14716 .29772 Mdot
.17111 .2885 Mdot
.22411 .27071 Mdot
.25341 .26218 Mdot
.28477 .25392 Mdot
.31834 .24596 Mdot
.35424 .23834 Mdot
.39261 .23109 Mdot
.43356 .22426 Mdot
.47718 .21788 Mdot
.52351 .21199 Mdot
.57255 .20665 Mdot
.62424 .2019 Mdot
.67844 .19778 Mdot
.73492 .19434 Mdot
.79337 .19162 Mdot
.8534 .18965 Mdot
.91451 .18845 Mdot
.03142 .3665 Mdot
.03925 .36144 Mdot
.04732 .35641 Mdot
.05562 .3514 Mdot
.07299 .34148 Mdot
.09144 .33168 Mdot
.10109 .32683 Mdot
.12133 .31725 Mdot
.13193 .31251 Mdot
.14288 .30782 Mdot
.1779 .29402 Mdot
.19034 .28952 Mdot
.2032 .28507 Mdot
.21649 .28067 Mdot
.25908 .26785 Mdot
.27423 .26371 Mdot
.2899 .25964 Mdot
.32282 .25173 Mdot
.34011 .2479 Mdot
.37641 .2405 Mdot
.4151 .23348 Mdot
.43537 .23013 Mdot
.45628 .22689 Mdot
.47782 .22376 Mdot
.52282 .21788 Mdot
.54629 .21513 Mdot
.57039 .21252 Mdot
.59512 .21005 Mdot
.6464 .20556 Mdot
.69997 .20171 Mdot
.72754 .20004 Mdot
.78407 .19721 Mdot
.81294 .19607 Mdot
.84215 .19511 Mdot
.93122 .19337 Mdot
.96119 .19318 Mdot
.0312 .3767 Mdot
.06297 .35661 Mdot
.08028 .34674 Mdot
.11816 .32739 Mdot
.13889 .31793 Mdot
.18441 .29955 Mdot
.20942 .29065 Mdot
.26449 .27354 Mdot
.29479 .26538 Mdot
.3615 .24997 Mdot
.39814 .24279 Mdot
.47843 .22965 Mdot
.52218 .22376 Mdot
.61691 .21357 Mdot
.66771 .20934 Mdot
.77533 .20282 Mdot
.83157 .2006 Mdot
.94703 .19835 Mdot
.031 .3869 Mdot
.03839 .38183 Mdot
.04599 .3768 Mdot
.0538 .37178 Mdot
.06184 .3668 Mdot
.0701 .36184 Mdot
.0786 .35692 Mdot
.08735 .35202 Mdot
.09636 .34716 Mdot
.10563 .34233 Mdot
.11518 .33753 Mdot
.12502 .33277 Mdot
.13515 .32805 Mdot
.14559 .32337 Mdot
.15636 .31873 Mdot
.16745 .31414 Mdot
.17889 .30959 Mdot
.19069 .30509 Mdot
.20285 .30064 Mdot
.2154 .29624 Mdot
.22834 .2919 Mdot
.24169 .28762 Mdot
.25546 .28339 Mdot
.26968 .27923 Mdot
.28434 .27514 Mdot
.29946 .27112 Mdot
.31507 .26717 Mdot
.33116 .26329 Mdot
.34776 .2595 Mdot
.36487 .25579 Mdot
.38252 .25217 Mdot
.4007 .24864 Mdot
.41943 .24521 Mdot
.43872 .24188 Mdot
.45858 .23865 Mdot
.479 .23553 Mdot
.52157 .22965 Mdot
.54372 .22689 Mdot
.56644 .22426 Mdot
.58972 .22176 Mdot
.61355 .21941 Mdot
.63792 .21719 Mdot
.66281 .21513 Mdot
.6882 .21322 Mdot
.71407 .21147 Mdot
.74037 .20987 Mdot
.76709 .20845 Mdot
.79417 .20719 Mdot
.82159 .20611 Mdot
.8493 .2052 Mdot
.87724 .20447 Mdot
.90538 .20392 Mdot
.93365 .20356 Mdot
.962 .20337 Mdot
.03081 .39709 Mdot
.04538 .38699 Mdot
.06077 .37699 Mdot
.07702 .3671 Mdot
.0942 .35732 Mdot
.11239 .34768 Mdot
.13165 .33817 Mdot
.15207 .32883 Mdot
.17373 .31964 Mdot
.22114 .30185 Mdot
.24708 .29327 Mdot
.27464 .28494 Mdot
.30393 .27686 Mdot
.33504 .26908 Mdot
.36808 .26162 Mdot
.40313 .25449 Mdot
.44027 .24775 Mdot
.47955 .24142 Mdot
.521 .23553 Mdot
.56463 .23013 Mdot
.61039 .22525 Mdot
.65819 .22093 Mdot
.7079 .21719 Mdot
.75931 .21409 Mdot
.81218 .21164 Mdot
.86619 .20987 Mdot
.92099 .20881 Mdot
.03063 .40729 Mdot
.03763 .40222 Mdot
.05218 .39217 Mdot
.05975 .38718 Mdot
.07552 .37727 Mdot
.08373 .37237 Mdot
.10084 .36264 Mdot
.10976 .35783 Mdot
.13806 .34359 Mdot
.15832 .33429 Mdot
.1689 .32971 Mdot
.19099 .32067 Mdot
.20253 .31622 Mdot
.22666 .30747 Mdot
.23926 .30317 Mdot
.26562 .29476 Mdot
.3082 .28262 Mdot
.32326 .27871 Mdot
.35472 .27112 Mdot
.37114 .26744 Mdot
.40545 .26035 Mdot
.42334 .25694 Mdot
.46064 .25041 Mdot
.48006 .2473 Mdot
.52045 .24142 Mdot
.54142 .23865 Mdot
.5849 .23348 Mdot
.60739 .23109 Mdot
.65382 .22673 Mdot
.67773 .22476 Mdot
.72683 .22126 Mdot
.75197 .21975 Mdot
.82938 .21616 Mdot
.88228 .21461 Mdot
.909 .21409 Mdot
.96273 .21357 Mdot
.03046 .41749 Mdot
.04426 .40738 Mdot
.07411 .38746 Mdot
.09025 .37766 Mdot
.10728 .36798 Mdot
.12527 .35843 Mdot
.16437 .33978 Mdot
.18563 .3307 Mdot
.20813 .3218 Mdot
.23197 .3131 Mdot
.28396 .29637 Mdot
.3123 .28838 Mdot
.34231 .28067 Mdot
.37407 .27327 Mdot
.44314 .2595 Mdot
.48056 .25319 Mdot
.51994 .2473 Mdot
.56128 .24188 Mdot
.64968 .23253 Mdot
.69656 .22868 Mdot
.74502 .22541 Mdot
.79486 .22277 Mdot
.89763 .21941 Mdot
.94994 .21873 Mdot
.03029 .42768 Mdot
.03694 .42262 Mdot
.04375 .41757 Mdot
.05073 .41255 Mdot
.05789 .40755 Mdot
.06523 .40258 Mdot
.07277 .39764 Mdot
.0805 .39272 Mdot
.08844 .38783 Mdot
.09658 .38297 Mdot
.10494 .37813 Mdot
.11353 .37333 Mdot
.12235 .36857 Mdot
.13141 .36383 Mdot
.14072 .35914 Mdot
.15028 .35448 Mdot
.16011 .34986 Mdot
.17021 .34528 Mdot
.18059 .34074 Mdot
.19127 .33624 Mdot
.20225 .33179 Mdot
.21354 .32739 Mdot
.22514 .32303 Mdot
.23708 .31873 Mdot
.24936 .31449 Mdot
.26199 .31029 Mdot
.27499 .30616 Mdot
.28835 .30209 Mdot
.30209 .29808 Mdot
.31623 .29414 Mdot
.33076 .29027 Mdot
.34571 .28647 Mdot
.36107 .28275 Mdot
.37686 .2791 Mdot
.39309 .27554 Mdot
.40977 .27206 Mdot
.42689 .26867 Mdot
.44447 .26538 Mdot
.46252 .26218 Mdot
.48102 .25908 Mdot
.51944 .25319 Mdot
.53936 .25041 Mdot
.55973 .24775 Mdot
.58057 .24521 Mdot
.60186 .24279 Mdot
.62359 .2405 Mdot
.64576 .23834 Mdot
.66835 .23632 Mdot
.69133 .23443 Mdot
.7147 .23269 Mdot
.73843 .23109 Mdot
.7625 .22965 Mdot
.78687 .22835 Mdot
.81153 .22722 Mdot
.83643 .22624 Mdot
.86154 .22541 Mdot
.88683 .22476 Mdot
.91226 .22426 Mdot
.93779 .22393 Mdot
.96338 .22376 Mdot
.03014 .43788 Mdot
.05703 .41774 Mdot
.10273 .38829 Mdot
.11959 .37871 Mdot
.1561 .35994 Mdot
.17586 .35079 Mdot
.21875 .33299 Mdot
.24201 .32438 Mdot
.29256 .30782 Mdot
.31999 .29991 Mdot
.41178 .27792 Mdot
.48147 .26496 Mdot
.51898 .25908 Mdot
.5993 .24864 Mdot
.73218 .23679 Mdot
.7793 .23396 Mdot
.87658 .23013 Mdot
.92623 .22916 Mdot
.02999 .44807 Mdot
.03631 .44301 Mdot
.04279 .43796 Mdot
.04942 .43294 Mdot
.05621 .42793 Mdot
.06317 .42296 Mdot
.07029 .418 Mdot
.0776 .41307 Mdot
.08509 .40817 Mdot
.09277 .4033 Mdot
.10064 .39845 Mdot
.10871 .39363 Mdot
.11699 .38885 Mdot
.12548 .38409 Mdot
.1342 .37937 Mdot
.14314 .37468 Mdot
.15231 .37003 Mdot
.16173 .36542 Mdot
.1714 .36084 Mdot
.18133 .35631 Mdot
.19152 .35182 Mdot
.20199 .34737 Mdot
.21274 .34296 Mdot
.22378 .3386 Mdot
.23512 .33429 Mdot
.24677 .33004 Mdot
.25873 .32583 Mdot
.27102 .32168 Mdot
.28364 .31759 Mdot
.29661 .31356 Mdot
.30993 .30959 Mdot
.32361 .30569 Mdot
.33766 .30185 Mdot
.35209 .29808 Mdot
.3669 .29439 Mdot
.3821 .29077 Mdot
.3977 .28724 Mdot
.41371 .28378 Mdot
.43013 .28041 Mdot
.44696 .27713 Mdot
.46422 .27394 Mdot
.4819 .27085 Mdot
.51853 .26496 Mdot
.53748 .26218 Mdot
.55686 .2595 Mdot
.57666 .25694 Mdot
.59687 .25449 Mdot
.61748 .25217 Mdot
.6385 .24997 Mdot
.65989 .2479 Mdot
.68166 .24596 Mdot
.70379 .24415 Mdot
.72625 .24249 Mdot
.74903 .24096 Mdot
.7721 .23958 Mdot
.79544 .23834 Mdot
.81903 .23725 Mdot
.84283 .23632 Mdot
.86682 .23553 Mdot
.89096 .2349 Mdot
.91523 .23443 Mdot
.93958 .23412 Mdot
.96398 .23396 Mdot
.02984 .45827 Mdot
.04234 .44815 Mdot
.05543 .43812 Mdot
.06915 .42818 Mdot
.08354 .41835 Mdot
.11452 .39899 Mdot
.1312 .3895 Mdot
.14874 .38013 Mdot
.1672 .37091 Mdot
.18662 .36184 Mdot
.20708 .35294 Mdot
.22864 .34422 Mdot
.25135 .3357 Mdot
.27529 .32739 Mdot
.30051 .3193 Mdot
.35508 .30389 Mdot
.38455 .29661 Mdot
.41555 .28964 Mdot
.44813 .28301 Mdot
.4823 .27673 Mdot
.5181 .27085 Mdot
.55553 .26538 Mdot
.59455 .26035 Mdot
.63513 .25579 Mdot
.67718 .25173 Mdot
.76525 .24521 Mdot
.81096 .24279 Mdot
.85753 .24096 Mdot
.90474 .23973 Mdot
.95233 .23911 Mdot
.02971 .46847 Mdot
.03574 .4634 Mdot
.04823 .45332 Mdot
.06806 .43837 Mdot
.07498 .43343 Mdot
.09676 .41877 Mdot
.11218 .40914 Mdot
.12017 .40437 Mdot
.13676 .39491 Mdot
.14536 .39023 Mdot
.16322 .38098 Mdot
.19175 .36739 Mdot
.20175 .36294 Mdot
.23333 .34986 Mdot
.25578 .34137 Mdot
.26744 .33721 Mdot
.29167 .32905 Mdot
.30426 .32505 Mdot
.33043 .31725 Mdot
.37226 .30604 Mdot
.38691 .30245 Mdot
.43309 .29215 Mdot
.46577 .28571 Mdot
.48269 .28262 Mdot
.5177 .27673 Mdot
.53578 .27394 Mdot
.57311 .26867 Mdot
.63192 .26162 Mdot
.65224 .2595 Mdot
.71523 .25392 Mdot
.75873 .25085 Mdot
.78089 .24953 Mdot
.82589 .2473 Mdot
.84868 .24641 Mdot
.89473 .24506 Mdot
.96452 .24415 Mdot
.02958 .47866 Mdot
.04151 .46854 Mdot
.05398 .4585 Mdot
.06702 .44855 Mdot
.08067 .4387 Mdot
.09497 .42894 Mdot
.10996 .41929 Mdot
.12567 .40975 Mdot
.14215 .40034 Mdot
.15946 .39106 Mdot
.17763 .38193 Mdot
.21678 .36413 Mdot
.23787 .3555 Mdot
.26005 .34705 Mdot
.28337 .33882 Mdot
.30788 .3308 Mdot
.33365 .32303 Mdot
.36073 .31552 Mdot
.38917 .30829 Mdot
.41902 .30137 Mdot
.45031 .29476 Mdot
.48306 .2885 Mdot
.51731 .28262 Mdot
.55304 .27713 Mdot
.59023 .27206 Mdot
.62886 .26744 Mdot
.66884 .26329 Mdot
.7101 .25964 Mdot
.75252 .25651 Mdot
.79595 .25392 Mdot
.84024 .25188 Mdot
.88518 .25041 Mdot
.93057 .24953 Mdot
.02946 .48886 Mdot
.03522 .48379 Mdot
.04112 .47874 Mdot
.04714 .47371 Mdot
.0533 .4687 Mdot
.05959 .46371 Mdot
.06603 .45874 Mdot
.07261 .45379 Mdot
.07934 .44887 Mdot
.08622 .44398 Mdot
.09326 .4391 Mdot
.10047 .43426 Mdot
.10784 .42944 Mdot
.11539 .42465 Mdot
.12311 .41989 Mdot
.13102 .41516 Mdot
.13911 .41045 Mdot
.1474 .40579 Mdot
.15589 .40115 Mdot
.16459 .39655 Mdot
.17349 .39198 Mdot
.18262 .38746 Mdot
.19196 .38297 Mdot
.20154 .37852 Mdot
.21135 .37411 Mdot
.2214 .36974 Mdot
.23171 .36542 Mdot
.24227 .36114 Mdot
.25309 .35692 Mdot
.26418 .35274 Mdot
.27555 .34861 Mdot
.2872 .34454 Mdot
.29914 .34052 Mdot
.31137 .33656 Mdot
.32391 .33266 Mdot
.33675 .32883 Mdot
.34991 .32505 Mdot
.3634 .32134 Mdot
.3772 .31771 Mdot
.39135 .31414 Mdot
.40582 .31065 Mdot
.42065 .30723 Mdot
.43581 .30389 Mdot
.45133 .30064 Mdot
.4672 .29747 Mdot
.48342 .29439 Mdot
.51694 .2885 Mdot
.53423 .28571 Mdot
.55187 .28301 Mdot
.56987 .28041 Mdot
.58822 .27792 Mdot
.60691 .27554 Mdot
.62593 .27327 Mdot
.64528 .27112 Mdot
.66496 .26908 Mdot
.68493 .26717 Mdot
.70521 .26538 Mdot
.72577 .26371 Mdot
.74659 .26218 Mdot
.76766 .26077 Mdot
.78896 .2595 Mdot
.81047 .25837 Mdot
.83217 .25737 Mdot
.85404 .25651 Mdot
.87605 .25579 Mdot
.89818 .25522 Mdot
.9204 .25478 Mdot
.94269 .25449 Mdot
.96502 .25435 Mdot
.02934 .49906 Mdot
.05265 .47889 Mdot
.06508 .46892 Mdot
.09163 .44927 Mdot
.10583 .43959 Mdot
.13622 .42057 Mdot
.15251 .41124 Mdot
.18745 .39299 Mdot
.20621 .38409 Mdot
.24652 .3668 Mdot
.26817 .35843 Mdot
.31474 .34233 Mdot
.33974 .33462 Mdot
.39344 .31998 Mdot
.42221 .3131 Mdot
.48376 .30028 Mdot
.51658 .29439 Mdot
.58629 .28378 Mdot
.62314 .2791 Mdot
.70054 .27112 Mdot
.74092 .26785 Mdot
.82447 .26288 Mdot
.86732 .26119 Mdot
.95432 .2595 Mdot
.02922 .50925 Mdot
.03475 .50418 Mdot
.04039 .49913 Mdot
.04615 .49409 Mdot
.05203 .48908 Mdot
.05804 .48408 Mdot
.07044 .47416 Mdot
.07685 .46923 Mdot
.08339 .46432 Mdot
.09008 .45944 Mdot
.09692 .45458 Mdot
.10391 .44975 Mdot
.11836 .44016 Mdot
.12583 .43541 Mdot
.13348 .43069 Mdot
.14129 .426 Mdot
.14929 .42134 Mdot
.15747 .41671 Mdot
.17441 .40755 Mdot
.18318 .40303 Mdot
.19215 .39854 Mdot
.20134 .39409 Mdot
.21074 .38968 Mdot
.22037 .38531 Mdot
.24031 .3767 Mdot
.25064 .37246 Mdot
.26121 .36827 Mdot
.27204 .36413 Mdot
.28312 .36004 Mdot
.29447 .356 Mdot
.31798 .34809 Mdot
.33016 .34422 Mdot
.34262 .34042 Mdot
.35538 .33667 Mdot
.36843 .33299 Mdot
.38179 .32938 Mdot
.40942 .32236 Mdot
.42371 .31896 Mdot
.43832 .31564 Mdot
.45325 .3124 Mdot
.46851 .30924 Mdot
.48409 .30616 Mdot
.51624 .30028 Mdot
.5328 .29747 Mdot
.54969 .29476 Mdot
.56691 .29215 Mdot
.58445 .28964 Mdot
.6023 .28724 Mdot
.63893 .28275 Mdot
.65769 .28067 Mdot
.67674 .27871 Mdot
.69607 .27686 Mdot
.71566 .27514 Mdot
.73551 .27354 Mdot
.77589 .27071 Mdot
.7964 .26949 Mdot
.8171 .2684 Mdot
.83796 .26744 Mdot
.85897 .26662 Mdot
.88011 .26593 Mdot
.92267 .26496 Mdot
.94406 .26468 Mdot
.96548 .26455 Mdot
.02911 .51945 Mdot
.04005 .50932 Mdot
.06331 .4893 Mdot
.07568 .47941 Mdot
.0886 .46961 Mdot
.10208 .45991 Mdot
.13086 .44082 Mdot
.14623 .43144 Mdot
.16231 .42219 Mdot
.17912 .41307 Mdot
.21514 .39528 Mdot
.23443 .38662 Mdot
.25463 .37813 Mdot
.27578 .36984 Mdot
.32112 .35387 Mdot
.3454 .34622 Mdot
.37081 .33882 Mdot
.39739 .33168 Mdot
.45416 .31828 Mdot
.48441 .31205 Mdot
.51591 .30616 Mdot
.54867 .30064 Mdot
.6179 .29077 Mdot
.65429 .28647 Mdot
.6918 .28262 Mdot
.73032 .27923 Mdot
.81004 .27394 Mdot
.85097 .27206 Mdot
.89242 .27071 Mdot
.93421 .2699 Mdot
.02901 .52964 Mdot
.03431 .52457 Mdot
.04524 .51448 Mdot
.05087 .50946 Mdot
.06248 .49948 Mdot
.06846 .49453 Mdot
.08081 .48467 Mdot
.08718 .47978 Mdot
.10032 .47007 Mdot
.10711 .46525 Mdot
.12112 .45568 Mdot
.14331 .44155 Mdot
.15104 .43689 Mdot
.16701 .42768 Mdot
.17526 .42312 Mdot
.19233 .41412 Mdot
.20116 .40967 Mdot
.21941 .40088 Mdot
.22885 .39655 Mdot
.24839 .38801 Mdot
.25849 .38381 Mdot
.29021 .37149 Mdot
.31259 .36354 Mdot
.32416 .35964 Mdot
.34809 .35202 Mdot
.36046 .3483 Mdot
.38604 .34105 Mdot
.39925 .33753 Mdot
.42655 .3307 Mdot
.44064 .32739 Mdot
.46972 .321 Mdot
.48471 .31793 Mdot
.51559 .31205 Mdot
.53149 .30924 Mdot
.56419 .30389 Mdot
.58098 .30137 Mdot
.61545 .29661 Mdot
.6331 .29439 Mdot
.66924 .29027 Mdot
.6877 .28838 Mdot
.72536 .28494 Mdot
.74454 .28339 Mdot
.78351 .28067 Mdot
.8433 .27753 Mdot
.86352 .27673 Mdot
.90427 .27554 Mdot
.92477 .27514 Mdot
.9659 .27474 Mdot
.02891 .53984 Mdot
.0394 .52971 Mdot
.05032 .51965 Mdot
.06168 .50967 Mdot
.0735 .49977 Mdot
.08582 .48995 Mdot
.11203 .4706 Mdot
.12598 .46107 Mdot
.14053 .45166 Mdot
.15573 .44236 Mdot
.17159 .43318 Mdot
.18816 .42414 Mdot
.20547 .41524 Mdot
.22356 .40649 Mdot
.24247 .39791 Mdot
.26224 .3895 Mdot
.28291 .38127 Mdot
.30451 .37324 Mdot
.35068 .35783 Mdot
.37533 .35048 Mdot
.40105 .34338 Mdot
.42789 .33656 Mdot
.45587 .33004 Mdot
.485 .32382 Mdot
.51529 .31793 Mdot
.54675 .3124 Mdot
.57935 .30723 Mdot
.61309 .30245 Mdot
.64791 .29808 Mdot
.68377 .29414 Mdot
.75831 .28762 Mdot
.7968 .28507 Mdot
.83595 .28301 Mdot
.87563 .28145 Mdot
.91569 .28041 Mdot
.956 .27989 Mdot
.02881 .55004 Mdot
.03391 .54496 Mdot
.0391 .53991 Mdot
.0444 .53487 Mdot
.0498 .52985 Mdot
.0553 .52484 Mdot
.06092 .51986 Mdot
.06664 .5149 Mdot
.07248 .50995 Mdot
.07844 .50503 Mdot
.08452 .50013 Mdot
.09072 .49525 Mdot
.09705 .49039 Mdot
.10351 .48556 Mdot
.1101 .48075 Mdot
.11683 .47597 Mdot
.1237 .47121 Mdot
.13072 .46648 Mdot
.13788 .46177 Mdot
.14519 .4571 Mdot
.15266 .45245 Mdot
.1603 .44783 Mdot
.16809 .44325 Mdot
.17605 .4387 Mdot
.18419 .43418 Mdot
.1925 .42969 Mdot
.20099 .42524 Mdot
.20967 .42083 Mdot
.21853 .41645 Mdot
.2276 .41211 Mdot
.23686 .40782 Mdot
.24632 .40356 Mdot
.25599 .39935 Mdot
.26588 .39519 Mdot
.27598 .39106 Mdot
.28631 .38699 Mdot
.29686 .38297 Mdot
.30764 .37899 Mdot
.31867 .37507 Mdot
.32993 .3712 Mdot
.34143 .36739 Mdot
.35319 .36364 Mdot
.3652 .35994 Mdot
.37747 .35631 Mdot
.39 .35274 Mdot
.40279 .34924 Mdot
.41585 .3458 Mdot
.42919 .34243 Mdot
.44279 .33914 Mdot
.45668 .33592 Mdot
.47084 .33277 Mdot
.48528 .32971 Mdot
.515 .32382 Mdot
.53028 .321 Mdot
.54584 .31828 Mdot
.56168 .31564 Mdot
.57779 .3131 Mdot
.59418 .31065 Mdot
.61083 .30829 Mdot
.62774 .30604 Mdot
.64492 .30389 Mdot
.66234 .30185 Mdot
.68001 .29991 Mdot
.69791 .29808 Mdot
.71604 .29637 Mdot
.73438 .29476 Mdot
.75292 .29327 Mdot
.77166 .2919 Mdot
.79058 .29065 Mdot
.80966 .28952 Mdot
.82889 .2885 Mdot
.84825 .28762 Mdot
.86773 .28685 Mdot
.88731 .28622 Mdot
.90697 .28571 Mdot
.9267 .28532 Mdot
.94648 .28507 Mdot
.96628 .28494 Mdot
.02872 .56023 Mdot
.04929 .54004 Mdot
.06018 .53005 Mdot
.08327 .5103 Mdot
.09551 .50056 Mdot
.12152 .48134 Mdot
.13534 .47189 Mdot
.16475 .45332 Mdot
.1804 .44422 Mdot
.21375 .42642 Mdot
.23152 .41774 Mdot
.26941 .40088 Mdot
.2896 .39272 Mdot
.33267 .37699 Mdot
.35562 .36945 Mdot
.40447 .35509 Mdot
.43044 .3483 Mdot
.48555 .33559 Mdot
.51472 .32971 Mdot
.57629 .31896 Mdot
.60865 .31414 Mdot
.67639 .30569 Mdot
.71165 .30209 Mdot
.7846 .29624 Mdot
.8221 .29402 Mdot
.89859 .29102 Mdot
.93732 .29027 Mdot
.02863 .57043 Mdot
.03353 .56535 Mdot
.03853 .5603 Mdot
.04362 .55525 Mdot
.05409 .54522 Mdot
.05948 .54024 Mdot
.06496 .53527 Mdot
.07056 .53032 Mdot
.088 .51559 Mdot
.09404 .51072 Mdot
.1002 .50588 Mdot
.1129 .49626 Mdot
.11944 .49148 Mdot
.12611 .48673 Mdot
.13291 .48201 Mdot
.14695 .47265 Mdot
.16156 .4634 Mdot
.16909 .45882 Mdot
.18464 .44975 Mdot
.19265 .44527 Mdot
.20083 .44082 Mdot
.20919 .4364 Mdot
.22643 .42768 Mdot
.23533 .42338 Mdot
.25369 .4149 Mdot
.27284 .40658 Mdot
.28271 .40249 Mdot
.2928 .39845 Mdot
.30311 .39446 Mdot
.32437 .38662 Mdot
.33534 .38278 Mdot
.34653 .37899 Mdot
.38154 .36798 Mdot
.39369 .36443 Mdot
.40609 .36094 Mdot
.41874 .35752 Mdot
.4448 .35089 Mdot
.45821 .34768 Mdot
.47188 .34454 Mdot
.48581 .34148 Mdot
.51445 .33559 Mdot
.52916 .33277 Mdot
.54413 .33004 Mdot
.55936 .32739 Mdot
.59058 .32236 Mdot
.60656 .31998 Mdot
.6228 .31771 Mdot
.63927 .31552 Mdot
.69007 .30959 Mdot
.70744 .30782 Mdot
.72501 .30616 Mdot
.76074 .30317 Mdot
.77886 .30185 Mdot
.79715 .30064 Mdot
.81559 .29955 Mdot
.85284 .29772 Mdot
.89052 .29637 Mdot
.90948 .29587 Mdot
.94756 .29526 Mdot
.96664 .29513 Mdot
.02854 .58063 Mdot
.03826 .57049 Mdot
.04834 .56042 Mdot
.06965 .5405 Mdot
.08092 .53065 Mdot
.09263 .52089 Mdot
.10479 .51121 Mdot
.11744 .50162 Mdot
.13059 .49214 Mdot
.15851 .47348 Mdot
.17333 .46432 Mdot
.18877 .45529 Mdot
.20484 .44639 Mdot
.22159 .43763 Mdot
.23903 .42902 Mdot
.27616 .41229 Mdot
.29591 .40419 Mdot
.31648 .39628 Mdot
.33791 .38857 Mdot
.36024 .38108 Mdot
.38348 .37382 Mdot
.43281 .36004 Mdot
.45894 .35356 Mdot
.48606 .34737 Mdot
.51419 .34148 Mdot
.54332 .33592 Mdot
.57345 .3307 Mdot
.6366 .32134 Mdot
.66957 .31725 Mdot
.70339 .31356 Mdot
.73801 .31029 Mdot
.77334 .30747 Mdot
.80931 .30509 Mdot
.88276 .30173 Mdot
.92 .30076 Mdot
.95744 .30028 Mdot
.02846 .59082 Mdot
.03319 .58575 Mdot
.0429 .57564 Mdot
.04789 .57062 Mdot
.05814 .56061 Mdot
.06341 .55564 Mdot
.07425 .54575 Mdot
.07981 .54083 Mdot
.09127 .53106 Mdot
.09716 .5262 Mdot
.10928 .51656 Mdot
.11552 .51177 Mdot
.13497 .49755 Mdot
.14858 .4882 Mdot
.15559 .48357 Mdot
.17003 .47439 Mdot
.17747 .46984 Mdot
.19279 .46084 Mdot
.20069 .45639 Mdot
.21696 .44759 Mdot
.22535 .44325 Mdot
.24264 .43467 Mdot
.25155 .43044 Mdot
.26993 .42211 Mdot
.29892 .40993 Mdot
.30899 .40596 Mdot
.32973 .39818 Mdot
.34041 .39436 Mdot
.36244 .3869 Mdot
.37378 .38325 Mdot
.39715 .37613 Mdot
.40917 .37266 Mdot
.43393 .36591 Mdot
.44666 .36264 Mdot
.47285 .35631 Mdot
.4863 .35325 Mdot
.51394 .34737 Mdot
.52812 .34454 Mdot
.55721 .33914 Mdot
.57211 .33656 Mdot
.60261 .33168 Mdot
.61821 .32938 Mdot
.65009 .32505 Mdot
.66635 .32303 Mdot
.69949 .3193 Mdot
.71636 .31759 Mdot
.75064 .31449 Mdot
.76803 .3131 Mdot
.82111 .30959 Mdot
.85712 .30782 Mdot
.87528 .30711 Mdot
.91182 .30604 Mdot
.93017 .30569 Mdot
.96698 .30533 Mdot
.02838 .60102 Mdot
.03775 .59088 Mdot
.04746 .58081 Mdot
.05751 .5708 Mdot
.06794 .56087 Mdot
.07875 .55101 Mdot
.08996 .54122 Mdot
.1016 .53153 Mdot
.11368 .52191 Mdot
.12622 .5124 Mdot
.13925 .50297 Mdot
.15279 .49366 Mdot
.16687 .48445 Mdot
.1815 .47536 Mdot
.21255 .45757 Mdot
.22902 .44887 Mdot
.24616 .44033 Mdot
.26399 .43194 Mdot
.28254 .42372 Mdot
.30185 .41567 Mdot
.32194 .40782 Mdot
.34284 .40016 Mdot
.36457 .39272 Mdot
.38717 .3855 Mdot
.41064 .37852 Mdot
.43502 .37178 Mdot
.46031 .36532 Mdot
.48654 .35914 Mdot
.5137 .35325 Mdot
.54179 .34768 Mdot
.57081 .34243 Mdot
.60075 .33753 Mdot
.63157 .33299 Mdot
.66325 .32883 Mdot
.69574 .32505 Mdot
.72898 .32168 Mdot
.76292 .31873 Mdot
.79747 .31622 Mdot
.83255 .31414 Mdot
.86807 .31251 Mdot
.90392 .31135 Mdot
.94 .31065 Mdot
.03751 .60108 Mdot
.04223 .59603 Mdot
.04704 .591 Mdot
.05193 .58599 Mdot
.05691 .58099 Mdot
.06197 .57601 Mdot
.06713 .57105 Mdot
.07238 .56611 Mdot
.07773 .56119 Mdot
.08317 .55628 Mdot
.08871 .55139 Mdot
.09435 .54653 Mdot
.1001 .54168 Mdot
.10595 .53686 Mdot
.11191 .53206 Mdot
.11798 .52728 Mdot
.12417 .52253 Mdot
.13048 .5178 Mdot
.1369 .51309 Mdot
.14344 .50841 Mdot
.15011 .50375 Mdot
.15691 .49913 Mdot
.16384 .49453 Mdot
.1709 .48995 Mdot
.1781 .48541 Mdot
.18544 .4809 Mdot
.19292 .47642 Mdot
.20055 .47197 Mdot
.20833 .46755 Mdot
.21626 .46317 Mdot
.22435 .45882 Mdot
.23259 .4545 Mdot
.241 .45023 Mdot
.24958 .44599 Mdot
.25832 .44179 Mdot
.26724 .43763 Mdot
.27633 .43351 Mdot
.2856 .42944 Mdot
.29506 .42541 Mdot
.3047 .42142 Mdot
.31453 .41749 Mdot
.32455 .41359 Mdot
.33477 .40975 Mdot
.34519 .40596 Mdot
.35581 .40222 Mdot
.36664 .39854 Mdot
.37768 .39491 Mdot
.38893 .39134 Mdot
.40039 .38783 Mdot
.41206 .38437 Mdot
.42396 .38098 Mdot
.43607 .37766 Mdot
.44841 .3744 Mdot
.46097 .3712 Mdot
.47375 .36808 Mdot
.48676 .36502 Mdot
.51346 .35914 Mdot
.52715 .35631 Mdot
.54106 .35356 Mdot
.5552 .35089 Mdot
.56956 .3483 Mdot
.58415 .3458 Mdot
.59895 .34338 Mdot
.61396 .34105 Mdot
.62919 .33882 Mdot
.64462 .33667 Mdot
.66026 .33462 Mdot
.67609 .33266 Mdot
.69212 .3308 Mdot
.70833 .32905 Mdot
.72471 .32739 Mdot
.74127 .32583 Mdot
.75799 .32438 Mdot
.77486 .32303 Mdot
.79187 .3218 Mdot
.80901 .32067 Mdot
.82627 .31964 Mdot
.84364 .31873 Mdot
.86111 .31793 Mdot
.87867 .31725 Mdot
.8963 .31667 Mdot
.91399 .31622 Mdot
.93173 .31587 Mdot
.9495 .31564 Mdot
.96729 .31552 Mdot
.05632 .59118 Mdot
.07674 .57136 Mdot
.0875 .56157 Mdot
.11021 .54221 Mdot
.1222 .53266 Mdot
.14754 .51386 Mdot
.18929 .48644 Mdot
.20429 .47754 Mdot
.25291 .45166 Mdot
.28857 .43517 Mdot
.30746 .42718 Mdot
.34748 .41177 Mdot
.36865 .40437 Mdot
.41344 .39023 Mdot
.48698 .37091 Mdot
.51324 .36502 Mdot
.59721 .34924 Mdot
.65738 .34042 Mdot
.68863 .33656 Mdot
.75323 .33004 Mdot
.78646 .32739 Mdot
.85441 .32337 Mdot
.95869 .32067 Mdot
.05576 .60137 Mdot
.06064 .59639 Mdot
.0656 .59142 Mdot
.07065 .58647 Mdot
.07579 .58154 Mdot
.08101 .57663 Mdot
.08633 .57174 Mdot
.09175 .56686 Mdot
.09726 .56201 Mdot
.10287 .55717 Mdot
.10858 .55236 Mdot
.11439 .54757 Mdot
.1203 .5428 Mdot
.12633 .53806 Mdot
.13246 .53333 Mdot
.13871 .52863 Mdot
.14507 .52396 Mdot
.15155 .51931 Mdot
.15815 .51469 Mdot
.16488 .51009 Mdot
.17172 .50552 Mdot
.1787 .50098 Mdot
.18581 .49647 Mdot
.19305 .49199 Mdot
.20043 .48754 Mdot
.20794 .48312 Mdot
.21561 .47874 Mdot
.22341 .47439 Mdot
.23137 .47007 Mdot
.23947 .46579 Mdot
.24773 .46154 Mdot
.25615 .45733 Mdot
.26473 .45316 Mdot
.27348 .44903 Mdot
.28239 .44494 Mdot
.29147 .4409 Mdot
.30072 .43689 Mdot
.31015 .43294 Mdot
.31976 .42902 Mdot
.32955 .42516 Mdot
.33953 .42134 Mdot
.3497 .41757 Mdot
.36005 .41386 Mdot
.3706 .41019 Mdot
.38134 .40658 Mdot
.39228 .40303 Mdot
.40343 .39953 Mdot
.41477 .39609 Mdot
.42632 .39272 Mdot
.43808 .3894 Mdot
.45004 .38615 Mdot
.46222 .38297 Mdot
.4746 .37985 Mdot
.48719 .3768 Mdot
.51302 .37091 Mdot
.52625 .36808 Mdot
.53969 .36532 Mdot
.55334 .36264 Mdot
.56719 .36004 Mdot
.58126 .35752 Mdot
.59553 .35509 Mdot
.61 .35274 Mdot
.62467 .35048 Mdot
.63954 .3483 Mdot
.6546 .34622 Mdot
.66984 .34422 Mdot
.68526 .34233 Mdot
.70086 .34052 Mdot
.71663 .33882 Mdot
.73256 .33721 Mdot
.74865 .3357 Mdot
.76488 .33429 Mdot
.78125 .33299 Mdot
.79775 .33179 Mdot
.81437 .3307 Mdot
.8311 .32971 Mdot
.84793 .32883 Mdot
.86485 .32805 Mdot
.88184 .32739 Mdot
.89891 .32683 Mdot
.91602 .32639 Mdot
.93318 .32605 Mdot
.95037 .32583 Mdot
.96758 .32572 Mdot
.06487 .60161 Mdot
.07487 .59172 Mdot
.08521 .58191 Mdot
.09592 .57217 Mdot
.107 .56252 Mdot
.11848 .55294 Mdot
.13037 .54346 Mdot
.1427 .53407 Mdot
.15548 .52478 Mdot
.16873 .51559 Mdot
.18247 .50651 Mdot
.21151 .48871 Mdot
.22686 .48001 Mdot
.24279 .47144 Mdot
.25932 .46301 Mdot
.27648 .45474 Mdot
.29429 .44663 Mdot
.31277 .4387 Mdot
.33195 .43094 Mdot
.35185 .42338 Mdot
.37249 .41602 Mdot
.39389 .40888 Mdot
.41607 .40196 Mdot
.43904 .39528 Mdot
.46281 .38885 Mdot
.4874 .38268 Mdot
.51281 .3768 Mdot
.53903 .3712 Mdot
.56607 .36591 Mdot
.59391 .36094 Mdot
.62253 .35631 Mdot
.65191 .35202 Mdot
.68202 .34809 Mdot
.7128 .34454 Mdot
.74422 .34137 Mdot
.77622 .3386 Mdot
.80873 .33624 Mdot
.84168 .33429 Mdot
.87498 .33277 Mdot
.90856 .33168 Mdot
.94233 .33102 Mdot
.07398 .6019 Mdot
.08413 .59208 Mdot
.10001 .57749 Mdot
.10548 .57267 Mdot
.11672 .56308 Mdot
.12249 .55832 Mdot
.13434 .54887 Mdot
.14042 .54418 Mdot
.1529 .53487 Mdot
.15932 .53025 Mdot
.17926 .51656 Mdot
.19316 .50757 Mdot
.20031 .50312 Mdot
.21499 .49431 Mdot
.23805 .48134 Mdot
.24602 .47709 Mdot
.2624 .4687 Mdot
.27082 .46455 Mdot
.28813 .45639 Mdot
.29703 .45237 Mdot
.31532 .44446 Mdot
.32471 .44057 Mdot
.35395 .42919 Mdot
.37433 .42185 Mdot
.38479 .41826 Mdot
.40629 .41124 Mdot
.43996 .40115 Mdot
.45158 .39791 Mdot
.47539 .39162 Mdot
.4876 .38857 Mdot
.5126 .38268 Mdot
.5254 .37985 Mdot
.55159 .3744 Mdot
.56498 .37178 Mdot
.60631 .36443 Mdot
.6348 .35994 Mdot
.64932 .35783 Mdot
.67888 .35387 Mdot
.72445 .34861 Mdot
.73995 .34705 Mdot
.77136 .34422 Mdot
.78726 .34296 Mdot
.81941 .34074 Mdot
.83563 .33978 Mdot
.86835 .33817 Mdot
.88482 .33753 Mdot
.93455 .33624 Mdot
.96786 .33592 Mdot
.08308 .60226 Mdot
.09337 .5925 Mdot
.10402 .58283 Mdot
.11503 .57323 Mdot
.12642 .56372 Mdot
.13822 .55429 Mdot
.15043 .54496 Mdot
.16308 .53573 Mdot
.17618 .52661 Mdot
.18975 .51759 Mdot
.20382 .50869 Mdot
.21839 .49991 Mdot
.2335 .49126 Mdot
.24917 .48275 Mdot
.26541 .47439 Mdot
.28225 .46617 Mdot
.2997 .45811 Mdot
.3178 .45023 Mdot
.33655 .44252 Mdot
.35598 .435 Mdot
.37611 .42768 Mdot
.39696 .42057 Mdot
.41854 .41368 Mdot
.44086 .40702 Mdot
.46395 .40061 Mdot
.48779 .39446 Mdot
.5124 .38857 Mdot
.53778 .38297 Mdot
.56393 .37766 Mdot
.59083 .37266 Mdot
.61846 .36798 Mdot
.64681 .36364 Mdot
.67584 .35964 Mdot
.70553 .356 Mdot
.73582 .35274 Mdot
.76667 .34986 Mdot
.79801 .34737 Mdot
.82979 .34528 Mdot
.86194 .34359 Mdot
.89437 .34233 Mdot
.92701 .34148 Mdot
.95978 .34105 Mdot
.09734 .59782 Mdot
.10261 .59298 Mdot
.10796 .58817 Mdot
.1134 .58337 Mdot
.12456 .57385 Mdot
.13028 .56912 Mdot
.1361 .56441 Mdot
.14202 .55972 Mdot
.16041 .54581 Mdot
.16676 .54122 Mdot
.17321 .53666 Mdot
.18647 .52762 Mdot
.19327 .52314 Mdot
.2002 .51869 Mdot
.20724 .51427 Mdot
.22172 .50552 Mdot
.22915 .5012 Mdot
.23671 .4969 Mdot
.26022 .48423 Mdot
.26834 .48008 Mdot
.27661 .47597 Mdot
.28502 .47189 Mdot
.30231 .46386 Mdot
.31118 .45991 Mdot
.32021 .456 Mdot
.3294 .45213 Mdot
.34828 .44454 Mdot
.36782 .43714 Mdot
.37785 .43351 Mdot
.39843 .42642 Mdot
.40899 .42296 Mdot
.41972 .41955 Mdot
.43064 .41619 Mdot
.45302 .40967 Mdot
.46449 .40649 Mdot
.47614 .40339 Mdot
.48798 .40034 Mdot
.51221 .39446 Mdot
.52461 .39162 Mdot
.53719 .38885 Mdot
.54996 .38615 Mdot
.57604 .38098 Mdot
.58936 .37852 Mdot
.60285 .37613 Mdot
.61652 .37382 Mdot
.64438 .36945 Mdot
.65857 .36739 Mdot
.68741 .36354 Mdot
.71688 .36004 Mdot
.73183 .35843 Mdot
.74691 .35692 Mdot
.76213 .3555 Mdot
.79292 .35294 Mdot
.80848 .35182 Mdot
.82414 .35079 Mdot
.83989 .34986 Mdot
.88761 .34768 Mdot
.90364 .34716 Mdot
.91972 .34674 Mdot
.95196 .34622 Mdot
.96811 .34611 Mdot
.10124 .60314 Mdot
.11182 .59352 Mdot
.13406 .57453 Mdot
.14575 .56517 Mdot
.17036 .54672 Mdot
.18332 .53766 Mdot
.2106 .51986 Mdot
.22497 .51114 Mdot
.25525 .49409 Mdot
.27121 .48578 Mdot
.30484 .46961 Mdot
.32256 .46177 Mdot
.35989 .44663 Mdot
.42087 .42541 Mdot
.44259 .41877 Mdot
.48816 .40623 Mdot
.51202 .40034 Mdot
.56192 .3894 Mdot
.58794 .38437 Mdot
.67007 .3712 Mdot
.72796 .36413 Mdot
.75773 .36114 Mdot
.81867 .35631 Mdot
.84972 .35448 Mdot
.91265 .35202 Mdot
.94438 .3514 Mdot
.11561 .59889 Mdot
.12102 .59412 Mdot
.12651 .58938 Mdot
.13209 .58465 Mdot
.13777 .57995 Mdot
.14354 .57527 Mdot
.14941 .57062 Mdot
.15538 .56598 Mdot
.16145 .56138 Mdot
.16762 .55679 Mdot
.1739 .55223 Mdot
.18028 .5477 Mdot
.18677 .5432 Mdot
.19338 .53872 Mdot
.20009 .53427 Mdot
.20693 .52985 Mdot
.21388 .52545 Mdot
.22095 .52109 Mdot
.22814 .51676 Mdot
.23546 .51246 Mdot
.24291 .5082 Mdot
.25048 .50397 Mdot
.25819 .49977 Mdot
.26603 .49561 Mdot
.274 .49148 Mdot
.28212 .48739 Mdot
.29038 .48335 Mdot
.29877 .47933 Mdot
.30732 .47536 Mdot
.31601 .47144 Mdot
.32485 .46755 Mdot
.33385 .46371 Mdot
.343 .45991 Mdot
.3523 .45616 Mdot
.36177 .45245 Mdot
.37139 .44879 Mdot
.38118 .44519 Mdot
.39113 .44163 Mdot
.40125 .43812 Mdot
.41153 .43467 Mdot
.42199 .43127 Mdot
.43261 .42793 Mdot
.44341 .42465 Mdot
.45438 .42142 Mdot
.46552 .41826 Mdot
.47684 .41516 Mdot
.48833 .41211 Mdot
.51184 .40623 Mdot
.52386 .40339 Mdot
.53605 .40061 Mdot
.54842 .39791 Mdot
.56096 .39528 Mdot
.57368 .39272 Mdot
.58656 .39023 Mdot
.59961 .38783 Mdot
.61283 .3855 Mdot
.62622 .38325 Mdot
.63976 .38108 Mdot
.65347 .37899 Mdot
.66733 .37699 Mdot
.68133 .37507 Mdot
.69549 .37324 Mdot
.70979 .37149 Mdot
.72422 .36984 Mdot
.73879 .36827 Mdot
.75348 .3668 Mdot
.76829 .36542 Mdot
.78322 .36413 Mdot
.79825 .36294 Mdot
.81338 .36184 Mdot
.8286 .36084 Mdot
.8439 .35994 Mdot
.85928 .35914 Mdot
.87473 .35843 Mdot
.89024 .35783 Mdot
.9058 .35732 Mdot
.9214 .35692 Mdot
.93703 .35661 Mdot
.95268 .35641 Mdot
.96835 .35631 Mdot
.1302 .59478 Mdot
.14141 .58538 Mdot
.153 .57608 Mdot
.16497 .56686 Mdot
.17736 .55775 Mdot
.19016 .54874 Mdot
.2034 .53984 Mdot
.21709 .53106 Mdot
.23125 .52239 Mdot
.2459 .51386 Mdot
.26105 .50545 Mdot
.27673 .49719 Mdot
.29295 .48908 Mdot
.30973 .48112 Mdot
.34504 .46571 Mdot
.3636 .45827 Mdot
.38278 .45102 Mdot
.4026 .44398 Mdot
.42308 .43714 Mdot
.44421 .43053 Mdot
.46602 .42414 Mdot
.4885 .418 Mdot
.51167 .41211 Mdot
.53551 .40649 Mdot
.56004 .40115 Mdot
.58523 .39609 Mdot
.61107 .39134 Mdot
.63756 .3869 Mdot
.66466 .38278 Mdot
.69236 .37899 Mdot
.74936 .37246 Mdot
.7786 .36974 Mdot
.80825 .36739 Mdot
.83827 .36542 Mdot
.86859 .36383 Mdot
.89916 .36264 Mdot
.9299 .36184 Mdot
.96075 .36144 Mdot
.13935 .59549 Mdot
.14498 .59082 Mdot
.15652 .58154 Mdot
.16242 .57694 Mdot
.17454 .5678 Mdot
.18075 .56327 Mdot
.19347 .55429 Mdot
.2 .54984 Mdot
.21337 .54103 Mdot
.22023 .53666 Mdot
.23429 .52803 Mdot
.24149 .52376 Mdot
.25628 .51531 Mdot
.26386 .51114 Mdot
.28737 .49884 Mdot
.30371 .49083 Mdot
.31209 .48688 Mdot
.32927 .47911 Mdot
.33807 .47529 Mdot
.35613 .46778 Mdot
.36538 .46409 Mdot
.38434 .45686 Mdot
.39405 .45332 Mdot
.41394 .44639 Mdot
.42413 .44301 Mdot
.44499 .4364 Mdot
.45566 .43318 Mdot
.4775 .42693 Mdot
.48867 .42389 Mdot
.5115 .418 Mdot
.52316 .41516 Mdot
.54698 .40967 Mdot
.55914 .40702 Mdot
.58393 .40196 Mdot
.59657 .39953 Mdot
.62232 .39491 Mdot
.63543 .39272 Mdot
.66209 .38857 Mdot
.67563 .38662 Mdot
.70314 .38297 Mdot
.71709 .38127 Mdot
.74537 .37813 Mdot
.75969 .3767 Mdot
.78865 .37411 Mdot
.8328 .37091 Mdot
.84769 .37003 Mdot
.87765 .36857 Mdot
.89272 .36798 Mdot
.92298 .3671 Mdot
.93816 .3668 Mdot
.96858 .3665 Mdot
.14848 .59627 Mdot
.15996 .58702 Mdot
.17182 .57786 Mdot
.18407 .56881 Mdot
.20979 .55101 Mdot
.2233 .54228 Mdot
.23726 .53367 Mdot
.25168 .52518 Mdot
.282 .50862 Mdot
.31439 .49264 Mdot
.33139 .4849 Mdot
.36711 .46992 Mdot
.38585 .4627 Mdot
.4052 .45568 Mdot
.46697 .43591 Mdot
.48883 .42977 Mdot
.51133 .42389 Mdot
.53448 .41826 Mdot
.60772 .40303 Mdot
.63336 .39854 Mdot
.65959 .39436 Mdot
.71369 .38699 Mdot
.74151 .38381 Mdot
.79846 .37852 Mdot
.85686 .37468 Mdot
.88647 .37333 Mdot
.91627 .37237 Mdot
.9462 .37178 Mdot
.15192 .60173 Mdot
.15759 .5971 Mdot
.16335 .5925 Mdot
.1692 .58793 Mdot
.17515 .58337 Mdot
.18119 .57885 Mdot
.18733 .57434 Mdot
.19356 .56987 Mdot
.1999 .56542 Mdot
.20634 .56099 Mdot
.21289 .5566 Mdot
.21955 .55223 Mdot
.22631 .5479 Mdot
.23318 .54359 Mdot
.24017 .53931 Mdot
.24726 .53507 Mdot
.25448 .53085 Mdot
.26181 .52667 Mdot
.26927 .52253 Mdot
.27685 .51842 Mdot
.28455 .51434 Mdot
.29237 .5103 Mdot
.30033 .5063 Mdot
.30841 .50234 Mdot
.31663 .49841 Mdot
.32498 .49453 Mdot
.33347 .49068 Mdot
.34209 .48688 Mdot
.35085 .48312 Mdot
.35976 .47941 Mdot
.3688 .47574 Mdot
.37799 .47212 Mdot
.38733 .46854 Mdot
.39681 .46502 Mdot
.40644 .46154 Mdot
.41623 .45811 Mdot
.42616 .45474 Mdot
.43624 .45142 Mdot
.44648 .44815 Mdot
.45688 .44494 Mdot
.46742 .44179 Mdot
.47813 .4387 Mdot
.48898 .43566 Mdot
.51117 .42977 Mdot
.5225 .42693 Mdot
.53398 .42414 Mdot
.54562 .42142 Mdot
.55741 .41877 Mdot
.56936 .41619 Mdot
.58146 .41368 Mdot
.59371 .41124 Mdot
.60611 .40888 Mdot
.61866 .40658 Mdot
.63135 .40437 Mdot
.64419 .40222 Mdot
.65716 .40016 Mdot
.67027 .39818 Mdot
.68352 .39628 Mdot
.69689 .39446 Mdot
.7104 .39272 Mdot
.72402 .39106 Mdot
.73776 .3895 Mdot
.75161 .38801 Mdot
.76557 .38662 Mdot
.77963 .38531 Mdot
.79379 .38409 Mdot
.80804 .38297 Mdot
.82237 .38193 Mdot
.83678 .38098 Mdot
.85126 .38013 Mdot
.8658 .37937 Mdot
.88041 .37871 Mdot
.89506 .37813 Mdot
.90975 .37766 Mdot
.92448 .37727 Mdot
.93923 .37699 Mdot
.95401 .3768 Mdot
.9688 .3767 Mdot
.16667 .598 Mdot
.19052 .57989 Mdot
.20302 .57099 Mdot
.22925 .55352 Mdot
.24301 .54496 Mdot
.27189 .52823 Mdot
.28703 .52007 Mdot
.31882 .50418 Mdot
.33549 .49647 Mdot
.37045 .48157 Mdot
.38877 .47439 Mdot
.42713 .46061 Mdot
.4472 .45403 Mdot
.48914 .44155 Mdot
.51102 .43566 Mdot
.55659 .42465 Mdot
.58028 .41955 Mdot
.6294 .41019 Mdot
.65481 .40596 Mdot
.7072 .39845 Mdot
.73412 .39519 Mdot
.78926 .38968 Mdot
.81738 .38746 Mdot
.87452 .38409 Mdot
.90342 .38297 Mdot
.96161 .38183 Mdot
.17572 .59895 Mdot
.1816 .59442 Mdot
.18758 .58992 Mdot
.19365 .58544 Mdot
.19982 .58099 Mdot
.20608 .57657 Mdot
.21244 .57217 Mdot
.2189 .5678 Mdot
.22547 .56346 Mdot
.23214 .55915 Mdot
.23891 .55487 Mdot
.2458 .55062 Mdot
.25279 .5464 Mdot
.25989 .54221 Mdot
.26711 .53806 Mdot
.27445 .53393 Mdot
.2819 .52985 Mdot
.28946 .52579 Mdot
.29715 .52178 Mdot
.30496 .5178 Mdot
.3129 .51386 Mdot
.32096 .50995 Mdot
.32915 .50609 Mdot
.33747 .50226 Mdot
.34591 .49848 Mdot
.35449 .49474 Mdot
.36321 .49105 Mdot
.37206 .48739 Mdot
.38105 .48379 Mdot
.39017 .48023 Mdot
.39944 .47672 Mdot
.40884 .47325 Mdot
.41839 .46984 Mdot
.42808 .46648 Mdot
.43792 .46317 Mdot
.4479 .45991 Mdot
.45802 .4567 Mdot
.4683 .45356 Mdot
.47872 .45047 Mdot
.48928 .44743 Mdot
.51086 .44155 Mdot
.52187 .4387 Mdot
.53303 .43591 Mdot
.54434 .43318 Mdot
.55579 .43053 Mdot
.56739 .42793 Mdot
.57913 .42541 Mdot
.59101 .42296 Mdot
.60304 .42057 Mdot
.61521 .41826 Mdot
.62751 .41602 Mdot
.63995 .41386 Mdot
.65252 .41177 Mdot
.66523 .40975 Mdot
.67806 .40782 Mdot
.69101 .40596 Mdot
.70409 .40419 Mdot
.71729 .40249 Mdot
.73059 .40088 Mdot
.74401 .39935 Mdot
.75753 .39791 Mdot
.77115 .39655 Mdot
.78486 .39528 Mdot
.79866 .39409 Mdot
.81255 .39299 Mdot
.82651 .39198 Mdot
.84054 .39106 Mdot
.85464 .39023 Mdot
.8688 .3895 Mdot
.88301 .38885 Mdot
.89727 .38829 Mdot
.91156 .38783 Mdot
.92589 .38746 Mdot
.94025 .38718 Mdot
.95462 .38699 Mdot
.969 .3869 Mdot
.18474 .59995 Mdot
.20907 .58215 Mdot
.22182 .57341 Mdot
.23496 .56479 Mdot
.24852 .55628 Mdot
.26251 .5479 Mdot
.27695 .53964 Mdot
.29184 .53153 Mdot
.3072 .52355 Mdot
.32305 .51573 Mdot
.33939 .50806 Mdot
.35625 .50056 Mdot
.37363 .49322 Mdot
.39154 .48607 Mdot
.41 .47911 Mdot
.42901 .47235 Mdot
.44858 .46579 Mdot
.46872 .45944 Mdot
.48943 .45332 Mdot
.51072 .44743 Mdot
.53258 .44179 Mdot
.55501 .4364 Mdot
.57801 .43127 Mdot
.60157 .42642 Mdot
.62567 .42185 Mdot
.6503 .41757 Mdot
.67545 .41359 Mdot
.70108 .40993 Mdot
.72716 .40658 Mdot
.75368 .40356 Mdot
.78059 .40088 Mdot
.80785 .39854 Mdot
.83541 .39655 Mdot
.86324 .39491 Mdot
.89129 .39363 Mdot
.9195 .39272 Mdot
.94782 .39217 Mdot
.19373 .60102 Mdot
.19973 .59657 Mdot
.21829 .58337 Mdot
.23115 .57472 Mdot
.23773 .57043 Mdot
.25119 .56195 Mdot
.25808 .55775 Mdot
.27218 .54945 Mdot
.29417 .53726 Mdot
.30172 .53327 Mdot
.32509 .5215 Mdot
.34128 .51386 Mdot
.34956 .51009 Mdot
.3665 .50269 Mdot
.37516 .49906 Mdot
.39287 .49192 Mdot
.42045 .48157 Mdot
.42991 .47821 Mdot
.45911 .46847 Mdot
.47928 .46224 Mdot
.48957 .45921 Mdot
.51057 .45332 Mdot
.52128 .45047 Mdot
.54312 .44494 Mdot
.57692 .43714 Mdot
.58847 .43467 Mdot
.62389 .42768 Mdot
.64815 .42338 Mdot
.66047 .42134 Mdot
.68547 .41749 Mdot
.69815 .41567 Mdot
.72384 .41229 Mdot
.76314 .40782 Mdot
.77644 .40649 Mdot
.81682 .40303 Mdot
.84411 .40115 Mdot
.85785 .40034 Mdot
.88548 .39899 Mdot
.89936 .39845 Mdot
.92723 .39764 Mdot
.96919 .39709 Mdot
.20269 .60214 Mdot
.21489 .59334 Mdot
.22747 .58465 Mdot
.24044 .57608 Mdot
.25381 .56762 Mdot
.26759 .55928 Mdot
.2818 .55107 Mdot
.29644 .543 Mdot
.31153 .53507 Mdot
.34312 .51965 Mdot
.35964 .51219 Mdot
.37665 .50489 Mdot
.39418 .49777 Mdot
.41222 .49083 Mdot
.43078 .48408 Mdot
.44988 .47754 Mdot
.46952 .47121 Mdot
.4897 .46509 Mdot
.51043 .45921 Mdot
.5317 .45356 Mdot
.55352 .44815 Mdot
.57587 .44301 Mdot
.59875 .43812 Mdot
.62215 .43351 Mdot
.64605 .42919 Mdot
.67045 .42516 Mdot
.6953 .42142 Mdot
.74631 .4149 Mdot
.7724 .41211 Mdot
.79884 .40967 Mdot
.82559 .40755 Mdot
.8526 .40579 Mdot
.87983 .40437 Mdot
.90723 .4033 Mdot
.93477 .40258 Mdot
.96237 .40222 Mdot
.21161 .60332 Mdot
.22392 .5946 Mdot
.23661 .58599 Mdot
.2431 .58173 Mdot
.24969 .57749 Mdot
.25637 .57329 Mdot
.27005 .56498 Mdot
.27704 .56087 Mdot
.28414 .55679 Mdot
.29135 .55275 Mdot
.29866 .54874 Mdot
.32128 .53693 Mdot
.32904 .53307 Mdot
.33692 .52924 Mdot
.34492 .52545 Mdot
.35304 .52171 Mdot
.36127 .518 Mdot
.37811 .51072 Mdot
.38672 .50714 Mdot
.4043 .50013 Mdot
.41329 .49669 Mdot
.4224 .4933 Mdot
.44101 .48666 Mdot
.45051 .48342 Mdot
.46015 .48023 Mdot
.46991 .47709 Mdot
.47981 .47401 Mdot
.48984 .47098 Mdot
.5103 .46509 Mdot
.52072 .46224 Mdot
.53128 .45944 Mdot
.54198 .4567 Mdot
.5528 .45403 Mdot
.56376 .45142 Mdot
.58606 .44639 Mdot
.5974 .44398 Mdot
.60887 .44163 Mdot
.63218 .43714 Mdot
.64402 .435 Mdot
.66805 .43094 Mdot
.68024 .42902 Mdot
.69254 .42718 Mdot
.70494 .42541 Mdot
.71746 .42372 Mdot
.73007 .42211 Mdot
.76848 .41774 Mdot
.78147 .41645 Mdot
.79453 .41524 Mdot
.80767 .41412 Mdot
.82088 .41307 Mdot
.84749 .41124 Mdot
.86089 .41045 Mdot
.87433 .40975 Mdot
.88782 .40914 Mdot
.91491 .40817 Mdot
.94211 .40755 Mdot
.95574 .40738 Mdot
.96937 .40729 Mdot
.24571 .58738 Mdot
.25889 .57897 Mdot
.28644 .56252 Mdot
.30084 .55449 Mdot
.33095 .53885 Mdot
.34668 .53126 Mdot
.39669 .50946 Mdot
.43248 .49582 Mdot
.45113 .4893 Mdot
.48997 .47687 Mdot
.51016 .47098 Mdot
.5521 .45991 Mdot
.57384 .45474 Mdot
.61882 .44519 Mdot
.68985 .43294 Mdot
.7144 .42944 Mdot
.76467 .42338 Mdot
.79033 .42083 Mdot
.84253 .41671 Mdot
.86898 .41516 Mdot
.9224 .41307 Mdot
.94927 .41255 Mdot
.23555 .60155 Mdot
.24186 .59728 Mdot
.24826 .59304 Mdot
.25476 .58883 Mdot
.26135 .58465 Mdot
.26803 .5805 Mdot
.27482 .57638 Mdot
.2817 .5723 Mdot
.28869 .56824 Mdot
.29578 .56422 Mdot
.30297 .56023 Mdot
.31027 .55628 Mdot
.31768 .55236 Mdot
.32519 .54848 Mdot
.33282 .54464 Mdot
.34055 .54083 Mdot
.3484 .53706 Mdot
.35636 .53333 Mdot
.36443 .52964 Mdot
.37262 .526 Mdot
.38093 .52239 Mdot
.38936 .51883 Mdot
.3979 .51531 Mdot
.40657 .51184 Mdot
.41535 .50841 Mdot
.42426 .50503 Mdot
.43329 .5017 Mdot
.44245 .49841 Mdot
.45173 .49518 Mdot
.46113 .49199 Mdot
.47066 .48886 Mdot
.48031 .48578 Mdot
.49009 .48275 Mdot
.51003 .47687 Mdot
.52019 .47401 Mdot
.53048 .47121 Mdot
.54089 .46847 Mdot
.55142 .46579 Mdot
.56208 .46317 Mdot
.57287 .46061 Mdot
.58377 .45811 Mdot
.5948 .45568 Mdot
.60595 .45332 Mdot
.61722 .45102 Mdot
.62861 .44879 Mdot
.64011 .44663 Mdot
.65172 .44454 Mdot
.66345 .44252 Mdot
.67529 .44057 Mdot
.68723 .4387 Mdot
.69928 .43689 Mdot
.71143 .43517 Mdot
.72367 .43351 Mdot
.73601 .43194 Mdot
.74845 .43044 Mdot
.76097 .42902 Mdot
.77357 .42768 Mdot
.78625 .42642 Mdot
.79901 .42524 Mdot
.81184 .42414 Mdot
.82474 .42312 Mdot
.83769 .42219 Mdot
.85071 .42134 Mdot
.86378 .42057 Mdot
.87689 .41989 Mdot
.89004 .41929 Mdot
.90324 .41877 Mdot
.91646 .41835 Mdot
.92971 .418 Mdot
.94297 .41774 Mdot
.95625 .41757 Mdot
.96954 .41749 Mdot
.25077 .59871 Mdot
.26376 .59034 Mdot
.27713 .58209 Mdot
.30506 .56598 Mdot
.31964 .55813 Mdot
.33464 .55042 Mdot
.35008 .54287 Mdot
.38229 .52823 Mdot
.39909 .52116 Mdot
.41635 .51427 Mdot
.43409 .50757 Mdot
.47102 .49474 Mdot
.49022 .48864 Mdot
.50991 .48275 Mdot
.53009 .47709 Mdot
.57192 .46648 Mdot
.59356 .46154 Mdot
.61566 .45686 Mdot
.63823 .45245 Mdot
.68468 .44446 Mdot
.70853 .4409 Mdot
.73276 .43763 Mdot
.75736 .43467 Mdot
.8075 .42969 Mdot
.83299 .42768 Mdot
.85871 .426 Mdot
.88461 .42465 Mdot
.93683 .42296 Mdot
.96306 .42262 Mdot
.26613 .59603 Mdot
.27271 .5919 Mdot
.28618 .58374 Mdot
.29306 .57971 Mdot
.30711 .57174 Mdot
.31428 .5678 Mdot
.32894 .56004 Mdot
.33643 .55622 Mdot
.35173 .54868 Mdot
.35953 .54496 Mdot
.37548 .53766 Mdot
.38362 .53407 Mdot
.40024 .52701 Mdot
.40872 .52355 Mdot
.42604 .51676 Mdot
.43487 .51344 Mdot
.45288 .50693 Mdot
.46206 .50375 Mdot
.48079 .49755 Mdot
.49034 .49453 Mdot
.50978 .48864 Mdot
.51969 .48578 Mdot
.53985 .48023 Mdot
.55012 .47754 Mdot
.57099 .47235 Mdot
.58161 .46984 Mdot
.60319 .46502 Mdot
.61415 .4627 Mdot
.6364 .45827 Mdot
.6477 .45616 Mdot
.6706 .45213 Mdot
.6822 .45023 Mdot
.70571 .44663 Mdot
.71761 .44494 Mdot
.74168 .44179 Mdot
.75384 .44033 Mdot
.77841 .43763 Mdot
.79081 .4364 Mdot
.81581 .43418 Mdot
.82841 .43318 Mdot
.85377 .43144 Mdot
.86652 .43069 Mdot
.89216 .42944 Mdot
.90503 .42894 Mdot
.93085 .42818 Mdot
.94379 .42793 Mdot
.96971 .42768 Mdot
.26844 .60173 Mdot
.28162 .59352 Mdot
.29517 .58544 Mdot
.30911 .57749 Mdot
.32344 .56968 Mdot
.33818 .56201 Mdot
.35334 .55449 Mdot
.36891 .54712 Mdot
.38492 .53991 Mdot
.40137 .53286 Mdot
.41827 .526 Mdot
.43562 .51931 Mdot
.45344 .51281 Mdot
.47171 .50651 Mdot
.49045 .50041 Mdot
.50966 .49453 Mdot
.52934 .48886 Mdot
.54949 .48342 Mdot
.57009 .47821 Mdot
.59116 .47325 Mdot
.61267 .46854 Mdot
.63462 .46409 Mdot
.657 .45991 Mdot
.67979 .456 Mdot
.70297 .45237 Mdot
.72652 .44903 Mdot
.75042 .44599 Mdot
.77465 .44325 Mdot
.79917 .44082 Mdot
.82395 .4387 Mdot
.84896 .43689 Mdot
.87417 .43541 Mdot
.89953 .43426 Mdot
.92502 .43343 Mdot
.95058 .43294 Mdot
.27721 .60332 Mdot
.2838 .59924 Mdot
.29048 .5952 Mdot
.29725 .59118 Mdot
.30411 .5872 Mdot
.31107 .58325 Mdot
.31813 .57934 Mdot
.32528 .57546 Mdot
.33254 .57161 Mdot
.33989 .5678 Mdot
.34735 .56403 Mdot
.35491 .5603 Mdot
.36257 .5566 Mdot
.37034 .55294 Mdot
.37822 .54932 Mdot
.3862 .54575 Mdot
.39428 .54221 Mdot
.40248 .53872 Mdot
.41079 .53527 Mdot
.4192 .53186 Mdot
.42773 .5285 Mdot
.43637 .52518 Mdot
.44512 .52191 Mdot
.45398 .51869 Mdot
.46296 .51552 Mdot
.47205 .5124 Mdot
.48125 .50932 Mdot
.49057 .5063 Mdot
.50955 .50041 Mdot
.51921 .49755 Mdot
.52898 .49474 Mdot
.53887 .49199 Mdot
.54887 .4893 Mdot
.55899 .48666 Mdot
.56922 .48408 Mdot
.57955 .48157 Mdot
.59 .47911 Mdot
.60056 .47672 Mdot
.61123 .47439 Mdot
.62201 .47212 Mdot
.63289 .46992 Mdot
.64387 .46778 Mdot
.65496 .46571 Mdot
.66615 .46371 Mdot
.67744 .46177 Mdot
.68882 .45991 Mdot
.7003 .45811 Mdot
.71187 .45639 Mdot
.72352 .45474 Mdot
.73527 .45316 Mdot
.74709 .45166 Mdot
.759 .45023 Mdot
.77098 .44887 Mdot
.78304 .44759 Mdot
.79516 .44639 Mdot
.80735 .44527 Mdot
.8196 .44422 Mdot
.83191 .44325 Mdot
.84427 .44236 Mdot
.85669 .44155 Mdot
.86914 .44082 Mdot
.88164 .44016 Mdot
.89417 .43959 Mdot
.90674 .4391 Mdot
.91933 .4387 Mdot
.93194 .43837 Mdot
.94457 .43812 Mdot
.95721 .43796 Mdot
.96986 .43788 Mdot
.29928 .59693 Mdot
.31299 .58901 Mdot
.34157 .5736 Mdot
.35645 .56611 Mdot
.38744 .55159 Mdot
.40356 .54457 Mdot
.45451 .52457 Mdot
.49068 .51219 Mdot
.50943 .5063 Mdot
.54827 .49518 Mdot
.60983 .48023 Mdot
.6312 .47574 Mdot
.67515 .46755 Mdot
.69769 .46386 Mdot
.74385 .45733 Mdot
.76741 .4545 Mdot
.81536 .44975 Mdot
.8397 .44783 Mdot
.91378 .44398 Mdot
.96369 .44301 Mdot
.30803 .59871 Mdot
.31488 .59478 Mdot
.32182 .59088 Mdot
.32886 .58702 Mdot
.34322 .5794 Mdot
.35054 .57564 Mdot
.36549 .56824 Mdot
.38083 .56099 Mdot
.38866 .55743 Mdot
.39659 .55391 Mdot
.40462 .55042 Mdot
.421 .54359 Mdot
.42935 .54024 Mdot
.4378 .53693 Mdot
.44636 .53367 Mdot
.46381 .52728 Mdot
.47269 .52416 Mdot
.48169 .52109 Mdot
.49079 .51807 Mdot
.50932 .51219 Mdot
.51875 .50932 Mdot
.52829 .50651 Mdot
.53794 .50375 Mdot
.55755 .49841 Mdot
.56752 .49582 Mdot
.5776 .4933 Mdot
.58778 .49083 Mdot
.60846 .48607 Mdot
.61895 .48379 Mdot
.62955 .48157 Mdot
.64024 .47941 Mdot
.66193 .47529 Mdot
.68399 .47144 Mdot
.69516 .46961 Mdot
.71775 .46617 Mdot
.72918 .46455 Mdot
.74068 .46301 Mdot
.75227 .46154 Mdot
.77565 .45882 Mdot
.78745 .45757 Mdot
.79931 .45639 Mdot
.81123 .45529 Mdot
.83525 .45332 Mdot
.84734 .45245 Mdot
.85947 .45166 Mdot
.89609 .44975 Mdot
.90837 .44927 Mdot
.92066 .44887 Mdot
.93298 .44855 Mdot
.95766 .44815 Mdot
.97001 .44807 Mdot
.31673 .60055 Mdot
.33059 .5928 Mdot
.34483 .5852 Mdot
.35944 .57774 Mdot
.37445 .57043 Mdot
.38985 .56327 Mdot
.40565 .55628 Mdot
.42187 .54945 Mdot
.43849 .5428 Mdot
.45554 .53633 Mdot
.473 .53005 Mdot
.49089 .52396 Mdot
.50921 .51807 Mdot
.52795 .5124 Mdot
.54712 .50693 Mdot
.56671 .5017 Mdot
.58671 .49669 Mdot
.60713 .49192 Mdot
.62794 .48739 Mdot
.64915 .48312 Mdot
.67073 .47911 Mdot
.69268 .47536 Mdot
.71498 .47189 Mdot
.7376 .4687 Mdot
.76053 .46579 Mdot
.78374 .46317 Mdot
.80721 .46084 Mdot
.83091 .45882 Mdot
.85481 .4571 Mdot
.87888 .45568 Mdot
.90308 .45458 Mdot
.92739 .45379 Mdot
.95177 .45332 Mdot
.32537 .60244 Mdot
.3393 .59478 Mdot
.3464 .591 Mdot
.36089 .58356 Mdot
.36828 .57989 Mdot
.38334 .57267 Mdot
.39102 .56912 Mdot
.40667 .56214 Mdot
.41464 .55871 Mdot
.43089 .55198 Mdot
.43917 .54868 Mdot
.45603 .54221 Mdot
.46462 .53905 Mdot
.4821 .53286 Mdot
.491 .52985 Mdot
.50911 .52396 Mdot
.51831 .52109 Mdot
.53704 .51552 Mdot
.54656 .51281 Mdot
.56591 .50757 Mdot
.57574 .50503 Mdot
.5957 .50013 Mdot
.60582 .49777 Mdot
.62637 .49322 Mdot
.63679 .49105 Mdot
.65791 .48688 Mdot
.66861 .4849 Mdot
.69027 .48112 Mdot
.70123 .47933 Mdot
.72339 .47597 Mdot
.73459 .47439 Mdot
.75721 .47144 Mdot
.76863 .47007 Mdot
.79167 .46755 Mdot
.82667 .46432 Mdot
.83844 .4634 Mdot
.86212 .46177 Mdot
.87402 .46107 Mdot
.89792 .45991 Mdot
.90992 .45944 Mdot
.93397 .45874 Mdot
.94602 .4585 Mdot
.97016 .45827 Mdot
.34795 .59681 Mdot
.36231 .58938 Mdot
.37705 .58209 Mdot
.39216 .57496 Mdot
.42355 .56119 Mdot
.43983 .55455 Mdot
.45652 .54809 Mdot
.47361 .54182 Mdot
.4911 .53573 Mdot
.509 .52985 Mdot
.52731 .52416 Mdot
.54602 .51869 Mdot
.56513 .51344 Mdot
.58465 .50841 Mdot
.62484 .49906 Mdot
.64551 .49474 Mdot
.66653 .49068 Mdot
.68791 .48688 Mdot
.70962 .48335 Mdot
.73166 .48008 Mdot
.75398 .47709 Mdot
.77659 .47439 Mdot
.79945 .47197 Mdot
.82253 .46984 Mdot
.86928 .46648 Mdot
.89289 .46525 Mdot
.91661 .46432 Mdot
.94041 .46371 Mdot
.96426 .4634 Mdot
.34946 .60261 Mdot
.35654 .59889 Mdot
.3637 .5952 Mdot
.37096 .59154 Mdot
.3783 .58793 Mdot
.38574 .58435 Mdot
.39328 .58081 Mdot
.40091 .57731 Mdot
.40863 .57385 Mdot
.41645 .57043 Mdot
.42436 .56705 Mdot
.43237 .56372 Mdot
.44048 .56042 Mdot
.44869 .55717 Mdot
.45699 .55397 Mdot
.4654 .55081 Mdot
.4739 .5477 Mdot
.4825 .54464 Mdot
.4912 .54162 Mdot
.5089 .53573 Mdot
.5179 .53286 Mdot
.527 .53005 Mdot
.53619 .52728 Mdot
.54549 .52457 Mdot
.55488 .52191 Mdot
.56438 .51931 Mdot
.57396 .51676 Mdot
.58365 .51427 Mdot
.59343 .51184 Mdot
.60331 .50946 Mdot
.61328 .50714 Mdot
.62335 .50489 Mdot
.6335 .50269 Mdot
.64375 .50056 Mdot
.65409 .49848 Mdot
.66451 .49647 Mdot
.67502 .49453 Mdot
.68561 .49264 Mdot
.69629 .49083 Mdot
.70705 .48908 Mdot
.71788 .48739 Mdot
.72879 .48578 Mdot
.73978 .48423 Mdot
.75083 .48275 Mdot
.76195 .48134 Mdot
.77314 .48001 Mdot
.78439 .47874 Mdot
.79571 .47754 Mdot
.80708 .47642 Mdot
.8185 .47536 Mdot
.82997 .47439 Mdot
.84149 .47348 Mdot
.85305 .47265 Mdot
.86466 .47189 Mdot
.8763 .47121 Mdot
.88797 .4706 Mdot
.89968 .47007 Mdot
.9114 .46961 Mdot
.92315 .46923 Mdot
.93492 .46892 Mdot
.9467 .4687 Mdot
.95849 .46854 Mdot
.97029 .46847 Mdot
.36507 .60102 Mdot
.39437 .58666 Mdot
.40958 .57971 Mdot
.44112 .5663 Mdot
.4913 .54751 Mdot
.5088 .54162 Mdot
.56363 .52518 Mdot
.6021 .51531 Mdot
.62189 .51072 Mdot
.66253 .50226 Mdot
.68337 .49841 Mdot
.726 .49148 Mdot
.79206 .48312 Mdot
.81456 .4809 Mdot
.88317 .47597 Mdot
.92956 .47416 Mdot
.95286 .47371 Mdot
.37353 .6032 Mdot
.38075 .5996 Mdot
.40293 .58901 Mdot
.41051 .58556 Mdot
.41818 .58215 Mdot
.42594 .57879 Mdot
.43379 .57546 Mdot
.44978 .56893 Mdot
.45791 .56573 Mdot
.46614 .56258 Mdot
.47446 .55947 Mdot
.48288 .55641 Mdot
.49139 .55339 Mdot
.5087 .54751 Mdot
.5175 .54464 Mdot
.52639 .54182 Mdot
.53538 .53905 Mdot
.54446 .53633 Mdot
.55364 .53367 Mdot
.57227 .5285 Mdot
.58173 .526 Mdot
.59128 .52355 Mdot
.60091 .52116 Mdot
.61064 .51883 Mdot
.64036 .51219 Mdot
.65044 .51009 Mdot
.66061 .50806 Mdot
.67085 .50609 Mdot
.68118 .50418 Mdot
.69159 .50234 Mdot
.71263 .49884 Mdot
.72327 .49719 Mdot
.73397 .49561 Mdot
.74475 .49409 Mdot
.7665 .49126 Mdot
.78849 .48871 Mdot
.79957 .48754 Mdot
.81071 .48644 Mdot
.8219 .48541 Mdot
.83313 .48445 Mdot
.84441 .48357 Mdot
.86709 .48201 Mdot
.87848 .48134 Mdot
.8899 .48075 Mdot
.91282 .47978 Mdot
.92432 .47941 Mdot
.94735 .47889 Mdot
.95888 .47874 Mdot
.97042 .47866 Mdot
.3965 .59835 Mdot
.41142 .59142 Mdot
.4267 .58465 Mdot
.44235 .57805 Mdot
.45836 .57161 Mdot
.47474 .56535 Mdot
.49149 .55928 Mdot
.50861 .55339 Mdot
.5261 .5477 Mdot
.54397 .54221 Mdot
.5622 .53693 Mdot
.5808 .53186 Mdot
.59976 .52701 Mdot
.61907 .52239 Mdot
.63873 .518 Mdot
.65872 .51386 Mdot
.67904 .50995 Mdot
.69967 .5063 Mdot
.74181 .49977 Mdot
.76329 .4969 Mdot
.78501 .49431 Mdot
.80695 .49199 Mdot
.8291 .48995 Mdot
.85142 .4882 Mdot
.87389 .48673 Mdot
.89649 .48556 Mdot
.91919 .48467 Mdot
.94196 .48408 Mdot
.96478 .48379 Mdot
.41232 .59728 Mdot
.41984 .59388 Mdot
.43515 .5872 Mdot
.44294 .58392 Mdot
.45879 .57749 Mdot
.46685 .57434 Mdot
.48324 .56818 Mdot
.49158 .56517 Mdot
.50851 .55928 Mdot
.51712 .55641 Mdot
.5346 .55081 Mdot
.54348 .54809 Mdot
.56151 .5428 Mdot
.57065 .54024 Mdot
.58921 .53527 Mdot
.59863 .53286 Mdot
.61771 .52823 Mdot
.62738 .526 Mdot
.64696 .52171 Mdot
.65688 .51965 Mdot
.67695 .51573 Mdot
.6871 .51386 Mdot
.70763 .5103 Mdot
.718 .50862 Mdot
.73895 .50545 Mdot
.74952 .50397 Mdot
.77085 .5012 Mdot
.78161 .49991 Mdot
.81419 .49647 Mdot
.83616 .49453 Mdot
.84721 .49366 Mdot
.86941 .49214 Mdot
.88056 .49148 Mdot
.90295 .49039 Mdot
.91418 .48995 Mdot
.93669 .4893 Mdot
.94797 .48908 Mdot
.97054 .48886 Mdot
.41319 .60314 Mdot
.42819 .59639 Mdot
.44353 .5898 Mdot
.45922 .58337 Mdot
.47526 .57712 Mdot
.49167 .57105 Mdot
.50842 .56517 Mdot
.52554 .55947 Mdot
.54301 .55397 Mdot
.56083 .54868 Mdot
.579 .54359 Mdot
.59752 .53872 Mdot
.61638 .53407 Mdot
.63557 .52964 Mdot
.65508 .52545 Mdot
.67491 .5215 Mdot
.69504 .5178 Mdot
.71545 .51434 Mdot
.73614 .51114 Mdot
.75709 .5082 Mdot
.77828 .50552 Mdot
.79969 .50312 Mdot
.8213 .50098 Mdot
.84309 .49913 Mdot
.86503 .49755 Mdot
.8871 .49626 Mdot
.90928 .49525 Mdot
.93154 .49453 Mdot
.95385 .49409 Mdot
.4289 .60226 Mdot
.43646 .59895 Mdot
.4441 .59567 Mdot
.45182 .59244 Mdot
.46753 .58611 Mdot
.47552 .58301 Mdot
.48359 .57995 Mdot
.49175 .57694 Mdot
.50833 .57105 Mdot
.51676 .56818 Mdot
.52526 .56535 Mdot
.53386 .56258 Mdot
.55131 .55717 Mdot
.56017 .55455 Mdot
.56911 .55198 Mdot
.57813 .54945 Mdot
.59644 .54457 Mdot
.60572 .54221 Mdot
.61508 .53991 Mdot
.62452 .53766 Mdot
.64364 .53333 Mdot
.65332 .53126 Mdot
.66308 .52924 Mdot
.6928 .52355 Mdot
.70285 .52178 Mdot
.71297 .52007 Mdot
.72315 .51842 Mdot
.74372 .51531 Mdot
.7541 .51386 Mdot
.76454 .51246 Mdot
.77503 .51114 Mdot
.79618 .50869 Mdot
.80684 .50757 Mdot
.81753 .50651 Mdot
.82828 .50552 Mdot
.84989 .50375 Mdot
.86075 .50297 Mdot
.88256 .50162 Mdot
.90449 .50056 Mdot
.91548 .50013 Mdot
.9265 .49977 Mdot
.93752 .49948 Mdot
.95961 .49913 Mdot
.97066 .49906 Mdot
.44466 .60155 Mdot
.46005 .59514 Mdot
.49184 .58283 Mdot
.50825 .57694 Mdot
.54209 .56573 Mdot
.55952 .56042 Mdot
.59538 .55042 Mdot
.6138 .54575 Mdot
.6516 .53706 Mdot
.67096 .53307 Mdot
.71054 .52579 Mdot
.73073 .52253 Mdot
.77186 .51676 Mdot
.79276 .51427 Mdot
.83512 .51009 Mdot
.85656 .50841 Mdot
.8998 .50588 Mdot
.92156 .50503 Mdot
.96525 .50418 Mdot
.46045 .60102 Mdot
.46819 .59788 Mdot
.47602 .59478 Mdot
.48393 .59172 Mdot
.49192 .58871 Mdot
.50816 .58283 Mdot
.51641 .57995 Mdot
.52474 .57712 Mdot
.53315 .57434 Mdot
.54164 .57161 Mdot
.55022 .56893 Mdot
.55888 .5663 Mdot
.56763 .56372 Mdot
.57645 .56119 Mdot
.58536 .55871 Mdot
.59435 .55628 Mdot
.60341 .55391 Mdot
.61256 .55159 Mdot
.62178 .54932 Mdot
.63109 .54712 Mdot
.64047 .54496 Mdot
.64992 .54287 Mdot
.65945 .54083 Mdot
.66905 .53885 Mdot
.67872 .53693 Mdot
.68847 .53507 Mdot
.69828 .53327 Mdot
.70816 .53153 Mdot
.7181 .52985 Mdot
.72811 .52823 Mdot
.73819 .52667 Mdot
.74832 .52518 Mdot
.75851 .52376 Mdot
.76875 .52239 Mdot
.77905 .52109 Mdot
.7894 .51986 Mdot
.7998 .51869 Mdot
.81025 .51759 Mdot
.82074 .51656 Mdot
.83127 .51559 Mdot
.84185 .51469 Mdot
.85246 .51386 Mdot
.8631 .51309 Mdot
.87378 .5124 Mdot
.88448 .51177 Mdot
.89521 .51121 Mdot
.90596 .51072 Mdot
.91673 .5103 Mdot
.92752 .50995 Mdot
.93832 .50967 Mdot
.94913 .50946 Mdot
.95995 .50932 Mdot
.97078 .50925 Mdot
.47626 .60066 Mdot
.492 .5946 Mdot
.50808 .58871 Mdot
.52448 .58301 Mdot
.54121 .57749 Mdot
.57564 .56705 Mdot
.59333 .56214 Mdot
.61134 .55743 Mdot
.62966 .55294 Mdot
.64827 .54868 Mdot
.66718 .54464 Mdot
.70583 .53726 Mdot
.72555 .53393 Mdot
.74552 .53085 Mdot
.76571 .52803 Mdot
.78612 .52545 Mdot
.80673 .52314 Mdot
.84845 .51931 Mdot
.86952 .5178 Mdot
.89072 .51656 Mdot
.912 .51559 Mdot
.93336 .5149 Mdot
.95476 .51448 Mdot
.49208 .60049 Mdot
.508 .5946 Mdot
.51607 .59172 Mdot
.53247 .58611 Mdot
.54078 .58337 Mdot
.55765 .57805 Mdot
.56621 .57546 Mdot
.58355 .57043 Mdot
.61015 .56327 Mdot
.61917 .56099 Mdot
.63743 .5566 Mdot
.64666 .55449 Mdot
.66536 .55042 Mdot
.67481 .54848 Mdot
.70356 .543 Mdot
.72305 .53964 Mdot
.73289 .53806 Mdot
.75274 .53507 Mdot
.76274 .53367 Mdot
.78291 .53106 Mdot
.79307 .52985 Mdot
.81353 .52762 Mdot
.82382 .52661 Mdot
.84452 .52478 Mdot
.85493 .52396 Mdot
.87583 .52253 Mdot
.88632 .52191 Mdot
.90737 .52089 Mdot
.93908 .51986 Mdot
.94968 .51965 Mdot
.97089 .51945 Mdot
.50792 .60049 Mdot
.52398 .59478 Mdot
.55706 .58392 Mdot
.57406 .57879 Mdot
.59137 .57385 Mdot
.60898 .56912 Mdot
.64509 .5603 Mdot
.66357 .55622 Mdot
.68232 .55236 Mdot
.70134 .54874 Mdot
.74011 .54221 Mdot
.75983 .53931 Mdot
.77977 .53666 Mdot
.79991 .53427 Mdot
.84068 .53025 Mdot
.86129 .52863 Mdot
.88202 .52728 Mdot
.90284 .5262 Mdot
.9447 .52484 Mdot
.96569 .52457 Mdot
.52374 .60066 Mdot
.53181 .59788 Mdot
.53995 .59514 Mdot
.54818 .59244 Mdot
.55647 .5898 Mdot
.56485 .5872 Mdot
.5733 .58465 Mdot
.58182 .58215 Mdot
.59042 .57971 Mdot
.59909 .57731 Mdot
.60784 .57496 Mdot
.61666 .57267 Mdot
.62555 .57043 Mdot
.63451 .56824 Mdot
.64355 .56611 Mdot
.65265 .56403 Mdot
.66182 .56201 Mdot
.67106 .56004 Mdot
.68036 .55813 Mdot
.68973 .55628 Mdot
.69916 .55449 Mdot
.70865 .55275 Mdot
.7182 .55107 Mdot
.72782 .54945 Mdot
.73749 .5479 Mdot
.74721 .5464 Mdot
.75699 .54496 Mdot
.76682 .54359 Mdot
.7767 .54228 Mdot
.78663 .54103 Mdot
.7966 .53984 Mdot
.80662 .53872 Mdot
.81668 .53766 Mdot
.82679 .53666 Mdot
.83692 .53573 Mdot
.8471 .53487 Mdot
.8573 .53407 Mdot
.86754 .53333 Mdot
.8778 .53266 Mdot
.88809 .53206 Mdot
.8984 .53153 Mdot
.90873 .53106 Mdot
.91908 .53065 Mdot
.92944 .53032 Mdot
.93982 .53005 Mdot
.9502 .52985 Mdot
.9606 .52971 Mdot
.97099 .52964 Mdot
.53955 .60102 Mdot
.5559 .59567 Mdot
.58949 .58556 Mdot
.60672 .58081 Mdot
.66011 .5678 Mdot
.69703 .56023 Mdot
.71586 .55679 Mdot
.7542 .55062 Mdot
.77369 .5479 Mdot
.81323 .5432 Mdot
.83324 .54122 Mdot
.87367 .53806 Mdot
.89405 .53686 Mdot
.93504 .53527 Mdot
.9556 .53487 Mdot
.55534 .60155 Mdot
.56354 .59895 Mdot
.57181 .59639 Mdot
.58016 .59388 Mdot
.58858 .59142 Mdot
.59707 .58901 Mdot
.60563 .58666 Mdot
.61426 .58435 Mdot
.62295 .58209 Mdot
.63172 .57989 Mdot
.64056 .57774 Mdot
.64946 .57564 Mdot
.65843 .5736 Mdot
.66746 .57161 Mdot
.67656 .56968 Mdot
.68572 .5678 Mdot
.69494 .56598 Mdot
.70422 .56422 Mdot
.71356 .56252 Mdot
.72296 .56087 Mdot
.73241 .55928 Mdot
.74192 .55775 Mdot
.75148 .55628 Mdot
.76109 .55487 Mdot
.77075 .55352 Mdot
.78045 .55223 Mdot
.79021 .55101 Mdot
.8 .54984 Mdot
.80984 .54874 Mdot
.81972 .5477 Mdot
.82964 .54672 Mdot
.83959 .54581 Mdot
.84957 .54496 Mdot
.85958 .54418 Mdot
.86963 .54346 Mdot
.8797 .5428 Mdot
.88979 .54221 Mdot
.8999 .54168 Mdot
.91004 .54122 Mdot
.92019 .54083 Mdot
.93035 .5405 Mdot
.94052 .54024 Mdot
.95071 .54004 Mdot
.9609 .53991 Mdot
.97109 .53984 Mdot
.5711 .60226 Mdot
.58768 .59728 Mdot
.6217 .58793 Mdot
.63911 .58356 Mdot
.65678 .5794 Mdot
.67472 .57546 Mdot
.69289 .57174 Mdot
.71131 .56824 Mdot
.72995 .56498 Mdot
.74881 .56195 Mdot
.76786 .55915 Mdot
.78711 .5566 Mdot
.80653 .55429 Mdot
.8261 .55223 Mdot
.86566 .54887 Mdot
.88561 .54757 Mdot
.90565 .54653 Mdot
.92575 .54575 Mdot
.94591 .54522 Mdot
.96609 .54496 Mdot
.58681 .60314 Mdot
.6035 .59835 Mdot
.62904 .59154 Mdot
.63769 .58938 Mdot
.65517 .5852 Mdot
.68187 .57934 Mdot
.69089 .57749 Mdot
.7183 .5723 Mdot
.74619 .56762 Mdot
.76504 .56479 Mdot
.77453 .56346 Mdot
.79366 .56099 Mdot
.82264 .55775 Mdot
.83238 .55679 Mdot
.86178 .55429 Mdot
.88152 .55294 Mdot
.89142 .55236 Mdot
.91129 .55139 Mdot
.92125 .55101 Mdot
.97119 .55004 Mdot
.61925 .5996 Mdot
.6363 .5952 Mdot
.6536 .591 Mdot
.67114 .58702 Mdot
.68893 .58325 Mdot
.70694 .57971 Mdot
.72518 .57638 Mdot
.74363 .57329 Mdot
.76227 .57043 Mdot
.7811 .5678 Mdot
.8001 .56542 Mdot
.81925 .56327 Mdot
.83855 .56138 Mdot
.85798 .55972 Mdot
.87751 .55832 Mdot
.89713 .55717 Mdot
.91683 .55628 Mdot
.93659 .55564 Mdot
.95638 .55525 Mdot
.62647 .6032 Mdot
.63493 .60102 Mdot
.64346 .59889 Mdot
.65205 .59681 Mdot
.6607 .59478 Mdot
.66941 .5928 Mdot
.67818 .59088 Mdot
.68701 .58901 Mdot
.69589 .5872 Mdot
.70483 .58544 Mdot
.71382 .58374 Mdot
.72287 .58209 Mdot
.73197 .5805 Mdot
.74111 .57897 Mdot
.75031 .57749 Mdot
.75956 .57608 Mdot
.76885 .57472 Mdot
.77818 .57341 Mdot
.78756 .57217 Mdot
.79698 .57099 Mdot
.80644 .56987 Mdot
.81593 .56881 Mdot
.82546 .5678 Mdot
.83503 .56686 Mdot
.84462 .56598 Mdot
.85425 .56517 Mdot
.8639 .56441 Mdot
.87358 .56372 Mdot
.88328 .56308 Mdot
.893 .56252 Mdot
.90274 .56201 Mdot
.9125 .56157 Mdot
.92227 .56119 Mdot
.93206 .56087 Mdot
.94186 .56061 Mdot
.95166 .56042 Mdot
.96147 .5603 Mdot
.97128 .56023 Mdot
.65054 .60261 Mdot
.68512 .59478 Mdot
.70275 .59118 Mdot
.73865 .58465 Mdot
.7569 .58173 Mdot
.79392 .57657 Mdot
.81267 .57434 Mdot
.85059 .57062 Mdot
.86972 .56912 Mdot
.90825 .56686 Mdot
.92762 .56611 Mdot
.96647 .56535 Mdot
.67463 .60244 Mdot
.68327 .60055 Mdot
.69197 .59871 Mdot
.70072 .59693 Mdot
.70952 .5952 Mdot
.71838 .59352 Mdot
.72729 .5919 Mdot
.73624 .59034 Mdot
.74524 .58883 Mdot
.75429 .58738 Mdot
.76339 .58599 Mdot
.77253 .58465 Mdot
.78171 .58337 Mdot
.79093 .58215 Mdot
.80018 .58099 Mdot
.80948 .57989 Mdot
.81881 .57885 Mdot
.82818 .57786 Mdot
.83758 .57694 Mdot
.847 .57608 Mdot
.85646 .57527 Mdot
.86594 .57453 Mdot
.87544 .57385 Mdot
.88497 .57323 Mdot
.89452 .57267 Mdot
.90408 .57217 Mdot
.91367 .57174 Mdot
.92326 .57136 Mdot
.93287 .57105 Mdot
.94249 .5708 Mdot
.95211 .57062 Mdot
.96174 .57049 Mdot
.97137 .57043 Mdot
.7162 .59924 Mdot
.73387 .59603 Mdot
.75174 .59304 Mdot
.80635 .58544 Mdot
.82485 .58337 Mdot
.84348 .58154 Mdot
.86223 .57995 Mdot
.89999 .57749 Mdot
.91899 .57663 Mdot
.93803 .57601 Mdot
.9571 .57564 Mdot
.72279 .60332 Mdot
.73156 .60173 Mdot
.74923 .59871 Mdot
.75814 .59728 Mdot
.77608 .5946 Mdot
.78511 .59334 Mdot
.81242 .58992 Mdot
.8308 .58793 Mdot
.84004 .58702 Mdot
.85859 .58538 Mdot
.86791 .58465 Mdot
.8866 .58337 Mdot
.89598 .58283 Mdot
.91479 .58191 Mdot
.92421 .58154 Mdot
.94309 .58099 Mdot
.95254 .58081 Mdot
.97146 .58063 Mdot
.76445 .60155 Mdot
.80027 .59657 Mdot
.8184 .59442 Mdot
.83665 .5925 Mdot
.85502 .59082 Mdot
.87349 .58938 Mdot
.89204 .58817 Mdot
.92935 .58647 Mdot
.94807 .58599 Mdot
.96681 .58575 Mdot
.78839 .60332 Mdot
.79731 .60214 Mdot
.80627 .60102 Mdot
.81526 .59995 Mdot
.82428 .59895 Mdot
.83333 .598 Mdot
.84241 .5971 Mdot
.85152 .59627 Mdot
.86065 .59549 Mdot
.8698 .59478 Mdot
.87898 .59412 Mdot
.88818 .59352 Mdot
.89739 .59298 Mdot
.90663 .5925 Mdot
.91587 .59208 Mdot
.92513 .59172 Mdot
.9344 .59142 Mdot
.94368 .59118 Mdot
.95296 .591 Mdot
.96225 .59088 Mdot
.97154 .59082 Mdot
.84808 .60173 Mdot
.88439 .59889 Mdot
.90266 .59782 Mdot
.93936 .59639 Mdot
.95777 .59603 Mdot
.89876 .60314 Mdot
.91692 .60226 Mdot
.92602 .6019 Mdot
.93513 .60161 Mdot
.94424 .60137 Mdot
.96249 .60108 Mdot
.97162 .60102 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{345, 213.125},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40005I0000eB000`400?l00000o`00003oo`3ooomJ
0?ooo`00o`3ooomJ0?ooo`00o`3ooomJ0?ooo`00o`3ooomJ0?ooo`00o`3ooomJ0?ooo`00o`3ooomJ
0?ooo`00o`3ooomJ0?ooo`00o`3ooomJ0?ooo`00o`3ooomJ0?ooo`00o`3ooomJ0?ooo`00@P3oool4
000000<0oooo0P00000T0?ooo`@000000`3oool2000002D0oooo0P0000040?ooo`8000009P3oool0
0`000000oooo0?ooo`020?ooo`800000903oool3000000@0oooo0P00000U0?ooo`800000103oool2
000002d0oooo00140?ooo`050000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`0Q
0?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool000008`3oool010000000oooo0?oo
o`0000020?ooo`040000003oool0oooo000002D0oooo00D000000?ooo`3oool0oooo000000020?oo
o`030000003oool0oooo02@0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`0Q
0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool00000;03oool004@0oooo00D0
00000?ooo`3oool0oooo000000020?ooo`030000003oool0oooo0280oooo00<000000?ooo`3oool0
0P3oool010000000oooo0?ooo`00000V0?ooo`040000003oool0oooo00000080oooo00<000000?oo
o`3oool0803oool500000080oooo00@000000?ooo`3oool000009P3oool010000000oooo0?ooo`00
00020?ooo`030000003oool0oooo0240oooo00@000000?ooo`3oool000000P3oool010000000oooo
0?ooo`00000/0?ooo`00A03oool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`3o
ool08`3oool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`3oool08P3oool20000
00<0oooo00@000000?ooo`3oool000008P3oool010000000oooo0?ooo`0000030?ooo`040000003o
ool0oooo000002<0oooo0`0000030?ooo`040000003oool0oooo000002<0oooo0`0000030?ooo`04
0000003oool0oooo000002`0oooo000J0?ooo`030000003oool0oooo02L0oooo00D000000?ooo`3o
ool0oooo000000020?ooo`030000003oool0oooo02@0oooo00@000000?ooo`3oool000000P3oool0
0`000000oooo0?ooo`0T0?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3oool08@3o
ool00`000000oooo000000030?ooo`040000003oool0oooo000002<0oooo00<000000?ooo`3oool0
0`3oool010000000oooo0?ooo`00000S0?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3o
ool00000;03oool001X0oooo00<000000?ooo`3oool09@3oool3000000<0oooo00@000000?ooo`3o
ool000008`3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000002<0oooo00@0
00000?ooo`3oool000000P3oool010000000oooo0?ooo`00000T0?ooo`8000000`3oool010000000
oooo0?ooo`00000S0?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool000008`3oool0
0`000000oooo0?ooo`030?ooo`040000003oool0oooo000002`0oooo000J0?ooo`030000003oool0
oooo02L0oooo00<000000?ooo`3oool00P3oool2000002D0oooo0P0000040?ooo`8000009@3oool2
000000@0oooo0P00000V0?ooo`030000003oool0oooo0080oooo0P00000T0?ooo`@000000`3oool2
000002D0oooo0`0000030?ooo`800000;@3oool001X0oooo00<000000?ooo`3oool0o`3ooolY0?oo
o`<000004@3oool001X0oooo00<000000?ooo`3oool0o`3oool/0?ooo`030000003oool0oooo00h0
oooo000C0?ooool000007@0000090?ooo`8000000P3oool5000000030?ooo`00000000000080oooo
0`0000020?ooo`D000000P3oool3000000@0oooo000J0?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool01P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo00<0
00000?ooo`3oool01P3oool00`000000oooo0?ooo`050?ooo`800000203oool00`000000oooo0?oo
o`060?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?oo
o`060?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?oo
o`060?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?oo
o`060?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?oo
o`060?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool03P3oool010000000oooo0?oo
o`0000020?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool00000103oool01@00
0000oooo0?ooo`3oool0000000L0oooo000I0?ooo`800000QP3oool00`000000oooo0?ooo`230?oo
o`800000403oool400000080oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`00
00040?ooo`040000003oool0oooo0?ooo`<000001@3oool001X0oooo00<000000?ooo`3oool0o`3o
ool<0?ooo`030000003oool0oooo00h0oooo00@000000?ooo`3oool000000P3oool010000000oooo
0?ooo`0000020?ooo`040000003oool0oooo000000@0oooo00D000000?ooo`3oool0oooo00000002
0?ooo`030000003oool0oooo0080oooo000J0?ooo`030000003oool0oooo02d0oooo0P00002Z0?oo
o`800000@P3oool01P000000oooo0000003oool000000?ooo`800000103oool2000000030?ooo`00
0000oooo0080oooo00<000000?ooo`3oool00P3oool2000000D0oooo000J0?ooo`030000003oool0
oooo02h0oooo00<000000?ooo`3oool04`3oool2000007`0oooo0P00000F0?ooo`030000003oool0
oooo03l0oooo0`00000A0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo
01X0oooo0P00000Y0?ooo`030000003oool0oooo07/0oooo00<000000?ooo`3oool09P3oool20000
0440oooo0P00000;0?ooo`006P3oool00`000000oooo0?ooo`0K0?ooo`030000003oool0oooo03@0
oooo0P00001P0?ooo`800000=`3oool00`000000oooo0?ooo`1<0?ooo`006P3oool2000005@0oooo
00<000000?ooo`3oool01@3oool2000004h0oooo0P0000080?ooo`030000003oool0oooo08@0oooo
000J0?ooo`030000003oool0oooo0140oooo0P0000190?ooo`030000003oool0oooo04d0oooo00<0
00000?ooo`3oool0AP3oool2000004D0oooo000J0?ooo`030000003oool0oooo0180oooo00<00000
0?ooo`3oool09@3oool2000002D0oooo0P0000120?ooo`800000903oool2000002T0oooo00<00000
0?ooo`3oool0@`3oool001X0oooo00<000000?ooo`3oool08@3oool2000001P0oooo00<000000?oo
o`3oool0903oool00`000000oooo0?ooo`110?ooo`030000003oool0oooo02<0oooo00<000000?oo
o`3oool05P3oool2000005D0oooo000J0?ooo`030000003oool0oooo00`0oooo0P00000D0?ooo`03
0000003oool0oooo0400oooo0P00000j0?ooo`800000@`3oool00`000000oooo0?ooo`0A0?ooo`80
0000@03oool001X0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0l0?ooo`800000
603oool00`000000oooo0?ooo`0200000380oooo0P0000030?ooo`030000003oool0oooo01D0oooo
0P00000o0?ooo`030000003oool0oooo03h0oooo000J0?ooo`030000003oool0oooo03H0oooo0P00
000E0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?oo
o`0J0?ooo`030000003oool0oooo0140oooo0P00001[0?ooo`006P3oool00`000000oooo0?ooo`09
0?ooo`8000006@3oool200000140oooo00<000000?ooo`3oool0<P3oool2000002`0oooo0P00000d
0?ooo`030000003oool0oooo00l0oooo0P00000I0?ooo`800000?@3oool001X0oooo0P00000;0?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool05P3oool2000001L0oooo0P00000D0?oo
o`030000003oool0oooo02/0oooo00<000000?ooo`3oool04@3oool2000001L0oooo0P00000I0?oo
o`030000003oool0oooo01P0oooo00<000000?ooo`3oool0>`3oool001X0oooo00<000000?ooo`3o
ool05@3oool2000002P0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0C0?ooo`80
0000:03oool2000001H0oooo00<000000?ooo`3oool05P3oool00`000000oooo0?ooo`0U0?ooo`80
0000B@3oool001X0oooo00<000000?ooo`3oool01`3oool2000000d0oooo00<000000?ooo`3oool0
1@3oool2000001@0oooo0P00000i0?ooo`<00000903oool2000000030?ooo`000000oooo03L0oooo
0P00000D0?ooo`800000203oool00`000000oooo0?ooo`0:0?ooo`800000>`3oool001X0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`0D0?ooo`030000003oool0oooo01<0oooo00<0
00000?ooo`3oool04P3oool200000180oooo0P00000A0?ooo`030000003oool0oooo02<0oooo00<0
00000?ooo`3oool03P3oool200000180oooo0P00000E0?ooo`030000003oool0oooo01<0oooo00<0
00000?ooo`3oool0503oool00`000000oooo0?ooo`0i0?ooo`006P3oool00`000000oooo0?ooo`0V
0?ooo`8000008`3oool00`000000oooo0?ooo`02000000l0oooo00<000000?ooo`3oool03`3oool2
00000280oooo0P00000B0?ooo`030000003oool0oooo00`0oooo0P0000030?ooo`030000003oool0
oooo0200oooo0P00001J0?ooo`006P3oool00`000000oooo0?ooo`060?ooo`8000007`3oool00`00
0000oooo0?ooo`0U0?ooo`030000003oool0oooo0200oooo00<000000?ooo`3oool08@3oool00`00
0000oooo0?ooo`0P0?ooo`030000003oool0oooo02D0oooo00<000000?ooo`3oool0703oool20000
03X0oooo000J0?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0103oool200000300
oooo0P00000`0?ooo`8000007`3oool200000340oooo0P00000_0?ooo`800000203oool00`000000
oooo0?ooo`0h0?ooo`006P3oool00`000000oooo0?ooo`0?0?ooo`030000003oool0oooo0240oooo
0P00000<0?ooo`030000003oool0oooo01l0oooo0P00000>0?ooo`<00000703oool3000000h0oooo
0P00000R0?ooo`030000003oool0oooo00T0oooo0P00000S0?ooo`030000003oool0oooo0440oooo
000J0?ooo`030000003oool0oooo00D0oooo0P00000?0?ooo`8000003`3oool2000000/0oooo00@0
00000?ooo`3oool0oooo0P00000K0?ooo`8000003@3oool00`000000oooo0?ooo`0>0?ooo`030000
003oool0oooo01/0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0:0?ooo`800000
6`3oool2000000@0oooo00<000000?ooo`3oool0203oool2000000l0oooo0P00000?0?ooo`800000
>@3oool001X0oooo0P0000070?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool00`3o
ool2000000T0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0<0?ooo`800000303o
ool010000000oooo00000000000K0?ooo`8000006P3oool2000001/0oooo0P0000020?ooo`030000
003oool0oooo00T0oooo0P00000?0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0
1P3oool2000000H0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0g0?ooo`006P3o
ool00`000000oooo0?ooo`0N0?ooo`030000003oool0oooo01`0oooo0P00000:0?ooo`030000003o
ool0oooo00d0oooo00<000000?ooo`3oool02`3oool2000000d0oooo00<000000?ooo`3oool06@3o
ool00`000000oooo0?ooo`0:0?ooo`8000003P3oool00`000000oooo0?ooo`0=0?ooo`030000003o
ool0oooo00L0oooo0P00000O0?ooo`030000003oool0oooo04l0oooo000J0?ooo`030000003oool0
oooo00@0oooo0P00000=0?ooo`8000003P3oool2000000d0oooo0P00000:0?ooo`030000003oool0
oooo02L0oooo00<000000?ooo`3oool02P3oool2000001T0oooo0P00000>0?ooo`030000003oool0
oooo02L0oooo00<000000?ooo`3oool01`3oool2000000d0oooo0P00000>0?ooo`8000003@3oool2
000003P0oooo000J0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool00`3oool20000
00L0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo
00/0oooo0P00000<0?ooo`8000009@3oool2000001P0oooo0P00000U0?ooo`800000303oool20000
00h0oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0=0?ooo`030000003oool0oooo
00@0oooo0P0000060?ooo`030000003oool0oooo03H0oooo000J0?ooo`030000003oool0oooo00`0
oooo00<000000?ooo`3oool06@3oool2000001P0oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`0:0?ooo`8000002`3oool2000000/0oooo00<000000?ooo`3oool05`3oool00`000000
oooo0?ooo`080?ooo`8000002`3oool2000000d0oooo00<000000?ooo`3oool02`3oool00`000000
oooo0?ooo`0E0?ooo`800000703oool00`000000oooo0?ooo`0m0?ooo`006P3oool00`000000oooo
0?ooo`030?ooo`8000006P3oool2000000P0oooo00<000000?ooo`3oool0<P3oool3000000X0oooo
00<000000?ooo`3oool02@3oool2000001H0oooo0P00000<0?ooo`030000003oool0oooo00L0oooo
0P0000000`3oool000000?ooo`0c0?ooo`030000003oool0oooo00D0oooo0P00000J0?ooo`800000
=`3oool001X0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0@0?ooo`8000001`3o
ool00`000000oooo0?ooo`0U0?ooo`8000002P3oool2000000X0oooo00<000000?ooo`3oool05@3o
ool00`000000oooo0?ooo`0E0?ooo`030000003oool0oooo01D0oooo00<000000?ooo`3oool01P3o
ool2000000/0oooo0P00000X0?ooo`030000003oool0oooo00@0oooo0P00000C0?ooo`030000003o
ool0oooo03D0oooo000;0?ooo`@000000`3oool2000000H0oooo00<000000?ooo`3oool00P3oool2
000001@0oooo00<000000?ooo`3oool05P3oool2000001H0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`0P0?ooo`800000503oool200000280oooo00<000000?ooo`3oool02P3oool0
0`000000oooo0?ooo`0C0?ooo`8000006@3oool00`000000oooo0?ooo`0A0?ooo`800000=P3oool0
00/0oooo00<000000?ooo`3oool00`3oool010000000oooo0?ooo`0000050?ooo`800000103oool0
0`000000oooo0?ooo`090?ooo`8000002`3oool2000000/0oooo0P0000070?ooo`030000003oool0
oooo00800000503oool200000200oooo0P0000090?ooo`800000503oool2000000T0oooo0P00000P
0?ooo`800000503oool2000000<0oooo00<000000?ooo`3oool0103oool2000000/0oooo0P00000;
0?ooo`800000303oool00`000000oooo0?ooo`0d0?ooo`00303oool00`000000oooo0?ooo`020?oo
o`040000003oool0oooo000000D0oooo00<000000?ooo`3oool0203oool2000000H0oooo00<00000
0?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0oooo00T0oooo00<00000
0?ooo`3oool0203oool2000000T0oooo0`00000C0?ooo`8000002P3oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo01<0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`06
0?ooo`800000503oool2000000030?ooo`000000oooo00L0oooo0P00000;0?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool02P3oool00`000000oooo0?ooo`0:0?ooo`030000003oool0
oooo00<0oooo0P00000l0?ooo`003@3oool01@000000oooo0?ooo`3oool000000080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`8000001@3oool00`000000oooo0?ooo`02
0?ooo`800000:@3oool2000000L0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`08
0?ooo`800000203oool3000001<0oooo0P00000B0?ooo`8000004`3oool3000000P0oooo0P00000;
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0103oool2000002T0oooo0P000005
0?ooo`030000003oool0oooo0080oooo0P00000f0?ooo`003P3oool010000000oooo0?ooo`000002
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo01<0oooo0P00000:0?ooo`8000001`3oool010000000oooo00000000
000B0?ooo`800000203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo0P00
00090?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool01P3oool2000000X0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`050?ooo`8000004P3oool200000080oooo00<0
00000?ooo`3oool0103oool2000000X0oooo0P00000F0?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`3oool0=03oool000/0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?oo
o`0000050?ooo`030000003oool0oooo01<0oooo0P00000A0?ooo`030000003oool0oooo00800000
1`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo0P0000080?ooo`030000
003oool0oooo01`0oooo00<000000?ooo`3oool01P3oool200000180oooo0P0000090?ooo`030000
003oool0oooo01`0oooo00<000000?ooo`3oool01@3oool2000000X0oooo00<000000?ooo`3oool0
203oool00`000000oooo0?ooo`040?ooo`8000000`3oool00`000000oooo0?ooo`0>0?ooo`800000
A`3oool000`0oooo0P0000040?ooo`8000001P3oool010000000oooo0?ooo`3oool200000140oooo
00<000000?ooo`3oool00P00000A0?ooo`030000003oool0oooo0180oooo0P0000070?ooo`030000
003oool0oooo0140oooo0P00000K0?ooo`8000004@3oool3000001/0oooo0P00000D0?ooo`030000
003oool0oooo00@0oooo0P00000E0?ooo`030000003oool0oooo00h0oooo0P0000030?ooo`030000
003oool0oooo00h0oooo0P00000e0?ooo`006P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo01<0oooo00<000000?ooo`3oool01`3oool2000001<0oooo0P0000070?ooo`030000003o
ool0oooo0140oooo0P0000070?ooo`030000003oool0oooo00L0oooo0P0000070?ooo`800000203o
ool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00H0oooo0P0000070?ooo`8000002P3o
ool00`000000oooo0?ooo`040?ooo`800000503oool00`000000oooo0?ooo`030?ooo`800000503o
ool2000000X0oooo00<000000?ooo`3oool04`3oool00`000000oooo0?ooo`0c0?ooo`006P3oool0
0`000000oooo0?ooo`070?ooo`8000004`3oool2000000D0oooo00<000000?ooo`3oool04P3oool0
0`000000oooo0?ooo`070?ooo`8000004@3oool00`000000oooo0?ooo`0@0?ooo`8000001`3oool0
0`000000oooo0?ooo`060?ooo`800000403oool2000000T0oooo00<000000?ooo`3oool0103oool3
00000180oooo00<000000?ooo`3oool03P3oool2000000T0oooo00<000000?ooo`3oool04`3oool0
0`000000oooo0?ooo`020?ooo`8000004`3oool2000003/0oooo000J0?ooo`8000000P3oool20000
00D0oooo00<000000?ooo`3oool00P0000080?ooo`8000001P3oool00`000000oooo0?ooo`020000
00h0oooo0P00000A0?ooo`030000003oool0oooo0100oooo0P0000070?ooo`800000203oool00`00
0000oooo0?ooo`0?0?ooo`030000003oool0oooo00l0oooo00<000000?ooo`3oool03`3oool00`00
0000oooo0?ooo`040?ooo`800000203oool2000001<0oooo00<000000?ooo`3oool03P3oool20000
00h0oooo0P0000030?ooo`030000003oool0oooo0080oooo0P0000090?ooo`8000000`3oool00`00
0000oooo0?ooo`020?ooo`800000=@3oool001X0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0203oool00`000000
oooo0?ooo`060?ooo`8000001@3oool010000000oooo0000000000080?ooo`800000603oool30000
00H0oooo00<000000?ooo`3oool05`3oool200000100oooo0P00000I0?ooo`030000003oool0oooo
00@0oooo0P0000000`3oool000000?ooo`0F0?ooo`8000001`3oool2000000<0oooo00<000000?oo
o`3oool00P3oool2000000T0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`080?oo
o`030000003oool0oooo00L0oooo00<000000?ooo`3oool0<`3oool001X0oooo00<000000?ooo`3o
ool0403oool200000140oooo0P0000060?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`060?ooo`8000001`3oool2000000L0oooo00<000000?ooo`3o
ool05P3oool2000000L0oooo0P00000>0?ooo`<000001P3oool2000001T0oooo00<000000?ooo`3o
ool0103oool2000000L0oooo0P0000090?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`030?ooo`8000004@3oool2000004@0oooo000J0?ooo`040000
003oool0oooo0?ooo`8000001`3oool2000000D0oooo00<000000?ooo`3oool00P0000070?ooo`80
00001@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3o
ool01@3oool2000000l0oooo0P0000070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3o
ool03@3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00@0oooo0P00000?0?ooo`80
0000203oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo0240oooo00<000000?ooo`3o
ool00P3oool2000000L0oooo0P0000030?ooo`030000003oool0oooo0080oooo0P0000070?ooo`80
0000=@3oool001X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`02000000@0oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00H0oooo
0P0000070?ooo`8000001`3oool2000000h0oooo0P00000?0?ooo`030000003oool0oooo00D0oooo
0P0000070?ooo`8000003P3oool2000000h0oooo0P00000>0?ooo`<000001P3oool2000000P0oooo
00<000000?ooo`3oool0303oool2000000h0oooo0P0000070?ooo`8000001`3oool2000000T0oooo
00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`070?ooo`040000003oool0oooo0?ooo`80
00000`3oool00`000000oooo0?ooo`0c0?ooo`006P3oool00`000000oooo0?ooo`060?ooo`030000
003oool0oooo00l0oooo0P00000>0?ooo`040000003oool00000000000D0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`050?ooo`8000001P3oool3000001H0oooo00<000000?ooo`3o
ool01P3oool00`000000oooo0?ooo`040?ooo`8000001`3oool00`000000oooo0?ooo`0=0?ooo`03
0000003oool0oooo00@0oooo0P0000070?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3o
ool04`3oool2000000030?ooo`000000oooo00@0oooo0P0000080?ooo`030000003oool0oooo00H0
oooo00<000000?ooo`3oool00P3oool200000080oooo00<000000?ooo`3oool02`3oool200000180
oooo00<000000?ooo`3oool0=`3oool001X0oooo00<000000?ooo`3oool00P00000@0?ooo`800000
1@3oool010000000oooo00000000000>0?ooo`030000003oool0oooo00h0oooo0P0000050?ooo`03
0000003oool0oooo00H0oooo00<000000?ooo`3oool01@3oool2000000H0oooo0P00000F0?ooo`03
0000003oool0oooo00@0oooo0P00000>0?ooo`8000001`3oool00`000000oooo0?ooo`0C0?ooo`80
00001P3oool2000000P0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`020?ooo`80
00004@3oool00`000000oooo0?ooo`0;0?ooo`8000000P3oool00`000000oooo0?ooo`020?ooo`80
0000403oool2000003@0oooo000J0?ooo`050000003oool0oooo0?ooo`0000002`3oool2000000@0
oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`0G0?ooo`8000001@3oool00`000000
oooo0?ooo`0=0?ooo`8000001P3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00@0
oooo0P00000E0?ooo`8000003P3oool2000001D0oooo0P0000070?ooo`030000003oool0oooo00D0
oooo00<000000?ooo`3oool00`3oool200000100oooo00<000000?ooo`3oool00P3oool2000001X0
oooo00<000000?ooo`3oool01P3oool010000000oooo0?ooo`3oool2000000`0oooo00<000000?oo
o`3oool0<P3oool001X0oooo00<000000?ooo`3oool03P3oool00`000000oooo0?ooo`0>0?ooo`80
00003`3oool2000000D0oooo00<000000?ooo`3oool0303oool2000000H0oooo0`00000<0?ooo`80
00001P3oool00`000000oooo0?ooo`040?ooo`8000001P3oool2000000H0oooo00<000000?ooo`3o
ool03@3oool00`000000oooo0?ooo`030?ooo`8000001P3oool2000000L0oooo00<000000?ooo`3o
ool00`3oool2000000`0oooo0P0000000`3oool000000?ooo`030?ooo`800000403oool00`000000
oooo0?ooo`020?ooo`8000003`3oool200000140oooo00<000000?ooo`3oool0?`3oool001X0oooo
0P0000000`3oool00000000000070?ooo`8000005@3oool010000000oooo00000000000=0?ooo`03
0000003oool0oooo00`0oooo0P0000050?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3o
ool02`3oool00`000000oooo0?ooo`0<0?ooo`8000001P3oool00`000000oooo0?ooo`040?ooo`80
0000303oool2000000L0oooo00<000000?ooo`3oool00`3oool3000000h0oooo00<000000?ooo`3o
ool02`3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00<0oooo0P00000?0?ooo`03
0000003oool0oooo00X0oooo0P0000020?ooo`030000003oool0oooo0180oooo0P0000070?ooo`80
0000=03oool001X0oooo00D000000?ooo`3oool0oooo000000020?ooo`800000103oool00`000000
oooo0?ooo`0F0?ooo`030000003oool0oooo00@0oooo0P0000060?ooo`8000003@3oool00`000000
oooo0?ooo`0<0?ooo`8000003@3oool2000000D0oooo0P0000060?ooo`030000003oool0oooo00`0
oooo00<000000?ooo`3oool02`3oool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00<0
oooo0P0000050?ooo`8000003@3oool2000000l0oooo00<000000?ooo`3oool02P3oool2000000H0
oooo0P0000070?ooo`030000003oool0oooo01H0oooo00@000000?ooo`3oool0oooo0P0000030?oo
o`030000003oool0oooo0380oooo000J0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool07@3oool2000000@0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0<0?ooo`80
00001P3oool2000000D0oooo00<000000?ooo`3oool0303oool2000000D0oooo00<000000?ooo`3o
ool04P3oool2000000`0oooo0P00000E0?ooo`030000003oool0oooo0080oooo0`00000>0?ooo`03
0000003oool0oooo0080oooo0P0000050?ooo`800000403oool00`000000oooo0?ooo`050?ooo`04
0000003oool0oooo0?ooo`800000803oool00`000000oooo0?ooo`0f0?ooo`006P3oool00`000000
oooo0?ooo`02000000H0oooo0P0000060?ooo`8000001P3oool2000000H0oooo0P0000040?ooo`04
0000003oool00000000000/0oooo0P00000<0?ooo`030000003oool0oooo00D0oooo00<000000?oo
o`3oool02`3oool2000000H0oooo00<000000?ooo`3oool04@3oool2000000D0oooo0P00000<0?oo
o`8000001@3oool2000001@0oooo00<000000?ooo`3oool00P3oool2000000l0oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`0:0?ooo`8000002`3oool200000080oooo00@000000?oo
o`3oool0oooo0P0000060?ooo`8000001P3oool2000000H0oooo0P0000060?ooo`800000=03oool0
01X0oooo00D000000?ooo`3oool0oooo000000070?ooo`030000003oool0oooo008000000`3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`040?ooo`800000103oool3000001X0oooo0P0000050?ooo`030000003oool0
oooo0140oooo0P0000060?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool02`3oool0
0`000000oooo0?ooo`040?ooo`030000003oool0oooo00<0oooo0P00000C0?ooo`030000003oool0
oooo00<0oooo0P00000J0?ooo`80000000<0oooo0000003oool00P3oool2000000L0oooo00<00000
0?ooo`3oool01@3oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<00000
0?ooo`3oool00P0000030?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0<P3oool0
01X0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0;0?ooo`8000003@3oool20000
00D0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`8000001@3oool20000
00H0oooo0P0000040?ooo`030000003oool0oooo0140oooo0P0000050?ooo`8000002`3oool20000
00`0oooo0P00000;0?ooo`<00000103oool2000001@0oooo00@000000?ooo`3oool0oooo0P000006
0?ooo`8000001@3oool2000000L0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`02
0?ooo`8000003@3oool2000000h0oooo00<000000?ooo`3oool0?P3oool001X0oooo00<000000?oo
o`3oool00P00000=0?ooo`8000001P3oool2000000<0oooo00<000000?ooo`3oool01`3oool20000
00<0oooo00<000000?ooo`3oool02`3oool2000000D0oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`050?ooo`030000003oool0oooo00<0oooo0P00000;0?ooo`8000001@3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo00<0oooo0P0000050?ooo`030000003oool0oooo
00/0oooo00<000000?ooo`3oool00P3oool2000000H0oooo00<000000?ooo`3oool0103oool00`00
0000oooo0?ooo`020?ooo`8000002`3oool2000000H0oooo00<000000?ooo`3oool01@3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo0080oooo0P00000>0?ooo`030000003oool0oooo
008000002P3oool00`000000oooo0?ooo`02000000H0oooo0P00000=0?ooo`800000=03oool001X0
oooo00H000000?ooo`3oool0oooo0000003oool2000000/0oooo00@000000?ooo`0000000000103o
ool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo0100oooo0`00000;0?ooo`800000303o
ool00`000000oooo0?ooo`030?ooo`8000001@3oool200000180oooo00<000000?ooo`3oool00P3o
ool2000000/0oooo0P0000060?ooo`030000003oool0oooo00l0oooo0`0000040?ooo`8000001P3o
ool00`000000oooo0?ooo`090?ooo`8000002P3oool200000080oooo00<000000?ooo`3oool0403o
ool00`000000oooo0?ooo`0<0?ooo`040000003oool0oooo0?ooo`8000000P3oool00`000000oooo
0?ooo`080?ooo`8000000P3oool00`000000oooo0?ooo`0b0?ooo`006P3oool2000000030?ooo`00
000000000080oooo00<000000?ooo`3oool0303oool00`000000oooo0?ooo`0I0?ooo`8000002`3o
ool00`000000oooo0?ooo`040?ooo`800000103oool00`000000oooo0?ooo`0:0?ooo`8000001@3o
ool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00<0oooo0P00000A0?ooo`8000002P3o
ool200000140oooo0P0000060?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P3o
ool2000000d0oooo00@000000?ooo`3oool0oooo0P0000060?ooo`030000003oool0oooo00T0oooo
0P00000L0?ooo`030000003oool0oooo00`0oooo00@000000?ooo`0000000000=03oool001X0oooo
00D000000?ooo`3oool0oooo000000050?ooo`8000001@3oool2000000D0oooo0P0000050?ooo`80
00001P3oool2000000D0oooo0P0000030?ooo`030000003oool0oooo00X0oooo0P0000050?ooo`03
0000003oool0oooo00X0oooo0P0000040?ooo`8000002`3oool2000000D0oooo00<000000?ooo`3o
ool00P3oool2000000D0oooo0P0000050?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3o
ool00P3oool2000000D0oooo0P0000050?ooo`030000003oool0oooo0080oooo0P00000;0?ooo`<0
00000`3oool2000000d0oooo00@000000?ooo`3oool0oooo0P00000>0?ooo`030000003oool0oooo
008000001@3oool2000000H0oooo0P0000050?ooo`8000001@3oool2000000D0oooo0P0000060?oo
o`030000003oool0oooo0380oooo000J0?ooo`030000003oool0oooo00P0oooo00@000000?ooo`00
000000000`3oool00`000000oooo0?ooo`040?ooo`040000003oool00000000000<0oooo00<00000
0?ooo`3oool01@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00<0oooo0P000004
0?ooo`800000103oool00`000000oooo0?ooo`0:0?ooo`8000001@3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool02@3oool2000000D0oooo00<00000
0?ooo`3oool00P3oool2000000X0oooo0P0000050?ooo`030000003oool0oooo0080oooo0`00000;
0?ooo`030000003oool0oooo00X0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`02
0?ooo`800000303oool010000000oooo0?ooo`3oool2000000D0oooo0P0000060?ooo`030000003o
ool0oooo00@0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`0200000080oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`0200000080oooo00<000000?ooo`3oool0>@3o
ool001X0oooo00@000000?ooo`00000000001@3oool2000000<0oooo00@000000?ooo`0000000000
2P3oool00`000000oooo0?ooo`0G0?ooo`<000000`3oool00`000000oooo0?ooo`030?ooo`800000
1@3oool2000000@0oooo0`0000090?ooo`8000002P3oool2000000D0oooo0P0000050?ooo`030000
003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000
003oool0oooo0080oooo0P0000040?ooo`8000002P3oool2000000X0oooo0P0000000`3oool00000
0?ooo`020?ooo`8000001@3oool2000000D0oooo00@000000?ooo`3oool0oooo0P0000000`3oool0
00000?ooo`0H0?ooo`030000003oool0oooo00L0oooo0P0000020?ooo`030000003oool0oooo0080
00001@3oool2000003<0oooo000J0?ooo`040000003oool0oooo000000H0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`030?ooo`8000001@3oool2000000D0oooo0P0000040?ooo`80
00002`3oool00`000000oooo0?ooo`080?ooo`<00000103oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo0080oooo0P0000040?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`0>0?ooo`8000002P3oool200000140oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo0080oooo0P0000050?ooo`03
0000003oool0oooo00@0oooo00@000000?ooo`3oool0oooo0P0000000`3oool000000?ooo`090?oo
o`030000003oool0oooo00L0oooo0P0000050?ooo`8000001@3oool2000000D0oooo0P0000060?oo
o`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0a0?oo
o`006P3oool00`000000oooo0?ooo`030?ooo`8000002`3oool2000000<0oooo00<000000?ooo`3o
ool0103oool010000000oooo0000000000030?ooo`030000003oool0oooo00<0oooo00<000000?oo
o`3oool00`3oool2000000D0oooo0P0000040?ooo`800000103oool00`000000oooo0?ooo`0?0?oo
o`030000003oool0oooo00<0oooo0P0000040?ooo`8000005@3oool2000000@0oooo0P00000:0?oo
o`800000103oool2000001D0oooo0P0000040?ooo`8000001P3oool00`000000oooo0?ooo`0?0?oo
o`040000003oool0oooo0?ooo`800000103oool2000000D0oooo0P0000050?ooo`030000003oool0
oooo00@0oooo00<000000?ooo`3oool00P0000020?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool00P00000;0?ooo`800000=`3oool001X0oooo00@000000?ooo`00000000000`3oool0
0`000000oooo0?ooo`0:0?ooo`040000003oool00000000000X0oooo00<000000?ooo`3oool02@3o
ool2000000<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00<0oooo0P0000040?ooo`800000403oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0080oooo0P00000:0?ooo`800000103oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo00T0oooo00<000000?ooo`3oool00`3oool010000000oooo0?ooo`3oool2000000X0oooo
0P0000050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool03@3oool2000000@0oooo
0P0000060?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`02000000`0oooo00<000000?ooo`3oool01`3oool200000080oooo00<000000?ooo`3oool0
2P3oool00`000000oooo0?ooo`02000003<0oooo000=0?ooo`030000003oool0oooo0080oooo0P00
00060?ooo`040000003oool0oooo000001<0oooo00<000000?ooo`3oool02@3oool2000000D0oooo
0P0000030?ooo`030000003oool0oooo00T0oooo0P00000:0?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool2000000D0oooo0P00000?0?ooo`030000003oool0oooo0080oooo
0P0000040?ooo`8000002@3oool2000000X0oooo0P0000090?ooo`800000103oool2000000D0oooo
00<000000?ooo`3oool0303oool2000000@0oooo0P0000060?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool01`3oool2000000`0oooo00<000000?ooo`3oool00P0000050?ooo`800000
303oool00`000000oooo0?ooo`0A0?ooo`030000003oool0oooo0340oooo000=0?ooo`050000003o
ool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`030?ooo`8000002P3oool2000000/0oooo
0P0000090?ooo`<00000103oool00`000000oooo0?ooo`080?ooo`800000103oool3000001@0oooo
0`0000040?ooo`030000003oool0oooo0080oooo0P0000090?ooo`800000103oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo0080oooo0P0000040?ooo`030000003oool0oooo00T0oooo
00@000000?ooo`3oool0oooo0P0000050?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3o
ool0oooo0P0000090?ooo`8000001@3oool010000000oooo0?ooo`3oool3000001@0oooo0P000000
0`3oool000000?ooo`020?ooo`8000002`3oool00`000000oooo0?ooo`0200000080oooo00<00000
0?ooo`3oool01P3oool2000000/0oooo0P00000m0?ooo`002P3oool500000080oooo00@000000?oo
o`3oool000001@3oool010000000oooo0000000000040?ooo`8000000`3oool010000000oooo0000
000000040?ooo`8000000`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo
0P0000030?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0203oool2000000X0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`020?ooo`8000000`3oool200000100oooo
00@000000?ooo`3oool0oooo0P0000090?ooo`8000001@3oool00`000000oooo0?ooo`0<0?ooo`<0
00000`3oool2000000D0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`80
00002`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo008000002`3oool00`000000
oooo0?ooo`0:0?ooo`030000003oool0oooo00800000103oool200000080oooo00<000000?ooo`3o
ool00P0000040?ooo`800000<`3oool000X0oooo00@000000?ooo`3oool000000`3oool010000000
oooo0?ooo`0000050?ooo`040000003oool0oooo000000D0oooo00<000000?ooo`3oool0103oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo0P0000050?ooo`800000103oool2
000000@0oooo0P0000030?ooo`030000003oool0oooo00P0oooo0P00000:0?ooo`030000003oool0
oooo01<0oooo0P0000040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00P3oool2
000000l0oooo0P0000080?ooo`8000003`3oool2000000@0oooo00<000000?ooo`3oool00`3oool0
10000000oooo0?ooo`3oool2000001H0oooo00<000000?ooo`3oool01`3oool2000000/0oooo00<0
00000?ooo`3oool00P0000040?ooo`800000103oool2000000D0oooo0P0000050?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0340oooo000;0?ooo`030000003oool0000000<0oooo00@000000?ooo`3oool000001@3o
ool00`000000oooo0?ooo`020?ooo`8000005@3oool010000000oooo0000000000030?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`800000
103oool2000000<0oooo0`0000080?ooo`8000003`3oool2000000@0oooo0P0000080?ooo`800000
103oool00`000000oooo0?ooo`020?ooo`8000000`3oool2000000D0oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`020?ooo`8000000`3oool2000000D0oooo00@000000?ooo`3oool0
oooo0P0000080?ooo`<000000`3oool2000000l0oooo0P0000080?ooo`80000000@0oooo0000003o
ool0oooo0P0000040?ooo`8000001@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00P0000020?ooo`030000003oool0oooo0180oooo0P00000f0?oo
o`00303oool2000000<0oooo00@000000?ooo`3oool000001@3oool010000000oooo000000000002
0?ooo`040000003oool00000000000@0oooo0P00000:0?ooo`8000000`3oool00`000000oooo0?oo
o`080?ooo`8000002@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool00P3oool2000000<0oooo0`0000080?ooo`800000103oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3oool2000000@0oooo00<0
00000?ooo`3oool00P3oool2000000P0oooo0P0000050?ooo`040000003oool0oooo0?ooo`800000
2P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool0
oooo0P0000080?ooo`80000000@0oooo0000003oool0oooo0P0000050?ooo`030000003oool0oooo
00<0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`060?ooo`8000002`3oool00`00
0000oooo0?ooo`02000000X0oooo0P0000040?ooo`8000000P3oool010000000oooo00000000000c
0?ooo`003@3oool00`000000oooo0?ooo`020?ooo`8000001P3oool010000000oooo0?ooo`000005
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`02
0?ooo`800000103oool2000000<0oooo0`0000040?ooo`8000003P3oool2000000<0oooo00<00000
0?ooo`3oool00`3oool010000000oooo0?ooo`3oool2000000@0oooo0P0000080?ooo`8000003P3o
ool2000000@0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo00P0oooo00@000000?ooo`3oool0oooo0P00000>0?ooo`800000203oool300000080oooo
0P0000050?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P00000>0?ooo`800000
103oool2000000040?ooo`000000oooo0?ooo`800000103oool2000000D0oooo00<000000?ooo`3o
ool02@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3o
ool0<@3oool001X0oooo00<000000?ooo`3oool0203oool2000000X0oooo0P0000080?ooo`<00000
0`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
0P3oool2000000T0oooo00<000000?ooo`3oool01`3oool2000000<0oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`020?ooo`8000000`3oool2000000P0oooo0P0000040?ooo`030000
003oool0oooo00`0oooo0P0000080?ooo`8000003`3oool010000000oooo0?ooo`3oool2000000P0
oooo0`0000020?ooo`8000001@3oool00`000000oooo0?ooo`020?ooo`040000003oool0oooo0?oo
o`8000002P3oool00`000000oooo0?ooo`060?ooo`8000001@3oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P0000000`3oool000000?ooo`060?oo
o`8000002P3oool2000003`0oooo000J0?ooo`040000003oool00000000000P0oooo0`0000040?oo
o`8000000`3oool00`000000oooo0?ooo`040?ooo`8000000P3oool00`000000oooo0?ooo`080?oo
o`8000002@3oool00`000000oooo0?ooo`070?ooo`800000103oool2000000<0oooo00<000000?oo
o`3oool03@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo0P0000030?oo
o`030000003oool0oooo00`0oooo0P0000030?ooo`800000203oool2000000<0oooo0P00000?0?oo
o`030000003oool0oooo008000001@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
00d0oooo00<000000?ooo`3oool00P0000040?ooo`8000002P3oool00`000000oooo0?ooo`060?oo
o`8000002`3oool010000000oooo0000000000070?ooo`030000003oool0oooo008000000`3oool2
00000080oooo00<000000?ooo`3oool01@3oool2000003<0oooo000J0?ooo`050000003oool0oooo
0000003oool00P0000070?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool02@3oool0
0`000000oooo0?ooo`020?ooo`800000203oool00`000000oooo0?ooo`0B0?ooo`<000000`3oool0
10000000oooo0?ooo`3oool2000000@0oooo0P00000=0?ooo`800000103oool010000000oooo0?oo
o`3oool2000000P0oooo0P0000040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0
1`3oool00`000000oooo0?ooo`020?ooo`040000003oool0oooo0?ooo`800000203oool2000000@0
oooo00@000000?ooo`3oool0oooo0P00000=0?ooo`800000103oool2000000@0oooo00<000000?oo
o`3oool00P0000000`3oool000000?ooo`0C0?ooo`030000003oool0oooo00D0oooo0P0000050?oo
o`030000003oool0oooo00T0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`050?oo
o`8000000P3oool00`000000oooo0?ooo`0a0?ooo`006P3oool2000000@0oooo00<000000?ooo`3o
ool01`3oool2000000T0oooo0P0000090?ooo`<000003@3oool2000000T0oooo0P0000030?ooo`03
0000003oool0oooo00H0oooo0`0000030?ooo`040000003oool0oooo0?ooo`800000103oool20000
00<0oooo00<000000?ooo`3oool01P3oool2000000<0oooo0P0000030?ooo`800000203oool20000
00P0oooo0P0000080?ooo`8000000`3oool200000080oooo0`0000090?ooo`030000003oool0oooo
00800000103oool2000000@0oooo00<000000?ooo`3oool00P0000000`3oool000000?ooo`070?oo
o`030000003oool0oooo008000002@3oool2000000d0oooo0P0000000`3oool000000?ooo`070?oo
o`8000002@3oool2000000X0oooo00<000000?ooo`3oool0=03oool001X0oooo00@000000?ooo`00
000000000`3oool2000000@0oooo0P0000020?ooo`040000003oool00000000000<0oooo0P000002
0?ooo`040000003oool00000000000P0oooo00<000000?ooo`3oool01`3oool2000000<0oooo00<0
00000?ooo`3oool0203oool010000000oooo0?ooo`3oool2000000T0oooo00<000000?ooo`3oool0
1P3oool2000000@0oooo00<000000?ooo`3oool01P3oool2000000<0oooo00<000000?ooo`3oool0
0P3oool00`000000oooo0?ooo`020?ooo`040000003oool0oooo0?ooo`800000103oool00`000000
oooo0?ooo`070?ooo`040000003oool0oooo0?ooo`800000103oool00`000000oooo0?ooo`020?oo
o`050000003oool0oooo0?ooo`0000000`3oool2000000T0oooo00@000000?ooo`3oool0oooo0`00
00080?ooo`030000003oool0oooo00H0oooo0P0000040?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool00P00000:0?ooo`030000003oool0oooo00D0oooo0P0000020?ooo`040000003o
ool00000000000<0oooo0P0000020?ooo`040000003oool00000000000@0oooo0P0000030?ooo`80
0000<`3oool001X0oooo00@000000?ooo`3oool00000103oool010000000oooo0000000000020?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`030?oo
o`040000003oool0oooo0?ooo`800000103oool2000000<0oooo0P0000030?ooo`<000001`3oool2
000000P0oooo00<000000?ooo`3oool01`3oool2000000<0oooo0P0000030?ooo`030000003oool0
oooo00L0oooo0P0000030?ooo`040000003oool0oooo0?ooo`8000003@3oool010000000oooo0?oo
o`3oool2000000P0oooo0P0000040?ooo`030000003oool0oooo00X0oooo0P0000040?ooo`030000
003oool0oooo008000002P3oool00`000000oooo0?ooo`02000000<0oooo0P00000:0?ooo`030000
003oool0oooo00D0oooo0P0000070?ooo`80000000@0oooo0000003oool0oooo0P0000030?ooo`80
0000103oool2000000@0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`020?ooo`03
0000003oool0oooo00<0oooo00@000000?ooo`00000000000P3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo0340oooo000J0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3o
ool01`3oool2000000d0oooo00<000000?ooo`3oool00`3oool300000080oooo00<000000?ooo`3o
ool00`3oool010000000oooo0?ooo`3oool2000000<0oooo0P00000B0?ooo`8000000`3oool00`00
0000oooo0?ooo`060?ooo`800000103oool00`000000oooo0?ooo`060?ooo`040000003oool0oooo
0?ooo`800000303oool2000000P0oooo0P00000<0?ooo`800000103oool00`000000oooo0?ooo`06
0?ooo`040000003oool0oooo0?ooo`8000002@3oool00`000000oooo0?ooo`0300000140oooo0`00
00020?ooo`800000103oool00`000000oooo0?ooo`030?ooo`060000003oool000000000003oool0
00001@3oool00`000000oooo0?ooo`0:0?ooo`8000002P3oool00`000000oooo0?ooo`0i0?ooo`00
6P3oool010000000oooo0000000000030?ooo`8000002@3oool200000080oooo0`0000090?ooo`80
0000203oool00`000000oooo0?ooo`060?ooo`8000000`3oool00`000000oooo0?ooo`020?ooo`04
0000003oool0oooo0?ooo`800000103oool2000000P0oooo00<000000?ooo`3oool02`3oool01000
0000oooo0?ooo`3oool2000000d0oooo00@000000?ooo`3oool0oooo0P0000030?ooo`8000000`3o
ool00`000000oooo0?ooo`070?ooo`030000003oool0oooo008000000`3oool2000000@0oooo00<0
00000?ooo`3oool02P3oool2000000@0oooo00<000000?ooo`3oool02`3oool00`000000oooo0?oo
o`050?ooo`8000000`3oool2000000D0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?oo
o`02000000T0oooo00<000000?ooo`3oool01@3oool2000000T0oooo0P0000000`3oool000000?oo
o`02000000T0oooo0P0000030?ooo`800000<`3oool001X0oooo00<000000?ooo`3oool00`000002
0?ooo`030000003oool0oooo00@0oooo0P0000020?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool0203oool00`000000oooo0?ooo`060?ooo`800000103oool200000080oooo00<00000
0?ooo`3oool02`3oool00`000000oooo0?ooo`030?ooo`040000003oool0oooo0?ooo`8000000`3o
ool2000000L0oooo0P0000080?ooo`030000003oool0oooo00H0oooo0P0000070?ooo`800000103o
ool00`000000oooo0?ooo`02000000P0oooo0P0000030?ooo`040000003oool0oooo0?ooo`800000
1`3oool2000000T0oooo00<000000?ooo`3oool01@3oool2000000L0oooo0P0000030?ooo`800000
103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00`0oooo00@000000?ooo`000000
00000`3oool2000000X0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`020?ooo`04
0000003oool00000000000L0oooo00H000000?ooo`00000000000?ooo`00000c0?ooo`006P3oool0
0`000000oooo0?ooo`020?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool01P3oool2
00000180oooo00<000000?ooo`3oool00`3oool010000000oooo0?ooo`3oool2000000<0oooo0P00
000<0?ooo`8000000`3oool00`000000oooo0?ooo`020?ooo`040000003oool0oooo0?ooo`800000
0`3oool2000000L0oooo0P0000080?ooo`040000003oool0oooo0?ooo`8000000`3oool00`000000
oooo0?ooo`060?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01P3oool00`000000
oooo0?ooo`02000000@0oooo00<000000?ooo`3oool01@3oool2000000L0oooo0`0000020?ooo`80
0000103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00800000303oool2000000<0
oooo0P0000040?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0403oool2000000T0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`0c0?ooo`006P3oool3000000T0oooo
0P0000030?ooo`8000001`3oool010000000oooo0000000000030?ooo`8000002`3oool2000000P0
oooo0P0000030?ooo`030000003oool0oooo00H0oooo0P0000030?ooo`8000001`3oool2000000<0
oooo00<000000?ooo`3oool00P3oool010000000oooo0?ooo`3oool2000000<0oooo0P0000060?oo
o`800000103oool00`000000oooo0?ooo`090?ooo`8000001`3oool2000000d0oooo00<000000?oo
o`3oool00P0000070?ooo`<000000P3oool2000000@0oooo00<000000?ooo`3oool00P3oool00`00
0000oooo0?ooo`02000000L0oooo0`0000020?ooo`8000002@3oool01P000000oooo000000000000
oooo000000L0oooo0P00000;0?ooo`8000000`3oool200000080oooo00<000000?ooo`3oool0103o
ool2000000<0oooo0P0000080?ooo`800000<`3oool001X0oooo0`0000060?ooo`8000000P3oool0
0`000000oooo0?ooo`020?ooo`<00000203oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00H0oooo0P0000020?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01P3oool2
000000<0oooo00<000000?ooo`3oool00P3oool010000000oooo0?ooo`3oool2000000<0oooo00<0
00000?ooo`3oool02`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo008000000`3o
ool00`000000oooo0?ooo`0:0?ooo`8000000`3oool2000000H0oooo0P0000030?ooo`800000303o
ool010000000oooo0?ooo`3oool2000000<0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`0;0?ooo`030000003oool0oooo00800000103oool00`000000oooo0?ooo`020?ooo`040000
003oool00000000000T0oooo00<000000?ooo`3oool0203oool010000000oooo0000000000090?oo
o`030000003oool0oooo0080oooo00<000000?ooo`3oool01@3oool2000000030?ooo`000000oooo
00<0oooo00@000000?ooo`00000000001P3oool00`000000oooo0?ooo`0a0?ooo`006P3oool30000
00L0oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`800000203oool20000
0080oooo00<000000?ooo`3oool02`3oool2000000<0oooo0P0000020?ooo`030000003oool0oooo
00H0oooo0P0000030?ooo`030000003oool0oooo00X0oooo0P0000070?ooo`8000000`3oool01000
0000oooo0?ooo`3oool2000000L0oooo0P0000030?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool01@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo008000001`3oool2
000000@0oooo00<000000?ooo`3oool00P0000060?ooo`8000003P3oool00`000000oooo0?ooo`02
000000P0oooo00<000000?ooo`3oool00P0000030?ooo`8000003P3oool010000000oooo00000000
00080?ooo`8000002P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00@0oooo0P00
000b0?ooo`006P3oool00`000000oooo000000030?ooo`8000000`3oool2000000@0oooo0P000003
0?ooo`8000000`3oool200000080oooo0`0000020?ooo`8000000`3oool00`000000oooo0?ooo`06
0?ooo`8000001`3oool2000000<0oooo00<000000?ooo`3oool01P3oool200000080oooo00<00000
0?ooo`3oool01P3oool2000000L0oooo00@000000?ooo`3oool0oooo0P0000030?ooo`030000003o
ool0oooo00D0oooo0P0000020?ooo`8000000`3oool2000000L0oooo0P0000060?ooo`8000001`3o
ool200000080oooo0P0000030?ooo`800000203oool00`000000oooo0?ooo`02000000<0oooo00<0
00000?ooo`3oool01@3oool2000000T0oooo00@000000?ooo`00000000002@3oool00`000000oooo
0?ooo`03000000H0oooo0P0000090?ooo`040000003oool00000000000<0oooo0P0000000`3oool0
00000?ooo`02000000<0oooo0P0000030?ooo`8000000`3oool2000000<0oooo0P0000050?ooo`03
0000003oool0oooo0300oooo000J0?ooo`040000003oool0oooo0?ooo`80000000<0oooo0000003o
ool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`0000000`3oool20000
00<0oooo0P0000030?ooo`8000001`3oool00`000000oooo0?ooo`060?ooo`8000000`3oool00`00
0000oooo0?ooo`02000000P0oooo00<000000?ooo`3oool01@3oool2000000<0oooo00@000000?oo
o`3oool0oooo0P0000020?ooo`8000000`3oool01@000000oooo0?ooo`3oool0000000@0oooo00<0
00000?ooo`3oool00P0000040?ooo`030000003oool0oooo00D0oooo00@000000?ooo`3oool0oooo
0P0000030?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`0200000080
oooo0P0000040?ooo`030000003oool0oooo00800000203oool00`000000oooo0?ooo`050?ooo`80
00000`3oool00`000000oooo0?ooo`02000000T0oooo00<000000?ooo`3oool0103oool300000080
oooo0P0000030?ooo`8000000`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00P3oool010000000oooo00000000000e0?ooo`006P3oool300000080
oooo0`0000030?ooo`800000503oool2000000L0oooo00<000000?ooo`3oool00P3oool00`000000
oooo0?ooo`020?ooo`040000003oool0oooo0?ooo`8000000`3oool2000000<0oooo0P0000020?oo
o`030000003oool0oooo00D0oooo0P0000070?ooo`8000000P3oool2000000<0oooo00<000000?oo
o`3oool01P3oool01@000000oooo0?ooo`3oool000000100oooo00@000000?ooo`3oool0oooo0P00
00060?ooo`800000103oool00`000000oooo0?ooo`0>0?ooo`050000003oool0oooo0?ooo`000000
203oool00`000000oooo0?ooo`0200000080oooo0P0000060?ooo`80000000<0oooo0000003oool0
1P3oool010000000oooo0000000000030?ooo`8000000`3oool2000000@0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo0P00000E0?ooo`80
00000`3oool2000000030?ooo`000000oooo00800000<P3oool001X0oooo00<000000?ooo`000000
103oool00`000000oooo0?ooo`020?ooo`040000003oool0oooo0?ooo`8000000`3oool2000000<0
oooo0P0000020?ooo`8000000P3oool300000100oooo0P0000020?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`02000000<0oooo0P0000020?ooo`03
0000003oool0oooo00H0oooo0P0000020?ooo`030000003oool0oooo00D0oooo0P00000?0?ooo`80
0000303oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00T0oooo0P00000>0?ooo`80
00002@3oool00`000000oooo00000002000000L0oooo00<000000?ooo`3oool00P0000030?ooo`80
00000`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0080oooo00@000000?ooo`00
00000000403oool3000000030?ooo`000000000000<0oooo0P0000030?ooo`8000000`3oool20000
00@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo
0300oooo000J0?ooo`030000003oool0oooo00D0oooo0P0000070?ooo`<000000P3oool00`000000
oooo0?ooo`020?ooo`<0000000<0oooo0000003oool00`3oool010000000oooo0?ooo`3oool20000
00<0oooo0P0000020?ooo`8000000`3oool2000000H0oooo0P0000070?ooo`030000003oool0oooo
0080oooo00<000000?ooo`3oool00P0000030?ooo`8000000`3oool00`000000oooo0?ooo`050?oo
o`8000000P3oool010000000oooo0?ooo`3oool2000000H0oooo0P0000030?ooo`030000003oool0
oooo008000000`3oool200000080oooo0P0000060?ooo`8000000P3oool200000080oooo0P000004
0?ooo`030000003oool0oooo008000001P3oool2000000<0oooo00<000000?ooo`3oool00P000008
0?ooo`030000003oool0oooo008000000P3oool2000000@0oooo00<000000?ooo`3oool00P3oool0
0`000000oooo0?ooo`040?ooo`8000001P3oool3000000030?ooo`000000000000<0oooo0P000003
0?ooo`800000103oool01P000000oooo0?ooo`3oool000000?ooo`80000000<0oooo0000003oool0
0`3oool01P000000oooo000000000000oooo000000D0oooo0P00000j0?ooo`006P3oool3000000H0
oooo00@000000?ooo`00000000001P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo
00D0oooo0P0000020?ooo`030000003oool0oooo0080oooo0P0000020?ooo`030000003oool0oooo
0080oooo00@000000?ooo`3oool0oooo0P0000020?ooo`8000003`3oool2000000<0oooo00<00000
0?ooo`3oool00P0000030?ooo`8000000`3oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo00@0oooo0P0000070?ooo`8000000`3oool00`000000oooo0?ooo`02000000H0oooo0P000003
0?ooo`030000003oool0oooo008000001P3oool3000000L0oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`0200000080oooo0P0000040?ooo`030000003oool00000008000003P3oool2
000000030?ooo`000000oooo00800000103oool01@000000oooo0?ooo`3oool000000080oooo0`00
00040?ooo`040000003oool00000000000P0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`030?ooo`80000000<0oooo0000003oool01@3oool200000380oooo000J0?ooo`<000002@3o
ool00`000000oooo0?ooo`0:0?ooo`8000000P3oool2000000L0oooo0`0000060?ooo`030000003o
ool0oooo00H0oooo0P0000020?ooo`050000003oool0oooo0?ooo`0000000`3oool2000000<0oooo
0P0000060?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool00P3oool00`000000oooo
0?ooo`02000000H0oooo0P0000070?ooo`040000003oool0oooo0?ooo`8000000`3oool00`000000
oooo0?ooo`050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool01@3oool00`000000
oooo0?ooo`02000000<0oooo00<000000?ooo`3oool01@3oool2000000H0oooo0P0000030?ooo`05
0000003oool0oooo0?ooo`0000001`3oool00`000000oooo0?ooo`040?ooo`8000000`3oool20000
00<0oooo00<000000?ooo`3oool00P3oool010000000oooo0000000000090?ooo`030000003oool0
oooo00<0oooo0P0000000`3oool000000?ooo`040?ooo`8000000`3oool2000000d0oooo00<00000
0?ooo`3oool01`3oool00`000000oooo0?ooo`0`0?ooo`006P3oool010000000oooo0?ooo`3oool2
000000L0oooo0P0000070?ooo`8000000P3oool010000000oooo0?ooo`3oool3000000L0oooo00<0
00000?ooo`3oool00P0000070?ooo`8000000`3oool00`000000oooo0?ooo`050?ooo`8000000P3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo008000000P3oool2000000H0oooo0P00
00070?ooo`040000003oool0oooo0?ooo`8000000P3oool2000000H0oooo0P0000030?ooo`030000
003oool0oooo00P0oooo0P0000060?ooo`8000002`3oool00`000000oooo0?ooo`02000000H0oooo
0`0000000`3oool00000000000040?ooo`030000003oool0oooo00@0oooo0P0000060?ooo`800000
0P3oool2000000<0oooo00<000000?ooo`3oool00P3oool010000000oooo0000000000080?ooo`04
0000003oool00000000000P0oooo0P0000030?ooo`030000003oool0oooo00@0oooo0`0000040?oo
o`040000003oool00000000000L0oooo0P0000070?ooo`800000=@3oool001X0oooo0`0000020?oo
o`<000001P3oool300000080oooo0P0000020?ooo`030000003oool0oooo00H0oooo00<000000?oo
o`3oool01@3oool200000080oooo00<000000?ooo`3oool01P3oool010000000oooo0?ooo`3oool2
000000L0oooo0P0000060?ooo`8000000P3oool01@000000oooo0?ooo`3oool0000000<0oooo0P00
00020?ooo`8000002`3oool01@000000oooo0?ooo`3oool000000080oooo0P0000030?ooo`030000
003oool0oooo00P0oooo0P0000030?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0
0P00000;0?ooo`030000003oool0oooo008000000`3oool01@000000oooo0?ooo`3oool0000000X0
oooo0`0000000`3oool00000000000040?ooo`060000003oool0oooo0?ooo`000000oooo0P000006
0?ooo`<000001P3oool2000000<0oooo00<000000?ooo`3oool01`3oool010000000oooo00000000
00080?ooo`030000003oool0oooo00H0oooo0`0000030?ooo`80000000<0oooo0000003oool0103o
ool2000000030?ooo`000000oooo00800000<P3oool001X0oooo00<000000?ooo`000000103oool0
0`000000oooo0?ooo`060?ooo`050000003oool0oooo0?ooo`0000000`3oool2000000`0oooo0P00
0000103oool000000000000000050?ooo`8000001`3oool2000000L0oooo00<000000?ooo`3oool0
0P0000030?ooo`030000003oool0oooo00@0oooo0P0000030?ooo`050000003oool0oooo0?ooo`00
00000P3oool2000000<0oooo0P0000050?ooo`8000000`3oool00`000000oooo0?ooo`02000000H0
oooo0P0000030?ooo`030000003oool0oooo008000001P3oool2000000<0oooo00<000000?ooo`3o
ool00P0000060?ooo`8000000`3oool00`000000oooo0?ooo`02000000D0oooo0P0000030?ooo`80
00000`3oool01@000000oooo0?ooo`3oool000000080oooo0P0000070?ooo`030000003oool0oooo
008000000`3oool00`000000oooo0?ooo`040?ooo`<000001@3oool2000000H0oooo0P000000103o
ool0000000000000000<0?ooo`8000000`3oool00`000000oooo0?ooo`020?ooo`030000003oool0
oooo00H0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`0`0?ooo`006P3oool00`00
0000oooo0?ooo`050?ooo`8000001`3oool2000000H0oooo00<000000?ooo`3oool02`3oool01@00
0000oooo0?ooo`3oool000000080oooo0P0000020?ooo`<000001P3oool00`000000oooo0?ooo`05
0?ooo`8000000P3oool00`000000oooo0?ooo`090?ooo`8000001P3oool200000080oooo00<00000
0?ooo`3oool00P3oool00`000000oooo0?ooo`0200000080oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`0200000080oooo0P0000060?ooo`8000001P3oool2000000H0oooo0P000002
0?ooo`8000000`3oool00`000000oooo0?ooo`050?ooo`040000003oool00000000000<0oooo00<0
00000?ooo`3oool00P3oool010000000oooo0000000000060?ooo`<000002`3oool010000000oooo
0000000000080?ooo`030000003oool0oooo00<0oooo0`0000000`3oool00000000000040?ooo`05
0000003oool0oooo0?ooo`0000003@3oool00`000000oooo0?ooo`030?ooo`8000001`3oool20000
03T0oooo000J0?ooo`<000000`3oool2000000040?ooo`000000000000000080oooo0P0000020?oo
o`<000000P3oool2000000X0oooo0P0000070?ooo`80000000<0oooo0000003oool00`3oool00`00
0000oooo0?ooo`050?ooo`8000000P3oool200000080oooo00<000000?ooo`3oool01@3oool20000
00H0oooo00<000000?ooo`3oool01@3oool2000000D0oooo0P0000030?ooo`030000003oool0oooo
008000000P3oool2000000H0oooo00D000000?ooo`3oool0oooo000000020?ooo`8000000`3oool0
0`000000oooo0?ooo`050?ooo`030000003oool0oooo008000000`3oool01@000000oooo0?ooo`3o
ool0000000D0oooo0P0000020?ooo`8000000`3oool00`000000oooo0?ooo`02000000D0oooo0`00
00070?ooo`030000003oool0oooo00<0oooo0P0000080?ooo`040000003oool0000000000080oooo
0P0000080?ooo`060000003oool0oooo0?ooo`000000oooo0P0000070?ooo`8000002P3oool20000
0080oooo0P000000103oool000000000000000030?ooo`80000000@0oooo00000000000000000`3o
ool200000380oooo000J0?ooo`030000003oool0000000@0oooo00D000000?ooo`3oool0oooo0000
00030?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool0oooo000000030?ooo`800000
0P3oool200000080oooo0P0000030?ooo`8000000P3oool00`000000oooo0?ooo`0=0?ooo`800000
0P3oool00`000000oooo0?ooo`0200000080oooo0P0000030?ooo`030000003oool0oooo00@0oooo
0P0000070?ooo`030000003oool0oooo008000000P3oool00`000000oooo0?ooo`040?ooo`800000
0`3oool00`000000oooo0?ooo`02000000X0oooo00<000000?ooo`3oool00P0000060?ooo`800000
0`3oool00`000000oooo0?ooo`070?ooo`8000000`3oool00`000000oooo0?ooo`02000000L0oooo
00@000000?ooo`00000000000`3oool00`000000oooo0?ooo`040?ooo`8000001`3oool010000000
oooo0000000000030?ooo`8000000`3oool00`000000oooo00000002000000l0oooo00@000000?oo
o`00000000000`3oool3000000030?ooo`00000000000080oooo0P0000040?ooo`050000003oool0
oooo0?ooo`0000000`3oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`000000
103oool00`000000oooo0?ooo`0`0?ooo`00303oool2000000@0oooo0P0000060?ooo`030000003o
ool0oooo008000001`3oool2000000L0oooo0P0000060?ooo`8000000P3oool01@000000oooo0?oo
o`3oool0000000@0oooo00<000000?ooo`3oool00P0000020?ooo`8000000`3oool2000000H0oooo
00<000000?ooo`3oool0103oool200000080oooo00<000000?ooo`3oool01@3oool200000080oooo
00<000000?ooo`3oool0103oool2000000<0oooo00<000000?ooo`3oool0103oool200000080oooo
00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`02000000X0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`070?ooo`8000000`3oool00`000000oooo0?ooo`050?ooo`040000
003oool00000000000L0oooo00<000000?ooo`3oool00P0000070?ooo`040000003oool000000000
00L0oooo00<000000?ooo`3oool00`0000060?ooo`030000003oool0oooo00<0oooo0P0000030?oo
o`8000000P3oool2000000<0oooo00<000000?ooo`3oool00P3oool01P000000oooo0?ooo`3oool0
00000?ooo`<000001@3oool2000000L0oooo0P0000070?ooo`800000=03oool000/0oooo00@00000
0?ooo`3oool000000P3oool010000000oooo0?ooo`0000050?ooo`<0000000<0oooo0000003oool0
103oool2000000040?ooo`0000000000000000H0oooo0`0000050?ooo`030000003oool0oooo00D0
oooo0P0000060?ooo`<0000000<0oooo0000003oool00`3oool00`000000oooo0?ooo`0200000080
oooo0P0000020?ooo`8000000P3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00@0
oooo0P0000020?ooo`030000003oool0oooo00@0oooo0P0000030?ooo`030000003oool0oooo00L0
oooo0P0000030?ooo`030000003oool0oooo008000000P3oool2000000030?ooo`000000000000H0
oooo0P0000000`3oool00000000000020?ooo`8000000`3oool00`000000oooo0?ooo`02000000X0
oooo00<000000?ooo`3oool00P0000070?ooo`040000003oool00000000000L0oooo00<000000?oo
o`3oool02@3oool010000000oooo0000000000020?ooo`8000000P3oool2000000<0oooo00<00000
0?ooo`3oool00P3oool3000000030?ooo`000000oooo00@0oooo0P0000080?ooo`030000003oool0
oooo0080oooo0P0000000`3oool000000?ooo`040?ooo`80000000@0oooo00000000000000001P3o
ool300000380oooo000;0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool00000
1@3oool3000000@0oooo0P0000020?ooo`050000003oool0oooo0?ooo`000000203oool00`000000
oooo0?ooo`02000000H0oooo0P0000020?ooo`<000001P3oool00`000000oooo0?ooo`040?ooo`80
00000P3oool01@000000oooo0?ooo`3oool0000000<0oooo00<000000?ooo`3oool00P0000020?oo
o`8000002P3oool200000080oooo00<000000?ooo`3oool00P0000060?ooo`030000003oool0oooo
00`0oooo0P0000050?ooo`8000000`3oool010000000oooo0000000000060?ooo`8000000P3oool0
0`000000oooo0?ooo`02000000D0oooo0`00000>0?ooo`030000003oool0oooo00<0oooo0P000003
0?ooo`040000003oool00000000000T0oooo0P0000030?ooo`8000000`3oool01@000000oooo0?oo
o`3oool0000000<0oooo00@000000?ooo`00000000001`3oool00`000000oooo0?ooo`030?ooo`80
000000<0oooo0000003oool00P0000060?ooo`8000000`3oool00`000000oooo0?ooo`060?ooo`06
0000003oool0oooo0?ooo`000000oooo0P0000050?ooo`030000003oool0oooo0300oooo000;0?oo
o`<000000`3oool010000000oooo0?ooo`0000050?ooo`030000003oool0oooo00D0oooo00<00000
0?ooo`3oool01@3oool2000000H0oooo0P0000020?ooo`030000003oool0oooo0080oooo0P000000
0`3oool000000?ooo`030?ooo`030000003oool0oooo00P0oooo0P0000020?ooo`030000003oool0
oooo00@0oooo0P0000060?ooo`050000003oool0oooo0?ooo`0000000`3oool200000080oooo0P00
00020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool2000000D0oooo0P00
00070?ooo`040000003oool00000000000<0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`050?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00P0000020?ooo`030000
003oool0oooo00@0oooo0P0000050?ooo`8000001`3oool00`000000oooo0?ooo`040?ooo`040000
003oool0000000000080oooo0P0000030?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool00`3oool2000000L0oooo00@000000?ooo`00000000002`3oool00`000000oooo0?ooo`020?oo
o`<000001@3oool010000000oooo0000000000060?ooo`800000203oool00`000000oooo0?ooo`0f
0?ooo`002`3oool00`000000oooo0?ooo`030?ooo`040000003oool0oooo000000D0oooo0`000002
0?ooo`8000000P3oool2000000<0oooo0P000000103oool000000000000000020?ooo`80000000@0
oooo00000000000000001P3oool00`000000oooo0?ooo`0<0?ooo`8000000P3oool00`000000oooo
0?ooo`040?ooo`8000000P3oool2000000d0oooo0P0000030?ooo`040000003oool0000000000080
oooo0P00000:0?ooo`030000003oool0oooo008000000P3oool2000000D0oooo0P0000020?ooo`03
0000003oool0oooo00L0oooo0P0000060?ooo`8000002P3oool010000000oooo0000000000050?oo
o`<0000000<0oooo0000000000000`3oool00`000000oooo0?ooo`060?ooo`8000000P3oool20000
00<0oooo00<000000?ooo`3oool00P00000=0?ooo`<0000000<0oooo0000000000001`3oool01000
0000oooo00000000000?0?ooo`030000003oool0oooo00<0oooo0P000000103oool0000000000000
00020?ooo`80000000@0oooo00000000000000000`3oool200000080oooo0P0000020?ooo`800000
<P3oool000/0oooo00<000000?ooo`3oool00`3oool010000000oooo0?ooo`0000050?ooo`030000
003oool000000080000000<0oooo0000003oool00P3oool00`000000oooo0?ooo`020?ooo`050000
003oool0oooo0?ooo`0000000`3oool01@000000oooo0?ooo`3oool000000080oooo0P0000020?oo
o`8000000P3oool2000000<0oooo0P0000050?ooo`030000003oool0oooo00P0oooo00D000000?oo
o`3oool0oooo000000060?ooo`8000001P3oool00`000000oooo0?ooo`040?ooo`050000003oool0
oooo0?ooo`0000000P3oool2000000X0oooo00H000000?ooo`3oool0oooo0000003oool2000000<0
oooo00<000000?ooo`3oool01`3oool200000080oooo00<000000?ooo`3oool01@3oool010000000
oooo00000000000:0?ooo`030000003oool0oooo008000000P3oool01@000000oooo0?ooo`3oool0
000000T0oooo0P0000020?ooo`050000003oool0oooo0?ooo`0000001`3oool00`000000oooo0?oo
o`030?ooo`8000001`3oool01@000000oooo0?ooo`3oool0000000X0oooo00<000000?ooo`3oool0
0P3oool200000080oooo0P0000030?ooo`8000000P3oool2000000<0oooo00D000000?ooo`3oool0
oooo000000030?ooo`050000003oool0oooo0?ooo`000000103oool010000000oooo0?ooo`3oool3
000000030?ooo`000000oooo0340oooo000<0?ooo`<000000`3oool2000000H0oooo00D000000?oo
o`3oool0oooo0000000?0?ooo`8000001P3oool2000000030?ooo`000000oooo0080oooo00D00000
0?ooo`3oool0oooo000000040?ooo`030000003oool0oooo008000000P3oool200000080oooo0P00
00080?ooo`8000001P3oool00`000000oooo0?ooo`02000000T0oooo0P0000060?ooo`030000003o
ool0oooo008000000P3oool2000000D0oooo0P0000020?ooo`030000003oool0oooo00L0oooo0P00
00030?ooo`040000003oool00000000000H0oooo0P0000020?ooo`030000003oool0oooo00800000
2P3oool010000000oooo0000000000050?ooo`8000000P3oool2000000<0oooo00<000000?ooo`3o
ool00`3oool2000000T0oooo0P0000030?ooo`030000003oool0oooo0080oooo0P0000090?ooo`80
00000P3oool200000080oooo0P0000030?ooo`050000003oool0oooo0?ooo`000000103oool01000
0000oooo0?ooo`3oool3000000H0oooo0P00000@0?ooo`030000003oool0oooo0380oooo000J0?oo
o`<000000P3oool200000080oooo0P0000020?ooo`8000000P3oool200000080oooo0P0000020?oo
o`8000000P3oool2000000d0oooo0P0000020?ooo`050000003oool0oooo0?ooo`0000000`3oool0
0`000000oooo0?ooo`0200000080oooo0P000000103oool000000000000000040?ooo`8000000P3o
ool00`000000oooo0?ooo`040?ooo`8000000P3oool2000000D0oooo0P0000020?ooo`050000003o
ool0oooo0?ooo`0000000P3oool200000080oooo00<000000?ooo`3oool00`3oool200000080oooo
0P0000000`3oool00000000000050?ooo`8000001@3oool3000000D0oooo0P0000000`3oool00000
000000020?ooo`8000001P3oool010000000oooo0000000000030?ooo`060000003oool0oooo0?oo
o`000000oooo0P0000050?ooo`<0000000<0oooo0000000000001`3oool010000000oooo00000000
00040?ooo`<0000000<0oooo0000000000000P3oool2000000<0oooo00D000000?ooo`3oool0oooo
000000030?ooo`040000003oool00000000000d0oooo0`0000000`3oool00000000000020?ooo`<0
000000<0oooo0000000000000P3oool200000080oooo0P0000020?ooo`8000000P3oool200000380
oooo000J0?ooo`030000003oool0000000<0oooo0`0000000`3oool000000?ooo`020?ooo`<00000
00<0oooo0000003oool00P3oool01@000000oooo0?ooo`3oool0000000<0oooo00<000000?ooo`3o
ool00P0000020?ooo`8000000P3oool200000080oooo0P0000050?ooo`8000001P3oool01@000000
oooo0?ooo`3oool0000000<0oooo00@000000?ooo`00000000000P3oool2000000T0oooo00D00000
0?ooo`3oool0oooo000000060?ooo`800000103oool2000000<0oooo00@000000?ooo`0000000000
0P3oool200000080oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003oool0oooo0080
00000P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo008000000`3oool010000000
oooo0?ooo`0000030?ooo`040000003oool0000000000080oooo0P0000020?ooo`040000003oool0
0000000000D0oooo0`0000060?ooo`050000003oool0oooo0?ooo`000000203oool5000000<0oooo
00D000000?ooo`3oool0oooo000000030?ooo`030000003oool0oooo00<0oooo0P0000050?ooo`<0
000000<0oooo0000000000000P3oool200000080oooo0P0000030?ooo`050000003oool0oooo0?oo
o`0000000`3oool010000000oooo0?ooo`3oool3000000030?ooo`000000oooo0080oooo0`000000
0`3oool000000?ooo`020?ooo`030000003oool0oooo0300oooo000J0?ooo`030000003oool0oooo
00D0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`040?ooo`8000001P3oool20000
00030?ooo`000000oooo0080oooo0`0000000`3oool000000?ooo`020?ooo`030000003oool0oooo
008000000P3oool200000080oooo0P0000090?ooo`80000000<0oooo0000003oool00P3oool00`00
0000oooo0?ooo`0200000080oooo0P0000050?ooo`8000001P3oool010000000oooo000000000002
0?ooo`030000003oool0oooo00@0oooo0P0000020?ooo`030000003oool0oooo008000002@3oool0
10000000oooo0000000000050?ooo`8000000`3oool00`000000oooo0?ooo`060?ooo`8000000`3o
ool00`000000oooo00000002000000D0oooo00<000000?ooo`3oool00P0000020?ooo`030000003o
ool0oooo00<0oooo0P0000050?ooo`8000000P3oool2000000<0oooo00D000000?ooo`3oool00000
0?ooo`02000000T0oooo0P0000020?ooo`<0000000<0oooo0000000000000`3oool010000000oooo
0?ooo`3oool3000000030?ooo`000000oooo0080oooo0`0000060?ooo`8000001`3oool00`000000
oooo0?ooo`050?ooo`030000003oool0oooo03H0oooo000J0?ooo`<000002P3oool2000000H0oooo
0P000000103oool000000000000000050?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool0103oool2000000030?ooo`000000oooo0080oooo00D000000?ooo`3oool0oooo000000020?oo
o`8000000P3oool200000080oooo00<000000?ooo`3oool00`3oool200000080oooo00D000000?oo
o`3oool0oooo000000020?ooo`80000000<0oooo0000000000001@3oool200000080oooo00<00000
0?ooo`3oool0103oool200000080oooo00<000000?ooo`3oool0103oool010000000oooo00000000
00090?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool01`3oool200000080oooo00<0
00000?ooo`3oool0103oool010000000oooo0000000000070?ooo`040000003oool00000000000D0
oooo0P0000000`3oool00000000000030?ooo`060000003oool0oooo0?ooo`000000oooo0P000006
0?ooo`040000003oool0000000000080oooo0P0000030?ooo`050000003oool0oooo0?ooo`000000
0`3oool3000000L0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`020?ooo`800000
00@0oooo00000000000000001P3oool2000000X0oooo0P00000b0?ooo`006P3oool00`000000oooo
00000002000000H0oooo0P0000000`3oool000000?ooo`020?ooo`8000000P3oool01@000000oooo
0?ooo`3oool0000000D0oooo0P0000020?ooo`8000001P3oool200000080oooo00<000000?ooo`3o
ool00`3oool2000000H0oooo0P0000020?ooo`030000003oool0oooo0080000000<0oooo00000000
00000P3oool00`000000oooo0?ooo`080?ooo`040000003oool0oooo00000080oooo0P0000020?oo
o`030000003oool0oooo00<0oooo0P0000030?ooo`040000003oool00000000000D0oooo0P000002
0?ooo`030000003oool0oooo0080000000<0oooo0000000000000P3oool2000000@0oooo0P000002
0?ooo`80000000<0oooo0000000000000`3oool010000000oooo0000000000050?ooo`8000000P3o
ool00`000000oooo0?ooo`02000000H0oooo00@000000?ooo`00000000000`3oool010000000oooo
0?ooo`00000:0?ooo`050000003oool000000000003oool00P0000030?ooo`030000003oool00000
008000001@3oool2000000H0oooo00@000000?ooo`00000000001P3oool2000000030?ooo`000000
000000L0oooo00@000000?ooo`3oool0oooo0`0000050?ooo`<000001P3oool2000000030?ooo`00
0000oooo0340oooo000J0?ooo`050000003oool0oooo0?ooo`0000001`3oool00`000000oooo0?oo
o`040?ooo`030000003oool0oooo00@0oooo0P0000050?ooo`<0000000<0oooo0000003oool01P3o
ool00`000000oooo0?ooo`070?ooo`<000001@3oool00`000000oooo0?ooo`040?ooo`80000000@0
oooo0000003oool0oooo0P0000020?ooo`800000203oool200000080oooo00<000000?ooo`3oool0
0P0000000`3oool00000000000020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
0`3oool2000000D0oooo0P0000020?ooo`030000003oool0oooo00800000103oool2000000<0oooo
00@000000?ooo`00000000001@3oool2000000H0oooo00<000000?ooo`3oool0103oool01@000000
oooo000000000000oooo008000000`3oool010000000oooo0000000000080?ooo`8000000P3oool2
000000<0oooo100000060?ooo`030000003oool0oooo0080oooo0P0000000`3oool000000?ooo`08
0?ooo`030000003oool0oooo00D0oooo100000050?ooo`8000001P3oool00`000000oooo0?ooo`05
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0<P3oool001X0oooo0`0000020?oo
o`8000000P3oool2000000D0oooo0P0000060?ooo`80000000@0oooo00000000000000000P3oool2
000000030?ooo`000000oooo00D0oooo0P0000050?ooo`8000001P3oool010000000oooo00000000
000:0?ooo`030000003oool0oooo00<0oooo0P0000020?ooo`040000003oool0000000000080oooo
0P0000020?ooo`030000003oool0oooo00<0oooo0P0000020?ooo`030000003oool0oooo00800000
303oool00`000000oooo0?ooo`0200000080oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00@0oooo00@000000?ooo`00000000000`3oool00`000000
oooo0?ooo`090?ooo`8000000`3oool010000000oooo0000000000060?ooo`050000003oool00000
0000003oool00P0000030?ooo`030000003oool00000008000001@3oool00`000000oooo0?ooo`07
0?ooo`8000000P3oool00`000000oooo0?ooo`030?ooo`8000001@3oool2000000L0oooo0`000002
0?ooo`80000000@0oooo00000000000000001P3oool2000000D0oooo0P0000020?ooo`8000000P3o
ool200000380oooo000J0?ooo`030000003oool0000000<0oooo0`0000000`3oool000000?ooo`02
0?ooo`80000000<0oooo0000003oool01P3oool01@000000oooo0?ooo`3oool0000000<0oooo00@0
00000?ooo`00000000001P3oool200000080oooo0P0000000`3oool000000?ooo`050?ooo`800000
00<0oooo0000003oool0203oool2000000H0oooo00<000000?ooo`3oool00`3oool200000080oooo
00<000000?ooo`3oool00P0000000`3oool00000000000020?ooo`030000003oool0oooo00@0oooo
00@000000?ooo`00000000000P3oool2000000T0oooo00<000000?ooo`3oool01P3oool2000000@0
oooo0P0000090?ooo`030000003oool0oooo00H0oooo0P0000020?ooo`8000000P3oool00`000000
oooo0?ooo`040?ooo`050000003oool000000000003oool00P0000030?ooo`040000003oool00000
000000H0oooo00<000000?ooo`3oool00P3oool2000000/0oooo0`0000070?ooo`<000000P3oool3
000000D0oooo0P0000020?ooo`050000003oool0oooo0?ooo`0000000`3oool00`000000oooo0?oo
o`050?ooo`<00000103oool3000000030?ooo`000000oooo0080oooo00<000000?ooo`3oool0<03o
ool001X0oooo0`0000020?ooo`80000000<0oooo0000000000001@3oool00`000000oooo0?ooo`0;
0?ooo`8000001@3oool3000000D0oooo00H000000?ooo`3oool0oooo0000003oool200000080oooo
0P0000020?ooo`800000103oool2000000D0oooo00<000000?ooo`3oool0103oool2000000030?oo
o`00000000000080oooo00<000000?ooo`3oool00`3oool200000080oooo00<000000?ooo`3oool0
0P0000080?ooo`060000003oool0oooo0?ooo`000000oooo0P00000;0?ooo`8000000`3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo008000002`3oool200000080oooo00D000000?oo
o`3oool0oooo000000070?ooo`8000000`3oool010000000oooo0000000000060?ooo`050000003o
ool000000000003oool00P0000060?ooo`030000003oool0oooo00<0oooo0P0000040?ooo`<00000
00<0oooo0000000000000P3oool200000080oooo00D000000?ooo`3oool0oooo000000040?ooo`80
000000<0oooo0000003oool00`3oool2000000h0oooo00<000000?ooo`3oool00P3oool500000080
oooo0P00000b0?ooo`006P3oool00`000000oooo000000030?ooo`040000003oool0oooo00000080
oooo0P0000020?ooo`8000000P3oool2000000030?ooo`00000000000080oooo0P0000001@3oool0
0000000000000000oooo008000000P3oool00`000000oooo0?ooo`0:0?ooo`<0000000<0oooo0000
003oool00P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00/0oooo00@000000?oo
o`3oool000001@3oool200000080oooo00<000000?ooo`3oool0103oool010000000oooo00000000
00020?ooo`800000103oool200000080oooo00<000000?ooo`3oool00P0000040?ooo`8000000P3o
ool00`000000oooo0?ooo`02000000@0oooo0P0000030?ooo`040000003oool00000000000@0oooo
0P0000030?ooo`040000003oool00000000000@0oooo0P0000020?ooo`8000000P3oool00`000000
oooo0?ooo`040?ooo`<000001`3oool010000000oooo0?ooo`00000=0?ooo`030000003oool0oooo
00<0oooo00@000000?ooo`3oool0oooo0`0000000`3oool000000?ooo`0;0?ooo`<000000P3oool2
000000040?ooo`000000000000000080oooo0P0000000`3oool00000000000020?ooo`8000000P3o
ool2000000<0oooo00@000000?ooo`3oool000000`3oool00`000000oooo0?ooo`0`0?ooo`006P3o
ool00`000000oooo0?ooo`02000000D0oooo0P0000000`3oool000000?ooo`020?ooo`050000003o
ool0oooo0?ooo`0000000P3oool3000000030?ooo`000000oooo0080oooo00@000000?ooo`3oool0
00000P3oool200000080oooo0P0000000`3oool00000000000060?ooo`030000003oool0oooo00<0
oooo0P0000080?ooo`80000000<0oooo0000000000001@3oool2000000P0oooo00<000000?ooo`3o
ool00P0000040?ooo`8000000P3oool01P000000oooo0?ooo`3oool000000?ooo`8000000P3oool0
0`000000oooo0?ooo`030?ooo`80000000@0oooo000000000000oooo0P0000050?ooo`800000103o
ool2000000D0oooo0P000000103oool000000000003oool2000000H0oooo00@000000?ooo`000000
00000P3oool01P000000oooo0?ooo`3oool000000?ooo`800000103oool200000080oooo00<00000
0?ooo`3oool01P3oool2000000D0oooo0P0000000`3oool00000000000080?ooo`8000001P3oool0
0`000000oooo0?ooo`020?ooo`8000000P3oool200000080oooo0P0000020?ooo`050000003oool0
oooo0?ooo`0000000`3oool3000000050?ooo`000000oooo0?ooo`0000000`3oool010000000oooo
0?ooo`3oool3000000D0oooo0P00000d0?ooo`006P3oool3000000030?ooo`000000oooo0080oooo
0P000000103oool0000000000000000<0?ooo`030000003oool0oooo00<0oooo0P0000050?ooo`80
00000P3oool010000000oooo0?ooo`0000020?ooo`8000000P3oool2000000030?ooo`0000000000
0080oooo00<000000?ooo`3oool01`3oool010000000oooo0?ooo`0000020?ooo`80000000<0oooo
0000000000001@3oool2000000D0oooo00<000000?ooo`3oool00`3oool2000000@0oooo0P000002
0?ooo`030000003oool0oooo0080000000<0oooo0000000000000P3oool010000000oooo0?ooo`00
00020?ooo`030000003oool0oooo008000000P3oool00`000000oooo0?ooo`030?ooo`040000003o
ool00000000000<0oooo00@000000?ooo`3oool000000P3oool01@000000oooo000000000000oooo
008000000`3oool010000000oooo0000000000040?ooo`<000001@3oool00`000000oooo0?ooo`02
0?ooo`8000001@3oool2000000030?ooo`000000000000<0oooo00@000000?ooo`3oool000002@3o
ool01@000000oooo000000000000oooo008000000P3oool200000080oooo00@000000?ooo`3oool0
oooo0`0000000`3oool000000?ooo`030?ooo`8000001P3oool00`000000oooo0?ooo`090?ooo`80
000000@0oooo0000000000000000103oool300000380oooo000J0?ooo`030000003oool0000000H0
oooo00D000000?ooo`3oool0oooo000000050?ooo`8000000P3oool2000000D0oooo0P000000103o
ool000000000000000040?ooo`030000003oool0oooo00@0oooo0P0000000`3oool000000?ooo`02
0?ooo`040000003oool0oooo00000080oooo0P0000000`3oool00000000000080?ooo`8000000P3o
ool010000000oooo0?ooo`0000020?ooo`80000000<0oooo0000000000001@3oool2000000D0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`8000000P3oool010000000oooo
0000000000090?ooo`040000003oool00000000000@0oooo0P0000020?ooo`030000003oool0oooo
00H0oooo0P0000020?ooo`040000003oool00000000000H0oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`020?ooo`8000001@3oool2000000030?ooo`000000000000<0oooo00D00000
0?ooo`3oool000000?ooo`02000000L0oooo0P0000020?ooo`8000000`3oool010000000oooo0?oo
o`0000030?ooo`<000001P3oool00`000000oooo0?ooo`020?ooo`80000000@0oooo000000000000
00001@3oool200000080oooo0P0000060?ooo`050000003oool0oooo0?ooo`0000001P3oool00`00
0000oooo0?ooo`0`0?ooo`006P3oool00`000000oooo0?ooo`030?ooo`8000001@3oool2000000D0
oooo0`0000000`3oool000000?ooo`020?ooo`80000000<0oooo0000003oool00P3oool00`000000
oooo0?ooo`030?ooo`80000000<0oooo0000000000000P3oool00`000000oooo0?ooo`0:0?ooo`80
000000@0oooo0000003oool0oooo0P0000020?ooo`80000000<0oooo000000000000103oool20000
0080oooo00@000000?ooo`3oool000001@3oool2000000@0oooo0P0000080?ooo`8000000P3oool0
0`000000oooo0?ooo`030?ooo`040000003oool00000000000T0oooo00<000000?ooo`3oool00`3o
ool00`000000oooo0?ooo`060?ooo`8000000P3oool00`000000oooo0?ooo`030?ooo`040000003o
ool00000000000P0oooo0P0000040?ooo`8000001P3oool01@000000oooo0?ooo`000000oooo0080
0000103oool500000080oooo0P0000020?ooo`030000003oool0000000800000303oool01@000000
oooo000000000000oooo008000001P3oool010000000oooo0?ooo`3oool3000000@0oooo0`000000
0`3oool000000?ooo`030?ooo`8000001@3oool2000003L0oooo000J0?ooo`<0000000@0oooo0000
00000000oooo0`000000103oool000000000003oool3000000D0oooo00<000000?ooo`3oool0103o
ool00`000000oooo0?ooo`030?ooo`8000001@3oool2000000030?ooo`000000oooo00@0oooo0P00
00020?ooo`8000001@3oool00`000000oooo0?ooo`030?ooo`8000000P3oool01@000000oooo0?oo
o`000000oooo008000000P3oool2000000/0oooo00@000000?ooo`00000000000P3oool00`000000
oooo0?ooo`030?ooo`8000000P3oool010000000oooo0000000000040?ooo`8000000P3oool00`00
0000oooo0?ooo`02000000040?ooo`00000000000?ooo`800000103oool2000000040?ooo`000000
00000?ooo`8000000`3oool010000000oooo0000000000040?ooo`8000000P3oool010000000oooo
0000000000060?ooo`<000000`3oool00`000000oooo0?ooo`080?ooo`D000000`3oool01@000000
oooo0?ooo`000000oooo00<000001@3oool00`000000oooo0?ooo`020?ooo`80000000<0oooo0000
000000001`3oool4000000@0oooo0P0000060?ooo`030000003oool0oooo00@0oooo00<000000?oo
o`3oool00P3oool2000000030?ooo`00000000000080oooo0P0000001@3oool00000000000000000
oooo00800000<P3oool001X0oooo00<000000?ooo`0000000P3oool01@000000oooo0?ooo`3oool0
00000080oooo00D000000?ooo`3oool0oooo000000020?ooo`80000000<0oooo0000000000000P3o
ool2000000030?ooo`00000000000080oooo0P0000000`3oool00000000000050?ooo`030000003o
ool0oooo00<0oooo0P0000000`3oool000000?ooo`020?ooo`030000003oool0oooo00<0oooo0P00
00000`3oool00000000000020?ooo`030000003oool0oooo00<0oooo0P0000000`3oool000000?oo
o`020?ooo`040000003oool00000000000/0oooo00<000000?ooo`3oool00`3oool200000080oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo
0P0000020?ooo`040000003oool00000000000@0oooo0P0000020?ooo`040000003oool000000000
00H0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`030?ooo`040000003oool00000
000000D0oooo00<000000?ooo`3oool02@3oool200000080oooo00D000000?ooo`3oool000000?oo
o`02000000H0oooo00D000000?ooo`00000000000?ooo`02000000H0oooo00D000000?ooo`3oool0
00000?ooo`02000000H0oooo00<000000?ooo`3oool00P3oool500000080oooo0P0000000`3oool0
0000000000020?ooo`80000000<0oooo0000000000000`3oool010000000oooo0?ooo`0000030?oo
o`050000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`0`0?ooo`006P3oool01000
0000oooo00000000000=0?ooo`80000000D0oooo0000003oool0oooo000000030?ooo`040000003o
ool0oooo000000<0oooo00@000000?ooo`3oool000000P3oool2000000030?ooo`00000000000080
oooo0P0000000`3oool000000?ooo`040?ooo`8000001@3oool2000000030?ooo`000000oooo00@0
oooo0P0000020?ooo`030000003oool0oooo00<0oooo0P000000103oool000000?ooo`3oool20000
00030?ooo`000000000000L0oooo0P0000020?ooo`040000003oool00000000000@0oooo0P000008
0?ooo`8000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool00`3oool010000000oooo0000000000070?ooo`8000001@3oool200000080oooo00@0
00000?ooo`00000000001`3oool2000000030?ooo`000000000000<0oooo0`0000060?ooo`040000
003oool00000000000H0oooo100000040?ooo`8000001P3oool300000080oooo0P0000000`3oool0
0000000000030?ooo`040000003oool0oooo000000<0oooo00@000000?ooo`3oool000000`3oool0
0`000000oooo0?ooo`03000000`0oooo0P00000d0?ooo`006P3oool6000000D0oooo0P0000020?oo
o`80000000<0oooo0000000000004P3oool2000000050?ooo`000000oooo0?ooo`0000000`3oool0
10000000oooo0000000000020?ooo`80000000<0oooo0000000000001@3oool00`000000oooo0?oo
o`030?ooo`80000000@0oooo0000003oool0oooo0P0000050?ooo`030000003oool0oooo00<0oooo
00D000000?ooo`3oool000000?ooo`02000000030?ooo`00000000000080oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`02000000030?ooo`000000oooo00@0oooo0P0000020?ooo`03
0000003oool0oooo00D0oooo0P0000040?ooo`800000203oool010000000oooo0000000000050?oo
o`040000003oool00000000000<0oooo00<000000?ooo`3oool00`3oool300000080oooo0P000002
0?ooo`040000003oool0oooo000000D0oooo00<000000?ooo`3oool00P3oool2000000<0oooo0`00
00060?ooo`030000003oool0oooo0080oooo1@0000020?ooo`8000000P3oool01@000000oooo0?oo
o`3oool000000080oooo0`00000B0?ooo`D000000P3oool2000000D0oooo1@00000b0?ooo`006P3o
ool00`000000oooo000000020?ooo`030000003oool0oooo00@0oooo00D000000?ooo`3oool0oooo
000000020?ooo`030000003oool0oooo00800000203oool2000000030?ooo`00000000000080oooo
00<000000?ooo`3oool00P3oool2000000D0oooo00D000000?ooo`3oool0oooo000000020?ooo`04
0000003oool0000000000080oooo0P0000050?ooo`030000003oool0oooo0080oooo0P0000040?oo
o`8000000P3oool2000000L0oooo0P000000103oool000000?ooo`3oool2000000P0oooo00<00000
0?ooo`3oool00P3oool200000080oooo00<000000?ooo`3oool01@3oool200000080oooo00<00000
0?ooo`3oool00`3oool010000000oooo0000000000080?ooo`040000003oool00000000000D0oooo
00<000000?ooo`3oool01@3oool200000080oooo00<000000?ooo`0000000P0000060?ooo`800000
0P3oool2000000@0oooo0P0000050?ooo`030000003oool0oooo0080oooo0P0000000`3oool00000
000000030?ooo`040000003oool0oooo000000<0oooo00@000000?ooo`3oool0oooo0P0000060?oo
o`<000000P3oool2000000P0oooo0P0000030?ooo`040000003oool0oooo000000<0oooo00<00000
0?ooo`3oool0103oool010000000oooo0?ooo`00000b0?ooo`006P3oool00`000000oooo0?ooo`03
0?ooo`8000001@3oool2000000D0oooo0P000000103oool000000?ooo`3oool2000000D0oooo00@0
00000?ooo`3oool000000P3oool2000000@0oooo0`0000000`3oool00000000000070?ooo`800000
00<0oooo0000003oool00P3oool010000000oooo0000000000020?ooo`80000000D0oooo00000000
0000oooo000000050?ooo`8000000P3oool00`000000oooo0?ooo`020?ooo`8000000P3oool00`00
0000oooo0?ooo`030?ooo`050000003oool000000000003oool00P0000040?ooo`8000000P3oool0
10000000oooo0000000000040?ooo`8000000P3oool010000000oooo0000000000040?ooo`800000
0P3oool010000000oooo0000000000040?ooo`8000000P3oool010000000oooo0000000000040?oo
o`80000000<0oooo0000000000000P3oool00`000000oooo0?ooo`030?ooo`040000003oool00000
000000D0oooo00<000000?ooo`0000000P0000050?ooo`<0000000<0oooo0000000000000P3oool2
00000080oooo00D000000?ooo`3oool000000?ooo`02000000L0oooo0P000000103oool000000000
000000040?ooo`8000000P3oool01@000000oooo0?ooo`3oool0000000<0oooo0P0000040?ooo`<0
00001@3oool2000000D0oooo0P00000g0?ooo`006P3oool3000000@0oooo0P0000020?ooo`800000
00<0oooo0000003oool00P3oool2000000030?ooo`000000000000@0oooo00<000000?ooo`3oool0
0`3oool2000000D0oooo0P0000050?ooo`050000003oool0oooo0000003oool00P0000020?ooo`80
000000<0oooo0000003oool0103oool2000000030?ooo`000000oooo0080oooo00D000000?ooo`3o
ool000000?ooo`02000000030?ooo`00000000000080oooo00<000000?ooo`3oool00P3oool20000
0080oooo00@000000?ooo`0000000000103oool200000080oooo00D000000?ooo`3oool000000?oo
o`0200000080oooo00<000000?ooo`3oool00P3oool2000000040?ooo`00000000000?ooo`800000
103oool2000000@0oooo0P0000040?ooo`80000000@0oooo000000000000oooo0P0000050?ooo`04
0000003oool0000000000080oooo00D000000?ooo`3oool000000?ooo`02000000@0oooo0P000002
0?ooo`<000001P3oool01@000000oooo000000000000oooo008000000P3oool010000000oooo0?oo
o`0000030?ooo`<000001P3oool3000000030?ooo`000000000000<0oooo00@000000?ooo`3oool0
0000103oool3000000@0oooo0P0000050?ooo`030000003oool0oooo0080oooo1@0000040?ooo`<0
000000D0oooo000000000000oooo000000030?ooo`800000<P3oool001X0oooo00@000000?ooo`00
00000000103oool01@000000oooo0?ooo`3oool0000000H0oooo00@000000?ooo`3oool000001@3o
ool2000000050?ooo`00000000000?ooo`0000000`3oool2000000030?ooo`000000oooo00@0oooo
0P0000040?ooo`8000000P3oool00`000000oooo0?ooo`020?ooo`8000000P3oool2000000X0oooo
00@000000?ooo`3oool000000P3oool2000000@0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`030?ooo`8000000`3oool200000080oooo00D000000?ooo`00000000000?ooo`020000
0080oooo00@000000?ooo`3oool000000P3oool010000000oooo0000000000020?ooo`030000003o
ool0oooo00<0oooo00@000000?ooo`00000000000P3oool010000000oooo0?ooo`0000020?ooo`05
0000003oool000000000003oool00P0000020?ooo`<00000103oool3000000D0oooo00<000000?oo
o`3oool00P3oool00`000000oooo0?ooo`020?ooo`8000000`3oool010000000oooo0?ooo`000009
0?ooo`<0000000<0oooo0000000000001@3oool4000000@0oooo0P0000060?ooo`<00000103oool3
000000030?ooo`000000000000H0oooo00@000000?ooo`3oool000001P3oool010000000oooo0?oo
o`0000040?ooo`<00000<P3oool001X0oooo00@000000?ooo`3oool000001@3oool2000000D0oooo
0P0000050?ooo`800000103oool2000000030?ooo`000000oooo00D0oooo00@000000?ooo`000000
00001@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00<0oooo0P0000000`3oool0
00000?ooo`020?ooo`050000003oool000000000003oool00P0000040?ooo`8000001@3oool01@00
0000oooo000000000000oooo00800000103oool2000000D0oooo00H000000?ooo`00000000000?oo
o`0000050?ooo`80000000<0oooo0000003oool00P0000080?ooo`040000003oool00000000000@0
oooo0P0000020?ooo`030000003oool0oooo00D0oooo0P0000020?ooo`@00000103oool010000000
oooo0000000000020?ooo`030000003oool0oooo0080oooo0P0000040?ooo`80000000<0oooo0000
000000000P3oool00`000000oooo0?ooo`020?ooo`800000103oool2000000030?ooo`0000000000
0080oooo00@000000?ooo`3oool0oooo0`0000050?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool00P3oool200000080oooo00<000000?ooo`3oool0103oool4000000<0oooo0P000005
0?ooo`8000001@3oool2000000H0oooo00<000000?ooo`3oool0<@3oool001X0oooo0`000000103o
ool000000000003oool2000000050?ooo`00000000000000003oool00P0000000`3oool000000000
00020?ooo`80000000<0oooo000000000000103oool00`000000oooo0?ooo`090?ooo`8000001`3o
ool2000000D0oooo0P0000000`3oool000000?ooo`040?ooo`80000000D0oooo0000003oool0oooo
000000050?ooo`8000000`3oool200000080oooo00D000000?ooo`3oool000000?ooo`0200000003
0?ooo`000000000000@0oooo0P0000020?ooo`030000003oool0oooo0080oooo0P0000020?ooo`03
0000003oool0oooo0080oooo00@000000?ooo`0000000000203oool00`000000oooo0?ooo`030?oo
o`030000003oool0oooo00D0oooo0P0000020?ooo`030000003oool0oooo0080oooo00@000000?oo
o`00000000001@3oool010000000oooo0000000000040?ooo`80000000<0oooo0000000000000P3o
ool01@000000oooo0?ooo`000000oooo008000000`3oool3000000D0oooo00<000000?ooo`3oool0
0`0000060?ooo`<000001@3oool2000000H0oooo0P0000000`3oool000000?ooo`0:0?ooo`040000
003oool0oooo0?ooo`D000000P3oool5000000040?ooo`00000000000?ooo`<0000000@0oooo0000
00000000oooo0P00000b0?ooo`006P3oool00`000000oooo000000020?ooo`80000000<0oooo0000
003oool00P3oool2000000050?ooo`000000oooo0?ooo`0000000`3oool01@000000oooo0?ooo`00
0000oooo008000000P3oool2000000040?ooo`00000000000?ooo`800000103oool00`000000oooo
0?ooo`030?ooo`80000000<0oooo0000003oool01@3oool010000000oooo0000000000050?ooo`03
0000003oool0oooo0080oooo0P0000050?ooo`060000003oool000000000003oool00000203oool0
1@000000oooo0?ooo`000000oooo008000000P3oool00`000000oooo0?ooo`020?ooo`8000000P3o
ool010000000oooo0000000000070?ooo`050000003oool000000000003oool00P0000000`3oool0
0000000000040?ooo`80000000@0oooo000000000000oooo0P0000020?ooo`030000003oool0oooo
00@0oooo0P0000020?ooo`040000003oool00000000000D0oooo00@000000?ooo`00000000000P3o
ool010000000oooo0?ooo`0000080?ooo`<000000P3oool00`000000oooo0?ooo`020?ooo`800000
1@3oool00`000000oooo0?ooo`020?ooo`8000000P3oool00`000000oooo0?ooo`040?ooo`<00000
1@3oool00`000000oooo0?ooo`020?ooo`80000000@0oooo000000000000oooo0P0000020?ooo`80
00000P3oool010000000oooo0?ooo`0000030?ooo`030000003oool0oooo00@000000`3oool20000
00050?ooo`000000oooo0?ooo`000000<P3oool001X0oooo00<000000?ooo`3oool00`3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo00T0oooo0P0000000`3oool000000?ooo`020?oo
o`040000003oool0oooo00000080oooo00@000000?ooo`00000000000P3oool2000000040?ooo`00
000000000?ooo`8000000`3oool2000000D0oooo0P0000030?ooo`8000000P3oool2000000030?oo
o`000000oooo00<0oooo0P0000020?ooo`030000003oool0oooo0080oooo0P0000070?ooo`800000
00<0oooo0000003oool00P0000000`3oool00000000000020?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool00P3oool2000000<0oooo0P0000020?ooo`040000003oool00000000000@0
oooo0P0000020?ooo`@000000`3oool2000000D0oooo00<000000?ooo`3oool00`3oool300000080
oooo0P0000020?ooo`<000001`3oool2000000D0oooo00@000000?ooo`00000000001@3oool30000
00030?ooo`000000000000@0oooo0`0000030?ooo`800000103oool5000000030?ooo`0000000000
0080oooo0P0000020?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool0oooo0`00
000<0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0=@3oool001X0oooo0`000000
103oool000000000003oool2000000D0oooo0P0000000`3oool00000000000040?ooo`80000000<0
oooo0000003oool03@3oool010000000oooo0?ooo`3oool2000000060?ooo`000000oooo0?ooo`00
0000oooo0P0000000`3oool00000000000050?ooo`030000003oool0oooo0080oooo0P0000020?oo
o`030000003oool0oooo0080oooo0P0000000`3oool00000000000040?ooo`80000000<0oooo0000
003oool00`3oool200000080oooo00<000000?ooo`3oool00P3oool2000000030?ooo`000000oooo
00800000103oool2000000D0oooo00H000000?ooo`00000000000?ooo`0000050?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool00P3oool010000000oooo0000000000020?ooo`030000
003oool0oooo0080oooo0P0000030?ooo`8000000P3oool00`000000oooo00000002000000@0oooo
00@000000?ooo`00000000001@3oool3000000@0oooo1@0000050?ooo`040000003oool000000000
00D0oooo00@000000?ooo`3oool0oooo0`0000000`3oool00000000000020?ooo`030000003oool0
oooo00@000000`3oool00`000000oooo0?ooo`0<0?ooo`<00000103oool2000000030?ooo`000000
000000D0oooo0P000000103oool000000000003oool200000380oooo000<0?ooo`800000103oool2
000000H0oooo00H000000?ooo`00000000000?ooo`0000020?ooo`<00000103oool2000000030?oo
o`000000oooo00@0oooo00<000000?ooo`3oool00P0000000`3oool00000000000040?ooo`800000
00<0oooo0000000000001@3oool00`000000oooo0?ooo`020?ooo`80000000D0oooo0000003oool0
oooo000000020?ooo`80000000@0oooo000000000000oooo0P0000000`3oool000000?ooo`030?oo
o`8000000P3oool010000000oooo0?ooo`0000050?ooo`030000003oool0oooo0080oooo0P000000
0`3oool000000?ooo`02000000030?ooo`00000000000080oooo00<000000?ooo`3oool00P3oool0
1@000000oooo000000000000oooo00800000103oool200000080oooo00<000000?ooo`3oool0103o
ool2000000@0oooo0P0000070?ooo`040000003oool00000000000@0oooo0P0000001@3oool00000
0000003oool0000000D0oooo00D000000?ooo`00000000000?ooo`0200000080oooo0`0000050?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool00`0000060?ooo`<0000000@0oooo0000
00000000oooo0P0000020?ooo`040000003oool0oooo0?ooo`<000001@3oool010000000oooo0?oo
o`3oool200000080oooo0P0000040?ooo`80000000<0oooo0000000000000`3oool00`000000oooo
0?ooo`030?ooo`@000000`3oool2000000040?ooo`000000oooo0?ooo`@00000<P3oool000/0oooo
00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`0000050?ooo`80000000<0oooo0000
003oool01@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0080oooo0P0000050?oo
o`80000000@0oooo0000003oool0oooo0P0000001@3oool000000?ooo`3oool000000080oooo0P00
0000103oool000000000003oool2000000030?ooo`000000oooo00L0oooo00@000000?ooo`3oool0
00000P3oool01@000000oooo000000000000oooo0080000000<0oooo0000003oool0103oool20000
00<0oooo0P0000000`3oool00000000000020?ooo`030000003oool0oooo0080oooo0P0000000`3o
ool000000?ooo`02000000L0oooo00D000000?ooo`3oool000000?ooo`0200000080oooo00<00000
0?ooo`3oool0103oool200000080oooo00<000000?ooo`3oool00`3oool010000000oooo00000000
00070?ooo`040000003oool0000000000080oooo00@000000?ooo`3oool000001P3oool200000080
oooo100000040?ooo`050000003oool000000000003oool00P0000030?ooo`8000001@3oool01@00
0000oooo000000000000oooo008000000P3oool010000000oooo0?ooo`0000020?ooo`030000003o
ool0oooo00H0oooo0`000000103oool000000000003oool200000080oooo00@000000?ooo`3oool0
oooo0`0000030?ooo`@000000`3oool2000000H0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo0340oooo000;0?ooo`040000003oool0oooo00000080
oooo00@000000?ooo`3oool000001@3oool2000000D0oooo0P0000000`3oool00000000000040?oo
o`8000000P3oool3000000<0oooo0P0000000`3oool000000?ooo`040?ooo`030000003oool0oooo
00<0oooo0P0000001@3oool000000?ooo`3oool000000080oooo00@000000?ooo`0000000000103o
ool2000000X0oooo0P0000000`3oool000000?ooo`02000000030?ooo`00000000000080oooo00<0
00000?ooo`3oool00P3oool2000000030?ooo`000000oooo00<0oooo0P0000020?ooo`030000003o
ool0oooo0080oooo00<000000?ooo`3oool00P3oool2000000L0oooo00@000000?ooo`0000000000
103oool2000000030?ooo`000000oooo00800000103oool200000080oooo0`0000040?ooo`800000
0P3oool00`000000oooo0?ooo`040?ooo`8000001@3oool00`000000oooo0?ooo`020?ooo`040000
003oool00000000000D0oooo100000040?ooo`050000003oool000000000003oool00P0000020?oo
o`@000002@3oool2000000@0oooo0P0000020?ooo`040000003oool0oooo00000080oooo0`000006
0?ooo`030000003oool0oooo00<0oooo0`0000030?ooo`<0000000<0oooo0000000000001@3oool2
000000030?ooo`000000000000@0oooo0P00000b0?ooo`00303oool2000000<0oooo00@000000?oo
o`3oool000001@3oool2000000<0oooo0P000000103oool000000?ooo`3oool3000000<0oooo0`00
00000`3oool000000?ooo`040?ooo`040000003oool00000000000@0oooo0P0000020?ooo`800000
00<0oooo0000003oool00`3oool2000000@0oooo0P0000040?ooo`030000003oool0oooo00<0oooo
0P0000040?ooo`030000003oool0oooo0080oooo00D000000?ooo`3oool000000?ooo`0200000003
0?ooo`00000000000080oooo00<000000?ooo`3oool00P3oool2000000030?ooo`000000oooo0080
0000103oool2000000@0oooo00@000000?ooo`00000000001P3oool2000000040?ooo`0000000000
0?ooo`8000000`3oool2000000@0oooo0P0000030?ooo`80000000@0oooo000000000000oooo0P00
00060?ooo`8000000P3oool010000000oooo0?ooo`3oool2000000@0oooo0P0000020?ooo`<00000
1@3oool01@000000oooo000000000000oooo008000000P3oool010000000oooo0?ooo`0000040?oo
o`040000003oool0oooo0?ooo`8000001P3oool010000000oooo0?ooo`3oool3000000<0oooo0P00
00050?ooo`<0000000<0oooo0000000000001@3oool200000080oooo00<000000?ooo`3oool00`3o
ool4000000<0oooo0P000000103oool000000?ooo`3oool3000000<0oooo00<000000?ooo`3oool0
<03oool000/0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`0000050?ooo`03
0000003oool0oooo00<0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0080oooo0P0000040?ooo`800000103oool01@000000oooo0?ooo`3oool00000
00@0oooo0P0000000`3oool000000?ooo`040?ooo`030000003oool0oooo0080oooo0P0000050?oo
o`050000003oool000000000003oool00P0000000`3oool00000000000060?ooo`80000000<0oooo
0000003oool00P0000000`3oool00000000000020?ooo`030000003oool0oooo0080oooo00<00000
0?ooo`3oool00P3oool2000000<0oooo0P0000020?ooo`050000003oool000000000003oool00P00
00001@3oool000000?ooo`3oool000000080oooo00H000000?ooo`00000000000?ooo`0000050?oo
o`<000000P3oool010000000oooo0?ooo`0000020?ooo`<0000000<0oooo0000000000000P3oool0
10000000oooo0000000000030?ooo`8000001@3oool00`000000oooo0?ooo`020?ooo`050000003o
ool000000000003oool00P0000020?ooo`@000001@3oool2000000040?ooo`00000000000?ooo`80
00000P3oool00`000000oooo0?ooo`020?ooo`8000001@3oool00`000000oooo0?ooo`030?ooo`<0
00001@3oool010000000oooo0?ooo`0000030?ooo`80000000<0oooo0000003oool00P3oool20000
00D0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo
03@0oooo000;0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool000001@3oool2
00000080oooo0P0000040?ooo`80000000@0oooo000000000000oooo0P0000000`3oool000000000
00040?ooo`800000103oool00`000000oooo0?ooo`030?ooo`800000103oool2000000030?ooo`00
0000oooo00H0oooo0P0000001@3oool000000000003oool0000000@0oooo0P0000020?ooo`040000
003oool0oooo00000080oooo00<000000?ooo`3oool00P3oool2000000030?ooo`000000oooo00<0
oooo0P0000000`3oool000000?ooo`02000000@0oooo0P0000040?ooo`060000003oool000000000
003oool00000103oool200000080oooo00<000000?ooo`3oool0203oool3000000@0oooo0P000000
0`3oool000000?ooo`090?ooo`040000003oool00000000000D0oooo0`0000020?ooo`040000003o
ool0oooo0?ooo`800000103oool200000080oooo100000040?ooo`<000001@3oool010000000oooo
0?ooo`0000020?ooo`040000003oool00000000000D0oooo0`0000000`3oool00000000000080?oo
o`<00000103oool2000000D0oooo00@000000?ooo`3oool0oooo0P0000000`3oool000000?ooo`02
0?ooo`80000000@0oooo000000000000oooo0P0000000`3oool00000000000040?ooo`80000000<0
oooo000000000000<P3oool000`0oooo0P0000040?ooo`8000001P3oool4000000030?ooo`000000
oooo00@0oooo00<000000?ooo`3oool00P000000103oool000000?ooo`3oool3000000<0oooo00<0
00000?ooo`3oool00`3oool2000000050?ooo`00000000000?ooo`0000000P3oool2000000030?oo
o`000000oooo00<0oooo0P0000040?ooo`80000000<0oooo0000003oool00`3oool2000000030?oo
o`000000000000d0oooo00@000000?ooo`0000000000103oool00`000000oooo0?ooo`020?ooo`05
0000003oool000000000003oool00P0000070?ooo`040000003oool0oooo0?ooo`8000000P3oool0
0`000000oooo0?ooo`040?ooo`8000001`3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00@0oooo0P0000070?ooo`<000001@3oool00`000000oooo0?ooo`040?ooo`80000000<0oooo
0000000000000P3oool00`000000oooo0?ooo`020?ooo`040000003oool0oooo0?ooo`8000000P3o
ool00`000000oooo0?ooo`0:0?ooo`D000001@3oool4000000<0oooo0P0000050?ooo`<000000`3o
ool3000000030?ooo`000000000000D0oooo00@000000?ooo`3oool0oooo0P000000103oool00000
0?ooo`3oool400000080oooo00<000000?ooo`3oool00`3oool400000380oooo000J0?ooo`800000
00<0oooo0000000000000P3oool2000000H0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`080?ooo`80000000<0oooo0000003oool0103oool01@000000oooo000000000000oooo0080
000000<0oooo000000000000103oool00`000000oooo0?ooo`020?ooo`80000000D0oooo0000003o
ool0oooo000000040?ooo`8000000`3oool2000000D0oooo0`0000000`3oool00000000000070?oo
o`050000003oool0oooo0000003oool00P0000000`3oool00000000000030?ooo`80000000<0oooo
0000003oool00P0000040?ooo`80000000<0oooo0000003oool00P0000000`3oool0000000000002
000000@0oooo100000000`3oool00000000000020?ooo`<00000103oool2000000030?ooo`000000
oooo008000000`3oool2000000030?ooo`00000000000080oooo00@000000?ooo`3oool000001P3o
ool400000080oooo00<000000?ooo`3oool00P3oool2000000<0oooo0P0000050?ooo`030000003o
ool0oooo00<000001@3oool010000000oooo0?ooo`3oool5000000030?ooo`00000000000080oooo
00<000000?ooo`3oool00`3oool4000000X0oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`030?ooo`80000000@0oooo000000000000oooo0P00000b0?ooo`006P3oool200000080oooo
00H000000?ooo`3oool0oooo0000003oool2000000040?ooo`00000000000?ooo`80000000@0oooo
000000000000oooo0P000000103oool000000000003oool2000000@0oooo00<000000?ooo`3oool0
0P3oool2000000@0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`020?ooo`80
000000D0oooo000000000000oooo000000040?ooo`800000103oool01P000000oooo000000000000
oooo000000@0oooo0P0000001P3oool000000?ooo`3oool000000?ooo`80000000<0oooo00000000
00000`3oool200000080oooo00<000000?ooo`3oool00`0000020?ooo`030000003oool0oooo0080
oooo00<000000?ooo`3oool00P3oool2000000<0oooo0P0000020?ooo`<00000103oool200000003
0?ooo`000000oooo008000000`3oool2000000D0oooo00<000000?ooo`3oool00P3oool300000080
oooo00D000000?ooo`3oool000000?ooo`02000000<0oooo0P0000000`3oool00000000000020?oo
o`040000003oool000000?ooo`8000001@3oool300000080oooo00@000000?ooo`3oool0oooo0P00
00050?ooo`<0000000<0oooo0000000000001@3oool010000000oooo0?ooo`0000020?ooo`040000
003oool0oooo0?ooo`8000001@3oool010000000oooo0?ooo`3oool2000000040?ooo`0000000000
0?ooo`80000000@0oooo000000000000oooo0P000000103oool000000000003oool200000080oooo
00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`0`0?ooo`006P3oool2000000<0oooo
0P0000040?ooo`80000000@0oooo0000003oool0oooo0P0000001@3oool000000?ooo`3oool00000
0080oooo00D000000?ooo`3oool000000?ooo`02000000040?ooo`00000000000?ooo`80000000<0
oooo0000000000002@3oool2000000@0oooo00D000000?ooo`3oool000000?ooo`02000000050?oo
o`00000000000?ooo`000000103oool2000000030?ooo`000000oooo00<0oooo0P0000000`3oool0
0000000000030?ooo`80000000D0oooo0000003oool0oooo000000040?ooo`030000003oool0oooo
0080oooo0P0000000`3oool000000?ooo`02000000<0oooo0P0000050?ooo`<000000P3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool00P3oool30000
0080oooo00@000000?ooo`3oool0oooo0P0000040?ooo`80000000<0oooo0000003oool00P000005
0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`0000030?ooo`D000001@3oool3
000000D0oooo0`0000000`3oool00000000000020?ooo`040000003oool0oooo000000<0oooo0P00
00090?ooo`D0000000@0oooo000000000000oooo0P0000020?ooo`040000003oool0oooo00000080
oooo00<000000?ooo`3oool0100000020?ooo`@000000`3oool2000003H0oooo000J0?ooo`800000
103oool2000000050?ooo`00000000000?ooo`0000001@3oool00`000000oooo0?ooo`020?ooo`80
0000103oool2000000040?ooo`000000oooo0?ooo`80000000H0oooo0000003oool0oooo0000003o
ool2000000040?ooo`00000000000?ooo`80000000<0oooo0000000000000`3oool2000000@0oooo
00@000000?ooo`3oool00000103oool2000000030?ooo`000000oooo00<0oooo0P0000001@3oool0
00000?ooo`3oool0000000<0oooo0P0000040?ooo`8000001P3oool2000000030?ooo`000000oooo
00<0oooo00D000000?ooo`00000000000?ooo`02000000<0oooo0P0000000`3oool000000?ooo`05
0?ooo`800000103oool2000000H0oooo00@000000?ooo`00000000000`3oool2000000030?ooo`00
000000000080oooo00<000000?ooo`3oool00P3oool3000000H0oooo0P0000030?ooo`<00000103o
ool00`000000oooo0?ooo`03000000D0oooo0`0000050?ooo`040000003oool0oooo000000<0oooo
0P0000030?ooo`D0000000@0oooo000000000000oooo0P0000020?ooo`030000003oool0oooo00@0
00000P3oool3000000@0oooo0P0000050?ooo`030000003oool0oooo00<0oooo0`000000103oool0
00000000000000030?ooo`800000<@3oool001X0oooo0P0000050?ooo`040000003oool0oooo0000
00L0oooo0P0000040?ooo`80000000D0oooo000000000000oooo000000050?ooo`030000003oool0
oooo00D0oooo00<000000?ooo`3oool00P0000001P3oool000000?ooo`3oool000000?ooo`800000
00<0oooo0000000000002@3oool2000000030?ooo`000000oooo00@0000000<0oooo000000000000
0`3oool2000000@0oooo00@000000?ooo`00000000000P3oool010000000oooo0?ooo`3oool20000
0080oooo0`0000000`3oool00000000000040?ooo`030000003oool0oooo00<000000P3oool00`00
0000oooo0?ooo`040?ooo`80000000<0oooo0000003oool0103oool3000000L0oooo00H000000?oo
o`00000000000?ooo`0000020?ooo`040000003oool0oooo0?ooo`80000000H0oooo000000000000
oooo0000003oool2000000@0oooo00H000000?ooo`00000000000?ooo`0000040?ooo`8000000`3o
ool4000000030?ooo`00000000000080oooo0`0000090?ooo`D000000P3oool00`000000oooo0?oo
o`0400000080oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`030?ooo`<0000000@0
oooo00000000000000000`3oool2000000P0oooo00@000000?ooo`3oool000001@3oool00`000000
oooo0?ooo`0_0?ooo`006P3oool010000000oooo0000000000040?ooo`8000002@3oool3000000<0
oooo00@000000?ooo`3oool00000103oool2000000030?ooo`000000000000@0oooo0P0000040?oo
o`030000003oool0oooo0080oooo0P0000001P3oool000000?ooo`3oool000000?ooo`80000000<0
oooo0000000000000P0000020?ooo`040000003oool0oooo0?ooo`80000000D0oooo0000003oool0
oooo000000040?ooo`030000003oool0oooo0080oooo0P0000000`3oool000000?ooo`030?ooo`80
000000<0oooo0000003oool00`3oool2000000030?ooo`000000oooo008000000`3oool200000003
0?ooo`000000oooo008000000`3oool200000080oooo0`0000040?ooo`80000000<0oooo0000003o
ool00P0000030?ooo`8000000P3oool3000000<0oooo0P0000020?ooo`@00000103oool3000000D0
oooo0`0000050?ooo`030000003oool0oooo0080oooo00@000000?ooo`000000oooo0P0000040?oo
o`030000003oool0000000<0000000<0oooo0000000000000P3oool00`000000oooo0?ooo`030000
00D0oooo00@000000?ooo`3oool0oooo0P0000040?ooo`80000000<0oooo0000000000001@3oool0
10000000oooo0?ooo`0000020?ooo`80000000<0oooo0000003oool01`3oool2000000@0oooo0P00
000c0?ooo`006P3oool2000000040?ooo`00000000000?ooo`80000000@0oooo000000000000oooo
0P0000070?ooo`030000003oool0oooo0080oooo0P0000030?ooo`80000000@0oooo0000003oool0
oooo0P0000040?ooo`040000003oool00000000000@0oooo0P0000000`3oool000000?ooo`030?oo
o`80000000P0oooo0000003oool0oooo0000003oool000000?ooo`80000000D0oooo000000000000
oooo0000000<0?ooo`80000000<0oooo0000003oool00P3oool200000080oooo00@000000?ooo`00
00000000103oool00`000000oooo0?ooo`020?ooo`<000000P3oool00`000000oooo0?ooo`020?oo
o`030000003oool0000000<000000`3oool2000000@0oooo0P0000030?ooo`@000000P3oool00`00
0000oooo0?ooo`020?ooo`<000000P3oool00`000000oooo0?ooo`020?ooo`040000003oool0oooo
0?ooo`8000000P3oool3000000D0oooo0`00000=0?ooo`<0000000<0oooo0000000000000P3oool0
0`000000oooo000000020?ooo`<000001@3oool3000000@0oooo0P0000020?ooo`040000003oool0
oooo0?ooo`<000000P3oool2000000@0oooo0P0000050?ooo`030000003oool0oooo00@0oooo0P00
00000`3oool0000000000003000000030?ooo`000000000000<00000<@3oool001X0oooo0P000002
0?ooo`040000003oool0oooo00000080oooo00D000000?ooo`3oool000000?ooo`02000000040?oo
o`00000000000?ooo`80000000@0oooo000000000000oooo0P0000000`3oool00000000000030000
00<0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`030?ooo`040000003oool00000
000000<0oooo0P0000030?ooo`800000103oool2000000030?ooo`000000oooo0080000000<0oooo
0000000000001@3oool200000080oooo0`0000001@3oool000000000003oool0000000@0oooo0P00
00030?ooo`8000000`3oool2000000030?ooo`000000oooo00<0oooo0P0000040?ooo`040000003o
ool000000?ooo`80000000<0oooo0000003oool0103oool300000080oooo00<000000?ooo`000000
0`3oool2000000@0oooo00@000000?ooo`00000000000`3oool2000000<0oooo0P0000040?ooo`03
0000003oool0000000<000000P3oool3000000H0oooo0P0000000`3oool00000000000020?ooo`@0
00000`3oool200000080oooo0`0000030?ooo`8000000P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00D0oooo00<000000?ooo`3oool01@0000000`3oool000000000000300000004
0?ooo`00000000000?ooo`80000000<0oooo0000000000000P3oool010000000oooo0?ooo`000002
0?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3oool0;`3oool001X0oooo00<00000
0?ooo`3oool00P3oool2000000@0oooo0P0000030?ooo`040000003oool0oooo00000080oooo0P00
00001@3oool000000?ooo`3oool000000080oooo00<000000?ooo`3oool00`000000103oool00000
0000003oool2000000030?ooo`000000000000<0oooo0P0000040?ooo`8000000`3oool01P000000
oooo000000000000oooo000000D0oooo00<000000?ooo`3oool00P3oool2000000050?ooo`000000
00000000003oool00P0000000`3oool000000?ooo`030?ooo`80000000<0oooo0000003oool00`3o
ool2000000030?ooo`000000oooo00<0oooo00H000000?ooo`00000000000?ooo`0000030?ooo`80
00000P3oool3000000L0oooo0`0000030?ooo`8000000P3oool00`000000oooo0?ooo`040?ooo`80
000000<0oooo0000003oool00P0000040?ooo`<000000P3oool00`000000oooo0?ooo`020?ooo`<0
00001@3oool3000000@0oooo00<000000?ooo`0000000`0000020?ooo`@00000103oool00`000000
oooo0?ooo`030?ooo`<0000000<0oooo0000003oool00P3oool3000000<0oooo0P0000030?ooo`80
000000@0oooo000000000000oooo100000020?ooo`040000003oool0oooo00000080oooo00<00000
0?ooo`3oool0100000020?ooo`040000003oool0oooo00000080oooo0P0000040?ooo`800000=P3o
ool001X0oooo0P000000103oool000000000003oool2000000040?ooo`00000000000?ooo`800000
00<0oooo0000000000001P3oool00`000000oooo0?ooo`020?ooo`8000001P3oool010000000oooo
0?ooo`0000020?ooo`050000003oool0oooo0000003oool00P000000103oool000000000003oool2
000000030?ooo`000000oooo0080oooo0P0000000`3oool000000?ooo`030?ooo`80000000<0oooo
000000000000103oool01@000000oooo0?ooo`3oool000000080oooo00D000000?ooo`0000000000
0?ooo`02000000030?ooo`000000oooo00H0oooo0`0000030?ooo`8000000P3oool010000000oooo
0?ooo`3oool2000000030?ooo`000000oooo00<0oooo00@000000?ooo`00000000001P3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo00@0oooo0P0000020?ooo`030000003oool0oooo
0080oooo0`0000040?ooo`040000003oool00000000000<0oooo0P0000000`3oool000000?ooo`06
0?ooo`<0000000<0oooo0000000000000P3oool00`000000oooo000000040?ooo`040000003oool0
oooo0?ooo`80000000<0oooo0000000000001@3oool3000000@0oooo0`0000000`3oool000000000
000300000080oooo00@000000?ooo`3oool000000P3oool00`000000oooo000000060?ooo`800000
1@3oool00`000000oooo0?ooo`030?ooo`80000000<0oooo0000000000000`0000000`3oool00000
00000003000000030?ooo`00000000000340oooo000J0?ooo`8000000P3oool010000000oooo0?oo
o`0000020?ooo`040000003oool0oooo00000080oooo00D000000?ooo`00000000000?ooo`040000
00040?ooo`00000000000?ooo`80000000<0oooo0000000000001@3oool2000000@0oooo0P000000
1@3oool000000?ooo`3oool000000080oooo0`0000001@3oool000000000003oool0000000@0oooo
0P0000001@3oool000000?ooo`3oool0000000<0oooo0P0000090?ooo`80000000@0oooo00000000
000000002@3oool01@000000oooo000000000000oooo008000000`3oool2000000030?ooo`000000
oooo008000000P3oool200000080oooo0`0000000`3oool0000000000002000000<0oooo0P000000
0`3oool000000000000200000080oooo0`0000030?ooo`8000000P3oool3000000<0oooo1@000002
0?ooo`030000003oool0oooo00D0oooo0P0000020?ooo`@00000203oool2000000@0oooo00<00000
0?ooo`3oool00`0000050?ooo`D000000P3oool010000000oooo0?ooo`0000020?ooo`<000000`3o
ool2000000H0oooo0P0000000`3oool0000000000003000000030?ooo`00000000000080000000<0
oooo0000000000000P3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo00000080
oooo00@000000?ooo`3oool00000<@3oool001X0oooo100000030?ooo`800000103oool2000000@0
oooo0P0000000`3oool000000?ooo`0400000080oooo00D000000?ooo`3oool000000?ooo`020000
00030?ooo`00000000000080000000D0oooo000000000000oooo000000090?ooo`80000000<0oooo
0000003oool00P000000103oool000000000003oool2000000P0oooo00<000000?ooo`3oool0203o
ool01P000000oooo0?ooo`3oool000000?ooo`80000000<0oooo0000000000000`3oool200000005
0?ooo`000000oooo0?ooo`000000103oool00`000000oooo0?ooo`020?ooo`040000003oool0oooo
0?ooo`8000000`3oool2000000040?ooo`0000000000000000<0oooo0P0000020?ooo`<000000`3o
ool2000000@0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`030000003o
ool0oooo00<000000`3oool2000000050?ooo`00000000000?ooo`000000103oool00`000000oooo
0?ooo`080?ooo`030000003oool0oooo00D0oooo1@0000000`3oool00000000000020?ooo`<00000
2P3oool8000000030?ooo`00000000000080oooo00@000000?ooo`3oool000000P3oool200000003
0?ooo`000000oooo00@000000`3oool2000000<0oooo0P0000040?ooo`800000<`3oool001X0oooo
0P0000000`3oool000000?ooo`030?ooo`030000003oool0oooo00<0oooo0P0000001@3oool00000
0000003oool0000000D0oooo00@000000?ooo`3oool0oooo0P0000040?ooo`060000003oool0oooo
0000003oool000000P3oool01@000000oooo000000000000oooo0080000000<0oooo000000000000
0`3oool00`000000oooo0?ooo`020?ooo`80000000@0oooo0000003oool0oooo0`0000000`3oool0
0000000000030?ooo`8000001@3oool2000000030?ooo`000000000000<0oooo0P000000103oool0
00000?ooo`3oool3000000030?ooo`000000000000<0oooo0P0000030?ooo`8000000`3oool20000
00@0oooo0`0000020?ooo`050000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`02
0?ooo`060000003oool000000000003oool000000`3oool2000000<0oooo0P0000030?ooo`800000
0`3oool400000080oooo00<000000?ooo`3oool00`0000030?ooo`80000000<0oooo000000000000
1@3oool2000000<0oooo0P0000001@3oool000000000003oool000000080oooo100000040?ooo`03
0000003oool0oooo0080000000@0oooo000000000000oooo0P0000020?ooo`030000003oool00000
0080oooo00@000000?ooo`3oool000000`3oool2000000@0oooo00<000000?ooo`3oool00`3oool5
000000030?ooo`000000oooo00<0oooo00<000000?ooo`3oool00`3oool300000340oooo000J0?oo
o`8000000P3oool2000000T0oooo00@000000?ooo`3oool00000103oool2000000030?ooo`000000
000000<0oooo0P0000080?ooo`800000103oool2000000040?ooo`000000oooo0?ooo`<0000000@0
oooo000000000000oooo0P0000000`3oool000000?ooo`050?ooo`050000003oool0oooo0000003o
ool0100000030?ooo`80000000D0oooo0000003oool0oooo000000030?ooo`8000000`3oool20000
00060?ooo`000000oooo0?ooo`000000oooo0P0000000`3oool000000?ooo`030?ooo`<0000000<0
oooo0000000000000`3oool2000000030?ooo`000000oooo00D0oooo0P0000020?ooo`8000001`3o
ool3000000<0oooo0P0000001@3oool000000000003oool0000000@0oooo0`0000020?ooo`040000
003oool000000?ooo`8000000`3oool2000000@0oooo00<000000?ooo`3oool00`0000020?ooo`D0
00000P3oool010000000oooo0?ooo`0000060?ooo`<0000000@0oooo000000000000oooo0P000000
103oool000000?ooo`3oool400000080oooo0P0000090?ooo`<000000P3oool2000000030?ooo`00
0000000000D0oooo00<000000?ooo`000000203oool2000000@0oooo00<000000?ooo`3oool0;`3o
ool001X0oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`020?ooo`800000103oool2
000000<0oooo0P000000103oool000000?ooo`3oool2000000<0oooo00<000000?ooo`3oool01@3o
ool4000000@0oooo00<000000?ooo`3oool00P3oool300000080oooo00D000000?ooo`3oool00000
0?ooo`04000000H0oooo0P0000001@3oool000000?ooo`000000oooo0080000000<0oooo00000000
00000`3oool2000000<0oooo00H000000?ooo`00000000000?ooo`0000030?ooo`8000000P3oool3
000000030?ooo`000000000000<0oooo00<000000?ooo`3oool00`0000020?ooo`030000003oool0
oooo00<0oooo0P0000020?ooo`060000003oool0oooo0?ooo`000000oooo0P0000060?ooo`060000
003oool000000000003oool000000P3oool00`000000oooo0?ooo`02000000060?ooo`0000000000
0?ooo`000000oooo0P0000040?ooo`<000000P3oool00`000000oooo0?ooo`02000000<0oooo1@00
0000103oool000000?ooo`3oool3000000H0oooo100000020?ooo`040000003oool0oooo00000080
oooo0P0000050?ooo`030000003oool0oooo00D00000203oool00`000000oooo0?ooo`0300000080
oooo0`0000030?ooo`800000103oool2000000@0oooo00<000000?ooo`3oool0=03oool001X0oooo
0P000000103oool000000000003oool2000000030?ooo`00000000000080000000@0oooo00000000
0000oooo0P000000103oool000000000000000050?ooo`040000003oool0oooo0?ooo`800000103o
ool2000000040?ooo`000000oooo000000L0oooo0P0000000`3oool000000?ooo`020?ooo`800000
0`3oool2000000030?ooo`000000oooo0080000000D0oooo000000000000oooo000000030?ooo`80
000000@0oooo0000003oool0oooo0`0000000`3oool00000000000030?ooo`80000000<0oooo0000
003oool00P3oool2000000030?ooo`000000oooo0080oooo0P0000020?ooo`<000000`3oool20000
00030?ooo`000000oooo00D0oooo0P0000000`3oool000000?ooo`0200000080oooo0P0000020?oo
o`<000001`3oool3000000<0oooo0P0000000`3oool0000000000002000000@0oooo0`0000040?oo
o`<000000`3oool400000080oooo00<000000?ooo`3oool00`0000040?ooo`<0000000<0oooo0000
000000000P3oool3000000<0oooo0P0000040?ooo`<000001`3oool00`000000oooo0?ooo`030000
00<0oooo0P0000050?ooo`030000003oool0oooo00<0oooo0P0000000`3oool00000000000030000
00030?ooo`00000000000080000000@0oooo000000000000oooo0P0000000`3oool000000000000a
0?ooo`006P3oool500000080oooo00H000000?ooo`3oool000000?ooo`0000020?ooo`040000003o
ool0oooo00000080oooo0`000000103oool000000000003oool2000000030?ooo`00000000000080
0000103oool010000000oooo0?ooo`3oool2000000@0oooo0P0000000`3oool000000?ooo`020?oo
o`80000000<0oooo0000000000000`3oool00`000000oooo0?ooo`020?ooo`80000000D0oooo0000
0000000000000?ooo`02000000030?ooo`000000oooo00D0oooo00D000000?ooo`3oool000000?oo
o`02000000030?ooo`000000oooo00D0oooo00@000000?ooo`00000000000`3oool00`000000oooo
0?ooo`020?ooo`060000003oool000000000003oool000000`3oool4000000030?ooo`0000000000
00<0oooo0P0000020?ooo`8000000`3oool2000000030?ooo`000000000000800000103oool30000
0080oooo00D000000?ooo`3oool0oooo000000030?ooo`8000000P3oool00`000000oooo0?ooo`04
0?ooo`<000000P3oool00`000000oooo000000070?ooo`<0000000@0oooo000000000000oooo1000
00040?ooo`030000003oool0oooo00D00000103oool3000000@0oooo0P0000040?ooo`030000003o
ool0oooo00D0000000@0oooo000000000000oooo100000020?ooo`040000003oool0oooo00000080
oooo00H000000?ooo`3oool000000?ooo`0000020?ooo`030000003oool0oooo0080000000<0oooo
0000003oool0<03oool001X0oooo00@000000?ooo`3oool000002@3oool2000000<0oooo0P000003
0?ooo`040000003oool0oooo00000080oooo0P0000001@3oool000000?ooo`000000oooo00800000
00@0oooo000000000000oooo100000040?ooo`040000003oool0oooo0?ooo`80000000D0oooo0000
003oool0oooo000000030?ooo`800000103oool010000000oooo0?ooo`3oool2000000030?ooo`00
0000oooo00@000001P3oool2000000050?ooo`00000000000000003oool00P0000050?ooo`800000
1`3oool200000080oooo0`0000001`3oool000000000003oool000000?ooo`0000000P3oool30000
0080oooo00H000000?ooo`3oool0oooo0000003oool2000000060?ooo`000000oooo0?ooo`000000
oooo0`000000103oool000000000003oool3000000P0oooo0P0000050?ooo`80000000@0oooo0000
00000000oooo0`0000060?ooo`@000000P3oool4000000<0oooo00@000000?ooo`3oool0oooo0P00
00040?ooo`030000003oool0oooo00<00000103oool010000000oooo0?ooo`3oool4000000040?oo
o`00000000000?ooo`80000000@0oooo0000003oool0oooo100000020?ooo`030000003oool00000
00<0oooo0P0000030?ooo`8000002P3oool00`000000oooo0?ooo`0a0?ooo`006P3oool200000004
0?ooo`00000000000?ooo`@0000000@0oooo000000000000oooo0P0000000`3oool0000000000002
000000030?ooo`000000000000H0oooo00<000000?ooo`3oool00P3oool300000080oooo00L00000
0?ooo`3oool000000?ooo`000000oooo0080000000<0oooo0000000000000P0000000`3oool00000
0?ooo`030?ooo`8000000`3oool00`000000oooo0?ooo`040?ooo`80000000<0oooo0000003oool0
0P3oool2000000030?ooo`000000oooo00@000000P3oool01@000000oooo0?ooo`3oool000000080
oooo0`0000001@3oool000000000003oool0000000D0oooo0P0000000`3oool000000?ooo`030?oo
o`030000003oool0oooo00<000001P3oool010000000oooo0000000000020?ooo`8000000P3oool0
0`000000oooo0?ooo`030?ooo`80000000D0oooo0000003oool0oooo000000030?ooo`040000003o
ool00000000000H0oooo0`0000001@3oool000000000003oool000000080oooo00@000000?ooo`3o
ool0oooo0`0000000`3oool00000000000020?ooo`<00000103oool3000000L0oooo00<000000?oo
o`3oool00P0000050?ooo`D0000000<0oooo0000000000000P3oool00`000000oooo000000020?oo
o`030000003oool0oooo008000001@3oool00`000000oooo0?ooo`030?ooo`80000000<0oooo0000
000000000P0000000`3oool0000000000005000000040?ooo`00000000000?ooo`80000000<0oooo
000000000000<@3oool001X0oooo0P0000020?ooo`80000000<0oooo0000003oool0100000020?oo
o`060000003oool0oooo0000003oool000000P3oool00`000000oooo00000003000000040?ooo`00
000000000?ooo`80000000<0oooo0000000000001`3oool2000000080?ooo`000000oooo0?ooo`00
0000oooo0000003oool2000000030?ooo`0000000000008000000`3oool2000000<0oooo0P000000
0`3oool000000?ooo`050?ooo`040000003oool0oooo0?ooo`80000000<0oooo0000003oool00P00
00000`3oool00000000000050?ooo`030000003oool0oooo00<000000`3oool2000000030?ooo`00
0000oooo0080oooo0P0000060?ooo`040000003oool00000000000H0oooo00D000000?ooo`3oool0
oooo000000050?ooo`8000000P3oool00`000000oooo0?ooo`030?ooo`800000103oool3000000<0
oooo0P0000001@3oool000000?ooo`3oool0000000@0oooo0P0000001P3oool000000000003oool0
00000?ooo`800000103oool00`000000oooo0?ooo`040?ooo`<000000`3oool2000000<0oooo1000
00000`3oool00000000000020?ooo`030000003oool000000080oooo0`0000070?ooo`D0000000<0
oooo0000000000000P0000000`3oool00000000000020?ooo`060000003oool0oooo0000003oool0
00000P3oool00`000000oooo0?ooo`0300000080oooo100000020?ooo`030000003oool0oooo02l0
oooo000J0?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool00`3oool010000000oooo
0?ooo`3oool2000000<0oooo0P0000040?ooo`030000003oool000000080oooo0P000000103oool0
00000?ooo`3oool3000000040?ooo`00000000000?ooo`@000000`3oool2000000<0oooo00L00000
0?ooo`3oool000000?ooo`000000oooo0080000000<0oooo0000000000000`3oool010000000oooo
0?ooo`3oool2000000<0oooo100000000`3oool000000?ooo`020?ooo`8000000P3oool300000003
0?ooo`000000000000D0oooo00<000000?ooo`0000000`0000030?ooo`80000000<0oooo0000003o
ool00P0000020?ooo`8000000P3oool5000000030?ooo`00000000000080oooo0P0000000`3oool0
000000000002000000030?ooo`000000oooo008000000P3oool200000080oooo0`0000030?ooo`@0
00000P3oool00`000000oooo0?ooo`020?ooo`80000000@0oooo000000000000oooo100000040?oo
o`D000000`3oool2000000@0oooo00<000000?ooo`3oool01@0000020?ooo`030000003oool00000
0080oooo00<000000?ooo`3oool00P0000030?ooo`@0000000@0oooo000000000000oooo0P000000
103oool000000?ooo`3oool4000000050?ooo`000000oooo0?ooo`0000000P3oool2000000@0oooo
0P0000040?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0<`3oool001X0oooo0P00
00040?ooo`@00000103oool4000000<0oooo1@0000060?ooo`030000003oool0oooo0080oooo0`00
00020?ooo`80000000D0oooo0000003oool000000?ooo`04000000030?ooo`000000000000D0oooo
0P000000103oool000000?ooo`3oool3000000030?ooo`000000000000<0oooo00@000000?ooo`3o
ool0oooo0P0000000`3oool000000?ooo`050?ooo`030000003oool0oooo0080oooo00<000000?oo
o`3oool00`0000030?ooo`80000000D0oooo0000003oool000000?ooo`02000000030?ooo`000000
oooo00<0oooo00<000000?ooo`3oool00P0000030?ooo`80000000<0oooo0000003oool00P000002
0?ooo`8000000P3oool3000000<0oooo0P0000030?ooo`030000003oool0oooo0080oooo0`000002
0?ooo`030000003oool00000008000000`3oool2000000050?ooo`000000oooo0?ooo`0000000`3o
ool00`000000oooo0?ooo`050?ooo`<00000103oool00`000000oooo0?ooo`02000000050?ooo`00
000000000?ooo`0000000P3oool3000000D0oooo1`0000020?ooo`030000003oool0000000<00000
0P3oool2000000D0oooo00<000000?ooo`3oool00`3oool5000000<0oooo100000040?ooo`@00000
103oool200000340oooo000J0?ooo`@000000`3oool300000080oooo0P000000103oool000000?oo
o`0000050?ooo`030000003oool0000000@0oooo100000030?ooo`800000103oool010000000oooo
0?ooo`3oool2000000040?ooo`000000oooo00000080oooo00<000000?ooo`0000000`0000000`3o
ool000000?ooo`020?ooo`80000000@0oooo0000003oool0oooo0`0000000`3oool0000000000003
0?ooo`040000003oool0oooo0?ooo`8000000`3oool200000080oooo0P0000040?ooo`030000003o
ool0000000<000000`3oool2000000040?ooo`0000000000000000<0oooo0P0000030?ooo`060000
003oool000000000003oool00000103oool01@000000oooo0?ooo`3oool0000000@0oooo0`000002
0?ooo`030000003oool0oooo008000000`3oool2000000040?ooo`0000000000000000<0oooo1000
00020?ooo`030000003oool0oooo008000000`3oool200000080oooo0P0000050?ooo`030000003o
ool0oooo0080000000D0oooo000000000000oooo000000020?ooo`<00000103oool500000080oooo
00<000000?ooo`0000000P3oool3000000@0oooo00<000000?ooo`3oool00`0000030?ooo`@00000
1@3oool00`000000oooo000000050?ooo`030000003oool00000008000000`3oool300000080oooo
0P0000000`3oool000000?ooo`0`0?ooo`006P3oool010000000oooo0?ooo`0000040?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool01`3oool2000000<0oooo0P000000103oool00000
0?ooo`0000040?ooo`040000003oool00000000000<0oooo0P0000000`3oool000000?ooo`080?oo
o`040000003oool000000?ooo`@0000000<0oooo0000000000001@3oool00`000000oooo0?ooo`03
000000030?ooo`000000000000<0oooo00<000000?ooo`3oool00P3oool010000000oooo0?ooo`3o
ool200000080oooo0P0000020?ooo`030000003oool00000008000000P3oool01P000000oooo0?oo
o`3oool000000?ooo`@000001P3oool00`000000oooo0?ooo`030?ooo`8000000P3oool2000000H0
oooo00<000000?ooo`3oool00`3oool400000080oooo00@000000?ooo`3oool0oooo0`0000020?oo
o`030000003oool00000008000000`3oool2000000@0oooo00D000000?ooo`3oool0oooo00000003
0?ooo`80000000D0oooo000000000000oooo000000020?ooo`030000003oool0oooo0080oooo1`00
00020?ooo`030000003oool0000000T0oooo0`0000030?ooo`80000000<0oooo0000003oool0103o
ool00`000000oooo00000002000000<0oooo0P00000:0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool00P3oool00`000000oooo0?ooo`0a0?ooo`006P3oool2000000030?ooo`000000
00000080000000@0oooo000000000000oooo10000000103oool000000000003oool4000000030?oo
o`000000000000<0oooo00@000000?ooo`3oool0oooo0P0000030?ooo`80000000<0oooo0000003o
ool00`3oool00`000000oooo0?ooo`040?ooo`800000203oool300000080oooo0`0000000`3oool0
0000000000020?ooo`80000000@0oooo0000003oool0oooo0`0000000`3oool00000000000020?oo
o`800000103oool3000000030?ooo`000000oooo00D0oooo00<000000?ooo`0000000`0000040?oo
o`030000003oool00000008000001`3oool200000080oooo00H000000?ooo`3oool0oooo0000003o
ool2000000L0oooo0P000000103oool000000?ooo`0000030?ooo`@0000000<0oooo0000003oool0
1@3oool01P000000oooo000000000000oooo000000<0oooo0P0000020?ooo`@000000P3oool00`00
0000oooo0?ooo`02000000<0oooo100000020?ooo`030000003oool0oooo00<000001`3oool20000
00L0oooo00<000000?ooo`3oool00P3oool2000000@0oooo0P0000040?ooo`030000003oool0oooo
00L0000000@0oooo000000000000oooo100000000`3oool0000000000002000000040?ooo`000000
00000?ooo`800000<@3oool001X0oooo0P0000020?ooo`<000000P3oool2000000040?ooo`000000
oooo00000080oooo00H000000?ooo`3oool000000?ooo`0000020?ooo`<0000000@0oooo00000000
0000oooo10000000103oool000000000000000040?ooo`8000000`3oool2000000<0oooo00@00000
0?ooo`00000000000`3oool2000000030?ooo`000000oooo0080oooo0`0000020?ooo`030000003o
ool0000000<000000P3oool2000000040?ooo`000000oooo0?ooo`<0000000<0oooo000000000000
0P3oool2000000030?ooo`000000oooo00T0oooo0P0000000`3oool000000?ooo`0200000080oooo
0P0000000`3oool000000?ooo`0200000080oooo0P0000000`3oool000000?ooo`0200000080oooo
0P0000020?ooo`<000000P3oool200000080oooo0`0000020?ooo`8000000P3oool3000000/0oooo
0`0000020?ooo`D0000000<0oooo0000003oool00`0000030?ooo`@000000P3oool00`000000oooo
00000002000000@0oooo0`0000030?ooo`8000000P3oool00`000000oooo0?ooo`02000000<0oooo
0P0000040?ooo`<0000000<0oooo0000000000000P0000000`3oool000000000000200000080oooo
00H000000?ooo`3oool000000?ooo`0000020?ooo`050000003oool0oooo0000003oool010000000
0`3oool000000000000200000080oooo00<000000?ooo`3oool0;`3oool000D0oooo100000030?oo
o`800000103oool2000000H0oooo0P000000103oool000000000000000040?ooo`030000003oool0
oooo00L0oooo0P0000030?ooo`<000000P3oool01P000000oooo0?ooo`000000oooo00000080oooo
0`000000103oool000000000003oool4000000030?ooo`000000oooo00D0oooo00<000000?ooo`3o
ool00P3oool3000000@0oooo00@000000?ooo`3oool0oooo0P0000001@3oool000000?ooo`000000
oooo00@000001@3oool00`000000oooo0?ooo`03000000030?ooo`00000000000080oooo0P000003
0?ooo`8000000`3oool00`000000oooo0?ooo`020?ooo`<0000000<0oooo0000003oool00P3oool6
000000<0oooo0P0000020?ooo`8000000`3oool6000000@0oooo0`0000000`3oool000000?ooo`03
0?ooo`030000003oool0oooo008000000`3oool200000080oooo1@0000001@3oool000000?ooo`3o
ool0000000@0oooo100000020?ooo`030000003oool0000000800000103oool010000000oooo0?oo
o`3oool2000000030?ooo`000000oooo00<0oooo00<000000?ooo`3oool0103oool8000000050?oo
o`00000000000?ooo`0000000P3oool00`000000oooo000000020?ooo`030000003oool000000080
00000`3oool2000000X0oooo00<000000?ooo`3oool00P3oool2000000030?ooo`00000000000340
oooo00070?ooo`050000003oool0oooo0?ooo`0000000P3oool010000000oooo0?ooo`0000020?oo
o`030000003oool0oooo00<0oooo0P0000020?ooo`040000003oool0oooo0?ooo`80000000<0oooo
0000000000000P0000000`3oool00000000000030?ooo`8000000`3oool00`000000oooo0?ooo`06
0?ooo`8000000`3oool010000000oooo0?ooo`0000020?ooo`040000003oool000000?ooo`@00000
1@3oool2000000<0oooo00@000000?ooo`3oool0oooo0P0000001@3oool000000000003oool00000
00H0oooo00<000000?ooo`0000000P0000050?ooo`80000000@0oooo0000003oool0oooo0`000000
0`3oool000000?ooo`030?ooo`030000003oool0oooo008000000`3oool2000000040?ooo`000000
0000000000@0oooo00D000000?ooo`000000oooo000000040?ooo`050000003oool0oooo0?ooo`00
0000103oool01@000000oooo0000003oool0000000<0oooo0P000000103oool00000000000000003
0?ooo`8000000`3oool00`000000oooo0?ooo`020?ooo`<0000000@0oooo0000003oool0oooo0`00
00040?ooo`8000000P3oool00`000000oooo000000060?ooo`<0000000<0oooo000000000000103o
ool00`000000oooo0?ooo`02000000D0oooo100000020?ooo`030000003oool000000080oooo00@0
00000?ooo`3oool000000P3oool2000000T0oooo00<000000?ooo`3oool00`0000020?ooo`800000
00<0oooo0000000000000P0000000`3oool00000000000040?ooo`040000003oool0oooo00000340
oooo00070?ooo`050000003oool0oooo0?ooo`0000000P3oool010000000oooo0?ooo`0000020?oo
o`030000003oool0oooo00<0oooo0P0000070?ooo`060000003oool0oooo0000003oool000000P3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00/0oooo00<000000?ooo`3oool0303o
ool00`000000oooo000000060?ooo`030000003oool0oooo00H0oooo00@000000?ooo`3oool00000
303oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
203oool00`000000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`0000007`3oool01@000000
oooo0?ooo`3oool0000000@0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool02`3oool010000000oooo0?ooo`000008
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`0000003P3oool00`000000oooo0?ooo`0;
0?ooo`030000003oool0oooo0080oooo00H000000?ooo`3oool000000?ooo`0000020?ooo`030000
003oool0oooo03L0oooo00070?ooo`050000003oool0oooo0?ooo`0000000P3oool010000000oooo
0?ooo`0000020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0o`3ooolm0?ooo`00
1`3oool01@000000oooo0?ooo`3oool000000080oooo00@000000?ooo`3oool000000P3oool00`00
0000oooo0?ooo`030?ooo`030000003oool0oooo0?l0oooo?@3oool000D0oooo0`0000030?ooo`04
0000003oool0oooo00000080oooo00@000000?ooo`3oool000001@3oool00`000000oooo0?ooo`3o
0?ooocd0oooo00070?ooo`030000003oool0oooo0080oooo0P0000040?ooo`8000001P3oool00`00
0000oooo0?ooo`3o0?ooocd0oooo003o0?oooeX0oooo003o0?oooeX0oooo003o0?oooeX0oooo003o
0?oooeX0oooo003o0?oooeX0oooo003o0?oooeX0oooo001S0?ooo`<00000>03oool300000;P0oooo
001V0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0]`3oool005H0oooo0P000002
0?ooo`D0000000<0oooo0000000000000P3oool300000080oooo1@0000020?ooo`<000002P3oool0
10000000oooo0?ooo`3oool3000000@0oooo00<000000?ooo`3oool01`3oool300000080oooo00<0
00000?ooo`3oool00`0000020?ooo`<00000103oool3000000<0oooo0P0000000`3oool000000?oo
o`020?ooo`8000001`3oool4000000<0oooo1@000000103oool000000000000000050?ooo`800000
0`3oool2000000030?ooo`00000000000080000000<0oooo0000000000000P3oool2000000@0oooo
0`00001^0?ooo`00E`3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo00000080
oooo00@000000?ooo`3oool00000103oool01@000000oooo0?ooo`3oool0000000`0oooo00<00000
0?ooo`0000001@3oool00`000000oooo0?ooo`0<0?ooo`040000003oool0oooo00000080oooo00<0
00000?ooo`3oool00`3oool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo00<0oooo
00@000000?ooo`3oool000000`3oool010000000oooo0?ooo`0000070?ooo`040000003oool0oooo
000000@0oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3oool000000080oooo00<0
00000?ooo`0000000P3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo00000080
oooo00@000000?ooo`3oool00000L@3oool005L0oooo100000020?ooo`040000003oool0oooo0000
0080oooo00@000000?ooo`3oool00000103oool010000000oooo0?ooo`3oool3000000X0oooo00<0
00000?ooo`0000000`3oool2000000l0oooo0`0000030?ooo`<000000`3oool010000000oooo0?oo
o`0000020?ooo`<00000103oool3000000<0oooo00<000000?ooo`3oool0203oool010000000oooo
0?ooo`0000040?ooo`030000003oool0oooo0080oooo0P0000040?ooo`030000003oool0oooo00<0
oooo0`0000020?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3oool00`3oool30000
06l0oooo00020?ooo`@000000`3oool5000000040?ooo`0000000000000000D0oooo0P0000030?oo
o`80000000<0oooo0000000000000P0000000`3oool00000000000020?ooo`800000103oool30000
02H0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`0000020?ooo`040000003o
ool0oooo000000@0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003oool0oooo00H0
oooo00L000000?ooo`3oool0oooo0000003oool000000140oooo00@000000?ooo`3oool000000P3o
ool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3o
ool000001@3oool01@000000oooo0?ooo`3oool0000000X0oooo00@000000?ooo`3oool00000103o
ool01@000000oooo0?ooo`3oool0000000H0oooo00<000000?ooo`3oool01@3oool010000000oooo
0?ooo`0000020?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool000000P3oool0
0`000000oooo0?ooo`1/0?ooo`000`3oool010000000oooo0?ooo`0000040?ooo`030000003oool0
oooo00@0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003oool000000080oooo00@0
00000?ooo`3oool000000P3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo0000
02X0oooo00H000000?ooo`000000oooo0000003oool2000000@0oooo0P0000000`3oool000000?oo
o`020?ooo`030000003oool0oooo0080oooo0P0000080?ooo`8000000P3oool200000080oooo0`00
000>0?ooo`040000003oool0oooo000000<0oooo0P0000020?ooo`040000003oool00000000000@0
oooo0P0000040?ooo`8000000`3oool4000000P0oooo00@000000?ooo`3oool000000`3oool20000
00@0oooo0`0000020?ooo`@000000`3oool200000080oooo00@000000?ooo`0000000000103oool2
000000@0oooo0P00001_0?ooo`000`3oool010000000oooo0?ooo`0000040?ooo`030000003oool0
oooo0080oooo0P0000040?ooo`030000003oool0oooo00<0oooo0`0000020?ooo`040000003oool0
oooo00000080oooo00<000000?ooo`3oool00`3oool3000002L0oooo0`00000A0?ooo`030000003o
ool0oooo02@0oooo1000000L0?ooo`030000003oool0oooo00L0oooo1000000@0?ooo`030000003o
ool0oooo08H0oooo00030?ooo`040000003oool0oooo000000@0oooo00D000000?ooo`3oool0oooo
000000060?ooo`030000003oool0oooo00D0oooo00@000000?ooo`3oool000000P3oool010000000
oooo0?ooo`0000020?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3oool0=`3oool2
000004H0oooo00<000000?ooo`3oool0403oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo08H0oooo00030?ooo`040000003oool0oooo000000<0oooo0P0000040?ooo`<000000P3oool4
000000<0oooo0P0000020?ooo`040000003oool00000000000@0oooo0P0000040?ooo`800000o`3o
oolZ0?ooo`000P3oool400000100oooo00<000000?ooo`3oool0o`3ooom10?ooo`002`3oool00`00
0000oooo0?ooo`080?ooo`030000003oool0oooo0?l0oooo@@3oool00?l0ooooFP3oool00?l0oooo
FP3oool00?l0ooooFP3oool00?l0ooooFP3oool00001\
\>"],
  ImageRangeCache->{{{0, 344}, {212.125, 0}} -> {-5.77482, -10.8741, 
  0.222097, 0.598845}}]
}, Open  ]]
}, Open  ]],

Cell[BoxData[
    \(AnglePairsAndRationalApproxs[SizeLimit_, StrainLimit_] := 
      Module[{AnglesAndApproxsList, AnglesAndRepeatsList, i, j, Approx, 
          Size}, \n\t\tAnglesAndApproxsList = {}; \n\t\t
        AnglesAndRepeatsList = Sort[FindAngles[SizeLimit]]; \n\t\t\n\t\t
        For[i = 1, \n\t\[Theta]1 = AnglesAndRepeatsList[\([i, 1]\)]; \n\t
            X1 = AnglesAndRepeatsList[\([i, 3]\)]; \n
            i < Length[AnglesAndRepeatsList]\t, \(i++\), \n\t\n\t
            For[j = 1, \n\t\t\[Theta]2 = AnglesAndRepeatsList[\([j, 1]\)]; 
              \t\t\t\t\t\n\t\tX2 = AnglesAndRepeatsList[\([j, 3]\)]; \n\t\t
              \[Theta]2 <= \[Theta]1\  && 
                \[Theta]2 <= \((60 - \[Theta]1 + 0.5)\), \n\t\t\(j++\), \n
              \t\t\t\t\t\t\t\t\n\t\t\t\t
              Approx = \t\tRationalize[X1/X2,  .05]; \n
              Size\  = \ 
                \((Numerator[Approx]*X2 + Denominator[Approx]*X1)\)/2; \n
              If[Approx != 0, 
                Strain\  = \ 
                  Abs[\((1/2)\) \((\((X1/X2)\)/\((N[Approx])\) - 1)\)], 
                Strain = 1]; \n\t\t\t\t\t\t\t\t\t
              If[Size < SizeLimit\ \  && \ Strain < \ StrainLimit, \n
                \t\t\t\t\tMI1 = AnglesAndRepeatsList[\([i, 2]\)]; \n\t\t\t\t\t
                MI2 = AnglesAndRepeatsList[\([j, 2]\)]; \n\t\t\t\t\t\t
                AnglesAndApproxsList = 
                  Append[AnglesAndApproxsList, {{\[Theta]1, \[Theta]2}, MI1, 
                      X1, MI2, X2, Approx, Size, Strain}]; \n\t\t\t\t\t\t\n
                \t\t\t\t] (*end\ of\ If\ Size\  < \ SizeLimit*) \n\t\t\t\t] (*
              end\ of\ For\ j*) \n\t\t\t] (*end\ of\ For\ i*) \n\t\t
          AnglesAndApproxsList; \n\t\tAnglesAndApproxsList\n\t\t]\)], "Input"],

Cell[BoxData[
    \(WriteData[anglefile_, angledata_] := 
      Module[{F, L, i}, \nF = OpenWrite[anglefile, PageWidth -> 200]; \n\t\t\t
        Write[F, Length[angledata]]; \n\t\t\t\t
        For[i = 1, i < Length[angledata], \(i++\), \n
            L = Flatten[angledata[\([i]\)]]; \nWrite[F, L]; \n\t\t\t]\n
          Close[F]; \n\t\t]\)], "Input"],

Cell[CellGroupData[{

Cell[BoxData[{
    \(X = AnglePairsAndRationalApproxs[10, 10^\(-3\)]; \nLength[X]\), 
    \(Angles = \(Transpose[X]\)[\([1]\)]; \n
    ListPlot[Angles, \ AspectRatio -> 1/2, Frame -> True, \ Axes -> False, \ 
      PlotRange -> {{\(-5\), 65}, {\(-5\), 35}}]; \n
    WriteData["\<~/ASP/MDWebServices/GrainBreaker/2DGeometries/Geo10.dat\>", 
      X]; WriteData[
      "\<~/ASP/MDWebServices/GrainBreaker/2DGeometries/Geo10forplot.dat\>", 
      Angles]; \)}], "Input"],

Cell[BoxData[
    \(41\)], "Output"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .5 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0714286 0.0142857 0.0625 0.0125 [
[.07143 -0.0125 -3 -9 ]
[.07143 -0.0125 3 0 ]
[.21429 -0.0125 -6 -9 ]
[.21429 -0.0125 6 0 ]
[.35714 -0.0125 -6 -9 ]
[.35714 -0.0125 6 0 ]
[.5 -0.0125 -6 -9 ]
[.5 -0.0125 6 0 ]
[.64286 -0.0125 -6 -9 ]
[.64286 -0.0125 6 0 ]
[.78571 -0.0125 -6 -9 ]
[.78571 -0.0125 6 0 ]
[.92857 -0.0125 -6 -9 ]
[.92857 -0.0125 6 0 ]
[ 0 0 -0.125 0 ]
[-0.0125 .0625 -6 -4.5 ]
[-0.0125 .0625 0 4.5 ]
[-0.0125 .125 -6 -4.5 ]
[-0.0125 .125 0 4.5 ]
[-0.0125 .1875 -12 -4.5 ]
[-0.0125 .1875 0 4.5 ]
[-0.0125 .25 -12 -4.5 ]
[-0.0125 .25 0 4.5 ]
[-0.0125 .3125 -12 -4.5 ]
[-0.0125 .3125 0 4.5 ]
[-0.0125 .375 -12 -4.5 ]
[-0.0125 .375 0 4.5 ]
[-0.0125 .4375 -12 -4.5 ]
[-0.0125 .4375 0 4.5 ]
[-0.0125 .5 -12 -4.5 ]
[-0.0125 .5 0 4.5 ]
[ 0 0 -0.125 0 ]
[ 0 .5 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .5 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.07143 0 m
.07143 .00625 L
s
[(0)] .07143 -0.0125 0 1 Mshowa
.21429 0 m
.21429 .00625 L
s
[(10)] .21429 -0.0125 0 1 Mshowa
.35714 0 m
.35714 .00625 L
s
[(20)] .35714 -0.0125 0 1 Mshowa
.5 0 m
.5 .00625 L
s
[(30)] .5 -0.0125 0 1 Mshowa
.64286 0 m
.64286 .00625 L
s
[(40)] .64286 -0.0125 0 1 Mshowa
.78571 0 m
.78571 .00625 L
s
[(50)] .78571 -0.0125 0 1 Mshowa
.92857 0 m
.92857 .00625 L
s
[(60)] .92857 -0.0125 0 1 Mshowa
.125 Mabswid
.1 0 m
.1 .00375 L
s
.12857 0 m
.12857 .00375 L
s
.15714 0 m
.15714 .00375 L
s
.18571 0 m
.18571 .00375 L
s
.24286 0 m
.24286 .00375 L
s
.27143 0 m
.27143 .00375 L
s
.3 0 m
.3 .00375 L
s
.32857 0 m
.32857 .00375 L
s
.38571 0 m
.38571 .00375 L
s
.41429 0 m
.41429 .00375 L
s
.44286 0 m
.44286 .00375 L
s
.47143 0 m
.47143 .00375 L
s
.52857 0 m
.52857 .00375 L
s
.55714 0 m
.55714 .00375 L
s
.58571 0 m
.58571 .00375 L
s
.61429 0 m
.61429 .00375 L
s
.67143 0 m
.67143 .00375 L
s
.7 0 m
.7 .00375 L
s
.72857 0 m
.72857 .00375 L
s
.75714 0 m
.75714 .00375 L
s
.81429 0 m
.81429 .00375 L
s
.84286 0 m
.84286 .00375 L
s
.87143 0 m
.87143 .00375 L
s
.9 0 m
.9 .00375 L
s
.04286 0 m
.04286 .00375 L
s
.01429 0 m
.01429 .00375 L
s
.95714 0 m
.95714 .00375 L
s
.98571 0 m
.98571 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
0 .0625 m
.00625 .0625 L
s
[(0)] -0.0125 .0625 1 0 Mshowa
0 .125 m
.00625 .125 L
s
[(5)] -0.0125 .125 1 0 Mshowa
0 .1875 m
.00625 .1875 L
s
[(10)] -0.0125 .1875 1 0 Mshowa
0 .25 m
.00625 .25 L
s
[(15)] -0.0125 .25 1 0 Mshowa
0 .3125 m
.00625 .3125 L
s
[(20)] -0.0125 .3125 1 0 Mshowa
0 .375 m
.00625 .375 L
s
[(25)] -0.0125 .375 1 0 Mshowa
0 .4375 m
.00625 .4375 L
s
[(30)] -0.0125 .4375 1 0 Mshowa
0 .5 m
.00625 .5 L
s
[(35)] -0.0125 .5 1 0 Mshowa
.125 Mabswid
0 .0125 m
.00375 .0125 L
s
0 .025 m
.00375 .025 L
s
0 .0375 m
.00375 .0375 L
s
0 .05 m
.00375 .05 L
s
0 .075 m
.00375 .075 L
s
0 .0875 m
.00375 .0875 L
s
0 .1 m
.00375 .1 L
s
0 .1125 m
.00375 .1125 L
s
0 .1375 m
.00375 .1375 L
s
0 .15 m
.00375 .15 L
s
0 .1625 m
.00375 .1625 L
s
0 .175 m
.00375 .175 L
s
0 .2 m
.00375 .2 L
s
0 .2125 m
.00375 .2125 L
s
0 .225 m
.00375 .225 L
s
0 .2375 m
.00375 .2375 L
s
0 .2625 m
.00375 .2625 L
s
0 .275 m
.00375 .275 L
s
0 .2875 m
.00375 .2875 L
s
0 .3 m
.00375 .3 L
s
0 .325 m
.00375 .325 L
s
0 .3375 m
.00375 .3375 L
s
0 .35 m
.00375 .35 L
s
0 .3625 m
.00375 .3625 L
s
0 .3875 m
.00375 .3875 L
s
0 .4 m
.00375 .4 L
s
0 .4125 m
.00375 .4125 L
s
0 .425 m
.00375 .425 L
s
0 .45 m
.00375 .45 L
s
0 .4625 m
.00375 .4625 L
s
0 .475 m
.00375 .475 L
s
0 .4875 m
.00375 .4875 L
s
.25 Mabswid
0 0 m
0 .5 L
s
.07143 .49375 m
.07143 .5 L
s
.21429 .49375 m
.21429 .5 L
s
.35714 .49375 m
.35714 .5 L
s
.5 .49375 m
.5 .5 L
s
.64286 .49375 m
.64286 .5 L
s
.78571 .49375 m
.78571 .5 L
s
.92857 .49375 m
.92857 .5 L
s
.125 Mabswid
.1 .49625 m
.1 .5 L
s
.12857 .49625 m
.12857 .5 L
s
.15714 .49625 m
.15714 .5 L
s
.18571 .49625 m
.18571 .5 L
s
.24286 .49625 m
.24286 .5 L
s
.27143 .49625 m
.27143 .5 L
s
.3 .49625 m
.3 .5 L
s
.32857 .49625 m
.32857 .5 L
s
.38571 .49625 m
.38571 .5 L
s
.41429 .49625 m
.41429 .5 L
s
.44286 .49625 m
.44286 .5 L
s
.47143 .49625 m
.47143 .5 L
s
.52857 .49625 m
.52857 .5 L
s
.55714 .49625 m
.55714 .5 L
s
.58571 .49625 m
.58571 .5 L
s
.61429 .49625 m
.61429 .5 L
s
.67143 .49625 m
.67143 .5 L
s
.7 .49625 m
.7 .5 L
s
.72857 .49625 m
.72857 .5 L
s
.75714 .49625 m
.75714 .5 L
s
.81429 .49625 m
.81429 .5 L
s
.84286 .49625 m
.84286 .5 L
s
.87143 .49625 m
.87143 .5 L
s
.9 .49625 m
.9 .5 L
s
.04286 .49625 m
.04286 .5 L
s
.01429 .49625 m
.01429 .5 L
s
.95714 .49625 m
.95714 .5 L
s
.98571 .49625 m
.98571 .5 L
s
.25 Mabswid
0 .5 m
1 .5 L
s
.99375 0 m
1 0 L
s
.99375 .0625 m
1 .0625 L
s
.99375 .125 m
1 .125 L
s
.99375 .1875 m
1 .1875 L
s
.99375 .25 m
1 .25 L
s
.99375 .3125 m
1 .3125 L
s
.99375 .375 m
1 .375 L
s
.99375 .4375 m
1 .4375 L
s
.125 Mabswid
.99625 .0125 m
1 .0125 L
s
.99625 .025 m
1 .025 L
s
.99625 .0375 m
1 .0375 L
s
.99625 .05 m
1 .05 L
s
.99625 .075 m
1 .075 L
s
.99625 .0875 m
1 .0875 L
s
.99625 .1 m
1 .1 L
s
.99625 .1125 m
1 .1125 L
s
.99625 .1375 m
1 .1375 L
s
.99625 .15 m
1 .15 L
s
.99625 .1625 m
1 .1625 L
s
.99625 .175 m
1 .175 L
s
.99625 .2 m
1 .2 L
s
.99625 .2125 m
1 .2125 L
s
.99625 .225 m
1 .225 L
s
.99625 .2375 m
1 .2375 L
s
.99625 .2625 m
1 .2625 L
s
.99625 .275 m
1 .275 L
s
.99625 .2875 m
1 .2875 L
s
.99625 .3 m
1 .3 L
s
.99625 .325 m
1 .325 L
s
.99625 .3375 m
1 .3375 L
s
.99625 .35 m
1 .35 L
s
.99625 .3625 m
1 .3625 L
s
.99625 .3875 m
1 .3875 L
s
.99625 .4 m
1 .4 L
s
.99625 .4125 m
1 .4125 L
s
.99625 .425 m
1 .425 L
s
.99625 .45 m
1 .45 L
s
.99625 .4625 m
1 .4625 L
s
.99625 .475 m
1 .475 L
s
.99625 .4875 m
1 .4875 L
s
.25 Mabswid
1 0 m
1 .5 L
s
0 0 m
1 0 L
1 .5 L
0 .5 L
closepath
clip
newpath
.008 w
.07143 .0625 Mdot
.11435 .10006 Mdot
.12386 .10838 Mdot
.13879 .12144 Mdot
.16553 .14483 Mdot
.18876 .16517 Mdot
.19926 .17435 Mdot
.22705 .19867 Mdot
.2572 .22505 Mdot
.26997 .23622 Mdot
.2815 .24631 Mdot
.30146 .26378 Mdot
.32548 .2848 Mdot
.34438 .30133 Mdot
.37217 .32565 Mdot
.39158 .34264 Mdot
.4059 .35517 Mdot
.41689 .36478 Mdot
.42559 .10006 Mdot
.42559 .37239 Mdot
.5 .16517 Mdot
.5 .4375 Mdot
.57441 .10006 Mdot
.57441 .37239 Mdot
.58311 .36478 Mdot
.5941 .35517 Mdot
.60842 .34264 Mdot
.62783 .32565 Mdot
.65562 .30133 Mdot
.67452 .2848 Mdot
.69854 .26378 Mdot
.7185 .24631 Mdot
.73003 .23622 Mdot
.7428 .22505 Mdot
.77295 .19867 Mdot
.80074 .17435 Mdot
.81124 .16517 Mdot
.83447 .14483 Mdot
.86121 .12144 Mdot
.87614 .10838 Mdot
.88565 .10006 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{288, 144},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40004P0000T2000`400?l00000o`00003oo`3ooolQ
0?ooo`00o`3ooolQ0?ooo`00o`3ooolQ0?ooo`00o`3ooolQ0?ooo`00:@3oool2000001h0oooo1000
00030?ooo`8000006`3oool4000000<0oooo0P00000M0?ooo`800000103oool2000001d0oooo00<0
00000?ooo`3oool00P3oool2000001/0oooo0`0000040?ooo`800000703oool2000000@0oooo0P00
000I0?ooo`00:03oool010000000oooo0?ooo`00000O0?ooo`050000003oool0oooo0?ooo`000000
0P3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool0
00006`3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000001`0oooo00D00000
0?ooo`3oool0oooo000000020?ooo`030000003oool0oooo01/0oooo00@000000?ooo`3oool00000
0P3oool00`000000oooo0?ooo`0H0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3o
ool00000603oool002P0oooo00@000000?ooo`3oool000007`3oool01@000000oooo0?ooo`3oool0
00000080oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`020?ooo`040000003oool0
oooo000001h0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`0G0?ooo`D00000
0P3oool010000000oooo0?ooo`00000M0?ooo`040000003oool0oooo00000080oooo00<000000?oo
o`3oool0603oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000001P0oooo000X
0?ooo`040000003oool0oooo000001l0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000
003oool0oooo01X0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003oool0oooo01X0
oooo0P0000030?ooo`040000003oool0oooo000001T0oooo00@000000?ooo`3oool000000`3oool0
10000000oooo0?ooo`00000J0?ooo`<000000`3oool010000000oooo0?ooo`00000J0?ooo`<00000
0`3oool010000000oooo0?ooo`00000H0?ooo`00:03oool010000000oooo0?ooo`00000O0?ooo`05
0000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`0K0?ooo`040000003oool0oooo
00000080oooo00<000000?ooo`3oool0703oool010000000oooo0?ooo`0000020?ooo`030000003o
ool0oooo01P0oooo00<000000?ooo`0000000`3oool010000000oooo0?ooo`00000J0?ooo`030000
003oool0oooo00<0oooo00@000000?ooo`3oool000006P3oool00`000000oooo0?ooo`030?ooo`04
0000003oool0oooo000001P0oooo000X0?ooo`040000003oool0oooo000001d0oooo0`0000030?oo
o`040000003oool0oooo000001X0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?oo
o`00000K0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool000006`3oool20000
00<0oooo00@000000?ooo`3oool000006P3oool00`000000oooo0?ooo`030?ooo`040000003oool0
oooo000001X0oooo00<000000?ooo`3oool00`3oool010000000oooo0?ooo`00000H0?ooo`00:@3o
ool200000200oooo00<000000?ooo`3oool00P3oool2000001`0oooo0P0000040?ooo`8000007@3o
ool2000000@0oooo0P00000M0?ooo`030000003oool0oooo0080oooo0P00000K0?ooo`@000000`3o
ool2000001`0oooo0`0000030?ooo`8000006@3oool00?l0oooo8@3oool00?l0oooo8@3oool001P0
oooooP00000:0?ooo`00603oool01@000000oooo0?ooo`3oool0000000H0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool01@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00D000000?ooo`3o
ool0oooo0000000:0?ooo`00603oool00`000000oooo0?ooo`3j0?ooo`030000003oool0oooo00P0
oooo000H0?ooo`800000nP3oool2000000X0oooo000H0?ooo`030000003oool0oooo0?X0oooo00<0
00000?ooo`3oool0203oool001P0oooo00<000000?ooo`3oool0nP3oool00`000000oooo0?ooo`08
0?ooo`00603oool200000?X0oooo0P00000:0?ooo`00603oool00`000000oooo0?ooo`3j0?ooo`03
0000003oool0oooo00P0oooo000H0?ooo`030000003oool0oooo0?X0oooo00<000000?ooo`3oool0
203oool001P0oooo0P00003j0?ooo`8000002P3oool001P0oooo00<000000?ooo`3oool0nP3oool0
0`000000oooo0?ooo`080?ooo`00603oool00`000000oooo0?ooo`3j0?ooo`030000003oool0oooo
00P0oooo000H0?ooo`800000nP3oool2000000X0oooo000H0?ooo`030000003oool0oooo0?X0oooo
00<000000?ooo`3oool0203oool00140oooo0P0000050?ooo`030000003oool0oooo0?X0oooo00<0
00000?ooo`3oool0203oool00100oooo00@000000?ooo`3oool00000103oool3000000l0oooo0P00
003W0?ooo`<000002P3oool00100oooo00@000000?ooo`3oool00000103oool00`000000oooo0?oo
o`0@0?ooo`030000003oool0oooo0>L0oooo00<000000?ooo`3oool0203oool00100oooo00@00000
0?ooo`3oool00000103oool00`000000oooo0?ooo`3j0?ooo`030000003oool0oooo00P0oooo000@
0?ooo`040000003oool0oooo000000@0oooo00<000000?ooo`3oool0nP3oool00`000000oooo0?oo
o`080?ooo`00403oool010000000oooo0?ooo`0000040?ooo`800000nP3oool2000000X0oooo000A
0?ooo`8000001@3oool00`000000oooo0?ooo`3j0?ooo`030000003oool0oooo00P0oooo000H0?oo
o`030000003oool0oooo0?X0oooo00<000000?ooo`3oool0203oool001P0oooo0P00003j0?ooo`80
00002P3oool001P0oooo00<000000?ooo`3oool0nP3oool00`000000oooo0?ooo`080?ooo`00603o
ool00`000000oooo0?ooo`0J0?ooo`800000C03oool2000002@0oooo0P00001<0?ooo`800000703o
ool00`000000oooo0?ooo`080?ooo`00603oool2000001`0oooo00<000000?ooo`3oool0B`3oool0
0`000000oooo0?ooo`0S0?ooo`030000003oool0oooo04/0oooo00<000000?ooo`3oool06@3oool2
000000X0oooo000H0?ooo`030000003oool0oooo0?X0oooo00<000000?ooo`3oool0203oool001P0
oooo00<000000?ooo`3oool0703oool200000;`0oooo0P00000N0?ooo`030000003oool0oooo00P0
oooo000H0?ooo`8000007P3oool00`000000oooo0?ooo`2k0?ooo`030000003oool0oooo01/0oooo
0P00000:0?ooo`00603oool00`000000oooo0?ooo`3j0?ooo`030000003oool0oooo00P0oooo000@
0?ooo`<000001@3oool00`000000oooo0?ooo`0P0?ooo`800000]03oool200000280oooo00<00000
0?ooo`3oool0203oool001<0oooo00<000000?ooo`3oool00P3oool300000240oooo00<000000?oo
o`3oool0/`3oool00`000000oooo0?ooo`0N0?ooo`<000002P3oool001<0oooo00<000000?ooo`3o
ool00P3oool00`000000oooo0?ooo`3j0?ooo`030000003oool0oooo00P0oooo000@0?ooo`<00000
1@3oool00`000000oooo0?ooo`3j0?ooo`030000003oool0oooo00P0oooo000@0?ooo`030000003o
ool0oooo00D0oooo0P00003j0?ooo`8000002P3oool00100oooo00<000000?ooo`3oool01@3oool0
0`000000oooo0?ooo`3j0?ooo`030000003oool0oooo00P0oooo000@0?ooo`@00000103oool00`00
0000oooo0?ooo`0W0?ooo`800000Y`3oool2000002P0oooo00<000000?ooo`3oool0203oool001P0
oooo00<000000?ooo`3oool0:03oool00`000000oooo0?ooo`2V0?ooo`030000003oool0oooo02H0
oooo00<000000?ooo`3oool0203oool001P0oooo0P00003j0?ooo`8000002P3oool001P0oooo00<0
00000?ooo`3oool0nP3oool00`000000oooo0?ooo`080?ooo`00603oool00`000000oooo0?ooo`3j
0?ooo`030000003oool0oooo00P0oooo000H0?ooo`800000;@3oool2000004d0oooo0P00001=0?oo
o`800000;@3oool2000000X0oooo000H0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3o
ool0C03oool00`000000oooo0?ooo`1<0?ooo`030000003oool0oooo02`0oooo00<000000?ooo`3o
ool0203oool001P0oooo00<000000?ooo`3oool0;`3oool2000009H0oooo0P00000a0?ooo`030000
003oool0oooo00P0oooo000H0?ooo`800000<@3oool00`000000oooo0?ooo`2E0?ooo`030000003o
ool0oooo02h0oooo0P00000:0?ooo`00603oool00`000000oooo0?ooo`3j0?ooo`030000003oool0
oooo00P0oooo000:0?ooo`@000000`3oool2000000D0oooo00<000000?ooo`3oool0nP3oool00`00
0000oooo0?ooo`080?ooo`00303oool01@000000oooo0?ooo`3oool000000080oooo00<000000?oo
o`3oool00P3oool300000?P0oooo0`00000:0?ooo`00303oool01@000000oooo0?ooo`3oool00000
0080oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`3j0?ooo`030000003oool0oooo
00P0oooo000<0?ooo`050000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`020?oo
o`030000003oool0oooo03H0oooo0P0000280?ooo`800000>03oool00`000000oooo0?ooo`080?oo
o`00303oool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`3oool00P3oool20000
03P0oooo00<000000?ooo`3oool0Q`3oool00`000000oooo0?ooo`0e0?ooo`8000002P3oool000X0
oooo0`0000030?ooo`040000003oool0oooo000000@0oooo00<000000?ooo`3oool0nP3oool00`00
0000oooo0?ooo`080?ooo`00303oool00`000000oooo0?ooo`020?ooo`8000001@3oool00`000000
oooo0?ooo`3j0?ooo`030000003oool0oooo00P0oooo000H0?ooo`800000nP3oool2000000X0oooo
000H0?ooo`030000003oool0oooo0?X0oooo00<000000?ooo`3oool0203oool001P0oooo00<00000
0?ooo`3oool0nP3oool00`000000oooo0?ooo`080?ooo`00603oool00`000000oooo0?ooo`0n0?oo
o`800000N03oool200000400oooo00<000000?ooo`3oool0203oool001P0oooo0P0000100?ooo`03
0000003oool0oooo07L0oooo00<000000?ooo`3oool0?@3oool2000000X0oooo000H0?ooo`030000
003oool0oooo0?X0oooo00<000000?ooo`3oool0203oool001P0oooo00<000000?ooo`3oool0@@3o
ool200000780oooo0P0000130?ooo`030000003oool0oooo00P0oooo000H0?ooo`800000@`3oool0
0`000000oooo0?ooo`1a0?ooo`030000003oool0oooo0400oooo0P00000:0?ooo`00603oool00`00
0000oooo0?ooo`140?ooo`800000K03oool2000004H0oooo00<000000?ooo`3oool0203oool000X0
oooo100000020?ooo`<000001@3oool00`000000oooo0?ooo`150?ooo`030000003oool0oooo06/0
oooo00<000000?ooo`3oool0A03oool00`000000oooo0?ooo`080?ooo`00303oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo0080oooo0`00003h0?ooo`<000002P3oool000`0oooo00<0
00000?ooo`3oool0103oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0?X0oooo00<0
00000?ooo`3oool0203oool000`0oooo00@000000?ooo`3oool0oooo0`0000050?ooo`030000003o
ool0oooo0?X0oooo00<000000?ooo`3oool0203oool000`0oooo00D000000?ooo`3oool0oooo0000
00070?ooo`800000BP3oool200000680oooo0P00001:0?ooo`8000002P3oool000X0oooo0`000003
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0BP3oool00`000000oooo0?ooo`1Q
0?ooo`030000003oool0oooo04T0oooo00<000000?ooo`3oool0203oool000`0oooo00@000000?oo
o`3oool0oooo100000040?ooo`030000003oool0oooo0?X0oooo00<000000?ooo`3oool0203oool0
01P0oooo0P00003j0?ooo`8000002P3oool001P0oooo00<000000?ooo`3oool0nP3oool00`000000
oooo0?ooo`080?ooo`00603oool00`000000oooo0?ooo`1?0?ooo`800000EP3oool200000540oooo
00<000000?ooo`3oool0203oool001P0oooo0P00001A0?ooo`030000003oool0oooo05D0oooo00<0
00000?ooo`3oool0CP3oool2000000X0oooo000H0?ooo`030000003oool0oooo0?X0oooo00<00000
0?ooo`3oool0203oool001P0oooo00<000000?ooo`3oool0nP3oool00`000000oooo0?ooo`080?oo
o`00603oool2000005D0oooo0P00001<0?ooo`800000E@3oool2000000X0oooo000H0?ooo`030000
003oool0oooo05D0oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`1D0?ooo`030000
003oool0oooo00P0oooo000:0?ooo`@000000`3oool2000000D0oooo00<000000?ooo`3oool0nP3o
ool00`000000oooo0?ooo`080?ooo`002P3oool00`000000oooo0?ooo`030?ooo`040000003oool0
oooo000000@0oooo00<000000?ooo`3oool0nP3oool00`000000oooo0?ooo`080?ooo`002`3oool0
0`000000oooo0?ooo`020?ooo`040000003oool0oooo000000@0oooo0`00003h0?ooo`<000002P3o
ool000`0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0nP3oool00`000000oooo0?ooo`080?ooo`003@3oool010000000oooo0?ooo`00
00020?ooo`030000003oool0oooo0080oooo00<000000?ooo`3oool0F`3oool2000003h0oooo0P00
001M0?ooo`030000003oool0oooo00P0oooo000:0?ooo`040000003oool0oooo00000080oooo00@0
00000?ooo`3oool00000103oool2000005d0oooo00<000000?ooo`3oool0?@3oool00`000000oooo
0?ooo`1J0?ooo`8000002P3oool000/0oooo0P0000040?ooo`8000001@3oool00`000000oooo0?oo
o`3j0?ooo`030000003oool0oooo00P0oooo000H0?ooo`030000003oool0oooo0?X0oooo00<00000
0?ooo`3oool0203oool001P0oooo0P00003j0?ooo`8000002P3oool001P0oooo00<000000?ooo`3o
ool0H03oool2000003@0oooo0P00001R0?ooo`030000003oool0oooo00P0oooo000H0?ooo`030000
003oool0oooo0640oooo00<000000?ooo`3oool0<`3oool00`000000oooo0?ooo`1P0?ooo`030000
003oool0oooo00P0oooo000H0?ooo`800000nP3oool2000000X0oooo000H0?ooo`030000003oool0
oooo06<0oooo0P00000^0?ooo`800000I@3oool00`000000oooo0?ooo`080?ooo`00603oool00`00
0000oooo0?ooo`1T0?ooo`030000003oool0oooo02d0oooo00<000000?ooo`3oool0H`3oool00`00
0000oooo0?ooo`080?ooo`00603oool2000006L0oooo0P00000X0?ooo`800000I`3oool2000000X0
oooo000H0?ooo`030000003oool0oooo06L0oooo00<000000?ooo`3oool09`3oool00`000000oooo
0?ooo`1V0?ooo`030000003oool0oooo00P0oooo000:0?ooo`@000000P3oool3000000D0oooo00<0
00000?ooo`3oool0J03oool2000002@0oooo0P00001Z0?ooo`030000003oool0oooo00P0oooo000:
0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool00P3oool3000006T0oooo00<00000
0?ooo`3oool08`3oool00`000000oooo0?ooo`1V0?ooo`<000002P3oool000/0oooo00<000000?oo
o`3oool01@3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo0?X0oooo00<000000?oo
o`3oool0203oool000`0oooo00@000000?ooo`3oool0oooo0`0000050?ooo`030000003oool0oooo
0?X0oooo00<000000?ooo`3oool0203oool000d0oooo00@000000?ooo`3oool000001`3oool00`00
0000oooo0?ooo`3j0?ooo`030000003oool0oooo00P0oooo000:0?ooo`040000003oool0oooo0000
0080oooo00<000000?ooo`3oool01@3oool200000?X0oooo0P00000:0?ooo`002`3oool2000000<0
oooo100000040?ooo`030000003oool0oooo0?X0oooo00<000000?ooo`3oool0203oool001P0oooo
00<000000?ooo`3oool0nP3oool00`000000oooo0?ooo`080?ooo`00603oool200000?X0oooo0P00
000:0?ooo`00603oool00`000000oooo0?ooo`3j0?ooo`030000003oool0oooo00P0oooo000H0?oo
o`030000003oool0oooo0?X0oooo00<000000?ooo`3oool0203oool001P0oooo0P00003j0?ooo`80
00002P3oool001P0oooo00<000000?ooo`3oool0nP3oool00`000000oooo0?ooo`080?ooo`00603o
ool00`000000oooo0?ooo`3j0?ooo`030000003oool0oooo00P0oooo000H0?ooo`800000nP3oool2
000000X0oooo000H0?ooo`030000003oool0oooo0?X0oooo00<000000?ooo`3oool0203oool000/0
oooo0P0000040?ooo`8000001@3oool00`000000oooo0?ooo`3j0?ooo`030000003oool0oooo00P0
oooo000:0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool00000103oool30000
07/0oooo0P00001k0?ooo`<000002P3oool000d0oooo00@000000?ooo`3oool000000P3oool00`00
0000oooo0?ooo`020?ooo`030000003oool0oooo07`0oooo00<000000?ooo`3oool0N`3oool00`00
0000oooo0?ooo`080?ooo`002`3oool2000000<0oooo00@000000?ooo`3oool00000103oool00`00
0000oooo0?ooo`3j0?ooo`030000003oool0oooo00P0oooo000=0?ooo`040000003oool0oooo0000
0080oooo00<000000?ooo`3oool00P3oool200000?X0oooo0P00000:0?ooo`002P3oool010000000
oooo0?ooo`0000020?ooo`040000003oool0oooo000000@0oooo00<000000?ooo`3oool0nP3oool0
0`000000oooo0?ooo`080?ooo`002`3oool2000000@0oooo0P0000050?ooo`030000003oool0oooo
0?X0oooo00<000000?ooo`3oool0203oool001P0oooo00<000000?ooo`3oool0nP3oool00`000000
oooo0?ooo`080?ooo`00603oool200000?X0oooo0P00000:0?ooo`00603oool00`000000oooo0?oo
o`3j0?ooo`030000003oool0oooo00P0oooo000H0?ooo`030000003oool0oooo0?X0oooo00<00000
0?ooo`3oool0203oool001P0oooo0P00003j0?ooo`8000002P3oool001P0oooo00<000000?ooo`3o
ool0nP3oool00`000000oooo0?ooo`080?ooo`00603oool00`000000oooo0?ooo`3j0?ooo`030000
003oool0oooo00P0oooo000H0?ooo`800000nP3oool2000000X0oooo000H0?ooo`030000003oool0
oooo00l0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0
oooo0280oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0Q0?ooo`030000003oool0
oooo0240oooo00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`080?ooo`002`3oool20000
00<0oooo0`0000050?ooo`050000003oool0oooo0?ooo`0000001P3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`05
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`05
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool01@3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0103oool01@000000oooo0?ooo`3o
ool0000000X0oooo000:0?ooo`040000003oool0oooo000000D0oooo00<000000?ooo`3oool00P3o
ooon000000X0oooo000=0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0o`3oool;
0?ooo`002`3oool2000000<0oooo0`00003o0?ooo`h0oooo000=0?ooo`040000003oool0oooo0000
0?l0oooo403oool000X0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`3o0?oo
o`h0oooo0000\
\>"],
  ImageRangeCache->{{{0, 287}, {143, 0}} -> {-11.8362, -8.84965, 0.27703, 
  0.316605}}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(X = AnglePairsAndRationalApproxs[15, 10^\(-3\)]; \nLength[X]\), 
    \(Angles = \(Transpose[X]\)[\([1]\)]; \n
    ListPlot[Angles, \ AspectRatio -> 1/2, Frame -> True, \ Axes -> False, \ 
      PlotRange -> {{\(-5\), 65}, {\(-5\), 35}}]; 
    WriteData["\<~/ASP/WebServices/GrainBreaker/2DGeometries/Geo15.dat\>", 
      X]; \)}], "Input"],

Cell[BoxData[
    \(94\)], "Output"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .5 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0714286 0.0142857 0.0625 0.0125 [
[.07143 -0.0125 -3 -9 ]
[.07143 -0.0125 3 0 ]
[.21429 -0.0125 -6 -9 ]
[.21429 -0.0125 6 0 ]
[.35714 -0.0125 -6 -9 ]
[.35714 -0.0125 6 0 ]
[.5 -0.0125 -6 -9 ]
[.5 -0.0125 6 0 ]
[.64286 -0.0125 -6 -9 ]
[.64286 -0.0125 6 0 ]
[.78571 -0.0125 -6 -9 ]
[.78571 -0.0125 6 0 ]
[.92857 -0.0125 -6 -9 ]
[.92857 -0.0125 6 0 ]
[ 0 0 -0.125 0 ]
[-0.0125 .0625 -6 -4.5 ]
[-0.0125 .0625 0 4.5 ]
[-0.0125 .125 -6 -4.5 ]
[-0.0125 .125 0 4.5 ]
[-0.0125 .1875 -12 -4.5 ]
[-0.0125 .1875 0 4.5 ]
[-0.0125 .25 -12 -4.5 ]
[-0.0125 .25 0 4.5 ]
[-0.0125 .3125 -12 -4.5 ]
[-0.0125 .3125 0 4.5 ]
[-0.0125 .375 -12 -4.5 ]
[-0.0125 .375 0 4.5 ]
[-0.0125 .4375 -12 -4.5 ]
[-0.0125 .4375 0 4.5 ]
[-0.0125 .5 -12 -4.5 ]
[-0.0125 .5 0 4.5 ]
[ 0 0 -0.125 0 ]
[ 0 .5 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .5 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.07143 0 m
.07143 .00625 L
s
[(0)] .07143 -0.0125 0 1 Mshowa
.21429 0 m
.21429 .00625 L
s
[(10)] .21429 -0.0125 0 1 Mshowa
.35714 0 m
.35714 .00625 L
s
[(20)] .35714 -0.0125 0 1 Mshowa
.5 0 m
.5 .00625 L
s
[(30)] .5 -0.0125 0 1 Mshowa
.64286 0 m
.64286 .00625 L
s
[(40)] .64286 -0.0125 0 1 Mshowa
.78571 0 m
.78571 .00625 L
s
[(50)] .78571 -0.0125 0 1 Mshowa
.92857 0 m
.92857 .00625 L
s
[(60)] .92857 -0.0125 0 1 Mshowa
.125 Mabswid
.1 0 m
.1 .00375 L
s
.12857 0 m
.12857 .00375 L
s
.15714 0 m
.15714 .00375 L
s
.18571 0 m
.18571 .00375 L
s
.24286 0 m
.24286 .00375 L
s
.27143 0 m
.27143 .00375 L
s
.3 0 m
.3 .00375 L
s
.32857 0 m
.32857 .00375 L
s
.38571 0 m
.38571 .00375 L
s
.41429 0 m
.41429 .00375 L
s
.44286 0 m
.44286 .00375 L
s
.47143 0 m
.47143 .00375 L
s
.52857 0 m
.52857 .00375 L
s
.55714 0 m
.55714 .00375 L
s
.58571 0 m
.58571 .00375 L
s
.61429 0 m
.61429 .00375 L
s
.67143 0 m
.67143 .00375 L
s
.7 0 m
.7 .00375 L
s
.72857 0 m
.72857 .00375 L
s
.75714 0 m
.75714 .00375 L
s
.81429 0 m
.81429 .00375 L
s
.84286 0 m
.84286 .00375 L
s
.87143 0 m
.87143 .00375 L
s
.9 0 m
.9 .00375 L
s
.04286 0 m
.04286 .00375 L
s
.01429 0 m
.01429 .00375 L
s
.95714 0 m
.95714 .00375 L
s
.98571 0 m
.98571 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
0 .0625 m
.00625 .0625 L
s
[(0)] -0.0125 .0625 1 0 Mshowa
0 .125 m
.00625 .125 L
s
[(5)] -0.0125 .125 1 0 Mshowa
0 .1875 m
.00625 .1875 L
s
[(10)] -0.0125 .1875 1 0 Mshowa
0 .25 m
.00625 .25 L
s
[(15)] -0.0125 .25 1 0 Mshowa
0 .3125 m
.00625 .3125 L
s
[(20)] -0.0125 .3125 1 0 Mshowa
0 .375 m
.00625 .375 L
s
[(25)] -0.0125 .375 1 0 Mshowa
0 .4375 m
.00625 .4375 L
s
[(30)] -0.0125 .4375 1 0 Mshowa
0 .5 m
.00625 .5 L
s
[(35)] -0.0125 .5 1 0 Mshowa
.125 Mabswid
0 .0125 m
.00375 .0125 L
s
0 .025 m
.00375 .025 L
s
0 .0375 m
.00375 .0375 L
s
0 .05 m
.00375 .05 L
s
0 .075 m
.00375 .075 L
s
0 .0875 m
.00375 .0875 L
s
0 .1 m
.00375 .1 L
s
0 .1125 m
.00375 .1125 L
s
0 .1375 m
.00375 .1375 L
s
0 .15 m
.00375 .15 L
s
0 .1625 m
.00375 .1625 L
s
0 .175 m
.00375 .175 L
s
0 .2 m
.00375 .2 L
s
0 .2125 m
.00375 .2125 L
s
0 .225 m
.00375 .225 L
s
0 .2375 m
.00375 .2375 L
s
0 .2625 m
.00375 .2625 L
s
0 .275 m
.00375 .275 L
s
0 .2875 m
.00375 .2875 L
s
0 .3 m
.00375 .3 L
s
0 .325 m
.00375 .325 L
s
0 .3375 m
.00375 .3375 L
s
0 .35 m
.00375 .35 L
s
0 .3625 m
.00375 .3625 L
s
0 .3875 m
.00375 .3875 L
s
0 .4 m
.00375 .4 L
s
0 .4125 m
.00375 .4125 L
s
0 .425 m
.00375 .425 L
s
0 .45 m
.00375 .45 L
s
0 .4625 m
.00375 .4625 L
s
0 .475 m
.00375 .475 L
s
0 .4875 m
.00375 .4875 L
s
.25 Mabswid
0 0 m
0 .5 L
s
.07143 .49375 m
.07143 .5 L
s
.21429 .49375 m
.21429 .5 L
s
.35714 .49375 m
.35714 .5 L
s
.5 .49375 m
.5 .5 L
s
.64286 .49375 m
.64286 .5 L
s
.78571 .49375 m
.78571 .5 L
s
.92857 .49375 m
.92857 .5 L
s
.125 Mabswid
.1 .49625 m
.1 .5 L
s
.12857 .49625 m
.12857 .5 L
s
.15714 .49625 m
.15714 .5 L
s
.18571 .49625 m
.18571 .5 L
s
.24286 .49625 m
.24286 .5 L
s
.27143 .49625 m
.27143 .5 L
s
.3 .49625 m
.3 .5 L
s
.32857 .49625 m
.32857 .5 L
s
.38571 .49625 m
.38571 .5 L
s
.41429 .49625 m
.41429 .5 L
s
.44286 .49625 m
.44286 .5 L
s
.47143 .49625 m
.47143 .5 L
s
.52857 .49625 m
.52857 .5 L
s
.55714 .49625 m
.55714 .5 L
s
.58571 .49625 m
.58571 .5 L
s
.61429 .49625 m
.61429 .5 L
s
.67143 .49625 m
.67143 .5 L
s
.7 .49625 m
.7 .5 L
s
.72857 .49625 m
.72857 .5 L
s
.75714 .49625 m
.75714 .5 L
s
.81429 .49625 m
.81429 .5 L
s
.84286 .49625 m
.84286 .5 L
s
.87143 .49625 m
.87143 .5 L
s
.9 .49625 m
.9 .5 L
s
.04286 .49625 m
.04286 .5 L
s
.01429 .49625 m
.01429 .5 L
s
.95714 .49625 m
.95714 .5 L
s
.98571 .49625 m
.98571 .5 L
s
.25 Mabswid
0 .5 m
1 .5 L
s
.99375 0 m
1 0 L
s
.99375 .0625 m
1 .0625 L
s
.99375 .125 m
1 .125 L
s
.99375 .1875 m
1 .1875 L
s
.99375 .25 m
1 .25 L
s
.99375 .3125 m
1 .3125 L
s
.99375 .375 m
1 .375 L
s
.99375 .4375 m
1 .4375 L
s
.125 Mabswid
.99625 .0125 m
1 .0125 L
s
.99625 .025 m
1 .025 L
s
.99625 .0375 m
1 .0375 L
s
.99625 .05 m
1 .05 L
s
.99625 .075 m
1 .075 L
s
.99625 .0875 m
1 .0875 L
s
.99625 .1 m
1 .1 L
s
.99625 .1125 m
1 .1125 L
s
.99625 .1375 m
1 .1375 L
s
.99625 .15 m
1 .15 L
s
.99625 .1625 m
1 .1625 L
s
.99625 .175 m
1 .175 L
s
.99625 .2 m
1 .2 L
s
.99625 .2125 m
1 .2125 L
s
.99625 .225 m
1 .225 L
s
.99625 .2375 m
1 .2375 L
s
.99625 .2625 m
1 .2625 L
s
.99625 .275 m
1 .275 L
s
.99625 .2875 m
1 .2875 L
s
.99625 .3 m
1 .3 L
s
.99625 .325 m
1 .325 L
s
.99625 .3375 m
1 .3375 L
s
.99625 .35 m
1 .35 L
s
.99625 .3625 m
1 .3625 L
s
.99625 .3875 m
1 .3875 L
s
.99625 .4 m
1 .4 L
s
.99625 .4125 m
1 .4125 L
s
.99625 .425 m
1 .425 L
s
.99625 .45 m
1 .45 L
s
.99625 .4625 m
1 .4625 L
s
.99625 .475 m
1 .475 L
s
.99625 .4875 m
1 .4875 L
s
.25 Mabswid
1 0 m
1 .5 L
s
0 0 m
1 0 L
1 .5 L
0 .5 L
closepath
clip
newpath
.008 w
.07143 .0625 Mdot
.09922 .08681 Mdot
.10292 .09005 Mdot
.10776 .09429 Mdot
.11435 .10006 Mdot
.12386 .10838 Mdot
.1304 .1141 Mdot
.13879 .12144 Mdot
.14995 .1312 Mdot
.15454 .13522 Mdot
.16553 .14483 Mdot
.17984 .15736 Mdot
.18876 .16517 Mdot
.19926 .17435 Mdot
.20524 .17959 Mdot
.20911 .18297 Mdot
.22705 .19867 Mdot
.24595 .2152 Mdot
.25028 .219 Mdot
.2572 .22505 Mdot
.26997 .23622 Mdot
.2815 .24631 Mdot
.28684 .25099 Mdot
.30146 .26378 Mdot
.31814 .27837 Mdot
.32548 .2848 Mdot
.33226 .29073 Mdot
.34438 .30133 Mdot
.35488 .08681 Mdot
.35488 .31052 Mdot
.35962 .31467 Mdot
.37217 .32565 Mdot
.38267 .0625 Mdot
.38267 .33483 Mdot
.39158 .34264 Mdot
.39925 .34934 Mdot
.4059 .35517 Mdot
.41689 .36478 Mdot
.42559 .10006 Mdot
.42559 .37239 Mdot
.43264 .37856 Mdot
.43848 .219 Mdot
.43848 .38367 Mdot
.44338 .38796 Mdot
.44756 .39162 Mdot
.45117 .39478 Mdot
.5 .09005 Mdot
.5 .16517 Mdot
.5 .4375 Mdot
.54883 .39478 Mdot
.55244 .39162 Mdot
.55662 .38796 Mdot
.56152 .219 Mdot
.56152 .38367 Mdot
.56736 .37856 Mdot
.57441 .10006 Mdot
.57441 .37239 Mdot
.58311 .36478 Mdot
.5941 .35517 Mdot
.60075 .34934 Mdot
.60842 .34264 Mdot
.61733 .0625 Mdot
.61733 .33483 Mdot
.62783 .32565 Mdot
.64038 .31467 Mdot
.64512 .08681 Mdot
.64512 .31052 Mdot
.65562 .30133 Mdot
.66774 .29073 Mdot
.67452 .2848 Mdot
.68186 .27837 Mdot
.69854 .26378 Mdot
.71316 .25099 Mdot
.7185 .24631 Mdot
.73003 .23622 Mdot
.7428 .22505 Mdot
.74972 .219 Mdot
.75405 .2152 Mdot
.77295 .19867 Mdot
.79089 .18297 Mdot
.79476 .17959 Mdot
.80074 .17435 Mdot
.81124 .16517 Mdot
.82016 .15736 Mdot
.83447 .14483 Mdot
.84546 .13522 Mdot
.85005 .1312 Mdot
.86121 .12144 Mdot
.8696 .1141 Mdot
.87614 .10838 Mdot
.88565 .10006 Mdot
.89224 .09429 Mdot
.89708 .09005 Mdot
.90078 .08681 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{407, 203.5},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006G0000bb000`400?l00000o`00003oo`3ooonH
0?ooo`00o`3ooonH0?ooo`00o`3ooonH0?ooo`00o`3ooonH0?ooo`00=03oool2000002h0oooo1000
00030?ooo`800000;03oool4000000<0oooo0P00000]0?ooo`800000103oool2000002d0oooo00<0
00000?ooo`3oool00P3oool2000002`0oooo0`0000040?ooo`800000;@3oool2000000@0oooo0P00
000R0?ooo`00<`3oool010000000oooo0?ooo`00000_0?ooo`050000003oool0oooo0?ooo`000000
0P3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool0
0000:`3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000002`0oooo00D00000
0?ooo`3oool0oooo000000020?ooo`030000003oool0oooo02`0oooo00@000000?ooo`3oool00000
0P3oool00`000000oooo0?ooo`0Y0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3o
ool000008@3oool003<0oooo00@000000?ooo`3oool00000;`3oool01@000000oooo0?ooo`3oool0
00000080oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`020?ooo`040000003oool0
oooo000002h0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`0W0?ooo`D00000
0P3oool010000000oooo0?ooo`00000^0?ooo`040000003oool0oooo00000080oooo00<000000?oo
o`3oool0:@3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo00000240oooo000c
0?ooo`040000003oool0oooo000002l0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000
003oool0oooo02/0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003oool0oooo02X0
oooo0P0000030?ooo`040000003oool0oooo000002T0oooo00@000000?ooo`3oool000000`3oool0
10000000oooo0?ooo`00000[0?ooo`<000000`3oool010000000oooo0?ooo`00000[0?ooo`<00000
0`3oool010000000oooo0?ooo`00000Q0?ooo`00<`3oool010000000oooo0?ooo`00000_0?ooo`05
0000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`0/0?ooo`040000003oool0oooo
00000080oooo00<000000?ooo`3oool0;03oool010000000oooo0?ooo`0000020?ooo`030000003o
ool0oooo02P0oooo00<000000?ooo`0000000`3oool010000000oooo0?ooo`00000[0?ooo`030000
003oool0oooo00<0oooo00@000000?ooo`3oool00000:`3oool00`000000oooo0?ooo`030?ooo`04
0000003oool0oooo00000240oooo000c0?ooo`040000003oool0oooo000002d0oooo0`0000030?oo
o`040000003oool0oooo000002/0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?oo
o`00000[0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool00000:`3oool20000
00<0oooo00@000000?ooo`3oool00000:`3oool00`000000oooo0?ooo`030?ooo`040000003oool0
oooo000002/0oooo00<000000?ooo`3oool00`3oool010000000oooo0?ooo`00000Q0?ooo`00=03o
ool200000300oooo00<000000?ooo`3oool00P3oool2000002d0oooo0P0000040?ooo`800000;@3o
ool2000000@0oooo0P00000]0?ooo`030000003oool0oooo0080oooo0P00000/0?ooo`@000000`3o
ool2000002d0oooo0`0000030?ooo`8000008P3oool00?l0ooooV03oool00?l0ooooV03oool00?l0
ooooV03oool001X0ooooo`00001c000000/0oooo000J0?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00T0oooo000J
0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`0b
0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`0b
0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`09
0?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofl0oooo00<000000?ooo`3oool02@3oool001X0
oooo0`00003o0?ooofh0oooo0P00000;0?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofl0oooo
00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0o`3ooom_0?ooo`030000003o
ool0oooo00T0oooo000J0?ooo`030000003oool0oooo0?l0ooooK`3oool00`000000oooo0?ooo`09
0?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofl0oooo00<000000?ooo`3oool02@3oool001X0
oooo0`00003o0?ooofh0oooo0P00000;0?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofl0oooo
00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0o`3ooom_0?ooo`030000003o
ool0oooo00T0oooo000J0?ooo`030000003oool0oooo0?l0ooooK`3oool00`000000oooo0?ooo`09
0?ooo`006P3oool300000?l0ooooKP3oool2000000/0oooo000J0?ooo`030000003oool0oooo0?l0
ooooK`3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofl0oooo
00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0o`3ooom_0?ooo`030000003o
ool0oooo00T0oooo000J0?ooo`030000003oool0oooo0?l0ooooK`3oool00`000000oooo0?ooo`09
0?ooo`006P3oool300000?l0ooooKP3oool2000000/0oooo000J0?ooo`030000003oool0oooo0?l0
ooooK`3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofl0oooo
00<000000?ooo`3oool02@3oool00180oooo0P0000060?ooo`030000003oool0oooo0?l0ooooK`3o
ool00`000000oooo0?ooo`090?ooo`004@3oool010000000oooo0?ooo`0000050?ooo`030000003o
ool0oooo0?l0ooooK`3oool00`000000oooo0?ooo`090?ooo`004@3oool010000000oooo0?ooo`00
00050?ooo`@000005P3oool200000740oooo0P00001D0?ooo`800000RP3oool3000000/0oooo000A
0?ooo`040000003oool0oooo000000D0oooo00<000000?ooo`3oool0603oool00`000000oooo0?oo
o`1`0?ooo`030000003oool0oooo05<0oooo00<000000?ooo`3oool0RP3oool00`000000oooo0?oo
o`090?ooo`004@3oool010000000oooo0?ooo`0000050?ooo`030000003oool0oooo0?l0ooooK`3o
ool00`000000oooo0?ooo`090?ooo`004@3oool010000000oooo0?ooo`0000050?ooo`030000003o
ool0oooo0?l0ooooK`3oool00`000000oooo0?ooo`090?ooo`004P3oool2000000H0oooo0`00003o
0?ooofh0oooo0P00000;0?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofl0oooo00<000000?oo
o`3oool02@3oool001X0oooo00<000000?ooo`3oool0o`3ooom_0?ooo`030000003oool0oooo00T0
oooo000J0?ooo`030000003oool0oooo0?l0ooooK`3oool00`000000oooo0?ooo`090?ooo`006P3o
ool00`000000oooo0?ooo`3o0?ooofl0oooo00<000000?ooo`3oool02@3oool001X0oooo0`00000Q
0?ooo`800000G03oool2000006T0oooo0P00001M0?ooo`8000008P3oool2000000/0oooo000J0?oo
o`030000003oool0oooo0280oooo0P00001L0?ooo`030000003oool0oooo0380oooo0P00000d0?oo
o`030000003oool0oooo05T0oooo0P0000000`3oool000000?ooo`0R0?ooo`030000003oool0oooo
00T0oooo000J0?ooo`030000003oool0oooo02<0oooo0`00002@0?ooo`030000003oool0oooo08`0
oooo0P0000000`3oool000000?ooo`0T0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000
003oool0oooo02D0oooo00<000000?ooo`3oool0o`3ooolO0?ooo`030000003oool0oooo02D0oooo
00<000000?ooo`3oool02@3oool001X0oooo0`00000W0?ooo`800000L03oool2000003D0oooo0P00
001a0?ooo`800000:03oool2000000/0oooo000J0?ooo`030000003oool0oooo02P0oooo00<00000
0?ooo`3oool0K`3oool00`000000oooo0?ooo`0d0?ooo`030000003oool0oooo0700oooo00<00000
0?ooo`3oool09`3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`3o0?oo
ofl0oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0o`3ooom_0?ooo`03
0000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo02X0oooo0P00003o0?oooaD0oooo
0P00000/0?ooo`030000003oool0oooo00T0oooo000J0?ooo`<00000:`3oool00`000000oooo0?oo
o`3o0?oooa@0oooo00<000000?ooo`3oool0:@3oool2000000/0oooo000J0?ooo`030000003oool0
oooo02d0oooo0P00003o0?ooo`l0oooo0P00000_0?ooo`030000003oool0oooo00T0oooo000J0?oo
o`030000003oool0oooo02h0oooo00<000000?ooo`3oool0o`3oool>0?ooo`030000003oool0oooo
02d0oooo00<000000?ooo`3oool02@3oool00140oooo0`0000060?ooo`030000003oool0oooo0300
oooo0P00003o0?ooo`T0oooo0P00000b0?ooo`030000003oool0oooo00T0oooo000D0?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`3o0?ooo`P0oooo
00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`090?ooo`00503oool00`000000oooo0?oo
o`030?ooo`@00000o`3ooom/0?ooo`<000002`3oool00140oooo0`0000060?ooo`030000003oool0
oooo0?l0ooooK`3oool00`000000oooo0?ooo`090?ooo`004@3oool00`000000oooo0?ooo`060?oo
o`030000003oool0oooo03@0oooo0P00003o0?ooo`40oooo0P00000f0?ooo`030000003oool0oooo
00T0oooo000A0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0=@3oool200000?d0
oooo0P0000000`3oool000000?ooo`0e0?ooo`030000003oool0oooo00T0oooo000A0?ooo`@00000
1@3oool3000003H0oooo00<000000?ooo`3oool0o03oool00`000000oooo0?ooo`0e0?ooo`800000
2`3oool001X0oooo00<000000?ooo`3oool0o`3ooom_0?ooo`030000003oool0oooo00T0oooo000J
0?ooo`030000003oool0oooo0?l0ooooK`3oool00`000000oooo0?ooo`090?ooo`006P3oool00`00
0000oooo0?ooo`0i0?ooo`800000m@3oool2000003`0oooo00<000000?ooo`3oool02@3oool001X0
oooo00<000000?ooo`3oool0>P3oool00`000000oooo0?ooo`3d0?ooo`030000003oool0oooo03X0
oooo00<000000?ooo`3oool02@3oool001X0oooo0`00003o0?ooofh0oooo0P00000;0?ooo`006P3o
ool00`000000oooo0?ooo`3o0?ooofl0oooo00<000000?ooo`3oool02@3oool001X0oooo00<00000
0?ooo`3oool0o`3ooom_0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo
03l0oooo0P00003Z0?ooo`800000@@3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000
oooo0?ooo`100?ooo`030000003oool0oooo0>T0oooo00<000000?ooo`3oool0?`3oool00`000000
oooo0?ooo`090?ooo`006P3oool300000480oooo0P00001a0?ooo`800000L@3oool2000004<0oooo
0P00000;0?ooo`006P3oool00`000000oooo0?ooo`130?ooo`030000003oool0oooo0700oooo00<0
00000?ooo`3oool0L03oool00`000000oooo0?ooo`120?ooo`030000003oool0oooo00T0oooo000J
0?ooo`030000003oool0oooo0?l0ooooK`3oool00`000000oooo0?ooo`090?ooo`006P3oool00`00
0000oooo0?ooo`3o0?ooofl0oooo00<000000?ooo`3oool02@3oool001X0oooo0`0000160?ooo`80
0000g03oool2000004L0oooo0P00000;0?ooo`006P3oool00`000000oooo0?ooo`170?ooo`030000
003oool0oooo0=/0oooo00<000000?ooo`3oool0AP3oool00`000000oooo0?ooo`090?ooo`006P3o
ool00`000000oooo0?ooo`180?ooo`800000f03oool2000004X0oooo00<000000?ooo`3oool02@3o
ool000/0oooo100000030?ooo`8000001P3oool00`000000oooo0?ooo`190?ooo`<00000e03oool2
000000030?ooo`000000oooo04T0oooo00<000000?ooo`3oool02@3oool000d0oooo00D000000?oo
o`3oool0oooo000000020?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0B`3oool0
0`000000oooo0?ooo`3C0?ooo`030000003oool0oooo04X0oooo00<000000?ooo`3oool02@3oool0
00d0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003oool0oooo00<0oooo1000003o
0?ooof`0oooo0`00000;0?ooo`003@3oool01@000000oooo0?ooo`3oool000000080oooo00<00000
0?ooo`3oool00`3oool00`000000oooo0?ooo`3o0?ooofl0oooo00<000000?ooo`3oool02@3oool0
00d0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool0o`3ooom_0?ooo`030000003oool0oooo00T0oooo000;0?ooo`<000000`3oool01000
0000oooo0?ooo`0000050?ooo`030000003oool0oooo0?l0ooooK`3oool00`000000oooo0?ooo`09
0?ooo`003@3oool00`000000oooo0?ooo`020?ooo`8000001P3oool300000500oooo0P0000370?oo
o`800000DP3oool2000000/0oooo000J0?ooo`030000003oool0oooo0540oooo00<000000?ooo`3o
ool0aP3oool00`000000oooo0?ooo`1A0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000
003oool0oooo0?l0ooooK`3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?oo
o`3o0?ooofl0oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0o`3ooom_
0?ooo`030000003oool0oooo00T0oooo000J0?ooo`<00000o`3ooom^0?ooo`8000002`3oool001X0
oooo00<000000?ooo`3oool0E`3oool200000;T0oooo0P00001J0?ooo`030000003oool0oooo00T0
oooo000J0?ooo`030000003oool0oooo05P0oooo0`0000130?ooo`800000:`3oool2000004@0oooo
0`00001J0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo05X0oooo00<0
00000?ooo`3oool0@P3oool00`000000oooo0?ooo`0Z0?ooo`030000003oool0oooo04<0oooo00<0
00000?ooo`3oool0F@3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`3o
0?ooofl0oooo00<000000?ooo`3oool02@3oool001X0oooo0`00001K0?ooo`800000/@3oool20000
05d0oooo0P00000;0?ooo`006P3oool00`000000oooo0?ooo`1L0?ooo`030000003oool0oooo0;00
oooo00<000000?ooo`3oool0G03oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo
0?ooo`3o0?ooofl0oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0o`3o
oom_0?ooo`030000003oool0oooo00T0oooo000J0?ooo`<00000H03oool200000:P0oooo0P00001Q
0?ooo`8000002`3oool001X0oooo00<000000?ooo`3oool0H@3oool00`000000oooo0?ooo`2W0?oo
o`030000003oool0oooo0600oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3o
ool0o`3ooom_0?ooo`030000003oool0oooo00T0oooo000;0?ooo`@000000P3oool3000000H0oooo
00<000000?ooo`3oool0I03oool2000009l0oooo0P00001W0?ooo`030000003oool0oooo00T0oooo
000=0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`1U0?ooo`030000003oool0oooo09h0oooo00<000000?ooo`3oool0I@3oool00`000000oooo0?oo
o`090?ooo`003@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00<0oooo1000001U
0?ooo`800000V`3oool2000006L0oooo0`00000;0?ooo`003@3oool010000000oooo0?ooo`3oool3
000000H0oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?ooo`2J0?ooo`030000003oool0
oooo06L0oooo00<000000?ooo`3oool02@3oool000d0oooo00D000000?ooo`3oool0oooo00000008
0?ooo`030000003oool0oooo0?l0ooooK`3oool00`000000oooo0?ooo`090?ooo`002`3oool30000
00<0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`3o0?ooofl0oooo00<000000?oo
o`3oool02@3oool000d0oooo00@000000?ooo`3oool0oooo100000050?ooo`<00000o`3ooom^0?oo
o`8000002`3oool001X0oooo00<000000?ooo`3oool0K03oool200000900oooo0P00001^0?ooo`03
0000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo06d0oooo00<000000?ooo`3oool0
S`3oool00`000000oooo0?ooo`1/0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003o
ool0oooo0?l0ooooK`3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`3o
0?ooofl0oooo00<000000?ooo`3oool02@3oool001X0oooo0`00003o0?ooofh0oooo0P00000;0?oo
o`006P3oool00`000000oooo0?ooo`1b0?ooo`800000Q03oool2000007@0oooo00<000000?ooo`3o
ool02@3oool001X0oooo00<000000?ooo`3oool0L`3oool00`000000oooo0?ooo`230?ooo`030000
003oool0oooo0780oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0o`3o
oom_0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo07@0oooo0P00001o
0?ooo`800000M`3oool00`000000oooo0?ooo`090?ooo`006P3oool3000007D0oooo00<000000?oo
o`3oool0OP3oool00`000000oooo0?ooo`1d0?ooo`8000002`3oool001X0oooo00<000000?ooo`3o
ool0M`3oool2000007X0oooo0P00001i0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000
003oool0oooo07P0oooo00<000000?ooo`3oool0N@3oool00`000000oooo0?ooo`1g0?ooo`030000
003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo0?l0ooooK`3oool00`000000oooo0?oo
o`090?ooo`006P3oool300000?l0ooooKP3oool2000000/0oooo000J0?ooo`030000003oool0oooo
07/0oooo0P00001a0?ooo`800000OP3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000
oooo0?ooo`1l0?ooo`030000003oool0oooo0700oooo00<000000?ooo`3oool0O03oool00`000000
oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofl0oooo00<000000?ooo`3oool0
2@3oool000/0oooo100000030?ooo`8000001P3oool00`000000oooo0?ooo`1o0?ooo`800000J@3o
ool200000880oooo00<000000?ooo`3oool02@3oool000/0oooo00<000000?ooo`3oool00`3oool0
10000000oooo0?ooo`0000050?ooo`@00000O`3oool00`000000oooo0?ooo`1X0?ooo`030000003o
ool0oooo07h0oooo0`00000;0?ooo`00303oool00`000000oooo0?ooo`020?ooo`040000003oool0
oooo000000D0oooo00<000000?ooo`3oool0P@3oool2000006H0oooo0P0000230?ooo`030000003o
ool0oooo00T0oooo000=0?ooo`050000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo0880oooo00<000000?ooo`3oool0I@3oool00`000000oooo0?oo
o`210?ooo`030000003oool0oooo00T0oooo000>0?ooo`040000003oool0oooo00000080oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`3o0?ooofl0oooo00<000000?ooo`3oool02@3o
ool000/0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?ooo`0000050?ooo`<00000
o`3ooom^0?ooo`8000002`3oool000`0oooo0P0000040?ooo`8000001P3oool00`000000oooo0?oo
o`260?ooo`800000G03oool2000008P0oooo00<000000?ooo`3oool02@3oool001X0oooo00<00000
0?ooo`3oool0Q`3oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo08H0oooo00<00000
0?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0o`3ooom_0?ooo`030000003oool0oooo
00T0oooo000J0?ooo`030000003oool0oooo08X0oooo0P00001D0?ooo`800000S03oool00`000000
oooo0?ooo`090?ooo`006P3oool3000008/0oooo00<000000?ooo`3oool0D`3oool00`000000oooo
0?ooo`290?ooo`8000002`3oool001X0oooo00<000000?ooo`3oool0o`3ooom_0?ooo`030000003o
ool0oooo00T0oooo000J0?ooo`030000003oool0oooo08d0oooo0P00001>0?ooo`800000S`3oool0
0`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`2>0?ooo`030000003oool0oooo
04d0oooo00<000000?ooo`3oool0S@3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000
oooo0?ooo`2@0?ooo`800000B03oool200000980oooo00<000000?ooo`3oool02@3oool001X0oooo
0`00002A0?ooo`030000003oool0oooo04L0oooo00<000000?ooo`3oool0S`3oool2000000/0oooo
000J0?ooo`030000003oool0oooo0?l0ooooK`3oool00`000000oooo0?ooo`090?ooo`006P3oool0
0`000000oooo0?ooo`2B0?ooo`800000A03oool2000009@0oooo00<000000?ooo`3oool02@3oool0
01X0oooo00<000000?ooo`3oool0T`3oool00`000000oooo0?ooo`130?ooo`030000003oool0oooo
0980oooo00<000000?ooo`3oool02@3oool001X0oooo0`00003o0?ooofh0oooo0P00000;0?ooo`00
6P3oool00`000000oooo0?ooo`2F0?ooo`800000>`3oool2000009T0oooo00<000000?ooo`3oool0
2@3oool001X0oooo00<000000?ooo`3oool0U`3oool00`000000oooo0?ooo`0j0?ooo`030000003o
ool0oooo09L0oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0o`3ooom_
0?ooo`030000003oool0oooo00T0oooo000;0?ooo`@000000P3oool3000000H0oooo00<000000?oo
o`3oool0V@3oool2000003D0oooo0P00002L0?ooo`030000003oool0oooo00T0oooo000;0?ooo`03
0000003oool0oooo00H0oooo00<000000?ooo`3oool00`3oool4000009T0oooo00<000000?ooo`3o
ool0=03oool00`000000oooo0?ooo`2H0?ooo`<000002`3oool000`0oooo00<000000?ooo`3oool0
1@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo09`0oooo0P00000`0?ooo`800000
WP3oool00`000000oooo0?ooo`090?ooo`003@3oool010000000oooo0?ooo`3oool3000000H0oooo
00<000000?ooo`3oool0W@3oool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo09`0oooo
00<000000?ooo`3oool02@3oool000h0oooo00@000000?ooo`3oool00000203oool00`000000oooo
0?ooo`2N0?ooo`800000:`3oool200000:40oooo00<000000?ooo`3oool02@3oool000/0oooo00@0
00000?ooo`3oool000000P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo09l0oooo
00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`2O0?ooo`030000003oool0oooo00T0oooo
000<0?ooo`8000000`3oool4000000D0oooo0`00002P0?ooo`800000:03oool200000:40oooo0P00
000;0?ooo`006P3oool00`000000oooo0?ooo`2Q0?ooo`8000009@3oool2000000030?ooo`000000
oooo0:40oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0XP3oool30000
0280oooo0`00002T0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo0:@0
oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`2S0?ooo`030000003oool0oooo00T0
oooo000J0?ooo`<00000o`3ooom^0?ooo`8000002`3oool001X0oooo00<000000?ooo`3oool0o`3o
oom_0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo0?l0ooooK`3oool0
0`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofl0oooo00<000000?oo
o`3oool02@3oool001X0oooo00<000000?ooo`3oool0o`3ooom_0?ooo`030000003oool0oooo00T0
oooo000J0?ooo`<00000o`3ooom^0?ooo`8000002`3oool001X0oooo00<000000?ooo`3oool0o`3o
oom_0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo0?l0ooooK`3oool0
0`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofl0oooo00<000000?oo
o`3oool02@3oool001X0oooo0`00003o0?ooofh0oooo0P00000;0?ooo`006P3oool00`000000oooo
0?ooo`3o0?ooofl0oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0o`3o
oom_0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo0?l0ooooK`3oool0
0`000000oooo0?ooo`090?ooo`00303oool2000000@0oooo0P0000060?ooo`030000003oool0oooo
0?l0ooooK`3oool00`000000oooo0?ooo`090?ooo`002`3oool010000000oooo0?ooo`0000020?oo
o`040000003oool0oooo000000D0oooo1000002d0?ooo`800000]@3oool3000000/0oooo000>0?oo
o`040000003oool0oooo00000080oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`2f
0?ooo`030000003oool0oooo0;D0oooo00<000000?ooo`3oool02@3oool000`0oooo0P0000030?oo
o`040000003oool0oooo000000D0oooo00<000000?ooo`3oool0o`3ooom_0?ooo`030000003oool0
oooo00T0oooo000>0?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3oool00`3oool0
0`000000oooo0?ooo`3o0?ooofl0oooo00<000000?ooo`3oool02@3oool000/0oooo00@000000?oo
o`3oool000000P3oool010000000oooo0?ooo`0000050?ooo`030000003oool0oooo0?l0ooooK`3o
ool00`000000oooo0?ooo`090?ooo`00303oool2000000@0oooo0P0000060?ooo`<00000o`3ooom^
0?ooo`8000002`3oool001X0oooo00<000000?ooo`3oool0o`3ooom_0?ooo`030000003oool0oooo
00T0oooo000J0?ooo`030000003oool0oooo0?l0ooooK`3oool00`000000oooo0?ooo`090?ooo`00
6P3oool00`000000oooo0?ooo`3o0?ooofl0oooo00<000000?ooo`3oool02@3oool001X0oooo0`00
003o0?ooofh0oooo0P00000;0?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofl0oooo00<00000
0?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0o`3ooom_0?ooo`030000003oool0oooo
00T0oooo000J0?ooo`030000003oool0oooo0?l0ooooK`3oool00`000000oooo0?ooo`090?ooo`00
6P3oool00`000000oooo0?ooo`3o0?ooofl0oooo00<000000?ooo`3oool02@3oool001X0oooo0`00
003o0?ooofh0oooo0P00000;0?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofl0oooo00<00000
0?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0o`3ooom_0?ooo`030000003oool0oooo
00T0oooo000J0?ooo`030000003oool0oooo0?l0ooooK`3oool00`000000oooo0?ooo`090?ooo`00
6P3oool300000?l0ooooKP3oool2000000/0oooo000J0?ooo`030000003oool0oooo0?l0ooooK`3o
ool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofl0oooo00<00000
0?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0603oool00`000000oooo0?ooo`0a0?oo
o`030000003oool0oooo0380oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`0a0?oo
o`030000003oool0oooo0380oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`0G0?oo
o`030000003oool0oooo00T0oooo000<0?ooo`8000000`3oool3000000H0oooo00<000000?ooo`3o
ool00`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo0080oooo00<000000?ooo`3o
ool02@3oool000/0oooo00@000000?ooo`3oool000001@3oool00`000000oooo0?ooo`030?ooool0
0000L`00000;0?ooo`003P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0?l0oooo
P@3oool000`0oooo0P0000030?ooo`<00000o`3ooon40?ooo`003P3oool010000000oooo0?ooo`00
003o0?ooohH0oooo000;0?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3oool0o`3o
oon40?ooo`00\
\>"],
  ImageRangeCache->{{{0, 406}, {202.5, 0}} -> {-10.0973, -7.95261, 0.189817, 
  0.216933}}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(X = AnglePairsAndRationalApproxs[20, 10^\(-3\)]; \nLength[X]\), 
    \(Angles = \(Transpose[X]\)[\([1]\)]; \n
    ListPlot[Angles, \ AspectRatio -> 1/2, Frame -> True, \ Axes -> False, \ 
      PlotRange -> {{\(-5\), 65}, {\(-5\), 35}}]; 
    WriteData["\<~/ASP/WebServices/GrainBreaker/2DGeometries/Geo20.dat\>", 
      X]; \)}], "Input"],

Cell[BoxData[
    \(187\)], "Output"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .5 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0714286 0.0142857 0.0625 0.0125 [
[.07143 -0.0125 -3 -9 ]
[.07143 -0.0125 3 0 ]
[.21429 -0.0125 -6 -9 ]
[.21429 -0.0125 6 0 ]
[.35714 -0.0125 -6 -9 ]
[.35714 -0.0125 6 0 ]
[.5 -0.0125 -6 -9 ]
[.5 -0.0125 6 0 ]
[.64286 -0.0125 -6 -9 ]
[.64286 -0.0125 6 0 ]
[.78571 -0.0125 -6 -9 ]
[.78571 -0.0125 6 0 ]
[.92857 -0.0125 -6 -9 ]
[.92857 -0.0125 6 0 ]
[ 0 0 -0.125 0 ]
[-0.0125 .0625 -6 -4.5 ]
[-0.0125 .0625 0 4.5 ]
[-0.0125 .125 -6 -4.5 ]
[-0.0125 .125 0 4.5 ]
[-0.0125 .1875 -12 -4.5 ]
[-0.0125 .1875 0 4.5 ]
[-0.0125 .25 -12 -4.5 ]
[-0.0125 .25 0 4.5 ]
[-0.0125 .3125 -12 -4.5 ]
[-0.0125 .3125 0 4.5 ]
[-0.0125 .375 -12 -4.5 ]
[-0.0125 .375 0 4.5 ]
[-0.0125 .4375 -12 -4.5 ]
[-0.0125 .4375 0 4.5 ]
[-0.0125 .5 -12 -4.5 ]
[-0.0125 .5 0 4.5 ]
[ 0 0 -0.125 0 ]
[ 0 .5 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .5 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.07143 0 m
.07143 .00625 L
s
[(0)] .07143 -0.0125 0 1 Mshowa
.21429 0 m
.21429 .00625 L
s
[(10)] .21429 -0.0125 0 1 Mshowa
.35714 0 m
.35714 .00625 L
s
[(20)] .35714 -0.0125 0 1 Mshowa
.5 0 m
.5 .00625 L
s
[(30)] .5 -0.0125 0 1 Mshowa
.64286 0 m
.64286 .00625 L
s
[(40)] .64286 -0.0125 0 1 Mshowa
.78571 0 m
.78571 .00625 L
s
[(50)] .78571 -0.0125 0 1 Mshowa
.92857 0 m
.92857 .00625 L
s
[(60)] .92857 -0.0125 0 1 Mshowa
.125 Mabswid
.1 0 m
.1 .00375 L
s
.12857 0 m
.12857 .00375 L
s
.15714 0 m
.15714 .00375 L
s
.18571 0 m
.18571 .00375 L
s
.24286 0 m
.24286 .00375 L
s
.27143 0 m
.27143 .00375 L
s
.3 0 m
.3 .00375 L
s
.32857 0 m
.32857 .00375 L
s
.38571 0 m
.38571 .00375 L
s
.41429 0 m
.41429 .00375 L
s
.44286 0 m
.44286 .00375 L
s
.47143 0 m
.47143 .00375 L
s
.52857 0 m
.52857 .00375 L
s
.55714 0 m
.55714 .00375 L
s
.58571 0 m
.58571 .00375 L
s
.61429 0 m
.61429 .00375 L
s
.67143 0 m
.67143 .00375 L
s
.7 0 m
.7 .00375 L
s
.72857 0 m
.72857 .00375 L
s
.75714 0 m
.75714 .00375 L
s
.81429 0 m
.81429 .00375 L
s
.84286 0 m
.84286 .00375 L
s
.87143 0 m
.87143 .00375 L
s
.9 0 m
.9 .00375 L
s
.04286 0 m
.04286 .00375 L
s
.01429 0 m
.01429 .00375 L
s
.95714 0 m
.95714 .00375 L
s
.98571 0 m
.98571 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
0 .0625 m
.00625 .0625 L
s
[(0)] -0.0125 .0625 1 0 Mshowa
0 .125 m
.00625 .125 L
s
[(5)] -0.0125 .125 1 0 Mshowa
0 .1875 m
.00625 .1875 L
s
[(10)] -0.0125 .1875 1 0 Mshowa
0 .25 m
.00625 .25 L
s
[(15)] -0.0125 .25 1 0 Mshowa
0 .3125 m
.00625 .3125 L
s
[(20)] -0.0125 .3125 1 0 Mshowa
0 .375 m
.00625 .375 L
s
[(25)] -0.0125 .375 1 0 Mshowa
0 .4375 m
.00625 .4375 L
s
[(30)] -0.0125 .4375 1 0 Mshowa
0 .5 m
.00625 .5 L
s
[(35)] -0.0125 .5 1 0 Mshowa
.125 Mabswid
0 .0125 m
.00375 .0125 L
s
0 .025 m
.00375 .025 L
s
0 .0375 m
.00375 .0375 L
s
0 .05 m
.00375 .05 L
s
0 .075 m
.00375 .075 L
s
0 .0875 m
.00375 .0875 L
s
0 .1 m
.00375 .1 L
s
0 .1125 m
.00375 .1125 L
s
0 .1375 m
.00375 .1375 L
s
0 .15 m
.00375 .15 L
s
0 .1625 m
.00375 .1625 L
s
0 .175 m
.00375 .175 L
s
0 .2 m
.00375 .2 L
s
0 .2125 m
.00375 .2125 L
s
0 .225 m
.00375 .225 L
s
0 .2375 m
.00375 .2375 L
s
0 .2625 m
.00375 .2625 L
s
0 .275 m
.00375 .275 L
s
0 .2875 m
.00375 .2875 L
s
0 .3 m
.00375 .3 L
s
0 .325 m
.00375 .325 L
s
0 .3375 m
.00375 .3375 L
s
0 .35 m
.00375 .35 L
s
0 .3625 m
.00375 .3625 L
s
0 .3875 m
.00375 .3875 L
s
0 .4 m
.00375 .4 L
s
0 .4125 m
.00375 .4125 L
s
0 .425 m
.00375 .425 L
s
0 .45 m
.00375 .45 L
s
0 .4625 m
.00375 .4625 L
s
0 .475 m
.00375 .475 L
s
0 .4875 m
.00375 .4875 L
s
.25 Mabswid
0 0 m
0 .5 L
s
.07143 .49375 m
.07143 .5 L
s
.21429 .49375 m
.21429 .5 L
s
.35714 .49375 m
.35714 .5 L
s
.5 .49375 m
.5 .5 L
s
.64286 .49375 m
.64286 .5 L
s
.78571 .49375 m
.78571 .5 L
s
.92857 .49375 m
.92857 .5 L
s
.125 Mabswid
.1 .49625 m
.1 .5 L
s
.12857 .49625 m
.12857 .5 L
s
.15714 .49625 m
.15714 .5 L
s
.18571 .49625 m
.18571 .5 L
s
.24286 .49625 m
.24286 .5 L
s
.27143 .49625 m
.27143 .5 L
s
.3 .49625 m
.3 .5 L
s
.32857 .49625 m
.32857 .5 L
s
.38571 .49625 m
.38571 .5 L
s
.41429 .49625 m
.41429 .5 L
s
.44286 .49625 m
.44286 .5 L
s
.47143 .49625 m
.47143 .5 L
s
.52857 .49625 m
.52857 .5 L
s
.55714 .49625 m
.55714 .5 L
s
.58571 .49625 m
.58571 .5 L
s
.61429 .49625 m
.61429 .5 L
s
.67143 .49625 m
.67143 .5 L
s
.7 .49625 m
.7 .5 L
s
.72857 .49625 m
.72857 .5 L
s
.75714 .49625 m
.75714 .5 L
s
.81429 .49625 m
.81429 .5 L
s
.84286 .49625 m
.84286 .5 L
s
.87143 .49625 m
.87143 .5 L
s
.9 .49625 m
.9 .5 L
s
.04286 .49625 m
.04286 .5 L
s
.01429 .49625 m
.01429 .5 L
s
.95714 .49625 m
.95714 .5 L
s
.98571 .49625 m
.98571 .5 L
s
.25 Mabswid
0 .5 m
1 .5 L
s
.99375 0 m
1 0 L
s
.99375 .0625 m
1 .0625 L
s
.99375 .125 m
1 .125 L
s
.99375 .1875 m
1 .1875 L
s
.99375 .25 m
1 .25 L
s
.99375 .3125 m
1 .3125 L
s
.99375 .375 m
1 .375 L
s
.99375 .4375 m
1 .4375 L
s
.125 Mabswid
.99625 .0125 m
1 .0125 L
s
.99625 .025 m
1 .025 L
s
.99625 .0375 m
1 .0375 L
s
.99625 .05 m
1 .05 L
s
.99625 .075 m
1 .075 L
s
.99625 .0875 m
1 .0875 L
s
.99625 .1 m
1 .1 L
s
.99625 .1125 m
1 .1125 L
s
.99625 .1375 m
1 .1375 L
s
.99625 .15 m
1 .15 L
s
.99625 .1625 m
1 .1625 L
s
.99625 .175 m
1 .175 L
s
.99625 .2 m
1 .2 L
s
.99625 .2125 m
1 .2125 L
s
.99625 .225 m
1 .225 L
s
.99625 .2375 m
1 .2375 L
s
.99625 .2625 m
1 .2625 L
s
.99625 .275 m
1 .275 L
s
.99625 .2875 m
1 .2875 L
s
.99625 .3 m
1 .3 L
s
.99625 .325 m
1 .325 L
s
.99625 .3375 m
1 .3375 L
s
.99625 .35 m
1 .35 L
s
.99625 .3625 m
1 .3625 L
s
.99625 .3875 m
1 .3875 L
s
.99625 .4 m
1 .4 L
s
.99625 .4125 m
1 .4125 L
s
.99625 .425 m
1 .425 L
s
.99625 .45 m
1 .45 L
s
.99625 .4625 m
1 .4625 L
s
.99625 .475 m
1 .475 L
s
.99625 .4875 m
1 .4875 L
s
.25 Mabswid
1 0 m
1 .5 L
s
0 0 m
1 0 L
1 .5 L
0 .5 L
closepath
clip
newpath
.008 w
.07143 .0625 Mdot
.09197 .08047 Mdot
.09393 .08219 Mdot
.09629 .08426 Mdot
.09922 .08681 Mdot
.10292 .09005 Mdot
.10776 .09429 Mdot
.11435 .0625 Mdot
.11435 .10006 Mdot
.11863 .0625 Mdot
.11863 .1038 Mdot
.12386 .10838 Mdot
.1304 .1141 Mdot
.13295 .11633 Mdot
.13879 .12144 Mdot
.14584 .12761 Mdot
.14995 .1312 Mdot
.15454 .13522 Mdot
.15704 .13741 Mdot
.16553 .14483 Mdot
.17587 .15389 Mdot
.17984 .15736 Mdot
.18324 .16034 Mdot
.18876 .16517 Mdot
.19484 .17049 Mdot
.19926 .17435 Mdot
.20524 .17959 Mdot
.20911 .18297 Mdot
.21181 .18533 Mdot
.22705 .19867 Mdot
.24081 .21071 Mdot
.24298 .2126 Mdot
.24595 .2152 Mdot
.25028 .219 Mdot
.2572 .22505 Mdot
.26247 .22966 Mdot
.26997 .23622 Mdot
.27702 .2424 Mdot
.2815 .24631 Mdot
.28684 .25099 Mdot
.28993 .25369 Mdot
.30146 .26378 Mdot
.31181 .27283 Mdot
.31423 .27495 Mdot
.31814 .27837 Mdot
.32115 .11633 Mdot
.32115 .281 Mdot
.32548 .2848 Mdot
.33226 .29073 Mdot
.33546 .0625 Mdot
.33546 .1038 Mdot
.33546 .29353 Mdot
.34438 .16034 Mdot
.34438 .30133 Mdot
.35239 .17435 Mdot
.35239 .30834 Mdot
.35488 .08681 Mdot
.35488 .31052 Mdot
.35962 .31467 Mdot
.36407 .15389 Mdot
.36407 .31856 Mdot
.37217 .32565 Mdot
.37936 .33194 Mdot
.38267 .0625 Mdot
.38267 .33483 Mdot
.3858 .33757 Mdot
.39158 .24631 Mdot
.39158 .34264 Mdot
.39925 .34934 Mdot
.4059 .35517 Mdot
.41174 .2424 Mdot
.41174 .36027 Mdot
.41689 .25369 Mdot
.41689 .36478 Mdot
.42148 .3688 Mdot
.42559 .0625 Mdot
.42559 .10006 Mdot
.42559 .37239 Mdot
.43264 .37856 Mdot
.43848 .219 Mdot
.43848 .38367 Mdot
.44338 .38796 Mdot
.44756 .39162 Mdot
.45117 .39478 Mdot
.45432 .39753 Mdot
.45708 .12761 Mdot
.45708 .39994 Mdot
.45953 .40209 Mdot
.46171 .19867 Mdot
.46171 .404 Mdot
.46367 .40571 Mdot
.5 .09005 Mdot
.5 .11633 Mdot
.5 .16517 Mdot
.5 .27283 Mdot
.5 .281 Mdot
.5 .4375 Mdot
.53633 .40571 Mdot
.53829 .19867 Mdot
.53829 .404 Mdot
.54047 .40209 Mdot
.54292 .12761 Mdot
.54292 .39994 Mdot
.54568 .39753 Mdot
.54883 .39478 Mdot
.55244 .39162 Mdot
.55662 .38796 Mdot
.56152 .219 Mdot
.56152 .38367 Mdot
.56736 .37856 Mdot
.57441 .0625 Mdot
.57441 .10006 Mdot
.57441 .37239 Mdot
.57852 .3688 Mdot
.58311 .25369 Mdot
.58311 .36478 Mdot
.58826 .2424 Mdot
.58826 .36027 Mdot
.5941 .35517 Mdot
.60075 .34934 Mdot
.60842 .24631 Mdot
.60842 .34264 Mdot
.6142 .33757 Mdot
.61733 .0625 Mdot
.61733 .33483 Mdot
.62064 .33194 Mdot
.62783 .32565 Mdot
.63593 .15389 Mdot
.63593 .31856 Mdot
.64038 .31467 Mdot
.64512 .08681 Mdot
.64512 .31052 Mdot
.64761 .17435 Mdot
.64761 .30834 Mdot
.65562 .16034 Mdot
.65562 .30133 Mdot
.66454 .0625 Mdot
.66454 .1038 Mdot
.66454 .29353 Mdot
.66774 .29073 Mdot
.67452 .2848 Mdot
.67885 .11633 Mdot
.67885 .281 Mdot
.68186 .27837 Mdot
.68577 .27495 Mdot
.68819 .27283 Mdot
.69854 .26378 Mdot
.71007 .25369 Mdot
.71316 .25099 Mdot
.7185 .24631 Mdot
.72298 .2424 Mdot
.73003 .23622 Mdot
.73753 .22966 Mdot
.7428 .22505 Mdot
.74972 .219 Mdot
.75405 .2152 Mdot
.75702 .2126 Mdot
.75919 .21071 Mdot
.77295 .19867 Mdot
.78819 .18533 Mdot
.79089 .18297 Mdot
.79476 .17959 Mdot
.80074 .17435 Mdot
.80516 .17049 Mdot
.81124 .16517 Mdot
.81676 .16034 Mdot
.82016 .15736 Mdot
.82413 .15389 Mdot
.83447 .14483 Mdot
.84296 .13741 Mdot
.84546 .13522 Mdot
.85005 .1312 Mdot
.85416 .12761 Mdot
.86121 .12144 Mdot
.86705 .11633 Mdot
.8696 .1141 Mdot
.87614 .10838 Mdot
.88137 .0625 Mdot
.88137 .1038 Mdot
.88565 .0625 Mdot
.88565 .10006 Mdot
.89224 .09429 Mdot
.89708 .09005 Mdot
.90078 .08681 Mdot
.90371 .08426 Mdot
.90607 .08219 Mdot
.90803 .08047 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{401, 200.5},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006A0000b2000`400?l00000o`00003oo`3ooonB
0?ooo`00o`3ooonB0?ooo`00o`3ooonB0?ooo`00o`3ooonB0?ooo`00<`3oool2000002h0oooo1000
00030?ooo`800000:`3oool4000000<0oooo0P00000/0?ooo`800000103oool2000002d0oooo00<0
00000?ooo`3oool00P3oool2000002X0oooo0`0000040?ooo`800000;03oool2000000@0oooo0P00
000R0?ooo`00<P3oool010000000oooo0?ooo`00000_0?ooo`050000003oool0oooo0?ooo`000000
0P3oool00`000000oooo0?ooo`0X0?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool0
0000:P3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000002`0oooo00D00000
0?ooo`3oool0oooo000000020?ooo`030000003oool0oooo02X0oooo00@000000?ooo`3oool00000
0P3oool00`000000oooo0?ooo`0X0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3o
ool000008@3oool00380oooo00@000000?ooo`3oool00000;`3oool01@000000oooo0?ooo`3oool0
00000080oooo00<000000?ooo`3oool0:@3oool00`000000oooo0?ooo`020?ooo`040000003oool0
oooo000002d0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`0W0?ooo`D00000
0P3oool010000000oooo0?ooo`00000/0?ooo`040000003oool0oooo00000080oooo00<000000?oo
o`3oool0:03oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo00000240oooo000b
0?ooo`040000003oool0oooo000002l0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000
003oool0oooo02X0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003oool0oooo02T0
oooo0P0000030?ooo`040000003oool0oooo000002T0oooo00@000000?ooo`3oool000000`3oool0
10000000oooo0?ooo`00000Y0?ooo`<000000`3oool010000000oooo0?ooo`00000Z0?ooo`<00000
0`3oool010000000oooo0?ooo`00000Q0?ooo`00<P3oool010000000oooo0?ooo`00000_0?ooo`05
0000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`0[0?ooo`040000003oool0oooo
00000080oooo00<000000?ooo`3oool0:`3oool010000000oooo0?ooo`0000020?ooo`030000003o
ool0oooo02P0oooo00<000000?ooo`0000000`3oool010000000oooo0?ooo`00000Y0?ooo`030000
003oool0oooo00<0oooo00@000000?ooo`3oool00000:P3oool00`000000oooo0?ooo`030?ooo`04
0000003oool0oooo00000240oooo000b0?ooo`040000003oool0oooo000002d0oooo0`0000030?oo
o`040000003oool0oooo000002X0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?oo
o`00000Z0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool00000:`3oool20000
00<0oooo00@000000?ooo`3oool00000:@3oool00`000000oooo0?ooo`030?ooo`040000003oool0
oooo000002X0oooo00<000000?ooo`3oool00`3oool010000000oooo0?ooo`00000Q0?ooo`00<`3o
ool200000300oooo00<000000?ooo`3oool00P3oool2000002`0oooo0P0000040?ooo`800000;03o
ool2000000@0oooo0P00000]0?ooo`030000003oool0oooo0080oooo0P00000Z0?ooo`@000000`3o
ool2000002`0oooo0`0000030?ooo`8000008P3oool00?l0ooooTP3oool00?l0ooooTP3oool00?l0
ooooTP3oool001X0ooooo`00001]000000/0oooo000J0?ooo`030000003oool0oooo0080oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00T0oooo000J
0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`0a
0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`0`
0?ooo`030000003oool0oooo0340oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`09
0?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofT0oooo00<000000?ooo`3oool02@3oool001X0
oooo0`00003o0?ooofP0oooo0P00000;0?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofT0oooo
00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0o`3ooomY0?ooo`030000003o
ool0oooo00T0oooo000J0?ooo`030000003oool0oooo0?l0ooooJ@3oool00`000000oooo0?ooo`09
0?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofT0oooo00<000000?ooo`3oool02@3oool001X0
oooo0`00003o0?ooofP0oooo0P00000;0?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofT0oooo
00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0o`3ooomY0?ooo`030000003o
ool0oooo00T0oooo000J0?ooo`030000003oool0oooo0?l0ooooJ@3oool00`000000oooo0?ooo`09
0?ooo`006P3oool300000?l0ooooJ03oool2000000/0oooo000J0?ooo`030000003oool0oooo0?l0
ooooJ@3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofT0oooo
00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0o`3ooomY0?ooo`030000003o
ool0oooo00T0oooo000J0?ooo`030000003oool0oooo0?l0ooooJ@3oool00`000000oooo0?ooo`09
0?ooo`006P3oool300000?l0ooooJ03oool2000000/0oooo000J0?ooo`030000003oool0oooo0?l0
ooooJ@3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofT0oooo
00<000000?ooo`3oool02@3oool00180oooo0P0000060?ooo`030000003oool0oooo0?l0ooooJ@3o
ool00`000000oooo0?ooo`090?ooo`004@3oool010000000oooo0?ooo`0000050?ooo`@000005@3o
ool2000000h0oooo0`00001=0?ooo`8000003`3oool2000000h0oooo0P00000d0?ooo`8000003@3o
ool2000000l0oooo0P00001=0?ooo`@000009P3oool3000000/0oooo000A0?ooo`040000003oool0
oooo000000D0oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`0=0?ooo`800000CP3o
ool00`000000oooo0?ooo`0>0?ooo`030000003oool0oooo00d0oooo00<000000?ooo`3oool0<`3o
ool00`000000oooo0?ooo`0<0?ooo`030000003oool0oooo00h0oooo00<000000?ooo`3oool0C03o
ool00`000000oooo0000000X0?ooo`030000003oool0oooo00T0oooo000A0?ooo`040000003oool0
oooo000000D0oooo00<000000?ooo`3oool0o`3ooomY0?ooo`030000003oool0oooo00T0oooo000A
0?ooo`040000003oool0oooo000000D0oooo00<000000?ooo`3oool0o`3ooomY0?ooo`030000003o
ool0oooo00T0oooo000A0?ooo`040000003oool0oooo000000D0oooo00<000000?ooo`3oool0o`3o
oomY0?ooo`030000003oool0oooo00T0oooo000B0?ooo`8000001P3oool300000?l0ooooJ03oool2
000000/0oooo000J0?ooo`030000003oool0oooo0?l0ooooJ@3oool00`000000oooo0?ooo`090?oo
o`006P3oool00`000000oooo0?ooo`0N0?ooo`800000o`3ooolV0?ooo`<00000803oool00`000000
oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`0O0?ooo`800000o`3ooolT0?ooo`@00000
803oool00`000000oooo0?ooo`090?ooo`006P3oool300000200oooo0P00001K0?ooo`800000I`3o
ool2000005/0oooo0`00000Q0?ooo`8000002`3oool001X0oooo00<000000?ooo`3oool08@3oool3
000005X0oooo00<000000?ooo`3oool0<@3oool2000003<0oooo00<000000?ooo`3oool0F03oool3
000002<0oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool08`3oool00`00
0000oooo0?ooo`2=0?ooo`030000003oool0oooo08d0oooo00<000000?ooo`3oool08P3oool00`00
0000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`0S0?ooo`800000o`3ooolL0?ooo`80
00009P3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`0T0?ooo`030000
003oool0oooo0?l0oooo6`3oool00`000000oooo0?ooo`0T0?ooo`030000003oool0oooo00T0oooo
000J0?ooo`<000009P3oool2000006l0oooo0P00000d0?ooo`800000K`3oool2000002L0oooo0P00
000;0?ooo`006P3oool00`000000oooo0?ooo`0W0?ooo`800000C@3oool200000200oooo00<00000
0?ooo`3oool0<`3oool00`000000oooo0?ooo`0L0?ooo`800000C@3oool2000000030?ooo`000000
oooo02L0oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0:03oool00`00
0000oooo0?ooo`1<0?ooo`030000003oool0oooo07@0oooo00<000000?ooo`3oool0C03oool00`00
0000oooo0?ooo`0X0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo02T0
oooo0P00003o0?oooa00oooo0P00000/0?ooo`030000003oool0oooo00T0oooo000J0?ooo`<00000
:P3oool00`000000oooo0?ooo`3o0?ooo`l0oooo00<000000?ooo`3oool0:@3oool2000000/0oooo
000J0?ooo`030000003oool0oooo02`0oooo0P00003o0?ooo`/0oooo0P00000^0?ooo`030000003o
ool0oooo00T0oooo000J0?ooo`030000003oool0oooo02d0oooo0P0000120?ooo`800000?`3oool2
000003l0oooo0P0000120?ooo`<00000;P3oool00`000000oooo0?ooo`090?ooo`006P3oool00`00
0000oooo0?ooo`0^0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool0?P3oool00`00
0000oooo0?ooo`0n0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool0;@3oool00`00
0000oooo0?ooo`090?ooo`004@3oool3000000H0oooo00<000000?ooo`3oool0;`3oool200000?l0
oooo1@3oool200000340oooo00<000000?ooo`3oool02@3oool001@0oooo00<000000?ooo`3oool0
0`3oool4000002l0oooo00<000000?ooo`3oool0o`3oool40?ooo`030000003oool0oooo02d0oooo
0`00000;0?ooo`00503oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0340oooo0P00
001_0?ooo`8000007@3oool2000006l0oooo0P00000d0?ooo`030000003oool0oooo00T0oooo000A
0?ooo`<000001P3oool00`000000oooo0?ooo`0b0?ooo`<00000KP3oool00`000000oooo0?ooo`0L
0?ooo`030000003oool0oooo06`0oooo0`00000d0?ooo`030000003oool0oooo00T0oooo000A0?oo
o`030000003oool0oooo00H0oooo00<000000?ooo`3oool0=03oool00`000000oooo0?ooo`3k0?oo
o`030000003oool0oooo03<0oooo00<000000?ooo`3oool02@3oool00140oooo00<000000?ooo`3o
ool01P3oool3000003@0oooo0`00003g0?ooo`<00000=P3oool2000000/0oooo000A0?ooo`@00000
1@3oool00`000000oooo0?ooo`0e0?ooo`800000n03oool2000003L0oooo00<000000?ooo`3oool0
2@3oool001X0oooo00<000000?ooo`3oool0o`3ooomY0?ooo`030000003oool0oooo00T0oooo000J
0?ooo`030000003oool0oooo03P0oooo0P00003a0?ooo`800000>`3oool00`000000oooo0?ooo`09
0?ooo`006P3oool00`000000oooo0?ooo`0i0?ooo`030000003oool0oooo0?00oooo00<000000?oo
o`3oool0>@3oool00`000000oooo0?ooo`090?ooo`006P3oool300000?l0ooooJ03oool2000000/0
oooo000J0?ooo`030000003oool0oooo03`0oooo0P0000120?ooo`800000H@3oool200000480oooo
0P00000o0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo03d0oooo00<0
00000?ooo`3oool0@@3oool00`000000oooo0?ooo`1P0?ooo`030000003oool0oooo0440oooo00<0
00000?ooo`3oool0?@3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`0n
0?ooo`800000iP3oool200000400oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?oo
o`3oool0?`3oool2000003P0oooo0P00001_0?ooo`800000>@3oool300000400oooo00<000000?oo
o`3oool02@3oool001X0oooo0`0000100?ooo`<00000=`3oool00`000000oooo0?ooo`0e0?ooo`80
0000=`3oool00`000000oooo0?ooo`0e0?ooo`80000000<0oooo0000003oool0?`3oool2000000/0
oooo000J0?ooo`030000003oool0oooo0480oooo00<000000?ooo`3oool0KP3oool00`000000oooo
0?ooo`1^0?ooo`030000003oool0oooo0440oooo00<000000?ooo`3oool02@3oool001X0oooo00<0
00000?ooo`3oool0@`3oool200000=`0oooo0P0000150?ooo`030000003oool0oooo00T0oooo000J
0?ooo`030000003oool0oooo04@0oooo00<000000?ooo`3oool0f`3oool00`000000oooo0?ooo`13
0?ooo`030000003oool0oooo00T0oooo000J0?ooo`<00000A@3oool2000003D0oooo0P00001Y0?oo
o`800000=P3oool2000004H0oooo0P00000;0?ooo`006P3oool00`000000oooo0?ooo`160?ooo`03
0000003oool0oooo03@0oooo00<000000?ooo`3oool0J03oool00`000000oooo0?ooo`0e0?ooo`03
0000003oool0oooo04D0oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0
A`3oool200000=@0oooo0P0000190?ooo`030000003oool0oooo00T0oooo000;0?ooo`@000000`3o
ool2000000H0oooo00<000000?ooo`3oool0B03oool200000=40oooo0P0000000`3oool000000?oo
o`180?ooo`030000003oool0oooo00T0oooo000=0?ooo`050000003oool0oooo0?ooo`0000000P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo04T0oooo0P00003?0?ooo`<00000B`3o
ool00`000000oooo0?ooo`090?ooo`003@3oool01@000000oooo0?ooo`3oool000000080oooo00<0
00000?ooo`3oool00`3oool4000004T0oooo00<000000?ooo`3oool0cP3oool00`000000oooo0?oo
o`180?ooo`<000002`3oool000d0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003o
ool0oooo00<0oooo00<000000?ooo`3oool0o`3ooomY0?ooo`030000003oool0oooo00T0oooo000=
0?ooo`050000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`030?ooo`030000003o
ool0oooo0?l0ooooJ@3oool00`000000oooo0?ooo`090?ooo`002`3oool3000000<0oooo00@00000
0?ooo`3oool000001@3oool00`000000oooo0?ooo`3o0?ooofT0oooo00<000000?ooo`3oool02@3o
ool000d0oooo00<000000?ooo`3oool00P3oool2000000H0oooo0`00001?0?ooo`800000D`3oool2
000001X0oooo0P00001C0?ooo`800000D03oool2000000/0oooo000J0?ooo`030000003oool0oooo
0500oooo00<000000?ooo`3oool0DP3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo
0580oooo00<000000?ooo`3oool0C`3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000
oooo0?ooo`3o0?ooofT0oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0
o`3ooomY0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo05@0oooo0P00
002j0?ooo`800000EP3oool00`000000oooo0?ooo`090?ooo`006P3oool3000005@0oooo0P00002i
0?ooo`<00000E@3oool2000000/0oooo000J0?ooo`030000003oool0oooo05D0oooo0`00002f0?oo
o`<00000E`3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`1G0?ooo`80
0000@P3oool2000002/0oooo0P0000120?ooo`80000000<0oooo0000003oool0E`3oool00`000000
oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo0440oooo
00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`110?ooo`030000003oool0oooo05P0oooo
00<000000?ooo`3oool02@3oool001X0oooo0`00001J0?ooo`800000[P3oool2000005/0oooo0P00
000;0?ooo`006P3oool00`000000oooo0?ooo`1K0?ooo`030000003oool0oooo0:d0oooo00<00000
0?ooo`3oool0FP3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`1L0?oo
o`800000ZP3oool2000005h0oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3o
ool0G@3oool00`000000oooo0?ooo`2Y0?ooo`030000003oool0oooo05`0oooo00<000000?ooo`3o
ool02@3oool001X0oooo00<000000?ooo`3oool0GP3oool200000:D0oooo0P00001Q0?ooo`030000
003oool0oooo00T0oooo000J0?ooo`<00000G`3oool00`000000oooo0?ooo`2T0?ooo`030000003o
ool0oooo05h0oooo0P00000;0?ooo`006P3oool00`000000oooo0?ooo`1Q0?ooo`800000;`3oool2
000003h0oooo0P00000_0?ooo`800000H`3oool00`000000oooo0?ooo`090?ooo`006P3oool00`00
0000oooo0?ooo`1R0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0?@3oool00`00
0000oooo0?ooo`0^0?ooo`030000003oool0oooo0640oooo00<000000?ooo`3oool02@3oool000/0
oooo100000020?ooo`<000001P3oool00`000000oooo0?ooo`1R0?ooo`8000009P3oool2000004d0
oooo0P00000V0?ooo`800000I@3oool00`000000oooo0?ooo`090?ooo`003@3oool00`000000oooo
0?ooo`040?ooo`030000003oool0oooo00<0oooo1000001R0?ooo`030000003oool0oooo02D0oooo
00<000000?ooo`3oool0C03oool00`000000oooo0?ooo`0U0?ooo`030000003oool0oooo0640oooo
0`00000;0?ooo`003@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool0I03oool2000009T0oooo0P00001W0?ooo`030000003oool0oooo00T0oooo000=
0?ooo`040000003oool0oooo0?ooo`<000001P3oool00`000000oooo0?ooo`1U0?ooo`<00000;03o
ool2000003X0oooo0P00000/0?ooo`<00000I`3oool00`000000oooo0?ooo`090?ooo`003@3oool0
1@000000oooo0?ooo`3oool0000000P0oooo00<000000?ooo`3oool0I`3oool00`000000oooo0?oo
o`0[0?ooo`030000003oool0oooo03T0oooo00<000000?ooo`3oool0:`3oool00`000000oooo0?oo
o`1V0?ooo`030000003oool0oooo00T0oooo000;0?ooo`<000000`3oool00`000000oooo0?ooo`06
0?ooo`030000003oool0oooo0?l0ooooJ@3oool00`000000oooo0?ooo`090?ooo`003@3oool01000
0000oooo0?ooo`3oool4000000D0oooo0`00001Z0?ooo`800000SP3oool2000006/0oooo0P00000;
0?ooo`006P3oool00`000000oooo0?ooo`1[0?ooo`030000003oool0oooo08d0oooo00<000000?oo
o`3oool0JP3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofT0
oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0K@3oool2000004<0oooo
0P0000120?ooo`800000L03oool00`000000oooo0?ooo`090?ooo`006P3oool3000006h0oooo0P00
00130?ooo`030000003oool0oooo03l0oooo0`00001_0?ooo`8000002`3oool001X0oooo00<00000
0?ooo`3oool0K`3oool300000880oooo0`00001a0?ooo`030000003oool0oooo00T0oooo000J0?oo
o`030000003oool0oooo0740oooo0P00000o0?ooo`800000?`3oool300000780oooo00<000000?oo
o`3oool02@3oool001X0oooo00<000000?ooo`3oool0LP3oool00`000000oooo0?ooo`0n0?ooo`03
0000003oool0oooo03h0oooo00<000000?ooo`3oool0L@3oool00`000000oooo0?ooo`090?ooo`00
6P3oool00`000000oooo0?ooo`1b0?ooo`800000O@3oool2000007D0oooo00<000000?ooo`3oool0
2@3oool001X0oooo0`00001c0?ooo`030000003oool0oooo07`0oooo00<000000?ooo`3oool0LP3o
ool2000000/0oooo000J0?ooo`030000003oool0oooo07D0oooo0P00001h0?ooo`800000M`3oool0
0`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`1f0?ooo`800000M@3oool20000
00030?ooo`000000oooo07H0oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3o
ool0M`3oool00`000000oooo0?ooo`1d0?ooo`030000003oool0oooo07L0oooo00<000000?ooo`3o
ool02@3oool001X0oooo0`00003o0?ooofP0oooo0P00000;0?ooo`006P3oool00`000000oooo0?oo
o`1i0?ooo`800000K`3oool2000007`0oooo00<000000?ooo`3oool02@3oool001X0oooo00<00000
0?ooo`3oool0NP3oool00`000000oooo0?ooo`1^0?ooo`030000003oool0oooo07X0oooo00<00000
0?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0O03oool2000006T0oooo0P00001o0?oo
o`030000003oool0oooo00T0oooo000;0?ooo`@000000`3oool2000000H0oooo00<000000?ooo`3o
ool0O@3oool2000006L0oooo0`00001o0?ooo`030000003oool0oooo00T0oooo000;0?ooo`030000
003oool0oooo00<0oooo00@000000?ooo`3oool000001@3oool4000007d0oooo00<000000?ooo`3o
ool0IP3oool00`000000oooo0?ooo`1l0?ooo`<000002`3oool000`0oooo00<000000?ooo`3oool0
0P3oool010000000oooo0?ooo`0000050?ooo`030000003oool0oooo07l0oooo0P00001T0?ooo`80
0000P@3oool00`000000oooo0?ooo`090?ooo`003@3oool01@000000oooo0?ooo`3oool000000080
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`200?ooo`800000H@3oool200000003
0?ooo`000000oooo0800oooo00<000000?ooo`3oool02@3oool000h0oooo00@000000?ooo`3oool0
00000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0840oooo00<000000?ooo`3o
ool0H03oool00`000000oooo0?ooo`210?ooo`030000003oool0oooo00T0oooo000;0?ooo`040000
003oool0oooo00000080oooo00@000000?ooo`3oool000001@3oool300000?l0ooooJ03oool20000
00/0oooo000<0?ooo`800000103oool2000000H0oooo00<000000?ooo`3oool0P`3oool2000005/0
oooo0P0000260?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo08@0oooo
00<000000?ooo`3oool0FP3oool00`000000oooo0?ooo`240?ooo`030000003oool0oooo00T0oooo
000J0?ooo`030000003oool0oooo08H0oooo0P00001F0?ooo`800000R03oool00`000000oooo0?oo
o`090?ooo`006P3oool00`000000oooo0?ooo`270?ooo`800000D`3oool2000000030?ooo`000000
oooo08L0oooo00<000000?ooo`3oool02@3oool001X0oooo0`0000280?ooo`800000D@3oool30000
08T0oooo0P00000;0?ooo`006P3oool00`000000oooo0?ooo`290?ooo`030000003oool0oooo0500
oooo00<000000?ooo`3oool0R@3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo
0?ooo`2:0?ooo`800000C@3oool2000008d0oooo00<000000?ooo`3oool02@3oool001X0oooo00<0
00000?ooo`3oool0R`3oool00`000000oooo0?ooo`1<0?ooo`030000003oool0oooo08/0oooo00<0
00000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0S@3oool2000004L0oooo0P00002@
0?ooo`030000003oool0oooo00T0oooo000J0?ooo`<00000SP3oool00`000000oooo0?ooo`160?oo
o`030000003oool0oooo08d0oooo0P00000;0?ooo`006P3oool00`000000oooo0?ooo`2@0?ooo`80
0000@P3oool200000980oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0
T@3oool00`000000oooo0?ooo`110?ooo`030000003oool0oooo0900oooo00<000000?ooo`3oool0
2@3oool001X0oooo00<000000?ooo`3oool0TP3oool2000003h0oooo0P00002D0?ooo`030000003o
ool0oooo00T0oooo000J0?ooo`<00000T`3oool00`000000oooo0?ooo`0m0?ooo`030000003oool0
oooo0940oooo0P00000;0?ooo`006P3oool00`000000oooo0?ooo`2D0?ooo`800000>P3oool20000
09H0oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0U@3oool2000003L0
oooo0P0000000`3oool000000?ooo`2E0?ooo`030000003oool0oooo00T0oooo000;0?ooo`@00000
0P3oool3000000H0oooo00<000000?ooo`3oool0UP3oool00`000000oooo0?ooo`0f0?ooo`030000
003oool0oooo09H0oooo00<000000?ooo`3oool02@3oool000/0oooo00<000000?ooo`3oool01P3o
ool00`000000oooo0?ooo`030?ooo`030000003oool0oooo09L0oooo0P00000d0?ooo`800000V@3o
ool00`000000oooo0?ooo`090?ooo`00303oool00`000000oooo0?ooo`050?ooo`030000003oool0
oooo00<0oooo1000002G0?ooo`030000003oool0oooo03<0oooo00<000000?ooo`3oool0U@3oool3
000000/0oooo000=0?ooo`040000003oool0oooo0?ooo`<000001P3oool00`000000oooo0?ooo`2I
0?ooo`800000;`3oool2000009`0oooo00<000000?ooo`3oool02@3oool000h0oooo00@000000?oo
o`3oool00000203oool00`000000oooo0?ooo`2J0?ooo`030000003oool0oooo02h0oooo00<00000
0?ooo`3oool0VP3oool00`000000oooo0?ooo`090?ooo`002`3oool010000000oooo0?ooo`000002
0?ooo`030000003oool0oooo00H0oooo00<000000?ooo`3oool0V`3oool2000002/0oooo0P00002N
0?ooo`030000003oool0oooo00T0oooo000<0?ooo`8000000`3oool4000000D0oooo0`00002L0?oo
o`<000009`3oool2000000030?ooo`000000oooo09`0oooo0P00000;0?ooo`006P3oool00`000000
oooo0?ooo`2N0?ooo`030000003oool0oooo02H0oooo00<000000?ooo`3oool0WP3oool00`000000
oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`2O0?ooo`800000903oool200000:40oooo
00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0X03oool200000240oooo0P00
00000`3oool000000?ooo`2P0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0
oooo0:40oooo0P00000O0?ooo`<00000X`3oool00`000000oooo0?ooo`090?ooo`006P3oool30000
0:80oooo0`00000K0?ooo`@00000X`3oool2000000/0oooo000J0?ooo`030000003oool0oooo0:<0
oooo0`00000J0?ooo`<00000Y@3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo
0?ooo`2U0?ooo`800000603oool300000:H0oooo00<000000?ooo`3oool02@3oool001X0oooo00<0
00000?ooo`3oool0YP3oool00`000000oooo0?ooo`0G0?ooo`030000003oool0oooo0:D0oooo00<0
00000?ooo`3oool02@3oool001X0oooo0`00003o0?ooofP0oooo0P00000;0?ooo`006P3oool00`00
0000oooo0?ooo`3o0?ooofT0oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3o
ool0o`3ooomY0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo0?l0oooo
J@3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofT0oooo00<0
00000?ooo`3oool02@3oool001X0oooo0`00003o0?ooofP0oooo0P00000;0?ooo`006P3oool00`00
0000oooo0?ooo`3o0?ooofT0oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3o
ool0o`3ooomY0?ooo`030000003oool0oooo00T0oooo000<0?ooo`800000103oool2000000H0oooo
00<000000?ooo`3oool0o`3ooomY0?ooo`030000003oool0oooo00T0oooo000;0?ooo`040000003o
ool0oooo00000080oooo00@000000?ooo`3oool000001@3oool400000;40oooo0P00002b0?ooo`<0
00002`3oool000h0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo0;<0oooo00<000000?ooo`3oool0/P3oool00`000000oooo0?ooo`090?ooo`00
303oool2000000<0oooo00@000000?ooo`3oool000001@3oool00`000000oooo0?ooo`3o0?ooofT0
oooo00<000000?ooo`3oool02@3oool000h0oooo00@000000?ooo`3oool000000P3oool00`000000
oooo0?ooo`030?ooo`030000003oool0oooo0?l0ooooJ@3oool00`000000oooo0?ooo`090?ooo`00
2`3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000000D0oooo00<000000?oo
o`3oool0o`3ooomY0?ooo`030000003oool0oooo00T0oooo000<0?ooo`800000103oool2000000H0
oooo0`00003o0?ooofP0oooo0P00000;0?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofT0oooo
00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0o`3ooomY0?ooo`030000003o
ool0oooo00T0oooo000J0?ooo`030000003oool0oooo0?l0ooooJ@3oool00`000000oooo0?ooo`09
0?ooo`006P3oool300000?l0ooooJ03oool2000000/0oooo000J0?ooo`030000003oool0oooo0?l0
ooooJ@3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofT0oooo
00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0o`3ooomY0?ooo`030000003o
ool0oooo00T0oooo000J0?ooo`030000003oool0oooo0?l0ooooJ@3oool00`000000oooo0?ooo`09
0?ooo`006P3oool300000?l0ooooJ03oool2000000/0oooo000J0?ooo`030000003oool0oooo0?l0
ooooJ@3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`3o0?ooofT0oooo
00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0o`3ooomY0?ooo`030000003o
ool0oooo00T0oooo000J0?ooo`<00000o`3ooomX0?ooo`8000002`3oool001X0oooo00<000000?oo
o`3oool0o`3ooomY0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo0?l0
ooooJ@3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`0G0?ooo`030000
003oool0oooo0340oooo00<000000?ooo`3oool0<@3oool00`000000oooo0?ooo`0a0?ooo`030000
003oool0oooo0340oooo00<000000?ooo`3oool0<03oool00`000000oooo0?ooo`0a0?ooo`030000
003oool0oooo01L0oooo00<000000?ooo`3oool02@3oool000`0oooo0P0000030?ooo`<000001P3o
ool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool01`3o
ool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3o
ool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3o
ool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3o
ool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0203o
ool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0203o
ool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool0203o
ool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool01`3o
ool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool01`3o
ool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool01`3o
ool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool01`3o
ool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3oool00P3o
ool00`000000oooo0?ooo`090?ooo`002`3oool010000000oooo0?ooo`0000050?ooo`030000003o
ool0oooo00<0ooooo`00001]000000/0oooo000>0?ooo`030000003oool0oooo00<0oooo00<00000
0?ooo`3oool0o`3ooomk0?ooo`00303oool2000000<0oooo0`00003o0?ooogh0oooo000>0?ooo`04
0000003oool0oooo00000?l0ooooP03oool000/0oooo00@000000?ooo`3oool000000P3oool00`00
0000oooo0?ooo`3o0?ooogh0oooo0000\
\>"],
  ImageRangeCache->{{{0, 400}, {199.5, 0}} -> {-10.1584, -7.9841, 0.192878, 
  0.220432}}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(X = AnglePairsAndRationalApproxs[30, 10^\(-3\)]; \nLength[X]\), 
    \(Angles = \(Transpose[X]\)[\([1]\)]; \n
    ListPlot[Angles, \ AspectRatio -> 1/2, Frame -> True, \ Axes -> False, \ 
      PlotRange -> {{\(-5\), 65}, {\(-5\), 35}}]; 
    WriteData["\<~/ASP/WebServices/GrainBreaker/2DGeometries/Geo30.dat\>", 
      X]; \)}], "Input"],

Cell[BoxData[
    \(539\)], "Output"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .5 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0714286 0.0142857 0.0625 0.0125 [
[.07143 -0.0125 -3 -9 ]
[.07143 -0.0125 3 0 ]
[.21429 -0.0125 -6 -9 ]
[.21429 -0.0125 6 0 ]
[.35714 -0.0125 -6 -9 ]
[.35714 -0.0125 6 0 ]
[.5 -0.0125 -6 -9 ]
[.5 -0.0125 6 0 ]
[.64286 -0.0125 -6 -9 ]
[.64286 -0.0125 6 0 ]
[.78571 -0.0125 -6 -9 ]
[.78571 -0.0125 6 0 ]
[.92857 -0.0125 -6 -9 ]
[.92857 -0.0125 6 0 ]
[ 0 0 -0.125 0 ]
[-0.0125 .0625 -6 -4.5 ]
[-0.0125 .0625 0 4.5 ]
[-0.0125 .125 -6 -4.5 ]
[-0.0125 .125 0 4.5 ]
[-0.0125 .1875 -12 -4.5 ]
[-0.0125 .1875 0 4.5 ]
[-0.0125 .25 -12 -4.5 ]
[-0.0125 .25 0 4.5 ]
[-0.0125 .3125 -12 -4.5 ]
[-0.0125 .3125 0 4.5 ]
[-0.0125 .375 -12 -4.5 ]
[-0.0125 .375 0 4.5 ]
[-0.0125 .4375 -12 -4.5 ]
[-0.0125 .4375 0 4.5 ]
[-0.0125 .5 -12 -4.5 ]
[-0.0125 .5 0 4.5 ]
[ 0 0 -0.125 0 ]
[ 0 .5 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .5 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.07143 0 m
.07143 .00625 L
s
[(0)] .07143 -0.0125 0 1 Mshowa
.21429 0 m
.21429 .00625 L
s
[(10)] .21429 -0.0125 0 1 Mshowa
.35714 0 m
.35714 .00625 L
s
[(20)] .35714 -0.0125 0 1 Mshowa
.5 0 m
.5 .00625 L
s
[(30)] .5 -0.0125 0 1 Mshowa
.64286 0 m
.64286 .00625 L
s
[(40)] .64286 -0.0125 0 1 Mshowa
.78571 0 m
.78571 .00625 L
s
[(50)] .78571 -0.0125 0 1 Mshowa
.92857 0 m
.92857 .00625 L
s
[(60)] .92857 -0.0125 0 1 Mshowa
.125 Mabswid
.1 0 m
.1 .00375 L
s
.12857 0 m
.12857 .00375 L
s
.15714 0 m
.15714 .00375 L
s
.18571 0 m
.18571 .00375 L
s
.24286 0 m
.24286 .00375 L
s
.27143 0 m
.27143 .00375 L
s
.3 0 m
.3 .00375 L
s
.32857 0 m
.32857 .00375 L
s
.38571 0 m
.38571 .00375 L
s
.41429 0 m
.41429 .00375 L
s
.44286 0 m
.44286 .00375 L
s
.47143 0 m
.47143 .00375 L
s
.52857 0 m
.52857 .00375 L
s
.55714 0 m
.55714 .00375 L
s
.58571 0 m
.58571 .00375 L
s
.61429 0 m
.61429 .00375 L
s
.67143 0 m
.67143 .00375 L
s
.7 0 m
.7 .00375 L
s
.72857 0 m
.72857 .00375 L
s
.75714 0 m
.75714 .00375 L
s
.81429 0 m
.81429 .00375 L
s
.84286 0 m
.84286 .00375 L
s
.87143 0 m
.87143 .00375 L
s
.9 0 m
.9 .00375 L
s
.04286 0 m
.04286 .00375 L
s
.01429 0 m
.01429 .00375 L
s
.95714 0 m
.95714 .00375 L
s
.98571 0 m
.98571 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
0 .0625 m
.00625 .0625 L
s
[(0)] -0.0125 .0625 1 0 Mshowa
0 .125 m
.00625 .125 L
s
[(5)] -0.0125 .125 1 0 Mshowa
0 .1875 m
.00625 .1875 L
s
[(10)] -0.0125 .1875 1 0 Mshowa
0 .25 m
.00625 .25 L
s
[(15)] -0.0125 .25 1 0 Mshowa
0 .3125 m
.00625 .3125 L
s
[(20)] -0.0125 .3125 1 0 Mshowa
0 .375 m
.00625 .375 L
s
[(25)] -0.0125 .375 1 0 Mshowa
0 .4375 m
.00625 .4375 L
s
[(30)] -0.0125 .4375 1 0 Mshowa
0 .5 m
.00625 .5 L
s
[(35)] -0.0125 .5 1 0 Mshowa
.125 Mabswid
0 .0125 m
.00375 .0125 L
s
0 .025 m
.00375 .025 L
s
0 .0375 m
.00375 .0375 L
s
0 .05 m
.00375 .05 L
s
0 .075 m
.00375 .075 L
s
0 .0875 m
.00375 .0875 L
s
0 .1 m
.00375 .1 L
s
0 .1125 m
.00375 .1125 L
s
0 .1375 m
.00375 .1375 L
s
0 .15 m
.00375 .15 L
s
0 .1625 m
.00375 .1625 L
s
0 .175 m
.00375 .175 L
s
0 .2 m
.00375 .2 L
s
0 .2125 m
.00375 .2125 L
s
0 .225 m
.00375 .225 L
s
0 .2375 m
.00375 .2375 L
s
0 .2625 m
.00375 .2625 L
s
0 .275 m
.00375 .275 L
s
0 .2875 m
.00375 .2875 L
s
0 .3 m
.00375 .3 L
s
0 .325 m
.00375 .325 L
s
0 .3375 m
.00375 .3375 L
s
0 .35 m
.00375 .35 L
s
0 .3625 m
.00375 .3625 L
s
0 .3875 m
.00375 .3875 L
s
0 .4 m
.00375 .4 L
s
0 .4125 m
.00375 .4125 L
s
0 .425 m
.00375 .425 L
s
0 .45 m
.00375 .45 L
s
0 .4625 m
.00375 .4625 L
s
0 .475 m
.00375 .475 L
s
0 .4875 m
.00375 .4875 L
s
.25 Mabswid
0 0 m
0 .5 L
s
.07143 .49375 m
.07143 .5 L
s
.21429 .49375 m
.21429 .5 L
s
.35714 .49375 m
.35714 .5 L
s
.5 .49375 m
.5 .5 L
s
.64286 .49375 m
.64286 .5 L
s
.78571 .49375 m
.78571 .5 L
s
.92857 .49375 m
.92857 .5 L
s
.125 Mabswid
.1 .49625 m
.1 .5 L
s
.12857 .49625 m
.12857 .5 L
s
.15714 .49625 m
.15714 .5 L
s
.18571 .49625 m
.18571 .5 L
s
.24286 .49625 m
.24286 .5 L
s
.27143 .49625 m
.27143 .5 L
s
.3 .49625 m
.3 .5 L
s
.32857 .49625 m
.32857 .5 L
s
.38571 .49625 m
.38571 .5 L
s
.41429 .49625 m
.41429 .5 L
s
.44286 .49625 m
.44286 .5 L
s
.47143 .49625 m
.47143 .5 L
s
.52857 .49625 m
.52857 .5 L
s
.55714 .49625 m
.55714 .5 L
s
.58571 .49625 m
.58571 .5 L
s
.61429 .49625 m
.61429 .5 L
s
.67143 .49625 m
.67143 .5 L
s
.7 .49625 m
.7 .5 L
s
.72857 .49625 m
.72857 .5 L
s
.75714 .49625 m
.75714 .5 L
s
.81429 .49625 m
.81429 .5 L
s
.84286 .49625 m
.84286 .5 L
s
.87143 .49625 m
.87143 .5 L
s
.9 .49625 m
.9 .5 L
s
.04286 .49625 m
.04286 .5 L
s
.01429 .49625 m
.01429 .5 L
s
.95714 .49625 m
.95714 .5 L
s
.98571 .49625 m
.98571 .5 L
s
.25 Mabswid
0 .5 m
1 .5 L
s
.99375 0 m
1 0 L
s
.99375 .0625 m
1 .0625 L
s
.99375 .125 m
1 .125 L
s
.99375 .1875 m
1 .1875 L
s
.99375 .25 m
1 .25 L
s
.99375 .3125 m
1 .3125 L
s
.99375 .375 m
1 .375 L
s
.99375 .4375 m
1 .4375 L
s
.125 Mabswid
.99625 .0125 m
1 .0125 L
s
.99625 .025 m
1 .025 L
s
.99625 .0375 m
1 .0375 L
s
.99625 .05 m
1 .05 L
s
.99625 .075 m
1 .075 L
s
.99625 .0875 m
1 .0875 L
s
.99625 .1 m
1 .1 L
s
.99625 .1125 m
1 .1125 L
s
.99625 .1375 m
1 .1375 L
s
.99625 .15 m
1 .15 L
s
.99625 .1625 m
1 .1625 L
s
.99625 .175 m
1 .175 L
s
.99625 .2 m
1 .2 L
s
.99625 .2125 m
1 .2125 L
s
.99625 .225 m
1 .225 L
s
.99625 .2375 m
1 .2375 L
s
.99625 .2625 m
1 .2625 L
s
.99625 .275 m
1 .275 L
s
.99625 .2875 m
1 .2875 L
s
.99625 .3 m
1 .3 L
s
.99625 .325 m
1 .325 L
s
.99625 .3375 m
1 .3375 L
s
.99625 .35 m
1 .35 L
s
.99625 .3625 m
1 .3625 L
s
.99625 .3875 m
1 .3875 L
s
.99625 .4 m
1 .4 L
s
.99625 .4125 m
1 .4125 L
s
.99625 .425 m
1 .425 L
s
.99625 .45 m
1 .45 L
s
.99625 .4625 m
1 .4625 L
s
.99625 .475 m
1 .475 L
s
.99625 .4875 m
1 .4875 L
s
.25 Mabswid
1 0 m
1 .5 L
s
0 0 m
1 0 L
1 .5 L
0 .5 L
closepath
clip
newpath
.008 w
.07143 .0625 Mdot
.08575 .07503 Mdot
.08667 .07584 Mdot
.08772 .07676 Mdot
.08893 .07781 Mdot
.09033 .07904 Mdot
.09197 .08047 Mdot
.09393 .08219 Mdot
.09629 .08426 Mdot
.09922 .0625 Mdot
.09922 .08681 Mdot
.10095 .0625 Mdot
.10095 .08833 Mdot
.10292 .0625 Mdot
.10292 .09005 Mdot
.10516 .0625 Mdot
.10516 .09202 Mdot
.10776 .0625 Mdot
.10776 .09429 Mdot
.11078 .0625 Mdot
.11078 .09693 Mdot
.11435 .0625 Mdot
.11435 .07503 Mdot
.11435 .10006 Mdot
.11711 .07584 Mdot
.11711 .10247 Mdot
.11863 .0625 Mdot
.11863 .1038 Mdot
.12026 .07676 Mdot
.12026 .10522 Mdot
.12386 .07781 Mdot
.12386 .10838 Mdot
.12694 .08681 Mdot
.12694 .11107 Mdot
.12805 .11204 Mdot
.1304 .08833 Mdot
.1304 .1141 Mdot
.13295 .11633 Mdot
.13431 .11752 Mdot
.13879 .12144 Mdot
.14285 .12499 Mdot
.14394 .12595 Mdot
.14584 .12761 Mdot
.14743 .129 Mdot
.14995 .1312 Mdot
.15264 .13356 Mdot
.15454 .13522 Mdot
.15704 .13741 Mdot
.15861 .13878 Mdot
.15969 .13973 Mdot
.16553 .14483 Mdot
.17117 .14978 Mdot
.17218 .15066 Mdot
.17363 .15192 Mdot
.17587 .15389 Mdot
.17754 .15535 Mdot
.17984 .15736 Mdot
.18194 .1592 Mdot
.18324 .16034 Mdot
.18477 .16167 Mdot
.18876 .16517 Mdot
.19304 .16891 Mdot
.19484 .17049 Mdot
.19646 .17191 Mdot
.19926 .17435 Mdot
.20261 .17729 Mdot
.20524 .17959 Mdot
.20736 .18144 Mdot
.20911 .18297 Mdot
.21181 .18533 Mdot
.2138 .18708 Mdot
.21533 .18842 Mdot
.21655 .18948 Mdot
.21753 .19034 Mdot
.22705 .19867 Mdot
.23683 .20722 Mdot
.23787 .20814 Mdot
.23916 .20927 Mdot
.24081 .21071 Mdot
.24298 .2126 Mdot
.24595 .2152 Mdot
.24789 .21691 Mdot
.25028 .219 Mdot
.25329 .22163 Mdot
.25448 .22267 Mdot
.2572 .22505 Mdot
.26052 .22795 Mdot
.26247 .22966 Mdot
.26466 .23158 Mdot
.26586 .23263 Mdot
.26997 .23622 Mdot
.27394 .07503 Mdot
.27394 .10006 Mdot
.27394 .2397 Mdot
.27505 .24067 Mdot
.27702 .2424 Mdot
.27872 .0625 Mdot
.27872 .24388 Mdot
.2815 .11107 Mdot
.2815 .13522 Mdot
.2815 .24631 Mdot
.28458 .24901 Mdot
.28684 .10247 Mdot
.28684 .25099 Mdot
.28993 .25369 Mdot
.29194 .25545 Mdot
.29336 .0625 Mdot
.29336 .09202 Mdot
.29336 .25669 Mdot
.29441 .13973 Mdot
.29441 .2576 Mdot
.30146 .07904 Mdot
.30146 .13356 Mdot
.30146 .14978 Mdot
.30146 .26378 Mdot
.30805 .18297 Mdot
.30805 .26955 Mdot
.30896 .27034 Mdot
.31016 .12595 Mdot
.31016 .27139 Mdot
.31181 .27283 Mdot
.31423 .27495 Mdot
.31592 .18144 Mdot
.31592 .27643 Mdot
.31814 .27837 Mdot
.32115 .11633 Mdot
.32115 .281 Mdot
.32241 .19034 Mdot
.32241 .2821 Mdot
.32548 .17959 Mdot
.32548 .2848 Mdot
.32845 .19867 Mdot
.32845 .2874 Mdot
.32961 .28841 Mdot
.33226 .18948 Mdot
.33226 .29073 Mdot
.33546 .0625 Mdot
.33546 .1038 Mdot
.33546 .29353 Mdot
.33733 .17729 Mdot
.33733 .29516 Mdot
.33855 .29623 Mdot
.34438 .16034 Mdot
.34438 .18842 Mdot
.34438 .30133 Mdot
.34981 .30609 Mdot
.35085 .307 Mdot
.35239 .17435 Mdot
.35239 .30834 Mdot
.35488 .0625 Mdot
.35488 .08681 Mdot
.35488 .11107 Mdot
.35488 .31052 Mdot
.35682 .31221 Mdot
.35962 .31467 Mdot
.36232 .31703 Mdot
.36407 .15389 Mdot
.36407 .31856 Mdot
.36619 .32041 Mdot
.36743 .07584 Mdot
.36743 .10247 Mdot
.36743 .3215 Mdot
.37217 .1141 Mdot
.37217 .13973 Mdot
.37217 .2576 Mdot
.37217 .32565 Mdot
.37658 .14978 Mdot
.37658 .26378 Mdot
.37658 .32951 Mdot
.37764 .24901 Mdot
.37764 .33044 Mdot
.37936 .33194 Mdot
.38071 .18297 Mdot
.38071 .26955 Mdot
.38071 .33312 Mdot
.38267 .0625 Mdot
.38267 .09202 Mdot
.38267 .25669 Mdot
.38267 .33483 Mdot
.3858 .33757 Mdot
.3873 .13356 Mdot
.3873 .26378 Mdot
.3873 .33889 Mdot
.39158 .24631 Mdot
.39158 .27034 Mdot
.39158 .34264 Mdot
.39556 .18144 Mdot
.39556 .27643 Mdot
.39556 .34611 Mdot
.39682 .25545 Mdot
.39682 .34721 Mdot
.39925 .19034 Mdot
.39925 .2821 Mdot
.39925 .34934 Mdot
.40157 .07904 Mdot
.40157 .26378 Mdot
.40157 .35137 Mdot
.40269 .35235 Mdot
.4059 .12595 Mdot
.4059 .27139 Mdot
.4059 .35517 Mdot
.40987 .16891 Mdot
.40987 .35864 Mdot
.41174 .2424 Mdot
.41174 .36027 Mdot
.41353 .36183 Mdot
.41689 .25369 Mdot
.41689 .36478 Mdot
.42001 .22505 Mdot
.42001 .36751 Mdot
.42148 .3688 Mdot
.4229 .23263 Mdot
.4229 .37004 Mdot
.42559 .0625 Mdot
.42559 .07503 Mdot
.42559 .10006 Mdot
.42559 .2397 Mdot
.42559 .37239 Mdot
.4281 .13522 Mdot
.4281 .37459 Mdot
.42929 .15192 Mdot
.42929 .37563 Mdot
.43264 .37856 Mdot
.43569 .38123 Mdot
.43848 .219 Mdot
.43848 .38367 Mdot
.44103 .22795 Mdot
.44103 .3859 Mdot
.44338 .38796 Mdot
.44555 .11752 Mdot
.44555 .38986 Mdot
.44756 .15535 Mdot
.44756 .39162 Mdot
.44943 .16517 Mdot
.44943 .39325 Mdot
.45117 .19867 Mdot
.45117 .39478 Mdot
.45432 .39753 Mdot
.45708 .12761 Mdot
.45708 .39994 Mdot
.45953 .40209 Mdot
.46171 .19867 Mdot
.46171 .404 Mdot
.46367 .40571 Mdot
.46544 .40726 Mdot
.46705 .40867 Mdot
.46851 .0625 Mdot
.46851 .09429 Mdot
.46851 .40995 Mdot
.46985 .13878 Mdot
.46985 .41112 Mdot
.47108 .2848 Mdot
.47108 .41219 Mdot
.47221 .18948 Mdot
.47221 .29073 Mdot
.47221 .41319 Mdot
.47326 .29623 Mdot
.47326 .4141 Mdot
.47423 .30133 Mdot
.47423 .41495 Mdot
.47514 .30609 Mdot
.47514 .41574 Mdot
.47598 .08681 Mdot
.47598 .11107 Mdot
.47598 .24631 Mdot
.47598 .31052 Mdot
.47598 .41648 Mdot
.5 .09005 Mdot
.5 .11633 Mdot
.5 .16517 Mdot
.5 .27283 Mdot
.5 .281 Mdot
.5 .4375 Mdot
.52402 .08681 Mdot
.52402 .11107 Mdot
.52402 .24631 Mdot
.52402 .31052 Mdot
.52402 .41648 Mdot
.52486 .30609 Mdot
.52486 .41574 Mdot
.52577 .30133 Mdot
.52577 .41495 Mdot
.52674 .29623 Mdot
.52674 .4141 Mdot
.52779 .18948 Mdot
.52779 .29073 Mdot
.52779 .41319 Mdot
.52892 .2848 Mdot
.52892 .41219 Mdot
.53015 .13878 Mdot
.53015 .41112 Mdot
.53149 .0625 Mdot
.53149 .09429 Mdot
.53149 .40995 Mdot
.53295 .40867 Mdot
.53456 .40726 Mdot
.53633 .40571 Mdot
.53829 .19867 Mdot
.53829 .404 Mdot
.54047 .40209 Mdot
.54292 .12761 Mdot
.54292 .39994 Mdot
.54568 .39753 Mdot
.54883 .19867 Mdot
.54883 .39478 Mdot
.55057 .16517 Mdot
.55057 .39325 Mdot
.55244 .15535 Mdot
.55244 .39162 Mdot
.55445 .11752 Mdot
.55445 .38986 Mdot
.55662 .38796 Mdot
.55897 .22795 Mdot
.55897 .3859 Mdot
.56152 .219 Mdot
.56152 .38367 Mdot
.56431 .38123 Mdot
.56736 .37856 Mdot
.57071 .15192 Mdot
.57071 .37563 Mdot
.5719 .13522 Mdot
.5719 .37459 Mdot
.57441 .0625 Mdot
.57441 .07503 Mdot
.57441 .10006 Mdot
.57441 .2397 Mdot
.57441 .37239 Mdot
.5771 .23263 Mdot
.5771 .37004 Mdot
.57852 .3688 Mdot
.57999 .22505 Mdot
.57999 .36751 Mdot
.58311 .25369 Mdot
.58311 .36478 Mdot
.58647 .36183 Mdot
.58826 .2424 Mdot
.58826 .36027 Mdot
.59013 .16891 Mdot
.59013 .35864 Mdot
.5941 .12595 Mdot
.5941 .27139 Mdot
.5941 .35517 Mdot
.59731 .35235 Mdot
.59843 .07904 Mdot
.59843 .26378 Mdot
.59843 .35137 Mdot
.60075 .19034 Mdot
.60075 .2821 Mdot
.60075 .34934 Mdot
.60318 .25545 Mdot
.60318 .34721 Mdot
.60444 .18144 Mdot
.60444 .27643 Mdot
.60444 .34611 Mdot
.60842 .24631 Mdot
.60842 .27034 Mdot
.60842 .34264 Mdot
.6127 .13356 Mdot
.6127 .26378 Mdot
.6127 .33889 Mdot
.6142 .33757 Mdot
.61733 .0625 Mdot
.61733 .09202 Mdot
.61733 .25669 Mdot
.61733 .33483 Mdot
.61929 .18297 Mdot
.61929 .26955 Mdot
.61929 .33312 Mdot
.62064 .33194 Mdot
.62236 .24901 Mdot
.62236 .33044 Mdot
.62342 .14978 Mdot
.62342 .26378 Mdot
.62342 .32951 Mdot
.62783 .1141 Mdot
.62783 .13973 Mdot
.62783 .2576 Mdot
.62783 .32565 Mdot
.63257 .07584 Mdot
.63257 .10247 Mdot
.63257 .3215 Mdot
.63381 .32041 Mdot
.63593 .15389 Mdot
.63593 .31856 Mdot
.63768 .31703 Mdot
.64038 .31467 Mdot
.64318 .31221 Mdot
.64512 .0625 Mdot
.64512 .08681 Mdot
.64512 .11107 Mdot
.64512 .31052 Mdot
.64761 .17435 Mdot
.64761 .30834 Mdot
.64915 .307 Mdot
.65019 .30609 Mdot
.65562 .16034 Mdot
.65562 .18842 Mdot
.65562 .30133 Mdot
.66145 .29623 Mdot
.66267 .17729 Mdot
.66267 .29516 Mdot
.66454 .0625 Mdot
.66454 .1038 Mdot
.66454 .29353 Mdot
.66774 .18948 Mdot
.66774 .29073 Mdot
.67039 .28841 Mdot
.67155 .19867 Mdot
.67155 .2874 Mdot
.67452 .17959 Mdot
.67452 .2848 Mdot
.67759 .19034 Mdot
.67759 .2821 Mdot
.67885 .11633 Mdot
.67885 .281 Mdot
.68186 .27837 Mdot
.68408 .18144 Mdot
.68408 .27643 Mdot
.68577 .27495 Mdot
.68819 .27283 Mdot
.68984 .12595 Mdot
.68984 .27139 Mdot
.69104 .27034 Mdot
.69195 .18297 Mdot
.69195 .26955 Mdot
.69854 .07904 Mdot
.69854 .13356 Mdot
.69854 .14978 Mdot
.69854 .26378 Mdot
.70559 .13973 Mdot
.70559 .2576 Mdot
.70664 .0625 Mdot
.70664 .09202 Mdot
.70664 .25669 Mdot
.70806 .25545 Mdot
.71007 .25369 Mdot
.71316 .10247 Mdot
.71316 .25099 Mdot
.71542 .24901 Mdot
.7185 .11107 Mdot
.7185 .13522 Mdot
.7185 .24631 Mdot
.72128 .0625 Mdot
.72128 .24388 Mdot
.72298 .2424 Mdot
.72495 .24067 Mdot
.72606 .07503 Mdot
.72606 .10006 Mdot
.72606 .2397 Mdot
.73003 .23622 Mdot
.73414 .23263 Mdot
.73534 .23158 Mdot
.73753 .22966 Mdot
.73948 .22795 Mdot
.7428 .22505 Mdot
.74552 .22267 Mdot
.74671 .22163 Mdot
.74972 .219 Mdot
.75211 .21691 Mdot
.75405 .2152 Mdot
.75702 .2126 Mdot
.75919 .21071 Mdot
.76084 .20927 Mdot
.76213 .20814 Mdot
.76317 .20722 Mdot
.77295 .19867 Mdot
.78247 .19034 Mdot
.78345 .18948 Mdot
.78467 .18842 Mdot
.7862 .18708 Mdot
.78819 .18533 Mdot
.79089 .18297 Mdot
.79264 .18144 Mdot
.79476 .17959 Mdot
.79739 .17729 Mdot
.80074 .17435 Mdot
.80354 .17191 Mdot
.80516 .17049 Mdot
.80696 .16891 Mdot
.81124 .16517 Mdot
.81523 .16167 Mdot
.81676 .16034 Mdot
.81806 .1592 Mdot
.82016 .15736 Mdot
.82246 .15535 Mdot
.82413 .15389 Mdot
.82637 .15192 Mdot
.82782 .15066 Mdot
.82883 .14978 Mdot
.83447 .14483 Mdot
.84031 .13973 Mdot
.84139 .13878 Mdot
.84296 .13741 Mdot
.84546 .13522 Mdot
.84736 .13356 Mdot
.85005 .1312 Mdot
.85257 .129 Mdot
.85416 .12761 Mdot
.85606 .12595 Mdot
.85715 .12499 Mdot
.86121 .12144 Mdot
.86569 .11752 Mdot
.86705 .11633 Mdot
.8696 .08833 Mdot
.8696 .1141 Mdot
.87195 .11204 Mdot
.87306 .08681 Mdot
.87306 .11107 Mdot
.87614 .07781 Mdot
.87614 .10838 Mdot
.87974 .07676 Mdot
.87974 .10522 Mdot
.88137 .0625 Mdot
.88137 .1038 Mdot
.88289 .07584 Mdot
.88289 .10247 Mdot
.88565 .0625 Mdot
.88565 .07503 Mdot
.88565 .10006 Mdot
.88922 .0625 Mdot
.88922 .09693 Mdot
.89224 .0625 Mdot
.89224 .09429 Mdot
.89484 .0625 Mdot
.89484 .09202 Mdot
.89708 .0625 Mdot
.89708 .09005 Mdot
.89905 .0625 Mdot
.89905 .08833 Mdot
.90078 .0625 Mdot
.90078 .08681 Mdot
.90371 .08426 Mdot
.90607 .08219 Mdot
.90803 .08047 Mdot
.90967 .07904 Mdot
.91107 .07781 Mdot
.91228 .07676 Mdot
.91333 .07584 Mdot
.91425 .07503 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{409, 204.5},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006I0000c2000`400?l00000o`00003oo`3ooonJ
0?ooo`00o`3ooonJ0?ooo`00o`3ooonJ0?ooo`00o`3ooonJ0?ooo`00=03oool2000002l0oooo1000
00030?ooo`800000;03oool4000000<0oooo0P00000]0?ooo`800000103oool2000002h0oooo00<0
00000?ooo`3oool00P3oool2000002`0oooo0`0000040?ooo`800000;@3oool2000000@0oooo0P00
000R0?ooo`00<`3oool010000000oooo0?ooo`00000`0?ooo`050000003oool0oooo0?ooo`000000
0P3oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool0
0000:`3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000002d0oooo00D00000
0?ooo`3oool0oooo000000020?ooo`030000003oool0oooo02`0oooo00@000000?ooo`3oool00000
0P3oool00`000000oooo0?ooo`0Y0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3o
ool000008@3oool003<0oooo00@000000?ooo`3oool00000<03oool01@000000oooo0?ooo`3oool0
00000080oooo00<000000?ooo`3oool0:P3oool00`000000oooo0?ooo`020?ooo`040000003oool0
oooo000002h0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`0X0?ooo`D00000
0P3oool010000000oooo0?ooo`00000^0?ooo`040000003oool0oooo00000080oooo00<000000?oo
o`3oool0:@3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo00000240oooo000c
0?ooo`040000003oool0oooo00000300oooo00D000000?ooo`3oool0oooo000000020?ooo`030000
003oool0oooo02/0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003oool0oooo02X0
oooo0P0000030?ooo`040000003oool0oooo000002X0oooo00@000000?ooo`3oool000000`3oool0
10000000oooo0?ooo`00000[0?ooo`<000000`3oool010000000oooo0?ooo`00000[0?ooo`<00000
0`3oool010000000oooo0?ooo`00000Q0?ooo`00<`3oool010000000oooo0?ooo`00000`0?ooo`05
0000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`0/0?ooo`040000003oool0oooo
00000080oooo00<000000?ooo`3oool0;03oool010000000oooo0?ooo`0000020?ooo`030000003o
ool0oooo02T0oooo00<000000?ooo`0000000`3oool010000000oooo0?ooo`00000[0?ooo`030000
003oool0oooo00<0oooo00@000000?ooo`3oool00000:`3oool00`000000oooo0?ooo`030?ooo`04
0000003oool0oooo00000240oooo000c0?ooo`040000003oool0oooo000002h0oooo0`0000030?oo
o`040000003oool0oooo000002/0oooo00@000000?ooo`3oool000000P3oool010000000oooo0?oo
o`00000[0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool00000;03oool20000
00<0oooo00@000000?ooo`3oool00000:`3oool00`000000oooo0?ooo`030?ooo`040000003oool0
oooo000002/0oooo00<000000?ooo`3oool00`3oool010000000oooo0?ooo`00000Q0?ooo`00=03o
ool200000340oooo00<000000?ooo`3oool00P3oool2000002d0oooo0P0000040?ooo`800000;@3o
ool2000000@0oooo0P00000^0?ooo`030000003oool0oooo0080oooo0P00000/0?ooo`@000000`3o
ool2000002d0oooo0`0000030?ooo`8000008P3oool00?l0ooooVP3oool00?l0ooooVP3oool00?l0
ooooVP3oool001X0ooooo`00001e000000/0oooo000J0?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`020?ooo`030000003oool0oooo00T0oooo000J
0?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`0b
0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0<P3oool00`000000oooo0?ooo`0b
0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool05`3oool00`000000oooo0?ooo`09
0?ooo`006P3oool00`000000oooo0?ooo`3o0?ooog40oooo00<000000?ooo`3oool02@3oool001X0
oooo0`00003o0?ooog00oooo0P00000;0?ooo`006P3oool00`000000oooo0?ooo`3o0?ooog40oooo
00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0o`3oooma0?ooo`030000003o
ool0oooo00T0oooo000J0?ooo`030000003oool0oooo0?l0ooooL@3oool00`000000oooo0?ooo`09
0?ooo`006P3oool00`000000oooo0?ooo`3o0?ooog40oooo00<000000?ooo`3oool02@3oool001X0
oooo0`00003o0?ooog00oooo0P00000;0?ooo`006P3oool00`000000oooo0?ooo`3o0?ooog40oooo
00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0o`3oooma0?ooo`030000003o
ool0oooo00T0oooo000J0?ooo`030000003oool0oooo0?l0ooooL@3oool00`000000oooo0?ooo`09
0?ooo`006P3oool00`000000oooo0?ooo`3o0?ooog40oooo00<000000?ooo`3oool02@3oool001X0
oooo0`00003o0?ooog00oooo0P00000;0?ooo`006P3oool00`000000oooo0?ooo`3o0?ooog40oooo
00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0o`3oooma0?ooo`030000003o
ool0oooo00T0oooo000J0?ooo`030000003oool0oooo0?l0ooooL@3oool00`000000oooo0?ooo`09
0?ooo`006P3oool300000?l0ooooL03oool2000000/0oooo000J0?ooo`030000003oool0oooo0?l0
ooooL@3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`3o0?ooog40oooo
00<000000?ooo`3oool02@3oool00180oooo0P0000060?ooo`030000003oool0oooo0?l0ooooL@3o
ool00`000000oooo0?ooo`090?ooo`004@3oool010000000oooo0?ooo`0000050?ooo`030000003o
ool0oooo0?l0ooooL@3oool00`000000oooo0?ooo`090?ooo`004@3oool010000000oooo0?ooo`00
00050?ooo`@000005P3oool2000000P0oooo2@00000j0?ooo`8000000`3oool2000000h0oooo0P00
00050?ooo`800000203oool2000000h0oooo0P00000>0?ooo`8000005@3oool2000000h0oooo0P00
000>0?ooo`8000002@3oool2000000D0oooo0P00000=0?ooo`800000103oool2000003T0oooo2@00
000R0?ooo`<000002`3oool00140oooo00@000000?ooo`3oool000001@3oool00`000000oooo0?oo
o`0H0?ooo`030000003oool0oooo00L0oooo10000000103oool0000000000000000k0?ooo`030000
003oool0oooo0080oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`040?ooo`030000
003oool0oooo00L0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000
003oool0oooo01@0oooo00<000000?ooo`3oool03@3oool00`000000oooo0?ooo`0=0?ooo`030000
003oool0oooo00P0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`0<0?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool0>03oool00`000000oooo00000005000002@0oooo
00<000000?ooo`3oool02@3oool00140oooo00@000000?ooo`3oool000001@3oool00`000000oooo
0?ooo`3o0?ooog40oooo00<000000?ooo`3oool02@3oool00140oooo00@000000?ooo`3oool00000
1@3oool00`000000oooo0?ooo`3o0?ooog40oooo00<000000?ooo`3oool02@3oool00180oooo0P00
00060?ooo`<00000703oool2000000T0oooo0P00000i0?ooo`800000=P3oool2000003D0oooo0P00
000g0?ooo`800000>@3oool2000000P0oooo0P00000N0?ooo`8000002`3oool001X0oooo00<00000
0?ooo`3oool07@3oool2000000T0oooo1000000g0?ooo`030000003oool0oooo01l0oooo0P00000D
0?ooo`030000003oool0oooo03@0oooo00<000000?ooo`3oool04P3oool200000280oooo00<00000
0?ooo`3oool0<`3oool6000000L0oooo0`00000O0?ooo`030000003oool0oooo00T0oooo000J0?oo
o`030000003oool0oooo01h0oooo0`0000080?ooo`<00000@03oool2000001P0oooo00<000000?oo
o`3oool0203oool2000004L0oooo0P00000<0?ooo`030000003oool0oooo01@0oooo0P0000110?oo
o`040000003oool00000000000L0oooo1000000O0?ooo`030000003oool0oooo00T0oooo000J0?oo
o`030000003oool0oooo01l0oooo0P00001<0?ooo`030000003oool0oooo0280oooo00<000000?oo
o`3oool0AP3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo04T0oooo1000000P0?oo
o`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo0200oooo0P00003o0?ooobX0
oooo0`00000R0?ooo`030000003oool0oooo00T0oooo000J0?ooo`<000008@3oool3000000P0oooo
0`00001A0?ooo`800000:`3oool200000100oooo0P00000[0?ooo`800000D@3oool3000000P0oooo
0`00000R0?ooo`8000002`3oool001X0oooo00<000000?ooo`3oool08P3oool3000000P0oooo0P00
001B0?ooo`030000003oool0oooo02X0oooo00<000000?ooo`3oool01@3oool2000000P0oooo00<0
00000?ooo`3oool0:P3oool00`000000oooo0?ooo`1@0?ooo`8000001`3oool3000002@0oooo00<0
00000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool08`3oool2000004@0oooo0P00000O
0?ooo`800000:`3oool00`000000oooo0?ooo`0W0?ooo`8000007`3oool2000004@0oooo0`00000U
0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo02@0oooo0P0000140?oo
o`030000003oool0oooo01h0oooo00<000000?ooo`3oool0703oool2000001D0oooo0P00000O0?oo
o`030000003oool0oooo01h0oooo00<000000?ooo`3oool0@@3oool3000002H0oooo00<000000?oo
o`3oool02@3oool001X0oooo00<000000?ooo`3oool09@3oool3000008<0oooo00<000000?ooo`3o
ool0503oool00`000000oooo0?ooo`210?ooo`<000009`3oool00`000000oooo0?ooo`090?ooo`00
6P3oool3000002L0oooo0P00000i0?ooo`800000=P3oool2000003D0oooo0P00000g0?ooo`800000
>@3oool3000002L0oooo0P00000;0?ooo`006P3oool00`000000oooo0?ooo`0X0?ooo`800000>@3o
ool010000000oooo0?ooo`3oool200000100oooo0P00000:0?ooo`800000503oool00`000000oooo
0?ooo`0d0?ooo`030000003oool0oooo0180oooo0P00000:0?ooo`800000403oool2000000@0oooo
00<000000?ooo`3oool0=@3oool4000002T0oooo00<000000?ooo`3oool02@3oool001X0oooo00<0
00000?ooo`3oool0:@3oool2000003d0oooo00<000000?ooo`3oool03`3oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo05l0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?oo
o`0?0?ooo`030000003oool0oooo03X0oooo0`00000Z0?ooo`030000003oool0oooo00T0oooo000J
0?ooo`030000003oool0oooo02X0oooo0P00003o0?oooaH0oooo0P0000000`3oool000000?ooo`0Z
0?ooo`030000003oool0oooo00T0oooo000J0?ooo`<00000:`3oool3000003L0oooo0P00000I0?oo
o`800000:`3oool200000100oooo0P00000[0?ooo`8000006@3oool2000003L0oooo0`00000/0?oo
o`8000002`3oool001X0oooo00<000000?ooo`3oool0;@3oool2000003L0oooo00<000000?ooo`3o
ool0603oool00`000000oooo0?ooo`020?ooo`8000009P3oool00`000000oooo0?ooo`0?0?ooo`03
0000003oool0oooo0280oooo0P0000060?ooo`030000003oool0oooo01P0oooo00<000000?ooo`3o
ool0=03oool3000002h0oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0
;P3oool2000004@0oooo0P00000A0?ooo`030000003oool0oooo01P0oooo0P00000B0?ooo`800000
4P3oool2000001X0oooo00<000000?ooo`3oool03`3oool2000004<0oooo1000000_0?ooo`030000
003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool0@`3o
ool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo0140oooo00<000000?ooo`3oool04@3o
ool00`000000oooo0?ooo`0[0?ooo`030000003oool0oooo0480oooo0P00000`0?ooo`030000003o
ool0oooo00T0oooo000A0?ooo`<000001P3oool00`000000oooo0?ooo`0`0?ooo`800000o`3oool;
0?ooo`800000<P3oool00`000000oooo0?ooo`090?ooo`00503oool00`000000oooo0?ooo`030?oo
o`@00000<03oool3000003/0oooo0P00000R0?ooo`800000A03oool200000240oooo0P00000l0?oo
o`80000000<0oooo0000003oool0;`3oool3000000/0oooo000D0?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool0<P3oool3000003/0oooo00<000000?ooo`3oool08@3oool00`000000
oooo0?ooo`0?0?ooo`8000007P3oool200000180oooo00<000000?ooo`3oool0803oool00`000000
oooo0?ooo`0i0?ooo`<00000=03oool00`000000oooo0?ooo`090?ooo`004@3oool3000000H0oooo
00<000000?ooo`3oool0=03oool200000740oooo00<000000?ooo`3oool07@3oool00`000000oooo
0?ooo`1^0?ooo`<00000=@3oool00`000000oooo0?ooo`090?ooo`004@3oool00`000000oooo0?oo
o`060?ooo`030000003oool0oooo03D0oooo0P00000e0?ooo`8000007P3oool200000580oooo0P00
000M0?ooo`800000=P3oool3000003H0oooo00<000000?ooo`3oool02@3oool00140oooo00<00000
0?ooo`3oool01P3oool00`000000oooo0?ooo`0f0?ooo`800000;@3oool2000000H0oooo00<00000
0?ooo`3oool07@3oool00`000000oooo0?ooo`0;0?ooo`800000<`3oool2000000l0oooo00<00000
0?ooo`3oool0703oool00`000000oooo0?ooo`040?ooo`800000;@3oool3000003L0oooo00<00000
0?ooo`3oool02@3oool00140oooo100000050?ooo`<00000=`3oool3000002`0oooo00@000000?oo
o`3oool0oooo0P00000J0?ooo`800000503oool00`000000oooo0?ooo`0<0?ooo`800000503oool2
000000h0oooo00<000000?ooo`3oool04@3oool2000001/0oooo0P0000040?ooo`030000003oool0
oooo02T0oooo1000000g0?ooo`8000002`3oool001X0oooo00<000000?ooo`3oool0>03oool20000
0340oooo00<000000?ooo`3oool06@3oool00`000000oooo0?ooo`0R0?ooo`030000003oool0oooo
01<0oooo00<000000?ooo`3oool08@3oool00`000000oooo0?ooo`0J0?ooo`030000003oool0oooo
02l0oooo0P00000i0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo03X0
oooo0P00003f0?ooo`800000?03oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo
0?ooo`0k0?ooo`030000003oool0oooo0?D0oooo00<000000?ooo`3oool0>P3oool00`000000oooo
0?ooo`090?ooo`006P3oool3000003`0oooo0P00000^0?ooo`8000006P3oool2000005X0oooo0P00
000I0?ooo`800000;P3oool3000003d0oooo0P00000;0?ooo`006P3oool00`000000oooo0?ooo`0m
0?ooo`800000;P3oool00`000000oooo0?ooo`0I0?ooo`030000003oool0oooo0100oooo0P00000b
0?ooo`8000004`3oool00`000000oooo0?ooo`0H0?ooo`030000003oool0oooo02`0oooo0`00000n
0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo03h0oooo0P0000130?oo
o`800000603oool00`000000oooo0?ooo`020?ooo`8000009@3oool2000000H0oooo00<000000?oo
o`3oool0503oool2000004<0oooo1000000o0?ooo`030000003oool0oooo00T0oooo000J0?ooo`03
0000003oool0oooo03l0oooo0P0000130?ooo`030000003oool0oooo01`0oooo00<000000?ooo`3o
ool0903oool00`000000oooo0?ooo`0L0?ooo`030000003oool0oooo0440oooo0`0000100?ooo`03
0000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo0400oooo0P00000j0?ooo`800000
LP3oool2000003T0oooo100000110?ooo`030000003oool0oooo00T0oooo000J0?ooo`<00000@@3o
ool2000003X0oooo00<000000?ooo`3oool0L@3oool00`000000oooo0?ooo`0g0?ooo`<00000@@3o
ool2000000/0oooo000J0?ooo`030000003oool0oooo0480oooo0P00001O0?ooo`8000004@3oool2
00000140oooo0P00001N0?ooo`80000000<0oooo0000003oool0@P3oool00`000000oooo0?ooo`09
0?ooo`006P3oool00`000000oooo0?ooo`130?ooo`<00000CP3oool2000000h0oooo00<000000?oo
o`3oool0403oool00`000000oooo0?ooo`0@0?ooo`030000003oool0oooo00X0oooo0P00001?0?oo
o`<00000A@3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`150?ooo`80
0000CP3oool00`000000oooo0?ooo`100?ooo`030000003oool0oooo04/0oooo100000160?ooo`03
0000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo04H0oooo0P00000g0?ooo`800000
K03oool2000003H0oooo100000170?ooo`030000003oool0oooo00T0oooo000J0?ooo`<00000A`3o
ool300000300oooo0P0000040?ooo`030000003oool0oooo06/0oooo00@000000?ooo`3oool0oooo
0P00000`0?ooo`<00000B03oool2000000/0oooo000J0?ooo`030000003oool0oooo04T0oooo0P00
000Z0?ooo`800000103oool00`000000oooo0?ooo`1e0?ooo`040000003oool0oooo0?ooo`800000
:P3oool3000004X0oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0B@3o
ool3000002<0oooo0P0000000`3oool00000000000020?ooo`030000003oool0oooo0140oooo0P00
00030?ooo`800000C03oool2000000<0oooo0P00000D0?ooo`050000003oool000000000003oool0
0P00000S0?ooo`<00000B`3oool00`000000oooo0?ooo`090?ooo`002`3oool4000000<0oooo0P00
00060?ooo`030000003oool0oooo04X0oooo0`00000S0?ooo`040000003oool0oooo000001L0oooo
00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`1;0?ooo`030000003oool0oooo0080oooo
00<000000?ooo`3oool05@3oool010000000oooo0?ooo`00000R0?ooo`<00000C03oool00`000000
oooo0?ooo`090?ooo`003@3oool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`3o
ool00`3oool4000004/0oooo0P00000^0?ooo`800000LP3oool2000002d0oooo1000001;0?ooo`<0
00002`3oool000d0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool0C@3oool2000002D0oooo0P0000020?ooo`8000000`3oool00`000000
oooo0?ooo`0@0?ooo`8000006P3oool200000180oooo0P00000I0?ooo`800000503oool00`000000
oooo0?ooo`0200000080oooo0P00000U0?ooo`<00000CP3oool00`000000oooo0?ooo`090?ooo`00
3@3oool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`1>0?ooo`030000003oool0oooo02@0oooo00D000000?ooo`3oool0oooo0000000G0?oo
o`030000003oool0oooo01T0oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0H0?oo
o`030000003oool0oooo01H0oooo00D000000?ooo`3oool0oooo0000000V0?ooo`030000003oool0
oooo04d0oooo00<000000?ooo`3oool02@3oool000d0oooo00D000000?ooo`3oool0oooo00000002
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0o`3oooma0?ooo`030000003oool0
oooo00T0oooo000;0?ooo`<000000`3oool010000000oooo0?ooo`0000050?ooo`030000003oool0
oooo0540oooo0P00000S0?ooo`800000;03oool200000080oooo0P00000J0?ooo`8000000P3oool2
000002/0oooo0P00000T0?ooo`800000D`3oool00`000000oooo0?ooo`090?ooo`003@3oool00`00
0000oooo0?ooo`020?ooo`8000001P3oool300000580oooo00<000000?ooo`3oool08P3oool00`00
0000oooo0?ooo`0[0?ooo`050000003oool0oooo0?ooo`0000006`3oool01@000000oooo0?ooo`3o
ool0000002`0oooo00<000000?ooo`3oool08`3oool00`000000oooo0?ooo`1@0?ooo`8000002`3o
ool001X0oooo00<000000?ooo`3oool0o`3oooma0?ooo`030000003oool0oooo00T0oooo000J0?oo
o`030000003oool0oooo05@0oooo0P0000310?ooo`800000E`3oool00`000000oooo0?ooo`090?oo
o`006P3oool00`000000oooo0?ooo`1E0?ooo`800000`03oool2000005L0oooo00<000000?ooo`3o
ool02@3oool001X0oooo0`00001F0?ooo`<00000_03oool4000005H0oooo0P00000;0?ooo`006P3o
ool00`000000oooo0?ooo`1G0?ooo`<00000^P3oool4000005P0oooo00<000000?ooo`3oool02@3o
ool001X0oooo00<000000?ooo`3oool0F03oool200000;T0oooo0`00001J0?ooo`030000003oool0
oooo00T0oooo000J0?ooo`030000003oool0oooo05T0oooo0P0000140?ooo`800000;03oool20000
04<0oooo0`00001K0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo05X0
oooo0`0000130?ooo`030000003oool0oooo02/0oooo00<000000?ooo`3oool0@03oool3000005`0
oooo00<000000?ooo`3oool02@3oool001X0oooo0`00001K0?ooo`<00000>P3oool2000003T0oooo
0P00000k0?ooo`<00000G03oool2000000/0oooo000J0?ooo`030000003oool0oooo05d0oooo0P00
000j0?ooo`030000003oool0oooo00@0oooo0P00000Z0?ooo`8000001P3oool00`000000oooo0?oo
o`0h0?ooo`<00000GP3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`1N
0?ooo`<00000@03oool00`000000oooo0?ooo`0Y0?ooo`030000003oool0oooo03d0oooo1000001O
0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo05l0oooo0P00000h0?oo
o`800000=`3oool2000003T0oooo0`00001P0?ooo`030000003oool0oooo00T0oooo000J0?ooo`03
0000003oool0oooo0600oooo0`00000g0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3o
ool0=@3oool2000000030?ooo`000000oooo0600oooo00<000000?ooo`3oool02@3oool001X0oooo
0`00001R0?ooo`800000=P3oool2000003D0oooo0P00000g0?ooo`<00000HP3oool2000000/0oooo
000J0?ooo`030000003oool0oooo0680oooo0`00000`0?ooo`800000103oool00`000000oooo0?oo
o`0d0?ooo`030000003oool0oooo0080oooo0P00000_0?ooo`@00000I03oool00`000000oooo0?oo
o`090?ooo`006P3oool00`000000oooo0?ooo`1S0?ooo`<00000<03oool00`000000oooo0?ooo`0o
0?ooo`030000003oool0oooo02d0oooo0`00001U0?ooo`030000003oool0oooo00T0oooo000;0?oo
o`@000000P3oool3000000H0oooo00<000000?ooo`3oool0I@3oool2000002L0oooo0P00000M0?oo
o`800000403oool2000001d0oooo0P00000W0?ooo`<00000IP3oool00`000000oooo0?ooo`090?oo
o`003@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00<0oooo1000001U0?ooo`80
0000803oool2000000D0oooo00<000000?ooo`3oool0703oool00`000000oooo0?ooo`0?0?ooo`03
0000003oool0oooo01`0oooo00@000000?ooo`3oool0oooo0P00000Q0?ooo`<00000I@3oool30000
00/0oooo000=0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool00`3oool00`000000
oooo0?ooo`1W0?ooo`800000803oool00`000000oooo0?ooo`1H0?ooo`030000003oool0oooo01h0
oooo0`00001X0?ooo`030000003oool0oooo00T0oooo000=0?ooo`040000003oool0oooo0?ooo`<0
00001P3oool00`000000oooo0?ooo`1X0?ooo`<000009@3oool2000000D0oooo0P00000l0?ooo`80
00001@3oool2000002D0oooo1000001Y0?ooo`030000003oool0oooo00T0oooo000=0?ooo`050000
003oool0oooo0?ooo`000000203oool00`000000oooo0?ooo`1Y0?ooo`<000006P3oool200000080
oooo0P0000050?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0>`3oool00`000000
oooo0?ooo`040?ooo`040000003oool0oooo0?ooo`8000000P3oool2000001/0oooo1000001Z0?oo
o`030000003oool0oooo00T0oooo000;0?ooo`<000000`3oool00`000000oooo0?ooo`060?ooo`03
0000003oool0oooo06X0oooo0P00000K0?ooo`050000003oool0oooo0?ooo`000000EP3oool01@00
0000oooo0?ooo`3oool0000001`0oooo00<000000?ooo`3oool0JP3oool00`000000oooo0?ooo`09
0?ooo`003@3oool010000000oooo0?ooo`3oool4000000D0oooo0`00001/0?ooo`8000006P3oool2
00000080oooo0P0000030?ooo`800000A`3oool2000000@0oooo0P0000020?ooo`8000006@3oool2
000006h0oooo0P00000;0?ooo`006P3oool00`000000oooo0?ooo`1]0?ooo`030000003oool0oooo
01T0oooo00D000000?ooo`3oool0oooo000000040?ooo`030000003oool0oooo04H0oooo00<00000
0?ooo`3oool00`3oool01@000000oooo0?ooo`3oool0000001X0oooo00<000000?ooo`3oool0K@3o
ool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`3o0?ooog40oooo00<00000
0?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0K`3oool2000001T0oooo0P0000020?oo
o`8000000`3oool2000004@0oooo0P0000030?ooo`8000000P3oool2000001P0oooo0`00001a0?oo
o`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo0700oooo0P00000I0?ooo`05
0000003oool0oooo0?ooo`000000103oool00`000000oooo0?ooo`0O0?ooo`8000008P3oool00`00
0000oooo0?ooo`020?ooo`050000003oool0oooo0?ooo`000000603oool300000740oooo00<00000
0?ooo`3oool02@3oool001X0oooo0`00001a0?ooo`<000006`3oool2000002H0oooo00<000000?oo
o`3oool08`3oool2000001/0oooo1000001a0?ooo`8000002`3oool001X0oooo00<000000?ooo`3o
ool0LP3oool2000001`0oooo00<000000?ooo`3oool0B`3oool00`000000oooo0?ooo`0H0?ooo`@0
0000L`3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`1c0?ooo`<00000
6P3oool2000002@0oooo0P00000S0?ooo`8000006`3oool3000007D0oooo00<000000?ooo`3oool0
2@3oool001X0oooo00<000000?ooo`3oool0M@3oool2000001X0oooo00<000000?ooo`3oool05`3o
ool2000000X0oooo00<000000?ooo`3oool01`3oool2000001T0oooo00<000000?ooo`3oool0603o
ool3000007H0oooo00<000000?ooo`3oool02@3oool001X0oooo0`00001f0?ooo`<00000<`3oool0
0`000000oooo0?ooo`0C0?ooo`030000003oool0oooo0300oooo0P0000000`3oool000000?ooo`1e
0?ooo`8000002`3oool001X0oooo00<000000?ooo`3oool0M`3oool300000380oooo0P00000B0?oo
o`800000<P3oool3000007T0oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3o
ool0N@3oool200000380oooo00<000000?ooo`3oool04@3oool00`000000oooo0?ooo`0_0?ooo`<0
0000NP3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`1j0?ooo`800000
<03oool200000180oooo0P00000`0?ooo`<00000N`3oool00`000000oooo0?ooo`090?ooo`006P3o
ool00`000000oooo0?ooo`1k0?ooo`030000003oool0oooo02l0oooo00<000000?ooo`3oool04@3o
ool00`000000oooo0?ooo`0_0?ooo`030000003oool0oooo07X0oooo00<000000?ooo`3oool02@3o
ool001X0oooo0`00001l0?ooo`800000;P3oool200000140oooo0P00000_0?ooo`800000O@3oool2
000000/0oooo000J0?ooo`030000003oool0oooo07d0oooo00<000000?ooo`3oool0;@3oool00`00
0000oooo0?ooo`0@0?ooo`030000003oool0oooo02h0oooo00<000000?ooo`3oool0O03oool00`00
0000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`1n0?ooo`<00000;03oool200000100
oooo0P00000/0?ooo`800000P@3oool00`000000oooo0?ooo`090?ooo`002`3oool4000000<0oooo
0P0000060?ooo`030000003oool0oooo07l0oooo0P00000]0?ooo`030000003oool0oooo00l0oooo
00<000000?ooo`3oool0:P3oool200000840oooo00<000000?ooo`3oool02@3oool000/0oooo00<0
00000?ooo`3oool00`3oool010000000oooo0?ooo`0000050?ooo`@00000O`3oool3000002X0oooo
0P00000@0?ooo`800000:P3oool4000007l0oooo0`00000;0?ooo`00303oool00`000000oooo0?oo
o`020?ooo`040000003oool0oooo000000D0oooo00<000000?ooo`3oool0P@3oool3000002X0oooo
00<000000?ooo`3oool03`3oool00`000000oooo0?ooo`0W0?ooo`@00000PP3oool00`000000oooo
0?ooo`090?ooo`003@3oool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`3oool0
0`3oool00`000000oooo0?ooo`230?ooo`800000I03oool3000008@0oooo00<000000?ooo`3oool0
2@3oool000h0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`030?ooo`030000
003oool0oooo08<0oooo0`00001Q0?ooo`@00000Q@3oool00`000000oooo0?ooo`090?ooo`002`3o
ool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000000D0oooo00<000000?ooo`3o
ool0Q03oool300000600oooo0`0000260?ooo`030000003oool0oooo00T0oooo000<0?ooo`800000
103oool2000000H0oooo0`0000260?ooo`800000G@3oool2000000030?ooo`000000oooo08D0oooo
0P00000;0?ooo`006P3oool00`000000oooo0?ooo`270?ooo`030000003oool0oooo05`0oooo00<0
00000?ooo`3oool0Q`3oool00`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`28
0?ooo`800000F@3oool3000008X0oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?oo
o`3oool0R@3oool3000005H0oooo1000002:0?ooo`030000003oool0oooo00T0oooo000J0?ooo`03
0000003oool0oooo08X0oooo0P00001E0?ooo`@00000R`3oool00`000000oooo0?ooo`090?ooo`00
6P3oool3000008/0oooo0`00001B0?ooo`<00000S03oool2000000/0oooo000J0?ooo`030000003o
ool0oooo08`0oooo0P00001C0?ooo`030000003oool0oooo08`0oooo00<000000?ooo`3oool02@3o
ool001X0oooo00<000000?ooo`3oool0SP3oool2000004h0oooo0P00002@0?ooo`030000003oool0
oooo00T0oooo000J0?ooo`030000003oool0oooo08l0oooo0`00001:0?ooo`@00000T03oool00`00
0000oooo0?ooo`090?ooo`006P3oool300000900oooo0P0000190?ooo`@00000T03oool2000000/0
oooo000J0?ooo`030000003oool0oooo0940oooo0`0000160?ooo`<00000T`3oool00`000000oooo
0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`2B0?ooo`<00000A03oool3000009@0oooo00<0
00000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0U03oool00`000000oooo0?ooo`13
0?ooo`030000003oool0oooo09<0oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?oo
o`3oool0U03oool300000400oooo0P00002G0?ooo`030000003oool0oooo00T0oooo000J0?ooo`<0
0000U@3oool3000003h0oooo0`00002F0?ooo`8000002`3oool001X0oooo00<000000?ooo`3oool0
U`3oool2000003`0oooo0`00002H0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003o
ool0oooo09P0oooo0`00000h0?ooo`80000000<0oooo0000003oool0V03oool00`000000oooo0?oo
o`090?ooo`002`3oool400000080oooo0`0000060?ooo`030000003oool0oooo09T0oooo0P00000g
0?ooo`<00000V`3oool00`000000oooo0?ooo`090?ooo`002`3oool00`000000oooo0?ooo`060?oo
o`030000003oool0oooo00<0oooo00<000000?ooo`3oool0VP3oool3000003<0oooo1000002L0?oo
o`030000003oool0oooo00T0oooo000<0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool00`3oool4000009X0oooo0`00000b0?ooo`<00000V`3oool3000000/0oooo000=0?ooo`040000
003oool0oooo0?ooo`<000001P3oool00`000000oooo0?ooo`2M0?ooo`800000<03oool3000009h0
oooo00<000000?ooo`3oool02@3oool000h0oooo00@000000?ooo`3oool00000203oool00`000000
oooo0?ooo`2N0?ooo`800000;P3oool3000009l0oooo00<000000?ooo`3oool02@3oool000/0oooo
00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo09l0
oooo0P00000/0?ooo`<00000X03oool00`000000oooo0?ooo`090?ooo`00303oool2000000<0oooo
100000050?ooo`<00000X03oool3000002P0oooo1000002P0?ooo`8000002`3oool001X0oooo00<0
00000?ooo`3oool0X@3oool3000002H0oooo1000002R0?ooo`030000003oool0oooo00T0oooo000J
0?ooo`030000003oool0oooo0:80oooo0`00000T0?ooo`<00000Y03oool00`000000oooo0?ooo`09
0?ooo`006P3oool00`000000oooo0?ooo`2S0?ooo`<000008P3oool300000:D0oooo00<000000?oo
o`3oool02@3oool001X0oooo00<000000?ooo`3oool0Y@3oool200000200oooo0`00002V0?ooo`03
0000003oool0oooo00T0oooo000J0?ooo`<00000YP3oool2000001h0oooo0`00002V0?ooo`800000
2`3oool001X0oooo00<000000?ooo`3oool0Y`3oool3000001X0oooo1000002X0?ooo`030000003o
ool0oooo00T0oooo000J0?ooo`030000003oool0oooo0:P0oooo0P00000I0?ooo`@00000Z@3oool0
0`000000oooo0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`2Y0?ooo`<000005P3oool30000
0:/0oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3oool0ZP3oool3000001@0
oooo1000002[0?ooo`030000003oool0oooo00T0oooo000J0?ooo`<00000Z`3oool300000140oooo
1000002/0?ooo`8000002`3oool001X0oooo00<000000?ooo`3oool0[@3oool200000100oooo0`00
002^0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo0:h0oooo00<00000
0?ooo`3oool03`3oool00`000000oooo0?ooo`2]0?ooo`030000003oool0oooo00T0oooo000J0?oo
o`030000003oool0oooo0?l0ooooL@3oool00`000000oooo0?ooo`090?ooo`006P3oool300000?l0
ooooL03oool2000000/0oooo000J0?ooo`030000003oool0oooo0?l0ooooL@3oool00`000000oooo
0?ooo`090?ooo`006P3oool00`000000oooo0?ooo`3o0?ooog40oooo00<000000?ooo`3oool02@3o
ool001X0oooo00<000000?ooo`3oool0o`3oooma0?ooo`030000003oool0oooo00T0oooo000<0?oo
o`800000103oool2000000H0oooo00<000000?ooo`3oool0o`3oooma0?ooo`030000003oool0oooo
00T0oooo000;0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool000001@3oool4
00000;D0oooo0P00002f0?ooo`<000002`3oool000h0oooo00@000000?ooo`3oool000000P3oool0
0`000000oooo0?ooo`030?ooo`030000003oool0oooo0;L0oooo00<000000?ooo`3oool0]P3oool0
0`000000oooo0?ooo`090?ooo`00303oool2000000<0oooo00@000000?ooo`3oool000001@3oool0
0`000000oooo0?ooo`3o0?ooog40oooo00<000000?ooo`3oool02@3oool000h0oooo00@000000?oo
o`3oool000000P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo0?l0ooooL@3oool0
0`000000oooo0?ooo`090?ooo`002`3oool010000000oooo0?ooo`0000020?ooo`040000003oool0
oooo000000D0oooo0`00003o0?ooog00oooo0P00000;0?ooo`00303oool2000000@0oooo0P000006
0?ooo`030000003oool0oooo0?l0ooooL@3oool00`000000oooo0?ooo`090?ooo`006P3oool00`00
0000oooo0?ooo`3o0?ooog40oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3o
ool0o`3oooma0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo0?l0oooo
L@3oool00`000000oooo0?ooo`090?ooo`006P3oool300000?l0ooooL03oool2000000/0oooo000J
0?ooo`030000003oool0oooo0?l0ooooL@3oool00`000000oooo0?ooo`090?ooo`006P3oool00`00
0000oooo0?ooo`3o0?ooog40oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3o
ool0o`3oooma0?ooo`030000003oool0oooo00T0oooo000J0?ooo`030000003oool0oooo0?l0oooo
L@3oool00`000000oooo0?ooo`090?ooo`006P3oool300000?l0ooooL03oool2000000/0oooo000J
0?ooo`030000003oool0oooo0?l0ooooL@3oool00`000000oooo0?ooo`090?ooo`006P3oool00`00
0000oooo0?ooo`3o0?ooog40oooo00<000000?ooo`3oool02@3oool001X0oooo00<000000?ooo`3o
ool0o`3oooma0?ooo`030000003oool0oooo00T0oooo000J0?ooo`<00000o`3ooom`0?ooo`800000
2`3oool001X0oooo00<000000?ooo`3oool0o`3oooma0?ooo`030000003oool0oooo00T0oooo000J
0?ooo`030000003oool0oooo0?l0ooooL@3oool00`000000oooo0?ooo`090?ooo`006P3oool00`00
0000oooo0?ooo`0H0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0<P3oool00`00
0000oooo0?ooo`0b0?ooo`030000003oool0oooo0380oooo00<000000?ooo`3oool0<P3oool00`00
0000oooo0?ooo`0b0?ooo`030000003oool0oooo01L0oooo00<000000?ooo`3oool02@3oool000`0
oooo0P0000030?ooo`<000001P3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00L0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00L0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool00P3oool00`000000oooo0?ooo`090?ooo`002`3oool010000000oooo
0?ooo`0000050?ooo`030000003oool0oooo00<0ooooo`00001e000000/0oooo000>0?ooo`030000
003oool0oooo00<0oooo00<000000?ooo`3oool0o`3ooon30?ooo`00303oool2000000<0oooo0`00
003o0?ooohH0oooo000>0?ooo`040000003oool0oooo00000?l0ooooR03oool000/0oooo00@00000
0?ooo`3oool000000P3oool00`000000oooo0?ooo`3o0?ooohH0oooo0000\
\>"],
  ImageRangeCache->{{{0, 408}, {203.5, 0}} -> {-10.0774, -7.94233, 0.188818, 
  0.215792}}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(X = AnglePairsAndRationalApproxs[50, 10^\(-3\)]; \nLength[X]\), 
    \(Angles = \(Transpose[X]\)[\([1]\)]; \n
    ListPlot[Angles, \ AspectRatio -> 1/2, Frame -> True, \ Axes -> False, \ 
      PlotRange -> {{\(-5\), 65}, {\(-5\), 35}}]; 
    WriteData["\<~/ASP/WebServices/GrainBreaker/2DGeometries/Geo50.dat\>", 
      X]; \)}], "Input"],

Cell[BoxData[
    \(2685\)], "Output"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .5 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0714286 0.0142857 0.0625 0.0125 [
[.07143 -0.0125 -3 -9 ]
[.07143 -0.0125 3 0 ]
[.21429 -0.0125 -6 -9 ]
[.21429 -0.0125 6 0 ]
[.35714 -0.0125 -6 -9 ]
[.35714 -0.0125 6 0 ]
[.5 -0.0125 -6 -9 ]
[.5 -0.0125 6 0 ]
[.64286 -0.0125 -6 -9 ]
[.64286 -0.0125 6 0 ]
[.78571 -0.0125 -6 -9 ]
[.78571 -0.0125 6 0 ]
[.92857 -0.0125 -6 -9 ]
[.92857 -0.0125 6 0 ]
[ 0 0 -0.125 0 ]
[-0.0125 .0625 -6 -4.5 ]
[-0.0125 .0625 0 4.5 ]
[-0.0125 .125 -6 -4.5 ]
[-0.0125 .125 0 4.5 ]
[-0.0125 .1875 -12 -4.5 ]
[-0.0125 .1875 0 4.5 ]
[-0.0125 .25 -12 -4.5 ]
[-0.0125 .25 0 4.5 ]
[-0.0125 .3125 -12 -4.5 ]
[-0.0125 .3125 0 4.5 ]
[-0.0125 .375 -12 -4.5 ]
[-0.0125 .375 0 4.5 ]
[-0.0125 .4375 -12 -4.5 ]
[-0.0125 .4375 0 4.5 ]
[-0.0125 .5 -12 -4.5 ]
[-0.0125 .5 0 4.5 ]
[ 0 0 -0.125 0 ]
[ 0 .5 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .5 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.07143 0 m
.07143 .00625 L
s
[(0)] .07143 -0.0125 0 1 Mshowa
.21429 0 m
.21429 .00625 L
s
[(10)] .21429 -0.0125 0 1 Mshowa
.35714 0 m
.35714 .00625 L
s
[(20)] .35714 -0.0125 0 1 Mshowa
.5 0 m
.5 .00625 L
s
[(30)] .5 -0.0125 0 1 Mshowa
.64286 0 m
.64286 .00625 L
s
[(40)] .64286 -0.0125 0 1 Mshowa
.78571 0 m
.78571 .00625 L
s
[(50)] .78571 -0.0125 0 1 Mshowa
.92857 0 m
.92857 .00625 L
s
[(60)] .92857 -0.0125 0 1 Mshowa
.125 Mabswid
.1 0 m
.1 .00375 L
s
.12857 0 m
.12857 .00375 L
s
.15714 0 m
.15714 .00375 L
s
.18571 0 m
.18571 .00375 L
s
.24286 0 m
.24286 .00375 L
s
.27143 0 m
.27143 .00375 L
s
.3 0 m
.3 .00375 L
s
.32857 0 m
.32857 .00375 L
s
.38571 0 m
.38571 .00375 L
s
.41429 0 m
.41429 .00375 L
s
.44286 0 m
.44286 .00375 L
s
.47143 0 m
.47143 .00375 L
s
.52857 0 m
.52857 .00375 L
s
.55714 0 m
.55714 .00375 L
s
.58571 0 m
.58571 .00375 L
s
.61429 0 m
.61429 .00375 L
s
.67143 0 m
.67143 .00375 L
s
.7 0 m
.7 .00375 L
s
.72857 0 m
.72857 .00375 L
s
.75714 0 m
.75714 .00375 L
s
.81429 0 m
.81429 .00375 L
s
.84286 0 m
.84286 .00375 L
s
.87143 0 m
.87143 .00375 L
s
.9 0 m
.9 .00375 L
s
.04286 0 m
.04286 .00375 L
s
.01429 0 m
.01429 .00375 L
s
.95714 0 m
.95714 .00375 L
s
.98571 0 m
.98571 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
0 .0625 m
.00625 .0625 L
s
[(0)] -0.0125 .0625 1 0 Mshowa
0 .125 m
.00625 .125 L
s
[(5)] -0.0125 .125 1 0 Mshowa
0 .1875 m
.00625 .1875 L
s
[(10)] -0.0125 .1875 1 0 Mshowa
0 .25 m
.00625 .25 L
s
[(15)] -0.0125 .25 1 0 Mshowa
0 .3125 m
.00625 .3125 L
s
[(20)] -0.0125 .3125 1 0 Mshowa
0 .375 m
.00625 .375 L
s
[(25)] -0.0125 .375 1 0 Mshowa
0 .4375 m
.00625 .4375 L
s
[(30)] -0.0125 .4375 1 0 Mshowa
0 .5 m
.00625 .5 L
s
[(35)] -0.0125 .5 1 0 Mshowa
.125 Mabswid
0 .0125 m
.00375 .0125 L
s
0 .025 m
.00375 .025 L
s
0 .0375 m
.00375 .0375 L
s
0 .05 m
.00375 .05 L
s
0 .075 m
.00375 .075 L
s
0 .0875 m
.00375 .0875 L
s
0 .1 m
.00375 .1 L
s
0 .1125 m
.00375 .1125 L
s
0 .1375 m
.00375 .1375 L
s
0 .15 m
.00375 .15 L
s
0 .1625 m
.00375 .1625 L
s
0 .175 m
.00375 .175 L
s
0 .2 m
.00375 .2 L
s
0 .2125 m
.00375 .2125 L
s
0 .225 m
.00375 .225 L
s
0 .2375 m
.00375 .2375 L
s
0 .2625 m
.00375 .2625 L
s
0 .275 m
.00375 .275 L
s
0 .2875 m
.00375 .2875 L
s
0 .3 m
.00375 .3 L
s
0 .325 m
.00375 .325 L
s
0 .3375 m
.00375 .3375 L
s
0 .35 m
.00375 .35 L
s
0 .3625 m
.00375 .3625 L
s
0 .3875 m
.00375 .3875 L
s
0 .4 m
.00375 .4 L
s
0 .4125 m
.00375 .4125 L
s
0 .425 m
.00375 .425 L
s
0 .45 m
.00375 .45 L
s
0 .4625 m
.00375 .4625 L
s
0 .475 m
.00375 .475 L
s
0 .4875 m
.00375 .4875 L
s
.25 Mabswid
0 0 m
0 .5 L
s
.07143 .49375 m
.07143 .5 L
s
.21429 .49375 m
.21429 .5 L
s
.35714 .49375 m
.35714 .5 L
s
.5 .49375 m
.5 .5 L
s
.64286 .49375 m
.64286 .5 L
s
.78571 .49375 m
.78571 .5 L
s
.92857 .49375 m
.92857 .5 L
s
.125 Mabswid
.1 .49625 m
.1 .5 L
s
.12857 .49625 m
.12857 .5 L
s
.15714 .49625 m
.15714 .5 L
s
.18571 .49625 m
.18571 .5 L
s
.24286 .49625 m
.24286 .5 L
s
.27143 .49625 m
.27143 .5 L
s
.3 .49625 m
.3 .5 L
s
.32857 .49625 m
.32857 .5 L
s
.38571 .49625 m
.38571 .5 L
s
.41429 .49625 m
.41429 .5 L
s
.44286 .49625 m
.44286 .5 L
s
.47143 .49625 m
.47143 .5 L
s
.52857 .49625 m
.52857 .5 L
s
.55714 .49625 m
.55714 .5 L
s
.58571 .49625 m
.58571 .5 L
s
.61429 .49625 m
.61429 .5 L
s
.67143 .49625 m
.67143 .5 L
s
.7 .49625 m
.7 .5 L
s
.72857 .49625 m
.72857 .5 L
s
.75714 .49625 m
.75714 .5 L
s
.81429 .49625 m
.81429 .5 L
s
.84286 .49625 m
.84286 .5 L
s
.87143 .49625 m
.87143 .5 L
s
.9 .49625 m
.9 .5 L
s
.04286 .49625 m
.04286 .5 L
s
.01429 .49625 m
.01429 .5 L
s
.95714 .49625 m
.95714 .5 L
s
.98571 .49625 m
.98571 .5 L
s
.25 Mabswid
0 .5 m
1 .5 L
s
.99375 0 m
1 0 L
s
.99375 .0625 m
1 .0625 L
s
.99375 .125 m
1 .125 L
s
.99375 .1875 m
1 .1875 L
s
.99375 .25 m
1 .25 L
s
.99375 .3125 m
1 .3125 L
s
.99375 .375 m
1 .375 L
s
.99375 .4375 m
1 .4375 L
s
.125 Mabswid
.99625 .0125 m
1 .0125 L
s
.99625 .025 m
1 .025 L
s
.99625 .0375 m
1 .0375 L
s
.99625 .05 m
1 .05 L
s
.99625 .075 m
1 .075 L
s
.99625 .0875 m
1 .0875 L
s
.99625 .1 m
1 .1 L
s
.99625 .1125 m
1 .1125 L
s
.99625 .1375 m
1 .1375 L
s
.99625 .15 m
1 .15 L
s
.99625 .1625 m
1 .1625 L
s
.99625 .175 m
1 .175 L
s
.99625 .2 m
1 .2 L
s
.99625 .2125 m
1 .2125 L
s
.99625 .225 m
1 .225 L
s
.99625 .2375 m
1 .2375 L
s
.99625 .2625 m
1 .2625 L
s
.99625 .275 m
1 .275 L
s
.99625 .2875 m
1 .2875 L
s
.99625 .3 m
1 .3 L
s
.99625 .325 m
1 .325 L
s
.99625 .3375 m
1 .3375 L
s
.99625 .35 m
1 .35 L
s
.99625 .3625 m
1 .3625 L
s
.99625 .3875 m
1 .3875 L
s
.99625 .4 m
1 .4 L
s
.99625 .4125 m
1 .4125 L
s
.99625 .425 m
1 .425 L
s
.99625 .45 m
1 .45 L
s
.99625 .4625 m
1 .4625 L
s
.99625 .475 m
1 .475 L
s
.99625 .4875 m
1 .4875 L
s
.25 Mabswid
1 0 m
1 .5 L
s
0 0 m
1 0 L
1 .5 L
0 .5 L
closepath
clip
newpath
.008 w
.07143 .0625 Mdot
.07972 .06975 Mdot
.08002 .07002 Mdot
.08034 .0703 Mdot
.08069 .07061 Mdot
.08107 .07094 Mdot
.08148 .0713 Mdot
.08193 .07169 Mdot
.08242 .07212 Mdot
.08295 .07258 Mdot
.08354 .0731 Mdot
.0842 .07367 Mdot
.08493 .07431 Mdot
.08575 .07503 Mdot
.08667 .07584 Mdot
.08772 .07676 Mdot
.0883 .0625 Mdot
.0883 .07727 Mdot
.08893 .0625 Mdot
.08893 .07781 Mdot
.0896 .0625 Mdot
.0896 .0784 Mdot
.09033 .0625 Mdot
.09033 .07904 Mdot
.09112 .0625 Mdot
.09112 .07973 Mdot
.09197 .0625 Mdot
.09197 .08047 Mdot
.0929 .0625 Mdot
.0929 .08129 Mdot
.09393 .0625 Mdot
.09393 .08219 Mdot
.09505 .0625 Mdot
.09505 .08317 Mdot
.09629 .0625 Mdot
.09629 .06975 Mdot
.09629 .08426 Mdot
.0972 .07002 Mdot
.0972 .08505 Mdot
.09767 .0625 Mdot
.09767 .08546 Mdot
.09817 .0703 Mdot
.09817 .0859 Mdot
.09922 .0625 Mdot
.09922 .07061 Mdot
.09922 .08681 Mdot
.10035 .07094 Mdot
.10035 .08781 Mdot
.10095 .0625 Mdot
.10095 .08833 Mdot
.10158 .0713 Mdot
.10158 .08888 Mdot
.10292 .0625 Mdot
.10292 .07169 Mdot
.10292 .09005 Mdot
.10438 .07212 Mdot
.10438 .09133 Mdot
.10516 .0625 Mdot
.10516 .07727 Mdot
.10516 .09202 Mdot
.10599 .07258 Mdot
.10599 .09274 Mdot
.10641 .0625 Mdot
.10641 .07781 Mdot
.10641 .09311 Mdot
.10776 .0625 Mdot
.10776 .0731 Mdot
.10776 .0784 Mdot
.10776 .09429 Mdot
.10921 .0625 Mdot
.10921 .07904 Mdot
.10921 .09556 Mdot
.10972 .07367 Mdot
.10972 .096 Mdot
.11078 .0625 Mdot
.11078 .07973 Mdot
.11078 .09693 Mdot
.1119 .07431 Mdot
.1119 .09791 Mdot
.11249 .0625 Mdot
.11249 .08047 Mdot
.11249 .09843 Mdot
.11285 .06975 Mdot
.11285 .08426 Mdot
.11285 .09874 Mdot
.11435 .0625 Mdot
.11435 .07002 Mdot
.11435 .07503 Mdot
.11435 .08129 Mdot
.11435 .08505 Mdot
.11435 .10006 Mdot
.11597 .0703 Mdot
.11597 .0859 Mdot
.11597 .10147 Mdot
.11639 .0625 Mdot
.11639 .08219 Mdot
.11639 .10184 Mdot
.11711 .07584 Mdot
.11711 .10247 Mdot
.11771 .07061 Mdot
.11771 .08681 Mdot
.11771 .103 Mdot
.11863 .0625 Mdot
.11863 .08317 Mdot
.11863 .1038 Mdot
.11959 .07094 Mdot
.11959 .08781 Mdot
.11959 .10464 Mdot
.12026 .07676 Mdot
.12026 .10522 Mdot
.12111 .0625 Mdot
.12111 .08426 Mdot
.12111 .10597 Mdot
.12164 .0713 Mdot
.12164 .08888 Mdot
.12164 .10643 Mdot
.122 .07727 Mdot
.122 .09202 Mdot
.122 .10675 Mdot
.12386 .07169 Mdot
.12386 .07781 Mdot
.12386 .08546 Mdot
.12386 .09005 Mdot
.12386 .09311 Mdot
.12386 .10838 Mdot
.12588 .0784 Mdot
.12588 .09429 Mdot
.12588 .11014 Mdot
.1263 .09133 Mdot
.1263 .11051 Mdot
.12694 .08681 Mdot
.12694 .11107 Mdot
.12805 .09556 Mdot
.12805 .11204 Mdot
.12896 .09274 Mdot
.12896 .11284 Mdot
.12937 .09874 Mdot
.12937 .1132 Mdot
.1304 .08833 Mdot
.1304 .09693 Mdot
.1304 .1141 Mdot
.13147 .10006 Mdot
.13147 .11503 Mdot
.1319 .09429 Mdot
.1319 .11542 Mdot
.13295 .09843 Mdot
.13295 .11633 Mdot
.13372 .10147 Mdot
.13372 .11701 Mdot
.13431 .11752 Mdot
.13516 .096 Mdot
.13516 .11826 Mdot
.13574 .10006 Mdot
.13574 .11877 Mdot
.13616 .103 Mdot
.13616 .11914 Mdot
.13879 .10184 Mdot
.13879 .10464 Mdot
.13879 .10675 Mdot
.13879 .12144 Mdot
.14127 .10838 Mdot
.14127 .12361 Mdot
.14164 .10643 Mdot
.14164 .12393 Mdot
.14214 .12437 Mdot
.14285 .12499 Mdot
.14394 .11014 Mdot
.14394 .12595 Mdot
.14474 .10838 Mdot
.14474 .12665 Mdot
.14584 .1132 Mdot
.14584 .12761 Mdot
.14683 .11204 Mdot
.14683 .12847 Mdot
.14743 .129 Mdot
.14813 .12962 Mdot
.14853 .11503 Mdot
.14853 .12996 Mdot
.14995 .1312 Mdot
.15142 .11701 Mdot
.15142 .13249 Mdot
.15185 .13287 Mdot
.15264 .13356 Mdot
.15334 .13417 Mdot
.15454 .13522 Mdot
.15552 .12144 Mdot
.15552 .13608 Mdot
.15595 .13645 Mdot
.15704 .13741 Mdot
.1579 .13817 Mdot
.15861 .13878 Mdot
.15969 .13973 Mdot
.16048 .14042 Mdot
.16108 .14095 Mdot
.16155 .14136 Mdot
.16194 .14169 Mdot
.16225 .14197 Mdot
.16553 .14483 Mdot
.16904 .14791 Mdot
.16941 .14823 Mdot
.16986 .14863 Mdot
.17043 .14913 Mdot
.17117 .14978 Mdot
.17218 .15066 Mdot
.17283 .15123 Mdot
.17363 .15192 Mdot
.17461 .15279 Mdot
.17587 .15389 Mdot
.17693 .15481 Mdot
.17754 .15535 Mdot
.17822 .15594 Mdot
.17859 .15627 Mdot
.17984 .15736 Mdot
.18136 .15869 Mdot
.18194 .1592 Mdot
.18244 .15964 Mdot
.18324 .16034 Mdot
.18413 .16111 Mdot
.18477 .16167 Mdot
.18563 .16243 Mdot
.18619 .16291 Mdot
.18658 .16325 Mdot
.18876 .16517 Mdot
.19103 .16715 Mdot
.19145 .16752 Mdot
.19206 .16806 Mdot
.19304 .16891 Mdot
.19379 .16956 Mdot
.19484 .17049 Mdot
.19584 .17136 Mdot
.19646 .17191 Mdot
.19722 .17256 Mdot
.19765 .17294 Mdot
.19926 .17435 Mdot
.20104 .17591 Mdot
.20159 .17639 Mdot
.20261 .17729 Mdot
.20356 .17811 Mdot
.204 .1785 Mdot
.20524 .17959 Mdot
.2067 .18087 Mdot
.20736 .18144 Mdot
.20798 .18198 Mdot
.20911 .18297 Mdot
.21011 .18384 Mdot
.21057 .18425 Mdot
.21181 .18533 Mdot
.21287 .18626 Mdot
.2138 .18708 Mdot
.21461 .18779 Mdot
.21533 .18842 Mdot
.21597 .18898 Mdot
.21655 .18948 Mdot
.21753 .19034 Mdot
.21835 .19106 Mdot
.21904 .19166 Mdot
.21963 .19217 Mdot
.22013 .19262 Mdot
.22058 .193 Mdot
.22096 .19334 Mdot
.22131 .0625 Mdot
.22131 .19365 Mdot
.22162 .06975 Mdot
.22162 .08426 Mdot
.22162 .19391 Mdot
.22705 .07002 Mdot
.22705 .0784 Mdot
.22705 .08505 Mdot
.22705 .19867 Mdot
.23256 .0625 Mdot
.23256 .07727 Mdot
.23256 .09202 Mdot
.23256 .10675 Mdot
.23256 .12144 Mdot
.23256 .20349 Mdot
.23288 .0703 Mdot
.23288 .0859 Mdot
.23288 .10147 Mdot
.23288 .20377 Mdot
.23324 .0625 Mdot
.23324 .07904 Mdot
.23324 .09556 Mdot
.23324 .20408 Mdot
.23364 .0713 Mdot
.23364 .20444 Mdot
.2341 .20484 Mdot
.23463 .2053 Mdot
.23524 .20584 Mdot
.23596 .20647 Mdot
.23683 .20722 Mdot
.23787 .09874 Mdot
.23787 .1132 Mdot
.23787 .12761 Mdot
.23787 .20814 Mdot
.23848 .09311 Mdot
.23848 .10838 Mdot
.23848 .12361 Mdot
.23848 .20867 Mdot
.23916 .07061 Mdot
.23916 .08681 Mdot
.23916 .103 Mdot
.23916 .11914 Mdot
.23916 .20927 Mdot
.23994 .0625 Mdot
.23994 .07973 Mdot
.23994 .09693 Mdot
.23994 .20994 Mdot
.24081 .07169 Mdot
.24081 .09005 Mdot
.24081 .21071 Mdot
.24181 .21159 Mdot
.24298 .2126 Mdot
.24386 .11503 Mdot
.24386 .12996 Mdot
.24386 .21337 Mdot
.24433 .21379 Mdot
.24484 .11014 Mdot
.24484 .12595 Mdot
.24484 .21424 Mdot
.24595 .08781 Mdot
.24595 .10464 Mdot
.24595 .12144 Mdot
.24595 .2152 Mdot
.2472 .0625 Mdot
.2472 .08047 Mdot
.2472 .09843 Mdot
.2472 .11633 Mdot
.2472 .2163 Mdot
.24789 .21691 Mdot
.24863 .07212 Mdot
.24863 .09133 Mdot
.24863 .21755 Mdot
.24902 .13608 Mdot
.24902 .15066 Mdot
.24902 .2179 Mdot
.25028 .0625 Mdot
.25028 .08317 Mdot
.25028 .13249 Mdot
.25028 .219 Mdot
.2517 .11204 Mdot
.2517 .12847 Mdot
.2517 .22023 Mdot
.25221 .22068 Mdot
.25329 .10643 Mdot
.25329 .12393 Mdot
.25329 .22163 Mdot
.25399 .14197 Mdot
.25399 .15627 Mdot
.25399 .22224 Mdot
.25448 .22267 Mdot
.25511 .10006 Mdot
.25511 .11877 Mdot
.25511 .22322 Mdot
.2555 .13878 Mdot
.2555 .15389 Mdot
.2555 .22357 Mdot
.2572 .07258 Mdot
.2572 .09274 Mdot
.2572 .11284 Mdot
.2572 .13522 Mdot
.2572 .15123 Mdot
.2572 .22505 Mdot
.25911 .1312 Mdot
.25911 .14823 Mdot
.25911 .22672 Mdot
.25962 .0625 Mdot
.25962 .08426 Mdot
.25962 .10597 Mdot
.25962 .22717 Mdot
.26052 .14483 Mdot
.26052 .15964 Mdot
.26052 .22795 Mdot
.26127 .12665 Mdot
.26127 .14483 Mdot
.26127 .22861 Mdot
.26247 .07431 Mdot
.26247 .14169 Mdot
.26247 .15736 Mdot
.26247 .22966 Mdot
.26375 .12144 Mdot
.26375 .23078 Mdot
.26466 .13817 Mdot
.26466 .15481 Mdot
.26466 .23158 Mdot
.26534 .16517 Mdot
.26534 .23217 Mdot
.26586 .23263 Mdot
.26662 .09429 Mdot
.26662 .11542 Mdot
.26662 .23329 Mdot
.26714 .13417 Mdot
.26714 .15192 Mdot
.26714 .23375 Mdot
.26752 .16325 Mdot
.26752 .23408 Mdot
.26997 .0625 Mdot
.26997 .08546 Mdot
.26997 .10838 Mdot
.26997 .12962 Mdot
.26997 .14863 Mdot
.26997 .16111 Mdot
.26997 .23622 Mdot
.27237 .16891 Mdot
.27237 .23832 Mdot
.27273 .15869 Mdot
.27273 .23864 Mdot
.27323 .12437 Mdot
.27323 .14483 Mdot
.27323 .23908 Mdot
.27394 .07503 Mdot
.27394 .10006 Mdot
.27394 .2397 Mdot
.27505 .16715 Mdot
.27505 .24067 Mdot
.27588 .15594 Mdot
.27588 .24139 Mdot
.27702 .11826 Mdot
.27702 .14042 Mdot
.27702 .17435 Mdot
.27702 .2424 Mdot
.27807 .16517 Mdot
.27807 .24331 Mdot
.27872 .0625 Mdot
.27872 .24388 Mdot
.27948 .15279 Mdot
.27948 .24455 Mdot
.27992 .17294 Mdot
.27992 .24493 Mdot
.2815 .11107 Mdot
.2815 .13522 Mdot
.2815 .16291 Mdot
.2815 .17959 Mdot
.2815 .19391 Mdot
.2815 .24631 Mdot
.28317 .17136 Mdot
.28317 .24777 Mdot
.28367 .14913 Mdot
.28367 .24821 Mdot
.28458 .19365 Mdot
.28458 .24901 Mdot
.28541 .16034 Mdot
.28541 .24973 Mdot
.28684 .10247 Mdot
.28684 .16956 Mdot
.28684 .25099 Mdot
.28804 .19334 Mdot
.28804 .25204 Mdot
.28857 .14483 Mdot
.28857 .2525 Mdot
.28906 .07002 Mdot
.28906 .08505 Mdot
.28906 .19867 Mdot
.28906 .25293 Mdot
.28993 .15736 Mdot
.28993 .25369 Mdot
.29104 .25466 Mdot
.29194 .193 Mdot
.29194 .25545 Mdot
.29271 .19867 Mdot
.29271 .25612 Mdot
.29336 .0625 Mdot
.29336 .07727 Mdot
.29336 .09202 Mdot
.29336 .10675 Mdot
.29336 .20349 Mdot
.29336 .25669 Mdot
.29441 .13973 Mdot
.29441 .2576 Mdot
.29521 .15389 Mdot
.29521 .25831 Mdot
.29586 .25887 Mdot
.29638 .19262 Mdot
.29638 .25933 Mdot
.29681 .19867 Mdot
.29681 .25971 Mdot
.29718 .0703 Mdot
.29718 .0859 Mdot
.29718 .10147 Mdot
.29718 .20377 Mdot
.29718 .26003 Mdot
.29749 .09874 Mdot
.29749 .1132 Mdot
.29749 .12761 Mdot
.29749 .20814 Mdot
.29749 .2603 Mdot
.30146 .0625 Mdot
.30146 .07904 Mdot
.30146 .09311 Mdot
.30146 .09556 Mdot
.30146 .12361 Mdot
.30146 .13356 Mdot
.30146 .14978 Mdot
.30146 .18425 Mdot
.30146 .19217 Mdot
.30146 .20408 Mdot
.30146 .20867 Mdot
.30146 .26378 Mdot
.30557 .11503 Mdot
.30557 .12996 Mdot
.30557 .21337 Mdot
.30557 .26737 Mdot
.3059 .07061 Mdot
.3059 .08681 Mdot
.3059 .103 Mdot
.3059 .11914 Mdot
.3059 .20927 Mdot
.3059 .26767 Mdot
.3063 .0713 Mdot
.3063 .20444 Mdot
.3063 .26801 Mdot
.30677 .19867 Mdot
.30677 .26842 Mdot
.30734 .19166 Mdot
.30734 .26892 Mdot
.30805 .18297 Mdot
.30805 .26955 Mdot
.30896 .27034 Mdot
.30951 .13608 Mdot
.30951 .15066 Mdot
.30951 .2179 Mdot
.30951 .27082 Mdot
.31016 .11014 Mdot
.31016 .12595 Mdot
.31016 .21424 Mdot
.31016 .27139 Mdot
.31091 .0625 Mdot
.31091 .07973 Mdot
.31091 .09693 Mdot
.31091 .1141 Mdot
.31091 .20994 Mdot
.31091 .27205 Mdot
.31181 .20484 Mdot
.31181 .27283 Mdot
.31289 .19867 Mdot
.31289 .27378 Mdot
.31331 .15627 Mdot
.31331 .22224 Mdot
.31331 .27414 Mdot
.31423 .13249 Mdot
.31423 .14791 Mdot
.31423 .19106 Mdot
.31423 .219 Mdot
.31423 .27495 Mdot
.31531 .10464 Mdot
.31531 .12144 Mdot
.31531 .2152 Mdot
.31531 .2759 Mdot
.31592 .18144 Mdot
.31592 .27643 Mdot
.31659 .07169 Mdot
.31659 .09005 Mdot
.31659 .10838 Mdot
.31659 .21071 Mdot
.31659 .27702 Mdot
.31814 .15389 Mdot
.31814 .2053 Mdot
.31814 .22357 Mdot
.31814 .27837 Mdot
.31952 .12847 Mdot
.31952 .14483 Mdot
.31952 .22023 Mdot
.31952 .27958 Mdot
.32003 .19867 Mdot
.32003 .28002 Mdot
.32115 .09843 Mdot
.32115 .11633 Mdot
.32115 .2163 Mdot
.32115 .281 Mdot
.32188 .15964 Mdot
.32188 .17435 Mdot
.32188 .22795 Mdot
.32188 .28165 Mdot
.32241 .19034 Mdot
.32241 .2821 Mdot
.3231 .0625 Mdot
.3231 .08219 Mdot
.3231 .21159 Mdot
.3231 .28271 Mdot
.32354 .15123 Mdot
.32354 .28309 Mdot
.32548 .12393 Mdot
.32548 .14136 Mdot
.32548 .17959 Mdot
.32548 .20584 Mdot
.32548 .2848 Mdot
.32738 .15736 Mdot
.32738 .17294 Mdot
.32738 .28646 Mdot
.3278 .09133 Mdot
.3278 .11051 Mdot
.3278 .28683 Mdot
.32845 .19867 Mdot
.32845 .2874 Mdot
.32961 .14823 Mdot
.32961 .28841 Mdot
.33062 .0731 Mdot
.33062 .28929 Mdot
.33107 .1785 Mdot
.33107 .28969 Mdot
.33226 .13741 Mdot
.33226 .18948 Mdot
.33226 .29073 Mdot
.33356 .17136 Mdot
.33356 .29186 Mdot
.33411 .29234 Mdot
.3346 .18384 Mdot
.3346 .29278 Mdot
.33546 .0625 Mdot
.33546 .08317 Mdot
.33546 .1038 Mdot
.33546 .29353 Mdot
.3365 .14483 Mdot
.3365 .16291 Mdot
.3365 .29444 Mdot
.33733 .17729 Mdot
.33733 .29516 Mdot
.33799 .0625 Mdot
.33799 .07727 Mdot
.33799 .09202 Mdot
.33799 .18898 Mdot
.33799 .25669 Mdot
.33799 .29574 Mdot
.33855 .29623 Mdot
.33941 .29698 Mdot
.34004 .13287 Mdot
.34004 .29754 Mdot
.34054 .16956 Mdot
.34054 .29797 Mdot
.34093 .18297 Mdot
.34093 .25612 Mdot
.34093 .29831 Mdot
.34125 .06975 Mdot
.34125 .08426 Mdot
.34125 .09874 Mdot
.34125 .1132 Mdot
.34125 .20814 Mdot
.34125 .2603 Mdot
.34125 .29859 Mdot
.34438 .0703 Mdot
.34438 .07367 Mdot
.34438 .096 Mdot
.34438 .11826 Mdot
.34438 .14095 Mdot
.34438 .16034 Mdot
.34438 .17591 Mdot
.34438 .18842 Mdot
.34438 .26003 Mdot
.34438 .30133 Mdot
.34768 .0625 Mdot
.34768 .07781 Mdot
.34768 .09311 Mdot
.34768 .10838 Mdot
.34768 .20867 Mdot
.34768 .26378 Mdot
.34768 .30422 Mdot
.34804 .18198 Mdot
.34804 .25971 Mdot
.34804 .30454 Mdot
.34849 .16752 Mdot
.34849 .30493 Mdot
.34906 .12761 Mdot
.34906 .30543 Mdot
.34981 .30609 Mdot
.35085 .10006 Mdot
.35085 .11503 Mdot
.35085 .12996 Mdot
.35085 .21337 Mdot
.35085 .26737 Mdot
.35085 .307 Mdot
.35154 .20408 Mdot
.35154 .26378 Mdot
.35154 .3076 Mdot
.35239 .17435 Mdot
.35239 .25933 Mdot
.35239 .30834 Mdot
.35347 .15736 Mdot
.35347 .30928 Mdot
.35389 .13608 Mdot
.35389 .15066 Mdot
.35389 .2179 Mdot
.35389 .27082 Mdot
.35389 .30966 Mdot
.35488 .0625 Mdot
.35488 .07061 Mdot
.35488 .08681 Mdot
.35488 .103 Mdot
.35488 .11107 Mdot
.35488 .20927 Mdot
.35488 .26767 Mdot
.35488 .31052 Mdot
.3561 .19867 Mdot
.3561 .26378 Mdot
.3561 .31158 Mdot
.35682 .14197 Mdot
.35682 .15627 Mdot
.35682 .22224 Mdot
.35682 .27414 Mdot
.35682 .31221 Mdot
.35763 .16517 Mdot
.35763 .25887 Mdot
.35763 .31292 Mdot
.35808 .11014 Mdot
.35808 .12595 Mdot
.35808 .21424 Mdot
.35808 .27139 Mdot
.35808 .31332 Mdot
.35962 .12144 Mdot
.35962 .20444 Mdot
.35962 .26801 Mdot
.35962 .31467 Mdot
.36114 .13249 Mdot
.36114 .14791 Mdot
.36114 .219 Mdot
.36114 .27495 Mdot
.36114 .31599 Mdot
.36156 .26378 Mdot
.36156 .31637 Mdot
.36232 .31703 Mdot
.36298 .0625 Mdot
.36298 .07973 Mdot
.36298 .09693 Mdot
.36298 .20994 Mdot
.36298 .27205 Mdot
.36298 .31761 Mdot
.36407 .15389 Mdot
.36407 .22357 Mdot
.36407 .25831 Mdot
.36407 .27837 Mdot
.36407 .31856 Mdot
.36529 .19867 Mdot
.36529 .26842 Mdot
.36529 .31962 Mdot
.36619 .10464 Mdot
.36619 .12144 Mdot
.36619 .2152 Mdot
.36619 .2759 Mdot
.36619 .32041 Mdot
.36688 .15964 Mdot
.36688 .17435 Mdot
.36688 .22795 Mdot
.36688 .28165 Mdot
.36688 .32102 Mdot
.36743 .07584 Mdot
.36743 .10247 Mdot
.36743 .3215 Mdot
.36824 .16243 Mdot
.36824 .26378 Mdot
.36824 .32221 Mdot
.36882 .20484 Mdot
.36882 .27283 Mdot
.36882 .32271 Mdot
.36924 .12847 Mdot
.36924 .14483 Mdot
.36924 .22023 Mdot
.36924 .27958 Mdot
.36924 .32309 Mdot
.36958 .17959 Mdot
.36958 .2848 Mdot
.36958 .32338 Mdot
.37217 .07169 Mdot
.37217 .09005 Mdot
.37217 .1141 Mdot
.37217 .13973 Mdot
.37217 .15123 Mdot
.37217 .19166 Mdot
.37217 .2576 Mdot
.37217 .26892 Mdot
.37217 .27702 Mdot
.37217 .28309 Mdot
.37217 .32565 Mdot
.37496 .17294 Mdot
.37496 .28646 Mdot
.37496 .32809 Mdot
.37535 .09843 Mdot
.37535 .11633 Mdot
.37535 .13417 Mdot
.37535 .2163 Mdot
.37535 .281 Mdot
.37535 .32843 Mdot
.37587 .19867 Mdot
.37587 .27378 Mdot
.37587 .32889 Mdot
.37658 .14978 Mdot
.37658 .26378 Mdot
.37658 .32951 Mdot
.37764 .1785 Mdot
.37764 .24901 Mdot
.37764 .28969 Mdot
.37764 .33044 Mdot
.37839 .14136 Mdot
.37839 .2848 Mdot
.37839 .33109 Mdot
.37936 .2053 Mdot
.37936 .27837 Mdot
.37936 .33194 Mdot
.38021 .18384 Mdot
.38021 .25293 Mdot
.38021 .29278 Mdot
.38021 .33268 Mdot
.38071 .18297 Mdot
.38071 .26955 Mdot
.38071 .33312 Mdot
.38127 .14823 Mdot
.38127 .16517 Mdot
.38127 .28841 Mdot
.38127 .33362 Mdot
.38267 .0625 Mdot
.38267 .07727 Mdot
.38267 .08219 Mdot
.38267 .09202 Mdot
.38267 .10184 Mdot
.38267 .10675 Mdot
.38267 .20349 Mdot
.38267 .25669 Mdot
.38267 .28271 Mdot
.38267 .29574 Mdot
.38267 .33483 Mdot
.38403 .17136 Mdot
.38403 .24777 Mdot
.38403 .29186 Mdot
.38403 .33603 Mdot
.38457 .19106 Mdot
.38457 .27495 Mdot
.38457 .3365 Mdot
.38503 .09874 Mdot
.38503 .1132 Mdot
.38503 .12761 Mdot
.38503 .20814 Mdot
.38503 .2603 Mdot
.38503 .29859 Mdot
.38503 .3369 Mdot
.3858 .11051 Mdot
.3858 .12962 Mdot
.3858 .28683 Mdot
.3858 .33757 Mdot
.38666 .17729 Mdot
.38666 .25204 Mdot
.38666 .29516 Mdot
.38666 .33833 Mdot
.3873 .13356 Mdot
.3873 .26378 Mdot
.3873 .33889 Mdot
.38818 .19867 Mdot
.38818 .28002 Mdot
.38818 .33966 Mdot
.38877 .13741 Mdot
.38877 .29073 Mdot
.38877 .34017 Mdot
.38918 .19867 Mdot
.38918 .25612 Mdot
.38918 .29831 Mdot
.38918 .34053 Mdot
.39158 .0703 Mdot
.39158 .0859 Mdot
.39158 .10147 Mdot
.39158 .16291 Mdot
.39158 .20377 Mdot
.39158 .24631 Mdot
.39158 .26003 Mdot
.39158 .27034 Mdot
.39158 .2848 Mdot
.39158 .29444 Mdot
.39158 .30133 Mdot
.39158 .34264 Mdot
.39389 .09311 Mdot
.39389 .10838 Mdot
.39389 .12361 Mdot
.39389 .20867 Mdot
.39389 .26378 Mdot
.39389 .30422 Mdot
.39389 .34465 Mdot
.39426 .16956 Mdot
.39426 .25099 Mdot
.39426 .29797 Mdot
.39426 .34498 Mdot
.39478 .0731 Mdot
.39478 .09429 Mdot
.39478 .28929 Mdot
.39478 .34544 Mdot
.39556 .18144 Mdot
.39556 .27643 Mdot
.39556 .34611 Mdot
.3961 .12996 Mdot
.3961 .14483 Mdot
.3961 .21337 Mdot
.3961 .26737 Mdot
.3961 .307 Mdot
.3961 .34659 Mdot
.39682 .193 Mdot
.39682 .25545 Mdot
.39682 .34721 Mdot
.3978 .12437 Mdot
.3978 .23908 Mdot
.3978 .34808 Mdot
.39822 .15066 Mdot
.39822 .27082 Mdot
.39822 .34844 Mdot
.39925 .19034 Mdot
.39925 .19867 Mdot
.39925 .25971 Mdot
.39925 .2821 Mdot
.39925 .34934 Mdot
.40026 .15627 Mdot
.40026 .17049 Mdot
.40026 .23622 Mdot
.40026 .27414 Mdot
.40026 .35022 Mdot
.40065 .15279 Mdot
.40065 .24455 Mdot
.40065 .35057 Mdot
.40157 .0625 Mdot
.40157 .07904 Mdot
.40157 .09556 Mdot
.40157 .20408 Mdot
.40157 .26378 Mdot
.40157 .35137 Mdot
.40269 .35235 Mdot
.40335 .16034 Mdot
.40335 .24973 Mdot
.40335 .35293 Mdot
.40378 .103 Mdot
.40378 .11914 Mdot
.40378 .20927 Mdot
.40378 .26767 Mdot
.40378 .35331 Mdot
.4059 .12595 Mdot
.4059 .14169 Mdot
.4059 .18626 Mdot
.4059 .25466 Mdot
.4059 .27139 Mdot
.4059 .35517 Mdot
.40793 .14791 Mdot
.40793 .16325 Mdot
.40793 .23408 Mdot
.40793 .35694 Mdot
.40833 .19262 Mdot
.40833 .25933 Mdot
.40833 .35729 Mdot
.40891 .0625 Mdot
.40891 .08546 Mdot
.40891 .10838 Mdot
.40891 .23622 Mdot
.40891 .3578 Mdot
.40987 .16891 Mdot
.40987 .23832 Mdot
.40987 .35864 Mdot
.41063 .19867 Mdot
.41063 .26378 Mdot
.41063 .3593 Mdot
.41174 .11826 Mdot
.41174 .14042 Mdot
.41174 .17435 Mdot
.41174 .2424 Mdot
.41174 .36027 Mdot
.41282 .0713 Mdot
.41282 .08888 Mdot
.41282 .10643 Mdot
.41282 .36122 Mdot
.41353 .06975 Mdot
.41353 .19391 Mdot
.41353 .24631 Mdot
.41353 .36183 Mdot
.41439 .14913 Mdot
.41439 .24821 Mdot
.41439 .36259 Mdot
.4149 .1141 Mdot
.4149 .1312 Mdot
.4149 .22672 Mdot
.4149 .36304 Mdot
.41689 .13817 Mdot
.41689 .15481 Mdot
.41689 .17811 Mdot
.41689 .23158 Mdot
.41689 .25369 Mdot
.41689 .36478 Mdot
.41879 .16111 Mdot
.41879 .23622 Mdot
.41879 .36644 Mdot
.41925 .18533 Mdot
.41925 .36684 Mdot
.42001 .22505 Mdot
.42001 .36751 Mdot
.4206 .16715 Mdot
.4206 .18297 Mdot
.4206 .24067 Mdot
.4206 .36803 Mdot
.42148 .19217 Mdot
.42148 .3688 Mdot
.42234 .18842 Mdot
.42234 .24493 Mdot
.42234 .36955 Mdot
.4229 .23263 Mdot
.4229 .37004 Mdot
.42359 .37064 Mdot
.424 .0625 Mdot
.424 .07781 Mdot
.424 .19365 Mdot
.424 .371 Mdot
.42559 .0625 Mdot
.42559 .07002 Mdot
.42559 .07503 Mdot
.42559 .08129 Mdot
.42559 .08505 Mdot
.42559 .10006 Mdot
.42559 .11503 Mdot
.42559 .11877 Mdot
.42559 .19867 Mdot
.42559 .22322 Mdot
.42559 .2397 Mdot
.42559 .307 Mdot
.42559 .37239 Mdot
.42712 .12144 Mdot
.42712 .13608 Mdot
.42712 .2179 Mdot
.42712 .30966 Mdot
.42712 .37373 Mdot
.42749 .12665 Mdot
.42749 .14483 Mdot
.42749 .22861 Mdot
.42749 .37405 Mdot
.4281 .13522 Mdot
.4281 .37459 Mdot
.42858 .14197 Mdot
.42858 .22224 Mdot
.42858 .31221 Mdot
.42858 .37501 Mdot
.42929 .15192 Mdot
.42929 .23375 Mdot
.42929 .37563 Mdot
.43044 .16806 Mdot
.43044 .37664 Mdot
.43101 .15869 Mdot
.43101 .17591 Mdot
.43101 .23864 Mdot
.43101 .37713 Mdot
.43264 .17639 Mdot
.43264 .18198 Mdot
.43264 .30454 Mdot
.43264 .37856 Mdot
.4342 .18779 Mdot
.4342 .3076 Mdot
.4342 .37993 Mdot
.43471 .38037 Mdot
.43569 .07061 Mdot
.43569 .08681 Mdot
.43569 .31052 Mdot
.43569 .38123 Mdot
.43665 .2126 Mdot
.43665 .38207 Mdot
.43712 .0784 Mdot
.43712 .09429 Mdot
.43712 .11014 Mdot
.43712 .12595 Mdot
.43712 .21424 Mdot
.43712 .27139 Mdot
.43712 .31332 Mdot
.43712 .38248 Mdot
.43848 .0625 Mdot
.43848 .08317 Mdot
.43848 .13249 Mdot
.43848 .219 Mdot
.43848 .27495 Mdot
.43848 .31599 Mdot
.43848 .38367 Mdot
.43978 .15389 Mdot
.43978 .22357 Mdot
.43978 .27837 Mdot
.43978 .31856 Mdot
.43978 .38481 Mdot
.4402 .09274 Mdot
.4402 .11284 Mdot
.4402 .22505 Mdot
.4402 .38518 Mdot
.44103 .15964 Mdot
.44103 .17435 Mdot
.44103 .22795 Mdot
.44103 .28165 Mdot
.44103 .32102 Mdot
.44103 .3859 Mdot
.44184 .14095 Mdot
.44184 .30133 Mdot
.44184 .38661 Mdot
.44223 .17959 Mdot
.44223 .2848 Mdot
.44223 .32338 Mdot
.44223 .38695 Mdot
.44338 .16752 Mdot
.44338 .30493 Mdot
.44338 .38796 Mdot
.44485 .17435 Mdot
.44485 .30834 Mdot
.44485 .38924 Mdot
.44555 .11752 Mdot
.44555 .38986 Mdot
.44624 .18087 Mdot
.44624 .26378 Mdot
.44624 .31158 Mdot
.44624 .39046 Mdot
.44756 .15535 Mdot
.44756 .20444 Mdot
.44756 .26801 Mdot
.44756 .31467 Mdot
.44756 .39162 Mdot
.44882 .0625 Mdot
.44882 .07973 Mdot
.44882 .09693 Mdot
.44882 .20994 Mdot
.44882 .27205 Mdot
.44882 .31761 Mdot
.44882 .39272 Mdot
.44943 .16517 Mdot
.44943 .39325 Mdot
.45003 .10464 Mdot
.45003 .12144 Mdot
.45003 .2152 Mdot
.45003 .2759 Mdot
.45003 .32041 Mdot
.45003 .39377 Mdot
.45117 .14483 Mdot
.45117 .19867 Mdot
.45117 .27958 Mdot
.45117 .32309 Mdot
.45117 .39478 Mdot
.45227 .15123 Mdot
.45227 .16715 Mdot
.45227 .28309 Mdot
.45227 .32565 Mdot
.45227 .39573 Mdot
.4528 .20647 Mdot
.4528 .3962 Mdot
.45331 .17294 Mdot
.45331 .24493 Mdot
.45331 .28646 Mdot
.45331 .32809 Mdot
.45331 .39665 Mdot
.45432 .1785 Mdot
.45432 .21379 Mdot
.45432 .24901 Mdot
.45432 .28969 Mdot
.45432 .29698 Mdot
.45432 .33044 Mdot
.45432 .39753 Mdot
.45528 .07002 Mdot
.45528 .08505 Mdot
.45528 .19867 Mdot
.45528 .25293 Mdot
.45528 .33268 Mdot
.45528 .39837 Mdot
.45574 .07367 Mdot
.45574 .096 Mdot
.45574 .30133 Mdot
.45574 .39877 Mdot
.4562 .0625 Mdot
.4562 .07727 Mdot
.4562 .09202 Mdot
.4562 .10675 Mdot
.4562 .12144 Mdot
.4562 .20349 Mdot
.4562 .25669 Mdot
.4562 .33483 Mdot
.4562 .39917 Mdot
.45708 .1132 Mdot
.45708 .12761 Mdot
.45708 .20814 Mdot
.45708 .2603 Mdot
.45708 .30543 Mdot
.45708 .3369 Mdot
.45708 .39994 Mdot
.45834 .15736 Mdot
.45834 .30928 Mdot
.45834 .40105 Mdot
.45953 .16517 Mdot
.45953 .25887 Mdot
.45953 .31292 Mdot
.45953 .40209 Mdot
.46065 .19217 Mdot
.46065 .26378 Mdot
.46065 .31637 Mdot
.46065 .40307 Mdot
.46171 .19867 Mdot
.46171 .26842 Mdot
.46171 .31962 Mdot
.46171 .404 Mdot
.46272 .0625 Mdot
.46272 .20484 Mdot
.46272 .27283 Mdot
.46272 .32271 Mdot
.46272 .40488 Mdot
.46367 .07169 Mdot
.46367 .09005 Mdot
.46367 .10838 Mdot
.46367 .27702 Mdot
.46367 .32565 Mdot
.46367 .40571 Mdot
.46458 .13417 Mdot
.46458 .23375 Mdot
.46458 .32843 Mdot
.46458 .40651 Mdot
.46544 .15869 Mdot
.46544 .23864 Mdot
.46544 .40726 Mdot
.46626 .16517 Mdot
.46626 .24331 Mdot
.46626 .40798 Mdot
.46705 .18779 Mdot
.46705 .24777 Mdot
.46705 .40867 Mdot
.4678 .19334 Mdot
.4678 .25204 Mdot
.4678 .40932 Mdot
.46851 .0625 Mdot
.46851 .0784 Mdot
.46851 .09429 Mdot
.46851 .19867 Mdot
.46851 .40995 Mdot
.46919 .10147 Mdot
.46919 .11701 Mdot
.46919 .13249 Mdot
.46919 .41055 Mdot
.46985 .13878 Mdot
.46985 .22357 Mdot
.46985 .41112 Mdot
.47048 .14483 Mdot
.47048 .15964 Mdot
.47048 .22795 Mdot
.47048 .3859 Mdot
.47048 .41167 Mdot
.47108 .16517 Mdot
.47108 .23217 Mdot
.47108 .2848 Mdot
.47108 .32338 Mdot
.47108 .38695 Mdot
.47108 .41219 Mdot
.47221 .18948 Mdot
.47221 .29073 Mdot
.47221 .41319 Mdot
.47326 .29623 Mdot
.47326 .4141 Mdot
.47423 .30133 Mdot
.47423 .41495 Mdot
.47514 .30609 Mdot
.47514 .41574 Mdot
.47598 .08681 Mdot
.47598 .11107 Mdot
.47598 .24631 Mdot
.47598 .31052 Mdot
.47598 .41648 Mdot
.47677 .14483 Mdot
.47677 .2525 Mdot
.47677 .41717 Mdot
.4775 .25831 Mdot
.4775 .41781 Mdot
.47819 .18425 Mdot
.47819 .26378 Mdot
.47819 .41842 Mdot
.47884 .19166 Mdot
.47884 .41899 Mdot
.47946 .41953 Mdot
.48004 .07258 Mdot
.48004 .09274 Mdot
.48004 .22505 Mdot
.48004 .42003 Mdot
.48058 .12144 Mdot
.48058 .23078 Mdot
.48058 .42051 Mdot
.4811 .14863 Mdot
.4811 .23622 Mdot
.4811 .42096 Mdot
.48159 .17435 Mdot
.48159 .42139 Mdot
.48206 .18087 Mdot
.48206 .39046 Mdot
.48206 .4218 Mdot
.4825 .18708 Mdot
.4825 .39162 Mdot
.4825 .42219 Mdot
.48292 .0625 Mdot
.48292 .07973 Mdot
.48292 .20994 Mdot
.48292 .31761 Mdot
.48292 .39272 Mdot
.48292 .42256 Mdot
.48332 .07094 Mdot
.48332 .08781 Mdot
.48332 .10464 Mdot
.48332 .12144 Mdot
.48332 .2152 Mdot
.48332 .2759 Mdot
.48332 .32041 Mdot
.48332 .39377 Mdot
.48332 .42291 Mdot
.48371 .12847 Mdot
.48371 .14483 Mdot
.48371 .22023 Mdot
.48371 .27958 Mdot
.48371 .32309 Mdot
.48371 .39478 Mdot
.48371 .42324 Mdot
.48407 .15123 Mdot
.48407 .22505 Mdot
.48407 .28309 Mdot
.48407 .32565 Mdot
.48407 .39573 Mdot
.48407 .42356 Mdot
.48442 .15736 Mdot
.48442 .17294 Mdot
.48442 .28646 Mdot
.48442 .32809 Mdot
.48442 .39665 Mdot
.48442 .42387 Mdot
.48476 .1785 Mdot
.48476 .28969 Mdot
.48476 .33044 Mdot
.48476 .39753 Mdot
.48476 .42416 Mdot
.48508 .18384 Mdot
.48508 .25293 Mdot
.48508 .29278 Mdot
.48508 .33268 Mdot
.48508 .39837 Mdot
.48508 .42444 Mdot
.48539 .0625 Mdot
.48539 .07727 Mdot
.48539 .09202 Mdot
.48539 .10675 Mdot
.48539 .20349 Mdot
.48539 .25669 Mdot
.48539 .29574 Mdot
.48539 .33483 Mdot
.48539 .39917 Mdot
.48539 .42471 Mdot
.48568 .09874 Mdot
.48568 .1132 Mdot
.48568 .12761 Mdot
.48568 .20814 Mdot
.48568 .2603 Mdot
.48568 .29859 Mdot
.48568 .3369 Mdot
.48568 .39994 Mdot
.48568 .42497 Mdot
.5 .09005 Mdot
.5 .11633 Mdot
.5 .16517 Mdot
.5 .27283 Mdot
.5 .281 Mdot
.5 .4375 Mdot
.51432 .09874 Mdot
.51432 .1132 Mdot
.51432 .12761 Mdot
.51432 .20814 Mdot
.51432 .2603 Mdot
.51432 .29859 Mdot
.51432 .3369 Mdot
.51432 .39994 Mdot
.51432 .42497 Mdot
.51461 .0625 Mdot
.51461 .07727 Mdot
.51461 .09202 Mdot
.51461 .10675 Mdot
.51461 .20349 Mdot
.51461 .25669 Mdot
.51461 .29574 Mdot
.51461 .33483 Mdot
.51461 .39917 Mdot
.51461 .42471 Mdot
.51492 .18384 Mdot
.51492 .25293 Mdot
.51492 .29278 Mdot
.51492 .33268 Mdot
.51492 .39837 Mdot
.51492 .42444 Mdot
.51524 .1785 Mdot
.51524 .28969 Mdot
.51524 .33044 Mdot
.51524 .39753 Mdot
.51524 .42416 Mdot
.51558 .15736 Mdot
.51558 .17294 Mdot
.51558 .28646 Mdot
.51558 .32809 Mdot
.51558 .39665 Mdot
.51558 .42387 Mdot
.51593 .15123 Mdot
.51593 .22505 Mdot
.51593 .28309 Mdot
.51593 .32565 Mdot
.51593 .39573 Mdot
.51593 .42356 Mdot
.51629 .12847 Mdot
.51629 .14483 Mdot
.51629 .22023 Mdot
.51629 .27958 Mdot
.51629 .32309 Mdot
.51629 .39478 Mdot
.51629 .42324 Mdot
.51668 .07094 Mdot
.51668 .08781 Mdot
.51668 .10464 Mdot
.51668 .12144 Mdot
.51668 .2152 Mdot
.51668 .2759 Mdot
.51668 .32041 Mdot
.51668 .39377 Mdot
.51668 .42291 Mdot
.51708 .0625 Mdot
.51708 .07973 Mdot
.51708 .20994 Mdot
.51708 .31761 Mdot
.51708 .39272 Mdot
.51708 .42256 Mdot
.5175 .18708 Mdot
.5175 .39162 Mdot
.5175 .42219 Mdot
.51794 .18087 Mdot
.51794 .39046 Mdot
.51794 .4218 Mdot
.51841 .17435 Mdot
.51841 .42139 Mdot
.5189 .14863 Mdot
.5189 .23622 Mdot
.5189 .42096 Mdot
.51942 .12144 Mdot
.51942 .23078 Mdot
.51942 .42051 Mdot
.51996 .07258 Mdot
.51996 .09274 Mdot
.51996 .22505 Mdot
.51996 .42003 Mdot
.52054 .41953 Mdot
.52116 .19166 Mdot
.52116 .41899 Mdot
.52181 .18425 Mdot
.52181 .26378 Mdot
.52181 .41842 Mdot
.5225 .25831 Mdot
.5225 .41781 Mdot
.52323 .14483 Mdot
.52323 .2525 Mdot
.52323 .41717 Mdot
.52402 .08681 Mdot
.52402 .11107 Mdot
.52402 .24631 Mdot
.52402 .31052 Mdot
.52402 .41648 Mdot
.52486 .30609 Mdot
.52486 .41574 Mdot
.52577 .30133 Mdot
.52577 .41495 Mdot
.52674 .29623 Mdot
.52674 .4141 Mdot
.52779 .18948 Mdot
.52779 .29073 Mdot
.52779 .41319 Mdot
.52892 .16517 Mdot
.52892 .23217 Mdot
.52892 .2848 Mdot
.52892 .32338 Mdot
.52892 .38695 Mdot
.52892 .41219 Mdot
.52952 .14483 Mdot
.52952 .15964 Mdot
.52952 .22795 Mdot
.52952 .3859 Mdot
.52952 .41167 Mdot
.53015 .13878 Mdot
.53015 .22357 Mdot
.53015 .41112 Mdot
.53081 .10147 Mdot
.53081 .11701 Mdot
.53081 .13249 Mdot
.53081 .41055 Mdot
.53149 .0625 Mdot
.53149 .0784 Mdot
.53149 .09429 Mdot
.53149 .19867 Mdot
.53149 .40995 Mdot
.5322 .19334 Mdot
.5322 .25204 Mdot
.5322 .40932 Mdot
.53295 .18779 Mdot
.53295 .24777 Mdot
.53295 .40867 Mdot
.53374 .16517 Mdot
.53374 .24331 Mdot
.53374 .40798 Mdot
.53456 .15869 Mdot
.53456 .23864 Mdot
.53456 .40726 Mdot
.53542 .13417 Mdot
.53542 .23375 Mdot
.53542 .32843 Mdot
.53542 .40651 Mdot
.53633 .07169 Mdot
.53633 .09005 Mdot
.53633 .10838 Mdot
.53633 .27702 Mdot
.53633 .32565 Mdot
.53633 .40571 Mdot
.53728 .0625 Mdot
.53728 .20484 Mdot
.53728 .27283 Mdot
.53728 .32271 Mdot
.53728 .40488 Mdot
.53829 .19867 Mdot
.53829 .26842 Mdot
.53829 .31962 Mdot
.53829 .404 Mdot
.53935 .19217 Mdot
.53935 .26378 Mdot
.53935 .31637 Mdot
.53935 .40307 Mdot
.54047 .16517 Mdot
.54047 .25887 Mdot
.54047 .31292 Mdot
.54047 .40209 Mdot
.54166 .15736 Mdot
.54166 .30928 Mdot
.54166 .40105 Mdot
.54292 .1132 Mdot
.54292 .12761 Mdot
.54292 .20814 Mdot
.54292 .2603 Mdot
.54292 .30543 Mdot
.54292 .3369 Mdot
.54292 .39994 Mdot
.5438 .0625 Mdot
.5438 .07727 Mdot
.5438 .09202 Mdot
.5438 .10675 Mdot
.5438 .12144 Mdot
.5438 .20349 Mdot
.5438 .25669 Mdot
.5438 .33483 Mdot
.5438 .39917 Mdot
.54426 .07367 Mdot
.54426 .096 Mdot
.54426 .30133 Mdot
.54426 .39877 Mdot
.54472 .07002 Mdot
.54472 .08505 Mdot
.54472 .19867 Mdot
.54472 .25293 Mdot
.54472 .33268 Mdot
.54472 .39837 Mdot
.54568 .1785 Mdot
.54568 .21379 Mdot
.54568 .24901 Mdot
.54568 .28969 Mdot
.54568 .29698 Mdot
.54568 .33044 Mdot
.54568 .39753 Mdot
.54669 .17294 Mdot
.54669 .24493 Mdot
.54669 .28646 Mdot
.54669 .32809 Mdot
.54669 .39665 Mdot
.5472 .20647 Mdot
.5472 .3962 Mdot
.54773 .15123 Mdot
.54773 .16715 Mdot
.54773 .28309 Mdot
.54773 .32565 Mdot
.54773 .39573 Mdot
.54883 .14483 Mdot
.54883 .19867 Mdot
.54883 .27958 Mdot
.54883 .32309 Mdot
.54883 .39478 Mdot
.54997 .10464 Mdot
.54997 .12144 Mdot
.54997 .2152 Mdot
.54997 .2759 Mdot
.54997 .32041 Mdot
.54997 .39377 Mdot
.55057 .16517 Mdot
.55057 .39325 Mdot
.55118 .0625 Mdot
.55118 .07973 Mdot
.55118 .09693 Mdot
.55118 .20994 Mdot
.55118 .27205 Mdot
.55118 .31761 Mdot
.55118 .39272 Mdot
.55244 .15535 Mdot
.55244 .20444 Mdot
.55244 .26801 Mdot
.55244 .31467 Mdot
.55244 .39162 Mdot
.55376 .18087 Mdot
.55376 .26378 Mdot
.55376 .31158 Mdot
.55376 .39046 Mdot
.55445 .11752 Mdot
.55445 .38986 Mdot
.55515 .17435 Mdot
.55515 .30834 Mdot
.55515 .38924 Mdot
.55662 .16752 Mdot
.55662 .30493 Mdot
.55662 .38796 Mdot
.55777 .17959 Mdot
.55777 .2848 Mdot
.55777 .32338 Mdot
.55777 .38695 Mdot
.55816 .14095 Mdot
.55816 .30133 Mdot
.55816 .38661 Mdot
.55897 .15964 Mdot
.55897 .17435 Mdot
.55897 .22795 Mdot
.55897 .28165 Mdot
.55897 .32102 Mdot
.55897 .3859 Mdot
.5598 .09274 Mdot
.5598 .11284 Mdot
.5598 .22505 Mdot
.5598 .38518 Mdot
.56022 .15389 Mdot
.56022 .22357 Mdot
.56022 .27837 Mdot
.56022 .31856 Mdot
.56022 .38481 Mdot
.56152 .0625 Mdot
.56152 .08317 Mdot
.56152 .13249 Mdot
.56152 .219 Mdot
.56152 .27495 Mdot
.56152 .31599 Mdot
.56152 .38367 Mdot
.56288 .0784 Mdot
.56288 .09429 Mdot
.56288 .11014 Mdot
.56288 .12595 Mdot
.56288 .21424 Mdot
.56288 .27139 Mdot
.56288 .31332 Mdot
.56288 .38248 Mdot
.56335 .2126 Mdot
.56335 .38207 Mdot
.56431 .07061 Mdot
.56431 .08681 Mdot
.56431 .31052 Mdot
.56431 .38123 Mdot
.56529 .38037 Mdot
.5658 .18779 Mdot
.5658 .3076 Mdot
.5658 .37993 Mdot
.56736 .17639 Mdot
.56736 .18198 Mdot
.56736 .30454 Mdot
.56736 .37856 Mdot
.56899 .15869 Mdot
.56899 .17591 Mdot
.56899 .23864 Mdot
.56899 .37713 Mdot
.56956 .16806 Mdot
.56956 .37664 Mdot
.57071 .15192 Mdot
.57071 .23375 Mdot
.57071 .37563 Mdot
.57142 .14197 Mdot
.57142 .22224 Mdot
.57142 .31221 Mdot
.57142 .37501 Mdot
.5719 .13522 Mdot
.5719 .37459 Mdot
.57251 .12665 Mdot
.57251 .14483 Mdot
.57251 .22861 Mdot
.57251 .37405 Mdot
.57288 .12144 Mdot
.57288 .13608 Mdot
.57288 .2179 Mdot
.57288 .30966 Mdot
.57288 .37373 Mdot
.57441 .0625 Mdot
.57441 .07002 Mdot
.57441 .07503 Mdot
.57441 .08129 Mdot
.57441 .08505 Mdot
.57441 .10006 Mdot
.57441 .11503 Mdot
.57441 .11877 Mdot
.57441 .19867 Mdot
.57441 .22322 Mdot
.57441 .2397 Mdot
.57441 .307 Mdot
.57441 .37239 Mdot
.576 .0625 Mdot
.576 .07781 Mdot
.576 .19365 Mdot
.576 .371 Mdot
.57641 .37064 Mdot
.5771 .23263 Mdot
.5771 .37004 Mdot
.57766 .18842 Mdot
.57766 .24493 Mdot
.57766 .36955 Mdot
.57852 .19217 Mdot
.57852 .3688 Mdot
.5794 .16715 Mdot
.5794 .18297 Mdot
.5794 .24067 Mdot
.5794 .36803 Mdot
.57999 .22505 Mdot
.57999 .36751 Mdot
.58075 .18533 Mdot
.58075 .36684 Mdot
.58121 .16111 Mdot
.58121 .23622 Mdot
.58121 .36644 Mdot
.58311 .13817 Mdot
.58311 .15481 Mdot
.58311 .17811 Mdot
.58311 .23158 Mdot
.58311 .25369 Mdot
.58311 .36478 Mdot
.5851 .1141 Mdot
.5851 .1312 Mdot
.5851 .22672 Mdot
.5851 .36304 Mdot
.58561 .14913 Mdot
.58561 .24821 Mdot
.58561 .36259 Mdot
.58647 .06975 Mdot
.58647 .19391 Mdot
.58647 .24631 Mdot
.58647 .36183 Mdot
.58718 .0713 Mdot
.58718 .08888 Mdot
.58718 .10643 Mdot
.58718 .36122 Mdot
.58826 .11826 Mdot
.58826 .14042 Mdot
.58826 .17435 Mdot
.58826 .2424 Mdot
.58826 .36027 Mdot
.58937 .19867 Mdot
.58937 .26378 Mdot
.58937 .3593 Mdot
.59013 .16891 Mdot
.59013 .23832 Mdot
.59013 .35864 Mdot
.59109 .0625 Mdot
.59109 .08546 Mdot
.59109 .10838 Mdot
.59109 .23622 Mdot
.59109 .3578 Mdot
.59167 .19262 Mdot
.59167 .25933 Mdot
.59167 .35729 Mdot
.59207 .14791 Mdot
.59207 .16325 Mdot
.59207 .23408 Mdot
.59207 .35694 Mdot
.5941 .12595 Mdot
.5941 .14169 Mdot
.5941 .18626 Mdot
.5941 .25466 Mdot
.5941 .27139 Mdot
.5941 .35517 Mdot
.59622 .103 Mdot
.59622 .11914 Mdot
.59622 .20927 Mdot
.59622 .26767 Mdot
.59622 .35331 Mdot
.59665 .16034 Mdot
.59665 .24973 Mdot
.59665 .35293 Mdot
.59731 .35235 Mdot
.59843 .0625 Mdot
.59843 .07904 Mdot
.59843 .09556 Mdot
.59843 .20408 Mdot
.59843 .26378 Mdot
.59843 .35137 Mdot
.59935 .15279 Mdot
.59935 .24455 Mdot
.59935 .35057 Mdot
.59974 .15627 Mdot
.59974 .17049 Mdot
.59974 .23622 Mdot
.59974 .27414 Mdot
.59974 .35022 Mdot
.60075 .19034 Mdot
.60075 .19867 Mdot
.60075 .25971 Mdot
.60075 .2821 Mdot
.60075 .34934 Mdot
.60178 .15066 Mdot
.60178 .27082 Mdot
.60178 .34844 Mdot
.6022 .12437 Mdot
.6022 .23908 Mdot
.6022 .34808 Mdot
.60318 .193 Mdot
.60318 .25545 Mdot
.60318 .34721 Mdot
.6039 .12996 Mdot
.6039 .14483 Mdot
.6039 .21337 Mdot
.6039 .26737 Mdot
.6039 .307 Mdot
.6039 .34659 Mdot
.60444 .18144 Mdot
.60444 .27643 Mdot
.60444 .34611 Mdot
.60522 .0731 Mdot
.60522 .09429 Mdot
.60522 .28929 Mdot
.60522 .34544 Mdot
.60574 .16956 Mdot
.60574 .25099 Mdot
.60574 .29797 Mdot
.60574 .34498 Mdot
.60611 .09311 Mdot
.60611 .10838 Mdot
.60611 .12361 Mdot
.60611 .20867 Mdot
.60611 .26378 Mdot
.60611 .30422 Mdot
.60611 .34465 Mdot
.60842 .0703 Mdot
.60842 .0859 Mdot
.60842 .10147 Mdot
.60842 .16291 Mdot
.60842 .20377 Mdot
.60842 .24631 Mdot
.60842 .26003 Mdot
.60842 .27034 Mdot
.60842 .2848 Mdot
.60842 .29444 Mdot
.60842 .30133 Mdot
.60842 .34264 Mdot
.61082 .19867 Mdot
.61082 .25612 Mdot
.61082 .29831 Mdot
.61082 .34053 Mdot
.61123 .13741 Mdot
.61123 .29073 Mdot
.61123 .34017 Mdot
.61182 .19867 Mdot
.61182 .28002 Mdot
.61182 .33966 Mdot
.6127 .13356 Mdot
.6127 .26378 Mdot
.6127 .33889 Mdot
.61334 .17729 Mdot
.61334 .25204 Mdot
.61334 .29516 Mdot
.61334 .33833 Mdot
.6142 .11051 Mdot
.6142 .12962 Mdot
.6142 .28683 Mdot
.6142 .33757 Mdot
.61497 .09874 Mdot
.61497 .1132 Mdot
.61497 .12761 Mdot
.61497 .20814 Mdot
.61497 .2603 Mdot
.61497 .29859 Mdot
.61497 .3369 Mdot
.61543 .19106 Mdot
.61543 .27495 Mdot
.61543 .3365 Mdot
.61597 .17136 Mdot
.61597 .24777 Mdot
.61597 .29186 Mdot
.61597 .33603 Mdot
.61733 .0625 Mdot
.61733 .07727 Mdot
.61733 .08219 Mdot
.61733 .09202 Mdot
.61733 .10184 Mdot
.61733 .10675 Mdot
.61733 .20349 Mdot
.61733 .25669 Mdot
.61733 .28271 Mdot
.61733 .29574 Mdot
.61733 .33483 Mdot
.61873 .14823 Mdot
.61873 .16517 Mdot
.61873 .28841 Mdot
.61873 .33362 Mdot
.61929 .18297 Mdot
.61929 .26955 Mdot
.61929 .33312 Mdot
.61979 .18384 Mdot
.61979 .25293 Mdot
.61979 .29278 Mdot
.61979 .33268 Mdot
.62064 .2053 Mdot
.62064 .27837 Mdot
.62064 .33194 Mdot
.62161 .14136 Mdot
.62161 .2848 Mdot
.62161 .33109 Mdot
.62236 .1785 Mdot
.62236 .24901 Mdot
.62236 .28969 Mdot
.62236 .33044 Mdot
.62342 .14978 Mdot
.62342 .26378 Mdot
.62342 .32951 Mdot
.62413 .19867 Mdot
.62413 .27378 Mdot
.62413 .32889 Mdot
.62465 .09843 Mdot
.62465 .11633 Mdot
.62465 .13417 Mdot
.62465 .2163 Mdot
.62465 .281 Mdot
.62465 .32843 Mdot
.62504 .17294 Mdot
.62504 .28646 Mdot
.62504 .32809 Mdot
.62783 .07169 Mdot
.62783 .09005 Mdot
.62783 .1141 Mdot
.62783 .13973 Mdot
.62783 .15123 Mdot
.62783 .19166 Mdot
.62783 .2576 Mdot
.62783 .26892 Mdot
.62783 .27702 Mdot
.62783 .28309 Mdot
.62783 .32565 Mdot
.63042 .17959 Mdot
.63042 .2848 Mdot
.63042 .32338 Mdot
.63076 .12847 Mdot
.63076 .14483 Mdot
.63076 .22023 Mdot
.63076 .27958 Mdot
.63076 .32309 Mdot
.63118 .20484 Mdot
.63118 .27283 Mdot
.63118 .32271 Mdot
.63176 .16243 Mdot
.63176 .26378 Mdot
.63176 .32221 Mdot
.63257 .07584 Mdot
.63257 .10247 Mdot
.63257 .3215 Mdot
.63312 .15964 Mdot
.63312 .17435 Mdot
.63312 .22795 Mdot
.63312 .28165 Mdot
.63312 .32102 Mdot
.63381 .10464 Mdot
.63381 .12144 Mdot
.63381 .2152 Mdot
.63381 .2759 Mdot
.63381 .32041 Mdot
.63471 .19867 Mdot
.63471 .26842 Mdot
.63471 .31962 Mdot
.63593 .15389 Mdot
.63593 .22357 Mdot
.63593 .25831 Mdot
.63593 .27837 Mdot
.63593 .31856 Mdot
.63702 .0625 Mdot
.63702 .07973 Mdot
.63702 .09693 Mdot
.63702 .20994 Mdot
.63702 .27205 Mdot
.63702 .31761 Mdot
.63768 .31703 Mdot
.63844 .26378 Mdot
.63844 .31637 Mdot
.63886 .13249 Mdot
.63886 .14791 Mdot
.63886 .219 Mdot
.63886 .27495 Mdot
.63886 .31599 Mdot
.64038 .12144 Mdot
.64038 .20444 Mdot
.64038 .26801 Mdot
.64038 .31467 Mdot
.64192 .11014 Mdot
.64192 .12595 Mdot
.64192 .21424 Mdot
.64192 .27139 Mdot
.64192 .31332 Mdot
.64237 .16517 Mdot
.64237 .25887 Mdot
.64237 .31292 Mdot
.64318 .14197 Mdot
.64318 .15627 Mdot
.64318 .22224 Mdot
.64318 .27414 Mdot
.64318 .31221 Mdot
.6439 .19867 Mdot
.6439 .26378 Mdot
.6439 .31158 Mdot
.64512 .0625 Mdot
.64512 .07061 Mdot
.64512 .08681 Mdot
.64512 .103 Mdot
.64512 .11107 Mdot
.64512 .20927 Mdot
.64512 .26767 Mdot
.64512 .31052 Mdot
.64611 .13608 Mdot
.64611 .15066 Mdot
.64611 .2179 Mdot
.64611 .27082 Mdot
.64611 .30966 Mdot
.64653 .15736 Mdot
.64653 .30928 Mdot
.64761 .17435 Mdot
.64761 .25933 Mdot
.64761 .30834 Mdot
.64846 .20408 Mdot
.64846 .26378 Mdot
.64846 .3076 Mdot
.64915 .10006 Mdot
.64915 .11503 Mdot
.64915 .12996 Mdot
.64915 .21337 Mdot
.64915 .26737 Mdot
.64915 .307 Mdot
.65019 .30609 Mdot
.65094 .12761 Mdot
.65094 .30543 Mdot
.65151 .16752 Mdot
.65151 .30493 Mdot
.65196 .18198 Mdot
.65196 .25971 Mdot
.65196 .30454 Mdot
.65232 .0625 Mdot
.65232 .07781 Mdot
.65232 .09311 Mdot
.65232 .10838 Mdot
.65232 .20867 Mdot
.65232 .26378 Mdot
.65232 .30422 Mdot
.65562 .0703 Mdot
.65562 .07367 Mdot
.65562 .096 Mdot
.65562 .11826 Mdot
.65562 .14095 Mdot
.65562 .16034 Mdot
.65562 .17591 Mdot
.65562 .18842 Mdot
.65562 .26003 Mdot
.65562 .30133 Mdot
.65875 .06975 Mdot
.65875 .08426 Mdot
.65875 .09874 Mdot
.65875 .1132 Mdot
.65875 .20814 Mdot
.65875 .2603 Mdot
.65875 .29859 Mdot
.65907 .18297 Mdot
.65907 .25612 Mdot
.65907 .29831 Mdot
.65946 .16956 Mdot
.65946 .29797 Mdot
.65996 .13287 Mdot
.65996 .29754 Mdot
.66059 .29698 Mdot
.66145 .29623 Mdot
.66201 .0625 Mdot
.66201 .07727 Mdot
.66201 .09202 Mdot
.66201 .18898 Mdot
.66201 .25669 Mdot
.66201 .29574 Mdot
.66267 .17729 Mdot
.66267 .29516 Mdot
.6635 .14483 Mdot
.6635 .16291 Mdot
.6635 .29444 Mdot
.66454 .0625 Mdot
.66454 .08317 Mdot
.66454 .1038 Mdot
.66454 .29353 Mdot
.6654 .18384 Mdot
.6654 .29278 Mdot
.66589 .29234 Mdot
.66644 .17136 Mdot
.66644 .29186 Mdot
.66774 .13741 Mdot
.66774 .18948 Mdot
.66774 .29073 Mdot
.66893 .1785 Mdot
.66893 .28969 Mdot
.66938 .0731 Mdot
.66938 .28929 Mdot
.67039 .14823 Mdot
.67039 .28841 Mdot
.67155 .19867 Mdot
.67155 .2874 Mdot
.6722 .09133 Mdot
.6722 .11051 Mdot
.6722 .28683 Mdot
.67262 .15736 Mdot
.67262 .17294 Mdot
.67262 .28646 Mdot
.67452 .12393 Mdot
.67452 .14136 Mdot
.67452 .17959 Mdot
.67452 .20584 Mdot
.67452 .2848 Mdot
.67646 .15123 Mdot
.67646 .28309 Mdot
.6769 .0625 Mdot
.6769 .08219 Mdot
.6769 .21159 Mdot
.6769 .28271 Mdot
.67759 .19034 Mdot
.67759 .2821 Mdot
.67812 .15964 Mdot
.67812 .17435 Mdot
.67812 .22795 Mdot
.67812 .28165 Mdot
.67885 .09843 Mdot
.67885 .11633 Mdot
.67885 .2163 Mdot
.67885 .281 Mdot
.67997 .19867 Mdot
.67997 .28002 Mdot
.68048 .12847 Mdot
.68048 .14483 Mdot
.68048 .22023 Mdot
.68048 .27958 Mdot
.68186 .15389 Mdot
.68186 .2053 Mdot
.68186 .22357 Mdot
.68186 .27837 Mdot
.68341 .07169 Mdot
.68341 .09005 Mdot
.68341 .10838 Mdot
.68341 .21071 Mdot
.68341 .27702 Mdot
.68408 .18144 Mdot
.68408 .27643 Mdot
.68469 .10464 Mdot
.68469 .12144 Mdot
.68469 .2152 Mdot
.68469 .2759 Mdot
.68577 .13249 Mdot
.68577 .14791 Mdot
.68577 .19106 Mdot
.68577 .219 Mdot
.68577 .27495 Mdot
.68669 .15627 Mdot
.68669 .22224 Mdot
.68669 .27414 Mdot
.68711 .19867 Mdot
.68711 .27378 Mdot
.68819 .20484 Mdot
.68819 .27283 Mdot
.68909 .0625 Mdot
.68909 .07973 Mdot
.68909 .09693 Mdot
.68909 .1141 Mdot
.68909 .20994 Mdot
.68909 .27205 Mdot
.68984 .11014 Mdot
.68984 .12595 Mdot
.68984 .21424 Mdot
.68984 .27139 Mdot
.69049 .13608 Mdot
.69049 .15066 Mdot
.69049 .2179 Mdot
.69049 .27082 Mdot
.69104 .27034 Mdot
.69195 .18297 Mdot
.69195 .26955 Mdot
.69266 .19166 Mdot
.69266 .26892 Mdot
.69323 .19867 Mdot
.69323 .26842 Mdot
.6937 .0713 Mdot
.6937 .20444 Mdot
.6937 .26801 Mdot
.6941 .07061 Mdot
.6941 .08681 Mdot
.6941 .103 Mdot
.6941 .11914 Mdot
.6941 .20927 Mdot
.6941 .26767 Mdot
.69443 .11503 Mdot
.69443 .12996 Mdot
.69443 .21337 Mdot
.69443 .26737 Mdot
.69854 .0625 Mdot
.69854 .07904 Mdot
.69854 .09311 Mdot
.69854 .09556 Mdot
.69854 .12361 Mdot
.69854 .13356 Mdot
.69854 .14978 Mdot
.69854 .18425 Mdot
.69854 .19217 Mdot
.69854 .20408 Mdot
.69854 .20867 Mdot
.69854 .26378 Mdot
.70251 .09874 Mdot
.70251 .1132 Mdot
.70251 .12761 Mdot
.70251 .20814 Mdot
.70251 .2603 Mdot
.70282 .0703 Mdot
.70282 .0859 Mdot
.70282 .10147 Mdot
.70282 .20377 Mdot
.70282 .26003 Mdot
.70319 .19867 Mdot
.70319 .25971 Mdot
.70362 .19262 Mdot
.70362 .25933 Mdot
.70414 .25887 Mdot
.70479 .15389 Mdot
.70479 .25831 Mdot
.70559 .13973 Mdot
.70559 .2576 Mdot
.70664 .0625 Mdot
.70664 .07727 Mdot
.70664 .09202 Mdot
.70664 .10675 Mdot
.70664 .20349 Mdot
.70664 .25669 Mdot
.70729 .19867 Mdot
.70729 .25612 Mdot
.70806 .193 Mdot
.70806 .25545 Mdot
.70896 .25466 Mdot
.71007 .15736 Mdot
.71007 .25369 Mdot
.71094 .07002 Mdot
.71094 .08505 Mdot
.71094 .19867 Mdot
.71094 .25293 Mdot
.71143 .14483 Mdot
.71143 .2525 Mdot
.71196 .19334 Mdot
.71196 .25204 Mdot
.71316 .10247 Mdot
.71316 .16956 Mdot
.71316 .25099 Mdot
.71459 .16034 Mdot
.71459 .24973 Mdot
.71542 .19365 Mdot
.71542 .24901 Mdot
.71633 .14913 Mdot
.71633 .24821 Mdot
.71683 .17136 Mdot
.71683 .24777 Mdot
.7185 .11107 Mdot
.7185 .13522 Mdot
.7185 .16291 Mdot
.7185 .17959 Mdot
.7185 .19391 Mdot
.7185 .24631 Mdot
.72008 .17294 Mdot
.72008 .24493 Mdot
.72052 .15279 Mdot
.72052 .24455 Mdot
.72128 .0625 Mdot
.72128 .24388 Mdot
.72193 .16517 Mdot
.72193 .24331 Mdot
.72298 .11826 Mdot
.72298 .14042 Mdot
.72298 .17435 Mdot
.72298 .2424 Mdot
.72412 .15594 Mdot
.72412 .24139 Mdot
.72495 .16715 Mdot
.72495 .24067 Mdot
.72606 .07503 Mdot
.72606 .10006 Mdot
.72606 .2397 Mdot
.72677 .12437 Mdot
.72677 .14483 Mdot
.72677 .23908 Mdot
.72727 .15869 Mdot
.72727 .23864 Mdot
.72763 .16891 Mdot
.72763 .23832 Mdot
.73003 .0625 Mdot
.73003 .08546 Mdot
.73003 .10838 Mdot
.73003 .12962 Mdot
.73003 .14863 Mdot
.73003 .16111 Mdot
.73003 .23622 Mdot
.73248 .16325 Mdot
.73248 .23408 Mdot
.73286 .13417 Mdot
.73286 .15192 Mdot
.73286 .23375 Mdot
.73338 .09429 Mdot
.73338 .11542 Mdot
.73338 .23329 Mdot
.73414 .23263 Mdot
.73466 .16517 Mdot
.73466 .23217 Mdot
.73534 .13817 Mdot
.73534 .15481 Mdot
.73534 .23158 Mdot
.73625 .12144 Mdot
.73625 .23078 Mdot
.73753 .07431 Mdot
.73753 .14169 Mdot
.73753 .15736 Mdot
.73753 .22966 Mdot
.73873 .12665 Mdot
.73873 .14483 Mdot
.73873 .22861 Mdot
.73948 .14483 Mdot
.73948 .15964 Mdot
.73948 .22795 Mdot
.74038 .0625 Mdot
.74038 .08426 Mdot
.74038 .10597 Mdot
.74038 .22717 Mdot
.74089 .1312 Mdot
.74089 .14823 Mdot
.74089 .22672 Mdot
.7428 .07258 Mdot
.7428 .09274 Mdot
.7428 .11284 Mdot
.7428 .13522 Mdot
.7428 .15123 Mdot
.7428 .22505 Mdot
.7445 .13878 Mdot
.7445 .15389 Mdot
.7445 .22357 Mdot
.74489 .10006 Mdot
.74489 .11877 Mdot
.74489 .22322 Mdot
.74552 .22267 Mdot
.74601 .14197 Mdot
.74601 .15627 Mdot
.74601 .22224 Mdot
.74671 .10643 Mdot
.74671 .12393 Mdot
.74671 .22163 Mdot
.74779 .22068 Mdot
.7483 .11204 Mdot
.7483 .12847 Mdot
.7483 .22023 Mdot
.74972 .0625 Mdot
.74972 .08317 Mdot
.74972 .13249 Mdot
.74972 .219 Mdot
.75098 .13608 Mdot
.75098 .15066 Mdot
.75098 .2179 Mdot
.75137 .07212 Mdot
.75137 .09133 Mdot
.75137 .21755 Mdot
.75211 .21691 Mdot
.7528 .0625 Mdot
.7528 .08047 Mdot
.7528 .09843 Mdot
.7528 .11633 Mdot
.7528 .2163 Mdot
.75405 .08781 Mdot
.75405 .10464 Mdot
.75405 .12144 Mdot
.75405 .2152 Mdot
.75516 .11014 Mdot
.75516 .12595 Mdot
.75516 .21424 Mdot
.75567 .21379 Mdot
.75614 .11503 Mdot
.75614 .12996 Mdot
.75614 .21337 Mdot
.75702 .2126 Mdot
.75819 .21159 Mdot
.75919 .07169 Mdot
.75919 .09005 Mdot
.75919 .21071 Mdot
.76006 .0625 Mdot
.76006 .07973 Mdot
.76006 .09693 Mdot
.76006 .20994 Mdot
.76084 .07061 Mdot
.76084 .08681 Mdot
.76084 .103 Mdot
.76084 .11914 Mdot
.76084 .20927 Mdot
.76152 .09311 Mdot
.76152 .10838 Mdot
.76152 .12361 Mdot
.76152 .20867 Mdot
.76213 .09874 Mdot
.76213 .1132 Mdot
.76213 .12761 Mdot
.76213 .20814 Mdot
.76317 .20722 Mdot
.76404 .20647 Mdot
.76476 .20584 Mdot
.76537 .2053 Mdot
.7659 .20484 Mdot
.76636 .0713 Mdot
.76636 .20444 Mdot
.76676 .0625 Mdot
.76676 .07904 Mdot
.76676 .09556 Mdot
.76676 .20408 Mdot
.76712 .0703 Mdot
.76712 .0859 Mdot
.76712 .10147 Mdot
.76712 .20377 Mdot
.76744 .0625 Mdot
.76744 .07727 Mdot
.76744 .09202 Mdot
.76744 .10675 Mdot
.76744 .12144 Mdot
.76744 .20349 Mdot
.77295 .07002 Mdot
.77295 .0784 Mdot
.77295 .08505 Mdot
.77295 .19867 Mdot
.77838 .06975 Mdot
.77838 .08426 Mdot
.77838 .19391 Mdot
.77869 .0625 Mdot
.77869 .19365 Mdot
.77904 .19334 Mdot
.77942 .193 Mdot
.77987 .19262 Mdot
.78037 .19217 Mdot
.78096 .19166 Mdot
.78165 .19106 Mdot
.78247 .19034 Mdot
.78345 .18948 Mdot
.78403 .18898 Mdot
.78467 .18842 Mdot
.78539 .18779 Mdot
.7862 .18708 Mdot
.78713 .18626 Mdot
.78819 .18533 Mdot
.78943 .18425 Mdot
.78989 .18384 Mdot
.79089 .18297 Mdot
.79202 .18198 Mdot
.79264 .18144 Mdot
.7933 .18087 Mdot
.79476 .17959 Mdot
.796 .1785 Mdot
.79644 .17811 Mdot
.79739 .17729 Mdot
.79841 .17639 Mdot
.79896 .17591 Mdot
.80074 .17435 Mdot
.80235 .17294 Mdot
.80278 .17256 Mdot
.80354 .17191 Mdot
.80416 .17136 Mdot
.80516 .17049 Mdot
.80621 .16956 Mdot
.80696 .16891 Mdot
.80794 .16806 Mdot
.80855 .16752 Mdot
.80897 .16715 Mdot
.81124 .16517 Mdot
.81342 .16325 Mdot
.81381 .16291 Mdot
.81437 .16243 Mdot
.81523 .16167 Mdot
.81587 .16111 Mdot
.81676 .16034 Mdot
.81756 .15964 Mdot
.81806 .1592 Mdot
.81864 .15869 Mdot
.82016 .15736 Mdot
.82141 .15627 Mdot
.82178 .15594 Mdot
.82246 .15535 Mdot
.82307 .15481 Mdot
.82413 .15389 Mdot
.82539 .15279 Mdot
.82637 .15192 Mdot
.82717 .15123 Mdot
.82782 .15066 Mdot
.82883 .14978 Mdot
.82957 .14913 Mdot
.83014 .14863 Mdot
.83059 .14823 Mdot
.83096 .14791 Mdot
.83447 .14483 Mdot
.83775 .14197 Mdot
.83806 .14169 Mdot
.83845 .14136 Mdot
.83892 .14095 Mdot
.83952 .14042 Mdot
.84031 .13973 Mdot
.84139 .13878 Mdot
.8421 .13817 Mdot
.84296 .13741 Mdot
.84405 .13645 Mdot
.84448 .12144 Mdot
.84448 .13608 Mdot
.84546 .13522 Mdot
.84666 .13417 Mdot
.84736 .13356 Mdot
.84815 .13287 Mdot
.84858 .11701 Mdot
.84858 .13249 Mdot
.85005 .1312 Mdot
.85147 .11503 Mdot
.85147 .12996 Mdot
.85187 .12962 Mdot
.85257 .129 Mdot
.85317 .11204 Mdot
.85317 .12847 Mdot
.85416 .1132 Mdot
.85416 .12761 Mdot
.85526 .10838 Mdot
.85526 .12665 Mdot
.85606 .11014 Mdot
.85606 .12595 Mdot
.85715 .12499 Mdot
.85786 .12437 Mdot
.85836 .10643 Mdot
.85836 .12393 Mdot
.85873 .10838 Mdot
.85873 .12361 Mdot
.86121 .10184 Mdot
.86121 .10464 Mdot
.86121 .10675 Mdot
.86121 .12144 Mdot
.86384 .103 Mdot
.86384 .11914 Mdot
.86426 .10006 Mdot
.86426 .11877 Mdot
.86484 .096 Mdot
.86484 .11826 Mdot
.86569 .11752 Mdot
.86628 .10147 Mdot
.86628 .11701 Mdot
.86705 .09843 Mdot
.86705 .11633 Mdot
.8681 .09429 Mdot
.8681 .11542 Mdot
.86853 .10006 Mdot
.86853 .11503 Mdot
.8696 .08833 Mdot
.8696 .09693 Mdot
.8696 .1141 Mdot
.87063 .09874 Mdot
.87063 .1132 Mdot
.87104 .09274 Mdot
.87104 .11284 Mdot
.87195 .09556 Mdot
.87195 .11204 Mdot
.87306 .08681 Mdot
.87306 .11107 Mdot
.8737 .09133 Mdot
.8737 .11051 Mdot
.87412 .0784 Mdot
.87412 .09429 Mdot
.87412 .11014 Mdot
.87614 .07169 Mdot
.87614 .07781 Mdot
.87614 .08546 Mdot
.87614 .09005 Mdot
.87614 .09311 Mdot
.87614 .10838 Mdot
.878 .07727 Mdot
.878 .09202 Mdot
.878 .10675 Mdot
.87836 .0713 Mdot
.87836 .08888 Mdot
.87836 .10643 Mdot
.87889 .0625 Mdot
.87889 .08426 Mdot
.87889 .10597 Mdot
.87974 .07676 Mdot
.87974 .10522 Mdot
.88041 .07094 Mdot
.88041 .08781 Mdot
.88041 .10464 Mdot
.88137 .0625 Mdot
.88137 .08317 Mdot
.88137 .1038 Mdot
.88229 .07061 Mdot
.88229 .08681 Mdot
.88229 .103 Mdot
.88289 .07584 Mdot
.88289 .10247 Mdot
.88361 .0625 Mdot
.88361 .08219 Mdot
.88361 .10184 Mdot
.88403 .0703 Mdot
.88403 .0859 Mdot
.88403 .10147 Mdot
.88565 .0625 Mdot
.88565 .07002 Mdot
.88565 .07503 Mdot
.88565 .08129 Mdot
.88565 .08505 Mdot
.88565 .10006 Mdot
.88715 .06975 Mdot
.88715 .08426 Mdot
.88715 .09874 Mdot
.88751 .0625 Mdot
.88751 .08047 Mdot
.88751 .09843 Mdot
.8881 .07431 Mdot
.8881 .09791 Mdot
.88922 .0625 Mdot
.88922 .07973 Mdot
.88922 .09693 Mdot
.89028 .07367 Mdot
.89028 .096 Mdot
.89079 .0625 Mdot
.89079 .07904 Mdot
.89079 .09556 Mdot
.89224 .0625 Mdot
.89224 .0731 Mdot
.89224 .0784 Mdot
.89224 .09429 Mdot
.89359 .0625 Mdot
.89359 .07781 Mdot
.89359 .09311 Mdot
.89401 .07258 Mdot
.89401 .09274 Mdot
.89484 .0625 Mdot
.89484 .07727 Mdot
.89484 .09202 Mdot
.89562 .07212 Mdot
.89562 .09133 Mdot
.89708 .0625 Mdot
.89708 .07169 Mdot
.89708 .09005 Mdot
.89842 .0713 Mdot
.89842 .08888 Mdot
.89905 .0625 Mdot
.89905 .08833 Mdot
.89965 .07094 Mdot
.89965 .08781 Mdot
.90078 .0625 Mdot
.90078 .07061 Mdot
.90078 .08681 Mdot
.90183 .0703 Mdot
.90183 .0859 Mdot
.90233 .0625 Mdot
.90233 .08546 Mdot
.9028 .07002 Mdot
.9028 .08505 Mdot
.90371 .0625 Mdot
.90371 .06975 Mdot
.90371 .08426 Mdot
.90495 .0625 Mdot
.90495 .08317 Mdot
.90607 .0625 Mdot
.90607 .08219 Mdot
.9071 .0625 Mdot
.9071 .08129 Mdot
.90803 .0625 Mdot
.90803 .08047 Mdot
.90888 .0625 Mdot
.90888 .07973 Mdot
.90967 .0625 Mdot
.90967 .07904 Mdot
.9104 .0625 Mdot
.9104 .0784 Mdot
.91107 .0625 Mdot
.91107 .07781 Mdot
.9117 .0625 Mdot
.9117 .07727 Mdot
.91228 .07676 Mdot
.91333 .07584 Mdot
.91425 .07503 Mdot
.91507 .07431 Mdot
.9158 .07367 Mdot
.91646 .0731 Mdot
.91705 .07258 Mdot
.91758 .07212 Mdot
.91807 .07169 Mdot
.91852 .0713 Mdot
.91893 .07094 Mdot
.91931 .07061 Mdot
.91966 .0703 Mdot
.91998 .07002 Mdot
.92028 .06975 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{446, 223},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006n0000gb000`400?l00000o`00003oo`3ooono
0?ooo`00o`3ooono0?ooo`00o`3ooono0?ooo`00o`3ooono0?ooo`00=`3oool2000003@0oooo1000
00030?ooo`800000<@3oool4000000<0oooo0P00000b0?ooo`800000103oool2000003@0oooo00<0
00000?ooo`3oool00P3oool200000340oooo0`0000040?ooo`800000<P3oool2000000@0oooo0P00
000U0?ooo`00=P3oool010000000oooo0?ooo`00000e0?ooo`050000003oool0oooo0?ooo`000000
0P3oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool0
0000<03oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000003<0oooo00D00000
0?ooo`3oool0oooo000000020?ooo`030000003oool0oooo0340oooo00@000000?ooo`3oool00000
0P3oool00`000000oooo0?ooo`0^0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3o
ool00000903oool003H0oooo00@000000?ooo`3oool00000=@3oool01@000000oooo0?ooo`3oool0
00000080oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`020?ooo`040000003oool0
oooo000003<0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`0^0?ooo`D00000
0P3oool010000000oooo0?ooo`00000c0?ooo`040000003oool0oooo00000080oooo00<000000?oo
o`3oool0;P3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000002@0oooo000f
0?ooo`040000003oool0oooo000003D0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000
003oool0oooo0300oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003oool0oooo02l0
oooo0P0000030?ooo`040000003oool0oooo00000300oooo00@000000?ooo`3oool000000`3oool0
10000000oooo0?ooo`00000`0?ooo`<000000`3oool010000000oooo0?ooo`00000`0?ooo`<00000
0`3oool010000000oooo0?ooo`00000T0?ooo`00=P3oool010000000oooo0?ooo`00000e0?ooo`05
0000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`0a0?ooo`040000003oool0oooo
00000080oooo00<000000?ooo`3oool0<@3oool010000000oooo0?ooo`0000020?ooo`030000003o
ool0oooo02l0oooo00<000000?ooo`0000000`3oool010000000oooo0?ooo`00000`0?ooo`030000
003oool0oooo00<0oooo00@000000?ooo`3oool00000<03oool00`000000oooo0?ooo`030?ooo`04
0000003oool0oooo000002@0oooo000f0?ooo`040000003oool0oooo000003<0oooo0`0000030?oo
o`040000003oool0oooo00000300oooo00@000000?ooo`3oool000000P3oool010000000oooo0?oo
o`00000`0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool00000<P3oool20000
00<0oooo00@000000?ooo`3oool00000<03oool00`000000oooo0?ooo`030?ooo`040000003oool0
oooo00000300oooo00<000000?ooo`3oool00`3oool010000000oooo0?ooo`00000T0?ooo`00=`3o
ool2000003H0oooo00<000000?ooo`3oool00P3oool200000380oooo0P0000040?ooo`800000<P3o
ool2000000@0oooo0P00000d0?ooo`030000003oool0oooo0080oooo0P00000a0?ooo`@000000`3o
ool200000380oooo0`0000030?ooo`8000009@3oool00?l0oooo_`3oool00?l0oooo_`3oool00?l0
oooo_`3oool00?l0oooo_`3oool001/0ooooo`00002I000000/0oooo000K0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00T0oooo000K0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0=`3oool00`00
0000oooo0?ooo`0g0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0>03oool00`00
0000oooo0?ooo`0g0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool06P3oool00`00
0000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`3o0?oooiD0oooo00<000000?ooo`3o
ool02@3oool001/0oooo00<000000?ooo`3oool0o`3ooonE0?ooo`030000003oool0oooo00T0oooo
000K0?ooo`<00000o`3ooonC0?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool0o`3ooonE
0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0?l0ooooU@3oool00`00
0000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`3o0?oooiD0oooo00<000000?ooo`3o
ool02@3oool001/0oooo00<000000?ooo`3oool0o`3ooonE0?ooo`030000003oool0oooo00T0oooo
000K0?ooo`<00000o`3ooonC0?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool0o`3ooonE
0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0?l0ooooU@3oool00`00
0000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`3o0?oooiD0oooo00<000000?ooo`3o
ool02@3oool001/0oooo00<000000?ooo`3oool0o`3ooonE0?ooo`030000003oool0oooo00T0oooo
000K0?ooo`<00000o`3ooonC0?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool0o`3ooonE
0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0?l0ooooU@3oool00`00
0000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`3o0?oooiD0oooo00<000000?ooo`3o
ool02@3oool001/0oooo00<000000?ooo`3oool0o`3ooonE0?ooo`030000003oool0oooo00T0oooo
000K0?ooo`<00000o`3ooonC0?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool0o`3ooonE
0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0?l0ooooU@3oool00`00
0000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`3o0?oooiD0oooo00<000000?ooo`3o
ool02@3oool00180oooo0P0000070?ooo`030000003oool0oooo01X0oooo0P0000050?ooo`l00000
9`3oool2000000<0oooo0P000000103oool000000000003oool300000080oooo0P0000020?ooo`80
000000<0oooo000000000000103oool200000080oooo0P0000000`3oool00000000000030?ooo`80
00000`3oool300000080oooo0P0000000`3oool00000000000020?ooo`8000001P3oool2000000D0
oooo0P0000000`3oool00000000000040?ooo`<000000`3oool2000000<0oooo0P0000000`3oool0
000000000002000000030?ooo`000000000000<0oooo0`00000:0?ooo`<00000103oool200000003
0?ooo`00000000000080000000<0oooo0000000000000P3oool2000000@0oooo0P0000040?ooo`80
000000<0oooo0000000000001P3oool2000000H0oooo0P000000103oool000000000003oool20000
0080oooo0`0000030?ooo`8000000`3oool200000080oooo0P0000000`3oool00000000000040?oo
o`8000000P3oool200000080oooo0P0000020?ooo`<0000000@0oooo000000000000oooo0P000003
0?ooo`8000009P3oool@00000280oooo00<000000?ooo`3oool02@3oool00140oooo00@000000?oo
o`3oool000001P3oool4000001T0oooo0P0000050?ooo`l000009`3oool2000000<0oooo0P000000
103oool000000000003oool300000080oooo0P0000020?ooo`80000000<0oooo000000000000103o
ool200000080oooo0P0000000`3oool00000000000030?ooo`8000000`3oool300000080oooo0P00
00000`3oool00000000000020?ooo`8000001P3oool2000000D0oooo0P0000000`3oool000000000
00040?ooo`<000000`3oool2000000<0oooo0P0000000`3oool0000000000002000000030?ooo`00
0000000000<0oooo0`00000:0?ooo`<00000103oool2000000030?ooo`00000000000080000000<0
oooo0000000000000P3oool2000000@0oooo0P0000040?ooo`80000000<0oooo0000000000001P3o
ool2000000H0oooo0P000000103oool000000000003oool200000080oooo0`0000030?ooo`800000
0`3oool200000080oooo0P0000000`3oool00000000000040?ooo`8000000P3oool200000080oooo
0P0000020?ooo`<0000000@0oooo000000000000oooo0P0000030?ooo`8000009P3oool@000001l0
oooo1000000;0?ooo`004@3oool010000000oooo0?ooo`0000060?ooo`030000003oool0oooo0?l0
ooooU@3oool00`000000oooo0?ooo`090?ooo`004@3oool010000000oooo0?ooo`0000060?ooo`03
0000003oool0oooo01d0oooo0`0000040?ooo`<00000103oool4000002P0oooo100000000`3oool0
000000000002000001<0oooo0P000000103oool000000000003oool2000000d0oooo0`0000020?oo
o`8000003@3oool2000000L0oooo0P0000030?ooo`8000000P3oool2000000H0oooo0P00000S0?oo
o`8000001P3oool200000080oooo0P0000030?ooo`8000001P3oool2000000d0oooo0P0000030?oo
o`<00000303oool200000080oooo0P0000000`3oool000000000000B0?ooo`80000000<0oooo0000
000000000P0000000`3oool000000000000X0?ooo`@000000`3oool4000000@0oooo0P00000O0?oo
o`030000003oool0oooo00T0oooo000A0?ooo`040000003oool0oooo000000H0oooo00<000000?oo
o`3oool07@3oool4000000<0oooo3@00000V0?ooo`@0000000<0oooo0000000000000`0000000`3o
ool00000000000020?ooo`8000002`3oool2000000040?ooo`00000000000?ooo`<000000P3oool2
000000<0oooo0P0000030?ooo`<000000P3oool2000000D0oooo0P0000060?ooo`@000001@3oool2
000000<0oooo0P0000020?ooo`8000001P3oool200000080oooo0P0000040?ooo`@000002`3oool3
000000D0oooo0P0000020?ooo`8000001P3oool200000080oooo0P0000030?ooo`8000001@3oool3
000000H0oooo0P0000050?ooo`8000000`3oool300000080oooo0P0000040?ooo`8000000P3oool2
00000080oooo0P0000000`3oool000000000000;0?ooo`8000000P3oool2000000040?ooo`000000
00000?ooo`@0000000<0oooo0000000000009@3oool>000000<0oooo0`00000O0?ooo`030000003o
ool0oooo00T0oooo000A0?ooo`040000003oool0oooo000000H0oooo00<000000?ooo`3oool07P3o
ool4000000@0oooo1`000000103oool0000000000000000[0?ooo`80000000@0oooo000000000000
oooo0P0000020?ooo`@000000P3oool2000000`0oooo0P0000020?ooo`8000000`3oool2000000@0
oooo0P0000090?ooo`800000203oool2000000D0oooo0P0000030?ooo`8000002P3oool200000080
oooo0P0000040?ooo`@000002`3oool3000000D0oooo0P0000000`3oool000000000000;0?ooo`80
00000`3oool2000000D0oooo0P0000070?ooo`8000002P3oool2000000<0oooo0P0000040?ooo`80
00000P3oool2000000/0oooo0P0000030?ooo`@000000P3oool2000000040?ooo`00000000000?oo
o`800000:P3oool<000000@0oooo1000000O0?ooo`030000003oool0oooo00T0oooo000B0?ooo`80
00001`3oool3000001l0oooo100000050?ooo`T00000:`3oool2000000X0oooo0P0000020?ooo`80
00001P3oool200000100oooo0P000000103oool000000000000000060?ooo`8000001@3oool20000
00h0oooo0`00000:0?ooo`<000002@3oool2000000X0oooo0P00000:0?ooo`8000002`3oool20000
00l0oooo0P0000040?ooo`8000001P3oool600000100oooo0P0000060?ooo`8000000`3oool20000
00X0oooo0P00000Z0?ooo`X000001@3oool4000001h0oooo0`00000;0?ooo`006`3oool00`000000
oooo0?ooo`0P0?ooo`@00000103oool:000002L0oooo0P000000103oool000000000003oool20000
00030?ooo`00000000000100oooo0P0000020?ooo`80000000<0oooo0000000000002@3oool20000
0080oooo0P0000050?ooo`<000001@3oool2000000D0oooo0P0000070?ooo`800000103oool20000
00<0oooo0P0000000`3oool00000000000030?ooo`8000000`3oool3000000X0oooo0`0000040?oo
o`8000000`3oool2000000030?ooo`000000000000<0oooo0P0000030?ooo`8000001`3oool20000
00H0oooo0P0000040?ooo`@00000103oool200000080oooo0P0000090?ooo`8000000P3oool20000
00030?ooo`00000000000140oooo0P000000103oool000000000003oool4000002P0oooo2P000003
0?ooo`D000008@3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`0Q0?oo
o`@00000103oool500000080oooo0P00000W0?ooo`80000000@0oooo000000000000oooo0P000000
0`3oool000000000000D0?ooo`80000000<0oooo0000000000000`3oool2000000l0oooo0P000006
0?ooo`8000001@3oool2000000P0oooo0P0000030?ooo`8000000`3oool2000000H0oooo0P000003
0?ooo`800000303oool2000000@0oooo0P0000060?ooo`8000000`3oool2000000<0oooo0P000007
0?ooo`8000001P3oool2000000H0oooo0P00000>0?ooo`8000000`3oool200000080oooo0P00000D
0?ooo`80000000@0oooo000000000000oooo1000000X0?ooo`80000000<0oooo0000000000001000
00040?ooo`<000008`3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`0S
0?ooo`@00000103oool6000002H0oooo100000080?ooo`8000000P3oool200000080oooo0P000006
0?ooo`8000002`3oool2000000<0oooo0P0000000`3oool000000000000?0?ooo`800000203oool2
000000D0oooo0P0000030?ooo`8000001@3oool2000002<0oooo0P0000040?ooo`800000103oool2
000000@0oooo0P0000090?ooo`8000003`3oool4000000<0oooo0P00000<0?ooo`8000001P3oool2
00000080oooo0P0000020?ooo`8000001`3oool2000000030?ooo`000000000002D0oooo1`000004
0?ooo`@00000903oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`0T0?oo
o`@000000`3oool8000002@0oooo100000000`3oool0000000000002000000030?ooo`0000000000
008000000P3oool200000080oooo0P0000060?ooo`80000000@0oooo000000000000oooo0P00000:
0?ooo`80000000<0oooo0000000000000`3oool2000000d0oooo0P0000050?ooo`8000001@3oool2
00000080oooo0`0000050?ooo`8000001`3oool2000000030?ooo`000000000000/0oooo0P000000
0`3oool00000000000070?ooo`800000103oool400000080oooo0P0000040?ooo`8000001@3oool2
000000d0oooo0P0000040?ooo`@000002P3oool200000080oooo0P0000000`3oool0000000000006
0?ooo`8000000P3oool200000080oooo1P0000000`3oool0000000000002000000030?ooo`000000
000002@0oooo200000030?ooo`@000009@3oool00`000000oooo0?ooo`090?ooo`006`3oool30000
02D0oooo100000030?ooo`P00000:03oool7000001<0oooo0P0000000`3oool00000000000030?oo
o`8000003@3oool2000000D0oooo0P0000060?ooo`8000001`3oool2000000L0oooo0P00000:0?oo
o`8000000`3oool2000000030?ooo`000000000000@0oooo0P0000050?ooo`80000000<0oooo0000
000000000`3oool2000000X0oooo0P0000070?ooo`8000001P3oool2000000H0oooo0P0000050?oo
o`8000003P3oool200000080oooo0P0000020?ooo`8000004`3oool4000000030?ooo`0000000000
02P0oooo200000030?ooo`@00000903oool3000000/0oooo000K0?ooo`030000003oool0oooo02H0
oooo100000040?ooo`H00000:03oool2000000040?ooo`00000000000?ooo`8000000P3oool20000
00`0oooo0P0000080?ooo`8000000P3oool200000080oooo0P00000<0?ooo`8000000`3oool20000
00X0oooo0P0000090?ooo`8000001@3oool2000000030?ooo`000000000000@0oooo100000040?oo
o`800000103oool4000000D0oooo0P0000000`3oool00000000000050?ooo`8000002@3oool20000
00X0oooo0P0000020?ooo`800000303oool200000080oooo0P0000030?ooo`8000001`3oool20000
00d0oooo0P0000020?ooo`80000000@0oooo000000000000oooo0P00000X0?ooo`D000001@3oool4
000002L0oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool09`3oool40000
00@0oooo1P00000W0?ooo`@000000P3oool200000080oooo0P0000000`3oool00000000000090?oo
o`8000000P3oool2000000P0oooo0P0000020?ooo`8000000P3oool2000000d0oooo0P0000020?oo
o`<000003`3oool3000000D0oooo0P0000030?ooo`8000000P3oool4000000X0oooo100000030?oo
o`8000000`3oool2000000D0oooo0`00000?0?ooo`<000000P3oool2000000`0oooo0P0000020?oo
o`8000000P3oool2000000T0oooo0P0000000`3oool00000000000090?ooo`8000000P3oool20000
0080oooo0P0000020?ooo`@000009`3oool6000000@0oooo0`00000Y0?ooo`030000003oool0oooo
00T0oooo000K0?ooo`030000003oool0oooo02P0oooo100000030?ooo`L000009P3oool5000000P0
oooo0P00000=0?ooo`80000000<0oooo000000000000303oool3000000D0oooo0P00000:0?ooo`D0
00003@3oool2000000<0oooo100000040?ooo`8000005`3oool2000000<0oooo0P0000000`3oool0
0000000000030?ooo`800000303oool2000000040?ooo`0000000000000000X0oooo0P0000040?oo
o`@000002`3oool200000080oooo0P00000<0?ooo`8000002@3oool5000002H0oooo1P0000040?oo
o`<00000:P3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`0Y0?ooo`@0
0000103oool5000002H0oooo1@000000103oool000000000003oool2000000D0oooo0P0000080?oo
o`@0000000<0oooo0000000000000`3oool2000000H0oooo0`0000000`3oool00000000000030?oo
o`8000000`3oool200000080oooo0P0000040?ooo`800000203oool2000000P0oooo1000000;0?oo
o`8000002@3oool2000000X0oooo0P0000000`3oool00000000000080?ooo`8000001`3oool20000
00D0oooo0P0000020?ooo`8000000`3oool2000000<0oooo0P000000103oool00000000000000006
0?ooo`8000000P3oool200000080oooo100000070?ooo`8000001P3oool2000000040?ooo`000000
00000?ooo`D000009P3oool5000000<0oooo1000000[0?ooo`030000003oool0oooo00T0oooo000K
0?ooo`<00000:P3oool5000000<0oooo1@00000U0?ooo`@000000P3oool2000000030?ooo`000000
000000D0oooo0P0000040?ooo`8000000P3oool2000000P0oooo0P0000060?ooo`8000000P3oool2
000000@0oooo0P0000020?ooo`80000000<0oooo0000000000000`00000<0?ooo`800000403oool2
000000D0oooo0P0000090?ooo`8000001@3oool200000100oooo0P00000;0?ooo`80000000D0oooo
00000000000000000?ooo`02000000030?ooo`000000000000D0oooo0P0000020?ooo`8000001P3o
ool2000000P0oooo0P0000020?ooo`8000000`3oool2000000H0oooo0P0000000`3oool000000000
00020?ooo`@00000903oool6000000<0oooo1000000Z0?ooo`<000002`3oool001/0oooo00<00000
0?ooo`3oool0;03oool4000000<0oooo1000000U0?ooo`@0000000<0oooo0000000000003`3oool2
00000080oooo0P0000000`3oool00000000000020?ooo`8000001P3oool2000000H0oooo0P000003
0?ooo`8000001@3oool2000000030?ooo`000000000000<0oooo0P00000A0?ooo`8000001P3oool2
000000@0oooo0P00000;0?ooo`800000103oool2000000H0oooo0P00000A0?ooo`8000000P3oool2
00000080oooo0P0000040?ooo`<000000P3oool2000000H0oooo0P0000060?ooo`8000000P3oool2
00000080oooo0P0000020?ooo`8000003`3oool4000000030?ooo`000000000002@0oooo10000004
0?ooo`@00000;@3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`0]0?oo
o`@000000`3oool5000002<0oooo10000000103oool000000000003oool2000000030?ooo`000000
000000/0oooo0P0000030?ooo`8000000P3oool2000000H0oooo0P0000060?ooo`8000000`3oool2
000000D0oooo0P0000060?ooo`8000000P3oool2000000d0oooo0P0000000`3oool0000000000009
0?ooo`8000002P3oool3000000T0oooo0P0000000`3oool000000000000=0?ooo`8000000P3oool2
000000H0oooo0P0000050?ooo`8000000P3oool2000000H0oooo0P0000060?ooo`8000000P3oool2
000000<0oooo0P00000<0?ooo`80000000@0oooo000000000000oooo100000000`3oool000000000
000S0?ooo`@00000103oool4000002h0oooo00<000000?ooo`3oool02@3oool001/0oooo00<00000
0?ooo`3oool0;P3oool3000000@0oooo1@00000R0?ooo`@00000103oool2000000030?ooo`000000
00000080oooo0P0000070?ooo`800000203oool2000000X0oooo0P00000=0?ooo`8000000P3oool2
000000@0oooo1000000@0?ooo`80000000<0oooo0000000000001P3oool2000000X0oooo0P000007
0?ooo`80000000<0oooo000000000000403oool3000000D0oooo0P0000020?ooo`800000303oool2
000000/0oooo0P0000070?ooo`800000203oool200000080oooo0P0000000`3oool0000000000004
0?ooo`@000008P3oool4000000@0oooo1000000_0?ooo`030000003oool0oooo00T0oooo000K0?oo
o`030000003oool0oooo02l0oooo100000030?ooo`D000008`3oool2000000030?ooo`0000000000
008000001P3oool200000080oooo0P00000:0?ooo`80000000<0oooo0000000000000P3oool20000
00H0oooo0P0000000`3oool0000000000002000000T0oooo0P0000000`3oool00000000000040?oo
o`8000002P3oool2000000T0oooo0P0000030?ooo`8000004@3oool2000000<0oooo0P0000090?oo
o`8000002@3oool2000000D0oooo0P0000000`3oool00000000000090?ooo`<0000000<0oooo0000
000000001P3oool2000000<0oooo0P0000000`3oool00000000000090?ooo`8000000`3oool20000
00D0oooo0P000000103oool000000000003oool2000002<0oooo100000040?ooo`<00000<@3oool0
0`000000oooo0?ooo`090?ooo`006`3oool300000300oooo100000040?ooo`<000008`3oool20000
00030?ooo`00000000000080000000<0oooo0000000000001`3oool2000000D0oooo0P0000030?oo
o`8000001@3oool2000000@0oooo0P0000030?ooo`@000000`3oool2000000@0oooo0P00000:0?oo
o`8000001`3oool3000000D0oooo0P0000060?ooo`8000000P3oool2000000T0oooo0P0000020?oo
o`8000001P3oool2000000D0oooo0`0000070?ooo`8000002P3oool2000000<0oooo0P0000040?oo
o`<00000103oool2000000<0oooo0P0000050?ooo`<000000`3oool2000000@0oooo0P0000080?oo
o`@0000000@0oooo000000000000oooo0P00000S0?ooo`<00000103oool4000002l0oooo0`00000;
0?ooo`006`3oool00`000000oooo0?ooo`0a0?ooo`@000000`3oool500000240oooo1P0000020?oo
o`80000000<0oooo0000000000002`3oool2000000030?ooo`0000000000008000000`3oool20000
00H0oooo0P0000020?ooo`8000001P3oool400000080oooo0P00000:0?ooo`8000000P3oool20000
00@0oooo0P0000050?ooo`8000000`3oool2000000D0oooo0P0000030?ooo`800000103oool20000
00D0oooo0P0000030?ooo`8000001@3oool2000000@0oooo0P0000020?ooo`8000002P3oool20000
0080oooo0`0000070?ooo`8000000P3oool2000000H0oooo0P0000020?ooo`@000000P3oool20000
00X0oooo0P0000020?ooo`8000000P3oool600000240oooo100000030?ooo`@00000<`3oool00`00
0000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`0b0?ooo`@000000`3oool400000240
oooo1P0000000`3oool00000000000020?ooo`8000000`3oool2000000T0oooo0P0000050?ooo`80
00001`3oool2000000030?ooo`000000000000P0oooo0P0000090?ooo`80000000<0oooo00000000
0000103oool2000000H0oooo0P0000080?ooo`8000002P3oool2000000/0oooo0P0000080?ooo`80
00001P3oool2000000<0oooo0P0000020?ooo`8000002@3oool2000000P0oooo0P0000000`3oool0
0000000000070?ooo`800000103oool2000000X0oooo0P0000020?ooo`8000000`3oool200000003
0?ooo`000000000000<000008P3oool3000000<0oooo1000000d0?ooo`030000003oool0oooo00T0
oooo000K0?ooo`030000003oool0oooo03@0oooo0`0000050?ooo`8000007P3oool4000000030?oo
o`00000000000080oooo0P0000000`3oool00000000000040?ooo`8000002@3oool200000080oooo
0P00000:0?ooo`800000103oool2000000030?ooo`000000000000d0oooo0P0000000`3oool00000
000000040?ooo`<000001@3oool4000000030?ooo`000000000000L0oooo1000000;0?ooo`<00000
203oool2000000030?ooo`0000000000008000001@3oool3000000<0oooo0P0000020?ooo`800000
3@3oool4000000D0oooo0P0000090?ooo`8000000P3oool2000000X0oooo0P0000030?ooo`800000
0P3oool200000080oooo100000000`3oool000000000000N0?ooo`800000103oool4000003D0oooo
00<000000?ooo`3oool02@3oool00140oooo0`0000070?ooo`030000003oool0oooo03D0oooo1000
00030?ooo`8000007P3oool4000000040?ooo`00000000000?ooo`8000000P3oool200000080oooo
0P00000:0?ooo`8000000`3oool200000080oooo0P00000<0?ooo`80000000<0oooo000000000000
2@3oool4000000X0oooo0P0000070?ooo`80000000<0oooo0000000000001`3oool4000000/0oooo
0`0000080?ooo`80000000<0oooo0000000000001`3oool2000000X0oooo100000090?ooo`@00000
303oool200000080oooo0P0000040?ooo`8000002P3oool2000000030?ooo`000000000000<0oooo
0P000000103oool000000000003oool4000001h0oooo0P0000030?ooo`<00000=`3oool00`000000
oooo0?ooo`090?ooo`00503oool00`000000oooo0?ooo`040?ooo`@00000=@3oool4000002@0oooo
0P000000103oool000000000003oool2000000030?ooo`000000000000<0oooo0P00000:0?ooo`80
000000<0oooo000000000000103oool2000000`0oooo0P00000<0?ooo`@0000000<0oooo00000000
00001`3oool200000080oooo0P00000a0?ooo`8000000P3oool2000000L0oooo0P0000000`3oool0
000000000002000000`0oooo0P00000;0?ooo`8000001@3oool2000000030?ooo`000000000000X0
oooo0P0000020?ooo`8000000P3oool2000000040?ooo`00000000000?ooo`800000903oool30000
03D0oooo1000000;0?ooo`00503oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo03L0
oooo1000000S0?ooo`80000000<0oooo0000000000000P0000020?ooo`8000000P3oool2000000T0
oooo0P0000030?ooo`8000000P3oool2000000X0oooo0P0000020?ooo`8000000P3oool2000000D0
oooo0P0000060?ooo`8000001`3oool200000080oooo0P0000060?ooo`8000002@3oool3000000T0
oooo0`0000090?ooo`8000001P3oool200000080oooo0P0000070?ooo`8000001P3oool2000000D0
oooo0P0000020?ooo`8000000P3oool2000000X0oooo0P0000020?ooo`8000000`3oool2000000T0
oooo0P0000000`3oool00000000000020?ooo`80000000@0oooo000000000000oooo0P00000S0?oo
o`@00000>03oool00`000000oooo0?ooo`090?ooo`004@3oool3000000L0oooo00<000000?ooo`3o
ool0>03oool400000280oooo10000000103oool000000000003oool2000000<0oooo0P0000090?oo
o`80000000<0oooo000000000000103oool2000000X0oooo0`0000050?ooo`8000001@3oool20000
0080oooo0P0000060?ooo`8000003`3oool2000000T0oooo0`0000090?ooo`<000002@3oool20000
00l0oooo0P0000060?ooo`8000000P3oool2000000D0oooo0P0000050?ooo`<000002P3oool20000
00<0oooo0P0000020?ooo`8000002@3oool200000080oooo0P0000000`3oool00000000000020?oo
o`@000008P3oool3000003X0oooo00<000000?ooo`3oool02@3oool00140oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`0i0?ooo`@000008`3oool2000000040?ooo`00000000000?oo
o`8000000P3oool2000000`0oooo0`0000020?ooo`800000203oool2000000<0oooo0P0000020?oo
o`800000103oool2000000<0oooo0P0000000`3oool00000000000060?ooo`8000001`3oool20000
00T0oooo1000000G0?ooo`@00000203oool2000000P0oooo0P0000060?ooo`80000000<0oooo0000
000000000`3oool2000000@0oooo0P0000020?ooo`8000000P3oool2000000T0oooo0P0000000`3o
ool0000000000002000000`0oooo0P0000000`3oool00000000000020?ooo`80000000<0oooo0000
000000008P3oool4000003/0oooo00<000000?ooo`3oool02@3oool00140oooo00<000000?ooo`3o
ool01`3oool00`000000oooo0?ooo`0k0?ooo`<00000903oool3000000030?ooo`00000000000080
oooo0P0000030?ooo`8000001`3oool2000000030?ooo`000000000000800000203oool2000000@0
oooo0P0000000`3oool00000000000040?ooo`8000000`3oool2000000h0oooo0P0000020?ooo`80
00002@3oool4000001L0oooo100000080?ooo`8000000`3oool2000000h0oooo0P0000030?ooo`80
0000103oool2000000030?ooo`000000000000<0oooo0P0000090?ooo`@0000000<0oooo00000000
00001P3oool2000000@0oooo0P0000020?ooo`80000000<0oooo000000000000903oool4000003`0
oooo00<000000?ooo`3oool02@3oool00140oooo100000060?ooo`030000003oool0oooo03`0oooo
0`00000S0?ooo`80000000D0oooo00000000000000000?ooo`02000000@0oooo0P00000:0?ooo`80
00001`3oool2000000L0oooo0P00000<0?ooo`8000002P3oool200000080oooo0P00000?0?ooo`80
00005`3oool2000000l0oooo0P0000020?ooo`8000002P3oool2000000`0oooo0P0000070?ooo`80
00001`3oool2000000T0oooo0P0000050?ooo`80000000<0oooo0000000000000`00000S0?ooo`@0
0000?@3oool00`000000oooo0?ooo`090?ooo`006`3oool3000003d0oooo1000000S0?ooo`<00000
00@0oooo00000000000000000`3oool2000000D0oooo0P00000:0?ooo`80000000<0oooo00000000
00000`3oool2000000<0oooo0P0000040?ooo`80000000<0oooo0000000000000P3oool2000000D0
oooo100000000`3oool00000000000020?ooo`800000103oool2000000T0oooo0P00000G0?ooo`80
00002@3oool2000000<0oooo0P0000030?ooo`@0000000<0oooo0000000000001@3oool200000080
oooo100000050?ooo`8000000`3oool2000000<0oooo1000000;0?ooo`8000001@3oool2000000<0
oooo0`0000000`3oool000000000000T0?ooo`@00000?03oool3000000/0oooo000K0?ooo`030000
003oool0oooo03h0oooo0`00000S0?ooo`8000000P3oool2000000@0oooo0P0000050?ooo`800000
2P3oool2000000H0oooo0P0000030?ooo`800000103oool2000000030?ooo`000000000000T0oooo
100000050?ooo`800000103oool2000002d0oooo0P0000030?ooo`8000001@3oool2000000030?oo
o`000000000000T0oooo100000050?ooo`8000000`3oool2000000D0oooo0P00000;0?ooo`800000
1@3oool2000000@0oooo0P0000000`3oool000000000000T0?ooo`<00000?`3oool00`000000oooo
0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`100?ooo`8000009@3oool2000000<0oooo0P00
00040?ooo`8000002`3oool2000000D0oooo0P00000;0?ooo`8000002@3oool2000000/0oooo0P00
00070?ooo`8000001P3oool2000000040?ooo`00000000000?ooo`8000002`3oool2000000030?oo
o`0000000000008000001P3oool2000000P0oooo0P00000;0?ooo`8000002@3oool2000000/0oooo
0P0000050?ooo`8000002P3oool2000000D0oooo0P0000020?ooo`<00000903oool200000480oooo
00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0@03oool400000280oooo0`00
00020?ooo`<000000P3oool4000000T0oooo100000020?ooo`80000000<0oooo000000000000203o
ool2000000030?ooo`000000000000<0oooo0P0000040?ooo`8000000`3oool2000000030?ooo`00
0000000000<0oooo0P0000070?ooo`8000001P3oool2000000030?ooo`000000000000<000002`3o
ool7000000H0oooo0P0000080?ooo`8000000`3oool2000000030?ooo`000000000000<0oooo0P00
00040?ooo`8000000`3oool2000000030?ooo`000000000000P0oooo0P0000000`3oool000000000
00020?ooo`@00000203oool4000000<0oooo0`000000103oool0000000000000000R0?ooo`@00000
@P3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`120?ooo`<000007@3o
ool200000080oooo0P0000030?ooo`8000000`3oool2000000H0oooo0P0000000`3oool000000000
000200000080oooo100000080?ooo`80000000<0oooo0000000000000P3oool6000000D0oooo0P00
00020?ooo`80000000<0oooo0000000000003@3oool2000000X0oooo0`00000;0?ooo`<000002P3o
ool2000000d0oooo0P0000000`3oool00000000000020?ooo`8000001@3oool5000000<0oooo0P00
00000`3oool00000000000080?ooo`@000000P3oool4000000030?ooo`000000000000D0oooo0P00
00040?ooo`8000000P3oool3000000030?ooo`000000000001d0oooo100000130?ooo`030000003o
ool0oooo00T0oooo000K0?ooo`<00000@`3oool3000001`0oooo0P000000103oool0000000000000
00020?ooo`8000000`3oool2000000@0oooo0P000000103oool000000000003oool2000000040?oo
o`00000000000?ooo`8000002P3oool200000080oooo0P0000000`3oool0000000000002000000L0
oooo0`00000:0?ooo`8000000P3oool2000000<0oooo0P00000;0?ooo`8000002`3oool2000000/0
oooo0P0000030?ooo`8000000P3oool2000000X0oooo0`0000070?ooo`<000000P3oool200000080
oooo0P00000:0?ooo`@000000P3oool2000000040?ooo`00000000000?ooo`800000103oool20000
00<0oooo0P0000020?ooo`D00000703oool400000480oooo0`00000;0?ooo`006`3oool00`000000
oooo0?ooo`140?ooo`@00000703oool300000080oooo0`0000000`3oool0000000000002000000@0
oooo0P0000050?ooo`@000003P3oool2000000030?ooo`000000000000d0oooo0P0000050?ooo`80
00000`3oool200000080oooo0P0000000`3oool000000000000Y0?ooo`80000000<0oooo00000000
00000P3oool2000000<0oooo0P0000050?ooo`8000003@3oool2000000030?ooo`000000000000d0
oooo100000060?ooo`800000103oool4000000040?ooo`000000000000000080oooo0P00000M0?oo
o`@00000A@3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`150?ooo`@0
00006`3oool200000080oooo0`000000103oool000000000000000040?ooo`8000001`3oool20000
00@0oooo0P0000090?ooo`<00000403oool2000000D0oooo0P0000030?ooo`8000001@3oool20000
0080oooo0P0000000`3oool00000000000060?ooo`8000002`3oool2000000D0oooo0P0000000`3o
ool00000000000030?ooo`800000103oool2000000@0oooo0P0000050?ooo`800000403oool30000
00T0oooo0P0000030?ooo`800000203oool2000000@0oooo0`0000001@3oool00000000000000000
oooo00800000703oool4000004H0oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?oo
o`3oool0AP3oool4000001d0oooo0`000000103oool000000000000000030?ooo`@000002`3oool4
000000D0oooo0P0000020?ooo`8000000`3oool2000000d0oooo0P0000040?ooo`8000000`3oool2
00000080oooo0P0000050?ooo`80000000<0oooo0000000000000P0000040?ooo`8000002`3oool2
000000<0oooo100000000`3oool00000000000050?ooo`8000000P3oool2000000<0oooo0P000005
0?ooo`8000003@3oool2000000<0oooo0P0000020?ooo`8000001@3oool4000000/0oooo10000003
0?ooo`<0000000@0oooo00000000000000007@3oool4000004L0oooo00<000000?ooo`3oool02@3o
ool001/0oooo00<000000?ooo`3oool0A`3oool4000001`0oooo0P000000103oool0000000000000
00020?ooo`@000003@3oool2000000@0oooo0P0000000`3oool00000000000070?ooo`<000001`3o
ool2000000<0oooo100000020?ooo`8000001`3oool2000000X0oooo0P00000F0?ooo`8000002P3o
ool2000000L0oooo0P0000030?ooo`@000000P3oool2000000P0oooo0`0000070?ooo`80000000<0
oooo000000000000103oool2000000d0oooo0`0000030?ooo`<0000000<0oooo000000000000703o
ool4000004P0oooo00<000000?ooo`3oool02@3oool001/0oooo0`0000180?ooo`@000007@3oool3
00000080oooo0`00000E0?ooo`8000001P3oool2000000<0oooo0P0000030?ooo`8000000P3oool2
000000D0oooo0P00000?0?ooo`8000000P3oool2000000030?ooo`0000000000008000002@3oool2
000000X0oooo100000000`3oool00000000000020?ooo`8000003`3oool2000000@0oooo0P000003
0?ooo`8000000`3oool200000080oooo0P0000070?ooo`800000503oool400000080oooo0`00000M
0?ooo`<00000B03oool3000000/0oooo000K0?ooo`030000003oool0oooo04X0oooo0`00000L0?oo
o`80000000<0oooo0000000000000P00000K0?ooo`80000000<0oooo000000000000203oool20000
00X0oooo0P0000020?ooo`8000000P3oool2000000<0oooo0P0000001@3oool00000000000000000
oooo0080000000<0oooo0000000000000P0000090?ooo`8000002P3oool4000000040?ooo`000000
00000?ooo`D00000103oool200000080oooo0P0000020?ooo`8000002P3oool2000000L0oooo0P00
00020?ooo`8000006`3oool4000000030?ooo`000000000001/0oooo1000001;0?ooo`030000003o
ool0oooo00T0oooo000K0?ooo`030000003oool0oooo04/0oooo1000000M0?ooo`<0000000<0oooo
0000000000000P00000A0?ooo`@000000P3oool2000000`0oooo0P0000020?ooo`80000000<0oooo
0000000000000P3oool200000080oooo0P0000020?ooo`8000000`3oool200000080oooo0P00000U
0?ooo`80000000<0oooo000000000000103oool200000080oooo0P0000020?ooo`8000000P3oool4
000000<0oooo0P00000<0?ooo`80000000@0oooo000000000000oooo0P00000A0?ooo`@0000000@0
oooo00000000000000007@3oool4000004`0oooo00<000000?ooo`3oool02@3oool001/0oooo00<0
00000?ooo`3oool0C03oool300000200oooo1@00000>0?ooo`80000000<0oooo0000000000000P00
000<0?ooo`8000000P3oool200000080oooo0P0000000`3oool000000000000C0?ooo`8000002`3o
ool2000000/0oooo0P00000:0?ooo`800000503oool4000000<0oooo0P0000000`3oool000000000
000<0?ooo`80000000@0oooo000000000000oooo0P00000>0?ooo`D000007`3oool4000004d0oooo
00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0C@3oool4000001d0oooo0`00
000?0?ooo`@000001@3oool2000000030?ooo`000000000000@0oooo0P0000000`3oool000000000
000=0?ooo`8000001P3oool3000000030?ooo`00000000000080000000<0oooo0000000000002P3o
ool3000000/0oooo0`0000090?ooo`8000000P3oool300000080oooo0P0000060?ooo`8000003@3o
ool200000080oooo0P0000030?ooo`8000000P3oool2000000D0oooo1000000?0?ooo`<000007@3o
ool4000004h0oooo00<000000?ooo`3oool02@3oool001/0oooo0`00001>0?ooo`@00000703oool2
00000100oooo0P0000020?ooo`@0000000@0oooo000000000000oooo0P0000040?ooo`8000000`3o
ool2000000030?ooo`000000000000P0oooo0P0000000`3oool00000000000030?ooo`<0000000<0
oooo0000000000000P0000020?ooo`8000002@3oool3000000/0oooo0`0000090?ooo`8000000P3o
ool300000080oooo0P0000040?ooo`@00000203oool200000080oooo0P0000030?ooo`8000000`3o
ool200000080oooo0P0000000`3oool000000000000200000080oooo0P00000@0?ooo`800000703o
ool4000004d0oooo0`00000;0?ooo`006`3oool00`000000oooo0?ooo`1?0?ooo`@00000703oool2
00000100oooo0P0000000`3oool0000000000002000000/0oooo0P0000020?ooo`80000000<0oooo
0000000000002`3oool2000000P0oooo0`0000020?ooo`8000002@3oool3000000/0oooo0`000009
0?ooo`80000000<0oooo0000000000000P0000080?ooo`8000002P3oool200000080oooo0P000000
0`3oool000000000000<0?ooo`H00000403oool2000001d0oooo0`00001A0?ooo`030000003oool0
oooo00T0oooo000K0?ooo`030000003oool0oooo0500oooo1000000K0?ooo`8000002@3oool20000
00030?ooo`00000000000080oooo0P0000050?ooo`80000000<0oooo0000000000001P3oool20000
00<0oooo0P0000040?ooo`800000203oool2000000<0oooo0P0000020?ooo`<000003@3oool20000
00d0oooo0P00000<0?ooo`@000000P3oool2000000<0oooo0P0000080?ooo`800000103oool20000
0080oooo0P0000070?ooo`80000000<0oooo000000000000103oool200000080oooo0P0000000`3o
ool00000000000090?ooo`800000703oool300000580oooo00<000000?ooo`3oool02@3oool001/0
oooo00<000000?ooo`3oool0D@3oool500000280oooo100000000`3oool00000000000060?ooo`80
000000@0oooo000000000000oooo0P00000;0?ooo`800000103oool200000080oooo0P0000040?oo
o`8000000`3oool200000100oooo0P0000000`3oool000000000000:0?ooo`80000000<0oooo0000
000000004@3oool2000000<0oooo0P0000040?ooo`8000000P3oool2000000@0oooo0P00000;0?oo
o`80000000@0oooo000000000000oooo0P0000050?ooo`80000000@0oooo000000000000oooo0P00
000R0?ooo`@00000D`3oool00`000000oooo0?ooo`090?ooo`002`3oool4000000<0oooo0P000007
0?ooo`030000003oool0oooo0580oooo1@00000Q0?ooo`8000002`3oool200000080oooo0P00000=
0?ooo`800000203oool2000000@0oooo0`0000030?ooo`8000002`3oool200000080oooo1@00000:
0?ooo`D000000`3oool2000000/0oooo0P0000030?ooo`<00000103oool2000000P0oooo0P00000=
0?ooo`8000000P3oool2000000/0oooo0P00000Q0?ooo`@00000E03oool00`000000oooo0?ooo`09
0?ooo`003@3oool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`3oool0103oool4
000005<0oooo1000000U0?ooo`80000000<0oooo0000000000000P3oool4000000030?ooo`000000
000000h0oooo0P0000040?ooo`800000203oool2000000<0oooo0P00000;0?ooo`@000000P3oool2
000000`0oooo0P0000030?ooo`@000002`3oool2000000<0oooo0P0000070?ooo`800000103oool2
000000l0oooo100000000`3oool00000000000020?ooo`80000000<0oooo0000000000009@3oool4
00000580oooo1000000;0?ooo`003@3oool01@000000oooo0?ooo`3oool000000080oooo00<00000
0?ooo`3oool0103oool00`000000oooo0?ooo`1E0?ooo`@00000603oool;000000040?ooo`000000
00000?ooo`8000000P3oool4000000`0oooo0P0000030?ooo`8000000`3oool300000080oooo0P00
00030?ooo`<000003@3oool2000000030?ooo`00000000000080000000<0oooo0000000000003`3o
ool2000000050?ooo`00000000000000003oool00P00000=0?ooo`<00000103oool2000000040?oo
o`0000000000000000<0oooo0P0000030?ooo`800000303oool2000000030?ooo`00000000000080
oooo0P000000103oool000000000003oool;000001P0oooo1000001F0?ooo`030000003oool0oooo
00T0oooo000=0?ooo`050000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo05H0oooo0`00000F0?ooo`d00000603oool2000000P0oooo0P0000030?oo
o`@0000000@0oooo00000000000000003@3oool2000000030?ooo`000000000000<0oooo0P00000?
0?ooo`8000000P3oool2000000030?ooo`000000000000d0oooo0`0000020?ooo`@000000P3oool2
000000P0oooo0P00000I0?ooo``000005`3oool3000005L0oooo00<000000?ooo`3oool02@3oool0
00d0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool0E`3oool2000001H0oooo0P00000d0?ooo`800000A@3oool2000003<0oooo0P00000G
0?ooo`800000F03oool00`000000oooo0?ooo`090?ooo`002`3oool3000000<0oooo00@000000?oo
o`3oool000001P3oool00`000000oooo0?ooo`1I0?ooo`800000603oool500000080oooo10000000
0`3oool00000000000020?ooo`8000002@3oool200000080oooo0P0000020?ooo`8000000`3oool2
00000080oooo0P0000030?ooo`800000103oool2000000P0oooo10000000103oool000000000003o
ool2000001L0oooo0P000000103oool000000000003oool3000000T0oooo0P0000040?ooo`800000
0P3oool200000080oooo0`0000030?ooo`8000000P3oool200000080oooo0P0000090?ooo`800000
0P3oool4000000030?ooo`00000000000080oooo1@00000G0?ooo`800000F`3oool00`000000oooo
0?ooo`090?ooo`003@3oool00`000000oooo0?ooo`020?ooo`8000001`3oool3000005T0oooo0P00
00000`3oool000000000000E0?ooo`D000000P3oool4000000030?ooo`00000000000080oooo0P00
00090?ooo`8000000P3oool200000080oooo0P0000000`3oool000000000000200000080oooo0P00
00030?ooo`800000103oool2000000P0oooo1`0000000`3oool00000000000040?ooo`8000002P3o
ool2000000D0oooo0P0000000`3oool0000000000004000000T0oooo0P0000040?ooo`8000000P3o
ool200000080oooo1@0000000`3oool00000000000020?ooo`8000000P3oool2000000T0oooo0P00
00020?ooo`@0000000<0oooo0000000000000P3oool5000001D0oooo1000001I0?ooo`<000002`3o
ool001/0oooo00<000000?ooo`3oool0G03oool3000001D0oooo300000000`3oool0000000000009
0?ooo`80000000<0oooo0000000000000P3oool200000080oooo100000000`3oool0000000000002
0?ooo`8000004@3oool200000080oooo100000070?ooo`8000002P3oool2000000P0oooo10000002
0?ooo`800000403oool200000080oooo0P0000020?ooo`<000000`3oool2000000030?ooo`000000
00000080oooo0P0000080?ooo`80000000@0oooo000000000000oooo2@00000F0?ooo`<00000G@3o
ool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`1L0?ooo`@000005P3oool:
000000030?ooo`000000000000D0oooo0P0000020?ooo`80000000<0oooo0000000000000P3oool2
00000080oooo0P0000000`3oool000000000000200000080oooo0P00000A0?ooo`P00000203oool2
000000T0oooo0P0000080?ooo`P00000403oool200000080oooo0P0000000`3oool0000000000002
000000<0oooo0P0000000`3oool00000000000020?ooo`8000000P3oool2000000@0oooo0P000000
103oool000000000003oool8000001H0oooo1000001M0?ooo`030000003oool0oooo00T0oooo000K
0?ooo`030000003oool0oooo05d0oooo1000000E0?ooo`L0000000<0oooo000000000000203oool4
000000030?ooo`00000000000080oooo0P0000070?ooo`80000000<0oooo0000000000000P3oool2
00000140oooo1@0000080?ooo`@000002@3oool3000000T0oooo1@00000A0?ooo`8000000P3oool2
000000030?ooo`000000000000L0oooo0P000000103oool000000000003oool2000000030?ooo`00
0000000000P0oooo2P00000E0?ooo`@00000GP3oool00`000000oooo0?ooo`090?ooo`006`3oool0
0`000000oooo0?ooo`1N0?ooo`@000005P3oool5000000030?ooo`000000000000800000203oool2
000000030?ooo`00000000000080oooo0P00000:0?ooo`8000000P3oool2000000`0oooo0P000003
0?ooo`8000002`3oool2000000`0oooo0P00000<0?ooo`8000000`3oool2000000`0oooo0P000002
0?ooo`8000002P3oool2000000040?ooo`00000000000?ooo`800000203oool2000000030?ooo`00
0000000000H000005P3oool4000005l0oooo00<000000?ooo`3oool02@3oool001/0oooo0`00001O
0?ooo`@000005P3oool6000000030?ooo`000000000000X0oooo0P000000103oool000000000003o
ool2000000X0oooo0P00000?0?ooo`8000000`3oool4000000X0oooo0P00000;0?ooo`8000002P3o
ool4000000<0oooo0P00000?0?ooo`8000002P3oool2000000040?ooo`00000000000?ooo`800000
2@3oool2000000040?ooo`00000000000?ooo`<000005`3oool3000005l0oooo0`00000;0?ooo`00
6`3oool00`000000oooo0?ooo`1P0?ooo`@000005@3oool6000000030?ooo`000000000000X0oooo
1@0000000`3oool00000000000020?ooo`8000001P3oool2000000/0oooo0P0000020?ooo`800000
0`3oool4000000X0oooo0P00000;0?ooo`8000002P3oool4000000<0oooo0P0000020?ooo`800000
2`3oool2000000H0oooo0P0000020?ooo`80000000<0oooo0000000000000`00000:0?ooo`@00000
00@0oooo00000000000000005@3oool500000640oooo00<000000?ooo`3oool02@3oool001/0oooo
00<000000?ooo`3oool0HP3oool4000001D0oooo1`00000;0?ooo`80000000@0oooo000000000000
oooo0P0000000`3oool000000000000C0?ooo`8000000P3oool200000140oooo0P00000;0?ooo`80
0000403oool2000000<0oooo0P00000C0?ooo`80000000@0oooo000000000000oooo0P0000000`3o
ool000000000000;0?ooo`L000005@3oool4000006<0oooo00<000000?ooo`3oool02@3oool001/0
oooo00<000000?ooo`3oool0H`3oool3000001H0oooo1@00000=0?ooo`@0000000<0oooo00000000
00005P3oool200000080oooo0P00000A0?ooo`8000002`3oool200000100oooo0P0000020?ooo`80
00005`3oool2000000030?ooo`0000000000008000003@3oool4000001H0oooo1000001T0?ooo`03
0000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo06@0oooo1000000D0?ooo`@00000
3P3oool2000000030?ooo`000000000001D0oooo1@0000030?ooo`8000002P3oool200000080oooo
1000000;0?ooo`<000000`3oool2000000X0oooo0P0000020?ooo`H000005@3oool2000000030?oo
o`000000000000d0oooo1000000E0?ooo`@00000I@3oool00`000000oooo0?ooo`090?ooo`006`3o
ool3000006D0oooo1000000E0?ooo`@000003`3oool300000180oooo1P0000040?ooo`8000002P3o
ool200000080oooo1000000;0?ooo`<000000P3oool3000000T0oooo0`0000040?ooo`H000004P3o
ool3000000l0oooo0`00000F0?ooo`@00000I03oool3000000/0oooo000K0?ooo`030000003oool0
oooo06H0oooo1000000F0?ooo`800000403oool200000180oooo0P0000030?ooo`8000000`3oool2
000000X0oooo0P00000F0?ooo`8000002P3oool2000000@0oooo0P0000030?ooo`8000004P3oool2
00000100oooo0P00000F0?ooo`<00000J03oool00`000000oooo0?ooo`090?ooo`006`3oool00`00
0000oooo0?ooo`1W0?ooo`@00000?03oool600000100oooo0P0000000`3oool000000000000>0?oo
o`8000000P3oool200000100oooo1P00000l0?ooo`<00000J@3oool00`000000oooo0?ooo`090?oo
o`006`3oool00`000000oooo0?ooo`1X0?ooo`@00000=`3oool200000080oooo100000000`3oool0
00000000000<0?ooo`80000000@0oooo000000000000oooo0P00000>0?ooo`8000000P3oool20000
00030?ooo`000000000000`0oooo0P0000000`3oool000000000000200000080oooo0P00000g0?oo
o`<00000JP3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`1Y0?ooo`@0
0000<`3oool2000000030?ooo`00000000000080oooo0P0000030?ooo`800000303oool2000000D0
oooo0P00000=0?ooo`8000001@3oool2000000`0oooo0P0000020?ooo`8000000P3oool300000003
0?ooo`000000000003<0oooo0`00001[0?ooo`030000003oool0oooo00T0oooo000K0?ooo`<00000
J`3oool300000340oooo0`0000001@3oool00000000000000000oooo0080000000<0oooo00000000
00000P00000<0?ooo`8000001@3oool2000000d0oooo0P0000040?ooo`800000303oool200000003
0?ooo`0000000000008000000P3oool200000080oooo0`00000`0?ooo`@00000JP3oool3000000/0
oooo000K0?ooo`030000003oool0oooo06`0oooo1000000_0?ooo`8000000`3oool200000080oooo
1P00000<0?ooo`8000006P3oool2000000`0oooo0P0000000`3oool0000000000002000000040?oo
o`0000000000000000<0oooo0P00000_0?ooo`@00000K@3oool00`000000oooo0?ooo`090?ooo`00
6`3oool00`000000oooo0?ooo`1]0?ooo`@00000;`3oool200000080oooo0P0000020?ooo`<00000
2P3oool2000000@0oooo0P00000I0?ooo`8000000`3oool2000000/0oooo0`0000000`3oool00000
000000030?ooo`800000;`3oool4000006h0oooo00<000000?ooo`3oool02@3oool001/0oooo00<0
00000?ooo`3oool0KP3oool3000002/0oooo0P0000020?ooo`8000000`3oool200000080oooo0P00
000:0?ooo`800000103oool200000080oooo0P00000A0?ooo`8000000P3oool2000000<0oooo0P00
000;0?ooo`8000000P3oool2000000<0oooo0P0000000`3oool000000000000[0?ooo`@00000K`3o
ool00`000000oooo0?ooo`090?ooo`002`3oool400000080oooo0`0000070?ooo`030000003oool0
oooo06l0oooo1000000T0?ooo`@0000000<0oooo0000000000000`3oool200000080oooo0`00000>
0?ooo`8000000`3oool200000080oooo0P00000A0?ooo`8000000P3oool2000000<0oooo0P00000>
0?ooo`<000000P3oool200000080oooo0P0000020?ooo`@00000903oool300000740oooo00<00000
0?ooo`3oool02@3oool000d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?oo
o`@00000K`3oool4000002<0oooo1@0000000`3oool00000000000020?ooo`8000000`3oool20000
00h0oooo0P0000030?ooo`8000000P3oool200000140oooo0P000000103oool00000000000000003
0?ooo`8000003P3oool2000000<0oooo0P000000103oool000000000003oool600000280oooo1@00
001^0?ooo`@000002`3oool000d0oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo0780oooo0`00000S0?ooo`@0000000<0oooo0000000000000`3oool2
000000<0oooo0P00000=0?ooo`8000000`3oool200000080oooo0P0000000`3oool000000000000:
0?ooo`8000000P3oool2000000030?ooo`000000000000@0oooo0P00000=0?ooo`8000000`3oool2
00000080oooo0P000000103oool000000000003oool200000280oooo1000001c0?ooo`030000003o
ool0oooo00T0oooo000=0?ooo`040000003oool0oooo0?ooo`<000001`3oool00`000000oooo0?oo
o`1c0?ooo`<00000403oool4000000h0oooo1@0000000`3oool00000000000020?ooo`8000000`3o
ool2000000d0oooo0`0000090?ooo`8000002P3oool2000000X0oooo0`00000=0?ooo`8000000`3o
ool2000000040?ooo`00000000000?ooo`80000000@0oooo00000000000000003P3oool300000100
oooo1000001d0?ooo`030000003oool0oooo00T0oooo000=0?ooo`050000003oool0oooo0?ooo`00
00002@3oool00`000000oooo0?ooo`1d0?ooo`@000003P3oool:000000040?ooo`00000000000?oo
o`8000000`3oool800000080oooo0P00000B0?ooo`<000001@3oool2000000030?ooo`0000000000
00X0oooo0P0000020?ooo`8000001@3oool300000180oooo0P0000000`3oool00000000000040000
00030?ooo`00000000000080oooo0P0000020?ooo``000003`3oool4000007D0oooo00<000000?oo
o`3oool02@3oool000/0oooo0`0000030?ooo`030000003oool0oooo00L0oooo00<000000?ooo`3o
ool0M@3oool300000100oooo20000000103oool000000000003oool2000000@0oooo100000000`3o
ool00000000000020?ooo`8000004P3oool3000000D0oooo0P0000020?ooo`8000002@3oool20000
0080oooo0P0000050?ooo`<000004P3oool2000000040?ooo`00000000000?ooo`80000000<0oooo
0000000000000`3oool200000080oooo2`00000@0?ooo`<00000MP3oool00`000000oooo0?ooo`09
0?ooo`003@3oool010000000oooo0?ooo`3oool4000000H0oooo0`00001f0?ooo`@000003P3oool:
000000040?ooo`00000000000?ooo`8000000P3oool5000000030?ooo`00000000000080oooo0P00
000<0?ooo`8000000`3oool3000000D0oooo0P0000020?ooo`8000002@3oool2000000030?ooo`00
0000000000D0oooo100000020?ooo`8000003@3oool2000000030?ooo`00000000000080oooo1@00
00020?ooo`80000000@0oooo000000000000oooo1`0000000`3oool000000000000>0?ooo`@00000
M@3oool3000000/0oooo000K0?ooo`030000003oool0oooo07P0oooo0`00000?0?ooo`P0000000@0
oooo000000000000oooo0P0000030?ooo`L000000P3oool2000000`0oooo0P0000040?ooo`800000
1@3oool200000100oooo0P0000050?ooo`800000103oool2000000d0oooo0P000000103oool00000
0000003oool5000000<0oooo0P000000103oool000000000003oool7000000l0oooo1000001i0?oo
o`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo07T0oooo0`00000@0?ooo`D0
000000<0oooo0000000000000P0000020?ooo`800000103oool2000000030?ooo`00000000000100
oooo0P0000040?ooo`8000007@3oool2000000@0oooo0P00000@0?ooo`80000000<0oooo00000000
0000103oool2000000040?ooo`00000000000?ooo`P000003`3oool3000007/0oooo00<000000?oo
o`3oool02@3oool001/0oooo00<000000?ooo`3oool0N@3oool400000100oooo2@0000020?ooo`80
00000P3oool2000000030?ooo`000000000000<000002`3oool200000080oooo0`0000030?ooo`80
00007@3oool2000000<0oooo0`0000020?ooo`8000002`3oool7000000<0oooo0P000000103oool0
00000000003oool7000000l0oooo1000001k0?ooo`030000003oool0oooo00T0oooo000K0?ooo`03
0000003oool0oooo07/0oooo0`00000?0?ooo`P0000000<0oooo000000000000103oool200000003
0?ooo`000000000000<000002`3oool2000000<0oooo0P0000030?ooo`8000003@3oool2000000h0
oooo0P0000030?ooo`8000000`3oool2000000/0oooo1`0000050?ooo`80000000<0oooo00000000
00001P00000>0?ooo`@00000O03oool00`000000oooo0?ooo`090?ooo`006`3oool3000007`0oooo
1000000>0?ooo`X0000000<0oooo0000000000000`3oool4000000d0oooo0P0000030?ooo`800000
0`3oool3000000H0oooo0P0000040?ooo`8000001@3oool2000000H0oooo0`0000030?ooo`800000
0P3oool2000000h0oooo100000020?ooo`8000000P3oool:000000h0oooo1000001k0?ooo`<00000
2`3oool001/0oooo00<000000?ooo`3oool0O@3oool4000000l0oooo1P0000000`3oool000000000
0002000000<0oooo0P00000?0?ooo`<000000P3oool2000000@0oooo0P0000060?ooo`8000002`3o
ool2000000H0oooo0P0000030?ooo`<000000P3oool200000100oooo0P0000020?ooo`@0000000<0
oooo0000000000001@00000?0?ooo`@00000OP3oool00`000000oooo0?ooo`090?ooo`006`3oool0
0`000000oooo0?ooo`1n0?ooo`@000003`3oool4000000040?ooo`000000000000000080oooo0P00
00020?ooo`8000003`3oool200000080oooo0P00000<0?ooo`800000103oool2000000D0oooo0P00
000;0?ooo`8000000P3oool3000000h0oooo0P0000030?ooo`80000000<0oooo0000000000000P00
00000`3oool0000000000002000000l0oooo0`0000200?ooo`030000003oool0oooo00T0oooo000K
0?ooo`030000003oool0oooo07l0oooo1000000?0?ooo`H0000000<0oooo0000000000000P000002
0?ooo`8000003`3oool2000000<0oooo0P00000;0?ooo`800000103oool2000000D0oooo0P00000;
0?ooo`8000000P3oool2000000l0oooo0P0000030?ooo`@0000000D0oooo00000000000000000?oo
o`02000000l0oooo100000200?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0
oooo0800oooo1000000?0?ooo`/000004`3oool200000080oooo0P0000060?ooo`8000000`3oool2
000000/0oooo0P0000030?ooo`8000001@3oool300000080oooo0P00000B0?ooo`80000000<0oooo
0000000000001P00000?0?ooo`@00000PP3oool00`000000oooo0?ooo`090?ooo`006`3oool00`00
0000oooo0?ooo`210?ooo`@000003P3oool=00000140oooo0P0000020?ooo`8000001P3oool20000
00<0oooo0P00000;0?ooo`8000000`3oool2000000D0oooo0P0000030?ooo`8000004@3oool30000
00030?ooo`000000000000H000003P3oool4000008<0oooo00<000000?ooo`3oool02@3oool001/0
oooo0`0000230?ooo`@00000403oool6000000030?ooo`000000000001H0oooo0P0000050?ooo`80
00000`3oool2000000/0oooo0P0000030?ooo`8000001@3oool2000001H0oooo0P0000000`3oool0
00000000000400000100oooo100000220?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool0
Q03oool4000000l0oooo1`00000H0?ooo`8000001@3oool2000000<0oooo0P00000:0?ooo`<00000
0`3oool2000000D0oooo0P00000G0?ooo`P000003`3oool4000008D0oooo00<000000?ooo`3oool0
2@3oool001/0oooo00<000000?ooo`3oool0Q@3oool4000000l0oooo1P00000O0?ooo`8000000`3o
ool2000000X0oooo0P0000030?ooo`8000007`3oool7000000l0oooo100000260?ooo`030000003o
ool0oooo00T0oooo000K0?ooo`030000003oool0oooo08H0oooo1000000?0?ooo`H000005`3oool2
000000D0oooo0P0000030?ooo`<000002@3oool2000000<0oooo0P0000060?ooo`8000005P3oool7
000000l0oooo100000270?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo
08L0oooo1000000?0?ooo`D000004P3oool2000000<0oooo0P0000060?ooo`8000000`3oool20000
00T0oooo0P0000030?ooo`8000001@3oool3000000<0oooo0P00000A0?ooo`H000003`3oool40000
08P0oooo00<000000?ooo`3oool02@3oool001/0oooo0`0000290?ooo`<00000403oool300000180
oooo0P0000030?ooo`8000001P3oool2000001<0oooo0P0000050?ooo`800000103oool200000180
oooo0P00000@0?ooo`@00000R03oool3000000/0oooo000K0?ooo`030000003oool0oooo08X0oooo
0`00000@0?ooo`8000003P3oool200000080oooo0P0000040?ooo`8000001@3oool200000180oooo
0P0000060?ooo`8000000`3oool2000000<0oooo0P00000>0?ooo`8000003`3oool3000008`0oooo
00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0R`3oool300000100oooo0P00
000:0?ooo`80000000D0oooo00000000000000000?ooo`03000000<0oooo0P0000050?ooo`800000
4P3oool2000000H0oooo0P0000030?ooo`8000000P3oool3000000030?ooo`000000000000X0oooo
0P00000?0?ooo`@00000S03oool00`000000oooo0?ooo`090?ooo`002`3oool4000000<0oooo0P00
00070?ooo`030000003oool0oooo08`0oooo1000000>0?ooo`8000002P3oool3000000030?ooo`00
000000000080oooo0P0000030?ooo`8000001P3oool200000140oooo0P0000050?ooo`8000000`3o
ool300000080oooo0P000000103oool0000000000000000:0?ooo`8000003P3oool4000008d0oooo
00<000000?ooo`3oool02@3oool000/0oooo00<000000?ooo`3oool00`3oool010000000oooo0?oo
o`0000060?ooo`030000003oool0oooo08d0oooo1000000J0?ooo`80000000D0oooo000000000000
00000?ooo`02000000<0oooo0`0000050?ooo`8000004@3oool2000000D0oooo0P0000030?ooo`80
00000P3oool6000001X0oooo1000002>0?ooo`030000003oool0oooo00T0oooo000<0?ooo`030000
003oool0oooo0080oooo00@000000?ooo`3oool000001P3oool4000008d0oooo1000000I0?ooo`80
00000P3oool200000080oooo0P0000030?ooo`8000007P3oool3000000<0oooo0P0000001@3oool0
0000000000000000oooo008000006P3oool4000008`0oooo1000000;0?ooo`003@3oool01@000000
oooo0?ooo`3oool000000080oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`2?0?oo
o`@00000703oool3000000040?ooo`000000000000000080oooo0P0000070?ooo`800000303oool2
000000L0oooo0P0000030?ooo`<0000000<0oooo0000000000007@3oool300000940oooo00<00000
0?ooo`3oool02@3oool000h0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`04
0?ooo`030000003oool0oooo0900oooo1000000L0?ooo`8000000P3oool2000000<0oooo0P000006
0?ooo`<000002`3oool2000000L0oooo0P0000030?ooo`80000000@0oooo0000000000000000703o
ool300000980oooo00<000000?ooo`3oool02@3oool000/0oooo00@000000?ooo`3oool000000P3o
ool010000000oooo0?ooo`0000060?ooo`030000003oool0oooo0980oooo0`00000K0?ooo`<00000
00<0oooo0000000000000`3oool200000080oooo0P0000030?ooo`8000002`3oool2000000<0oooo
0P0000020?ooo`8000000P3oool3000000030?ooo`000000000001/0oooo1000002C0?ooo`030000
003oool0oooo00T0oooo000<0?ooo`800000103oool2000000L0oooo00<000000?ooo`3oool0TP3o
ool4000001/0oooo0P000000103oool000000000000000020?ooo`<0000000<0oooo000000000000
0`3oool2000000/0oooo0P0000030?ooo`80000000@0oooo00000000000000000P3oool200000080
oooo0P00000J0?ooo`@00000U03oool00`000000oooo0?ooo`090?ooo`006`3oool3000009@0oooo
1000000M0?ooo`8000000`3oool2000000H0oooo0P00000;0?ooo`8000001P3oool200000080oooo
0`00000M0?ooo`<00000U03oool3000000/0oooo000K0?ooo`030000003oool0oooo09D0oooo1000
000L0?ooo`<000000P3oool2000000H0oooo0P00000;0?ooo`8000001P3oool200000080oooo0P00
000M0?ooo`<00000U`3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`2F
0?ooo`@00000703oool2000000X0oooo0P00000:0?ooo`<000002P3oool2000001`0oooo1000002G
0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo09L0oooo1000000K0?oo
o`<000002@3oool2000000X0oooo0P00000:0?ooo`<000006`3oool3000009T0oooo00<000000?oo
o`3oool02@3oool001/0oooo00<000000?ooo`3oool0V@3oool3000001/0oooo0P0000090?ooo`<0
00002@3oool2000000X0oooo0P00000J0?ooo`@00000VP3oool00`000000oooo0?ooo`090?ooo`00
6`3oool3000009T0oooo1000000J0?ooo`8000002P3oool2000000T0oooo0P00000:0?ooo`800000
6@3oool4000009T0oooo0`00000;0?ooo`006`3oool00`000000oooo0?ooo`2K0?ooo`<00000EP3o
ool4000009`0oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0W03oool4
000005<0oooo0`00002N0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo
09d0oooo1000001A0?ooo`@00000WP3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000
oooo0?ooo`2N0?ooo`@00000C`3oool4000009l0oooo00<000000?ooo`3oool02@3oool001/0oooo
0`00002O0?ooo`@00000C@3oool3000009l0oooo0`00000;0?ooo`006`3oool00`000000oooo0?oo
o`2P0?ooo`@00000B`3oool300000:80oooo00<000000?ooo`3oool02@3oool001/0oooo00<00000
0?ooo`3oool0XP3oool3000004P0oooo1000002S0?ooo`030000003oool0oooo00T0oooo000K0?oo
o`030000003oool0oooo0:<0oooo0`0000160?ooo`@00000Y03oool00`000000oooo0?ooo`090?oo
o`006`3oool00`000000oooo0?ooo`2T0?ooo`@00000@`3oool400000:D0oooo00<000000?ooo`3o
ool02@3oool001/0oooo0`00002U0?ooo`@00000@@3oool400000:@0oooo0`00000;0?ooo`006`3o
ool00`000000oooo0?ooo`2W0?ooo`<00000?`3oool300000:P0oooo00<000000?ooo`3oool02@3o
ool001/0oooo00<000000?ooo`3oool0Y`3oool4000003`0oooo1000002Y0?ooo`030000003oool0
oooo00T0oooo000K0?ooo`030000003oool0oooo0:P0oooo1000000k0?ooo`<00000ZP3oool00`00
0000oooo0?ooo`090?ooo`002`3oool400000080oooo0`0000070?ooo`030000003oool0oooo0:T0
oooo1000000h0?ooo`@00000Z`3oool00`000000oooo0?ooo`090?ooo`002`3oool00`000000oooo
0?ooo`060?ooo`030000003oool0oooo00@0oooo1000002Z0?ooo`<00000=P3oool400000:T0oooo
1000000;0?ooo`00303oool00`000000oooo0?ooo`050?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool0[03oool4000003<0oooo1000002]0?ooo`030000003oool0oooo00T0oooo000=
0?ooo`040000003oool0oooo0?ooo`<000001`3oool00`000000oooo0?ooo`2]0?ooo`@00000<@3o
ool400000:h0oooo00<000000?ooo`3oool02@3oool000h0oooo00@000000?ooo`3oool000002@3o
ool00`000000oooo0?ooo`2^0?ooo`@00000;`3oool400000:l0oooo00<000000?ooo`3oool02@3o
ool000/0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`070?ooo`030000003o
ool0oooo0:l0oooo100000090?ooo`<000005@3oool2000000X0oooo0`00002a0?ooo`030000003o
ool0oooo00T0oooo000<0?ooo`8000000`3oool4000000H0oooo0`00002`0?ooo`@00000203oool3
000001D0oooo0P0000090?ooo`<00000/03oool3000000/0oooo000K0?ooo`030000003oool0oooo
0;40oooo1@00000;0?ooo`800000303oool3000000/0oooo1000002c0?ooo`030000003oool0oooo
00T0oooo000K0?ooo`030000003oool0oooo0;80oooo1000000;0?ooo`<000002`3oool3000000X0
oooo1000002d0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0;@0oooo
1000000:0?ooo`8000002`3oool2000000X0oooo1000002e0?ooo`030000003oool0oooo00T0oooo
000K0?ooo`030000003oool0oooo0;D0oooo100000090?ooo`<000002@3oool3000000T0oooo1000
002f0?ooo`030000003oool0oooo00T0oooo000K0?ooo`<00000]P3oool4000000P0oooo0`000009
0?ooo`8000002@3oool400000;D0oooo0`00000;0?ooo`006`3oool00`000000oooo0?ooo`2g0?oo
o`@000007@3oool300000;T0oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3o
ool0^03oool4000001X0oooo1000002j0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000
003oool0oooo0;X0oooo1000000G0?ooo`@00000^`3oool00`000000oooo0?ooo`090?ooo`006`3o
ool00`000000oooo0?ooo`2k0?ooo`@000005@3oool400000;`0oooo00<000000?ooo`3oool02@3o
ool001/0oooo0`00002l0?ooo`<00000503oool400000;/0oooo0`00000;0?ooo`006`3oool00`00
0000oooo0?ooo`2m0?ooo`@000004@3oool400000;h0oooo00<000000?ooo`3oool02@3oool001/0
oooo00<000000?ooo`3oool0_P3oool4000000l0oooo1000002o0?ooo`030000003oool0oooo00T0
oooo000K0?ooo`030000003oool0oooo0;l0oooo1000000<0?ooo`D00000`03oool00`000000oooo
0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`300?ooo`@000002P3oool400000<80oooo00<0
00000?ooo`3oool02@3oool001/0oooo0`0000310?ooo`@000002@3oool300000<40oooo0`00000;
0?ooo`006`3oool00`000000oooo0?ooo`330?ooo`8000002@3oool200000<@0oooo00<000000?oo
o`3oool02@3oool001/0oooo00<000000?ooo`3oool0o`3ooonE0?ooo`030000003oool0oooo00T0
oooo000K0?ooo`030000003oool0oooo0?l0ooooU@3oool00`000000oooo0?ooo`090?ooo`00303o
ool2000000@0oooo0P0000070?ooo`030000003oool0oooo0?l0ooooU@3oool00`000000oooo0?oo
o`090?ooo`002`3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000000H0oooo
00<000000?ooo`3oool0b03oool200000<X0oooo00<000000?ooo`3oool02@3oool000h0oooo00@0
00000?ooo`3oool000000P3oool00`000000oooo0?ooo`040?ooo`@00000a`3oool200000<L0oooo
1000000;0?ooo`00303oool2000000<0oooo00@000000?ooo`3oool000001P3oool00`000000oooo
0?ooo`3o0?oooiD0oooo00<000000?ooo`3oool02@3oool000h0oooo00@000000?ooo`3oool00000
0P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0?l0ooooU@3oool00`000000oooo
0?ooo`090?ooo`002`3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000000H0
oooo00<000000?ooo`3oool0o`3ooonE0?ooo`030000003oool0oooo00T0oooo000<0?ooo`800000
103oool2000000L0oooo00<000000?ooo`3oool0o`3ooonE0?ooo`030000003oool0oooo00T0oooo
000K0?ooo`<00000o`3ooonC0?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool0o`3ooonE
0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0?l0ooooU@3oool00`00
0000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`3o0?oooiD0oooo00<000000?ooo`3o
ool02@3oool001/0oooo00<000000?ooo`3oool0o`3ooonE0?ooo`030000003oool0oooo00T0oooo
000K0?ooo`<00000o`3ooonC0?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool0o`3ooonE
0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0?l0ooooU@3oool00`00
0000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`3o0?oooiD0oooo00<000000?ooo`3o
ool02@3oool001/0oooo00<000000?ooo`3oool0o`3ooonE0?ooo`030000003oool0oooo00T0oooo
000K0?ooo`<00000o`3ooonC0?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool0o`3ooonE
0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0?l0ooooU@3oool00`00
0000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`3o0?oooiD0oooo00<000000?ooo`3o
ool02@3oool001/0oooo00<000000?ooo`3oool0o`3ooonE0?ooo`030000003oool0oooo00T0oooo
000K0?ooo`<00000o`3ooonC0?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool0o`3ooonE
0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool0=`3oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo03L0oooo00<000000?oo
o`3oool0>03oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo03L0oooo00<000000?oo
o`3oool06P3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`030?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`03
0000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`03
0000003oool0oooo00T0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`090?ooo`00
303oool2000000<0oooo0`0000070?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0
203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool0
2@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00T0oooo000;0?ooo`040000003o
ool0oooo000000D0oooo00<000000?ooo`3oool0103ooooo000009T000002`3oool000h0oooo00<0
00000?ooo`3oool00`3oool00`000000oooo0?ooo`3o0?ooojP0oooo000<0?ooo`8000000`3oool3
00000?l0ooooZ`3oool000h0oooo00@000000?ooo`3oool00000o`3ooon]0?ooo`002`3oool01000
0000oooo0?ooo`0000020?ooo`030000003oool0oooo0?l0ooooZ`3oool00001\
\>"],
  ImageRangeCache->{{{0, 445}, {222, 0}} -> {-9.7434, -7.77001, 0.172064, 
  0.196645}}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(X = AnglePairsAndRationalApproxs[60, 10^\(-3\)]; \nLength[X]\), 
    \(Angles = \(Transpose[X]\)[\([1]\)]; \n
    ListPlot[Angles, \ AspectRatio -> 1/2, Frame -> True, \ Axes -> False, \ 
      PlotRange -> {{\(-5\), 65}, {\(-5\), 35}}]; 
    WriteData["\<~/ASP/WebServices/GrainBreaker/2DGeometries/Geo60.dat\>", 
      X]; \)}], "Input"],

Cell[BoxData[
    \(5113\)], "Output"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .5 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0714286 0.0142857 0.0625 0.0125 [
[.07143 -0.0125 -3 -9 ]
[.07143 -0.0125 3 0 ]
[.21429 -0.0125 -6 -9 ]
[.21429 -0.0125 6 0 ]
[.35714 -0.0125 -6 -9 ]
[.35714 -0.0125 6 0 ]
[.5 -0.0125 -6 -9 ]
[.5 -0.0125 6 0 ]
[.64286 -0.0125 -6 -9 ]
[.64286 -0.0125 6 0 ]
[.78571 -0.0125 -6 -9 ]
[.78571 -0.0125 6 0 ]
[.92857 -0.0125 -6 -9 ]
[.92857 -0.0125 6 0 ]
[ 0 0 -0.125 0 ]
[-0.0125 .0625 -6 -4.5 ]
[-0.0125 .0625 0 4.5 ]
[-0.0125 .125 -6 -4.5 ]
[-0.0125 .125 0 4.5 ]
[-0.0125 .1875 -12 -4.5 ]
[-0.0125 .1875 0 4.5 ]
[-0.0125 .25 -12 -4.5 ]
[-0.0125 .25 0 4.5 ]
[-0.0125 .3125 -12 -4.5 ]
[-0.0125 .3125 0 4.5 ]
[-0.0125 .375 -12 -4.5 ]
[-0.0125 .375 0 4.5 ]
[-0.0125 .4375 -12 -4.5 ]
[-0.0125 .4375 0 4.5 ]
[-0.0125 .5 -12 -4.5 ]
[-0.0125 .5 0 4.5 ]
[ 0 0 -0.125 0 ]
[ 0 .5 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .5 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.07143 0 m
.07143 .00625 L
s
[(0)] .07143 -0.0125 0 1 Mshowa
.21429 0 m
.21429 .00625 L
s
[(10)] .21429 -0.0125 0 1 Mshowa
.35714 0 m
.35714 .00625 L
s
[(20)] .35714 -0.0125 0 1 Mshowa
.5 0 m
.5 .00625 L
s
[(30)] .5 -0.0125 0 1 Mshowa
.64286 0 m
.64286 .00625 L
s
[(40)] .64286 -0.0125 0 1 Mshowa
.78571 0 m
.78571 .00625 L
s
[(50)] .78571 -0.0125 0 1 Mshowa
.92857 0 m
.92857 .00625 L
s
[(60)] .92857 -0.0125 0 1 Mshowa
.125 Mabswid
.1 0 m
.1 .00375 L
s
.12857 0 m
.12857 .00375 L
s
.15714 0 m
.15714 .00375 L
s
.18571 0 m
.18571 .00375 L
s
.24286 0 m
.24286 .00375 L
s
.27143 0 m
.27143 .00375 L
s
.3 0 m
.3 .00375 L
s
.32857 0 m
.32857 .00375 L
s
.38571 0 m
.38571 .00375 L
s
.41429 0 m
.41429 .00375 L
s
.44286 0 m
.44286 .00375 L
s
.47143 0 m
.47143 .00375 L
s
.52857 0 m
.52857 .00375 L
s
.55714 0 m
.55714 .00375 L
s
.58571 0 m
.58571 .00375 L
s
.61429 0 m
.61429 .00375 L
s
.67143 0 m
.67143 .00375 L
s
.7 0 m
.7 .00375 L
s
.72857 0 m
.72857 .00375 L
s
.75714 0 m
.75714 .00375 L
s
.81429 0 m
.81429 .00375 L
s
.84286 0 m
.84286 .00375 L
s
.87143 0 m
.87143 .00375 L
s
.9 0 m
.9 .00375 L
s
.04286 0 m
.04286 .00375 L
s
.01429 0 m
.01429 .00375 L
s
.95714 0 m
.95714 .00375 L
s
.98571 0 m
.98571 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
0 .0625 m
.00625 .0625 L
s
[(0)] -0.0125 .0625 1 0 Mshowa
0 .125 m
.00625 .125 L
s
[(5)] -0.0125 .125 1 0 Mshowa
0 .1875 m
.00625 .1875 L
s
[(10)] -0.0125 .1875 1 0 Mshowa
0 .25 m
.00625 .25 L
s
[(15)] -0.0125 .25 1 0 Mshowa
0 .3125 m
.00625 .3125 L
s
[(20)] -0.0125 .3125 1 0 Mshowa
0 .375 m
.00625 .375 L
s
[(25)] -0.0125 .375 1 0 Mshowa
0 .4375 m
.00625 .4375 L
s
[(30)] -0.0125 .4375 1 0 Mshowa
0 .5 m
.00625 .5 L
s
[(35)] -0.0125 .5 1 0 Mshowa
.125 Mabswid
0 .0125 m
.00375 .0125 L
s
0 .025 m
.00375 .025 L
s
0 .0375 m
.00375 .0375 L
s
0 .05 m
.00375 .05 L
s
0 .075 m
.00375 .075 L
s
0 .0875 m
.00375 .0875 L
s
0 .1 m
.00375 .1 L
s
0 .1125 m
.00375 .1125 L
s
0 .1375 m
.00375 .1375 L
s
0 .15 m
.00375 .15 L
s
0 .1625 m
.00375 .1625 L
s
0 .175 m
.00375 .175 L
s
0 .2 m
.00375 .2 L
s
0 .2125 m
.00375 .2125 L
s
0 .225 m
.00375 .225 L
s
0 .2375 m
.00375 .2375 L
s
0 .2625 m
.00375 .2625 L
s
0 .275 m
.00375 .275 L
s
0 .2875 m
.00375 .2875 L
s
0 .3 m
.00375 .3 L
s
0 .325 m
.00375 .325 L
s
0 .3375 m
.00375 .3375 L
s
0 .35 m
.00375 .35 L
s
0 .3625 m
.00375 .3625 L
s
0 .3875 m
.00375 .3875 L
s
0 .4 m
.00375 .4 L
s
0 .4125 m
.00375 .4125 L
s
0 .425 m
.00375 .425 L
s
0 .45 m
.00375 .45 L
s
0 .4625 m
.00375 .4625 L
s
0 .475 m
.00375 .475 L
s
0 .4875 m
.00375 .4875 L
s
.25 Mabswid
0 0 m
0 .5 L
s
.07143 .49375 m
.07143 .5 L
s
.21429 .49375 m
.21429 .5 L
s
.35714 .49375 m
.35714 .5 L
s
.5 .49375 m
.5 .5 L
s
.64286 .49375 m
.64286 .5 L
s
.78571 .49375 m
.78571 .5 L
s
.92857 .49375 m
.92857 .5 L
s
.125 Mabswid
.1 .49625 m
.1 .5 L
s
.12857 .49625 m
.12857 .5 L
s
.15714 .49625 m
.15714 .5 L
s
.18571 .49625 m
.18571 .5 L
s
.24286 .49625 m
.24286 .5 L
s
.27143 .49625 m
.27143 .5 L
s
.3 .49625 m
.3 .5 L
s
.32857 .49625 m
.32857 .5 L
s
.38571 .49625 m
.38571 .5 L
s
.41429 .49625 m
.41429 .5 L
s
.44286 .49625 m
.44286 .5 L
s
.47143 .49625 m
.47143 .5 L
s
.52857 .49625 m
.52857 .5 L
s
.55714 .49625 m
.55714 .5 L
s
.58571 .49625 m
.58571 .5 L
s
.61429 .49625 m
.61429 .5 L
s
.67143 .49625 m
.67143 .5 L
s
.7 .49625 m
.7 .5 L
s
.72857 .49625 m
.72857 .5 L
s
.75714 .49625 m
.75714 .5 L
s
.81429 .49625 m
.81429 .5 L
s
.84286 .49625 m
.84286 .5 L
s
.87143 .49625 m
.87143 .5 L
s
.9 .49625 m
.9 .5 L
s
.04286 .49625 m
.04286 .5 L
s
.01429 .49625 m
.01429 .5 L
s
.95714 .49625 m
.95714 .5 L
s
.98571 .49625 m
.98571 .5 L
s
.25 Mabswid
0 .5 m
1 .5 L
s
.99375 0 m
1 0 L
s
.99375 .0625 m
1 .0625 L
s
.99375 .125 m
1 .125 L
s
.99375 .1875 m
1 .1875 L
s
.99375 .25 m
1 .25 L
s
.99375 .3125 m
1 .3125 L
s
.99375 .375 m
1 .375 L
s
.99375 .4375 m
1 .4375 L
s
.125 Mabswid
.99625 .0125 m
1 .0125 L
s
.99625 .025 m
1 .025 L
s
.99625 .0375 m
1 .0375 L
s
.99625 .05 m
1 .05 L
s
.99625 .075 m
1 .075 L
s
.99625 .0875 m
1 .0875 L
s
.99625 .1 m
1 .1 L
s
.99625 .1125 m
1 .1125 L
s
.99625 .1375 m
1 .1375 L
s
.99625 .15 m
1 .15 L
s
.99625 .1625 m
1 .1625 L
s
.99625 .175 m
1 .175 L
s
.99625 .2 m
1 .2 L
s
.99625 .2125 m
1 .2125 L
s
.99625 .225 m
1 .225 L
s
.99625 .2375 m
1 .2375 L
s
.99625 .2625 m
1 .2625 L
s
.99625 .275 m
1 .275 L
s
.99625 .2875 m
1 .2875 L
s
.99625 .3 m
1 .3 L
s
.99625 .325 m
1 .325 L
s
.99625 .3375 m
1 .3375 L
s
.99625 .35 m
1 .35 L
s
.99625 .3625 m
1 .3625 L
s
.99625 .3875 m
1 .3875 L
s
.99625 .4 m
1 .4 L
s
.99625 .4125 m
1 .4125 L
s
.99625 .425 m
1 .425 L
s
.99625 .45 m
1 .45 L
s
.99625 .4625 m
1 .4625 L
s
.99625 .475 m
1 .475 L
s
.99625 .4875 m
1 .4875 L
s
.25 Mabswid
1 0 m
1 .5 L
s
0 0 m
1 0 L
1 .5 L
0 .5 L
closepath
clip
newpath
.008 w
.07143 .0625 Mdot
.07828 .06849 Mdot
.07848 .06867 Mdot
.0787 .06886 Mdot
.07893 .06906 Mdot
.07918 .06928 Mdot
.07944 .06951 Mdot
.07972 .06975 Mdot
.08002 .07002 Mdot
.08034 .0703 Mdot
.08069 .07061 Mdot
.08107 .07094 Mdot
.08148 .0713 Mdot
.08193 .07169 Mdot
.08242 .07212 Mdot
.08295 .07258 Mdot
.08354 .0731 Mdot
.0842 .07367 Mdot
.08493 .07431 Mdot
.08533 .0625 Mdot
.08533 .07466 Mdot
.08575 .0625 Mdot
.08575 .07503 Mdot
.08619 .0625 Mdot
.08619 .07542 Mdot
.08667 .0625 Mdot
.08667 .07584 Mdot
.08718 .0625 Mdot
.08718 .07628 Mdot
.08772 .0625 Mdot
.08772 .07676 Mdot
.0883 .0625 Mdot
.0883 .07727 Mdot
.08893 .0625 Mdot
.08893 .07781 Mdot
.0896 .0625 Mdot
.0896 .0784 Mdot
.09033 .0625 Mdot
.09033 .07904 Mdot
.09112 .0625 Mdot
.09112 .07973 Mdot
.09197 .0625 Mdot
.09197 .06849 Mdot
.09197 .08047 Mdot
.09258 .06867 Mdot
.09258 .08101 Mdot
.0929 .0625 Mdot
.0929 .08129 Mdot
.09323 .06886 Mdot
.09323 .08158 Mdot
.09393 .0625 Mdot
.09393 .06906 Mdot
.09393 .08219 Mdot
.09466 .06928 Mdot
.09466 .08283 Mdot
.09505 .0625 Mdot
.09505 .08317 Mdot
.09545 .06951 Mdot
.09545 .08352 Mdot
.09629 .0625 Mdot
.09629 .06975 Mdot
.09629 .08426 Mdot
.0972 .07002 Mdot
.0972 .08505 Mdot
.09767 .0625 Mdot
.09767 .08546 Mdot
.09817 .0703 Mdot
.09817 .0859 Mdot
.09922 .0625 Mdot
.09922 .07061 Mdot
.09922 .07466 Mdot
.09922 .08681 Mdot
.10006 .0625 Mdot
.10006 .07503 Mdot
.10006 .08755 Mdot
.10035 .07094 Mdot
.10035 .08781 Mdot
.10095 .0625 Mdot
.10095 .07542 Mdot
.10095 .08833 Mdot
.10158 .0713 Mdot
.10158 .08888 Mdot
.1019 .0625 Mdot
.1019 .07584 Mdot
.1019 .08916 Mdot
.10292 .0625 Mdot
.10292 .07169 Mdot
.10292 .07628 Mdot
.10292 .09005 Mdot
.104 .0625 Mdot
.104 .07676 Mdot
.104 .091 Mdot
.10438 .07212 Mdot
.10438 .09133 Mdot
.10516 .0625 Mdot
.10516 .07727 Mdot
.10516 .09202 Mdot
.10565 .06849 Mdot
.10565 .08047 Mdot
.10565 .09245 Mdot
.10599 .07258 Mdot
.10599 .09274 Mdot
.10641 .0625 Mdot
.10641 .07781 Mdot
.10641 .09311 Mdot
.10667 .06867 Mdot
.10667 .08101 Mdot
.10667 .09334 Mdot
.10776 .0625 Mdot
.10776 .06886 Mdot
.10776 .0731 Mdot
.10776 .0784 Mdot
.10776 .08158 Mdot
.10776 .09429 Mdot
.10891 .06906 Mdot
.10891 .08219 Mdot
.10891 .09529 Mdot
.10921 .0625 Mdot
.10921 .07904 Mdot
.10921 .09556 Mdot
.10972 .07367 Mdot
.10972 .096 Mdot
.11013 .06928 Mdot
.11013 .08283 Mdot
.11013 .09637 Mdot
.11078 .0625 Mdot
.11078 .07973 Mdot
.11078 .09693 Mdot
.11144 .06951 Mdot
.11144 .08352 Mdot
.11144 .09751 Mdot
.1119 .07431 Mdot
.1119 .09791 Mdot
.11249 .0625 Mdot
.11249 .08047 Mdot
.11249 .09843 Mdot
.11285 .06975 Mdot
.11285 .08426 Mdot
.11285 .09874 Mdot
.11309 .0625 Mdot
.11309 .07466 Mdot
.11309 .08681 Mdot
.11309 .09895 Mdot
.11435 .0625 Mdot
.11435 .07002 Mdot
.11435 .07503 Mdot
.11435 .08129 Mdot
.11435 .08505 Mdot
.11435 .08755 Mdot
.11435 .10006 Mdot
.11569 .0625 Mdot
.11569 .07542 Mdot
.11569 .08833 Mdot
.11569 .10123 Mdot
.11597 .0703 Mdot
.11597 .0859 Mdot
.11597 .10147 Mdot
.11639 .0625 Mdot
.11639 .08219 Mdot
.11639 .10184 Mdot
.11711 .0625 Mdot
.11711 .07584 Mdot
.11711 .08916 Mdot
.11711 .10247 Mdot
.11771 .07061 Mdot
.11771 .08681 Mdot
.11771 .103 Mdot
.11863 .0625 Mdot
.11863 .07628 Mdot
.11863 .08317 Mdot
.11863 .09005 Mdot
.11863 .1038 Mdot
.11932 .06849 Mdot
.11932 .08047 Mdot
.11932 .09245 Mdot
.11932 .1044 Mdot
.11959 .07094 Mdot
.11959 .08781 Mdot
.11959 .10464 Mdot
.12026 .0625 Mdot
.12026 .07676 Mdot
.12026 .091 Mdot
.12026 .10522 Mdot
.12074 .06867 Mdot
.12074 .08101 Mdot
.12074 .09334 Mdot
.12074 .10565 Mdot
.12111 .0625 Mdot
.12111 .08426 Mdot
.12111 .10597 Mdot
.12164 .0713 Mdot
.12164 .08888 Mdot
.12164 .10643 Mdot
.122 .07727 Mdot
.122 .09202 Mdot
.122 .10675 Mdot
.12226 .06886 Mdot
.12226 .08158 Mdot
.12226 .09429 Mdot
.12226 .10697 Mdot
.12386 .07169 Mdot
.12386 .07781 Mdot
.12386 .08219 Mdot
.12386 .08546 Mdot
.12386 .09005 Mdot
.12386 .09311 Mdot
.12386 .09529 Mdot
.12386 .10838 Mdot
.12558 .08283 Mdot
.12558 .09637 Mdot
.12558 .10988 Mdot
.12588 .0784 Mdot
.12588 .09429 Mdot
.12588 .11014 Mdot
.1263 .09133 Mdot
.1263 .11051 Mdot
.12694 .08681 Mdot
.12694 .09895 Mdot
.12694 .11107 Mdot
.12741 .08352 Mdot
.12741 .09751 Mdot
.12741 .11148 Mdot
.12805 .09556 Mdot
.12805 .11204 Mdot
.12862 .08755 Mdot
.12862 .10006 Mdot
.12862 .11254 Mdot
.12896 .09274 Mdot
.12896 .11284 Mdot
.12937 .09874 Mdot
.12937 .1132 Mdot
.1304 .08833 Mdot
.1304 .09693 Mdot
.1304 .10123 Mdot
.1304 .1141 Mdot
.13147 .10006 Mdot
.13147 .11503 Mdot
.1319 .09429 Mdot
.1319 .11542 Mdot
.13229 .10247 Mdot
.13229 .11576 Mdot
.13295 .09245 Mdot
.13295 .09843 Mdot
.13295 .1044 Mdot
.13295 .11633 Mdot
.13372 .10147 Mdot
.13372 .11701 Mdot
.13431 .1038 Mdot
.13431 .11752 Mdot
.13478 .10565 Mdot
.13478 .11793 Mdot
.13516 .096 Mdot
.13516 .11826 Mdot
.13574 .10006 Mdot
.13574 .11877 Mdot
.13616 .103 Mdot
.13616 .11914 Mdot
.13647 .10522 Mdot
.13647 .11941 Mdot
.13672 .10697 Mdot
.13672 .11963 Mdot
.13879 .10184 Mdot
.13879 .10464 Mdot
.13879 .10675 Mdot
.13879 .12144 Mdot
.14076 .11107 Mdot
.14076 .12316 Mdot
.14098 .10988 Mdot
.14098 .12336 Mdot
.14127 .10838 Mdot
.14127 .12361 Mdot
.14164 .10643 Mdot
.14164 .12393 Mdot
.14214 .12437 Mdot
.14285 .11254 Mdot
.14285 .12499 Mdot
.14333 .11148 Mdot
.14333 .12541 Mdot
.14394 .11014 Mdot
.14394 .12595 Mdot
.14474 .10838 Mdot
.14474 .12665 Mdot
.14507 .1141 Mdot
.14507 .12693 Mdot
.14584 .1132 Mdot
.14584 .12761 Mdot
.14683 .11204 Mdot
.14683 .12847 Mdot
.14743 .11576 Mdot
.14743 .129 Mdot
.14813 .12962 Mdot
.14853 .11503 Mdot
.14853 .12996 Mdot
.14878 .11793 Mdot
.14878 .13019 Mdot
.14995 .11752 Mdot
.14995 .1312 Mdot
.15115 .11963 Mdot
.15115 .13225 Mdot
.15142 .11701 Mdot
.15142 .13249 Mdot
.15185 .13287 Mdot
.15264 .11941 Mdot
.15264 .13356 Mdot
.15334 .13417 Mdot
.15366 .12144 Mdot
.15366 .13445 Mdot
.15454 .12316 Mdot
.15454 .13522 Mdot
.15552 .12144 Mdot
.15552 .13608 Mdot
.15595 .13645 Mdot
.15634 .12336 Mdot
.15634 .1368 Mdot
.15704 .12499 Mdot
.15704 .13741 Mdot
.1579 .13817 Mdot
.15861 .13878 Mdot
.1592 .12541 Mdot
.1592 .1393 Mdot
.15969 .12693 Mdot
.15969 .13973 Mdot
.16048 .14042 Mdot
.16108 .14095 Mdot
.16155 .14136 Mdot
.16194 .14169 Mdot
.16225 .14197 Mdot
.16252 .129 Mdot
.16252 .1422 Mdot
.16274 .13019 Mdot
.16274 .1424 Mdot
.16553 .13225 Mdot
.16553 .14483 Mdot
.16827 .13522 Mdot
.16827 .14723 Mdot
.16849 .13445 Mdot
.16849 .14742 Mdot
.16874 .14765 Mdot
.16904 .14791 Mdot
.16941 .14823 Mdot
.16986 .14863 Mdot
.17043 .14913 Mdot
.17117 .13741 Mdot
.17117 .14978 Mdot
.17164 .15018 Mdot
.17218 .15066 Mdot
.17283 .15123 Mdot
.17363 .15192 Mdot
.17426 .15248 Mdot
.17461 .15279 Mdot
.175 .15312 Mdot
.17587 .15389 Mdot
.17664 .15456 Mdot
.17693 .15481 Mdot
.17754 .15535 Mdot
.17822 .15594 Mdot
.17859 .15627 Mdot
.17984 .15736 Mdot
.18103 .15841 Mdot
.18136 .15869 Mdot
.18194 .1592 Mdot
.18244 .15964 Mdot
.18324 .16034 Mdot
.18413 .16111 Mdot
.18477 .16167 Mdot
.18525 .16209 Mdot
.18563 .16243 Mdot
.18619 .16291 Mdot
.18658 .16325 Mdot
.18686 .1635 Mdot
.18876 .16517 Mdot
.19049 .16668 Mdot
.19072 .16688 Mdot
.19103 .16715 Mdot
.19145 .16752 Mdot
.19206 .16806 Mdot
.19249 .16843 Mdot
.19304 .16891 Mdot
.19379 .16956 Mdot
.1941 .16983 Mdot
.19484 .17049 Mdot
.19556 .17111 Mdot
.19584 .17136 Mdot
.19646 .17191 Mdot
.19722 .17256 Mdot
.19765 .17294 Mdot
.19793 .17319 Mdot
.19926 .17435 Mdot
.20071 .17562 Mdot
.20104 .17591 Mdot
.20159 .17639 Mdot
.20201 .17676 Mdot
.20261 .17729 Mdot
.20319 .17779 Mdot
.20356 .17811 Mdot
.204 .1785 Mdot
.20426 .07466 Mdot
.20426 .17873 Mdot
.20524 .17959 Mdot
.20636 .18056 Mdot
.2067 .18087 Mdot
.20736 .18144 Mdot
.20798 .18198 Mdot
.20827 .0625 Mdot
.20827 .07503 Mdot
.20827 .08755 Mdot
.20827 .18224 Mdot
.20911 .06849 Mdot
.20911 .08047 Mdot
.20911 .09245 Mdot
.20911 .18297 Mdot
.21011 .18384 Mdot
.21057 .18425 Mdot
.211 .18463 Mdot
.21181 .06928 Mdot
.21181 .18533 Mdot
.21253 .0625 Mdot
.21253 .07542 Mdot
.21253 .08833 Mdot
.21253 .18597 Mdot
.21287 .18626 Mdot
.2132 .06867 Mdot
.2132 .08101 Mdot
.2132 .09334 Mdot
.2132 .10565 Mdot
.2132 .18655 Mdot
.2138 .18708 Mdot
.21461 .18779 Mdot
.21533 .18842 Mdot
.21597 .18898 Mdot
.21655 .06951 Mdot
.21655 .08352 Mdot
.21655 .18948 Mdot
.21707 .0625 Mdot
.21707 .07584 Mdot
.21707 .08916 Mdot
.21707 .10247 Mdot
.21707 .18993 Mdot
.21753 .06886 Mdot
.21753 .08158 Mdot
.21753 .09429 Mdot
.21753 .10697 Mdot
.21753 .19034 Mdot
.21796 .0625 Mdot
.21796 .07466 Mdot
.21796 .08681 Mdot
.21796 .09895 Mdot
.21796 .11107 Mdot
.21796 .19072 Mdot
.21835 .19106 Mdot
.21904 .19166 Mdot
.21963 .19217 Mdot
.22013 .19262 Mdot
.22058 .193 Mdot
.22096 .19334 Mdot
.22131 .0625 Mdot
.22131 .19365 Mdot
.22162 .06975 Mdot
.22162 .08426 Mdot
.22162 .19391 Mdot
.22189 .0625 Mdot
.22189 .07628 Mdot
.22189 .09005 Mdot
.22189 .1038 Mdot
.22189 .19416 Mdot
.22214 .06906 Mdot
.22214 .08219 Mdot
.22214 .09529 Mdot
.22214 .10838 Mdot
.22214 .19438 Mdot
.22237 .07503 Mdot
.22237 .08755 Mdot
.22237 .10006 Mdot
.22237 .11254 Mdot
.22237 .12499 Mdot
.22237 .19457 Mdot
.22258 .09245 Mdot
.22258 .1044 Mdot
.22258 .11633 Mdot
.22258 .19475 Mdot
.22705 .06928 Mdot
.22705 .07002 Mdot
.22705 .0784 Mdot
.22705 .08283 Mdot
.22705 .08505 Mdot
.22705 .091 Mdot
.22705 .09637 Mdot
.22705 .19867 Mdot
.23158 .12316 Mdot
.23158 .13522 Mdot
.23158 .20263 Mdot
.23179 .10697 Mdot
.23179 .11963 Mdot
.23179 .13225 Mdot
.23179 .20282 Mdot
.23202 .10247 Mdot
.23202 .11576 Mdot
.23202 .129 Mdot
.23202 .20302 Mdot
.23228 .08352 Mdot
.23228 .09751 Mdot
.23228 .11148 Mdot
.23228 .12541 Mdot
.23228 .20324 Mdot
.23256 .0625 Mdot
.23256 .07727 Mdot
.23256 .09202 Mdot
.23256 .10675 Mdot
.23256 .12144 Mdot
.23256 .20349 Mdot
.23288 .0703 Mdot
.23288 .0859 Mdot
.23288 .10147 Mdot
.23288 .20377 Mdot
.23324 .0625 Mdot
.23324 .07904 Mdot
.23324 .09556 Mdot
.23324 .20408 Mdot
.23364 .0713 Mdot
.23364 .20444 Mdot
.2341 .20484 Mdot
.23463 .2053 Mdot
.23524 .20584 Mdot
.23596 .20647 Mdot
.23638 .12499 Mdot
.23638 .13741 Mdot
.23638 .20683 Mdot
.23683 .12144 Mdot
.23683 .13445 Mdot
.23683 .20722 Mdot
.23732 .11752 Mdot
.23732 .1312 Mdot
.23732 .20766 Mdot
.23787 .09874 Mdot
.23787 .1132 Mdot
.23787 .12761 Mdot
.23787 .20814 Mdot
.23848 .09311 Mdot
.23848 .10838 Mdot
.23848 .12361 Mdot
.23848 .20867 Mdot
.23916 .07061 Mdot
.23916 .08681 Mdot
.23916 .103 Mdot
.23916 .11914 Mdot
.23916 .20927 Mdot
.23994 .0625 Mdot
.23994 .07973 Mdot
.23994 .09693 Mdot
.23994 .20994 Mdot
.24081 .07169 Mdot
.24081 .09005 Mdot
.24081 .1424 Mdot
.24081 .15456 Mdot
.24081 .21071 Mdot
.24146 .12693 Mdot
.24146 .13973 Mdot
.24146 .15248 Mdot
.24146 .21128 Mdot
.24181 .21159 Mdot
.24218 .12336 Mdot
.24218 .1368 Mdot
.24218 .21191 Mdot
.24298 .11941 Mdot
.24298 .13356 Mdot
.24298 .2126 Mdot
.24386 .11503 Mdot
.24386 .12996 Mdot
.24386 .21337 Mdot
.24433 .21379 Mdot
.24484 .11014 Mdot
.24484 .12595 Mdot
.24484 .21424 Mdot
.24511 .14723 Mdot
.24511 .1592 Mdot
.24511 .21447 Mdot
.24595 .08781 Mdot
.24595 .10464 Mdot
.24595 .12144 Mdot
.24595 .14483 Mdot
.24595 .15736 Mdot
.24595 .2152 Mdot
.24687 .1422 Mdot
.24687 .15535 Mdot
.24687 .21601 Mdot
.2472 .0625 Mdot
.2472 .08047 Mdot
.2472 .09843 Mdot
.2472 .11633 Mdot
.2472 .2163 Mdot
.24789 .1393 Mdot
.24789 .15312 Mdot
.24789 .21691 Mdot
.24863 .07212 Mdot
.24863 .09133 Mdot
.24863 .21755 Mdot
.24902 .13608 Mdot
.24902 .15066 Mdot
.24902 .2179 Mdot
.25028 .0625 Mdot
.25028 .08317 Mdot
.25028 .13249 Mdot
.25028 .14978 Mdot
.25028 .16209 Mdot
.25028 .219 Mdot
.2514 .14742 Mdot
.2514 .16034 Mdot
.2514 .21998 Mdot
.2517 .11204 Mdot
.2517 .12847 Mdot
.2517 .22023 Mdot
.25221 .22068 Mdot
.25263 .14483 Mdot
.25263 .15841 Mdot
.25263 .22105 Mdot
.25329 .10643 Mdot
.25329 .12393 Mdot
.25329 .22163 Mdot
.25399 .14197 Mdot
.25399 .15627 Mdot
.25399 .22224 Mdot
.25448 .15456 Mdot
.25448 .16668 Mdot
.25448 .22267 Mdot
.25511 .10006 Mdot
.25511 .11877 Mdot
.25511 .22322 Mdot
.2555 .13878 Mdot
.2555 .15389 Mdot
.2555 .22357 Mdot
.25577 .15248 Mdot
.25577 .16517 Mdot
.25577 .2238 Mdot
.2572 .07258 Mdot
.2572 .09274 Mdot
.2572 .11284 Mdot
.2572 .13522 Mdot
.2572 .15018 Mdot
.2572 .15123 Mdot
.2572 .1635 Mdot
.2572 .22505 Mdot
.25854 .17111 Mdot
.25854 .22622 Mdot
.25877 .14765 Mdot
.25877 .16167 Mdot
.25877 .22643 Mdot
.25911 .1312 Mdot
.25911 .14823 Mdot
.25911 .22672 Mdot
.25962 .0625 Mdot
.25962 .08426 Mdot
.25962 .10597 Mdot
.25962 .22717 Mdot
.26 .16983 Mdot
.26 .2275 Mdot
.26052 .14483 Mdot
.26052 .15964 Mdot
.26052 .22795 Mdot
.26127 .12665 Mdot
.26127 .14483 Mdot
.26127 .22861 Mdot
.26161 .16843 Mdot
.26161 .22891 Mdot
.26247 .07431 Mdot
.26247 .14169 Mdot
.26247 .15736 Mdot
.26247 .22966 Mdot
.26338 .16688 Mdot
.26338 .23045 Mdot
.26375 .12144 Mdot
.26375 .23078 Mdot
.26409 .17435 Mdot
.26409 .23108 Mdot
.26466 .13817 Mdot
.26466 .15481 Mdot
.26466 .23158 Mdot
.26534 .16517 Mdot
.26534 .23217 Mdot
.26586 .17319 Mdot
.26586 .23263 Mdot
.26662 .09429 Mdot
.26662 .11542 Mdot
.26662 .23329 Mdot
.26714 .13417 Mdot
.26714 .15192 Mdot
.26714 .23375 Mdot
.26752 .16325 Mdot
.26752 .23408 Mdot
.26781 .17191 Mdot
.26781 .23434 Mdot
.26804 .0625 Mdot
.26804 .07466 Mdot
.26804 .08681 Mdot
.26804 .17873 Mdot
.26804 .23454 Mdot
.26997 .0625 Mdot
.26997 .06886 Mdot
.26997 .08158 Mdot
.26997 .08546 Mdot
.26997 .10838 Mdot
.26997 .12962 Mdot
.26997 .14863 Mdot
.26997 .16111 Mdot
.26997 .17779 Mdot
.26997 .23622 Mdot
.27187 .06849 Mdot
.27187 .08047 Mdot
.27187 .09245 Mdot
.27187 .1044 Mdot
.27187 .11633 Mdot
.27187 .18297 Mdot
.27187 .19475 Mdot
.27187 .23788 Mdot
.27209 .17676 Mdot
.27209 .23808 Mdot
.27237 .16891 Mdot
.27237 .23832 Mdot
.27273 .15869 Mdot
.27273 .23864 Mdot
.27323 .12437 Mdot
.27323 .14483 Mdot
.27323 .23908 Mdot
.27394 .0625 Mdot
.27394 .07503 Mdot
.27394 .08755 Mdot
.27394 .10006 Mdot
.27394 .18224 Mdot
.27394 .19457 Mdot
.27394 .2397 Mdot
.27443 .17562 Mdot
.27443 .24012 Mdot
.27505 .16715 Mdot
.27505 .24067 Mdot
.27588 .15594 Mdot
.27588 .24139 Mdot
.27622 .06906 Mdot
.27622 .08219 Mdot
.27622 .09529 Mdot
.27622 .18144 Mdot
.27622 .19438 Mdot
.27622 .24169 Mdot
.27702 .11826 Mdot
.27702 .14042 Mdot
.27702 .17435 Mdot
.27702 .2424 Mdot
.27778 .06867 Mdot
.27778 .08101 Mdot
.27778 .09334 Mdot
.27778 .10565 Mdot
.27778 .11793 Mdot
.27778 .18655 Mdot
.27778 .19867 Mdot
.27778 .24306 Mdot
.27807 .16517 Mdot
.27807 .24331 Mdot
.27872 .0625 Mdot
.27872 .07628 Mdot
.27872 .18056 Mdot
.27872 .19416 Mdot
.27872 .24388 Mdot
.27948 .15279 Mdot
.27948 .24455 Mdot
.27992 .17294 Mdot
.27992 .24493 Mdot
.2802 .0625 Mdot
.2802 .07542 Mdot
.2802 .08833 Mdot
.2802 .10123 Mdot
.2802 .1141 Mdot
.2802 .18597 Mdot
.2802 .19867 Mdot
.2802 .24517 Mdot
.2815 .11107 Mdot
.2815 .12316 Mdot
.2815 .13522 Mdot
.2815 .16291 Mdot
.2815 .17959 Mdot
.2815 .19391 Mdot
.2815 .20263 Mdot
.2815 .24631 Mdot
.28286 .06928 Mdot
.28286 .08283 Mdot
.28286 .09637 Mdot
.28286 .10988 Mdot
.28286 .19867 Mdot
.28286 .2475 Mdot
.28317 .17136 Mdot
.28317 .24777 Mdot
.28367 .14913 Mdot
.28367 .24821 Mdot
.28404 .09429 Mdot
.28404 .10697 Mdot
.28404 .11963 Mdot
.28404 .13225 Mdot
.28404 .20282 Mdot
.28404 .24854 Mdot
.28458 .19365 Mdot
.28458 .24901 Mdot
.28541 .16034 Mdot
.28541 .24973 Mdot
.2858 .0625 Mdot
.2858 .07676 Mdot
.2858 .091 Mdot
.2858 .19867 Mdot
.2858 .25007 Mdot
.28684 .08916 Mdot
.28684 .10247 Mdot
.28684 .11576 Mdot
.28684 .16956 Mdot
.28684 .20302 Mdot
.28684 .25099 Mdot
.28776 .12499 Mdot
.28776 .13741 Mdot
.28776 .20683 Mdot
.28776 .25179 Mdot
.28804 .19334 Mdot
.28804 .25204 Mdot
.28857 .14483 Mdot
.28857 .2525 Mdot
.28906 .07002 Mdot
.28906 .08505 Mdot
.28906 .19867 Mdot
.28906 .25293 Mdot
.28993 .06951 Mdot
.28993 .08352 Mdot
.28993 .09751 Mdot
.28993 .11148 Mdot
.28993 .15736 Mdot
.28993 .20324 Mdot
.28993 .25369 Mdot
.29069 .12144 Mdot
.29069 .13445 Mdot
.29069 .20722 Mdot
.29069 .25436 Mdot
.29104 .25466 Mdot
.29136 .13019 Mdot
.29136 .1424 Mdot
.29136 .21071 Mdot
.29136 .25494 Mdot
.29194 .193 Mdot
.29194 .25545 Mdot
.29271 .19867 Mdot
.29271 .25612 Mdot
.29336 .0625 Mdot
.29336 .07727 Mdot
.29336 .09202 Mdot
.29336 .10675 Mdot
.29336 .20349 Mdot
.29336 .25669 Mdot
.29392 .1038 Mdot
.29392 .11752 Mdot
.29392 .1312 Mdot
.29392 .20766 Mdot
.29392 .25718 Mdot
.29441 .12693 Mdot
.29441 .13973 Mdot
.29441 .21128 Mdot
.29441 .2576 Mdot
.29483 .14723 Mdot
.29483 .1592 Mdot
.29483 .21447 Mdot
.29483 .25798 Mdot
.29521 .15389 Mdot
.29521 .25831 Mdot
.29586 .25887 Mdot
.29638 .19262 Mdot
.29638 .25933 Mdot
.29681 .19867 Mdot
.29681 .25971 Mdot
.29718 .0703 Mdot
.29718 .0859 Mdot
.29718 .10147 Mdot
.29718 .20377 Mdot
.29718 .26003 Mdot
.29749 .09874 Mdot
.29749 .1132 Mdot
.29749 .12761 Mdot
.29749 .20814 Mdot
.29749 .2603 Mdot
.29776 .12336 Mdot
.29776 .1368 Mdot
.29776 .21191 Mdot
.29776 .26054 Mdot
.29799 .14483 Mdot
.29799 .15736 Mdot
.29799 .2152 Mdot
.29799 .26074 Mdot
.30146 .0625 Mdot
.30146 .07904 Mdot
.30146 .09311 Mdot
.30146 .09556 Mdot
.30146 .11941 Mdot
.30146 .12361 Mdot
.30146 .13356 Mdot
.30146 .1422 Mdot
.30146 .14978 Mdot
.30146 .16209 Mdot
.30146 .18425 Mdot
.30146 .19217 Mdot
.30146 .20408 Mdot
.30146 .20867 Mdot
.30146 .21601 Mdot
.30146 .26378 Mdot
.30481 .16668 Mdot
.30481 .22267 Mdot
.30481 .26671 Mdot
.30503 .14742 Mdot
.30503 .16034 Mdot
.30503 .21998 Mdot
.30503 .2669 Mdot
.30528 .1393 Mdot
.30528 .15312 Mdot
.30528 .21691 Mdot
.30528 .26712 Mdot
.30557 .11503 Mdot
.30557 .12996 Mdot
.30557 .21337 Mdot
.30557 .26737 Mdot
.3059 .07061 Mdot
.3059 .08681 Mdot
.3059 .103 Mdot
.3059 .11914 Mdot
.3059 .20927 Mdot
.3059 .26767 Mdot
.3063 .0713 Mdot
.3063 .20444 Mdot
.3063 .26801 Mdot
.30677 .19867 Mdot
.30677 .26842 Mdot
.30734 .19166 Mdot
.30734 .26892 Mdot
.30805 .17111 Mdot
.30805 .18297 Mdot
.30805 .22622 Mdot
.30805 .26955 Mdot
.30848 .16517 Mdot
.30848 .2238 Mdot
.30848 .26992 Mdot
.30896 .15841 Mdot
.30896 .22105 Mdot
.30896 .27034 Mdot
.30951 .13608 Mdot
.30951 .15066 Mdot
.30951 .2179 Mdot
.30951 .27082 Mdot
.31016 .11014 Mdot
.31016 .12595 Mdot
.31016 .21424 Mdot
.31016 .27139 Mdot
.31091 .0625 Mdot
.31091 .07973 Mdot
.31091 .09693 Mdot
.31091 .1141 Mdot
.31091 .20994 Mdot
.31091 .27205 Mdot
.31181 .16983 Mdot
.31181 .18224 Mdot
.31181 .20484 Mdot
.31181 .2275 Mdot
.31181 .27283 Mdot
.31251 .1635 Mdot
.31251 .22505 Mdot
.31251 .27344 Mdot
.31289 .19867 Mdot
.31289 .27378 Mdot
.31331 .15627 Mdot
.31331 .22224 Mdot
.31331 .27414 Mdot
.31423 .13249 Mdot
.31423 .14791 Mdot
.31423 .19106 Mdot
.31423 .219 Mdot
.31423 .27495 Mdot
.31502 .06867 Mdot
.31502 .08101 Mdot
.31502 .09334 Mdot
.31502 .17435 Mdot
.31502 .18655 Mdot
.31502 .23108 Mdot
.31502 .27565 Mdot
.31531 .10464 Mdot
.31531 .12144 Mdot
.31531 .2152 Mdot
.31531 .2759 Mdot
.31592 .16843 Mdot
.31592 .18144 Mdot
.31592 .22891 Mdot
.31592 .27643 Mdot
.31659 .07169 Mdot
.31659 .09005 Mdot
.31659 .10838 Mdot
.31659 .21071 Mdot
.31659 .27702 Mdot
.31695 .16167 Mdot
.31695 .17562 Mdot
.31695 .22643 Mdot
.31695 .27733 Mdot
.31814 .0625 Mdot
.31814 .07466 Mdot
.31814 .08681 Mdot
.31814 .09895 Mdot
.31814 .11107 Mdot
.31814 .15389 Mdot
.31814 .19072 Mdot
.31814 .2053 Mdot
.31814 .22357 Mdot
.31814 .23454 Mdot
.31814 .27837 Mdot
.31922 .0625 Mdot
.31922 .07542 Mdot
.31922 .18597 Mdot
.31922 .23263 Mdot
.31922 .27932 Mdot
.31952 .12847 Mdot
.31952 .14483 Mdot
.31952 .22023 Mdot
.31952 .27958 Mdot
.32003 .19867 Mdot
.32003 .28002 Mdot
.32046 .16688 Mdot
.32046 .18056 Mdot
.32046 .23045 Mdot
.32046 .2804 Mdot
.32115 .09245 Mdot
.32115 .09843 Mdot
.32115 .1044 Mdot
.32115 .11633 Mdot
.32115 .19475 Mdot
.32115 .2163 Mdot
.32115 .23788 Mdot
.32115 .281 Mdot
.32188 .15964 Mdot
.32188 .17435 Mdot
.32188 .22795 Mdot
.32188 .28165 Mdot
.32241 .06886 Mdot
.32241 .08158 Mdot
.32241 .09429 Mdot
.32241 .10697 Mdot
.32241 .19034 Mdot
.32241 .2821 Mdot
.3231 .0625 Mdot
.3231 .08219 Mdot
.3231 .21159 Mdot
.3231 .28271 Mdot
.32354 .15123 Mdot
.32354 .28309 Mdot
.32384 .18533 Mdot
.32384 .28336 Mdot
.32548 .08755 Mdot
.32548 .10006 Mdot
.32548 .11254 Mdot
.32548 .12393 Mdot
.32548 .14136 Mdot
.32548 .17959 Mdot
.32548 .19457 Mdot
.32548 .20584 Mdot
.32548 .2848 Mdot
.32709 .0625 Mdot
.32709 .07584 Mdot
.32709 .08916 Mdot
.32709 .10247 Mdot
.32709 .18993 Mdot
.32709 .28621 Mdot
.32738 .15736 Mdot
.32738 .17294 Mdot
.32738 .28646 Mdot
.3278 .09133 Mdot
.3278 .11051 Mdot
.3278 .28683 Mdot
.32845 .11793 Mdot
.32845 .13019 Mdot
.32845 .1424 Mdot
.32845 .19867 Mdot
.32845 .25494 Mdot
.32845 .2874 Mdot
.32894 .18463 Mdot
.32894 .28782 Mdot
.32961 .14823 Mdot
.32961 .28841 Mdot
.33023 .08219 Mdot
.33023 .09529 Mdot
.33023 .10838 Mdot
.33023 .12144 Mdot
.33023 .19438 Mdot
.33023 .28895 Mdot
.33062 .0731 Mdot
.33062 .28929 Mdot
.33107 .1785 Mdot
.33107 .28969 Mdot
.33133 .13522 Mdot
.33133 .14723 Mdot
.33133 .20263 Mdot
.33133 .21447 Mdot
.33133 .25798 Mdot
.33133 .28991 Mdot
.33226 .06951 Mdot
.33226 .08352 Mdot
.33226 .09751 Mdot
.33226 .13741 Mdot
.33226 .18948 Mdot
.33226 .29073 Mdot
.33326 .1141 Mdot
.33326 .12693 Mdot
.33326 .13973 Mdot
.33326 .19867 Mdot
.33326 .21128 Mdot
.33326 .2576 Mdot
.33326 .2916 Mdot
.33356 .17136 Mdot
.33356 .29186 Mdot
.33411 .29234 Mdot
.3346 .18384 Mdot
.3346 .29278 Mdot
.33546 .0625 Mdot
.33546 .07628 Mdot
.33546 .08317 Mdot
.33546 .09005 Mdot
.33546 .1038 Mdot
.33546 .11752 Mdot
.33546 .19416 Mdot
.33546 .25718 Mdot
.33546 .29353 Mdot
.33619 .13225 Mdot
.33619 .14483 Mdot
.33619 .2152 Mdot
.33619 .26074 Mdot
.33619 .29416 Mdot
.3365 .14483 Mdot
.3365 .16291 Mdot
.3365 .29444 Mdot
.33733 .17729 Mdot
.33733 .29516 Mdot
.33799 .0625 Mdot
.33799 .07727 Mdot
.33799 .09202 Mdot
.33799 .18898 Mdot
.33799 .25669 Mdot
.33799 .29574 Mdot
.33855 .10988 Mdot
.33855 .12336 Mdot
.33855 .1368 Mdot
.33855 .21191 Mdot
.33855 .26054 Mdot
.33855 .29623 Mdot
.33901 .14978 Mdot
.33901 .16209 Mdot
.33901 .219 Mdot
.33901 .26378 Mdot
.33901 .29663 Mdot
.33941 .29698 Mdot
.34004 .13287 Mdot
.34004 .29754 Mdot
.34054 .16956 Mdot
.34054 .29797 Mdot
.34093 .18297 Mdot
.34093 .25612 Mdot
.34093 .29831 Mdot
.34125 .06975 Mdot
.34125 .08426 Mdot
.34125 .09874 Mdot
.34125 .1132 Mdot
.34125 .20814 Mdot
.34125 .2603 Mdot
.34125 .29859 Mdot
.34152 .129 Mdot
.34152 .1422 Mdot
.34152 .21601 Mdot
.34152 .26378 Mdot
.34152 .29883 Mdot
.34174 .15456 Mdot
.34174 .16668 Mdot
.34174 .22267 Mdot
.34174 .26671 Mdot
.34174 .29902 Mdot
.34438 .0703 Mdot
.34438 .07367 Mdot
.34438 .096 Mdot
.34438 .11826 Mdot
.34438 .11941 Mdot
.34438 .14095 Mdot
.34438 .14742 Mdot
.34438 .16034 Mdot
.34438 .17111 Mdot
.34438 .17591 Mdot
.34438 .18842 Mdot
.34438 .21998 Mdot
.34438 .22622 Mdot
.34438 .26003 Mdot
.34438 .2669 Mdot
.34438 .30133 Mdot
.34714 .15248 Mdot
.34714 .16517 Mdot
.34714 .2238 Mdot
.34714 .26992 Mdot
.34714 .30375 Mdot
.34739 .12541 Mdot
.34739 .1393 Mdot
.34739 .21691 Mdot
.34739 .26712 Mdot
.34739 .30397 Mdot
.34768 .0625 Mdot
.34768 .07781 Mdot
.34768 .09311 Mdot
.34768 .10838 Mdot
.34768 .20867 Mdot
.34768 .26378 Mdot
.34768 .30422 Mdot
.34804 .18198 Mdot
.34804 .25971 Mdot
.34804 .30454 Mdot
.34849 .16752 Mdot
.34849 .30493 Mdot
.34906 .12761 Mdot
.34906 .30543 Mdot
.34981 .16983 Mdot
.34981 .2275 Mdot
.34981 .27283 Mdot
.34981 .30609 Mdot
.35029 .14483 Mdot
.35029 .15841 Mdot
.35029 .22105 Mdot
.35029 .27034 Mdot
.35029 .3065 Mdot
.35085 .10006 Mdot
.35085 .11503 Mdot
.35085 .12996 Mdot
.35085 .21337 Mdot
.35085 .26737 Mdot
.35085 .307 Mdot
.35154 .20408 Mdot
.35154 .26378 Mdot
.35154 .3076 Mdot
.35239 .06867 Mdot
.35239 .08101 Mdot
.35239 .17435 Mdot
.35239 .18655 Mdot
.35239 .23108 Mdot
.35239 .25933 Mdot
.35239 .27565 Mdot
.35239 .30834 Mdot
.35308 .1635 Mdot
.35308 .22505 Mdot
.35308 .27344 Mdot
.35308 .30894 Mdot
.35347 .15736 Mdot
.35347 .30928 Mdot
.35389 .13608 Mdot
.35389 .15066 Mdot
.35389 .2179 Mdot
.35389 .27082 Mdot
.35389 .30966 Mdot
.35488 .0625 Mdot
.35488 .07061 Mdot
.35488 .07466 Mdot
.35488 .08681 Mdot
.35488 .09895 Mdot
.35488 .103 Mdot
.35488 .11107 Mdot
.35488 .19072 Mdot
.35488 .20927 Mdot
.35488 .23454 Mdot
.35488 .26767 Mdot
.35488 .27837 Mdot
.35488 .31052 Mdot
.35577 .16843 Mdot
.35577 .18144 Mdot
.35577 .22891 Mdot
.35577 .27643 Mdot
.35577 .3113 Mdot
.3561 .19867 Mdot
.3561 .26378 Mdot
.3561 .31158 Mdot
.35682 .14197 Mdot
.35682 .15627 Mdot
.35682 .22224 Mdot
.35682 .27414 Mdot
.35682 .31221 Mdot
.35729 .09245 Mdot
.35729 .1044 Mdot
.35729 .11633 Mdot
.35729 .19475 Mdot
.35729 .23788 Mdot
.35729 .281 Mdot
.35729 .31263 Mdot
.35763 .16517 Mdot
.35763 .25887 Mdot
.35763 .31292 Mdot
.35808 .11014 Mdot
.35808 .12595 Mdot
.35808 .21424 Mdot
.35808 .27139 Mdot
.35808 .31332 Mdot
.35836 .07542 Mdot
.35836 .17319 Mdot
.35836 .18597 Mdot
.35836 .23263 Mdot
.35836 .27932 Mdot
.35836 .31357 Mdot
.35962 .12144 Mdot
.35962 .16167 Mdot
.35962 .20444 Mdot
.35962 .22643 Mdot
.35962 .26801 Mdot
.35962 .27733 Mdot
.35962 .31467 Mdot
.36086 .06886 Mdot
.36086 .08158 Mdot
.36086 .09429 Mdot
.36086 .10697 Mdot
.36086 .19034 Mdot
.36086 .23622 Mdot
.36086 .2821 Mdot
.36086 .31575 Mdot
.36114 .13249 Mdot
.36114 .14791 Mdot
.36114 .219 Mdot
.36114 .27495 Mdot
.36114 .31599 Mdot
.36156 .26378 Mdot
.36156 .31637 Mdot
.36232 .16688 Mdot
.36232 .18056 Mdot
.36232 .23045 Mdot
.36232 .2804 Mdot
.36232 .31703 Mdot
.36298 .0625 Mdot
.36298 .07973 Mdot
.36298 .09693 Mdot
.36298 .20994 Mdot
.36298 .27205 Mdot
.36298 .31761 Mdot
.36328 .08755 Mdot
.36328 .10006 Mdot
.36328 .11254 Mdot
.36328 .12499 Mdot
.36328 .19457 Mdot
.36328 .2848 Mdot
.36328 .31787 Mdot
.36407 .15389 Mdot
.36407 .22357 Mdot
.36407 .25831 Mdot
.36407 .27837 Mdot
.36407 .31856 Mdot
.36492 .18533 Mdot
.36492 .28336 Mdot
.36492 .31931 Mdot
.36529 .19867 Mdot
.36529 .26842 Mdot
.36529 .31962 Mdot
.36561 .11793 Mdot
.36561 .13019 Mdot
.36561 .1424 Mdot
.36561 .19867 Mdot
.36561 .25494 Mdot
.36561 .2874 Mdot
.36561 .31991 Mdot
.36619 .10464 Mdot
.36619 .12144 Mdot
.36619 .2152 Mdot
.36619 .2759 Mdot
.36619 .32041 Mdot
.36688 .15964 Mdot
.36688 .17435 Mdot
.36688 .22795 Mdot
.36688 .28165 Mdot
.36688 .32102 Mdot
.36743 .0625 Mdot
.36743 .07584 Mdot
.36743 .08916 Mdot
.36743 .10247 Mdot
.36743 .18993 Mdot
.36743 .28621 Mdot
.36743 .3215 Mdot
.36787 .13522 Mdot
.36787 .14723 Mdot
.36787 .20263 Mdot
.36787 .21447 Mdot
.36787 .25798 Mdot
.36787 .28991 Mdot
.36787 .32189 Mdot
.36824 .16243 Mdot
.36824 .26378 Mdot
.36824 .32221 Mdot
.36882 .20484 Mdot
.36882 .27283 Mdot
.36882 .32271 Mdot
.36924 .12847 Mdot
.36924 .14483 Mdot
.36924 .22023 Mdot
.36924 .27958 Mdot
.36924 .32309 Mdot
.36958 .17959 Mdot
.36958 .2848 Mdot
.36958 .32338 Mdot
.36984 .09529 Mdot
.36984 .10838 Mdot
.36984 .12144 Mdot
.36984 .19438 Mdot
.36984 .25436 Mdot
.36984 .28895 Mdot
.36984 .32361 Mdot
.37217 .07169 Mdot
.37217 .09005 Mdot
.37217 .1141 Mdot
.37217 .12693 Mdot
.37217 .13973 Mdot
.37217 .15123 Mdot
.37217 .18463 Mdot
.37217 .19166 Mdot
.37217 .21128 Mdot
.37217 .25007 Mdot
.37217 .2576 Mdot
.37217 .26892 Mdot
.37217 .27702 Mdot
.37217 .28309 Mdot
.37217 .28782 Mdot
.37217 .2916 Mdot
.37217 .32565 Mdot
.37441 .13225 Mdot
.37441 .14483 Mdot
.37441 .2152 Mdot
.37441 .26074 Mdot
.37441 .29416 Mdot
.37441 .32761 Mdot
.37466 .06951 Mdot
.37466 .08352 Mdot
.37466 .09751 Mdot
.37466 .18948 Mdot
.37466 .25369 Mdot
.37466 .29073 Mdot
.37466 .32783 Mdot
.37496 .17294 Mdot
.37496 .28646 Mdot
.37496 .32809 Mdot
.37535 .09843 Mdot
.37535 .11633 Mdot
.37535 .13417 Mdot
.37535 .2163 Mdot
.37535 .281 Mdot
.37535 .32843 Mdot
.37587 .19867 Mdot
.37587 .27378 Mdot
.37587 .32889 Mdot
.37658 .14978 Mdot
.37658 .16209 Mdot
.37658 .219 Mdot
.37658 .26378 Mdot
.37658 .29663 Mdot
.37658 .32951 Mdot
.37706 .09005 Mdot
.37706 .1038 Mdot
.37706 .11752 Mdot
.37706 .20766 Mdot
.37706 .25718 Mdot
.37706 .29353 Mdot
.37706 .32992 Mdot
.37764 .1785 Mdot
.37764 .24901 Mdot
.37764 .28969 Mdot
.37764 .33044 Mdot
.37839 .14136 Mdot
.37839 .2848 Mdot
.37839 .33109 Mdot
.37868 .16668 Mdot
.37868 .22267 Mdot
.37868 .26671 Mdot
.37868 .29902 Mdot
.37868 .33135 Mdot
.37936 .12336 Mdot
.37936 .1368 Mdot
.37936 .2053 Mdot
.37936 .21191 Mdot
.37936 .26054 Mdot
.37936 .27837 Mdot
.37936 .29623 Mdot
.37936 .33194 Mdot
.38021 .18384 Mdot
.38021 .25293 Mdot
.38021 .29278 Mdot
.38021 .33268 Mdot
.38071 .17111 Mdot
.38071 .18297 Mdot
.38071 .22622 Mdot
.38071 .26955 Mdot
.38071 .30133 Mdot
.38071 .33312 Mdot
.38127 .14823 Mdot
.38127 .16517 Mdot
.38127 .28841 Mdot
.38127 .33362 Mdot
.38159 .1422 Mdot
.38159 .15535 Mdot
.38159 .21601 Mdot
.38159 .26378 Mdot
.38159 .29883 Mdot
.38159 .33389 Mdot
.38267 .0625 Mdot
.38267 .07727 Mdot
.38267 .08219 Mdot
.38267 .09202 Mdot
.38267 .10184 Mdot
.38267 .10675 Mdot
.38267 .20349 Mdot
.38267 .25669 Mdot
.38267 .28271 Mdot
.38267 .29574 Mdot
.38267 .33483 Mdot
.38373 .14742 Mdot
.38373 .16034 Mdot
.38373 .21998 Mdot
.38373 .2669 Mdot
.38373 .30133 Mdot
.38373 .33576 Mdot
.38403 .17136 Mdot
.38403 .24777 Mdot
.38403 .29186 Mdot
.38403 .33603 Mdot
.38457 .19106 Mdot
.38457 .27495 Mdot
.38457 .3365 Mdot
.38503 .09874 Mdot
.38503 .1132 Mdot
.38503 .12761 Mdot
.38503 .20814 Mdot
.38503 .2603 Mdot
.38503 .29859 Mdot
.38503 .3369 Mdot
.3858 .11051 Mdot
.3858 .12962 Mdot
.3858 .16517 Mdot
.3858 .17779 Mdot
.3858 .2238 Mdot
.3858 .26992 Mdot
.3858 .28683 Mdot
.3858 .30375 Mdot
.3858 .33757 Mdot
.38666 .17729 Mdot
.38666 .25204 Mdot
.38666 .29516 Mdot
.38666 .33833 Mdot
.3873 .11941 Mdot
.3873 .13356 Mdot
.3873 .2126 Mdot
.3873 .26378 Mdot
.3873 .30133 Mdot
.3873 .33889 Mdot
.38779 .16983 Mdot
.38779 .18224 Mdot
.38779 .2275 Mdot
.38779 .27283 Mdot
.38779 .30609 Mdot
.38779 .33932 Mdot
.38818 .19867 Mdot
.38818 .28002 Mdot
.38818 .33966 Mdot
.38877 .13741 Mdot
.38877 .29073 Mdot
.38877 .34017 Mdot
.38918 .19867 Mdot
.38918 .25612 Mdot
.38918 .29831 Mdot
.38918 .34053 Mdot
.38948 .1393 Mdot
.38948 .15312 Mdot
.38948 .21691 Mdot
.38948 .26712 Mdot
.38948 .30397 Mdot
.38948 .3408 Mdot
.38972 .06867 Mdot
.38972 .08101 Mdot
.38972 .09334 Mdot
.38972 .10565 Mdot
.38972 .18655 Mdot
.38972 .24306 Mdot
.38972 .27565 Mdot
.38972 .30834 Mdot
.38972 .34101 Mdot
.39158 .0703 Mdot
.39158 .0859 Mdot
.39158 .08681 Mdot
.39158 .09895 Mdot
.39158 .10147 Mdot
.39158 .12316 Mdot
.39158 .15841 Mdot
.39158 .16291 Mdot
.39158 .19072 Mdot
.39158 .20377 Mdot
.39158 .22105 Mdot
.39158 .24631 Mdot
.39158 .26003 Mdot
.39158 .27034 Mdot
.39158 .27837 Mdot
.39158 .2848 Mdot
.39158 .29444 Mdot
.39158 .30133 Mdot
.39158 .3065 Mdot
.39158 .31052 Mdot
.39158 .34264 Mdot
.39361 .1635 Mdot
.39361 .17676 Mdot
.39361 .27344 Mdot
.39361 .30894 Mdot
.39361 .34441 Mdot
.39389 .09311 Mdot
.39389 .10838 Mdot
.39389 .12361 Mdot
.39389 .20867 Mdot
.39389 .26378 Mdot
.39389 .30422 Mdot
.39389 .34465 Mdot
.39426 .16956 Mdot
.39426 .25099 Mdot
.39426 .29797 Mdot
.39426 .34498 Mdot
.39478 .0731 Mdot
.39478 .09429 Mdot
.39478 .28929 Mdot
.39478 .34544 Mdot
.39556 .18144 Mdot
.39556 .24169 Mdot
.39556 .27643 Mdot
.39556 .3113 Mdot
.39556 .34611 Mdot
.3961 .12996 Mdot
.3961 .14483 Mdot
.3961 .21337 Mdot
.3961 .26737 Mdot
.3961 .307 Mdot
.3961 .34659 Mdot
.39682 .193 Mdot
.39682 .25545 Mdot
.39682 .34721 Mdot
.39743 .0625 Mdot
.39743 .07542 Mdot
.39743 .08833 Mdot
.39743 .10123 Mdot
.39743 .18597 Mdot
.39743 .24517 Mdot
.39743 .27932 Mdot
.39743 .31357 Mdot
.39743 .34776 Mdot
.3978 .12437 Mdot
.3978 .23908 Mdot
.3978 .34808 Mdot
.39822 .15066 Mdot
.39822 .27082 Mdot
.39822 .34844 Mdot
.39925 .09429 Mdot
.39925 .10697 Mdot
.39925 .11963 Mdot
.39925 .19034 Mdot
.39925 .19867 Mdot
.39925 .20282 Mdot
.39925 .24854 Mdot
.39925 .25971 Mdot
.39925 .2821 Mdot
.39925 .34934 Mdot
.40026 .15627 Mdot
.40026 .17049 Mdot
.40026 .23622 Mdot
.40026 .27414 Mdot
.40026 .35022 Mdot
.40065 .15279 Mdot
.40065 .24455 Mdot
.40065 .35057 Mdot
.401 .12499 Mdot
.401 .13741 Mdot
.401 .20683 Mdot
.401 .25179 Mdot
.401 .35087 Mdot
.40157 .0625 Mdot
.40157 .07904 Mdot
.40157 .09556 Mdot
.40157 .20408 Mdot
.40157 .26378 Mdot
.40157 .35137 Mdot
.40221 .17562 Mdot
.40221 .24012 Mdot
.40221 .27733 Mdot
.40221 .35193 Mdot
.40269 .1424 Mdot
.40269 .15456 Mdot
.40269 .21071 Mdot
.40269 .25494 Mdot
.40269 .35235 Mdot
.40335 .16034 Mdot
.40335 .24973 Mdot
.40335 .35293 Mdot
.40378 .103 Mdot
.40378 .11914 Mdot
.40378 .20927 Mdot
.40378 .26767 Mdot
.40378 .35331 Mdot
.40409 .18056 Mdot
.40409 .24388 Mdot
.40409 .35358 Mdot
.40432 .14723 Mdot
.40432 .1592 Mdot
.40432 .21447 Mdot
.40432 .25798 Mdot
.40432 .35378 Mdot
.4059 .06928 Mdot
.4059 .08283 Mdot
.4059 .09637 Mdot
.4059 .10988 Mdot
.4059 .12595 Mdot
.4059 .14169 Mdot
.4059 .18626 Mdot
.4059 .19867 Mdot
.4059 .2475 Mdot
.4059 .25466 Mdot
.4059 .27139 Mdot
.4059 .35517 Mdot
.40765 .10247 Mdot
.40765 .11576 Mdot
.40765 .129 Mdot
.40765 .20302 Mdot
.40765 .25099 Mdot
.40765 .35669 Mdot
.40793 .14791 Mdot
.40793 .16325 Mdot
.40793 .23408 Mdot
.40793 .35694 Mdot
.40833 .19262 Mdot
.40833 .25933 Mdot
.40833 .35729 Mdot
.40891 .0625 Mdot
.40891 .08546 Mdot
.40891 .10838 Mdot
.40891 .23622 Mdot
.40891 .3578 Mdot
.40933 .12144 Mdot
.40933 .13445 Mdot
.40933 .14742 Mdot
.40933 .20722 Mdot
.40933 .25436 Mdot
.40933 .35816 Mdot
.40987 .16891 Mdot
.40987 .23832 Mdot
.40987 .35864 Mdot
.41063 .19867 Mdot
.41063 .26378 Mdot
.41063 .3593 Mdot
.41095 .13973 Mdot
.41095 .15248 Mdot
.41095 .21128 Mdot
.41095 .2576 Mdot
.41095 .35958 Mdot
.41174 .11826 Mdot
.41174 .14042 Mdot
.41174 .17435 Mdot
.41174 .2424 Mdot
.41174 .36027 Mdot
.41251 .15736 Mdot
.41251 .16983 Mdot
.41251 .2275 Mdot
.41251 .26074 Mdot
.41251 .36095 Mdot
.41282 .0713 Mdot
.41282 .08888 Mdot
.41282 .10643 Mdot
.41282 .36122 Mdot
.41353 .06975 Mdot
.41353 .19391 Mdot
.41353 .24631 Mdot
.41353 .36183 Mdot
.41402 .16209 Mdot
.41402 .17435 Mdot
.41402 .23108 Mdot
.41402 .36227 Mdot
.41439 .14913 Mdot
.41439 .24821 Mdot
.41439 .36259 Mdot
.4149 .1141 Mdot
.4149 .1312 Mdot
.4149 .22672 Mdot
.4149 .36304 Mdot
.41524 .0625 Mdot
.41524 .07676 Mdot
.41524 .091 Mdot
.41524 .10522 Mdot
.41524 .19867 Mdot
.41524 .25007 Mdot
.41524 .36334 Mdot
.41548 .0625 Mdot
.41548 .07466 Mdot
.41548 .08681 Mdot
.41548 .17873 Mdot
.41548 .23454 Mdot
.41548 .31052 Mdot
.41548 .36355 Mdot
.41689 .06849 Mdot
.41689 .08047 Mdot
.41689 .09245 Mdot
.41689 .1044 Mdot
.41689 .11148 Mdot
.41689 .11633 Mdot
.41689 .12541 Mdot
.41689 .13817 Mdot
.41689 .15481 Mdot
.41689 .17811 Mdot
.41689 .18297 Mdot
.41689 .19475 Mdot
.41689 .20324 Mdot
.41689 .23158 Mdot
.41689 .23788 Mdot
.41689 .25369 Mdot
.41689 .281 Mdot
.41689 .31263 Mdot
.41689 .36478 Mdot
.41848 .1312 Mdot
.41848 .14483 Mdot
.41848 .20766 Mdot
.41848 .36617 Mdot
.41879 .16111 Mdot
.41879 .23622 Mdot
.41879 .36644 Mdot
.41925 .18533 Mdot
.41925 .36684 Mdot
.42001 .15018 Mdot
.42001 .1635 Mdot
.42001 .22505 Mdot
.42001 .30894 Mdot
.42001 .36751 Mdot
.4206 .16715 Mdot
.4206 .18297 Mdot
.4206 .24067 Mdot
.4206 .36803 Mdot
.42148 .15535 Mdot
.42148 .16843 Mdot
.42148 .19217 Mdot
.42148 .22891 Mdot
.42148 .3113 Mdot
.42148 .3688 Mdot
.42234 .18842 Mdot
.42234 .24493 Mdot
.42234 .36955 Mdot
.4229 .17319 Mdot
.4229 .23263 Mdot
.4229 .27932 Mdot
.4229 .31357 Mdot
.4229 .37004 Mdot
.42359 .37064 Mdot
.424 .0625 Mdot
.424 .07781 Mdot
.424 .19365 Mdot
.424 .371 Mdot
.42427 .06886 Mdot
.42427 .08158 Mdot
.42427 .09429 Mdot
.42427 .17779 Mdot
.42427 .19034 Mdot
.42427 .23622 Mdot
.42427 .2821 Mdot
.42427 .31575 Mdot
.42427 .37123 Mdot
.42559 .0625 Mdot
.42559 .07002 Mdot
.42559 .07503 Mdot
.42559 .08129 Mdot
.42559 .08505 Mdot
.42559 .08755 Mdot
.42559 .10006 Mdot
.42559 .11254 Mdot
.42559 .11503 Mdot
.42559 .11877 Mdot
.42559 .19457 Mdot
.42559 .19867 Mdot
.42559 .22322 Mdot
.42559 .2397 Mdot
.42559 .2848 Mdot
.42559 .307 Mdot
.42559 .31787 Mdot
.42559 .37239 Mdot
.42687 .10565 Mdot
.42687 .11793 Mdot
.42687 .13019 Mdot
.42687 .19867 Mdot
.42687 .24306 Mdot
.42687 .2874 Mdot
.42687 .31991 Mdot
.42687 .37351 Mdot
.42712 .12144 Mdot
.42712 .13608 Mdot
.42712 .2179 Mdot
.42712 .30966 Mdot
.42712 .37373 Mdot
.42749 .12665 Mdot
.42749 .14483 Mdot
.42749 .22861 Mdot
.42749 .37405 Mdot
.4281 .13522 Mdot
.4281 .14723 Mdot
.4281 .20263 Mdot
.4281 .28991 Mdot
.4281 .32189 Mdot
.4281 .37459 Mdot
.42858 .14197 Mdot
.42858 .22224 Mdot
.42858 .31221 Mdot
.42858 .37501 Mdot
.42929 .15192 Mdot
.42929 .23375 Mdot
.42929 .37563 Mdot
.42999 .14765 Mdot
.42999 .16167 Mdot
.42999 .22643 Mdot
.42999 .27733 Mdot
.42999 .31467 Mdot
.42999 .37624 Mdot
.43044 .16806 Mdot
.43044 .37664 Mdot
.43101 .15869 Mdot
.43101 .17591 Mdot
.43101 .23864 Mdot
.43101 .37713 Mdot
.43134 .16688 Mdot
.43134 .23045 Mdot
.43134 .2804 Mdot
.43134 .31703 Mdot
.43134 .37742 Mdot
.43264 .17191 Mdot
.43264 .17639 Mdot
.43264 .18198 Mdot
.43264 .18533 Mdot
.43264 .23434 Mdot
.43264 .28336 Mdot
.43264 .30454 Mdot
.43264 .31931 Mdot
.43264 .37856 Mdot
.4339 .0625 Mdot
.4339 .07584 Mdot
.4339 .08916 Mdot
.4339 .10247 Mdot
.4339 .18993 Mdot
.4339 .28621 Mdot
.4339 .3215 Mdot
.4339 .37966 Mdot
.4342 .18779 Mdot
.4342 .3076 Mdot
.4342 .37993 Mdot
.43471 .38037 Mdot
.4351 .08219 Mdot
.4351 .09529 Mdot
.4351 .10838 Mdot
.4351 .12144 Mdot
.4351 .19438 Mdot
.4351 .28895 Mdot
.4351 .32361 Mdot
.4351 .38072 Mdot
.43569 .07061 Mdot
.43569 .08681 Mdot
.43569 .31052 Mdot
.43569 .38123 Mdot
.43627 .1141 Mdot
.43627 .12693 Mdot
.43627 .13973 Mdot
.43627 .19867 Mdot
.43627 .21128 Mdot
.43627 .2576 Mdot
.43627 .2916 Mdot
.43627 .32565 Mdot
.43627 .38174 Mdot
.43665 .2126 Mdot
.43665 .38207 Mdot
.43712 .0784 Mdot
.43712 .09429 Mdot
.43712 .11014 Mdot
.43712 .12595 Mdot
.43712 .21424 Mdot
.43712 .27139 Mdot
.43712 .31332 Mdot
.43712 .38248 Mdot
.43739 .13225 Mdot
.43739 .14483 Mdot
.43739 .2152 Mdot
.43739 .26074 Mdot
.43739 .29416 Mdot
.43739 .32761 Mdot
.43739 .38272 Mdot
.43848 .0625 Mdot
.43848 .08317 Mdot
.43848 .13249 Mdot
.43848 .14978 Mdot
.43848 .16209 Mdot
.43848 .219 Mdot
.43848 .26378 Mdot
.43848 .27495 Mdot
.43848 .29663 Mdot
.43848 .31599 Mdot
.43848 .32951 Mdot
.43848 .38367 Mdot
.43952 .16668 Mdot
.43952 .22267 Mdot
.43952 .26671 Mdot
.43952 .29902 Mdot
.43952 .33135 Mdot
.43952 .38458 Mdot
.43978 .15389 Mdot
.43978 .22357 Mdot
.43978 .27837 Mdot
.43978 .31856 Mdot
.43978 .38481 Mdot
.4402 .09274 Mdot
.4402 .11284 Mdot
.4402 .22505 Mdot
.4402 .38518 Mdot
.44054 .06849 Mdot
.44054 .08047 Mdot
.44054 .17111 Mdot
.44054 .18297 Mdot
.44054 .22622 Mdot
.44054 .26955 Mdot
.44054 .30133 Mdot
.44054 .33312 Mdot
.44054 .38547 Mdot
.44103 .15964 Mdot
.44103 .17435 Mdot
.44103 .22795 Mdot
.44103 .28165 Mdot
.44103 .32102 Mdot
.44103 .3859 Mdot
.44184 .14095 Mdot
.44184 .30133 Mdot
.44184 .38661 Mdot
.44223 .17959 Mdot
.44223 .2848 Mdot
.44223 .32338 Mdot
.44223 .38695 Mdot
.44338 .16752 Mdot
.44338 .18463 Mdot
.44338 .25007 Mdot
.44338 .28782 Mdot
.44338 .30493 Mdot
.44338 .32565 Mdot
.44338 .38796 Mdot
.44449 .06951 Mdot
.44449 .08352 Mdot
.44449 .09751 Mdot
.44449 .11148 Mdot
.44449 .18948 Mdot
.44449 .20324 Mdot
.44449 .25369 Mdot
.44449 .29073 Mdot
.44449 .32783 Mdot
.44449 .38893 Mdot
.44485 .17435 Mdot
.44485 .30834 Mdot
.44485 .38924 Mdot
.44555 .1038 Mdot
.44555 .11752 Mdot
.44555 .1312 Mdot
.44555 .20766 Mdot
.44555 .25718 Mdot
.44555 .29353 Mdot
.44555 .32992 Mdot
.44555 .38986 Mdot
.44624 .18087 Mdot
.44624 .26378 Mdot
.44624 .31158 Mdot
.44624 .39046 Mdot
.44658 .12336 Mdot
.44658 .1368 Mdot
.44658 .21191 Mdot
.44658 .26054 Mdot
.44658 .29623 Mdot
.44658 .33194 Mdot
.44658 .39076 Mdot
.44756 .1422 Mdot
.44756 .15535 Mdot
.44756 .20444 Mdot
.44756 .21601 Mdot
.44756 .26378 Mdot
.44756 .26801 Mdot
.44756 .29883 Mdot
.44756 .31467 Mdot
.44756 .33389 Mdot
.44756 .39162 Mdot
.44852 .16034 Mdot
.44852 .17319 Mdot
.44852 .21998 Mdot
.44852 .2669 Mdot
.44852 .33576 Mdot
.44852 .39245 Mdot
.44882 .0625 Mdot
.44882 .07973 Mdot
.44882 .09693 Mdot
.44882 .20994 Mdot
.44882 .27205 Mdot
.44882 .31761 Mdot
.44882 .39272 Mdot
.44943 .16517 Mdot
.44943 .17779 Mdot
.44943 .23622 Mdot
.44943 .26992 Mdot
.44943 .33757 Mdot
.44943 .39325 Mdot
.45003 .10464 Mdot
.45003 .12144 Mdot
.45003 .2152 Mdot
.45003 .2759 Mdot
.45003 .32041 Mdot
.45003 .39377 Mdot
.45032 .0625 Mdot
.45032 .07503 Mdot
.45032 .08755 Mdot
.45032 .18224 Mdot
.45032 .2397 Mdot
.45032 .27283 Mdot
.45032 .33932 Mdot
.45032 .39403 Mdot
.45117 .06867 Mdot
.45117 .08101 Mdot
.45117 .09334 Mdot
.45117 .10565 Mdot
.45117 .11793 Mdot
.45117 .14483 Mdot
.45117 .18655 Mdot
.45117 .19867 Mdot
.45117 .24306 Mdot
.45117 .27958 Mdot
.45117 .32309 Mdot
.45117 .34101 Mdot
.45117 .39478 Mdot
.452 .11107 Mdot
.452 .12316 Mdot
.452 .13522 Mdot
.452 .20263 Mdot
.452 .24631 Mdot
.452 .34264 Mdot
.452 .3955 Mdot
.45227 .15123 Mdot
.45227 .16715 Mdot
.45227 .28309 Mdot
.45227 .32565 Mdot
.45227 .39573 Mdot
.4528 .20647 Mdot
.4528 .3962 Mdot
.45331 .17294 Mdot
.45331 .24493 Mdot
.45331 .28646 Mdot
.45331 .32809 Mdot
.45331 .39665 Mdot
.45432 .1785 Mdot
.45432 .21379 Mdot
.45432 .24901 Mdot
.45432 .28969 Mdot
.45432 .29698 Mdot
.45432 .33044 Mdot
.45432 .39753 Mdot
.45528 .07002 Mdot
.45528 .08505 Mdot
.45528 .19867 Mdot
.45528 .25293 Mdot
.45528 .33268 Mdot
.45528 .39837 Mdot
.45574 .07367 Mdot
.45574 .096 Mdot
.45574 .30133 Mdot
.45574 .39877 Mdot
.4562 .0625 Mdot
.4562 .07727 Mdot
.4562 .09202 Mdot
.4562 .10675 Mdot
.4562 .12144 Mdot
.4562 .20349 Mdot
.4562 .25669 Mdot
.4562 .33483 Mdot
.4562 .39917 Mdot
.45708 .1132 Mdot
.45708 .12761 Mdot
.45708 .20814 Mdot
.45708 .2603 Mdot
.45708 .30543 Mdot
.45708 .3369 Mdot
.45708 .39994 Mdot
.45793 .13356 Mdot
.45793 .14765 Mdot
.45793 .2126 Mdot
.45793 .26378 Mdot
.45793 .33889 Mdot
.45793 .40069 Mdot
.45834 .15736 Mdot
.45834 .30928 Mdot
.45834 .40105 Mdot
.45874 .15312 Mdot
.45874 .16688 Mdot
.45874 .23045 Mdot
.45874 .3408 Mdot
.45874 .4014 Mdot
.45953 .16517 Mdot
.45953 .17191 Mdot
.45953 .23434 Mdot
.45953 .25887 Mdot
.45953 .31292 Mdot
.45953 .34264 Mdot
.45953 .40209 Mdot
.46028 .17676 Mdot
.46028 .23808 Mdot
.46028 .40275 Mdot
.46065 .19217 Mdot
.46065 .26378 Mdot
.46065 .31637 Mdot
.46065 .40307 Mdot
.46101 .06906 Mdot
.46101 .08219 Mdot
.46101 .09529 Mdot
.46101 .19438 Mdot
.46101 .24169 Mdot
.46101 .40338 Mdot
.46171 .08833 Mdot
.46171 .10123 Mdot
.46171 .1141 Mdot
.46171 .12693 Mdot
.46171 .19867 Mdot
.46171 .24517 Mdot
.46171 .26842 Mdot
.46171 .2916 Mdot
.46171 .31962 Mdot
.46171 .38174 Mdot
.46171 .404 Mdot
.46239 .11963 Mdot
.46239 .13225 Mdot
.46239 .14483 Mdot
.46239 .20282 Mdot
.46239 .24854 Mdot
.46239 .29416 Mdot
.46239 .32761 Mdot
.46239 .38272 Mdot
.46239 .40459 Mdot
.46272 .0625 Mdot
.46272 .20484 Mdot
.46272 .27283 Mdot
.46272 .32271 Mdot
.46272 .40488 Mdot
.46304 .13741 Mdot
.46304 .14978 Mdot
.46304 .20683 Mdot
.46304 .219 Mdot
.46304 .29663 Mdot
.46304 .32951 Mdot
.46304 .38367 Mdot
.46304 .40516 Mdot
.46367 .07169 Mdot
.46367 .09005 Mdot
.46367 .10838 Mdot
.46367 .15456 Mdot
.46367 .16668 Mdot
.46367 .22267 Mdot
.46367 .26671 Mdot
.46367 .27702 Mdot
.46367 .29902 Mdot
.46367 .32565 Mdot
.46367 .33135 Mdot
.46367 .38458 Mdot
.46367 .40571 Mdot
.46428 .17111 Mdot
.46428 .22622 Mdot
.46428 .26955 Mdot
.46428 .30133 Mdot
.46428 .33312 Mdot
.46428 .38547 Mdot
.46428 .40625 Mdot
.46458 .13417 Mdot
.46458 .23375 Mdot
.46458 .32843 Mdot
.46458 .40651 Mdot
.46544 .15869 Mdot
.46544 .23864 Mdot
.46544 .40726 Mdot
.46626 .16517 Mdot
.46626 .24331 Mdot
.46626 .40798 Mdot
.46705 .18779 Mdot
.46705 .24777 Mdot
.46705 .40867 Mdot
.4678 .19334 Mdot
.4678 .25204 Mdot
.4678 .40932 Mdot
.46851 .0625 Mdot
.46851 .0784 Mdot
.46851 .09429 Mdot
.46851 .19867 Mdot
.46851 .40995 Mdot
.46919 .10147 Mdot
.46919 .11701 Mdot
.46919 .13249 Mdot
.46919 .41055 Mdot
.46985 .13878 Mdot
.46985 .22357 Mdot
.46985 .41112 Mdot
.47048 .14483 Mdot
.47048 .15964 Mdot
.47048 .22795 Mdot
.47048 .3859 Mdot
.47048 .41167 Mdot
.47108 .16517 Mdot
.47108 .23217 Mdot
.47108 .2848 Mdot
.47108 .32338 Mdot
.47108 .38695 Mdot
.47108 .41219 Mdot
.47166 .18463 Mdot
.47166 .23622 Mdot
.47166 .28782 Mdot
.47166 .32565 Mdot
.47166 .38796 Mdot
.47166 .4127 Mdot
.47221 .06951 Mdot
.47221 .08352 Mdot
.47221 .18948 Mdot
.47221 .29073 Mdot
.47221 .32783 Mdot
.47221 .38893 Mdot
.47221 .41319 Mdot
.47275 .0625 Mdot
.47275 .07628 Mdot
.47275 .09005 Mdot
.47275 .1038 Mdot
.47275 .11752 Mdot
.47275 .19416 Mdot
.47275 .29353 Mdot
.47275 .32992 Mdot
.47275 .38986 Mdot
.47275 .41365 Mdot
.47326 .10988 Mdot
.47326 .12336 Mdot
.47326 .1368 Mdot
.47326 .21191 Mdot
.47326 .26054 Mdot
.47326 .29623 Mdot
.47326 .33194 Mdot
.47326 .39076 Mdot
.47326 .4141 Mdot
.47376 .129 Mdot
.47376 .1422 Mdot
.47376 .21601 Mdot
.47376 .26378 Mdot
.47376 .29883 Mdot
.47376 .33389 Mdot
.47376 .39162 Mdot
.47376 .41454 Mdot
.47423 .14742 Mdot
.47423 .16034 Mdot
.47423 .21998 Mdot
.47423 .2669 Mdot
.47423 .30133 Mdot
.47423 .33576 Mdot
.47423 .39245 Mdot
.47423 .41495 Mdot
.47469 .16517 Mdot
.47469 .2238 Mdot
.47469 .26992 Mdot
.47469 .30375 Mdot
.47469 .33757 Mdot
.47469 .39325 Mdot
.47469 .41536 Mdot
.47514 .16983 Mdot
.47514 .18224 Mdot
.47514 .2275 Mdot
.47514 .27283 Mdot
.47514 .30609 Mdot
.47514 .33932 Mdot
.47514 .39403 Mdot
.47514 .41574 Mdot
.47556 .06867 Mdot
.47556 .08101 Mdot
.47556 .09334 Mdot
.47556 .10565 Mdot
.47556 .18655 Mdot
.47556 .24306 Mdot
.47556 .27565 Mdot
.47556 .30834 Mdot
.47556 .34101 Mdot
.47556 .39478 Mdot
.47556 .41612 Mdot
.47598 .08681 Mdot
.47598 .09895 Mdot
.47598 .11107 Mdot
.47598 .12316 Mdot
.47598 .19072 Mdot
.47598 .24631 Mdot
.47598 .27837 Mdot
.47598 .31052 Mdot
.47598 .34264 Mdot
.47598 .3955 Mdot
.47598 .41648 Mdot
.47677 .14483 Mdot
.47677 .2525 Mdot
.47677 .41717 Mdot
.4775 .25831 Mdot
.4775 .41781 Mdot
.47819 .18425 Mdot
.47819 .26378 Mdot
.47819 .41842 Mdot
.47884 .19166 Mdot
.47884 .41899 Mdot
.47946 .41953 Mdot
.48004 .07258 Mdot
.48004 .09274 Mdot
.48004 .22505 Mdot
.48004 .42003 Mdot
.48058 .12144 Mdot
.48058 .23078 Mdot
.48058 .42051 Mdot
.4811 .14863 Mdot
.4811 .23622 Mdot
.4811 .42096 Mdot
.48159 .17435 Mdot
.48159 .42139 Mdot
.48206 .18087 Mdot
.48206 .39046 Mdot
.48206 .4218 Mdot
.4825 .18708 Mdot
.4825 .39162 Mdot
.4825 .42219 Mdot
.48292 .0625 Mdot
.48292 .07973 Mdot
.48292 .20994 Mdot
.48292 .31761 Mdot
.48292 .39272 Mdot
.48292 .42256 Mdot
.48332 .07094 Mdot
.48332 .08781 Mdot
.48332 .10464 Mdot
.48332 .12144 Mdot
.48332 .2152 Mdot
.48332 .2759 Mdot
.48332 .32041 Mdot
.48332 .39377 Mdot
.48332 .42291 Mdot
.48371 .12847 Mdot
.48371 .14483 Mdot
.48371 .22023 Mdot
.48371 .27958 Mdot
.48371 .32309 Mdot
.48371 .39478 Mdot
.48371 .42324 Mdot
.48407 .15123 Mdot
.48407 .22505 Mdot
.48407 .28309 Mdot
.48407 .32565 Mdot
.48407 .39573 Mdot
.48407 .42356 Mdot
.48442 .15736 Mdot
.48442 .17294 Mdot
.48442 .28646 Mdot
.48442 .32809 Mdot
.48442 .39665 Mdot
.48442 .42387 Mdot
.48476 .1785 Mdot
.48476 .28969 Mdot
.48476 .33044 Mdot
.48476 .39753 Mdot
.48476 .42416 Mdot
.48508 .18384 Mdot
.48508 .25293 Mdot
.48508 .29278 Mdot
.48508 .33268 Mdot
.48508 .39837 Mdot
.48508 .42444 Mdot
.48539 .0625 Mdot
.48539 .07727 Mdot
.48539 .09202 Mdot
.48539 .10675 Mdot
.48539 .20349 Mdot
.48539 .25669 Mdot
.48539 .29574 Mdot
.48539 .33483 Mdot
.48539 .39917 Mdot
.48539 .42471 Mdot
.48568 .09874 Mdot
.48568 .1132 Mdot
.48568 .12761 Mdot
.48568 .20814 Mdot
.48568 .2603 Mdot
.48568 .29859 Mdot
.48568 .3369 Mdot
.48568 .39994 Mdot
.48568 .42497 Mdot
.48596 .13356 Mdot
.48596 .14765 Mdot
.48596 .2126 Mdot
.48596 .26378 Mdot
.48596 .30133 Mdot
.48596 .33889 Mdot
.48596 .40069 Mdot
.48596 .42522 Mdot
.48624 .15312 Mdot
.48624 .21691 Mdot
.48624 .26712 Mdot
.48624 .3408 Mdot
.48624 .4014 Mdot
.48624 .42546 Mdot
.4865 .15841 Mdot
.4865 .17191 Mdot
.4865 .23434 Mdot
.4865 .27034 Mdot
.4865 .34264 Mdot
.4865 .40209 Mdot
.4865 .42569 Mdot
.48675 .17676 Mdot
.48675 .23808 Mdot
.48675 .34441 Mdot
.48675 .40275 Mdot
.48675 .42591 Mdot
.48699 .06906 Mdot
.48699 .08219 Mdot
.48699 .18144 Mdot
.48699 .24169 Mdot
.48699 .34611 Mdot
.48699 .40338 Mdot
.48699 .42612 Mdot
.48723 .0625 Mdot
.48723 .07542 Mdot
.48723 .08833 Mdot
.48723 .10123 Mdot
.48723 .1141 Mdot
.48723 .19867 Mdot
.48723 .24517 Mdot
.48723 .34776 Mdot
.48723 .404 Mdot
.48723 .42633 Mdot
.48745 .10697 Mdot
.48745 .11963 Mdot
.48745 .13225 Mdot
.48745 .20282 Mdot
.48745 .24854 Mdot
.48745 .34934 Mdot
.48745 .40459 Mdot
.48745 .42652 Mdot
.48767 .13741 Mdot
.48767 .14978 Mdot
.48767 .20683 Mdot
.48767 .25179 Mdot
.48767 .29663 Mdot
.48767 .35087 Mdot
.48767 .40516 Mdot
.48767 .42671 Mdot
.48788 .15456 Mdot
.48788 .21071 Mdot
.48788 .22267 Mdot
.48788 .29902 Mdot
.48788 .35235 Mdot
.48788 .40571 Mdot
.48788 .4269 Mdot
.48809 .1592 Mdot
.48809 .17111 Mdot
.48809 .22622 Mdot
.48809 .30133 Mdot
.48809 .35378 Mdot
.48809 .40625 Mdot
.48809 .42708 Mdot
.5 .09005 Mdot
.5 .11633 Mdot
.5 .16517 Mdot
.5 .27283 Mdot
.5 .281 Mdot
.5 .4375 Mdot
.51191 .1592 Mdot
.51191 .17111 Mdot
.51191 .22622 Mdot
.51191 .30133 Mdot
.51191 .35378 Mdot
.51191 .40625 Mdot
.51191 .42708 Mdot
.51212 .15456 Mdot
.51212 .21071 Mdot
.51212 .22267 Mdot
.51212 .29902 Mdot
.51212 .35235 Mdot
.51212 .40571 Mdot
.51212 .4269 Mdot
.51233 .13741 Mdot
.51233 .14978 Mdot
.51233 .20683 Mdot
.51233 .25179 Mdot
.51233 .29663 Mdot
.51233 .35087 Mdot
.51233 .40516 Mdot
.51233 .42671 Mdot
.51255 .10697 Mdot
.51255 .11963 Mdot
.51255 .13225 Mdot
.51255 .20282 Mdot
.51255 .24854 Mdot
.51255 .34934 Mdot
.51255 .40459 Mdot
.51255 .42652 Mdot
.51277 .0625 Mdot
.51277 .07542 Mdot
.51277 .08833 Mdot
.51277 .10123 Mdot
.51277 .1141 Mdot
.51277 .19867 Mdot
.51277 .24517 Mdot
.51277 .34776 Mdot
.51277 .404 Mdot
.51277 .42633 Mdot
.51301 .06906 Mdot
.51301 .08219 Mdot
.51301 .18144 Mdot
.51301 .24169 Mdot
.51301 .34611 Mdot
.51301 .40338 Mdot
.51301 .42612 Mdot
.51325 .17676 Mdot
.51325 .23808 Mdot
.51325 .34441 Mdot
.51325 .40275 Mdot
.51325 .42591 Mdot
.5135 .15841 Mdot
.5135 .17191 Mdot
.5135 .23434 Mdot
.5135 .27034 Mdot
.5135 .34264 Mdot
.5135 .40209 Mdot
.5135 .42569 Mdot
.51376 .15312 Mdot
.51376 .21691 Mdot
.51376 .26712 Mdot
.51376 .3408 Mdot
.51376 .4014 Mdot
.51376 .42546 Mdot
.51404 .13356 Mdot
.51404 .14765 Mdot
.51404 .2126 Mdot
.51404 .26378 Mdot
.51404 .30133 Mdot
.51404 .33889 Mdot
.51404 .40069 Mdot
.51404 .42522 Mdot
.51432 .09874 Mdot
.51432 .1132 Mdot
.51432 .12761 Mdot
.51432 .20814 Mdot
.51432 .2603 Mdot
.51432 .29859 Mdot
.51432 .3369 Mdot
.51432 .39994 Mdot
.51432 .42497 Mdot
.51461 .0625 Mdot
.51461 .07727 Mdot
.51461 .09202 Mdot
.51461 .10675 Mdot
.51461 .20349 Mdot
.51461 .25669 Mdot
.51461 .29574 Mdot
.51461 .33483 Mdot
.51461 .39917 Mdot
.51461 .42471 Mdot
.51492 .18384 Mdot
.51492 .25293 Mdot
.51492 .29278 Mdot
.51492 .33268 Mdot
.51492 .39837 Mdot
.51492 .42444 Mdot
.51524 .1785 Mdot
.51524 .28969 Mdot
.51524 .33044 Mdot
.51524 .39753 Mdot
.51524 .42416 Mdot
.51558 .15736 Mdot
.51558 .17294 Mdot
.51558 .28646 Mdot
.51558 .32809 Mdot
.51558 .39665 Mdot
.51558 .42387 Mdot
.51593 .15123 Mdot
.51593 .22505 Mdot
.51593 .28309 Mdot
.51593 .32565 Mdot
.51593 .39573 Mdot
.51593 .42356 Mdot
.51629 .12847 Mdot
.51629 .14483 Mdot
.51629 .22023 Mdot
.51629 .27958 Mdot
.51629 .32309 Mdot
.51629 .39478 Mdot
.51629 .42324 Mdot
.51668 .07094 Mdot
.51668 .08781 Mdot
.51668 .10464 Mdot
.51668 .12144 Mdot
.51668 .2152 Mdot
.51668 .2759 Mdot
.51668 .32041 Mdot
.51668 .39377 Mdot
.51668 .42291 Mdot
.51708 .0625 Mdot
.51708 .07973 Mdot
.51708 .20994 Mdot
.51708 .31761 Mdot
.51708 .39272 Mdot
.51708 .42256 Mdot
.5175 .18708 Mdot
.5175 .39162 Mdot
.5175 .42219 Mdot
.51794 .18087 Mdot
.51794 .39046 Mdot
.51794 .4218 Mdot
.51841 .17435 Mdot
.51841 .42139 Mdot
.5189 .14863 Mdot
.5189 .23622 Mdot
.5189 .42096 Mdot
.51942 .12144 Mdot
.51942 .23078 Mdot
.51942 .42051 Mdot
.51996 .07258 Mdot
.51996 .09274 Mdot
.51996 .22505 Mdot
.51996 .42003 Mdot
.52054 .41953 Mdot
.52116 .19166 Mdot
.52116 .41899 Mdot
.52181 .18425 Mdot
.52181 .26378 Mdot
.52181 .41842 Mdot
.5225 .25831 Mdot
.5225 .41781 Mdot
.52323 .14483 Mdot
.52323 .2525 Mdot
.52323 .41717 Mdot
.52402 .08681 Mdot
.52402 .09895 Mdot
.52402 .11107 Mdot
.52402 .12316 Mdot
.52402 .19072 Mdot
.52402 .24631 Mdot
.52402 .27837 Mdot
.52402 .31052 Mdot
.52402 .34264 Mdot
.52402 .3955 Mdot
.52402 .41648 Mdot
.52444 .06867 Mdot
.52444 .08101 Mdot
.52444 .09334 Mdot
.52444 .10565 Mdot
.52444 .18655 Mdot
.52444 .24306 Mdot
.52444 .27565 Mdot
.52444 .30834 Mdot
.52444 .34101 Mdot
.52444 .39478 Mdot
.52444 .41612 Mdot
.52486 .16983 Mdot
.52486 .18224 Mdot
.52486 .2275 Mdot
.52486 .27283 Mdot
.52486 .30609 Mdot
.52486 .33932 Mdot
.52486 .39403 Mdot
.52486 .41574 Mdot
.52531 .16517 Mdot
.52531 .2238 Mdot
.52531 .26992 Mdot
.52531 .30375 Mdot
.52531 .33757 Mdot
.52531 .39325 Mdot
.52531 .41536 Mdot
.52577 .14742 Mdot
.52577 .16034 Mdot
.52577 .21998 Mdot
.52577 .2669 Mdot
.52577 .30133 Mdot
.52577 .33576 Mdot
.52577 .39245 Mdot
.52577 .41495 Mdot
.52624 .129 Mdot
.52624 .1422 Mdot
.52624 .21601 Mdot
.52624 .26378 Mdot
.52624 .29883 Mdot
.52624 .33389 Mdot
.52624 .39162 Mdot
.52624 .41454 Mdot
.52674 .10988 Mdot
.52674 .12336 Mdot
.52674 .1368 Mdot
.52674 .21191 Mdot
.52674 .26054 Mdot
.52674 .29623 Mdot
.52674 .33194 Mdot
.52674 .39076 Mdot
.52674 .4141 Mdot
.52725 .0625 Mdot
.52725 .07628 Mdot
.52725 .09005 Mdot
.52725 .1038 Mdot
.52725 .11752 Mdot
.52725 .19416 Mdot
.52725 .29353 Mdot
.52725 .32992 Mdot
.52725 .38986 Mdot
.52725 .41365 Mdot
.52779 .06951 Mdot
.52779 .08352 Mdot
.52779 .18948 Mdot
.52779 .29073 Mdot
.52779 .32783 Mdot
.52779 .38893 Mdot
.52779 .41319 Mdot
.52834 .18463 Mdot
.52834 .23622 Mdot
.52834 .28782 Mdot
.52834 .32565 Mdot
.52834 .38796 Mdot
.52834 .4127 Mdot
.52892 .16517 Mdot
.52892 .23217 Mdot
.52892 .2848 Mdot
.52892 .32338 Mdot
.52892 .38695 Mdot
.52892 .41219 Mdot
.52952 .14483 Mdot
.52952 .15964 Mdot
.52952 .22795 Mdot
.52952 .3859 Mdot
.52952 .41167 Mdot
.53015 .13878 Mdot
.53015 .22357 Mdot
.53015 .41112 Mdot
.53081 .10147 Mdot
.53081 .11701 Mdot
.53081 .13249 Mdot
.53081 .41055 Mdot
.53149 .0625 Mdot
.53149 .0784 Mdot
.53149 .09429 Mdot
.53149 .19867 Mdot
.53149 .40995 Mdot
.5322 .19334 Mdot
.5322 .25204 Mdot
.5322 .40932 Mdot
.53295 .18779 Mdot
.53295 .24777 Mdot
.53295 .40867 Mdot
.53374 .16517 Mdot
.53374 .24331 Mdot
.53374 .40798 Mdot
.53456 .15869 Mdot
.53456 .23864 Mdot
.53456 .40726 Mdot
.53542 .13417 Mdot
.53542 .23375 Mdot
.53542 .32843 Mdot
.53542 .40651 Mdot
.53572 .17111 Mdot
.53572 .22622 Mdot
.53572 .26955 Mdot
.53572 .30133 Mdot
.53572 .33312 Mdot
.53572 .38547 Mdot
.53572 .40625 Mdot
.53633 .07169 Mdot
.53633 .09005 Mdot
.53633 .10838 Mdot
.53633 .15456 Mdot
.53633 .16668 Mdot
.53633 .22267 Mdot
.53633 .26671 Mdot
.53633 .27702 Mdot
.53633 .29902 Mdot
.53633 .32565 Mdot
.53633 .33135 Mdot
.53633 .38458 Mdot
.53633 .40571 Mdot
.53696 .13741 Mdot
.53696 .14978 Mdot
.53696 .20683 Mdot
.53696 .219 Mdot
.53696 .29663 Mdot
.53696 .32951 Mdot
.53696 .38367 Mdot
.53696 .40516 Mdot
.53728 .0625 Mdot
.53728 .20484 Mdot
.53728 .27283 Mdot
.53728 .32271 Mdot
.53728 .40488 Mdot
.53761 .11963 Mdot
.53761 .13225 Mdot
.53761 .14483 Mdot
.53761 .20282 Mdot
.53761 .24854 Mdot
.53761 .29416 Mdot
.53761 .32761 Mdot
.53761 .38272 Mdot
.53761 .40459 Mdot
.53829 .08833 Mdot
.53829 .10123 Mdot
.53829 .1141 Mdot
.53829 .12693 Mdot
.53829 .19867 Mdot
.53829 .24517 Mdot
.53829 .26842 Mdot
.53829 .2916 Mdot
.53829 .31962 Mdot
.53829 .38174 Mdot
.53829 .404 Mdot
.53899 .06906 Mdot
.53899 .08219 Mdot
.53899 .09529 Mdot
.53899 .19438 Mdot
.53899 .24169 Mdot
.53899 .40338 Mdot
.53935 .19217 Mdot
.53935 .26378 Mdot
.53935 .31637 Mdot
.53935 .40307 Mdot
.53972 .17676 Mdot
.53972 .23808 Mdot
.53972 .40275 Mdot
.54047 .16517 Mdot
.54047 .17191 Mdot
.54047 .23434 Mdot
.54047 .25887 Mdot
.54047 .31292 Mdot
.54047 .34264 Mdot
.54047 .40209 Mdot
.54126 .15312 Mdot
.54126 .16688 Mdot
.54126 .23045 Mdot
.54126 .3408 Mdot
.54126 .4014 Mdot
.54166 .15736 Mdot
.54166 .30928 Mdot
.54166 .40105 Mdot
.54207 .13356 Mdot
.54207 .14765 Mdot
.54207 .2126 Mdot
.54207 .26378 Mdot
.54207 .33889 Mdot
.54207 .40069 Mdot
.54292 .1132 Mdot
.54292 .12761 Mdot
.54292 .20814 Mdot
.54292 .2603 Mdot
.54292 .30543 Mdot
.54292 .3369 Mdot
.54292 .39994 Mdot
.5438 .0625 Mdot
.5438 .07727 Mdot
.5438 .09202 Mdot
.5438 .10675 Mdot
.5438 .12144 Mdot
.5438 .20349 Mdot
.5438 .25669 Mdot
.5438 .33483 Mdot
.5438 .39917 Mdot
.54426 .07367 Mdot
.54426 .096 Mdot
.54426 .30133 Mdot
.54426 .39877 Mdot
.54472 .07002 Mdot
.54472 .08505 Mdot
.54472 .19867 Mdot
.54472 .25293 Mdot
.54472 .33268 Mdot
.54472 .39837 Mdot
.54568 .1785 Mdot
.54568 .21379 Mdot
.54568 .24901 Mdot
.54568 .28969 Mdot
.54568 .29698 Mdot
.54568 .33044 Mdot
.54568 .39753 Mdot
.54669 .17294 Mdot
.54669 .24493 Mdot
.54669 .28646 Mdot
.54669 .32809 Mdot
.54669 .39665 Mdot
.5472 .20647 Mdot
.5472 .3962 Mdot
.54773 .15123 Mdot
.54773 .16715 Mdot
.54773 .28309 Mdot
.54773 .32565 Mdot
.54773 .39573 Mdot
.548 .11107 Mdot
.548 .12316 Mdot
.548 .13522 Mdot
.548 .20263 Mdot
.548 .24631 Mdot
.548 .34264 Mdot
.548 .3955 Mdot
.54883 .06867 Mdot
.54883 .08101 Mdot
.54883 .09334 Mdot
.54883 .10565 Mdot
.54883 .11793 Mdot
.54883 .14483 Mdot
.54883 .18655 Mdot
.54883 .19867 Mdot
.54883 .24306 Mdot
.54883 .27958 Mdot
.54883 .32309 Mdot
.54883 .34101 Mdot
.54883 .39478 Mdot
.54968 .0625 Mdot
.54968 .07503 Mdot
.54968 .08755 Mdot
.54968 .18224 Mdot
.54968 .2397 Mdot
.54968 .27283 Mdot
.54968 .33932 Mdot
.54968 .39403 Mdot
.54997 .10464 Mdot
.54997 .12144 Mdot
.54997 .2152 Mdot
.54997 .2759 Mdot
.54997 .32041 Mdot
.54997 .39377 Mdot
.55057 .16517 Mdot
.55057 .17779 Mdot
.55057 .23622 Mdot
.55057 .26992 Mdot
.55057 .33757 Mdot
.55057 .39325 Mdot
.55118 .0625 Mdot
.55118 .07973 Mdot
.55118 .09693 Mdot
.55118 .20994 Mdot
.55118 .27205 Mdot
.55118 .31761 Mdot
.55118 .39272 Mdot
.55148 .16034 Mdot
.55148 .17319 Mdot
.55148 .21998 Mdot
.55148 .2669 Mdot
.55148 .33576 Mdot
.55148 .39245 Mdot
.55244 .1422 Mdot
.55244 .15535 Mdot
.55244 .20444 Mdot
.55244 .21601 Mdot
.55244 .26378 Mdot
.55244 .26801 Mdot
.55244 .29883 Mdot
.55244 .31467 Mdot
.55244 .33389 Mdot
.55244 .39162 Mdot
.55342 .12336 Mdot
.55342 .1368 Mdot
.55342 .21191 Mdot
.55342 .26054 Mdot
.55342 .29623 Mdot
.55342 .33194 Mdot
.55342 .39076 Mdot
.55376 .18087 Mdot
.55376 .26378 Mdot
.55376 .31158 Mdot
.55376 .39046 Mdot
.55445 .1038 Mdot
.55445 .11752 Mdot
.55445 .1312 Mdot
.55445 .20766 Mdot
.55445 .25718 Mdot
.55445 .29353 Mdot
.55445 .32992 Mdot
.55445 .38986 Mdot
.55515 .17435 Mdot
.55515 .30834 Mdot
.55515 .38924 Mdot
.55551 .06951 Mdot
.55551 .08352 Mdot
.55551 .09751 Mdot
.55551 .11148 Mdot
.55551 .18948 Mdot
.55551 .20324 Mdot
.55551 .25369 Mdot
.55551 .29073 Mdot
.55551 .32783 Mdot
.55551 .38893 Mdot
.55662 .16752 Mdot
.55662 .18463 Mdot
.55662 .25007 Mdot
.55662 .28782 Mdot
.55662 .30493 Mdot
.55662 .32565 Mdot
.55662 .38796 Mdot
.55777 .17959 Mdot
.55777 .2848 Mdot
.55777 .32338 Mdot
.55777 .38695 Mdot
.55816 .14095 Mdot
.55816 .30133 Mdot
.55816 .38661 Mdot
.55897 .15964 Mdot
.55897 .17435 Mdot
.55897 .22795 Mdot
.55897 .28165 Mdot
.55897 .32102 Mdot
.55897 .3859 Mdot
.55946 .06849 Mdot
.55946 .08047 Mdot
.55946 .17111 Mdot
.55946 .18297 Mdot
.55946 .22622 Mdot
.55946 .26955 Mdot
.55946 .30133 Mdot
.55946 .33312 Mdot
.55946 .38547 Mdot
.5598 .09274 Mdot
.5598 .11284 Mdot
.5598 .22505 Mdot
.5598 .38518 Mdot
.56022 .15389 Mdot
.56022 .22357 Mdot
.56022 .27837 Mdot
.56022 .31856 Mdot
.56022 .38481 Mdot
.56048 .16668 Mdot
.56048 .22267 Mdot
.56048 .26671 Mdot
.56048 .29902 Mdot
.56048 .33135 Mdot
.56048 .38458 Mdot
.56152 .0625 Mdot
.56152 .08317 Mdot
.56152 .13249 Mdot
.56152 .14978 Mdot
.56152 .16209 Mdot
.56152 .219 Mdot
.56152 .26378 Mdot
.56152 .27495 Mdot
.56152 .29663 Mdot
.56152 .31599 Mdot
.56152 .32951 Mdot
.56152 .38367 Mdot
.56261 .13225 Mdot
.56261 .14483 Mdot
.56261 .2152 Mdot
.56261 .26074 Mdot
.56261 .29416 Mdot
.56261 .32761 Mdot
.56261 .38272 Mdot
.56288 .0784 Mdot
.56288 .09429 Mdot
.56288 .11014 Mdot
.56288 .12595 Mdot
.56288 .21424 Mdot
.56288 .27139 Mdot
.56288 .31332 Mdot
.56288 .38248 Mdot
.56335 .2126 Mdot
.56335 .38207 Mdot
.56373 .1141 Mdot
.56373 .12693 Mdot
.56373 .13973 Mdot
.56373 .19867 Mdot
.56373 .21128 Mdot
.56373 .2576 Mdot
.56373 .2916 Mdot
.56373 .32565 Mdot
.56373 .38174 Mdot
.56431 .07061 Mdot
.56431 .08681 Mdot
.56431 .31052 Mdot
.56431 .38123 Mdot
.5649 .08219 Mdot
.5649 .09529 Mdot
.5649 .10838 Mdot
.5649 .12144 Mdot
.5649 .19438 Mdot
.5649 .28895 Mdot
.5649 .32361 Mdot
.5649 .38072 Mdot
.56529 .38037 Mdot
.5658 .18779 Mdot
.5658 .3076 Mdot
.5658 .37993 Mdot
.5661 .0625 Mdot
.5661 .07584 Mdot
.5661 .08916 Mdot
.5661 .10247 Mdot
.5661 .18993 Mdot
.5661 .28621 Mdot
.5661 .3215 Mdot
.5661 .37966 Mdot
.56736 .17191 Mdot
.56736 .17639 Mdot
.56736 .18198 Mdot
.56736 .18533 Mdot
.56736 .23434 Mdot
.56736 .28336 Mdot
.56736 .30454 Mdot
.56736 .31931 Mdot
.56736 .37856 Mdot
.56866 .16688 Mdot
.56866 .23045 Mdot
.56866 .2804 Mdot
.56866 .31703 Mdot
.56866 .37742 Mdot
.56899 .15869 Mdot
.56899 .17591 Mdot
.56899 .23864 Mdot
.56899 .37713 Mdot
.56956 .16806 Mdot
.56956 .37664 Mdot
.57001 .14765 Mdot
.57001 .16167 Mdot
.57001 .22643 Mdot
.57001 .27733 Mdot
.57001 .31467 Mdot
.57001 .37624 Mdot
.57071 .15192 Mdot
.57071 .23375 Mdot
.57071 .37563 Mdot
.57142 .14197 Mdot
.57142 .22224 Mdot
.57142 .31221 Mdot
.57142 .37501 Mdot
.5719 .13522 Mdot
.5719 .14723 Mdot
.5719 .20263 Mdot
.5719 .28991 Mdot
.5719 .32189 Mdot
.5719 .37459 Mdot
.57251 .12665 Mdot
.57251 .14483 Mdot
.57251 .22861 Mdot
.57251 .37405 Mdot
.57288 .12144 Mdot
.57288 .13608 Mdot
.57288 .2179 Mdot
.57288 .30966 Mdot
.57288 .37373 Mdot
.57313 .10565 Mdot
.57313 .11793 Mdot
.57313 .13019 Mdot
.57313 .19867 Mdot
.57313 .24306 Mdot
.57313 .2874 Mdot
.57313 .31991 Mdot
.57313 .37351 Mdot
.57441 .0625 Mdot
.57441 .07002 Mdot
.57441 .07503 Mdot
.57441 .08129 Mdot
.57441 .08505 Mdot
.57441 .08755 Mdot
.57441 .10006 Mdot
.57441 .11254 Mdot
.57441 .11503 Mdot
.57441 .11877 Mdot
.57441 .19457 Mdot
.57441 .19867 Mdot
.57441 .22322 Mdot
.57441 .2397 Mdot
.57441 .2848 Mdot
.57441 .307 Mdot
.57441 .31787 Mdot
.57441 .37239 Mdot
.57573 .06886 Mdot
.57573 .08158 Mdot
.57573 .09429 Mdot
.57573 .17779 Mdot
.57573 .19034 Mdot
.57573 .23622 Mdot
.57573 .2821 Mdot
.57573 .31575 Mdot
.57573 .37123 Mdot
.576 .0625 Mdot
.576 .07781 Mdot
.576 .19365 Mdot
.576 .371 Mdot
.57641 .37064 Mdot
.5771 .17319 Mdot
.5771 .23263 Mdot
.5771 .27932 Mdot
.5771 .31357 Mdot
.5771 .37004 Mdot
.57766 .18842 Mdot
.57766 .24493 Mdot
.57766 .36955 Mdot
.57852 .15535 Mdot
.57852 .16843 Mdot
.57852 .19217 Mdot
.57852 .22891 Mdot
.57852 .3113 Mdot
.57852 .3688 Mdot
.5794 .16715 Mdot
.5794 .18297 Mdot
.5794 .24067 Mdot
.5794 .36803 Mdot
.57999 .15018 Mdot
.57999 .1635 Mdot
.57999 .22505 Mdot
.57999 .30894 Mdot
.57999 .36751 Mdot
.58075 .18533 Mdot
.58075 .36684 Mdot
.58121 .16111 Mdot
.58121 .23622 Mdot
.58121 .36644 Mdot
.58152 .1312 Mdot
.58152 .14483 Mdot
.58152 .20766 Mdot
.58152 .36617 Mdot
.58311 .06849 Mdot
.58311 .08047 Mdot
.58311 .09245 Mdot
.58311 .1044 Mdot
.58311 .11148 Mdot
.58311 .11633 Mdot
.58311 .12541 Mdot
.58311 .13817 Mdot
.58311 .15481 Mdot
.58311 .17811 Mdot
.58311 .18297 Mdot
.58311 .19475 Mdot
.58311 .20324 Mdot
.58311 .23158 Mdot
.58311 .23788 Mdot
.58311 .25369 Mdot
.58311 .281 Mdot
.58311 .31263 Mdot
.58311 .36478 Mdot
.58452 .0625 Mdot
.58452 .07466 Mdot
.58452 .08681 Mdot
.58452 .17873 Mdot
.58452 .23454 Mdot
.58452 .31052 Mdot
.58452 .36355 Mdot
.58476 .0625 Mdot
.58476 .07676 Mdot
.58476 .091 Mdot
.58476 .10522 Mdot
.58476 .19867 Mdot
.58476 .25007 Mdot
.58476 .36334 Mdot
.5851 .1141 Mdot
.5851 .1312 Mdot
.5851 .22672 Mdot
.5851 .36304 Mdot
.58561 .14913 Mdot
.58561 .24821 Mdot
.58561 .36259 Mdot
.58598 .16209 Mdot
.58598 .17435 Mdot
.58598 .23108 Mdot
.58598 .36227 Mdot
.58647 .06975 Mdot
.58647 .19391 Mdot
.58647 .24631 Mdot
.58647 .36183 Mdot
.58718 .0713 Mdot
.58718 .08888 Mdot
.58718 .10643 Mdot
.58718 .36122 Mdot
.58749 .15736 Mdot
.58749 .16983 Mdot
.58749 .2275 Mdot
.58749 .26074 Mdot
.58749 .36095 Mdot
.58826 .11826 Mdot
.58826 .14042 Mdot
.58826 .17435 Mdot
.58826 .2424 Mdot
.58826 .36027 Mdot
.58905 .13973 Mdot
.58905 .15248 Mdot
.58905 .21128 Mdot
.58905 .2576 Mdot
.58905 .35958 Mdot
.58937 .19867 Mdot
.58937 .26378 Mdot
.58937 .3593 Mdot
.59013 .16891 Mdot
.59013 .23832 Mdot
.59013 .35864 Mdot
.59067 .12144 Mdot
.59067 .13445 Mdot
.59067 .14742 Mdot
.59067 .20722 Mdot
.59067 .25436 Mdot
.59067 .35816 Mdot
.59109 .0625 Mdot
.59109 .08546 Mdot
.59109 .10838 Mdot
.59109 .23622 Mdot
.59109 .3578 Mdot
.59167 .19262 Mdot
.59167 .25933 Mdot
.59167 .35729 Mdot
.59207 .14791 Mdot
.59207 .16325 Mdot
.59207 .23408 Mdot
.59207 .35694 Mdot
.59235 .10247 Mdot
.59235 .11576 Mdot
.59235 .129 Mdot
.59235 .20302 Mdot
.59235 .25099 Mdot
.59235 .35669 Mdot
.5941 .06928 Mdot
.5941 .08283 Mdot
.5941 .09637 Mdot
.5941 .10988 Mdot
.5941 .12595 Mdot
.5941 .14169 Mdot
.5941 .18626 Mdot
.5941 .19867 Mdot
.5941 .2475 Mdot
.5941 .25466 Mdot
.5941 .27139 Mdot
.5941 .35517 Mdot
.59568 .14723 Mdot
.59568 .1592 Mdot
.59568 .21447 Mdot
.59568 .25798 Mdot
.59568 .35378 Mdot
.59591 .18056 Mdot
.59591 .24388 Mdot
.59591 .35358 Mdot
.59622 .103 Mdot
.59622 .11914 Mdot
.59622 .20927 Mdot
.59622 .26767 Mdot
.59622 .35331 Mdot
.59665 .16034 Mdot
.59665 .24973 Mdot
.59665 .35293 Mdot
.59731 .1424 Mdot
.59731 .15456 Mdot
.59731 .21071 Mdot
.59731 .25494 Mdot
.59731 .35235 Mdot
.59779 .17562 Mdot
.59779 .24012 Mdot
.59779 .27733 Mdot
.59779 .35193 Mdot
.59843 .0625 Mdot
.59843 .07904 Mdot
.59843 .09556 Mdot
.59843 .20408 Mdot
.59843 .26378 Mdot
.59843 .35137 Mdot
.599 .12499 Mdot
.599 .13741 Mdot
.599 .20683 Mdot
.599 .25179 Mdot
.599 .35087 Mdot
.59935 .15279 Mdot
.59935 .24455 Mdot
.59935 .35057 Mdot
.59974 .15627 Mdot
.59974 .17049 Mdot
.59974 .23622 Mdot
.59974 .27414 Mdot
.59974 .35022 Mdot
.60075 .09429 Mdot
.60075 .10697 Mdot
.60075 .11963 Mdot
.60075 .19034 Mdot
.60075 .19867 Mdot
.60075 .20282 Mdot
.60075 .24854 Mdot
.60075 .25971 Mdot
.60075 .2821 Mdot
.60075 .34934 Mdot
.60178 .15066 Mdot
.60178 .27082 Mdot
.60178 .34844 Mdot
.6022 .12437 Mdot
.6022 .23908 Mdot
.6022 .34808 Mdot
.60257 .0625 Mdot
.60257 .07542 Mdot
.60257 .08833 Mdot
.60257 .10123 Mdot
.60257 .18597 Mdot
.60257 .24517 Mdot
.60257 .27932 Mdot
.60257 .31357 Mdot
.60257 .34776 Mdot
.60318 .193 Mdot
.60318 .25545 Mdot
.60318 .34721 Mdot
.6039 .12996 Mdot
.6039 .14483 Mdot
.6039 .21337 Mdot
.6039 .26737 Mdot
.6039 .307 Mdot
.6039 .34659 Mdot
.60444 .18144 Mdot
.60444 .24169 Mdot
.60444 .27643 Mdot
.60444 .3113 Mdot
.60444 .34611 Mdot
.60522 .0731 Mdot
.60522 .09429 Mdot
.60522 .28929 Mdot
.60522 .34544 Mdot
.60574 .16956 Mdot
.60574 .25099 Mdot
.60574 .29797 Mdot
.60574 .34498 Mdot
.60611 .09311 Mdot
.60611 .10838 Mdot
.60611 .12361 Mdot
.60611 .20867 Mdot
.60611 .26378 Mdot
.60611 .30422 Mdot
.60611 .34465 Mdot
.60639 .1635 Mdot
.60639 .17676 Mdot
.60639 .27344 Mdot
.60639 .30894 Mdot
.60639 .34441 Mdot
.60842 .0703 Mdot
.60842 .0859 Mdot
.60842 .08681 Mdot
.60842 .09895 Mdot
.60842 .10147 Mdot
.60842 .12316 Mdot
.60842 .15841 Mdot
.60842 .16291 Mdot
.60842 .19072 Mdot
.60842 .20377 Mdot
.60842 .22105 Mdot
.60842 .24631 Mdot
.60842 .26003 Mdot
.60842 .27034 Mdot
.60842 .27837 Mdot
.60842 .2848 Mdot
.60842 .29444 Mdot
.60842 .30133 Mdot
.60842 .3065 Mdot
.60842 .31052 Mdot
.60842 .34264 Mdot
.61028 .06867 Mdot
.61028 .08101 Mdot
.61028 .09334 Mdot
.61028 .10565 Mdot
.61028 .18655 Mdot
.61028 .24306 Mdot
.61028 .27565 Mdot
.61028 .30834 Mdot
.61028 .34101 Mdot
.61052 .1393 Mdot
.61052 .15312 Mdot
.61052 .21691 Mdot
.61052 .26712 Mdot
.61052 .30397 Mdot
.61052 .3408 Mdot
.61082 .19867 Mdot
.61082 .25612 Mdot
.61082 .29831 Mdot
.61082 .34053 Mdot
.61123 .13741 Mdot
.61123 .29073 Mdot
.61123 .34017 Mdot
.61182 .19867 Mdot
.61182 .28002 Mdot
.61182 .33966 Mdot
.61221 .16983 Mdot
.61221 .18224 Mdot
.61221 .2275 Mdot
.61221 .27283 Mdot
.61221 .30609 Mdot
.61221 .33932 Mdot
.6127 .11941 Mdot
.6127 .13356 Mdot
.6127 .2126 Mdot
.6127 .26378 Mdot
.6127 .30133 Mdot
.6127 .33889 Mdot
.61334 .17729 Mdot
.61334 .25204 Mdot
.61334 .29516 Mdot
.61334 .33833 Mdot
.6142 .11051 Mdot
.6142 .12962 Mdot
.6142 .16517 Mdot
.6142 .17779 Mdot
.6142 .2238 Mdot
.6142 .26992 Mdot
.6142 .28683 Mdot
.6142 .30375 Mdot
.6142 .33757 Mdot
.61497 .09874 Mdot
.61497 .1132 Mdot
.61497 .12761 Mdot
.61497 .20814 Mdot
.61497 .2603 Mdot
.61497 .29859 Mdot
.61497 .3369 Mdot
.61543 .19106 Mdot
.61543 .27495 Mdot
.61543 .3365 Mdot
.61597 .17136 Mdot
.61597 .24777 Mdot
.61597 .29186 Mdot
.61597 .33603 Mdot
.61627 .14742 Mdot
.61627 .16034 Mdot
.61627 .21998 Mdot
.61627 .2669 Mdot
.61627 .30133 Mdot
.61627 .33576 Mdot
.61733 .0625 Mdot
.61733 .07727 Mdot
.61733 .08219 Mdot
.61733 .09202 Mdot
.61733 .10184 Mdot
.61733 .10675 Mdot
.61733 .20349 Mdot
.61733 .25669 Mdot
.61733 .28271 Mdot
.61733 .29574 Mdot
.61733 .33483 Mdot
.61841 .1422 Mdot
.61841 .15535 Mdot
.61841 .21601 Mdot
.61841 .26378 Mdot
.61841 .29883 Mdot
.61841 .33389 Mdot
.61873 .14823 Mdot
.61873 .16517 Mdot
.61873 .28841 Mdot
.61873 .33362 Mdot
.61929 .17111 Mdot
.61929 .18297 Mdot
.61929 .22622 Mdot
.61929 .26955 Mdot
.61929 .30133 Mdot
.61929 .33312 Mdot
.61979 .18384 Mdot
.61979 .25293 Mdot
.61979 .29278 Mdot
.61979 .33268 Mdot
.62064 .12336 Mdot
.62064 .1368 Mdot
.62064 .2053 Mdot
.62064 .21191 Mdot
.62064 .26054 Mdot
.62064 .27837 Mdot
.62064 .29623 Mdot
.62064 .33194 Mdot
.62132 .16668 Mdot
.62132 .22267 Mdot
.62132 .26671 Mdot
.62132 .29902 Mdot
.62132 .33135 Mdot
.62161 .14136 Mdot
.62161 .2848 Mdot
.62161 .33109 Mdot
.62236 .1785 Mdot
.62236 .24901 Mdot
.62236 .28969 Mdot
.62236 .33044 Mdot
.62294 .09005 Mdot
.62294 .1038 Mdot
.62294 .11752 Mdot
.62294 .20766 Mdot
.62294 .25718 Mdot
.62294 .29353 Mdot
.62294 .32992 Mdot
.62342 .14978 Mdot
.62342 .16209 Mdot
.62342 .219 Mdot
.62342 .26378 Mdot
.62342 .29663 Mdot
.62342 .32951 Mdot
.62413 .19867 Mdot
.62413 .27378 Mdot
.62413 .32889 Mdot
.62465 .09843 Mdot
.62465 .11633 Mdot
.62465 .13417 Mdot
.62465 .2163 Mdot
.62465 .281 Mdot
.62465 .32843 Mdot
.62504 .17294 Mdot
.62504 .28646 Mdot
.62504 .32809 Mdot
.62534 .06951 Mdot
.62534 .08352 Mdot
.62534 .09751 Mdot
.62534 .18948 Mdot
.62534 .25369 Mdot
.62534 .29073 Mdot
.62534 .32783 Mdot
.62559 .13225 Mdot
.62559 .14483 Mdot
.62559 .2152 Mdot
.62559 .26074 Mdot
.62559 .29416 Mdot
.62559 .32761 Mdot
.62783 .07169 Mdot
.62783 .09005 Mdot
.62783 .1141 Mdot
.62783 .12693 Mdot
.62783 .13973 Mdot
.62783 .15123 Mdot
.62783 .18463 Mdot
.62783 .19166 Mdot
.62783 .21128 Mdot
.62783 .25007 Mdot
.62783 .2576 Mdot
.62783 .26892 Mdot
.62783 .27702 Mdot
.62783 .28309 Mdot
.62783 .28782 Mdot
.62783 .2916 Mdot
.62783 .32565 Mdot
.63016 .09529 Mdot
.63016 .10838 Mdot
.63016 .12144 Mdot
.63016 .19438 Mdot
.63016 .25436 Mdot
.63016 .28895 Mdot
.63016 .32361 Mdot
.63042 .17959 Mdot
.63042 .2848 Mdot
.63042 .32338 Mdot
.63076 .12847 Mdot
.63076 .14483 Mdot
.63076 .22023 Mdot
.63076 .27958 Mdot
.63076 .32309 Mdot
.63118 .20484 Mdot
.63118 .27283 Mdot
.63118 .32271 Mdot
.63176 .16243 Mdot
.63176 .26378 Mdot
.63176 .32221 Mdot
.63213 .13522 Mdot
.63213 .14723 Mdot
.63213 .20263 Mdot
.63213 .21447 Mdot
.63213 .25798 Mdot
.63213 .28991 Mdot
.63213 .32189 Mdot
.63257 .0625 Mdot
.63257 .07584 Mdot
.63257 .08916 Mdot
.63257 .10247 Mdot
.63257 .18993 Mdot
.63257 .28621 Mdot
.63257 .3215 Mdot
.63312 .15964 Mdot
.63312 .17435 Mdot
.63312 .22795 Mdot
.63312 .28165 Mdot
.63312 .32102 Mdot
.63381 .10464 Mdot
.63381 .12144 Mdot
.63381 .2152 Mdot
.63381 .2759 Mdot
.63381 .32041 Mdot
.63439 .11793 Mdot
.63439 .13019 Mdot
.63439 .1424 Mdot
.63439 .19867 Mdot
.63439 .25494 Mdot
.63439 .2874 Mdot
.63439 .31991 Mdot
.63471 .19867 Mdot
.63471 .26842 Mdot
.63471 .31962 Mdot
.63508 .18533 Mdot
.63508 .28336 Mdot
.63508 .31931 Mdot
.63593 .15389 Mdot
.63593 .22357 Mdot
.63593 .25831 Mdot
.63593 .27837 Mdot
.63593 .31856 Mdot
.63672 .08755 Mdot
.63672 .10006 Mdot
.63672 .11254 Mdot
.63672 .12499 Mdot
.63672 .19457 Mdot
.63672 .2848 Mdot
.63672 .31787 Mdot
.63702 .0625 Mdot
.63702 .07973 Mdot
.63702 .09693 Mdot
.63702 .20994 Mdot
.63702 .27205 Mdot
.63702 .31761 Mdot
.63768 .16688 Mdot
.63768 .18056 Mdot
.63768 .23045 Mdot
.63768 .2804 Mdot
.63768 .31703 Mdot
.63844 .26378 Mdot
.63844 .31637 Mdot
.63886 .13249 Mdot
.63886 .14791 Mdot
.63886 .219 Mdot
.63886 .27495 Mdot
.63886 .31599 Mdot
.63914 .06886 Mdot
.63914 .08158 Mdot
.63914 .09429 Mdot
.63914 .10697 Mdot
.63914 .19034 Mdot
.63914 .23622 Mdot
.63914 .2821 Mdot
.63914 .31575 Mdot
.64038 .12144 Mdot
.64038 .16167 Mdot
.64038 .20444 Mdot
.64038 .22643 Mdot
.64038 .26801 Mdot
.64038 .27733 Mdot
.64038 .31467 Mdot
.64164 .07542 Mdot
.64164 .17319 Mdot
.64164 .18597 Mdot
.64164 .23263 Mdot
.64164 .27932 Mdot
.64164 .31357 Mdot
.64192 .11014 Mdot
.64192 .12595 Mdot
.64192 .21424 Mdot
.64192 .27139 Mdot
.64192 .31332 Mdot
.64237 .16517 Mdot
.64237 .25887 Mdot
.64237 .31292 Mdot
.64271 .09245 Mdot
.64271 .1044 Mdot
.64271 .11633 Mdot
.64271 .19475 Mdot
.64271 .23788 Mdot
.64271 .281 Mdot
.64271 .31263 Mdot
.64318 .14197 Mdot
.64318 .15627 Mdot
.64318 .22224 Mdot
.64318 .27414 Mdot
.64318 .31221 Mdot
.6439 .19867 Mdot
.6439 .26378 Mdot
.6439 .31158 Mdot
.64423 .16843 Mdot
.64423 .18144 Mdot
.64423 .22891 Mdot
.64423 .27643 Mdot
.64423 .3113 Mdot
.64512 .0625 Mdot
.64512 .07061 Mdot
.64512 .07466 Mdot
.64512 .08681 Mdot
.64512 .09895 Mdot
.64512 .103 Mdot
.64512 .11107 Mdot
.64512 .19072 Mdot
.64512 .20927 Mdot
.64512 .23454 Mdot
.64512 .26767 Mdot
.64512 .27837 Mdot
.64512 .31052 Mdot
.64611 .13608 Mdot
.64611 .15066 Mdot
.64611 .2179 Mdot
.64611 .27082 Mdot
.64611 .30966 Mdot
.64653 .15736 Mdot
.64653 .30928 Mdot
.64692 .1635 Mdot
.64692 .22505 Mdot
.64692 .27344 Mdot
.64692 .30894 Mdot
.64761 .06867 Mdot
.64761 .08101 Mdot
.64761 .17435 Mdot
.64761 .18655 Mdot
.64761 .23108 Mdot
.64761 .25933 Mdot
.64761 .27565 Mdot
.64761 .30834 Mdot
.64846 .20408 Mdot
.64846 .26378 Mdot
.64846 .3076 Mdot
.64915 .10006 Mdot
.64915 .11503 Mdot
.64915 .12996 Mdot
.64915 .21337 Mdot
.64915 .26737 Mdot
.64915 .307 Mdot
.64971 .14483 Mdot
.64971 .15841 Mdot
.64971 .22105 Mdot
.64971 .27034 Mdot
.64971 .3065 Mdot
.65019 .16983 Mdot
.65019 .2275 Mdot
.65019 .27283 Mdot
.65019 .30609 Mdot
.65094 .12761 Mdot
.65094 .30543 Mdot
.65151 .16752 Mdot
.65151 .30493 Mdot
.65196 .18198 Mdot
.65196 .25971 Mdot
.65196 .30454 Mdot
.65232 .0625 Mdot
.65232 .07781 Mdot
.65232 .09311 Mdot
.65232 .10838 Mdot
.65232 .20867 Mdot
.65232 .26378 Mdot
.65232 .30422 Mdot
.65261 .12541 Mdot
.65261 .1393 Mdot
.65261 .21691 Mdot
.65261 .26712 Mdot
.65261 .30397 Mdot
.65286 .15248 Mdot
.65286 .16517 Mdot
.65286 .2238 Mdot
.65286 .26992 Mdot
.65286 .30375 Mdot
.65562 .0703 Mdot
.65562 .07367 Mdot
.65562 .096 Mdot
.65562 .11826 Mdot
.65562 .11941 Mdot
.65562 .14095 Mdot
.65562 .14742 Mdot
.65562 .16034 Mdot
.65562 .17111 Mdot
.65562 .17591 Mdot
.65562 .18842 Mdot
.65562 .21998 Mdot
.65562 .22622 Mdot
.65562 .26003 Mdot
.65562 .2669 Mdot
.65562 .30133 Mdot
.65826 .15456 Mdot
.65826 .16668 Mdot
.65826 .22267 Mdot
.65826 .26671 Mdot
.65826 .29902 Mdot
.65848 .129 Mdot
.65848 .1422 Mdot
.65848 .21601 Mdot
.65848 .26378 Mdot
.65848 .29883 Mdot
.65875 .06975 Mdot
.65875 .08426 Mdot
.65875 .09874 Mdot
.65875 .1132 Mdot
.65875 .20814 Mdot
.65875 .2603 Mdot
.65875 .29859 Mdot
.65907 .18297 Mdot
.65907 .25612 Mdot
.65907 .29831 Mdot
.65946 .16956 Mdot
.65946 .29797 Mdot
.65996 .13287 Mdot
.65996 .29754 Mdot
.66059 .29698 Mdot
.66099 .14978 Mdot
.66099 .16209 Mdot
.66099 .219 Mdot
.66099 .26378 Mdot
.66099 .29663 Mdot
.66145 .10988 Mdot
.66145 .12336 Mdot
.66145 .1368 Mdot
.66145 .21191 Mdot
.66145 .26054 Mdot
.66145 .29623 Mdot
.66201 .0625 Mdot
.66201 .07727 Mdot
.66201 .09202 Mdot
.66201 .18898 Mdot
.66201 .25669 Mdot
.66201 .29574 Mdot
.66267 .17729 Mdot
.66267 .29516 Mdot
.6635 .14483 Mdot
.6635 .16291 Mdot
.6635 .29444 Mdot
.66381 .13225 Mdot
.66381 .14483 Mdot
.66381 .2152 Mdot
.66381 .26074 Mdot
.66381 .29416 Mdot
.66454 .0625 Mdot
.66454 .07628 Mdot
.66454 .08317 Mdot
.66454 .09005 Mdot
.66454 .1038 Mdot
.66454 .11752 Mdot
.66454 .19416 Mdot
.66454 .25718 Mdot
.66454 .29353 Mdot
.6654 .18384 Mdot
.6654 .29278 Mdot
.66589 .29234 Mdot
.66644 .17136 Mdot
.66644 .29186 Mdot
.66674 .1141 Mdot
.66674 .12693 Mdot
.66674 .13973 Mdot
.66674 .19867 Mdot
.66674 .21128 Mdot
.66674 .2576 Mdot
.66674 .2916 Mdot
.66774 .06951 Mdot
.66774 .08352 Mdot
.66774 .09751 Mdot
.66774 .13741 Mdot
.66774 .18948 Mdot
.66774 .29073 Mdot
.66867 .13522 Mdot
.66867 .14723 Mdot
.66867 .20263 Mdot
.66867 .21447 Mdot
.66867 .25798 Mdot
.66867 .28991 Mdot
.66893 .1785 Mdot
.66893 .28969 Mdot
.66938 .0731 Mdot
.66938 .28929 Mdot
.66977 .08219 Mdot
.66977 .09529 Mdot
.66977 .10838 Mdot
.66977 .12144 Mdot
.66977 .19438 Mdot
.66977 .28895 Mdot
.67039 .14823 Mdot
.67039 .28841 Mdot
.67106 .18463 Mdot
.67106 .28782 Mdot
.67155 .11793 Mdot
.67155 .13019 Mdot
.67155 .1424 Mdot
.67155 .19867 Mdot
.67155 .25494 Mdot
.67155 .2874 Mdot
.6722 .09133 Mdot
.6722 .11051 Mdot
.6722 .28683 Mdot
.67262 .15736 Mdot
.67262 .17294 Mdot
.67262 .28646 Mdot
.67291 .0625 Mdot
.67291 .07584 Mdot
.67291 .08916 Mdot
.67291 .10247 Mdot
.67291 .18993 Mdot
.67291 .28621 Mdot
.67452 .08755 Mdot
.67452 .10006 Mdot
.67452 .11254 Mdot
.67452 .12393 Mdot
.67452 .14136 Mdot
.67452 .17959 Mdot
.67452 .19457 Mdot
.67452 .20584 Mdot
.67452 .2848 Mdot
.67616 .18533 Mdot
.67616 .28336 Mdot
.67646 .15123 Mdot
.67646 .28309 Mdot
.6769 .0625 Mdot
.6769 .08219 Mdot
.6769 .21159 Mdot
.6769 .28271 Mdot
.67759 .06886 Mdot
.67759 .08158 Mdot
.67759 .09429 Mdot
.67759 .10697 Mdot
.67759 .19034 Mdot
.67759 .2821 Mdot
.67812 .15964 Mdot
.67812 .17435 Mdot
.67812 .22795 Mdot
.67812 .28165 Mdot
.67885 .09245 Mdot
.67885 .09843 Mdot
.67885 .1044 Mdot
.67885 .11633 Mdot
.67885 .19475 Mdot
.67885 .2163 Mdot
.67885 .23788 Mdot
.67885 .281 Mdot
.67954 .16688 Mdot
.67954 .18056 Mdot
.67954 .23045 Mdot
.67954 .2804 Mdot
.67997 .19867 Mdot
.67997 .28002 Mdot
.68048 .12847 Mdot
.68048 .14483 Mdot
.68048 .22023 Mdot
.68048 .27958 Mdot
.68078 .0625 Mdot
.68078 .07542 Mdot
.68078 .18597 Mdot
.68078 .23263 Mdot
.68078 .27932 Mdot
.68186 .0625 Mdot
.68186 .07466 Mdot
.68186 .08681 Mdot
.68186 .09895 Mdot
.68186 .11107 Mdot
.68186 .15389 Mdot
.68186 .19072 Mdot
.68186 .2053 Mdot
.68186 .22357 Mdot
.68186 .23454 Mdot
.68186 .27837 Mdot
.68305 .16167 Mdot
.68305 .17562 Mdot
.68305 .22643 Mdot
.68305 .27733 Mdot
.68341 .07169 Mdot
.68341 .09005 Mdot
.68341 .10838 Mdot
.68341 .21071 Mdot
.68341 .27702 Mdot
.68408 .16843 Mdot
.68408 .18144 Mdot
.68408 .22891 Mdot
.68408 .27643 Mdot
.68469 .10464 Mdot
.68469 .12144 Mdot
.68469 .2152 Mdot
.68469 .2759 Mdot
.68498 .06867 Mdot
.68498 .08101 Mdot
.68498 .09334 Mdot
.68498 .17435 Mdot
.68498 .18655 Mdot
.68498 .23108 Mdot
.68498 .27565 Mdot
.68577 .13249 Mdot
.68577 .14791 Mdot
.68577 .19106 Mdot
.68577 .219 Mdot
.68577 .27495 Mdot
.68669 .15627 Mdot
.68669 .22224 Mdot
.68669 .27414 Mdot
.68711 .19867 Mdot
.68711 .27378 Mdot
.68749 .1635 Mdot
.68749 .22505 Mdot
.68749 .27344 Mdot
.68819 .16983 Mdot
.68819 .18224 Mdot
.68819 .20484 Mdot
.68819 .2275 Mdot
.68819 .27283 Mdot
.68909 .0625 Mdot
.68909 .07973 Mdot
.68909 .09693 Mdot
.68909 .1141 Mdot
.68909 .20994 Mdot
.68909 .27205 Mdot
.68984 .11014 Mdot
.68984 .12595 Mdot
.68984 .21424 Mdot
.68984 .27139 Mdot
.69049 .13608 Mdot
.69049 .15066 Mdot
.69049 .2179 Mdot
.69049 .27082 Mdot
.69104 .15841 Mdot
.69104 .22105 Mdot
.69104 .27034 Mdot
.69152 .16517 Mdot
.69152 .2238 Mdot
.69152 .26992 Mdot
.69195 .17111 Mdot
.69195 .18297 Mdot
.69195 .22622 Mdot
.69195 .26955 Mdot
.69266 .19166 Mdot
.69266 .26892 Mdot
.69323 .19867 Mdot
.69323 .26842 Mdot
.6937 .0713 Mdot
.6937 .20444 Mdot
.6937 .26801 Mdot
.6941 .07061 Mdot
.6941 .08681 Mdot
.6941 .103 Mdot
.6941 .11914 Mdot
.6941 .20927 Mdot
.6941 .26767 Mdot
.69443 .11503 Mdot
.69443 .12996 Mdot
.69443 .21337 Mdot
.69443 .26737 Mdot
.69472 .1393 Mdot
.69472 .15312 Mdot
.69472 .21691 Mdot
.69472 .26712 Mdot
.69497 .14742 Mdot
.69497 .16034 Mdot
.69497 .21998 Mdot
.69497 .2669 Mdot
.69519 .16668 Mdot
.69519 .22267 Mdot
.69519 .26671 Mdot
.69854 .0625 Mdot
.69854 .07904 Mdot
.69854 .09311 Mdot
.69854 .09556 Mdot
.69854 .11941 Mdot
.69854 .12361 Mdot
.69854 .13356 Mdot
.69854 .1422 Mdot
.69854 .14978 Mdot
.69854 .16209 Mdot
.69854 .18425 Mdot
.69854 .19217 Mdot
.69854 .20408 Mdot
.69854 .20867 Mdot
.69854 .21601 Mdot
.69854 .26378 Mdot
.70201 .14483 Mdot
.70201 .15736 Mdot
.70201 .2152 Mdot
.70201 .26074 Mdot
.70224 .12336 Mdot
.70224 .1368 Mdot
.70224 .21191 Mdot
.70224 .26054 Mdot
.70251 .09874 Mdot
.70251 .1132 Mdot
.70251 .12761 Mdot
.70251 .20814 Mdot
.70251 .2603 Mdot
.70282 .0703 Mdot
.70282 .0859 Mdot
.70282 .10147 Mdot
.70282 .20377 Mdot
.70282 .26003 Mdot
.70319 .19867 Mdot
.70319 .25971 Mdot
.70362 .19262 Mdot
.70362 .25933 Mdot
.70414 .25887 Mdot
.70479 .15389 Mdot
.70479 .25831 Mdot
.70517 .14723 Mdot
.70517 .1592 Mdot
.70517 .21447 Mdot
.70517 .25798 Mdot
.70559 .12693 Mdot
.70559 .13973 Mdot
.70559 .21128 Mdot
.70559 .2576 Mdot
.70608 .1038 Mdot
.70608 .11752 Mdot
.70608 .1312 Mdot
.70608 .20766 Mdot
.70608 .25718 Mdot
.70664 .0625 Mdot
.70664 .07727 Mdot
.70664 .09202 Mdot
.70664 .10675 Mdot
.70664 .20349 Mdot
.70664 .25669 Mdot
.70729 .19867 Mdot
.70729 .25612 Mdot
.70806 .193 Mdot
.70806 .25545 Mdot
.70864 .13019 Mdot
.70864 .1424 Mdot
.70864 .21071 Mdot
.70864 .25494 Mdot
.70896 .25466 Mdot
.70931 .12144 Mdot
.70931 .13445 Mdot
.70931 .20722 Mdot
.70931 .25436 Mdot
.71007 .06951 Mdot
.71007 .08352 Mdot
.71007 .09751 Mdot
.71007 .11148 Mdot
.71007 .15736 Mdot
.71007 .20324 Mdot
.71007 .25369 Mdot
.71094 .07002 Mdot
.71094 .08505 Mdot
.71094 .19867 Mdot
.71094 .25293 Mdot
.71143 .14483 Mdot
.71143 .2525 Mdot
.71196 .19334 Mdot
.71196 .25204 Mdot
.71224 .12499 Mdot
.71224 .13741 Mdot
.71224 .20683 Mdot
.71224 .25179 Mdot
.71316 .08916 Mdot
.71316 .10247 Mdot
.71316 .11576 Mdot
.71316 .16956 Mdot
.71316 .20302 Mdot
.71316 .25099 Mdot
.7142 .0625 Mdot
.7142 .07676 Mdot
.7142 .091 Mdot
.7142 .19867 Mdot
.7142 .25007 Mdot
.71459 .16034 Mdot
.71459 .24973 Mdot
.71542 .19365 Mdot
.71542 .24901 Mdot
.71596 .09429 Mdot
.71596 .10697 Mdot
.71596 .11963 Mdot
.71596 .13225 Mdot
.71596 .20282 Mdot
.71596 .24854 Mdot
.71633 .14913 Mdot
.71633 .24821 Mdot
.71683 .17136 Mdot
.71683 .24777 Mdot
.71714 .06928 Mdot
.71714 .08283 Mdot
.71714 .09637 Mdot
.71714 .10988 Mdot
.71714 .19867 Mdot
.71714 .2475 Mdot
.7185 .11107 Mdot
.7185 .12316 Mdot
.7185 .13522 Mdot
.7185 .16291 Mdot
.7185 .17959 Mdot
.7185 .19391 Mdot
.7185 .20263 Mdot
.7185 .24631 Mdot
.7198 .0625 Mdot
.7198 .07542 Mdot
.7198 .08833 Mdot
.7198 .10123 Mdot
.7198 .1141 Mdot
.7198 .18597 Mdot
.7198 .19867 Mdot
.7198 .24517 Mdot
.72008 .17294 Mdot
.72008 .24493 Mdot
.72052 .15279 Mdot
.72052 .24455 Mdot
.72128 .0625 Mdot
.72128 .07628 Mdot
.72128 .18056 Mdot
.72128 .19416 Mdot
.72128 .24388 Mdot
.72193 .16517 Mdot
.72193 .24331 Mdot
.72222 .06867 Mdot
.72222 .08101 Mdot
.72222 .09334 Mdot
.72222 .10565 Mdot
.72222 .11793 Mdot
.72222 .18655 Mdot
.72222 .19867 Mdot
.72222 .24306 Mdot
.72298 .11826 Mdot
.72298 .14042 Mdot
.72298 .17435 Mdot
.72298 .2424 Mdot
.72378 .06906 Mdot
.72378 .08219 Mdot
.72378 .09529 Mdot
.72378 .18144 Mdot
.72378 .19438 Mdot
.72378 .24169 Mdot
.72412 .15594 Mdot
.72412 .24139 Mdot
.72495 .16715 Mdot
.72495 .24067 Mdot
.72557 .17562 Mdot
.72557 .24012 Mdot
.72606 .0625 Mdot
.72606 .07503 Mdot
.72606 .08755 Mdot
.72606 .10006 Mdot
.72606 .18224 Mdot
.72606 .19457 Mdot
.72606 .2397 Mdot
.72677 .12437 Mdot
.72677 .14483 Mdot
.72677 .23908 Mdot
.72727 .15869 Mdot
.72727 .23864 Mdot
.72763 .16891 Mdot
.72763 .23832 Mdot
.72791 .17676 Mdot
.72791 .23808 Mdot
.72813 .06849 Mdot
.72813 .08047 Mdot
.72813 .09245 Mdot
.72813 .1044 Mdot
.72813 .11633 Mdot
.72813 .18297 Mdot
.72813 .19475 Mdot
.72813 .23788 Mdot
.73003 .0625 Mdot
.73003 .06886 Mdot
.73003 .08158 Mdot
.73003 .08546 Mdot
.73003 .10838 Mdot
.73003 .12962 Mdot
.73003 .14863 Mdot
.73003 .16111 Mdot
.73003 .17779 Mdot
.73003 .23622 Mdot
.73196 .0625 Mdot
.73196 .07466 Mdot
.73196 .08681 Mdot
.73196 .17873 Mdot
.73196 .23454 Mdot
.73219 .17191 Mdot
.73219 .23434 Mdot
.73248 .16325 Mdot
.73248 .23408 Mdot
.73286 .13417 Mdot
.73286 .15192 Mdot
.73286 .23375 Mdot
.73338 .09429 Mdot
.73338 .11542 Mdot
.73338 .23329 Mdot
.73414 .17319 Mdot
.73414 .23263 Mdot
.73466 .16517 Mdot
.73466 .23217 Mdot
.73534 .13817 Mdot
.73534 .15481 Mdot
.73534 .23158 Mdot
.73591 .17435 Mdot
.73591 .23108 Mdot
.73625 .12144 Mdot
.73625 .23078 Mdot
.73662 .16688 Mdot
.73662 .23045 Mdot
.73753 .07431 Mdot
.73753 .14169 Mdot
.73753 .15736 Mdot
.73753 .22966 Mdot
.73839 .16843 Mdot
.73839 .22891 Mdot
.73873 .12665 Mdot
.73873 .14483 Mdot
.73873 .22861 Mdot
.73948 .14483 Mdot
.73948 .15964 Mdot
.73948 .22795 Mdot
.74 .16983 Mdot
.74 .2275 Mdot
.74038 .0625 Mdot
.74038 .08426 Mdot
.74038 .10597 Mdot
.74038 .22717 Mdot
.74089 .1312 Mdot
.74089 .14823 Mdot
.74089 .22672 Mdot
.74123 .14765 Mdot
.74123 .16167 Mdot
.74123 .22643 Mdot
.74146 .17111 Mdot
.74146 .22622 Mdot
.7428 .07258 Mdot
.7428 .09274 Mdot
.7428 .11284 Mdot
.7428 .13522 Mdot
.7428 .15018 Mdot
.7428 .15123 Mdot
.7428 .1635 Mdot
.7428 .22505 Mdot
.74423 .15248 Mdot
.74423 .16517 Mdot
.74423 .2238 Mdot
.7445 .13878 Mdot
.7445 .15389 Mdot
.7445 .22357 Mdot
.74489 .10006 Mdot
.74489 .11877 Mdot
.74489 .22322 Mdot
.74552 .15456 Mdot
.74552 .16668 Mdot
.74552 .22267 Mdot
.74601 .14197 Mdot
.74601 .15627 Mdot
.74601 .22224 Mdot
.74671 .10643 Mdot
.74671 .12393 Mdot
.74671 .22163 Mdot
.74737 .14483 Mdot
.74737 .15841 Mdot
.74737 .22105 Mdot
.74779 .22068 Mdot
.7483 .11204 Mdot
.7483 .12847 Mdot
.7483 .22023 Mdot
.7486 .14742 Mdot
.7486 .16034 Mdot
.7486 .21998 Mdot
.74972 .0625 Mdot
.74972 .08317 Mdot
.74972 .13249 Mdot
.74972 .14978 Mdot
.74972 .16209 Mdot
.74972 .219 Mdot
.75098 .13608 Mdot
.75098 .15066 Mdot
.75098 .2179 Mdot
.75137 .07212 Mdot
.75137 .09133 Mdot
.75137 .21755 Mdot
.75211 .1393 Mdot
.75211 .15312 Mdot
.75211 .21691 Mdot
.7528 .0625 Mdot
.7528 .08047 Mdot
.7528 .09843 Mdot
.7528 .11633 Mdot
.7528 .2163 Mdot
.75313 .1422 Mdot
.75313 .15535 Mdot
.75313 .21601 Mdot
.75405 .08781 Mdot
.75405 .10464 Mdot
.75405 .12144 Mdot
.75405 .14483 Mdot
.75405 .15736 Mdot
.75405 .2152 Mdot
.75489 .14723 Mdot
.75489 .1592 Mdot
.75489 .21447 Mdot
.75516 .11014 Mdot
.75516 .12595 Mdot
.75516 .21424 Mdot
.75567 .21379 Mdot
.75614 .11503 Mdot
.75614 .12996 Mdot
.75614 .21337 Mdot
.75702 .11941 Mdot
.75702 .13356 Mdot
.75702 .2126 Mdot
.75782 .12336 Mdot
.75782 .1368 Mdot
.75782 .21191 Mdot
.75819 .21159 Mdot
.75854 .12693 Mdot
.75854 .13973 Mdot
.75854 .15248 Mdot
.75854 .21128 Mdot
.75919 .07169 Mdot
.75919 .09005 Mdot
.75919 .1424 Mdot
.75919 .15456 Mdot
.75919 .21071 Mdot
.76006 .0625 Mdot
.76006 .07973 Mdot
.76006 .09693 Mdot
.76006 .20994 Mdot
.76084 .07061 Mdot
.76084 .08681 Mdot
.76084 .103 Mdot
.76084 .11914 Mdot
.76084 .20927 Mdot
.76152 .09311 Mdot
.76152 .10838 Mdot
.76152 .12361 Mdot
.76152 .20867 Mdot
.76213 .09874 Mdot
.76213 .1132 Mdot
.76213 .12761 Mdot
.76213 .20814 Mdot
.76268 .11752 Mdot
.76268 .1312 Mdot
.76268 .20766 Mdot
.76317 .12144 Mdot
.76317 .13445 Mdot
.76317 .20722 Mdot
.76362 .12499 Mdot
.76362 .13741 Mdot
.76362 .20683 Mdot
.76404 .20647 Mdot
.76476 .20584 Mdot
.76537 .2053 Mdot
.7659 .20484 Mdot
.76636 .0713 Mdot
.76636 .20444 Mdot
.76676 .0625 Mdot
.76676 .07904 Mdot
.76676 .09556 Mdot
.76676 .20408 Mdot
.76712 .0703 Mdot
.76712 .0859 Mdot
.76712 .10147 Mdot
.76712 .20377 Mdot
.76744 .0625 Mdot
.76744 .07727 Mdot
.76744 .09202 Mdot
.76744 .10675 Mdot
.76744 .12144 Mdot
.76744 .20349 Mdot
.76772 .08352 Mdot
.76772 .09751 Mdot
.76772 .11148 Mdot
.76772 .12541 Mdot
.76772 .20324 Mdot
.76798 .10247 Mdot
.76798 .11576 Mdot
.76798 .129 Mdot
.76798 .20302 Mdot
.76821 .10697 Mdot
.76821 .11963 Mdot
.76821 .13225 Mdot
.76821 .20282 Mdot
.76842 .12316 Mdot
.76842 .13522 Mdot
.76842 .20263 Mdot
.77295 .06928 Mdot
.77295 .07002 Mdot
.77295 .0784 Mdot
.77295 .08283 Mdot
.77295 .08505 Mdot
.77295 .091 Mdot
.77295 .09637 Mdot
.77295 .19867 Mdot
.77742 .09245 Mdot
.77742 .1044 Mdot
.77742 .11633 Mdot
.77742 .19475 Mdot
.77763 .07503 Mdot
.77763 .08755 Mdot
.77763 .10006 Mdot
.77763 .11254 Mdot
.77763 .12499 Mdot
.77763 .19457 Mdot
.77786 .06906 Mdot
.77786 .08219 Mdot
.77786 .09529 Mdot
.77786 .10838 Mdot
.77786 .19438 Mdot
.77811 .0625 Mdot
.77811 .07628 Mdot
.77811 .09005 Mdot
.77811 .1038 Mdot
.77811 .19416 Mdot
.77838 .06975 Mdot
.77838 .08426 Mdot
.77838 .19391 Mdot
.77869 .0625 Mdot
.77869 .19365 Mdot
.77904 .19334 Mdot
.77942 .193 Mdot
.77987 .19262 Mdot
.78037 .19217 Mdot
.78096 .19166 Mdot
.78165 .19106 Mdot
.78204 .0625 Mdot
.78204 .07466 Mdot
.78204 .08681 Mdot
.78204 .09895 Mdot
.78204 .11107 Mdot
.78204 .19072 Mdot
.78247 .06886 Mdot
.78247 .08158 Mdot
.78247 .09429 Mdot
.78247 .10697 Mdot
.78247 .19034 Mdot
.78293 .0625 Mdot
.78293 .07584 Mdot
.78293 .08916 Mdot
.78293 .10247 Mdot
.78293 .18993 Mdot
.78345 .06951 Mdot
.78345 .08352 Mdot
.78345 .18948 Mdot
.78403 .18898 Mdot
.78467 .18842 Mdot
.78539 .18779 Mdot
.7862 .18708 Mdot
.7868 .06867 Mdot
.7868 .08101 Mdot
.7868 .09334 Mdot
.7868 .10565 Mdot
.7868 .18655 Mdot
.78713 .18626 Mdot
.78747 .0625 Mdot
.78747 .07542 Mdot
.78747 .08833 Mdot
.78747 .18597 Mdot
.78819 .06928 Mdot
.78819 .18533 Mdot
.789 .18463 Mdot
.78943 .18425 Mdot
.78989 .18384 Mdot
.79089 .06849 Mdot
.79089 .08047 Mdot
.79089 .09245 Mdot
.79089 .18297 Mdot
.79173 .0625 Mdot
.79173 .07503 Mdot
.79173 .08755 Mdot
.79173 .18224 Mdot
.79202 .18198 Mdot
.79264 .18144 Mdot
.7933 .18087 Mdot
.79364 .18056 Mdot
.79476 .17959 Mdot
.79574 .07466 Mdot
.79574 .17873 Mdot
.796 .1785 Mdot
.79644 .17811 Mdot
.79681 .17779 Mdot
.79739 .17729 Mdot
.79799 .17676 Mdot
.79841 .17639 Mdot
.79896 .17591 Mdot
.79929 .17562 Mdot
.80074 .17435 Mdot
.80207 .17319 Mdot
.80235 .17294 Mdot
.80278 .17256 Mdot
.80354 .17191 Mdot
.80416 .17136 Mdot
.80444 .17111 Mdot
.80516 .17049 Mdot
.8059 .16983 Mdot
.80621 .16956 Mdot
.80696 .16891 Mdot
.80751 .16843 Mdot
.80794 .16806 Mdot
.80855 .16752 Mdot
.80897 .16715 Mdot
.80928 .16688 Mdot
.80951 .16668 Mdot
.81124 .16517 Mdot
.81314 .1635 Mdot
.81342 .16325 Mdot
.81381 .16291 Mdot
.81437 .16243 Mdot
.81475 .16209 Mdot
.81523 .16167 Mdot
.81587 .16111 Mdot
.81676 .16034 Mdot
.81756 .15964 Mdot
.81806 .1592 Mdot
.81864 .15869 Mdot
.81897 .15841 Mdot
.82016 .15736 Mdot
.82141 .15627 Mdot
.82178 .15594 Mdot
.82246 .15535 Mdot
.82307 .15481 Mdot
.82336 .15456 Mdot
.82413 .15389 Mdot
.825 .15312 Mdot
.82539 .15279 Mdot
.82574 .15248 Mdot
.82637 .15192 Mdot
.82717 .15123 Mdot
.82782 .15066 Mdot
.82836 .15018 Mdot
.82883 .13741 Mdot
.82883 .14978 Mdot
.82957 .14913 Mdot
.83014 .14863 Mdot
.83059 .14823 Mdot
.83096 .14791 Mdot
.83126 .14765 Mdot
.83151 .13445 Mdot
.83151 .14742 Mdot
.83173 .13522 Mdot
.83173 .14723 Mdot
.83447 .13225 Mdot
.83447 .14483 Mdot
.83726 .13019 Mdot
.83726 .1424 Mdot
.83748 .129 Mdot
.83748 .1422 Mdot
.83775 .14197 Mdot
.83806 .14169 Mdot
.83845 .14136 Mdot
.83892 .14095 Mdot
.83952 .14042 Mdot
.84031 .12693 Mdot
.84031 .13973 Mdot
.8408 .12541 Mdot
.8408 .1393 Mdot
.84139 .13878 Mdot
.8421 .13817 Mdot
.84296 .12499 Mdot
.84296 .13741 Mdot
.84366 .12336 Mdot
.84366 .1368 Mdot
.84405 .13645 Mdot
.84448 .12144 Mdot
.84448 .13608 Mdot
.84546 .12316 Mdot
.84546 .13522 Mdot
.84634 .12144 Mdot
.84634 .13445 Mdot
.84666 .13417 Mdot
.84736 .11941 Mdot
.84736 .13356 Mdot
.84815 .13287 Mdot
.84858 .11701 Mdot
.84858 .13249 Mdot
.84885 .11963 Mdot
.84885 .13225 Mdot
.85005 .11752 Mdot
.85005 .1312 Mdot
.85122 .11793 Mdot
.85122 .13019 Mdot
.85147 .11503 Mdot
.85147 .12996 Mdot
.85187 .12962 Mdot
.85257 .11576 Mdot
.85257 .129 Mdot
.85317 .11204 Mdot
.85317 .12847 Mdot
.85416 .1132 Mdot
.85416 .12761 Mdot
.85493 .1141 Mdot
.85493 .12693 Mdot
.85526 .10838 Mdot
.85526 .12665 Mdot
.85606 .11014 Mdot
.85606 .12595 Mdot
.85667 .11148 Mdot
.85667 .12541 Mdot
.85715 .11254 Mdot
.85715 .12499 Mdot
.85786 .12437 Mdot
.85836 .10643 Mdot
.85836 .12393 Mdot
.85873 .10838 Mdot
.85873 .12361 Mdot
.85902 .10988 Mdot
.85902 .12336 Mdot
.85924 .11107 Mdot
.85924 .12316 Mdot
.86121 .10184 Mdot
.86121 .10464 Mdot
.86121 .10675 Mdot
.86121 .12144 Mdot
.86328 .10697 Mdot
.86328 .11963 Mdot
.86353 .10522 Mdot
.86353 .11941 Mdot
.86384 .103 Mdot
.86384 .11914 Mdot
.86426 .10006 Mdot
.86426 .11877 Mdot
.86484 .096 Mdot
.86484 .11826 Mdot
.86522 .10565 Mdot
.86522 .11793 Mdot
.86569 .1038 Mdot
.86569 .11752 Mdot
.86628 .10147 Mdot
.86628 .11701 Mdot
.86705 .09245 Mdot
.86705 .09843 Mdot
.86705 .1044 Mdot
.86705 .11633 Mdot
.86771 .10247 Mdot
.86771 .11576 Mdot
.8681 .09429 Mdot
.8681 .11542 Mdot
.86853 .10006 Mdot
.86853 .11503 Mdot
.8696 .08833 Mdot
.8696 .09693 Mdot
.8696 .10123 Mdot
.8696 .1141 Mdot
.87063 .09874 Mdot
.87063 .1132 Mdot
.87104 .09274 Mdot
.87104 .11284 Mdot
.87138 .08755 Mdot
.87138 .10006 Mdot
.87138 .11254 Mdot
.87195 .09556 Mdot
.87195 .11204 Mdot
.87259 .08352 Mdot
.87259 .09751 Mdot
.87259 .11148 Mdot
.87306 .08681 Mdot
.87306 .09895 Mdot
.87306 .11107 Mdot
.8737 .09133 Mdot
.8737 .11051 Mdot
.87412 .0784 Mdot
.87412 .09429 Mdot
.87412 .11014 Mdot
.87442 .08283 Mdot
.87442 .09637 Mdot
.87442 .10988 Mdot
.87614 .07169 Mdot
.87614 .07781 Mdot
.87614 .08219 Mdot
.87614 .08546 Mdot
.87614 .09005 Mdot
.87614 .09311 Mdot
.87614 .09529 Mdot
.87614 .10838 Mdot
.87774 .06886 Mdot
.87774 .08158 Mdot
.87774 .09429 Mdot
.87774 .10697 Mdot
.878 .07727 Mdot
.878 .09202 Mdot
.878 .10675 Mdot
.87836 .0713 Mdot
.87836 .08888 Mdot
.87836 .10643 Mdot
.87889 .0625 Mdot
.87889 .08426 Mdot
.87889 .10597 Mdot
.87926 .06867 Mdot
.87926 .08101 Mdot
.87926 .09334 Mdot
.87926 .10565 Mdot
.87974 .0625 Mdot
.87974 .07676 Mdot
.87974 .091 Mdot
.87974 .10522 Mdot
.88041 .07094 Mdot
.88041 .08781 Mdot
.88041 .10464 Mdot
.88068 .06849 Mdot
.88068 .08047 Mdot
.88068 .09245 Mdot
.88068 .1044 Mdot
.88137 .0625 Mdot
.88137 .07628 Mdot
.88137 .08317 Mdot
.88137 .09005 Mdot
.88137 .1038 Mdot
.88229 .07061 Mdot
.88229 .08681 Mdot
.88229 .103 Mdot
.88289 .0625 Mdot
.88289 .07584 Mdot
.88289 .08916 Mdot
.88289 .10247 Mdot
.88361 .0625 Mdot
.88361 .08219 Mdot
.88361 .10184 Mdot
.88403 .0703 Mdot
.88403 .0859 Mdot
.88403 .10147 Mdot
.88431 .0625 Mdot
.88431 .07542 Mdot
.88431 .08833 Mdot
.88431 .10123 Mdot
.88565 .0625 Mdot
.88565 .07002 Mdot
.88565 .07503 Mdot
.88565 .08129 Mdot
.88565 .08505 Mdot
.88565 .08755 Mdot
.88565 .10006 Mdot
.88691 .0625 Mdot
.88691 .07466 Mdot
.88691 .08681 Mdot
.88691 .09895 Mdot
.88715 .06975 Mdot
.88715 .08426 Mdot
.88715 .09874 Mdot
.88751 .0625 Mdot
.88751 .08047 Mdot
.88751 .09843 Mdot
.8881 .07431 Mdot
.8881 .09791 Mdot
.88856 .06951 Mdot
.88856 .08352 Mdot
.88856 .09751 Mdot
.88922 .0625 Mdot
.88922 .07973 Mdot
.88922 .09693 Mdot
.88987 .06928 Mdot
.88987 .08283 Mdot
.88987 .09637 Mdot
.89028 .07367 Mdot
.89028 .096 Mdot
.89079 .0625 Mdot
.89079 .07904 Mdot
.89079 .09556 Mdot
.89109 .06906 Mdot
.89109 .08219 Mdot
.89109 .09529 Mdot
.89224 .0625 Mdot
.89224 .06886 Mdot
.89224 .0731 Mdot
.89224 .0784 Mdot
.89224 .08158 Mdot
.89224 .09429 Mdot
.89333 .06867 Mdot
.89333 .08101 Mdot
.89333 .09334 Mdot
.89359 .0625 Mdot
.89359 .07781 Mdot
.89359 .09311 Mdot
.89401 .07258 Mdot
.89401 .09274 Mdot
.89435 .06849 Mdot
.89435 .08047 Mdot
.89435 .09245 Mdot
.89484 .0625 Mdot
.89484 .07727 Mdot
.89484 .09202 Mdot
.89562 .07212 Mdot
.89562 .09133 Mdot
.896 .0625 Mdot
.896 .07676 Mdot
.896 .091 Mdot
.89708 .0625 Mdot
.89708 .07169 Mdot
.89708 .07628 Mdot
.89708 .09005 Mdot
.8981 .0625 Mdot
.8981 .07584 Mdot
.8981 .08916 Mdot
.89842 .0713 Mdot
.89842 .08888 Mdot
.89905 .0625 Mdot
.89905 .07542 Mdot
.89905 .08833 Mdot
.89965 .07094 Mdot
.89965 .08781 Mdot
.89994 .0625 Mdot
.89994 .07503 Mdot
.89994 .08755 Mdot
.90078 .0625 Mdot
.90078 .07061 Mdot
.90078 .07466 Mdot
.90078 .08681 Mdot
.90183 .0703 Mdot
.90183 .0859 Mdot
.90233 .0625 Mdot
.90233 .08546 Mdot
.9028 .07002 Mdot
.9028 .08505 Mdot
.90371 .0625 Mdot
.90371 .06975 Mdot
.90371 .08426 Mdot
.90455 .06951 Mdot
.90455 .08352 Mdot
.90495 .0625 Mdot
.90495 .08317 Mdot
.90534 .06928 Mdot
.90534 .08283 Mdot
.90607 .0625 Mdot
.90607 .06906 Mdot
.90607 .08219 Mdot
.90677 .06886 Mdot
.90677 .08158 Mdot
.9071 .0625 Mdot
.9071 .08129 Mdot
.90742 .06867 Mdot
.90742 .08101 Mdot
.90803 .0625 Mdot
.90803 .06849 Mdot
.90803 .08047 Mdot
.90888 .0625 Mdot
.90888 .07973 Mdot
.90967 .0625 Mdot
.90967 .07904 Mdot
.9104 .0625 Mdot
.9104 .0784 Mdot
.91107 .0625 Mdot
.91107 .07781 Mdot
.9117 .0625 Mdot
.9117 .07727 Mdot
.91228 .0625 Mdot
.91228 .07676 Mdot
.91282 .0625 Mdot
.91282 .07628 Mdot
.91333 .0625 Mdot
.91333 .07584 Mdot
.91381 .0625 Mdot
.91381 .07542 Mdot
.91425 .0625 Mdot
.91425 .07503 Mdot
.91467 .0625 Mdot
.91467 .07466 Mdot
.91507 .07431 Mdot
.9158 .07367 Mdot
.91646 .0731 Mdot
.91705 .07258 Mdot
.91758 .07212 Mdot
.91807 .07169 Mdot
.91852 .0713 Mdot
.91893 .07094 Mdot
.91931 .07061 Mdot
.91966 .0703 Mdot
.91998 .07002 Mdot
.92028 .06975 Mdot
.92056 .06951 Mdot
.92082 .06928 Mdot
.92107 .06906 Mdot
.9213 .06886 Mdot
.92152 .06867 Mdot
.92172 .06849 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{442, 221},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`40006j0000gB000`400?l00000o`00003oo`3ooonk
0?ooo`00o`3ooonk0?ooo`00o`3ooonk0?ooo`00o`3ooonk0?ooo`00=`3oool2000003<0oooo1000
00030?ooo`800000<@3oool4000000<0oooo0P00000a0?ooo`800000103oool2000003<0oooo00<0
00000?ooo`3oool00P3oool200000340oooo0`0000040?ooo`800000<@3oool2000000@0oooo0P00
000U0?ooo`00=P3oool010000000oooo0?ooo`00000d0?ooo`050000003oool0oooo0?ooo`000000
0P3oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool0
0000;`3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo00000380oooo00D00000
0?ooo`3oool0oooo000000020?ooo`030000003oool0oooo0340oooo00@000000?ooo`3oool00000
0P3oool00`000000oooo0?ooo`0]0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3o
ool00000903oool003H0oooo00@000000?ooo`3oool00000=03oool01@000000oooo0?ooo`3oool0
00000080oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`020?ooo`040000003oool0
oooo00000380oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`0]0?ooo`D00000
0P3oool010000000oooo0?ooo`00000c0?ooo`040000003oool0oooo00000080oooo00<000000?oo
o`3oool0;@3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000002@0oooo000f
0?ooo`040000003oool0oooo000003@0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000
003oool0oooo0300oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003oool0oooo02h0
oooo0P0000030?ooo`040000003oool0oooo000002l0oooo00@000000?ooo`3oool000000`3oool0
10000000oooo0?ooo`00000`0?ooo`<000000`3oool010000000oooo0?ooo`00000_0?ooo`<00000
0`3oool010000000oooo0?ooo`00000T0?ooo`00=P3oool010000000oooo0?ooo`00000d0?ooo`05
0000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`0a0?ooo`040000003oool0oooo
00000080oooo00<000000?ooo`3oool0<03oool010000000oooo0?ooo`0000020?ooo`030000003o
ool0oooo02h0oooo00<000000?ooo`0000000`3oool010000000oooo0?ooo`00000`0?ooo`030000
003oool0oooo00<0oooo00@000000?ooo`3oool00000;`3oool00`000000oooo0?ooo`030?ooo`04
0000003oool0oooo000002@0oooo000f0?ooo`040000003oool0oooo00000380oooo0`0000030?oo
o`040000003oool0oooo00000300oooo00@000000?ooo`3oool000000P3oool010000000oooo0?oo
o`00000_0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool00000<@3oool20000
00<0oooo00@000000?ooo`3oool00000<03oool00`000000oooo0?ooo`030?ooo`040000003oool0
oooo000002l0oooo00<000000?ooo`3oool00`3oool010000000oooo0?ooo`00000T0?ooo`00=`3o
ool2000003D0oooo00<000000?ooo`3oool00P3oool200000380oooo0P0000040?ooo`800000<@3o
ool2000000@0oooo0P00000c0?ooo`030000003oool0oooo0080oooo0P00000a0?ooo`@000000`3o
ool200000340oooo0`0000030?ooo`8000009@3oool00?l0oooo^`3oool00?l0oooo^`3oool00?l0
oooo^`3oool00?l0oooo^`3oool001/0ooooo`00002E000000/0oooo000K0?ooo`030000003oool0
oooo00<0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0
oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0
oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`030?ooo`030000003oool0
oooo00T0oooo000K0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool0=P3oool00`00
0000oooo0?ooo`0g0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0=`3oool00`00
0000oooo0?ooo`0g0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool06P3oool00`00
0000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`3o0?oooi40oooo00<000000?ooo`3o
ool02@3oool001/0oooo00<000000?ooo`3oool0o`3ooonA0?ooo`030000003oool0oooo00T0oooo
000K0?ooo`<00000o`3ooon?0?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool0o`3ooonA
0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0?l0ooooT@3oool00`00
0000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`3o0?oooi40oooo00<000000?ooo`3o
ool02@3oool001/0oooo00<000000?ooo`3oool0o`3ooonA0?ooo`030000003oool0oooo00T0oooo
000K0?ooo`<00000o`3ooon?0?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool0o`3ooonA
0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0?l0ooooT@3oool00`00
0000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`3o0?oooi40oooo00<000000?ooo`3o
ool02@3oool001/0oooo00<000000?ooo`3oool0o`3ooonA0?ooo`030000003oool0oooo00T0oooo
000K0?ooo`<00000o`3ooon?0?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool0o`3ooonA
0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0?l0ooooT@3oool00`00
0000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`3o0?oooi40oooo00<000000?ooo`3o
ool02@3oool001/0oooo00<000000?ooo`3oool0o`3ooonA0?ooo`030000003oool0oooo00T0oooo
000K0?ooo`<00000o`3ooon?0?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool0o`3ooonA
0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0?l0ooooT@3oool00`00
0000oooo0?ooo`090?ooo`004P3oool2000000L0oooo00<000000?ooo`3oool0o`3ooonA0?ooo`03
0000003oool0oooo00T0oooo000A0?ooo`040000003oool0oooo000000H0oooo00<000000?ooo`3o
ool06P3oool2000000<0oooo4@00000Q0?ooo`L000000`3oool2000000040?ooo`00000000000?oo
o`<000000P3oool2000000030?ooo`000000000000L0000000@0oooo000000000000oooo0P000002
0?ooo`80000000<0oooo00000000000010000000103oool000000000000000020?ooo`80000000@0
oooo000000000000oooo100000040?ooo`800000103oool4000000030?ooo`000000000000800000
0P3oool200000080oooo100000020?ooo`80000000<0oooo0000000000000P000000103oool00000
0000000000030?ooo`<000002@3oool300000080oooo1P0000000`3oool000000000000300000080
oooo100000001@3oool00000000000000000oooo00D0000000<0oooo0000000000000P0000040?oo
o`800000103oool4000000040?ooo`00000000000?ooo`8000000P3oool3000000030?ooo`000000
000000@0000000@0oooo000000000000oooo0P0000020?ooo`80000000<0oooo0000000000001`00
00000`3oool00000000000020?ooo`<0000000@0oooo000000000000oooo0P0000020?ooo`P00000
8@3oool@00000240oooo00<000000?ooo`3oool02@3oool00140oooo00@000000?ooo`3oool00000
1P3oool4000001T0oooo0P0000030?oooa4000008@3oool7000000<0oooo0P000000103oool00000
0000003oool300000080oooo0P0000000`3oool0000000000007000000040?ooo`00000000000?oo
o`8000000P3oool2000000030?ooo`000000000000@0000000@0oooo00000000000000000P3oool2
000000040?ooo`00000000000?ooo`@00000103oool2000000@0oooo100000000`3oool000000000
000200000080oooo0P0000020?ooo`@000000P3oool2000000030?ooo`00000000000080000000@0
oooo00000000000000000`3oool3000000T0oooo0`0000020?ooo`H0000000<0oooo000000000000
0`0000020?ooo`@0000000D0oooo00000000000000000?ooo`05000000030?ooo`00000000000080
0000103oool2000000@0oooo10000000103oool000000000003oool200000080oooo0`0000000`3o
ool0000000000004000000040?ooo`00000000000?ooo`8000000P3oool2000000030?ooo`000000
000000L0000000<0oooo0000000000000P3oool3000000040?ooo`00000000000?ooo`8000000P3o
ool800000240oooo4000000N0?ooo`@000002`3oool00140oooo00@000000?ooo`3oool000001P3o
ool00`000000oooo0?ooo`0L0?ooo`<000000`3oool3000000<0oooo0`0000020?ooo`<000008@3o
ool700000180oooo1@00000=0?ooo`80000000<0oooo0000000000002P3oool2000000030?ooo`00
0000000000X0oooo0P0000090?ooo`80000000<0oooo000000000000103oool2000000<0oooo0P00
00020?ooo`800000103oool200000080oooo0P0000090?ooo`8000000P3oool2000000@0oooo0P00
00020?ooo`8000000P3oool2000000D0oooo0P0000000`3oool00000000000090?ooo`8000002@3o
ool200000080oooo0P00000:0?ooo`80000000<0oooo0000000000003@3oool500000140oooo2000
000Q0?ooo`<000000P3oool3000000<0oooo0`0000030?ooo`8000007P3oool00`000000oooo0?oo
o`090?ooo`004@3oool010000000oooo0?ooo`0000060?ooo`030000003oool0oooo01`0oooo1000
00020?ooo`l00000803oool9000000030?ooo`000000000000<000002P3oool:000000040?ooo`00
000000000?ooo`8000000P3oool2000000030?ooo`00000000000080oooo0P0000000`3oool00000
00000002000000030?ooo`000000000000<000000`3oool3000000@0oooo0`0000030?ooo`800000
00<0oooo0000000000000P0000000`3oool00000000000020?ooo`H0000000D0oooo000000000000
00000?ooo`03000000030?ooo`00000000000080000000@0oooo00000000000000002@3oool30000
00030?ooo`00000000000080000000<0oooo0000000000001@0000000`3oool00000000000030000
0080oooo0`0000000`3oool0000000000004000000@0oooo0`0000040?ooo`<000000P3oool20000
00050?ooo`00000000000000003oool00`0000020?ooo`8000000P3oool5000000050?ooo`000000
00000000003oool00P000000103oool000000000003oool7000000X0oooo0P0000000`3oool00000
0000000:00000200oooo3`0000020?ooo`<000007P3oool00`000000oooo0?ooo`090?ooo`004@3o
ool010000000oooo0?ooo`0000060?ooo`030000003oool0oooo01d0oooo100000020?ooo`h00000
8@3oool8000000030?ooo`000000000000<0000000<0oooo0000000000000P3oool2000000P0oooo
1@000000103oool000000000003oool200000080oooo0P0000040?ooo`<0000000<0oooo00000000
00000P0000020?ooo`8000001@3oool3000000D0oooo0`0000020?ooo`80000000@0oooo00000000
000000000P3oool200000080oooo0P0000020?ooo`8000000P3oool3000000040?ooo`0000000000
0?ooo`80000000<0oooo0000000000000P00000;0?ooo`<000000P3oool2000000040?ooo`000000
00000?ooo`8000000`3oool2000000030?ooo`00000000000080oooo0P0000030?ooo`D000000`3o
ool3000000D0oooo0`0000050?ooo`8000000P3oool300000080oooo0`0000030?ooo`8000000P3o
ool3000000040?ooo`00000000000?ooo`80000000<0oooo000000000000203oool200000080oooo
0P000000103oool000000000003oool4000000030?ooo`000000000000@000008@3oool>00000080
oooo1000000N0?ooo`030000003oool0oooo00T0oooo000B0?ooo`8000001`3oool00`000000oooo
0?ooo`0N0?ooo`@000000`3oool:00000200oooo2P0000080?ooo`8000000P3oool;000000d0oooo
0`0000000`3oool000000000000300000080oooo0P0000020?ooo`<000000P3oool2000000T0oooo
0`0000050?ooo`8000000P3oool200000080oooo0P0000040?ooo`80000000<0oooo000000000000
103oool2000000040?ooo`00000000000?ooo`8000002@3oool4000000030?ooo`000000000000D0
oooo100000050?ooo`80000000<0oooo0000000000000P3oool2000000H0oooo0`0000090?ooo`80
000000<0oooo0000000000000P0000020?ooo`8000000P3oool5000000040?ooo`00000000000000
00d0oooo100000000`3oool000000000000400000080oooo0P0000080?ooo`X00000803oool:0000
00<0oooo1000000O0?ooo`030000003oool0oooo00T0oooo000K0?ooo`<000007`3oool400000080
oooo3@00000M0?ooo`/0000000<0oooo0000000000002P3oool;000000040?ooo`00000000000?oo
o`8000001@3oool3000000040?ooo`00000000000?ooo`<000000P3oool2000000040?ooo`000000
000000000080oooo0P0000040?ooo`800000103oool4000000<0oooo0P0000020?ooo`8000000P3o
ool3000000<0oooo0P0000000`3oool00000000000030?ooo`<00000103oool2000000T0oooo0P00
00030?ooo`@000000`3oool4000000@0oooo0`0000001@3oool00000000000000000oooo00<00000
0`3oool4000000@0oooo0P0000040?ooo`80000000<0oooo0000000000000P0000000`3oool00000
000000020?ooo`<0000000@0oooo000000000000oooo0`0000040?ooo`8000000P3oool200000003
0?ooo`000000000000@0000000<0oooo0000000000000P00000:0?ooo`h000007@3oool=00000003
0?ooo`000000000000<000007P3oool3000000/0oooo000K0?ooo`030000003oool0oooo0200oooo
100000030?ooo`/000007`3oool6000000040?ooo`00000000000?ooo`80000000@0oooo00000000
0000oooo0P0000070?ooo`80000000@0oooo000000000000oooo0P000000103oool000000000003o
ool200000080oooo100000000`3oool00000000000040?ooo`8000000P3oool4000000040?ooo`00
00000000000000H0oooo0P0000000`3oool00000000000030?ooo`8000000`3oool3000000030?oo
o`000000000000<0oooo0`0000020?ooo`D000000`3oool2000000040?ooo`00000000000?ooo`<0
00002@3oool3000000040?ooo`00000000000?ooo`8000000`3oool5000000030?ooo`0000000000
008000000P3oool3000000040?ooo`0000000000000000<0oooo0P0000030?ooo`80000000<0oooo
0000000000001P3oool200000080oooo100000020?ooo`800000103oool2000000030?ooo`000000
00000080000000<0oooo0000000000000P3oool2000000040?ooo`00000000000?ooo`D000001`3o
ool2000000040?ooo`00000000000?ooo`@000000P3oool6000001l0oooo2`0000020?ooo`D00000
8@3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`0R0?ooo`<000000`3o
ool:000001l0oooo2@000000103oool000000000003oool2000000040?ooo`0000000000000000H0
oooo1`0000000`3oool00000000000060?ooo`@0000000@0oooo000000000000oooo100000050?oo
o`80000000@0oooo00000000000000000`3oool2000000040?ooo`00000000000?ooo`800000103o
ool2000000<0oooo0P0000000`3oool00000000000020?ooo`T000000P3oool200000080oooo1000
0000103oool000000000000000090?ooo`<0000000<0oooo0000000000000P0000020?ooo`800000
0P3oool9000000050?ooo`00000000000000003oool00P0000020?ooo`8000001@3oool200000004
0?ooo`00000000000?ooo`8000000`3oool200000080oooo0P0000050?ooo`@0000000@0oooo0000
00000000oooo100000060?ooo`80000000<0oooo0000000000001@0000060?ooo`<0000000@0oooo
000000000000oooo3000000O0?ooo`X000000P3oool4000002<0oooo00<000000?ooo`3oool02@3o
ool001/0oooo00<000000?ooo`3oool08`3oool3000000<0oooo2@00000R0?ooo`H0000000<0oooo
0000000000001@3oool200000080oooo0P0000020?ooo`@0000000<0oooo0000000000000`000000
0`3oool00000000000080?ooo`80000000<0oooo0000000000001000000<0?ooo`80000000<0oooo
0000000000000P3oool2000000<0oooo100000040?ooo`8000000P3oool600000080oooo0P000000
0`3oool00000000000020?ooo`800000103oool2000000T0oooo0P0000040?ooo`8000000P3oool4
000000<0oooo1P0000000`3oool00000000000050?ooo`<00000103oool200000080oooo0P000000
0`3oool000000000000;0?ooo`80000000<0oooo0000000000000P0000000`3oool0000000000008
0?ooo`80000000@0oooo000000000000oooo0P0000000`3oool000000000000200000080oooo0P00
00020?ooo`8000001@3oool900000280oooo2@0000020?ooo`@00000903oool00`000000oooo0?oo
o`090?ooo`006`3oool00`000000oooo0?ooo`0T0?ooo`@000000P3oool9000001h0oooo2@000000
0`3oool0000000000002000000030?ooo`000000000000@0oooo0P0000000`3oool0000000000002
000000040?ooo`00000000000?ooo`8000000P3oool2000000030?ooo`000000000000<0oooo0P00
00000`3oool00000000000040?ooo`800000103oool2000000030?ooo`000000000000X0oooo1000
00030?ooo`@000000P3oool200000080oooo0P0000040?ooo`@0000000<0oooo000000000000103o
ool2000000040?ooo`0000000000000000T0oooo0`0000000`3oool00000000000040?ooo`H00000
103oool200000080oooo0P0000020?ooo`D000000`3oool4000000T0oooo0P0000020?ooo`800000
0`3oool2000000D0oooo0P0000000`3oool00000000000030?ooo`80000000@0oooo000000000000
oooo0P0000020?ooo`L0000000<0oooo000000000000103oool4000000030?ooo`000000000000X0
00007@3oool9000000<0oooo1000000U0?ooo`030000003oool0oooo00T0oooo000K0?ooo`<00000
9@3oool400000080oooo2000000N0?ooo`T000000`3oool6000000H0oooo100000000`3oool00000
00000003000000D0oooo0P0000020?ooo`<0000000D0oooo00000000000000000?ooo`02000000H0
oooo0P0000000`3oool0000000000006000000@0oooo100000040?ooo`<000000P3oool200000080
oooo0P0000040?ooo`8000000`3oool3000000030?ooo`00000000000080000000@0oooo00000000
000000000`3oool2000000@0oooo0`000000103oool000000000000000020?ooo`<000000P3oool2
000000@0oooo0`0000000`3oool00000000000020?ooo`@00000103oool4000000@0oooo1`000002
0?ooo`8000001P3oool2000000050?ooo`00000000000000003oool00P0000030?ooo`8000001@3o
ool7000000030?ooo`000000000000H0oooo1@0000030?ooo`X000007@3oool900000080oooo1000
000T0?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool09P3oool400000080oooo2@00000L
0?ooo`T0000000<0oooo0000000000000`0000000`3oool00000000000020?ooo`8000000`3oool2
000000040?ooo`00000000000?ooo`D0000000<0oooo000000000000103oool4000000040?ooo`00
000000000?ooo`<000000P3oool200000080oooo0P0000020?ooo`P0000000<0oooo000000000000
0P0000060?ooo`<000001@3oool4000000<0oooo10000000103oool000000000003oool500000003
0?ooo`000000000000<0oooo0P0000040?ooo`L000000P3oool2000000030?ooo`00000000000080
00000P3oool2000000030?ooo`000000000000D0oooo0`0000060?ooo`@0000000<0oooo00000000
00001P0000020?ooo`8000000P3oool200000080oooo0`0000000`3oool0000000000005000000<0
oooo0P0000020?ooo`D0000000<0oooo0000000000000P0000040?ooo`8000000P3oool200000003
0?ooo`000000000000d00000703oool900000080oooo1000000W0?ooo`030000003oool0oooo00T0
oooo000K0?ooo`030000003oool0oooo02L0oooo100000020?ooo`P00000703oool9000000030?oo
o`0000000000008000001P3oool2000000030?ooo`000000000000L000000P3oool2000000030?oo
o`000000000000@0oooo1@000000103oool000000000003oool2000000040?ooo`00000000000000
0080oooo0`0000020?ooo`8000000`3oool2000000030?ooo`000000000000H000000`3oool20000
00030?ooo`00000000000080oooo100000030?ooo`H0000000@0oooo000000000000oooo0`000000
0`3oool00000000000090?ooo`H0000000@0oooo000000000000oooo1P0000020?ooo`D000000P3o
ool2000000030?ooo`00000000000080oooo2@0000000`3oool00000000000030?ooo`80000000<0
oooo0000000000000P0000020?ooo`<0000000@0oooo000000000000oooo0P000000103oool00000
0000003oool2000000<0oooo0P0000020?ooo`80000000<0oooo0000000000001`0000020?ooo`80
00001P3oool>000001/0oooo2@0000000`3oool0000000000002000002T0oooo00<000000?ooo`3o
ool02@3oool001/0oooo00<000000?ooo`3oool0:03oool5000000030?ooo`000000000000D00000
803oool5000000030?ooo`000000000000<0000000<0oooo0000000000001@3oool200000080oooo
0P000000103oool000000000003oool2000000040?ooo`000000000000000080oooo0P0000000`3o
ool0000000000002000000050?ooo`00000000000000003oool0100000020?ooo`D0000000<0oooo
0000000000000P0000020?ooo`80000000<0oooo0000000000001@0000060?ooo`8000000P3oool3
000000030?ooo`00000000000080000000<0oooo0000000000000P000000103oool000000000003o
ool200000080oooo0P0000090?ooo`8000000P3oool2000000040?ooo`00000000000?ooo`@00000
00<0oooo0000000000000P000000103oool000000000000000020?ooo`8000001@3oool800000003
0?ooo`00000000000080oooo10000000103oool000000000003oool200000080oooo0`0000020?oo
o`<0000000<0oooo0000000000000P000000103oool000000000003oool4000000030?ooo`000000
000000<0000000@0oooo000000000000oooo0P0000060?ooo`80000000<0oooo0000000000002@00
000O0?ooo`P0000000<0oooo0000000000000P00000Z0?ooo`030000003oool0oooo00T0oooo000K
0?ooo`030000003oool0oooo02X0oooo100000020?ooo`H000007`3oool400000080oooo1@000000
103oool000000000003oool2000000D0oooo0P0000000`3oool00000000000020?ooo`80000000<0
oooo0000000000000`3oool2000000030?ooo`000000000000<0000000@0oooo000000000000oooo
0P0000020?ooo`@0000000<0oooo0000000000000`3oool200000080oooo0P0000000`3oool00000
00000002000000T0oooo0P0000060?ooo`@000000`3oool2000000040?ooo`00000000000?ooo`80
00000P3oool2000000T0oooo0P0000020?ooo`80000000@0oooo000000000000oooo0P0000030?oo
o`@000001@3oool2000000X0oooo100000000`3oool00000000000020?ooo`8000000P3oool30000
00040?ooo`000000000000000080oooo0P0000020?ooo`80000000<0oooo0000000000000`000000
0`3oool00000000000030?ooo`80000000<0oooo0000000000000P3oool4000000H0oooo0P000000
103oool000000000003oool500000080oooo1000000O0?ooo`H000000P3oool4000002/0oooo00<0
00000?ooo`3oool02@3oool001/0oooo0`00000[0?ooo`@000000P3oool6000001d0oooo1@000000
0`3oool0000000000003000000D0oooo0P0000050?ooo`80000000@0oooo000000000000oooo1P00
00000`3oool00000000000060?ooo`<0000000<0oooo000000000000103oool4000000030?ooo`00
00000000008000000P3oool400000080oooo10000000103oool000000000000000050?ooo`800000
0P3oool200000080oooo0P0000050?ooo`80000000@0oooo0000000000000000103oool2000000T0
oooo0P0000030?ooo`@0000000<0oooo000000000000103oool2000000<0oooo0P0000000`3oool0
0000000000060?ooo`<0000000<0oooo0000000000000P0000020?ooo`@000000P3oool300000080
oooo0`0000050?ooo`80000000@0oooo00000000000000001P3oool2000000050?ooo`0000000000
0000003oool00P0000000`3oool0000000000002000000H0oooo0P0000040?ooo`80000000<0oooo
0000000000000P3oool4000001h0oooo1P0000020?ooo`@00000:P3oool3000000/0oooo000K0?oo
o`030000003oool0oooo02`0oooo100000020?ooo`H000006`3oool6000000040?ooo`0000000000
0?ooo`80000000@0oooo000000000000oooo0P0000000`3oool00000000000020?ooo`80000000<0
oooo0000000000000P3oool4000000<0oooo0P0000020?ooo`@0000000@0oooo000000000000oooo
0P0000060?ooo`<0000000@0oooo00000000000000000P3oool4000000030?ooo`000000000000@0
oooo1`0000020?ooo`80000000<0oooo0000000000000P3oool5000000H0oooo100000000`3oool0
00000000000;0?ooo`80000000@0oooo00000000000000001`3oool4000000<0oooo0P0000000`3o
ool00000000000020?ooo`L00000103oool2000000030?ooo`0000000000008000000P3oool20000
0080oooo0`0000060?ooo`80000000<0oooo0000000000000P0000000`3oool00000000000020?oo
o`8000000P3oool2000000030?ooo`00000000000080oooo100000030?ooo`80000000@0oooo0000
00000000oooo100000000`3oool00000000000020?ooo`H000006`3oool600000080oooo1000000]
0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo02d0oooo100000020?oo
o`H000006P3oool6000000030?ooo`000000000000<0000000@0oooo000000000000oooo0P000000
0`3oool00000000000020?ooo`80000000<0oooo0000000000000P0000020?ooo`8000001`3oool5
000000030?ooo`000000000000D0oooo0P0000020?ooo`L000000`3oool2000000030?ooo`000000
000000@000000P3oool500000080oooo0P0000000`3oool00000000000040?ooo`D0000000<0oooo
0000000000000`3oool2000000040?ooo`0000000000000000P0oooo100000000`3oool000000000
00030?ooo`80000000<0oooo0000000000000`0000040?ooo`80000000<0oooo0000000000000P3o
ool500000080oooo1P0000000`3oool00000000000030?ooo`L000000P3oool2000000D0oooo0P00
00000`3oool0000000000003000000L0oooo0P000000103oool000000000003oool500000080oooo
0P000000103oool000000000003oool2000000030?ooo`0000000000008000000P3oool6000001X0
oooo1P0000020?ooo`@00000;P3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo
0?ooo`0^0?ooo`@000000P3oool5000001`0oooo0`0000020?ooo`D0000000<0oooo000000000000
203oool200000080oooo0`0000020?ooo`8000001@3oool:000000030?ooo`00000000000080oooo
0P0000000`3oool000000000000200000080oooo0P0000030?ooo`<000000P3oool8000000P0oooo
0`0000040?ooo`@0000000<0oooo0000000000000P3oool300000080oooo0P0000080?ooo`<00000
0P3oool300000080oooo0P0000000`3oool0000000000002000000@0oooo0`0000080?ooo`<00000
00<0oooo0000000000000P0000000`3oool0000000000002000000<0oooo0P0000000`3oool00000
00000003000000040?ooo`00000000000?ooo`8000000P3oool7000000030?ooo`000000000000D0
oooo0P000000103oool000000000000000030?ooo`800000203oool2000000030?ooo`0000000000
008000000P3oool4000001`0oooo1@0000020?ooo`@00000;`3oool00`000000oooo0?ooo`090?oo
o`006`3oool00`000000oooo0?ooo`0`0?ooo`<000000`3oool4000001/0oooo100000020?ooo`@0
000000<0oooo0000000000000P0000000`3oool00000000000070?ooo`<0000000@0oooo00000000
0000oooo0P0000030?ooo`80000000@0oooo000000000000oooo0`0000020?ooo`<00000103oool5
000000L0oooo0P0000060?ooo`8000000`3oool2000000030?ooo`000000000000<0oooo1@000000
0`3oool0000000000002000000D0oooo0`0000020?ooo`8000002@3oool200000080oooo0`000005
0?ooo`@0000000<0oooo0000000000000`0000020?ooo`8000000P3oool200000080oooo0P000006
0?ooo`<000001P3oool2000000040?ooo`0000000000000000<0oooo0`0000030?ooo`<0000000@0
oooo000000000000oooo0P0000030?ooo`80000000@0oooo000000000000oooo0P0000080?ooo`@0
000000@0oooo000000000000oooo100000020?ooo`@000006`3oool4000000<0oooo0`00000a0?oo
o`030000003oool0oooo00T0oooo000K0?ooo`<00000<03oool5000000030?ooo`000000000000<0
0000703oool2000000030?ooo`000000000000P0000000@0oooo000000000000oooo100000020?oo
o`80000000D0oooo00000000000000000?ooo`02000000030?ooo`0000000000008000000P3oool4
000000040?ooo`00000000000?ooo`8000000P3oool2000000030?ooo`0000000000008000000P3o
ool300000080oooo0P0000070?ooo`80000000D0oooo00000000000000000?ooo`02000000<0oooo
1@0000030?ooo`@00000203oool2000000<0oooo0P0000040?ooo`800000203oool4000000<0oooo
1@0000020?ooo`8000000P3oool3000000030?ooo`000000000000L0oooo0P0000020?ooo`<00000
00@0oooo000000000000oooo100000020?ooo`8000000P3oool2000000040?ooo`00000000000000
00<0oooo10000000103oool000000000003oool3000000040?ooo`00000000000?ooo`@000000P3o
ool<00000080oooo0P00000L0?ooo`@000000P3oool5000002l0oooo0`00000;0?ooo`006`3oool0
0`000000oooo0?ooo`0a0?ooo`D0000000<0oooo0000000000000`00000K0?ooo`80000000@0oooo
000000000000oooo1P0000000`3oool00000000000020?ooo`@0000000D0oooo0000000000000000
0?ooo`04000000<0oooo100000020?ooo`80000000<0oooo0000000000000`0000020?ooo`@00000
00@0oooo000000000000oooo0P0000000`3oool0000000000002000000X0oooo1P000000103oool0
00000000000000020?ooo`80000000@0oooo000000000000oooo0P0000020?ooo`80000000@0oooo
0000000000000000103oool2000000<0oooo0P0000040?ooo`8000000`3oool4000000030?ooo`00
000000000080oooo100000020?ooo`8000000P3oool200000080oooo1P00000:0?ooo`@0000000@0
oooo000000000000oooo100000000`3oool00000000000020?ooo`L000000`3oool400000080oooo
0P000000103oool000000000003oool7000000<0oooo0P0000000`3oool000000000000600000080
oooo0P00000K0?ooo`@000000P3oool4000003<0oooo00<000000?ooo`3oool02@3oool001/0oooo
00<000000?ooo`3oool0<`3oool3000000<0oooo0`00000N0?ooo`80000000<0oooo000000000000
0P0000030?ooo`8000001`3oool400000080oooo0P0000000`3oool0000000000002000000L0oooo
0P0000000`3oool00000000000020?ooo`8000001P3oool2000000<0oooo0P0000020?ooo`800000
0`3oool4000000030?ooo`000000000000<0oooo0`0000050?ooo`80000000<0oooo000000000000
0P3oool2000000040?ooo`0000000000000000@0oooo0P0000080?ooo`800000103oool400000003
0?ooo`00000000000080oooo100000060?ooo`800000103oool2000000030?ooo`00000000000080
00000`3oool200000080oooo0P0000030?ooo`8000001P3oool200000080oooo0P0000000`3oool0
0000000000070?ooo`<0000000<0oooo0000000000000P3oool2000000030?ooo`000000000000L0
oooo0P0000030?ooo`H000007P3oool4000000<0oooo0`00000d0?ooo`030000003oool0oooo00T0
oooo000K0?ooo`030000003oool0oooo03@0oooo100000020?ooo`@00000703oool8000000D0oooo
0P0000050?ooo`<0000000@0oooo000000000000oooo1@0000020?ooo`800000103oool200000004
0?ooo`00000000000?ooo`800000103oool600000080oooo0P0000000`3oool00000000000090000
00D0oooo0P0000000`3oool00000000000030?ooo`P000000`3oool7000000P0oooo200000030?oo
o`P000000`3oool2000000030?ooo`000000000000D0oooo2`0000000`3oool00000000000020?oo
o`<0000000<0oooo000000000000103oool400000080oooo0P0000040?ooo`8000000P3oool50000
00030?ooo`000000000000<000001P3oool2000000D0oooo2000000L0?ooo`@000000P3oool40000
03D0oooo00<000000?ooo`3oool02@3oool00140oooo0`0000070?ooo`030000003oool0oooo03D0
oooo100000020?ooo`@00000603oool2000000030?ooo`000000000000H0000000<0oooo00000000
00000P3oool200000080oooo0P000000103oool000000000003oool3000000040?ooo`0000000000
00000080oooo100000020?ooo`@0000000<0oooo0000000000000P3oool200000080oooo1`000002
0?ooo`8000000`3oool9000000030?ooo`00000000000080oooo0P000000103oool0000000000000
00020?ooo`H000001@3oool6000000/0oooo1P0000050?ooo`H000000P3oool3000000030?ooo`00
000000000080oooo0P0000000`3oool0000000000007000000<0oooo0P0000020?ooo`L000000P3o
ool2000000030?ooo`00000000000080oooo100000020?ooo`@0000000<0oooo0000000000000P00
00000`3oool000000000000300000080oooo0P0000020?ooo`8000000P3oool2000000030?ooo`00
0000000000H0000000<0oooo000000000000603oool400000080oooo1000000f0?ooo`030000003o
ool0oooo00T0oooo000D0?ooo`030000003oool0oooo00@0oooo1000000e0?ooo`@000000P3oool3
000001P0oooo0P0000020?ooo`X0000000<0oooo0000000000000`3oool2000000@0oooo0P000000
103oool000000000000000030?ooo`8000000P3oool4000000030?ooo`000000000000@0oooo0`00
00000`3oool0000000000002000000040?ooo`0000000000000000<0oooo0P0000030?ooo`D00000
0`3oool200000080oooo0P0000020?ooo`8000001P3oool4000000L0oooo0`0000090?ooo`<00000
1`3oool4000000H0oooo0P0000020?ooo`8000000P3oool200000080oooo1P0000030?ooo`800000
0`3oool5000000040?ooo`00000000000?ooo`<00000103oool2000000030?ooo`00000000000080
00000P3oool2000000<0oooo0`0000000`3oool00000000000040?ooo`8000000`3oool200000004
0?ooo`00000000000?ooo`L000000P3oool2000001P0oooo0`0000020?ooo`@00000=03oool40000
00/0oooo000D0?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0=`3oool400000080
oooo1000000I0?ooo`80000000<0oooo0000000000000P000000103oool000000000003oool30000
0080oooo0P0000060?ooo`D0000000<0oooo000000000000103oool2000000030?ooo`0000000000
008000000P3oool2000000030?ooo`000000000000@0oooo1@0000030?ooo`8000000`3oool20000
0080oooo0P0000000`3oool0000000000002000000030?ooo`00000000000080oooo0P0000000`3o
ool00000000000030?ooo`@000000`3oool200000080oooo0`0000090?ooo`<000000P3oool20000
00<0oooo100000020?ooo`8000000P3oool200000080oooo0P0000000`3oool00000000000020000
00030?ooo`00000000000080oooo0P0000020?ooo`<000000`3oool5000000<0oooo0`0000000`3o
ool00000000000020?ooo`@0000000<0oooo000000000000103oool2000000030?ooo`0000000000
008000001`3oool200000080oooo0P0000000`3oool00000000000020?ooo`H000006P3oool40000
00030?ooo`000000000000<00000>03oool00`000000oooo0?ooo`090?ooo`004@3oool3000000L0
oooo00<000000?ooo`3oool0>03oool400000080oooo1000000H0?ooo`80000000<0oooo00000000
00000P000000103oool000000000003oool2000000<0oooo0P0000030?ooo`80000000D0oooo0000
0000000000000?ooo`0400000080oooo0P0000030?ooo`80000000<0oooo0000000000000`000000
0`3oool00000000000020?ooo`@000000P3oool200000080oooo0`0000020?ooo`8000000P3oool2
000000030?ooo`00000000000080000000<0oooo0000000000000P3oool5000000<0oooo10000000
0`3oool0000000000002000000<0oooo0P0000080?ooo`<000000`3oool4000000040?ooo`000000
0000000000<0oooo0P000000103oool000000000000000020?ooo`80000000<0oooo000000000000
0P0000000`3oool00000000000020?ooo`8000000P3oool2000000<0oooo0P0000020?ooo`<00000
0P3oool200000080oooo100000020?ooo`8000000`3oool200000080oooo0`0000001@3oool00000
000000000000oooo00800000103oool200000080oooo0P0000020?ooo`80000000<0oooo00000000
00001000000I0?ooo`<000000P3oool4000003X0oooo00<000000?ooo`3oool02@3oool00140oooo
00<000000?ooo`3oool01`3oool00`000000oooo0?ooo`0j0?ooo`@000000P3oool3000001L0oooo
100000000`3oool0000000000003000000030?ooo`00000000000080oooo0P0000030?ooo`<00000
00<0oooo0000000000000P3oool200000080oooo100000040?ooo`H00000103oool4000000040?oo
o`00000000000?ooo`8000000`3oool2000000L0oooo0P0000050?ooo`8000000P3oool3000000<0
oooo2@0000050?ooo`800000203oool3000000D0oooo1P0000000`3oool00000000000030?ooo`<0
000000@0oooo00000000000000001@3oool2000000L0oooo0P0000030?ooo`80000000<0oooo0000
000000000P0000000`3oool00000000000040?ooo`D000001@3oool4000000030?ooo`0000000000
00<0oooo1@0000040?ooo`8000000P3oool2000000030?ooo`000000000000P000005`3oool30000
0080oooo1000000k0?ooo`030000003oool0oooo00T0oooo000A0?ooo`030000003oool0oooo00L0
oooo00<000000?ooo`3oool0>`3oool400000080oooo0`00000F0?ooo`@0000000<0oooo00000000
00000P0000020?ooo`80000000@0oooo00000000000000000`3oool2000000050?ooo`0000000000
0000003oool00P0000030?ooo`8000001P3oool5000000D0oooo0P0000030?ooo`80000000<0oooo
0000000000000P0000020?ooo`8000000`3oool2000000040?ooo`00000000000?ooo`8000000P3o
ool2000000H0oooo100000030?ooo`8000000P3oool2000000<0oooo0P0000080?ooo`800000103o
ool200000080oooo0P0000030?ooo`@000001@3oool300000080oooo0P000000103oool000000000
003oool2000000<0oooo0P0000020?ooo`@0000000<0oooo0000000000000`3oool2000000@0oooo
0P0000000`3oool00000000000070?ooo`8000000`3oool2000000030?ooo`000000000000<00000
103oool3000000030?ooo`00000000000080oooo2@00000F0?ooo`<000000P3oool4000003`0oooo
00<000000?ooo`3oool02@3oool00140oooo100000060?ooo`<00000?03oool400000080oooo0P00
000H0?ooo`L0000000<0oooo0000000000000P3oool2000000<0oooo0P0000020?ooo`80000000D0
oooo00000000000000000?ooo`02000000T0oooo10000000103oool000000000000000080?ooo`80
000000<0oooo0000000000000P3oool2000000<0oooo0P0000020?ooo`@000001P3oool6000000D0
oooo1P0000030?ooo`800000203oool2000000@0oooo1P0000050?ooo`H000001P3oool400000080
oooo0P0000030?ooo`8000000P3oool2000000030?ooo`000000000000P0oooo1@0000000`3oool0
0000000000090?ooo`80000000D0oooo00000000000000000?ooo`0200000080oooo0P0000030?oo
o`8000000P3oool2000000030?ooo`000000000000D00000603oool200000080oooo1000000k0?oo
o`<000002`3oool001/0oooo00<000000?ooo`3oool0?@3oool4000001d0oooo200000000`3oool0
0000000000040?ooo`8000000`3oool8000000H0oooo1@0000020?ooo`@000000P3oool200000004
0?ooo`00000000000?ooo`H0000000<0oooo0000000000000`3oool3000000030?ooo`0000000000
00D0oooo0P0000000`3oool0000000000004000000L0oooo1000000C0?ooo`@000001`3oool60000
00030?ooo`000000000000D0oooo1P0000030?ooo`80000000D0oooo00000000000000000?ooo`02
000000040?ooo`00000000000?ooo`8000000P3oool400000080oooo1@0000060?ooo`<0000000@0
oooo0000000000000000103oool2000000@0oooo0P0000000`3oool0000000000003000000030?oo
o`000000000001d0oooo1000000n0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003o
ool0oooo03h0oooo1000000L0?ooo`@0000000@0oooo000000000000oooo0`0000020?ooo`800000
103oool3000000040?ooo`0000000000000000H0oooo1@0000000`3oool000000000000300000004
0?ooo`00000000000?ooo`80000000<0oooo0000000000001`0000040?ooo`D00000103oool20000
00050?ooo`00000000000000003oool00P0000020?ooo`@000000P3oool200000080oooo10000000
0`3oool000000000000;0?ooo`80000000<0oooo0000000000000P0000020?ooo`8000000P3oool4
00000080oooo0P0000001@3oool00000000000000000oooo008000000`3oool6000000@0oooo2@00
0000103oool000000000003oool2000000030?ooo`0000000000008000000P3oool5000000D0oooo
200000040?ooo`8000000P3oool3000000030?ooo`000000000000D00000703oool4000003l0oooo
00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0@03oool3000001d0oooo2000
00020?ooo`<00000103oool2000000050?ooo`00000000000000003oool00P0000020?ooo`@00000
0P3oool5000000030?ooo`0000000000008000000P3oool2000000030?ooo`000000000000<00000
00@0oooo00000000000000000`3oool2000000030?ooo`0000000000008000000P3oool200000005
0?ooo`00000000000000003oool00P0000040?ooo`H000000P3oool8000000T0oooo200000020?oo
o`<0000000<0oooo000000000000103oool2000000050?ooo`00000000000000003oool00P000002
0?ooo`@0000000<0oooo0000000000000`3oool3000000030?ooo`000000000000D000000`3oool4
000000030?ooo`000000000000<000000P3oool400000080oooo0P0000001@3oool0000000000000
0000oooo00800000103oool300000080oooo2000000L0?ooo`@00000@@3oool00`000000oooo0?oo
o`090?ooo`006`3oool00`000000oooo0?ooo`110?ooo`<00000703oool4000000030?ooo`000000
000000<0oooo0P0000030?ooo`8000000`3oool:000000@0oooo0P0000020?ooo`@000000P3oool4
000000030?ooo`00000000000080oooo1@0000030?ooo`80000000<0oooo0000000000001P000000
103oool000000000000000020?ooo`8000001@3oool2000000030?ooo`00000000000080oooo0P00
00000`3oool0000000000002000000P0oooo1@0000000`3oool00000000000020?ooo`@000001P3o
ool200000080oooo0P0000020?ooo`X00000103oool500000080oooo100000000`3oool000000000
00020?ooo`@0000000@0oooo0000000000000000103oool7000000030?ooo`000000000000<0oooo
0P0000030?ooo`8000000P3oool3000000030?ooo`0000000000008000006`3oool400000480oooo
00<000000?ooo`3oool02@3oool001/0oooo0`0000120?ooo`<000006@3oool2000000030?ooo`00
0000000000@000000P3oool2000000<0oooo0`0000050?ooo`H000000`3oool2000000D0oooo0P00
0000103oool000000000003oool2000000D0oooo100000030?ooo`80000000@0oooo000000000000
00000`3oool500000080oooo0P0000020?ooo`8000000`3oool2000000030?ooo`00000000000080
00001P3oool3000000P0oooo100000060?ooo`@0000000<0oooo0000000000000`3oool200000080
oooo0P0000020?ooo`D000000P3oool300000080oooo0P0000030?ooo`@000001@3oool200000004
0?ooo`00000000000?ooo`800000103oool2000000@0oooo1@0000060?ooo`<000000`3oool20000
0080oooo1P0000000`3oool000000000000H0?ooo`@00000@@3oool3000000/0oooo000K0?ooo`03
0000003oool0oooo04<0oooo1000000G0?ooo`T0000000D0oooo00000000000000000?ooo`040000
00@0oooo0P0000020?ooo`8000000`3oool4000000H0oooo100000050?ooo`80000000<0oooo0000
000000000P3oool2000000030?ooo`00000000000080oooo0`0000020?ooo`H0000000<0oooo0000
000000000P3oool2000000030?ooo`00000000000080000000<0oooo0000000000000P0000060?oo
o`@000001`3oool4000000H0oooo100000000`3oool0000000000002000000030?ooo`0000000000
0080oooo0P0000000`3oool000000000000400000080oooo0P0000030?ooo`80000000<0oooo0000
000000000P3oool2000000030?ooo`000000000000D0oooo100000050?ooo`80000000<0oooo0000
000000000`3oool200000080oooo0P0000040?ooo`@0000000D0oooo00000000000000000?ooo`03
000000040?ooo`00000000000?ooo`8000005`3oool4000004@0oooo00<000000?ooo`3oool02@3o
ool001/0oooo00<000000?ooo`3oool0A03oool4000001H0oooo100000001@3oool0000000000000
0000oooo00<000000P3oool2000000@0oooo1@0000020?ooo`H000000P3oool2000000@0oooo0P00
00000`3oool0000000000003000000030?ooo`000000000000D0oooo0P0000020?ooo`80000000@0
oooo00000000000000000P3oool6000000D0oooo0P0000000`3oool00000000000030?ooo`@00000
1P3oool4000000L0oooo100000060?ooo`@000000`3oool2000000030?ooo`000000000000D0oooo
1P0000020?ooo`<0000000<0oooo0000000000000P3oool2000000D0oooo0P0000000`3oool00000
00000003000000030?ooo`000000000000@0oooo0P0000020?ooo`D000000`3oool5000000@0oooo
0P0000020?ooo`<0000000<0oooo0000000000001P00000F0?ooo`@00000A@3oool00`000000oooo
0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`150?ooo`@000005`3oool5000000040?ooo`00
0000000000000080oooo0P0000030?ooo`L0000000<0oooo0000000000000`0000020?ooo`@00000
1@3oool700000080oooo0P0000050?ooo`80000000@0oooo000000000000oooo100000000`3oool0
0000000000060?ooo`8000000P3oool2000000030?ooo`00000000000080oooo10000000103oool0
00000000000000020?ooo`@000001`3oool400000080oooo1@0000000`3oool00000000000020?oo
o`80000000<0oooo0000000000000P3oool2000000H0oooo0P0000000`3oool00000000000020000
00040?ooo`00000000000?ooo`8000001@3oool200000080oooo1`0000050?ooo`@000000P3oool5
000000030?ooo`000000000000D000000`3oool200000080oooo0`0000000`3oool0000000000003
000001L0oooo100000160?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo
04H0oooo1000000F0?ooo`@0000000D0oooo00000000000000000?ooo`04000000030?ooo`000000
0000008000000P3oool6000000030?ooo`000000000000@00000103oool5000000040?ooo`000000
00000?ooo`80000000<0oooo0000000000000P3oool2000000040?ooo`00000000000?ooo`<00000
00<0oooo0000000000000P0000000`3oool000000000000200000080oooo0`0000001@3oool00000
000000000000oooo00800000103oool2000000040?ooo`0000000000000000@0oooo0P0000070?oo
o`800000103oool5000000D0oooo0P0000001@3oool00000000000000000oooo00<000000P3oool8
00000080oooo0`000000103oool000000000003oool2000000050?ooo`00000000000000003oool0
0P000000103oool000000000003oool5000000@0oooo1P0000001@3oool00000000000000000oooo
008000000P3oool300000080oooo100000001@3oool00000000000000000oooo00@000005P3oool4
000004L0oooo00<000000?ooo`3oool02@3oool001/0oooo0`0000170?ooo`@000005`3oool50000
00030?ooo`00000000000080000000@0oooo00000000000000001P3oool2000000030?ooo`000000
000000@000001P3oool3000000030?ooo`000000000000D0000000<0oooo0000000000000P3oool5
000000040?ooo`0000000000000000<0oooo0P0000000`3oool000000000000200000080oooo0P00
00020?ooo`8000000P3oool200000080oooo0P0000000`3oool0000000000003000000P0oooo0P00
00090?ooo`D0000000<0oooo0000000000000P3oool200000080oooo0P0000020?ooo`8000000P3o
ool6000000@0oooo2@0000000`3oool00000000000020?ooo`L0000000@0oooo0000000000000000
1P3oool8000000H0oooo100000000`3oool0000000000002000000030?ooo`000000000000<00000
5`3oool4000004H0oooo0`00000;0?ooo`006`3oool00`000000oooo0?ooo`190?ooo`<000005`3o
ool700000080oooo0`0000090?ooo`<0000000<0oooo0000000000000P0000070?ooo`@0000000<0
oooo0000000000000`0000050?ooo`@00000303oool3000000050?ooo`00000000000000003oool0
1`0000000`3oool0000000000006000000P0oooo0P0000090?ooo`P0000000<0oooo000000000000
1@0000000`3oool00000000000020?ooo`800000303oool5000000D0oooo2P0000070?ooo`@00000
00@0oooo00000000000000002@3oool300000080oooo0`0000000`3oool000000000000G0?ooo`@0
0000BP3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`1:0?ooo`<00000
5P3oool2000000040?ooo`000000000000000080oooo0`0000030?ooo`8000001@3oool200000003
0?ooo`000000000000<000001P3oool8000000030?ooo`000000000000D0oooo0P000000103oool0
00000000003oool400000080oooo0`0000001@3oool00000000000000000oooo00<0000000<0oooo
0000000000000P000000103oool000000000003oool400000080oooo0P00000B0?ooo`8000000`3o
ool4000000040?ooo`00000000000?ooo`@0000000<0oooo0000000000000P3oool200000080oooo
0`0000020?ooo`@0000000@0oooo000000000000oooo0P0000050?ooo`80000000<0oooo00000000
00000P000000103oool000000000000000060?ooo`D0000000<0oooo0000000000001@3oool20000
00<0oooo0`0000020?ooo`D000005P3oool4000004/0oooo00<000000?ooo`3oool02@3oool001/0
oooo00<000000?ooo`3oool0B`3oool4000001H0oooo200000000`3oool0000000000003000000H0
oooo100000040?ooo`@0000000<0oooo000000000000100000000`3oool00000000000050?ooo`L0
000000<0oooo0000000000000P0000020?ooo`<000000P3oool200000080oooo0P0000000`3oool0
0000000000020?ooo`P000000P3oool200000080oooo100000070?ooo`@0000000<0oooo00000000
00000`3oool8000000030?ooo`00000000000080oooo0P0000000`3oool00000000000030?ooo`<0
00000P3oool4000000030?ooo`00000000000080000000<0oooo000000000000103oool200000080
oooo200000000`3oool00000000000040?ooo`@000001P3oool5000000030?ooo`000000000000H0
00005P3oool4000004`0oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0
C03oool4000001D0oooo1P0000000`3oool0000000000003000000P0oooo0P0000000`3oool00000
00000007000000<0oooo0P0000000`3oool000000000000200000080oooo0P0000000`3oool00000
00000004000000D0oooo0P0000020?ooo`<0000000<0oooo0000000000000P3oool2000000030?oo
o`000000000000X000001@3oool5000000L0oooo1@0000050?ooo`/000000P3oool2000000030?oo
o`00000000000080oooo0`0000020?ooo`8000001@3oool3000000040?ooo`00000000000?ooo`80
000000@0oooo000000000000oooo0P0000000`3oool00000000000020?ooo`80000000<0oooo0000
000000001@0000000`3oool00000000000080?ooo`D0000000D0oooo00000000000000000?ooo`02
000001D0oooo1000001=0?ooo`030000003oool0oooo00T0oooo000K0?ooo`<00000C@3oool40000
01H0oooo0P0000000`3oool0000000000003000000d0oooo1@0000040?ooo`80000000@0oooo0000
00000000oooo0P0000040?ooo`8000001@3oool3000000040?ooo`00000000000?ooo`8000000P3o
ool4000000040?ooo`00000000000?ooo`8000000P3oool500000080oooo0P0000070?ooo`@00000
2@3oool4000000H0oooo0P0000030?ooo`D000000P3oool2000000040?ooo`00000000000?ooo`@0
00000P3oool2000000040?ooo`00000000000?ooo`8000001P3oool2000000@0oooo0P000000103o
ool000000000003oool2000000@0oooo1@00000=0?ooo`L000005`3oool4000004`0oooo0`00000;
0?ooo`006`3oool00`000000oooo0?ooo`1>0?ooo`@000005`3oool7000000h0oooo1P0000000`3o
ool00000000000080?ooo`80000000@0oooo000000000000oooo0P0000001@3oool0000000000000
0000oooo008000000P3oool200000080oooo0`000000103oool000000000003oool200000080oooo
2@0000070?ooo`@000002@3oool4000000H0oooo0P0000000`3oool000000000000500000080oooo
0P000000103oool000000000003oool2000000<0oooo0P0000020?ooo`80000000<0oooo00000000
00000P3oool2000000040?ooo`00000000000?ooo`800000203oool4000000030?ooo`0000000000
008000003@3oool8000001L0oooo1000001?0?ooo`030000003oool0oooo00T0oooo000K0?ooo`03
0000003oool0oooo04l0oooo1000000F0?ooo`L000002@3oool;00000080oooo0P000000103oool0
00000000003oool2000000040?ooo`00000000000?ooo`80000000D0oooo00000000000000000?oo
o`02000000030?ooo`00000000000080oooo0P0000020?ooo`@000000`3oool2000000030?ooo`00
0000000000H000001P3oool2000000030?ooo`0000000000008000002@3oool6000000L0oooo1`00
00020?ooo`8000000P3oool4000000<0oooo0P0000000`3oool00000000000020?ooo`H0000000@0
oooo000000000000oooo0P000000103oool000000000003oool400000080oooo0P0000000`3oool0
000000000007000000P0oooo2000000F0?ooo`@00000D03oool00`000000oooo0?ooo`090?ooo`00
6`3oool00`000000oooo0?ooo`1@0?ooo`@000005P3oool4000000P0oooo0P0000000`3oool00000
00000003000000030?ooo`00000000000080000000<0oooo0000000000000P000000103oool00000
0000003oool200000080oooo0P000000103oool000000000003oool2000000040?ooo`0000000000
0?ooo`8000001`3oool3000000<0oooo0P0000000`3oool0000000000004000000L0oooo20000009
0?ooo`80000000<0oooo0000000000000`0000060?ooo`80000000@0oooo00000000000000000P3o
ool200000080oooo100000060?ooo`8000000P3oool5000000040?ooo`00000000000?ooo`800000
0P3oool2000000030?ooo`00000000000080000000<0oooo000000000000100000020?ooo`L00000
2@3oool3000001L0oooo0`00001B0?ooo`030000003oool0oooo00T0oooo000;0?ooo`@000000`3o
ool2000000L0oooo00<000000?ooo`3oool0DP3oool3000001P0oooo0`0000060?ooo`800000103o
ool7000000030?ooo`000000000000D0oooo10000000103oool000000000003oool2000000030?oo
o`00000000000080oooo0P000000103oool000000000003oool2000000@0oooo0P0000030?ooo`<0
00000P3oool2000000030?ooo`000000000000@0oooo2@00000:0?ooo`X00000103oool200000003
0?ooo`00000000000080oooo0`0000030?ooo`8000000`3oool200000080oooo0P0000000`3oool0
0000000000020?ooo`80000000@0oooo000000000000oooo100000000`3oool00000000000050?oo
o`X000000`3oool2000000L0oooo0`00000G0?ooo`@00000D`3oool00`000000oooo0?ooo`090?oo
o`003@3oool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`3oool0103oool40000
0580oooo1000000F0?ooo`<00000303oool;000000040?ooo`00000000000?ooo`D0000000@0oooo
000000000000oooo0P0000040?ooo`P000001@3oool300000080oooo0P0000020?ooo`80000000<0
oooo000000000000103oool8000000`0oooo0P0000000`3oool0000000000004000000@0oooo0P00
00000`3oool00000000000020?ooo`8000000P3oool3000000@0oooo0P0000000`3oool000000000
0004000000@0oooo0P0000000`3oool0000000000006000000040?ooo`00000000000?ooo`/00000
303oool3000001H0oooo1000001A0?ooo`@000002`3oool000d0oooo00D000000?ooo`3oool0oooo
000000020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0E03oool4000001T0oooo
1`000000103oool000000000003oool;000000030?ooo`00000000000080oooo100000000`3oool0
00000000000300000080oooo0P0000000`3oool0000000000003000000030?ooo`000000000000@0
oooo0`0000020?ooo`8000000P3oool2000000D0oooo2@00000?0?ooo`H0000000<0oooo00000000
00001@3oool200000080oooo0P0000020?ooo`<000000`3oool200000080oooo1@0000000`3oool0
0000000000020?ooo`L0000000<0oooo0000000000000P3oool2000000030?ooo`000000000000T0
000000<0oooo0000000000002000000I0?ooo`@00000E@3oool00`000000oooo0?ooo`090?ooo`00
3@3oool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`3oool0103oool00`000000
oooo0?ooo`1E0?ooo`@000004@3oool>000000040?ooo`00000000000?ooo`X000001`3oool40000
00040?ooo`0000000000000000<0oooo0P0000030?ooo`8000000P3oool2000000030?ooo`000000
000000@000000P3oool2000000T0oooo1P0000000`3oool000000000000?0?ooo`H0000000<0oooo
0000000000002@3oool2000000030?ooo`000000000000L000000`3oool2000000<0oooo0P000003
0?ooo`D0000000<0oooo0000000000001`3oool7000000040?ooo`00000000000?oooa0000004P3o
ool4000005H0oooo00<000000?ooo`3oool02@3oool000d0oooo00D000000?ooo`3oool0oooo0000
00020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0EP3oool300000140oooo2000
000<0?ooo`P000001`3oool4000000030?ooo`00000000000100oooo0`0000000`3oool000000000
00020?ooo`8000002@3oool200000080oooo0P00000D0?ooo`8000000`3oool2000000T0oooo0P00
00001@3oool00000000000000000oooo00<00000403oool4000000030?ooo`000000000000L0oooo
1`00000=0?ooo`L000004P3oool3000005L0oooo00<000000?ooo`3oool02@3oool000/0oooo0`00
00030?ooo`040000003oool0oooo000000H0oooo00<000000?ooo`3oool0F03oool2000001<0oooo
2P0000020?ooo`@0000000@0oooo000000000000oooo100000070?ooo`8000000P3oool200000080
oooo0P0000030?ooo`<000000P3oool800000080oooo0`0000020?ooo`800000103oool300000004
0?ooo`00000000000?ooo`8000001@3oool2000000T0oooo0P0000050?ooo`80000000<0oooo0000
00000000100000040?ooo`8000000P3oool2000000<0oooo1@0000000`3oool00000000000020?oo
o`8000000`3oool2000000<0oooo0P0000000`3oool00000000000080?ooo`@0000000@0oooo0000
00000000oooo100000020?ooo`X000004P3oool2000005X0oooo00<000000?ooo`3oool02@3oool0
00d0oooo00<000000?ooo`3oool00P3oool2000000L0oooo0`00001H0?ooo`D00000403oool:0000
0080oooo10000000103oool000000000003oool4000000L0oooo0P0000020?ooo`<0000000<0oooo
0000000000000`3oool300000080oooo2@000000103oool000000000000000020?ooo`80000000@0
oooo000000000000oooo0`000000103oool000000000003oool2000000D0oooo0P0000080?ooo`<0
00001@3oool2000000030?ooo`000000000000@0000000@0oooo000000000000oooo0P000000103o
ool000000000000000020?ooo`H0000000<0oooo0000000000000P3oool2000000<0oooo0P000002
0?ooo`<0000000<0oooo000000000000203oool4000000040?ooo`00000000000?ooo`@000000P3o
ool:00000100oooo1000001H0?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool0FP3oool3
00000140oooo300000000`3oool0000000000002000000<0oooo0P0000070?ooo`80000000@0oooo
000000000000oooo0`0000020?ooo`<000000P3oool2000000030?ooo`000000000000<000000P3o
ool200000080oooo0P0000050?ooo`T00000203oool2000000P0oooo0`0000070?ooo`X00000103o
ool2000000<0oooo0P0000020?ooo`D0000000<0oooo0000000000000P3oool300000080oooo0`00
0000103oool000000000003oool2000000H0oooo0P0000040?oooa0000004@3oool4000005/0oooo
00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0F`3oool400000180oooo2@00
00000`3oool0000000000002000000030?ooo`000000000000@0oooo0P0000030?ooo`80000000<0
oooo0000000000000P3oool2000000030?ooo`000000000000<0000000<0oooo0000000000000P3o
ool2000000030?ooo`00000000000080oooo0P0000080?ooo`X000001`3oool2000000P0oooo0`00
00070?ooo`80000000D0oooo00000000000000000?ooo`02000000T0oooo0P0000020?ooo`800000
00<0oooo0000000000000P3oool2000000030?ooo`000000000000<0000000<0oooo000000000000
0P3oool2000000030?ooo`00000000000080oooo0P0000050?ooo`80000000<0oooo000000000000
3000000B0?ooo`@00000G03oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?oo
o`1L0?ooo`@000004@3oool=00000080oooo0P0000040?ooo`80000000@0oooo000000000000oooo
0P0000000`3oool00000000000040?ooo`80000000<0oooo0000000000000P3oool2000000030?oo
o`000000000000<000000P3oool2000000P0oooo1`0000000`3oool00000000000060?ooo`@00000
1`3oool4000000H0oooo0P0000000`3oool0000000000004000000T0oooo0P0000020?ooo`D00000
00<0oooo0000000000000P3oool2000000030?ooo`000000000000@0oooo0P000000103oool00000
0000003oool4000000D0oooo0P000000103oool000000000003oool;00000140oooo1000001M0?oo
o`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo05d0oooo1000000A0?ooo``0
000000<0oooo0000000000000P3oool400000080oooo0P000000103oool000000000003oool20000
0080oooo0P000000103oool000000000003oool2000000040?ooo`00000000000?ooo`<0000000<0
oooo000000000000203oool200000080oooo0`0000000`3oool00000000000050?ooo`8000000P3o
ool4000000L0oooo100000020?ooo`800000103oool200000080oooo0`0000020?ooo`800000203o
ool2000000050?ooo`00000000000000003oool00P000000103oool000000000003oool200000003
0?ooo`00000000000080oooo0P000000103oool000000000003oool2000000040?ooo`0000000000
0?ooo`8000000P3oool4000000030?ooo`000000000000L000004P3oool4000005h0oooo00<00000
0?ooo`3oool02@3oool001/0oooo0`00001N0?ooo`@000004P3oool3000000030?ooo`0000000000
00@0000000@0oooo000000000000oooo1@0000030?ooo`80000000@0oooo000000000000oooo2000
00000`3oool00000000000020?ooo`80000000@0oooo000000000000oooo0P0000080?ooo`800000
0P3oool6000000D0oooo0P0000020?ooo`<000002@3oool300000080oooo0P0000040?ooo`L00000
0P3oool2000000P0oooo0P000000103oool000000000003oool200000080oooo0P0000000`3oool0
00000000000500000080oooo100000030?ooo`D000000P3oool2000000030?ooo`000000000000@0
000000@0oooo00000000000000004P3oool4000005d0oooo0`00000;0?ooo`006`3oool00`000000
oooo0?ooo`1O0?ooo`@000004@3oool<00000080oooo100000000`3oool000000000000600000005
0?ooo`00000000000000003oool00P000000103oool000000000003oool2000000040?ooo`000000
00000?ooo`8000001`3oool200000080oooo0P0000020?ooo`D000001P3oool200000080oooo0`00
00090?ooo`<000000P3oool2000000H0oooo1@0000020?ooo`8000000P3oool2000000L0oooo0P00
0000103oool000000000003oool2000000040?ooo`00000000000?ooo`80000000<0oooo00000000
00000P3oool<00000080oooo0P0000000`3oool000000000000800000140oooo1000001P0?ooo`03
0000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0600oooo1000000B0?ooo`X00000
1P3oool5000000030?ooo`0000000000008000000P3oool:000000d0oooo0P0000030?ooo`800000
00@0oooo0000000000000000103oool200000080oooo0P0000020?ooo`<000002@3oool300000080
oooo0P0000020?ooo`800000103oool3000000030?ooo`000000000000<0oooo0P00000=0?ooo`X0
000000@0oooo000000000000oooo0P0000000`3oool0000000000003000000D0oooo2P00000B0?oo
o`@00000HP3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`1R0?ooo`<0
00004`3oool8000000H0oooo1P000000103oool000000000000000020?ooo`80000000<0oooo0000
000000000`0000000`3oool000000000000=0?ooo`8000000P3oool200000080oooo0P0000040?oo
o`8000000P3oool200000080oooo100000070?ooo`@000000P3oool200000080oooo0P0000040?oo
o`8000000P3oool200000080oooo0P00000=0?ooo`80000000<0oooo0000000000000`000000103o
ool000000000003oool4000000030?ooo`000000000000@000001P3oool800000180oooo1000001S
0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo06<0oooo1000000A0?oo
o`L00000203oool8000000030?ooo`000000000000@0oooo100000000`3oool00000000000090?oo
o`H000000P3oool2000000P0oooo300000070?ooo`D0000000<0oooo000000000000100000070?oo
o`<000000P3oool3000000030?ooo`000000000000T0oooo100000000`3oool00000000000040?oo
o`80000000<0oooo0000000000001P0000080?ooo`L000004@3oool4000006@0oooo00<000000?oo
o`3oool02@3oool001/0oooo0`00001T0?ooo`@000004@3oool8000000L0oooo1@0000000`3oool0
000000000003000000<0oooo1@0000080?ooo`T000000P3oool3000000L0oooo300000070?ooo``0
00001`3oool300000080oooo0`0000000`3oool0000000000003000000P0oooo100000040?ooo`D0
000000<0oooo0000000000000`0000070?ooo`P000004@3oool4000006<0oooo0`00000;0?ooo`00
6`3oool00`000000oooo0?ooo`1U0?ooo`@00000403oool8000000L0oooo2`0000030?ooo`800000
00<0oooo000000000000203oool3000000030?ooo`000000000000@0000000@0oooo000000000000
00001@3oool4000000040?ooo`0000000000000000@0oooo0P0000070?ooo`8000000`3oool40000
00030?ooo`0000000000008000001@3oool200000080oooo1P000000103oool00000000000000008
0?ooo`@00000103oool;000000L0oooo2000000@0?ooo`@00000IP3oool00`000000oooo0?ooo`09
0?ooo`006`3oool00`000000oooo0?ooo`1V0?ooo`@000004P3oool4000000/0oooo1P00000C0?oo
o`T000002@3oool2000000<0oooo0P0000020?ooo`8000003@3oool200000080oooo0P0000030?oo
o`8000002@3oool9000001<0oooo1P00000;0?ooo`@000004P3oool4000006L0oooo00<000000?oo
o`3oool02@3oool001/0oooo00<000000?ooo`3oool0I`3oool400000140oooo1000000;0?ooo`@0
00004P3oool2000000030?ooo`000000000000<0000000@0oooo00000000000000002@3oool40000
00030?ooo`00000000000080oooo100000090?ooo`@000000P3oool2000000030?ooo`0000000000
008000002@3oool5000000030?ooo`000000000000<000004P3oool5000000/0oooo1000000A0?oo
o`@00000J03oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`1X0?ooo`@0
00004@3oool3000000`0oooo1000000>0?ooo`80000000<0oooo0000000000002P0000050?ooo`80
00000P3oool4000000030?ooo`00000000000080oooo100000090?ooo`@000000P3oool200000003
0?ooo`0000000000008000000P3oool2000000D0oooo200000000`3oool00000000000020?ooo`80
00003@3oool5000000/0oooo1000000@0?ooo`@00000JP3oool00`000000oooo0?ooo`090?ooo`00
6`3oool3000006X0oooo1000000@0?ooo`8000003@3oool3000000d0oooo10000000103oool00000
0000003oool5000000030?ooo`000000000000D0oooo0P0000020?ooo`@0000000<0oooo00000000
00000P3oool4000000T0oooo100000020?ooo`H000000P3oool3000000D0oooo20000000103oool0
00000000003oool4000000`0oooo1000000<0?ooo`8000004@3oool4000006T0oooo0`00000;0?oo
o`006`3oool00`000000oooo0?ooo`1[0?ooo`@00000;@3oool5000000050?ooo`00000000000000
003oool0100000000`3oool00000000000050?ooo`8000000P3oool4000000L0oooo0P0000090?oo
o`8000001P3oool400000080oooo0P0000060?ooo`H000000P3oool3000000030?ooo`0000000000
00800000;P3oool4000006`0oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3o
ool0K03oool4000002X0oooo20000000103oool000000000003oool5000000P0oooo0`0000000`3o
ool000000000000200000080oooo0P0000020?ooo`8000002@3oool200000080oooo0P0000020?oo
o`@0000000@0oooo0000000000000000203oool400000080oooo0P0000000`3oool0000000000003
000000030?ooo`000000000002X0oooo1000001]0?ooo`030000003oool0oooo00T0oooo000K0?oo
o`030000003oool0oooo06d0oooo1000000W0?ooo`X000000P3oool2000000030?ooo`0000000000
00800000203oool3000000030?ooo`0000000000008000000P3oool200000080oooo0P0000090?oo
o`8000000P3oool200000080oooo10000000103oool000000000000000080?ooo`@0000000@0oooo
000000000000oooo2`00000W0?ooo`@00000KP3oool00`000000oooo0?ooo`090?ooo`002`3oool4
00000080oooo0`0000070?ooo`030000003oool0oooo06h0oooo1000000Q0?ooo`@0000000@0oooo
000000000000oooo0P0000000`3oool000000000000200000080oooo0P00000:0?ooo`80000000D0
oooo00000000000000000?ooo`0400000080oooo0P0000020?ooo`800000203oool2000000<0oooo
0P0000020?ooo`@0000000D0oooo00000000000000000?ooo`02000000X0oooo0P0000000`3oool0
000000000003000000040?ooo`00000000000?ooo`80000000<0oooo0000000000000P00000Q0?oo
o`<00000L03oool00`000000oooo0?ooo`090?ooo`003@3oool00`000000oooo0?ooo`040?ooo`03
0000003oool0oooo00@0oooo1000001^0?ooo`@00000803oool400000080oooo0P0000000`3oool0
000000000005000000030?ooo`000000000000X0oooo0P0000020?ooo`80000000<0oooo00000000
00000P0000020?ooo`8000000P3oool2000000P0oooo0P0000030?ooo`80000000@0oooo00000000
0000oooo0P0000000`3oool00000000000020?ooo`8000002P3oool2000000030?ooo`0000000000
00D0000000<0oooo0000000000000P3oool400000200oooo1000001]0?ooo`@000002`3oool000d0
oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo0740
oooo0`00000=0?ooo`8000003@3oool8000000040?ooo`00000000000?ooo`80000000<0oooo0000
000000000`0000000`3oool00000000000090?ooo`8000000P3oool2000000<0oooo0P0000020?oo
o`80000000@0oooo0000000000000000203oool300000080oooo0P0000000`3oool0000000000003
0?ooo`8000000`3oool2000000T0oooo0P0000000`3oool0000000000003000000040?ooo`000000
00000?ooo`80000000<0oooo0000000000001P00000=0?ooo`800000303oool400000780oooo00<0
00000?ooo`3oool02@3oool000d0oooo00@000000?ooo`3oool0oooo0`0000070?ooo`030000003o
ool0oooo0740oooo1@00000;0?ooo`L00000203oool9000000040?ooo`00000000000?ooo`X00000
1P3oool2000000030?ooo`00000000000080oooo0`0000090?ooo`<000002@3oool2000000X0oooo
0P0000020?ooo`<0000000<0oooo0000000000001P3oool:000000040?ooo`00000000000?ooo`T0
00001`3oool8000000/0oooo1000001c0?ooo`030000003oool0oooo00T0oooo000=0?ooo`050000
003oool0oooo0?ooo`0000002@3oool00`000000oooo0?ooo`1c0?ooo`@000002`3oool=00000003
0?ooo`000000000001<00000203oool2000000030?ooo`000000000000<0oooo0`0000050?ooo`80
00000P3oool2000000T0oooo0P0000000`3oool00000000000060?ooo`<000000P3oool200000080
oooo0P0000080?oooaD0000000<0oooo0000000000002P00000<0?ooo`@00000M03oool00`000000
oooo0?ooo`090?ooo`002`3oool3000000<0oooo00<000000?ooo`3oool01`3oool00`000000oooo
0?ooo`1d0?ooo`<000002`3oool=000000050?ooo`00000000000000003oool01P000000103oool0
00000000003oool7000000P0oooo0P0000020?ooo`8000000P3oool3000000@0oooo0`0000020?oo
o`8000002@3oool2000000030?ooo`000000000000800000103oool300000080oooo0P0000020?oo
o`800000203oool7000000040?ooo`00000000000?ooo`H0000000D0oooo00000000000000000?oo
o`0<000000`0oooo0`00001e0?ooo`030000003oool0oooo00T0oooo000=0?ooo`040000003oool0
oooo0?ooo`@000001P3oool3000007D0oooo0`00000<0?ooo`X0000000<0oooo0000000000000P00
00000`3oool0000000000005000000040?ooo`00000000000?ooo`80000000@0oooo000000000000
0000203oool3000000030?ooo`00000000000080oooo100000030?ooo`@0000000<0oooo00000000
00002@3oool2000000030?ooo`0000000000008000000`3oool3000000<0oooo0P000000103oool0
00000000000000080?ooo`8000000P3oool2000000040?ooo`00000000000?ooo`T0000000<0oooo
0000000000002P00000<0?ooo`@00000M03oool3000000/0oooo000K0?ooo`030000003oool0oooo
07H0oooo1000000<0?ooo`P0000000<0oooo0000000000000P0000020?ooo`l000002P3oool20000
00050?ooo`00000000000000003oool01@0000020?ooo`@0000000<0oooo0000000000002@3oool2
000000030?ooo`0000000000008000000P3oool400000080oooo0`0000000`3oool0000000000009
0?ooo`80000000<0oooo0000000000002`000000103oool000000000003oool;000000`0oooo0`00
001h0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo07P0oooo0`00000<
0?ooo`/0000000@0oooo000000000000oooo1P000000103oool000000000003oool2000000`0oooo
0P000000103oool000000000000000030?ooo`<000000P3oool2000000<0oooo0P0000090?ooo`80
00000`3oool200000080oooo0`0000030?ooo`H00000303oool2000000040?ooo`00000000000?oo
o`H0000000@0oooo000000000000oooo2`00000<0?ooo`<00000N@3oool00`000000oooo0?ooo`09
0?ooo`006`3oool00`000000oooo0?ooo`1h0?ooo`@000003@3oool9000000040?ooo`0000000000
0?ooo`@0000000<0oooo0000000000000P0000000`3oool000000000000;0?ooo`<0000000@0oooo
00000000000000000`3oool300000080oooo0P0000030?ooo`8000002@3oool2000000<0oooo0P00
00020?ooo`<000000`3oool7000000X0oooo0P000000103oool000000000003oool600000080oooo
0P0000000`3oool0000000000007000000d0oooo0`00001j0?ooo`030000003oool0oooo00T0oooo
000K0?ooo`030000003oool0oooo07T0oooo1@00000<0?ooo`L0000000<0oooo0000000000000P00
00020?ooo`X000002`3oool300000080oooo0P0000030?ooo`8000000`3oool2000000<0oooo0P00
00030?ooo`800000103oool200000080oooo0P0000030?ooo`8000000`3oool300000080oooo0`00
000:0?ooo`80000000<0oooo0000000000001P000000103oool000000000003oool2000000030?oo
o`000000000000D00000303oool4000007/0oooo00<000000?ooo`3oool02@3oool001/0oooo0`00
001k0?ooo`@000002`3oool<00000080oooo2@0000090?ooo`8000000P3oool200000080oooo0P00
00030?ooo`<000000P3oool3000000030?ooo`000000000000@0oooo0P0000050?ooo`80000000<0
oooo0000000000000`3oool2000000<0oooo0`0000020?ooo`8000000P3oool2000000T0oooo2@00
00000`3oool000000000000;000000/0oooo1000001j0?ooo`<000002`3oool001/0oooo00<00000
0?ooo`3oool0O03oool4000000/0oooo2P000000103oool000000000003oool8000000H0oooo0P00
00000`3oool00000000000020?ooo`8000000P3oool3000000<0oooo0P0000030?ooo`80000000<0
oooo0000000000002`3oool2000000030?ooo`000000000000<0oooo0P0000030?ooo`<000000P3o
ool200000080oooo100000070?ooo`L000000P3oool2000000030?ooo`000000000000P000002`3o
ool4000007d0oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0O@3oool4
000000/0oooo300000000`3oool0000000000005000000D0oooo1@000000103oool000000000003o
ool300000080oooo0P0000080?ooo`80000000<0oooo000000000000103oool2000000D0oooo0P00
00000`3oool00000000000080?ooo`8000000P3oool3000000040?ooo`00000000000?ooo`800000
00<0oooo0000000000001@3oool600000080oooo3000000;0?ooo`@00000OP3oool00`000000oooo
0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`1n0?ooo`@000002`3oool<00000080oooo0P00
00000`3oool00000000000050?ooo`80000000@0oooo000000000000oooo0P0000020?ooo`800000
0P3oool2000000H0oooo0P0000030?ooo`800000103oool2000000D0oooo0P0000030?ooo`800000
1P3oool200000080oooo0P0000020?ooo`D0000000<0oooo0000000000001@3oool2000000030?oo
o`00000000000080oooo3000000;0?ooo`@00000O`3oool00`000000oooo0?ooo`090?ooo`006`3o
ool00`000000oooo0?ooo`1o0?ooo`@00000303oool;000000030?ooo`000000000000/0oooo1P00
00000`3oool00000000000020?ooo`<000001@3oool2000000<0oooo0P00000;0?ooo`8000000`3o
ool2000000D0oooo0`0000020?ooo`80000000<0oooo0000000000000`00000<0?ooo`d000003@3o
ool400000800oooo00<000000?ooo`3oool02@3oool001/0oooo0`0000210?ooo`<00000303oool:
00000080oooo0P00000;0?ooo`80000000<0oooo0000000000000P3oool200000080oooo0P000005
0?ooo`8000000`3oool2000000/0oooo0P0000030?ooo`8000001@3oool200000080oooo0P000002
0?ooo`D000002`3oool2000000030?ooo`000000000000T000002`3oool400000800oooo0`00000;
0?ooo`006`3oool00`000000oooo0?ooo`210?ooo`D000002`3oool=000000/0oooo0P0000001@3o
ool00000000000000000oooo008000000P3oool2000000030?ooo`00000000000080oooo0P000003
0?ooo`8000002`3oool2000000<0oooo0P0000020?ooo`80000000<0oooo0000000000000P3oool2
000000030?ooo`000000000000<00000303oool=000000/0oooo100000230?ooo`030000003oool0
oooo00T0oooo000K0?ooo`030000003oool0oooo08<0oooo1000000<0?ooo`X00000403oool20000
00030?ooo`000000000000D0oooo0P0000020?ooo`8000000`3oool2000000X0oooo0P0000030?oo
o`<000000P3oool2000000@0oooo0`0000000`3oool000000000000@0?ooo`X00000303oool40000
08@0oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0Q03oool4000000`0
oooo2@00000@0?ooo`H0000000@0oooo000000000000oooo0`000000103oool00000000000000002
0?ooo`<00000203oool3000000<0oooo0P0000020?ooo`<0000000@0oooo000000000000oooo0P00
0000103oool0000000000000000@0?ooo`T00000303oool4000008D0oooo00<000000?ooo`3oool0
2@3oool001/0oooo00<000000?ooo`3oool0Q@3oool4000000`0oooo2@00000@0?ooo`80000000@0
oooo000000000000oooo0P0000020?ooo`8000000P3oool2000000<0oooo0`0000070?ooo`<00000
0`3oool200000080oooo0P0000020?ooo`80000000@0oooo000000000000oooo0P00000@0?ooo`T0
0000303oool4000008H0oooo00<000000?ooo`3oool02@3oool001/0oooo0`0000260?ooo`@00000
303oool800000100oooo1@0000020?ooo`80000000<0oooo0000000000000P3oool2000000<0oooo
0`0000070?ooo`<000000`3oool200000080oooo0P0000000`3oool00000000000020?ooo`D00000
403oool8000000`0oooo100000250?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool0R03o
ool3000000/0oooo2000000=0?ooo`80000000<0oooo0000000000000P0000030?ooo`80000000<0
oooo0000000000000P3oool2000000<0oooo0`0000070?ooo`<000000`3oool200000080oooo0P00
00000`3oool00000000000030?ooo`<000000P3oool2000000d0oooo1`00000;0?ooo`@00000R@3o
ool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`290?ooo`<00000303oool7
000000T0oooo0P0000000`3oool00000000000030?ooo`8000000`3oool2000000D0oooo0P00000B
0?ooo`<000001@3oool2000000<0oooo0P0000030?ooo`@000002P3oool6000000`0oooo1000002:
0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo08X0oooo1000000;0?oo
o`H000001`3oool5000000030?ooo`00000000000080oooo0P0000030?ooo`<00000103oool30000
0140oooo0P0000050?ooo`<000000`3oool200000080oooo0P000000103oool000000000003oool2
000000L0oooo1P00000;0?ooo`@00000R`3oool00`000000oooo0?ooo`090?ooo`002`3oool40000
00<0oooo0P0000070?ooo`030000003oool0oooo08/0oooo1000000;0?ooo`@000001P3oool:0000
0080oooo0`0000030?ooo`8000001@3oool200000140oooo0P0000050?ooo`8000000`3oool30000
00030?ooo`000000000000P000001P3oool5000000/0oooo1000002<0?ooo`030000003oool0oooo
00T0oooo000;0?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool000001P3oool40000
08/0oooo1000000;0?ooo`@000001@3oool5000000030?ooo`000000000000<000000P3oool20000
00<0oooo0P0000050?ooo`8000004@3oool2000000D0oooo0P0000030?ooo`8000000P3oool20000
00030?ooo`000000000000D000001P3oool4000000/0oooo1000002:0?ooo`@000002`3oool000`0
oooo00<000000?ooo`3oool00P3oool010000000oooo0?ooo`0000060?ooo`030000003oool0oooo
08d0oooo1000000<0?ooo`8000001P3oool;000000030?ooo`000000000000<0oooo0`00000M0?oo
o`<000000`3oool2000000030?ooo`000000000000H0000000<0oooo0000000000001P3oool20000
00`0oooo1000002>0?ooo`030000003oool0oooo00T0oooo000=0?ooo`050000003oool0oooo0?oo
o`0000000P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo08h0oooo1000000F0?oo
o`D0000000<0oooo0000000000000P3oool2000000<0oooo0P0000070?ooo`8000002`3oool20000
00L0oooo0P0000030?ooo`8000000P3oool2000000030?ooo`000000000000<000005P3oool30000
0900oooo00<000000?ooo`3oool02@3oool000h0oooo00@000000?ooo`3oool000000P3oool00`00
0000oooo0?ooo`040?ooo`030000003oool0oooo08l0oooo1000000E0?ooo`T0000000<0oooo0000
000000000`3oool2000000L0oooo0P00000;0?ooo`8000001`3oool2000000<0oooo0P0000000`3o
ool0000000000006000001D0oooo1@00002@0?ooo`030000003oool0oooo00T0oooo000;0?ooo`04
0000003oool0oooo00000080oooo00@000000?ooo`3oool000001P3oool00`000000oooo0?ooo`2@
0?ooo`@000005@3oool2000000040?ooo`00000000000?ooo`80000000@0oooo0000000000000000
0P3oool200000080oooo0P0000030?ooo`8000002`3oool2000000<0oooo0P0000000`3oool00000
000000030?ooo`<0000000@0oooo000000000000oooo1000000F0?ooo`<00000TP3oool00`000000
oooo0?ooo`090?ooo`00303oool2000000@0oooo0P0000070?ooo`<00000TP3oool3000001L0oooo
1P0000000`3oool00000000000020?ooo`<0000000<0oooo0000000000000`3oool2000000/0oooo
0P0000030?ooo`80000000<0oooo0000000000000`3oool2000000030?ooo`000000000000@00000
5`3oool300000940oooo0`00000;0?ooo`006`3oool00`000000oooo0?ooo`2C0?ooo`@000005P3o
ool5000000050?ooo`00000000000000003oool00`0000000`3oool00000000000030?ooo`800000
2`3oool2000000<0oooo0P0000001@3oool00000000000000000oooo00<0000000<0oooo00000000
00000`00000F0?ooo`@00000U03oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo
0?ooo`2D0?ooo`@000005@3oool6000000040?ooo`00000000000?ooo`<0000000@0oooo00000000
000000000P3oool2000000/0oooo0P0000020?ooo`<0000000D0oooo00000000000000000?ooo`02
000000030?ooo`000000000000@000005@3oool4000009D0oooo00<000000?ooo`3oool02@3oool0
01/0oooo00<000000?ooo`3oool0U@3oool4000001D0oooo1@0000000`3oool00000000000020?oo
o`80000000@0oooo00000000000000000P3oool2000000X0oooo0`0000020?ooo`8000000P3oool2
000000050?ooo`00000000000000003oool01@00000E0?ooo`@00000UP3oool00`000000oooo0?oo
o`090?ooo`006`3oool00`000000oooo0?ooo`2F0?ooo`@00000503oool2000000030?ooo`000000
000000@0000000<0oooo0000000000000P3oool200000080oooo0`0000090?ooo`8000000`3oool2
00000080oooo0P000000103oool000000000003oool5000001D0oooo1000002G0?ooo`030000003o
ool0oooo00T0oooo000K0?ooo`030000003oool0oooo09L0oooo1000000G0?ooo`80000000<0oooo
0000000000001@3oool2000000<0oooo0P0000090?ooo`8000000P3oool3000000@0oooo1P00000F
0?ooo`D00000V03oool00`000000oooo0?ooo`090?ooo`006`3oool3000009P0oooo1000000F0?oo
o`H00000103oool300000080oooo0P0000090?ooo`8000000P3oool3000000@0oooo0P000000103o
ool0000000000000000F0?ooo`<00000V03oool3000000/0oooo000K0?ooo`030000003oool0oooo
09X0oooo1000000E0?ooo`80000000<0oooo0000000000001@3oool200000080oooo0P0000090?oo
o`8000000P3oool2000000D0oooo0P0000000`3oool000000000000E0?ooo`@00000V`3oool00`00
0000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`2K0?ooo`@00000503oool200000003
0?ooo`000000000000D0oooo0P0000020?ooo`8000002@3oool200000080oooo0P0000050?ooo`80
000000<0oooo000000000000503oool4000009`0oooo00<000000?ooo`3oool02@3oool001/0oooo
00<000000?ooo`3oool0W03oool400000240oooo0P0000080?ooo`<000008@3oool4000009d0oooo
00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0W@3oool400000200oooo0P00
00080?ooo`<00000803oool4000009h0oooo00<000000?ooo`3oool02@3oool001/0oooo0`00002N
0?ooo`@000007`3oool3000000L0oooo0P00000P0?ooo`@00000W@3oool3000000/0oooo000K0?oo
o`030000003oool0oooo09l0oooo1000000O0?ooo`8000001`3oool2000001h0oooo1@00002P0?oo
o`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0:00oooo1@0000150?ooo`@0
0000XP3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`2Q0?ooo`D00000
@`3oool400000:<0oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0X`3o
ool300000480oooo1000002T0?ooo`030000003oool0oooo00T0oooo000K0?ooo`<00000Y03oool4
000003l0oooo1000002S0?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool0Y@3oool40000
03d0oooo1000002V0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0:H0
oooo1000000k0?ooo`@00000Y`3oool00`000000oooo0?ooo`090?ooo`002`3oool400000080oooo
0`0000070?ooo`030000003oool0oooo0:L0oooo1000000i0?ooo`<00000Z@3oool00`000000oooo
0?ooo`090?ooo`002`3oool00`000000oooo0?ooo`060?ooo`030000003oool0oooo00@0oooo00<0
00000?ooo`3oool0Z03oool4000003H0oooo1000002Z0?ooo`030000003oool0oooo00T0oooo000<
0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3oool0103oool400000:P0oooo1000000e
0?ooo`<00000Z03oool4000000/0oooo000=0?ooo`040000003oool0oooo0?ooo`<000001`3oool0
0`000000oooo0?ooo`2[0?ooo`@00000203oool2000001d0oooo0P0000080?ooo`@00000[03oool0
0`000000oooo0?ooo`090?ooo`003P3oool010000000oooo0?ooo`0000090?ooo`030000003oool0
oooo0:`0oooo100000070?ooo`<000006`3oool3000000L0oooo1000002]0?ooo`030000003oool0
oooo00T0oooo000;0?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3oool01`3oool0
0`000000oooo0?ooo`2]0?ooo`@000001P3oool3000000030?ooo`000000000001D0oooo0P000000
103oool000000000000000060?ooo`@00000[P3oool00`000000oooo0?ooo`090?ooo`00303oool2
000000<0oooo100000060?ooo`030000003oool0oooo0:h0oooo100000060?ooo`80000000<0oooo
0000000000005@3oool2000000030?ooo`000000000000H0oooo1000002_0?ooo`030000003oool0
oooo00T0oooo000K0?ooo`<00000[`3oool4000000P0oooo0`0000000`3oool000000000000<0?oo
o`<0000000@0oooo0000000000000000203oool400000:h0oooo0`00000;0?ooo`006`3oool00`00
0000oooo0?ooo`2`0?ooo`@000001`3oool3000000040?ooo`0000000000000000/0oooo1P000007
0?ooo`@00000/P3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`2b0?oo
o`@000001P3oool3000000030?ooo`000000000000/0oooo0P000000103oool00000000000000006
0?ooo`@00000/`3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`2c0?oo
o`@000001P3oool2000000040?ooo`0000000000000000T0oooo0`0000000`3oool0000000000006
0?ooo`@00000]03oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`2d0?oo
o`@00000203oool3000000T0oooo0`0000080?ooo`@00000]@3oool00`000000oooo0?ooo`090?oo
o`006`3oool300000;D0oooo100000080?ooo`8000002@3oool2000000P0oooo1000002d0?ooo`<0
00002`3oool001/0oooo00<000000?ooo`3oool0]P3oool4000000L0oooo0`0000070?ooo`<00000
1`3oool400000;L0oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0]`3o
ool4000000H0oooo0`0000070?ooo`8000001`3oool400000;P0oooo00<000000?ooo`3oool02@3o
ool001/0oooo00<000000?ooo`3oool0^03oool5000001D0oooo1000002j0?ooo`030000003oool0
oooo00T0oooo000K0?ooo`030000003oool0oooo0;T0oooo1@00000C0?ooo`@00000^`3oool00`00
0000oooo0?ooo`090?ooo`006`3oool300000;/0oooo1000000A0?ooo`@00000^P3oool3000000/0
oooo000K0?ooo`030000003oool0oooo0;`0oooo1000000?0?ooo`@00000_@3oool00`000000oooo
0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`2m0?ooo`@000003@3oool400000;h0oooo00<0
00000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0_P3oool4000000/0oooo1000002o
0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0;l0oooo100000080?oo
o`D00000`03oool00`000000oooo0?ooo`090?ooo`006`3oool300000<00oooo100000070?ooo`@0
0000_`3oool3000000/0oooo000K0?ooo`030000003oool0oooo0<40oooo0`0000070?ooo`800000
``3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`3o0?oooi40oooo00<0
00000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0o`3ooonA0?ooo`030000003oool0
oooo00T0oooo000<0?ooo`800000103oool2000000L0oooo00<000000?ooo`3oool0aP3oool20000
0<P0oooo00<000000?ooo`3oool02@3oool000/0oooo00@000000?ooo`3oool000000P3oool01000
0000oooo0?ooo`0000060?ooo`@00000a@3oool200000<D0oooo1000000;0?ooo`003P3oool01000
0000oooo0?ooo`0000020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0o`3ooonA
0?ooo`030000003oool0oooo00T0oooo000<0?ooo`8000000`3oool010000000oooo0?ooo`000006
0?ooo`030000003oool0oooo0?l0ooooT@3oool00`000000oooo0?ooo`090?ooo`003P3oool01000
0000oooo0?ooo`0000020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0o`3ooonA
0?ooo`030000003oool0oooo00T0oooo000;0?ooo`040000003oool0oooo00000080oooo00@00000
0?ooo`3oool000001P3oool00`000000oooo0?ooo`3o0?oooi40oooo00<000000?ooo`3oool02@3o
ool000`0oooo0P0000040?ooo`8000001`3oool300000?l0ooooS`3oool3000000/0oooo000K0?oo
o`030000003oool0oooo0?l0ooooT@3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000
oooo0?ooo`3o0?oooi40oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0
o`3ooonA0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0?l0ooooT@3o
ool00`000000oooo0?ooo`090?ooo`006`3oool300000?l0ooooS`3oool3000000/0oooo000K0?oo
o`030000003oool0oooo0?l0ooooT@3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000
oooo0?ooo`3o0?oooi40oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0
o`3ooonA0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0?l0ooooT@3o
ool00`000000oooo0?ooo`090?ooo`006`3oool300000?l0ooooS`3oool3000000/0oooo000K0?oo
o`030000003oool0oooo0?l0ooooT@3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000
oooo0?ooo`3o0?oooi40oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0
o`3ooonA0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0?l0ooooT@3o
ool00`000000oooo0?ooo`090?ooo`006`3oool300000?l0ooooS`3oool3000000/0oooo000K0?oo
o`030000003oool0oooo0?l0ooooT@3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000
oooo0?ooo`0J0?ooo`030000003oool0oooo03H0oooo00<000000?ooo`3oool0=`3oool00`000000
oooo0?ooo`0f0?ooo`030000003oool0oooo03L0oooo00<000000?ooo`3oool0=`3oool00`000000
oooo0?ooo`0f0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool02@3oool001/0oooo
00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo
00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo
00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo
00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00<0oooo
00<000000?ooo`3oool02@3oool000`0oooo0P0000030?ooo`<000001`3oool00`000000oooo0?oo
o`030?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?oo
o`080?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?oo
o`090?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool00`3oool00`000000oooo0?oo
o`090?ooo`002`3oool010000000oooo0?ooo`0000050?ooo`030000003oool0oooo00@0ooooo`00
002E000000/0oooo000>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0o`3ooonT
0?ooo`00303oool2000000<0oooo0`00003o0?ooojL0oooo000>0?ooo`040000003oool0oooo0000
0?l0ooooZ@3oool000/0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`3o0?oo
ojL0oooo0000\
\>"],
  ImageRangeCache->{{{0, 441}, {220, 0}} -> {-9.77662, -7.78715, 0.173731, 
  0.198549}}]
}, Open  ]],

Cell[CellGroupData[{

Cell[BoxData[{
    \(X = AnglePairsAndRationalApproxs[75, 10^\(-3\)]; \nLength[X]\), 
    \(Angles = \(Transpose[X]\)[\([1]\)]; \n
    ListPlot[Angles, \ AspectRatio -> 1/2, Frame -> True, \ Axes -> False, \ 
      PlotRange -> {{\(-5\), 65}, {\(-5\), 35}}]; 
    WriteData["\<~/ASP/MDWebServices/GrainBreaker/2DGeometries/Geo70.dat\>", 
      X]; WriteData[
      "\<~/ASP/MDWebServices/GrainBreaker/2DGeometries/Geo70forplot.dat\>", 
      Angles]; \)}], "Input"],

Cell[BoxData[
    \(11647\)], "Output"],

Cell[GraphicsData["PostScript", "\<\
%!
%%Creator: Mathematica
%%AspectRatio: .5 
MathPictureStart
/Mabs {
Mgmatrix idtransform
Mtmatrix dtransform
} bind def
/Mabsadd { Mabs
3 -1 roll add
3 1 roll add
exch } bind def
%% Graphics
%%IncludeResource: font Courier
%%IncludeFont: Courier
/Courier findfont 10  scalefont  setfont
% Scaling calculations
0.0714286 0.0142857 0.0625 0.0125 [
[.07143 -0.0125 -3 -9 ]
[.07143 -0.0125 3 0 ]
[.21429 -0.0125 -6 -9 ]
[.21429 -0.0125 6 0 ]
[.35714 -0.0125 -6 -9 ]
[.35714 -0.0125 6 0 ]
[.5 -0.0125 -6 -9 ]
[.5 -0.0125 6 0 ]
[.64286 -0.0125 -6 -9 ]
[.64286 -0.0125 6 0 ]
[.78571 -0.0125 -6 -9 ]
[.78571 -0.0125 6 0 ]
[.92857 -0.0125 -6 -9 ]
[.92857 -0.0125 6 0 ]
[ 0 0 -0.125 0 ]
[-0.0125 .0625 -6 -4.5 ]
[-0.0125 .0625 0 4.5 ]
[-0.0125 .125 -6 -4.5 ]
[-0.0125 .125 0 4.5 ]
[-0.0125 .1875 -12 -4.5 ]
[-0.0125 .1875 0 4.5 ]
[-0.0125 .25 -12 -4.5 ]
[-0.0125 .25 0 4.5 ]
[-0.0125 .3125 -12 -4.5 ]
[-0.0125 .3125 0 4.5 ]
[-0.0125 .375 -12 -4.5 ]
[-0.0125 .375 0 4.5 ]
[-0.0125 .4375 -12 -4.5 ]
[-0.0125 .4375 0 4.5 ]
[-0.0125 .5 -12 -4.5 ]
[-0.0125 .5 0 4.5 ]
[ 0 0 -0.125 0 ]
[ 0 .5 .125 0 ]
[ 1 0 .125 0 ]
[ 0 0 0 0 ]
[ 1 .5 0 0 ]
] MathScale
% Start of Graphics
1 setlinecap
1 setlinejoin
newpath
0 g
.25 Mabswid
[ ] 0 setdash
.07143 0 m
.07143 .00625 L
s
[(0)] .07143 -0.0125 0 1 Mshowa
.21429 0 m
.21429 .00625 L
s
[(10)] .21429 -0.0125 0 1 Mshowa
.35714 0 m
.35714 .00625 L
s
[(20)] .35714 -0.0125 0 1 Mshowa
.5 0 m
.5 .00625 L
s
[(30)] .5 -0.0125 0 1 Mshowa
.64286 0 m
.64286 .00625 L
s
[(40)] .64286 -0.0125 0 1 Mshowa
.78571 0 m
.78571 .00625 L
s
[(50)] .78571 -0.0125 0 1 Mshowa
.92857 0 m
.92857 .00625 L
s
[(60)] .92857 -0.0125 0 1 Mshowa
.125 Mabswid
.1 0 m
.1 .00375 L
s
.12857 0 m
.12857 .00375 L
s
.15714 0 m
.15714 .00375 L
s
.18571 0 m
.18571 .00375 L
s
.24286 0 m
.24286 .00375 L
s
.27143 0 m
.27143 .00375 L
s
.3 0 m
.3 .00375 L
s
.32857 0 m
.32857 .00375 L
s
.38571 0 m
.38571 .00375 L
s
.41429 0 m
.41429 .00375 L
s
.44286 0 m
.44286 .00375 L
s
.47143 0 m
.47143 .00375 L
s
.52857 0 m
.52857 .00375 L
s
.55714 0 m
.55714 .00375 L
s
.58571 0 m
.58571 .00375 L
s
.61429 0 m
.61429 .00375 L
s
.67143 0 m
.67143 .00375 L
s
.7 0 m
.7 .00375 L
s
.72857 0 m
.72857 .00375 L
s
.75714 0 m
.75714 .00375 L
s
.81429 0 m
.81429 .00375 L
s
.84286 0 m
.84286 .00375 L
s
.87143 0 m
.87143 .00375 L
s
.9 0 m
.9 .00375 L
s
.04286 0 m
.04286 .00375 L
s
.01429 0 m
.01429 .00375 L
s
.95714 0 m
.95714 .00375 L
s
.98571 0 m
.98571 .00375 L
s
.25 Mabswid
0 0 m
1 0 L
s
0 .0625 m
.00625 .0625 L
s
[(0)] -0.0125 .0625 1 0 Mshowa
0 .125 m
.00625 .125 L
s
[(5)] -0.0125 .125 1 0 Mshowa
0 .1875 m
.00625 .1875 L
s
[(10)] -0.0125 .1875 1 0 Mshowa
0 .25 m
.00625 .25 L
s
[(15)] -0.0125 .25 1 0 Mshowa
0 .3125 m
.00625 .3125 L
s
[(20)] -0.0125 .3125 1 0 Mshowa
0 .375 m
.00625 .375 L
s
[(25)] -0.0125 .375 1 0 Mshowa
0 .4375 m
.00625 .4375 L
s
[(30)] -0.0125 .4375 1 0 Mshowa
0 .5 m
.00625 .5 L
s
[(35)] -0.0125 .5 1 0 Mshowa
.125 Mabswid
0 .0125 m
.00375 .0125 L
s
0 .025 m
.00375 .025 L
s
0 .0375 m
.00375 .0375 L
s
0 .05 m
.00375 .05 L
s
0 .075 m
.00375 .075 L
s
0 .0875 m
.00375 .0875 L
s
0 .1 m
.00375 .1 L
s
0 .1125 m
.00375 .1125 L
s
0 .1375 m
.00375 .1375 L
s
0 .15 m
.00375 .15 L
s
0 .1625 m
.00375 .1625 L
s
0 .175 m
.00375 .175 L
s
0 .2 m
.00375 .2 L
s
0 .2125 m
.00375 .2125 L
s
0 .225 m
.00375 .225 L
s
0 .2375 m
.00375 .2375 L
s
0 .2625 m
.00375 .2625 L
s
0 .275 m
.00375 .275 L
s
0 .2875 m
.00375 .2875 L
s
0 .3 m
.00375 .3 L
s
0 .325 m
.00375 .325 L
s
0 .3375 m
.00375 .3375 L
s
0 .35 m
.00375 .35 L
s
0 .3625 m
.00375 .3625 L
s
0 .3875 m
.00375 .3875 L
s
0 .4 m
.00375 .4 L
s
0 .4125 m
.00375 .4125 L
s
0 .425 m
.00375 .425 L
s
0 .45 m
.00375 .45 L
s
0 .4625 m
.00375 .4625 L
s
0 .475 m
.00375 .475 L
s
0 .4875 m
.00375 .4875 L
s
.25 Mabswid
0 0 m
0 .5 L
s
.07143 .49375 m
.07143 .5 L
s
.21429 .49375 m
.21429 .5 L
s
.35714 .49375 m
.35714 .5 L
s
.5 .49375 m
.5 .5 L
s
.64286 .49375 m
.64286 .5 L
s
.78571 .49375 m
.78571 .5 L
s
.92857 .49375 m
.92857 .5 L
s
.125 Mabswid
.1 .49625 m
.1 .5 L
s
.12857 .49625 m
.12857 .5 L
s
.15714 .49625 m
.15714 .5 L
s
.18571 .49625 m
.18571 .5 L
s
.24286 .49625 m
.24286 .5 L
s
.27143 .49625 m
.27143 .5 L
s
.3 .49625 m
.3 .5 L
s
.32857 .49625 m
.32857 .5 L
s
.38571 .49625 m
.38571 .5 L
s
.41429 .49625 m
.41429 .5 L
s
.44286 .49625 m
.44286 .5 L
s
.47143 .49625 m
.47143 .5 L
s
.52857 .49625 m
.52857 .5 L
s
.55714 .49625 m
.55714 .5 L
s
.58571 .49625 m
.58571 .5 L
s
.61429 .49625 m
.61429 .5 L
s
.67143 .49625 m
.67143 .5 L
s
.7 .49625 m
.7 .5 L
s
.72857 .49625 m
.72857 .5 L
s
.75714 .49625 m
.75714 .5 L
s
.81429 .49625 m
.81429 .5 L
s
.84286 .49625 m
.84286 .5 L
s
.87143 .49625 m
.87143 .5 L
s
.9 .49625 m
.9 .5 L
s
.04286 .49625 m
.04286 .5 L
s
.01429 .49625 m
.01429 .5 L
s
.95714 .49625 m
.95714 .5 L
s
.98571 .49625 m
.98571 .5 L
s
.25 Mabswid
0 .5 m
1 .5 L
s
.99375 0 m
1 0 L
s
.99375 .0625 m
1 .0625 L
s
.99375 .125 m
1 .125 L
s
.99375 .1875 m
1 .1875 L
s
.99375 .25 m
1 .25 L
s
.99375 .3125 m
1 .3125 L
s
.99375 .375 m
1 .375 L
s
.99375 .4375 m
1 .4375 L
s
.125 Mabswid
.99625 .0125 m
1 .0125 L
s
.99625 .025 m
1 .025 L
s
.99625 .0375 m
1 .0375 L
s
.99625 .05 m
1 .05 L
s
.99625 .075 m
1 .075 L
s
.99625 .0875 m
1 .0875 L
s
.99625 .1 m
1 .1 L
s
.99625 .1125 m
1 .1125 L
s
.99625 .1375 m
1 .1375 L
s
.99625 .15 m
1 .15 L
s
.99625 .1625 m
1 .1625 L
s
.99625 .175 m
1 .175 L
s
.99625 .2 m
1 .2 L
s
.99625 .2125 m
1 .2125 L
s
.99625 .225 m
1 .225 L
s
.99625 .2375 m
1 .2375 L
s
.99625 .2625 m
1 .2625 L
s
.99625 .275 m
1 .275 L
s
.99625 .2875 m
1 .2875 L
s
.99625 .3 m
1 .3 L
s
.99625 .325 m
1 .325 L
s
.99625 .3375 m
1 .3375 L
s
.99625 .35 m
1 .35 L
s
.99625 .3625 m
1 .3625 L
s
.99625 .3875 m
1 .3875 L
s
.99625 .4 m
1 .4 L
s
.99625 .4125 m
1 .4125 L
s
.99625 .425 m
1 .425 L
s
.99625 .45 m
1 .45 L
s
.99625 .4625 m
1 .4625 L
s
.99625 .475 m
1 .475 L
s
.99625 .4875 m
1 .4875 L
s
.25 Mabswid
1 0 m
1 .5 L
s
0 0 m
1 0 L
1 .5 L
0 .5 L
closepath
clip
newpath
.008 w
.07143 .0625 Mdot
.07699 .06736 Mdot
.07712 .06748 Mdot
.07726 .0676 Mdot
.07741 .06773 Mdot
.07757 .06787 Mdot
.07773 .06801 Mdot
.0779 .06816 Mdot
.07808 .06832 Mdot
.07828 .06849 Mdot
.07848 .06867 Mdot
.0787 .06886 Mdot
.07893 .06906 Mdot
.07918 .06928 Mdot
.07944 .06951 Mdot
.07972 .06975 Mdot
.08002 .07002 Mdot
.08034 .0703 Mdot
.08069 .07061 Mdot
.08107 .07094 Mdot
.08148 .0713 Mdot
.08193 .07169 Mdot
.08242 .0625 Mdot
.08242 .07212 Mdot
.08268 .0625 Mdot
.08268 .07234 Mdot
.08295 .0625 Mdot
.08295 .07258 Mdot
.08324 .0625 Mdot
.08324 .07284 Mdot
.08354 .0625 Mdot
.08354 .0731 Mdot
.08386 .0625 Mdot
.08386 .07338 Mdot
.0842 .0625 Mdot
.0842 .07367 Mdot
.08455 .0625 Mdot
.08455 .07399 Mdot
.08493 .0625 Mdot
.08493 .07431 Mdot
.08533 .0625 Mdot
.08533 .07466 Mdot
.08575 .0625 Mdot
.08575 .07503 Mdot
.08619 .0625 Mdot
.08619 .07542 Mdot
.08667 .0625 Mdot
.08667 .07584 Mdot
.08718 .0625 Mdot
.08718 .07628 Mdot
.08772 .0625 Mdot
.08772 .07676 Mdot
.08811 .06736 Mdot
.08811 .07709 Mdot
.0883 .0625 Mdot
.0883 .07727 Mdot
.08851 .06748 Mdot
.08851 .07744 Mdot
.08893 .0625 Mdot
.08893 .0676 Mdot
.08893 .07781 Mdot
.08937 .06773 Mdot
.08937 .0782 Mdot
.0896 .0625 Mdot
.0896 .0784 Mdot
.08984 .06787 Mdot
.08984 .07861 Mdot
.09033 .0625 Mdot
.09033 .06801 Mdot
.09033 .07904 Mdot
.09085 .06816 Mdot
.09085 .07949 Mdot
.09112 .0625 Mdot
.09112 .07973 Mdot
.09139 .06832 Mdot
.09139 .07997 Mdot
.09197 .0625 Mdot
.09197 .06849 Mdot
.09197 .08047 Mdot
.09258 .06867 Mdot
.09258 .08101 Mdot
.0929 .0625 Mdot
.0929 .08129 Mdot
.09323 .06886 Mdot
.09323 .08158 Mdot
.0934 .0625 Mdot
.0934 .07212 Mdot
.0934 .08173 Mdot
.09393 .0625 Mdot
.09393 .06906 Mdot
.09393 .07234 Mdot
.09393 .08219 Mdot
.09447 .0625 Mdot
.09447 .07258 Mdot
.09447 .08267 Mdot
.09466 .06928 Mdot
.09466 .08283 Mdot
.09505 .0625 Mdot
.09505 .07284 Mdot
.09505 .08317 Mdot
.09545 .06951 Mdot
.09545 .08352 Mdot
.09566 .0625 Mdot
.09566 .0731 Mdot
.09566 .0837 Mdot
.09629 .0625 Mdot
.09629 .06975 Mdot
.09629 .07338 Mdot
.09629 .08426 Mdot
.09696 .0625 Mdot
.09696 .07367 Mdot
.09696 .08484 Mdot
.0972 .07002 Mdot
.0972 .08505 Mdot
.09767 .0625 Mdot
.09767 .07399 Mdot
.09767 .08546 Mdot
.09817 .0703 Mdot
.09817 .0859 Mdot
.09842 .0625 Mdot
.09842 .07431 Mdot
.09842 .08612 Mdot
.09922 .0625 Mdot
.09922 .06736 Mdot
.09922 .07061 Mdot
.09922 .07466 Mdot
.09922 .07709 Mdot
.09922 .08681 Mdot
.09989 .06748 Mdot
.09989 .07744 Mdot
.09989 .0874 Mdot
.10006 .0625 Mdot
.10006 .07503 Mdot
.10006 .08755 Mdot
.10035 .07094 Mdot
.10035 .08781 Mdot
.10059 .0676 Mdot
.10059 .07781 Mdot
.10059 .08801 Mdot
.10095 .0625 Mdot
.10095 .07542 Mdot
.10095 .08833 Mdot
.10132 .06773 Mdot
.10132 .0782 Mdot
.10132 .08866 Mdot
.10158 .0713 Mdot
.10158 .08888 Mdot
.1019 .0625 Mdot
.1019 .07584 Mdot
.1019 .08916 Mdot
.1021 .06787 Mdot
.1021 .07861 Mdot
.1021 .08934 Mdot
.10292 .0625 Mdot
.10292 .06801 Mdot
.10292 .07169 Mdot
.10292 .07628 Mdot
.10292 .07904 Mdot
.10292 .09005 Mdot
.10378 .06816 Mdot
.10378 .07949 Mdot
.10378 .09081 Mdot
.104 .0625 Mdot
.104 .07676 Mdot
.104 .091 Mdot
.10438 .0625 Mdot
.10438 .07212 Mdot
.10438 .08173 Mdot
.10438 .09133 Mdot
.10469 .06832 Mdot
.10469 .07997 Mdot
.10469 .0916 Mdot
.10516 .0625 Mdot
.10516 .07234 Mdot
.10516 .07727 Mdot
.10516 .08219 Mdot
.10516 .09202 Mdot
.10565 .06849 Mdot
.10565 .08047 Mdot
.10565 .09245 Mdot
.10599 .0625 Mdot
.10599 .07258 Mdot
.10599 .08267 Mdot
.10599 .09274 Mdot
.10641 .0625 Mdot
.10641 .07781 Mdot
.10641 .09311 Mdot
.10667 .06867 Mdot
.10667 .08101 Mdot
.10667 .09334 Mdot
.10685 .0625 Mdot
.10685 .07284 Mdot
.10685 .08317 Mdot
.10685 .09349 Mdot
.10776 .0625 Mdot
.10776 .06886 Mdot
.10776 .0731 Mdot
.10776 .0784 Mdot
.10776 .08158 Mdot
.10776 .0837 Mdot
.10776 .09429 Mdot
.10871 .0625 Mdot
.10871 .07338 Mdot
.10871 .08426 Mdot
.10871 .09512 Mdot
.10891 .06906 Mdot
.10891 .08219 Mdot
.10891 .09529 Mdot
.10921 .0625 Mdot
.10921 .07904 Mdot
.10921 .09556 Mdot
.10972 .0625 Mdot
.10972 .07367 Mdot
.10972 .08484 Mdot
.10972 .096 Mdot
.11013 .06928 Mdot
.11013 .08283 Mdot
.11013 .09637 Mdot
.11032 .06736 Mdot
.11032 .07709 Mdot
.11032 .08681 Mdot
.11032 .09653 Mdot
.11078 .0625 Mdot
.11078 .07399 Mdot
.11078 .07973 Mdot
.11078 .08546 Mdot
.11078 .09693 Mdot
.11125 .06748 Mdot
.11125 .07744 Mdot
.11125 .0874 Mdot
.11125 .09735 Mdot
.11144 .06951 Mdot
.11144 .08352 Mdot
.11144 .09751 Mdot
.1119 .0625 Mdot
.1119 .07431 Mdot
.1119 .08612 Mdot
.1119 .09791 Mdot
.11223 .0676 Mdot
.11223 .07781 Mdot
.11223 .08801 Mdot
.11223 .0982 Mdot
.11249 .0625 Mdot
.11249 .08047 Mdot
.11249 .09843 Mdot
.11285 .06975 Mdot
.11285 .08426 Mdot
.11285 .09874 Mdot
.11309 .0625 Mdot
.11309 .07466 Mdot
.11309 .08681 Mdot
.11309 .09895 Mdot
.11327 .06773 Mdot
.11327 .0782 Mdot
.11327 .08866 Mdot
.11327 .09911 Mdot
.11435 .0625 Mdot
.11435 .06787 Mdot
.11435 .07002 Mdot
.11435 .07503 Mdot
.11435 .07861 Mdot
.11435 .08129 Mdot
.11435 .08505 Mdot
.11435 .08755 Mdot
.11435 .08934 Mdot
.11435 .10006 Mdot
.11535 .07212 Mdot
.11535 .08173 Mdot
.11535 .09133 Mdot
.11535 .10093 Mdot
.11549 .06801 Mdot
.11549 .07904 Mdot
.11549 .09005 Mdot
.11549 .10106 Mdot
.11569 .0625 Mdot
.11569 .07542 Mdot
.11569 .08833 Mdot
.11569 .10123 Mdot
.11597 .0703 Mdot
.11597 .0859 Mdot
.11597 .10147 Mdot
.11639 .0625 Mdot
.11639 .07234 Mdot
.11639 .08219 Mdot
.11639 .09202 Mdot
.11639 .10184 Mdot
.1167 .06816 Mdot
.1167 .07949 Mdot
.1167 .09081 Mdot
.1167 .10211 Mdot
.11711 .0625 Mdot
.11711 .07584 Mdot
.11711 .08916 Mdot
.11711 .10247 Mdot
.11748 .07258 Mdot
.11748 .08267 Mdot
.11748 .09274 Mdot
.11748 .1028 Mdot
.11771 .07061 Mdot
.11771 .08681 Mdot
.11771 .103 Mdot
.11797 .06832 Mdot
.11797 .07997 Mdot
.11797 .0916 Mdot
.11797 .10322 Mdot
.11863 .0625 Mdot
.11863 .07284 Mdot
.11863 .07628 Mdot
.11863 .08317 Mdot
.11863 .09005 Mdot
.11863 .09349 Mdot
.11863 .1038 Mdot
.11932 .06849 Mdot
.11932 .08047 Mdot
.11932 .09245 Mdot
.11932 .1044 Mdot
.11959 .07094 Mdot
.11959 .08781 Mdot
.11959 .10464 Mdot
.11984 .0731 Mdot
.11984 .0837 Mdot
.11984 .09429 Mdot
.11984 .10486 Mdot
.12026 .0625 Mdot
.12026 .07676 Mdot
.12026 .091 Mdot
.12026 .10522 Mdot
.12074 .06867 Mdot
.12074 .08101 Mdot
.12074 .09334 Mdot
.12074 .10565 Mdot
.12111 .0625 Mdot
.12111 .07338 Mdot
.12111 .08426 Mdot
.12111 .09512 Mdot
.12111 .10597 Mdot
.1214 .06736 Mdot
.1214 .07709 Mdot
.1214 .08681 Mdot
.1214 .09653 Mdot
.1214 .10623 Mdot
.12164 .0713 Mdot
.12164 .08888 Mdot
.12164 .10643 Mdot
.122 .07727 Mdot
.122 .09202 Mdot
.122 .10675 Mdot
.12226 .06886 Mdot
.12226 .08158 Mdot
.12226 .09429 Mdot
.12226 .10697 Mdot
.12245 .07367 Mdot
.12245 .08484 Mdot
.12245 .096 Mdot
.12245 .10714 Mdot
.1226 .06748 Mdot
.1226 .07744 Mdot
.1226 .0874 Mdot
.1226 .09735 Mdot
.1226 .10728 Mdot
.12386 .07169 Mdot
.12386 .07399 Mdot
.12386 .07781 Mdot
.12386 .08219 Mdot
.12386 .08546 Mdot
.12386 .08801 Mdot
.12386 .09005 Mdot
.12386 .09311 Mdot
.12386 .09529 Mdot
.12386 .0982 Mdot
.12386 .10838 Mdot
.12519 .0782 Mdot
.12519 .08866 Mdot
.12519 .09911 Mdot
.12519 .10954 Mdot
.12536 .08612 Mdot
.12536 .09791 Mdot
.12536 .10969 Mdot
.12558 .08283 Mdot
.12558 .09637 Mdot
.12558 .10988 Mdot
.12588 .0784 Mdot
.12588 .09429 Mdot
.12588 .11014 Mdot
.1263 .08173 Mdot
.1263 .09133 Mdot
.1263 .10093 Mdot
.1263 .11051 Mdot
.12658 .08934 Mdot
.12658 .10006 Mdot
.12658 .11076 Mdot
.12694 .08681 Mdot
.12694 .09895 Mdot
.12694 .11107 Mdot
.12741 .08352 Mdot
.12741 .09751 Mdot
.12741 .11148 Mdot
.1276 .08219 Mdot
.1276 .09202 Mdot
.1276 .10184 Mdot
.1276 .11165 Mdot
.12805 .09005 Mdot
.12805 .09556 Mdot
.12805 .10106 Mdot
.12805 .11204 Mdot
.12862 .08755 Mdot
.12862 .10006 Mdot
.12862 .11254 Mdot
.12896 .09274 Mdot
.12896 .1028 Mdot
.12896 .11284 Mdot
.12937 .09874 Mdot
.12937 .1132 Mdot
.12959 .09081 Mdot
.12959 .10211 Mdot
.12959 .11339 Mdot
.1304 .08833 Mdot
.1304 .09349 Mdot
.1304 .09693 Mdot
.1304 .10123 Mdot
.1304 .1038 Mdot
.1304 .1141 Mdot
.13122 .0916 Mdot
.13122 .10322 Mdot
.13122 .11482 Mdot
.13147 .10006 Mdot
.13147 .11503 Mdot
.1319 .09429 Mdot
.1319 .10486 Mdot
.1319 .11542 Mdot
.13229 .10247 Mdot
.13229 .11576 Mdot
.13247 .09653 Mdot
.13247 .10623 Mdot
.13247 .11591 Mdot
.13295 .09245 Mdot
.13295 .09843 Mdot
.13295 .1044 Mdot
.13295 .11633 Mdot
.13349 .09512 Mdot
.13349 .10597 Mdot
.13349 .1168 Mdot
.13372 .10147 Mdot
.13372 .11701 Mdot
.13394 .09735 Mdot
.13394 .10728 Mdot
.13394 .11719 Mdot
.13431 .1038 Mdot
.13431 .11752 Mdot
.13478 .10565 Mdot
.13478 .11793 Mdot
.13516 .096 Mdot
.13516 .10714 Mdot
.13516 .11826 Mdot
.13547 .0982 Mdot
.13547 .10838 Mdot
.13547 .11854 Mdot
.13574 .10006 Mdot
.13574 .11877 Mdot
.13616 .103 Mdot
.13616 .11914 Mdot
.13647 .10522 Mdot
.13647 .11941 Mdot
.13672 .10697 Mdot
.13672 .11963 Mdot
.13692 .10838 Mdot
.13692 .11981 Mdot
.13709 .09911 Mdot
.13709 .10954 Mdot
.13709 .11995 Mdot
.13723 .10093 Mdot
.13723 .11051 Mdot
.13723 .12007 Mdot
.13879 .10184 Mdot
.13879 .10464 Mdot
.13879 .10675 Mdot
.13879 .10969 Mdot
.13879 .11076 Mdot
.13879 .11165 Mdot
.13879 .12144 Mdot
.14042 .1028 Mdot
.14042 .11284 Mdot
.14042 .12287 Mdot
.14057 .11204 Mdot
.14057 .123 Mdot
.14076 .11107 Mdot
.14076 .12316 Mdot
.14098 .10988 Mdot
.14098 .12336 Mdot
.14127 .10838 Mdot
.14127 .12361 Mdot
.14164 .10643 Mdot
.14164 .12393 Mdot
.14214 .1141 Mdot
.14214 .12437 Mdot
.14246 .11339 Mdot
.14246 .12465 Mdot
.14285 .11254 Mdot
.14285 .12499 Mdot
.14333 .11148 Mdot
.14333 .12541 Mdot
.14352 .11591 Mdot
.14352 .12558 Mdot
.14394 .11014 Mdot
.14394 .11542 Mdot
.14394 .12595 Mdot
.14445 .11482 Mdot
.14445 .12639 Mdot
.14474 .10838 Mdot
.14474 .12665 Mdot
.14507 .1141 Mdot
.14507 .12693 Mdot
.14524 .11719 Mdot
.14524 .12709 Mdot
.14584 .1132 Mdot
.14584 .1168 Mdot
.14584 .12761 Mdot
.14655 .11633 Mdot
.14655 .12823 Mdot
.14683 .11204 Mdot
.14683 .12847 Mdot
.14706 .11854 Mdot
.14706 .12867 Mdot
.14743 .11576 Mdot
.14743 .129 Mdot
.14784 .11826 Mdot
.14784 .12936 Mdot
.14813 .12007 Mdot
.14813 .12962 Mdot
.14853 .11503 Mdot
.14853 .12996 Mdot
.14878 .11793 Mdot
.14878 .13019 Mdot
.14896 .11995 Mdot
.14896 .13034 Mdot
.14995 .11752 Mdot
.14995 .11981 Mdot
.14995 .1312 Mdot
.15096 .12144 Mdot
.15096 .13209 Mdot
.15115 .11963 Mdot
.15115 .13225 Mdot
.15142 .11701 Mdot
.15142 .13249 Mdot
.15185 .12287 Mdot
.15185 .13287 Mdot
.15218 .12144 Mdot
.15218 .13316 Mdot
.15264 .11941 Mdot
.15264 .13356 Mdot
.15307 .123 Mdot
.15307 .13393 Mdot
.15334 .13417 Mdot
.15366 .12144 Mdot
.15366 .13445 Mdot
.15385 .12437 Mdot
.15385 .13462 Mdot
.15454 .12316 Mdot
.15454 .12558 Mdot
.15454 .13522 Mdot
.15529 .12465 Mdot
.15529 .13588 Mdot
.15552 .12144 Mdot
.15552 .13608 Mdot
.15595 .12595 Mdot
.15595 .13645 Mdot
.15634 .12336 Mdot
.15634 .1368 Mdot
.15652 .12709 Mdot
.15652 .13696 Mdot
.15704 .12499 Mdot
.15704 .13741 Mdot
.15763 .12639 Mdot
.15763 .13793 Mdot
.1579 .13817 Mdot
.15815 .12761 Mdot
.15815 .13839 Mdot
.15861 .12867 Mdot
.15861 .13878 Mdot
.15901 .12962 Mdot
.15901 .13914 Mdot
.1592 .12541 Mdot
.1592 .1393 Mdot
.15969 .12693 Mdot
.15969 .13973 Mdot
.16011 .12823 Mdot
.16011 .1401 Mdot
.16048 .12936 Mdot
.16048 .14042 Mdot
.1608 .13034 Mdot
.1608 .1407 Mdot
.16108 .1312 Mdot
.16108 .14095 Mdot
.16155 .14136 Mdot
.16194 .14169 Mdot
.16225 .14197 Mdot
.16252 .129 Mdot
.16252 .1422 Mdot
.16274 .13019 Mdot
.16274 .1424 Mdot
.16293 .1312 Mdot
.16293 .14257 Mdot
.1631 .13209 Mdot
.1631 .14271 Mdot
.16325 .13287 Mdot
.16325 .14284 Mdot
.16553 .13225 Mdot
.16553 .13316 Mdot
.16553 .13393 Mdot
.16553 .13462 Mdot
.16553 .14483 Mdot
.16777 .13696 Mdot
.16777 .1468 Mdot
.16792 .13645 Mdot
.16792 .14693 Mdot
.16808 .13588 Mdot
.16808 .14707 Mdot
.16827 .13522 Mdot
.16827 .14723 Mdot
.16849 .13445 Mdot
.16849 .14742 Mdot
.16874 .14765 Mdot
.16904 .14791 Mdot
.16941 .14823 Mdot
.16986 .13914 Mdot
.16986 .14863 Mdot
.17013 .13878 Mdot
.17013 .14886 Mdot
.17043 .13839 Mdot
.17043 .14913 Mdot
.17078 .13793 Mdot
.17078 .14943 Mdot
.17117 .13741 Mdot
.17117 .14978 Mdot
.17164 .15018 Mdot
.17218 .14095 Mdot
.17218 .15066 Mdot
.1726 .1407 Mdot
.1726 .15103 Mdot
.17283 .15123 Mdot
.17308 .14042 Mdot
.17308 .15144 Mdot
.17363 .1401 Mdot
.17363 .15192 Mdot
.17426 .15248 Mdot
.17461 .14284 Mdot
.17461 .15279 Mdot
.175 .15312 Mdot
.1752 .14271 Mdot
.1752 .1533 Mdot
.17587 .14257 Mdot
.17587 .15389 Mdot
.17648 .14483 Mdot
.17648 .15442 Mdot
.17664 .15456 Mdot
.17693 .15481 Mdot
.17716 .14483 Mdot
.17716 .15502 Mdot
.17754 .15535 Mdot
.17794 .14483 Mdot
.17794 .1557 Mdot
.17822 .15594 Mdot
.17859 .15627 Mdot
.17882 .15647 Mdot
.17898 .1468 Mdot
.17898 .15661 Mdot
.17984 .14693 Mdot
.17984 .15736 Mdot
.18082 .14707 Mdot
.18082 .15822 Mdot
.18103 .15841 Mdot
.18136 .15869 Mdot
.18161 .14886 Mdot
.18161 .15891 Mdot
.18194 .1592 Mdot
.18244 .15964 Mdot
.18266 .14913 Mdot
.18266 .15983 Mdot
.18324 .15066 Mdot
.18324 .16034 Mdot
.18387 .16088 Mdot
.18413 .16111 Mdot
.18436 .15103 Mdot
.18436 .16132 Mdot
.18477 .16167 Mdot
.18525 .16209 Mdot
.18563 .16243 Mdot
.18594 .15279 Mdot
.18594 .16269 Mdot
.18619 .16291 Mdot
.18658 .16325 Mdot
.18686 .1635 Mdot
.18708 .1637 Mdot
.18726 .1533 Mdot
.18726 .16385 Mdot
.1874 .07212 Mdot
.1874 .08173 Mdot
.1874 .15442 Mdot
.1874 .16397 Mdot
.18876 .15502 Mdot
.18876 .16517 Mdot
.19015 .07234 Mdot
.19015 .08219 Mdot
.19015 .15661 Mdot
.19015 .16638 Mdot
.1903 .16651 Mdot
.19049 .16668 Mdot
.19072 .16688 Mdot
.19103 .16715 Mdot
.19145 .16752 Mdot
.19173 .06773 Mdot
.19173 .16776 Mdot
.19206 .16806 Mdot
.19249 .16843 Mdot
.19304 .0625 Mdot
.19304 .07258 Mdot
.19304 .08267 Mdot
.19304 .15891 Mdot
.19304 .16891 Mdot
.19351 .16932 Mdot
.19379 .16956 Mdot
.1941 .16983 Mdot
.19427 .06736 Mdot
.19427 .07709 Mdot
.19427 .08681 Mdot
.19427 .09653 Mdot
.19427 .16034 Mdot
.19427 .16998 Mdot
.19484 .06787 Mdot
.19484 .07861 Mdot
.19484 .17049 Mdot
.19556 .17111 Mdot
.19584 .17136 Mdot
.19608 .0625 Mdot
.19608 .07284 Mdot
.19608 .08317 Mdot
.19608 .09349 Mdot
.19608 .17157 Mdot
.19646 .17191 Mdot
.1969 .17229 Mdot
.19722 .06748 Mdot
.19722 .07744 Mdot
.19722 .0874 Mdot
.19722 .09735 Mdot
.19722 .17256 Mdot
.19765 .17294 Mdot
.19793 .17319 Mdot
.19813 .06801 Mdot
.19813 .07904 Mdot
.19813 .17336 Mdot
.19827 .0625 Mdot
.19827 .07212 Mdot
.19827 .08173 Mdot
.19827 .09133 Mdot
.19827 .10093 Mdot
.19827 .11051 Mdot
.19827 .16397 Mdot
.19827 .17349 Mdot
.19926 .0625 Mdot
.19926 .0837 Mdot
.19926 .09429 Mdot
.19926 .17435 Mdot
.20031 .0676 Mdot
.20031 .07781 Mdot
.20031 .08801 Mdot
.20031 .0982 Mdot
.20031 .17527 Mdot
.20048 .17542 Mdot
.20071 .17562 Mdot
.20104 .17591 Mdot
.20128 .0625 Mdot
.20128 .07234 Mdot
.20128 .08219 Mdot
.20128 .09202 Mdot
.20128 .10184 Mdot
.20128 .11165 Mdot
.20128 .17612 Mdot
.20159 .06816 Mdot
.20159 .07949 Mdot
.20159 .09081 Mdot
.20159 .17639 Mdot
.20201 .17676 Mdot
.20261 .0625 Mdot
.20261 .07338 Mdot
.20261 .08426 Mdot
.20261 .09512 Mdot
.20261 .17729 Mdot
.20319 .17779 Mdot
.20356 .06773 Mdot
.20356 .0782 Mdot
.20356 .08866 Mdot
.20356 .09911 Mdot
.20356 .10954 Mdot
.20356 .17811 Mdot
.204 .1785 Mdot
.20426 .07466 Mdot
.20426 .17873 Mdot
.20443 .0625 Mdot
.20443 .07258 Mdot
.20443 .08267 Mdot
.20443 .09274 Mdot
.20443 .1028 Mdot
.20443 .11284 Mdot
.20443 .17888 Mdot
.20524 .06736 Mdot
.20524 .06832 Mdot
.20524 .07709 Mdot
.20524 .07997 Mdot
.20524 .08681 Mdot
.20524 .0916 Mdot
.20524 .09653 Mdot
.20524 .10623 Mdot
.20524 .11591 Mdot
.20524 .17959 Mdot
.20614 .0625 Mdot
.20614 .07367 Mdot
.20614 .08484 Mdot
.20614 .096 Mdot
.20614 .10714 Mdot
.20614 .18038 Mdot
.20636 .18056 Mdot
.2067 .18087 Mdot
.20697 .06787 Mdot
.20697 .07861 Mdot
.20697 .08934 Mdot
.20697 .10006 Mdot
.20697 .11076 Mdot
.20697 .1811 Mdot
.20736 .18144 Mdot
.20774 .07284 Mdot
.20774 .08317 Mdot
.20774 .09349 Mdot
.20774 .1038 Mdot
.20774 .1141 Mdot
.20774 .18177 Mdot
.20798 .18198 Mdot
.20827 .0625 Mdot
.20827 .07503 Mdot
.20827 .08755 Mdot
.20827 .18224 Mdot
.20845 .06748 Mdot
.20845 .07744 Mdot
.20845 .0874 Mdot
.20845 .09735 Mdot
.20845 .10728 Mdot
.20845 .11719 Mdot
.20845 .18239 Mdot
.20911 .06849 Mdot
.20911 .08047 Mdot
.20911 .09133 Mdot
.20911 .09245 Mdot
.20911 .10093 Mdot
.20911 .11051 Mdot
.20911 .12007 Mdot
.20911 .12962 Mdot
.20911 .18297 Mdot
.20987 .0625 Mdot
.20987 .07399 Mdot
.20987 .08546 Mdot
.20987 .09693 Mdot
.20987 .10838 Mdot
.20987 .18363 Mdot
.21011 .18384 Mdot
.21057 .06801 Mdot
.21057 .07904 Mdot
.21057 .09005 Mdot
.21057 .10106 Mdot
.21057 .11204 Mdot
.21057 .18425 Mdot
.211 .18463 Mdot
.21121 .0731 Mdot
.21121 .0837 Mdot
.21121 .09429 Mdot
.21121 .10486 Mdot
.21121 .11542 Mdot
.21121 .18481 Mdot
.21181 .06928 Mdot
.21181 .08801 Mdot
.21181 .0982 Mdot
.21181 .10838 Mdot
.21181 .11854 Mdot
.21181 .12867 Mdot
.21181 .18533 Mdot
.21236 .10184 Mdot
.21236 .11165 Mdot
.21236 .12144 Mdot
.21236 .1312 Mdot
.21236 .18581 Mdot
.21253 .0625 Mdot
.21253 .07542 Mdot
.21253 .08833 Mdot
.21253 .18597 Mdot
.21287 .18626 Mdot
.2132 .06867 Mdot
.2132 .08101 Mdot
.2132 .09334 Mdot
.2132 .10565 Mdot
.2132 .18655 Mdot
.2138 .0625 Mdot
.2138 .07431 Mdot
.2138 .08612 Mdot
.2138 .09791 Mdot
.2138 .10969 Mdot
.2138 .18708 Mdot
.21435 .06816 Mdot
.21435 .07949 Mdot
.21435 .09081 Mdot
.21435 .10211 Mdot
.21435 .11339 Mdot
.21435 .18756 Mdot
.21461 .18779 Mdot
.21486 .08426 Mdot
.21486 .09512 Mdot
.21486 .10597 Mdot
.21486 .1168 Mdot
.21486 .12761 Mdot
.21486 .188 Mdot
.21533 .09911 Mdot
.21533 .10954 Mdot
.21533 .11995 Mdot
.21533 .13034 Mdot
.21533 .18842 Mdot
.21577 .1028 Mdot
.21577 .11284 Mdot
.21577 .12287 Mdot
.21577 .13287 Mdot
.21577 .1888 Mdot
.21597 .18898 Mdot
.21617 .11591 Mdot
.21617 .12558 Mdot
.21617 .13522 Mdot
.21617 .18915 Mdot
.21655 .06951 Mdot
.21655 .08352 Mdot
.21655 .18948 Mdot
.21707 .0625 Mdot
.21707 .07584 Mdot
.21707 .08916 Mdot
.21707 .10247 Mdot
.21707 .18993 Mdot
.21753 .06886 Mdot
.21753 .08158 Mdot
.21753 .09429 Mdot
.21753 .10697 Mdot
.21753 .19034 Mdot
.21796 .0625 Mdot
.21796 .07466 Mdot
.21796 .08681 Mdot
.21796 .09895 Mdot
.21796 .11107 Mdot
.21796 .19072 Mdot
.21835 .06832 Mdot
.21835 .07997 Mdot
.21835 .0916 Mdot
.21835 .10322 Mdot
.21835 .11482 Mdot
.21835 .12639 Mdot
.21835 .19106 Mdot
.21871 .096 Mdot
.21871 .10714 Mdot
.21871 .11826 Mdot
.21871 .12936 Mdot
.21871 .19137 Mdot
.21904 .10006 Mdot
.21904 .11076 Mdot
.21904 .12144 Mdot
.21904 .13209 Mdot
.21904 .19166 Mdot
.21934 .1141 Mdot
.21934 .12437 Mdot
.21934 .13462 Mdot
.21934 .19193 Mdot
.21963 .11719 Mdot
.21963 .12709 Mdot
.21963 .13696 Mdot
.21963 .19217 Mdot
.21989 .12007 Mdot
.21989 .12962 Mdot
.21989 .13914 Mdot
.21989 .1924 Mdot
.22013 .19262 Mdot
.22058 .193 Mdot
.22096 .19334 Mdot
.22131 .0625 Mdot
.22131 .19365 Mdot
.22162 .06975 Mdot
.22162 .08426 Mdot
.22162 .19391 Mdot
.22189 .0625 Mdot
.22189 .07628 Mdot
.22189 .09005 Mdot
.22189 .1038 Mdot
.22189 .19416 Mdot
.22214 .06906 Mdot
.22214 .08219 Mdot
.22214 .09529 Mdot
.22214 .10838 Mdot
.22214 .19438 Mdot
.22237 .07503 Mdot
.22237 .08755 Mdot
.22237 .10006 Mdot
.22237 .11254 Mdot
.22237 .12499 Mdot
.22237 .19457 Mdot
.22258 .09245 Mdot
.22258 .1044 Mdot
.22258 .11633 Mdot
.22258 .12823 Mdot
.22258 .19475 Mdot
.22277 .10838 Mdot
.22277 .11981 Mdot
.22277 .1312 Mdot
.22277 .19492 Mdot
.22294 .11204 Mdot
.22294 .123 Mdot
.22294 .13393 Mdot
.22294 .19507 Mdot
.2231 .11542 Mdot
.2231 .12595 Mdot
.2231 .13645 Mdot
.2231 .19521 Mdot
.22325 .11854 Mdot
.22325 .12867 Mdot
.22325 .13878 Mdot
.22325 .19534 Mdot
.22339 .12144 Mdot
.22339 .1312 Mdot
.22339 .14095 Mdot
.22339 .19546 Mdot
.22705 .06928 Mdot
.22705 .07002 Mdot
.22705 .0784 Mdot
.22705 .08283 Mdot
.22705 .08505 Mdot
.22705 .091 Mdot
.22705 .09637 Mdot
.22705 .19867 Mdot
.23075 .13696 Mdot
.23075 .1468 Mdot
.23075 .15661 Mdot
.23075 .20191 Mdot
.23089 .13462 Mdot
.23089 .14483 Mdot
.23089 .15502 Mdot
.23089 .20203 Mdot
.23104 .13209 Mdot
.23104 .14271 Mdot
.23104 .1533 Mdot
.23104 .20216 Mdot
.23121 .12936 Mdot
.23121 .14042 Mdot
.23121 .20231 Mdot
.23138 .12639 Mdot
.23138 .13793 Mdot
.23138 .20246 Mdot
.23158 .12316 Mdot
.23158 .13522 Mdot
.23158 .20263 Mdot
.23179 .10697 Mdot
.23179 .11963 Mdot
.23179 .13225 Mdot
.23179 .20282 Mdot
.23202 .10247 Mdot
.23202 .11576 Mdot
.23202 .129 Mdot
.23202 .20302 Mdot
.23228 .08352 Mdot
.23228 .09751 Mdot
.23228 .11148 Mdot
.23228 .12541 Mdot
.23228 .20324 Mdot
.23256 .0625 Mdot
.23256 .07727 Mdot
.23256 .09202 Mdot
.23256 .10675 Mdot
.23256 .12144 Mdot
.23256 .20349 Mdot
.23288 .0703 Mdot
.23288 .0859 Mdot
.23288 .10147 Mdot
.23288 .20377 Mdot
.23324 .0625 Mdot
.23324 .07904 Mdot
.23324 .09556 Mdot
.23324 .20408 Mdot
.23364 .0713 Mdot
.23364 .20444 Mdot
.2341 .20484 Mdot
.23436 .14095 Mdot
.23436 .15066 Mdot
.23436 .16034 Mdot
.23436 .20506 Mdot
.23463 .13878 Mdot
.23463 .14886 Mdot
.23463 .15891 Mdot
.23463 .2053 Mdot
.23492 .13645 Mdot
.23492 .14693 Mdot
.23492 .15736 Mdot
.23492 .20556 Mdot
.23524 .13393 Mdot
.23524 .14483 Mdot
.23524 .1557 Mdot
.23524 .20584 Mdot
.23559 .1312 Mdot
.23559 .14257 Mdot
.23559 .15389 Mdot
.23559 .20614 Mdot
.23596 .12823 Mdot
.23596 .1401 Mdot
.23596 .20647 Mdot
.23638 .12499 Mdot
.23638 .13741 Mdot
.23638 .20683 Mdot
.23683 .12144 Mdot
.23683 .13445 Mdot
.23683 .20722 Mdot
.23732 .11752 Mdot
.23732 .1312 Mdot
.23732 .20766 Mdot
.23787 .09874 Mdot
.23787 .1132 Mdot
.23787 .12761 Mdot
.23787 .15442 Mdot
.23787 .16397 Mdot
.23787 .20814 Mdot
.23827 .15279 Mdot
.23827 .16269 Mdot
.23827 .20849 Mdot
.23848 .09311 Mdot
.23848 .10838 Mdot
.23848 .12361 Mdot
.23848 .20867 Mdot
.2387 .15103 Mdot
.2387 .16132 Mdot
.2387 .20886 Mdot
.23916 .07061 Mdot
.23916 .08681 Mdot
.23916 .103 Mdot
.23916 .11914 Mdot
.23916 .14913 Mdot
.23916 .15983 Mdot
.23916 .20927 Mdot
.23967 .14707 Mdot
.23967 .15822 Mdot
.23967 .20971 Mdot
.23994 .0625 Mdot
.23994 .07973 Mdot
.23994 .09693 Mdot
.23994 .20994 Mdot
.24021 .14483 Mdot
.24021 .15647 Mdot
.24021 .21019 Mdot
.24081 .07169 Mdot
.24081 .09005 Mdot
.24081 .1424 Mdot
.24081 .15456 Mdot
.24081 .21071 Mdot
.24146 .12693 Mdot
.24146 .13973 Mdot
.24146 .15248 Mdot
.24146 .21128 Mdot
.24181 .15661 Mdot
.24181 .16638 Mdot
.24181 .21159 Mdot
.24218 .12336 Mdot
.24218 .1368 Mdot
.24218 .21191 Mdot
.24237 .15502 Mdot
.24237 .16517 Mdot
.24237 .21208 Mdot
.24298 .11941 Mdot
.24298 .13356 Mdot
.24298 .1533 Mdot
.24298 .16385 Mdot
.24298 .2126 Mdot
.24363 .15144 Mdot
.24363 .16243 Mdot
.24363 .21317 Mdot
.24386 .11503 Mdot
.24386 .12996 Mdot
.24386 .21337 Mdot
.24433 .14943 Mdot
.24433 .16088 Mdot
.24433 .21379 Mdot
.24484 .11014 Mdot
.24484 .12595 Mdot
.24484 .21424 Mdot
.24511 .14723 Mdot
.24511 .1592 Mdot
.24511 .21447 Mdot
.24527 .06736 Mdot
.24527 .07709 Mdot
.24527 .08681 Mdot
.24527 .16034 Mdot
.24527 .16998 Mdot
.24527 .21461 Mdot
.24595 .0625 Mdot
.24595 .07258 Mdot
.24595 .08781 Mdot
.24595 .10464 Mdot
.24595 .12144 Mdot
.24595 .14483 Mdot
.24595 .15736 Mdot
.24595 .15891 Mdot
.24595 .16891 Mdot
.24595 .2152 Mdot
.24668 .15736 Mdot
.24668 .16776 Mdot
.24668 .21585 Mdot
.24687 .1422 Mdot
.24687 .15535 Mdot
.24687 .21601 Mdot
.2472 .0625 Mdot
.2472 .08047 Mdot
.2472 .09843 Mdot
.2472 .11633 Mdot
.2472 .2163 Mdot
.24747 .1557 Mdot
.24747 .16651 Mdot
.24747 .21654 Mdot
.24789 .1393 Mdot
.24789 .15312 Mdot
.24789 .21691 Mdot
.24833 .15389 Mdot
.24833 .16517 Mdot
.24833 .21729 Mdot
.24863 .0625 Mdot
.24863 .07212 Mdot
.24863 .08173 Mdot
.24863 .09133 Mdot
.24863 .10093 Mdot
.24863 .16397 Mdot
.24863 .17349 Mdot
.24863 .21755 Mdot
.24902 .13608 Mdot
.24902 .15066 Mdot
.24902 .2179 Mdot
.24926 .15192 Mdot
.24926 .1637 Mdot
.24926 .21811 Mdot
.24943 .06748 Mdot
.24943 .07744 Mdot
.24943 .0874 Mdot
.24943 .09735 Mdot
.24943 .16269 Mdot
.24943 .17256 Mdot
.24943 .21825 Mdot
.25028 .0625 Mdot
.25028 .07284 Mdot
.25028 .08317 Mdot
.25028 .13249 Mdot
.25028 .14978 Mdot
.25028 .16132 Mdot
.25028 .16209 Mdot
.25028 .17157 Mdot
.25028 .219 Mdot
.25121 .15983 Mdot
.25121 .17049 Mdot
.25121 .21981 Mdot
.2514 .14742 Mdot
.2514 .16034 Mdot
.2514 .21998 Mdot
.2517 .11204 Mdot
.2517 .12847 Mdot
.2517 .22023 Mdot
.25221 .15822 Mdot
.25221 .16932 Mdot
.25221 .22068 Mdot
.25263 .14483 Mdot
.25263 .15841 Mdot
.25263 .22105 Mdot
.25282 .0625 Mdot
.25282 .07234 Mdot
.25282 .08219 Mdot
.25282 .09202 Mdot
.25282 .10184 Mdot
.25282 .11165 Mdot
.25282 .17612 Mdot
.25282 .22121 Mdot
.25329 .10643 Mdot
.25329 .12393 Mdot
.25329 .15647 Mdot
.25329 .16806 Mdot
.25329 .22163 Mdot
.25379 .0676 Mdot
.25379 .07781 Mdot
.25379 .08801 Mdot
.25379 .0982 Mdot
.25379 .17527 Mdot
.25379 .22206 Mdot
.25399 .14197 Mdot
.25399 .15627 Mdot
.25399 .22224 Mdot
.25448 .15456 Mdot
.25448 .16668 Mdot
.25448 .22267 Mdot
.25484 .0625 Mdot
.25484 .0731 Mdot
.25484 .0837 Mdot
.25484 .09429 Mdot
.25484 .17435 Mdot
.25484 .22298 Mdot
.25511 .10006 Mdot
.25511 .11877 Mdot
.25511 .22322 Mdot
.2555 .13878 Mdot
.2555 .15389 Mdot
.2555 .22357 Mdot
.25577 .15248 Mdot
.25577 .16517 Mdot
.25577 .2238 Mdot
.25597 .06801 Mdot
.25597 .07904 Mdot
.25597 .17336 Mdot
.25597 .22397 Mdot
.25612 .06736 Mdot
.25612 .07709 Mdot
.25612 .08681 Mdot
.25612 .09653 Mdot
.25612 .10623 Mdot
.25612 .11591 Mdot
.25612 .17959 Mdot
.25612 .2241 Mdot
.2572 .0625 Mdot
.2572 .07258 Mdot
.2572 .08267 Mdot
.2572 .09274 Mdot
.2572 .1028 Mdot
.2572 .11284 Mdot
.2572 .13522 Mdot
.2572 .15018 Mdot
.2572 .15123 Mdot
.2572 .1635 Mdot
.2572 .17229 Mdot
.2572 .17888 Mdot
.2572 .22505 Mdot
.25836 .06773 Mdot
.25836 .0782 Mdot
.25836 .08866 Mdot
.25836 .09911 Mdot
.25836 .10954 Mdot
.25836 .17811 Mdot
.25836 .22607 Mdot
.25854 .17111 Mdot
.25854 .22622 Mdot
.25877 .14765 Mdot
.25877 .16167 Mdot
.25877 .22643 Mdot
.25911 .1312 Mdot
.25911 .14823 Mdot
.25911 .22672 Mdot
.25933 .10093 Mdot
.25933 .11051 Mdot
.25933 .12007 Mdot
.25933 .12962 Mdot
.25933 .18297 Mdot
.25933 .1924 Mdot
.25933 .22692 Mdot
.25962 .0625 Mdot
.25962 .07338 Mdot
.25962 .08426 Mdot
.25962 .09512 Mdot
.25962 .10597 Mdot
.25962 .17729 Mdot
.25962 .22717 Mdot
.26 .16983 Mdot
.26 .2275 Mdot
.26052 .0874 Mdot
.26052 .09735 Mdot
.26052 .10728 Mdot
.26052 .11719 Mdot
.26052 .12709 Mdot
.26052 .14483 Mdot
.26052 .15964 Mdot
.26052 .18239 Mdot
.26052 .19217 Mdot
.26052 .22795 Mdot
.26099 .06816 Mdot
.26099 .07949 Mdot
.26099 .09081 Mdot
.26099 .17639 Mdot
.26099 .22836 Mdot
.26127 .12665 Mdot
.26127 .14483 Mdot
.26127 .22861 Mdot
.26161 .16843 Mdot
.26161 .22891 Mdot
.26179 .07284 Mdot
.26179 .08317 Mdot
.26179 .09349 Mdot
.26179 .1038 Mdot
.26179 .1141 Mdot
.26179 .12437 Mdot
.26179 .18177 Mdot
.26179 .19193 Mdot
.26179 .22907 Mdot
.26247 .0625 Mdot
.26247 .07431 Mdot
.26247 .14169 Mdot
.26247 .15736 Mdot
.26247 .17542 Mdot
.26247 .22966 Mdot
.26317 .06787 Mdot
.26317 .07861 Mdot
.26317 .08934 Mdot
.26317 .10006 Mdot
.26317 .11076 Mdot
.26317 .1811 Mdot
.26317 .19166 Mdot
.26317 .23027 Mdot
.26338 .16688 Mdot
.26338 .23045 Mdot
.26375 .11165 Mdot
.26375 .12144 Mdot
.26375 .1312 Mdot
.26375 .18581 Mdot
.26375 .19546 Mdot
.26375 .23078 Mdot
.26409 .17435 Mdot
.26409 .23108 Mdot
.26466 .0625 Mdot
.26466 .07367 Mdot
.26466 .08484 Mdot
.26466 .096 Mdot
.26466 .10714 Mdot
.26466 .13817 Mdot
.26466 .15481 Mdot
.26466 .18038 Mdot
.26466 .19137 Mdot
.26466 .23158 Mdot
.26513 .10838 Mdot
.26513 .11854 Mdot
.26513 .12867 Mdot
.26513 .18533 Mdot
.26513 .19534 Mdot
.26513 .23199 Mdot
.26534 .16517 Mdot
.26534 .23217 Mdot
.26586 .17319 Mdot
.26586 .23263 Mdot
.26628 .06832 Mdot
.26628 .07997 Mdot
.26628 .0916 Mdot
.26628 .10322 Mdot
.26628 .17959 Mdot
.26628 .19106 Mdot
.26628 .23299 Mdot
.26662 .09429 Mdot
.26662 .10486 Mdot
.26662 .11542 Mdot
.26662 .12595 Mdot
.26662 .18481 Mdot
.26662 .19521 Mdot
.26662 .23329 Mdot
.2669 .12558 Mdot
.2669 .13522 Mdot
.2669 .14483 Mdot
.2669 .18915 Mdot
.2669 .19867 Mdot
.2669 .23354 Mdot
.26714 .13417 Mdot
.26714 .15192 Mdot
.26714 .23375 Mdot
.26752 .16325 Mdot
.26752 .23408 Mdot
.26781 .17191 Mdot
.26781 .23434 Mdot
.26804 .0625 Mdot
.26804 .07466 Mdot
.26804 .08681 Mdot
.26804 .17873 Mdot
.26804 .23454 Mdot
.26823 .07904 Mdot
.26823 .09005 Mdot
.26823 .10106 Mdot
.26823 .11204 Mdot
.26823 .123 Mdot
.26823 .18425 Mdot
.26823 .19507 Mdot
.26823 .2347 Mdot
.26838 .12287 Mdot
.26838 .13287 Mdot
.26838 .14284 Mdot
.26838 .1888 Mdot
.26838 .19867 Mdot
.26838 .23483 Mdot
.26997 .0625 Mdot
.26997 .06886 Mdot
.26997 .07399 Mdot
.26997 .08158 Mdot
.26997 .08546 Mdot
.26997 .10838 Mdot
.26997 .10954 Mdot
.26997 .11981 Mdot
.26997 .11995 Mdot
.26997 .12962 Mdot
.26997 .13034 Mdot
.26997 .13914 Mdot
.26997 .1407 Mdot
.26997 .14863 Mdot
.26997 .16111 Mdot
.26997 .17779 Mdot
.26997 .18363 Mdot
.26997 .1924 Mdot
.26997 .19492 Mdot
.26997 .23622 Mdot
.27154 .12709 Mdot
.27154 .13696 Mdot
.27154 .1468 Mdot
.27154 .19217 Mdot
.27154 .20191 Mdot
.27154 .2376 Mdot
.27169 .10597 Mdot
.27169 .1168 Mdot
.27169 .12761 Mdot
.27169 .188 Mdot
.27169 .19867 Mdot
.27169 .23773 Mdot
.27187 .06849 Mdot
.27187 .08047 Mdot
.27187 .09245 Mdot
.27187 .1044 Mdot
.27187 .11633 Mdot
.27187 .18297 Mdot
.27187 .19475 Mdot
.27187 .23788 Mdot
.27209 .17676 Mdot
.27209 .23808 Mdot
.27237 .16891 Mdot
.27237 .23832 Mdot
.27273 .15869 Mdot
.27273 .23864 Mdot
.27323 .12437 Mdot
.27323 .13462 Mdot
.27323 .14483 Mdot
.27323 .19193 Mdot
.27323 .20203 Mdot
.27323 .23908 Mdot
.27355 .10211 Mdot
.27355 .11339 Mdot
.27355 .12465 Mdot
.27355 .18756 Mdot
.27355 .19867 Mdot
.27355 .23936 Mdot
.27394 .0625 Mdot
.27394 .07503 Mdot
.27394 .08755 Mdot
.27394 .10006 Mdot
.27394 .18224 Mdot
.27394 .19457 Mdot
.27394 .2397 Mdot
.27443 .17562 Mdot
.27443 .24012 Mdot
.27462 .14095 Mdot
.27462 .15066 Mdot
.27462 .19546 Mdot
.27462 .20506 Mdot
.27462 .24029 Mdot
.27505 .12144 Mdot
.27505 .13209 Mdot
.27505 .14271 Mdot
.27505 .16715 Mdot
.27505 .19166 Mdot
.27505 .20216 Mdot
.27505 .24067 Mdot
.27557 .08612 Mdot
.27557 .09791 Mdot
.27557 .10969 Mdot
.27557 .12144 Mdot
.27557 .18708 Mdot
.27557 .19867 Mdot
.27557 .24113 Mdot
.27588 .15594 Mdot
.27588 .24139 Mdot
.27622 .06906 Mdot
.27622 .08219 Mdot
.27622 .09529 Mdot
.27622 .18144 Mdot
.27622 .19438 Mdot
.27622 .24169 Mdot
.2764 .13878 Mdot
.2764 .14886 Mdot
.2764 .19534 Mdot
.2764 .2053 Mdot
.2764 .24185 Mdot
.27702 .11826 Mdot
.27702 .12936 Mdot
.27702 .14042 Mdot
.27702 .17435 Mdot
.27702 .19137 Mdot
.27702 .20231 Mdot
.27702 .2424 Mdot
.27762 .14483 Mdot
.27762 .15442 Mdot
.27762 .16397 Mdot
.27762 .20814 Mdot
.27762 .24291 Mdot
.27778 .06867 Mdot
.27778 .08101 Mdot
.27778 .09334 Mdot
.27778 .10565 Mdot
.27778 .11793 Mdot
.27778 .18655 Mdot
.27778 .19867 Mdot
.27778 .24306 Mdot
.27807 .16517 Mdot
.27807 .24331 Mdot
.27832 .13645 Mdot
.27832 .14693 Mdot
.27832 .19521 Mdot
.27832 .20556 Mdot
.27832 .24353 Mdot
.27872 .0625 Mdot
.27872 .07628 Mdot
.27872 .18056 Mdot
.27872 .19416 Mdot
.27872 .24388 Mdot
.27916 .11482 Mdot
.27916 .12639 Mdot
.27916 .13793 Mdot
.27916 .19106 Mdot
.27916 .20246 Mdot
.27916 .24427 Mdot
.27948 .14284 Mdot
.27948 .15279 Mdot
.27948 .16269 Mdot
.27948 .20849 Mdot
.27948 .24455 Mdot
.27992 .17294 Mdot
.27992 .24493 Mdot
.2802 .0625 Mdot
.2802 .07542 Mdot
.2802 .08833 Mdot
.2802 .10123 Mdot
.2802 .1141 Mdot
.2802 .18597 Mdot
.2802 .19867 Mdot
.2802 .24517 Mdot
.28039 .13393 Mdot
.28039 .14483 Mdot
.28039 .20584 Mdot
.28039 .24534 Mdot
.2815 .11107 Mdot
.2815 .12316 Mdot
.2815 .13522 Mdot
.2815 .1407 Mdot
.2815 .15103 Mdot
.2815 .16132 Mdot
.2815 .16291 Mdot
.2815 .17959 Mdot
.2815 .19391 Mdot
.2815 .20263 Mdot
.2815 .20886 Mdot
.2815 .24631 Mdot
.28249 .15661 Mdot
.28249 .16638 Mdot
.28249 .21159 Mdot
.28249 .24718 Mdot
.28264 .1312 Mdot
.28264 .14257 Mdot
.28264 .20614 Mdot
.28264 .24731 Mdot
.28286 .06928 Mdot
.28286 .08283 Mdot
.28286 .09637 Mdot
.28286 .10988 Mdot
.28286 .19867 Mdot
.28286 .2475 Mdot
.28317 .17136 Mdot
.28317 .24777 Mdot
.28367 .13839 Mdot
.28367 .14913 Mdot
.28367 .15983 Mdot
.28367 .20927 Mdot
.28367 .24821 Mdot
.28404 .09429 Mdot
.28404 .10697 Mdot
.28404 .11963 Mdot
.28404 .13225 Mdot
.28404 .20282 Mdot
.28404 .24854 Mdot
.28458 .15502 Mdot
.28458 .16517 Mdot
.28458 .19365 Mdot
.28458 .21208 Mdot
.28458 .24901 Mdot
.28509 .12823 Mdot
.28509 .1401 Mdot
.28509 .20647 Mdot
.28509 .24945 Mdot
.28541 .06736 Mdot
.28541 .07709 Mdot
.28541 .08681 Mdot
.28541 .16034 Mdot
.28541 .16998 Mdot
.28541 .21461 Mdot
.28541 .24973 Mdot
.2858 .0625 Mdot
.2858 .07676 Mdot
.2858 .091 Mdot
.2858 .19867 Mdot
.2858 .25007 Mdot
.28602 .13588 Mdot
.28602 .14707 Mdot
.28602 .15822 Mdot
.28602 .20971 Mdot
.28602 .25027 Mdot
.28684 .08916 Mdot
.28684 .10247 Mdot
.28684 .11576 Mdot
.28684 .1533 Mdot
.28684 .16385 Mdot
.28684 .16956 Mdot
.28684 .20302 Mdot
.28684 .2126 Mdot
.28684 .25099 Mdot
.28759 .15891 Mdot
.28759 .16891 Mdot
.28759 .2152 Mdot
.28759 .25164 Mdot
.28776 .12499 Mdot
.28776 .13741 Mdot
.28776 .20683 Mdot
.28776 .25179 Mdot
.28804 .19334 Mdot
.28804 .25204 Mdot
.28826 .0625 Mdot
.28826 .07212 Mdot
.28826 .08173 Mdot
.28826 .09133 Mdot
.28826 .10093 Mdot
.28826 .16397 Mdot
.28826 .17349 Mdot
.28826 .21755 Mdot
.28826 .25223 Mdot
.28857 .13316 Mdot
.28857 .14483 Mdot
.28857 .21019 Mdot
.28857 .2525 Mdot
.28906 .07002 Mdot
.28906 .08505 Mdot
.28906 .19867 Mdot
.28906 .25293 Mdot
.28929 .15144 Mdot
.28929 .16243 Mdot
.28929 .21317 Mdot
.28929 .25313 Mdot
.28993 .06951 Mdot
.28993 .08352 Mdot
.28993 .09751 Mdot
.28993 .11148 Mdot
.28993 .15736 Mdot
.28993 .16776 Mdot
.28993 .20324 Mdot
.28993 .21585 Mdot
.28993 .25369 Mdot
.29051 .06748 Mdot
.29051 .07744 Mdot
.29051 .0874 Mdot
.29051 .09735 Mdot
.29051 .16269 Mdot
.29051 .17256 Mdot
.29051 .21825 Mdot
.29051 .2542 Mdot
.29069 .12144 Mdot
.29069 .13445 Mdot
.29069 .20722 Mdot
.29069 .25436 Mdot
.29104 .25466 Mdot
.29136 .13019 Mdot
.29136 .1424 Mdot
.29136 .21071 Mdot
.29136 .25494 Mdot
.29194 .14943 Mdot
.29194 .16088 Mdot
.29194 .193 Mdot
.29194 .21379 Mdot
.29194 .25545 Mdot
.29247 .1557 Mdot
.29247 .16651 Mdot
.29247 .21654 Mdot
.29247 .25591 Mdot
.29271 .19867 Mdot
.29271 .25612 Mdot
.29294 .0625 Mdot
.29294 .07284 Mdot
.29294 .08317 Mdot
.29294 .16132 Mdot
.29294 .17157 Mdot
.29294 .219 Mdot
.29294 .25632 Mdot
.29336 .0625 Mdot
.29336 .07234 Mdot
.29336 .07727 Mdot
.29336 .08219 Mdot
.29336 .09202 Mdot
.29336 .10184 Mdot
.29336 .10675 Mdot
.29336 .11165 Mdot
.29336 .17612 Mdot
.29336 .20349 Mdot
.29336 .22121 Mdot
.29336 .25669 Mdot
.29392 .1038 Mdot
.29392 .11752 Mdot
.29392 .1312 Mdot
.29392 .20766 Mdot
.29392 .25718 Mdot
.29441 .12693 Mdot
.29441 .13973 Mdot
.29441 .21128 Mdot
.29441 .2576 Mdot
.29483 .14723 Mdot
.29483 .1592 Mdot
.29483 .21447 Mdot
.29483 .25798 Mdot
.29521 .15389 Mdot
.29521 .16517 Mdot
.29521 .21729 Mdot
.29521 .25831 Mdot
.29555 .06787 Mdot
.29555 .17049 Mdot
.29555 .21981 Mdot
.29555 .25861 Mdot
.29586 .0676 Mdot
.29586 .07781 Mdot
.29586 .08801 Mdot
.29586 .0982 Mdot
.29586 .10838 Mdot
.29586 .17527 Mdot
.29586 .22206 Mdot
.29586 .25887 Mdot
.29613 .08681 Mdot
.29613 .09653 Mdot
.29613 .10623 Mdot
.29613 .11591 Mdot
.29613 .12558 Mdot
.29613 .17959 Mdot
.29613 .18915 Mdot
.29613 .2241 Mdot
.29613 .25911 Mdot
.29638 .19262 Mdot
.29638 .25933 Mdot
.29681 .19867 Mdot
.29681 .25971 Mdot
.29718 .0703 Mdot
.29718 .0859 Mdot
.29718 .10147 Mdot
.29718 .20377 Mdot
.29718 .26003 Mdot
.29749 .09874 Mdot
.29749 .1132 Mdot
.29749 .12761 Mdot
.29749 .20814 Mdot
.29749 .2603 Mdot
.29776 .12336 Mdot
.29776 .1368 Mdot
.29776 .21191 Mdot
.29776 .26054 Mdot
.29799 .14483 Mdot
.29799 .15736 Mdot
.29799 .2152 Mdot
.29799 .26074 Mdot
.2982 .15192 Mdot
.2982 .1637 Mdot
.2982 .21811 Mdot
.2982 .26092 Mdot
.29838 .16932 Mdot
.29838 .22068 Mdot
.29838 .26109 Mdot
.29855 .0625 Mdot
.29855 .0731 Mdot
.29855 .0837 Mdot
.29855 .09429 Mdot
.29855 .10486 Mdot
.29855 .17435 Mdot
.29855 .18481 Mdot
.29855 .22298 Mdot
.29855 .26123 Mdot
.2987 .07258 Mdot
.2987 .08267 Mdot
.2987 .09274 Mdot
.2987 .1028 Mdot
.2987 .11284 Mdot
.2987 .12287 Mdot
.2987 .17888 Mdot
.2987 .1888 Mdot
.2987 .22505 Mdot
.2987 .26136 Mdot
.29883 .11051 Mdot
.29883 .12007 Mdot
.29883 .12962 Mdot
.29883 .18297 Mdot
.29883 .1924 Mdot
.29883 .22692 Mdot
.29883 .26147 Mdot
.30146 .0625 Mdot
.30146 .06773 Mdot
.30146 .06801 Mdot
.30146 .0782 Mdot
.30146 .07904 Mdot
.30146 .08866 Mdot
.30146 .09311 Mdot
.30146 .09556 Mdot
.30146 .09911 Mdot
.30146 .10728 Mdot
.30146 .10954 Mdot
.30146 .11719 Mdot
.30146 .11941 Mdot
.30146 .11995 Mdot
.30146 .12361 Mdot
.30146 .13356 Mdot
.30146 .1422 Mdot
.30146 .14978 Mdot
.30146 .16209 Mdot
.30146 .17336 Mdot
.30146 .18239 Mdot
.30146 .18425 Mdot
.30146 .19217 Mdot
.30146 .20408 Mdot
.30146 .20867 Mdot
.30146 .21601 Mdot
.30146 .22397 Mdot
.30146 .22607 Mdot
.30146 .26378 Mdot
.30415 .12144 Mdot
.30415 .1312 Mdot
.30415 .14095 Mdot
.30415 .18581 Mdot
.30415 .19546 Mdot
.30415 .23078 Mdot
.30415 .26613 Mdot
.30429 .09349 Mdot
.30429 .1038 Mdot
.30429 .1141 Mdot
.30429 .12437 Mdot
.30429 .18177 Mdot
.30429 .19193 Mdot
.30429 .22907 Mdot
.30429 .26625 Mdot
.30444 .0625 Mdot
.30444 .07338 Mdot
.30444 .08426 Mdot
.30444 .09512 Mdot
.30444 .10597 Mdot
.30444 .17729 Mdot
.30444 .188 Mdot
.30444 .22717 Mdot
.30444 .26639 Mdot
.30462 .07399 Mdot
.30462 .08546 Mdot
.30462 .17229 Mdot
.30462 .18363 Mdot
.30462 .22505 Mdot
.30462 .26654 Mdot
.30481 .16668 Mdot
.30481 .22267 Mdot
.30481 .26671 Mdot
.30503 .14742 Mdot
.30503 .16034 Mdot
.30503 .21998 Mdot
.30503 .2669 Mdot
.30528 .1393 Mdot
.30528 .15312 Mdot
.30528 .21691 Mdot
.30528 .26712 Mdot
.30557 .11503 Mdot
.30557 .12996 Mdot
.30557 .21337 Mdot
.30557 .26737 Mdot
.3059 .07061 Mdot
.3059 .08681 Mdot
.3059 .103 Mdot
.3059 .11914 Mdot
.3059 .20927 Mdot
.3059 .26767 Mdot
.3063 .0713 Mdot
.3063 .20444 Mdot
.3063 .26801 Mdot
.30677 .13522 Mdot
.30677 .14483 Mdot
.30677 .19867 Mdot
.30677 .23354 Mdot
.30677 .26842 Mdot
.30704 .11854 Mdot
.30704 .12867 Mdot
.30704 .13878 Mdot
.30704 .18533 Mdot
.30704 .19534 Mdot
.30704 .23199 Mdot
.30704 .26866 Mdot
.30734 .08934 Mdot
.30734 .10006 Mdot
.30734 .11076 Mdot
.30734 .12144 Mdot
.30734 .1811 Mdot
.30734 .19166 Mdot
.30734 .23027 Mdot
.30734 .26892 Mdot
.30768 .06816 Mdot
.30768 .07949 Mdot
.30768 .09081 Mdot
.30768 .10211 Mdot
.30768 .17639 Mdot
.30768 .18756 Mdot
.30768 .22836 Mdot
.30768 .26922 Mdot
.30805 .17111 Mdot
.30805 .18297 Mdot
.30805 .22622 Mdot
.30805 .26955 Mdot
.30848 .16517 Mdot
.30848 .2238 Mdot
.30848 .26992 Mdot
.30896 .15841 Mdot
.30896 .22105 Mdot
.30896 .27034 Mdot
.30951 .13608 Mdot
.30951 .15066 Mdot
.30951 .2179 Mdot
.30951 .27082 Mdot
.30972 .13287 Mdot
.30972 .14284 Mdot
.30972 .19867 Mdot
.30972 .23483 Mdot
.30972 .271 Mdot
.31016 .11014 Mdot
.31016 .11542 Mdot
.31016 .12595 Mdot
.31016 .13645 Mdot
.31016 .19521 Mdot
.31016 .21424 Mdot
.31016 .23329 Mdot
.31016 .27139 Mdot
.31064 .07367 Mdot
.31064 .08484 Mdot
.31064 .096 Mdot
.31064 .10714 Mdot
.31064 .11826 Mdot
.31064 .18038 Mdot
.31064 .19137 Mdot
.31064 .23158 Mdot
.31064 .27181 Mdot
.31091 .0625 Mdot
.31091 .07973 Mdot
.31091 .09693 Mdot
.31091 .1141 Mdot
.31091 .20994 Mdot
.31091 .27205 Mdot
.31119 .0625 Mdot
.31119 .07431 Mdot
.31119 .08612 Mdot
.31119 .09791 Mdot
.31119 .17542 Mdot
.31119 .18708 Mdot
.31119 .22966 Mdot
.31119 .27229 Mdot
.31181 .16983 Mdot
.31181 .18224 Mdot
.31181 .20484 Mdot
.31181 .2275 Mdot
.31181 .27283 Mdot
.31232 .13696 Mdot
.31232 .1468 Mdot
.31232 .15661 Mdot
.31232 .20191 Mdot
.31232 .2376 Mdot
.31232 .27328 Mdot
.31251 .1635 Mdot
.31251 .22505 Mdot
.31251 .27344 Mdot
.31289 .13034 Mdot
.31289 .1407 Mdot
.31289 .19867 Mdot
.31289 .23622 Mdot
.31289 .27378 Mdot
.31331 .15627 Mdot
.31331 .22224 Mdot
.31331 .27414 Mdot
.31352 .11204 Mdot
.31352 .123 Mdot
.31352 .13393 Mdot
.31352 .19507 Mdot
.31352 .2347 Mdot
.31352 .27433 Mdot
.31423 .06832 Mdot
.31423 .07997 Mdot
.31423 .0916 Mdot
.31423 .10322 Mdot
.31423 .11482 Mdot
.31423 .13249 Mdot
.31423 .14791 Mdot
.31423 .19106 Mdot
.31423 .219 Mdot
.31423 .23299 Mdot
.31423 .27495 Mdot
.31486 .15066 Mdot
.31486 .16034 Mdot
.31486 .20506 Mdot
.31486 .24029 Mdot
.31486 .24973 Mdot
.31486 .2755 Mdot
.31502 .06867 Mdot
.31502 .08101 Mdot
.31502 .09334 Mdot
.31502 .17435 Mdot
.31502 .18655 Mdot
.31502 .23108 Mdot
.31502 .27565 Mdot
.31531 .10464 Mdot
.31531 .12144 Mdot
.31531 .2152 Mdot
.31531 .2759 Mdot
.31555 .13462 Mdot
.31555 .14483 Mdot
.31555 .15502 Mdot
.31555 .20203 Mdot
.31555 .23908 Mdot
.31555 .27611 Mdot
.31592 .16843 Mdot
.31592 .18144 Mdot
.31592 .22891 Mdot
.31592 .27643 Mdot
.31632 .12761 Mdot
.31632 .13839 Mdot
.31632 .19867 Mdot
.31632 .23773 Mdot
.31632 .27678 Mdot
.31659 .07169 Mdot
.31659 .09005 Mdot
.31659 .10838 Mdot
.31659 .21071 Mdot
.31659 .27702 Mdot
.31695 .16167 Mdot
.31695 .17562 Mdot
.31695 .22643 Mdot
.31695 .27733 Mdot
.31717 .10838 Mdot
.31717 .11981 Mdot
.31717 .1312 Mdot
.31717 .19492 Mdot
.31717 .23622 Mdot
.31717 .27753 Mdot
.31733 .07212 Mdot
.31733 .15442 Mdot
.31733 .16397 Mdot
.31733 .20814 Mdot
.31733 .21755 Mdot
.31733 .24291 Mdot
.31733 .25223 Mdot
.31733 .27766 Mdot
.31814 .0625 Mdot
.31814 .07466 Mdot
.31814 .08681 Mdot
.31814 .09895 Mdot
.31814 .11107 Mdot
.31814 .14886 Mdot
.31814 .15389 Mdot
.31814 .15891 Mdot
.31814 .19072 Mdot
.31814 .2053 Mdot
.31814 .22357 Mdot
.31814 .23454 Mdot
.31814 .24185 Mdot
.31814 .25164 Mdot
.31814 .27837 Mdot
.31903 .14271 Mdot
.31903 .1533 Mdot
.31903 .20216 Mdot
.31903 .24067 Mdot
.31903 .25099 Mdot
.31903 .27915 Mdot
.31922 .0625 Mdot
.31922 .07542 Mdot
.31922 .18597 Mdot
.31922 .23263 Mdot
.31922 .27932 Mdot
.31952 .12847 Mdot
.31952 .14483 Mdot
.31952 .22023 Mdot
.31952 .27958 Mdot
.32003 .12465 Mdot
.32003 .13588 Mdot
.32003 .19867 Mdot
.32003 .23936 Mdot
.32003 .28002 Mdot
.32046 .16688 Mdot
.32046 .18056 Mdot
.32046 .23045 Mdot
.32046 .2804 Mdot
.32065 .15279 Mdot
.32065 .16269 Mdot
.32065 .20849 Mdot
.32065 .21825 Mdot
.32065 .24455 Mdot
.32065 .2542 Mdot
.32065 .28057 Mdot
.32115 .09245 Mdot
.32115 .09843 Mdot
.32115 .1044 Mdot
.32115 .11633 Mdot
.32115 .12823 Mdot
.32115 .19475 Mdot
.32115 .2163 Mdot
.32115 .23788 Mdot
.32115 .281 Mdot
.32167 .14693 Mdot
.32167 .15736 Mdot
.32167 .20556 Mdot
.32167 .21585 Mdot
.32167 .24353 Mdot
.32167 .25369 Mdot
.32167 .28146 Mdot
.32188 .15964 Mdot
.32188 .17435 Mdot
.32188 .22795 Mdot
.32188 .28165 Mdot
.32241 .06886 Mdot
.32241 .08158 Mdot
.32241 .09429 Mdot
.32241 .10697 Mdot
.32241 .19034 Mdot
.32241 .2821 Mdot
.3228 .14042 Mdot
.3228 .15144 Mdot
.3228 .20231 Mdot
.3228 .25313 Mdot
.3228 .28245 Mdot
.3231 .0625 Mdot
.3231 .07234 Mdot
.3231 .08219 Mdot
.3231 .09202 Mdot
.3231 .16638 Mdot
.3231 .17612 Mdot
.3231 .21159 Mdot
.3231 .22121 Mdot
.3231 .24718 Mdot
.3231 .25669 Mdot
.3231 .28271 Mdot
.32354 .15123 Mdot
.32354 .28309 Mdot
.32384 .18533 Mdot
.32384 .28336 Mdot
.32406 .12144 Mdot
.32406 .13316 Mdot
.32406 .19867 Mdot
.32406 .28355 Mdot
.32423 .16132 Mdot
.32423 .17157 Mdot
.32423 .20886 Mdot
.32423 .219 Mdot
.32423 .25632 Mdot
.32423 .2837 Mdot
.32548 .06736 Mdot
.32548 .07709 Mdot
.32548 .08681 Mdot
.32548 .08755 Mdot
.32548 .09653 Mdot
.32548 .10006 Mdot
.32548 .10623 Mdot
.32548 .11254 Mdot
.32548 .12393 Mdot
.32548 .14136 Mdot
.32548 .14483 Mdot
.32548 .1557 Mdot
.32548 .16998 Mdot
.32548 .17959 Mdot
.32548 .19457 Mdot
.32548 .20584 Mdot
.32548 .21461 Mdot
.32548 .21654 Mdot
.32548 .2241 Mdot
.32548 .25591 Mdot
.32548 .25911 Mdot
.32548 .2848 Mdot
.32672 .0676 Mdot
.32672 .07781 Mdot
.32672 .08801 Mdot
.32672 .16517 Mdot
.32672 .17527 Mdot
.32672 .21208 Mdot
.32672 .22206 Mdot
.32672 .25887 Mdot
.32672 .28588 Mdot
.32688 .13793 Mdot
.32688 .14943 Mdot
.32688 .20246 Mdot
.32688 .21379 Mdot
.32688 .25545 Mdot
.32688 .28602 Mdot
.32709 .0625 Mdot
.32709 .07584 Mdot
.32709 .08916 Mdot
.32709 .10247 Mdot
.32709 .18993 Mdot
.32709 .28621 Mdot
.32738 .15736 Mdot
.32738 .17294 Mdot
.32738 .28646 Mdot
.3278 .08173 Mdot
.3278 .09133 Mdot
.3278 .10093 Mdot
.3278 .11051 Mdot
.3278 .12007 Mdot
.3278 .17349 Mdot
.3278 .18297 Mdot
.3278 .22692 Mdot
.3278 .26147 Mdot
.3278 .28683 Mdot
.32809 .15983 Mdot
.32809 .17049 Mdot
.32809 .20927 Mdot
.32809 .21981 Mdot
.32809 .25861 Mdot
.32809 .28708 Mdot
.32845 .11793 Mdot
.32845 .13019 Mdot
.32845 .1424 Mdot
.32845 .19867 Mdot
.32845 .25494 Mdot
.32845 .2874 Mdot
.32894 .18463 Mdot
.32894 .28782 Mdot
.32914 .0625 Mdot
.32914 .07258 Mdot
.32914 .08267 Mdot
.32914 .09274 Mdot
.32914 .1028 Mdot
.32914 .16891 Mdot
.32914 .17888 Mdot
.32914 .22505 Mdot
.32914 .26136 Mdot
.32914 .288 Mdot
.32961 .14257 Mdot
.32961 .14823 Mdot
.32961 .15389 Mdot
.32961 .20614 Mdot
.32961 .21729 Mdot
.32961 .25831 Mdot
.32961 .28841 Mdot
.33023 .08219 Mdot
.33023 .09529 Mdot
.33023 .10838 Mdot
.33023 .12144 Mdot
.33023 .19438 Mdot
.33023 .28895 Mdot
.33062 .0625 Mdot
.33062 .0731 Mdot
.33062 .16385 Mdot
.33062 .17435 Mdot
.33062 .22298 Mdot
.33062 .26123 Mdot
.33062 .28929 Mdot
.33107 .1785 Mdot
.33107 .28969 Mdot
.33133 .13522 Mdot
.33133 .14723 Mdot
.33133 .20263 Mdot
.33133 .21447 Mdot
.33133 .25798 Mdot
.33133 .28991 Mdot
.33149 .06748 Mdot
.33149 .07744 Mdot
.33149 .0874 Mdot
.33149 .09735 Mdot
.33149 .10728 Mdot
.33149 .11719 Mdot
.33149 .18239 Mdot
.33149 .22795 Mdot
.33149 .26378 Mdot
.33149 .29006 Mdot
.33226 .06951 Mdot
.33226 .08352 Mdot
.33226 .09751 Mdot
.33226 .13741 Mdot
.33226 .15822 Mdot
.33226 .16932 Mdot
.33226 .18948 Mdot
.33226 .22068 Mdot
.33226 .26109 Mdot
.33226 .29073 Mdot
.33308 .06773 Mdot
.33308 .0782 Mdot
.33308 .08866 Mdot
.33308 .09911 Mdot
.33308 .16776 Mdot
.33308 .17811 Mdot
.33308 .22607 Mdot
.33308 .26378 Mdot
.33308 .29144 Mdot
.33326 .1141 Mdot
.33326 .12693 Mdot
.33326 .13973 Mdot
.33326 .19867 Mdot
.33326 .21128 Mdot
.33326 .2576 Mdot
.33326 .2916 Mdot
.33356 .17136 Mdot
.33356 .29186 Mdot
.33378 .10184 Mdot
.33378 .11165 Mdot
.33378 .12144 Mdot
.33378 .1312 Mdot
.33378 .18581 Mdot
.33378 .19546 Mdot
.33378 .23078 Mdot
.33378 .26613 Mdot
.33378 .29206 Mdot
.33411 .15192 Mdot
.33411 .21811 Mdot
.33411 .26092 Mdot
.33411 .29234 Mdot
.3346 .18384 Mdot
.3346 .29278 Mdot
.33483 .16243 Mdot
.33483 .17336 Mdot
.33483 .22397 Mdot
.33483 .26378 Mdot
.33483 .29298 Mdot
.33546 .0625 Mdot
.33546 .07284 Mdot
.33546 .07628 Mdot
.33546 .08317 Mdot
.33546 .09005 Mdot
.33546 .09349 Mdot
.33546 .1038 Mdot
.33546 .1141 Mdot
.33546 .11752 Mdot
.33546 .18177 Mdot
.33546 .19416 Mdot
.33546 .22907 Mdot
.33546 .25718 Mdot
.33546 .26625 Mdot
.33546 .29353 Mdot
.33602 .12558 Mdot
.33602 .13522 Mdot
.33602 .14483 Mdot
.33602 .18915 Mdot
.33602 .19867 Mdot
.33602 .23354 Mdot
.33602 .26842 Mdot
.33602 .29401 Mdot
.33619 .13225 Mdot
.33619 .14483 Mdot
.33619 .2152 Mdot
.33619 .26074 Mdot
.33619 .29416 Mdot
.3365 .14483 Mdot
.3365 .16291 Mdot
.3365 .29444 Mdot
.3368 .15647 Mdot
.3368 .16806 Mdot
.3368 .22163 Mdot
.3368 .26378 Mdot
.3368 .2947 Mdot
.33733 .0625 Mdot
.33733 .07338 Mdot
.33733 .08426 Mdot
.33733 .09512 Mdot
.33733 .17729 Mdot
.33733 .22717 Mdot
.33733 .26639 Mdot
.33733 .29516 Mdot
.33779 .10838 Mdot
.33779 .11854 Mdot
.33779 .12867 Mdot
.33779 .18533 Mdot
.33779 .19534 Mdot
.33779 .23199 Mdot
.33779 .26866 Mdot
.33779 .29556 Mdot
.33799 .0625 Mdot
.33799 .07727 Mdot
.33799 .09202 Mdot
.33799 .18898 Mdot
.33799 .25669 Mdot
.33799 .29574 Mdot
.33819 .13914 Mdot
.33819 .14863 Mdot
.33819 .1924 Mdot
.33819 .23622 Mdot
.33819 .29592 Mdot
.33855 .10988 Mdot
.33855 .12336 Mdot
.33855 .1368 Mdot
.33855 .21191 Mdot
.33855 .26054 Mdot
.33855 .29623 Mdot
.33901 .14978 Mdot
.33901 .16209 Mdot
.33901 .219 Mdot
.33901 .26378 Mdot
.33901 .29663 Mdot
.33941 .16088 Mdot
.33941 .17229 Mdot
.33941 .22505 Mdot
.33941 .26654 Mdot
.33941 .29698 Mdot
.33975 .06787 Mdot
.33975 .07861 Mdot
.33975 .08934 Mdot
.33975 .10006 Mdot
.33975 .11076 Mdot
.33975 .1811 Mdot
.33975 .19166 Mdot
.33975 .23027 Mdot
.33975 .26892 Mdot
.33975 .29728 Mdot
.34004 .12287 Mdot
.34004 .13287 Mdot
.34004 .14284 Mdot
.34004 .1888 Mdot
.34004 .19867 Mdot
.34004 .23483 Mdot
.34004 .271 Mdot
.34004 .29754 Mdot
.34054 .16956 Mdot
.34054 .29797 Mdot
.34093 .18297 Mdot
.34093 .25612 Mdot
.34093 .29831 Mdot
.34125 .06975 Mdot
.34125 .08426 Mdot
.34125 .09874 Mdot
.34125 .1132 Mdot
.34125 .20814 Mdot
.34125 .2603 Mdot
.34125 .29859 Mdot
.34152 .129 Mdot
.34152 .1422 Mdot
.34152 .21601 Mdot
.34152 .26378 Mdot
.34152 .29883 Mdot
.34174 .15456 Mdot
.34174 .16668 Mdot
.34174 .22267 Mdot
.34174 .26671 Mdot
.34174 .29902 Mdot
.34193 .06816 Mdot
.34193 .07949 Mdot
.34193 .09081 Mdot
.34193 .17639 Mdot
.34193 .22836 Mdot
.34193 .26922 Mdot
.34193 .29919 Mdot
.3421 .10486 Mdot
.3421 .11542 Mdot
.3421 .12595 Mdot
.3421 .18481 Mdot
.3421 .19521 Mdot
.3421 .23329 Mdot
.3421 .27139 Mdot
.3421 .29934 Mdot
.34224 .13696 Mdot
.34224 .1468 Mdot
.34224 .20191 Mdot
.34224 .2376 Mdot
.34224 .27328 Mdot
.34224 .29946 Mdot
.34438 .0703 Mdot
.34438 .07367 Mdot
.34438 .08484 Mdot
.34438 .096 Mdot
.34438 .10714 Mdot
.34438 .11826 Mdot
.34438 .11941 Mdot
.34438 .11995 Mdot
.34438 .13034 Mdot
.34438 .1407 Mdot
.34438 .14095 Mdot
.34438 .14742 Mdot
.34438 .16034 Mdot
.34438 .17111 Mdot
.34438 .17591 Mdot
.34438 .18038 Mdot
.34438 .18842 Mdot
.34438 .19137 Mdot
.34438 .20506 Mdot
.34438 .21998 Mdot
.34438 .22622 Mdot
.34438 .24029 Mdot
.34438 .26003 Mdot
.34438 .2669 Mdot
.34438 .27181 Mdot
.34438 .2755 Mdot
.34438 .30133 Mdot
.34646 .15442 Mdot
.34646 .16397 Mdot
.34646 .20814 Mdot
.34646 .24291 Mdot
.34646 .27766 Mdot
.34646 .30315 Mdot
.3466 .13462 Mdot
.3466 .14483 Mdot
.3466 .20203 Mdot
.3466 .23908 Mdot
.3466 .27611 Mdot
.3466 .30327 Mdot
.34675 .10106 Mdot
.34675 .11204 Mdot
.34675 .123 Mdot
.34675 .18425 Mdot
.34675 .19507 Mdot
.34675 .2347 Mdot
.34675 .27433 Mdot
.34675 .30341 Mdot
.34693 .0625 Mdot
.34693 .07431 Mdot
.34693 .08612 Mdot
.34693 .17542 Mdot
.34693 .18708 Mdot
.34693 .22966 Mdot
.34693 .27229 Mdot
.34693 .30357 Mdot
.34714 .15248 Mdot
.34714 .16517 Mdot
.34714 .2238 Mdot
.34714 .26992 Mdot
.34714 .30375 Mdot
.34739 .12541 Mdot
.34739 .1393 Mdot
.34739 .21691 Mdot
.34739 .26712 Mdot
.34739 .30397 Mdot
.34768 .0625 Mdot
.34768 .07781 Mdot
.34768 .09311 Mdot
.34768 .10838 Mdot
.34768 .20867 Mdot
.34768 .26378 Mdot
.34768 .30422 Mdot
.34804 .18198 Mdot
.34804 .25971 Mdot
.34804 .30454 Mdot
.34849 .16752 Mdot
.34849 .30493 Mdot
.34876 .14886 Mdot
.34876 .15891 Mdot
.34876 .2053 Mdot
.34876 .24185 Mdot
.34876 .27837 Mdot
.34876 .30516 Mdot
.34906 .1168 Mdot
.34906 .12761 Mdot
.34906 .13839 Mdot
.34906 .19867 Mdot
.34906 .23773 Mdot
.34906 .27678 Mdot
.34906 .30543 Mdot
.34941 .06832 Mdot
.34941 .07997 Mdot
.34941 .0916 Mdot
.34941 .10322 Mdot
.34941 .11482 Mdot
.34941 .17959 Mdot
.34941 .19106 Mdot
.34941 .23299 Mdot
.34941 .27495 Mdot
.34941 .30573 Mdot
.34981 .16983 Mdot
.34981 .2275 Mdot
.34981 .27283 Mdot
.34981 .30609 Mdot
.35029 .14483 Mdot
.35029 .15841 Mdot
.35029 .22105 Mdot
.35029 .27034 Mdot
.35029 .3065 Mdot
.35085 .10006 Mdot
.35085 .11503 Mdot
.35085 .12996 Mdot
.35085 .15279 Mdot
.35085 .16269 Mdot
.35085 .20849 Mdot
.35085 .21337 Mdot
.35085 .24455 Mdot
.35085 .26737 Mdot
.35085 .28057 Mdot
.35085 .307 Mdot
.3513 .13209 Mdot
.3513 .14271 Mdot
.3513 .1533 Mdot
.3513 .20216 Mdot
.3513 .24067 Mdot
.3513 .27915 Mdot
.3513 .30738 Mdot
.35154 .20408 Mdot
.35154 .26378 Mdot
.35154 .3076 Mdot
.3518 .09693 Mdot
.3518 .10838 Mdot
.3518 .11981 Mdot
.3518 .1312 Mdot
.3518 .19492 Mdot
.3518 .23622 Mdot
.3518 .27753 Mdot
.3518 .30783 Mdot
.35239 .06867 Mdot
.35239 .08101 Mdot
.35239 .17435 Mdot
.35239 .18655 Mdot
.35239 .23108 Mdot
.35239 .25933 Mdot
.35239 .27565 Mdot
.35239 .30834 Mdot
.35289 .0625 Mdot
.35289 .07234 Mdot
.35289 .08219 Mdot
.35289 .09202 Mdot
.35289 .16638 Mdot
.35289 .17612 Mdot
.35289 .21159 Mdot
.35289 .22121 Mdot
.35289 .24718 Mdot
.35289 .25669 Mdot
.35289 .28271 Mdot
.35289 .30878 Mdot
.35308 .1635 Mdot
.35308 .22505 Mdot
.35308 .27344 Mdot
.35308 .30894 Mdot
.35347 .14693 Mdot
.35347 .15736 Mdot
.35347 .20556 Mdot
.35347 .24353 Mdot
.35347 .28146 Mdot
.35347 .30928 Mdot
.35389 .13608 Mdot
.35389 .15066 Mdot
.35389 .2179 Mdot
.35389 .27082 Mdot
.35389 .30966 Mdot
.35412 .11339 Mdot
.35412 .12465 Mdot
.35412 .13588 Mdot
.35412 .19867 Mdot
.35412 .23936 Mdot
.35412 .28002 Mdot
.35412 .30986 Mdot
.35488 .0625 Mdot
.35488 .06736 Mdot
.35488 .07061 Mdot
.35488 .07466 Mdot
.35488 .07709 Mdot
.35488 .08681 Mdot
.35488 .09653 Mdot
.35488 .09895 Mdot
.35488 .103 Mdot
.35488 .10623 Mdot
.35488 .11107 Mdot
.35488 .16998 Mdot
.35488 .17959 Mdot
.35488 .19072 Mdot
.35488 .20927 Mdot
.35488 .21461 Mdot
.35488 .2241 Mdot
.35488 .23454 Mdot
.35488 .24973 Mdot
.35488 .25911 Mdot
.35488 .26767 Mdot
.35488 .27837 Mdot
.35488 .2848 Mdot
.35488 .31052 Mdot
.35558 .15103 Mdot
.35558 .16132 Mdot
.35558 .20886 Mdot
.35558 .219 Mdot
.35558 .24631 Mdot
.35558 .25632 Mdot
.35558 .2837 Mdot
.35558 .31113 Mdot
.35577 .16843 Mdot
.35577 .18144 Mdot
.35577 .22891 Mdot
.35577 .27643 Mdot
.35577 .3113 Mdot
.3561 .19867 Mdot
.3561 .26378 Mdot
.3561 .31158 Mdot
.35637 .12936 Mdot
.35637 .14042 Mdot
.35637 .15144 Mdot
.35637 .20231 Mdot
.35637 .2424 Mdot
.35637 .28245 Mdot
.35637 .31183 Mdot
.35682 .08173 Mdot
.35682 .09133 Mdot
.35682 .10093 Mdot
.35682 .11051 Mdot
.35682 .12007 Mdot
.35682 .14197 Mdot
.35682 .15627 Mdot
.35682 .17349 Mdot
.35682 .18297 Mdot
.35682 .22224 Mdot
.35682 .22692 Mdot
.35682 .26147 Mdot
.35682 .27414 Mdot
.35682 .28683 Mdot
.35682 .31221 Mdot
.35729 .09245 Mdot
.35729 .1044 Mdot
.35729 .11633 Mdot
.35729 .12823 Mdot
.35729 .19475 Mdot
.35729 .23788 Mdot
.35729 .281 Mdot
.35729 .31263 Mdot
.35763 .0676 Mdot
.35763 .07781 Mdot
.35763 .08801 Mdot
.35763 .16517 Mdot
.35763 .17527 Mdot
.35763 .21208 Mdot
.35763 .22206 Mdot
.35763 .25887 Mdot
.35763 .28588 Mdot
.35763 .31292 Mdot
.35808 .11014 Mdot
.35808 .12595 Mdot
.35808 .21424 Mdot
.35808 .27139 Mdot
.35808 .31332 Mdot
.35836 .07542 Mdot
.35836 .17319 Mdot
.35836 .18597 Mdot
.35836 .23263 Mdot
.35836 .27932 Mdot
.35836 .31357 Mdot
.35856 .14483 Mdot
.35856 .1557 Mdot
.35856 .20584 Mdot
.35856 .21654 Mdot
.35856 .25591 Mdot
.35856 .2848 Mdot
.35856 .31374 Mdot
.35962 .0625 Mdot
.35962 .08267 Mdot
.35962 .1028 Mdot
.35962 .12144 Mdot
.35962 .13316 Mdot
.35962 .16167 Mdot
.35962 .16891 Mdot
.35962 .17888 Mdot
.35962 .19867 Mdot
.35962 .20444 Mdot
.35962 .22505 Mdot
.35962 .22643 Mdot
.35962 .26136 Mdot
.35962 .26801 Mdot
.35962 .27733 Mdot
.35962 .28355 Mdot
.35962 .288 Mdot
.35962 .31467 Mdot
.36067 .15983 Mdot
.36067 .17049 Mdot
.36067 .20927 Mdot
.36067 .21981 Mdot
.36067 .25861 Mdot
.36067 .28708 Mdot
.36067 .31559 Mdot
.36086 .06886 Mdot
.36086 .08158 Mdot
.36086 .09429 Mdot
.36086 .10697 Mdot
.36086 .19034 Mdot
.36086 .23622 Mdot
.36086 .2821 Mdot
.36086 .31575 Mdot
.36114 .13249 Mdot
.36114 .14791 Mdot
.36114 .219 Mdot
.36114 .27495 Mdot
.36114 .31599 Mdot
.36156 .07744 Mdot
.36156 .0874 Mdot
.36156 .09735 Mdot
.36156 .10728 Mdot
.36156 .11719 Mdot
.36156 .18239 Mdot
.36156 .22795 Mdot
.36156 .26378 Mdot
.36156 .29006 Mdot
.36156 .31637 Mdot
.36188 .13793 Mdot
.36188 .14943 Mdot
.36188 .20246 Mdot
.36188 .21379 Mdot
.36188 .25545 Mdot
.36188 .28602 Mdot
.36188 .31664 Mdot
.36232 .16688 Mdot
.36232 .18056 Mdot
.36232 .23045 Mdot
.36232 .2804 Mdot
.36232 .31703 Mdot
.36273 .0731 Mdot
.36273 .0837 Mdot
.36273 .16385 Mdot
.36273 .17435 Mdot
.36273 .22298 Mdot
.36273 .26123 Mdot
.36273 .28929 Mdot
.36273 .31739 Mdot
.36298 .0625 Mdot
.36298 .07973 Mdot
.36298 .09693 Mdot
.36298 .20994 Mdot
.36298 .27205 Mdot
.36298 .31761 Mdot
.36328 .08755 Mdot
.36328 .10006 Mdot
.36328 .11254 Mdot
.36328 .12499 Mdot
.36328 .19457 Mdot
.36328 .2848 Mdot
.36328 .31787 Mdot
.36345 .11165 Mdot
.36345 .12144 Mdot
.36345 .1312 Mdot
.36345 .18581 Mdot
.36345 .19546 Mdot
.36345 .23078 Mdot
.36345 .26613 Mdot
.36345 .29206 Mdot
.36345 .31802 Mdot
.36407 .14257 Mdot
.36407 .15389 Mdot
.36407 .20614 Mdot
.36407 .21729 Mdot
.36407 .22357 Mdot
.36407 .25831 Mdot
.36407 .27837 Mdot
.36407 .28841 Mdot
.36407 .31856 Mdot
.36473 .06773 Mdot
.36473 .0782 Mdot
.36473 .08866 Mdot
.36473 .09911 Mdot
.36473 .10954 Mdot
.36473 .17811 Mdot
.36473 .22607 Mdot
.36473 .26378 Mdot
.36473 .29144 Mdot
.36473 .31913 Mdot
.36492 .18533 Mdot
.36492 .28336 Mdot
.36492 .31931 Mdot
.36529 .12558 Mdot
.36529 .13522 Mdot
.36529 .14483 Mdot
.36529 .18915 Mdot
.36529 .19867 Mdot
.36529 .23354 Mdot
.36529 .26842 Mdot
.36529 .29401 Mdot
.36529 .31962 Mdot
.36561 .11793 Mdot
.36561 .13019 Mdot
.36561 .1424 Mdot
.36561 .19867 Mdot
.36561 .25494 Mdot
.36561 .2874 Mdot
.36561 .31991 Mdot
.36619 .10464 Mdot
.36619 .12144 Mdot
.36619 .15822 Mdot
.36619 .16932 Mdot
.36619 .2152 Mdot
.36619 .22068 Mdot
.36619 .26109 Mdot
.36619 .2759 Mdot
.36619 .29073 Mdot
.36619 .32041 Mdot
.36667 .07284 Mdot
.36667 .08317 Mdot
.36667 .09349 Mdot
.36667 .1038 Mdot
.36667 .1141 Mdot
.36667 .12437 Mdot
.36667 .18177 Mdot
.36667 .19193 Mdot
.36667 .22907 Mdot
.36667 .26625 Mdot
.36667 .29353 Mdot
.36667 .32083 Mdot
.36688 .15964 Mdot
.36688 .17435 Mdot
.36688 .22795 Mdot
.36688 .28165 Mdot
.36688 .32102 Mdot
.36707 .13914 Mdot
.36707 .1924 Mdot
.36707 .23622 Mdot
.36707 .29592 Mdot
.36707 .32119 Mdot
.36743 .0625 Mdot
.36743 .07584 Mdot
.36743 .08916 Mdot
.36743 .10247 Mdot
.36743 .18993 Mdot
.36743 .28621 Mdot
.36743 .3215 Mdot
.36787 .13522 Mdot
.36787 .14723 Mdot
.36787 .20263 Mdot
.36787 .21447 Mdot
.36787 .25798 Mdot
.36787 .28991 Mdot
.36787 .32189 Mdot
.36824 .06801 Mdot
.36824 .16243 Mdot
.36824 .17336 Mdot
.36824 .22397 Mdot
.36824 .26378 Mdot
.36824 .29298 Mdot
.36824 .32221 Mdot
.36855 .10838 Mdot
.36855 .11854 Mdot
.36855 .12867 Mdot
.36855 .18533 Mdot
.36855 .19534 Mdot
.36855 .23199 Mdot
.36855 .26866 Mdot
.36855 .29556 Mdot
.36855 .32248 Mdot
.36882 .20484 Mdot
.36882 .27283 Mdot
.36882 .32271 Mdot
.36924 .12847 Mdot
.36924 .14483 Mdot
.36924 .22023 Mdot
.36924 .27958 Mdot
.36924 .32309 Mdot
.36958 .17959 Mdot
.36958 .2848 Mdot
.36958 .32338 Mdot
.36984 .09529 Mdot
.36984 .10838 Mdot
.36984 .12144 Mdot
.36984 .19438 Mdot
.36984 .25436 Mdot
.36984 .28895 Mdot
.36984 .32361 Mdot
.37005 .15192 Mdot
.37005 .1637 Mdot
.37005 .21811 Mdot
.37005 .26092 Mdot
.37005 .29234 Mdot
.37005 .3238 Mdot
.37023 .0625 Mdot
.37023 .07338 Mdot
.37023 .08426 Mdot
.37023 .09512 Mdot
.37023 .10597 Mdot
.37023 .17729 Mdot
.37023 .22717 Mdot
.37023 .26639 Mdot
.37023 .29516 Mdot
.37023 .32395 Mdot
.37038 .12287 Mdot
.37038 .13287 Mdot
.37038 .14284 Mdot
.37038 .1888 Mdot
.37038 .19867 Mdot
.37038 .23483 Mdot
.37038 .271 Mdot
.37038 .29754 Mdot
.37038 .32409 Mdot
.37217 .06787 Mdot
.37217 .07169 Mdot
.37217 .07861 Mdot
.37217 .08934 Mdot
.37217 .09005 Mdot
.37217 .11076 Mdot
.37217 .1141 Mdot
.37217 .12144 Mdot
.37217 .12693 Mdot
.37217 .13696 Mdot
.37217 .13973 Mdot
.37217 .1468 Mdot
.37217 .15123 Mdot
.37217 .15647 Mdot
.37217 .15661 Mdot
.37217 .1811 Mdot
.37217 .18463 Mdot
.37217 .19166 Mdot
.37217 .20191 Mdot
.37217 .21128 Mdot
.37217 .23027 Mdot
.37217 .2376 Mdot
.37217 .25007 Mdot
.37217 .2576 Mdot
.37217 .26892 Mdot
.37217 .27328 Mdot
.37217 .27702 Mdot
.37217 .28309 Mdot
.37217 .28782 Mdot
.37217 .2916 Mdot
.37217 .2947 Mdot
.37217 .29728 Mdot
.37217 .29946 Mdot
.37217 .32565 Mdot
.3739 .15066 Mdot
.3739 .16034 Mdot
.3739 .20506 Mdot
.3739 .24029 Mdot
.3739 .24973 Mdot
.3739 .2755 Mdot
.3739 .30133 Mdot
.3739 .32716 Mdot
.37405 .10486 Mdot
.37405 .11542 Mdot
.37405 .12595 Mdot
.37405 .13645 Mdot
.37405 .18481 Mdot
.37405 .19521 Mdot
.37405 .23329 Mdot
.37405 .27139 Mdot
.37405 .29934 Mdot
.37405 .32729 Mdot
.37421 .17229 Mdot
.37421 .22505 Mdot
.37421 .26654 Mdot
.37421 .29698 Mdot
.37421 .32744 Mdot
.37441 .13225 Mdot
.37441 .14483 Mdot
.37441 .2152 Mdot
.37441 .26074 Mdot
.37441 .29416 Mdot
.37441 .32761 Mdot
.37466 .06951 Mdot
.37466 .08352 Mdot
.37466 .09751 Mdot
.37466 .18948 Mdot
.37466 .25369 Mdot
.37466 .29073 Mdot
.37466 .32783 Mdot
.37496 .17294 Mdot
.37496 .28646 Mdot
.37496 .32809 Mdot
.37535 .09843 Mdot
.37535 .11633 Mdot
.37535 .13417 Mdot
.37535 .2163 Mdot
.37535 .281 Mdot
.37535 .32843 Mdot
.37559 .07212 Mdot
.37559 .08173 Mdot
.37559 .15442 Mdot
.37559 .16397 Mdot
.37559 .17349 Mdot
.37559 .20814 Mdot
.37559 .21755 Mdot
.37559 .24291 Mdot
.37559 .25223 Mdot
.37559 .27766 Mdot
.37559 .30315 Mdot
.37559 .32864 Mdot
.37587 .13034 Mdot
.37587 .1407 Mdot
.37587 .19867 Mdot
.37587 .23622 Mdot
.37587 .27378 Mdot
.37587 .30133 Mdot
.37587 .32889 Mdot
.3762 .06816 Mdot
.3762 .07949 Mdot
.3762 .09081 Mdot
.3762 .10211 Mdot
.3762 .17639 Mdot
.3762 .18756 Mdot
.3762 .22836 Mdot
.3762 .26922 Mdot
.3762 .29919 Mdot
.3762 .32917 Mdot
.37658 .14978 Mdot
.37658 .16209 Mdot
.37658 .219 Mdot
.37658 .26378 Mdot
.37658 .29663 Mdot
.37658 .32951 Mdot
.37706 .09005 Mdot
.37706 .1038 Mdot
.37706 .11752 Mdot
.37706 .20766 Mdot
.37706 .25718 Mdot
.37706 .29353 Mdot
.37706 .32992 Mdot
.37764 .14483 Mdot
.37764 .15502 Mdot
.37764 .1785 Mdot
.37764 .20203 Mdot
.37764 .23908 Mdot
.37764 .24901 Mdot
.37764 .27611 Mdot
.37764 .28969 Mdot
.37764 .30327 Mdot
.37764 .33044 Mdot
.37812 .07367 Mdot
.37812 .08484 Mdot
.37812 .096 Mdot
.37812 .10714 Mdot
.37812 .11826 Mdot
.37812 .18038 Mdot
.37812 .19137 Mdot
.37812 .23158 Mdot
.37812 .27181 Mdot
.37812 .30133 Mdot
.37812 .33085 Mdot
.37839 .14136 Mdot
.37839 .2848 Mdot
.37839 .33109 Mdot
.37868 .16668 Mdot
.37868 .22267 Mdot
.37868 .26671 Mdot
.37868 .29902 Mdot
.37868 .33135 Mdot
.37936 .12336 Mdot
.37936 .1368 Mdot
.37936 .14886 Mdot
.37936 .15891 Mdot
.37936 .2053 Mdot
.37936 .21191 Mdot
.37936 .2152 Mdot
.37936 .25164 Mdot
.37936 .26054 Mdot
.37936 .27837 Mdot
.37936 .29623 Mdot
.37936 .30516 Mdot
.37936 .33194 Mdot
.37998 .11204 Mdot
.37998 .123 Mdot
.37998 .13393 Mdot
.37998 .19507 Mdot
.37998 .2347 Mdot
.37998 .27433 Mdot
.37998 .30341 Mdot
.37998 .33248 Mdot
.38021 .18384 Mdot
.38021 .25293 Mdot
.38021 .29278 Mdot
.38021 .33268 Mdot
.38071 .17111 Mdot
.38071 .18297 Mdot
.38071 .22622 Mdot
.38071 .26955 Mdot
.38071 .30133 Mdot
.38071 .33312 Mdot
.38104 .06748 Mdot
.38104 .07744 Mdot
.38104 .16269 Mdot
.38104 .17256 Mdot
.38104 .20849 Mdot
.38104 .21825 Mdot
.38104 .2542 Mdot
.38104 .28057 Mdot
.38104 .307 Mdot
.38104 .33341 Mdot
.38127 .14823 Mdot
.38127 .16517 Mdot
.38127 .28841 Mdot
.38127 .33362 Mdot
.38159 .1422 Mdot
.38159 .15535 Mdot
.38159 .21601 Mdot
.38159 .26378 Mdot
.38159 .29883 Mdot
.38159 .33389 Mdot
.38179 .12761 Mdot
.38179 .13839 Mdot
.38179 .14913 Mdot
.38179 .19867 Mdot
.38179 .24821 Mdot
.38179 .27678 Mdot
.38179 .30543 Mdot
.38179 .33406 Mdot
.38267 .0625 Mdot
.38267 .07234 Mdot
.38267 .07431 Mdot
.38267 .07727 Mdot
.38267 .08219 Mdot
.38267 .08612 Mdot
.38267 .09202 Mdot
.38267 .09791 Mdot
.38267 .10184 Mdot
.38267 .10675 Mdot
.38267 .16638 Mdot
.38267 .17542 Mdot
.38267 .17612 Mdot
.38267 .18708 Mdot
.38267 .20349 Mdot
.38267 .22121 Mdot
.38267 .22966 Mdot
.38267 .25669 Mdot
.38267 .27229 Mdot
.38267 .28271 Mdot
.38267 .29574 Mdot
.38267 .30357 Mdot
.38267 .30878 Mdot
.38267 .33483 Mdot
.38354 .14271 Mdot
.38354 .1533 Mdot
.38354 .20216 Mdot
.38354 .2126 Mdot
.38354 .25099 Mdot
.38354 .27915 Mdot
.38354 .30738 Mdot
.38354 .3356 Mdot
.38373 .14742 Mdot
.38373 .16034 Mdot
.38373 .21998 Mdot
.38373 .2669 Mdot
.38373 .30133 Mdot
.38373 .33576 Mdot
.38403 .17136 Mdot
.38403 .24777 Mdot
.38403 .29186 Mdot
.38403 .33603 Mdot
.38425 .06736 Mdot
.38425 .07709 Mdot
.38425 .08681 Mdot
.38425 .09653 Mdot
.38425 .10623 Mdot
.38425 .11591 Mdot
.38425 .17959 Mdot
.38425 .2241 Mdot
.38425 .25911 Mdot
.38425 .2848 Mdot
.38425 .31052 Mdot
.38425 .33622 Mdot
.38457 .06832 Mdot
.38457 .07997 Mdot
.38457 .0916 Mdot
.38457 .10322 Mdot
.38457 .11482 Mdot
.38457 .12639 Mdot
.38457 .19106 Mdot
.38457 .23299 Mdot
.38457 .24427 Mdot
.38457 .27495 Mdot
.38457 .30573 Mdot
.38457 .3365 Mdot
.38503 .09874 Mdot
.38503 .1132 Mdot
.38503 .12761 Mdot
.38503 .20814 Mdot
.38503 .2603 Mdot
.38503 .29859 Mdot
.38503 .3369 Mdot
.38524 .15736 Mdot
.38524 .16776 Mdot
.38524 .20556 Mdot
.38524 .21585 Mdot
.38524 .25369 Mdot
.38524 .28146 Mdot
.38524 .30928 Mdot
.38524 .33709 Mdot
.3858 .10093 Mdot
.3858 .11051 Mdot
.3858 .12007 Mdot
.3858 .12962 Mdot
.3858 .16517 Mdot
.3858 .17779 Mdot
.3858 .18297 Mdot
.3858 .1924 Mdot
.3858 .2238 Mdot
.3858 .22692 Mdot
.3858 .26147 Mdot
.3858 .26992 Mdot
.3858 .28683 Mdot
.3858 .30375 Mdot
.3858 .31221 Mdot
.3858 .33757 Mdot
.3864 .10838 Mdot
.3864 .11981 Mdot
.3864 .1312 Mdot
.3864 .19492 Mdot
.3864 .24731 Mdot
.3864 .27753 Mdot
.3864 .30783 Mdot
.3864 .3381 Mdot
.38666 .17729 Mdot
.38666 .25204 Mdot
.38666 .29516 Mdot
.38666 .33833 Mdot
.3869 .07284 Mdot
.3869 .16132 Mdot
.3869 .17157 Mdot
.3869 .219 Mdot
.3869 .25632 Mdot
.3869 .2837 Mdot
.3869 .31113 Mdot
.3869 .33853 Mdot
.3873 .11941 Mdot
.3873 .13356 Mdot
.3873 .2126 Mdot
.3873 .26378 Mdot
.3873 .30133 Mdot
.3873 .33889 Mdot
.38779 .16983 Mdot
.38779 .18224 Mdot
.38779 .2275 Mdot
.38779 .27283 Mdot
.38779 .30609 Mdot
.38779 .33932 Mdot
.38818 .13588 Mdot
.38818 .14707 Mdot
.38818 .19867 Mdot
.38818 .20971 Mdot
.38818 .25027 Mdot
.38818 .28002 Mdot
.38818 .30986 Mdot
.38818 .33966 Mdot
.3885 .0676 Mdot
.3885 .07781 Mdot
.3885 .08801 Mdot
.3885 .0982 Mdot
.3885 .17527 Mdot
.3885 .22206 Mdot
.3885 .25887 Mdot
.3885 .28588 Mdot
.3885 .31292 Mdot
.3885 .33994 Mdot
.38877 .13741 Mdot
.38877 .29073 Mdot
.38877 .34017 Mdot
.38918 .19867 Mdot
.38918 .25612 Mdot
.38918 .29831 Mdot
.38918 .34053 Mdot
.38948 .1393 Mdot
.38948 .15312 Mdot
.38948 .21691 Mdot
.38948 .26712 Mdot
.38948 .30397 Mdot
.38948 .3408 Mdot
.38972 .06867 Mdot
.38972 .08101 Mdot
.38972 .09334 Mdot
.38972 .10565 Mdot
.38972 .18655 Mdot
.38972 .24306 Mdot
.38972 .27565 Mdot
.38972 .30834 Mdot
.38972 .34101 Mdot
.38991 .14042 Mdot
.38991 .15144 Mdot
.38991 .21317 Mdot
.38991 .25313 Mdot
.38991 .28245 Mdot
.38991 .31183 Mdot
.38991 .34117 Mdot
.39006 .07258 Mdot
.39006 .08267 Mdot
.39006 .09274 Mdot
.39006 .1028 Mdot
.39006 .11284 Mdot
.39006 .12287 Mdot
.39006 .17888 Mdot
.39006 .1888 Mdot
.39006 .22505 Mdot
.39006 .26136 Mdot
.39006 .288 Mdot
.39006 .31467 Mdot
.39006 .34131 Mdot
.39158 .0703 Mdot
.39158 .0859 Mdot
.39158 .08681 Mdot
.39158 .09895 Mdot
.39158 .10147 Mdot
.39158 .10728 Mdot
.39158 .11719 Mdot
.39158 .12316 Mdot
.39158 .12709 Mdot
.39158 .13696 Mdot
.39158 .1557 Mdot
.39158 .15841 Mdot
.39158 .16291 Mdot
.39158 .16651 Mdot
.39158 .18239 Mdot
.39158 .19072 Mdot
.39158 .20377 Mdot
.39158 .21654 Mdot
.39158 .22105 Mdot
.39158 .24631 Mdot
.39158 .25591 Mdot
.39158 .26003 Mdot
.39158 .27034 Mdot
.39158 .27837 Mdot
.39158 .2848 Mdot
.39158 .29006 Mdot
.39158 .29444 Mdot
.39158 .30133 Mdot
.39158 .3065 Mdot
.39158 .31052 Mdot
.39158 .31374 Mdot
.39158 .34264 Mdot
.39306 .12144 Mdot
.39306 .1312 Mdot
.39306 .14095 Mdot
.39306 .19546 Mdot
.39306 .23078 Mdot
.39306 .24029 Mdot
.39306 .26613 Mdot
.39306 .29206 Mdot
.39306 .31802 Mdot
.39306 .34393 Mdot
.39321 .06787 Mdot
.39321 .17049 Mdot
.39321 .21981 Mdot
.39321 .25861 Mdot
.39321 .28708 Mdot
.39321 .31559 Mdot
.39321 .34406 Mdot
.39339 .11633 Mdot
.39339 .12823 Mdot
.39339 .1401 Mdot
.39339 .19475 Mdot
.39339 .20647 Mdot
.39339 .24945 Mdot
.39339 .281 Mdot
.39339 .31263 Mdot
.39339 .34421 Mdot
.39361 .1635 Mdot
.39361 .17676 Mdot
.39361 .27344 Mdot
.39361 .30894 Mdot
.39361 .34441 Mdot
.39389 .09311 Mdot
.39389 .10838 Mdot
.39389 .12361 Mdot
.39389 .20867 Mdot
.39389 .26378 Mdot
.39389 .30422 Mdot
.39389 .34465 Mdot
.39426 .16956 Mdot
.39426 .25099 Mdot
.39426 .29797 Mdot
.39426 .34498 Mdot
.3945 .13522 Mdot
.3945 .14483 Mdot
.3945 .15442 Mdot
.3945 .19867 Mdot
.3945 .23354 Mdot
.3945 .24291 Mdot
.3945 .26842 Mdot
.3945 .29401 Mdot
.3945 .31962 Mdot
.3945 .34519 Mdot
.39478 .0625 Mdot
.39478 .0731 Mdot
.39478 .0837 Mdot
.39478 .09429 Mdot
.39478 .10486 Mdot
.39478 .17435 Mdot
.39478 .18481 Mdot
.39478 .22298 Mdot
.39478 .26123 Mdot
.39478 .28929 Mdot
.39478 .31739 Mdot
.39478 .34544 Mdot
.39513 .13316 Mdot
.39513 .14483 Mdot
.39513 .21019 Mdot
.39513 .2525 Mdot
.39513 .28355 Mdot
.39513 .31467 Mdot
.39513 .34574 Mdot
.39556 .18144 Mdot
.39556 .24169 Mdot
.39556 .27643 Mdot
.39556 .3113 Mdot
.39556 .34611 Mdot
.3961 .12996 Mdot
.3961 .14483 Mdot
.3961 .21337 Mdot
.3961 .26737 Mdot
.3961 .307 Mdot
.3961 .34659 Mdot
.39632 .06773 Mdot
.39632 .0782 Mdot
.39632 .08866 Mdot
.39632 .09911 Mdot
.39632 .10954 Mdot
.39632 .11995 Mdot
.39632 .17811 Mdot
.39632 .18842 Mdot
.39632 .22607 Mdot
.39632 .26378 Mdot
.39632 .29144 Mdot
.39632 .31913 Mdot
.39632 .34678 Mdot
.39682 .14943 Mdot
.39682 .16088 Mdot
.39682 .193 Mdot
.39682 .21379 Mdot
.39682 .25545 Mdot
.39682 .28602 Mdot
.39682 .31664 Mdot
.39682 .34721 Mdot
.39743 .0625 Mdot
.39743 .07542 Mdot
.39743 .08833 Mdot
.39743 .10123 Mdot
.39743 .18597 Mdot
.39743 .24517 Mdot
.39743 .27932 Mdot
.39743 .31357 Mdot
.39743 .34776 Mdot
.3978 .1141 Mdot
.3978 .12437 Mdot
.3978 .13462 Mdot
.3978 .19193 Mdot
.3978 .22907 Mdot
.3978 .23908 Mdot
.3978 .26625 Mdot
.3978 .32083 Mdot
.3978 .34808 Mdot
.39822 .15066 Mdot
.39822 .27082 Mdot
.39822 .34844 Mdot
.39845 .15389 Mdot
.39845 .16517 Mdot
.39845 .21729 Mdot
.39845 .25831 Mdot
.39845 .31856 Mdot
.39845 .34864 Mdot
.39925 .09429 Mdot
.39925 .10697 Mdot
.39925 .11963 Mdot
.39925 .12867 Mdot
.39925 .13878 Mdot
.39925 .14886 Mdot
.39925 .19034 Mdot
.39925 .19534 Mdot
.39925 .19867 Mdot
.39925 .20282 Mdot
.39925 .24185 Mdot
.39925 .24854 Mdot
.39925 .25971 Mdot
.39925 .26866 Mdot
.39925 .2821 Mdot
.39925 .32248 Mdot
.39925 .34934 Mdot
.40003 .07367 Mdot
.40003 .16932 Mdot
.40003 .18038 Mdot
.40003 .22068 Mdot
.40003 .26109 Mdot
.40003 .32041 Mdot
.40003 .35003 Mdot
.40026 .15627 Mdot
.40026 .17049 Mdot
.40026 .23622 Mdot
.40026 .27414 Mdot
.40026 .35022 Mdot
.40065 .14284 Mdot
.40065 .15279 Mdot
.40065 .19867 Mdot
.40065 .20849 Mdot
.40065 .24455 Mdot
.40065 .271 Mdot
.40065 .32409 Mdot
.40065 .35057 Mdot
.401 .12499 Mdot
.401 .13741 Mdot
.401 .20683 Mdot
.401 .25179 Mdot
.401 .35087 Mdot
.40157 .0625 Mdot
.40157 .06801 Mdot
.40157 .07904 Mdot
.40157 .09005 Mdot
.40157 .09556 Mdot
.40157 .10106 Mdot
.40157 .17336 Mdot
.40157 .18425 Mdot
.40157 .20408 Mdot
.40157 .22397 Mdot
.40157 .2347 Mdot
.40157 .26378 Mdot
.40157 .35137 Mdot
.40202 .1468 Mdot
.40202 .15661 Mdot
.40202 .16638 Mdot
.40202 .20191 Mdot
.40202 .21159 Mdot
.40202 .24718 Mdot
.40202 .27328 Mdot
.40202 .32565 Mdot
.40202 .35177 Mdot
.40221 .17562 Mdot
.40221 .24012 Mdot
.40221 .27733 Mdot
.40221 .35193 Mdot
.40269 .1424 Mdot
.40269 .15456 Mdot
.40269 .21071 Mdot
.40269 .25494 Mdot
.40269 .35235 Mdot
.40306 .08426 Mdot
.40306 .09512 Mdot
.40306 .10597 Mdot
.40306 .1168 Mdot
.40306 .12761 Mdot
.40306 .188 Mdot
.40306 .23773 Mdot
.40306 .26639 Mdot
.40306 .35267 Mdot
.40335 .06736 Mdot
.40335 .07709 Mdot
.40335 .08681 Mdot
.40335 .16034 Mdot
.40335 .16998 Mdot
.40335 .21461 Mdot
.40335 .24973 Mdot
.40335 .2755 Mdot
.40335 .31052 Mdot
.40335 .32716 Mdot
.40335 .35293 Mdot
.40378 .103 Mdot
.40378 .11914 Mdot
.40378 .20927 Mdot
.40378 .26767 Mdot
.40378 .35331 Mdot
.40409 .18056 Mdot
.40409 .24388 Mdot
.40409 .35358 Mdot
.40432 .14723 Mdot
.40432 .1592 Mdot
.40432 .21447 Mdot
.40432 .25798 Mdot
.40432 .35378 Mdot
.4045 .11076 Mdot
.4045 .12144 Mdot
.4045 .13209 Mdot
.4045 .19166 Mdot
.4045 .24067 Mdot
.4045 .26892 Mdot
.4045 .35394 Mdot
.40464 .0625 Mdot
.40464 .07212 Mdot
.40464 .08173 Mdot
.40464 .09133 Mdot
.40464 .10093 Mdot
.40464 .11051 Mdot
.40464 .16397 Mdot
.40464 .17349 Mdot
.40464 .21755 Mdot
.40464 .25223 Mdot
.40464 .28683 Mdot
.40464 .31221 Mdot
.40464 .32864 Mdot
.40464 .35406 Mdot
.4059 .06928 Mdot
.4059 .08283 Mdot
.4059 .09637 Mdot
.4059 .10988 Mdot
.4059 .12595 Mdot
.4059 .14169 Mdot
.4059 .14693 Mdot
.4059 .1637 Mdot
.4059 .17542 Mdot
.4059 .18626 Mdot
.4059 .19521 Mdot
.4059 .19867 Mdot
.4059 .20556 Mdot
.4059 .21811 Mdot
.4059 .24353 Mdot
.4059 .2475 Mdot
.4059 .25466 Mdot
.4059 .26092 Mdot
.4059 .27139 Mdot
.4059 .35517 Mdot
.40727 .1407 Mdot
.40727 .15103 Mdot
.40727 .16132 Mdot
.40727 .20886 Mdot
.40727 .24631 Mdot
.40727 .31113 Mdot
.40727 .35636 Mdot
.40743 .06832 Mdot
.40743 .07997 Mdot
.40743 .16806 Mdot
.40743 .17959 Mdot
.40743 .23299 Mdot
.40743 .26378 Mdot
.40743 .3565 Mdot
.40765 .10247 Mdot
.40765 .11576 Mdot
.40765 .129 Mdot
.40765 .20302 Mdot
.40765 .25099 Mdot
.40765 .35669 Mdot
.40793 .14791 Mdot
.40793 .16325 Mdot
.40793 .23408 Mdot
.40793 .35694 Mdot
.40833 .19262 Mdot
.40833 .25933 Mdot
.40833 .35729 Mdot
.40859 .15502 Mdot
.40859 .16517 Mdot
.40859 .21208 Mdot
.40859 .24901 Mdot
.40859 .28588 Mdot
.40859 .31292 Mdot
.40859 .35751 Mdot
.40891 .0625 Mdot
.40891 .07399 Mdot
.40891 .08546 Mdot
.40891 .09693 Mdot
.40891 .10838 Mdot
.40891 .18363 Mdot
.40891 .23622 Mdot
.40891 .26654 Mdot
.40891 .3578 Mdot
.40933 .12144 Mdot
.40933 .13445 Mdot
.40933 .14742 Mdot
.40933 .20722 Mdot
.40933 .25436 Mdot
.40933 .35816 Mdot
.40987 .0625 Mdot
.40987 .07258 Mdot
.40987 .08267 Mdot
.40987 .15891 Mdot
.40987 .16891 Mdot
.40987 .2152 Mdot
.40987 .23832 Mdot
.40987 .25164 Mdot
.40987 .288 Mdot
.40987 .31467 Mdot
.40987 .35864 Mdot
.41035 .09081 Mdot
.41035 .10211 Mdot
.41035 .11339 Mdot
.41035 .12465 Mdot
.41035 .18756 Mdot
.41035 .19867 Mdot
.41035 .23936 Mdot
.41035 .30986 Mdot
.41035 .35905 Mdot
.41063 .19867 Mdot
.41063 .26378 Mdot
.41063 .3593 Mdot
.41095 .13973 Mdot
.41095 .15248 Mdot
.41095 .21128 Mdot
.41095 .2576 Mdot
.41095 .35958 Mdot
.41112 .06748 Mdot
.41112 .07744 Mdot
.41112 .0874 Mdot
.41112 .09735 Mdot
.41112 .10728 Mdot
.41112 .17256 Mdot
.41112 .18239 Mdot
.41112 .21825 Mdot
.41112 .22795 Mdot
.41112 .2542 Mdot
.41112 .29006 Mdot
.41112 .31637 Mdot
.41112 .35973 Mdot
.41174 .11826 Mdot
.41174 .12936 Mdot
.41174 .14042 Mdot
.41174 .17435 Mdot
.41174 .19137 Mdot
.41174 .20231 Mdot
.41174 .2424 Mdot
.41174 .31183 Mdot
.41174 .36027 Mdot
.41234 .08219 Mdot
.41234 .09202 Mdot
.41234 .10184 Mdot
.41234 .11165 Mdot
.41234 .12144 Mdot
.41234 .17612 Mdot
.41234 .18581 Mdot
.41234 .22121 Mdot
.41234 .23078 Mdot
.41234 .25669 Mdot
.41234 .29206 Mdot
.41234 .31802 Mdot
.41234 .3608 Mdot
.41251 .15736 Mdot
.41251 .16983 Mdot
.41251 .2275 Mdot
.41251 .26074 Mdot
.41251 .36095 Mdot
.41282 .0713 Mdot
.41282 .08888 Mdot
.41282 .10643 Mdot
.41282 .36122 Mdot
.41308 .13393 Mdot
.41308 .14483 Mdot
.41308 .20584 Mdot
.41308 .24534 Mdot
.41308 .2848 Mdot
.41308 .31374 Mdot
.41308 .36145 Mdot
.41353 .06975 Mdot
.41353 .11591 Mdot
.41353 .12558 Mdot
.41353 .13522 Mdot
.41353 .18915 Mdot
.41353 .19391 Mdot
.41353 .23354 Mdot
.41353 .24631 Mdot
.41353 .25911 Mdot
.41353 .26842 Mdot
.41353 .29401 Mdot
.41353 .31962 Mdot
.41353 .36183 Mdot
.41402 .16209 Mdot
.41402 .17435 Mdot
.41402 .23108 Mdot
.41402 .36227 Mdot
.41439 .14913 Mdot
.41439 .15983 Mdot
.41439 .20927 Mdot
.41439 .24821 Mdot
.41439 .28708 Mdot
.41439 .31559 Mdot
.41439 .36259 Mdot
.41468 .12962 Mdot
.41468 .13914 Mdot
.41468 .14863 Mdot
.41468 .1924 Mdot
.41468 .23622 Mdot
.41468 .29592 Mdot
.41468 .32119 Mdot
.41468 .36284 Mdot
.4149 .1141 Mdot
.4149 .1312 Mdot
.4149 .22672 Mdot
.4149 .36304 Mdot
.41524 .0625 Mdot
.41524 .07676 Mdot
.41524 .091 Mdot
.41524 .10522 Mdot
.41524 .19867 Mdot
.41524 .25007 Mdot
.41524 .36334 Mdot
.41548 .0625 Mdot
.41548 .07466 Mdot
.41548 .08681 Mdot
.41548 .17873 Mdot
.41548 .23454 Mdot
.41548 .31052 Mdot
.41548 .36355 Mdot
.41566 .1533 Mdot
.41566 .16385 Mdot
.41566 .2126 Mdot
.41566 .25099 Mdot
.41566 .28929 Mdot
.41566 .31739 Mdot
.41566 .3637 Mdot
.41689 .06773 Mdot
.41689 .06849 Mdot
.41689 .0782 Mdot
.41689 .08047 Mdot
.41689 .08866 Mdot
.41689 .09245 Mdot
.41689 .1044 Mdot
.41689 .11148 Mdot
.41689 .11633 Mdot
.41689 .12541 Mdot
.41689 .13817 Mdot
.41689 .15481 Mdot
.41689 .16776 Mdot
.41689 .17811 Mdot
.41689 .18297 Mdot
.41689 .19475 Mdot
.41689 .20324 Mdot
.41689 .21585 Mdot
.41689 .22607 Mdot
.41689 .23158 Mdot
.41689 .23788 Mdot
.41689 .25369 Mdot
.41689 .281 Mdot
.41689 .29144 Mdot
.41689 .31263 Mdot
.41689 .31913 Mdot
.41689 .36478 Mdot
.41809 .0625 Mdot
.41809 .07284 Mdot
.41809 .08317 Mdot
.41809 .09349 Mdot
.41809 .1038 Mdot
.41809 .1141 Mdot
.41809 .17157 Mdot
.41809 .18177 Mdot
.41809 .22907 Mdot
.41809 .26625 Mdot
.41809 .29353 Mdot
.41809 .32083 Mdot
.41809 .36583 Mdot
.41826 .09791 Mdot
.41826 .10969 Mdot
.41826 .12144 Mdot
.41826 .13316 Mdot
.41826 .19867 Mdot
.41826 .24113 Mdot
.41826 .28355 Mdot
.41826 .31467 Mdot
.41826 .36597 Mdot
.41848 .1312 Mdot
.41848 .14483 Mdot
.41848 .20766 Mdot
.41848 .36617 Mdot
.41879 .16111 Mdot
.41879 .23622 Mdot
.41879 .36644 Mdot
.41925 .0982 Mdot
.41925 .10838 Mdot
.41925 .11854 Mdot
.41925 .12867 Mdot
.41925 .18533 Mdot
.41925 .23199 Mdot
.41925 .26866 Mdot
.41925 .29556 Mdot
.41925 .32248 Mdot
.41925 .36684 Mdot
.41958 .12639 Mdot
.41958 .13793 Mdot
.41958 .20246 Mdot
.41958 .24427 Mdot
.41958 .28602 Mdot
.41958 .31664 Mdot
.41958 .36713 Mdot
.42001 .15018 Mdot
.42001 .1635 Mdot
.42001 .22505 Mdot
.42001 .30894 Mdot
.42001 .36751 Mdot
.42038 .12287 Mdot
.42038 .13287 Mdot
.42038 .14284 Mdot
.42038 .1888 Mdot
.42038 .19867 Mdot
.42038 .23483 Mdot
.42038 .271 Mdot
.42038 .29754 Mdot
.42038 .32409 Mdot
.42038 .36783 Mdot
.4206 .16715 Mdot
.4206 .18297 Mdot
.4206 .24067 Mdot
.4206 .36803 Mdot
.42086 .14257 Mdot
.42086 .15389 Mdot
.42086 .20614 Mdot
.42086 .24731 Mdot
.42086 .28841 Mdot
.42086 .31856 Mdot
.42086 .36825 Mdot
.42148 .13696 Mdot
.42148 .1468 Mdot
.42148 .15535 Mdot
.42148 .16843 Mdot
.42148 .19217 Mdot
.42148 .20191 Mdot
.42148 .22891 Mdot
.42148 .2376 Mdot
.42148 .27328 Mdot
.42148 .29946 Mdot
.42148 .3113 Mdot
.42148 .32565 Mdot
.42148 .3688 Mdot
.4221 .14707 Mdot
.4221 .15822 Mdot
.4221 .20971 Mdot
.4221 .22068 Mdot
.4221 .25027 Mdot
.4221 .29073 Mdot
.4221 .32041 Mdot
.4221 .36933 Mdot
.42234 .18842 Mdot
.42234 .24493 Mdot
.42234 .36955 Mdot
.42255 .14095 Mdot
.42255 .15066 Mdot
.42255 .16034 Mdot
.42255 .20506 Mdot
.42255 .24029 Mdot
.42255 .2755 Mdot
.42255 .30133 Mdot
.42255 .32716 Mdot
.42255 .36973 Mdot
.4229 .17319 Mdot
.4229 .23263 Mdot
.4229 .27932 Mdot
.4229 .31357 Mdot
.4229 .37004 Mdot
.4233 .16243 Mdot
.4233 .17336 Mdot
.4233 .22397 Mdot
.4233 .26378 Mdot
.4233 .29298 Mdot
.4233 .32221 Mdot
.4233 .37038 Mdot
.42359 .15442 Mdot
.42359 .16397 Mdot
.42359 .20814 Mdot
.42359 .24291 Mdot
.42359 .27766 Mdot
.42359 .30315 Mdot
.42359 .32864 Mdot
.42359 .37064 Mdot
.424 .0625 Mdot
.424 .07781 Mdot
.424 .19365 Mdot
.424 .371 Mdot
.42427 .06886 Mdot
.42427 .08158 Mdot
.42427 .09429 Mdot
.42427 .17779 Mdot
.42427 .19034 Mdot
.42427 .23622 Mdot
.42427 .2821 Mdot
.42427 .31575 Mdot
.42427 .37123 Mdot
.42446 .0625 Mdot
.42446 .07338 Mdot
.42446 .08426 Mdot
.42446 .09512 Mdot
.42446 .17729 Mdot
.42446 .22717 Mdot
.42446 .26639 Mdot
.42446 .29516 Mdot
.42446 .32395 Mdot
.42446 .3714 Mdot
.42559 .0625 Mdot
.42559 .06787 Mdot
.42559 .07002 Mdot
.42559 .07503 Mdot
.42559 .07861 Mdot
.42559 .08129 Mdot
.42559 .08505 Mdot
.42559 .08755 Mdot
.42559 .08934 Mdot
.42559 .10006 Mdot
.42559 .11076 Mdot
.42559 .11254 Mdot
.42559 .11503 Mdot
.42559 .11877 Mdot
.42559 .1811 Mdot
.42559 .19457 Mdot
.42559 .19867 Mdot
.42559 .22322 Mdot
.42559 .23027 Mdot
.42559 .2397 Mdot
.42559 .2848 Mdot
.42559 .29728 Mdot
.42559 .307 Mdot
.42559 .31787 Mdot
.42559 .37239 Mdot
.42669 .10486 Mdot
.42669 .11542 Mdot
.42669 .12595 Mdot
.42669 .18481 Mdot
.42669 .19521 Mdot
.42669 .23329 Mdot
.42669 .27139 Mdot
.42669 .29934 Mdot
.42669 .32729 Mdot
.42669 .37335 Mdot
.42687 .10565 Mdot
.42687 .11793 Mdot
.42687 .13019 Mdot
.42687 .19867 Mdot
.42687 .24306 Mdot
.42687 .2874 Mdot
.42687 .31991 Mdot
.42687 .37351 Mdot
.42712 .12144 Mdot
.42712 .13608 Mdot
.42712 .2179 Mdot
.42712 .30966 Mdot
.42712 .37373 Mdot
.42749 .12665 Mdot
.42749 .14483 Mdot
.42749 .22861 Mdot
.42749 .37405 Mdot
.42775 .11995 Mdot
.42775 .13034 Mdot
.42775 .1407 Mdot
.42775 .19867 Mdot
.42775 .23622 Mdot
.42775 .27378 Mdot
.42775 .30133 Mdot
.42775 .32889 Mdot
.42775 .37428 Mdot
.4281 .13522 Mdot
.4281 .14723 Mdot
.4281 .20263 Mdot
.4281 .28991 Mdot
.4281 .32189 Mdot
.4281 .37459 Mdot
.42858 .14197 Mdot
.42858 .22224 Mdot
.42858 .31221 Mdot
.42858 .37501 Mdot
.42879 .13462 Mdot
.42879 .14483 Mdot
.42879 .15502 Mdot
.42879 .20203 Mdot
.42879 .23908 Mdot
.42879 .27611 Mdot
.42879 .30327 Mdot
.42879 .33044 Mdot
.42879 .37519 Mdot
.42929 .1401 Mdot
.42929 .15192 Mdot
.42929 .21811 Mdot
.42929 .23375 Mdot
.42929 .26092 Mdot
.42929 .29234 Mdot
.42929 .3238 Mdot
.42929 .37563 Mdot
.42979 .14886 Mdot
.42979 .15891 Mdot
.42979 .2053 Mdot
.42979 .2152 Mdot
.42979 .25164 Mdot
.42979 .27837 Mdot
.42979 .30516 Mdot
.42979 .33194 Mdot
.42979 .37607 Mdot
.42999 .14765 Mdot
.42999 .16167 Mdot
.42999 .22643 Mdot
.42999 .27733 Mdot
.42999 .31467 Mdot
.42999 .37624 Mdot
.43044 .15647 Mdot
.43044 .16806 Mdot
.43044 .22163 Mdot
.43044 .26378 Mdot
.43044 .2947 Mdot
.43044 .32565 Mdot
.43044 .37664 Mdot
.43077 .06748 Mdot
.43077 .07744 Mdot
.43077 .16269 Mdot
.43077 .17256 Mdot
.43077 .20849 Mdot
.43077 .21825 Mdot
.43077 .2542 Mdot
.43077 .28057 Mdot
.43077 .307 Mdot
.43077 .33341 Mdot
.43077 .37692 Mdot
.43101 .15869 Mdot
.43101 .17591 Mdot
.43101 .23864 Mdot
.43101 .37713 Mdot
.43134 .16688 Mdot
.43134 .23045 Mdot
.43134 .2804 Mdot
.43134 .31703 Mdot
.43134 .37742 Mdot
.43156 .17229 Mdot
.43156 .22505 Mdot
.43156 .26654 Mdot
.43156 .29698 Mdot
.43156 .32744 Mdot
.43156 .37762 Mdot
.43172 .0625 Mdot
.43172 .07234 Mdot
.43172 .08219 Mdot
.43172 .09202 Mdot
.43172 .10184 Mdot
.43172 .16638 Mdot
.43172 .17612 Mdot
.43172 .22121 Mdot
.43172 .25669 Mdot
.43172 .28271 Mdot
.43172 .30878 Mdot
.43172 .33483 Mdot
.43172 .37775 Mdot
.43264 .06736 Mdot
.43264 .06816 Mdot
.43264 .07709 Mdot
.43264 .07949 Mdot
.43264 .09081 Mdot
.43264 .09653 Mdot
.43264 .10211 Mdot
.43264 .10623 Mdot
.43264 .11591 Mdot
.43264 .17191 Mdot
.43264 .17639 Mdot
.43264 .17959 Mdot
.43264 .18198 Mdot
.43264 .18533 Mdot
.43264 .18756 Mdot
.43264 .2241 Mdot
.43264 .22836 Mdot
.43264 .23434 Mdot
.43264 .25911 Mdot
.43264 .26922 Mdot
.43264 .28336 Mdot
.43264 .29919 Mdot
.43264 .30454 Mdot
.43264 .31931 Mdot
.43264 .32917 Mdot
.43264 .33622 Mdot
.43264 .37856 Mdot
.43354 .11051 Mdot
.43354 .12007 Mdot
.43354 .12962 Mdot
.43354 .18297 Mdot
.43354 .1924 Mdot
.43354 .22692 Mdot
.43354 .26147 Mdot
.43354 .28683 Mdot
.43354 .31221 Mdot
.43354 .33757 Mdot
.43354 .37935 Mdot
.43369 .07367 Mdot
.43369 .08484 Mdot
.43369 .096 Mdot
.43369 .10714 Mdot
.43369 .11826 Mdot
.43369 .18038 Mdot
.43369 .19137 Mdot
.43369 .23158 Mdot
.43369 .27181 Mdot
.43369 .30133 Mdot
.43369 .33085 Mdot
.43369 .37948 Mdot
.4339 .0625 Mdot
.4339 .07584 Mdot
.4339 .08916 Mdot
.4339 .10247 Mdot
.4339 .18993 Mdot
.4339 .28621 Mdot
.4339 .3215 Mdot
.4339 .37966 Mdot
.4342 .18779 Mdot
.4342 .3076 Mdot
.4342 .37993 Mdot
.43471 .11204 Mdot
.43471 .123 Mdot
.43471 .13393 Mdot
.43471 .19507 Mdot
.43471 .2347 Mdot
.43471 .24534 Mdot
.43471 .27433 Mdot
.43471 .30341 Mdot
.43471 .33248 Mdot
.43471 .38037 Mdot
.4351 .08219 Mdot
.4351 .09529 Mdot
.4351 .10838 Mdot
.4351 .12144 Mdot
.4351 .19438 Mdot
.4351 .28895 Mdot
.4351 .32361 Mdot
.4351 .38072 Mdot
.43569 .07061 Mdot
.43569 .08681 Mdot
.43569 .12761 Mdot
.43569 .13839 Mdot
.43569 .14913 Mdot
.43569 .19867 Mdot
.43569 .24821 Mdot
.43569 .27678 Mdot
.43569 .30543 Mdot
.43569 .31052 Mdot
.43569 .33406 Mdot
.43569 .38123 Mdot
.43627 .1141 Mdot
.43627 .12693 Mdot
.43627 .13973 Mdot
.43627 .19867 Mdot
.43627 .21128 Mdot
.43627 .2576 Mdot
.43627 .2916 Mdot
.43627 .32565 Mdot
.43627 .38174 Mdot
.43665 .14271 Mdot
.43665 .1533 Mdot
.43665 .20216 Mdot
.43665 .2126 Mdot
.43665 .25099 Mdot
.43665 .27915 Mdot
.43665 .30738 Mdot
.43665 .3356 Mdot
.43665 .38207 Mdot
.43712 .0784 Mdot
.43712 .09429 Mdot
.43712 .11014 Mdot
.43712 .12595 Mdot
.43712 .21424 Mdot
.43712 .27139 Mdot
.43712 .31332 Mdot
.43712 .38248 Mdot
.43739 .13225 Mdot
.43739 .14483 Mdot
.43739 .2152 Mdot
.43739 .26074 Mdot
.43739 .29416 Mdot
.43739 .32761 Mdot
.43739 .38272 Mdot
.43758 .15736 Mdot
.43758 .16776 Mdot
.43758 .21585 Mdot
.43758 .25369 Mdot
.43758 .28146 Mdot
.43758 .30928 Mdot
.43758 .33709 Mdot
.43758 .38288 Mdot
.43848 .0625 Mdot
.43848 .07284 Mdot
.43848 .08317 Mdot
.43848 .13249 Mdot
.43848 .14978 Mdot
.43848 .16132 Mdot
.43848 .16209 Mdot
.43848 .17157 Mdot
.43848 .219 Mdot
.43848 .25632 Mdot
.43848 .26378 Mdot
.43848 .27495 Mdot
.43848 .2837 Mdot
.43848 .29663 Mdot
.43848 .31113 Mdot
.43848 .31599 Mdot
.43848 .32951 Mdot
.43848 .33853 Mdot
.43848 .38367 Mdot
.43935 .0676 Mdot
.43935 .07781 Mdot
.43935 .08801 Mdot
.43935 .0982 Mdot
.43935 .10838 Mdot
.43935 .17527 Mdot
.43935 .18533 Mdot
.43935 .22206 Mdot
.43935 .25887 Mdot
.43935 .28588 Mdot
.43935 .31292 Mdot
.43935 .33994 Mdot
.43935 .38443 Mdot
.43952 .16668 Mdot
.43952 .22267 Mdot
.43952 .26671 Mdot
.43952 .29902 Mdot
.43952 .33135 Mdot
.43952 .38458 Mdot
.43978 .15389 Mdot
.43978 .22357 Mdot
.43978 .27837 Mdot
.43978 .31856 Mdot
.43978 .38481 Mdot
.4402 .09274 Mdot
.4402 .1028 Mdot
.4402 .11284 Mdot
.4402 .12287 Mdot
.4402 .17888 Mdot
.4402 .1888 Mdot
.4402 .22505 Mdot
.4402 .26136 Mdot
.4402 .34131 Mdot
.4402 .38518 Mdot
.44054 .06849 Mdot
.44054 .08047 Mdot
.44054 .17111 Mdot
.44054 .18297 Mdot
.44054 .22622 Mdot
.44054 .26955 Mdot
.44054 .30133 Mdot
.44054 .33312 Mdot
.44054 .38547 Mdot
.44103 .11719 Mdot
.44103 .12709 Mdot
.44103 .13696 Mdot
.44103 .15964 Mdot
.44103 .17435 Mdot
.44103 .19217 Mdot
.44103 .22795 Mdot
.44103 .2376 Mdot
.44103 .26378 Mdot
.44103 .28165 Mdot
.44103 .32102 Mdot
.44103 .34264 Mdot
.44103 .3859 Mdot
.44152 .0625 Mdot
.44152 .07431 Mdot
.44152 .08612 Mdot
.44152 .09791 Mdot
.44152 .10969 Mdot
.44152 .18708 Mdot
.44152 .24113 Mdot
.44152 .27229 Mdot
.44152 .30357 Mdot
.44152 .33483 Mdot
.44152 .38633 Mdot
.44184 .1312 Mdot
.44184 .14095 Mdot
.44184 .15066 Mdot
.44184 .19546 Mdot
.44184 .20506 Mdot
.44184 .24029 Mdot
.44184 .26613 Mdot
.44184 .30133 Mdot
.44184 .34393 Mdot
.44184 .36973 Mdot
.44184 .38661 Mdot
.44223 .17959 Mdot
.44223 .2848 Mdot
.44223 .32338 Mdot
.44223 .38695 Mdot
.44246 .0916 Mdot
.44246 .10322 Mdot
.44246 .11482 Mdot
.44246 .12639 Mdot
.44246 .19106 Mdot
.44246 .24427 Mdot
.44246 .27495 Mdot
.44246 .30573 Mdot
.44246 .3365 Mdot
.44246 .38716 Mdot
.44262 .14483 Mdot
.44262 .15442 Mdot
.44262 .19867 Mdot
.44262 .20814 Mdot
.44262 .24291 Mdot
.44262 .26842 Mdot
.44262 .30315 Mdot
.44262 .34519 Mdot
.44262 .37064 Mdot
.44262 .38729 Mdot
.44338 .11981 Mdot
.44338 .1312 Mdot
.44338 .14257 Mdot
.44338 .16752 Mdot
.44338 .18463 Mdot
.44338 .19492 Mdot
.44338 .20614 Mdot
.44338 .24731 Mdot
.44338 .25007 Mdot
.44338 .27753 Mdot
.44338 .28782 Mdot
.44338 .30493 Mdot
.44338 .32565 Mdot
.44338 .3381 Mdot
.44338 .38796 Mdot
.44427 .13588 Mdot
.44427 .14707 Mdot
.44427 .20971 Mdot
.44427 .25027 Mdot
.44427 .28002 Mdot
.44427 .33966 Mdot
.44427 .38874 Mdot
.44449 .06951 Mdot
.44449 .08352 Mdot
.44449 .09751 Mdot
.44449 .11148 Mdot
.44449 .18948 Mdot
.44449 .20324 Mdot
.44449 .25369 Mdot
.44449 .29073 Mdot
.44449 .32783 Mdot
.44449 .38893 Mdot
.44485 .17435 Mdot
.44485 .30834 Mdot
.44485 .38924 Mdot
.44513 .15144 Mdot
.44513 .16243 Mdot
.44513 .21317 Mdot
.44513 .25313 Mdot
.44513 .34117 Mdot
.44513 .38949 Mdot
.44555 .1038 Mdot
.44555 .11752 Mdot
.44555 .1312 Mdot
.44555 .20766 Mdot
.44555 .25718 Mdot
.44555 .29353 Mdot
.44555 .32992 Mdot
.44555 .38986 Mdot
.44597 .16651 Mdot
.44597 .21654 Mdot
.44597 .25591 Mdot
.44597 .34264 Mdot
.44597 .39022 Mdot
.44624 .18087 Mdot
.44624 .26378 Mdot
.44624 .31158 Mdot
.44624 .39046 Mdot
.44658 .12336 Mdot
.44658 .1368 Mdot
.44658 .21191 Mdot
.44658 .26054 Mdot
.44658 .29623 Mdot
.44658 .33194 Mdot
.44658 .39076 Mdot
.44678 .06787 Mdot
.44678 .07861 Mdot
.44678 .08934 Mdot
.44678 .17049 Mdot
.44678 .1811 Mdot
.44678 .21981 Mdot
.44678 .25861 Mdot
.44678 .34406 Mdot
.44678 .39093 Mdot
.44756 .0625 Mdot
.44756 .0731 Mdot
.44756 .0837 Mdot
.44756 .09429 Mdot
.44756 .10486 Mdot
.44756 .11542 Mdot
.44756 .1422 Mdot
.44756 .15535 Mdot
.44756 .17435 Mdot
.44756 .18481 Mdot
.44756 .20444 Mdot
.44756 .21601 Mdot
.44756 .22298 Mdot
.44756 .23329 Mdot
.44756 .26123 Mdot
.44756 .26378 Mdot
.44756 .26801 Mdot
.44756 .29883 Mdot
.44756 .29934 Mdot
.44756 .31467 Mdot
.44756 .33389 Mdot
.44756 .34544 Mdot
.44756 .37335 Mdot
.44756 .39162 Mdot
.44833 .09911 Mdot
.44833 .10954 Mdot
.44833 .11995 Mdot
.44833 .13034 Mdot
.44833 .18842 Mdot
.44833 .23622 Mdot
.44833 .26378 Mdot
.44833 .30133 Mdot
.44833 .34678 Mdot
.44833 .37428 Mdot
.44833 .39229 Mdot
.44852 .16034 Mdot
.44852 .17319 Mdot
.44852 .21998 Mdot
.44852 .2669 Mdot
.44852 .33576 Mdot
.44852 .39245 Mdot
.44882 .0625 Mdot
.44882 .07973 Mdot
.44882 .09693 Mdot
.44882 .20994 Mdot
.44882 .27205 Mdot
.44882 .31761 Mdot
.44882 .39272 Mdot
.44907 .12437 Mdot
.44907 .13462 Mdot
.44907 .14483 Mdot
.44907 .19193 Mdot
.44907 .20203 Mdot
.44907 .23908 Mdot
.44907 .27611 Mdot
.44907 .30327 Mdot
.44907 .34808 Mdot
.44907 .37519 Mdot
.44907 .39294 Mdot
.44943 .16517 Mdot
.44943 .17779 Mdot
.44943 .23622 Mdot
.44943 .26992 Mdot
.44943 .33757 Mdot
.44943 .39325 Mdot
.44979 .13878 Mdot
.44979 .14886 Mdot
.44979 .2053 Mdot
.44979 .24185 Mdot
.44979 .27837 Mdot
.44979 .30516 Mdot
.44979 .33194 Mdot
.44979 .34934 Mdot
.44979 .37607 Mdot
.44979 .39357 Mdot
.45003 .10464 Mdot
.45003 .12144 Mdot
.45003 .2152 Mdot
.45003 .2759 Mdot
.45003 .32041 Mdot
.45003 .39377 Mdot
.45032 .0625 Mdot
.45032 .07503 Mdot
.45032 .08755 Mdot
.45032 .18224 Mdot
.45032 .2397 Mdot
.45032 .27283 Mdot
.45032 .33932 Mdot
.45032 .39403 Mdot
.45049 .15279 Mdot
.45049 .16269 Mdot
.45049 .20849 Mdot
.45049 .24455 Mdot
.45049 .28057 Mdot
.45049 .307 Mdot
.45049 .33341 Mdot
.45049 .35057 Mdot
.45049 .37692 Mdot
.45049 .39418 Mdot
.45117 .0625 Mdot
.45117 .06867 Mdot
.45117 .07234 Mdot
.45117 .08101 Mdot
.45117 .08219 Mdot
.45117 .09334 Mdot
.45117 .10565 Mdot
.45117 .11793 Mdot
.45117 .14483 Mdot
.45117 .15661 Mdot
.45117 .16638 Mdot
.45117 .18655 Mdot
.45117 .19867 Mdot
.45117 .21159 Mdot
.45117 .24306 Mdot
.45117 .24718 Mdot
.45117 .27958 Mdot
.45117 .28271 Mdot
.45117 .30878 Mdot
.45117 .32309 Mdot
.45117 .33483 Mdot
.45117 .34101 Mdot
.45117 .35177 Mdot
.45117 .37775 Mdot
.45117 .39478 Mdot
.45183 .06736 Mdot
.45183 .07709 Mdot
.45183 .08681 Mdot
.45183 .09653 Mdot
.45183 .10623 Mdot
.45183 .16998 Mdot
.45183 .17959 Mdot
.45183 .21461 Mdot
.45183 .2241 Mdot
.45183 .24973 Mdot
.45183 .2848 Mdot
.45183 .31052 Mdot
.45183 .33622 Mdot
.45183 .35293 Mdot
.45183 .37856 Mdot
.45183 .39536 Mdot
.452 .11107 Mdot
.452 .12316 Mdot
.452 .13522 Mdot
.452 .20263 Mdot
.452 .24631 Mdot
.452 .34264 Mdot
.452 .3955 Mdot
.45227 .15123 Mdot
.45227 .16715 Mdot
.45227 .28309 Mdot
.45227 .32565 Mdot
.45227 .39573 Mdot
.45248 .07212 Mdot
.45248 .08173 Mdot
.45248 .09133 Mdot
.45248 .10093 Mdot
.45248 .11051 Mdot
.45248 .12007 Mdot
.45248 .17349 Mdot
.45248 .18297 Mdot
.45248 .21755 Mdot
.45248 .22692 Mdot
.45248 .25223 Mdot
.45248 .26147 Mdot
.45248 .28683 Mdot
.45248 .31221 Mdot
.45248 .33757 Mdot
.45248 .35406 Mdot
.45248 .37935 Mdot
.45248 .39592 Mdot
.4528 .12823 Mdot
.4528 .1401 Mdot
.4528 .20647 Mdot
.4528 .24945 Mdot
.4528 .34421 Mdot
.4528 .3962 Mdot
.45331 .17294 Mdot
.45331 .24493 Mdot
.45331 .28646 Mdot
.45331 .32809 Mdot
.45331 .39665 Mdot
.45357 .14483 Mdot
.45357 .15647 Mdot
.45357 .21019 Mdot
.45357 .2525 Mdot
.45357 .2947 Mdot
.45357 .34574 Mdot
.45357 .39687 Mdot
.45432 .16088 Mdot
.45432 .17229 Mdot
.45432 .1785 Mdot
.45432 .21379 Mdot
.45432 .22505 Mdot
.45432 .24901 Mdot
.45432 .25545 Mdot
.45432 .28969 Mdot
.45432 .29698 Mdot
.45432 .33044 Mdot
.45432 .34721 Mdot
.45432 .37762 Mdot
.45432 .39753 Mdot
.45504 .16517 Mdot
.45504 .17639 Mdot
.45504 .22836 Mdot
.45504 .29919 Mdot
.45504 .34864 Mdot
.45504 .37856 Mdot
.45504 .39816 Mdot
.45528 .07002 Mdot
.45528 .08505 Mdot
.45528 .19867 Mdot
.45528 .25293 Mdot
.45528 .33268 Mdot
.45528 .39837 Mdot
.45574 .0625 Mdot
.45574 .07367 Mdot
.45574 .08484 Mdot
.45574 .096 Mdot
.45574 .10714 Mdot
.45574 .18038 Mdot
.45574 .23158 Mdot
.45574 .27181 Mdot
.45574 .30133 Mdot
.45574 .33085 Mdot
.45574 .35003 Mdot
.45574 .37948 Mdot
.45574 .39877 Mdot
.4562 .0625 Mdot
.4562 .07727 Mdot
.4562 .09202 Mdot
.4562 .10675 Mdot
.4562 .12144 Mdot
.4562 .20349 Mdot
.4562 .25669 Mdot
.4562 .33483 Mdot
.4562 .39917 Mdot
.45642 .07904 Mdot
.45642 .09005 Mdot
.45642 .10106 Mdot
.45642 .11204 Mdot
.45642 .123 Mdot
.45642 .18425 Mdot
.45642 .19507 Mdot
.45642 .2347 Mdot
.45642 .27433 Mdot
.45642 .30341 Mdot
.45642 .33248 Mdot
.45642 .35137 Mdot
.45642 .38037 Mdot
.45642 .39937 Mdot
.45708 .1132 Mdot
.45708 .1168 Mdot
.45708 .12761 Mdot
.45708 .13839 Mdot
.45708 .188 Mdot
.45708 .19867 Mdot
.45708 .20814 Mdot
.45708 .23773 Mdot
.45708 .2603 Mdot
.45708 .27678 Mdot
.45708 .30543 Mdot
.45708 .33406 Mdot
.45708 .3369 Mdot
.45708 .35267 Mdot
.45708 .38123 Mdot
.45708 .39994 Mdot
.45772 .13209 Mdot
.45772 .14271 Mdot
.45772 .1533 Mdot
.45772 .20216 Mdot
.45772 .24067 Mdot
.45772 .27915 Mdot
.45772 .30738 Mdot
.45772 .3356 Mdot
.45772 .38207 Mdot
.45772 .4005 Mdot
.45793 .13356 Mdot
.45793 .14765 Mdot
.45793 .2126 Mdot
.45793 .26378 Mdot
.45793 .33889 Mdot
.45793 .40069 Mdot
.45834 .14693 Mdot
.45834 .15736 Mdot
.45834 .20556 Mdot
.45834 .24353 Mdot
.45834 .28146 Mdot
.45834 .30928 Mdot
.45834 .33709 Mdot
.45834 .38288 Mdot
.45834 .40105 Mdot
.45874 .15312 Mdot
.45874 .16688 Mdot
.45874 .23045 Mdot
.45874 .3408 Mdot
.45874 .4014 Mdot
.45894 .16132 Mdot
.45894 .17157 Mdot
.45894 .20886 Mdot
.45894 .219 Mdot
.45894 .25632 Mdot
.45894 .2837 Mdot
.45894 .31113 Mdot
.45894 .33853 Mdot
.45894 .38367 Mdot
.45894 .40157 Mdot
.45953 .0676 Mdot
.45953 .07781 Mdot
.45953 .08801 Mdot
.45953 .16517 Mdot
.45953 .17191 Mdot
.45953 .17527 Mdot
.45953 .22206 Mdot
.45953 .23434 Mdot
.45953 .25887 Mdot
.45953 .28588 Mdot
.45953 .31292 Mdot
.45953 .33994 Mdot
.45953 .34264 Mdot
.45953 .38443 Mdot
.45953 .40209 Mdot
.4601 .0625 Mdot
.4601 .07258 Mdot
.4601 .08267 Mdot
.4601 .09274 Mdot
.4601 .1028 Mdot
.4601 .11284 Mdot
.4601 .17888 Mdot
.4601 .22505 Mdot
.4601 .26136 Mdot
.4601 .288 Mdot
.4601 .31467 Mdot
.4601 .34131 Mdot
.4601 .38518 Mdot
.4601 .40258 Mdot
.46028 .17676 Mdot
.46028 .23808 Mdot
.46028 .40275 Mdot
.46065 .09735 Mdot
.46065 .10728 Mdot
.46065 .11719 Mdot
.46065 .12709 Mdot
.46065 .18239 Mdot
.46065 .19217 Mdot
.46065 .22795 Mdot
.46065 .26378 Mdot
.46065 .29006 Mdot
.46065 .31637 Mdot
.46065 .34264 Mdot
.46065 .3859 Mdot
.46065 .40307 Mdot
.46101 .06906 Mdot
.46101 .08219 Mdot
.46101 .09529 Mdot
.46101 .19438 Mdot
.46101 .24169 Mdot
.46101 .40338 Mdot
.46119 .12144 Mdot
.46119 .1312 Mdot
.46119 .14095 Mdot
.46119 .18581 Mdot
.46119 .19546 Mdot
.46119 .23078 Mdot
.46119 .26613 Mdot
.46119 .29206 Mdot
.46119 .31802 Mdot
.46119 .34393 Mdot
.46119 .38661 Mdot
.46119 .40354 Mdot
.46171 .08833 Mdot
.46171 .10123 Mdot
.46171 .1141 Mdot
.46171 .12693 Mdot
.46171 .13522 Mdot
.46171 .14483 Mdot
.46171 .15442 Mdot
.46171 .19867 Mdot
.46171 .23354 Mdot
.46171 .24291 Mdot
.46171 .24517 Mdot
.46171 .26842 Mdot
.46171 .2916 Mdot
.46171 .29401 Mdot
.46171 .31962 Mdot
.46171 .34519 Mdot
.46171 .38174 Mdot
.46171 .38729 Mdot
.46171 .404 Mdot
.46239 .11963 Mdot
.46239 .13225 Mdot
.46239 .14483 Mdot
.46239 .20282 Mdot
.46239 .24854 Mdot
.46239 .29416 Mdot
.46239 .32761 Mdot
.46239 .38272 Mdot
.46239 .40459 Mdot
.46272 .0625 Mdot
.46272 .20484 Mdot
.46272 .27283 Mdot
.46272 .32271 Mdot
.46272 .40488 Mdot
.46304 .13741 Mdot
.46304 .14978 Mdot
.46304 .20683 Mdot
.46304 .219 Mdot
.46304 .29663 Mdot
.46304 .32951 Mdot
.46304 .38367 Mdot
.46304 .40516 Mdot
.46367 .07169 Mdot
.46367 .09005 Mdot
.46367 .10838 Mdot
.46367 .15456 Mdot
.46367 .16668 Mdot
.46367 .22267 Mdot
.46367 .26671 Mdot
.46367 .27702 Mdot
.46367 .29902 Mdot
.46367 .32565 Mdot
.46367 .33135 Mdot
.46367 .38458 Mdot
.46367 .40571 Mdot
.46428 .17111 Mdot
.46428 .22622 Mdot
.46428 .26955 Mdot
.46428 .30133 Mdot
.46428 .33312 Mdot
.46428 .38547 Mdot
.46428 .40625 Mdot
.46458 .13417 Mdot
.46458 .23375 Mdot
.46458 .32843 Mdot
.46458 .40651 Mdot
.46487 .0625 Mdot
.46487 .07431 Mdot
.46487 .08612 Mdot
.46487 .17542 Mdot
.46487 .18708 Mdot
.46487 .22966 Mdot
.46487 .27229 Mdot
.46487 .30357 Mdot
.46487 .33483 Mdot
.46487 .38633 Mdot
.46487 .40676 Mdot
.46544 .06832 Mdot
.46544 .07997 Mdot
.46544 .0916 Mdot
.46544 .10322 Mdot
.46544 .11482 Mdot
.46544 .15869 Mdot
.46544 .17959 Mdot
.46544 .19106 Mdot
.46544 .23299 Mdot
.46544 .23864 Mdot
.46544 .27495 Mdot
.46544 .30573 Mdot
.46544 .3365 Mdot
.46544 .38716 Mdot
.46544 .40726 Mdot
.46599 .10838 Mdot
.46599 .11981 Mdot
.46599 .1312 Mdot
.46599 .19492 Mdot
.46599 .23622 Mdot
.46599 .27753 Mdot
.46599 .30783 Mdot
.46599 .3381 Mdot
.46599 .38796 Mdot
.46599 .40775 Mdot
.46626 .16517 Mdot
.46626 .24331 Mdot
.46626 .40798 Mdot
.46653 .12465 Mdot
.46653 .13588 Mdot
.46653 .14707 Mdot
.46653 .19867 Mdot
.46653 .25027 Mdot
.46653 .28002 Mdot
.46653 .30986 Mdot
.46653 .33966 Mdot
.46653 .38874 Mdot
.46653 .40821 Mdot
.46705 .14042 Mdot
.46705 .15144 Mdot
.46705 .18779 Mdot
.46705 .20231 Mdot
.46705 .21317 Mdot
.46705 .24777 Mdot
.46705 .25313 Mdot
.46705 .28245 Mdot
.46705 .31183 Mdot
.46705 .34117 Mdot
.46705 .38949 Mdot
.46705 .40867 Mdot
.46755 .1557 Mdot
.46755 .16651 Mdot
.46755 .21654 Mdot
.46755 .25591 Mdot
.46755 .2848 Mdot
.46755 .31374 Mdot
.46755 .34264 Mdot
.46755 .39022 Mdot
.46755 .40911 Mdot
.4678 .19334 Mdot
.4678 .25204 Mdot
.4678 .40932 Mdot
.46804 .15983 Mdot
.46804 .17049 Mdot
.46804 .21981 Mdot
.46804 .25861 Mdot
.46804 .28708 Mdot
.46804 .31559 Mdot
.46804 .34406 Mdot
.46804 .39093 Mdot
.46804 .40953 Mdot
.46851 .0625 Mdot
.46851 .0731 Mdot
.46851 .0784 Mdot
.46851 .0837 Mdot
.46851 .09429 Mdot
.46851 .10486 Mdot
.46851 .17435 Mdot
.46851 .18481 Mdot
.46851 .19867 Mdot
.46851 .22298 Mdot
.46851 .26123 Mdot
.46851 .28929 Mdot
.46851 .31739 Mdot
.46851 .34544 Mdot
.46851 .39162 Mdot
.46851 .40995 Mdot
.46897 .06773 Mdot
.46897 .0782 Mdot
.46897 .08866 Mdot
.46897 .09911 Mdot
.46897 .10954 Mdot
.46897 .11995 Mdot
.46897 .17811 Mdot
.46897 .18842 Mdot
.46897 .22607 Mdot
.46897 .26378 Mdot
.46897 .29144 Mdot
.46897 .31913 Mdot
.46897 .34678 Mdot
.46897 .39229 Mdot
.46897 .41035 Mdot
.46919 .10147 Mdot
.46919 .11701 Mdot
.46919 .13249 Mdot
.46919 .41055 Mdot
.46942 .1141 Mdot
.46942 .12437 Mdot
.46942 .13462 Mdot
.46942 .19193 Mdot
.46942 .22907 Mdot
.46942 .23908 Mdot
.46942 .26625 Mdot
.46942 .32083 Mdot
.46942 .34808 Mdot
.46942 .39294 Mdot
.46942 .41074 Mdot
.46985 .12867 Mdot
.46985 .13878 Mdot
.46985 .14886 Mdot
.46985 .19534 Mdot
.46985 .2053 Mdot
.46985 .22357 Mdot
.46985 .24185 Mdot
.46985 .26866 Mdot
.46985 .34934 Mdot
.46985 .39357 Mdot
.46985 .41112 Mdot
.47027 .14284 Mdot
.47027 .15279 Mdot
.47027 .19867 Mdot
.47027 .20849 Mdot
.47027 .24455 Mdot
.47027 .271 Mdot
.47027 .307 Mdot
.47027 .35057 Mdot
.47027 .39418 Mdot
.47027 .41149 Mdot
.47048 .14483 Mdot
.47048 .15964 Mdot
.47048 .22795 Mdot
.47048 .3859 Mdot
.47048 .41167 Mdot
.47068 .15661 Mdot
.47068 .16638 Mdot
.47068 .21159 Mdot
.47068 .24718 Mdot
.47068 .28271 Mdot
.47068 .30878 Mdot
.47068 .35177 Mdot
.47068 .39478 Mdot
.47068 .41185 Mdot
.47108 .06736 Mdot
.47108 .07709 Mdot
.47108 .08681 Mdot
.47108 .16034 Mdot
.47108 .16517 Mdot
.47108 .16998 Mdot
.47108 .21461 Mdot
.47108 .23217 Mdot
.47108 .24973 Mdot
.47108 .2848 Mdot
.47108 .31052 Mdot
.47108 .32338 Mdot
.47108 .35293 Mdot
.47108 .38695 Mdot
.47108 .39536 Mdot
.47108 .41219 Mdot
.47147 .0625 Mdot
.47147 .07212 Mdot
.47147 .08173 Mdot
.47147 .09133 Mdot
.47147 .10093 Mdot
.47147 .11051 Mdot
.47147 .17349 Mdot
.47147 .18297 Mdot
.47147 .21755 Mdot
.47147 .25223 Mdot
.47147 .28683 Mdot
.47147 .31221 Mdot
.47147 .35406 Mdot
.47147 .39592 Mdot
.47147 .41253 Mdot
.47166 .18463 Mdot
.47166 .23622 Mdot
.47166 .28782 Mdot
.47166 .32565 Mdot
.47166 .38796 Mdot
.47166 .4127 Mdot
.47221 .06951 Mdot
.47221 .08352 Mdot
.47221 .18948 Mdot
.47221 .29073 Mdot
.47221 .32783 Mdot
.47221 .38893 Mdot
.47221 .41319 Mdot
.47275 .0625 Mdot
.47275 .07628 Mdot
.47275 .09005 Mdot
.47275 .1038 Mdot
.47275 .11752 Mdot
.47275 .19416 Mdot
.47275 .29353 Mdot
.47275 .32992 Mdot
.47275 .38986 Mdot
.47275 .41365 Mdot
.47326 .10988 Mdot
.47326 .12336 Mdot
.47326 .1368 Mdot
.47326 .21191 Mdot
.47326 .26054 Mdot
.47326 .29623 Mdot
.47326 .33194 Mdot
.47326 .39076 Mdot
.47326 .4141 Mdot
.47376 .129 Mdot
.47376 .1422 Mdot
.47376 .21601 Mdot
.47376 .26378 Mdot
.47376 .29883 Mdot
.47376 .33389 Mdot
.47376 .39162 Mdot
.47376 .41454 Mdot
.47423 .14742 Mdot
.47423 .16034 Mdot
.47423 .21998 Mdot
.47423 .2669 Mdot
.47423 .30133 Mdot
.47423 .33576 Mdot
.47423 .39245 Mdot
.47423 .41495 Mdot
.47469 .16517 Mdot
.47469 .2238 Mdot
.47469 .26992 Mdot
.47469 .30375 Mdot
.47469 .33757 Mdot
.47469 .39325 Mdot
.47469 .41536 Mdot
.47514 .16983 Mdot
.47514 .18224 Mdot
.47514 .2275 Mdot
.47514 .27283 Mdot
.47514 .30609 Mdot
.47514 .33932 Mdot
.47514 .39403 Mdot
.47514 .41574 Mdot
.47556 .06867 Mdot
.47556 .08101 Mdot
.47556 .09334 Mdot
.47556 .10565 Mdot
.47556 .18655 Mdot
.47556 .24306 Mdot
.47556 .27565 Mdot
.47556 .30834 Mdot
.47556 .34101 Mdot
.47556 .39478 Mdot
.47556 .41612 Mdot
.47598 .08681 Mdot
.47598 .09895 Mdot
.47598 .11107 Mdot
.47598 .12316 Mdot
.47598 .19072 Mdot
.47598 .24631 Mdot
.47598 .27837 Mdot
.47598 .31052 Mdot
.47598 .34264 Mdot
.47598 .3955 Mdot
.47598 .41648 Mdot
.47638 .11633 Mdot
.47638 .12823 Mdot
.47638 .1401 Mdot
.47638 .19475 Mdot
.47638 .20647 Mdot
.47638 .24945 Mdot
.47638 .281 Mdot
.47638 .31263 Mdot
.47638 .34421 Mdot
.47638 .3962 Mdot
.47638 .41683 Mdot
.47677 .13316 Mdot
.47677 .14483 Mdot
.47677 .21019 Mdot
.47677 .2525 Mdot
.47677 .34574 Mdot
.47677 .39687 Mdot
.47677 .41717 Mdot
.47714 .14943 Mdot
.47714 .16088 Mdot
.47714 .21379 Mdot
.47714 .25545 Mdot
.47714 .34721 Mdot
.47714 .39753 Mdot
.47714 .4175 Mdot
.4775 .16517 Mdot
.4775 .21729 Mdot
.4775 .25831 Mdot
.4775 .34864 Mdot
.4775 .39816 Mdot
.4775 .41781 Mdot
.47785 .0625 Mdot
.47785 .07367 Mdot
.47785 .08484 Mdot
.47785 .16932 Mdot
.47785 .18038 Mdot
.47785 .22068 Mdot
.47785 .23158 Mdot
.47785 .26109 Mdot
.47785 .35003 Mdot
.47785 .39877 Mdot
.47785 .41812 Mdot
.47819 .06801 Mdot
.47819 .07904 Mdot
.47819 .09005 Mdot
.47819 .10106 Mdot
.47819 .11204 Mdot
.47819 .18425 Mdot
.47819 .2347 Mdot
.47819 .26378 Mdot
.47819 .30341 Mdot
.47819 .35137 Mdot
.47819 .39937 Mdot
.47819 .41842 Mdot
.47852 .10597 Mdot
.47852 .1168 Mdot
.47852 .12761 Mdot
.47852 .188 Mdot
.47852 .19867 Mdot
.47852 .23773 Mdot
.47852 .30543 Mdot
.47852 .35267 Mdot
.47852 .39994 Mdot
.47852 .41871 Mdot
.47884 .12144 Mdot
.47884 .13209 Mdot
.47884 .14271 Mdot
.47884 .19166 Mdot
.47884 .20216 Mdot
.47884 .24067 Mdot
.47884 .27915 Mdot
.47884 .30738 Mdot
.47884 .35394 Mdot
.47884 .4005 Mdot
.47884 .41899 Mdot
.47916 .13645 Mdot
.47916 .14693 Mdot
.47916 .15736 Mdot
.47916 .20556 Mdot
.47916 .24353 Mdot
.47916 .28146 Mdot
.47916 .30928 Mdot
.47916 .35517 Mdot
.47916 .40105 Mdot
.47916 .41926 Mdot
.47946 .15103 Mdot
.47946 .16132 Mdot
.47946 .20886 Mdot
.47946 .24631 Mdot
.47946 .2837 Mdot
.47946 .31113 Mdot
.47946 .33853 Mdot
.47946 .35636 Mdot
.47946 .40157 Mdot
.47946 .41953 Mdot
.47975 .0676 Mdot
.47975 .16517 Mdot
.47975 .17527 Mdot
.47975 .21208 Mdot
.47975 .22206 Mdot
.47975 .24901 Mdot
.47975 .28588 Mdot
.47975 .31292 Mdot
.47975 .33994 Mdot
.47975 .35751 Mdot
.47975 .40209 Mdot
.47975 .41978 Mdot
.48004 .0625 Mdot
.48004 .07258 Mdot
.48004 .08267 Mdot
.48004 .09274 Mdot
.48004 .1028 Mdot
.48004 .16891 Mdot
.48004 .17888 Mdot
.48004 .2152 Mdot
.48004 .22505 Mdot
.48004 .26136 Mdot
.48004 .288 Mdot
.48004 .31467 Mdot
.48004 .34131 Mdot
.48004 .35864 Mdot
.48004 .40258 Mdot
.48004 .42003 Mdot
.48031 .06748 Mdot
.48031 .07744 Mdot
.48031 .0874 Mdot
.48031 .09735 Mdot
.48031 .10728 Mdot
.48031 .11719 Mdot
.48031 .18239 Mdot
.48031 .22795 Mdot
.48031 .26378 Mdot
.48031 .29006 Mdot
.48031 .31637 Mdot
.48031 .34264 Mdot
.48031 .35973 Mdot
.48031 .40307 Mdot
.48031 .42027 Mdot
.48058 .11165 Mdot
.48058 .12144 Mdot
.48058 .1312 Mdot
.48058 .18581 Mdot
.48058 .19546 Mdot
.48058 .23078 Mdot
.48058 .26613 Mdot
.48058 .29206 Mdot
.48058 .31802 Mdot
.48058 .34393 Mdot
.48058 .3608 Mdot
.48058 .40354 Mdot
.48058 .42051 Mdot
.48085 .12558 Mdot
.48085 .13522 Mdot
.48085 .14483 Mdot
.48085 .18915 Mdot
.48085 .19867 Mdot
.48085 .23354 Mdot
.48085 .26842 Mdot
.48085 .29401 Mdot
.48085 .31962 Mdot
.48085 .34519 Mdot
.48085 .36183 Mdot
.48085 .404 Mdot
.48085 .42074 Mdot
.4811 .13914 Mdot
.4811 .14863 Mdot
.4811 .23622 Mdot
.4811 .29592 Mdot
.4811 .36284 Mdot
.4811 .42096 Mdot
.48159 .17435 Mdot
.48159 .42139 Mdot
.48206 .18087 Mdot
.48206 .39046 Mdot
.48206 .4218 Mdot
.4825 .18708 Mdot
.4825 .39162 Mdot
.4825 .42219 Mdot
.48292 .0625 Mdot
.48292 .07973 Mdot
.48292 .20994 Mdot
.48292 .31761 Mdot
.48292 .39272 Mdot
.48292 .42256 Mdot
.48332 .07094 Mdot
.48332 .08781 Mdot
.48332 .10464 Mdot
.48332 .12144 Mdot
.48332 .2152 Mdot
.48332 .2759 Mdot
.48332 .32041 Mdot
.48332 .39377 Mdot
.48332 .42291 Mdot
.48371 .12847 Mdot
.48371 .14483 Mdot
.48371 .22023 Mdot
.48371 .27958 Mdot
.48371 .32309 Mdot
.48371 .39478 Mdot
.48371 .42324 Mdot
.48407 .15123 Mdot
.48407 .22505 Mdot
.48407 .28309 Mdot
.48407 .32565 Mdot
.48407 .39573 Mdot
.48407 .42356 Mdot
.48442 .15736 Mdot
.48442 .17294 Mdot
.48442 .28646 Mdot
.48442 .32809 Mdot
.48442 .39665 Mdot
.48442 .42387 Mdot
.48476 .1785 Mdot
.48476 .28969 Mdot
.48476 .33044 Mdot
.48476 .39753 Mdot
.48476 .42416 Mdot
.48508 .18384 Mdot
.48508 .25293 Mdot
.48508 .29278 Mdot
.48508 .33268 Mdot
.48508 .39837 Mdot
.48508 .42444 Mdot
.48539 .0625 Mdot
.48539 .07727 Mdot
.48539 .09202 Mdot
.48539 .10675 Mdot
.48539 .20349 Mdot
.48539 .25669 Mdot
.48539 .29574 Mdot
.48539 .33483 Mdot
.48539 .39917 Mdot
.48539 .42471 Mdot
.48568 .09874 Mdot
.48568 .1132 Mdot
.48568 .12761 Mdot
.48568 .20814 Mdot
.48568 .2603 Mdot
.48568 .29859 Mdot
.48568 .3369 Mdot
.48568 .39994 Mdot
.48568 .42497 Mdot
.48596 .13356 Mdot
.48596 .14765 Mdot
.48596 .2126 Mdot
.48596 .26378 Mdot
.48596 .30133 Mdot
.48596 .33889 Mdot
.48596 .40069 Mdot
.48596 .42522 Mdot
.48624 .15312 Mdot
.48624 .21691 Mdot
.48624 .26712 Mdot
.48624 .3408 Mdot
.48624 .4014 Mdot
.48624 .42546 Mdot
.4865 .15841 Mdot
.4865 .17191 Mdot
.4865 .23434 Mdot
.4865 .27034 Mdot
.4865 .34264 Mdot
.4865 .40209 Mdot
.4865 .42569 Mdot
.48675 .17676 Mdot
.48675 .23808 Mdot
.48675 .34441 Mdot
.48675 .40275 Mdot
.48675 .42591 Mdot
.48699 .06906 Mdot
.48699 .08219 Mdot
.48699 .18144 Mdot
.48699 .24169 Mdot
.48699 .34611 Mdot
.48699 .40338 Mdot
.48699 .42612 Mdot
.48723 .0625 Mdot
.48723 .07542 Mdot
.48723 .08833 Mdot
.48723 .10123 Mdot
.48723 .1141 Mdot
.48723 .19867 Mdot
.48723 .24517 Mdot
.48723 .34776 Mdot
.48723 .404 Mdot
.48723 .42633 Mdot
.48745 .10697 Mdot
.48745 .11963 Mdot
.48745 .13225 Mdot
.48745 .20282 Mdot
.48745 .24854 Mdot
.48745 .34934 Mdot
.48745 .40459 Mdot
.48745 .42652 Mdot
.48767 .13741 Mdot
.48767 .14978 Mdot
.48767 .20683 Mdot
.48767 .25179 Mdot
.48767 .29663 Mdot
.48767 .35087 Mdot
.48767 .40516 Mdot
.48767 .42671 Mdot
.48788 .15456 Mdot
.48788 .21071 Mdot
.48788 .22267 Mdot
.48788 .29902 Mdot
.48788 .35235 Mdot
.48788 .40571 Mdot
.48788 .4269 Mdot
.48809 .1592 Mdot
.48809 .17111 Mdot
.48809 .22622 Mdot
.48809 .30133 Mdot
.48809 .35378 Mdot
.48809 .40625 Mdot
.48809 .42708 Mdot
.48828 .17542 Mdot
.48828 .22966 Mdot
.48828 .27229 Mdot
.48828 .30357 Mdot
.48828 .33483 Mdot
.48828 .35517 Mdot
.48828 .40676 Mdot
.48828 .42725 Mdot
.48847 .06832 Mdot
.48847 .07997 Mdot
.48847 .0916 Mdot
.48847 .10322 Mdot
.48847 .17959 Mdot
.48847 .19106 Mdot
.48847 .23299 Mdot
.48847 .27495 Mdot
.48847 .30573 Mdot
.48847 .3365 Mdot
.48847 .3565 Mdot
.48847 .40726 Mdot
.48847 .42742 Mdot
.48866 .08546 Mdot
.48866 .09693 Mdot
.48866 .10838 Mdot
.48866 .11981 Mdot
.48866 .19492 Mdot
.48866 .23622 Mdot
.48866 .27753 Mdot
.48866 .30783 Mdot
.48866 .3381 Mdot
.48866 .3578 Mdot
.48866 .40775 Mdot
.48866 .42758 Mdot
.48884 .11339 Mdot
.48884 .12465 Mdot
.48884 .13588 Mdot
.48884 .19867 Mdot
.48884 .23936 Mdot
.48884 .28002 Mdot
.48884 .30986 Mdot
.48884 .33966 Mdot
.48884 .35905 Mdot
.48884 .40821 Mdot
.48884 .42773 Mdot
.48901 .14042 Mdot
.48901 .15144 Mdot
.48901 .20231 Mdot
.48901 .25313 Mdot
.48901 .28245 Mdot
.48901 .31183 Mdot
.48901 .34117 Mdot
.48901 .40867 Mdot
.48901 .42788 Mdot
.48918 .14483 Mdot
.48918 .1557 Mdot
.48918 .20584 Mdot
.48918 .21654 Mdot
.48918 .25591 Mdot
.48918 .2848 Mdot
.48918 .31374 Mdot
.48918 .34264 Mdot
.48918 .40911 Mdot
.48918 .42803 Mdot
.48934 .15983 Mdot
.48934 .17049 Mdot
.48934 .21981 Mdot
.48934 .25861 Mdot
.48934 .28708 Mdot
.48934 .31559 Mdot
.48934 .34406 Mdot
.48934 .40953 Mdot
.48934 .42817 Mdot
.4895 .0625 Mdot
.4895 .0731 Mdot
.4895 .0837 Mdot
.4895 .09429 Mdot
.4895 .17435 Mdot
.4895 .22298 Mdot
.4895 .26123 Mdot
.4895 .28929 Mdot
.4895 .31739 Mdot
.4895 .34544 Mdot
.4895 .39162 Mdot
.4895 .40995 Mdot
.4895 .42831 Mdot
.48965 .06773 Mdot
.48965 .0782 Mdot
.48965 .08866 Mdot
.48965 .09911 Mdot
.48965 .10954 Mdot
.48965 .11995 Mdot
.48965 .17811 Mdot
.48965 .18842 Mdot
.48965 .22607 Mdot
.48965 .26378 Mdot
.48965 .29144 Mdot
.48965 .31913 Mdot
.48965 .34678 Mdot
.48965 .39229 Mdot
.48965 .41035 Mdot
.48965 .42845 Mdot
.4898 .1038 Mdot
.4898 .1141 Mdot
.4898 .12437 Mdot
.4898 .13462 Mdot
.4898 .18177 Mdot
.4898 .19193 Mdot
.4898 .22907 Mdot
.4898 .26625 Mdot
.4898 .29353 Mdot
.4898 .32083 Mdot
.4898 .34808 Mdot
.4898 .39294 Mdot
.4898 .41074 Mdot
.4898 .42858 Mdot
.48995 .12867 Mdot
.48995 .13878 Mdot
.48995 .19534 Mdot
.48995 .23199 Mdot
.48995 .24185 Mdot
.48995 .26866 Mdot
.48995 .29556 Mdot
.48995 .32248 Mdot
.48995 .34934 Mdot
.48995 .39357 Mdot
.48995 .41112 Mdot
.48995 .4287 Mdot
.49009 .14284 Mdot
.49009 .15279 Mdot
.49009 .19867 Mdot
.49009 .20849 Mdot
.49009 .24455 Mdot
.49009 .271 Mdot
.49009 .32409 Mdot
.49009 .35057 Mdot
.49009 .39418 Mdot
.49009 .41149 Mdot
.49009 .42883 Mdot
.49022 .1468 Mdot
.49022 .15661 Mdot
.49022 .16638 Mdot
.49022 .20191 Mdot
.49022 .21159 Mdot
.49022 .24718 Mdot
.49022 .27328 Mdot
.49022 .32565 Mdot
.49022 .35177 Mdot
.49022 .39478 Mdot
.49022 .41185 Mdot
.49022 .42895 Mdot
.49036 .06736 Mdot
.49036 .07709 Mdot
.49036 .16034 Mdot
.49036 .16998 Mdot
.49036 .21461 Mdot
.49036 .24973 Mdot
.49036 .2755 Mdot
.49036 .32716 Mdot
.49036 .35293 Mdot
.49036 .39536 Mdot
.49036 .41219 Mdot
.49036 .42906 Mdot
.49049 .0625 Mdot
.49049 .07212 Mdot
.49049 .08173 Mdot
.49049 .09133 Mdot
.49049 .10093 Mdot
.49049 .11051 Mdot
.49049 .16397 Mdot
.49049 .17349 Mdot
.49049 .21755 Mdot
.49049 .25223 Mdot
.49049 .28683 Mdot
.49049 .31221 Mdot
.49049 .32864 Mdot
.49049 .35406 Mdot
.49049 .39592 Mdot
.49049 .41253 Mdot
.49049 .42917 Mdot
.5 .09005 Mdot
.5 .11633 Mdot
.5 .16517 Mdot
.5 .27283 Mdot
.5 .281 Mdot
.5 .4375 Mdot
.50951 .0625 Mdot
.50951 .07212 Mdot
.50951 .08173 Mdot
.50951 .09133 Mdot
.50951 .10093 Mdot
.50951 .11051 Mdot
.50951 .16397 Mdot
.50951 .17349 Mdot
.50951 .21755 Mdot
.50951 .25223 Mdot
.50951 .28683 Mdot
.50951 .31221 Mdot
.50951 .32864 Mdot
.50951 .35406 Mdot
.50951 .39592 Mdot
.50951 .41253 Mdot
.50951 .42917 Mdot
.50964 .06736 Mdot
.50964 .07709 Mdot
.50964 .16034 Mdot
.50964 .16998 Mdot
.50964 .21461 Mdot
.50964 .24973 Mdot
.50964 .2755 Mdot
.50964 .32716 Mdot
.50964 .35293 Mdot
.50964 .39536 Mdot
.50964 .41219 Mdot
.50964 .42906 Mdot
.50978 .1468 Mdot
.50978 .15661 Mdot
.50978 .16638 Mdot
.50978 .20191 Mdot
.50978 .21159 Mdot
.50978 .24718 Mdot
.50978 .27328 Mdot
.50978 .32565 Mdot
.50978 .35177 Mdot
.50978 .39478 Mdot
.50978 .41185 Mdot
.50978 .42895 Mdot
.50991 .14284 Mdot
.50991 .15279 Mdot
.50991 .19867 Mdot
.50991 .20849 Mdot
.50991 .24455 Mdot
.50991 .271 Mdot
.50991 .32409 Mdot
.50991 .35057 Mdot
.50991 .39418 Mdot
.50991 .41149 Mdot
.50991 .42883 Mdot
.51005 .12867 Mdot
.51005 .13878 Mdot
.51005 .19534 Mdot
.51005 .23199 Mdot
.51005 .24185 Mdot
.51005 .26866 Mdot
.51005 .29556 Mdot
.51005 .32248 Mdot
.51005 .34934 Mdot
.51005 .39357 Mdot
.51005 .41112 Mdot
.51005 .4287 Mdot
.5102 .1038 Mdot
.5102 .1141 Mdot
.5102 .12437 Mdot
.5102 .13462 Mdot
.5102 .18177 Mdot
.5102 .19193 Mdot
.5102 .22907 Mdot
.5102 .26625 Mdot
.5102 .29353 Mdot
.5102 .32083 Mdot
.5102 .34808 Mdot
.5102 .39294 Mdot
.5102 .41074 Mdot
.5102 .42858 Mdot
.51035 .06773 Mdot
.51035 .0782 Mdot
.51035 .08866 Mdot
.51035 .09911 Mdot
.51035 .10954 Mdot
.51035 .11995 Mdot
.51035 .17811 Mdot
.51035 .18842 Mdot
.51035 .22607 Mdot
.51035 .26378 Mdot
.51035 .29144 Mdot
.51035 .31913 Mdot
.51035 .34678 Mdot
.51035 .39229 Mdot
.51035 .41035 Mdot
.51035 .42845 Mdot
.5105 .0625 Mdot
.5105 .0731 Mdot
.5105 .0837 Mdot
.5105 .09429 Mdot
.5105 .17435 Mdot
.5105 .22298 Mdot
.5105 .26123 Mdot
.5105 .28929 Mdot
.5105 .31739 Mdot
.5105 .34544 Mdot
.5105 .39162 Mdot
.5105 .40995 Mdot
.5105 .42831 Mdot
.51066 .15983 Mdot
.51066 .17049 Mdot
.51066 .21981 Mdot
.51066 .25861 Mdot
.51066 .28708 Mdot
.51066 .31559 Mdot
.51066 .34406 Mdot
.51066 .40953 Mdot
.51066 .42817 Mdot
.51082 .14483 Mdot
.51082 .1557 Mdot
.51082 .20584 Mdot
.51082 .21654 Mdot
.51082 .25591 Mdot
.51082 .2848 Mdot
.51082 .31374 Mdot
.51082 .34264 Mdot
.51082 .40911 Mdot
.51082 .42803 Mdot
.51099 .14042 Mdot
.51099 .15144 Mdot
.51099 .20231 Mdot
.51099 .25313 Mdot
.51099 .28245 Mdot
.51099 .31183 Mdot
.51099 .34117 Mdot
.51099 .40867 Mdot
.51099 .42788 Mdot
.51116 .11339 Mdot
.51116 .12465 Mdot
.51116 .13588 Mdot
.51116 .19867 Mdot
.51116 .23936 Mdot
.51116 .28002 Mdot
.51116 .30986 Mdot
.51116 .33966 Mdot
.51116 .35905 Mdot
.51116 .40821 Mdot
.51116 .42773 Mdot
.51134 .08546 Mdot
.51134 .09693 Mdot
.51134 .10838 Mdot
.51134 .11981 Mdot
.51134 .19492 Mdot
.51134 .23622 Mdot
.51134 .27753 Mdot
.51134 .30783 Mdot
.51134 .3381 Mdot
.51134 .3578 Mdot
.51134 .40775 Mdot
.51134 .42758 Mdot
.51153 .06832 Mdot
.51153 .07997 Mdot
.51153 .0916 Mdot
.51153 .10322 Mdot
.51153 .17959 Mdot
.51153 .19106 Mdot
.51153 .23299 Mdot
.51153 .27495 Mdot
.51153 .30573 Mdot
.51153 .3365 Mdot
.51153 .3565 Mdot
.51153 .40726 Mdot
.51153 .42742 Mdot
.51172 .17542 Mdot
.51172 .22966 Mdot
.51172 .27229 Mdot
.51172 .30357 Mdot
.51172 .33483 Mdot
.51172 .35517 Mdot
.51172 .40676 Mdot
.51172 .42725 Mdot
.51191 .1592 Mdot
.51191 .17111 Mdot
.51191 .22622 Mdot
.51191 .30133 Mdot
.51191 .35378 Mdot
.51191 .40625 Mdot
.51191 .42708 Mdot
.51212 .15456 Mdot
.51212 .21071 Mdot
.51212 .22267 Mdot
.51212 .29902 Mdot
.51212 .35235 Mdot
.51212 .40571 Mdot
.51212 .4269 Mdot
.51233 .13741 Mdot
.51233 .14978 Mdot
.51233 .20683 Mdot
.51233 .25179 Mdot
.51233 .29663 Mdot
.51233 .35087 Mdot
.51233 .40516 Mdot
.51233 .42671 Mdot
.51255 .10697 Mdot
.51255 .11963 Mdot
.51255 .13225 Mdot
.51255 .20282 Mdot
.51255 .24854 Mdot
.51255 .34934 Mdot
.51255 .40459 Mdot
.51255 .42652 Mdot
.51277 .0625 Mdot
.51277 .07542 Mdot
.51277 .08833 Mdot
.51277 .10123 Mdot
.51277 .1141 Mdot
.51277 .19867 Mdot
.51277 .24517 Mdot
.51277 .34776 Mdot
.51277 .404 Mdot
.51277 .42633 Mdot
.51301 .06906 Mdot
.51301 .08219 Mdot
.51301 .18144 Mdot
.51301 .24169 Mdot
.51301 .34611 Mdot
.51301 .40338 Mdot
.51301 .42612 Mdot
.51325 .17676 Mdot
.51325 .23808 Mdot
.51325 .34441 Mdot
.51325 .40275 Mdot
.51325 .42591 Mdot
.5135 .15841 Mdot
.5135 .17191 Mdot
.5135 .23434 Mdot
.5135 .27034 Mdot
.5135 .34264 Mdot
.5135 .40209 Mdot
.5135 .42569 Mdot
.51376 .15312 Mdot
.51376 .21691 Mdot
.51376 .26712 Mdot
.51376 .3408 Mdot
.51376 .4014 Mdot
.51376 .42546 Mdot
.51404 .13356 Mdot
.51404 .14765 Mdot
.51404 .2126 Mdot
.51404 .26378 Mdot
.51404 .30133 Mdot
.51404 .33889 Mdot
.51404 .40069 Mdot
.51404 .42522 Mdot
.51432 .09874 Mdot
.51432 .1132 Mdot
.51432 .12761 Mdot
.51432 .20814 Mdot
.51432 .2603 Mdot
.51432 .29859 Mdot
.51432 .3369 Mdot
.51432 .39994 Mdot
.51432 .42497 Mdot
.51461 .0625 Mdot
.51461 .07727 Mdot
.51461 .09202 Mdot
.51461 .10675 Mdot
.51461 .20349 Mdot
.51461 .25669 Mdot
.51461 .29574 Mdot
.51461 .33483 Mdot
.51461 .39917 Mdot
.51461 .42471 Mdot
.51492 .18384 Mdot
.51492 .25293 Mdot
.51492 .29278 Mdot
.51492 .33268 Mdot
.51492 .39837 Mdot
.51492 .42444 Mdot
.51524 .1785 Mdot
.51524 .28969 Mdot
.51524 .33044 Mdot
.51524 .39753 Mdot
.51524 .42416 Mdot
.51558 .15736 Mdot
.51558 .17294 Mdot
.51558 .28646 Mdot
.51558 .32809 Mdot
.51558 .39665 Mdot
.51558 .42387 Mdot
.51593 .15123 Mdot
.51593 .22505 Mdot
.51593 .28309 Mdot
.51593 .32565 Mdot
.51593 .39573 Mdot
.51593 .42356 Mdot
.51629 .12847 Mdot
.51629 .14483 Mdot
.51629 .22023 Mdot
.51629 .27958 Mdot
.51629 .32309 Mdot
.51629 .39478 Mdot
.51629 .42324 Mdot
.51668 .07094 Mdot
.51668 .08781 Mdot
.51668 .10464 Mdot
.51668 .12144 Mdot
.51668 .2152 Mdot
.51668 .2759 Mdot
.51668 .32041 Mdot
.51668 .39377 Mdot
.51668 .42291 Mdot
.51708 .0625 Mdot
.51708 .07973 Mdot
.51708 .20994 Mdot
.51708 .31761 Mdot
.51708 .39272 Mdot
.51708 .42256 Mdot
.5175 .18708 Mdot
.5175 .39162 Mdot
.5175 .42219 Mdot
.51794 .18087 Mdot
.51794 .39046 Mdot
.51794 .4218 Mdot
.51841 .17435 Mdot
.51841 .42139 Mdot
.5189 .13914 Mdot
.5189 .14863 Mdot
.5189 .23622 Mdot
.5189 .29592 Mdot
.5189 .36284 Mdot
.5189 .42096 Mdot
.51915 .12558 Mdot
.51915 .13522 Mdot
.51915 .14483 Mdot
.51915 .18915 Mdot
.51915 .19867 Mdot
.51915 .23354 Mdot
.51915 .26842 Mdot
.51915 .29401 Mdot
.51915 .31962 Mdot
.51915 .34519 Mdot
.51915 .36183 Mdot
.51915 .404 Mdot
.51915 .42074 Mdot
.51942 .11165 Mdot
.51942 .12144 Mdot
.51942 .1312 Mdot
.51942 .18581 Mdot
.51942 .19546 Mdot
.51942 .23078 Mdot
.51942 .26613 Mdot
.51942 .29206 Mdot
.51942 .31802 Mdot
.51942 .34393 Mdot
.51942 .3608 Mdot
.51942 .40354 Mdot
.51942 .42051 Mdot
.51969 .06748 Mdot
.51969 .07744 Mdot
.51969 .0874 Mdot
.51969 .09735 Mdot
.51969 .10728 Mdot
.51969 .11719 Mdot
.51969 .18239 Mdot
.51969 .22795 Mdot
.51969 .26378 Mdot
.51969 .29006 Mdot
.51969 .31637 Mdot
.51969 .34264 Mdot
.51969 .35973 Mdot
.51969 .40307 Mdot
.51969 .42027 Mdot
.51996 .0625 Mdot
.51996 .07258 Mdot
.51996 .08267 Mdot
.51996 .09274 Mdot
.51996 .1028 Mdot
.51996 .16891 Mdot
.51996 .17888 Mdot
.51996 .2152 Mdot
.51996 .22505 Mdot
.51996 .26136 Mdot
.51996 .288 Mdot
.51996 .31467 Mdot
.51996 .34131 Mdot
.51996 .35864 Mdot
.51996 .40258 Mdot
.51996 .42003 Mdot
.52025 .0676 Mdot
.52025 .16517 Mdot
.52025 .17527 Mdot
.52025 .21208 Mdot
.52025 .22206 Mdot
.52025 .24901 Mdot
.52025 .28588 Mdot
.52025 .31292 Mdot
.52025 .33994 Mdot
.52025 .35751 Mdot
.52025 .40209 Mdot
.52025 .41978 Mdot
.52054 .15103 Mdot
.52054 .16132 Mdot
.52054 .20886 Mdot
.52054 .24631 Mdot
.52054 .2837 Mdot
.52054 .31113 Mdot
.52054 .33853 Mdot
.52054 .35636 Mdot
.52054 .40157 Mdot
.52054 .41953 Mdot
.52084 .13645 Mdot
.52084 .14693 Mdot
.52084 .15736 Mdot
.52084 .20556 Mdot
.52084 .24353 Mdot
.52084 .28146 Mdot
.52084 .30928 Mdot
.52084 .35517 Mdot
.52084 .40105 Mdot
.52084 .41926 Mdot
.52116 .12144 Mdot
.52116 .13209 Mdot
.52116 .14271 Mdot
.52116 .19166 Mdot
.52116 .20216 Mdot
.52116 .24067 Mdot
.52116 .27915 Mdot
.52116 .30738 Mdot
.52116 .35394 Mdot
.52116 .4005 Mdot
.52116 .41899 Mdot
.52148 .10597 Mdot
.52148 .1168 Mdot
.52148 .12761 Mdot
.52148 .188 Mdot
.52148 .19867 Mdot
.52148 .23773 Mdot
.52148 .30543 Mdot
.52148 .35267 Mdot
.52148 .39994 Mdot
.52148 .41871 Mdot
.52181 .06801 Mdot
.52181 .07904 Mdot
.52181 .09005 Mdot
.52181 .10106 Mdot
.52181 .11204 Mdot
.52181 .18425 Mdot
.52181 .2347 Mdot
.52181 .26378 Mdot
.52181 .30341 Mdot
.52181 .35137 Mdot
.52181 .39937 Mdot
.52181 .41842 Mdot
.52215 .0625 Mdot
.52215 .07367 Mdot
.52215 .08484 Mdot
.52215 .16932 Mdot
.52215 .18038 Mdot
.52215 .22068 Mdot
.52215 .23158 Mdot
.52215 .26109 Mdot
.52215 .35003 Mdot
.52215 .39877 Mdot
.52215 .41812 Mdot
.5225 .16517 Mdot
.5225 .21729 Mdot
.5225 .25831 Mdot
.5225 .34864 Mdot
.5225 .39816 Mdot
.5225 .41781 Mdot
.52286 .14943 Mdot
.52286 .16088 Mdot
.52286 .21379 Mdot
.52286 .25545 Mdot
.52286 .34721 Mdot
.52286 .39753 Mdot
.52286 .4175 Mdot
.52323 .13316 Mdot
.52323 .14483 Mdot
.52323 .21019 Mdot
.52323 .2525 Mdot
.52323 .34574 Mdot
.52323 .39687 Mdot
.52323 .41717 Mdot
.52362 .11633 Mdot
.52362 .12823 Mdot
.52362 .1401 Mdot
.52362 .19475 Mdot
.52362 .20647 Mdot
.52362 .24945 Mdot
.52362 .281 Mdot
.52362 .31263 Mdot
.52362 .34421 Mdot
.52362 .3962 Mdot
.52362 .41683 Mdot
.52402 .08681 Mdot
.52402 .09895 Mdot
.52402 .11107 Mdot
.52402 .12316 Mdot
.52402 .19072 Mdot
.52402 .24631 Mdot
.52402 .27837 Mdot
.52402 .31052 Mdot
.52402 .34264 Mdot
.52402 .3955 Mdot
.52402 .41648 Mdot
.52444 .06867 Mdot
.52444 .08101 Mdot
.52444 .09334 Mdot
.52444 .10565 Mdot
.52444 .18655 Mdot
.52444 .24306 Mdot
.52444 .27565 Mdot
.52444 .30834 Mdot
.52444 .34101 Mdot
.52444 .39478 Mdot
.52444 .41612 Mdot
.52486 .16983 Mdot
.52486 .18224 Mdot
.52486 .2275 Mdot
.52486 .27283 Mdot
.52486 .30609 Mdot
.52486 .33932 Mdot
.52486 .39403 Mdot
.52486 .41574 Mdot
.52531 .16517 Mdot
.52531 .2238 Mdot
.52531 .26992 Mdot
.52531 .30375 Mdot
.52531 .33757 Mdot
.52531 .39325 Mdot
.52531 .41536 Mdot
.52577 .14742 Mdot
.52577 .16034 Mdot
.52577 .21998 Mdot
.52577 .2669 Mdot
.52577 .30133 Mdot
.52577 .33576 Mdot
.52577 .39245 Mdot
.52577 .41495 Mdot
.52624 .129 Mdot
.52624 .1422 Mdot
.52624 .21601 Mdot
.52624 .26378 Mdot
.52624 .29883 Mdot
.52624 .33389 Mdot
.52624 .39162 Mdot
.52624 .41454 Mdot
.52674 .10988 Mdot
.52674 .12336 Mdot
.52674 .1368 Mdot
.52674 .21191 Mdot
.52674 .26054 Mdot
.52674 .29623 Mdot
.52674 .33194 Mdot
.52674 .39076 Mdot
.52674 .4141 Mdot
.52725 .0625 Mdot
.52725 .07628 Mdot
.52725 .09005 Mdot
.52725 .1038 Mdot
.52725 .11752 Mdot
.52725 .19416 Mdot
.52725 .29353 Mdot
.52725 .32992 Mdot
.52725 .38986 Mdot
.52725 .41365 Mdot
.52779 .06951 Mdot
.52779 .08352 Mdot
.52779 .18948 Mdot
.52779 .29073 Mdot
.52779 .32783 Mdot
.52779 .38893 Mdot
.52779 .41319 Mdot
.52834 .18463 Mdot
.52834 .23622 Mdot
.52834 .28782 Mdot
.52834 .32565 Mdot
.52834 .38796 Mdot
.52834 .4127 Mdot
.52853 .0625 Mdot
.52853 .07212 Mdot
.52853 .08173 Mdot
.52853 .09133 Mdot
.52853 .10093 Mdot
.52853 .11051 Mdot
.52853 .17349 Mdot
.52853 .18297 Mdot
.52853 .21755 Mdot
.52853 .25223 Mdot
.52853 .28683 Mdot
.52853 .31221 Mdot
.52853 .35406 Mdot
.52853 .39592 Mdot
.52853 .41253 Mdot
.52892 .06736 Mdot
.52892 .07709 Mdot
.52892 .08681 Mdot
.52892 .16034 Mdot
.52892 .16517 Mdot
.52892 .16998 Mdot
.52892 .21461 Mdot
.52892 .23217 Mdot
.52892 .24973 Mdot
.52892 .2848 Mdot
.52892 .31052 Mdot
.52892 .32338 Mdot
.52892 .35293 Mdot
.52892 .38695 Mdot
.52892 .39536 Mdot
.52892 .41219 Mdot
.52932 .15661 Mdot
.52932 .16638 Mdot
.52932 .21159 Mdot
.52932 .24718 Mdot
.52932 .28271 Mdot
.52932 .30878 Mdot
.52932 .35177 Mdot
.52932 .39478 Mdot
.52932 .41185 Mdot
.52952 .14483 Mdot
.52952 .15964 Mdot
.52952 .22795 Mdot
.52952 .3859 Mdot
.52952 .41167 Mdot
.52973 .14284 Mdot
.52973 .15279 Mdot
.52973 .19867 Mdot
.52973 .20849 Mdot
.52973 .24455 Mdot
.52973 .271 Mdot
.52973 .307 Mdot
.52973 .35057 Mdot
.52973 .39418 Mdot
.52973 .41149 Mdot
.53015 .12867 Mdot
.53015 .13878 Mdot
.53015 .14886 Mdot
.53015 .19534 Mdot
.53015 .2053 Mdot
.53015 .22357 Mdot
.53015 .24185 Mdot
.53015 .26866 Mdot
.53015 .34934 Mdot
.53015 .39357 Mdot
.53015 .41112 Mdot
.53058 .1141 Mdot
.53058 .12437 Mdot
.53058 .13462 Mdot
.53058 .19193 Mdot
.53058 .22907 Mdot
.53058 .23908 Mdot
.53058 .26625 Mdot
.53058 .32083 Mdot
.53058 .34808 Mdot
.53058 .39294 Mdot
.53058 .41074 Mdot
.53081 .10147 Mdot
.53081 .11701 Mdot
.53081 .13249 Mdot
.53081 .41055 Mdot
.53103 .06773 Mdot
.53103 .0782 Mdot
.53103 .08866 Mdot
.53103 .09911 Mdot
.53103 .10954 Mdot
.53103 .11995 Mdot
.53103 .17811 Mdot
.53103 .18842 Mdot
.53103 .22607 Mdot
.53103 .26378 Mdot
.53103 .29144 Mdot
.53103 .31913 Mdot
.53103 .34678 Mdot
.53103 .39229 Mdot
.53103 .41035 Mdot
.53149 .0625 Mdot
.53149 .0731 Mdot
.53149 .0784 Mdot
.53149 .0837 Mdot
.53149 .09429 Mdot
.53149 .10486 Mdot
.53149 .17435 Mdot
.53149 .18481 Mdot
.53149 .19867 Mdot
.53149 .22298 Mdot
.53149 .26123 Mdot
.53149 .28929 Mdot
.53149 .31739 Mdot
.53149 .34544 Mdot
.53149 .39162 Mdot
.53149 .40995 Mdot
.53196 .15983 Mdot
.53196 .17049 Mdot
.53196 .21981 Mdot
.53196 .25861 Mdot
.53196 .28708 Mdot
.53196 .31559 Mdot
.53196 .34406 Mdot
.53196 .39093 Mdot
.53196 .40953 Mdot
.5322 .19334 Mdot
.5322 .25204 Mdot
.5322 .40932 Mdot
.53245 .1557 Mdot
.53245 .16651 Mdot
.53245 .21654 Mdot
.53245 .25591 Mdot
.53245 .2848 Mdot
.53245 .31374 Mdot
.53245 .34264 Mdot
.53245 .39022 Mdot
.53245 .40911 Mdot
.53295 .14042 Mdot
.53295 .15144 Mdot
.53295 .18779 Mdot
.53295 .20231 Mdot
.53295 .21317 Mdot
.53295 .24777 Mdot
.53295 .25313 Mdot
.53295 .28245 Mdot
.53295 .31183 Mdot
.53295 .34117 Mdot
.53295 .38949 Mdot
.53295 .40867 Mdot
.53347 .12465 Mdot
.53347 .13588 Mdot
.53347 .14707 Mdot
.53347 .19867 Mdot
.53347 .25027 Mdot
.53347 .28002 Mdot
.53347 .30986 Mdot
.53347 .33966 Mdot
.53347 .38874 Mdot
.53347 .40821 Mdot
.53374 .16517 Mdot
.53374 .24331 Mdot
.53374 .40798 Mdot
.53401 .10838 Mdot
.53401 .11981 Mdot
.53401 .1312 Mdot
.53401 .19492 Mdot
.53401 .23622 Mdot
.53401 .27753 Mdot
.53401 .30783 Mdot
.53401 .3381 Mdot
.53401 .38796 Mdot
.53401 .40775 Mdot
.53456 .06832 Mdot
.53456 .07997 Mdot
.53456 .0916 Mdot
.53456 .10322 Mdot
.53456 .11482 Mdot
.53456 .15869 Mdot
.53456 .17959 Mdot
.53456 .19106 Mdot
.53456 .23299 Mdot
.53456 .23864 Mdot
.53456 .27495 Mdot
.53456 .30573 Mdot
.53456 .3365 Mdot
.53456 .38716 Mdot
.53456 .40726 Mdot
.53513 .0625 Mdot
.53513 .07431 Mdot
.53513 .08612 Mdot
.53513 .17542 Mdot
.53513 .18708 Mdot
.53513 .22966 Mdot
.53513 .27229 Mdot
.53513 .30357 Mdot
.53513 .33483 Mdot
.53513 .38633 Mdot
.53513 .40676 Mdot
.53542 .13417 Mdot
.53542 .23375 Mdot
.53542 .32843 Mdot
.53542 .40651 Mdot
.53572 .17111 Mdot
.53572 .22622 Mdot
.53572 .26955 Mdot
.53572 .30133 Mdot
.53572 .33312 Mdot
.53572 .38547 Mdot
.53572 .40625 Mdot
.53633 .07169 Mdot
.53633 .09005 Mdot
.53633 .10838 Mdot
.53633 .15456 Mdot
.53633 .16668 Mdot
.53633 .22267 Mdot
.53633 .26671 Mdot
.53633 .27702 Mdot
.53633 .29902 Mdot
.53633 .32565 Mdot
.53633 .33135 Mdot
.53633 .38458 Mdot
.53633 .40571 Mdot
.53696 .13741 Mdot
.53696 .14978 Mdot
.53696 .20683 Mdot
.53696 .219 Mdot
.53696 .29663 Mdot
.53696 .32951 Mdot
.53696 .38367 Mdot
.53696 .40516 Mdot
.53728 .0625 Mdot
.53728 .20484 Mdot
.53728 .27283 Mdot
.53728 .32271 Mdot
.53728 .40488 Mdot
.53761 .11963 Mdot
.53761 .13225 Mdot
.53761 .14483 Mdot
.53761 .20282 Mdot
.53761 .24854 Mdot
.53761 .29416 Mdot
.53761 .32761 Mdot
.53761 .38272 Mdot
.53761 .40459 Mdot
.53829 .08833 Mdot
.53829 .10123 Mdot
.53829 .1141 Mdot
.53829 .12693 Mdot
.53829 .13522 Mdot
.53829 .14483 Mdot
.53829 .15442 Mdot
.53829 .19867 Mdot
.53829 .23354 Mdot
.53829 .24291 Mdot
.53829 .24517 Mdot
.53829 .26842 Mdot
.53829 .2916 Mdot
.53829 .29401 Mdot
.53829 .31962 Mdot
.53829 .34519 Mdot
.53829 .38174 Mdot
.53829 .38729 Mdot
.53829 .404 Mdot
.53881 .12144 Mdot
.53881 .1312 Mdot
.53881 .14095 Mdot
.53881 .18581 Mdot
.53881 .19546 Mdot
.53881 .23078 Mdot
.53881 .26613 Mdot
.53881 .29206 Mdot
.53881 .31802 Mdot
.53881 .34393 Mdot
.53881 .38661 Mdot
.53881 .40354 Mdot
.53899 .06906 Mdot
.53899 .08219 Mdot
.53899 .09529 Mdot
.53899 .19438 Mdot
.53899 .24169 Mdot
.53899 .40338 Mdot
.53935 .09735 Mdot
.53935 .10728 Mdot
.53935 .11719 Mdot
.53935 .12709 Mdot
.53935 .18239 Mdot
.53935 .19217 Mdot
.53935 .22795 Mdot
.53935 .26378 Mdot
.53935 .29006 Mdot
.53935 .31637 Mdot
.53935 .34264 Mdot
.53935 .3859 Mdot
.53935 .40307 Mdot
.53972 .17676 Mdot
.53972 .23808 Mdot
.53972 .40275 Mdot
.5399 .0625 Mdot
.5399 .07258 Mdot
.5399 .08267 Mdot
.5399 .09274 Mdot
.5399 .1028 Mdot
.5399 .11284 Mdot
.5399 .17888 Mdot
.5399 .22505 Mdot
.5399 .26136 Mdot
.5399 .288 Mdot
.5399 .31467 Mdot
.5399 .34131 Mdot
.5399 .38518 Mdot
.5399 .40258 Mdot
.54047 .0676 Mdot
.54047 .07781 Mdot
.54047 .08801 Mdot
.54047 .16517 Mdot
.54047 .17191 Mdot
.54047 .17527 Mdot
.54047 .22206 Mdot
.54047 .23434 Mdot
.54047 .25887 Mdot
.54047 .28588 Mdot
.54047 .31292 Mdot
.54047 .33994 Mdot
.54047 .34264 Mdot
.54047 .38443 Mdot
.54047 .40209 Mdot
.54106 .16132 Mdot
.54106 .17157 Mdot
.54106 .20886 Mdot
.54106 .219 Mdot
.54106 .25632 Mdot
.54106 .2837 Mdot
.54106 .31113 Mdot
.54106 .33853 Mdot
.54106 .38367 Mdot
.54106 .40157 Mdot
.54126 .15312 Mdot
.54126 .16688 Mdot
.54126 .23045 Mdot
.54126 .3408 Mdot
.54126 .4014 Mdot
.54166 .14693 Mdot
.54166 .15736 Mdot
.54166 .20556 Mdot
.54166 .24353 Mdot
.54166 .28146 Mdot
.54166 .30928 Mdot
.54166 .33709 Mdot
.54166 .38288 Mdot
.54166 .40105 Mdot
.54207 .13356 Mdot
.54207 .14765 Mdot
.54207 .2126 Mdot
.54207 .26378 Mdot
.54207 .33889 Mdot
.54207 .40069 Mdot
.54228 .13209 Mdot
.54228 .14271 Mdot
.54228 .1533 Mdot
.54228 .20216 Mdot
.54228 .24067 Mdot
.54228 .27915 Mdot
.54228 .30738 Mdot
.54228 .3356 Mdot
.54228 .38207 Mdot
.54228 .4005 Mdot
.54292 .1132 Mdot
.54292 .1168 Mdot
.54292 .12761 Mdot
.54292 .13839 Mdot
.54292 .188 Mdot
.54292 .19867 Mdot
.54292 .20814 Mdot
.54292 .23773 Mdot
.54292 .2603 Mdot
.54292 .27678 Mdot
.54292 .30543 Mdot
.54292 .33406 Mdot
.54292 .3369 Mdot
.54292 .35267 Mdot
.54292 .38123 Mdot
.54292 .39994 Mdot
.54358 .07904 Mdot
.54358 .09005 Mdot
.54358 .10106 Mdot
.54358 .11204 Mdot
.54358 .123 Mdot
.54358 .18425 Mdot
.54358 .19507 Mdot
.54358 .2347 Mdot
.54358 .27433 Mdot
.54358 .30341 Mdot
.54358 .33248 Mdot
.54358 .35137 Mdot
.54358 .38037 Mdot
.54358 .39937 Mdot
.5438 .0625 Mdot
.5438 .07727 Mdot
.5438 .09202 Mdot
.5438 .10675 Mdot
.5438 .12144 Mdot
.5438 .20349 Mdot
.5438 .25669 Mdot
.5438 .33483 Mdot
.5438 .39917 Mdot
.54426 .0625 Mdot
.54426 .07367 Mdot
.54426 .08484 Mdot
.54426 .096 Mdot
.54426 .10714 Mdot
.54426 .18038 Mdot
.54426 .23158 Mdot
.54426 .27181 Mdot
.54426 .30133 Mdot
.54426 .33085 Mdot
.54426 .35003 Mdot
.54426 .37948 Mdot
.54426 .39877 Mdot
.54472 .07002 Mdot
.54472 .08505 Mdot
.54472 .19867 Mdot
.54472 .25293 Mdot
.54472 .33268 Mdot
.54472 .39837 Mdot
.54496 .16517 Mdot
.54496 .17639 Mdot
.54496 .22836 Mdot
.54496 .29919 Mdot
.54496 .34864 Mdot
.54496 .37856 Mdot
.54496 .39816 Mdot
.54568 .16088 Mdot
.54568 .17229 Mdot
.54568 .1785 Mdot
.54568 .21379 Mdot
.54568 .22505 Mdot
.54568 .24901 Mdot
.54568 .25545 Mdot
.54568 .28969 Mdot
.54568 .29698 Mdot
.54568 .33044 Mdot
.54568 .34721 Mdot
.54568 .37762 Mdot
.54568 .39753 Mdot
.54643 .14483 Mdot
.54643 .15647 Mdot
.54643 .21019 Mdot
.54643 .2525 Mdot
.54643 .2947 Mdot
.54643 .34574 Mdot
.54643 .39687 Mdot
.54669 .17294 Mdot
.54669 .24493 Mdot
.54669 .28646 Mdot
.54669 .32809 Mdot
.54669 .39665 Mdot
.5472 .12823 Mdot
.5472 .1401 Mdot
.5472 .20647 Mdot
.5472 .24945 Mdot
.5472 .34421 Mdot
.5472 .3962 Mdot
.54752 .07212 Mdot
.54752 .08173 Mdot
.54752 .09133 Mdot
.54752 .10093 Mdot
.54752 .11051 Mdot
.54752 .12007 Mdot
.54752 .17349 Mdot
.54752 .18297 Mdot
.54752 .21755 Mdot
.54752 .22692 Mdot
.54752 .25223 Mdot
.54752 .26147 Mdot
.54752 .28683 Mdot
.54752 .31221 Mdot
.54752 .33757 Mdot
.54752 .35406 Mdot
.54752 .37935 Mdot
.54752 .39592 Mdot
.54773 .15123 Mdot
.54773 .16715 Mdot
.54773 .28309 Mdot
.54773 .32565 Mdot
.54773 .39573 Mdot
.548 .11107 Mdot
.548 .12316 Mdot
.548 .13522 Mdot
.548 .20263 Mdot
.548 .24631 Mdot
.548 .34264 Mdot
.548 .3955 Mdot
.54817 .06736 Mdot
.54817 .07709 Mdot
.54817 .08681 Mdot
.54817 .09653 Mdot
.54817 .10623 Mdot
.54817 .16998 Mdot
.54817 .17959 Mdot
.54817 .21461 Mdot
.54817 .2241 Mdot
.54817 .24973 Mdot
.54817 .2848 Mdot
.54817 .31052 Mdot
.54817 .33622 Mdot
.54817 .35293 Mdot
.54817 .37856 Mdot
.54817 .39536 Mdot
.54883 .0625 Mdot
.54883 .06867 Mdot
.54883 .07234 Mdot
.54883 .08101 Mdot
.54883 .08219 Mdot
.54883 .09334 Mdot
.54883 .10565 Mdot
.54883 .11793 Mdot
.54883 .14483 Mdot
.54883 .15661 Mdot
.54883 .16638 Mdot
.54883 .18655 Mdot
.54883 .19867 Mdot
.54883 .21159 Mdot
.54883 .24306 Mdot
.54883 .24718 Mdot
.54883 .27958 Mdot
.54883 .28271 Mdot
.54883 .30878 Mdot
.54883 .32309 Mdot
.54883 .33483 Mdot
.54883 .34101 Mdot
.54883 .35177 Mdot
.54883 .37775 Mdot
.54883 .39478 Mdot
.54951 .15279 Mdot
.54951 .16269 Mdot
.54951 .20849 Mdot
.54951 .24455 Mdot
.54951 .28057 Mdot
.54951 .307 Mdot
.54951 .33341 Mdot
.54951 .35057 Mdot
.54951 .37692 Mdot
.54951 .39418 Mdot
.54968 .0625 Mdot
.54968 .07503 Mdot
.54968 .08755 Mdot
.54968 .18224 Mdot
.54968 .2397 Mdot
.54968 .27283 Mdot
.54968 .33932 Mdot
.54968 .39403 Mdot
.54997 .10464 Mdot
.54997 .12144 Mdot
.54997 .2152 Mdot
.54997 .2759 Mdot
.54997 .32041 Mdot
.54997 .39377 Mdot
.55021 .13878 Mdot
.55021 .14886 Mdot
.55021 .2053 Mdot
.55021 .24185 Mdot
.55021 .27837 Mdot
.55021 .30516 Mdot
.55021 .33194 Mdot
.55021 .34934 Mdot
.55021 .37607 Mdot
.55021 .39357 Mdot
.55057 .16517 Mdot
.55057 .17779 Mdot
.55057 .23622 Mdot
.55057 .26992 Mdot
.55057 .33757 Mdot
.55057 .39325 Mdot
.55093 .12437 Mdot
.55093 .13462 Mdot
.55093 .14483 Mdot
.55093 .19193 Mdot
.55093 .20203 Mdot
.55093 .23908 Mdot
.55093 .27611 Mdot
.55093 .30327 Mdot
.55093 .34808 Mdot
.55093 .37519 Mdot
.55093 .39294 Mdot
.55118 .0625 Mdot
.55118 .07973 Mdot
.55118 .09693 Mdot
.55118 .20994 Mdot
.55118 .27205 Mdot
.55118 .31761 Mdot
.55118 .39272 Mdot
.55148 .16034 Mdot
.55148 .17319 Mdot
.55148 .21998 Mdot
.55148 .2669 Mdot
.55148 .33576 Mdot
.55148 .39245 Mdot
.55167 .09911 Mdot
.55167 .10954 Mdot
.55167 .11995 Mdot
.55167 .13034 Mdot
.55167 .18842 Mdot
.55167 .23622 Mdot
.55167 .26378 Mdot
.55167 .30133 Mdot
.55167 .34678 Mdot
.55167 .37428 Mdot
.55167 .39229 Mdot
.55244 .0625 Mdot
.55244 .0731 Mdot
.55244 .0837 Mdot
.55244 .09429 Mdot
.55244 .10486 Mdot
.55244 .11542 Mdot
.55244 .1422 Mdot
.55244 .15535 Mdot
.55244 .17435 Mdot
.55244 .18481 Mdot
.55244 .20444 Mdot
.55244 .21601 Mdot
.55244 .22298 Mdot
.55244 .23329 Mdot
.55244 .26123 Mdot
.55244 .26378 Mdot
.55244 .26801 Mdot
.55244 .29883 Mdot
.55244 .29934 Mdot
.55244 .31467 Mdot
.55244 .33389 Mdot
.55244 .34544 Mdot
.55244 .37335 Mdot
.55244 .39162 Mdot
.55322 .06787 Mdot
.55322 .07861 Mdot
.55322 .08934 Mdot
.55322 .17049 Mdot
.55322 .1811 Mdot
.55322 .21981 Mdot
.55322 .25861 Mdot
.55322 .34406 Mdot
.55322 .39093 Mdot
.55342 .12336 Mdot
.55342 .1368 Mdot
.55342 .21191 Mdot
.55342 .26054 Mdot
.55342 .29623 Mdot
.55342 .33194 Mdot
.55342 .39076 Mdot
.55376 .18087 Mdot
.55376 .26378 Mdot
.55376 .31158 Mdot
.55376 .39046 Mdot
.55403 .16651 Mdot
.55403 .21654 Mdot
.55403 .25591 Mdot
.55403 .34264 Mdot
.55403 .39022 Mdot
.55445 .1038 Mdot
.55445 .11752 Mdot
.55445 .1312 Mdot
.55445 .20766 Mdot
.55445 .25718 Mdot
.55445 .29353 Mdot
.55445 .32992 Mdot
.55445 .38986 Mdot
.55487 .15144 Mdot
.55487 .16243 Mdot
.55487 .21317 Mdot
.55487 .25313 Mdot
.55487 .34117 Mdot
.55487 .38949 Mdot
.55515 .17435 Mdot
.55515 .30834 Mdot
.55515 .38924 Mdot
.55551 .06951 Mdot
.55551 .08352 Mdot
.55551 .09751 Mdot
.55551 .11148 Mdot
.55551 .18948 Mdot
.55551 .20324 Mdot
.55551 .25369 Mdot
.55551 .29073 Mdot
.55551 .32783 Mdot
.55551 .38893 Mdot
.55573 .13588 Mdot
.55573 .14707 Mdot
.55573 .20971 Mdot
.55573 .25027 Mdot
.55573 .28002 Mdot
.55573 .33966 Mdot
.55573 .38874 Mdot
.55662 .11981 Mdot
.55662 .1312 Mdot
.55662 .14257 Mdot
.55662 .16752 Mdot
.55662 .18463 Mdot
.55662 .19492 Mdot
.55662 .20614 Mdot
.55662 .24731 Mdot
.55662 .25007 Mdot
.55662 .27753 Mdot
.55662 .28782 Mdot
.55662 .30493 Mdot
.55662 .32565 Mdot
.55662 .3381 Mdot
.55662 .38796 Mdot
.55738 .14483 Mdot
.55738 .15442 Mdot
.55738 .19867 Mdot
.55738 .20814 Mdot
.55738 .24291 Mdot
.55738 .26842 Mdot
.55738 .30315 Mdot
.55738 .34519 Mdot
.55738 .37064 Mdot
.55738 .38729 Mdot
.55754 .0916 Mdot
.55754 .10322 Mdot
.55754 .11482 Mdot
.55754 .12639 Mdot
.55754 .19106 Mdot
.55754 .24427 Mdot
.55754 .27495 Mdot
.55754 .30573 Mdot
.55754 .3365 Mdot
.55754 .38716 Mdot
.55777 .17959 Mdot
.55777 .2848 Mdot
.55777 .32338 Mdot
.55777 .38695 Mdot
.55816 .1312 Mdot
.55816 .14095 Mdot
.55816 .15066 Mdot
.55816 .19546 Mdot
.55816 .20506 Mdot
.55816 .24029 Mdot
.55816 .26613 Mdot
.55816 .30133 Mdot
.55816 .34393 Mdot
.55816 .36973 Mdot
.55816 .38661 Mdot
.55848 .0625 Mdot
.55848 .07431 Mdot
.55848 .08612 Mdot
.55848 .09791 Mdot
.55848 .10969 Mdot
.55848 .18708 Mdot
.55848 .24113 Mdot
.55848 .27229 Mdot
.55848 .30357 Mdot
.55848 .33483 Mdot
.55848 .38633 Mdot
.55897 .11719 Mdot
.55897 .12709 Mdot
.55897 .13696 Mdot
.55897 .15964 Mdot
.55897 .17435 Mdot
.55897 .19217 Mdot
.55897 .22795 Mdot
.55897 .2376 Mdot
.55897 .26378 Mdot
.55897 .28165 Mdot
.55897 .32102 Mdot
.55897 .34264 Mdot
.55897 .3859 Mdot
.55946 .06849 Mdot
.55946 .08047 Mdot
.55946 .17111 Mdot
.55946 .18297 Mdot
.55946 .22622 Mdot
.55946 .26955 Mdot
.55946 .30133 Mdot
.55946 .33312 Mdot
.55946 .38547 Mdot
.5598 .09274 Mdot
.5598 .1028 Mdot
.5598 .11284 Mdot
.5598 .12287 Mdot
.5598 .17888 Mdot
.5598 .1888 Mdot
.5598 .22505 Mdot
.5598 .26136 Mdot
.5598 .34131 Mdot
.5598 .38518 Mdot
.56022 .15389 Mdot
.56022 .22357 Mdot
.56022 .27837 Mdot
.56022 .31856 Mdot
.56022 .38481 Mdot
.56048 .16668 Mdot
.56048 .22267 Mdot
.56048 .26671 Mdot
.56048 .29902 Mdot
.56048 .33135 Mdot
.56048 .38458 Mdot
.56065 .0676 Mdot
.56065 .07781 Mdot
.56065 .08801 Mdot
.56065 .0982 Mdot
.56065 .10838 Mdot
.56065 .17527 Mdot
.56065 .18533 Mdot
.56065 .22206 Mdot
.56065 .25887 Mdot
.56065 .28588 Mdot
.56065 .31292 Mdot
.56065 .33994 Mdot
.56065 .38443 Mdot
.56152 .0625 Mdot
.56152 .07284 Mdot
.56152 .08317 Mdot
.56152 .13249 Mdot
.56152 .14978 Mdot
.56152 .16132 Mdot
.56152 .16209 Mdot
.56152 .17157 Mdot
.56152 .219 Mdot
.56152 .25632 Mdot
.56152 .26378 Mdot
.56152 .27495 Mdot
.56152 .2837 Mdot
.56152 .29663 Mdot
.56152 .31113 Mdot
.56152 .31599 Mdot
.56152 .32951 Mdot
.56152 .33853 Mdot
.56152 .38367 Mdot
.56242 .15736 Mdot
.56242 .16776 Mdot
.56242 .21585 Mdot
.56242 .25369 Mdot
.56242 .28146 Mdot
.56242 .30928 Mdot
.56242 .33709 Mdot
.56242 .38288 Mdot
.56261 .13225 Mdot
.56261 .14483 Mdot
.56261 .2152 Mdot
.56261 .26074 Mdot
.56261 .29416 Mdot
.56261 .32761 Mdot
.56261 .38272 Mdot
.56288 .0784 Mdot
.56288 .09429 Mdot
.56288 .11014 Mdot
.56288 .12595 Mdot
.56288 .21424 Mdot
.56288 .27139 Mdot
.56288 .31332 Mdot
.56288 .38248 Mdot
.56335 .14271 Mdot
.56335 .1533 Mdot
.56335 .20216 Mdot
.56335 .2126 Mdot
.56335 .25099 Mdot
.56335 .27915 Mdot
.56335 .30738 Mdot
.56335 .3356 Mdot
.56335 .38207 Mdot
.56373 .1141 Mdot
.56373 .12693 Mdot
.56373 .13973 Mdot
.56373 .19867 Mdot
.56373 .21128 Mdot
.56373 .2576 Mdot
.56373 .2916 Mdot
.56373 .32565 Mdot
.56373 .38174 Mdot
.56431 .07061 Mdot
.56431 .08681 Mdot
.56431 .12761 Mdot
.56431 .13839 Mdot
.56431 .14913 Mdot
.56431 .19867 Mdot
.56431 .24821 Mdot
.56431 .27678 Mdot
.56431 .30543 Mdot
.56431 .31052 Mdot
.56431 .33406 Mdot
.56431 .38123 Mdot
.5649 .08219 Mdot
.5649 .09529 Mdot
.5649 .10838 Mdot
.5649 .12144 Mdot
.5649 .19438 Mdot
.5649 .28895 Mdot
.5649 .32361 Mdot
.5649 .38072 Mdot
.56529 .11204 Mdot
.56529 .123 Mdot
.56529 .13393 Mdot
.56529 .19507 Mdot
.56529 .2347 Mdot
.56529 .24534 Mdot
.56529 .27433 Mdot
.56529 .30341 Mdot
.56529 .33248 Mdot
.56529 .38037 Mdot
.5658 .18779 Mdot
.5658 .3076 Mdot
.5658 .37993 Mdot
.5661 .0625 Mdot
.5661 .07584 Mdot
.5661 .08916 Mdot
.5661 .10247 Mdot
.5661 .18993 Mdot
.5661 .28621 Mdot
.5661 .3215 Mdot
.5661 .37966 Mdot
.56631 .07367 Mdot
.56631 .08484 Mdot
.56631 .096 Mdot
.56631 .10714 Mdot
.56631 .11826 Mdot
.56631 .18038 Mdot
.56631 .19137 Mdot
.56631 .23158 Mdot
.56631 .27181 Mdot
.56631 .30133 Mdot
.56631 .33085 Mdot
.56631 .37948 Mdot
.56646 .11051 Mdot
.56646 .12007 Mdot
.56646 .12962 Mdot
.56646 .18297 Mdot
.56646 .1924 Mdot
.56646 .22692 Mdot
.56646 .26147 Mdot
.56646 .28683 Mdot
.56646 .31221 Mdot
.56646 .33757 Mdot
.56646 .37935 Mdot
.56736 .06736 Mdot
.56736 .06816 Mdot
.56736 .07709 Mdot
.56736 .07949 Mdot
.56736 .09081 Mdot
.56736 .09653 Mdot
.56736 .10211 Mdot
.56736 .10623 Mdot
.56736 .11591 Mdot
.56736 .17191 Mdot
.56736 .17639 Mdot
.56736 .17959 Mdot
.56736 .18198 Mdot
.56736 .18533 Mdot
.56736 .18756 Mdot
.56736 .2241 Mdot
.56736 .22836 Mdot
.56736 .23434 Mdot
.56736 .25911 Mdot
.56736 .26922 Mdot
.56736 .28336 Mdot
.56736 .29919 Mdot
.56736 .30454 Mdot
.56736 .31931 Mdot
.56736 .32917 Mdot
.56736 .33622 Mdot
.56736 .37856 Mdot
.56828 .0625 Mdot
.56828 .07234 Mdot
.56828 .08219 Mdot
.56828 .09202 Mdot
.56828 .10184 Mdot
.56828 .16638 Mdot
.56828 .17612 Mdot
.56828 .22121 Mdot
.56828 .25669 Mdot
.56828 .28271 Mdot
.56828 .30878 Mdot
.56828 .33483 Mdot
.56828 .37775 Mdot
.56844 .17229 Mdot
.56844 .22505 Mdot
.56844 .26654 Mdot
.56844 .29698 Mdot
.56844 .32744 Mdot
.56844 .37762 Mdot
.56866 .16688 Mdot
.56866 .23045 Mdot
.56866 .2804 Mdot
.56866 .31703 Mdot
.56866 .37742 Mdot
.56899 .15869 Mdot
.56899 .17591 Mdot
.56899 .23864 Mdot
.56899 .37713 Mdot
.56923 .06748 Mdot
.56923 .07744 Mdot
.56923 .16269 Mdot
.56923 .17256 Mdot
.56923 .20849 Mdot
.56923 .21825 Mdot
.56923 .2542 Mdot
.56923 .28057 Mdot
.56923 .307 Mdot
.56923 .33341 Mdot
.56923 .37692 Mdot
.56956 .15647 Mdot
.56956 .16806 Mdot
.56956 .22163 Mdot
.56956 .26378 Mdot
.56956 .2947 Mdot
.56956 .32565 Mdot
.56956 .37664 Mdot
.57001 .14765 Mdot
.57001 .16167 Mdot
.57001 .22643 Mdot
.57001 .27733 Mdot
.57001 .31467 Mdot
.57001 .37624 Mdot
.57021 .14886 Mdot
.57021 .15891 Mdot
.57021 .2053 Mdot
.57021 .2152 Mdot
.57021 .25164 Mdot
.57021 .27837 Mdot
.57021 .30516 Mdot
.57021 .33194 Mdot
.57021 .37607 Mdot
.57071 .1401 Mdot
.57071 .15192 Mdot
.57071 .21811 Mdot
.57071 .23375 Mdot
.57071 .26092 Mdot
.57071 .29234 Mdot
.57071 .3238 Mdot
.57071 .37563 Mdot
.57121 .13462 Mdot
.57121 .14483 Mdot
.57121 .15502 Mdot
.57121 .20203 Mdot
.57121 .23908 Mdot
.57121 .27611 Mdot
.57121 .30327 Mdot
.57121 .33044 Mdot
.57121 .37519 Mdot
.57142 .14197 Mdot
.57142 .22224 Mdot
.57142 .31221 Mdot
.57142 .37501 Mdot
.5719 .13522 Mdot
.5719 .14723 Mdot
.5719 .20263 Mdot
.5719 .28991 Mdot
.5719 .32189 Mdot
.5719 .37459 Mdot
.57225 .11995 Mdot
.57225 .13034 Mdot
.57225 .1407 Mdot
.57225 .19867 Mdot
.57225 .23622 Mdot
.57225 .27378 Mdot
.57225 .30133 Mdot
.57225 .32889 Mdot
.57225 .37428 Mdot
.57251 .12665 Mdot
.57251 .14483 Mdot
.57251 .22861 Mdot
.57251 .37405 Mdot
.57288 .12144 Mdot
.57288 .13608 Mdot
.57288 .2179 Mdot
.57288 .30966 Mdot
.57288 .37373 Mdot
.57313 .10565 Mdot
.57313 .11793 Mdot
.57313 .13019 Mdot
.57313 .19867 Mdot
.57313 .24306 Mdot
.57313 .2874 Mdot
.57313 .31991 Mdot
.57313 .37351 Mdot
.57331 .10486 Mdot
.57331 .11542 Mdot
.57331 .12595 Mdot
.57331 .18481 Mdot
.57331 .19521 Mdot
.57331 .23329 Mdot
.57331 .27139 Mdot
.57331 .29934 Mdot
.57331 .32729 Mdot
.57331 .37335 Mdot
.57441 .0625 Mdot
.57441 .06787 Mdot
.57441 .07002 Mdot
.57441 .07503 Mdot
.57441 .07861 Mdot
.57441 .08129 Mdot
.57441 .08505 Mdot
.57441 .08755 Mdot
.57441 .08934 Mdot
.57441 .10006 Mdot
.57441 .11076 Mdot
.57441 .11254 Mdot
.57441 .11503 Mdot
.57441 .11877 Mdot
.57441 .1811 Mdot
.57441 .19457 Mdot
.57441 .19867 Mdot
.57441 .22322 Mdot
.57441 .23027 Mdot
.57441 .2397 Mdot
.57441 .2848 Mdot
.57441 .29728 Mdot
.57441 .307 Mdot
.57441 .31787 Mdot
.57441 .37239 Mdot
.57554 .0625 Mdot
.57554 .07338 Mdot
.57554 .08426 Mdot
.57554 .09512 Mdot
.57554 .17729 Mdot
.57554 .22717 Mdot
.57554 .26639 Mdot
.57554 .29516 Mdot
.57554 .32395 Mdot
.57554 .3714 Mdot
.57573 .06886 Mdot
.57573 .08158 Mdot
.57573 .09429 Mdot
.57573 .17779 Mdot
.57573 .19034 Mdot
.57573 .23622 Mdot
.57573 .2821 Mdot
.57573 .31575 Mdot
.57573 .37123 Mdot
.576 .0625 Mdot
.576 .07781 Mdot
.576 .19365 Mdot
.576 .371 Mdot
.57641 .15442 Mdot
.57641 .16397 Mdot
.57641 .20814 Mdot
.57641 .24291 Mdot
.57641 .27766 Mdot
.57641 .30315 Mdot
.57641 .32864 Mdot
.57641 .37064 Mdot
.5767 .16243 Mdot
.5767 .17336 Mdot
.5767 .22397 Mdot
.5767 .26378 Mdot
.5767 .29298 Mdot
.5767 .32221 Mdot
.5767 .37038 Mdot
.5771 .17319 Mdot
.5771 .23263 Mdot
.5771 .27932 Mdot
.5771 .31357 Mdot
.5771 .37004 Mdot
.57745 .14095 Mdot
.57745 .15066 Mdot
.57745 .16034 Mdot
.57745 .20506 Mdot
.57745 .24029 Mdot
.57745 .2755 Mdot
.57745 .30133 Mdot
.57745 .32716 Mdot
.57745 .36973 Mdot
.57766 .18842 Mdot
.57766 .24493 Mdot
.57766 .36955 Mdot
.5779 .14707 Mdot
.5779 .15822 Mdot
.5779 .20971 Mdot
.5779 .22068 Mdot
.5779 .25027 Mdot
.5779 .29073 Mdot
.5779 .32041 Mdot
.5779 .36933 Mdot
.57852 .13696 Mdot
.57852 .1468 Mdot
.57852 .15535 Mdot
.57852 .16843 Mdot
.57852 .19217 Mdot
.57852 .20191 Mdot
.57852 .22891 Mdot
.57852 .2376 Mdot
.57852 .27328 Mdot
.57852 .29946 Mdot
.57852 .3113 Mdot
.57852 .32565 Mdot
.57852 .3688 Mdot
.57914 .14257 Mdot
.57914 .15389 Mdot
.57914 .20614 Mdot
.57914 .24731 Mdot
.57914 .28841 Mdot
.57914 .31856 Mdot
.57914 .36825 Mdot
.5794 .16715 Mdot
.5794 .18297 Mdot
.5794 .24067 Mdot
.5794 .36803 Mdot
.57962 .12287 Mdot
.57962 .13287 Mdot
.57962 .14284 Mdot
.57962 .1888 Mdot
.57962 .19867 Mdot
.57962 .23483 Mdot
.57962 .271 Mdot
.57962 .29754 Mdot
.57962 .32409 Mdot
.57962 .36783 Mdot
.57999 .15018 Mdot
.57999 .1635 Mdot
.57999 .22505 Mdot
.57999 .30894 Mdot
.57999 .36751 Mdot
.58042 .12639 Mdot
.58042 .13793 Mdot
.58042 .20246 Mdot
.58042 .24427 Mdot
.58042 .28602 Mdot
.58042 .31664 Mdot
.58042 .36713 Mdot
.58075 .0982 Mdot
.58075 .10838 Mdot
.58075 .11854 Mdot
.58075 .12867 Mdot
.58075 .18533 Mdot
.58075 .23199 Mdot
.58075 .26866 Mdot
.58075 .29556 Mdot
.58075 .32248 Mdot
.58075 .36684 Mdot
.58121 .16111 Mdot
.58121 .23622 Mdot
.58121 .36644 Mdot
.58152 .1312 Mdot
.58152 .14483 Mdot
.58152 .20766 Mdot
.58152 .36617 Mdot
.58174 .09791 Mdot
.58174 .10969 Mdot
.58174 .12144 Mdot
.58174 .13316 Mdot
.58174 .19867 Mdot
.58174 .24113 Mdot
.58174 .28355 Mdot
.58174 .31467 Mdot
.58174 .36597 Mdot
.58191 .0625 Mdot
.58191 .07284 Mdot
.58191 .08317 Mdot
.58191 .09349 Mdot
.58191 .1038 Mdot
.58191 .1141 Mdot
.58191 .17157 Mdot
.58191 .18177 Mdot
.58191 .22907 Mdot
.58191 .26625 Mdot
.58191 .29353 Mdot
.58191 .32083 Mdot
.58191 .36583 Mdot
.58311 .06773 Mdot
.58311 .06849 Mdot
.58311 .0782 Mdot
.58311 .08047 Mdot
.58311 .08866 Mdot
.58311 .09245 Mdot
.58311 .1044 Mdot
.58311 .11148 Mdot
.58311 .11633 Mdot
.58311 .12541 Mdot
.58311 .13817 Mdot
.58311 .15481 Mdot
.58311 .16776 Mdot
.58311 .17811 Mdot
.58311 .18297 Mdot
.58311 .19475 Mdot
.58311 .20324 Mdot
.58311 .21585 Mdot
.58311 .22607 Mdot
.58311 .23158 Mdot
.58311 .23788 Mdot
.58311 .25369 Mdot
.58311 .281 Mdot
.58311 .29144 Mdot
.58311 .31263 Mdot
.58311 .31913 Mdot
.58311 .36478 Mdot
.58434 .1533 Mdot
.58434 .16385 Mdot
.58434 .2126 Mdot
.58434 .25099 Mdot
.58434 .28929 Mdot
.58434 .31739 Mdot
.58434 .3637 Mdot
.58452 .0625 Mdot
.58452 .07466 Mdot
.58452 .08681 Mdot
.58452 .17873 Mdot
.58452 .23454 Mdot
.58452 .31052 Mdot
.58452 .36355 Mdot
.58476 .0625 Mdot
.58476 .07676 Mdot
.58476 .091 Mdot
.58476 .10522 Mdot
.58476 .19867 Mdot
.58476 .25007 Mdot
.58476 .36334 Mdot
.5851 .1141 Mdot
.5851 .1312 Mdot
.5851 .22672 Mdot
.5851 .36304 Mdot
.58532 .12962 Mdot
.58532 .13914 Mdot
.58532 .14863 Mdot
.58532 .1924 Mdot
.58532 .23622 Mdot
.58532 .29592 Mdot
.58532 .32119 Mdot
.58532 .36284 Mdot
.58561 .14913 Mdot
.58561 .15983 Mdot
.58561 .20927 Mdot
.58561 .24821 Mdot
.58561 .28708 Mdot
.58561 .31559 Mdot
.58561 .36259 Mdot
.58598 .16209 Mdot
.58598 .17435 Mdot
.58598 .23108 Mdot
.58598 .36227 Mdot
.58647 .06975 Mdot
.58647 .11591 Mdot
.58647 .12558 Mdot
.58647 .13522 Mdot
.58647 .18915 Mdot
.58647 .19391 Mdot
.58647 .23354 Mdot
.58647 .24631 Mdot
.58647 .25911 Mdot
.58647 .26842 Mdot
.58647 .29401 Mdot
.58647 .31962 Mdot
.58647 .36183 Mdot
.58692 .13393 Mdot
.58692 .14483 Mdot
.58692 .20584 Mdot
.58692 .24534 Mdot
.58692 .2848 Mdot
.58692 .31374 Mdot
.58692 .36145 Mdot
.58718 .0713 Mdot
.58718 .08888 Mdot
.58718 .10643 Mdot
.58718 .36122 Mdot
.58749 .15736 Mdot
.58749 .16983 Mdot
.58749 .2275 Mdot
.58749 .26074 Mdot
.58749 .36095 Mdot
.58766 .08219 Mdot
.58766 .09202 Mdot
.58766 .10184 Mdot
.58766 .11165 Mdot
.58766 .12144 Mdot
.58766 .17612 Mdot
.58766 .18581 Mdot
.58766 .22121 Mdot
.58766 .23078 Mdot
.58766 .25669 Mdot
.58766 .29206 Mdot
.58766 .31802 Mdot
.58766 .3608 Mdot
.58826 .11826 Mdot
.58826 .12936 Mdot
.58826 .14042 Mdot
.58826 .17435 Mdot
.58826 .19137 Mdot
.58826 .20231 Mdot
.58826 .2424 Mdot
.58826 .31183 Mdot
.58826 .36027 Mdot
.58888 .06748 Mdot
.58888 .07744 Mdot
.58888 .0874 Mdot
.58888 .09735 Mdot
.58888 .10728 Mdot
.58888 .17256 Mdot
.58888 .18239 Mdot
.58888 .21825 Mdot
.58888 .22795 Mdot
.58888 .2542 Mdot
.58888 .29006 Mdot
.58888 .31637 Mdot
.58888 .35973 Mdot
.58905 .13973 Mdot
.58905 .15248 Mdot
.58905 .21128 Mdot
.58905 .2576 Mdot
.58905 .35958 Mdot
.58937 .19867 Mdot
.58937 .26378 Mdot
.58937 .3593 Mdot
.58965 .09081 Mdot
.58965 .10211 Mdot
.58965 .11339 Mdot
.58965 .12465 Mdot
.58965 .18756 Mdot
.58965 .19867 Mdot
.58965 .23936 Mdot
.58965 .30986 Mdot
.58965 .35905 Mdot
.59013 .0625 Mdot
.59013 .07258 Mdot
.59013 .08267 Mdot
.59013 .15891 Mdot
.59013 .16891 Mdot
.59013 .2152 Mdot
.59013 .23832 Mdot
.59013 .25164 Mdot
.59013 .288 Mdot
.59013 .31467 Mdot
.59013 .35864 Mdot
.59067 .12144 Mdot
.59067 .13445 Mdot
.59067 .14742 Mdot
.59067 .20722 Mdot
.59067 .25436 Mdot
.59067 .35816 Mdot
.59109 .0625 Mdot
.59109 .07399 Mdot
.59109 .08546 Mdot
.59109 .09693 Mdot
.59109 .10838 Mdot
.59109 .18363 Mdot
.59109 .23622 Mdot
.59109 .26654 Mdot
.59109 .3578 Mdot
.59141 .15502 Mdot
.59141 .16517 Mdot
.59141 .21208 Mdot
.59141 .24901 Mdot
.59141 .28588 Mdot
.59141 .31292 Mdot
.59141 .35751 Mdot
.59167 .19262 Mdot
.59167 .25933 Mdot
.59167 .35729 Mdot
.59207 .14791 Mdot
.59207 .16325 Mdot
.59207 .23408 Mdot
.59207 .35694 Mdot
.59235 .10247 Mdot
.59235 .11576 Mdot
.59235 .129 Mdot
.59235 .20302 Mdot
.59235 .25099 Mdot
.59235 .35669 Mdot
.59257 .06832 Mdot
.59257 .07997 Mdot
.59257 .16806 Mdot
.59257 .17959 Mdot
.59257 .23299 Mdot
.59257 .26378 Mdot
.59257 .3565 Mdot
.59273 .1407 Mdot
.59273 .15103 Mdot
.59273 .16132 Mdot
.59273 .20886 Mdot
.59273 .24631 Mdot
.59273 .31113 Mdot
.59273 .35636 Mdot
.5941 .06928 Mdot
.5941 .08283 Mdot
.5941 .09637 Mdot
.5941 .10988 Mdot
.5941 .12595 Mdot
.5941 .14169 Mdot
.5941 .14693 Mdot
.5941 .1637 Mdot
.5941 .17542 Mdot
.5941 .18626 Mdot
.5941 .19521 Mdot
.5941 .19867 Mdot
.5941 .20556 Mdot
.5941 .21811 Mdot
.5941 .24353 Mdot
.5941 .2475 Mdot
.5941 .25466 Mdot
.5941 .26092 Mdot
.5941 .27139 Mdot
.5941 .35517 Mdot
.59536 .0625 Mdot
.59536 .07212 Mdot
.59536 .08173 Mdot
.59536 .09133 Mdot
.59536 .10093 Mdot
.59536 .11051 Mdot
.59536 .16397 Mdot
.59536 .17349 Mdot
.59536 .21755 Mdot
.59536 .25223 Mdot
.59536 .28683 Mdot
.59536 .31221 Mdot
.59536 .32864 Mdot
.59536 .35406 Mdot
.5955 .11076 Mdot
.5955 .12144 Mdot
.5955 .13209 Mdot
.5955 .19166 Mdot
.5955 .24067 Mdot
.5955 .26892 Mdot
.5955 .35394 Mdot
.59568 .14723 Mdot
.59568 .1592 Mdot
.59568 .21447 Mdot
.59568 .25798 Mdot
.59568 .35378 Mdot
.59591 .18056 Mdot
.59591 .24388 Mdot
.59591 .35358 Mdot
.59622 .103 Mdot
.59622 .11914 Mdot
.59622 .20927 Mdot
.59622 .26767 Mdot
.59622 .35331 Mdot
.59665 .06736 Mdot
.59665 .07709 Mdot
.59665 .08681 Mdot
.59665 .16034 Mdot
.59665 .16998 Mdot
.59665 .21461 Mdot
.59665 .24973 Mdot
.59665 .2755 Mdot
.59665 .31052 Mdot
.59665 .32716 Mdot
.59665 .35293 Mdot
.59694 .08426 Mdot
.59694 .09512 Mdot
.59694 .10597 Mdot
.59694 .1168 Mdot
.59694 .12761 Mdot
.59694 .188 Mdot
.59694 .23773 Mdot
.59694 .26639 Mdot
.59694 .35267 Mdot
.59731 .1424 Mdot
.59731 .15456 Mdot
.59731 .21071 Mdot
.59731 .25494 Mdot
.59731 .35235 Mdot
.59779 .17562 Mdot
.59779 .24012 Mdot
.59779 .27733 Mdot
.59779 .35193 Mdot
.59798 .1468 Mdot
.59798 .15661 Mdot
.59798 .16638 Mdot
.59798 .20191 Mdot
.59798 .21159 Mdot
.59798 .24718 Mdot
.59798 .27328 Mdot
.59798 .32565 Mdot
.59798 .35177 Mdot
.59843 .0625 Mdot
.59843 .06801 Mdot
.59843 .07904 Mdot
.59843 .09005 Mdot
.59843 .09556 Mdot
.59843 .10106 Mdot
.59843 .17336 Mdot
.59843 .18425 Mdot
.59843 .20408 Mdot
.59843 .22397 Mdot
.59843 .2347 Mdot
.59843 .26378 Mdot
.59843 .35137 Mdot
.599 .12499 Mdot
.599 .13741 Mdot
.599 .20683 Mdot
.599 .25179 Mdot
.599 .35087 Mdot
.59935 .14284 Mdot
.59935 .15279 Mdot
.59935 .19867 Mdot
.59935 .20849 Mdot
.59935 .24455 Mdot
.59935 .271 Mdot
.59935 .32409 Mdot
.59935 .35057 Mdot
.59974 .15627 Mdot
.59974 .17049 Mdot
.59974 .23622 Mdot
.59974 .27414 Mdot
.59974 .35022 Mdot
.59997 .07367 Mdot
.59997 .16932 Mdot
.59997 .18038 Mdot
.59997 .22068 Mdot
.59997 .26109 Mdot
.59997 .32041 Mdot
.59997 .35003 Mdot
.60075 .09429 Mdot
.60075 .10697 Mdot
.60075 .11963 Mdot
.60075 .12867 Mdot
.60075 .13878 Mdot
.60075 .14886 Mdot
.60075 .19034 Mdot
.60075 .19534 Mdot
.60075 .19867 Mdot
.60075 .20282 Mdot
.60075 .24185 Mdot
.60075 .24854 Mdot
.60075 .25971 Mdot
.60075 .26866 Mdot
.60075 .2821 Mdot
.60075 .32248 Mdot
.60075 .34934 Mdot
.60155 .15389 Mdot
.60155 .16517 Mdot
.60155 .21729 Mdot
.60155 .25831 Mdot
.60155 .31856 Mdot
.60155 .34864 Mdot
.60178 .15066 Mdot
.60178 .27082 Mdot
.60178 .34844 Mdot
.6022 .1141 Mdot
.6022 .12437 Mdot
.6022 .13462 Mdot
.6022 .19193 Mdot
.6022 .22907 Mdot
.6022 .23908 Mdot
.6022 .26625 Mdot
.6022 .32083 Mdot
.6022 .34808 Mdot
.60257 .0625 Mdot
.60257 .07542 Mdot
.60257 .08833 Mdot
.60257 .10123 Mdot
.60257 .18597 Mdot
.60257 .24517 Mdot
.60257 .27932 Mdot
.60257 .31357 Mdot
.60257 .34776 Mdot
.60318 .14943 Mdot
.60318 .16088 Mdot
.60318 .193 Mdot
.60318 .21379 Mdot
.60318 .25545 Mdot
.60318 .28602 Mdot
.60318 .31664 Mdot
.60318 .34721 Mdot
.60368 .06773 Mdot
.60368 .0782 Mdot
.60368 .08866 Mdot
.60368 .09911 Mdot
.60368 .10954 Mdot
.60368 .11995 Mdot
.60368 .17811 Mdot
.60368 .18842 Mdot
.60368 .22607 Mdot
.60368 .26378 Mdot
.60368 .29144 Mdot
.60368 .31913 Mdot
.60368 .34678 Mdot
.6039 .12996 Mdot
.6039 .14483 Mdot
.6039 .21337 Mdot
.6039 .26737 Mdot
.6039 .307 Mdot
.6039 .34659 Mdot
.60444 .18144 Mdot
.60444 .24169 Mdot
.60444 .27643 Mdot
.60444 .3113 Mdot
.60444 .34611 Mdot
.60487 .13316 Mdot
.60487 .14483 Mdot
.60487 .21019 Mdot
.60487 .2525 Mdot
.60487 .28355 Mdot
.60487 .31467 Mdot
.60487 .34574 Mdot
.60522 .0625 Mdot
.60522 .0731 Mdot
.60522 .0837 Mdot
.60522 .09429 Mdot
.60522 .10486 Mdot
.60522 .17435 Mdot
.60522 .18481 Mdot
.60522 .22298 Mdot
.60522 .26123 Mdot
.60522 .28929 Mdot
.60522 .31739 Mdot
.60522 .34544 Mdot
.6055 .13522 Mdot
.6055 .14483 Mdot
.6055 .15442 Mdot
.6055 .19867 Mdot
.6055 .23354 Mdot
.6055 .24291 Mdot
.6055 .26842 Mdot
.6055 .29401 Mdot
.6055 .31962 Mdot
.6055 .34519 Mdot
.60574 .16956 Mdot
.60574 .25099 Mdot
.60574 .29797 Mdot
.60574 .34498 Mdot
.60611 .09311 Mdot
.60611 .10838 Mdot
.60611 .12361 Mdot
.60611 .20867 Mdot
.60611 .26378 Mdot
.60611 .30422 Mdot
.60611 .34465 Mdot
.60639 .1635 Mdot
.60639 .17676 Mdot
.60639 .27344 Mdot
.60639 .30894 Mdot
.60639 .34441 Mdot
.60661 .11633 Mdot
.60661 .12823 Mdot
.60661 .1401 Mdot
.60661 .19475 Mdot
.60661 .20647 Mdot
.60661 .24945 Mdot
.60661 .281 Mdot
.60661 .31263 Mdot
.60661 .34421 Mdot
.60679 .06787 Mdot
.60679 .17049 Mdot
.60679 .21981 Mdot
.60679 .25861 Mdot
.60679 .28708 Mdot
.60679 .31559 Mdot
.60679 .34406 Mdot
.60694 .12144 Mdot
.60694 .1312 Mdot
.60694 .14095 Mdot
.60694 .19546 Mdot
.60694 .23078 Mdot
.60694 .24029 Mdot
.60694 .26613 Mdot
.60694 .29206 Mdot
.60694 .31802 Mdot
.60694 .34393 Mdot
.60842 .0703 Mdot
.60842 .0859 Mdot
.60842 .08681 Mdot
.60842 .09895 Mdot
.60842 .10147 Mdot
.60842 .10728 Mdot
.60842 .11719 Mdot
.60842 .12316 Mdot
.60842 .12709 Mdot
.60842 .13696 Mdot
.60842 .1557 Mdot
.60842 .15841 Mdot
.60842 .16291 Mdot
.60842 .16651 Mdot
.60842 .18239 Mdot
.60842 .19072 Mdot
.60842 .20377 Mdot
.60842 .21654 Mdot
.60842 .22105 Mdot
.60842 .24631 Mdot
.60842 .25591 Mdot
.60842 .26003 Mdot
.60842 .27034 Mdot
.60842 .27837 Mdot
.60842 .2848 Mdot
.60842 .29006 Mdot
.60842 .29444 Mdot
.60842 .30133 Mdot
.60842 .3065 Mdot
.60842 .31052 Mdot
.60842 .31374 Mdot
.60842 .34264 Mdot
.60994 .07258 Mdot
.60994 .08267 Mdot
.60994 .09274 Mdot
.60994 .1028 Mdot
.60994 .11284 Mdot
.60994 .12287 Mdot
.60994 .17888 Mdot
.60994 .1888 Mdot
.60994 .22505 Mdot
.60994 .26136 Mdot
.60994 .288 Mdot
.60994 .31467 Mdot
.60994 .34131 Mdot
.61009 .14042 Mdot
.61009 .15144 Mdot
.61009 .21317 Mdot
.61009 .25313 Mdot
.61009 .28245 Mdot
.61009 .31183 Mdot
.61009 .34117 Mdot
.61028 .06867 Mdot
.61028 .08101 Mdot
.61028 .09334 Mdot
.61028 .10565 Mdot
.61028 .18655 Mdot
.61028 .24306 Mdot
.61028 .27565 Mdot
.61028 .30834 Mdot
.61028 .34101 Mdot
.61052 .1393 Mdot
.61052 .15312 Mdot
.61052 .21691 Mdot
.61052 .26712 Mdot
.61052 .30397 Mdot
.61052 .3408 Mdot
.61082 .19867 Mdot
.61082 .25612 Mdot
.61082 .29831 Mdot
.61082 .34053 Mdot
.61123 .13741 Mdot
.61123 .29073 Mdot
.61123 .34017 Mdot
.6115 .0676 Mdot
.6115 .07781 Mdot
.6115 .08801 Mdot
.6115 .0982 Mdot
.6115 .17527 Mdot
.6115 .22206 Mdot
.6115 .25887 Mdot
.6115 .28588 Mdot
.6115 .31292 Mdot
.6115 .33994 Mdot
.61182 .13588 Mdot
.61182 .14707 Mdot
.61182 .19867 Mdot
.61182 .20971 Mdot
.61182 .25027 Mdot
.61182 .28002 Mdot
.61182 .30986 Mdot
.61182 .33966 Mdot
.61221 .16983 Mdot
.61221 .18224 Mdot
.61221 .2275 Mdot
.61221 .27283 Mdot
.61221 .30609 Mdot
.61221 .33932 Mdot
.6127 .11941 Mdot
.6127 .13356 Mdot
.6127 .2126 Mdot
.6127 .26378 Mdot
.6127 .30133 Mdot
.6127 .33889 Mdot
.6131 .07284 Mdot
.6131 .16132 Mdot
.6131 .17157 Mdot
.6131 .219 Mdot
.6131 .25632 Mdot
.6131 .2837 Mdot
.6131 .31113 Mdot
.6131 .33853 Mdot
.61334 .17729 Mdot
.61334 .25204 Mdot
.61334 .29516 Mdot
.61334 .33833 Mdot
.6136 .10838 Mdot
.6136 .11981 Mdot
.6136 .1312 Mdot
.6136 .19492 Mdot
.6136 .24731 Mdot
.6136 .27753 Mdot
.6136 .30783 Mdot
.6136 .3381 Mdot
.6142 .10093 Mdot
.6142 .11051 Mdot
.6142 .12007 Mdot
.6142 .12962 Mdot
.6142 .16517 Mdot
.6142 .17779 Mdot
.6142 .18297 Mdot
.6142 .1924 Mdot
.6142 .2238 Mdot
.6142 .22692 Mdot
.6142 .26147 Mdot
.6142 .26992 Mdot
.6142 .28683 Mdot
.6142 .30375 Mdot
.6142 .31221 Mdot
.6142 .33757 Mdot
.61476 .15736 Mdot
.61476 .16776 Mdot
.61476 .20556 Mdot
.61476 .21585 Mdot
.61476 .25369 Mdot
.61476 .28146 Mdot
.61476 .30928 Mdot
.61476 .33709 Mdot
.61497 .09874 Mdot
.61497 .1132 Mdot
.61497 .12761 Mdot
.61497 .20814 Mdot
.61497 .2603 Mdot
.61497 .29859 Mdot
.61497 .3369 Mdot
.61543 .06832 Mdot
.61543 .07997 Mdot
.61543 .0916 Mdot
.61543 .10322 Mdot
.61543 .11482 Mdot
.61543 .12639 Mdot
.61543 .19106 Mdot
.61543 .23299 Mdot
.61543 .24427 Mdot
.61543 .27495 Mdot
.61543 .30573 Mdot
.61543 .3365 Mdot
.61575 .06736 Mdot
.61575 .07709 Mdot
.61575 .08681 Mdot
.61575 .09653 Mdot
.61575 .10623 Mdot
.61575 .11591 Mdot
.61575 .17959 Mdot
.61575 .2241 Mdot
.61575 .25911 Mdot
.61575 .2848 Mdot
.61575 .31052 Mdot
.61575 .33622 Mdot
.61597 .17136 Mdot
.61597 .24777 Mdot
.61597 .29186 Mdot
.61597 .33603 Mdot
.61627 .14742 Mdot
.61627 .16034 Mdot
.61627 .21998 Mdot
.61627 .2669 Mdot
.61627 .30133 Mdot
.61627 .33576 Mdot
.61646 .14271 Mdot
.61646 .1533 Mdot
.61646 .20216 Mdot
.61646 .2126 Mdot
.61646 .25099 Mdot
.61646 .27915 Mdot
.61646 .30738 Mdot
.61646 .3356 Mdot
.61733 .0625 Mdot
.61733 .07234 Mdot
.61733 .07431 Mdot
.61733 .07727 Mdot
.61733 .08219 Mdot
.61733 .08612 Mdot
.61733 .09202 Mdot
.61733 .09791 Mdot
.61733 .10184 Mdot
.61733 .10675 Mdot
.61733 .16638 Mdot
.61733 .17542 Mdot
.61733 .17612 Mdot
.61733 .18708 Mdot
.61733 .20349 Mdot
.61733 .22121 Mdot
.61733 .22966 Mdot
.61733 .25669 Mdot
.61733 .27229 Mdot
.61733 .28271 Mdot
.61733 .29574 Mdot
.61733 .30357 Mdot
.61733 .30878 Mdot
.61733 .33483 Mdot
.61821 .12761 Mdot
.61821 .13839 Mdot
.61821 .14913 Mdot
.61821 .19867 Mdot
.61821 .24821 Mdot
.61821 .27678 Mdot
.61821 .30543 Mdot
.61821 .33406 Mdot
.61841 .1422 Mdot
.61841 .15535 Mdot
.61841 .21601 Mdot
.61841 .26378 Mdot
.61841 .29883 Mdot
.61841 .33389 Mdot
.61873 .14823 Mdot
.61873 .16517 Mdot
.61873 .28841 Mdot
.61873 .33362 Mdot
.61896 .06748 Mdot
.61896 .07744 Mdot
.61896 .16269 Mdot
.61896 .17256 Mdot
.61896 .20849 Mdot
.61896 .21825 Mdot
.61896 .2542 Mdot
.61896 .28057 Mdot
.61896 .307 Mdot
.61896 .33341 Mdot
.61929 .17111 Mdot
.61929 .18297 Mdot
.61929 .22622 Mdot
.61929 .26955 Mdot
.61929 .30133 Mdot
.61929 .33312 Mdot
.61979 .18384 Mdot
.61979 .25293 Mdot
.61979 .29278 Mdot
.61979 .33268 Mdot
.62002 .11204 Mdot
.62002 .123 Mdot
.62002 .13393 Mdot
.62002 .19507 Mdot
.62002 .2347 Mdot
.62002 .27433 Mdot
.62002 .30341 Mdot
.62002 .33248 Mdot
.62064 .12336 Mdot
.62064 .1368 Mdot
.62064 .14886 Mdot
.62064 .15891 Mdot
.62064 .2053 Mdot
.62064 .21191 Mdot
.62064 .2152 Mdot
.62064 .25164 Mdot
.62064 .26054 Mdot
.62064 .27837 Mdot
.62064 .29623 Mdot
.62064 .30516 Mdot
.62064 .33194 Mdot
.62132 .16668 Mdot
.62132 .22267 Mdot
.62132 .26671 Mdot
.62132 .29902 Mdot
.62132 .33135 Mdot
.62161 .14136 Mdot
.62161 .2848 Mdot
.62161 .33109 Mdot
.62188 .07367 Mdot
.62188 .08484 Mdot
.62188 .096 Mdot
.62188 .10714 Mdot
.62188 .11826 Mdot
.62188 .18038 Mdot
.62188 .19137 Mdot
.62188 .23158 Mdot
.62188 .27181 Mdot
.62188 .30133 Mdot
.62188 .33085 Mdot
.62236 .14483 Mdot
.62236 .15502 Mdot
.62236 .1785 Mdot
.62236 .20203 Mdot
.62236 .23908 Mdot
.62236 .24901 Mdot
.62236 .27611 Mdot
.62236 .28969 Mdot
.62236 .30327 Mdot
.62236 .33044 Mdot
.62294 .09005 Mdot
.62294 .1038 Mdot
.62294 .11752 Mdot
.62294 .20766 Mdot
.62294 .25718 Mdot
.62294 .29353 Mdot
.62294 .32992 Mdot
.62342 .14978 Mdot
.62342 .16209 Mdot
.62342 .219 Mdot
.62342 .26378 Mdot
.62342 .29663 Mdot
.62342 .32951 Mdot
.6238 .06816 Mdot
.6238 .07949 Mdot
.6238 .09081 Mdot
.6238 .10211 Mdot
.6238 .17639 Mdot
.6238 .18756 Mdot
.6238 .22836 Mdot
.6238 .26922 Mdot
.6238 .29919 Mdot
.6238 .32917 Mdot
.62413 .13034 Mdot
.62413 .1407 Mdot
.62413 .19867 Mdot
.62413 .23622 Mdot
.62413 .27378 Mdot
.62413 .30133 Mdot
.62413 .32889 Mdot
.62441 .07212 Mdot
.62441 .08173 Mdot
.62441 .15442 Mdot
.62441 .16397 Mdot
.62441 .17349 Mdot
.62441 .20814 Mdot
.62441 .21755 Mdot
.62441 .24291 Mdot
.62441 .25223 Mdot
.62441 .27766 Mdot
.62441 .30315 Mdot
.62441 .32864 Mdot
.62465 .09843 Mdot
.62465 .11633 Mdot
.62465 .13417 Mdot
.62465 .2163 Mdot
.62465 .281 Mdot
.62465 .32843 Mdot
.62504 .17294 Mdot
.62504 .28646 Mdot
.62504 .32809 Mdot
.62534 .06951 Mdot
.62534 .08352 Mdot
.62534 .09751 Mdot
.62534 .18948 Mdot
.62534 .25369 Mdot
.62534 .29073 Mdot
.62534 .32783 Mdot
.62559 .13225 Mdot
.62559 .14483 Mdot
.62559 .2152 Mdot
.62559 .26074 Mdot
.62559 .29416 Mdot
.62559 .32761 Mdot
.62579 .17229 Mdot
.62579 .22505 Mdot
.62579 .26654 Mdot
.62579 .29698 Mdot
.62579 .32744 Mdot
.62595 .10486 Mdot
.62595 .11542 Mdot
.62595 .12595 Mdot
.62595 .13645 Mdot
.62595 .18481 Mdot
.62595 .19521 Mdot
.62595 .23329 Mdot
.62595 .27139 Mdot
.62595 .29934 Mdot
.62595 .32729 Mdot
.6261 .15066 Mdot
.6261 .16034 Mdot
.6261 .20506 Mdot
.6261 .24029 Mdot
.6261 .24973 Mdot
.6261 .2755 Mdot
.6261 .30133 Mdot
.6261 .32716 Mdot
.62783 .06787 Mdot
.62783 .07169 Mdot
.62783 .07861 Mdot
.62783 .08934 Mdot
.62783 .09005 Mdot
.62783 .11076 Mdot
.62783 .1141 Mdot
.62783 .12144 Mdot
.62783 .12693 Mdot
.62783 .13696 Mdot
.62783 .13973 Mdot
.62783 .1468 Mdot
.62783 .15123 Mdot
.62783 .15647 Mdot
.62783 .15661 Mdot
.62783 .1811 Mdot
.62783 .18463 Mdot
.62783 .19166 Mdot
.62783 .20191 Mdot
.62783 .21128 Mdot
.62783 .23027 Mdot
.62783 .2376 Mdot
.62783 .25007 Mdot
.62783 .2576 Mdot
.62783 .26892 Mdot
.62783 .27328 Mdot
.62783 .27702 Mdot
.62783 .28309 Mdot
.62783 .28782 Mdot
.62783 .2916 Mdot
.62783 .2947 Mdot
.62783 .29728 Mdot
.62783 .29946 Mdot
.62783 .32565 Mdot
.62962 .12287 Mdot
.62962 .13287 Mdot
.62962 .14284 Mdot
.62962 .1888 Mdot
.62962 .19867 Mdot
.62962 .23483 Mdot
.62962 .271 Mdot
.62962 .29754 Mdot
.62962 .32409 Mdot
.62977 .0625 Mdot
.62977 .07338 Mdot
.62977 .08426 Mdot
.62977 .09512 Mdot
.62977 .10597 Mdot
.62977 .17729 Mdot
.62977 .22717 Mdot
.62977 .26639 Mdot
.62977 .29516 Mdot
.62977 .32395 Mdot
.62995 .15192 Mdot
.62995 .1637 Mdot
.62995 .21811 Mdot
.62995 .26092 Mdot
.62995 .29234 Mdot
.62995 .3238 Mdot
.63016 .09529 Mdot
.63016 .10838 Mdot
.63016 .12144 Mdot
.63016 .19438 Mdot
.63016 .25436 Mdot
.63016 .28895 Mdot
.63016 .32361 Mdot
.63042 .17959 Mdot
.63042 .2848 Mdot
.63042 .32338 Mdot
.63076 .12847 Mdot
.63076 .14483 Mdot
.63076 .22023 Mdot
.63076 .27958 Mdot
.63076 .32309 Mdot
.63118 .20484 Mdot
.63118 .27283 Mdot
.63118 .32271 Mdot
.63145 .10838 Mdot
.63145 .11854 Mdot
.63145 .12867 Mdot
.63145 .18533 Mdot
.63145 .19534 Mdot
.63145 .23199 Mdot
.63145 .26866 Mdot
.63145 .29556 Mdot
.63145 .32248 Mdot
.63176 .06801 Mdot
.63176 .16243 Mdot
.63176 .17336 Mdot
.63176 .22397 Mdot
.63176 .26378 Mdot
.63176 .29298 Mdot
.63176 .32221 Mdot
.63213 .13522 Mdot
.63213 .14723 Mdot
.63213 .20263 Mdot
.63213 .21447 Mdot
.63213 .25798 Mdot
.63213 .28991 Mdot
.63213 .32189 Mdot
.63257 .0625 Mdot
.63257 .07584 Mdot
.63257 .08916 Mdot
.63257 .10247 Mdot
.63257 .18993 Mdot
.63257 .28621 Mdot
.63257 .3215 Mdot
.63293 .13914 Mdot
.63293 .1924 Mdot
.63293 .23622 Mdot
.63293 .29592 Mdot
.63293 .32119 Mdot
.63312 .15964 Mdot
.63312 .17435 Mdot
.63312 .22795 Mdot
.63312 .28165 Mdot
.63312 .32102 Mdot
.63333 .07284 Mdot
.63333 .08317 Mdot
.63333 .09349 Mdot
.63333 .1038 Mdot
.63333 .1141 Mdot
.63333 .12437 Mdot
.63333 .18177 Mdot
.63333 .19193 Mdot
.63333 .22907 Mdot
.63333 .26625 Mdot
.63333 .29353 Mdot
.63333 .32083 Mdot
.63381 .10464 Mdot
.63381 .12144 Mdot
.63381 .15822 Mdot
.63381 .16932 Mdot
.63381 .2152 Mdot
.63381 .22068 Mdot
.63381 .26109 Mdot
.63381 .2759 Mdot
.63381 .29073 Mdot
.63381 .32041 Mdot
.63439 .11793 Mdot
.63439 .13019 Mdot
.63439 .1424 Mdot
.63439 .19867 Mdot
.63439 .25494 Mdot
.63439 .2874 Mdot
.63439 .31991 Mdot
.63471 .12558 Mdot
.63471 .13522 Mdot
.63471 .14483 Mdot
.63471 .18915 Mdot
.63471 .19867 Mdot
.63471 .23354 Mdot
.63471 .26842 Mdot
.63471 .29401 Mdot
.63471 .31962 Mdot
.63508 .18533 Mdot
.63508 .28336 Mdot
.63508 .31931 Mdot
.63527 .06773 Mdot
.63527 .0782 Mdot
.63527 .08866 Mdot
.63527 .09911 Mdot
.63527 .10954 Mdot
.63527 .17811 Mdot
.63527 .22607 Mdot
.63527 .26378 Mdot
.63527 .29144 Mdot
.63527 .31913 Mdot
.63593 .14257 Mdot
.63593 .15389 Mdot
.63593 .20614 Mdot
.63593 .21729 Mdot
.63593 .22357 Mdot
.63593 .25831 Mdot
.63593 .27837 Mdot
.63593 .28841 Mdot
.63593 .31856 Mdot
.63655 .11165 Mdot
.63655 .12144 Mdot
.63655 .1312 Mdot
.63655 .18581 Mdot
.63655 .19546 Mdot
.63655 .23078 Mdot
.63655 .26613 Mdot
.63655 .29206 Mdot
.63655 .31802 Mdot
.63672 .08755 Mdot
.63672 .10006 Mdot
.63672 .11254 Mdot
.63672 .12499 Mdot
.63672 .19457 Mdot
.63672 .2848 Mdot
.63672 .31787 Mdot
.63702 .0625 Mdot
.63702 .07973 Mdot
.63702 .09693 Mdot
.63702 .20994 Mdot
.63702 .27205 Mdot
.63702 .31761 Mdot
.63727 .0731 Mdot
.63727 .0837 Mdot
.63727 .16385 Mdot
.63727 .17435 Mdot
.63727 .22298 Mdot
.63727 .26123 Mdot
.63727 .28929 Mdot
.63727 .31739 Mdot
.63768 .16688 Mdot
.63768 .18056 Mdot
.63768 .23045 Mdot
.63768 .2804 Mdot
.63768 .31703 Mdot
.63812 .13793 Mdot
.63812 .14943 Mdot
.63812 .20246 Mdot
.63812 .21379 Mdot
.63812 .25545 Mdot
.63812 .28602 Mdot
.63812 .31664 Mdot
.63844 .07744 Mdot
.63844 .0874 Mdot
.63844 .09735 Mdot
.63844 .10728 Mdot
.63844 .11719 Mdot
.63844 .18239 Mdot
.63844 .22795 Mdot
.63844 .26378 Mdot
.63844 .29006 Mdot
.63844 .31637 Mdot
.63886 .13249 Mdot
.63886 .14791 Mdot
.63886 .219 Mdot
.63886 .27495 Mdot
.63886 .31599 Mdot
.63914 .06886 Mdot
.63914 .08158 Mdot
.63914 .09429 Mdot
.63914 .10697 Mdot
.63914 .19034 Mdot
.63914 .23622 Mdot
.63914 .2821 Mdot
.63914 .31575 Mdot
.63933 .15983 Mdot
.63933 .17049 Mdot
.63933 .20927 Mdot
.63933 .21981 Mdot
.63933 .25861 Mdot
.63933 .28708 Mdot
.63933 .31559 Mdot
.64038 .0625 Mdot
.64038 .08267 Mdot
.64038 .1028 Mdot
.64038 .12144 Mdot
.64038 .13316 Mdot
.64038 .16167 Mdot
.64038 .16891 Mdot
.64038 .17888 Mdot
.64038 .19867 Mdot
.64038 .20444 Mdot
.64038 .22505 Mdot
.64038 .22643 Mdot
.64038 .26136 Mdot
.64038 .26801 Mdot
.64038 .27733 Mdot
.64038 .28355 Mdot
.64038 .288 Mdot
.64038 .31467 Mdot
.64144 .14483 Mdot
.64144 .1557 Mdot
.64144 .20584 Mdot
.64144 .21654 Mdot
.64144 .25591 Mdot
.64144 .2848 Mdot
.64144 .31374 Mdot
.64164 .07542 Mdot
.64164 .17319 Mdot
.64164 .18597 Mdot
.64164 .23263 Mdot
.64164 .27932 Mdot
.64164 .31357 Mdot
.64192 .11014 Mdot
.64192 .12595 Mdot
.64192 .21424 Mdot
.64192 .27139 Mdot
.64192 .31332 Mdot
.64237 .0676 Mdot
.64237 .07781 Mdot
.64237 .08801 Mdot
.64237 .16517 Mdot
.64237 .17527 Mdot
.64237 .21208 Mdot
.64237 .22206 Mdot
.64237 .25887 Mdot
.64237 .28588 Mdot
.64237 .31292 Mdot
.64271 .09245 Mdot
.64271 .1044 Mdot
.64271 .11633 Mdot
.64271 .12823 Mdot
.64271 .19475 Mdot
.64271 .23788 Mdot
.64271 .281 Mdot
.64271 .31263 Mdot
.64318 .08173 Mdot
.64318 .09133 Mdot
.64318 .10093 Mdot
.64318 .11051 Mdot
.64318 .12007 Mdot
.64318 .14197 Mdot
.64318 .15627 Mdot
.64318 .17349 Mdot
.64318 .18297 Mdot
.64318 .22224 Mdot
.64318 .22692 Mdot
.64318 .26147 Mdot
.64318 .27414 Mdot
.64318 .28683 Mdot
.64318 .31221 Mdot
.64363 .12936 Mdot
.64363 .14042 Mdot
.64363 .15144 Mdot
.64363 .20231 Mdot
.64363 .2424 Mdot
.64363 .28245 Mdot
.64363 .31183 Mdot
.6439 .19867 Mdot
.6439 .26378 Mdot
.6439 .31158 Mdot
.64423 .16843 Mdot
.64423 .18144 Mdot
.64423 .22891 Mdot
.64423 .27643 Mdot
.64423 .3113 Mdot
.64442 .15103 Mdot
.64442 .16132 Mdot
.64442 .20886 Mdot
.64442 .219 Mdot
.64442 .24631 Mdot
.64442 .25632 Mdot
.64442 .2837 Mdot
.64442 .31113 Mdot
.64512 .0625 Mdot
.64512 .06736 Mdot
.64512 .07061 Mdot
.64512 .07466 Mdot
.64512 .07709 Mdot
.64512 .08681 Mdot
.64512 .09653 Mdot
.64512 .09895 Mdot
.64512 .103 Mdot
.64512 .10623 Mdot
.64512 .11107 Mdot
.64512 .16998 Mdot
.64512 .17959 Mdot
.64512 .19072 Mdot
.64512 .20927 Mdot
.64512 .21461 Mdot
.64512 .2241 Mdot
.64512 .23454 Mdot
.64512 .24973 Mdot
.64512 .25911 Mdot
.64512 .26767 Mdot
.64512 .27837 Mdot
.64512 .2848 Mdot
.64512 .31052 Mdot
.64588 .11339 Mdot
.64588 .12465 Mdot
.64588 .13588 Mdot
.64588 .19867 Mdot
.64588 .23936 Mdot
.64588 .28002 Mdot
.64588 .30986 Mdot
.64611 .13608 Mdot
.64611 .15066 Mdot
.64611 .2179 Mdot
.64611 .27082 Mdot
.64611 .30966 Mdot
.64653 .14693 Mdot
.64653 .15736 Mdot
.64653 .20556 Mdot
.64653 .24353 Mdot
.64653 .28146 Mdot
.64653 .30928 Mdot
.64692 .1635 Mdot
.64692 .22505 Mdot
.64692 .27344 Mdot
.64692 .30894 Mdot
.64711 .0625 Mdot
.64711 .07234 Mdot
.64711 .08219 Mdot
.64711 .09202 Mdot
.64711 .16638 Mdot
.64711 .17612 Mdot
.64711 .21159 Mdot
.64711 .22121 Mdot
.64711 .24718 Mdot
.64711 .25669 Mdot
.64711 .28271 Mdot
.64711 .30878 Mdot
.64761 .06867 Mdot
.64761 .08101 Mdot
.64761 .17435 Mdot
.64761 .18655 Mdot
.64761 .23108 Mdot
.64761 .25933 Mdot
.64761 .27565 Mdot
.64761 .30834 Mdot
.6482 .09693 Mdot
.6482 .10838 Mdot
.6482 .11981 Mdot
.6482 .1312 Mdot
.6482 .19492 Mdot
.6482 .23622 Mdot
.6482 .27753 Mdot
.6482 .30783 Mdot
.64846 .20408 Mdot
.64846 .26378 Mdot
.64846 .3076 Mdot
.6487 .13209 Mdot
.6487 .14271 Mdot
.6487 .1533 Mdot
.6487 .20216 Mdot
.6487 .24067 Mdot
.6487 .27915 Mdot
.6487 .30738 Mdot
.64915 .10006 Mdot
.64915 .11503 Mdot
.64915 .12996 Mdot
.64915 .15279 Mdot
.64915 .16269 Mdot
.64915 .20849 Mdot
.64915 .21337 Mdot
.64915 .24455 Mdot
.64915 .26737 Mdot
.64915 .28057 Mdot
.64915 .307 Mdot
.64971 .14483 Mdot
.64971 .15841 Mdot
.64971 .22105 Mdot
.64971 .27034 Mdot
.64971 .3065 Mdot
.65019 .16983 Mdot
.65019 .2275 Mdot
.65019 .27283 Mdot
.65019 .30609 Mdot
.65059 .06832 Mdot
.65059 .07997 Mdot
.65059 .0916 Mdot
.65059 .10322 Mdot
.65059 .11482 Mdot
.65059 .17959 Mdot
.65059 .19106 Mdot
.65059 .23299 Mdot
.65059 .27495 Mdot
.65059 .30573 Mdot
.65094 .1168 Mdot
.65094 .12761 Mdot
.65094 .13839 Mdot
.65094 .19867 Mdot
.65094 .23773 Mdot
.65094 .27678 Mdot
.65094 .30543 Mdot
.65124 .14886 Mdot
.65124 .15891 Mdot
.65124 .2053 Mdot
.65124 .24185 Mdot
.65124 .27837 Mdot
.65124 .30516 Mdot
.65151 .16752 Mdot
.65151 .30493 Mdot
.65196 .18198 Mdot
.65196 .25971 Mdot
.65196 .30454 Mdot
.65232 .0625 Mdot
.65232 .07781 Mdot
.65232 .09311 Mdot
.65232 .10838 Mdot
.65232 .20867 Mdot
.65232 .26378 Mdot
.65232 .30422 Mdot
.65261 .12541 Mdot
.65261 .1393 Mdot
.65261 .21691 Mdot
.65261 .26712 Mdot
.65261 .30397 Mdot
.65286 .15248 Mdot
.65286 .16517 Mdot
.65286 .2238 Mdot
.65286 .26992 Mdot
.65286 .30375 Mdot
.65307 .0625 Mdot
.65307 .07431 Mdot
.65307 .08612 Mdot
.65307 .17542 Mdot
.65307 .18708 Mdot
.65307 .22966 Mdot
.65307 .27229 Mdot
.65307 .30357 Mdot
.65325 .10106 Mdot
.65325 .11204 Mdot
.65325 .123 Mdot
.65325 .18425 Mdot
.65325 .19507 Mdot
.65325 .2347 Mdot
.65325 .27433 Mdot
.65325 .30341 Mdot
.6534 .13462 Mdot
.6534 .14483 Mdot
.6534 .20203 Mdot
.6534 .23908 Mdot
.6534 .27611 Mdot
.6534 .30327 Mdot
.65354 .15442 Mdot
.65354 .16397 Mdot
.65354 .20814 Mdot
.65354 .24291 Mdot
.65354 .27766 Mdot
.65354 .30315 Mdot
.65562 .0703 Mdot
.65562 .07367 Mdot
.65562 .08484 Mdot
.65562 .096 Mdot
.65562 .10714 Mdot
.65562 .11826 Mdot
.65562 .11941 Mdot
.65562 .11995 Mdot
.65562 .13034 Mdot
.65562 .1407 Mdot
.65562 .14095 Mdot
.65562 .14742 Mdot
.65562 .16034 Mdot
.65562 .17111 Mdot
.65562 .17591 Mdot
.65562 .18038 Mdot
.65562 .18842 Mdot
.65562 .19137 Mdot
.65562 .20506 Mdot
.65562 .21998 Mdot
.65562 .22622 Mdot
.65562 .24029 Mdot
.65562 .26003 Mdot
.65562 .2669 Mdot
.65562 .27181 Mdot
.65562 .2755 Mdot
.65562 .30133 Mdot
.65776 .13696 Mdot
.65776 .1468 Mdot
.65776 .20191 Mdot
.65776 .2376 Mdot
.65776 .27328 Mdot
.65776 .29946 Mdot
.6579 .10486 Mdot
.6579 .11542 Mdot
.6579 .12595 Mdot
.6579 .18481 Mdot
.6579 .19521 Mdot
.6579 .23329 Mdot
.6579 .27139 Mdot
.6579 .29934 Mdot
.65807 .06816 Mdot
.65807 .07949 Mdot
.65807 .09081 Mdot
.65807 .17639 Mdot
.65807 .22836 Mdot
.65807 .26922 Mdot
.65807 .29919 Mdot
.65826 .15456 Mdot
.65826 .16668 Mdot
.65826 .22267 Mdot
.65826 .26671 Mdot
.65826 .29902 Mdot
.65848 .129 Mdot
.65848 .1422 Mdot
.65848 .21601 Mdot
.65848 .26378 Mdot
.65848 .29883 Mdot
.65875 .06975 Mdot
.65875 .08426 Mdot
.65875 .09874 Mdot
.65875 .1132 Mdot
.65875 .20814 Mdot
.65875 .2603 Mdot
.65875 .29859 Mdot
.65907 .18297 Mdot
.65907 .25612 Mdot
.65907 .29831 Mdot
.65946 .16956 Mdot
.65946 .29797 Mdot
.65996 .12287 Mdot
.65996 .13287 Mdot
.65996 .14284 Mdot
.65996 .1888 Mdot
.65996 .19867 Mdot
.65996 .23483 Mdot
.65996 .271 Mdot
.65996 .29754 Mdot
.66025 .06787 Mdot
.66025 .07861 Mdot
.66025 .08934 Mdot
.66025 .10006 Mdot
.66025 .11076 Mdot
.66025 .1811 Mdot
.66025 .19166 Mdot
.66025 .23027 Mdot
.66025 .26892 Mdot
.66025 .29728 Mdot
.66059 .16088 Mdot
.66059 .17229 Mdot
.66059 .22505 Mdot
.66059 .26654 Mdot
.66059 .29698 Mdot
.66099 .14978 Mdot
.66099 .16209 Mdot
.66099 .219 Mdot
.66099 .26378 Mdot
.66099 .29663 Mdot
.66145 .10988 Mdot
.66145 .12336 Mdot
.66145 .1368 Mdot
.66145 .21191 Mdot
.66145 .26054 Mdot
.66145 .29623 Mdot
.66181 .13914 Mdot
.66181 .14863 Mdot
.66181 .1924 Mdot
.66181 .23622 Mdot
.66181 .29592 Mdot
.66201 .0625 Mdot
.66201 .07727 Mdot
.66201 .09202 Mdot
.66201 .18898 Mdot
.66201 .25669 Mdot
.66201 .29574 Mdot
.66221 .10838 Mdot
.66221 .11854 Mdot
.66221 .12867 Mdot
.66221 .18533 Mdot
.66221 .19534 Mdot
.66221 .23199 Mdot
.66221 .26866 Mdot
.66221 .29556 Mdot
.66267 .0625 Mdot
.66267 .07338 Mdot
.66267 .08426 Mdot
.66267 .09512 Mdot
.66267 .17729 Mdot
.66267 .22717 Mdot
.66267 .26639 Mdot
.66267 .29516 Mdot
.6632 .15647 Mdot
.6632 .16806 Mdot
.6632 .22163 Mdot
.6632 .26378 Mdot
.6632 .2947 Mdot
.6635 .14483 Mdot
.6635 .16291 Mdot
.6635 .29444 Mdot
.66381 .13225 Mdot
.66381 .14483 Mdot
.66381 .2152 Mdot
.66381 .26074 Mdot
.66381 .29416 Mdot
.66398 .12558 Mdot
.66398 .13522 Mdot
.66398 .14483 Mdot
.66398 .18915 Mdot
.66398 .19867 Mdot
.66398 .23354 Mdot
.66398 .26842 Mdot
.66398 .29401 Mdot
.66454 .0625 Mdot
.66454 .07284 Mdot
.66454 .07628 Mdot
.66454 .08317 Mdot
.66454 .09005 Mdot
.66454 .09349 Mdot
.66454 .1038 Mdot
.66454 .1141 Mdot
.66454 .11752 Mdot
.66454 .18177 Mdot
.66454 .19416 Mdot
.66454 .22907 Mdot
.66454 .25718 Mdot
.66454 .26625 Mdot
.66454 .29353 Mdot
.66517 .16243 Mdot
.66517 .17336 Mdot
.66517 .22397 Mdot
.66517 .26378 Mdot
.66517 .29298 Mdot
.6654 .18384 Mdot
.6654 .29278 Mdot
.66589 .15192 Mdot
.66589 .21811 Mdot
.66589 .26092 Mdot
.66589 .29234 Mdot
.66622 .10184 Mdot
.66622 .11165 Mdot
.66622 .12144 Mdot
.66622 .1312 Mdot
.66622 .18581 Mdot
.66622 .19546 Mdot
.66622 .23078 Mdot
.66622 .26613 Mdot
.66622 .29206 Mdot
.66644 .17136 Mdot
.66644 .29186 Mdot
.66674 .1141 Mdot
.66674 .12693 Mdot
.66674 .13973 Mdot
.66674 .19867 Mdot
.66674 .21128 Mdot
.66674 .2576 Mdot
.66674 .2916 Mdot
.66692 .06773 Mdot
.66692 .0782 Mdot
.66692 .08866 Mdot
.66692 .09911 Mdot
.66692 .16776 Mdot
.66692 .17811 Mdot
.66692 .22607 Mdot
.66692 .26378 Mdot
.66692 .29144 Mdot
.66774 .06951 Mdot
.66774 .08352 Mdot
.66774 .09751 Mdot
.66774 .13741 Mdot
.66774 .15822 Mdot
.66774 .16932 Mdot
.66774 .18948 Mdot
.66774 .22068 Mdot
.66774 .26109 Mdot
.66774 .29073 Mdot
.66851 .06748 Mdot
.66851 .07744 Mdot
.66851 .0874 Mdot
.66851 .09735 Mdot
.66851 .10728 Mdot
.66851 .11719 Mdot
.66851 .18239 Mdot
.66851 .22795 Mdot
.66851 .26378 Mdot
.66851 .29006 Mdot
.66867 .13522 Mdot
.66867 .14723 Mdot
.66867 .20263 Mdot
.66867 .21447 Mdot
.66867 .25798 Mdot
.66867 .28991 Mdot
.66893 .1785 Mdot
.66893 .28969 Mdot
.66938 .0625 Mdot
.66938 .0731 Mdot
.66938 .16385 Mdot
.66938 .17435 Mdot
.66938 .22298 Mdot
.66938 .26123 Mdot
.66938 .28929 Mdot
.66977 .08219 Mdot
.66977 .09529 Mdot
.66977 .10838 Mdot
.66977 .12144 Mdot
.66977 .19438 Mdot
.66977 .28895 Mdot
.67039 .14257 Mdot
.67039 .14823 Mdot
.67039 .15389 Mdot
.67039 .20614 Mdot
.67039 .21729 Mdot
.67039 .25831 Mdot
.67039 .28841 Mdot
.67086 .0625 Mdot
.67086 .07258 Mdot
.67086 .08267 Mdot
.67086 .09274 Mdot
.67086 .1028 Mdot
.67086 .16891 Mdot
.67086 .17888 Mdot
.67086 .22505 Mdot
.67086 .26136 Mdot
.67086 .288 Mdot
.67106 .18463 Mdot
.67106 .28782 Mdot
.67155 .11793 Mdot
.67155 .13019 Mdot
.67155 .1424 Mdot
.67155 .19867 Mdot
.67155 .25494 Mdot
.67155 .2874 Mdot
.67191 .15983 Mdot
.67191 .17049 Mdot
.67191 .20927 Mdot
.67191 .21981 Mdot
.67191 .25861 Mdot
.67191 .28708 Mdot
.6722 .08173 Mdot
.6722 .09133 Mdot
.6722 .10093 Mdot
.6722 .11051 Mdot
.6722 .12007 Mdot
.6722 .17349 Mdot
.6722 .18297 Mdot
.6722 .22692 Mdot
.6722 .26147 Mdot
.6722 .28683 Mdot
.67262 .15736 Mdot
.67262 .17294 Mdot
.67262 .28646 Mdot
.67291 .0625 Mdot
.67291 .07584 Mdot
.67291 .08916 Mdot
.67291 .10247 Mdot
.67291 .18993 Mdot
.67291 .28621 Mdot
.67312 .13793 Mdot
.67312 .14943 Mdot
.67312 .20246 Mdot
.67312 .21379 Mdot
.67312 .25545 Mdot
.67312 .28602 Mdot
.67328 .0676 Mdot
.67328 .07781 Mdot
.67328 .08801 Mdot
.67328 .16517 Mdot
.67328 .17527 Mdot
.67328 .21208 Mdot
.67328 .22206 Mdot
.67328 .25887 Mdot
.67328 .28588 Mdot
.67452 .06736 Mdot
.67452 .07709 Mdot
.67452 .08681 Mdot
.67452 .08755 Mdot
.67452 .09653 Mdot
.67452 .10006 Mdot
.67452 .10623 Mdot
.67452 .11254 Mdot
.67452 .12393 Mdot
.67452 .14136 Mdot
.67452 .14483 Mdot
.67452 .1557 Mdot
.67452 .16998 Mdot
.67452 .17959 Mdot
.67452 .19457 Mdot
.67452 .20584 Mdot
.67452 .21461 Mdot
.67452 .21654 Mdot
.67452 .2241 Mdot
.67452 .25591 Mdot
.67452 .25911 Mdot
.67452 .2848 Mdot
.67577 .16132 Mdot
.67577 .17157 Mdot
.67577 .20886 Mdot
.67577 .219 Mdot
.67577 .25632 Mdot
.67577 .2837 Mdot
.67594 .12144 Mdot
.67594 .13316 Mdot
.67594 .19867 Mdot
.67594 .28355 Mdot
.67616 .18533 Mdot
.67616 .28336 Mdot
.67646 .15123 Mdot
.67646 .28309 Mdot
.6769 .0625 Mdot
.6769 .07234 Mdot
.6769 .08219 Mdot
.6769 .09202 Mdot
.6769 .16638 Mdot
.6769 .17612 Mdot
.6769 .21159 Mdot
.6769 .22121 Mdot
.6769 .24718 Mdot
.6769 .25669 Mdot
.6769 .28271 Mdot
.6772 .14042 Mdot
.6772 .15144 Mdot
.6772 .20231 Mdot
.6772 .25313 Mdot
.6772 .28245 Mdot
.67759 .06886 Mdot
.67759 .08158 Mdot
.67759 .09429 Mdot
.67759 .10697 Mdot
.67759 .19034 Mdot
.67759 .2821 Mdot
.67812 .15964 Mdot
.67812 .17435 Mdot
.67812 .22795 Mdot
.67812 .28165 Mdot
.67833 .14693 Mdot
.67833 .15736 Mdot
.67833 .20556 Mdot
.67833 .21585 Mdot
.67833 .24353 Mdot
.67833 .25369 Mdot
.67833 .28146 Mdot
.67885 .09245 Mdot
.67885 .09843 Mdot
.67885 .1044 Mdot
.67885 .11633 Mdot
.67885 .12823 Mdot
.67885 .19475 Mdot
.67885 .2163 Mdot
.67885 .23788 Mdot
.67885 .281 Mdot
.67935 .15279 Mdot
.67935 .16269 Mdot
.67935 .20849 Mdot
.67935 .21825 Mdot
.67935 .24455 Mdot
.67935 .2542 Mdot
.67935 .28057 Mdot
.67954 .16688 Mdot
.67954 .18056 Mdot
.67954 .23045 Mdot
.67954 .2804 Mdot
.67997 .12465 Mdot
.67997 .13588 Mdot
.67997 .19867 Mdot
.67997 .23936 Mdot
.67997 .28002 Mdot
.68048 .12847 Mdot
.68048 .14483 Mdot
.68048 .22023 Mdot
.68048 .27958 Mdot
.68078 .0625 Mdot
.68078 .07542 Mdot
.68078 .18597 Mdot
.68078 .23263 Mdot
.68078 .27932 Mdot
.68097 .14271 Mdot
.68097 .1533 Mdot
.68097 .20216 Mdot
.68097 .24067 Mdot
.68097 .25099 Mdot
.68097 .27915 Mdot
.68186 .0625 Mdot
.68186 .07466 Mdot
.68186 .08681 Mdot
.68186 .09895 Mdot
.68186 .11107 Mdot
.68186 .14886 Mdot
.68186 .15389 Mdot
.68186 .15891 Mdot
.68186 .19072 Mdot
.68186 .2053 Mdot
.68186 .22357 Mdot
.68186 .23454 Mdot
.68186 .24185 Mdot
.68186 .25164 Mdot
.68186 .27837 Mdot
.68267 .07212 Mdot
.68267 .15442 Mdot
.68267 .16397 Mdot
.68267 .20814 Mdot
.68267 .21755 Mdot
.68267 .24291 Mdot
.68267 .25223 Mdot
.68267 .27766 Mdot
.68283 .10838 Mdot
.68283 .11981 Mdot
.68283 .1312 Mdot
.68283 .19492 Mdot
.68283 .23622 Mdot
.68283 .27753 Mdot
.68305 .16167 Mdot
.68305 .17562 Mdot
.68305 .22643 Mdot
.68305 .27733 Mdot
.68341 .07169 Mdot
.68341 .09005 Mdot
.68341 .10838 Mdot
.68341 .21071 Mdot
.68341 .27702 Mdot
.68368 .12761 Mdot
.68368 .13839 Mdot
.68368 .19867 Mdot
.68368 .23773 Mdot
.68368 .27678 Mdot
.68408 .16843 Mdot
.68408 .18144 Mdot
.68408 .22891 Mdot
.68408 .27643 Mdot
.68445 .13462 Mdot
.68445 .14483 Mdot
.68445 .15502 Mdot
.68445 .20203 Mdot
.68445 .23908 Mdot
.68445 .27611 Mdot
.68469 .10464 Mdot
.68469 .12144 Mdot
.68469 .2152 Mdot
.68469 .2759 Mdot
.68498 .06867 Mdot
.68498 .08101 Mdot
.68498 .09334 Mdot
.68498 .17435 Mdot
.68498 .18655 Mdot
.68498 .23108 Mdot
.68498 .27565 Mdot
.68514 .15066 Mdot
.68514 .16034 Mdot
.68514 .20506 Mdot
.68514 .24029 Mdot
.68514 .24973 Mdot
.68514 .2755 Mdot
.68577 .06832 Mdot
.68577 .07997 Mdot
.68577 .0916 Mdot
.68577 .10322 Mdot
.68577 .11482 Mdot
.68577 .13249 Mdot
.68577 .14791 Mdot
.68577 .19106 Mdot
.68577 .219 Mdot
.68577 .23299 Mdot
.68577 .27495 Mdot
.68648 .11204 Mdot
.68648 .123 Mdot
.68648 .13393 Mdot
.68648 .19507 Mdot
.68648 .2347 Mdot
.68648 .27433 Mdot
.68669 .15627 Mdot
.68669 .22224 Mdot
.68669 .27414 Mdot
.68711 .13034 Mdot
.68711 .1407 Mdot
.68711 .19867 Mdot
.68711 .23622 Mdot
.68711 .27378 Mdot
.68749 .1635 Mdot
.68749 .22505 Mdot
.68749 .27344 Mdot
.68768 .13696 Mdot
.68768 .1468 Mdot
.68768 .15661 Mdot
.68768 .20191 Mdot
.68768 .2376 Mdot
.68768 .27328 Mdot
.68819 .16983 Mdot
.68819 .18224 Mdot
.68819 .20484 Mdot
.68819 .2275 Mdot
.68819 .27283 Mdot
.68881 .0625 Mdot
.68881 .07431 Mdot
.68881 .08612 Mdot
.68881 .09791 Mdot
.68881 .17542 Mdot
.68881 .18708 Mdot
.68881 .22966 Mdot
.68881 .27229 Mdot
.68909 .0625 Mdot
.68909 .07973 Mdot
.68909 .09693 Mdot
.68909 .1141 Mdot
.68909 .20994 Mdot
.68909 .27205 Mdot
.68936 .07367 Mdot
.68936 .08484 Mdot
.68936 .096 Mdot
.68936 .10714 Mdot
.68936 .11826 Mdot
.68936 .18038 Mdot
.68936 .19137 Mdot
.68936 .23158 Mdot
.68936 .27181 Mdot
.68984 .11014 Mdot
.68984 .11542 Mdot
.68984 .12595 Mdot
.68984 .13645 Mdot
.68984 .19521 Mdot
.68984 .21424 Mdot
.68984 .23329 Mdot
.68984 .27139 Mdot
.69028 .13287 Mdot
.69028 .14284 Mdot
.69028 .19867 Mdot
.69028 .23483 Mdot
.69028 .271 Mdot
.69049 .13608 Mdot
.69049 .15066 Mdot
.69049 .2179 Mdot
.69049 .27082 Mdot
.69104 .15841 Mdot
.69104 .22105 Mdot
.69104 .27034 Mdot
.69152 .16517 Mdot
.69152 .2238 Mdot
.69152 .26992 Mdot
.69195 .17111 Mdot
.69195 .18297 Mdot
.69195 .22622 Mdot
.69195 .26955 Mdot
.69232 .06816 Mdot
.69232 .07949 Mdot
.69232 .09081 Mdot
.69232 .10211 Mdot
.69232 .17639 Mdot
.69232 .18756 Mdot
.69232 .22836 Mdot
.69232 .26922 Mdot
.69266 .08934 Mdot
.69266 .10006 Mdot
.69266 .11076 Mdot
.69266 .12144 Mdot
.69266 .1811 Mdot
.69266 .19166 Mdot
.69266 .23027 Mdot
.69266 .26892 Mdot
.69296 .11854 Mdot
.69296 .12867 Mdot
.69296 .13878 Mdot
.69296 .18533 Mdot
.69296 .19534 Mdot
.69296 .23199 Mdot
.69296 .26866 Mdot
.69323 .13522 Mdot
.69323 .14483 Mdot
.69323 .19867 Mdot
.69323 .23354 Mdot
.69323 .26842 Mdot
.6937 .0713 Mdot
.6937 .20444 Mdot
.6937 .26801 Mdot
.6941 .07061 Mdot
.6941 .08681 Mdot
.6941 .103 Mdot
.6941 .11914 Mdot
.6941 .20927 Mdot
.6941 .26767 Mdot
.69443 .11503 Mdot
.69443 .12996 Mdot
.69443 .21337 Mdot
.69443 .26737 Mdot
.69472 .1393 Mdot
.69472 .15312 Mdot
.69472 .21691 Mdot
.69472 .26712 Mdot
.69497 .14742 Mdot
.69497 .16034 Mdot
.69497 .21998 Mdot
.69497 .2669 Mdot
.69519 .16668 Mdot
.69519 .22267 Mdot
.69519 .26671 Mdot
.69538 .07399 Mdot
.69538 .08546 Mdot
.69538 .17229 Mdot
.69538 .18363 Mdot
.69538 .22505 Mdot
.69538 .26654 Mdot
.69556 .0625 Mdot
.69556 .07338 Mdot
.69556 .08426 Mdot
.69556 .09512 Mdot
.69556 .10597 Mdot
.69556 .17729 Mdot
.69556 .188 Mdot
.69556 .22717 Mdot
.69556 .26639 Mdot
.69571 .09349 Mdot
.69571 .1038 Mdot
.69571 .1141 Mdot
.69571 .12437 Mdot
.69571 .18177 Mdot
.69571 .19193 Mdot
.69571 .22907 Mdot
.69571 .26625 Mdot
.69585 .12144 Mdot
.69585 .1312 Mdot
.69585 .14095 Mdot
.69585 .18581 Mdot
.69585 .19546 Mdot
.69585 .23078 Mdot
.69585 .26613 Mdot
.69854 .0625 Mdot
.69854 .06773 Mdot
.69854 .06801 Mdot
.69854 .0782 Mdot
.69854 .07904 Mdot
.69854 .08866 Mdot
.69854 .09311 Mdot
.69854 .09556 Mdot
.69854 .09911 Mdot
.69854 .10728 Mdot
.69854 .10954 Mdot
.69854 .11719 Mdot
.69854 .11941 Mdot
.69854 .11995 Mdot
.69854 .12361 Mdot
.69854 .13356 Mdot
.69854 .1422 Mdot
.69854 .14978 Mdot
.69854 .16209 Mdot
.69854 .17336 Mdot
.69854 .18239 Mdot
.69854 .18425 Mdot
.69854 .19217 Mdot
.69854 .20408 Mdot
.69854 .20867 Mdot
.69854 .21601 Mdot
.69854 .22397 Mdot
.69854 .22607 Mdot
.69854 .26378 Mdot
.70117 .11051 Mdot
.70117 .12007 Mdot
.70117 .12962 Mdot
.70117 .18297 Mdot
.70117 .1924 Mdot
.70117 .22692 Mdot
.70117 .26147 Mdot
.7013 .07258 Mdot
.7013 .08267 Mdot
.7013 .09274 Mdot
.7013 .1028 Mdot
.7013 .11284 Mdot
.7013 .12287 Mdot
.7013 .17888 Mdot
.7013 .1888 Mdot
.7013 .22505 Mdot
.7013 .26136 Mdot
.70145 .0625 Mdot
.70145 .0731 Mdot
.70145 .0837 Mdot
.70145 .09429 Mdot
.70145 .10486 Mdot
.70145 .17435 Mdot
.70145 .18481 Mdot
.70145 .22298 Mdot
.70145 .26123 Mdot
.70162 .16932 Mdot
.70162 .22068 Mdot
.70162 .26109 Mdot
.7018 .15192 Mdot
.7018 .1637 Mdot
.7018 .21811 Mdot
.7018 .26092 Mdot
.70201 .14483 Mdot
.70201 .15736 Mdot
.70201 .2152 Mdot
.70201 .26074 Mdot
.70224 .12336 Mdot
.70224 .1368 Mdot
.70224 .21191 Mdot
.70224 .26054 Mdot
.70251 .09874 Mdot
.70251 .1132 Mdot
.70251 .12761 Mdot
.70251 .20814 Mdot
.70251 .2603 Mdot
.70282 .0703 Mdot
.70282 .0859 Mdot
.70282 .10147 Mdot
.70282 .20377 Mdot
.70282 .26003 Mdot
.70319 .19867 Mdot
.70319 .25971 Mdot
.70362 .19262 Mdot
.70362 .25933 Mdot
.70387 .08681 Mdot
.70387 .09653 Mdot
.70387 .10623 Mdot
.70387 .11591 Mdot
.70387 .12558 Mdot
.70387 .17959 Mdot
.70387 .18915 Mdot
.70387 .2241 Mdot
.70387 .25911 Mdot
.70414 .0676 Mdot
.70414 .07781 Mdot
.70414 .08801 Mdot
.70414 .0982 Mdot
.70414 .10838 Mdot
.70414 .17527 Mdot
.70414 .22206 Mdot
.70414 .25887 Mdot
.70445 .06787 Mdot
.70445 .17049 Mdot
.70445 .21981 Mdot
.70445 .25861 Mdot
.70479 .15389 Mdot
.70479 .16517 Mdot
.70479 .21729 Mdot
.70479 .25831 Mdot
.70517 .14723 Mdot
.70517 .1592 Mdot
.70517 .21447 Mdot
.70517 .25798 Mdot
.70559 .12693 Mdot
.70559 .13973 Mdot
.70559 .21128 Mdot
.70559 .2576 Mdot
.70608 .1038 Mdot
.70608 .11752 Mdot
.70608 .1312 Mdot
.70608 .20766 Mdot
.70608 .25718 Mdot
.70664 .0625 Mdot
.70664 .07234 Mdot
.70664 .07727 Mdot
.70664 .08219 Mdot
.70664 .09202 Mdot
.70664 .10184 Mdot
.70664 .10675 Mdot
.70664 .11165 Mdot
.70664 .17612 Mdot
.70664 .20349 Mdot
.70664 .22121 Mdot
.70664 .25669 Mdot
.70706 .0625 Mdot
.70706 .07284 Mdot
.70706 .08317 Mdot
.70706 .16132 Mdot
.70706 .17157 Mdot
.70706 .219 Mdot
.70706 .25632 Mdot
.70729 .19867 Mdot
.70729 .25612 Mdot
.70753 .1557 Mdot
.70753 .16651 Mdot
.70753 .21654 Mdot
.70753 .25591 Mdot
.70806 .14943 Mdot
.70806 .16088 Mdot
.70806 .193 Mdot
.70806 .21379 Mdot
.70806 .25545 Mdot
.70864 .13019 Mdot
.70864 .1424 Mdot
.70864 .21071 Mdot
.70864 .25494 Mdot
.70896 .25466 Mdot
.70931 .12144 Mdot
.70931 .13445 Mdot
.70931 .20722 Mdot
.70931 .25436 Mdot
.70949 .06748 Mdot
.70949 .07744 Mdot
.70949 .0874 Mdot
.70949 .09735 Mdot
.70949 .16269 Mdot
.70949 .17256 Mdot
.70949 .21825 Mdot
.70949 .2542 Mdot
.71007 .06951 Mdot
.71007 .08352 Mdot
.71007 .09751 Mdot
.71007 .11148 Mdot
.71007 .15736 Mdot
.71007 .16776 Mdot
.71007 .20324 Mdot
.71007 .21585 Mdot
.71007 .25369 Mdot
.71071 .15144 Mdot
.71071 .16243 Mdot
.71071 .21317 Mdot
.71071 .25313 Mdot
.71094 .07002 Mdot
.71094 .08505 Mdot
.71094 .19867 Mdot
.71094 .25293 Mdot
.71143 .13316 Mdot
.71143 .14483 Mdot
.71143 .21019 Mdot
.71143 .2525 Mdot
.71174 .0625 Mdot
.71174 .07212 Mdot
.71174 .08173 Mdot
.71174 .09133 Mdot
.71174 .10093 Mdot
.71174 .16397 Mdot
.71174 .17349 Mdot
.71174 .21755 Mdot
.71174 .25223 Mdot
.71196 .19334 Mdot
.71196 .25204 Mdot
.71224 .12499 Mdot
.71224 .13741 Mdot
.71224 .20683 Mdot
.71224 .25179 Mdot
.71241 .15891 Mdot
.71241 .16891 Mdot
.71241 .2152 Mdot
.71241 .25164 Mdot
.71316 .08916 Mdot
.71316 .10247 Mdot
.71316 .11576 Mdot
.71316 .1533 Mdot
.71316 .16385 Mdot
.71316 .16956 Mdot
.71316 .20302 Mdot
.71316 .2126 Mdot
.71316 .25099 Mdot
.71398 .13588 Mdot
.71398 .14707 Mdot
.71398 .15822 Mdot
.71398 .20971 Mdot
.71398 .25027 Mdot
.7142 .0625 Mdot
.7142 .07676 Mdot
.7142 .091 Mdot
.7142 .19867 Mdot
.7142 .25007 Mdot
.71459 .06736 Mdot
.71459 .07709 Mdot
.71459 .08681 Mdot
.71459 .16034 Mdot
.71459 .16998 Mdot
.71459 .21461 Mdot
.71459 .24973 Mdot
.71491 .12823 Mdot
.71491 .1401 Mdot
.71491 .20647 Mdot
.71491 .24945 Mdot
.71542 .15502 Mdot
.71542 .16517 Mdot
.71542 .19365 Mdot
.71542 .21208 Mdot
.71542 .24901 Mdot
.71596 .09429 Mdot
.71596 .10697 Mdot
.71596 .11963 Mdot
.71596 .13225 Mdot
.71596 .20282 Mdot
.71596 .24854 Mdot
.71633 .13839 Mdot
.71633 .14913 Mdot
.71633 .15983 Mdot
.71633 .20927 Mdot
.71633 .24821 Mdot
.71683 .17136 Mdot
.71683 .24777 Mdot
.71714 .06928 Mdot
.71714 .08283 Mdot
.71714 .09637 Mdot
.71714 .10988 Mdot
.71714 .19867 Mdot
.71714 .2475 Mdot
.71736 .1312 Mdot
.71736 .14257 Mdot
.71736 .20614 Mdot
.71736 .24731 Mdot
.71751 .15661 Mdot
.71751 .16638 Mdot
.71751 .21159 Mdot
.71751 .24718 Mdot
.7185 .11107 Mdot
.7185 .12316 Mdot
.7185 .13522 Mdot
.7185 .1407 Mdot
.7185 .15103 Mdot
.7185 .16132 Mdot
.7185 .16291 Mdot
.7185 .17959 Mdot
.7185 .19391 Mdot
.7185 .20263 Mdot
.7185 .20886 Mdot
.7185 .24631 Mdot
.71961 .13393 Mdot
.71961 .14483 Mdot
.71961 .20584 Mdot
.71961 .24534 Mdot
.7198 .0625 Mdot
.7198 .07542 Mdot
.7198 .08833 Mdot
.7198 .10123 Mdot
.7198 .1141 Mdot
.7198 .18597 Mdot
.7198 .19867 Mdot
.7198 .24517 Mdot
.72008 .17294 Mdot
.72008 .24493 Mdot
.72052 .14284 Mdot
.72052 .15279 Mdot
.72052 .16269 Mdot
.72052 .20849 Mdot
.72052 .24455 Mdot
.72084 .11482 Mdot
.72084 .12639 Mdot
.72084 .13793 Mdot
.72084 .19106 Mdot
.72084 .20246 Mdot
.72084 .24427 Mdot
.72128 .0625 Mdot
.72128 .07628 Mdot
.72128 .18056 Mdot
.72128 .19416 Mdot
.72128 .24388 Mdot
.72168 .13645 Mdot
.72168 .14693 Mdot
.72168 .19521 Mdot
.72168 .20556 Mdot
.72168 .24353 Mdot
.72193 .16517 Mdot
.72193 .24331 Mdot
.72222 .06867 Mdot
.72222 .08101 Mdot
.72222 .09334 Mdot
.72222 .10565 Mdot
.72222 .11793 Mdot
.72222 .18655 Mdot
.72222 .19867 Mdot
.72222 .24306 Mdot
.72238 .14483 Mdot
.72238 .15442 Mdot
.72238 .16397 Mdot
.72238 .20814 Mdot
.72238 .24291 Mdot
.72298 .11826 Mdot
.72298 .12936 Mdot
.72298 .14042 Mdot
.72298 .17435 Mdot
.72298 .19137 Mdot
.72298 .20231 Mdot
.72298 .2424 Mdot
.7236 .13878 Mdot
.7236 .14886 Mdot
.7236 .19534 Mdot
.7236 .2053 Mdot
.7236 .24185 Mdot
.72378 .06906 Mdot
.72378 .08219 Mdot
.72378 .09529 Mdot
.72378 .18144 Mdot
.72378 .19438 Mdot
.72378 .24169 Mdot
.72412 .15594 Mdot
.72412 .24139 Mdot
.72443 .08612 Mdot
.72443 .09791 Mdot
.72443 .10969 Mdot
.72443 .12144 Mdot
.72443 .18708 Mdot
.72443 .19867 Mdot
.72443 .24113 Mdot
.72495 .12144 Mdot
.72495 .13209 Mdot
.72495 .14271 Mdot
.72495 .16715 Mdot
.72495 .19166 Mdot
.72495 .20216 Mdot
.72495 .24067 Mdot
.72538 .14095 Mdot
.72538 .15066 Mdot
.72538 .19546 Mdot
.72538 .20506 Mdot
.72538 .24029 Mdot
.72557 .17562 Mdot
.72557 .24012 Mdot
.72606 .0625 Mdot
.72606 .07503 Mdot
.72606 .08755 Mdot
.72606 .10006 Mdot
.72606 .18224 Mdot
.72606 .19457 Mdot
.72606 .2397 Mdot
.72645 .10211 Mdot
.72645 .11339 Mdot
.72645 .12465 Mdot
.72645 .18756 Mdot
.72645 .19867 Mdot
.72645 .23936 Mdot
.72677 .12437 Mdot
.72677 .13462 Mdot
.72677 .14483 Mdot
.72677 .19193 Mdot
.72677 .20203 Mdot
.72677 .23908 Mdot
.72727 .15869 Mdot
.72727 .23864 Mdot
.72763 .16891 Mdot
.72763 .23832 Mdot
.72791 .17676 Mdot
.72791 .23808 Mdot
.72813 .06849 Mdot
.72813 .08047 Mdot
.72813 .09245 Mdot
.72813 .1044 Mdot
.72813 .11633 Mdot
.72813 .18297 Mdot
.72813 .19475 Mdot
.72813 .23788 Mdot
.72831 .10597 Mdot
.72831 .1168 Mdot
.72831 .12761 Mdot
.72831 .188 Mdot
.72831 .19867 Mdot
.72831 .23773 Mdot
.72846 .12709 Mdot
.72846 .13696 Mdot
.72846 .1468 Mdot
.72846 .19217 Mdot
.72846 .20191 Mdot
.72846 .2376 Mdot
.73003 .0625 Mdot
.73003 .06886 Mdot
.73003 .07399 Mdot
.73003 .08158 Mdot
.73003 .08546 Mdot
.73003 .10838 Mdot
.73003 .10954 Mdot
.73003 .11981 Mdot
.73003 .11995 Mdot
.73003 .12962 Mdot
.73003 .13034 Mdot
.73003 .13914 Mdot
.73003 .1407 Mdot
.73003 .14863 Mdot
.73003 .16111 Mdot
.73003 .17779 Mdot
.73003 .18363 Mdot
.73003 .1924 Mdot
.73003 .19492 Mdot
.73003 .23622 Mdot
.73162 .12287 Mdot
.73162 .13287 Mdot
.73162 .14284 Mdot
.73162 .1888 Mdot
.73162 .19867 Mdot
.73162 .23483 Mdot
.73177 .07904 Mdot
.73177 .09005 Mdot
.73177 .10106 Mdot
.73177 .11204 Mdot
.73177 .123 Mdot
.73177 .18425 Mdot
.73177 .19507 Mdot
.73177 .2347 Mdot
.73196 .0625 Mdot
.73196 .07466 Mdot
.73196 .08681 Mdot
.73196 .17873 Mdot
.73196 .23454 Mdot
.73219 .17191 Mdot
.73219 .23434 Mdot
.73248 .16325 Mdot
.73248 .23408 Mdot
.73286 .13417 Mdot
.73286 .15192 Mdot
.73286 .23375 Mdot
.7331 .12558 Mdot
.7331 .13522 Mdot
.7331 .14483 Mdot
.7331 .18915 Mdot
.7331 .19867 Mdot
.7331 .23354 Mdot
.73338 .09429 Mdot
.73338 .10486 Mdot
.73338 .11542 Mdot
.73338 .12595 Mdot
.73338 .18481 Mdot
.73338 .19521 Mdot
.73338 .23329 Mdot
.73372 .06832 Mdot
.73372 .07997 Mdot
.73372 .0916 Mdot
.73372 .10322 Mdot
.73372 .17959 Mdot
.73372 .19106 Mdot
.73372 .23299 Mdot
.73414 .17319 Mdot
.73414 .23263 Mdot
.73466 .16517 Mdot
.73466 .23217 Mdot
.73487 .10838 Mdot
.73487 .11854 Mdot
.73487 .12867 Mdot
.73487 .18533 Mdot
.73487 .19534 Mdot
.73487 .23199 Mdot
.73534 .0625 Mdot
.73534 .07367 Mdot
.73534 .08484 Mdot
.73534 .096 Mdot
.73534 .10714 Mdot
.73534 .13817 Mdot
.73534 .15481 Mdot
.73534 .18038 Mdot
.73534 .19137 Mdot
.73534 .23158 Mdot
.73591 .17435 Mdot
.73591 .23108 Mdot
.73625 .11165 Mdot
.73625 .12144 Mdot
.73625 .1312 Mdot
.73625 .18581 Mdot
.73625 .19546 Mdot
.73625 .23078 Mdot
.73662 .16688 Mdot
.73662 .23045 Mdot
.73683 .06787 Mdot
.73683 .07861 Mdot
.73683 .08934 Mdot
.73683 .10006 Mdot
.73683 .11076 Mdot
.73683 .1811 Mdot
.73683 .19166 Mdot
.73683 .23027 Mdot
.73753 .0625 Mdot
.73753 .07431 Mdot
.73753 .14169 Mdot
.73753 .15736 Mdot
.73753 .17542 Mdot
.73753 .22966 Mdot
.73821 .07284 Mdot
.73821 .08317 Mdot
.73821 .09349 Mdot
.73821 .1038 Mdot
.73821 .1141 Mdot
.73821 .12437 Mdot
.73821 .18177 Mdot
.73821 .19193 Mdot
.73821 .22907 Mdot
.73839 .16843 Mdot
.73839 .22891 Mdot
.73873 .12665 Mdot
.73873 .14483 Mdot
.73873 .22861 Mdot
.73901 .06816 Mdot
.73901 .07949 Mdot
.73901 .09081 Mdot
.73901 .17639 Mdot
.73901 .22836 Mdot
.73948 .0874 Mdot
.73948 .09735 Mdot
.73948 .10728 Mdot
.73948 .11719 Mdot
.73948 .12709 Mdot
.73948 .14483 Mdot
.73948 .15964 Mdot
.73948 .18239 Mdot
.73948 .19217 Mdot
.73948 .22795 Mdot
.74 .16983 Mdot
.74 .2275 Mdot
.74038 .0625 Mdot
.74038 .07338 Mdot
.74038 .08426 Mdot
.74038 .09512 Mdot
.74038 .10597 Mdot
.74038 .17729 Mdot
.74038 .22717 Mdot
.74067 .10093 Mdot
.74067 .11051 Mdot
.74067 .12007 Mdot
.74067 .12962 Mdot
.74067 .18297 Mdot
.74067 .1924 Mdot
.74067 .22692 Mdot
.74089 .1312 Mdot
.74089 .14823 Mdot
.74089 .22672 Mdot
.74123 .14765 Mdot
.74123 .16167 Mdot
.74123 .22643 Mdot
.74146 .17111 Mdot
.74146 .22622 Mdot
.74164 .06773 Mdot
.74164 .0782 Mdot
.74164 .08866 Mdot
.74164 .09911 Mdot
.74164 .10954 Mdot
.74164 .17811 Mdot
.74164 .22607 Mdot
.7428 .0625 Mdot
.7428 .07258 Mdot
.7428 .08267 Mdot
.7428 .09274 Mdot
.7428 .1028 Mdot
.7428 .11284 Mdot
.7428 .13522 Mdot
.7428 .15018 Mdot
.7428 .15123 Mdot
.7428 .1635 Mdot
.7428 .17229 Mdot
.7428 .17888 Mdot
.7428 .22505 Mdot
.74388 .06736 Mdot
.74388 .07709 Mdot
.74388 .08681 Mdot
.74388 .09653 Mdot
.74388 .10623 Mdot
.74388 .11591 Mdot
.74388 .17959 Mdot
.74388 .2241 Mdot
.74403 .06801 Mdot
.74403 .07904 Mdot
.74403 .17336 Mdot
.74403 .22397 Mdot
.74423 .15248 Mdot
.74423 .16517 Mdot
.74423 .2238 Mdot
.7445 .13878 Mdot
.7445 .15389 Mdot
.7445 .22357 Mdot
.74489 .10006 Mdot
.74489 .11877 Mdot
.74489 .22322 Mdot
.74516 .0625 Mdot
.74516 .0731 Mdot
.74516 .0837 Mdot
.74516 .09429 Mdot
.74516 .17435 Mdot
.74516 .22298 Mdot
.74552 .15456 Mdot
.74552 .16668 Mdot
.74552 .22267 Mdot
.74601 .14197 Mdot
.74601 .15627 Mdot
.74601 .22224 Mdot
.74621 .0676 Mdot
.74621 .07781 Mdot
.74621 .08801 Mdot
.74621 .0982 Mdot
.74621 .17527 Mdot
.74621 .22206 Mdot
.74671 .10643 Mdot
.74671 .12393 Mdot
.74671 .15647 Mdot
.74671 .16806 Mdot
.74671 .22163 Mdot
.74718 .0625 Mdot
.74718 .07234 Mdot
.74718 .08219 Mdot
.74718 .09202 Mdot
.74718 .10184 Mdot
.74718 .11165 Mdot
.74718 .17612 Mdot
.74718 .22121 Mdot
.74737 .14483 Mdot
.74737 .15841 Mdot
.74737 .22105 Mdot
.74779 .15822 Mdot
.74779 .16932 Mdot
.74779 .22068 Mdot
.7483 .11204 Mdot
.7483 .12847 Mdot
.7483 .22023 Mdot
.7486 .14742 Mdot
.7486 .16034 Mdot
.7486 .21998 Mdot
.74879 .15983 Mdot
.74879 .17049 Mdot
.74879 .21981 Mdot
.74972 .0625 Mdot
.74972 .07284 Mdot
.74972 .08317 Mdot
.74972 .13249 Mdot
.74972 .14978 Mdot
.74972 .16132 Mdot
.74972 .16209 Mdot
.74972 .17157 Mdot
.74972 .219 Mdot
.75057 .06748 Mdot
.75057 .07744 Mdot
.75057 .0874 Mdot
.75057 .09735 Mdot
.75057 .16269 Mdot
.75057 .17256 Mdot
.75057 .21825 Mdot
.75074 .15192 Mdot
.75074 .1637 Mdot
.75074 .21811 Mdot
.75098 .13608 Mdot
.75098 .15066 Mdot
.75098 .2179 Mdot
.75137 .0625 Mdot
.75137 .07212 Mdot
.75137 .08173 Mdot
.75137 .09133 Mdot
.75137 .10093 Mdot
.75137 .16397 Mdot
.75137 .17349 Mdot
.75137 .21755 Mdot
.75167 .15389 Mdot
.75167 .16517 Mdot
.75167 .21729 Mdot
.75211 .1393 Mdot
.75211 .15312 Mdot
.75211 .21691 Mdot
.75253 .1557 Mdot
.75253 .16651 Mdot
.75253 .21654 Mdot
.7528 .0625 Mdot
.7528 .08047 Mdot
.7528 .09843 Mdot
.7528 .11633 Mdot
.7528 .2163 Mdot
.75313 .1422 Mdot
.75313 .15535 Mdot
.75313 .21601 Mdot
.75332 .15736 Mdot
.75332 .16776 Mdot
.75332 .21585 Mdot
.75405 .0625 Mdot
.75405 .07258 Mdot
.75405 .08781 Mdot
.75405 .10464 Mdot
.75405 .12144 Mdot
.75405 .14483 Mdot
.75405 .15736 Mdot
.75405 .15891 Mdot
.75405 .16891 Mdot
.75405 .2152 Mdot
.75473 .06736 Mdot
.75473 .07709 Mdot
.75473 .08681 Mdot
.75473 .16034 Mdot
.75473 .16998 Mdot
.75473 .21461 Mdot
.75489 .14723 Mdot
.75489 .1592 Mdot
.75489 .21447 Mdot
.75516 .11014 Mdot
.75516 .12595 Mdot
.75516 .21424 Mdot
.75567 .14943 Mdot
.75567 .16088 Mdot
.75567 .21379 Mdot
.75614 .11503 Mdot
.75614 .12996 Mdot
.75614 .21337 Mdot
.75637 .15144 Mdot
.75637 .16243 Mdot
.75637 .21317 Mdot
.75702 .11941 Mdot
.75702 .13356 Mdot
.75702 .1533 Mdot
.75702 .16385 Mdot
.75702 .2126 Mdot
.75763 .15502 Mdot
.75763 .16517 Mdot
.75763 .21208 Mdot
.75782 .12336 Mdot
.75782 .1368 Mdot
.75782 .21191 Mdot
.75819 .15661 Mdot
.75819 .16638 Mdot
.75819 .21159 Mdot
.75854 .12693 Mdot
.75854 .13973 Mdot
.75854 .15248 Mdot
.75854 .21128 Mdot
.75919 .07169 Mdot
.75919 .09005 Mdot
.75919 .1424 Mdot
.75919 .15456 Mdot
.75919 .21071 Mdot
.75979 .14483 Mdot
.75979 .15647 Mdot
.75979 .21019 Mdot
.76006 .0625 Mdot
.76006 .07973 Mdot
.76006 .09693 Mdot
.76006 .20994 Mdot
.76033 .14707 Mdot
.76033 .15822 Mdot
.76033 .20971 Mdot
.76084 .07061 Mdot
.76084 .08681 Mdot
.76084 .103 Mdot
.76084 .11914 Mdot
.76084 .14913 Mdot
.76084 .15983 Mdot
.76084 .20927 Mdot
.7613 .15103 Mdot
.7613 .16132 Mdot
.7613 .20886 Mdot
.76152 .09311 Mdot
.76152 .10838 Mdot
.76152 .12361 Mdot
.76152 .20867 Mdot
.76173 .15279 Mdot
.76173 .16269 Mdot
.76173 .20849 Mdot
.76213 .09874 Mdot
.76213 .1132 Mdot
.76213 .12761 Mdot
.76213 .15442 Mdot
.76213 .16397 Mdot
.76213 .20814 Mdot
.76268 .11752 Mdot
.76268 .1312 Mdot
.76268 .20766 Mdot
.76317 .12144 Mdot
.76317 .13445 Mdot
.76317 .20722 Mdot
.76362 .12499 Mdot
.76362 .13741 Mdot
.76362 .20683 Mdot
.76404 .12823 Mdot
.76404 .1401 Mdot
.76404 .20647 Mdot
.76441 .1312 Mdot
.76441 .14257 Mdot
.76441 .15389 Mdot
.76441 .20614 Mdot
.76476 .13393 Mdot
.76476 .14483 Mdot
.76476 .1557 Mdot
.76476 .20584 Mdot
.76508 .13645 Mdot
.76508 .14693 Mdot
.76508 .15736 Mdot
.76508 .20556 Mdot
.76537 .13878 Mdot
.76537 .14886 Mdot
.76537 .15891 Mdot
.76537 .2053 Mdot
.76564 .14095 Mdot
.76564 .15066 Mdot
.76564 .16034 Mdot
.76564 .20506 Mdot
.7659 .20484 Mdot
.76636 .0713 Mdot
.76636 .20444 Mdot
.76676 .0625 Mdot
.76676 .07904 Mdot
.76676 .09556 Mdot
.76676 .20408 Mdot
.76712 .0703 Mdot
.76712 .0859 Mdot
.76712 .10147 Mdot
.76712 .20377 Mdot
.76744 .0625 Mdot
.76744 .07727 Mdot
.76744 .09202 Mdot
.76744 .10675 Mdot
.76744 .12144 Mdot
.76744 .20349 Mdot
.76772 .08352 Mdot
.76772 .09751 Mdot
.76772 .11148 Mdot
.76772 .12541 Mdot
.76772 .20324 Mdot
.76798 .10247 Mdot
.76798 .11576 Mdot
.76798 .129 Mdot
.76798 .20302 Mdot
.76821 .10697 Mdot
.76821 .11963 Mdot
.76821 .13225 Mdot
.76821 .20282 Mdot
.76842 .12316 Mdot
.76842 .13522 Mdot
.76842 .20263 Mdot
.76862 .12639 Mdot
.76862 .13793 Mdot
.76862 .20246 Mdot
.76879 .12936 Mdot
.76879 .14042 Mdot
.76879 .20231 Mdot
.76896 .13209 Mdot
.76896 .14271 Mdot
.76896 .1533 Mdot
.76896 .20216 Mdot
.76911 .13462 Mdot
.76911 .14483 Mdot
.76911 .15502 Mdot
.76911 .20203 Mdot
.76925 .13696 Mdot
.76925 .1468 Mdot
.76925 .15661 Mdot
.76925 .20191 Mdot
.77295 .06928 Mdot
.77295 .07002 Mdot
.77295 .0784 Mdot
.77295 .08283 Mdot
.77295 .08505 Mdot
.77295 .091 Mdot
.77295 .09637 Mdot
.77295 .19867 Mdot
.77661 .12144 Mdot
.77661 .1312 Mdot
.77661 .14095 Mdot
.77661 .19546 Mdot
.77675 .11854 Mdot
.77675 .12867 Mdot
.77675 .13878 Mdot
.77675 .19534 Mdot
.7769 .11542 Mdot
.7769 .12595 Mdot
.7769 .13645 Mdot
.7769 .19521 Mdot
.77706 .11204 Mdot
.77706 .123 Mdot
.77706 .13393 Mdot
.77706 .19507 Mdot
.77723 .10838 Mdot
.77723 .11981 Mdot
.77723 .1312 Mdot
.77723 .19492 Mdot
.77742 .09245 Mdot
.77742 .1044 Mdot
.77742 .11633 Mdot
.77742 .12823 Mdot
.77742 .19475 Mdot
.77763 .07503 Mdot
.77763 .08755 Mdot
.77763 .10006 Mdot
.77763 .11254 Mdot
.77763 .12499 Mdot
.77763 .19457 Mdot
.77786 .06906 Mdot
.77786 .08219 Mdot
.77786 .09529 Mdot
.77786 .10838 Mdot
.77786 .19438 Mdot
.77811 .0625 Mdot
.77811 .07628 Mdot
.77811 .09005 Mdot
.77811 .1038 Mdot
.77811 .19416 Mdot
.77838 .06975 Mdot
.77838 .08426 Mdot
.77838 .19391 Mdot
.77869 .0625 Mdot
.77869 .19365 Mdot
.77904 .19334 Mdot
.77942 .193 Mdot
.77987 .19262 Mdot
.78011 .12007 Mdot
.78011 .12962 Mdot
.78011 .13914 Mdot
.78011 .1924 Mdot
.78037 .11719 Mdot
.78037 .12709 Mdot
.78037 .13696 Mdot
.78037 .19217 Mdot
.78066 .1141 Mdot
.78066 .12437 Mdot
.78066 .13462 Mdot
.78066 .19193 Mdot
.78096 .10006 Mdot
.78096 .11076 Mdot
.78096 .12144 Mdot
.78096 .13209 Mdot
.78096 .19166 Mdot
.78129 .096 Mdot
.78129 .10714 Mdot
.78129 .11826 Mdot
.78129 .12936 Mdot
.78129 .19137 Mdot
.78165 .06832 Mdot
.78165 .07997 Mdot
.78165 .0916 Mdot
.78165 .10322 Mdot
.78165 .11482 Mdot
.78165 .12639 Mdot
.78165 .19106 Mdot
.78204 .0625 Mdot
.78204 .07466 Mdot
.78204 .08681 Mdot
.78204 .09895 Mdot
.78204 .11107 Mdot
.78204 .19072 Mdot
.78247 .06886 Mdot
.78247 .08158 Mdot
.78247 .09429 Mdot
.78247 .10697 Mdot
.78247 .19034 Mdot
.78293 .0625 Mdot
.78293 .07584 Mdot
.78293 .08916 Mdot
.78293 .10247 Mdot
.78293 .18993 Mdot
.78345 .06951 Mdot
.78345 .08352 Mdot
.78345 .18948 Mdot
.78383 .11591 Mdot
.78383 .12558 Mdot
.78383 .13522 Mdot
.78383 .18915 Mdot
.78403 .18898 Mdot
.78423 .1028 Mdot
.78423 .11284 Mdot
.78423 .12287 Mdot
.78423 .13287 Mdot
.78423 .1888 Mdot
.78467 .09911 Mdot
.78467 .10954 Mdot
.78467 .11995 Mdot
.78467 .13034 Mdot
.78467 .18842 Mdot
.78514 .08426 Mdot
.78514 .09512 Mdot
.78514 .10597 Mdot
.78514 .1168 Mdot
.78514 .12761 Mdot
.78514 .188 Mdot
.78539 .18779 Mdot
.78565 .06816 Mdot
.78565 .07949 Mdot
.78565 .09081 Mdot
.78565 .10211 Mdot
.78565 .11339 Mdot
.78565 .18756 Mdot
.7862 .0625 Mdot
.7862 .07431 Mdot
.7862 .08612 Mdot
.7862 .09791 Mdot
.7862 .10969 Mdot
.7862 .18708 Mdot
.7868 .06867 Mdot
.7868 .08101 Mdot
.7868 .09334 Mdot
.7868 .10565 Mdot
.7868 .18655 Mdot
.78713 .18626 Mdot
.78747 .0625 Mdot
.78747 .07542 Mdot
.78747 .08833 Mdot
.78747 .18597 Mdot
.78764 .10184 Mdot
.78764 .11165 Mdot
.78764 .12144 Mdot
.78764 .1312 Mdot
.78764 .18581 Mdot
.78819 .06928 Mdot
.78819 .08801 Mdot
.78819 .0982 Mdot
.78819 .10838 Mdot
.78819 .11854 Mdot
.78819 .12867 Mdot
.78819 .18533 Mdot
.78879 .0731 Mdot
.78879 .0837 Mdot
.78879 .09429 Mdot
.78879 .10486 Mdot
.78879 .11542 Mdot
.78879 .18481 Mdot
.789 .18463 Mdot
.78943 .06801 Mdot
.78943 .07904 Mdot
.78943 .09005 Mdot
.78943 .10106 Mdot
.78943 .11204 Mdot
.78943 .18425 Mdot
.78989 .18384 Mdot
.79013 .0625 Mdot
.79013 .07399 Mdot
.79013 .08546 Mdot
.79013 .09693 Mdot
.79013 .10838 Mdot
.79013 .18363 Mdot
.79089 .06849 Mdot
.79089 .08047 Mdot
.79089 .09133 Mdot
.79089 .09245 Mdot
.79089 .10093 Mdot
.79089 .11051 Mdot
.79089 .12007 Mdot
.79089 .12962 Mdot
.79089 .18297 Mdot
.79155 .06748 Mdot
.79155 .07744 Mdot
.79155 .0874 Mdot
.79155 .09735 Mdot
.79155 .10728 Mdot
.79155 .11719 Mdot
.79155 .18239 Mdot
.79173 .0625 Mdot
.79173 .07503 Mdot
.79173 .08755 Mdot
.79173 .18224 Mdot
.79202 .18198 Mdot
.79226 .07284 Mdot
.79226 .08317 Mdot
.79226 .09349 Mdot
.79226 .1038 Mdot
.79226 .1141 Mdot
.79226 .18177 Mdot
.79264 .18144 Mdot
.79303 .06787 Mdot
.79303 .07861 Mdot
.79303 .08934 Mdot
.79303 .10006 Mdot
.79303 .11076 Mdot
.79303 .1811 Mdot
.7933 .18087 Mdot
.79364 .18056 Mdot
.79386 .0625 Mdot
.79386 .07367 Mdot
.79386 .08484 Mdot
.79386 .096 Mdot
.79386 .10714 Mdot
.79386 .18038 Mdot
.79476 .06736 Mdot
.79476 .06832 Mdot
.79476 .07709 Mdot
.79476 .07997 Mdot
.79476 .08681 Mdot
.79476 .0916 Mdot
.79476 .09653 Mdot
.79476 .10623 Mdot
.79476 .11591 Mdot
.79476 .17959 Mdot
.79557 .0625 Mdot
.79557 .07258 Mdot
.79557 .08267 Mdot
.79557 .09274 Mdot
.79557 .1028 Mdot
.79557 .11284 Mdot
.79557 .17888 Mdot
.79574 .07466 Mdot
.79574 .17873 Mdot
.796 .1785 Mdot
.79644 .06773 Mdot
.79644 .0782 Mdot
.79644 .08866 Mdot
.79644 .09911 Mdot
.79644 .10954 Mdot
.79644 .17811 Mdot
.79681 .17779 Mdot
.79739 .0625 Mdot
.79739 .07338 Mdot
.79739 .08426 Mdot
.79739 .09512 Mdot
.79739 .17729 Mdot
.79799 .17676 Mdot
.79841 .06816 Mdot
.79841 .07949 Mdot
.79841 .09081 Mdot
.79841 .17639 Mdot
.79872 .0625 Mdot
.79872 .07234 Mdot
.79872 .08219 Mdot
.79872 .09202 Mdot
.79872 .10184 Mdot
.79872 .11165 Mdot
.79872 .17612 Mdot
.79896 .17591 Mdot
.79929 .17562 Mdot
.79952 .17542 Mdot
.79969 .0676 Mdot
.79969 .07781 Mdot
.79969 .08801 Mdot
.79969 .0982 Mdot
.79969 .17527 Mdot
.80074 .0625 Mdot
.80074 .0837 Mdot
.80074 .09429 Mdot
.80074 .17435 Mdot
.80173 .0625 Mdot
.80173 .07212 Mdot
.80173 .08173 Mdot
.80173 .09133 Mdot
.80173 .10093 Mdot
.80173 .11051 Mdot
.80173 .16397 Mdot
.80173 .17349 Mdot
.80187 .06801 Mdot
.80187 .07904 Mdot
.80187 .17336 Mdot
.80207 .17319 Mdot
.80235 .17294 Mdot
.80278 .06748 Mdot
.80278 .07744 Mdot
.80278 .0874 Mdot
.80278 .09735 Mdot
.80278 .17256 Mdot
.8031 .17229 Mdot
.80354 .17191 Mdot
.80392 .0625 Mdot
.80392 .07284 Mdot
.80392 .08317 Mdot
.80392 .09349 Mdot
.80392 .17157 Mdot
.80416 .17136 Mdot
.80444 .17111 Mdot
.80516 .06787 Mdot
.80516 .07861 Mdot
.80516 .17049 Mdot
.80573 .06736 Mdot
.80573 .07709 Mdot
.80573 .08681 Mdot
.80573 .09653 Mdot
.80573 .16034 Mdot
.80573 .16998 Mdot
.8059 .16983 Mdot
.80621 .16956 Mdot
.80649 .16932 Mdot
.80696 .0625 Mdot
.80696 .07258 Mdot
.80696 .08267 Mdot
.80696 .15891 Mdot
.80696 .16891 Mdot
.80751 .16843 Mdot
.80794 .16806 Mdot
.80827 .06773 Mdot
.80827 .16776 Mdot
.80855 .16752 Mdot
.80897 .16715 Mdot
.80928 .16688 Mdot
.80951 .16668 Mdot
.8097 .16651 Mdot
.80985 .07234 Mdot
.80985 .08219 Mdot
.80985 .15661 Mdot
.80985 .16638 Mdot
.81124 .15502 Mdot
.81124 .16517 Mdot
.8126 .07212 Mdot
.8126 .08173 Mdot
.8126 .15442 Mdot
.8126 .16397 Mdot
.81274 .1533 Mdot
.81274 .16385 Mdot
.81292 .1637 Mdot
.81314 .1635 Mdot
.81342 .16325 Mdot
.81381 .16291 Mdot
.81406 .15279 Mdot
.81406 .16269 Mdot
.81437 .16243 Mdot
.81475 .16209 Mdot
.81523 .16167 Mdot
.81564 .15103 Mdot
.81564 .16132 Mdot
.81587 .16111 Mdot
.81613 .16088 Mdot
.81676 .15066 Mdot
.81676 .16034 Mdot
.81734 .14913 Mdot
.81734 .15983 Mdot
.81756 .15964 Mdot
.81806 .1592 Mdot
.81839 .14886 Mdot
.81839 .15891 Mdot
.81864 .15869 Mdot
.81897 .15841 Mdot
.81918 .14707 Mdot
.81918 .15822 Mdot
.82016 .14693 Mdot
.82016 .15736 Mdot
.82102 .1468 Mdot
.82102 .15661 Mdot
.82118 .15647 Mdot
.82141 .15627 Mdot
.82178 .15594 Mdot
.82206 .14483 Mdot
.82206 .1557 Mdot
.82246 .15535 Mdot
.82284 .14483 Mdot
.82284 .15502 Mdot
.82307 .15481 Mdot
.82336 .15456 Mdot
.82352 .14483 Mdot
.82352 .15442 Mdot
.82413 .14257 Mdot
.82413 .15389 Mdot
.8248 .14271 Mdot
.8248 .1533 Mdot
.825 .15312 Mdot
.82539 .14284 Mdot
.82539 .15279 Mdot
.82574 .15248 Mdot
.82637 .1401 Mdot
.82637 .15192 Mdot
.82692 .14042 Mdot
.82692 .15144 Mdot
.82717 .15123 Mdot
.8274 .1407 Mdot
.8274 .15103 Mdot
.82782 .14095 Mdot
.82782 .15066 Mdot
.82836 .15018 Mdot
.82883 .13741 Mdot
.82883 .14978 Mdot
.82922 .13793 Mdot
.82922 .14943 Mdot
.82957 .13839 Mdot
.82957 .14913 Mdot
.82987 .13878 Mdot
.82987 .14886 Mdot
.83014 .13914 Mdot
.83014 .14863 Mdot
.83059 .14823 Mdot
.83096 .14791 Mdot
.83126 .14765 Mdot
.83151 .13445 Mdot
.83151 .14742 Mdot
.83173 .13522 Mdot
.83173 .14723 Mdot
.83192 .13588 Mdot
.83192 .14707 Mdot
.83208 .13645 Mdot
.83208 .14693 Mdot
.83223 .13696 Mdot
.83223 .1468 Mdot
.83447 .13225 Mdot
.83447 .13316 Mdot
.83447 .13393 Mdot
.83447 .13462 Mdot
.83447 .14483 Mdot
.83675 .13287 Mdot
.83675 .14284 Mdot
.8369 .13209 Mdot
.8369 .14271 Mdot
.83707 .1312 Mdot
.83707 .14257 Mdot
.83726 .13019 Mdot
.83726 .1424 Mdot
.83748 .129 Mdot
.83748 .1422 Mdot
.83775 .14197 Mdot
.83806 .14169 Mdot
.83845 .14136 Mdot
.83892 .1312 Mdot
.83892 .14095 Mdot
.8392 .13034 Mdot
.8392 .1407 Mdot
.83952 .12936 Mdot
.83952 .14042 Mdot
.83989 .12823 Mdot
.83989 .1401 Mdot
.84031 .12693 Mdot
.84031 .13973 Mdot
.8408 .12541 Mdot
.8408 .1393 Mdot
.84099 .12962 Mdot
.84099 .13914 Mdot
.84139 .12867 Mdot
.84139 .13878 Mdot
.84185 .12761 Mdot
.84185 .13839 Mdot
.8421 .13817 Mdot
.84237 .12639 Mdot
.84237 .13793 Mdot
.84296 .12499 Mdot
.84296 .13741 Mdot
.84348 .12709 Mdot
.84348 .13696 Mdot
.84366 .12336 Mdot
.84366 .1368 Mdot
.84405 .12595 Mdot
.84405 .13645 Mdot
.84448 .12144 Mdot
.84448 .13608 Mdot
.84471 .12465 Mdot
.84471 .13588 Mdot
.84546 .12316 Mdot
.84546 .12558 Mdot
.84546 .13522 Mdot
.84615 .12437 Mdot
.84615 .13462 Mdot
.84634 .12144 Mdot
.84634 .13445 Mdot
.84666 .13417 Mdot
.84693 .123 Mdot
.84693 .13393 Mdot
.84736 .11941 Mdot
.84736 .13356 Mdot
.84782 .12144 Mdot
.84782 .13316 Mdot
.84815 .12287 Mdot
.84815 .13287 Mdot
.84858 .11701 Mdot
.84858 .13249 Mdot
.84885 .11963 Mdot
.84885 .13225 Mdot
.84904 .12144 Mdot
.84904 .13209 Mdot
.85005 .11752 Mdot
.85005 .11981 Mdot
.85005 .1312 Mdot
.85104 .11995 Mdot
.85104 .13034 Mdot
.85122 .11793 Mdot
.85122 .13019 Mdot
.85147 .11503 Mdot
.85147 .12996 Mdot
.85187 .12007 Mdot
.85187 .12962 Mdot
.85216 .11826 Mdot
.85216 .12936 Mdot
.85257 .11576 Mdot
.85257 .129 Mdot
.85294 .11854 Mdot
.85294 .12867 Mdot
.85317 .11204 Mdot
.85317 .12847 Mdot
.85345 .11633 Mdot
.85345 .12823 Mdot
.85416 .1132 Mdot
.85416 .1168 Mdot
.85416 .12761 Mdot
.85476 .11719 Mdot
.85476 .12709 Mdot
.85493 .1141 Mdot
.85493 .12693 Mdot
.85526 .10838 Mdot
.85526 .12665 Mdot
.85555 .11482 Mdot
.85555 .12639 Mdot
.85606 .11014 Mdot
.85606 .11542 Mdot
.85606 .12595 Mdot
.85648 .11591 Mdot
.85648 .12558 Mdot
.85667 .11148 Mdot
.85667 .12541 Mdot
.85715 .11254 Mdot
.85715 .12499 Mdot
.85754 .11339 Mdot
.85754 .12465 Mdot
.85786 .1141 Mdot
.85786 .12437 Mdot
.85836 .10643 Mdot
.85836 .12393 Mdot
.85873 .10838 Mdot
.85873 .12361 Mdot
.85902 .10988 Mdot
.85902 .12336 Mdot
.85924 .11107 Mdot
.85924 .12316 Mdot
.85943 .11204 Mdot
.85943 .123 Mdot
.85958 .1028 Mdot
.85958 .11284 Mdot
.85958 .12287 Mdot
.86121 .10184 Mdot
.86121 .10464 Mdot
.86121 .10675 Mdot
.86121 .10969 Mdot
.86121 .11076 Mdot
.86121 .11165 Mdot
.86121 .12144 Mdot
.86277 .10093 Mdot
.86277 .11051 Mdot
.86277 .12007 Mdot
.86291 .09911 Mdot
.86291 .10954 Mdot
.86291 .11995 Mdot
.86308 .10838 Mdot
.86308 .11981 Mdot
.86328 .10697 Mdot
.86328 .11963 Mdot
.86353 .10522 Mdot
.86353 .11941 Mdot
.86384 .103 Mdot
.86384 .11914 Mdot
.86426 .10006 Mdot
.86426 .11877 Mdot
.86453 .0982 Mdot
.86453 .10838 Mdot
.86453 .11854 Mdot
.86484 .096 Mdot
.86484 .10714 Mdot
.86484 .11826 Mdot
.86522 .10565 Mdot
.86522 .11793 Mdot
.86569 .1038 Mdot
.86569 .11752 Mdot
.86606 .09735 Mdot
.86606 .10728 Mdot
.86606 .11719 Mdot
.86628 .10147 Mdot
.86628 .11701 Mdot
.86651 .09512 Mdot
.86651 .10597 Mdot
.86651 .1168 Mdot
.86705 .09245 Mdot
.86705 .09843 Mdot
.86705 .1044 Mdot
.86705 .11633 Mdot
.86753 .09653 Mdot
.86753 .10623 Mdot
.86753 .11591 Mdot
.86771 .10247 Mdot
.86771 .11576 Mdot
.8681 .09429 Mdot
.8681 .10486 Mdot
.8681 .11542 Mdot
.86853 .10006 Mdot
.86853 .11503 Mdot
.86878 .0916 Mdot
.86878 .10322 Mdot
.86878 .11482 Mdot
.8696 .08833 Mdot
.8696 .09349 Mdot
.8696 .09693 Mdot
.8696 .10123 Mdot
.8696 .1038 Mdot
.8696 .1141 Mdot
.87041 .09081 Mdot
.87041 .10211 Mdot
.87041 .11339 Mdot
.87063 .09874 Mdot
.87063 .1132 Mdot
.87104 .09274 Mdot
.87104 .1028 Mdot
.87104 .11284 Mdot
.87138 .08755 Mdot
.87138 .10006 Mdot
.87138 .11254 Mdot
.87195 .09005 Mdot
.87195 .09556 Mdot
.87195 .10106 Mdot
.87195 .11204 Mdot
.8724 .08219 Mdot
.8724 .09202 Mdot
.8724 .10184 Mdot
.8724 .11165 Mdot
.87259 .08352 Mdot
.87259 .09751 Mdot
.87259 .11148 Mdot
.87306 .08681 Mdot
.87306 .09895 Mdot
.87306 .11107 Mdot
.87342 .08934 Mdot
.87342 .10006 Mdot
.87342 .11076 Mdot
.8737 .08173 Mdot
.8737 .09133 Mdot
.8737 .10093 Mdot
.8737 .11051 Mdot
.87412 .0784 Mdot
.87412 .09429 Mdot
.87412 .11014 Mdot
.87442 .08283 Mdot
.87442 .09637 Mdot
.87442 .10988 Mdot
.87464 .08612 Mdot
.87464 .09791 Mdot
.87464 .10969 Mdot
.87481 .0782 Mdot
.87481 .08866 Mdot
.87481 .09911 Mdot
.87481 .10954 Mdot
.87614 .07169 Mdot
.87614 .07399 Mdot
.87614 .07781 Mdot
.87614 .08219 Mdot
.87614 .08546 Mdot
.87614 .08801 Mdot
.87614 .09005 Mdot
.87614 .09311 Mdot
.87614 .09529 Mdot
.87614 .0982 Mdot
.87614 .10838 Mdot
.8774 .06748 Mdot
.8774 .07744 Mdot
.8774 .0874 Mdot
.8774 .09735 Mdot
.8774 .10728 Mdot
.87755 .07367 Mdot
.87755 .08484 Mdot
.87755 .096 Mdot
.87755 .10714 Mdot
.87774 .06886 Mdot
.87774 .08158 Mdot
.87774 .09429 Mdot
.87774 .10697 Mdot
.878 .07727 Mdot
.878 .09202 Mdot
.878 .10675 Mdot
.87836 .0713 Mdot
.87836 .08888 Mdot
.87836 .10643 Mdot
.8786 .06736 Mdot
.8786 .07709 Mdot
.8786 .08681 Mdot
.8786 .09653 Mdot
.8786 .10623 Mdot
.87889 .0625 Mdot
.87889 .07338 Mdot
.87889 .08426 Mdot
.87889 .09512 Mdot
.87889 .10597 Mdot
.87926 .06867 Mdot
.87926 .08101 Mdot
.87926 .09334 Mdot
.87926 .10565 Mdot
.87974 .0625 Mdot
.87974 .07676 Mdot
.87974 .091 Mdot
.87974 .10522 Mdot
.88016 .0731 Mdot
.88016 .0837 Mdot
.88016 .09429 Mdot
.88016 .10486 Mdot
.88041 .07094 Mdot
.88041 .08781 Mdot
.88041 .10464 Mdot
.88068 .06849 Mdot
.88068 .08047 Mdot
.88068 .09245 Mdot
.88068 .1044 Mdot
.88137 .0625 Mdot
.88137 .07284 Mdot
.88137 .07628 Mdot
.88137 .08317 Mdot
.88137 .09005 Mdot
.88137 .09349 Mdot
.88137 .1038 Mdot
.88203 .06832 Mdot
.88203 .07997 Mdot
.88203 .0916 Mdot
.88203 .10322 Mdot
.88229 .07061 Mdot
.88229 .08681 Mdot
.88229 .103 Mdot
.88252 .07258 Mdot
.88252 .08267 Mdot
.88252 .09274 Mdot
.88252 .1028 Mdot
.88289 .0625 Mdot
.88289 .07584 Mdot
.88289 .08916 Mdot
.88289 .10247 Mdot
.8833 .06816 Mdot
.8833 .07949 Mdot
.8833 .09081 Mdot
.8833 .10211 Mdot
.88361 .0625 Mdot
.88361 .07234 Mdot
.88361 .08219 Mdot
.88361 .09202 Mdot
.88361 .10184 Mdot
.88403 .0703 Mdot
.88403 .0859 Mdot
.88403 .10147 Mdot
.88431 .0625 Mdot
.88431 .07542 Mdot
.88431 .08833 Mdot
.88431 .10123 Mdot
.88451 .06801 Mdot
.88451 .07904 Mdot
.88451 .09005 Mdot
.88451 .10106 Mdot
.88465 .07212 Mdot
.88465 .08173 Mdot
.88465 .09133 Mdot
.88465 .10093 Mdot
.88565 .0625 Mdot
.88565 .06787 Mdot
.88565 .07002 Mdot
.88565 .07503 Mdot
.88565 .07861 Mdot
.88565 .08129 Mdot
.88565 .08505 Mdot
.88565 .08755 Mdot
.88565 .08934 Mdot
.88565 .10006 Mdot
.88673 .06773 Mdot
.88673 .0782 Mdot
.88673 .08866 Mdot
.88673 .09911 Mdot
.88691 .0625 Mdot
.88691 .07466 Mdot
.88691 .08681 Mdot
.88691 .09895 Mdot
.88715 .06975 Mdot
.88715 .08426 Mdot
.88715 .09874 Mdot
.88751 .0625 Mdot
.88751 .08047 Mdot
.88751 .09843 Mdot
.88777 .0676 Mdot
.88777 .07781 Mdot
.88777 .08801 Mdot
.88777 .0982 Mdot
.8881 .0625 Mdot
.8881 .07431 Mdot
.8881 .08612 Mdot
.8881 .09791 Mdot
.88856 .06951 Mdot
.88856 .08352 Mdot
.88856 .09751 Mdot
.88875 .06748 Mdot
.88875 .07744 Mdot
.88875 .0874 Mdot
.88875 .09735 Mdot
.88922 .0625 Mdot
.88922 .07399 Mdot
.88922 .07973 Mdot
.88922 .08546 Mdot
.88922 .09693 Mdot
.88968 .06736 Mdot
.88968 .07709 Mdot
.88968 .08681 Mdot
.88968 .09653 Mdot
.88987 .06928 Mdot
.88987 .08283 Mdot
.88987 .09637 Mdot
.89028 .0625 Mdot
.89028 .07367 Mdot
.89028 .08484 Mdot
.89028 .096 Mdot
.89079 .0625 Mdot
.89079 .07904 Mdot
.89079 .09556 Mdot
.89109 .06906 Mdot
.89109 .08219 Mdot
.89109 .09529 Mdot
.89129 .0625 Mdot
.89129 .07338 Mdot
.89129 .08426 Mdot
.89129 .09512 Mdot
.89224 .0625 Mdot
.89224 .06886 Mdot
.89224 .0731 Mdot
.89224 .0784 Mdot
.89224 .08158 Mdot
.89224 .0837 Mdot
.89224 .09429 Mdot
.89315 .0625 Mdot
.89315 .07284 Mdot
.89315 .08317 Mdot
.89315 .09349 Mdot
.89333 .06867 Mdot
.89333 .08101 Mdot
.89333 .09334 Mdot
.89359 .0625 Mdot
.89359 .07781 Mdot
.89359 .09311 Mdot
.89401 .0625 Mdot
.89401 .07258 Mdot
.89401 .08267 Mdot
.89401 .09274 Mdot
.89435 .06849 Mdot
.89435 .08047 Mdot
.89435 .09245 Mdot
.89484 .0625 Mdot
.89484 .07234 Mdot
.89484 .07727 Mdot
.89484 .08219 Mdot
.89484 .09202 Mdot
.89531 .06832 Mdot
.89531 .07997 Mdot
.89531 .0916 Mdot
.89562 .0625 Mdot
.89562 .07212 Mdot
.89562 .08173 Mdot
.89562 .09133 Mdot
.896 .0625 Mdot
.896 .07676 Mdot
.896 .091 Mdot
.89622 .06816 Mdot
.89622 .07949 Mdot
.89622 .09081 Mdot
.89708 .0625 Mdot
.89708 .06801 Mdot
.89708 .07169 Mdot
.89708 .07628 Mdot
.89708 .07904 Mdot
.89708 .09005 Mdot
.8979 .06787 Mdot
.8979 .07861 Mdot
.8979 .08934 Mdot
.8981 .0625 Mdot
.8981 .07584 Mdot
.8981 .08916 Mdot
.89842 .0713 Mdot
.89842 .08888 Mdot
.89868 .06773 Mdot
.89868 .0782 Mdot
.89868 .08866 Mdot
.89905 .0625 Mdot
.89905 .07542 Mdot
.89905 .08833 Mdot
.89941 .0676 Mdot
.89941 .07781 Mdot
.89941 .08801 Mdot
.89965 .07094 Mdot
.89965 .08781 Mdot
.89994 .0625 Mdot
.89994 .07503 Mdot
.89994 .08755 Mdot
.90011 .06748 Mdot
.90011 .07744 Mdot
.90011 .0874 Mdot
.90078 .0625 Mdot
.90078 .06736 Mdot
.90078 .07061 Mdot
.90078 .07466 Mdot
.90078 .07709 Mdot
.90078 .08681 Mdot
.90158 .0625 Mdot
.90158 .07431 Mdot
.90158 .08612 Mdot
.90183 .0703 Mdot
.90183 .0859 Mdot
.90233 .0625 Mdot
.90233 .07399 Mdot
.90233 .08546 Mdot
.9028 .07002 Mdot
.9028 .08505 Mdot
.90304 .0625 Mdot
.90304 .07367 Mdot
.90304 .08484 Mdot
.90371 .0625 Mdot
.90371 .06975 Mdot
.90371 .07338 Mdot
.90371 .08426 Mdot
.90434 .0625 Mdot
.90434 .0731 Mdot
.90434 .0837 Mdot
.90455 .06951 Mdot
.90455 .08352 Mdot
.90495 .0625 Mdot
.90495 .07284 Mdot
.90495 .08317 Mdot
.90534 .06928 Mdot
.90534 .08283 Mdot
.90553 .0625 Mdot
.90553 .07258 Mdot
.90553 .08267 Mdot
.90607 .0625 Mdot
.90607 .06906 Mdot
.90607 .07234 Mdot
.90607 .08219 Mdot
.9066 .0625 Mdot
.9066 .07212 Mdot
.9066 .08173 Mdot
.90677 .06886 Mdot
.90677 .08158 Mdot
.9071 .0625 Mdot
.9071 .08129 Mdot
.90742 .06867 Mdot
.90742 .08101 Mdot
.90803 .0625 Mdot
.90803 .06849 Mdot
.90803 .08047 Mdot
.90861 .06832 Mdot
.90861 .07997 Mdot
.90888 .0625 Mdot
.90888 .07973 Mdot
.90915 .06816 Mdot
.90915 .07949 Mdot
.90967 .0625 Mdot
.90967 .06801 Mdot
.90967 .07904 Mdot
.91016 .06787 Mdot
.91016 .07861 Mdot
.9104 .0625 Mdot
.9104 .0784 Mdot
.91063 .06773 Mdot
.91063 .0782 Mdot
.91107 .0625 Mdot
.91107 .0676 Mdot
.91107 .07781 Mdot
.91149 .06748 Mdot
.91149 .07744 Mdot
.9117 .0625 Mdot
.9117 .07727 Mdot
.91189 .06736 Mdot
.91189 .07709 Mdot
.91228 .0625 Mdot
.91228 .07676 Mdot
.91282 .0625 Mdot
.91282 .07628 Mdot
.91333 .0625 Mdot
.91333 .07584 Mdot
.91381 .0625 Mdot
.91381 .07542 Mdot
.91425 .0625 Mdot
.91425 .07503 Mdot
.91467 .0625 Mdot
.91467 .07466 Mdot
.91507 .0625 Mdot
.91507 .07431 Mdot
.91545 .0625 Mdot
.91545 .07399 Mdot
.9158 .0625 Mdot
.9158 .07367 Mdot
.91614 .0625 Mdot
.91614 .07338 Mdot
.91646 .0625 Mdot
.91646 .0731 Mdot
.91676 .0625 Mdot
.91676 .07284 Mdot
.91705 .0625 Mdot
.91705 .07258 Mdot
.91732 .0625 Mdot
.91732 .07234 Mdot
.91758 .0625 Mdot
.91758 .07212 Mdot
.91807 .07169 Mdot
.91852 .0713 Mdot
.91893 .07094 Mdot
.91931 .07061 Mdot
.91966 .0703 Mdot
.91998 .07002 Mdot
.92028 .06975 Mdot
.92056 .06951 Mdot
.92082 .06928 Mdot
.92107 .06906 Mdot
.9213 .06886 Mdot
.92152 .06867 Mdot
.92172 .06849 Mdot
.92192 .06832 Mdot
.9221 .06816 Mdot
.92227 .06801 Mdot
.92243 .06787 Mdot
.92259 .06773 Mdot
.92274 .0676 Mdot
.92288 .06748 Mdot
.92301 .06736 Mdot
% End of Graphics
MathPictureEnd
\
\>"], "Graphics",
  ImageSize->{449, 224.5},
  ImageMargins->{{43, 0}, {0, 0}},
  ImageRegion->{{0, 1}, {0, 1}},
  ImageCache->GraphicsData["Bitmap", "\<\
CF5dJ6E]HGAYHf4PAg9QL6QYHg<PAVmbKF5d0`4000710000h2000`400?l00000o`00003oo`3oooo2
0?ooo`00o`3oooo20?ooo`00o`3oooo20?ooo`00o`3oooo20?ooo`00=`3oool2000003D0oooo1000
00030?ooo`800000<@3oool4000000<0oooo0P00000c0?ooo`800000103oool2000003@0oooo00<0
00000?ooo`3oool00P3oool200000340oooo0`0000040?ooo`800000<`3oool2000000@0oooo0P00
000U0?ooo`00=P3oool010000000oooo0?ooo`00000f0?ooo`050000003oool0oooo0?ooo`000000
0P3oool00`000000oooo0?ooo`0^0?ooo`030000003oool0oooo00<0oooo00@000000?ooo`3oool0
0000<@3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000003<0oooo00D00000
0?ooo`3oool0oooo000000020?ooo`030000003oool0oooo0340oooo00@000000?ooo`3oool00000
0P3oool00`000000oooo0?ooo`0_0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3o
ool00000903oool003H0oooo00@000000?ooo`3oool00000=P3oool01@000000oooo0?ooo`3oool0
00000080oooo00<000000?ooo`3oool0;`3oool00`000000oooo0?ooo`020?ooo`040000003oool0
oooo000003@0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`0^0?ooo`D00000
0P3oool010000000oooo0?ooo`00000c0?ooo`040000003oool0oooo00000080oooo00<000000?oo
o`3oool0;`3oool010000000oooo0?ooo`0000020?ooo`040000003oool0oooo000002@0oooo000f
0?ooo`040000003oool0oooo000003H0oooo00D000000?ooo`3oool0oooo000000020?ooo`030000
003oool0oooo0300oooo00D000000?ooo`3oool0oooo000000020?ooo`030000003oool0oooo0300
oooo0P0000030?ooo`040000003oool0oooo00000300oooo00@000000?ooo`3oool000000`3oool0
10000000oooo0?ooo`00000`0?ooo`<000000`3oool010000000oooo0?ooo`00000a0?ooo`<00000
0`3oool010000000oooo0?ooo`00000T0?ooo`00=P3oool010000000oooo0?ooo`00000f0?ooo`05
0000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`0a0?ooo`040000003oool0oooo
00000080oooo00<000000?ooo`3oool0<P3oool010000000oooo0?ooo`0000020?ooo`030000003o
ool0oooo02l0oooo00<000000?ooo`0000000`3oool010000000oooo0?ooo`00000`0?ooo`030000
003oool0oooo00<0oooo00@000000?ooo`3oool00000<@3oool00`000000oooo0?ooo`030?ooo`04
0000003oool0oooo000002@0oooo000f0?ooo`040000003oool0oooo000003@0oooo0`0000030?oo
o`040000003oool0oooo00000300oooo00@000000?ooo`3oool000000P3oool010000000oooo0?oo
o`00000a0?ooo`040000003oool0oooo00000080oooo00@000000?ooo`3oool00000<P3oool20000
00<0oooo00@000000?ooo`3oool00000<03oool00`000000oooo0?ooo`030?ooo`040000003oool0
oooo00000340oooo00<000000?ooo`3oool00`3oool010000000oooo0?ooo`00000T0?ooo`00=`3o
ool2000003L0oooo00<000000?ooo`3oool00P3oool200000380oooo0P0000040?ooo`800000<`3o
ool2000000@0oooo0P00000d0?ooo`030000003oool0oooo0080oooo0P00000a0?ooo`@000000`3o
ool2000003<0oooo0`0000030?ooo`8000009@3oool00?l0oooo`P3oool00?l0oooo`P3oool00?l0
oooo`P3oool001/0ooooo`00002L000000/0oooo000K0?ooo`030000003oool0oooo00<0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0oooo00<0
00000?ooo`3oool02@3oool00`000000oooo0?ooo`030?ooo`030000003oool0oooo00T0oooo000K
0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`08
0?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`09
0?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`03
0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo01X0oooo00<000000?oo
o`3oool0>03oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo03P0oooo00<000000?oo
o`3oool0>03oool00`000000oooo0?ooo`0g0?ooo`030000003oool0oooo03P0oooo00<000000?oo
o`3oool06P3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`3o0?oooiP0
oooo00<000000?ooo`3oool02@3oool001/0oooo0`00003o0?oooiH0oooo0`00000;0?ooo`006`3o
ool00`000000oooo0?ooo`3o0?oooiP0oooo00<000000?ooo`3oool02@3oool001/0oooo00<00000
0?ooo`3oool0o`3ooonH0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo
0?l0ooooV03oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`3o0?oooiP0
oooo00<000000?ooo`3oool02@3oool001/0oooo0`00003o0?oooiH0oooo0`00000;0?ooo`006`3o
ool00`000000oooo0?ooo`3o0?oooiP0oooo00<000000?ooo`3oool02@3oool001/0oooo00<00000
0?ooo`3oool0o`3ooonH0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo
0?l0ooooV03oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`3o0?oooiP0
oooo00<000000?ooo`3oool02@3oool001/0oooo0`00003o0?oooiH0oooo0`00000;0?ooo`006`3o
ool00`000000oooo0?ooo`3o0?oooiP0oooo00<000000?ooo`3oool02@3oool001/0oooo00<00000
0?ooo`3oool0o`3ooonH0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo
0?l0ooooV03oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`3o0?oooiP0
oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0o`3ooonH0?ooo`030000
003oool0oooo00T0oooo000K0?ooo`<00000o`3ooonF0?ooo`<000002`3oool001/0oooo00<00000
0?ooo`3oool0o`3ooonH0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo
0?l0ooooV03oool00`000000oooo0?ooo`090?ooo`004P3oool2000000L0oooo00<000000?ooo`3o
ool0o`3ooonH0?ooo`030000003oool0oooo00T0oooo000A0?ooo`040000003oool0oooo000000H0
oooo00<000000?ooo`3oool06P3oool2000000<0oooo4P00000K0?ooo`h000000P3oool500000003
0?ooo`000000000002D000000P3oool2000000030?ooo`000000000000L000000`3oool2000000<0
oooo2`000000103oool000000000003ooolJ000000H0oooo4`0000000`3oool000000000000C0000
00<0oooo0P0000030?ooo``0000000<0oooo00000000000020000000103oool000000000003oool4
000000030?ooo`000000000001D0000000<0oooo0000000000000`3oool>000001/0oooo4P00000P
0?ooo`030000003oool0oooo00T0oooo000A0?ooo`040000003oool0oooo000000H0oooo1000000I
0?ooo`8000000`3ooolB000001/0oooo3P0000020?ooo`D0000000<0oooo0000000000009@000002
0?ooo`80000000<0oooo0000000000001`0000030?ooo`8000000`3oool;000000040?ooo`000000
00000?oooaX000001P3ooolC000000030?ooo`000000000001<000000`3oool2000000<0oooo3000
00000`3oool0000000000008000000040?ooo`00000000000?ooo`@0000000<0oooo000000000000
5@0000000`3oool00000000000030?ooo`h000006`3ooolB000001d0oooo1000000;0?ooo`004@3o
ool010000000oooo0?ooo`0000060?ooo`030000003oool0oooo01d0oooo0P0000020?ooo`@00000
00<0oooo0000000000002@00000K0?ooo``000002@3oool=000000040?ooo`00000000000?ooo`H0
000000<0oooo0000000000000P000000103oool000000000003oool6000000050?ooo`0000000000
0000003oool05@0000000`3oool0000000000003000000040?ooo`00000000000?ooo`80000000@0
oooo000000000000oooo1`000000103oool000000000003oool8000000040?ooo`00000000000000
00H0oooo0`0000000`3oool000000000000800000080oooo3@0000000`3oool000000000000L0000
00030?ooo`000000000000P0000000@0oooo000000000000oooo100000000`3oool0000000000004
000000030?ooo`000000000000d000002@3oool=000001/0oooo2`0000000`3oool0000000000002
00000080oooo0P00000N0?ooo`030000003oool0oooo00T0oooo000A0?ooo`040000003oool0oooo
000000H0oooo00<000000?ooo`3oool07@3oool3000000030?ooo`000000000000h000006`3ooola
000000040?ooo`00000000000?ooo`H0000000<0oooo0000000000005`0000000`3oool000000000
0006000000040?ooo`00000000000?oooaX000001P3ooolM000000030?ooo`000000000002L00000
00@0oooo000000000000oooo6`0000000`3oool0000000000002000000030?ooo`000000000000h0
00006`3ooolD000001h0oooo00<000000?ooo`3oool02@3oool00140oooo00@000000?ooo`3oool0
00001P3oool00`000000oooo0?ooo`0M0?ooo`D0000000<0oooo0000000000003@00000H0?ooob40
00000`3oool2000000030?ooo`000000000000H0000000<0oooo0000000000000P0000000`3oool0
000000000002000000030?ooo`000000000000H000000P3oool200000080oooo60000000103oool0
00000000003ooolB000000030?ooo`000000000000D000001P3ooolf000000040?ooo`0000000000
00000080oooo2P0000000`3oool000000000000>000000030?ooo`00000000000080oooo3`000000
0`3oool000000000000@000001P0oooo3P0000020?ooo`@000007`3oool00`000000oooo0?ooo`09
0?ooo`004P3oool2000000L0oooo00<000000?ooo`3oool07`3ooolC000001P0oooo400000060?oo
o`l0000000@0oooo000000000000oooo1P0000000`3oool000000000000=000000050?ooo`000000
00000000003oool070000000103oool000000000003ooolB000000050?ooo`00000000000000003o
ool00`0000060?ooo`<0000000D0oooo00000000000000000?ooo`0;000000030?ooo`0000000000
02@0000000D0oooo00000000000000000?ooo`0<000000030?ooo`00000000000080000000<0oooo
0000000000000P000000103oool000000000003oool?000000H0oooo4000000H0?ooo`h0000000<0
oooo0000000000000P00000P0?ooo`030000003oool0oooo00T0oooo000K0?ooo`<00000803oool4
000000030?ooo`000000000000`000006P3ooolA000000<0oooo5`0000001@3oool0000000000000
0000oooo00d0000000@0oooo000000000000oooo2@0000020?ooo`D0000000<0oooo000000000000
0`0000000`3oool0000000000003000000040?ooo`00000000000?oooa<0000000<0oooo00000000
0000100000060?ooo`H0000000<0oooo000000000000500000000`3oool000000000000500000003
0?ooo`00000000000080oooo1@0000000`3oool0000000000007000000050?ooo`00000000000000
003oool01`000000103oool000000000003oool2000000050?ooo`00000000000000003oool05P00
00030?ooo`80000000<0oooo0000000000003@00000J0?ooo`h0000000<0oooo0000000000000P00
000O0?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool08@3oool4000000030?ooo`000000
000000/000006P3ooolQ000000040?ooo`00000000000?ooo`H0000000<0oooo0000000000001@00
00020?oooa80000000<0oooo0000000000002P0000000`3oool0000000000006000000040?ooo`00
000000000?ooo`d0000000<0oooo0000000000002P0000060?oooad0000000<0oooo000000000000
200000000`3oool0000000000007000000030?ooo`000000000000L0000000<0oooo000000000000
1P0000030?ooo`L0000000<0oooo000000000000100000000`3oool000000000000?000000040?oo
o`00000000000?ooo`l000006P3oool=000000030?ooo`0000000000008000008P3oool00`000000
oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`0R0?ooo`@0000000<0oooo000000000000
2P00000G0?oooaL0000000<0oooo000000000000700000000`3oool0000000000005000000040?oo
o`00000000000?ooo`T000000P3oool2000000030?ooo`000000000000<000000`3oool800000004
0?ooo`00000000000?oooaX000001P3ooolV000000<0oooo1@0000000`3oool00000000000020?oo
o`T0000000<0oooo000000000000200000000`3oool0000000000006000000030?ooo`0000000000
0180000000@0oooo000000000000oooo5@00000F0?ooo`d0000000<0oooo0000000000000P00000S
0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo02<0oooo100000000`3o
ool0000000000009000001L0oooo500000050?ooo`P0000000<0oooo0000000000000P0000000`3o
ool0000000000006000000030?ooo`0000000000008000000`3oool;00000080oooo3P0000000`3o
ool0000000000002000000030?ooo`000000000000H0000000@0oooo000000000000oooo3P000000
1@3oool00000000000000000oooo00<0000000@0oooo00000000000000001P3oool3000000030?oo
o`000000000000D0000000<0oooo0000000000004@0000000`3oool0000000000004000000030?oo
o`00000000000080000000<0oooo000000000000300000020?ooo`/000000P3oool2000000030?oo
o`000000000000T0000000<0oooo0000000000002`0000040?oooaD000005P3oool<000000030?oo
o`000000000000800000903oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?oo
o`0T0?ooo`@0000000<0oooo0000000000002@00000I0?oooa40000000<0oooo0000000000006000
00000`3oool0000000000005000000030?ooo`000000000000T0000000<0oooo0000000000001000
0000103oool000000000003oool<000000030?ooo`0000000000008000000P3oool200000080oooo
6@0000060?oooa<0000000<0oooo0000000000000`000000103oool000000000000000020?oooa40
000000@0oooo000000000000oooo1@0000020?ooo`X0000000@0oooo000000000000oooo7@000000
0`3oool0000000000002000000030?ooo`000000000000d00000603ooolA000002D0oooo00<00000
0?ooo`3oool02@3oool001/0oooo0`00000U0?ooo`l000006@3oool=000000D0oooo6@0000000`3o
ool0000000000006000000030?ooo`000000000000T0000000<0oooo000000000000500000000`3o
ool0000000000003000000040?ooo`00000000000?oooaX000000P3oool200000080oooo6@000000
0`3oool00000000000020?oooa/000000P3oool:000000030?ooo`0000000000024000001@3oool=
000001P0oooo3`00000U0?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool09`3oool>0000
01T0oooo400000001@3oool00000000000000000oooo00/0000000@0oooo000000000000oooo2`00
00000`3oool000000000000A000000030?ooo`000000000000d0000000<0oooo0000000000001P00
0000103oool000000000003ooolJ00000080oooo0P0000020?oooaH0000000<0oooo000000000000
0`0000020?oooaL0000000<0oooo000000000000100000000`3oool000000000000;000000030?oo
o`000000000000<0000000<0oooo0000000000000`0000000`3oool000000000000<000000030?oo
o`000000000000<0000000<0oooo0000000000003000000H0?ooo`X0000000<0oooo000000000000
0P00000X0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo02P0oooo3P00
000G0?oooa@000000P3ooolI00000080oooo0P0000000`3oool000000000000600000080oooo3000
00030?ooo`/0000000<0oooo0000000000000P000000103oool000000000003ooolJ000000H0oooo
6P0000000`3oool0000000000005000000030?ooo`000000000000T000000`3oool;00000080oooo
2@0000000`3oool00000000000020?ooo`@0000000<0oooo0000000000000`0000000`3oool00000
00000002000000030?ooo`000000000000L000000P3oool4000000030?ooo`000000000000`00000
603oool9000000030?ooo`000000000000800000:@3oool00`000000oooo0?ooo`090?ooo`006`3o
ool00`000000oooo0?ooo`0Y0?ooo`@0000000<0oooo0000000000001`00000F0?oooa@000000P3o
ool900000080oooo0P000000103oool000000000003oool;00000080oooo20000000103oool00000
0000003oool3000000030?ooo`000000000000/0000000<0oooo0000000000001P0000000`3oool0
000000000003000000030?ooo`000000000000`0000000<0oooo0000000000000`3oool200000004
0?ooo`0000000000000000H0oooo0`0000000`3oool00000000000030?ooo`80000000<0oooo0000
000000002`0000020?ooo`@0000000<0oooo0000000000001`0000000`3oool000000000000?0000
00040?ooo`00000000000?ooo`P000000P3oool2000000030?ooo`000000000000T0000000<0oooo
0000000000000P3oool9000000040?ooo`00000000000?oooa8000005P3oool9000000030?ooo`00
0000000000800000:P3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`0Z
0?ooo`@0000000<0oooo0000000000001P00000F0?oooa@0000000<0oooo0000000000001`000002
0?ooo`<0000000@0oooo000000000000oooo40000000103oool000000000003oool4000000050?oo
o`00000000000000003oool00`0000030?ooo`X000000P3oool4000000030?ooo`000000000000<0
000000<0oooo00000000000030000000103oool000000000003oool8000000H0oooo20000000103o
ool000000000003oool=000000040?ooo`00000000000?ooo`80000000<0oooo0000000000000P00
00020?ooo`/000000P3oool8000000030?ooo`000000000000X0000000<0oooo0000000000002@00
0000103oool000000000003oool3000000030?ooo`000000000000L0000000D0oooo000000000000
00000?ooo`0A000001H0oooo200000000`3oool0000000000002000002/0oooo00<000000?ooo`3o
ool02@3oool001/0oooo0`00000[0?ooo`@0000000<0oooo0000000000001`00000E0?ooo``00000
0P3oool5000000030?ooo`000000000000/0000000@0oooo000000000000oooo2P0000020?oooaL0
000000@0oooo000000000000oooo3P0000030?ooo`80000000<0oooo0000000000000`0000020?oo
o`L0000000<0oooo0000000000002@0000060?ooo`/0000000<0oooo0000000000001@000000103o
ool000000000003oool5000000@0oooo3P000000103oool000000000003ooolF000000<0oooo2P00
0000103oool000000000003oool<00000080oooo1@0000020?ooo``000005@3oool800000080oooo
1000000Z0?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool0;03oool4000000030?ooo`00
0000000000H000005@3oool<00000080oooo0P000000103oool000000000003oool=000000040?oo
o`00000000000?ooo`H0000000D0oooo00000000000000000?ooo`02000000030?ooo`0000000000
00D0000000<0oooo0000000000001`0000000`3oool000000000000E00000080oooo700000060?oo
oaH0000000<0oooo0000000000000`0000000`3oool000000000000B000000030?ooo`0000000000
00D0000000<0oooo0000000000002`000000103oool000000000003oool:000000040?ooo`000000
00000?ooo`d0000000<0oooo0000000000000`0000020?ooo``000005@3oool<000002h0oooo00<0
00000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0;@3oool4000000030?ooo`000000
000000D00000603oool900000080oooo0P0000040?ooo`80000000<0oooo0000000000001`000000
0`3oool000000000000200000080oooo2`0000001@3oool00000000000000000oooo00<0000000<0
oooo0000000000006@0000000`3oool000000000000300000080oooo1@0000000`3oool000000000
000D000000H0oooo5@0000020?ooo`80000000@0oooo000000000000oooo1@0000020?oooaD00000
00<0oooo0000000000002`0000001@3oool00000000000000000oooo00L0000000@0oooo00000000
0000oooo100000000`3oool0000000000005000000030?ooo`000000000000<0oooo0`0000020?oo
o`T00000603oool600000080oooo0`00000_0?ooo`030000003oool0oooo00T0oooo000K0?ooo`03
0000003oool0oooo02h0oooo3000000F0?ooo`X000000P3oool2000000030?ooo`00000000000080
00000P3oool>00000080oooo0`0000000`3oool0000000000003000000030?ooo`000000000001H0
000000<0oooo0000000000004@0000000`3oool0000000000006000000030?ooo`00000000000140
00001P3ooolC000000030?ooo`000000000000<0000000@0oooo000000000000oooo0P0000000`3o
ool000000000000>000000030?ooo`000000000000d0000000<0oooo000000000000400000000`3o
ool000000000000=000000040?ooo`00000000000?ooo`D0000000<0oooo0000000000002000000F
0?ooo`L0000000<0oooo0000000000000P00000`0?ooo`030000003oool0oooo00T0oooo000K0?oo
o`030000003oool0oooo0300oooo100000000`3oool0000000000004000001<0oooo300000020?oo
o`80000000<0oooo0000000000000P0000000`3oool000000000000=000000030?ooo`0000000000
00/0000000<0oooo0000000000001P0000000`3oool000000000000:00000080oooo200000020?oo
o`X000000P3ooolE000000040?ooo`0000000000000000H0oooo0`0000000`3oool000000000000C
000000040?ooo`00000000000?ooo`P0000000<0oooo0000000000001`0000000`3oool000000000
000;000000030?ooo`000000000000D0000000<0oooo00000000000030000000103oool000000000
003oool>000000040?ooo`00000000000?ooo`80000000<0oooo0000000000002`00000C0?ooo`H0
000000<0oooo0000000000000P00000a0?ooo`030000003oool0oooo00T0oooo000K0?ooo`<00000
<@3oool;00000180oooo300000020?ooo`@0000000@0oooo000000000000oooo3P0000020?ooo`D0
000000<0oooo0000000000001@0000001@3oool00000000000000000oooo00/0000000<0oooo0000
000000000`0000020?ooo`H0000000@0oooo000000000000oooo1`0000000`3oool0000000000002
0?ooo`H0000000<0oooo000000000000400000060?oooa80000000<0oooo00000000000010000000
103oool000000000003oool8000000040?ooo`00000000000?ooo`H0000000@0oooo000000000000
oooo0`0000000`3oool000000000000<000000030?ooo`000000000000`0000000D0oooo00000000
000000000?ooo`0;000000030?ooo`00000000000080000000@0oooo000000000000oooo3@00000B
0?ooo`/00000<03oool3000000/0oooo000K0?ooo`030000003oool0oooo0380oooo100000000`3o
ool0000000000004000001@0oooo2@0000020?ooo`@0000000@0oooo00000000000000000P3oool<
000000030?ooo`000000000000/0000000<0oooo0000000000000P3oool3000000040?ooo`000000
00000?oooah0000000<0oooo0000000000002@0000020?oooa0000000P3oool200000080oooo3`00
00020?ooo`P0000000@0oooo000000000000oooo3@000000103oool000000000003oool400000003
0?ooo`000000000000L0000000<0oooo000000000000100000020?ooo`80000000<0oooo00000000
000020000000103oool000000000003oool2000000040?ooo`00000000000?ooo`H0000000<0oooo
0000000000001`0000000`3oool0000000000008000001@0oooo1P0000000`3oool0000000000002
000003<0oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0<`3oool40000
00030?ooo`000000000000@000004`3oool900000080oooo200000020?ooo`P0000000<0oooo0000
000000003@0000000`3oool000000000000`000000030?ooo`00000000000100000000@0oooo0000
0000000000000P3oool200000080oooo0`0000000`3oool000000000000@000000030?ooo`000000
000000<0000000<0oooo0000000000003P0000001@3oool00000000000000000oooo0180000000<0
oooo0000000000002`0000001@3oool00000000000000000oooo00/0000000<0oooo000000000000
1`0000000`3oool0000000000008000001<0oooo1P0000000`3oool0000000000002000003@0oooo
00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0=03oool4000000030?ooo`00
0000000000@00000503oool8000000030?ooo`000000000000H000000P3oool=000000040?ooo`00
000000000?ooo`X0000000<0oooo0000000000001@000000103oool000000000003oool:00000004
0?ooo`00000000000?oooa<0000000<0oooo0000000000000P3oool<00000080oooo1`0000060?oo
o`L000000P3oool<000000030?ooo`000000000000@0000000@0oooo000000000000oooo3@000000
0`3oool000000000000;000000030?ooo`000000000000P0000000<0oooo0000000000001`000002
0?ooo`@0000000<0oooo0000000000002@0000020?ooo`P0000000<0oooo0000000000001P00000D
0?ooo`D000000P3oool4000003D0oooo00<000000?ooo`3oool02@3oool00140oooo0`0000070?oo
o`030000003oool0oooo03H0oooo100000000`3oool0000000000002000001D0oooo1`0000000`3o
ool0000000000006000000040?ooo`00000000000?ooo`X0000000D0oooo00000000000000000?oo
o`0C000000040?ooo`00000000000?ooo`T0000000@0oooo00000000000000000P3oool>00000004
0?ooo`00000000000?ooo`@0000000<0oooo0000000000004P0000060?ooo`80000000<0oooo0000
00000000500000000`3oool0000000000003000000030?ooo`000000000000T000000P3oool20000
00030?ooo`000000000000D0000000@0oooo000000000000oooo3`0000000`3oool0000000000004
000000030?ooo`000000000000<0000000<0oooo0000000000001@0000020?ooo`80000000<0oooo
0000000000001P0000000`3oool0000000000005000001@0oooo1@0000000`3oool0000000000002
000003L0oooo00<000000?ooo`3oool02@3oool001@0oooo00<000000?ooo`3oool0103oool40000
03H0oooo2@00000E0?ooo`D000000P3oool700000080oooo0P0000000`3oool00000000000080000
00040?ooo`00000000000?ooo`T0000000<0oooo0000000000000P0000020?ooo`L0000000<0oooo
0000000000006`000000103oool000000000003ooolE000000040?ooo`0000000000000000H0oooo
0P0000020?oooaD0000000<0oooo0000000000009P000000103oool000000000003oool<00000003
0?ooo`000000000000<0000000<0oooo0000000000001@000000103oool000000000003oool80000
00030?ooo`000000000000<000005@3oool:000003D0oooo1000000;0?ooo`00503oool00`000000
oooo0?ooo`040?ooo`030000003oool0oooo03P0oooo2P00000A0?ooo`L000000P3oool700000004
0?ooo`00000000000?ooo`L0000000@0oooo000000000000oooo0P0000020?ooo`<0000000<0oooo
0000000000000P0000000`3oool0000000000002000000<0oooo300000000`3oool0000000000003
000000030?ooo`00000000000100000000@0oooo000000000000oooo1@0000030?ooo`D000000P3o
ool:000000H0oooo2P0000020?ooo`D000000P3oool6000000040?ooo`00000000000?ooo`d00000
00D0oooo00000000000000000?ooo`06000000030?ooo`000000000000@0000000<0oooo00000000
00000`0000030?ooo`@0000000<0oooo0000000000001P0000020?ooo``000000P3oool:00000003
0?ooo`000000000000H000004@3oool5000000030?ooo`000000000000800000>@3oool00`000000
oooo0?ooo`090?ooo`004@3oool3000000L0oooo00<000000?ooo`3oool0>@3oool4000000030?oo
o`000000000000<00000403oool8000000030?ooo`00000000000080000000@0oooo000000000000
oooo0P0000000`3oool000000000000@000000030?ooo`000000000000L0000000<0oooo00000000
00002`0000000`3oool0000000000009000000030?ooo`000000000000/0000000@0oooo00000000
0000oooo6@0000060?oooa`0000000<0oooo0000000000003`0000000`3oool000000000000C0000
00030?ooo`000000000000`0000000<0oooo0000000000002`000000103oool000000000003oool7
000000030?ooo`000000000000H00000403oool5000000030?ooo`000000000000800000>P3oool0
0`000000oooo0?ooo`090?ooo`004@3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo
03X0oooo100000000`3oool000000000000300000100oooo1`0000000`3oool00000000000020000
00040?ooo`00000000000?ooo`80000000<0oooo0000000000003P0000020?ooo`P0000000<0oooo
0000000000002`0000000`3oool000000000000@000000030?ooo`000000000000H0000000<0oooo
0000000000002@0000020?ooo`H0000000<0oooo0000000000001P0000060?ooo`L0000000<0oooo
0000000000001@0000020?oooa@0000000<0oooo000000000000100000000`3oool0000000000008
00000080oooo3@0000000`3oool0000000000006000000030?ooo`000000000000l0000000@0oooo
000000000000oooo0P0000000`3oool0000000000002000000030?ooo`000000000000D00000403o
ool5000000030?ooo`000000000000800000>`3oool00`000000oooo0?ooo`090?ooo`004@3oool0
0`000000oooo0?ooo`070?ooo`030000003oool0oooo03/0oooo100000000`3oool0000000000002
00000140oooo1@0000020?ooo``000000P3oool<000000030?ooo`000000000000/0000000<0oooo
0000000000001P0000000`3oool000000000000;000000030?ooo`000000000000@000000`3ooolC
000000030?ooo`000000000000d000001P3oool?000000030?ooo`0000000000014000000`3oool6
000000030?ooo`000000000000P0000000@0oooo000000000000oooo200000000`3oool000000000
000;000000030?ooo`000000000000T000000P3oool2000000030?ooo`000000000000P0000000<0
oooo0000000000000`00000B0?ooo`@0000000<0oooo0000000000000P00000l0?ooo`030000003o
ool0oooo00T0oooo000A0?ooo`@000001P3oool3000003`0oooo100000020?ooo`<000004@3oool4
00000080oooo1P000000103oool000000000003oool2000000030?ooo`000000000000X000000P3o
ool2000000030?ooo`000000000000H0000000<0oooo0000000000002`0000000`3oool000000000
0008000000030?ooo`000000000000D000000`3oool800000080oooo50000000103oool000000000
000000060?ooo`<0000000@0oooo000000000000oooo200000000`3oool000000000000600000003
0?ooo`000000000000H000000`3ooolB00000080oooo100000000`3oool000000000000600000003
0?ooo`000000000000H0000000<0oooo0000000000000P3oool;000000050?ooo`00000000000000
003oool00P0000000`3oool0000000000004000000030?ooo`000000000000<000004@3oool30000
0080oooo0`00000l0?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool0?@3oool400000080
oooo0`00000A0?ooo`@0000000<0oooo000000000000100000000`3oool000000000000300000004
0?ooo`00000000000?ooo`T0000000<0oooo0000000000000P3ooolD000000030?ooo`0000000000
00X0000000<0oooo0000000000000P0000000`3oool000000000000<000000030?ooo`0000000000
0080000000<0oooo0000000000003@000000103oool000000000000000060?ooo`<0000000<0oooo
000000000000300000020?ooo`@0000000@0oooo000000000000oooo2`0000001@3oool000000000
00000000oooo00X0000000@0oooo000000000000oooo300000000`3oool000000000000500000080
oooo0P0000000`3oool000000000000:000000030?ooo`000000000000<0000000<0oooo00000000
0000100000000`3oool000000000000200000140oooo0`0000000`3oool0000000000002000003l0
oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0?P3oool400000080oooo
0`00000B0?ooo`80000000<0oooo0000000000001P0000000`3oool00000000000020?ooo`l00000
00<0oooo000000000000100000000`3oool000000000000:000000030?ooo`000000000000800000
00<0oooo000000000000200000001@3oool00000000000000000oooo00L0000000<0oooo00000000
0000100000000`3oool0000000000004000000030?ooo`000000000000X000000P3oool3000000H0
oooo0P0000030?ooo`H0000000<0oooo0000000000002P000000103oool000000000003oool20000
0080oooo1`0000001@3oool00000000000000000oooo00X0000000<0oooo0000000000000P000000
1@3oool00000000000000000oooo00P0000000<0oooo000000000000100000001@3oool000000000
00000000oooo00/000000P3oool2000000030?ooo`000000000000H0000000<0oooo000000000000
4P3oool300000080oooo0`0000100?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003o
ool0oooo0400oooo100000000`3oool0000000000002000001<0oooo20000000103oool000000000
003oool>000000030?ooo`000000000000d0000000<0oooo0000000000004`0000030?ooo`H00000
00<0oooo0000000000001`0000000`3oool000000000000:00000080oooo2P0000060?ooo`800000
00<0oooo0000000000001@0000000`3oool0000000000008000000040?ooo`00000000000?ooo`P0
000000<0oooo0000000000001@0000030?oooaD0000000<0oooo0000000000006@000000103oool0
00000000003oool2000000030?ooo`000000000000H000004`3oool4000000030?ooo`0000000000
00800000@@3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`110?ooo`T0
00004P3ooolM000000030?ooo`00000000000280000000<0oooo0000000000003P0000000`3oool0
00000000000F000000H0oooo5@000000103oool000000000003oool@000000030?ooo`0000000000
0280000000<0oooo000000000000300000000`3oool000000000000<00000180oooo100000000`3o
ool000000000000200000480oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3o
ool0@P3oool4000000030?ooo`0000000000008000004P3oool<000000030?ooo`000000000002L0
000000<0oooo0000000000002@0000000`3oool000000000000=000000030?ooo`000000000000H0
000000<0oooo000000000000300000070?oooaP0000000@0oooo000000000000oooo300000000`3o
ool000000000000b00000080oooo0P0000000`3oool000000000000700000180oooo100000000`3o
ool0000000000002000004<0oooo00<000000?ooo`3oool02@3oool001/0oooo0`0000130?ooo`@0
000000<0oooo0000000000000P00000@0?ooo``000000P3oool200000080oooo100000000`3oool0
00000000000E000000030?ooo`000000000000@0000000@0oooo000000000000oooo10000000103o
ool000000000003oool2000000050?ooo`00000000000000003oool020000000103oool000000000
003oool5000000040?ooo`00000000000?ooo`L0000000<0oooo0000000000001@0000060?ooo`H0
00000P3oool@000000040?ooo`00000000000?ooo`P0000000D0oooo00000000000000000?ooo`02
000000030?ooo`000000000000D0000000@0oooo000000000000oooo1P0000000`3oool000000000
000J000000030?ooo`00000000000080oooo3@00000@0?ooo`@0000000<0oooo0000000000000P00
00120?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool0A03oool4000000030?ooo`000000
0000008000003`3oool<000000030?ooo`000000000000<0oooo2P0000020?ooo`80000000<0oooo
0000000000001`0000000`3oool000000000000;000000030?ooo`000000000000D0000000<0oooo
000000000000400000000`3oool000000000000>000000@0oooo100000060?ooo`@00000103ooolK
000000030?ooo`000000000000D0000000<0oooo000000000000100000020?ooo`d0000000<0oooo
0000000000001`0000000`3oool00000000000020?ooo`X000000P3oool200000080oooo3000000?
0?ooo`P00000AP3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`150?oo
o`T000003P3oool;00000080oooo0P0000020?ooo``000000`3ooolF000000030?ooo`0000000000
02/0000000<0oooo0000000000001@0000060?ooo`H000000P3oool?000000030?ooo`0000000000
0100000000<0oooo0000000000001@000000103oool000000000003ooolF000000<0oooo30000000
1@3oool00000000000000000oooo00`000003P3oool8000004L0oooo00<000000?ooo`3oool02@3o
ool001/0oooo00<000000?ooo`3oool0A`3oool4000000040?ooo`0000000000000000h0oooo3P00
00001@3oool00000000000000000oooo00T0000000<0oooo0000000000002`0000000`3oool00000
0000000;000000030?ooo`000000000000L0000000<0oooo0000000000002P0000000`3oool00000
00000005000000030?ooo`000000000000@0000000<0oooo0000000000002P0000060?ooo``00000
00<0oooo000000000000100000001@3oool00000000000000000oooo00<0000000<0oooo00000000
00001`0000000`3oool0000000000007000000030?ooo`000000000000h0000000<0oooo00000000
00002`0000000`3oool000000000000600000080oooo0`0000000`3oool000000000000<000000h0
oooo0`0000000`3oool0000000000002000004P0oooo00<000000?ooo`3oool02@3oool001/0oooo
00<000000?ooo`3oool0B03oool4000000040?ooo`0000000000000000d0oooo3@0000001@3oool0
0000000000000000oooo00d0000000<0oooo0000000000009`0000000`3oool00000000000070000
00030?ooo`000000000000P000000P3oool6000000<0oooo0P0000060?ooo`8000000P3oool70000
0080oooo2P0000000`3oool0000000000007000000050?ooo`00000000000000003oool09@000000
0`3oool000000000000;000000050?ooo`00000000000000003oool00`0000000`3oool000000000
0007000000d0oooo0`0000000`3oool0000000000002000004T0oooo00<000000?ooo`3oool02@3o
ool001/0oooo0`0000190?ooo`@0000000<0oooo0000000000003P3oool:00000080oooo0`000002
0?oooa/0000000<0oooo0000000000006@0000000`3oool0000000000004000000040?ooo`000000
00000?ooo`X0000000<0oooo0000000000001P0000020?ooo`8000000P3oool200000080oooo0P00
00020?oooa00000000@0oooo000000000000oooo0P0000000`3oool000000000000=000000030?oo
o`00000000000100000000<0oooo000000000000300000000`3oool000000000000:00000080oooo
0`0000000`3oool0000000000008000000l0oooo0P0000000`3oool0000000000002000004P0oooo
0`00000;0?ooo`006`3oool00`000000oooo0?ooo`1:0?ooo`@000004P3oool<000000030?ooo`00
0000000000T000000P3oool3000000030?ooo`000000000001<000001@3oool7000000030?ooo`00
0000000000D0000000<0oooo0000000000000P0000000`3oool000000000000E00000080oooo0P00
00020?ooo`8000000P3oool200000080oooo4`0000000`3oool00000000000020?ooo`@0000000<0
oooo00000000000010000000103oool000000000003oool5000000@0oooo5@0000000`3oool00000
0000000200000080oooo2`0000000`3oool000000000000:00000180oooo0`00001<0?ooo`030000
003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo04/0oooo1000000B0?ooo`X000000P3o
ool3000000030?ooo`000000000000H0000000<0oooo0000000000006@0000030?oooa@0000000<0
oooo0000000000002@0000000`3oool0000000000009000000040?ooo`0000000000000000H0oooo
0`0000000`3oool000000000000A000000030?ooo`00000000000080oooo500000030?ooob@00000
00@0oooo00000000000000000P3oool9000001<0oooo1000001<0?ooo`030000003oool0oooo00T0
oooo000K0?ooo`030000003oool0oooo04`0oooo1000000A0?ooo``000000`3oool?000000<0oooo
:P0000020?ooo`80000000<0oooo0000000000005@0000060?ooo`@0000000<0oooo000000000000
4`0000000`3oool0000000000002000000040?ooo`00000000000?ooo`h0000000<0oooo00000000
00001`0000000`3oool0000000000008000000<0oooo3`0000030?ooo`/000004P3oool3000004h0
oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0C@3oool400000180oooo
1@0000000`3oool0000000000007000000030?ooo`000000000000L0000000<0oooo000000000000
7`000000103oool000000000003oool:00000080oooo0P0000000`3oool000000000000=00000080
oooo1P0000060?ooo`H000000P3ooolB000000030?ooo`000000000000T0000000@0oooo00000000
0000oooo1`0000000`3oool0000000000006000000030?ooo`000000000000h0000000<0oooo0000
000000001`0000000`3oool0000000000002000000030?ooo`000000000000P000004@3oool40000
04l0oooo00<000000?ooo`3oool02@3oool001/0oooo0`00001?0?ooo`@000004P3oool<000000@0
oooo100000000`3oool000000000000U000000050?ooo`00000000000000003oool010000000103o
ool000000000003oool2000000030?ooo`000000000000`000000`3oool6000000H0oooo1P000003
0?oooa40000000@0oooo000000000000oooo2`0000000`3oool000000000000>000000030?ooo`00
0000000001H00000103oool<00000180oooo1000001>0?ooo`<000002`3oool001/0oooo00<00000
0?ooo`3oool0D03oool400000180oooo20000000103oool000000000000000040?ooo`<0000000<0
oooo0000000000000P0000020?ooobH0000000D0oooo00000000000000000?ooo`02000000030?oo
o`000000000000l000000P3oool7000000H0oooo1`0000020?ooo`d0000000@0oooo000000000000
oooo0`0000001@3oool00000000000000000oooo00/0000000<0oooo0000000000002@0000000`3o
ool000000000000;00000080oooo200000040?ooo`d000004P3oool400000540oooo00<000000?oo
o`3oool02@3oool001/0oooo00<000000?ooo`3oool0D@3oool400000140oooo300000040?ooob40
000000<0oooo0000000000003@0000000`3oool0000000000003000000030?ooo`000000000001P0
00001P3ooolF000000040?ooo`00000000000?oooal0000000<0oooo0000000000005@0000040?oo
o`d000004@3oool400000580oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3o
ool0DP3oool400000140oooo2`0000050?oooaL0000000<0oooo0000000000001`0000000`3oool0
000000000009000000030?ooo`000000000000@0000000<0oooo0000000000003P0000000`3oool0
000000000009000000H0oooo2`0000000`3oool000000000000;000000030?ooo`000000000000L0
000000<0oooo0000000000004`0000000`3oool000000000000D000000H0oooo2P00000B0?ooo`@0
0000D`3oool00`000000oooo0?ooo`090?ooo`002`3oool4000000<0oooo0P0000070?ooo`030000
003oool0oooo05<0oooo1000000B0?ooo`T000001@3ooolH000000030?ooo`000000000000X00000
00<0oooo0000000000002@0000000`3oool000000000000M000000H0oooo6@0000000`3oool00000
00000003000000030?ooo`000000000000T0000000<0oooo000000000000300000000`3oool00000
0000000C000000H0oooo2@00000B0?ooo`@00000E03oool00`000000oooo0?ooo`090?ooo`003@3o
ool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`3oool0103oool4000005<0oooo
1000000B0?ooo`L000001@3oool3000000030?ooo`000000000001@0000000<0oooo000000000000
1P000000103oool000000000003oool;000000030?ooo`00000000000080000000<0oooo00000000
00002P0000000`3oool0000000000008000000030?ooo`000000000000H0oooo0P0000000`3oool0
000000000007000000040?ooo`00000000000?ooo`H0000000<0oooo0000000000000P3oool30000
00030?ooo`000000000000X0000000@0oooo000000000000oooo2`0000000`3oool000000000000D
000000H0oooo1P00000B0?ooo`D00000DP3oool4000000/0oooo000=0?ooo`050000003oool0oooo
0?ooo`0000000P3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo05D0oooo1@00000=
0?ooo`X000000P3ooolM000000030?ooo`000000000000L0000000<0oooo0000000000000P000000
0`3oool000000000000;00000080oooo200000030?ooo`X0000000@0oooo00000000000000001P3o
ool3000000030?ooo`000000000000P000000`3oool5000000030?ooo`00000000000080oooo0`00
00000`3oool000000000000<000000030?ooo`000000000000T0000000<0oooo0000000000006@00
00000`3oool0000000000008000000h0oooo1000001G0?ooo`030000003oool0oooo00T0oooo000=
0?ooo`050000003oool0oooo0?ooo`0000000P3oool00`000000oooo0?ooo`040?ooo`030000003o
ool0oooo05H0oooo1@00000<0?ooobX0000000<0oooo0000000000002@0000000`3oool000000000
000>000000040?ooo`00000000000?ooo`D0000000<0oooo0000000000002P000000103oool00000
0000000000060?ooo`<0000000<0oooo0000000000002P0000000`3oool000000000000600000003
0?ooo`000000000001`0000000<0oooo0000000000009@00000=0?ooo`@00000F03oool00`000000
oooo0?ooo`090?ooo`003@3oool01@000000oooo0?ooo`3oool000000080oooo00<000000?ooo`3o
ool0103oool00`000000oooo0?ooo`1H0?ooo`<000003P3oool;000000H0oooo5`0000000`3oool0
000000000004000000030?ooo`00000000000080000000@0oooo000000000000oooo100000000`3o
ool0000000000002000000050?ooo`00000000000000003oool00P000000103oool0000000000000
00030?ooo``0000000@0oooo00000000000000001P3oool3000000030?ooo`000000000000X00000
0`3oool6000000050?ooo`00000000000000003oool00`000000103oool000000000003oool20000
00030?ooo`00000000000080000000<0oooo0000000000002@0000000`3oool000000000000C0000
00H0oooo2`00000>0?ooo`<00000F@3oool00`000000oooo0?ooo`090?ooo`002`3oool3000000<0
oooo00@000000?ooo`3oool000001P3oool00`000000oooo0?ooo`1J0?ooo`8000003P3oool?0000
0080oooo3`0000020?ooo`H0000000<0oooo0000000000001P000000103oool000000000003oool?
00000080oooo100000020?ooo`X0000000D0oooo00000000000000000?ooo`03000000H0oooo0`00
00001@3oool00000000000000000oooo00X0000000@0oooo000000000000oooo0P0000000`3oool0
000000000009000000030?ooo`000000000000L0000000<0oooo0000000000002`0000000`3oool0
00000000000>00000080oooo3P00000?0?ooo`800000F`3oool00`000000oooo0?ooo`090?ooo`00
3@3oool00`000000oooo0?ooo`020?ooo`8000001`3oool3000005X0oooo1000000<0?ooo`l00000
0P3ooolS000000030?ooo`000000000000h0000000<0oooo0000000000000P0000000`3oool00000
00000009000000050?ooo`00000000000000003oool00`0000060?ooo`<0000000D0oooo00000000
000000000?ooo`0=000000040?ooo`00000000000?oooc@000000P3oool>000000d0oooo1000001I
0?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool0G03oool3000000d0oooo6`0000020?oo
o`X0000000<0oooo0000000000002@0000020?ooo`h0000000<0oooo0000000000003P000000103o
ool000000000003oool4000000H0oooo1`0000000`3oool000000000000>000000040?ooo`000000
00000?ooo`/0000000@0oooo000000000000oooo500000020?oooaX000003@3oool4000005d0oooo
00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0G03oool4000000h0oooo6000
00040?oooaH0000000<0oooo0000000000002@0000000`3oool00000000000030?ooo`/0000000<0
oooo0000000000000`000000103oool000000000000000070?ooo`X0000000<0oooo000000000000
2@0000030?ooo`80000000<0oooo000000000000200000020?ooo`80000000<0oooo000000000000
4@0000040?oooaP000003P3oool4000005d0oooo00<000000?ooo`3oool02@3oool001/0oooo00<0
00000?ooo`3oool0G@3oool4000000h0oooo3`0000000`3oool0000000000005000000<0oooo2`00
00030?oooaH0000000<0oooo0000000000000P3oool;000000030?ooo`000000000000P000001P3o
ool:000000030?ooo`000000000000P000000`3oool2000000030?ooo`000000000001@000000`3o
ool;000000<0oooo5`00000=0?ooo`@00000G`3oool00`000000oooo0?ooo`090?ooo`006`3oool0
0`000000oooo0?ooo`1N0?ooo`D00000303oool?000000030?ooo`000000000000@0000000<0oooo
000000000000300000000`3oool000000000000B000000050?ooo`00000000000000003oool03@00
00030?ooo`X000001P3oool:000000<0oooo200000000`3oool0000000000002000000030?ooo`00
0000000000`0000000<0oooo0000000000001P0000020?ooobD00000303oool400000600oooo00<0
00000?ooo`3oool02@3oool001/0oooo0`00001P0?ooo`@000003@3oool8000000030?ooo`000000
000000h000000`3ooolL000000@0oooo0P000000103oool000000000003oool8000000030?ooo`00
0000000000X000001P3oool;00000080oooo1`0000020?ooo`80000000<0oooo0000000000000`3o
ool:000000030?ooo`0000000000010000000`3oool8000000030?ooo`000000000000D0000000<0
oooo0000000000001P00000=0?ooo`@00000G`3oool3000000/0oooo000K0?ooo`030000003oool0
oooo0640oooo1000000=0?ooocD0000000<0oooo0000000000000`3oool2000000040?ooo`000000
00000?ooo`H000000`3oool<000000H0oooo2`0000040?ooo`D0000000D0oooo0000000000000000
0?ooo`02000000<0oooo0P0000000`3oool000000000000>000000030?ooo`000000000000P00000
00<0oooo0000000000005`00000=0?ooo`@00000HP3oool00`000000oooo0?ooo`090?ooo`006`3o
ool00`000000oooo0?ooo`1R0?ooo`@000003@3ooolJ000000030?ooo`00000000000180000000<0
oooo0000000000000P0000000`3oool00000000000020?ooo`L0000000<0oooo0000000000000P00
00000`3oool000000000000=000000H0oooo50000000103oool000000000003oool400000080oooo
2@0000000`3oool0000000000009000000030?ooo`000000000001l000003@3oool4000006<0oooo
00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0H`3oool4000000d0oooo1@00
00000`3oool000000000000@000000030?ooo`00000000000140000000<0oooo0000000000001@00
00020?ooo`H0000000@0oooo000000000000oooo0`0000020?ooo`l000001P3oool8000000030?oo
o`000000000000@000000P3oool3000000040?ooo`00000000000?ooo`H000000P3oool400000004
0?ooo`00000000000?ooo`P0000000<0oooo000000000000200000000`3oool000000000000F0000
00`0oooo1@00001T0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo06@0
oooo1000000=0?ooobl000000P3oool2000000030?ooo`000000000000D0000000@0oooo00000000
0000oooo500000060?ooo`h0000000<0oooo0000000000000P0000000`3oool00000000000090000
00030?ooo`00000000000080oooo3@0000000`3oool000000000000O000000`0oooo1000001V0?oo
o`030000003oool0oooo00T0oooo000K0?ooo`<00000I@3oool5000000`0oooo2`0000000`3oool0
00000000000D000000030?ooo`000000000000H0000000<0oooo0000000000000P3ooolC00000003
0?ooo`000000000000`000001P3oool>000000030?ooo`00000000000080000000<0oooo00000000
0000300000020?ooob40000000<0oooo0000000000002@00000=0?ooo`@00000I@3oool3000000/0
oooo000K0?ooo`030000003oool0oooo06L0oooo1000000<0?oooaL0000000<0oooo000000000000
3@000000103oool000000000000000030?ooo``0000000<0oooo0000000000000P3oool300000003
0?ooo`000000000000P0000000@0oooo00000000000000001P3oool3000000030?ooo`0000000000
00/000000`3oool2000000030?ooo`000000000000T00000103oool200000080oooo6P0000000`3o
ool0000000000009000000d0oooo1000001X0?ooo`030000003oool0oooo00T0oooo000K0?ooo`03
0000003oool0oooo06P0oooo1000000>0?ooo`T000000P3ooolG00000080oooo100000030?ooo``0
000000<0oooo0000000000000`3oool9000000050?ooo`00000000000000003oool00`0000060?oo
o`<0000000D0oooo00000000000000000?ooo`08000000@0oooo0P0000000`3oool0000000000009
000000@0oooo100000020?ooo`h0000000<0oooo0000000000001P0000020?ooo`T000003P3oool4
000006T0oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0J@3oool40000
00d0oooo200000040?oooaL0000000D0oooo00000000000000000?ooo`0?000000<0oooo0P000002
0?ooo`P0000000<0oooo0000000000001@0000060?ooo`<0000000D0oooo00000000000000000?oo
o`08000000030?ooo`000000000000@0oooo3`0000001@3oool00000000000000000oooo01L00000
0`3oool9000000d0oooo1000001Z0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003o
ool0oooo06X0oooo1000000=0?ooo`L000001@3ooolF00000080oooo4P0000030?ooo`<0000000<0
oooo0000000000001P0000000`3oool0000000000004000000P0oooo0P0000000`3oool000000000
0003000000030?ooo`000000000000@0000000<0oooo000000000000103ooolB00000080oooo5P00
00050?ooo`L000003@3oool4000006/0oooo00<000000?ooo`3oool02@3oool001/0oooo0`00001/
0?ooo`<000003P3oool6000000D0oooo4P0000060?ooo`@0000000<0oooo0000000000002@000002
0?ooo`80000000<0oooo0000000000002@0000000`3oool0000000000004000000L0oooo0`000000
0`3oool0000000000009000000050?ooo`00000000000000003oool00P0000020?ooo`/0000000<0
oooo0000000000000P0000060?ooo`H0000000<0oooo0000000000002@0000050?ooo`H000003P3o
ool3000006/0oooo0`00000;0?ooo`006`3oool00`000000oooo0?ooo`1/0?ooo`@000003@3oool6
000000H0oooo200000040?ooo`D00000103oool9000000030?ooo`000000000000H000000P3oool2
000000030?ooo`000000000000T0000000@0oooo000000000000oooo100000060?ooo`<000000P3o
ool2000000030?ooo`000000000000D0000000D0oooo00000000000000000?ooo`0200000080oooo
200000000`3oool0000000000007000000@0oooo100000040?ooo`T000001P3oool6000000`0oooo
1000001^0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo06d0oooo1@00
000<0?ooo`D000001`3oool7000000D0oooo0`0000020?ooo`X0000000<0oooo0000000000001P00
00040?ooo`80000000D0oooo00000000000000000?ooo`0:00000080oooo0`0000060?ooo`<00000
00<0oooo0000000000002@000000103oool000000000003oool3000000@0oooo1@0000000`3oool0
00000000000;00000080oooo0`0000050?ooo`L000001`3oool5000000`0oooo1000001_0?ooo`03
0000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo06l0oooo0`00000=0?ooo`@00000
203oool6000000H0oooo0P0000020?ooo`d0000000<0oooo0000000000000`000000103oool00000
0000003oool200000080oooo3@0000020?ooo`<000001P3oool3000000030?ooo`000000000000`0
000000<0oooo0000000000000P3oool2000000030?ooo`0000000000014000000P3oool2000000H0
oooo1P0000080?ooo`@00000303oool400000700oooo00<000000?ooo`3oool02@3oool000/0oooo
100000020?ooo`<000001`3oool00`000000oooo0?ooo`1`0?ooo`@000003@3oool2000000/0oooo
0`0000070?oooa@000000`3oool3000000040?ooo`00000000000?ooo`<0000000<0oooo00000000
000020000000103oool000000000000000060?ooo`<0000000<0oooo000000000000200000001@3o
ool00000000000000000oooo0080000000@0oooo00000000000000000`3oool2000000030?ooo`00
0000000000l000001`3oool3000000X0oooo0P00000>0?ooo`@00000L@3oool00`000000oooo0?oo
o`090?ooo`003@3oool00`000000oooo0?ooo`040?ooo`030000003oool0oooo00@0oooo00<00000
0?ooo`3oool0L@3oool4000000T0oooo1@00000;0?ooo`8000001@3ooolG000000040?ooo`000000
00000?ooo`80000000@0oooo000000000000oooo0`0000000`3oool0000000000008000000040?oo
o`0000000000000000H0oooo0`0000000`3oool000000000000700000080oooo0`000000103oool0
00000000003oool2000000040?ooo`00000000000?oooaH000001@3oool3000000X0oooo1P000009
0?ooo`@00000LP3oool00`000000oooo0?ooo`090?ooo`003@3oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo00@0oooo1000001a0?ooo`@00000203oool5000000/0oooo0P0000050?oo
oaL0000000<0oooo0000000000000`000000103oool000000000003oool400000080oooo1P000002
0?ooo`@000001P3oool4000000040?ooo`00000000000?ooo`@000000P3oool=000000030?ooo`00
0000000000P0000000<0oooo0000000000002P0000040?ooo`8000002`3oool6000000P0oooo1000
001`0?ooo`@000002`3oool000d0oooo00@000000?ooo`3oool0oooo0`0000070?ooo`030000003o
ool0oooo07<0oooo100000080?ooo`L0000000<0oooo0000000000000P0000030?oooa`00000103o
ool=000000030?ooo`000000000000D0000000<0oooo0000000000000P0000060?ooo`@0000000@0
oooo000000000000oooo0`0000020?ooo`@0000000<0oooo0000000000000`0000000`3oool00000
000000040?oooa`000000`3oool300000080oooo1`0000080?ooo`@00000M03oool00`000000oooo
0?ooo`090?ooo`003@3oool01@000000oooo0?ooo`3oool0000000T0oooo00<000000?ooo`3oool0
M03oool4000000T0oooo2P0000030?ooo`T0000000<0oooo0000000000003P0000060?ooo`D00000
00@0oooo000000000000oooo100000020?ooo`8000000P3oool2000000030?ooo`00000000000080
00001P3oool4000000030?ooo`00000000000080oooo0P0000000`3oool000000000000300000004
0?ooo`00000000000?ooo`D000001@3ooolK000000<0oooo2@0000090?ooo`@00000MP3oool00`00
0000oooo0?ooo`090?ooo`002`3oool3000000<0oooo00<000000?ooo`3oool01`3oool00`000000
oooo0?ooo`1e0?ooo`@000002@3ooolV000000D0oooo1P0000000`3oool000000000000600000003
0?ooo`000000000000T000001P3oool;000000030?ooo`000000000000d00000103ooolV000000T0
oooo1000001g0?ooo`030000003oool0oooo00T0oooo000=0?ooo`040000003oool0oooo0?ooo`@0
00001P3oool00`000000oooo0?ooo`1f0?ooo`D00000203ooolU00000080oooo0P0000000`3oool0
00000000000=000000030?ooo`000000000000T000001P3oool;000000030?ooo`000000000000l0
00000`3ooolT000000T0oooo1000001h0?ooo`030000003oool0oooo00T0oooo000K0?ooo`<00000
N@3oool3000000T0oooo5P0000000`3oool0000000000009000000030?ooo`000000000000H00000
00<0oooo000000000000100000020?ooo`<0000000<0oooo0000000000002@0000060?ooo`/00000
00@0oooo00000000000000000P3oool6000000040?ooo`00000000000?ooo`@0000000<0oooo0000
000000006P0000000`3oool0000000000004000000T0oooo0`00001h0?ooo`<000002`3oool001/0
oooo00<000000?ooo`3oool0NP3oool3000000T0oooo5P0000000`3oool000000000000>00000004
0?ooo`00000000000?ooo`H000000`3oool7000000030?ooo`000000000000@000001P3oool:0000
00040?ooo`0000000000000000<0oooo1P000000103oool000000000003oool@000000030?ooo`00
0000000001@000002@3oool3000007/0oooo00<000000?ooo`3oool02@3oool001/0oooo00<00000
0?ooo`3oool0NP3oool4000000T0oooo2`0000000`3oool000000000000@000000040?ooo`000000
00000?ooo`80000000<0oooo0000000000002@0000030?ooo`<0000000<0oooo0000000000002000
00060?ooo`X0000000@0oooo00000000000000000`3oool8000000030?ooo`000000000000D00000
00<0oooo0000000000007@0000090?ooo`@00000N`3oool00`000000oooo0?ooo`090?ooo`006`3o
ool00`000000oooo0?ooo`1k0?ooo`@000002@3ooolM000000@0oooo2`0000000`3oool000000000
0003000000040?ooo`00000000000?ooo`@000000P3oool400000080oooo0P0000020?ooo`<00000
0`3oool7000000030?ooo`000000000000<0000000<0oooo000000000000200000040?oooad00000
2P3oool4000007`0oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0O03o
ool4000000T0oooo5`000000103oool000000000000000060?ooo`X0000000<0oooo000000000000
0`000000103oool000000000000000020?ooo`80000000<0oooo0000000000000`0000020?ooo`80
00000P3oool5000000030?ooo`00000000000080oooo0`0000000`3oool000000000000;00000003
0?ooo`000000000000D0oooo700000080?ooo`@00000OP3oool00`000000oooo0?ooo`090?ooo`00
6`3oool3000007d0oooo100000090?oooaX000001P3oool=000000040?ooo`00000000000?ooo`<0
00000P3oool2000000030?ooo`000000000000<000001P3oool5000000030?ooo`00000000000080
oooo500000050?ooo`l0000000<0oooo0000000000002@0000080?ooo`@00000O@3oool3000000/0
oooo000K0?ooo`030000003oool0oooo07h0oooo1@0000080?oooaT00000103ooolF00000080oooo
1`0000030?ooo`8000000P3oool800000080oooo1@000000103oool000000000003oool:00000003
0?ooo`000000000000@0oooo3`0000000`3oool0000000000007000000P0oooo100000200?ooo`03
0000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0800oooo100000090?oooaH00000
1@3oool5000000030?ooo`000000000000D0000000@0oooo000000000000oooo0P0000000`3oool0
000000000009000000<0oooo0P0000020?ooo``0000000<0oooo0000000000000`0000000`3oool0
000000000005000000030?ooo`000000000000<000001@3ooolF000000T0oooo100000210?ooo`03
0000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0840oooo100000090?ooo`T00000
00<0oooo0000000000002@0000000`3oool0000000000007000000030?ooo`000000000000@00000
0P3oool2000000040?ooo`00000000000?ooo`@0000000<0oooo000000000000100000070?ooo`H0
00000P3oool3000000030?ooo`000000000000@000000P3oool?00000080oooo5@0000090?ooo`@0
0000PP3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`220?ooo`@00000
203ooolD000000030?ooo`000000000000T0000000<0oooo000000000000100000000`3oool00000
00000004000000040?ooo`000000000000000080oooo1P0000060?ooo`H0000000<0oooo00000000
00000P000000103oool000000000003oool3000000030?ooo`000000000000@0000000<0oooo0000
000000002@0000000`3oool000000000000A000000T0oooo0`0000240?ooo`030000003oool0oooo
00T0oooo000K0?ooo`<00000P`3oool4000000P0oooo4P0000020?ooo`/0000000<0oooo00000000
00000P3oool2000000050?ooo`00000000000000003oool00P000000103oool00000000000000002
0?ooo`H000001P3oool6000000030?ooo`00000000000080000000@0oooo000000000000oooo0P00
00020?ooo`80000000<0oooo0000000000003@0000020?oooa800000203oool400000880oooo0`00
000;0?ooo`006`3oool00`000000oooo0?ooo`240?ooo`@000002@3oool@000000@0oooo3@000000
0`3oool00000000000020?ooo`80000000@0oooo000000000000oooo0`0000020?ooo`H000001P3o
ool6000000030?ooo`00000000000080000000@0oooo000000000000oooo0P0000020?ooo`800000
00@0oooo000000000000oooo2@0000040?oooa4000002@3oool3000008H0oooo00<000000?ooo`3o
ool02@3oool001/0oooo00<000000?ooo`3oool0Q@3oool5000000P0oooo3`0000050?ooo`T00000
00<0oooo0000000000000P3oool2000000040?ooo`00000000000?ooo`80000000<0oooo00000000
00000P0000000`3oool0000000000004000000H0oooo1P0000000`3oool000000000000500000004
0?ooo`00000000000?ooo`8000000P3oool<000000@0oooo400000080?ooo`@00000Q`3oool00`00
0000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`260?ooo`D00000203oool>000000D0
oooo3@000000103oool000000000003oool2000000040?ooo`000000000000000080oooo0P000000
0`3oool0000000000004000000H0oooo1P0000000`3oool00000000000020?ooo`<0000000@0oooo
000000000000oooo0P0000000`3oool000000000000:000000H0oooo3P0000080?ooo`@00000R03o
ool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`280?ooo`<000002P3oool9
000000030?ooo`000000000000P0oooo1`000000103oool000000000003oool2000000030?ooo`00
000000000080oooo0P0000020?ooo`8000000`3oool3000000P0oooo0`0000030?ooo`8000000P3o
ool2000000030?ooo`000000000000@0000000<0oooo0000000000001`0000090?ooo``000002P3o
ool3000008T0oooo00<000000?ooo`3oool02@3oool001/0oooo0`0000290?ooo`@00000203oool:
000000/0oooo400000000`3oool00000000000020?ooo`8000000`3oool3000000P0oooo0`000002
0?ooo`<000000P3oool2000000040?ooo`00000000000?ooo`d000002P3oool:000000T0oooo1000
00280?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool0RP3oool4000000P0oooo2P00000:
0?ooo`d0000000@0oooo000000000000oooo0P0000020?ooo`@0000000@0oooo0000000000000000
203oool3000000040?ooo`000000000000000080oooo500000090?ooo`/00000203oool4000008/0
oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0R`3oool4000000T0oooo
2@00000:0?oooaP000000P3oool2000000P0oooo0P0000020?ooo`L0000000<0oooo000000000000
3@00000:0?ooo`X000002@3oool4000008`0oooo00<000000?ooo`3oool02@3oool001/0oooo00<0
00000?ooo`3oool0S03oool4000000T0oooo20000000103oool000000000003oool@000000040?oo
o`00000000000?ooo`80000000<0oooo0000000000001@0000020?ooo`8000001`3oool300000003
0?ooo`000000000000H0000000@0oooo000000000000oooo300000000`3oool00000000000060000
00030?ooo`000000000000L00000203oool5000008d0oooo00<000000?ooo`3oool02@3oool000/0
oooo100000030?ooo`8000001`3oool00`000000oooo0?ooo`2=0?ooo`D00000203oool700000003
0?ooo`000000000001@0000000@0oooo000000000000oooo20000000103oool00000000000000006
0?ooo`8000000P3oool@000000030?ooo`000000000000h000000P3oool7000000P0oooo1000002?
0?ooo`030000003oool0oooo00T0oooo000;0?ooo`030000003oool0oooo00<0oooo00@000000?oo
o`3oool000001P3oool4000008h0oooo0`0000090?ooo`L0000000<0oooo0000000000001`000000
0`3oool0000000000004000000030?ooo`00000000000080000000<0oooo0000000000002@000000
103oool000000000000000060?ooo`8000000P3oool7000000030?ooo`000000000000H0000000<0
oooo0000000000003P0000000`3oool0000000000005000000P0oooo1000002=0?ooo`@000002`3o
ool000`0oooo00<000000?ooo`3oool00P3oool010000000oooo0?ooo`0000060?ooo`030000003o
ool0oooo0900oooo100000090?ooo`D000000`3oool>000000040?ooo`0000000000000000<0oooo
0P0000000`3oool00000000000030?ooo`H000001P3oool2000000040?ooo`0000000000000000<0
oooo0P0000000`3oool00000000000030?ooo`8000000P3oool>000000<0oooo1@0000090?ooo`@0
0000T@3oool00`000000oooo0?ooo`090?ooo`003@3oool01@000000oooo0?ooo`3oool000000080
oooo00<000000?ooo`3oool0103oool00`000000oooo0?ooo`2A0?ooo`@000002@3oool4000000<0
oooo2`0000000`3oool00000000000020?ooo`8000000`3oool2000000030?ooo`000000000000<0
oooo0`0000000`3oool00000000000060?ooo`80000000@0oooo00000000000000000`3oool20000
00030?ooo`000000000000<0oooo0P0000020?ooo`d000000`3oool5000000T0oooo1000002B0?oo
o`030000003oool0oooo00T0oooo000>0?ooo`040000003oool0oooo00000080oooo00<000000?oo
o`3oool0103oool00`000000oooo0?ooo`2B0?ooo`@00000203oool5000000<0oooo3P0000000`3o
ool00000000000030?ooo`D000000`3oool3000000030?ooo`000000000000H0oooo0P000000103o
ool000000000000000020?ooo`H000000P3oool200000080oooo3P0000030?ooo`D00000203oool4
000009<0oooo00<000000?ooo`3oool02@3oool000/0oooo00@000000?ooo`3oool000000P3oool0
10000000oooo0?ooo`0000060?ooo`030000003oool0oooo09<0oooo100000090?ooo`<00000103o
ool:000000040?ooo`00000000000?ooo`<000000`3oool5000000<0oooo0P0000000`3oool00000
000000060?ooo`80000000<0oooo0000000000000`3oool2000000030?ooo`000000000000<0oooo
0P0000020?ooo`d00000103oool3000000T0oooo1000002D0?ooo`030000003oool0oooo00T0oooo
000<0?ooo`800000103oool2000000L0oooo0`00002D0?ooo`@000002@3oool3000000D0oooo3000
00000`3oool00000000000020?ooo`<0000000<0oooo0000000000000`3oool2000000030?ooo`00
0000000000H0oooo1@0000020?ooo`<0000000<0oooo0000000000000P3oool3000000030?ooo`00
0000000000T000001@3oool4000000P0oooo1000002D0?ooo`<000002`3oool001/0oooo00<00000
0?ooo`3oool0U@3oool5000000P0oooo0`0000050?ooo`l0000000D0oooo00000000000000000?oo
o`0300000080oooo0P0000000`3oool00000000000060?ooo`@000000`3oool2000000040?ooo`00
0000000000000080oooo0P0000000`3oool000000000000:000000D0oooo0`0000080?ooo`@00000
U`3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`2G0?ooo`@00000203o
ool2000000H0oooo3`000000103oool000000000003oool300000080oooo1@0000060?ooo`@00000
0`3oool2000000050?ooo`00000000000000003oool03`0000060?ooo`800000203oool4000009P0
oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0V03oool400000140oooo
3@0000000`3oool00000000000020?ooo`8000000`3oool3000000P0oooo0`0000030?ooo`800000
00<0oooo0000000000003`00000A0?ooo`@00000V@3oool00`000000oooo0?ooo`090?ooo`006`3o
ool00`000000oooo0?ooo`2I0?ooo`@000004@3oool@000000030?ooo`000000000000<0oooo0`00
00080?ooo`<000000P3oool3000000030?ooo`000000000000h000004@3oool4000009X0oooo00<0
00000?ooo`3oool02@3oool001/0oooo0`00002J0?ooo`@00000403oool=000000040?ooo`000000
00000?ooo`@0000000@0oooo00000000000000001`3oool800000080oooo0P0000000`3oool00000
0000000:00000140oooo1000002I0?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool0V`3o
ool400000100oooo2@0000000`3oool000000000000300000080oooo1`0000070?ooo`P0000000<0
oooo0000000000003P00000@0?ooo`@00000W03oool00`000000oooo0?ooo`090?ooo`006`3oool0
0`000000oooo0?ooo`2L0?ooo`@000004@3oool>000000030?ooo`000000000000H000001P3oool8
000000040?ooo`00000000000?ooo`/00000403oool5000009d0oooo00<000000?ooo`3oool02@3o
ool001/0oooo00<000000?ooo`3oool0W@3oool400000140oooo2P000000103oool000000000003o
ool8000000H0oooo0`0000000`3oool0000000000002000000040?ooo`00000000000?ooo`800000
00<0oooo0000000000001@00000@0?ooo`@00000W`3oool00`000000oooo0?ooo`090?ooo`006`3o
ool00`000000oooo0?ooo`2O0?ooo`@000003`3oool7000000040?ooo`00000000000?ooo`L00000
00@0oooo00000000000000001P3oool3000000050?ooo`00000000000000003oool00`0000000`3o
ool000000000000700000100oooo1000002P0?ooo`030000003oool0oooo00T0oooo000K0?ooo`03
0000003oool0oooo0:00oooo1000000A0?ooo`D000000`3oool3000000030?ooo`00000000000080
oooo0`0000060?ooo`<000000P3oool2000000030?ooo`000000000000@0oooo1@00000A0?ooo`@0
0000X@3oool00`000000oooo0?ooo`090?ooo`006`3oool300000:40oooo1000000A0?ooo`@00000
103oool2000000030?ooo`00000000000080oooo0`0000060?ooo`<000000P3oool2000000030?oo
o`000000000000<0oooo1@00000A0?ooo`@00000X03oool3000000/0oooo000K0?ooo`030000003o
ool0oooo0:80oooo1000000@0?ooo`@00000103oool2000000030?ooo`00000000000080oooo0`00
00060?ooo`<0000000D0oooo00000000000000000?ooo`02000000<0oooo1@00000@0?ooo`@00000
X`3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`2S0?ooo`@000006P3o
ool3000000030?ooo`000000000000L0oooo0`0000000`3oool000000000000J0?ooo`@00000Y@3o
ool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`2T0?ooo`@000006P3oool2
000000030?ooo`000000000000L0oooo0`0000000`3oool000000000000I0?ooo`D00000Y@3oool0
0`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`2U0?ooo`@000006@3oool20000
00h0oooo0P00000H0?ooo`@00000Y`3oool00`000000oooo0?ooo`090?ooo`006`3oool300000:H0
oooo1@00000G0?ooo`8000003P3oool2000001L0oooo1000002V0?ooo`<000002`3oool001/0oooo
00<000000?ooo`3oool0Z03oool4000003h0oooo1000002Y0?ooo`030000003oool0oooo00T0oooo
000K0?ooo`030000003oool0oooo0:T0oooo100000050?ooo`800000;@3oool3000000D0oooo1000
002Z0?ooo`030000003oool0oooo00T0oooo000;0?ooo`@000000P3oool3000000L0oooo00<00000
0?ooo`3oool0ZP3oool4000000@0oooo1000000Y0?ooo`D00000103oool400000:/0oooo00<00000
0?ooo`3oool02@3oool000/0oooo00<000000?ooo`3oool01P3oool00`000000oooo0?ooo`040?oo
o`030000003oool0oooo0:/0oooo100000050?ooo`<000009`3oool3000000H0oooo0`00002]0?oo
o`030000003oool0oooo00T0oooo000<0?ooo`030000003oool0oooo00D0oooo00<000000?ooo`3o
ool0103oool400000:/0oooo100000050?ooo`@000008`3oool5000000@0oooo1@00002Z0?ooo`@0
00002`3oool000d0oooo00@000000?ooo`3oool0oooo0`0000070?ooo`030000003oool0oooo0:d0
oooo1@0000040?ooo`@000008@3oool5000000@0oooo1000002_0?ooo`030000003oool0oooo00T0
oooo000>0?ooo`040000003oool0oooo000000T0oooo00<000000?ooo`3oool0[`3oool4000000<0
oooo1`00000L0?ooo`H00000103oool400000;00oooo00<000000?ooo`3oool02@3oool000/0oooo
00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`070?ooo`030000003oool0oooo0;00
oooo100000050?ooo`@0000000<0oooo0000000000005P3oool2000000030?ooo`00000000000080
00001@3oool400000;40oooo00<000000?ooo`3oool02@3oool000`0oooo0P0000030?ooo`@00000
1P3oool00`000000oooo0?ooo`2a0?ooo`@00000103oool7000001D0oooo1`0000050?ooo`@00000
/P3oool00`000000oooo0?ooo`090?ooo`006`3oool300000;80oooo100000040?ooo`L000000P3o
ool2000000`0oooo0P0000020?ooo`L00000103oool400000;40oooo0`00000;0?ooo`006`3oool0
0`000000oooo0?ooo`2c0?ooo`@000001@3oool6000000040?ooo`00000000000?ooo`8000001P3o
ool2000000040?ooo`00000000000?ooo`H000001@3oool400000;@0oooo00<000000?ooo`3oool0
2@3oool001/0oooo00<000000?ooo`3oool0]03oool4000000D0oooo1@000000103oool000000000
003oool2000000H0oooo1@0000000`3oool0000000000003000000@0oooo1000002f0?ooo`030000
003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0;D0oooo1@0000040?ooo`X000001P3o
ool:000000@0oooo1000002g0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0
oooo0;L0oooo100000050?ooo`H000002P3oool2000000040?ooo`0000000000000000D0oooo1000
002h0?ooo`030000003oool0oooo00T0oooo000K0?ooo`<00000^03oool4000000D0oooo1@000009
0?ooo`H000001@3oool400000;L0oooo0`00000;0?ooo`006`3oool00`000000oooo0?ooo`2i0?oo
o`@00000103oool6000000P0oooo1@0000050?ooo`@00000^P3oool00`000000oooo0?ooo`090?oo
o`006`3oool00`000000oooo0?ooo`2j0?ooo`@00000103oool2000000030?ooo`000000000000P0
oooo0P0000000`3oool00000000000040?ooo`@00000^`3oool00`000000oooo0?ooo`090?ooo`00
6`3oool00`000000oooo0?ooo`2k0?ooo`@000001P3oool3000000H0oooo0`0000050?ooo`D00000
_03oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`2l0?ooo`@000001@3o
ool3000000H0oooo0P0000050?ooo`D00000_@3oool00`000000oooo0?ooo`090?ooo`006`3oool3
00000;d0oooo1@0000040?ooo`8000001P3oool2000000@0oooo1000002m0?ooo`<000002`3oool0
01/0oooo00<000000?ooo`3oool0_P3oool500000100oooo100000300?ooo`030000003oool0oooo
00T0oooo000K0?ooo`030000003oool0oooo0<00oooo1000000>0?ooo`@00000`@3oool00`000000
oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`310?ooo`@00000303oool400000<80oooo
00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3oool0`P3oool4000000X0oooo1000
00330?ooo`030000003oool0oooo00T0oooo000K0?ooo`<00000``3oool4000000P0oooo10000032
0?ooo`<000002`3oool001/0oooo00<000000?ooo`3oool0a03oool4000000H0oooo100000350?oo
o`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0<D0oooo0`0000060?ooo`80
0000a`3oool00`000000oooo0?ooo`090?ooo`006`3oool00`000000oooo0?ooo`3o0?oooiP0oooo
00<000000?ooo`3oool02@3oool000`0oooo0P0000040?ooo`8000001`3oool00`000000oooo0?oo
o`3o0?oooiP0oooo00<000000?ooo`3oool02@3oool000/0oooo00@000000?ooo`3oool000000P3o
ool010000000oooo0?ooo`0000060?ooo`@00000b@3oool200000<P0oooo1000000;0?ooo`003P3o
ool010000000oooo0?ooo`0000020?ooo`030000003oool0oooo00@0oooo00<000000?ooo`3oool0
bP3oool200000</0oooo00<000000?ooo`3oool02@3oool000`0oooo0P0000030?ooo`040000003o
ool0oooo000000H0oooo00<000000?ooo`3oool0o`3ooonH0?ooo`030000003oool0oooo00T0oooo
000>0?ooo`040000003oool0oooo00000080oooo00<000000?ooo`3oool0103oool00`000000oooo
0?ooo`3o0?oooiP0oooo00<000000?ooo`3oool02@3oool000/0oooo00@000000?ooo`3oool00000
0P3oool010000000oooo0?ooo`0000060?ooo`030000003oool0oooo0?l0ooooV03oool00`000000
oooo0?ooo`090?ooo`00303oool2000000@0oooo0P0000070?ooo`030000003oool0oooo0?l0oooo
V03oool00`000000oooo0?ooo`090?ooo`006`3oool300000?l0ooooUP3oool3000000/0oooo000K
0?ooo`030000003oool0oooo0?l0ooooV03oool00`000000oooo0?ooo`090?ooo`006`3oool00`00
0000oooo0?ooo`3o0?oooiP0oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3o
ool0o`3ooonH0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0?l0oooo
V03oool00`000000oooo0?ooo`090?ooo`006`3oool300000?l0ooooUP3oool3000000/0oooo000K
0?ooo`030000003oool0oooo0?l0ooooV03oool00`000000oooo0?ooo`090?ooo`006`3oool00`00
0000oooo0?ooo`3o0?oooiP0oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3o
ool0o`3ooonH0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0?l0oooo
V03oool00`000000oooo0?ooo`090?ooo`006`3oool300000?l0ooooUP3oool3000000/0oooo000K
0?ooo`030000003oool0oooo0?l0ooooV03oool00`000000oooo0?ooo`090?ooo`006`3oool00`00
0000oooo0?ooo`3o0?oooiP0oooo00<000000?ooo`3oool02@3oool001/0oooo00<000000?ooo`3o
ool0o`3ooonH0?ooo`030000003oool0oooo00T0oooo000K0?ooo`030000003oool0oooo0?l0oooo
V03oool00`000000oooo0?ooo`090?ooo`006`3oool300000?l0ooooUP3oool3000000/0oooo000K
0?ooo`030000003oool0oooo0?l0ooooV03oool00`000000oooo0?ooo`090?ooo`006`3oool00`00
0000oooo0?ooo`0J0?ooo`030000003oool0oooo03P0oooo00<000000?ooo`3oool0=`3oool00`00
0000oooo0?ooo`0h0?ooo`030000003oool0oooo03P0oooo00<000000?ooo`3oool0=`3oool00`00
0000oooo0?ooo`0h0?ooo`030000003oool0oooo01X0oooo00<000000?ooo`3oool02@3oool001/0
oooo00<000000?ooo`3oool00`3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`080?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00P0
oooo00<000000?ooo`3oool02@3oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00T0
oooo00<000000?ooo`3oool0203oool00`000000oooo0?ooo`090?ooo`030000003oool0oooo00<0
oooo00<000000?ooo`3oool02@3oool000`0oooo0P0000030?ooo`<000001`3oool00`000000oooo
0?ooo`030?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo
0?ooo`080?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo
0?ooo`080?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo
0?ooo`080?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo00P0oooo00<000000?ooo`3oool02@3oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool0203oool00`000000oooo
0?ooo`090?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool02@3oool00`000000oooo
0?ooo`080?ooo`030000003oool0oooo00T0oooo00<000000?ooo`3oool00`3oool00`000000oooo
0?ooo`090?ooo`002`3oool010000000oooo0?ooo`0000050?ooo`030000003oool0oooo00@0oooo
o`00002L000000/0oooo000>0?ooo`030000003oool0oooo00<0oooo00<000000?ooo`3oool0o`3o
oon[0?ooo`00303oool2000000<0oooo0`00003o0?ooojh0oooo000>0?ooo`040000003oool0oooo
00000?l0oooo/03oool000/0oooo00@000000?ooo`3oool000000P3oool00`000000oooo0?ooo`3o
0?ooojh0oooo0000\
\>"],
  ImageRangeCache->{{{0, 448}, {223.5, 0}} -> {-9.71889, -7.75737, 0.170835, 
  0.19524}}]
}, Open  ]]
},
FrontEndVersion->"X 3.0",
ScreenRectangle->{{0, 1152}, {0, 864}},
WindowSize->{880, 600},
WindowMargins->{{Automatic, 72}, {83, Automatic}},
PrintingPageRange->{Automatic, Automatic},
PrintingOptions->{"PaperSize"->{612, 792},
"PaperOrientation"->"Portrait",
"Magnification"->1}
]


(***********************************************************************
Cached data follows.  If you edit this Notebook file directly, not using
Mathematica, you must remove the line containing CacheID at the top of 
the file.  The cache data will then be recreated when you save this file 
from within Mathematica.
***********************************************************************)

(*CellTagsOutline
CellTagsIndex->{}
*)

(*CellTagsIndex
CellTagsIndex->{}
*)

(*NotebookFileOutline
Notebook[{
Cell[1709, 49, 1971, 33, 539, "Input"],

Cell[CellGroupData[{
Cell[3705, 86, 477, 10, 107, "Input"],
Cell[4185, 98, 2763, 39, 187, "Output"],
Cell[6951, 139, 36, 1, 27, "Output"],

Cell[CellGroupData[{
Cell[7012, 144, 19393, 528, 186, 6359, 363, "GraphicsData", 
"PostScript", "Graphics"],
Cell[26408, 674, 24223, 639, 186, 8203, 437, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[50680, 1319, 470, 9, 107, "Input"],
Cell[51153, 1330, 37, 1, 27, "Output"],

Cell[CellGroupData[{
Cell[51215, 1335, 40586, 1464, 231, 22873, 1241, "GraphicsData", 
"PostScript", "Graphics"],
Cell[91804, 2801, 78175, 1993, 231, 24847, 1330, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}, Open  ]],

Cell[CellGroupData[{
Cell[170028, 4800, 471, 9, 107, "Input"],
Cell[170502, 4811, 38, 1, 27, "Output"],

Cell[CellGroupData[{
Cell[170565, 4816, 96034, 4333, 222, 75808, 4079, "GraphicsData", 
"PostScript", "Graphics"],
Cell[266602, 9151, 200840, 5623, 222, 76787, 4087, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}, Open  ]],
Cell[467469, 14778, 1767, 28, 491, "Input"],
Cell[469239, 14808, 351, 6, 155, "Input"],

Cell[CellGroupData[{
Cell[469615, 14818, 467, 8, 107, "Input"],
Cell[470085, 14828, 36, 1, 27, "Output"],
Cell[470124, 14831, 22401, 798, 152, 6767, 600, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],

Cell[CellGroupData[{
Cell[492562, 15634, 360, 6, 91, "Input"],
Cell[492925, 15642, 36, 1, 27, "Output"],
Cell[492964, 15645, 29952, 932, 212, 7752, 653, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],

Cell[CellGroupData[{
Cell[522953, 16582, 360, 6, 91, "Input"],
Cell[523316, 16590, 37, 1, 27, "Output"],
Cell[523356, 16593, 33557, 1048, 209, 9475, 746, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],

Cell[CellGroupData[{
Cell[556950, 17646, 360, 6, 91, "Input"],
Cell[557313, 17654, 37, 1, 27, "Output"],
Cell[557353, 17657, 45702, 1468, 213, 16083, 1098, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],

Cell[CellGroupData[{
Cell[603092, 19130, 360, 6, 91, "Input"],
Cell[603455, 19138, 38, 1, 27, "Output"],
Cell[603496, 19141, 115130, 3975, 231, 56271, 3244, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],

Cell[CellGroupData[{
Cell[718663, 23121, 360, 6, 91, "Input"],
Cell[719026, 23129, 38, 1, 27, "Output"],
Cell[719067, 23132, 176574, 6600, 229, 101809, 5672, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]],

Cell[CellGroupData[{
Cell[895678, 29737, 465, 8, 107, "Input"],
Cell[896146, 29747, 39, 1, 27, "Output"],
Cell[896188, 29750, 290121, 13022, 233, 224421, 12206, "GraphicsData", 
"PostScript", "Graphics"]
}, Open  ]]
}
]
*)




(***********************************************************************
End of Mathematica Notebook file.
***********************************************************************)

